#!/bin/bash

unalias -a

usage()
{
cat << EOF

	Usage: `basename $0` [-t|--timeout <timeout>] [--force_restart] [--daemon] [-v]

                -t|--timeout timeout:   Time to wait for SA response in seconds
                --force_restart:        Restart OpenSM if failed on timeout (/etc/init.d/opensmd restart)
                --daemon:               Run in infinite loop
                -v:                     Verbose

EOF
}


OSM_TIMEOUT=10
RETRIES=4
FORCE_RESTART=0
DAEMON=0
VERBOSE=0

##########
#  MAIN  #
##########
main()
{
        [ -z "$1" ] && usage && exit 0
        
        # Parse parameters
	while [ ! -z "$1" ]
	do
		case "$1" in
                        -t|--timeout)
                                OSM_TIMEOUT=$2
                                shift 2
                                ;;
                        --force_restart)
                                FORCE_RESTART=1
                                shift
                                ;;
                        --daemon)
                                DAEMON=1
                                shift
                                ;;
                        -v)
                                VERBOSE=1
                                shift
                                ;;        
			-h|--help)
				usage
				exit 0
			        ;;
                         *)
                                printf "\nWrong parameter: $1\n"
                                usage
                                exit 1
                                ;;
		esac	
	done

        # Locate sa_ping program
        SA_PING=$(which sa_ping 2> /dev/null)

        if [ ! -x ${SA_PING} ]; then
            # Check through RPM
            SA_PING=$(rpm -ql ib-verbs | grep -w "bin.sa_ping" 2> /dev/null)
        fi

        # sa_ping is not found. Exit
        if [ ! -x ${SA_PING} ]; then
            printf "\nsa_ping utility is not found. Exiting ...\n"
            exit 1
        fi

        # Check vstat
        if ! ( vstat  > /dev/null 2>&1 ); then
            printf "\nHCA driver is not loaded. Exiting ...\n"
            exit 2
        fi
        
        if ! ( ${SA_PING} -t ${OSM_TIMEOUT} > /dev/null 2>&1 ); then
            [ $VERBOSE -eq 1 ] && printf "\nSA does not respond.\n"
            if [ $FORCE_RESTART -eq 1 ]; then
                if [ ! -x /etc/init.d/opensmd ]; then
                   printf "\n/etc/init.d/opensmd does not exist\n"
                   exit 3
                fi
                [ $VERBOSE -eq 1 ] && printf "\nRestarting OpenSM. Please wait...\n"
                /etc/init.d/opensmd restart > /dev/null 2>&1
                if [ $? -ne 0 ]; then
                    [ $VERBOSE -eq 1 ] && printf "\nFailed to restart OpenSM.\n"
                fi
            fi
        else
            [ $VERBOSE -eq 1 ] && printf "\nSA is alive.\n"
        fi

        if [ $DAEMON -eq 1 ]; then
	    declare -i failed_cnt=0
            while true
            do
                if ! ( ${SA_PING} -t ${OSM_TIMEOUT} > /dev/null 2>&1 ); then
		    let failed_cnt++	
                    [ $VERBOSE -eq 1 ] && printf "\nSA does not respond.\n"
                    if [ $FORCE_RESTART -eq 1 ]; then
                        if [ ! -x /etc/init.d/opensmd ]; then
                           printf "\n/etc/init.d/opensmd does not exist\n"
                           exit 3
                        fi
			if [ $failed_cnt -ge $RETRIES ]; then
                        	[ $VERBOSE -eq 1 ] && printf "\nRestarting OpenSM. Please wait...\n"
                        	/etc/init.d/opensmd restart > /dev/null 2>&1
                        	if [ $? -ne 0 ]; then
                            		[ $VERBOSE -eq 1 ] && printf "\nFailed to restart OpenSM.\n"
                        	fi
                    	fi
		    fi
                else
                    [ $VERBOSE -eq 1 ] && printf "\nSA is alive.\n"
		    failed_cnt=0
                fi
                sleep ${OSM_TIMEOUT}
            done
        fi

}

main $@
