/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: useraccess_cm.c 477 2004-07-16 16:40:34Z roland $
*/

/* FIXME: We should validate the QP handle given back from userspace */

//#define MT_MEMTRACK

#ifdef MT_MEMTRACK
	#include <memtrack.h>
	#define XMALLOC(bsize, flags)  ({                                                                                \
									void *p;                                                                          \
									p = kmalloc(bsize, flags);                                                        \
									if ( p ) {                                                                        \
									  memtrack_alloc(MEMTRACK_KMALLOC, (unsigned long)p, bsize, __FILE__, __LINE__);  \
									}                                                                                 \
									p;                                                                                \
                                  })


	#define 	XFREE(ptr) do {                                                    \
							 if ( ptr ) {                                           \
                                memtrack_free(MEMTRACK_KMALLOC, (unsigned long)ptr, __FILE__, __LINE__);  \
                                kfree(ptr);                                           \
							 } \
                          }                                                       \
                          while(0)

#else
	#define XMALLOC(bsize, flags) kmalloc(bsize, flags)
	#define XFREE(ptr)  kfree(ptr);
#endif

#define IB_QP_STATE_INVALID (-1)
#define TS_IB_UCM_MRA_TIMEOUT 27	/* around 9 minutes */

#include "useraccess_cm.h"

#include "ts_kernel_trace.h"
#include "ts_kernel_services.h"
#include "ts_ib_tavor_provider.h"

/* for path record lookup code */
#include "ip2pr_export.h"
#include "openib_version.h"
#include <linux/version.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/errno.h>
#if defined(TS_KERNEL_2_6)
#include <linux/cdev.h>
#endif

MODULE_AUTHOR("Johannes Erdfelt");
MODULE_DESCRIPTION("kernel IB CM userspace access" OPENIB_REVISION);
MODULE_LICENSE("Dual BSD/GPL");

/* This is used to queue up completions out to userspace */
struct ib_cm_user_compeletion {
	struct list_head list;

	struct ib_qp *qp;
	tTS_IB_CM_COMM_ID comm_id;
	tTS_IB_LISTEN_HANDLE listen_handle;
	unsigned long cm_arg;
	tTS_IB_CM_EVENT event;

	void *params;
	ssize_t params_size;

	enum ib_qp_state qp_state;
};

enum {
	TS_USERACCESS_CM_NUM_DEVICE = 16
};

#if defined(TS_KERNEL_2_6)
static dev_t useraccess_cm_devnum;
static struct cdev useraccess_cm_cdev;
#else
static int useraccess_cm_major;
static devfs_handle_t useraccess_cm_devfs_dir;
#endif

static const char TS_USERACCESS_CM_NAME[] = "ts_ib_useraccess_cm";

static struct ib_useraccess_cm_device
    useraccess_cm_dev_list[TS_USERACCESS_CM_NUM_DEVICE];

/* FIXME: We need to handle this cleaner */
int tsIbTavorDeviceQueryVapiHandle(struct ib_device *device,
				   VAPI_hca_hndl_t * vapi_handle);

static struct ib_cm_user_connection *
tsIbCmUserConnectionAlloc(struct ib_useraccess_cm_private *priv, int flags)
{
	struct ib_cm_user_connection *cm_conn;

	/* Get any memory allocate errors out of the way first */
	cm_conn = XMALLOC(sizeof(*cm_conn), flags);
	if (NULL == cm_conn) {
		return NULL;
	}

	atomic_set(&cm_conn->refcnt, 12);	/* One for UCM, one for calling func */
	INIT_LIST_HEAD(&cm_conn->list);

	cm_conn->priv = priv;
	cm_conn->comm_id = TS_IB_CM_COMM_ID_INVALID;
	cm_conn->listen_handle = NULL;
	cm_conn->cm_arg = 0;
	cm_conn->v_device = VAPI_INVAL_HNDL;
	cm_conn->qp = NULL;
	cm_conn->vk_qp = VAPI_INVAL_HNDL;

	return cm_conn;
}

static void tsIbCmUserConnectionGet(struct ib_cm_user_connection *cm_conn)
{
	atomic_inc(&cm_conn->refcnt);
}

static void tsIbCmUserConnectionPut(struct ib_cm_user_connection *cm_conn)
{
	if (atomic_dec_and_test(&cm_conn->refcnt)) {
		cm_conn->priv = (void *)0x5a5a5a5a;
		XFREE(cm_conn);
	}
}

static struct ib_cm_user_connection *
tsIbCmUserConnectionFind(struct ib_useraccess_cm_private *priv,
			 tTS_IB_CM_COMM_ID comm_id)
{
	struct list_head *ptr;

	down(&priv->cm_conn_sem);
	list_for_each(ptr, &priv->cm_conn_list) {
		struct ib_cm_user_connection *cm_conn;

		cm_conn = list_entry(ptr, struct ib_cm_user_connection, list);

		if (cm_conn->comm_id == comm_id) {
			tsIbCmUserConnectionGet(cm_conn);
			up(&priv->cm_conn_sem);

			return cm_conn;
		}
	}
	up(&priv->cm_conn_sem);

	return NULL;
}

static void _tsIbCmUserConnectionDestroy(struct ib_cm_user_connection *cm_conn)
{
	struct ib_useraccess_cm_private *priv = cm_conn->priv;
	unsigned long flags;
	struct list_head *ptr, *tmp;
	if (cm_conn->vk_qp != VAPI_INVAL_HNDL) {
		EVAPI_k_clear_destroy_qp_cbk(cm_conn->v_device, cm_conn->vk_qp);
		cm_conn->vk_qp = VAPI_INVAL_HNDL;
	}

	if (cm_conn->listen_handle) {
		tsIbCmListenStop(cm_conn->listen_handle);
		cm_conn->listen_handle = NULL;
	}

	if (cm_conn->comm_id != TS_IB_CM_COMM_ID_INVALID) {
		/* Free up any completions */
		spin_lock_irqsave(&priv->cm_comp_lock, flags);
		list_for_each_safe(ptr, tmp, &priv->cm_comp_list) {
			struct ib_cm_user_compeletion *cm_comp;

			cm_comp = list_entry(ptr, struct ib_cm_user_compeletion,
					     list);

			if (cm_comp->comm_id == cm_conn->comm_id) {
				list_del_init(&cm_comp->list);
				kfree(cm_comp);
			}
		}
		spin_unlock_irqrestore(&priv->cm_comp_lock, flags);

		/* This will drop all state for this connection */
		tsIbCmKill(cm_conn->comm_id);
	}
	
}

static void tsIbCmUserConnectionDestroy(struct ib_cm_user_connection *cm_conn)
{
	struct ib_useraccess_cm_private *priv = cm_conn->priv;

	down(&priv->cm_conn_sem);
    if (!list_empty(&cm_conn->list)){

        list_del_init(&cm_conn->list);
		list_add(&cm_conn->list, &priv->free_conn_list);

        _tsIbCmUserConnectionDestroy(cm_conn);

        tsIbCmUserConnectionPut(cm_conn);	/* drop the structure */
    } else {
        _tsIbCmUserConnectionDestroy(cm_conn);
    }
	
	up(&priv->cm_conn_sem);
}

/* User Mode CM Filter Routines */
int tsIbCmUserGetServiceId(unsigned long arg)
{
	tTS_IB_SERVICE_ID service_id;

	service_id = (tTS_IB_SERVICE_ID) tsIbCmServiceAssign();

	return copy_to_user((tTS_IB_SERVICE_ID *) arg,
			    &service_id, sizeof service_id) ? -EFAULT : 0;
}

static tTS_IB_CM_CALLBACK_RETURN tsIbCmUserCompletionHandler(tTS_IB_CM_EVENT
							     event,
							     tTS_IB_CM_COMM_ID
							     comm_id,
							     void *params,
							     void *arg)
{
	struct ib_cm_user_connection *cm_conn =
					(struct ib_cm_user_connection *) arg;
	struct ib_cm_user_compeletion *cm_comp;
	unsigned long flags;
	tTS_IB_CM_CALLBACK_RETURN ret = TS_IB_CM_CALLBACK_PROCEED;
    int freeing_conn = 0;
	int rc;
	struct ib_qp_attribute *qp_attr=NULL;

	if (!cm_conn) {
		return TS_IB_CM_CALLBACK_ABORT;
	}

	tsIbCmUserConnectionGet(cm_conn);	/* since we can possibly sleep */

	/* If we've destroyed the conn, don't worry about completions to userspace */
	if (list_empty(&cm_conn->list)) {
		goto out;
	}
    
	/* The CM will be destroying the connection anyway, so clean up our state */
	if (event == TS_IB_CM_IDLE) {
		down(&cm_conn->priv->cm_conn_sem);
		if (!list_empty(&cm_conn->list)) {
            freeing_conn = 1;
			list_del_init(&cm_conn->list);
			list_add(&cm_conn->list, &cm_conn->priv->free_conn_list);
			up(&cm_conn->priv->cm_conn_sem);

			/* Don't free up the completions. They need to go down to userspace */

			EVAPI_k_clear_destroy_qp_cbk(cm_conn->v_device,
						     cm_conn->vk_qp);

			/* Cleanup finished below */
		} else if (cm_conn->vk_qp != VAPI_INVAL_HNDL) {
			/* FIXME: Can we clean this code up a little bit? */
			up(&cm_conn->priv->cm_conn_sem);

			/* Drop this completion on the floor, we don't exist anymore */
			goto out;
		}
	}

	cm_comp = kmalloc(sizeof(*cm_comp), GFP_KERNEL);
	if (!cm_comp) {
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "Unable to allocate memory for CM completion buffer");
		ret = TS_IB_CM_CALLBACK_ABORT;
		goto out;
	}

	cm_comp->qp = cm_conn->qp;
	cm_comp->comm_id = comm_id;
	cm_comp->listen_handle = cm_conn->listen_handle;
	cm_comp->cm_arg = cm_conn->cm_arg;
	cm_comp->event = event;

	if (cm_conn->qp) {

		qp_attr = kmalloc(sizeof *qp_attr, GFP_KERNEL);
		if ( !qp_attr ) {
			ret = TS_IB_CM_CALLBACK_ABORT;
			goto out;
		}
		qp_attr->valid_fields = IB_QP_ATTRIBUTE_STATE;
		rc = ib_qp_query(cm_conn->qp, qp_attr);
		if (rc == 0) {
			cm_comp->qp_state = qp_attr->state;
		} else {
			cm_comp->qp_state = IB_QP_STATE_INVALID;
		}
	} else {
		cm_comp->qp_state = IB_QP_STATE_INVALID;
	}

	switch (event) {
	case TS_IB_CM_REQ_RECEIVED:
		{
			struct ib_cm_req_received_param *req_param =
			    (struct ib_cm_req_received_param *) params;
			struct ib_u_cm_req_received_param *u_req_param;
			struct ib_cm_user_connection *ncm_conn;

            cm_comp->params_size = sizeof(*u_req_param) +
			    req_param->remote_private_data_len;
			cm_comp->params =
			    kmalloc(cm_comp->params_size, GFP_KERNEL);
			if (!cm_comp->params) {
				TS_REPORT_WARN(MOD_KERNEL_IB,
					       "Unable to allocate memory for REQ_RECV param struct");
				kfree(cm_comp);
				goto out;
			}

			u_req_param =
			  (struct ib_u_cm_req_received_param *) cm_comp->params;

			u_req_param->listen_handle = req_param->listen_handle;
			u_req_param->service_id = req_param->service_id;
			u_req_param->local_qpn = req_param->local_qpn;
			u_req_param->remote_qpn = req_param->remote_qpn;

			memcpy(u_req_param->remote_guid, req_param->remote_guid,
			       sizeof(u_req_param->remote_guid));
			memcpy(u_req_param->dgid, req_param->dgid,
			       sizeof(u_req_param->dgid));
			memcpy(u_req_param->sgid, req_param->sgid,
			       sizeof(u_req_param->sgid));

			u_req_param->dlid = req_param->dlid;
			u_req_param->slid = req_param->slid;
			u_req_param->port = req_param->port;

			u_req_param->remote_private_data_len =
			    req_param->remote_private_data_len;

			memcpy(cm_comp->params + sizeof(*u_req_param),
			       req_param->remote_private_data,
			       req_param->remote_private_data_len);
			/*
			 * We also need to create a new connection structure and update the
			 * callback argument.
			 */
			ncm_conn =
			    tsIbCmUserConnectionAlloc(cm_conn->priv,
						      GFP_KERNEL);

			/*
			 * Send an MRA
			 */
			rc = tsIbCmDelayResponse(comm_id, TS_IB_UCM_MRA_TIMEOUT,
						 NULL, 0);
			if (rc)
				TS_REPORT_WARN(MOD_KERNEL_IB,
					       "tsIbCmDelayResponse() failed: %d",
					       rc);

			tsIbCmCallbackModify(comm_id,
					     tsIbCmUserCompletionHandler,
					     ncm_conn);
			if (NULL != ncm_conn) {
				struct ib_useraccess_cm_private *priv =
				    ncm_conn->priv;

				ncm_conn->comm_id = comm_id;

				down(&priv->cm_conn_sem);
				list_add(&ncm_conn->list, &priv->cm_conn_list);
				up(&priv->cm_conn_sem);

				tsIbCmUserConnectionPut(ncm_conn);	/* for the Alloc() */
			}

			ret = TS_IB_CM_CALLBACK_DEFER;
		}
		break;
	case TS_IB_CM_IDLE:
		{
            struct ib_cm_idle_param *idle_param =
			    (struct ib_cm_idle_param *) params;
			struct ib_u_cm_idle_param *u_idle_param;

			cm_comp->params_size = sizeof(*u_idle_param) +
			    idle_param->rej_info_len;
			cm_comp->params =
			    kmalloc(cm_comp->params_size, GFP_KERNEL);
			if (!cm_comp->params) {
				TS_REPORT_WARN(MOD_KERNEL_IB,
					       "Unable to allocate memory for IDLE param struct");
				kfree(cm_comp);
				goto out;
			}

			u_idle_param =
			          (struct ib_u_cm_idle_param *) cm_comp->params;

			u_idle_param->reason = idle_param->reason;
			u_idle_param->rej_reason = idle_param->rej_reason;

			u_idle_param->rej_info_len = idle_param->rej_info_len;

			memcpy(cm_comp->params + sizeof(*u_idle_param),
			       idle_param->rej_info, idle_param->rej_info_len);

			/* Final cleanup of connection */
			cm_conn->qp = NULL;

			EVAPI_k_clear_destroy_qp_cbk(cm_conn->v_device,
						     cm_conn->vk_qp);
			cm_conn->vk_qp = VAPI_INVAL_HNDL;

			/*
			 * We don't use ConnectionDestroy() since it will delete the completion
			 * for the IDLE state change that we want to be delivered to userspace
             * Freeing from the conn list is done in the beginning of the function
			 */

			
			/* We shouldn't get anymore callbacks, but just in case... */
			/* No more just in case, since we can deadlock here */
#if 0
			tsIbCmCallbackModify(cm_conn->comm_id, NULL, NULL);
#endif
            if (freeing_conn){
                tsIbCmUserConnectionPut(cm_conn);	/* drop the structure */
            }
			
		}
		break;
	case TS_IB_CM_REP_RECEIVED:
		{
			struct ib_cm_rep_received_param *rep_param =
			    (struct ib_cm_rep_received_param *) params;
			struct ib_u_cm_rep_received_param *u_rep_param;

			cm_comp->params_size = sizeof(*u_rep_param) +
			    rep_param->remote_private_data_len;
			cm_comp->params =
			    kmalloc(cm_comp->params_size, GFP_KERNEL);
			if (!cm_comp->params) {
				TS_REPORT_WARN(MOD_KERNEL_IB,
					       "Unable to allocate memory for REP_RECV param struct");
				kfree(cm_comp);
				goto out;
			}

			u_rep_param =
			  (struct ib_u_cm_rep_received_param *) cm_comp->params;

			u_rep_param->local_qpn = rep_param->local_qpn;
			u_rep_param->remote_qpn = rep_param->remote_qpn;

			u_rep_param->remote_private_data_len =
			    rep_param->remote_private_data_len;

			memcpy(cm_comp->params + sizeof(*u_rep_param),
			       rep_param->remote_private_data,
			       rep_param->remote_private_data_len);
			ret = TS_IB_CM_CALLBACK_DEFER;
		}
		break;
	case TS_IB_CM_ESTABLISHED:
		/* struct ib_cm_established_param * is 0 bytes */
		cm_comp->params = NULL;
		cm_comp->params_size = 0;
		break;
	case TS_IB_CM_DISCONNECTED:
		{
			struct ib_cm_disconnected_param *disconnected_param =
			    (struct ib_cm_disconnected_param *) params;
			struct ib_u_cm_disconnected_param *u_disconnected_param;

			cm_comp->params_size = sizeof(*u_disconnected_param);
			cm_comp->params =
			    kmalloc(cm_comp->params_size, GFP_KERNEL);
			if (!cm_comp->params) {
				TS_REPORT_WARN(MOD_KERNEL_IB,
					       "Unable to allocate memory for DISCONNECTED param struct");
				kfree(cm_comp);
				goto out;
			}

			u_disconnected_param =
			  (struct ib_u_cm_disconnected_param *) cm_comp->params;

			u_disconnected_param->reason =
			    disconnected_param->reason;
		}
		break;
		/* FIXME: Probably should do something on receiving these messages */
	case TS_IB_CM_LAP_RECEIVED:
	case TS_IB_CM_APR_RECEIVED:
		cm_comp->params = NULL;
		cm_comp->params_size = 0;
		break;
	}

	spin_lock_irqsave(&cm_conn->priv->cm_comp_lock, flags);
	list_add_tail(&cm_comp->list, &cm_conn->priv->cm_comp_list);
	spin_unlock_irqrestore(&cm_conn->priv->cm_comp_lock, flags);

    wake_up_interruptible(&cm_conn->priv->cm_comp_wait);

      out:
	tsIbCmUserConnectionPut(cm_conn);	/* for the Get() in the beginning */
	  if (qp_attr) kfree(qp_attr);

	return ret;
}

struct pathrecordlookup {
	struct semaphore sem;
	int status;
	struct ib_path_record path_record;
};

static int tsIbUserPathRecordCompletion(tIP2PR_PATH_LOOKUP_ID plid,
					s32 status,
					u32 src_addr,
					u32 dst_addr,
					tTS_IB_PORT hw_port,
					struct ib_device *ca,
					struct ib_path_record *path,
					tPTR usr_arg)
{
	struct pathrecordlookup *prl = (struct pathrecordlookup *)usr_arg;

	prl->status = status;
	if (!status) {
		memcpy(&prl->path_record, path, sizeof(prl->path_record));
	}

	up(&prl->sem);

	return 0;
}

int tsIbUserPathRecord(struct ib_useraccess_cm_private *priv, unsigned long arg)
{
	struct ib_path_record_ioctl pathrecord_ioctl;
	tIP2PR_PATH_LOOKUP_ID plid;
	struct pathrecordlookup *prl;
	int rc;

	prl = kmalloc(sizeof *prl, GFP_KERNEL);
	if ( !prl ) {
		return -ENOMEM;
	}

	if (copy_from_user(&pathrecord_ioctl,
			   (struct ib_path_record_ioctl *) arg,
			   sizeof pathrecord_ioctl)) {
		rc = -EFAULT;
		goto exit;
	}

	sema_init(&prl->sem, 0);

	rc = tsIp2prPathRecordLookup(pathrecord_ioctl.dst_addr,
				     0, 0, 0,
				     tsIbUserPathRecordCompletion, prl, &plid);
	if (rc < 0) {
		goto exit;
	}

	rc = down_interruptible(&prl->sem);
	if (rc) {
		tsIp2prPathRecordCancel(plid);
	}

	if (prl->status) {
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "Path record lookup completion status: %d",
			       prl->status);
		rc = prl->status;
		goto exit;
	}

	rc = copy_to_user(pathrecord_ioctl.path_record,
			    &prl->path_record,
			    sizeof prl->path_record) ? -EFAULT : 0;
exit:
	if ( prl ) kfree(prl);
	return rc;
}

void tsIbUserQpDestroyCallback(VAPI_hca_hndl_t hca_hndl,
			       VAPI_k_qp_hndl_t k_qp_hndl, void *private_data)
{
	struct ib_cm_user_connection *cm_conn =
	    (struct ib_cm_user_connection *) private_data;

	/* Destroying the callback is implicit when we received it */
	cm_conn->qp    = NULL;
	cm_conn->vk_qp = VAPI_INVAL_HNDL;

	tsIbCmUserConnectionDestroy(cm_conn);
}

int tsIbCmUserConnect(struct ib_useraccess_cm_private *priv, unsigned long arg)
{
	struct ib_cm_connect_ioctl *connect_ioctl=NULL;
	void *req_private_data = NULL;
	tTS_IB_CM_COMM_ID comm_id;
	struct ib_cm_user_connection *cm_conn;
	struct ib_path_record *primary_path=NULL, *alternate_path=NULL;
	struct ib_qp_attribute *qp_attr=NULL;
	int rc, ret;
	VAPI_ret_t status;


	connect_ioctl = kmalloc(sizeof *connect_ioctl, GFP_KERNEL);
	if ( !connect_ioctl ) {
		rc = -ENOMEM;
		goto out_free;
	}

	primary_path = kmalloc(sizeof *primary_path, GFP_KERNEL);
	if ( !primary_path ) {
		rc = -ENOMEM;
		goto out_free;
	}


	alternate_path = kmalloc(sizeof *alternate_path, GFP_KERNEL);
	if ( !alternate_path ) {
		rc = -ENOMEM;
		goto out_free;
	}

	qp_attr = kmalloc(sizeof *qp_attr, GFP_KERNEL);
	if ( !qp_attr ) {
		rc = -ENOMEM;
		goto out_free;
	}

	if (copy_from_user(connect_ioctl,
			   (struct ib_cm_connect_ioctl *) arg,
			   sizeof *connect_ioctl)) {
		rc = -EFAULT;
		goto out_free;
	}

	if (connect_ioctl->primary_path) {
		if (copy_from_user(primary_path,
				   (struct ib_path_record *) connect_ioctl->
				   primary_path, sizeof *primary_path)) {
			rc = -EFAULT;
			goto out_free;
		}
	}

	if (connect_ioctl->alternate_path) {
		if (copy_from_user(alternate_path,
				   (struct ib_path_record *) connect_ioctl->
				   alternate_path, sizeof *alternate_path)) {
			rc = -EFAULT;
			goto out_free;
		}
	}

	if (connect_ioctl->req_private_data_len < 0) {
		rc = -EINVAL;
		goto out_free;
	}

	/* FIXME: Arbitrary limit */
	if (connect_ioctl->req_private_data_len > PAGE_SIZE) {
		rc = -E2BIG;
		goto out_free;
	}


	if (connect_ioctl->req_private_data_len != 0) {
		/* Allocate memory and copy it */
		req_private_data =
		    kmalloc(connect_ioctl->req_private_data_len, GFP_KERNEL);
		if (req_private_data == NULL) {
			rc = -ENOMEM;
			goto out_free;
		}

		if (copy_from_user(req_private_data,
				   connect_ioctl->req_private_data,
				   connect_ioctl->req_private_data_len)) {
			rc = -EFAULT;
			goto out_free;
		}
	} else {
		req_private_data = NULL;
	}

	/* Get any memory allocate errors out of the way first */
	cm_conn = tsIbCmUserConnectionAlloc(priv, GFP_KERNEL);
	if (NULL == cm_conn) {
		rc = -ENOMEM;
		goto out_free;
	}

	tsIbTavorDeviceQueryVapiHandle(priv->device->ib_device,
				       &cm_conn->v_device);
	cm_conn->cm_arg = connect_ioctl->cm_arg;

	status = EVAPI_k_set_destroy_qp_cbk(cm_conn->v_device,
					    connect_ioctl->k_qp_handle,
					    tsIbUserQpDestroyCallback, cm_conn);

	switch (status) {
	case VAPI_OK:
		break;
	default:
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "EVAPI_k_set_destroy_qp_cbk, return code = %d (%s)",
			       status, VAPI_strerror(status));
		rc = -EINVAL;

		goto out_destroy;
	}
	cm_conn->qp = connect_ioctl->qp_handle;
	cm_conn->vk_qp = connect_ioctl->k_qp_handle;

	/* set RC timeout to a value of 16 */
	primary_path->packet_life = 15;

	{
		struct ib_cm_active_param active_param = {
			.qp = connect_ioctl->qp_handle,
			.req_private_data = req_private_data,
			.req_private_data_len =
			    connect_ioctl->req_private_data_len,
			.responder_resources =
			    connect_ioctl->responder_resources,
			.initiator_depth = connect_ioctl->initiator_depth,
			.retry_count = connect_ioctl->retry_count,
			.rnr_retry_count = connect_ioctl->rnr_retry_count,
			.cm_response_timeout =
			    connect_ioctl->cm_response_timeout,
			.max_cm_retries = connect_ioctl->max_cm_retries
		};

		rc = tsIbCmConnect(&active_param,
				   connect_ioctl->
				   primary_path ? primary_path : NULL,
				   connect_ioctl->
				   alternate_path ? alternate_path : NULL,
				   connect_ioctl->service_id, 0,
				   tsIbCmUserCompletionHandler, cm_conn,
				   &comm_id);
	}

	if (0 == rc) {
		connect_ioctl->comm_id = comm_id;

		/* We can only add it to the list after we get the comm_id */
		cm_conn->comm_id = comm_id;

		down(&priv->cm_conn_sem);
		list_add(&cm_conn->list, &priv->cm_conn_list);
		up(&priv->cm_conn_sem);
	}

	/* use ret so we don't trash rc */
	qp_attr->valid_fields = IB_QP_ATTRIBUTE_STATE;
	ret = ib_qp_query(cm_conn->qp, qp_attr);
	if (ret == 0 && qp_attr->state >= 0) {
		connect_ioctl->qp_state = qp_attr->state;
	} else {
		connect_ioctl->qp_state = IB_QP_STATE_INVALID;
	}

	if (copy_to_user((struct ib_cm_connect_ioctl *) arg,
			 connect_ioctl, sizeof *connect_ioctl)) {
		rc = -EFAULT;
	}

      out_destroy:
	if (rc) {
		tsIbCmUserConnectionDestroy(cm_conn);	/* drop the structure */
	}

	tsIbCmUserConnectionPut(cm_conn);	/* for the Alloc() */

out_free:
	if ( alternate_path ) kfree(alternate_path);
	if ( primary_path ) kfree(primary_path);
	if ( connect_ioctl ) kfree(connect_ioctl);
	if ( req_private_data ) kfree(req_private_data);
	if ( qp_attr ) kfree(qp_attr);

	return rc;
}

int tsIbCmUserListen(struct ib_useraccess_cm_private *priv, unsigned long arg)
{
	struct ib_cm_listen_ioctl listen_ioctl;
	tTS_IB_LISTEN_HANDLE listen_handle;
	struct ib_cm_user_connection *cm_conn;
	int rc;

	if (copy_from_user(&listen_ioctl,
			   (struct ib_cm_listen_ioctl *) arg,
			   sizeof listen_ioctl)) {
		return -EFAULT;
	}

	/* Get any memory allocate errors out of the way first */
	cm_conn = tsIbCmUserConnectionAlloc(priv, GFP_KERNEL);
	if (NULL == cm_conn) {
		return -ENOMEM;
	}

	cm_conn->cm_arg = listen_ioctl.cm_arg;

	rc = tsIbCmListen(listen_ioctl.service_id,
			  listen_ioctl.service_mask,
			  tsIbCmUserCompletionHandler, cm_conn, &listen_handle);
	if (0 == rc) {
		listen_ioctl.listen_handle = listen_handle;

		cm_conn->listen_handle = listen_handle;

		/* We can only add it to the list after we get the comm_id */
		down(&priv->cm_conn_sem);
		list_add(&cm_conn->list, &priv->cm_conn_list);
		up(&priv->cm_conn_sem);
	}

	if (copy_to_user((struct ib_cm_listen_ioctl *) arg,
			 &listen_ioctl, sizeof listen_ioctl)) {
		rc = -EFAULT;
	}

	if (rc) {
		tsIbCmUserConnectionDestroy(cm_conn);	/* drop the structure */
	}

	tsIbCmUserConnectionPut(cm_conn);	/* for the Alloc() */

	return rc;
}

int
tsIbCmUserListenStop(struct ib_useraccess_cm_private *priv, unsigned long arg)
{
	tTS_IB_LISTEN_HANDLE listen_handle = (tTS_IB_LISTEN_HANDLE) arg;
	struct list_head *ptr, *tmp;

	if (!listen_handle) {
		return -EINVAL;
	}

	down(&priv->cm_conn_sem);
	list_for_each_safe(ptr, tmp, &priv->cm_conn_list) {
		struct ib_cm_user_connection *cm_conn;
		int ret;

		cm_conn = list_entry(ptr, struct ib_cm_user_connection, list);

		if (cm_conn->listen_handle == listen_handle) {
			cm_conn->listen_handle = NULL;
			list_del_init(&cm_conn->list);
			list_add(&cm_conn->list, &priv->free_conn_list);

			up(&priv->cm_conn_sem);

			ret = tsIbCmListenStop(listen_handle);

			tsIbCmUserConnectionPut(cm_conn);	/* drop the structure */

			return ret;
		}
	}
	up(&priv->cm_conn_sem);

	return -ENOENT;
}

int tsIbCmUserAccept(struct ib_useraccess_cm_private *priv, unsigned long arg)
{
	struct ib_cm_accept_ioctl accept_ioctl;
	void *reply_data=NULL;
	struct ib_cm_user_connection *cm_conn;
	struct ib_qp_attribute *qp_attr=NULL;
	int rc, ret;
	VAPI_ret_t status;


	qp_attr = kmalloc(sizeof(struct ib_qp_attribute), GFP_KERNEL);
	if ( !qp_attr ) {
		rc = -ENOMEM;
		goto out_free;
	}

	if (copy_from_user(&accept_ioctl,
			   (struct ib_cm_accept_ioctl *) arg,
			   sizeof accept_ioctl)) {
		rc = -EFAULT;
		goto out_free;
	}

	if (accept_ioctl.reply_size < 0) {
		rc = -EINVAL;
		goto out_free;
	}

	/* FIXME: Arbitrary limit */
	if (accept_ioctl.reply_size > PAGE_SIZE) {
		rc = -E2BIG;
		goto out_free;
	}

	if (accept_ioctl.reply_size != 0) {
		/* Allocate memory and copy it */
		reply_data = kmalloc(accept_ioctl.reply_size, GFP_KERNEL);
		if (reply_data == NULL) {
			rc = -ENOMEM;
			goto out_free;
		}

		if (copy_from_user(reply_data,
				   accept_ioctl.reply_data,
				   accept_ioctl.reply_size)) {
			rc = -EFAULT;
			goto out_free;
		}
	} else {
		reply_data = NULL;
	}

	cm_conn = tsIbCmUserConnectionFind(priv, accept_ioctl.comm_id);
	if (!cm_conn) {
		rc = -ENOENT;
		goto out_free;
	}

	tsIbTavorDeviceQueryVapiHandle(priv->device->ib_device,
				       &cm_conn->v_device);
	cm_conn->cm_arg = accept_ioctl.cm_arg;

	status = EVAPI_k_set_destroy_qp_cbk(cm_conn->v_device,
					    accept_ioctl.k_qp_handle,
					    tsIbUserQpDestroyCallback, cm_conn);
	switch (status) {
	case VAPI_OK:
		break;
	default:
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "EVAPI_k_set_destroy_qp_cbk, return code = %d (%s)",
			       status, VAPI_strerror(status));
		rc = -EINVAL;

		goto out_free_conn;
	}
	cm_conn->qp = accept_ioctl.qp_handle;
	cm_conn->vk_qp = accept_ioctl.k_qp_handle;

	{
		struct ib_cm_passive_param passive_params = {
			.qp = accept_ioctl.qp_handle,
			.reply_private_data = reply_data,
			.reply_private_data_len = accept_ioctl.reply_size,
		};

		rc = tsIbCmAccept(cm_conn->comm_id, &passive_params);
	}

	/* use ret so we don't trash rc */
	qp_attr->valid_fields = IB_QP_ATTRIBUTE_STATE;
	ret = ib_qp_query(cm_conn->qp, qp_attr);
	if (ret == 0 && qp_attr->state >= 0) {
		accept_ioctl.qp_state = qp_attr->state;
	} else {
		accept_ioctl.qp_state = IB_QP_STATE_INVALID;
	}

	if (copy_to_user((struct ib_cm_accept_ioctl *) arg,
			 &accept_ioctl, sizeof accept_ioctl)) {
		rc = -EFAULT;
	}

	if (rc) {
		tsIbCmUserConnectionDestroy(cm_conn);	/* drop the structure */
	}

      out_free_conn:
	tsIbCmUserConnectionPut(cm_conn);	/* for the Find() */

out_free:
	if ( reply_data) kfree(reply_data);
	if ( qp_attr ) kfree(qp_attr);

	return rc;
}

int tsIbCmUserReject(struct ib_useraccess_cm_private *priv, unsigned long arg)
{
	struct ib_cm_reject_ioctl reject_ioctl;
	struct ib_cm_user_connection *cm_conn;
	void *reply_data;
	int rc;

	if (copy_from_user(&reject_ioctl,
			   (struct ib_cm_reject_ioctl *) arg,
			   sizeof reject_ioctl)) {
		return -EFAULT;
	}

	if (reject_ioctl.reply_size < 0) {
		return -EINVAL;
	}

	/* FIXME: Arbitrary limit */
	if (reject_ioctl.reply_size > PAGE_SIZE) {
		return -E2BIG;
	}

	if (reject_ioctl.reply_size != 0) {
		/* Allocate memory and copy it */
		reply_data = kmalloc(reject_ioctl.reply_size, GFP_KERNEL);
		if (reply_data == NULL) {
			return -ENOMEM;
		}

		if (copy_from_user(reply_data,
				   reject_ioctl.reply_data,
				   reject_ioctl.reply_size)) {
			rc = -EFAULT;
			goto out_free;
		}
	} else {
		reply_data = NULL;
	}

	cm_conn = tsIbCmUserConnectionFind(priv, reject_ioctl.comm_id);
	if (!cm_conn) {
		rc = -ENOENT;
		goto out_free;
	}

	rc = tsIbCmReject(cm_conn->comm_id,
			  reply_data, reject_ioctl.reply_size);

	tsIbCmCallbackModify(cm_conn->comm_id, NULL, NULL);
	cm_conn->comm_id = TS_IB_CM_COMM_ID_INVALID;

	tsIbCmUserConnectionDestroy(cm_conn);
	tsIbCmUserConnectionPut(cm_conn);	/* for the Find() */

      out_free:
	kfree(reply_data);

	return rc;
}

int tsIbCmUserConfirm(struct ib_useraccess_cm_private *priv, unsigned long arg)
{
	struct ib_cm_confirm_ioctl confirm_ioctl;
	struct ib_qp_attribute *qp_attr=NULL;
	struct ib_cm_user_connection *cm_conn;
	void *reply_data=NULL;
	int rc, ret;

	qp_attr = kmalloc(sizeof(struct ib_qp_attribute), GFP_KERNEL);
	if ( !qp_attr ) {
		rc = -ENOMEM;
		goto out_free;
	}

	if (copy_from_user(&confirm_ioctl,
			   (struct ib_cm_confirm_ioctl *) arg,
			   sizeof confirm_ioctl)) {
		rc = -EFAULT;
		goto out_free;
	}

	if (confirm_ioctl.reply_size < 0) {
		rc = -EINVAL;
		goto out_free;
	}

	/* FIXME: Arbitrary limit */
	if (confirm_ioctl.reply_size > PAGE_SIZE) {
		rc = -E2BIG;
		goto out_free;
	}

	if (confirm_ioctl.reply_size != 0) {
		/* Allocate memory and copy it */
		reply_data = kmalloc(confirm_ioctl.reply_size, GFP_KERNEL);
		if (reply_data == NULL) {
			rc = -ENOMEM;
			goto out_free;
		}

		if (copy_from_user(reply_data,
				   confirm_ioctl.reply_data,
				   confirm_ioctl.reply_size)) {
			rc = -EFAULT;
			goto out_free;
		}
	} else {
		reply_data = NULL;
	}

	cm_conn = tsIbCmUserConnectionFind(priv, confirm_ioctl.comm_id);
	if (!cm_conn) {
		rc = -ENOENT;
		goto out_free;
	}

	rc = tsIbCmConfirm(cm_conn->comm_id,
			   reply_data, confirm_ioctl.reply_size);

	/* use ret so we don't trash rc */
	qp_attr->valid_fields = IB_QP_ATTRIBUTE_STATE;
	ret = ib_qp_query(cm_conn->qp, qp_attr);
	if (ret == 0 && qp_attr->state >= 0) {
		confirm_ioctl.qp_state = qp_attr->state;
	} else {
		confirm_ioctl.qp_state = IB_QP_STATE_INVALID;
	}

    if (rc) {
		tsIbCmUserConnectionDestroy(cm_conn);	/* drop the structure */
	}

	tsIbCmUserConnectionPut(cm_conn);	/* for the Find() */

	if (copy_to_user((struct ib_cm_confirm_ioctl *) arg,
			 &confirm_ioctl, sizeof confirm_ioctl)) {
		rc = -EFAULT;
	}

out_free:
	if (reply_data) kfree(reply_data);
	if ( qp_attr ) kfree(qp_attr);

	return rc;
}

int
tsIbCmUserEstablish(struct ib_useraccess_cm_private *priv, unsigned long arg)
{
	struct ib_cm_establist_ioctl establish_ioctl;
	struct ib_cm_user_connection *cm_conn;
	struct ib_qp_attribute *qp_attr=NULL;
	int rc, ret;


	qp_attr = kmalloc(sizeof *qp_attr, GFP_KERNEL);
	if ( !qp_attr ) {
		rc = -ENOMEM;
		goto exit;
	}

	if (copy_from_user(&establish_ioctl,
			   (struct ib_cm_establist_ioctl *) arg,
			   sizeof establish_ioctl)) {
		rc = -EFAULT;
		goto exit;
	}

	cm_conn = tsIbCmUserConnectionFind(priv, establish_ioctl.comm_id);
	if (!cm_conn) {
		rc = -ENOENT;
		goto exit;
	}

	/* Always immediate for now atleast */
	rc = tsIbCmEstablish(cm_conn->comm_id, 1);

	/* use ret so we don't trash rc */
	qp_attr->valid_fields = IB_QP_ATTRIBUTE_STATE;
	ret = ib_qp_query(cm_conn->qp, qp_attr);
	if (ret == 0 && qp_attr->state >= 0) {
		establish_ioctl.qp_state = qp_attr->state;
	} else {
		establish_ioctl.qp_state = IB_QP_STATE_INVALID;
	}

	tsIbCmUserConnectionPut(cm_conn);	/* for the Find() */

	if (copy_to_user((struct ib_cm_establist_ioctl *) arg,
			 &establish_ioctl, sizeof establish_ioctl)) {
		rc = -EFAULT;
	}

exit:
	if ( qp_attr ) kfree(qp_attr);
	return rc;
}

int tsIbCmUserDelayResponse(struct ib_useraccess_cm_private *priv,
			    unsigned long arg)
{
	struct ib_cm_delayresponse_ioctl delay_ioctl;
	struct ib_cm_user_connection *cm_conn;
	void *mra_data;
	int rc;

	if (copy_from_user(&delay_ioctl,
			   (struct ib_cm_delayresponse_ioctl *) arg,
			   sizeof delay_ioctl)) {
		return -EFAULT;
	}

	if (delay_ioctl.mra_data_size < 0) {
		return -EINVAL;
	}

	/* FIXME: Arbitrary limit */
	if (delay_ioctl.mra_data_size > PAGE_SIZE) {
		return -E2BIG;
	}

	if (delay_ioctl.mra_data_size != 0) {
		/* Allocate memory and copy it */
		mra_data = kmalloc(delay_ioctl.mra_data_size, GFP_KERNEL);
		if (mra_data == NULL) {
			return -ENOMEM;
		}

		if (copy_from_user(mra_data,
				   delay_ioctl.mra_data,
				   delay_ioctl.mra_data_size)) {
			rc = -EFAULT;
			goto out_free;
		}
	} else {
		mra_data = NULL;
	}

	cm_conn = tsIbCmUserConnectionFind(priv, delay_ioctl.comm_id);
	if (!cm_conn) {
		rc = -ENOENT;
		goto out_free;
	}

	rc = tsIbCmDelayResponse(cm_conn->comm_id,
				 delay_ioctl.service_timeout,
				 mra_data, delay_ioctl.mra_data_size);

	tsIbCmUserConnectionPut(cm_conn);	/* for the Find() */

      out_free:
	kfree(mra_data);

	return rc;
}

int
tsIbCmUserDropConsumer(struct ib_useraccess_cm_private *priv, unsigned long arg)
{
	tTS_IB_CM_COMM_ID comm_id = (tTS_IB_CM_COMM_ID) arg;
	struct ib_cm_user_connection *cm_conn;
    unsigned long flags;
    struct list_head *ptr, *tmp;

    spin_lock_irqsave(&priv->cm_comp_lock, flags);
    list_for_each_safe(ptr, tmp, &priv->cm_comp_list) {
        struct ib_cm_user_compeletion *cm_comp;

        cm_comp = list_entry(ptr, struct ib_cm_user_compeletion,
                     list);

        if (cm_comp->comm_id == comm_id) {
            list_del_init(&cm_comp->list);
            kfree(cm_comp);
        }
    }
    spin_unlock_irqrestore(&priv->cm_comp_lock, flags);

    wait_event_interruptible(priv->cleanup_wait, 
                             priv->last_comm_id != comm_id || priv->callbck_holder == current);
    
	cm_conn = tsIbCmUserConnectionFind(priv, comm_id);
	if (!cm_conn) {
		return -ENOENT;
	}

	tsIbCmUserConnectionDestroy(cm_conn);
	tsIbCmUserConnectionPut(cm_conn);	/* for the Find() */

	return 0;
}

int tsIbCmUserGetCompletion(struct ib_useraccess_cm_private *priv,
			    unsigned long arg)
{
	struct ib_u_cm_get_completion_ioctl getcomp_ioctl;
	struct ib_cm_user_compeletion *cm_comp;
	unsigned long flags;

	if (copy_from_user(&getcomp_ioctl,
			   (struct ib_u_cm_get_completion_ioctl *) arg,
			   sizeof getcomp_ioctl)) {
		return -EFAULT;
	}

	/* Wake all user thread waiting for last callback to finish */
	priv->last_comm_id = 0;
    priv->callbck_holder = NULL;

	wake_up_interruptible_all(&priv->cleanup_wait);
	
	spin_lock_irqsave(&priv->cm_comp_lock, flags);
	while (list_empty(&priv->cm_comp_list)) {
		spin_unlock_irqrestore(&priv->cm_comp_lock, flags);

		if (wait_event_interruptible(priv->cm_comp_wait,
					     !list_empty(&priv->
							 cm_comp_list))) {
			return -ERESTARTSYS;
		}

		spin_lock_irqsave(&priv->cm_comp_lock, flags);
	}

	cm_comp = list_entry(priv->cm_comp_list.next,
			     struct ib_cm_user_compeletion, list);
	list_del_init(&cm_comp->list);

	 priv->last_comm_id = getcomp_ioctl.comm_id;
     priv->callbck_holder = current;

	spin_unlock_irqrestore(&priv->cm_comp_lock, flags);

	getcomp_ioctl.qp_handle = cm_comp->qp;
	getcomp_ioctl.comm_id = cm_comp->comm_id;
	getcomp_ioctl.listen_handle = cm_comp->listen_handle;
	getcomp_ioctl.event = cm_comp->event;
	getcomp_ioctl.qp_state = cm_comp->qp_state;
	getcomp_ioctl.cm_arg = cm_comp->cm_arg;

	if (getcomp_ioctl.params_size < cm_comp->params_size) {
		/* FIXME: Should we throw away the buffer here? */
		kfree(cm_comp->params);
		kfree(cm_comp);
		return -ENOSPC;
	}

	if (cm_comp->params_size > 0) {
		if (copy_to_user(getcomp_ioctl.params,
				 cm_comp->params, cm_comp->params_size)) {
			kfree(cm_comp->params);
			kfree(cm_comp);
			return -EFAULT;
		}
	}

	getcomp_ioctl.params_size = cm_comp->params_size;

	kfree(cm_comp->params);
	kfree(cm_comp);

	return copy_to_user((struct ib_u_cm_get_completion_ioctl *) arg,
			    &getcomp_ioctl, sizeof getcomp_ioctl) ? -EFAULT : 0;
}

int tsIbCmUserClose(struct ib_useraccess_cm_private *priv)
{
	struct list_head *ptr, *tmp;
	unsigned long flags;

    TS_REPORT_DATA(MOD_KERNEL_IB, "DEBUG - start closing user-prive %p",priv);

	spin_lock_irqsave(&priv->cm_comp_lock, flags);
	list_for_each_safe(ptr, tmp, &priv->cm_comp_list) {
		struct ib_cm_user_compeletion *cm_comp;

		cm_comp = list_entry(ptr, struct ib_cm_user_compeletion, list);
		list_del_init(&cm_comp->list);
		kfree(cm_comp);
	}
	spin_unlock_irqrestore(&priv->cm_comp_lock, flags);

	/*
	 * We need to drop priv->com_conn_sem because we can deadlock with the
	 * completion handler (locks are obtained in opposite order from each
	 * other, a "deadly embrace").
	 */
	down(&priv->cm_conn_sem);
	while (!list_empty(&priv->cm_conn_list)) {
		struct ib_cm_user_connection *cm_conn;

		cm_conn = list_entry(priv->cm_conn_list.next,
			       struct ib_cm_user_connection, list);

		list_del_init(&cm_conn->list);
		list_add(&cm_conn->list, &priv->free_conn_list);

		/* Drop and reacquire lock since destroying the connection may callback */
		up(&priv->cm_conn_sem);

		_tsIbCmUserConnectionDestroy(cm_conn);

        tsIbCmUserConnectionPut(cm_conn); /*legal after list_del_init */

		down(&priv->cm_conn_sem);
	}
	up(&priv->cm_conn_sem);

    spin_lock_irqsave(&priv->cm_comp_lock, flags);
	list_for_each_safe(ptr, tmp, &priv->cm_comp_list) {
		struct ib_cm_user_compeletion *cm_comp;

		cm_comp = list_entry(ptr, struct ib_cm_user_compeletion, list);
		list_del_init(&cm_comp->list);
		kfree(cm_comp);
	}
	spin_unlock_irqrestore(&priv->cm_comp_lock, flags);


	down(&priv->cm_conn_sem);
	while (!list_empty(&priv->free_conn_list)) {
		struct ib_cm_user_connection *cm_conn;

		cm_conn = list_entry(priv->free_conn_list.next,
			       struct ib_cm_user_connection, list);
		list_del_init(&cm_conn->list);
		XFREE(cm_conn);
	}
	up(&priv->cm_conn_sem);

    TS_REPORT_DATA(MOD_KERNEL_IB, "DEBUG - finish closing user-prive %p",priv);

	return 0;
}

/* =============================================================== */
/*..tsIbCmUserIoctl -                                              */
int tsIbCmUserIoctl(struct inode *inode,
		    struct file *filp, unsigned int cmd, unsigned long arg)
{
	struct ib_useraccess_cm_private *priv =
	    (struct ib_useraccess_cm_private *) filp->private_data;
	int rc;

	switch (cmd) {
	case TS_IB_IOCCMCONNECT:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "initiate connection ioctl");
		return tsIbCmUserConnect(priv, arg);

	case TS_IB_IOCCMLISTEN:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "connection listen ioctl");
		return tsIbCmUserListen(priv, arg);

	case TS_IB_IOCCMLISTENSTOP:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "stop listening connection ioctl");
		return tsIbCmUserListenStop(priv, arg);

	case TS_IB_IOCCMDISCONNECT:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "disconnect CM connection ioctl");
		/* We'll cleanup everything when see the connection go to IDLE */
		/* FIXME: We should validate the connection id from userspace */
		return tsIbCmDisconnect((tTS_IB_CM_COMM_ID) arg);

	case TS_IB_IOCCMALTPATHLOAD:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "load alternate path ioctl");
		{
			struct ib_cm_alternate_path_load_ioctl
			    altpathload_ioctl;
			struct ib_path_record *path_record=NULL;

			path_record = kmalloc(sizeof(struct ib_path_record), GFP_KERNEL);
			if ( !path_record ) {
				return -ENOMEM;
			}

			if (copy_from_user(&altpathload_ioctl,
					   (struct
					   ib_cm_alternate_path_load_ioctl *)
					   arg, sizeof altpathload_ioctl)) {
				rc = -EFAULT;
				goto exit_TS_IB_IOCCMALTPATHLOAD;
			}

			if (copy_from_user(path_record,
					   (struct ib_path_record *)
					   altpathload_ioctl.alternate_path,
					   sizeof *path_record)) {
				rc = -EFAULT;
				goto exit_TS_IB_IOCCMALTPATHLOAD;
			}

			/* FIXME: We should validate the connection id from userspace */
			rc = tsIbCmAlternatePathLoad(altpathload_ioctl.
						       comm_id, path_record);
		exit_TS_IB_IOCCMALTPATHLOAD:
			if ( path_record ) kfree(path_record);
			return rc;
		}

	case TS_IB_IOCCMACCEPT:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "accept CM connection ioctl");
		return tsIbCmUserAccept(priv, arg);

	case TS_IB_IOCCMREJECT:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "reject CM connection ioctl");
		return tsIbCmUserReject(priv, arg);

	case TS_IB_IOCCMCONFIRM:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "confirm CM connection ioctl");
		return tsIbCmUserConfirm(priv, arg);

	case TS_IB_IOCCMESTABLISH:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "notify CM connection is establed ioctl");
		return tsIbCmUserEstablish(priv, arg);

	case TS_IB_IOCCMDELAYRESPONSE:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "delay response ioctl");
		return tsIbCmUserDelayResponse(priv, arg);

	case TS_IB_IOCCMDROPCONSUMER:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "drop CM consumer ioctl");
		return tsIbCmUserDropConsumer(priv, arg);

	case TS_IB_IOCCMGETCOMPLETION:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "get completion ioctl");
		return tsIbCmUserGetCompletion(priv, arg);

	case TS_IB_IOCQPREGISTER:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "register QP with kernel ioctl");
		{
			struct ib_qp_register_ioctl qpregister_ioctl;
			struct ib_tavor_qp_create_param tavor_qp_create;
			struct ib_qp *qp_handle;
			u32 qpn;
			int rc;
			struct ib_qp_create_param *qp_create=NULL;

			qp_create = kmalloc(sizeof(*qp_create), GFP_KERNEL);
			if ( !qp_create ) {
				rc = -ENOMEM;
				goto exit_TS_IB_IOCQPREGISTER;
			}
			memset(qp_create, 0, sizeof(*qp_create));
			qp_create->pd = priv->device->pd;
			qp_create->transport = qpregister_ioctl.transport;
			qp_create->device_specific = &tavor_qp_create;

			if (copy_from_user(&qpregister_ioctl,
					   (struct ib_qp_register_ioctl *) arg,
					   sizeof qpregister_ioctl)) {
				rc = -EFAULT;
				goto exit_TS_IB_IOCQPREGISTER;
			}

			tavor_qp_create.is_user     = 1;
			tavor_qp_create.vapi_handle =
			    (void *) (unsigned long) qpregister_ioctl.vk_qp_handle;
			tavor_qp_create.qpn         = qpregister_ioctl.qpn;
			tavor_qp_create.qp_state    = qpregister_ioctl.qp_state;

			rc = ib_qp_create(qp_create, &qp_handle, &qpn);
			if (rc) {
				goto exit_TS_IB_IOCQPREGISTER;
			}

			qpregister_ioctl.qp_handle = qp_handle;

			rc = copy_to_user((struct ib_qp_register_ioctl *) arg,
					    &qpregister_ioctl,
					    sizeof qpregister_ioctl) ? -EFAULT :
			    0;

			exit_TS_IB_IOCQPREGISTER:
				if ( qp_create ) kfree(qp_create);
				return rc;
		}

	case TS_IB_IOCGPATHRECORD:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "perform path record lookup ioctl");
		return tsIbUserPathRecord(priv, arg);

	case TS_IB_IOCCMSERVIDASSIGN:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "assign a service id ioctl");
		{
			tTS_IB_SERVICE_ID servid = tsIbCmServiceAssign();

			return copy_to_user((tTS_IB_SERVICE_ID *) arg,
					    &servid,
					    sizeof servid) ? -EFAULT : 0;
		}

        case TS_IB_IOCMGETREMOTEGID:
                TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
                         "remote gid get ioctl");

                {
                    int rc;
                    struct ib_cm_get_remote_gid_ioctl rgid_ioctl;
                    tTS_IB_GID gid;

                    
                    if (copy_from_user(&rgid_ioctl,
                                       (struct ib_cm_get_remote_gid_ioctl *) arg,
                                       sizeof rgid_ioctl)){
                        return -EFAULT;
                    }

                    
                    rc = ib_cm_get_remote_gid(rgid_ioctl.comm_id, &gid);
                    if(rc < 0)
                    {
                        return rc;
                    }

                    return copy_to_user(rgid_ioctl.gid, gid, sizeof gid);

                   
                }



	default:
		TS_REPORT_WARN(MOD_KERNEL_IB, "Unimplemented ioctl %d", cmd);
		return -ENOIOCTLCMD;
	}

	return 0;
}

/* =============================================================== */
/*.._tsIbCmUserOpen -                                              */
static int _tsIbCmUserOpen(struct inode *inode, struct file *filp)
{

	struct ib_useraccess_cm_device *dev;
	struct ib_useraccess_cm_private *priv;
	int dev_num;

	if (filp->private_data) {
		dev = (struct ib_useraccess_cm_device *) filp->private_data;
	} else {
		dev_num = MINOR(inode->i_rdev);

		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "device number %d", dev_num);

		/* not using devfs, use minor number */
		dev = &useraccess_cm_dev_list[dev_num];
	}

	if (!dev->ib_device) {
		return -ENODEV;
	}

	priv = kmalloc(sizeof(*priv), GFP_KERNEL);
	if (!priv) {
		return -ENOMEM;
	}

	priv->device = dev;

	priv->last_comm_id = 0;
    priv->callbck_holder = NULL;
	init_waitqueue_head(&priv->cleanup_wait);


	INIT_LIST_HEAD(&priv->cm_comp_list);
	init_waitqueue_head(&priv->cm_comp_wait);
	spin_lock_init(&priv->cm_comp_lock);

	INIT_LIST_HEAD(&priv->cm_conn_list);
	INIT_LIST_HEAD(&priv->free_conn_list);
	sema_init(&priv->cm_conn_sem, 1);

	filp->private_data = priv;

	return 0;
}

/* =============================================================== */
/*.._tsIbCmUserClose -                                             */
static int _tsIbCmUserClose(struct inode *inode, struct file *filp)
{
	struct ib_useraccess_cm_private *priv =
	    (struct ib_useraccess_cm_private *) filp->private_data;

	tsIbCmUserClose(priv);
	kfree(priv);

#ifdef MT_MEMTRACK
	memtrack_report();
#endif
	return 0;
}

static struct file_operations useraccess_cm_fops = {
	.owner = THIS_MODULE,
	.ioctl = tsIbCmUserIoctl,
	.open = _tsIbCmUserOpen,
	.release = _tsIbCmUserClose,
};

#if defined(TS_KERNEL_2_6)

/* Create devices for kernel 2.4 */
static int __init _tsIbCmUserCreateDevices(void
    )
{
	int ret;

	ret = alloc_chrdev_region(&useraccess_cm_devnum,
				  0,
				  TS_USERACCESS_CM_NUM_DEVICE,
				  (char *)TS_USERACCESS_CM_NAME);
	if (ret) {
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"Couldn't allocate device numbers for useraccess_cm module");
		return ret;
	}

	cdev_init(&useraccess_cm_cdev, &useraccess_cm_fops);
	useraccess_cm_cdev.owner = THIS_MODULE;
	kobject_set_name(&useraccess_cm_cdev.kobj, TS_USERACCESS_CM_NAME);
	ret = cdev_add(&useraccess_cm_cdev,
		       useraccess_cm_devnum, TS_USERACCESS_CM_NUM_DEVICE);
	if (ret) {
		kobject_put(&useraccess_cm_cdev.kobj);
		unregister_chrdev_region(useraccess_cm_devnum,
					 TS_USERACCESS_CM_NUM_DEVICE);
	}

	return ret;
}

#else				/* TS_KERNEL_2_6 */

/* Create devices for kernel 2.4 */
static int __init _tsIbCmUserCreateDevices(void
    )
{
	useraccess_cm_major = devfs_register_chrdev(0,
						    TS_USERACCESS_CM_NAME,
						    &useraccess_cm_fops);
	if (useraccess_cm_major < 0) {
		TS_REPORT_FATAL(MOD_KERNEL_IB, "Failed to register device");
		return useraccess_cm_major;
	}

	TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
		 "TS IB user access major %d", useraccess_cm_major);

	useraccess_cm_devfs_dir = devfs_mk_dir(NULL, "ts_ua", NULL);

	{
		int i;
		char name[4];

		for (i = 0; i < TS_USERACCESS_CM_NUM_DEVICE; ++i) {
			snprintf(name, sizeof name, "%02d", i);

			useraccess_cm_dev_list[i].devfs_handle =
			    devfs_register(useraccess_cm_devfs_dir,
					   name,
					   DEVFS_FL_DEFAULT,
					   useraccess_cm_major,
					   i,
					   S_IFCHR | S_IRUSR | S_IWUSR,
					   &useraccess_cm_fops,
					   &useraccess_cm_dev_list[i]);

			TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
				 "TS IB add using major %d, minor %d",
				 useraccess_cm_major, i);
		}
	}

	return 0;
}

#endif				/* TS_KERNEL_2_6 */

static int __init _tsIbCmUserInitModule(void
    )
{
	int i;
	int ret;

	for (i = 0; i < TS_USERACCESS_CM_NUM_DEVICE; ++i) {
		useraccess_cm_dev_list[i].ib_device = ib_device_get_by_index(i);
		if (useraccess_cm_dev_list[i].ib_device) {
			ret = ib_pd_create(useraccess_cm_dev_list[i].ib_device,
                                           NULL, &useraccess_cm_dev_list[i].pd);
			if (ret) {
				TS_REPORT_WARN(MOD_KERNEL_IB,
					       "couldn't create PD (ret = %d)",
					       ret);
				useraccess_cm_dev_list[i].ib_device = NULL;
				continue;
			}
		}
	}

	ret = _tsIbCmUserCreateDevices();
	if (ret) {
		for (i = 0; i < TS_USERACCESS_CM_NUM_DEVICE; ++i) {
			if (useraccess_cm_dev_list[i].pd) {
				ib_pd_destroy(useraccess_cm_dev_list[i].pd);
			}
		}
	}

	return ret;
}

#if defined(TS_KERNEL_2_6)

/* Delete devices for kernel 2.6 */
static void __exit _tsIbCmUserDeleteDevices(void
    )
{
	unregister_chrdev_region(useraccess_cm_devnum,
				 TS_USERACCESS_CM_NUM_DEVICE);
	cdev_del(&useraccess_cm_cdev);
}

#else				/* TS_KERNEL_2_6 */

/* Delete devices for kernel 2.4 */
static void __exit _tsIbCmUserDeleteDevices(void
    )
{
	devfs_unregister_chrdev(useraccess_cm_major, TS_USERACCESS_CM_NAME);

	{
		int i;

		for (i = 0; i < TS_USERACCESS_CM_NUM_DEVICE; ++i) {
			if (useraccess_cm_dev_list[i].devfs_handle) {
				devfs_unregister(useraccess_cm_dev_list[i].
						 devfs_handle);
			}
		}
	}

	devfs_unregister(useraccess_cm_devfs_dir);
}

#endif				/* TS_KERNEL_2_6 */

static void __exit _tsIbCmUserCleanupModule(void
    )
{
	int i;

	TS_REPORT_CLEANUP(MOD_KERNEL_IB, "Unloading IB userspace access");

	_tsIbCmUserDeleteDevices();

	for (i = 0; i < TS_USERACCESS_CM_NUM_DEVICE; ++i) {
		if (useraccess_cm_dev_list[i].pd) {
			ib_pd_destroy(useraccess_cm_dev_list[i].pd);
		}
	}

	TS_REPORT_CLEANUP(MOD_KERNEL_IB, "IB userspace access unloaded");
}

module_init(_tsIbCmUserInitModule);
module_exit(_tsIbCmUserCleanupModule);
