/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: cm.c,v 1.2 2004/08/01 10:59:42 rabin Exp $
*/

/*
 * System includes
 */
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>

/*
 * Local includes
 */
#include "list.h"


/*
 * TS specific includes
 */
#include "ts_ib_useraccess_cm.h"	/* ioctl API to kernel */



/* Mellanox stuff */
#include "vapi_types.h"
#include "vapi.h"
#include "evapi.h"


/* FIXME: Convert connection_list to something faster (keyed on comm_id) */

struct cm_connection {
  struct list_head              list;
  unsigned int                  refcnt;

  tTS_IB_CM_COMM_ID             comm_id;
  tTS_IB_U_CM_COMPLETION_FUNC   cm_func;
  void                         *cm_arg;

  VAPI_hca_hndl_t               hca_handle;
  //tTS_IB_QP_HANDLE              t_qp_handle;	/* TS API handle */  // EFIX
  void * t_qp_handle;
  
  tTS_IB_U_QP_HANDLE            v_qp_handle;	/* VAPI Userspace handle */
  tTS_IB_U_K_QP_HANDLE          vk_qp_handle;	/* VAPI Kernel handle */
};

struct cm_listen {
  struct list_head              list;
  unsigned int                  refcnt;

  tTS_IB_LISTEN_HANDLE          listen_handle;
  tTS_IB_U_CM_COMPLETION_FUNC   cm_func;
  void                         *cm_arg;
};

static int ua_fd = -1;
static pthread_t completion_thread;

static pthread_mutex_t connection_lock = PTHREAD_MUTEX_INITIALIZER;
static struct list_head connection_list;

static pthread_mutex_t listen_lock = PTHREAD_MUTEX_INITIALIZER;
static struct list_head listen_list;

static int _tsIbUQpRegister(
                            struct ib_device    *device,
                            u_int32_t            qpn,
                            enum ib_transport    transport,
                            tTS_IB_U_K_QP_HANDLE vk_qp_handle,
                            void                **qp_handle
                            );

/* ========================================================================= */
/*..tsIbUCmListenFind -- Find listening entry by listen_handle (and get ref) */
static struct cm_listen *tsIbUCmListenFind(
                                           tTS_IB_LISTEN_HANDLE listen_handle
                                           )
{
  struct list_head *tmp;

  pthread_mutex_lock(&listen_lock);
  tmp = listen_list.next;
  while (tmp != &listen_list) {
    struct cm_listen *cm_listen;

    cm_listen = list_entry(tmp, struct cm_listen, list);
    if (cm_listen->listen_handle == listen_handle) {
      cm_listen->refcnt++;
      pthread_mutex_unlock(&listen_lock);
      return cm_listen;
    }

    tmp = tmp->next;
  }
  pthread_mutex_unlock(&listen_lock);

  return NULL;
}

/* ========================================================================= */
/*..tsIbUCmListenPut -- Put back reference to listening entry */
static void tsIbUCmListenPut(
                             struct cm_listen *cm_listen
                             )
{
  pthread_mutex_lock(&listen_lock);
  cm_listen->refcnt--;
  if (!cm_listen->refcnt) {
    free(cm_listen);
  }
  pthread_mutex_unlock(&listen_lock);
}

/* ========================================================================= */
/*..tsIbUCmConnectionFind -- Find connection by comm_id (and get reference) */
static struct cm_connection *tsIbUCmConnectionFind(
                                                   tTS_IB_CM_COMM_ID comm_id
                                                   )
{
  struct list_head *tmp;

  pthread_mutex_lock(&connection_lock);
  tmp = connection_list.next;
  while (tmp != &connection_list) {
    struct cm_connection *cm_conn;

    cm_conn = list_entry(tmp, struct cm_connection, list);
    if (cm_conn->comm_id == comm_id) {
      cm_conn->refcnt++;
      pthread_mutex_unlock(&connection_lock);
      return cm_conn;
    }

    tmp = tmp->next;
  }
  pthread_mutex_unlock(&connection_lock);

  return NULL;
}

/* ========================================================================= */
/*..tsIbUCmConnectionFindRelease -- Find and release conn by comm_id  from DB (unlocked, doesn't get reference) */
static struct cm_connection *_tsIbUCmConnectionFindRelease(
                                                   tTS_IB_CM_COMM_ID comm_id
                                                   )
{
  struct list_head *tmp;

  tmp = connection_list.next;
  while (tmp != &connection_list) {
    struct cm_connection *cm_conn;

    cm_conn = list_entry(tmp, struct cm_connection, list);
    if (cm_conn->comm_id == comm_id) {
      list_del(&cm_conn->list);
      return cm_conn;
    }

    tmp = tmp->next;
  }
  
  return NULL;
}

/* ========================================================================= */
/*..tsIbUCmConnectionFindRelease -- Find and release conn by comm_id  from DB */
static struct cm_connection *tsIbUCmConnectionFindRelease(
                                                   tTS_IB_CM_COMM_ID comm_id
                                                   )
{
  struct cm_connection *cm_conn;
  pthread_mutex_lock(&connection_lock);
  cm_conn = _tsIbUCmConnectionFindRelease(comm_id);
  pthread_mutex_unlock(&connection_lock);

  return cm_conn;
}

/* ========================================================================= */
/*.._tsIbUCmConnectionPut -- Put back reference to connection (unlocked) */
static void _tsIbUCmConnectionPut(
                                  struct cm_connection *cm_conn
                                  )
{
  cm_conn->refcnt--;
  if (!cm_conn->refcnt) {
    free(cm_conn);
  }
}

/* ========================================================================= */
/*..tsIbUCmConnectionPut -- Put back reference to connection */
static void tsIbUCmConnectionPut(
                                 struct cm_connection *cm_conn
                                 )
{
  pthread_mutex_lock(&connection_lock);
  _tsIbUCmConnectionPut(cm_conn);
  pthread_mutex_unlock(&connection_lock);
}

/* ========================================================================= */
/*..tsIbUCmProcessCompletions -- Loop processing completions from the kernel */
static void *tsIbUCmProcessCompletions(void *arg)
{
  struct ib_u_cm_get_completion_ioctl getcomp_ioctl;
  unsigned int params_size = 8192;	/* FIXME: Arbitrary */

  getcomp_ioctl.params = malloc(params_size);
  if (!getcomp_ioctl.params) {
    fprintf(stderr, "couldn't allocate %d bytes for params buffer\n",
            (int) getcomp_ioctl.params_size);
    return NULL;
  }

  while (1) {
    /* params_size can get trashed each time through, so reset it */
    getcomp_ioctl.params_size = params_size;
    if (ioctl(ua_fd, TS_IB_IOCCMGETCOMPLETION, &getcomp_ioctl)) {
      if (errno != EINTR) {
        fprintf(stderr, "TS_IB_IOCCMGETCOMPLETION failed: %m\n");
      }

      continue;
    }

    
    /* Setup pointers correctly */
    switch (getcomp_ioctl.event) {
    case TS_IB_CM_REQ_RECEIVED:
      {
        struct ib_u_cm_req_received_param *req_param =
         (struct ib_u_cm_req_received_param *)getcomp_ioctl.params;

        req_param->remote_private_data = getcomp_ioctl.params +
                                         sizeof(*req_param);
      }
      break;
    case TS_IB_CM_IDLE:
      {
        struct ib_u_cm_idle_param *idle_param =
         (struct ib_u_cm_idle_param *)getcomp_ioctl.params;

        idle_param->rej_info = getcomp_ioctl.params +
                               sizeof(*idle_param);
      }
      break;
    case TS_IB_CM_REP_RECEIVED:
      {
        struct ib_u_cm_rep_received_param *rep_param =
         (struct ib_u_cm_rep_received_param *)getcomp_ioctl.params;

        rep_param->remote_private_data = getcomp_ioctl.params +
                                         sizeof(*rep_param);
      }
      break;
    default:
      /* Nothing required for other states */
      break;
    }

    if (getcomp_ioctl.listen_handle != NULL) {
      struct cm_listen *cm_listen;

      cm_listen = (struct cm_listen *)getcomp_ioctl.cm_arg;
      if (cm_listen) {
        cm_listen->cm_func(VAPI_INVAL_HNDL,
                           getcomp_ioctl.comm_id,
                           getcomp_ioctl.event,
                           getcomp_ioctl.params,
                           cm_listen->cm_arg);
      }
    } else {
      struct cm_connection *cm_conn;


      // ETBD - We dont think if new QP state is reset
      // we need to find if this is the right thing to do
      cm_conn = (struct cm_connection *)getcomp_ioctl.cm_arg;
      if (cm_conn) {
        if (getcomp_ioctl.qp_handle
            && getcomp_ioctl.qp_state > 0) {
          /*
           * We need to lock here to make sure all QP state changes are
           * synchronized and we don't end up out of sync with the kernel
           */
          pthread_mutex_lock(&connection_lock);


          EVAPI_k_sync_qp_state(cm_conn->hca_handle,
                                cm_conn->v_qp_handle,
                                getcomp_ioctl.qp_state);

          pthread_mutex_unlock(&connection_lock);
        }

        cm_conn->cm_func(cm_conn->v_qp_handle,
                         getcomp_ioctl.comm_id,
                         getcomp_ioctl.event,
                         getcomp_ioctl.params,
                         cm_conn->cm_arg);

        if (getcomp_ioctl.event == TS_IB_CM_IDLE) {
          cm_conn = tsIbUCmConnectionFindRelease(getcomp_ioctl.comm_id);
          if (cm_conn) {
            tsIbUCmConnectionPut(cm_conn);	/* Once for the find release*/
          }
        }
      } else {
        fprintf(stderr, "unable to find completion for comm_id %u\n",
                getcomp_ioctl.comm_id);
      }
    }
  }

  free(getcomp_ioctl.params);

  return NULL;
}

/* ========================================================================= */
/*..tsIbUCmServiceAssign -- Allocate locally administered service ID */
tTS_IB_SERVICE_ID tsIbUCmServiceAssign(
                                       void
                                       )
{
  tTS_IB_SERVICE_ID service_id;
  int rc;

  rc = ioctl(ua_fd, TS_IB_IOCCMSERVIDASSIGN, &service_id);
  if (rc < 0) {
    return -errno;
  }

  return service_id;
}

/* ========================================================================= */
/*..tsIbUCmConnect -- Establish a connection */
int tsIbUCmConnect(
                   VAPI_hca_hndl_t                hca_handle,
                   tTS_IB_U_QP_HANDLE             v_qp_handle,
                   u_int32_t                      qpn,
                   struct ib_u_cm_qp_param       *qp_param,
                   struct ib_path_record         *primary_path,
                   struct ib_path_record         *alternate_path,
                   tTS_IB_SERVICE_ID              service_id,
                   void                          *req_private_data,
                   int                            req_private_data_len,
                   tTS_IB_U_CM_COMPLETION_FUNC    comp_func,
                   void                          *comp_arg,
                   tTS_IB_CM_COMM_ID             *comm_id
                   )
{
  struct ib_cm_connect_ioctl connect_ioctl;
  tTS_IB_U_K_QP_HANDLE vk_qp_handle;
  // tTS_IB_QP_HANDLE qp_handle; // EFIX
  void * qp_handle;
  
  VAPI_hca_hndl_t vhca_handle = hca_handle;
  struct ib_device *device = 0;
  struct cm_connection *cm_conn;
  int ret = 0, rc;

  /* First get the kernel QP handle from VAPI */
  rc = EVAPI_k_get_qp_hndl(vhca_handle, v_qp_handle, &vk_qp_handle);
  if (rc != VAPI_OK) {
    fprintf(stderr, "EVAPI_k_get_qp_hndl returns %d (%s)",
            rc, VAPI_strerror(rc));
    return -EIO;
  }

  /* Then register it with the kernel */
  rc = _tsIbUQpRegister(device, qpn, IB_TRANSPORT_RC, vk_qp_handle, &qp_handle);
  if (rc) {
    fprintf(stderr, "couldn't register QP: %d\n", rc);
    return -EIO;
  }

  cm_conn = malloc(sizeof(*cm_conn));
  if (!cm_conn) {
    return -ENOMEM;
  }

  cm_conn->refcnt = 1;		/* Start at 1 */
  cm_conn->cm_func = comp_func;
  cm_conn->cm_arg = comp_arg;
  cm_conn->hca_handle = hca_handle;
  cm_conn->t_qp_handle = qp_handle;
  cm_conn->v_qp_handle = v_qp_handle;
  cm_conn->vk_qp_handle = vk_qp_handle;

  connect_ioctl.cm_arg = (unsigned long)cm_conn;
  connect_ioctl.responder_resources = qp_param->responder_resources;
  connect_ioctl.initiator_depth = qp_param->initiator_depth;
  connect_ioctl.retry_count = qp_param->retry_count;
  connect_ioctl.rnr_retry_count = qp_param->rnr_retry_count;
  connect_ioctl.cm_response_timeout = qp_param->cm_response_timeout;
  connect_ioctl.max_cm_retries = qp_param->max_cm_retries;
  connect_ioctl.qp_handle = qp_handle;
  connect_ioctl.k_qp_handle = vk_qp_handle;
  connect_ioctl.primary_path = primary_path;
  connect_ioctl.alternate_path = alternate_path;
  connect_ioctl.service_id = service_id;
  connect_ioctl.req_private_data = req_private_data;
  connect_ioctl.req_private_data_len = req_private_data_len;
  connect_ioctl.qp_state = TS_IB_QP_STATE_INVALID;

  pthread_mutex_lock(&connection_lock);

  rc = ioctl(ua_fd, TS_IB_IOCCMCONNECT, &connect_ioctl);
  if (rc < 0) {
    _tsIbUCmConnectionPut(cm_conn);

    ret = -errno;
  } else {
    list_add_tail(&cm_conn->list, &connection_list);

    /* Store the return variables now */
    cm_conn->comm_id = *comm_id = connect_ioctl.comm_id;
  }

  if (connect_ioctl.qp_state > 0) {
    EVAPI_k_sync_qp_state(hca_handle, v_qp_handle, connect_ioctl.qp_state);
  }

  pthread_mutex_unlock(&connection_lock);

  return ret;
}

/* ========================================================================= */
/*..tsIbUCmListen -- Begin listening for connections to a set of services */
int tsIbUCmListen(
                  tTS_IB_SERVICE_ID           service_id,
                  tTS_IB_SERVICE_ID           service_mask,
                  tTS_IB_U_CM_COMPLETION_FUNC comp_func,
                  void                       *comp_arg,
                  tTS_IB_LISTEN_HANDLE       *listen_handle
                  )
{
  struct ib_cm_listen_ioctl listen_ioctl;
  struct cm_listen *cm_listen;
  int ret = 0, rc;

  cm_listen = malloc(sizeof(*cm_listen));
  if (!cm_listen) {
    return -ENOMEM;
  }

  cm_listen->refcnt = 1;		/* Start at 1 */
  cm_listen->cm_func = comp_func;
  cm_listen->cm_arg = comp_arg;

  listen_ioctl.cm_arg = (unsigned long)cm_listen;
  listen_ioctl.service_id = service_id;
  listen_ioctl.service_mask = service_mask;

  pthread_mutex_lock(&listen_lock);

  rc = ioctl(ua_fd, TS_IB_IOCCMLISTEN, &listen_ioctl);
  if (rc < 0) {

    free(cm_listen);

    ret = -errno;
  } else {
    list_add_tail(&cm_listen->list, &listen_list);

    /* Store the return variables now */
    cm_listen->listen_handle = *listen_handle = listen_ioctl.listen_handle;
  }

  pthread_mutex_unlock(&listen_lock);

  return ret;
}

/* ========================================================================= */
/*..tsIbUCmListenStop -- Cancel a listen request */
int tsIbUCmListenStop(
                      tTS_IB_LISTEN_HANDLE listen_handle
                      )
{
  struct cm_listen *cm_listen;
  int ret = 0, rc;

  cm_listen = tsIbUCmListenFind(listen_handle);
  if (!cm_listen) {
    return -ENOENT;
  }

  rc = ioctl(ua_fd, TS_IB_IOCCMLISTENSTOP, listen_handle);
  if (rc < 0) {
    ret = -errno;
  } else {
    pthread_mutex_lock(&listen_lock);
    list_del(&cm_listen->list);
    pthread_mutex_unlock(&listen_lock);

    tsIbUCmListenPut(cm_listen);	/* Once for the call */
   }

  tsIbUCmListenPut(cm_listen);		/* Once for the find */

  return ret;
}

/* ========================================================================= */
/*..tsIbUCmDisconnect -- Disconnect an established connection */
int tsIbUCmDisconnect(
                      tTS_IB_CM_COMM_ID comm_id
                      )
{
  int ret = 0, rc;

  rc = ioctl(ua_fd, TS_IB_IOCCMDISCONNECT, comm_id);
  if (rc < 0) {
    ret = -errno;
  }

  return ret;
}

/* ========================================================================= */
/*..tsIbUCmAlternatePathLoad -- Load a new althernate path for a connection */
int tsIbUCmAlternatePathLoad(
                             tTS_IB_CM_COMM_ID  comm_id,
                             struct ib_path_record *alternate_path
                             )
{
  struct ib_cm_alternate_path_load_ioctl altpathload_ioctl;
  int ret = 0, rc;

  altpathload_ioctl.comm_id = comm_id;
  altpathload_ioctl.alternate_path = alternate_path;

  rc = ioctl(ua_fd, TS_IB_IOCCMALTPATHLOAD, &altpathload_ioctl);
  if (rc < 0) {
    ret = -errno;
  }

  return ret;
}

/* ========================================================================= */
/*..tsIbUCmCallbackUpdate -- Update the callback context for this connection */
/*
int tsIbUCmCallbackUpdate(
                          tTS_IB_CM_COMM_ID           comm_id,
                          tTS_IB_U_CM_COMPLETION_FUNC comp_func,
                          void                       *comp_arg
                          )
{
  struct cm_connection *cm_conn;

  cm_conn = tsIbUCmConnectionFind(comm_id);
  if (!cm_conn) {
    return -EINVAL;
  }

  //pthread_mutex_lock(&connection_lock); should also have locks in tsIbUCmProcessCompletions
  cm_conn->cm_func = comp_func;
  cm_conn->cm_arg = comp_arg;
  //pthread_mutex_unlock(&connection_lock);

  tsIbUCmConnectionPut(cm_conn);

  return 0;
}
*/
/* ========================================================================= */
/*..tsIbUCmAccept -- Accept a connection request */
int tsIbUCmAccept(
                  tTS_IB_CM_COMM_ID              comm_id,
                  VAPI_hca_hndl_t                hca_handle,
                  tTS_IB_U_QP_HANDLE             v_qp_handle,
                  u_int32_t                     qpn,
                  tTS_IB_U_CM_COMPLETION_FUNC    comp_func,
                  void                          *comp_arg,
                  void                          *reply_data,
                  int                            reply_size
                  )
{
  struct ib_cm_accept_ioctl accept_ioctl;
  tTS_IB_U_K_QP_HANDLE vk_qp_handle;
  void * qp_handle;
  /* FIXME: Don't hardcode HCA handle for EVAPI_k_get_qp_hndl and _tsIbUQpRegister */
  VAPI_hca_hndl_t vhca_handle = 0;
  struct ib_device *device = 0;
  struct cm_connection *cm_conn;
  int ret = 0, rc;

  /* First get the kernel QP handle from VAPI */
  rc = EVAPI_k_get_qp_hndl(vhca_handle, v_qp_handle, &vk_qp_handle);
  if (rc != VAPI_OK) {
    fprintf(stderr, "EVAPI_k_get_qp_hndl returns %d (%s)",
            rc, VAPI_strerror(rc));
    return -EIO;
  }

  /* Then register it with the kernel */
  rc = _tsIbUQpRegister(device, qpn, IB_TRANSPORT_RC, vk_qp_handle, &qp_handle);
  if (rc) {
    fprintf(stderr, "couldn't register QP: %d\n", rc);
    return -EIO;
  }

  cm_conn = malloc(sizeof(*cm_conn));
  if (!cm_conn) {
    return -ENOMEM;
  }

  cm_conn->refcnt = 1;		/* Start at 1 */
  cm_conn->cm_func = comp_func;
  cm_conn->cm_arg = comp_arg;
  cm_conn->comm_id = comm_id;
  cm_conn->hca_handle = hca_handle;
  cm_conn->t_qp_handle = qp_handle;
  cm_conn->v_qp_handle = v_qp_handle;
  cm_conn->vk_qp_handle = vk_qp_handle;

  accept_ioctl.cm_arg = (unsigned long)cm_conn;
  accept_ioctl.comm_id = comm_id;
  accept_ioctl.qp_handle = qp_handle;
  accept_ioctl.k_qp_handle = vk_qp_handle;
  accept_ioctl.reply_data = reply_data;
  accept_ioctl.reply_size = reply_size;
  accept_ioctl.qp_state = TS_IB_QP_STATE_INVALID;



  pthread_mutex_lock(&connection_lock);

  rc = ioctl(ua_fd, TS_IB_IOCCMACCEPT, &accept_ioctl);
  if (rc < 0) {
    _tsIbUCmConnectionPut(cm_conn);

    ret = -errno;
  } else {
    list_add_tail(&cm_conn->list, &connection_list);
  }


  // ETBD - We don't sync if we are in reset.
  // need to check if this is the right thing to do
  //
  if (accept_ioctl.qp_state > 0) {
      EVAPI_k_sync_qp_state(hca_handle, v_qp_handle, accept_ioctl.qp_state);
  }

  pthread_mutex_unlock(&connection_lock);

  return ret;
}

/* ========================================================================= */
/*..tsIbUCmReject -- Reject a connection request */
int tsIbUCmReject(
                  tTS_IB_CM_COMM_ID    comm_id,
                  void                *reply_data,
                  int                  reply_size
                  )
{
  struct cm_connection *cm_conn;
  struct ib_cm_reject_ioctl reject_ioctl;
  int ret = 0, rc;

  cm_conn = tsIbUCmConnectionFindRelease(comm_id);

  reject_ioctl.comm_id = comm_id;
  reject_ioctl.reply_data = reply_data;
  reject_ioctl.reply_size = reply_size;

  rc = ioctl(ua_fd, TS_IB_IOCCMREJECT, &reject_ioctl);
  if (rc < 0) {
    ret = -errno;
  }

  if (cm_conn) {
    tsIbUCmConnectionPut(cm_conn);	/* Once for the find release*/
  }

  return ret;
}

/* ========================================================================= */
/*..tsIbUCmConfirm -- Confirm a connection request */
int tsIbUCmConfirm(
                   tTS_IB_CM_COMM_ID    comm_id,
                   void                *reply_data,
                   int                  reply_size
                   )
{
  struct cm_connection *cm_conn;
  struct ib_cm_confirm_ioctl confirm_ioctl;
  int ret = 0, rc;

  cm_conn = tsIbUCmConnectionFind(comm_id);
  if (!cm_conn) {
    return -ENOENT;
  }

  confirm_ioctl.comm_id = comm_id;
  confirm_ioctl.qp_handle = cm_conn->t_qp_handle;
  confirm_ioctl.reply_data = reply_data;
  confirm_ioctl.reply_size = reply_size;
  confirm_ioctl.qp_state = TS_IB_QP_STATE_INVALID;

  pthread_mutex_lock(&connection_lock);

  rc = ioctl(ua_fd, TS_IB_IOCCMCONFIRM, &confirm_ioctl);
  if (rc < 0) {
    if(_tsIbUCmConnectionFindRelease(comm_id))
        _tsIbUCmConnectionPut(cm_conn);

    ret = -errno;
  }

  // ETBD - We dont think if new QP state is reset
  // we need to find if this is the right thing to do
  if (confirm_ioctl.qp_state > 0) {
      EVAPI_k_sync_qp_state(cm_conn->hca_handle,
                          cm_conn->v_qp_handle,
                          confirm_ioctl.qp_state);
  }

  pthread_mutex_unlock(&connection_lock);

  tsIbUCmConnectionPut(cm_conn);

  return ret;
}

/* ========================================================================= */
/*..tsIbUCmDelayResponse -- Delay response to REQ */
int tsIbUCmDelayResponse(
                         tTS_IB_CM_COMM_ID    comm_id,
                         int                  service_timeout,
                         void                *mra_data,
                         int                  mra_data_size
                         )
{
  struct ib_cm_delayresponse_ioctl delay_ioctl;
  int ret = 0, rc;

  delay_ioctl.comm_id = comm_id;
  delay_ioctl.service_timeout = service_timeout;
  delay_ioctl.mra_data = mra_data;
  delay_ioctl.mra_data_size = mra_data_size;

  rc = ioctl(ua_fd, TS_IB_IOCCMDELAYRESPONSE, &delay_ioctl);
  if (rc < 0) {
      ret = -errno;
  }

  return ret;
}

/* ========================================================================= */
/*..tsIbUCmEstablish -- Inform kernel CM connection has been established */
int tsIbUCmEstablish(
                     tTS_IB_CM_COMM_ID comm_id
                     )
{
  struct cm_connection *cm_conn;
  struct ib_cm_establist_ioctl establish_ioctl;
  int ret = 0, rc;

  cm_conn = tsIbUCmConnectionFind(comm_id);
  if (!cm_conn) {
    return -ENOENT;
  }

  establish_ioctl.comm_id = comm_id;
  establish_ioctl.qp_state = TS_IB_QP_STATE_INVALID;

  pthread_mutex_lock(&connection_lock);

  rc = ioctl(ua_fd, TS_IB_IOCCMESTABLISH, &establish_ioctl);
  if (rc < 0) {
    ret = -errno;
  }

  // ETBD - We dont think if new QP state is reset
  // we need to find if this is the right thing to do
  if (establish_ioctl.qp_state > 0) {
      EVAPI_k_sync_qp_state(cm_conn->hca_handle,
                          cm_conn->v_qp_handle,
                          establish_ioctl.qp_state);
  }

  pthread_mutex_unlock(&connection_lock);

  tsIbUCmConnectionPut(cm_conn);

  return ret;
}

/* ========================================================================= */
/*..tsIbUCmDropConsumer -- Drop all library references to a connection */
int tsIbUCmDropConsumer(
                        tTS_IB_CM_COMM_ID comm_id
                        )
{
  struct cm_connection *cm_conn;
  int ret = 0, rc;

  cm_conn = tsIbUCmConnectionFindRelease(comm_id);
  if (!cm_conn) {
    return -ENOENT;
  }

  rc = ioctl(ua_fd, TS_IB_IOCCMDROPCONSUMER, comm_id);
  if (rc < 0) {
    ret = -errno;
  }

  tsIbUCmConnectionPut(cm_conn);	/* Once for the find release*/

  return ret;
}

/* ========================================================================= */
/*..tsIbUCmPathRecord -- Obtain the path record for a remote destination */
int tsIbUCmPathRecord(
                      in_addr_t dst_addr,
                      struct ib_path_record *path_record
                      )
{
  struct ib_path_record_ioctl pathrecord_ioctl;
  int ret = 0, rc;

  pathrecord_ioctl.dst_addr = dst_addr;
  pathrecord_ioctl.path_record = (struct ib_path_record*)path_record;

  rc = ioctl(ua_fd, TS_IB_IOCGPATHRECORD, &pathrecord_ioctl);
  if (rc < 0) {
    ret = -errno;
  }

  return ret;
}

/* ========================================================================= */
/*.._tsIbUQpRegister -- Register a userlevel QP with the kernel */
static int _tsIbUQpRegister(
                            struct ib_device *device,
                            u_int32_t            qpn,
                            enum ib_transport    transport,
                            tTS_IB_U_K_QP_HANDLE vk_qp_handle,
                            void    **qp_handle
                            )
{
  struct ib_qp_register_ioctl qpregister_ioctl;
  int ret = 0, rc;

  qpregister_ioctl.qpn = qpn;
  qpregister_ioctl.transport = transport;
  qpregister_ioctl.vk_qp_handle = vk_qp_handle;

  rc = ioctl(ua_fd, TS_IB_IOCQPREGISTER, &qpregister_ioctl);
  if (rc < 0) {
    ret = -errno;
  } else {
    *qp_handle = qpregister_ioctl.qp_handle;
  }

  return ret;
}


/* ========================================================================= */
/*..ib_cm_get_remote_gid */
int ib_cm_get_remote_gid(tTS_IB_CM_COMM_ID comm_id, tTS_IB_GID *gid)
{
    int ret = 0, rc;

    struct ib_cm_get_remote_gid_ioctl rgid_ioctl;

    rgid_ioctl.comm_id = comm_id;
    rgid_ioctl.gid = gid;


    rc = ioctl(ua_fd, TS_IB_IOCMGETREMOTEGID, &rgid_ioctl);
    if (rc < 0) {
        ret = -errno;
    }

    
    return ret;
}







/* ========================================================================= */
/*.._init -- intialize the library */
void libts_ib_cm_user_init(
           void
           )
{

  if (ua_fd != -1) {
    return;
  }

  init_list_head(&connection_list);
  init_list_head(&listen_list);

  /* FIXME: This shouldn't be hardcoded */
  ua_fd = open("/dev/ts_ua_cm0", O_RDWR);
  if (ua_fd < 0) {
    fprintf(stderr, "couldn't open /dev/ts_ua_cm0: %m\n");
    return;
  }

  if (pthread_create(&completion_thread, NULL, tsIbUCmProcessCompletions, NULL)) {
    fprintf(stderr, "couldn't create completion thread: %m\n");
    return;
  }
}



