/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 *
 * MODULE: dat_osd.c
 *
 * PURPOSE: Operating System Dependent layer
 * Description:
 *	Provide OS dependent functions with a canonical DAT
 *	interface. Designed to be portable and hide OS specific quirks
 *	of common functions.
 *
 *
 * $Id: dat_osd.c,v 1.8 2005/01/12 16:39:52 jlentini Exp $
 **********************************************************************/

#include <linux/module.h>
#include <linux/sched.h>

#include "dat_osd.h"
#include <dat/dat.h>


/*********************************************************************
 *                                                                   *
 * Constants                                                         *
 *                                                                   *
 *********************************************************************/

#define DAT_DBG_TYPE_ENV 	"DAT_DBG_TYPE"
#define DAT_DBG_DEST_ENV 	"DAT_DBG_DEST"


/*********************************************************************
 *                                                                   *
 * Enumerations                                                      *
 *                                                                   *
 *********************************************************************/

typedef int 			DAT_OS_DBG_DEST;

typedef enum
{
    DAT_OS_DBG_DEST_STDOUT  		= 0x1,
    DAT_OS_DBG_DEST_SYSLOG  		= 0x2,
    DAT_OS_DBG_DEST_ALL  		= 0x3
} DAT_OS_DBG_DEST_TYPE;


/*********************************************************************
 *                                                                   *
 * Global Variables                                                  *
 *                                                                   *
 *********************************************************************/

static DAT_OS_DBG_TYPE_VAL 	g_dbg_type = 0;
static DAT_OS_DBG_DEST 		g_dbg_dest = DAT_OS_DBG_DEST_STDOUT;


/***********************************************************************
 * Function: dat_os_dbg_init
 ***********************************************************************/

void
dat_os_dbg_init ( void )
{
    g_dbg_type = dat_os_getenv (DAT_DBG_TYPE_ENV);
    g_dbg_dest = dat_os_getenv (DAT_DBG_DEST_ENV);
}


/***********************************************************************
 * Function: dat_os_dbg_print
 ***********************************************************************/

void
dat_os_dbg_print (
	DAT_OS_DBG_TYPE_VAL		type,
	const char *			fmt,
    ...)
{
    static char buf[1024];

    if ( (DAT_OS_DBG_TYPE_ERROR == type) || (type & g_dbg_type) )
    {
	va_list args;

	va_start (args, fmt);

	if ( DAT_OS_DBG_DEST_STDOUT & g_dbg_dest )
	{
	    vsnprintf (buf, sizeof (buf), fmt, args);
	    printk ("kDAPL: %s", buf);
	}

	va_end (args);
    }
}



/***********************************************************************
 * Function: dapl_os_get_env_val
 ***********************************************************************/
/*
 * dapl_os_get_env_val
 *
 * Update val to  value of passed in environment variable if present
 *
 * Input:
 *      env_str
 *
 * Returns:
 *	value of environment variable
 */
int
dat_os_getenv (
	char		*env_str )
{
    extern int DbgLvl;			/* defined in dapl_module.c */
    int def_val;

    def_val = DAT_OS_DBG_TYPE_ERROR;

    if ( strcmp (env_str, DAT_DBG_TYPE_ENV) == 0 )
    {
	if (DbgLvl != 0)
	{
	    def_val = DbgLvl;
	}
    }
    else if ( strcmp (env_str, DAT_DBG_DEST_ENV) == 0 )
    {
	;			       /* do nothing */
    }

    return  def_val;
}


/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */
