/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 *
 * HEADER: dat_osd.h
 *
 * PURPOSE: Operating System Dependent layer
 * Description:
 *	Provide OS dependent data structures & functions with
 *	a canonical DAT interface. Designed to be portable
 *	and hide OS specific quirks of common functions.
 *
 * $Id: dat_osd.h,v 1.7 2004/07/14 14:06:56 sjs2 Exp $
 **********************************************************************/

#ifndef _DAT_OSD_H_
#define _DAT_OSD_H_

/*
 * This file is defined for Linux systems only, including it on any
 * other build will cause an error
 */
#ifndef __linux__
#error UNDEFINED OS TYPE
#endif /* __linux__ */

#ifndef  __KERNEL__
#error "Must compile for the kernel"
#endif /* __KERNEL__ */

#ifndef  __KDAPL__
#error "Must compile for KDAPL"
#endif /* __KDAPL__ */

/* Useful debug definitions */
#ifndef STATIC
#define STATIC static
#endif /* STATIC */
#ifndef INLINE
#define INLINE __inline__
#endif /* INLINE */

#include <dat/kdat.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <asm/types.h>



/*********************************************************************
 *                                                                   *
 * Debuging                                                          *
 *                                                                   *
 *********************************************************************/

typedef int 			DAT_OS_DBG_TYPE_VAL;

typedef enum
{
    DAT_OS_DBG_TYPE_ERROR 		= 0x1,
    DAT_OS_DBG_TYPE_GENERIC 		= 0x2,
    DAT_OS_DBG_TYPE_SR  		= 0x4,
    DAT_OS_DBG_TYPE_DR  		= 0x8,
    DAT_OS_DBG_TYPE_PROVIDER_API 	= 0x10,
    DAT_OS_DBG_TYPE_CONSUMER_API 	= 0x20,
    DAT_OS_DBG_TYPE_ALL 		= 0xff
} DAT_OS_DBG_TYPE;

extern void dat_os_dbg_init ( void );

extern void dat_os_dbg_print (
	DAT_OS_DBG_TYPE_VAL	type,
	const char *		fmt,
	...);

extern void dapl_os_vprintf (
	const char *		fmt,
	va_list			args);

/*
 *  Debug helper routines in the Linux kernel
 */
#define ASSERT(expr) \
if(!(expr)) { \
        printk( "Assertion failed! %s:%s:%d %s\n", \
        __FILE__, __FUNCTION__, __LINE__, (#expr));  \
    }

#define dat_os_printf printk
#define dat_os_assert ASSERT
#define dat_os_breakpoint breakpoint


/*********************************************************************
 *                                                                   *
 * Utility Functions                                                 *
 *                                                                   *
 *********************************************************************/

typedef size_t 			DAT_OS_SIZE;

#define DAT_ERROR(Type, SubType) ((DAT_RETURN)(DAT_CLASS_ERROR | Type | SubType))

int dat_os_getenv (char		*env_str);


/*********************************************************************
 *                                                                   *
 * Lock Functions                                                    *
 *                                                                   *
 * Simply use spinlocks at this point.                               *
 *                                                                   *
 * N.B. No lock can be taken in both common code and an upcall,      *
 * 	which has the possibility of being called in interrupt       *
 *	context. If this is a requirement we need to change          *
 *      from buzz locks (spinlock) to semaphores.                    *
 *********************************************************************/

typedef spinlock_t      DAT_OS_LOCK;


STATIC INLINE DAT_RETURN
dat_os_lock_init (
    IN	DAT_OS_LOCK *m)
{
    spin_lock_init ( m );

    return DAT_SUCCESS;
}

STATIC INLINE DAT_RETURN
dat_os_lock_destroy (
    IN	DAT_OS_LOCK *m)
{
    /* Nothing */
    return DAT_SUCCESS;
}

STATIC INLINE DAT_RETURN
dat_os_lock (
    IN	DAT_OS_LOCK *m)
{
    spin_lock (m); /* down(mutex); */
    return DAT_SUCCESS;
}

STATIC INLINE DAT_RETURN
dat_os_unlock (
    IN	DAT_OS_LOCK *m)
{
    spin_unlock (m); /* up(mutex); */
    return DAT_SUCCESS;
}


/*********************************************************************
 *                                                                   *
 * Memory Functions                                                  *
 *                                                                   *
 * Some memory functions are already inlined by the compiler, so we  *
 * use simple #defines to avoid confusing double inlines             *
 *********************************************************************/
#define dat_os_alloc(size)		kmalloc(size, GFP_ATOMIC)
#define dat_os_free(ptr, size)		kfree(ptr)
#define dat_os_memset(loc, c, size) 	memset(loc, c, size)
#define dat_os_strlen(s)		strlen(s)
#define dat_os_strncmp(s1, s2, n)	strncmp(s1, s2, n)
#define dat_os_strncpy(dest, src, len)	strncpy(dest, src, len)


#endif /*  _DAT_OSD_H_ */
