/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 * 
 * MODULE: dapl_lmr_free.c
 *
 * PURPOSE: Memory management
 * Description: Interfaces in this file are completely described in
 *		the DAPL 1.1 API, Chapter 6, section 6
 *
 * $Id$
 **********************************************************************/

#include "dapl_lmr_util.h"
#include "dapl_adapter_util.h"
#include "dapl_ia_util.h"
#include "dapl_openib.h"

/*
 * dapl_lmr_free
 *
 * DAPL Requirements Version xxx, 6.6.3.2
 *
 * Destroy an instance of the Local Memory Region
 *
 * Input:
 * 	lmr_handle
 *
 * Output:
 *
 * Returns:
 * 	DAT_SUCCESS
 *      DAT_INVALID_HANDLE
 * 	DAT_INVALID_PARAMETER
 * 	DAT_INVALID_STATE 
 */


DAT_RETURN
dapl_lmr_free (
	IN	DAT_LMR_HANDLE	   lmr_handle )
{
    DAPL_LMR 		*lmr;
    DAT_RETURN		dat_status;
#ifndef __KDAPL__
    int                 ts_status;
#endif

    dapl_dbg_log (DAPL_DBG_TYPE_API, "dapl_lmr_free (%p)\n", lmr_handle);

    if ( DAPL_BAD_HANDLE (lmr_handle, DAPL_MAGIC_LMR) )
    {
	dat_status = DAT_ERROR (DAT_INVALID_HANDLE,DAT_INVALID_HANDLE_LMR);
	goto bail;
    }

    lmr = (DAPL_LMR *) lmr_handle;

    switch (lmr->param.mem_type)
    {
    case DAT_MEM_TYPE_VIRTUAL:
	/* fall through */
#if defined(__KDAPL__)
    case DAT_MEM_TYPE_PHYSICAL:
    case DAT_MEM_TYPE_PLATFORM:
    case DAT_MEM_TYPE_IA:
#else
    case DAT_MEM_TYPE_SHARED_VIRTUAL:
#endif
	/* fall through */
	case DAT_MEM_TYPE_LMR:
	{
	    DAPL_PZ 	*pz;

	    if ( 0 != dapl_os_atomic_read (&lmr->lmr_ref_count) )
	    {
		return DAT_INVALID_STATE;
	    }

	    dat_status = dapls_hash_remove (lmr->header.owner_ia->hca_ptr->lmr_hash_table, 
					    lmr->param.lmr_context, NULL);
	    if (dat_status != DAT_SUCCESS)
	    {
		goto bail;
	    }
#ifndef __KDAPL__
            if (lmr->param.mem_type == DAT_MEM_TYPE_SHARED_VIRTUAL) {

                DAT_LMR_COOKIE lmr_cookie;
                int            ts_status;

                lmr_cookie =
                    lmr->param.region_desc.for_shared_memory.shared_memory_id;
                ts_status = ts_smr_del_mrh(lmr_cookie, lmr->mr_handle);

                if (ts_status) {
                    /*                                                                                                           
                    syslog(LOG_NOTICE,                                                                                           
                           "ts_smr_del_mrh failed in dapl_lmr_free cookie = %llx mr_handle = %x\n",                              
                           lmr_cookie.as_64, lmr->mr_handle);                                                                    
                    */
                    return DAT_INSUFFICIENT_RESOURCES;
                }

                ts_status = ts_smr_dec(lmr_cookie);

                if (ts_status) {
                    /*                                                                                                           
                    syslog(LOG_NOTICE,                                                                                           
                           "ts_smr_dec failed in dapl_lmr_free cookie = %llx\n", lmr_cookie.as_64);                              
                    */
                    return DAT_INSUFFICIENT_RESOURCES;
                }   

		ts_status = ts_smr_mutex_lock();
		if (ts_status) {
		    dat_status = DAT_ERROR (DAT_INTERNAL_ERROR, 0);
		    goto bail;
		}
            
	    }
#endif

	    dat_status = dapls_ib_mr_deregister (lmr);
#ifndef __KDAPL__
            if (lmr->param.mem_type == DAT_MEM_TYPE_SHARED_VIRTUAL) 
	    {
		ts_status = ts_smr_mutex_unlock();
		if (ts_status) {
		    dat_status = DAT_ERROR (DAT_INTERNAL_ERROR, 0);
		    goto bail;
		}
	    }
#endif
		
	    if (dat_status == DAT_SUCCESS)
	    {

		pz = (DAPL_PZ *) lmr->param.pz_handle;
		dapl_os_atomic_dec (&pz->pz_ref_count);

		dapl_lmr_dealloc (lmr);
	    }
	    else
	    {
		/*
		 * Deregister failed; put it back in the
		 * hash table.
		 */
		dapls_hash_insert (lmr->header.owner_ia->hca_ptr->lmr_hash_table, 
				   lmr->param.lmr_context, lmr);
	    }

	    break;
	}
#if defined(__KDAPL__)
    case DAT_MEM_TYPE_BYPASS:
	{
	    return DAT_ERROR (DAT_NOT_IMPLEMENTED,0);
	}
#endif	/* defined(__KDAPL__) */
	default:
	{
	    dat_status = DAT_ERROR (DAT_INVALID_PARAMETER,DAT_INVALID_ARG1);
	    break;
	}
    }
 bail:
    return dat_status;
}
