/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 * 
 * MODULE: dapl_ring_buffer_util.c
 *
 * PURPOSE: Ring buffer management
 * Description: Support and management functions for ring buffers
 *
 * $Id$
 **********************************************************************/

#include "dapl_ring_buffer_util.h"


/*
 * dapls_rbuf_alloc
 *
 * Given a DAPL_RING_BUFFER, initialize it and provide memory for
 * the ringbuf itself. A passed in size will be adjusted to the next
 * largest power of two number to simplify management.
 *
 * Input:
 *	rbuf		pointer to DAPL_RING_BUFFER
 *	size		number of elements to allocate & manage
 *
 * Output:
 *	none
 *
 * Returns:
 *	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN
dapls_rbuf_alloc (
        INOUT  DAPL_RING_BUFFER		*rbuf,
	IN     DAT_COUNT		 size )
{
    unsigned int			rsize;	/* real size */

    /* The circular buffer must be allocated one too large.
     * This eliminates any need for a distinct counter, as that
     * having the two pointers equal always means "empty" -- never "full"
     */
    size++;

    /* Put size on a power of 2 boundary */
    rsize = 1;
    while ( (DAT_COUNT)rsize < size)
    {
	rsize <<= 1;
    }

    rbuf->base = (void *) dapl_os_alloc (rsize * sizeof (void *));
    if ( rbuf->base != NULL )
    {
	rbuf->lim = rsize - 1;
	dapl_os_atomic_set (&rbuf->head, 0);
	dapl_os_atomic_set (&rbuf->tail, 0);
    }
    else
    {
	return DAT_INSUFFICIENT_RESOURCES | DAT_RESOURCE_MEMORY;
    }

    return DAT_SUCCESS;
}


/*
 * dapls_rbuf_realloc
 *
 * Resizes an empty DAPL_RING_BUFFER. This function is not thread safe;
 * adding or removing elements from a ring buffer while resizing 
 * will have indeterminate results.
 *
 * Input:
 *	rbuf		pointer to DAPL_RING_BUFFER
 *	size		number of elements to allocate & manage
 *
 * Output:
 *	none
 *
 * Returns:
 *	DAT_SUCCESS
 *	DAT_INVALID_STATE
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN
dapls_rbuf_realloc (
        INOUT  DAPL_RING_BUFFER		*rbuf,
	IN     DAT_COUNT		 size )
{
    int			rsize;		/* real size */
    int         tail;
    int         tail_mod;
    int         head;
    int         i;
    int         current_tail;
//    int         current_head;
    DAT_RETURN		dat_status;

    dat_status = DAT_SUCCESS;

    /* rbuf sizes are power of 2, so a user may ask for an increase
     * that is still less than what has been allocated
     */
    if (size <= rbuf->lim + 1)
    {
    /* This fix is from dapl plugfest founf by crload 1.2 (vu)*/
	/*dat_status = DAT_ERROR (DAT_INVALID_PARAMETER, DAT_INVALID_ARG2); */
	goto bail;
    }

    /* Put size on a power of 2 boundary */
    rsize = 1;
    while (rsize < size)
    {
	rsize <<= 1;
    }

    rbuf->base = (void *) dapl_os_realloc (rbuf->base, rsize * sizeof (void *));
    if ( NULL == rbuf->base )
    {
	dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);
	goto bail;
    }
    head = (dapl_os_atomic_read (&rbuf->head) & rbuf->lim);
    tail = dapl_os_atomic_read (&rbuf->tail);
    tail_mod = tail & rbuf->lim;
    if (tail_mod  > head) {
        for (i=0;i<(rbuf->lim + 1 - tail_mod);i++) {
            rbuf->base[rsize-1-i] = rbuf->base[rbuf->lim-i];
            rbuf->base[rbuf->lim-i] = NULL;
        }

    
    current_tail = dapl_os_atomic_assign (&rbuf->tail, tail, ((tail & rbuf->lim)+ rsize - 1- rbuf->lim));
    if ((current_tail != tail)) {
      dat_status = DAT_ERROR (DAT_INVALID_PARAMETER, DAT_INVALID_ARG2);
    }
    }
    rbuf->lim = rsize - 1;

    


 bail:
    return dat_status;
}


/*
 * dapls_rbuf_destroy
 *
 * Release the buffer and reset pointers to a DAPL_RING_BUFFER
 *
 * Input:
 *	rbuf		pointer to DAPL_RING_BUFFER
 *
 * Output:
 *	none
 *
 * Returns:
 *	none
 *
 */
void
dapls_rbuf_destroy (
	IN  DAPL_RING_BUFFER		*rbuf)
{
    if ( (NULL == rbuf) || 
         (NULL == rbuf->base) )
    {
        return;
    }

    dapl_os_free (rbuf->base, (rbuf->lim + 1) * sizeof (void *));
    rbuf->base = NULL;
    rbuf->lim = 0;

    return;
}

/*
 * dapls_rbuf_add
 *
 * Add an entry to the ring buffer
 *
 * Input:
 *	rbuf		pointer to DAPL_RING_BUFFER
 *	entry		entry to add
 *
 * Output:
 *	none
 *
 * Returns:
 *	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES         (queue full)
 *
 */
DAT_RETURN
dapls_rbuf_add (
	IN  DAPL_RING_BUFFER		*rbuf,
	IN  void			*entry)
{
    int			pos;
    int			val;   
    int         head;
    int         tail;

    while ( ((dapl_os_atomic_read (&rbuf->head) + 1) & rbuf->lim) != (dapl_os_atomic_read (&rbuf->tail) & rbuf->lim) )
    {
	pos = dapl_os_atomic_read (&rbuf->head);
	val = dapl_os_atomic_assign (&rbuf->head, pos, (pos + 1 ) & rbuf->lim);
	if ( val == pos )
	{
	    pos =  (pos + 1) & rbuf->lim; /* verify in range */
	    rbuf->base[pos] = entry;
	    return DAT_SUCCESS;
	}
    }
    head =   dapl_os_atomic_read (&rbuf->head);
    tail =      dapl_os_atomic_read (&rbuf->tail);
    if (((dapl_os_atomic_read (&rbuf->head) + 1) & rbuf->lim) == (dapl_os_atomic_read (&rbuf->tail) & rbuf->lim)) {
    }
    
    return DAT_ERROR (DAT_INSUFFICIENT_RESOURCES,DAT_RESOURCE_MEMORY);

}


/*
 * dapls_rbuf_remove
 *
 * Remove an entry from the ring buffer
 *
 * Input:
 *	rbuf		pointer to DAPL_RING_BUFFER
 *
 * Output:
 *	entry		entry removed from the ring buffer
 *
 * Returns:
 *	DAT_SUCCESS
 *	DAT_QUEUE_EMPTY
   
 */
void *
dapls_rbuf_remove (
	IN  DAPL_RING_BUFFER	*rbuf )
{
    int			pos;
    int			val;

    while ( dapl_os_atomic_read (&rbuf->head) != dapl_os_atomic_read (&rbuf->tail) )
    {
	pos = dapl_os_atomic_read (&rbuf->tail);
	val = dapl_os_atomic_assign (&rbuf->tail, pos, (pos + 1) & rbuf->lim);
	if ( val == pos )
	{
	    pos =  (pos + 1) & rbuf->lim; /* verify in range */

        rbuf->base[((pos-1)&rbuf->lim)] = NULL;
	    return (rbuf->base[pos]);
	}
    }
    return NULL;

}


/*
 * dapli_rbuf_count
 *
 * Return the number of entries in use in the ring buffer
 *
 * Input:
 *      rbuf            pointer to DAPL_RING_BUFFER
 *
 *
 * Output:
 *	none
 *
 * Returns:
 *	count of entries
 *
 */
DAT_COUNT
dapls_rbuf_count (
	IN DAPL_RING_BUFFER		*rbuf )
{
    DAT_COUNT count;
    int head;
    int tail;

    head = dapl_os_atomic_read (&rbuf->head) & rbuf->lim;
    tail = dapl_os_atomic_read (&rbuf->tail) & rbuf->lim;
    if ( head > tail )
    {
	count = head - tail;
    }
    else
    {
	/* add 1 to lim as it is a mask, number of entries - 1 */
	count = (rbuf->lim + 1 - tail + head) & rbuf->lim;
    }

    return count;
}


/*
 * dapls_rbuf_contains
 *
 * Return TRUE or FALSE if an element exists in a ring buffer
 *
 * Input:
 *	rbuf		pointer to DAPL_RING_BUFFER
 *	entry		entry to match
 *
 * Output:
 *	none
 *
 * Returns:
 *	DAT_TRUE
 *	DAT_FALSE
   
 */
//DAT_BOOLEAN
//dapls_rbuf_contains (
//	IN  DAPL_RING_BUFFER	*rbuf,
//	IN  void		*entry)
//{
//    int			pos;
//
//    pos = dapl_os_atomic_read (&rbuf->head);
//    while ( pos != dapl_os_atomic_read (&rbuf->tail) )
//    {
//	if (rbuf->base[pos] == entry)
//    {
//           return DAT_TRUE;
//    }
//pos =  (pos + 1) & rbuf->lim; /* verify in range */
//    }
////    if ( pos == dapl_os_atomic_read (&rbuf->tail) ) {
////        return DAT_TRUE;
//  //  }
//    return DAT_FALSE;
//
//}

DAT_BOOLEAN
dapls_rbuf_contains (
	IN  DAPL_RING_BUFFER	*rbuf,
	IN  void		*entry,
    IN  void		*old_entry)
{
    int			pos;

    pos = dapl_os_atomic_read (&rbuf->head);
    while ( pos != dapl_os_atomic_read (&rbuf->tail) )
    {
	if (rbuf->base[pos] == old_entry)
    {
          rbuf->base[pos] = entry;
           return DAT_TRUE;
    }
pos =  (pos - 1) & rbuf->lim; /* verify in range */
    }
    return DAT_FALSE;

}

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */
