/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 *
 * MODULE: dapl_evd_kcreate.c
 *
 * PURPOSE: EVENT management
 *
 * Description: Interfaces in this file are completely defined in
 * 		the uDAPL 1.1 API, Chapter 6, section 3
 *
 * $Id: dapl_evd_kcreate.c,v 1.2 2004/08/02 15:14:28 rabin Exp $
 **********************************************************************/

#include "dapl.h"
#include "dapl_evd_util.h"
#include "dapl_ia_util.h"
#include "dapl_cno_util.h"
#include "dapl_adapter_util.h"

/*
 * dapl_evd_kcreate
 *
 * DAPL Requirements Version xxx, 6.3.2.1
 *
 * Create and instance of Event Dispatcher.
 *
 * Input:
 *    ia_handle
 *    upcall_policy
 *    upcall
 *    evd_min_qlen
 *    evd_flags
 *
 * Output:
 *    evd_handle
 *
 * Returns:
 *     DAT_SUCCESS
 *     DAT_INSUFFICIENT_RESOURCES
 *     DAT_INVALID_HANDLE
 *     DAT_INVALID_PARAMETER
 */

/* ** REVISIT **
 *
 * Selecting the cqe handing domain must still be done.
 * We *probably* want one per hca, but we could have one
 * per provider or one per consumer.
 */
/* Note that if there already is a cq, it is not deleted
 * even if it is not required. However, it will not be armed.
 */

DAT_RETURN dapl_evd_kcreate (
    IN    DAT_IA_HANDLE		ia_handle,
    IN    DAT_COUNT		evd_min_qlen,
    IN    DAT_UPCALL_POLICY	upcall_policy,	/* ignore */
    IN    const DAT_UPCALL_OBJECT  *upcall,
    IN    DAT_EVD_FLAGS		evd_flags,
    OUT   DAT_EVD_HANDLE	*evd_handle )
{
    DAPL_IA	*ia_ptr;
    DAPL_EVD	*evd_ptr;
    DAPL_CNO	*cno_ptr;
    DAT_RETURN	dat_status;
    DAT_EVD_HANDLE  evdhandle = (DAT_EVD_HANDLE) NULL;

    dapl_dbg_log (DAPL_DBG_TYPE_API,
		  "dapl_evd_kcreate (%p, %d, %x, upcall, 0x%x, %p)\n",
		  ia_handle,
		  evd_min_qlen,
		  upcall_policy,
		  evd_flags,
		  evd_handle);

    ia_ptr      = (DAPL_IA *)ia_handle;
    cno_ptr     = NULL;
    evd_ptr     = NULL;
    *evd_handle = NULL;
    dat_status  = DAT_SUCCESS;

    if (DAPL_BAD_HANDLE (ia_handle, DAPL_MAGIC_IA))
    {
	dat_status = DAT_ERROR (DAT_INVALID_HANDLE, DAT_INVALID_HANDLE_IA);
	goto bail;
    }

    if ( evd_min_qlen <= 0 )
    {
	dat_status = DAT_ERROR (DAT_INVALID_PARAMETER, DAT_INVALID_ARG2);
	goto bail;
    }
    if ( evd_min_qlen > ia_ptr->hca_ptr->ia_attr.max_evd_qlen )
    {
	dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_TEVD);
	goto bail;
    }

    if (evd_flags & DAT_EVD_ASYNC_FLAG)
    {
	dat_status = DAT_ERROR (DAT_INVALID_PARAMETER, DAT_INVALID_ARG4);
	goto bail;
    }

    cno_ptr = dapl_cno_alloc (ia_ptr, upcall);

    if (!cno_ptr)
    {
	dat_status = DAT_INSUFFICIENT_RESOURCES;
	goto bail;
    }

    cno_ptr->cno_state = DAPL_CNO_STATE_TRIGGERED;
    cno_ptr->cno_upcall_policy = upcall_policy;

    dapl_ia_link_cno (ia_ptr, cno_ptr);

    dat_status = dapls_evd_internal_create (ia_ptr,
					  cno_ptr,
					  evd_min_qlen,
					  evd_flags,
					  &evd_ptr);
    if (dat_status != DAT_SUCCESS)
    {
	goto bail;
    }

    evd_ptr->evd_state   = DAPL_EVD_STATE_OPEN;
    evd_ptr->evd_enabled = DAT_TRUE;

    evdhandle = *evd_handle = (DAT_EVD_HANDLE) evd_ptr;

bail:
    if (dat_status != DAT_SUCCESS && cno_ptr != NULL)
    {
	dapl_ia_unlink_cno (ia_ptr, cno_ptr);
	dapl_cno_dealloc (cno_ptr);
    }

    if (dat_status != DAT_SUCCESS)
    {
	if (evd_ptr)
	{
	    dapl_evd_free (evd_ptr);
	}
    }

    return dat_status;
}

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */
