/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 *
 * MODULE: dapl_evd_kquery.c
 *
 * PURPOSE: Event management
 * Description: Interfaces in this file are completely described in
 *		the kDAPL 1.1 API, Chapter 6, section 3
 *
 * $Id: dapl_evd_kquery.c,v 1.2 2004/08/02 15:14:28 rabin Exp $
 **********************************************************************/

#include "dapl.h"

/*
 * dapl_evd_kquery
 *
 * kDAPL Requirements Version 1.1, 6.3.2.3
 *
 * Provides the consumer with arguments of the Event Dispatcher.
 *
 * Input:
 * 	evd_handle
 * 	evd_mask
 *
 * Output:
 * 	evd_param
 *
 * Returns:
 * 	DAT_SUCCESS
 * 	DAT_INVALID_PARAMETER
 */
DAT_RETURN
dapl_evd_kquery (
	IN	DAT_EVD_HANDLE		evd_handle,
	IN	DAT_EVD_PARAM_MASK	evd_param_mask,
	OUT	DAT_EVD_PARAM		*evd_param )
{
    DAPL_EVD	    *evd_ptr;
    DAPL_CNO	    *cno_ptr;
    DAT_RETURN	    dat_status;

    dat_status = DAT_SUCCESS;

    if ( NULL == evd_param )
    {
	dat_status = DAT_ERROR (DAT_INVALID_PARAMETER, DAT_INVALID_ARG3);
	goto bail;
    }

    /* Note: the spec. allows for events to be directed to a NULL EVD */
    /* with handle of type DAT_HANDLE_NULL. See 6.3.1                 */
    if ( DAT_HANDLE_NULL == evd_handle )
    {
	dapl_os_memzero (evd_param, sizeof (DAT_EVD_PARAM));
    }
    else
    {
	if ( DAPL_BAD_HANDLE (evd_handle, DAPL_MAGIC_EVD) )
	{
	    dat_status = DAT_ERROR (DAT_INVALID_HANDLE, 0);
	    goto bail;
	}

	evd_ptr = (DAPL_EVD *) evd_handle;
	cno_ptr = evd_ptr->cno_ptr;
	/*
	 * We may be racing against the thread safe modify
	 * calls here (dat_evd_{enable,disable,{set,clear}_unwaitable}).
	 * They are thread safe, so our reads need to be atomic with
	 * regard to those calls.  The below is ok (a single bit
	 * read counts as atomic; if it's in transition you'll get one
	 * of the correct values) but we'll need to be careful
	 * about reading the state variable atomically when we add
	 * in waitable/unwaitable.
	 */
	evd_param->ia_handle = evd_ptr->header.owner_ia;
	evd_param->evd_qlen = evd_ptr->qlen;

	if (cno_ptr != NULL)
	{
	    evd_param->upcall_policy = cno_ptr->cno_upcall_policy;
	    evd_param->upcall        = cno_ptr->cno_upcall;
	}
	evd_param->evd_flags  = evd_ptr->evd_flags;
    }

bail:
    return dat_status;
}


/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */
