/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 *
 * MODULE: dapl_init.c
 *
 * PURPOSE: Interface Adapter management
 * Description: Interfaces in this file are completely described in
 *		the DAPL 1.1 API, Chapter 6, section 2
 *
 **********************************************************************/

#include "dapl.h"
#include "dapl_hca_util.h"
#include "dapl_init.h"
#include "dapl_provider.h"
#include "dapl_mr_util.h"
#include "dapl_osd.h"
#include "dapl_adapter_util.h"
#include "dapl_name_service.h"
#include "dapl_timer_util.h"
#include "dapl_vendor.h"

#define DHCA_PORT_COUNT	2

/* Globally maintained variables */
static int    g_hca_count;
static DAT_PROVIDER_INFO   **g_dapl_provider_info;

void dapli_register_providers (void);


/*
 * dapl_init
 *
 * initialize this provider
 * includes initialization of all global variables
 * as well as registering all supported IAs with the dat registry
 *
 * This function needs to be called once when the provider is loaded.
 *
 * Input:
 *	none
 *
 * Output:
 *	none
 *
 * Return Values:
 */
void
dapl_init ( void )
{
    DAT_RETURN		dat_status;

    dapl_dbg_log (DAPL_DBG_TYPE_UTIL, "DAPL: Started (dapl_init)\n");

#if defined(DAPL_DBG)
    /* set up debug level */
    g_dapl_dbg_type = dapl_os_get_env_val ( "DAPL_DBG_TYPE",
					    DAPL_DBG_TYPE_ERR | DAPL_DBG_TYPE_WARN);
    /* set up debug level */
    g_dapl_dbg_dest = dapl_os_get_env_val ( "DAPL_DBG_DEST",
					    DAPL_DBG_DEST_STDOUT );
#endif /* DAPL_DBG */

    /* See if the user is on a loopback setup */
    g_dapl_loopback_connection = dapl_os_get_env_bool ( "DAPL_LOOPBACK" );

    dapl_dbg_log (DAPL_DBG_TYPE_UTIL, "DAPL: %s Setting Loopback\n",
		  g_dapl_loopback_connection ? "" : "NOT");

    /* initialize verbs library */
    dapls_ib_init ();

    /* initialize the timer */
    dapls_timer_init ();

    /* Set up name services */
    dat_status = dapls_ns_init ();
    if (DAT_SUCCESS != dat_status)
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR,
		      "dapls_ns_init failed %d\n", dat_status);
	goto bail;
    }

    /* initialize the provider list */
    dat_status = dapl_provider_list_create ();
    if (DAT_SUCCESS != dat_status)
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR,
		      "dapl_provider_list_create failed %d\n", dat_status);
	goto bail;
    }

#if defined(__KDAPL__)
    /* insert providers into the list and register with the registry */
    dapli_register_providers ();
#endif /* __KDAPL__ */

    return;

bail:
    dapl_dbg_log (DAPL_DBG_TYPE_ERR, "ERROR: dapl_init failed\n");
    dapl_fini ();
    return;
}


/*
 * dapl_fini
 *
 * finalize this provider
 * includes freeing of all global variables
 * as well as deregistering all supported IAs from the dat registry
 *
 * This function needs to be called once when the provider is loaded.
 *
 * Input:
 *	none
 *
 * Output:
 *	none
 *
 * Return Values:
 */
void
dapl_fini ( void )
{
    DAT_RETURN		dat_status;

    dapl_dbg_log (DAPL_DBG_TYPE_UTIL, "DAPL: Stopped (dapl_fini)\n");

#if defined(__KDAPL__)
    {
	DAT_PROVIDER_INFO   *provider_info;
	int		    i;
	int		    port_num;

	provider_info = (DAT_PROVIDER_INFO *)g_dapl_provider_info;

	for (i = g_hca_count; i > 0; i--)
	{
	    for (port_num = 0; port_num < DHCA_PORT_COUNT; port_num++)
	    {
		DAT_PROVIDER_FINI_FUNC_NAME (provider_info);
		provider_info++;
	    }
	}
    }

#endif /* __KDAPL__ */

    dat_status = dapl_provider_list_destroy ();
    if (DAT_SUCCESS != dat_status)
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR,
		      "dapl_provider_list_destroy failed %d\n", dat_status);
    }

    dapls_ib_release ();
    dapl_os_free ((void *)g_dapl_provider_info,
		 g_hca_count * DHCA_PORT_COUNT * sizeof (DAT_PROVIDER_INFO));

    return;
}


/*
 *
 * This function is called by the registry to initialize a provider
 *
 * The instance data string is expected to have the following form:
 *
 * <hca name> <port number>
 *
 */
void
DAT_PROVIDER_INIT_FUNC_NAME (
    IN const DAT_PROVIDER_INFO 	*provider_info,
    IN const char 		*instance_data )
{
    DAT_PROVIDER	*provider;
    DAPL_HCA		*hca_ptr;
    DAT_RETURN		dat_status;
    char *        	data;
    char *		name;
    char *		port;
    unsigned int 	len;
    unsigned int 	i;

    data = NULL;
    provider = NULL;
    hca_ptr = NULL;

    dat_status = dapl_provider_list_insert (provider_info->ia_name, &provider);
    if ( DAT_SUCCESS != dat_status )
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR,
		      "dat_provider_list_insert failed: %x\n", dat_status);
	goto bail;
    }

    data = dapl_os_strdup (instance_data);
    if ( NULL == data )
    {
	goto bail;
    }

    len = dapl_os_strlen (data);

    for ( i = 0; i < len; i++ )
    {
	if ( ' ' == data[i] )
	{
	    data[i] = '\0';
	    break;
	}
    }

    /* if the instance data did not have a valid format */
    if ( i == len )
    {
	goto bail;
    }

    name = data;
    port = data + (i + 1);

    hca_ptr = dapl_hca_alloc (name, port);
    if ( NULL == hca_ptr )
    {
	goto bail;
    }

    provider->extension = hca_ptr;

    /* register providers with dat_registry */
    dat_status = dat_registry_add_provider (provider, provider_info);
    /* dat_status = dat_registry_add_provider (provider); */
    if ( DAT_SUCCESS != dat_status )
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR,
		      "dat_registry_add_provider failed: %x\n", dat_status);
	goto bail;
    }

bail:
    if ( NULL != data )
    {
	dapl_os_free (data, len + 1);
    }

    if ( DAT_SUCCESS != dat_status )
    {
	if ( NULL != provider )
	{
	    (void) dapl_provider_list_remove (provider_info->ia_name);
	}

	if ( NULL != hca_ptr )
	{
	    dapl_hca_free (hca_ptr);
	}
    }
}


/*
 *
 * This function is called by the registry to de-initialize a provider
 *
 */
void
DAT_PROVIDER_FINI_FUNC_NAME (
    IN const DAT_PROVIDER_INFO 	*provider_info )
{
    DAT_PROVIDER	*provider;
    DAT_RETURN		dat_status;

    dat_status = dapl_provider_list_search (provider_info->ia_name, &provider);
    if ( DAT_SUCCESS != dat_status )
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR,
		      "dat_provider_list_search failed: %x\n", dat_status);
	return;
    }

    dat_status = dat_registry_remove_provider (provider, provider_info);
    /* dat_status = dat_registry_remove_provider (provider); */
    if ( DAT_SUCCESS != dat_status )
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR,
		      "dat_registry_remove_provider failed: %x\n", dat_status);
    }

    /*
    * free HCA memory
    */
    dapl_hca_free (provider->extension);
    
    (void) dapl_provider_list_remove (provider_info->ia_name);
}


/*
 * Deterimine what HCAs are available and register with the DAT
 * driver.
 * This might be better done by the provider driver invoking a
 * routine for each HCA, but for now we'll call into the provider
 * driver for services
 */

char *dev_name_suffix_table[3] =
{
    "",
    "a",
    "b"
};

//void
//dapli_register_providers ()
//{
//    DAPL_HCA_NAME	*hca_names;
//    DAT_COUNT		hca_count;
//    DAT_COUNT		hcaidx;
//    int			port_num;
//    int			provider_num;
//    DAT_PROVIDER_INFO   *provider_info;
//    DAT_RETURN		dat_status;
//    char		instance[DAT_NAME_MAX_LENGTH];
//    char		portstr[4];
//
//    /*
//     * Obtain a list of HCA names and their count
//     */
//    dat_status = dapls_ib_enum_hcas (&hca_names,
//				     &hca_count,
//				     VN_PREFIX );
//
//    if (dat_status != DAT_SUCCESS)
//    {
//	dapl_os_printf ("dapls_ib_enum_hcas () failed: %x\n", dat_status);
//	goto bail;
//    }
//
//    /* allocate provider info area */
//    g_hca_count = hca_count;
//    g_dapl_provider_info =
//	dapl_os_alloc (hca_count * DHCA_PORT_COUNT * sizeof (DAT_PROVIDER_INFO));
//
//    if (g_dapl_provider_info == NULL)
//    {
//	dapl_dbg_log (DAPL_DBG_TYPE_ERR,
//	      "dapli_register_providers() cannot allocate provider_info memory\n");
//	goto bail;
//    }
//    provider_num = 0;
//    provider_info = (DAT_PROVIDER_INFO *)g_dapl_provider_info;
//    for ( hcaidx = 0; hcaidx < hca_count; hcaidx++)
//    {
//	for (port_num = 0; port_num < DHCA_PORT_COUNT; port_num++)
//	{
//	    strncpy (provider_info->ia_name, hca_names[hcaidx],
//		     DAT_NAME_MAX_LENGTH);
//	    (void)strcat (provider_info->ia_name,
//			 dev_name_suffix_table[port_num+1]);
//
//	    provider_info->dapl_version_major = DAT_VERSION_MAJOR;
//	    provider_info->dapl_version_minor = DAT_VERSION_MINOR;
//	    provider_info->is_thread_safe = DAT_FALSE;
//
//	    /* build the instance string so we can call PROVIDER_INIT */
//	    strncpy (instance, provider_info->ia_name, DAT_NAME_MAX_LENGTH);
//	    portstr[0] = ' ';
//	    portstr[1] = '0' + port_num + 1;
//	    portstr[2] = '\0';
//	    strcat (instance, portstr);
//	    DAT_PROVIDER_INIT_FUNC_NAME (provider_info, instance);
//	    provider_info++;
//	}
//    }
//
//bail:
//    return;
//}

void                                                                              
dapli_register_providers ()                                                       
{                                                                                 
    DAPL_HCA_NAME     *hca_names;                                                 
    DAT_COUNT          hca_count;                                                  
    DAT_COUNT          hcaidx;                                                     
    int                provider_num;                                       
    DAT_PROVIDER_INFO *provider_info;                                           
    DAT_RETURN         dat_status;                                         
    char               *instance =NULL;

    instance = dapl_os_alloc(DAT_NAME_MAX_LENGTH);
    if (instance ==NULL) {
        return;
    }
    /*                                                                            
     * Obtain a list of HCA names and their count                                 
     */                                                                           
    dat_status = dapls_ib_enum_hcas (&hca_names,                                  
                                     &hca_count,                                    
                                     VN_PREFIX );                                   

    if (dat_status != DAT_SUCCESS)
    {
        dapl_os_printf ("dapls_ib_enum_hcas () failed: %x\n", dat_status);          
        goto bail;                                                                  
    }

    /* allocate provider info area */
    g_hca_count = hca_count;                                                      
    g_dapl_provider_info =                                                        
        dapl_os_alloc (hca_count * sizeof(DAT_PROVIDER_INFO));   

    if (g_dapl_provider_info == NULL)
    {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR,                                            
                      "dapli_register_providers() cannot allocate provider_info memory\n"); 
        goto bail;                                                                  
    }

    provider_num = 0;                                                             
    provider_info = (DAT_PROVIDER_INFO *)g_dapl_provider_info;                    

    for ( hcaidx = 0; hcaidx < hca_count; hcaidx++)
    {

        dapls_ib_get_instance_data(instance, hca_names[hcaidx]);

        strncpy (provider_info->ia_name, hca_names[hcaidx],
                 DAT_NAME_MAX_LENGTH);

        provider_info->dapl_version_major = DAT_VERSION_MAJOR;
        provider_info->dapl_version_minor = DAT_VERSION_MINOR;
        provider_info->is_thread_safe = DAT_FALSE; 

        DAT_PROVIDER_INIT_FUNC_NAME(provider_info, instance);

        provider_info++;
    }


 bail:
     dapl_os_free(instance,DAT_NAME_MAX_LENGTH);
    return;                                                                       
}                                                                                 


/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */
