/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 *
 * MODULE: dapl_lmr_kcreate.c
 *
 * PURPOSE: Memory management
 * Description: Interfaces in this file are completely described in
 *		the DAPL 1.1 API, Chapter 6, section 6
 *
 * $Id: dapl_lmr_kcreate.c,v 1.3 2004/08/02 15:14:28 rabin Exp $
 **********************************************************************/

#include "dapl_lmr_util.h"
#include "dapl_adapter_util.h"
#include "dapl_mem_fmr.h"

/*********************************************************************
 *                                                                   *
 * Function Prototypes                                               *
 *                                                                   *
 *********************************************************************/

DAT_RETURN
dapli_lmr_create_virtual (
    IN	DAPL_IA			*ia,
    IN	DAT_PVOID		virt_addr,
    IN	DAT_VLEN		length,
    IN	DAPL_PZ			*pz,
    IN	DAT_MEM_PRIV_FLAGS	privileges,
    OUT	DAT_LMR_HANDLE		*lmr_handle,
    OUT	DAT_LMR_CONTEXT		*lmr_context,
    OUT DAT_RMR_CONTEXT		*rmr_context,
    OUT	DAT_VLEN		*registered_length,
    OUT	DAT_VADDR		*registered_address );


DAT_RETURN
dapli_lmr_create_physical (
    IN	DAPL_IA			    *ia,
    IN	DAT_REGION_DESCRIPTION phys_addr,
    IN  DAT_VLEN            size,
    IN  DAT_MEM_TYPE        mem_type,
    IN	DAPL_PZ			    *pz,
    IN	DAT_MEM_PRIV_FLAGS	privileges,
    OUT	DAT_LMR_HANDLE		*lmr_handle,
    OUT	DAT_LMR_CONTEXT		*lmr_context,
    OUT DAT_RMR_CONTEXT		*rmr_context,
    OUT	DAT_VLEN		    *registered_length,
    OUT	DAT_VADDR		    *registered_address );


STATIC _INLINE_ DAT_RETURN
dapl_lmr_create_lmr (
    IN  DAPL_IA                 *ia,
    IN  DAPL_LMR	        *original_lmr,
    IN  DAPL_PZ                 *pz,
    IN  DAT_MEM_PRIV_FLAGS      privileges,
    OUT DAT_LMR_HANDLE          *lmr_handle,
    OUT DAT_LMR_CONTEXT         *lmr_context,
    OUT DAT_RMR_CONTEXT		*rmr_context,
    OUT DAT_VLEN                *registered_length,
    OUT DAT_VADDR               *registered_address );

/*********************************************************************
 *                                                                   *
 * Function Definitions                                              *
 *                                                                   *
 *********************************************************************/

DAT_RETURN
dapli_lmr_create_virtual (
    IN	DAPL_IA			*ia,
    IN	DAT_PVOID		virt_addr,
    IN	DAT_VLEN		length,
    IN	DAPL_PZ			*pz,
    IN	DAT_MEM_PRIV_FLAGS	privileges,
    OUT	DAT_LMR_HANDLE		*lmr_handle,
    OUT	DAT_LMR_CONTEXT		*lmr_context,
    OUT DAT_RMR_CONTEXT		*rmr_context,
    OUT	DAT_VLEN		*registered_length,
    OUT	DAT_VADDR		*registered_address )
{
    DAPL_LMR 			*lmr;
    DAT_REGION_DESCRIPTION 	reg_desc;
    DAT_RETURN			dat_status;

    reg_desc.for_va = virt_addr;
    dat_status = DAT_SUCCESS;

    lmr = dapl_lmr_alloc (ia,
			 DAT_MEM_TYPE_VIRTUAL,
			 reg_desc,
			 length,
			 (DAT_PZ_HANDLE) pz,
			 privileges);

    if ( NULL == lmr )
    {
	dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
	goto bail;
    }

    dat_status = dapls_ib_mr_register (ia,
				       lmr,
				       virt_addr,
				       length,
				       privileges);

    if (DAT_SUCCESS != dat_status)
    {
	dapl_lmr_dealloc (lmr);
	goto bail;
    }

    /* if the LMR context is already in the hash table */
    dat_status =  dapls_hash_search (ia->hca_ptr->lmr_hash_table,
				     lmr->param.lmr_context,
				     NULL);
    if (dat_status == DAT_SUCCESS)
    {
	(void)dapls_ib_mr_deregister (lmr);
	dapl_lmr_dealloc (lmr);

	dat_status = DAT_ERROR (DAT_INVALID_STATE, DAT_INVALID_STATE_LMR_IN_USE);
	goto bail;
    }

    dat_status = dapls_hash_insert (ia->hca_ptr->lmr_hash_table,
				    lmr->param.lmr_context,
				    lmr);
    if (dat_status != DAT_SUCCESS)
    {
	(void)dapls_ib_mr_deregister (lmr);
	dapl_lmr_dealloc (lmr);

	/* The value returned by dapls_hash_insert(.) is not    */
	/* returned to the consumer because the spec. requires */
	/* that dat_lmr_create(.) return only certain values.  */
	dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
	goto bail;
    }

    dapl_os_atomic_inc (&pz->pz_ref_count);

    if (NULL != lmr_handle)
    {
	*lmr_handle = (DAT_LMR_HANDLE) lmr;
    }
    if (NULL != lmr_context)
    {
	*lmr_context = (DAT_LMR_CONTEXT) lmr->param.lmr_context;
    }
    if (NULL != rmr_context)
    {
	*rmr_context = (DAT_LMR_CONTEXT) lmr->param.rmr_context;
    }
    if (NULL != registered_address)
    {
	*registered_address = (DAT_VADDR) (uintptr_t) virt_addr;
    }
    if (NULL != registered_length)
    {
	*registered_length = length;
    }

bail:
    return dat_status;


}
struct my_buf {
    u_int64_t start;
    u_int64_t size;
    };

DAT_RETURN
dapli_lmr_create_physical (
    IN	DAPL_IA			    *ia,
    IN	DAT_REGION_DESCRIPTION phys_addr,
    IN  DAT_VLEN            size,
    IN  DAT_MEM_TYPE        mem_type,
    IN	DAPL_PZ			    *pz,
    IN	DAT_MEM_PRIV_FLAGS	privileges,
    OUT	DAT_LMR_HANDLE		*lmr_handle,
    OUT	DAT_LMR_CONTEXT		*lmr_context,
    OUT DAT_RMR_CONTEXT		*rmr_context,
    OUT	DAT_VLEN		    *registered_length,
    OUT	DAT_VADDR		    *registered_address )
{
    DAPL_LMR 			    *lmr;
    DAT_REGION_DESCRIPTION 	reg_desc;
    DAT_RETURN			    dat_status;
    DAT_UINT32              ii;
    DAT_VLEN		        length, page_count;
    struct my_buf              buf_4_type_ia;
    struct my_buf *            buf_array_4_type_physical = NULL;
    dapl_dbg_log (DAPL_DBG_TYPE_API,"%s: ia=%p phys_addr=%p size=0x%llx "
                  "mem_type=0x%x pz=%p privileges=0x%x\n", __FUNCTION__, 
                  ia, phys_addr, size, mem_type, pz, privileges);

    if (!active_fmr && (DAT_MEM_TYPE_PLATFORM == mem_type))
    {
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "%s: mem_type FMR not activated\n", 
                      __FUNCTION__);
        dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
        goto bail;
    }

    if (DAT_MEM_TYPE_IA == mem_type) 
    {
        length = size;
        page_count = 1;
        buf_4_type_ia.start = phys_addr.for_pa;
        buf_4_type_ia.size  = size;
        reg_desc.for_va = (void *)&buf_4_type_ia;
    } 
    else //DAT_MEM_TYPE_PHYSICAL or DAT_MEM_TYPE_PLATFORM (FMR)
    {
        length = 0;
        page_count = size;
        buf_array_4_type_physical = dapl_os_alloc(page_count * sizeof(struct my_buf));
        if (!buf_array_4_type_physical) {
            dapl_dbg_log (DAPL_DBG_TYPE_ERR, "%s: no memory to alloc buf array\n", 
                          __FUNCTION__);
            dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
            goto bail;
        }
        for (ii=0; ii<page_count; ii++)
        {
            buf_array_4_type_physical[ii].start = ((u64*)(phys_addr.for_array))[ii];
            buf_array_4_type_physical[ii].size = PAGE_SIZE;
            length += buf_array_4_type_physical[ii].size;
        }
        reg_desc.for_va = (void *)buf_array_4_type_physical;
        dapl_dbg_log (DAPL_DBG_TYPE_UTIL, "%s: page_count=0x%llx length=0x%llx \n",
                  __FUNCTION__, page_count, length);
    }
    dat_status = DAT_SUCCESS;

    lmr = dapl_lmr_alloc (ia,
			 mem_type,
			 reg_desc,
			 length,
			 (DAT_PZ_HANDLE) pz,
			 privileges);

    if ( NULL == lmr )
    {
        dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
        goto bail;
    }

    if (mem_type == DAT_MEM_TYPE_PLATFORM) 
    {
        
        dat_status = dapls_ib_fmr (ia,
                        lmr,
                        reg_desc.for_va,
						page_count,
                        privileges);
    }
    else
    {
        dat_status = dapls_ib_mr_register_physical (ia,
						lmr,
                        reg_desc.for_va,
						page_count,
                        privileges);
    }
    
    if (DAT_SUCCESS != dat_status)
    {
        dapl_lmr_dealloc (lmr);
        goto bail;
    }

    /* if the LMR context is already in the hash table */
    dat_status =  dapls_hash_search (ia->hca_ptr->lmr_hash_table,
				     lmr->param.lmr_context,
				     NULL);
    if (dat_status == DAT_SUCCESS)
    {
        (void)dapls_ib_mr_deregister (lmr);
        dapl_lmr_dealloc (lmr);

        dat_status = DAT_ERROR (DAT_INVALID_STATE, DAT_INVALID_STATE_LMR_IN_USE);
        goto bail;
    }

    dat_status = dapls_hash_insert (ia->hca_ptr->lmr_hash_table,
				    lmr->param.lmr_context,
				    lmr);
    if (dat_status != DAT_SUCCESS)
    {
        (void)dapls_ib_mr_deregister (lmr);
        dapl_lmr_dealloc (lmr);

    	/* The value returned by dapls_hash_insert(.) is not    */
    	/* returned to the consumer because the spec. requires */
    	/* that dat_lmr_create(.) return only certain values.  */
    	dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
    	goto bail;
    }

    dapl_os_atomic_inc (&pz->pz_ref_count);
    
    if (NULL != lmr_handle)
    {
        *lmr_handle = (DAT_LMR_HANDLE) lmr;
    }
    if (NULL != lmr_context)
    {
        *lmr_context = (DAT_LMR_CONTEXT) lmr->param.lmr_context;
    }
    if (NULL != rmr_context)
    {
        *rmr_context = (DAT_LMR_CONTEXT) lmr->param.rmr_context;
    }
    if (NULL != registered_address)
    {
        *registered_address = lmr->param.registered_address;
    }
    if (NULL != registered_length)
    {
        *registered_length = lmr->param.registered_size;
    }

bail:
    if(!buf_array_4_type_physical) {
        dapl_os_free(buf_array_4_type_physical,0);
    }
    return dat_status;
}


STATIC _INLINE_ DAT_RETURN
dapl_lmr_create_lmr (
    IN  DAPL_IA                 *ia,
    IN  DAPL_LMR	        *original_lmr,
    IN  DAPL_PZ                 *pz,
    IN  DAT_MEM_PRIV_FLAGS      privileges,
    OUT DAT_LMR_HANDLE          *lmr_handle,
    OUT DAT_LMR_CONTEXT         *lmr_context,
    OUT DAT_RMR_CONTEXT		*rmr_context,
    OUT DAT_VLEN                *registered_length,
    OUT DAT_VADDR               *registered_address )
{
    DAPL_LMR 			*lmr;
    DAT_REGION_DESCRIPTION 	reg_desc;
    DAT_RETURN			dat_status;

    dat_status = dapls_hash_search (ia->hca_ptr->lmr_hash_table,
				    original_lmr->param.lmr_context,
				    (DAPL_HASH_DATA *) &lmr);
    if ( dat_status != DAT_SUCCESS )
    {
	dat_status = DAT_ERROR (DAT_INVALID_PARAMETER, DAT_INVALID_ARG2);
	goto bail;
    }

    reg_desc.for_lmr_handle = (DAT_LMR_HANDLE) original_lmr;

    lmr = dapl_lmr_alloc (ia,
			 DAT_MEM_TYPE_LMR,
			 reg_desc,
			 0, /* length is meaningless */
			 (DAT_PZ_HANDLE) pz,
			 privileges);

    if (NULL == lmr)
    {
	dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
	goto bail;
    }

    dat_status = dapls_ib_mr_register_shared (ia,
					      lmr,
					      privileges);

    if (DAT_SUCCESS != dat_status)
    {
	dapl_lmr_dealloc (lmr);
	goto bail;
    }

    /* if the LMR context is already in the hash table */
    dat_status = dapls_hash_search (ia->hca_ptr->lmr_hash_table,
				    lmr->param.lmr_context,
				    NULL);
    if (dat_status == DAT_SUCCESS)
    {
	dapls_ib_mr_deregister (lmr);
	dapl_lmr_dealloc (lmr);

	dat_status = DAT_ERROR (DAT_INVALID_STATE, DAT_INVALID_STATE_LMR_IN_USE);
	goto bail;
    }

    dat_status = dapls_hash_insert (ia->hca_ptr->lmr_hash_table,
				    lmr->param.lmr_context,
				    lmr);
    if (dat_status != DAT_SUCCESS)
    {
	dapls_ib_mr_deregister (lmr);
	dapl_lmr_dealloc (lmr);

	/* The value returned by dapls_hash_insert(.) is not    */
	/* returned to the consumer because the spec. requires */
	/* that dat_lmr_create(.) return only certain values.  */
	dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
	goto bail;
    }

    dapl_os_atomic_inc (&pz->pz_ref_count);

    if (NULL != lmr_handle)
    {
	*lmr_handle = (DAT_LMR_HANDLE) lmr;
    }
    if (NULL != lmr_context)
    {
	*lmr_context = (DAT_LMR_CONTEXT) lmr->param.lmr_context;
    }
    if (NULL != rmr_context)
    {
	*rmr_context = (DAT_LMR_CONTEXT) lmr->param.rmr_context;
    }
    if (NULL != registered_address)
    {
	*registered_address = (DAT_VADDR) (uintptr_t) original_lmr->param.region_desc.for_va;
    }
    if (NULL != registered_length)
    {
	*registered_length = original_lmr->param.length;
    }
bail:
    return dat_status;
}


/*
 * dapl_lmr_kcreate
 *
 * DAPL Requirements Version xxx, 6.6.3.1
 *
 * Register a memory region with an Interface Adaptor.
 *
 * Input:
 *	ia_handle
 *	mem_type
 *	region_description
 *	length
 *	pz_handle
 *	privileges
 *	optimization
 *
 * Output:
 *	lmr_handle
 *	lmr_context
 *	registered_length
 *	registered_address
 *
 * Returns:
 * 	DAT_SUCCESS
 * 	DAT_INSUFFICIENT_RESOURCES
 * 	DAT_INVALID_PARAMETER
 * 	DAT_INVALID_STATE
 * 	DAT_MODEL_NOT_SUPPORTED
 *
 */
DAT_RETURN
dapl_lmr_kcreate (
	IN	DAT_IA_HANDLE		ia_handle,
	IN	DAT_MEM_TYPE		mem_type,
	IN	DAT_REGION_DESCRIPTION	region_description,
	IN	DAT_VLEN		size,
	IN	DAT_PZ_HANDLE		pz_handle,
	IN	DAT_MEM_PRIV_FLAGS	privileges,
	IN	DAT_MEM_OPT		optimization,
	OUT	DAT_LMR_HANDLE		*lmr_handle,
	OUT	DAT_LMR_CONTEXT		*lmr_context,
	OUT     DAT_RMR_CONTEXT		*rmr_context,
	OUT	DAT_VLEN		*registered_length,
	OUT	DAT_VADDR		*registered_address )
{
    DAPL_IA 		*ia;
    DAPL_PZ 		*pz;
    DAT_RETURN		dat_status;

    dapl_dbg_log (DAPL_DBG_TYPE_API,
        "%s: ia=%p mem_type=0x%x region_description=%p size=0x%llx pz=%p privileges=0x%x\n",
		__FUNCTION__, ia_handle, mem_type, region_description, size, pz_handle, privileges);
    
    if ( DAPL_BAD_HANDLE (ia_handle, DAPL_MAGIC_IA) )
    {
	dat_status = DAT_ERROR (DAT_INVALID_HANDLE, DAT_INVALID_HANDLE_IA);
	goto bail;
    }
    if ( DAPL_BAD_HANDLE (pz_handle, DAPL_MAGIC_PZ) )
    {
	dat_status = DAT_ERROR (DAT_INVALID_HANDLE, DAT_INVALID_HANDLE_PZ);
	goto bail;
    }

    ia = (DAPL_IA *) ia_handle;
    pz = (DAPL_PZ *) pz_handle;

    switch (mem_type)
    {
	case DAT_MEM_TYPE_VIRTUAL:
	{
	    dat_status = dapli_lmr_create_virtual (
		ia, region_description.for_va, size, pz, privileges,
		lmr_handle, lmr_context, rmr_context, registered_length,
		registered_address);
	    break;
	}
    case DAT_MEM_TYPE_PLATFORM: // using as proprietary FMR type
    case DAT_MEM_TYPE_IA:
	case DAT_MEM_TYPE_PHYSICAL:
	{
	    /* FIXME!!
	     * XXX Need to fix for 64 bit phys addrs! use
	     * XXX region_description.for_pointer, but this neeeds a fix
	     * XXX in the spec to deal with > 4G memories, common to
	     * XXX modern servers.
	     */
	    dat_status = dapli_lmr_create_physical (
            ia, region_description, size, mem_type, 
            pz, privileges, lmr_handle, lmr_context, rmr_context, 
            registered_length, registered_address);
	    break;
	}
	case DAT_MEM_TYPE_LMR:
	{
	    DAPL_LMR *lmr;

	    if ( DAPL_BAD_HANDLE (region_description.for_lmr_handle, DAPL_MAGIC_LMR) )
	    {
		dat_status = DAT_ERROR (DAT_INVALID_HANDLE, DAT_INVALID_HANDLE_LMR);
		goto bail;
	    }

	    lmr = (DAPL_LMR *) region_description.for_lmr_handle;

	    dat_status = dapl_lmr_create_lmr (
		ia, lmr, pz, privileges, lmr_handle,
		lmr_context, rmr_context, registered_length, registered_address);
	    break;
	}
    case DAT_MEM_TYPE_BYPASS:
	{
	    dat_status = DAT_ERROR (DAT_NOT_IMPLEMENTED, DAT_NO_SUBTYPE);
	    break;
	}
	default:
	{
	    dat_status = DAT_ERROR (DAT_INVALID_PARAMETER, DAT_INVALID_ARG2);
	    break;
	}
    }

bail:
    return dat_status;
}

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */
