/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 *
 * MODULE: dapl_osd.c
 *
 * PURPOSE: Operating System Dependent layer
 * Description:
 *	Provide OS dependent functions with a canonical DAPL
 *	interface. Designed to be portable and hide OS specific quirks
 *	of common functions.
 *
 *
 * $Id: dapl_osd.c,v 1.2 2004/08/02 15:14:28 rabin Exp $
 **********************************************************************/

#include "dapl_osd.h"
#include "dapl.h"
#include "dapl_hca_util.h"
#include "dapl_ia_util.h"
#include "dapl_rmr_util.h"
#include "dapl_lmr_util.h"
#include "dapl_pz_util.h"
#include "dapl_ep_util.h"
#include "dapl_cr_util.h"
#include "dapl_evd_util.h"
#include "dapl_sp_util.h"
#include "dapl_adapter_util.h"
#include "dapl_provider.h"
#include "dapl_hash.h"
#include "dapl_timer_util.h"

/*
 * dapl_osd_init
 *
 * Do Linux initialization:
 * - Set up fork handler to clean up DAPL resources in the child
 *   process after a fork().
 *
 * Input:
 *      none
 *
 * Returns:
 *	DAT_SUCCESS
 */
void
dapl_os_init ( )
{
    return;
}


/*
 * dapl_os_get_time
 *
 * Return 64 bit value of current time in microseconds.
 *
 * Input:
 *      loc       User location to place current time
 *
 * Returns:
 *	DAT_SUCCESS
 */

DAT_RETURN
dapl_os_get_time (
    OUT DAPL_OS_TIMEVAL * loc)
{
    struct timeval	tv;

    do_gettimeofday (&tv);
    *loc = ((DAT_UINT64) (tv.tv_sec) * 1000000L) + (DAT_UINT64) tv.tv_usec;

    return DAT_SUCCESS;
}


/*
 * get_os_env_bool
 *
 * Return boolean value of passed in environment variable: 1 if present,
 * 0 if not
 *
 * Input:
 *
 *
 * Returns:
 *	TRUE or FALSE
 */
int
dapl_os_get_env_bool (char		*env_str )
{
    return 0;
}


/*
 * dapl_os_get_env_val
 *
 * Update val to  value of passed in environment variable if present
 *
 * Input:
 *      env_str
 *	def_val		default value if environment variable does not exist
 *
 * Returns:
 *	TRUE or FALSE
 */
int
dapl_os_get_env_val (
	char		*env_str,
	int		def_val )
{
    extern int DbgLvl;			/* defined in dapl_module.c */

    if ( strcmp (env_str, "DAPL_DBG_TYPE") == 0 )
    {
	if (DbgLvl != 0)
	{
	    def_val = DbgLvl;
	}
    }
    else if ( strcmp (env_str, "DAPL_DBG_DEST") == 0 )
    {
	;			       /* do nothing */
    }

    return  def_val;
}


/*
 * Wait object routines
 */

/*
 * dapl_os_wait_object_init
 *
 * Initialize a wait object
 *
 * Input:
 *	wait_obj
 *
 * Returns:
 *	DAT_SUCCESS
 *	DAT_INTERNAL_ERROR
 */
DAT_RETURN
dapl_os_wait_object_init (
    IN DAPL_OS_WAIT_OBJECT *wait_obj)
{
    init_waitqueue_head (&wait_obj->wait_queue);

    return DAT_SUCCESS;
}


/* Wait on the supplied wait object, up to the specified time_out.
 * A timeout of DAT_TIMEOUT_INFINITE will wait indefinitely.
 * Timeout should be specified in micro seconds.
 *
 * Functional returns:
 *	DAT_SUCCESS -- another thread invoked dapl_os_wait object_wakeup
 * 	DAT_INVALID_STATE -- someone else is already waiting in this wait
 * 	object.
 *			     only one waiter is allowed at a time.
 *	DAT_ABORT -- another thread invoked dapl_os_wait_object_destroy
 *	DAT_TIMEOUT -- the specified time limit was reached.
 */

DAT_RETURN
dapl_os_wait_object_wait (
    IN	DAPL_OS_WAIT_OBJECT *wait_obj,
    IN  DAT_TIMEOUT timeout_val,
    IN  int * flag)
{
    DAT_RETURN 		dat_status;

    dat_status = DAT_SUCCESS;

    if ( DAT_TIMEOUT_INFINITE == timeout_val )
    {
        wait_event_interruptible(wait_obj->wait_queue,(*flag==1));
    }
    else
    {
	int     expire;

	expire = timeout_val * HZ / 1000000;

	while ( expire )
	{
	    current->state = TASK_INTERRUPTIBLE;
	    expire = schedule_timeout (expire);
	}
	dat_status = DAT_ERROR (DAT_TIMEOUT_EXPIRED, 0);
    }
    *flag = 0;
    return dat_status;
}


/*
 * dapl_os_wait_object_wakeup
 *
 * Wakeup a thread waiting on a wait object
 *
 * Input:
 *      wait_obj
 *
 * Returns:
 *	DAT_SUCCESS
 *	DAT_INTERNAL_ERROR
 */
DAT_RETURN
dapl_os_wait_object_wakeup (
    IN	DAPL_OS_WAIT_OBJECT *wait_obj)
{
    wake_up_interruptible (&wait_obj->wait_queue);

    return DAT_SUCCESS;
}


/*
 * dapl_os_wait_object_destroy
 *
 * Destroy a wait object
 *
 * Input:
 *      wait_obj
 *
 * Returns:
 *	DAT_SUCCESS
 *	DAT_INTERNAL_ERROR
 */
DAT_RETURN
dapl_os_wait_object_destroy (
    IN	DAPL_OS_WAIT_OBJECT *wait_obj)
{

    return DAT_SUCCESS;
}




/*
 * Structure to contain all elements of a thread in order to enable a
 * local routine to intercept and do some necessary initialization.
 */
struct thread_draft
{
    void		(*func) (void *);		/* start routine */
    void 		*data;		/* argument to start routine */
};

void dapli_thread_init (	struct thread_draft	*thread_draft);

/*
 * dapls_os_thread_create
 *
 * Create a thread for dapl
 *
 * Input:
 *	func		function to invoke thread
 *	f_arg		argument to pass to function
 *
 * Output
 *	thread_id	handle for thread
 *
 * Returns:
 *	DAT_SUCCESS
 */
DAT_RETURN
dapl_os_thread_create (
	IN  void			(*func) (void *),
	IN  void 			*data,
	OUT DAPL_OS_THREAD		*thread_id )
{
    struct thread_draft		*thread_draft;

    thread_draft = dapl_os_alloc (sizeof ( struct thread_draft));

    thread_draft->func = func;
    thread_draft->data = data;

    *thread_id = kernel_thread ((void *)dapli_thread_init, (void *)thread_draft, 0);

    return DAT_SUCCESS;
}

/*
 * dapli_thread_init
 *
 * Need to mask all signals from this thread in order to be a good
 * citizen. Signals will arrive randomly and will be processed by
 * whatever thread is running unless they are specifically blocked; and
 * this should be a user thread, not a dapl thread
 */

void
dapli_thread_init (
	struct thread_draft		*thread_draft)
{
    void		(*func) (void *);
    void 		*data;

    func = thread_draft->func;
    data = thread_draft->data;
    dapl_os_free (thread_draft, sizeof ( struct thread_draft ));

    (*func) (data);

    return;
}

void
dapl_os_vprintf (
    const char *		fmt,
    va_list			args)
{
    static char buf[1024];

    vsnprintf (buf, sizeof (buf), fmt, args);
    printk ("kDAPL: %s", buf);
    return;
}

#if defined(DAPL_DBG)
void
dapl_os_dbg_print (
	DAPL_DBG_TYPE		type,
	const char *		fmt,
	...)
{
    static char buf[1024];

    if ( type & g_dapl_dbg_type )
    {
	va_list args;

	va_start (args, fmt);
	vsnprintf (buf, sizeof (buf), fmt, args);
	va_end (args);
	printk ("kDAPL: %s", buf);
    }
}
#endif /* DAPL_DBG */


/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */
