/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 *
 * HEADER: dapl_osd.h
 *
 * PURPOSE: Operating System Dependent layer
 * Description:
 *	Provide OS dependent data structures & functions with
 *	a canonical DAPL interface. Designed to be portable
 *	and hide OS specific quirks of common functions.
 *
 * $Id$
 **********************************************************************/

#ifndef _DAPL_OSD_H_
#define _DAPL_OSD_H_

/*
 * This file is defined for Linux systems only, including it on any
 * other build will cause an error
 */
#ifndef __linux__
#error UNDEFINED OS TYPE
#endif /* __linux__ */

#ifndef __KERNEL__
#error "Must compile for the kernel"
#endif	/* __KERNEL__ */

#ifndef __KDAPL__
#error "Must compile for KDAPL"
#endif	/* __KDAPL__ */

#if !defined (__i386__) && !defined (__ia64__) && !defined(__PPC__)  && !defined (__x86_64__)
#error UNDEFINED ARCH
#endif


#include <dat/kdat.h>

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <asm/types.h>
#include <asm/atomic.h>
#include <linux/delay.h>

#include "dapl_debug.h"
#include <asm/div64.h>			/* needed by hash functions */

#ifdef __ia64__
#include <linux/version.h>
#include <asm/atomic.h>
#include <asm/system.h>
#endif


/* Useful debug definitions */
#ifndef STATIC
#define STATIC static
#endif /* STATIC */
#ifndef _INLINE_
#define _INLINE_ __inline__
#endif /* _INLINE_ */

typedef unsigned long   uintptr_t;

void dapl_os_init ( void );	/* initialization function */

#define dapl_os_panic(fmt, args...) 			\
	do {					\
	     printk("PANIC in %s:%i:%s\n", __FILE__, __LINE__, __func__); \
	     panic(fmt, ## args);		\
	} while(0)

#define dapl_ip_addr6(sockaddr) (((struct sockaddr_in6 *)sockaddr)->sin6_addr.s6_addr32)

/*
 * Atomic operations
 */

typedef atomic_t DAPL_ATOMIC;

/* atomic function prototypes */
STATIC _INLINE_ void
dapl_os_atomic_inc (
	INOUT	DAPL_ATOMIC *v);

STATIC _INLINE_ void
dapl_os_atomic_dec (
	INOUT	DAPL_ATOMIC *v);

STATIC _INLINE_ DAT_COUNT
dapl_os_atomic_assign (
    INOUT DAPL_ATOMIC *v,
    IN	DAT_COUNT match_value,
    IN	DAT_COUNT new_value );

#define dapl_os_atomic_read(v)		atomic_read(v)
#define dapl_os_atomic_set(v,i)		atomic_set(v,i)

int dapl_os_get_env_bool (
	char		*env_str );

int dapl_os_get_env_val (
	char		*env_str,
	int		def_val );



/* atomic functions */

/* dapl_os_atomic_inc
 *
 * get the current value of '*v', and then increment it.
 *
 * This is equivalent to an IB atomic fetch and add of 1,
 * except that a DAT_COUNT might be 32 bits, rather than 64
 * and it occurs in local memory.
 */

STATIC _INLINE_ void
dapl_os_atomic_inc (
	INOUT	DAPL_ATOMIC *v)
{
#ifdef __ia64__
    DAT_COUNT	old_value;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    IA64_FETCHADD  (old_value,v,1,4);
#else
    IA64_FETCHADD  (old_value,v,1,4,rel);
#endif

#else
	atomic_inc(v);
#endif
    return;
}


/* dapl_os_atomic_dec
 *
 * decrement the current value of '*v'. No return value is required.
 */

STATIC _INLINE_ void
dapl_os_atomic_dec (
	INOUT	DAPL_ATOMIC *v)
{
#ifdef __ia64__
    DAT_COUNT	old_value;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    IA64_FETCHADD  (old_value,v,-1,4);
#else
    IA64_FETCHADD  (old_value,v,-1,4,rel);
#endif

#else
	atomic_dec(v);
#endif
    return;
}


/* dapl_os_atomic_assign
 *
 * assign 'new_value' to '*v' if the current value
 * matches the provided 'match_value'.
 *
 * Make no assignment if there is no match.
 *
 * Return the current value in any case.
 *
 * This matches the IBTA atomic operation compare & swap
 * except that it is for local memory and a DAT_COUNT may
 * be only 32 bits, rather than 64.
 */

STATIC _INLINE_ DAT_COUNT
dapl_os_atomic_assign (
    INOUT DAPL_ATOMIC *v,
    IN	DAT_COUNT match_value,
    IN	DAT_COUNT new_value )
{
    DAT_COUNT	current_value;

    /*
     * Use the Pentium compare and exchange instruction
     */

#ifdef __ia64__


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
current_value = ia64_cmpxchg("acq",v,match_value,new_value,4);
#else
current_value = ia64_cmpxchg(acq,v,match_value,new_value,4);
#endif

#elif defined (__PPC__)
    
    current_value = __cmpxchg_u32((volatile int *)v, (int) match_value, (int)new_value);

#else
    current_value = __cmpxchg((volatile void *)v, 
                              (unsigned long)match_value,
                              (unsigned long)new_value, 
                              (int)4);
#endif
    return current_value;
}


/*
 * Thread Functions
 */
typedef int			DAPL_OS_THREAD;

DAT_RETURN
dapl_os_thread_create (
	IN  void			(*func)	(void *),
	IN  void			*data,
	OUT DAPL_OS_THREAD		*thread_id );


/*
 * Lock Functions
 */

typedef spinlock_t 	DAPL_OS_LOCK;

/* function prototypes */
STATIC _INLINE_ DAT_RETURN
dapl_os_lock_init (
    IN	DAPL_OS_LOCK *m);

STATIC _INLINE_ DAT_RETURN
dapl_os_lock (
    IN	DAPL_OS_LOCK *m);

STATIC _INLINE_ DAT_RETURN
dapl_os_unlock (
    IN	DAPL_OS_LOCK *m);

STATIC _INLINE_ DAT_RETURN
dapl_os_lock_destroy (
    IN	DAPL_OS_LOCK *m);

/* lock functions */
STATIC _INLINE_ DAT_RETURN
dapl_os_lock_init (
    IN	DAPL_OS_LOCK *m)
{
    spin_lock_init ( m );

    return DAT_SUCCESS;
}

STATIC _INLINE_ DAT_RETURN
dapl_os_lock (
    IN	DAPL_OS_LOCK *m)
{
    spin_lock ( m );
    return DAT_SUCCESS;
}

STATIC _INLINE_ DAT_RETURN
dapl_os_unlock (
    IN	DAPL_OS_LOCK *m)
{
    spin_unlock ( m );
    return DAT_SUCCESS;
}

STATIC _INLINE_ DAT_RETURN
dapl_os_lock_destroy (
    IN	DAPL_OS_LOCK *m)
{
    /* Nothing */
    return DAT_SUCCESS;
}


/*
 * Wait Objects
 */

/*
 * The wait object invariant: Presuming a call to dapl_os_wait_object_wait
 * occurs at some point, there will be at least one wakeup after each call
 * to dapl_os_wait_object_signal.  I.e. Signals are not ignored, though
 * they may be coallesced.
 */
typedef struct _dapl_os_wait_object_
{
    wait_queue_head_t       wait_queue;
} DAPL_OS_WAIT_OBJECT;


/* function prototypes */
DAT_RETURN
dapl_os_wait_object_init (
    IN DAPL_OS_WAIT_OBJECT *wait_obj);

DAT_RETURN
dapl_os_wait_object_wait (
    IN	DAPL_OS_WAIT_OBJECT *wait_obj,
    IN  DAT_TIMEOUT timeout_val,
    IN  int * flag);

DAT_RETURN
dapl_os_wait_object_wakeup (
    IN	DAPL_OS_WAIT_OBJECT *wait_obj);

DAT_RETURN
dapl_os_wait_object_destroy (
    IN	DAPL_OS_WAIT_OBJECT *wait_obj);

/*
 * Memory Functions
 */

/* function prototypes */
STATIC _INLINE_ void *dapl_os_alloc (int size);

STATIC _INLINE_ void *dapl_os_realloc (void *ptr, int size);

STATIC _INLINE_ void dapl_os_free (void *ptr, int size);

STATIC _INLINE_ void * dapl_os_memzero (void *loc, int size);

STATIC _INLINE_ void * dapl_os_memcpy (void *dest, const void *src, int len);

STATIC _INLINE_ int dapl_os_memcmp (const void *mem1, const void *mem2, int len);

/* memory functions */


STATIC _INLINE_ void *dapl_os_alloc (int size)
{
    return kmalloc (size, GFP_ATOMIC);
}

STATIC _INLINE_ void *dapl_os_realloc (void *ptr, int size)
{
    void	*newptr;

    newptr = kmalloc (size, GFP_ATOMIC);
    if (newptr != NULL)
    {
	/* copy the mem array to the new */
	memcpy (newptr, ptr, size);
	kfree (ptr);
    }
    return newptr;
}


STATIC _INLINE_ void dapl_os_free (void *ptr, int size)
{
    kfree (ptr);
}

STATIC _INLINE_ void * dapl_os_memzero (void *loc, int size)
{
    return memset (loc, 0, size);
}

STATIC _INLINE_ void * dapl_os_memcpy (void *dest, const void *src, int len)
{
    return memcpy (dest, src, len);
}

STATIC _INLINE_ int dapl_os_memcmp (const void *mem1, const void *mem2, int len)
{
    return memcmp (mem1, mem2, len);
}


/*
 * Memory coherency functions
 * For i386 Linux, there are no coherency issues so we just return success.
 */
STATIC _INLINE_  DAT_RETURN
dapl_os_sync_rdma_read (
    IN      const DAT_LMR_TRIPLET	*local_segments,
    IN      DAT_VLEN			num_segments)
{
    return DAT_SUCCESS;
}

STATIC _INLINE_  DAT_RETURN
dapl_os_sync_rdma_write (
    IN      const DAT_LMR_TRIPLET	*local_segments,
    IN      DAT_VLEN			num_segments)
{
    return DAT_SUCCESS;
}



/*
 * String Functions
 */

STATIC _INLINE_ unsigned int dapl_os_strlen (const char *str)
{
    return strlen (str);
}

STATIC _INLINE_ char * dapl_os_strdup (const char *str)
{
    char *ns = kmalloc (strlen (str)+1, GFP_KERNEL);
    if (ns)
    {
	strcpy (ns, str);
    }
    return ns;
}


/*
 * Timer Functions
 */

typedef DAT_UINT64		    DAPL_OS_TIMEVAL;
typedef struct dapl_timer_entry	    DAPL_OS_TIMER;
typedef unsigned long long int	    DAPL_OS_TICKS;


/* timer function prototype */
DAT_RETURN dapl_os_get_time (DAPL_OS_TIMEVAL *);


/*
 *
 * Name Service Helper functions
 *
 */
#if defined(IBHOSTS_NAMING) || defined(CM_BUSTED)
struct addrinfo
{
  int ai_flags;                 /* Input flags.  */
  int ai_family;                /* Protocol family for socket.  */
  int ai_socktype;              /* Socket type.  */
  int ai_protocol;              /* Protocol for socket.  */
  int ai_addrlen;               /* Length of socket address.  */
  struct sockaddr *ai_addr;     /* Socket address for socket.  */
  char *ai_canonname;           /* Canonical name for service location.  */
  struct addrinfo *ai_next;     /* Pointer to next in list.  */
};


// #define dapls_osd_getaddrinfo(name, addr_ptr) getaddrinfo(name,NULL,NULL,addr_ptr)
// #define dapls_osd_freeaddrinfo(addr) freeaddrinfo (addr)
int dapls_osd_getaddrinfo(const char *name, struct addrinfo **addr_ptr);
void dapls_osd_freeaddrinfo(struct addrinfo *addr);

#endif /* IBHOSTS_NAMING */

/*
 * *printf format helpers. We use the C string constant concatenation
 * ability to define 64 bit formats, which unfortunatly are non standard
 * in the C compiler world. E.g. %llx for gcc, %I64x for Windows
 */
#ifdef __x86_64__
#define F64d   "%ld"
#define F64u   "%lu"
#define F64x   "%lx"
#define F64X   "%lX"
#else
#define F64d   "%lld"
#define F64u   "%llu"
#define F64x   "%llx"
#define F64X   "%llX"
#endif

/*
 *  Conversion Functions
 */

STATIC _INLINE_ long
dapl_os_strtol (const char *nptr, char **endptr, int base)
{
    return simple_strtol (nptr, endptr, base);
}

/*
 * helper routines
 */

/*
 * dapl_os_mod64
 *
 * Returne the modulo of a 64 bit number. Given that this is running
 * on a 32 bit platform, we need to use the kernel macro to prevent
 * exceptions from killing the machine.
 */
STATIC _INLINE_ long
dapl_os_mod64(uint64_t	key, unsigned long hashsize)
{
#ifdef __ia64__
    return ((uint64_t)((key) % (hashsize)));
#else
    return do_div(key, hashsize);
#endif
}


/*
 *  Debug Helper Functions
 */

#define dapl_os_assert(expression)                      \
        if (!(expression)) {                            \
             panic("ASSERTION fail in %s:%i:%s\n", __FILE__, __LINE__, __func__); \
        }
#define dapl_os_printf 			printk


#if defined(DAPL_DBG)

void dapl_os_dbg_print (
	DAPL_DBG_TYPE		type,
	const char *		fmt,
	...);

#else  /* !DBG */
#define dapl_os_dbg_print(...)
#endif /* !DBG */

void dapl_os_vprintf (const char *		fmt,
		      va_list			 args);

#define dapl_os_syslog(fmt, args)	/* do nothing, printk will log */

#define MACRO_START     do{
#ifdef lint
#define MACRO_END       }while (fooLint ())
#else
#define MACRO_END       }while (0)
#endif


#endif /*  _DAPL_OSD_H_ */

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */
