/*
 * Copyright (C) Voltaire Ltd. 2001-2005.  ALL RIGHTS RESERVED.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * Maintained by: vtrmaint1@voltaire.com
 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */
 /**********************************************************************
 * MODULE: dapl_mem_fmr.c                                 
 *
 * PURPOSE: Memory management support routines for FMR       
 * Description: This file contains the implementaion of all  
 *              FMR functionuality.                          
 *
 * $Id: dapl_mem_fmr.c,v 1.1.2.17 2005/02/15 14:49:11 guyg Exp $
 **********************************************************************/

/*******************
 * kernel include  *
 *******************/
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/list.h>
#include <linux/spinlock.h>
#include <linux/ctype.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/moduleparam.h>
#endif

/********************
 * voltaire include *
 ********************/
#include "dapl_mem_fmr.h"
#include "dapl.h"

/***********************************
 * forward function declarations   *
 ***********************************/
mem_err_t print_pool_entry(fmr_pool_entry  *p_pool_entry); // for debug
mem_err_t print_fmr_hndl(fmr_hndl_impl   *p_fmr_hndl);     // for debug
void mem_fmr_print_all(fmr_ctrl_struct      *mem_fmr_ctrl);
void mem_fmr_show_param (fmr_ctrl_struct    *mem_fmr_ctrl);
void mem_show_fmr_pool(char *func, fmr_ctrl_struct *mem_fmr_ctrl);


/***********************************
 * private definitions of mem_fmr  *
 ***********************************/

/* define memory locks */
#define MEM_LOCK(x) do { \
spin_lock(&x);\
} while(0)

#define MEM_UNLOCK(x) do {\
spin_unlock(&x);\
} while(0)

#define MEM_LOCK_FL(x,y) do { \
spin_lock_irqsave(&x,y);\
} while(0)

#define MEM_UNLOCK_FL(x,y) do { \
spin_unlock_irqrestore(&x,y);\
} while(0)
/* end of memory locks */


// module parameters
int num_of_fmr_hndls = FMR_DEFUALT_NUMBER_OF_HANDLES;
int max_unmap_at_once = FMR_DEFUALT_MAX_UNMAP_AT_ONCE;
int fmr_water_mark = FMR_DEFUALT_WATER_MARK;
int avg_recv_size = 0  ;
int avg_recv_size_percent = 0;          // relative size of total allocations
int active_fmr = 0;                  // the FMR mechanism is not set by default
int g_unmap_cnt = 0;                    // counts the unmap operation that has been made


// the handlers pool allocation will be  built according to one of the following tables
// (default values will be taken unless user states otherwise):
//
//       default values:                   user defined values
//       ===============                   ======================
//  mapped_pages |   % of resource         mapped_pages |   % of resource
//  -------------|----------------         -------------|----------------
//                                          user value  |  50%
//        1      |  20%                          1      |  10%
//        4      |  30%                          4      |  15%
//        16     |  20%                          16     |  10%
//        64     |  20%                          64     |  10%
//------------------------------        -------------------------------

#define    DEFAULT_VALUES               0
#define    USER_VALUES                  1
#define    INITIATIVE                   2   // either user or default values are set
#define    DEFAULT_AVG_RECV_SIZE_PERC   50


int        alloc_fmr_resource_tbl[FMR_MAX_OPTIONS_TO_MAP][INITIATIVE]= { [1] = {20,10},
                                                                         [4] = {30,15},
                                                                         [16] = {30,15},
                                                                         [FMR_MAX_OPTIONS_TO_MAP-1] = {20,10}};

/************************************
 * END of internal FMR structures   *
 ************************************/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
module_param(avg_recv_size, int, 0444);
module_param(avg_recv_size_percent, int, 0444);
module_param(active_fmr, int, 0444);
module_param(num_of_fmr_hndls, int, 0444);
module_param(max_unmap_at_once, int, 0444);
module_param(fmr_water_mark, int, 0444);
#else
MODULE_PARM(avg_recv_size, "i");
MODULE_PARM(avg_recv_size_percent, "i");
MODULE_PARM(active_fmr, "i");
MODULE_PARM(num_of_fmr_hndls, "i");
MODULE_PARM(max_unmap_at_once, "i");
MODULE_PARM(fmr_water_mark, "i");
#endif

MODULE_PARM_DESC(avg_recv_size," Average size of recieved data");
MODULE_PARM_DESC(avg_recv_size_percent," Reletive amount of allocation");
MODULE_PARM_DESC(active_fmr," Create an fmr pool at pz_create or not (default=0 - off)");
MODULE_PARM_DESC(num_of_fmr_hndls, " Number of FMR handlers to be created (default=2000)");
MODULE_PARM_DESC(max_unmap_at_once, " Maximum number of handler that can be unmapped at once (default=2000)");
MODULE_PARM_DESC(fmr_water_mark," The water mark of fmr handlers (default= 10% of total handlers)");

mem_err_t init_mem_fmr_ctrl(fmr_ctrl_struct *mem_fmr_ctrl)
{
	int i;

	for (i=0 ; i < FMR_MAX_OPTIONS_TO_MAP; ++i)
	{
		mem_fmr_ctrl->fmr_pool[i].allocted_pages = 0;
		mem_fmr_ctrl->fmr_pool[i].num_free_hndls = 0;
		mem_fmr_ctrl->fmr_pool[i].total_hndls = 0;
		spin_lock_init(&(mem_fmr_ctrl->fmr_pool[i].lock));
	}

	mem_fmr_ctrl->unmap_vector.idx = 0;
	mem_fmr_ctrl->unmap_vector.urgent_flag = FMR_UNMAP_URGENT_FLAG_OFF;
	mem_fmr_ctrl->unmap_vector.max_unmap_at_once = max_unmap_at_once;
	mem_fmr_ctrl->current_fmr_num = 0;
	mem_fmr_ctrl->max_fmr_num = 0;                           // this value will be set on the first init function.
	mem_fmr_ctrl->max_outstanding_maps= 0;                   // this value will be set on the first init function.

	dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: max map at once:%d\n", 
                  __FUNCTION__, max_unmap_at_once);
    
    /* init locks */
	spin_lock_init(&mem_fmr_ctrl->unmap_vector.hndl_vec_lock);
	spin_lock_init(&mem_fmr_ctrl->unmap_vector.urgent_flag_lock);

	return MEM_RET_SUCCESS;
}

mem_err_t mem_fmr_create_pool_entry(IN DAPL_PZ		        *pz,
                                    IN VAPI_mrw_acl_t       acl_flags,
									IN u_int8_t             log2_page_size,
									IN int                  num_pages,        // argument to the module
									IN int                  num_of_fmr_hndl,
                                    IN u_int32_t            max_outstanding_maps)
{                                                        
	fmr_pool_entry      *p_fmr_pool_entry = NULL;
	fmr_hndl_impl       *p_fmr_hndl_impl  = NULL;
	EVAPI_fmr_t         *fmr_props = NULL;
	VAPI_ret_t          retcode;
	VAPI_hca_cap_t      *hca_attr = NULL;
	int                 i;
	mem_err_t           err = MEM_RET_SUCCESS;
    VAPI_hca_vendor_t   *hca_vendor = NULL;
    VAPI_hca_hndl_t     hca_hdnl;
    VAPI_pd_hndl_t      pd_hndl;
    fmr_ctrl_struct     *mem_fmr_ctrl;
    
    pd_hndl = pz->pd_handle;
    hca_hdnl = pz->header.owner_ia->hca_ptr->ib_hca_handle;
    mem_fmr_ctrl = &pz->mem_fmr_ctrl;

	if (num_of_fmr_hndl == 0)
		goto endl;
    
    if (num_pages > FMR_MAX_OPTIONS_TO_MAP)
	{
        dapl_dbg_log (DAPL_DBG_TYPE_ERR, "%s: num pages:%d, is greater than allowed (%d)\n",
					  __FUNCTION__, num_pages, FMR_MAX_OPTIONS_TO_MAP);

		err = MEM_RET_INVALID_ARG;
		goto endl;
	}
    hca_vendor = dapl_os_alloc (sizeof(VAPI_hca_vendor_t));
    if ( hca_vendor == NULL )
    {
        err = MEM_RET_FMR_LOWER_LEVEL_ERROR;
    goto endl;
    }
    memset(hca_vendor,0,sizeof(hca_vendor));
    hca_attr = dapl_os_alloc (sizeof(VAPI_hca_cap_t));
    if ( hca_attr == NULL )
    {
		err = MEM_RET_FMR_LOWER_LEVEL_ERROR;
    goto endl;
    }
    memset(hca_attr,0,sizeof(hca_attr));
	
    retcode = VAPI_query_hca_cap(hca_hdnl, hca_vendor, hca_attr);
	
    if (retcode != VAPI_OK)
	{
		dapl_dbg_log (DAPL_DBG_TYPE_ERR, "%s: hca_query failed!",
                      __FUNCTION__);
		err = MEM_RET_FMR_LOWER_LEVEL_ERROR;
		goto endl;
	}

	dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: max_fmr_num:%d, max_maps_per_fmr:%d\n",
                  __FUNCTION__, hca_attr->max_num_fmr, hca_attr->max_num_map_per_fmr);

	mem_fmr_ctrl->max_fmr_num  = hca_attr->max_num_fmr;
	mem_fmr_ctrl->max_outstanding_maps = hca_attr->max_num_map_per_fmr;//max_outstanding_maps; // GG> ignore sw defaults - take hca attributes
    
    /* create and initialize fmr_hndl pool from page size given */
	p_fmr_pool_entry = &(mem_fmr_ctrl->fmr_pool[num_pages]); 
	p_fmr_pool_entry->allocted_pages = num_pages;
	p_fmr_pool_entry->total_hndls=num_of_fmr_hndl;
	p_fmr_pool_entry->num_free_hndls =0;
	p_fmr_pool_entry->min_free_hndl_needed = (fmr_water_mark * num_of_fmr_hndl )/ 100; //currently the watermark is 10% of total allocations

	dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: fmr_water_mark(%d) * num_of_fmr_hndl(%d)\n",
					  __FUNCTION__, (int)fmr_water_mark,num_of_fmr_hndl);
    
	INIT_LIST_HEAD(&(p_fmr_pool_entry->free_fmr_hndl_list));
	INIT_LIST_HEAD(&(p_fmr_pool_entry->used_fmr_hndl_list));
    
    /* create free fmr handlers */
	for (i=0 ; i < num_of_fmr_hndl; ++i)
	{
        // make sure we don't exceed number of FMR allocation allowed
		if (mem_fmr_ctrl->current_fmr_num == mem_fmr_ctrl->max_fmr_num)
		{
			dapl_dbg_log (DAPL_DBG_TYPE_ERR, "%s: ERROR. run out of fmr allocations!\n",
                          __FUNCTION__);
			err = MEM_RET_FMR_OUT_OF_RESOURCE;
			goto endl;
		}

        p_fmr_hndl_impl = kmalloc(sizeof(fmr_hndl_impl),GFP_KERNEL);
        
        /* dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: allocated fmr handler:%p\n",
                      __FUNCTION__, p_fmr_hndl_impl); */

		p_fmr_hndl_impl->hca_hdnl = hca_hdnl;
		p_fmr_hndl_impl->map_count = mem_fmr_ctrl->max_outstanding_maps;
		p_fmr_hndl_impl->mapped_pages = num_pages;
        fmr_props = dapl_os_alloc (sizeof(EVAPI_fmr_t));
        if ( fmr_props == NULL )
        {
            err = MEM_RET_FMR_LOWER_LEVEL_ERROR;
        goto endl;
        }
        fmr_props->pd_hndl = pd_hndl;

		ACCESS_CONTROL_MASK_CLR_ALL(fmr_props->acl);
		if (acl_flags & VAPI_EN_MEMREG_BIND)
		{
            // the flags can not include bind.
			dapl_dbg_log (DAPL_DBG_TYPE_ERR, "%s: ERROR. acl can not include bind flag\n",
                          __FUNCTION__);
			err =  MEM_RET_INVALID_ARG;
			goto endl;
		}
		if (acl_flags & VAPI_EN_LOCAL_WRITE)
			ACCESS_CONTROL_MASK_SET (fmr_props->acl, VAPI_EN_LOCAL_WRITE);
		if (acl_flags & VAPI_EN_REMOTE_WRITE)
			ACCESS_CONTROL_MASK_SET (fmr_props->acl, VAPI_EN_REMOTE_WRITE);
		if (acl_flags & VAPI_EN_REMOTE_READ)
			ACCESS_CONTROL_MASK_SET(fmr_props->acl, VAPI_EN_REMOTE_READ);

		fmr_props->max_pages= num_pages;
		fmr_props->log2_page_sz = log2_page_size;
		fmr_props->max_outstanding_maps = mem_fmr_ctrl->max_outstanding_maps;

        retcode = EVAPI_alloc_fmr(hca_hdnl, fmr_props, 
                                  &(p_fmr_hndl_impl->fmr_hndl));
        
        if (retcode != VAPI_OK )
		{
			dapl_dbg_log (DAPL_DBG_TYPE_ERR, "%s: ERROR. The alloc fmr operation failed.\n",
                          __FUNCTION__);
			err =  MEM_RET_FMR_LOWER_LEVEL_ERROR;
			goto endl;
		}

        /* add the new fmr_handle to free list of fmr_handlers  */
		list_add(&(p_fmr_hndl_impl->list),&(p_fmr_pool_entry->free_fmr_hndl_list));
		++p_fmr_pool_entry->num_free_hndls;
		++mem_fmr_ctrl->current_fmr_num;
	} /* end of for on num_of_fmr_hndl */

    dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: current_fmr_num=%d num_free_hndls=%d p_fmr_pool_entry=%p fmr_hndl_impl=%p\n",
                  __FUNCTION__, mem_fmr_ctrl->current_fmr_num, 
                  p_fmr_pool_entry->num_free_hndls, p_fmr_pool_entry, p_fmr_hndl_impl);
    if (hca_attr) {
        dapl_os_free(hca_attr,0);
    }
    if (fmr_props) {
        dapl_os_free(fmr_props,0);
    }
    if (hca_vendor) {
        dapl_os_free(hca_vendor,0);
    }
    return err;
	endl:
    if (hca_attr) {
        dapl_os_free(hca_attr,0);
    }
    if (fmr_props) {
        dapl_os_free(fmr_props,0);
    }
    if (hca_vendor) {
        dapl_os_free(hca_vendor,0);
    }
	if (p_fmr_hndl_impl)
		kfree (p_fmr_hndl_impl);
	return err;
}  



mem_err_t  
mem_fmr_get_fmr(IN      int                 num_pages,
                IN      fmr_ctrl_struct     *mem_fmr_ctrl,
				OUT     mem_fmr_hndl_t      *mem_fmr_hndl)
{
	fmr_pool_entry  *p_fmr_pool_entry = NULL;
	fmr_hndl_impl   *p_fmr_hndl;
	mem_err_t       ret = MEM_RET_SUCCESS;
	int             i=0;
    
    mem_show_fmr_pool("mem_fmr_get_fmr - begin", mem_fmr_ctrl);
	if (num_pages > FMR_MAX_OPTIONS_TO_MAP)
	{
		dapl_dbg_log (DAPL_DBG_TYPE_ERR, "%s: number of pages:%d is greater than allowed(%d)",
                      __FUNCTION__, num_pages, FMR_MAX_OPTIONS_TO_MAP);
		ret = MEM_RET_INVALID_ARG;
		goto endl;
	}

    p_fmr_pool_entry = &(mem_fmr_ctrl->fmr_pool[num_pages]);

    // validation checks  
	MEM_LOCK(p_fmr_pool_entry->lock);

	if (!p_fmr_pool_entry->num_free_hndls)
	{
        // find another entry with greater number of pages
		MEM_UNLOCK(p_fmr_pool_entry->lock);
		for (i= num_pages+1; i < FMR_MAX_OPTIONS_TO_MAP; ++i)
		{
			p_fmr_pool_entry = &(mem_fmr_ctrl->fmr_pool[i]);
            MEM_LOCK(p_fmr_pool_entry->lock);
            if (p_fmr_pool_entry->num_free_hndls)
			{
                // find fmr pool entry
				break;
			}
			MEM_UNLOCK(p_fmr_pool_entry->lock);
            p_fmr_pool_entry = NULL;

		}
		if (!p_fmr_pool_entry)
		{
            // no entry is found in pool that matches the number of pages required
			dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: no available handlers found\n",
                          __FUNCTION__);
			ret = MEM_RET_FMR_OUT_OF_RESOURCE;
			goto endl;
		}
	}

    //  do the following:
    //  1. take the first element in the free list
    //  2. delete that element from free handles list
    //  3. add the element to the used handles list
    //  4. update number of free handlers left.
    //  4. return handler to user


	p_fmr_hndl = list_entry(p_fmr_pool_entry->free_fmr_hndl_list.next,
							fmr_hndl_impl,
							list);

	list_del(&(p_fmr_hndl->list));
	--p_fmr_pool_entry->num_free_hndls ;
        
	if (p_fmr_pool_entry->num_free_hndls <=  p_fmr_pool_entry->min_free_hndl_needed)
	{
        // set urgent unmap flag on
		dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: Turning on the urgent flag. current free=%d, min needed=%d\n",
                      __FUNCTION__, 
					  p_fmr_pool_entry->num_free_hndls,
					  p_fmr_pool_entry->min_free_hndl_needed);

		MEM_LOCK(mem_fmr_ctrl->unmap_vector.urgent_flag_lock);
		mem_fmr_ctrl->unmap_vector.urgent_flag = FMR_UNMAP_URGENT_FLAG_ON;
		MEM_UNLOCK(mem_fmr_ctrl->unmap_vector.urgent_flag_lock);
    }

	list_add(&p_fmr_hndl->list,&p_fmr_pool_entry->used_fmr_hndl_list);

	MEM_UNLOCK(p_fmr_pool_entry->lock);
    
	*mem_fmr_hndl = p_fmr_hndl;
    
#ifdef DEBUG
	print_fmr_hndl(p_fmr_hndl);
	mem_fmr_print_all(mem_fmr_ctrl);
#endif /*DEBUG*/ 

	endl:
        mem_show_fmr_pool("mem_fmr_get_fmr - end", mem_fmr_ctrl);
	    return ret;
}


mem_err_t
mem_fmr_free_fmr (IN     int                num_to_free,
				  IN     mem_fmr_hndl_t     fmr_hndl_arr,
                  IN     fmr_ctrl_struct    *mem_fmr_ctrl,
				  OUT    int                *num_success_free)
{
	int             i, j, index;
	int             total_free =  0;
	fmr_hndl_impl   *p_fmr_hndl;
	fmr_pool_entry  *p_fmr_pool_entry = NULL;
	mem_err_t       ret = MEM_RET_SUCCESS;
	VAPI_ret_t      retcode;
	int             num_to_unmap;
    
    mem_show_fmr_pool("mem_fmr_free_fmr - begin", mem_fmr_ctrl);
	for (i=0; i< num_to_free; ++i)
	{
		p_fmr_hndl = (fmr_hndl_arr + i);

        // validation check
		if (!p_fmr_hndl)
		{
			dapl_dbg_log (DAPL_DBG_TYPE_ERR, "%s: Validation check failed: no matching fmr handler found\n",
                          __FUNCTION__);
			ret = MEM_RET_INVALID_ARG;
			goto endl;
		}

		if (p_fmr_hndl->mapped_pages > FMR_MAX_OPTIONS_TO_MAP)
		{
            // no entry is found in pool that matches the number of pages required
			dapl_dbg_log (DAPL_DBG_TYPE_ERR, "%s: Validation check failed: no matching allocated pages found\n",
                          __FUNCTION__);
			ret = MEM_RET_INVALID_ARG;
			goto endl;
		}

        p_fmr_pool_entry = &(mem_fmr_ctrl->fmr_pool[p_fmr_hndl->mapped_pages]);

		dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: mapped_pages:%d fmr_hndl:%p, map_count:%d \n",
                      __FUNCTION__, p_fmr_hndl->mapped_pages,
					  p_fmr_hndl, p_fmr_hndl->map_count); 
		
        dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: unmap details: total_hndls:%d, idx:%d, max_map_at_once:%d \n",
                      __FUNCTION__, p_fmr_pool_entry->total_hndls, 
					  mem_fmr_ctrl->unmap_vector.idx,
                      max_unmap_at_once); 

        
        // do the following:
        // 1. remove element from used FMR list.
        // 2. if map_count =0, add fmr handler to need_to_unmap vector. 
        //    if the vector is full, perform unmap and add handlers to the head of the free FMR list
        // 3. else, return handler to the tail of the list.

		MEM_LOCK(p_fmr_pool_entry->lock);
		list_del(&p_fmr_hndl->list);//removing element from used FMR list.

        if (p_fmr_hndl->map_count > 0)
		{
			dapl_dbg_log (DAPL_DBG_TYPE_FMR, 
                          "%s: since map count is not zero, p_fmr_pool_entry=%p num_free_hndls=%d allocted_pages=%d\n",
                          __FUNCTION__, p_fmr_pool_entry, 
                          p_fmr_pool_entry->num_free_hndls, p_fmr_pool_entry->allocted_pages);
            
            list_add_tail(&p_fmr_hndl->list,&p_fmr_pool_entry->free_fmr_hndl_list);
			++p_fmr_pool_entry->num_free_hndls;
			MEM_UNLOCK(p_fmr_pool_entry->lock);
        }
		else
		{
            MEM_UNLOCK(p_fmr_pool_entry->lock);
            
            dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: need_to_unmap_idx=%d, p_fmr_hndl=%p fmr_hndl=%p\n",
                          __FUNCTION__, 
						  mem_fmr_ctrl->unmap_vector.idx,
						  p_fmr_hndl, p_fmr_hndl->fmr_hndl);

			MEM_LOCK(mem_fmr_ctrl->unmap_vector.hndl_vec_lock);
			index = mem_fmr_ctrl->unmap_vector.idx;
			mem_fmr_ctrl->unmap_vector.fmr_hndl[index] = p_fmr_hndl;
			mem_fmr_ctrl->unmap_vector.v_hndl[index] = p_fmr_hndl->fmr_hndl;
			++ mem_fmr_ctrl->unmap_vector.idx;
            
			MEM_LOCK(mem_fmr_ctrl->unmap_vector.urgent_flag_lock);

			if (mem_fmr_ctrl->unmap_vector.idx == mem_fmr_ctrl->unmap_vector.max_unmap_at_once ||
				mem_fmr_ctrl->unmap_vector.urgent_flag == FMR_UNMAP_URGENT_FLAG_ON)
			{
				MEM_UNLOCK(mem_fmr_ctrl->unmap_vector.urgent_flag_lock);
                // unmap all waiting handler.
				num_to_unmap = mem_fmr_ctrl->unmap_vector.idx;
                
				retcode = EVAPI_unmap_fmr(p_fmr_hndl->hca_hdnl,
                                          num_to_unmap,
                                          mem_fmr_ctrl->unmap_vector.v_hndl);

				MEM_UNLOCK(mem_fmr_ctrl->unmap_vector.hndl_vec_lock);

				if (retcode != VAPI_OK)
				{
					dapl_dbg_log (DAPL_DBG_TYPE_ERR, "%s: ERROR, unmap_fmr failed!",
                                  __FUNCTION__);
					ret = MEM_RET_FMR_LOWER_LEVEL_ERROR;
					goto endl; // merav. 1.note the error. 2.erase?.
				}

				++ g_unmap_cnt;

				MEM_LOCK(mem_fmr_ctrl->unmap_vector.urgent_flag_lock);
				mem_fmr_ctrl->unmap_vector.urgent_flag = FMR_UNMAP_URGENT_FLAG_OFF;
				MEM_UNLOCK(mem_fmr_ctrl->unmap_vector.urgent_flag_lock);

				dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: successful unmap of %d handlers.\n",
                              __FUNCTION__, num_to_unmap);

                // if we got here it means that all waiting handlers were successfully unmapped. 
                // now we need to return all handlers to the pool for reuse.
				for (j = 0;j < num_to_unmap; ++j)
				{
					fmr_hndl_impl*  p_fmr_hndl =  mem_fmr_ctrl->unmap_vector.fmr_hndl[j];
					fmr_pool_entry*   p_fmr_pool_entry =  &mem_fmr_ctrl->fmr_pool[p_fmr_hndl->mapped_pages];

					dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: returning the handler:%p to free fmr pool after unmap operation ",
                                  __FUNCTION__, p_fmr_hndl);

#ifdef DEBUG         
					print_fmr_hndl(p_fmr_hndl);
#endif /*DEBUG*/

					p_fmr_hndl->map_count = mem_fmr_ctrl->max_outstanding_maps;

					MEM_LOCK(p_fmr_pool_entry->lock);
					list_add(&p_fmr_hndl->list,&p_fmr_pool_entry->free_fmr_hndl_list);
					++p_fmr_pool_entry->num_free_hndls;
					MEM_UNLOCK(p_fmr_pool_entry->lock);

					MEM_LOCK(mem_fmr_ctrl->unmap_vector.hndl_vec_lock);
					--mem_fmr_ctrl->unmap_vector.idx;
					mem_fmr_ctrl->unmap_vector.fmr_hndl[j] = 0;
					MEM_UNLOCK(mem_fmr_ctrl->unmap_vector.hndl_vec_lock);

				}


			}/* end of if idx == MAX_UNMAP_AT_ONCE */
			MEM_UNLOCK(mem_fmr_ctrl->unmap_vector.urgent_flag_lock);
			MEM_UNLOCK(mem_fmr_ctrl->unmap_vector.hndl_vec_lock);

		} /* end of else */
        ++total_free;
	} /* end of for */
    mem_show_fmr_pool("mem_fmr_free_fmr - end", mem_fmr_ctrl);
#ifdef DEBUG
	mem_fmr_print_all(mem_fmr_ctrl);
#endif /*DEBUG*/

	endl: 
	*num_success_free = total_free;
	return ret;
}


mem_err_t
mem_fmr_map_fmr (IN  mem_fmr_hndl_t     mem_fmr_hndl,
                 OUT VAPI_lkey_t        *l_key,
                 OUT VAPI_rkey_t        *r_key)
{
	fmr_hndl_impl   *fmr_hndl = (fmr_hndl_impl*)mem_fmr_hndl;
	VAPI_ret_t      retcode;
    EVAPI_fmr_map_t map_props;
    mem_err_t       ret = MEM_RET_SUCCESS;

#ifdef DEBUG  
	dapl_dbg_log (DAPL_DBG_TYPE_FMR, "FMR: going to map the follwoing handler:\n");
	print_fmr_hndl(fmr_hndl);
#endif /*DEBUG*/

    //mem_show_fmr_pool("mem_fmr_map_fmr - begin", mem_fmr_ctrl);
	if (!fmr_hndl->map_count)
	{
        // you have reached maximum outstanding maps without doing unmap (unmap is 
        // being performed automatically by using the mem_fmr_free_fmr function).  
		dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
                      "%s: ERROR. map count=0. maximum maps were done without unmap!\n",
                      __FUNCTION__);

		ret =  MEM_RET_INVALID_ARG;
		goto endl;
	}

    map_props.start = fmr_hndl->page_arr[0];
	map_props.size = fmr_hndl->data_size;
	map_props.page_array_len = fmr_hndl->mapped_pages_req;
	map_props.page_array = fmr_hndl->page_arr;
    
	retcode = EVAPI_map_fmr(fmr_hndl->hca_hdnl,
                            fmr_hndl->fmr_hndl,
                            &map_props,
                            l_key,
                            r_key);
    
    if (retcode != VAPI_OK)
	{
		dapl_dbg_log (DAPL_DBG_TYPE_ERR, "%s: map_fmr operation failed (error=0x%x)\n",
                      __FUNCTION__, retcode);

		ret =  MEM_RET_FMR_LOWER_LEVEL_ERROR;
		goto endl;
	}

	--fmr_hndl->map_count;
	dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: current map count = %d\n",
                  __FUNCTION__, fmr_hndl->map_count);
	dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: lkey=0x%x, rkey=0x%x\n", 
                  __FUNCTION__, *l_key, *r_key);
    
    endl: 
        //mem_show_fmr_pool("mem_fmr_map_fmr - end", mem_fmr_ctrl);
        return ret;
}



mem_err_t 
mem_fmr_cleanup (IN   fmr_ctrl_struct     *mem_fmr_ctrl)
{

	struct list_head *fmr_hndl_iter;

	fmr_pool_entry  *p_fmr_pool_entry = NULL;
	fmr_hndl_impl   *p_fmr_hndl= NULL;
	VAPI_ret_t      ret;
    mem_err_t       retcode;
	int             i, f_counter;
    
    ret = VAPI_OK;
    retcode = MEM_RET_SUCCESS;
    
    dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: Starting to free all memory allocated (MAX_MAPED=%d) \n",
                  __FUNCTION__, FMR_MAX_OPTIONS_TO_MAP);
    
    mem_show_fmr_pool("mem_fmr_cleanup - begin", mem_fmr_ctrl);

	for (i = 0; i<FMR_MAX_OPTIONS_TO_MAP ; ++i )
	{
		p_fmr_pool_entry = &(mem_fmr_ctrl->fmr_pool[i]);

		if (!(p_fmr_pool_entry->allocted_pages))
		{
			continue;
		}

#ifdef DEBUG    
		print_pool_entry(p_fmr_pool_entry);
#endif /* DEBUG */
        
        MEM_LOCK(p_fmr_pool_entry->lock);
		f_counter = 0;
        // delete all elements in free FMR handlers list
        list_for_each(fmr_hndl_iter, &(p_fmr_pool_entry->free_fmr_hndl_list))
		{
			p_fmr_hndl = list_entry(fmr_hndl_iter, fmr_hndl_impl ,list);

			fmr_hndl_iter = fmr_hndl_iter->prev;

			ret = EVAPI_free_fmr(p_fmr_hndl->hca_hdnl,
                                     p_fmr_hndl->fmr_hndl);
            if (ret != VAPI_OK)
            {
                retcode = MEM_RET_FMR_LOWER_LEVEL_ERROR;
                dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: in free handles list - EVAPI_free_fmr error = 0x%x on handle %p\n",
                          __FUNCTION__, ret, p_fmr_hndl->fmr_hndl);
            }
            else
                f_counter++;
			list_del(&p_fmr_hndl->list);
            p_fmr_pool_entry->num_free_hndls--; // GG
			kfree(p_fmr_hndl);
		} /* end of loop on free fmr handlers list. */

        dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: [%d] in free handles list f_counter=%d\n",
                      __FUNCTION__, i, f_counter);
        
        // delete all elements in used FMR handlers list
        f_counter = 0;
        list_for_each(fmr_hndl_iter, &(p_fmr_pool_entry->used_fmr_hndl_list))
		{
			p_fmr_hndl = list_entry(fmr_hndl_iter,  fmr_hndl_impl ,list );
			fmr_hndl_iter=fmr_hndl_iter->prev; // otherwise, list_for_each will fall

			ret = EVAPI_free_fmr(p_fmr_hndl->hca_hdnl,
                                     p_fmr_hndl->fmr_hndl);
            if (ret != VAPI_OK)
            {
                retcode = MEM_RET_FMR_LOWER_LEVEL_ERROR;
                dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: in used handles list - EVAPI_free_fmr error = 0x%x on handle %p\n",
                          __FUNCTION__, ret, p_fmr_hndl->fmr_hndl);
            }
            else
                f_counter++;

			dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: freeing fmr handle:%p\n",
                          __FUNCTION__, p_fmr_hndl);
			list_del(&p_fmr_hndl->list);
			kfree(p_fmr_hndl);
		} /* end of loop on used fmr handlers list. */
		MEM_UNLOCK(p_fmr_pool_entry->lock);
        dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: [%d] in used handles list f_counter=%d\n",
                      __FUNCTION__, i, f_counter);
	} /* end of loop on fmr pool */

    dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: now releasing all handlers in need to unmap vector (index=%d):\n",
                  __FUNCTION__, mem_fmr_ctrl->unmap_vector.idx);

	/* release all handlers that wait on need_to_unmap vector */
    MEM_LOCK(mem_fmr_ctrl->unmap_vector.hndl_vec_lock);
    f_counter = 0;
    for (i = 0; i <  mem_fmr_ctrl->unmap_vector.idx ; ++i)
	{
		p_fmr_hndl = mem_fmr_ctrl->unmap_vector.fmr_hndl[i];
		dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: going to free handler:%p ,location:%d\n",
                      __FUNCTION__, p_fmr_hndl, i);

		ret = EVAPI_free_fmr(p_fmr_hndl->hca_hdnl,
                                 p_fmr_hndl->fmr_hndl); 
        if (retcode != VAPI_OK)
        {
            retcode = MEM_RET_FMR_LOWER_LEVEL_ERROR;
            dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: in unmap_vector - EVAPI_free_fmr error = 0x%x on handle %p\n",
                      __FUNCTION__, ret, p_fmr_hndl->fmr_hndl);
        }
        else
            f_counter++;

		kfree(p_fmr_hndl);
	}
    dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: [%d] in unmap_vector f_counter=%d\n",
                  __FUNCTION__, i, f_counter);
	MEM_UNLOCK(mem_fmr_ctrl->unmap_vector.hndl_vec_lock);
    mem_show_fmr_pool("mem_fmr_cleanup - end", mem_fmr_ctrl);


    return retcode;
}


mem_err_t
print_pool_entry (fmr_pool_entry  *p_pool_entry)
{
	MEM_LOCK(p_pool_entry->lock);

	dapl_dbg_log (DAPL_DBG_TYPE_FMR, "FMR: Pool Entry:%p,of pages:%d, free handles:%d(total: %d), min fmr needed:%d\n",
		   p_pool_entry,
		   p_pool_entry->allocted_pages,
		   p_pool_entry->num_free_hndls,
		   p_pool_entry->total_hndls,
		   p_pool_entry->min_free_hndl_needed);

	MEM_UNLOCK(p_pool_entry->lock);

	return MEM_RET_SUCCESS;
}


mem_err_t 
print_fmr_hndl (fmr_hndl_impl   *p_fmr_hndl)
{

	dapl_dbg_log (DAPL_DBG_TYPE_FMR, "FMR: ptr=%p  fmr handle=%d  pages=%d  map_count=%d\n",
           p_fmr_hndl, 
           (int)(p_fmr_hndl->fmr_hndl),
           p_fmr_hndl->mapped_pages,
           p_fmr_hndl->map_count);  
	
	return MEM_RET_SUCCESS;
}


/* This function is only for debug use */
void 
mem_fmr_print_all (fmr_ctrl_struct    *mem_fmr_ctrl)
{

	struct list_head *fmr_hndl_iter;

	fmr_pool_entry  *p_fmr_pool_entry = NULL;
	fmr_hndl_impl   *p_fmr_hndl = NULL;
	int             i;

	dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: Starting to print all FMR handles available\n",
                  __FUNCTION__);

	for (i=0; i< FMR_MAX_OPTIONS_TO_MAP; ++i)
	{

		p_fmr_pool_entry = &(mem_fmr_ctrl->fmr_pool[i]);

		if (!(p_fmr_pool_entry->allocted_pages))
		{
			continue;
		}

		dapl_dbg_log (DAPL_DBG_TYPE_FMR, "------------------------------------\n");

		print_pool_entry(p_fmr_pool_entry);


        // print all elements in free FMR handles list
		list_for_each(fmr_hndl_iter, &(p_fmr_pool_entry->free_fmr_hndl_list))
		{
			p_fmr_hndl = list_entry(fmr_hndl_iter,  fmr_hndl_impl ,list );
			dapl_dbg_log (DAPL_DBG_TYPE_FMR, "FMR: free handles:\n");
        }/*end of loop on free fmr handles list.*/

        //print all elements in used FMR handles list
		list_for_each(fmr_hndl_iter, &(p_fmr_pool_entry->used_fmr_hndl_list))
		{
			p_fmr_hndl = list_entry(fmr_hndl_iter,  fmr_hndl_impl ,list );
			dapl_dbg_log (DAPL_DBG_TYPE_FMR, "FMR: used handles:\n");
            print_fmr_hndl(p_fmr_hndl);       
		}/*end of loop on used fmr handles list.*/

    } /*end of loop on fmr pool*/

}


mem_err_t mem_fmr_init_module (IN void  *pz_h)
{
	mem_err_t               ret;
    IN VAPI_mrw_acl_t       acl_flags;
	int                     i, j, pages_mapped = 0;
	int                     num_of_hndls;
    u_int32_t               max_outstanding_maps;
    VAPI_pd_hndl_t          pd_hndl;
    VAPI_hca_hndl_t         hca_hdnl;
    DAPL_PZ                 *pz;

    pz = (DAPL_PZ *)pz_h;
    
    pd_hndl = pz->pd_handle;
    hca_hdnl = pz->header.owner_ia->hca_ptr->ib_hca_handle;

    max_outstanding_maps = FMR_DEFUALT_MAX_OUTSTANDING_MAPS; 	// == hca_attr.max_maps_per_fmr

    dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: initial values:  num_of_fmr_hndls=%d\n",
				  __FUNCTION__, num_of_fmr_hndls);

    dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: calling init_mem_fmr_ctrl. \n",
                  __FUNCTION__); 
                                      
    init_mem_fmr_ctrl(&pz->mem_fmr_ctrl);
    
    /* memroy pool access privileges are set to maximum */
    /* BUGBUG : need to be determined by the consumer   */
    ACCESS_CONTROL_MASK_CLR_ALL(acl_flags);
    ACCESS_CONTROL_MASK_SET (acl_flags, VAPI_EN_LOCAL_WRITE);
    ACCESS_CONTROL_MASK_SET (acl_flags, VAPI_EN_REMOTE_WRITE);
    ACCESS_CONTROL_MASK_SET (acl_flags, VAPI_EN_REMOTE_READ);
    
    //need to add validity check to make sure it's 100% of allocation

    if (avg_recv_size)
	{
		j = USER_VALUES;
		pages_mapped = avg_recv_size / PAGE_SIZE + 1;

		if (avg_recv_size_percent != 0)
			alloc_fmr_resource_tbl[pages_mapped][USER_VALUES] += avg_recv_size_percent;
		else
			alloc_fmr_resource_tbl[pages_mapped][USER_VALUES] += DEFAULT_AVG_RECV_SIZE_PERC  ;
	}
	else
		j = DEFAULT_VALUES;


	for (i = 1; i< FMR_MAX_OPTIONS_TO_MAP; ++i)
	{
		pages_mapped = i;
		num_of_hndls = (alloc_fmr_resource_tbl[i][j] * num_of_fmr_hndls) / 100;

		if (alloc_fmr_resource_tbl[i][j]) {
            dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: going to allocate the following :pages_mapped=%d, percent=%d, total amount=%d\n",
                      __FUNCTION__, i, alloc_fmr_resource_tbl[i][j], num_of_hndls);

    		ret  = mem_fmr_create_pool_entry(pz,
    										 acl_flags,
    										 PAGE_SHIFT,    // taken from kernel definitions
    										 pages_mapped,  // max pages mapped
    										 num_of_hndls,  // number of handles
                                             max_outstanding_maps);
    
    		if (ret != MEM_RET_SUCCESS) {
    			dapl_dbg_log (DAPL_DBG_TYPE_ERR, "%s: creating pool entry for pages_mapped=%d failed! exiting \n",
                              __FUNCTION__, pages_mapped);
    			return ret;
    		}
        }
		
	}//end of  for

    mem_show_fmr_pool("mem_fmr_init_module - end", &pz->mem_fmr_ctrl);
    
#ifdef DEBUG
	mem_fmr_print_all(mem_fmr_ctrl);
#endif /*DEBUG*/
	return 0;
}


void fmr_set_watermark (int                 mapped_pages, 
                        int                 new_watermark,
                        fmr_ctrl_struct     *mem_fmr_ctrl)
{
	fmr_pool_entry  *p_fmr_pool_entry = NULL;

    
    dapl_dbg_log (DAPL_DBG_TYPE_FMR, "FMR: mapped_pages=%d, new_watermark=%d\n",
		   mapped_pages, new_watermark);

	if (mapped_pages>FMR_MAX_OPTIONS_TO_MAP)
	{
		dapl_dbg_log (DAPL_DBG_TYPE_ERR, "%s: number of pages:%d is greater than allowed(%d)",
                      __FUNCTION__, mapped_pages, FMR_MAX_OPTIONS_TO_MAP);
		return;
	}

	if (new_watermark >= 100)
	{
		dapl_dbg_log (DAPL_DBG_TYPE_ERR, "%s: Watermark is not valid! [greater/equals 100%%]",
                      __FUNCTION__);
		return ;
	}

	p_fmr_pool_entry = &(mem_fmr_ctrl->fmr_pool[mapped_pages]);

	if (p_fmr_pool_entry->total_hndls == 0)
	{
		dapl_dbg_log (DAPL_DBG_TYPE_ERR, "%s: No matching entry found that matches mapped_pages=%d",
                      __FUNCTION__, mapped_pages);
		return ;
	}

	MEM_LOCK(p_fmr_pool_entry->lock);
	p_fmr_pool_entry->min_free_hndl_needed= new_watermark * p_fmr_pool_entry->total_hndls /100;
	MEM_UNLOCK(p_fmr_pool_entry->lock);
}

void fmr_set_unmap_at_once (int                 new_unmap_value,
                            fmr_ctrl_struct     *mem_fmr_ctrl)
{
    MEM_LOCK(mem_fmr_ctrl->unmap_vector.hndl_vec_lock);
	mem_fmr_ctrl->unmap_vector.max_unmap_at_once = new_unmap_value;
	MEM_UNLOCK(mem_fmr_ctrl->unmap_vector.hndl_vec_lock);

	dapl_dbg_log (DAPL_DBG_TYPE_FMR, "FMR: new value of mem_fmr_ctrl.unmap_vector.max_unmap_at_once=%d\n",
		   mem_fmr_ctrl->unmap_vector.max_unmap_at_once);
}


void mem_show_fmr_pool(char             *func,
                       fmr_ctrl_struct  *mem_fmr_ctrl)
{
    int                 jj, free_counter, used_counter;
    struct list_head    *fmr_hndl_iter;
    fmr_pool_entry      *p_fmr_pool_entry = NULL;

    for (jj = 0;jj < FMR_MAX_OPTIONS_TO_MAP; ++jj)
    {
        p_fmr_pool_entry = &(mem_fmr_ctrl->fmr_pool[jj]);
        
        if (!(p_fmr_pool_entry->allocted_pages))
            continue;
        
        free_counter = 0; used_counter=0;
        list_for_each (fmr_hndl_iter, &(p_fmr_pool_entry->free_fmr_hndl_list))
        {
            free_counter++;
        }
            
        list_for_each (fmr_hndl_iter, &(p_fmr_pool_entry->used_fmr_hndl_list))
        {
            used_counter++;
        }
            
        dapl_dbg_log (DAPL_DBG_TYPE_FMR, "%s: fmr_pool[%d] - num_free_hndls=%d used_counter=%d free_counter=%d\n",
                      func, jj, p_fmr_pool_entry->num_free_hndls, used_counter, free_counter);
    }

}
#if 0
int show_proc_fmr_pool(char             *page, 
                       fmr_ctrl_struct  *mem_fmr_ctrl)
{
    char                buf[256];
    int                 nc=0, jj, ii, free_counter, used_counter;
    struct list_head    *fmr_hndl_iter;
    fmr_pool_entry      *p_fmr_pool_entry = NULL;
    int                 pages_mapped;
	int                 num_of_hndls;

    
	if (avg_recv_size)
	{
		nc += sprintf(buf, "\t\tuser requested size of :%d ",avg_recv_size);
        strcat(page, buf);
        
		if (avg_recv_size_percent)
		{
            nc += sprintf(buf, ", percent of allocations:%d\n",avg_recv_size_percent);
            strcat(page, buf);
        }
		else
        {
            nc += sprintf(buf, "\n");
            strcat(page, buf);
        }
        ii=USER_VALUES;
	}
	else
		ii = DEFAULT_VALUES;

    nc += sprintf(buf, "\t\tMax unmap at once = %d\n", mem_fmr_ctrl->unmap_vector.max_unmap_at_once);
    strcat(page, buf);
    
    nc += sprintf(buf, "\t\tNeed to unmap vector: index=%d (urgent=%d)\n", 
                  mem_fmr_ctrl->unmap_vector.idx, mem_fmr_ctrl->unmap_vector.urgent_flag);
    strcat(page, buf);

    nc += sprintf(buf, "\t\tTotal unmap done = %d\n", g_unmap_cnt);
    strcat(page, buf);

    for (jj = 0;jj < FMR_MAX_OPTIONS_TO_MAP; ++jj)
    {
        p_fmr_pool_entry = &(mem_fmr_ctrl->fmr_pool[jj]);
        
        pages_mapped = jj;
		num_of_hndls = (alloc_fmr_resource_tbl[jj][ii] * num_of_fmr_hndls) / 100;
		if (num_of_hndls)
        {
            nc += sprintf(buf, "\t\tAllocated %d pages pool (%d%% of overall handles), total amount=%d\n",
                          jj, alloc_fmr_resource_tbl[jj][ii], num_of_hndls);
            strcat(page, buf);
        }
			
        
        if (!(p_fmr_pool_entry->allocted_pages))
            continue;
        
        free_counter = 0; used_counter=0;
        list_for_each (fmr_hndl_iter, &(p_fmr_pool_entry->free_fmr_hndl_list))
        {
            free_counter++;
        }
            
        list_for_each (fmr_hndl_iter, &(p_fmr_pool_entry->used_fmr_hndl_list))
        {
            used_counter++;
        }
            
        nc += sprintf(buf, "\t\t[%d pages pool] : used_handles=%d free_handles=%d\n",
                      jj, used_counter, free_counter);
        strcat(page, buf);
    } 

    return nc;
}
#endif
