/*
 * Copyright (c) 1999-2002, Mellanox Technologies, Inc. All rights reserved. 
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under either one of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 * 
 * HEADER: dapl_vapi_dto.h
 *
 * PURPOSE: Utility routines for data transfer operations using the
 * MELLANOX Tavor Verbs
 *
 * $Id$
 *
 **********************************************************************/

#ifndef _DAPL_TAVOR_DTO_H
#define _DAPL_TAVOR_DTO_H

#include "vapi_types.h"
#include "vapi.h"
#include "evapi.h"
#include "dapl_openib_util.h"

STATIC _INLINE_ int
dapls_cqe_opcode_convert (ib_work_completion_t *cqe_p);

/*
 * dapls_ib_post_recv
 *
 * Provider specific Post RECV function
 */
STATIC _INLINE_ DAT_RETURN 
dapls_ib_post_recv (
	IN  DAPL_EP		   	*ep_ptr,
	IN  DAPL_COOKIE			*cookie,
	IN  DAT_COUNT	   		num_segments,
	IN  DAT_LMR_TRIPLET	   	*local_iov)
{
    VAPI_ret_t          ib_status;
    VAPI_rr_desc_t	rr_desc;
    VAPI_sg_lst_entry_t *sg_list_p;
    DAT_COUNT           i, total_len;

    rr_desc.id        = (VAPI_wr_id_t) (uintptr_t) cookie;
    rr_desc.opcode    = VAPI_RECEIVE;

    if (cookie != NULL)
    {
	rr_desc.comp_type = VAPI_SIGNALED;
    }
    else
    {
	rr_desc.comp_type = VAPI_UNSIGNALED;
    }

    dapl_dbg_log (DAPL_DBG_TYPE_EVD, 
                  " comp_type = %d\n",
		  rr_desc.comp_type);

    rr_desc.sg_lst_len = num_segments;
    rr_desc.sg_lst_p   = ep_ptr->recv_iov;

    sg_list_p          = rr_desc.sg_lst_p;
    total_len                   = 0;

    for (i = 0; i < num_segments; i++, sg_list_p++)
    {
	sg_list_p->len  = (unsigned int) local_iov[i].segment_length;
	sg_list_p->lkey = local_iov[i].lmr_context;
	sg_list_p->addr = local_iov[i].virtual_address;
	total_len      += sg_list_p->len;
    }

    if (cookie != NULL)
    {
	cookie->val.dto.size            =  total_len;

	dapl_dbg_log (DAPL_DBG_TYPE_EVD, 
                      " op_type= %d, cookie= %p, num_seg= %d\n", 
                      VAPI_RECEIVE, 
		      cookie, 
		      num_segments);
    }

    ib_status = VAPI_post_rr (ep_ptr->header.owner_ia->hca_ptr->ib_hca_handle,
                              ep_ptr->qp_handle,
                              &rr_desc);

    dapl_dbg_log (DAPL_DBG_TYPE_EVD, " post_rr ib_status = %d\n", ib_status);

    if (VAPI_OK == ib_status)
    {
	return DAT_SUCCESS;
    }
    else
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
                      " dat_status = %d\n", 
		      DAT_INTERNAL_ERROR);
	return DAT_INTERNAL_ERROR;
    }
}


/*
 * dapls_ib_post_send
 *
 * Provider specific Post SEND function
 */
STATIC _INLINE_ DAT_RETURN 
dapls_ib_post_send (
	IN  DAPL_EP		   	*ep_ptr,
	IN  int				op_type,
	IN  DAPL_COOKIE			*cookie,
	IN  DAT_COUNT	   		num_segments,
	IN  DAT_LMR_TRIPLET	   	*local_iov,
	IN  const DAT_RMR_TRIPLET	*remote_iov,
	IN  DAT_COMPLETION_FLAGS	completion_flags)
{
    VAPI_ret_t 		ib_status;
    VAPI_sr_desc_t      send_desc;
    VAPI_sg_lst_entry_t	*sg_list_p;
    DAT_COUNT           i;
    DAT_UINT32          total_len;

    send_desc.opcode		= op_type;
    send_desc.sg_lst_len	= num_segments;
    send_desc.sg_lst_p	    = ep_ptr->send_iov;
    total_len               = 0;
    sg_list_p = send_desc.sg_lst_p;

    for (i = 0; i < num_segments; i++, sg_list_p++)
    {
	sg_list_p->len  = (unsigned int) local_iov[i].segment_length;
	sg_list_p->lkey = local_iov[i].lmr_context;
	sg_list_p->addr = local_iov[i].virtual_address;
	total_len      += sg_list_p->len;
    }

    if (cookie != NULL)
    {
	cookie->val.dto.size            =  total_len;

	dapl_dbg_log (DAPL_DBG_TYPE_EVD, 
                      " op_type = %d, cookie = %p num_seg = %d\n", 
                      op_type, 
		      cookie, 
		      num_segments);
    }

    send_desc.id		= (VAPI_wr_id_t) (uintptr_t) cookie;
    send_desc.imm_data		= 0;
    send_desc.remote_qkey	= 0;
    send_desc.remote_ah		= 0;
    send_desc.ethertype		= 0;

    send_desc.set_se		= (completion_flags &DAT_COMPLETION_SOLICITED_WAIT_FLAG) ? TRUE : FALSE;

    if (num_segments > 0)
    {
	send_desc.remote_addr	= remote_iov->target_address;
	send_desc.r_key		= remote_iov->rmr_context;
    }
    else
    {
	send_desc.remote_addr	= 0;
	send_desc.r_key		= 0;
    }

    send_desc.fence	= (DAT_COMPLETION_BARRIER_FENCE_FLAG & 
		    	   completion_flags) ? TRUE : FALSE;
    send_desc.comp_type	= (DAT_COMPLETION_SUPPRESS_FLAG & 
		    	   completion_flags) ? VAPI_UNSIGNALED : VAPI_SIGNALED;

    dapl_dbg_log (DAPL_DBG_TYPE_EVD, 
                  " comp_type = %d, rem_addr = %p rem_rkey = 0x%x\n", 
                  send_desc.comp_type, 
		  (void *) (uintptr_t) send_desc.remote_addr,
		  send_desc.r_key);

    ib_status = VAPI_post_sr (ep_ptr->header.owner_ia->hca_ptr->ib_hca_handle,
                              ep_ptr->qp_handle,
                              &send_desc);

    dapl_dbg_log (DAPL_DBG_TYPE_EVD, " post_sr ib_status = %d\n", ib_status);

    if (VAPI_OK == ib_status)
    {
	return DAT_SUCCESS;
    }
    else
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
                      " dat_status = %d\n", 
		      DAT_INTERNAL_ERROR);
	return DAT_INTERNAL_ERROR;
    }
}

/*
 * dapls_ib_optional_prv_dat
 *
 * Allocate space for private data to be used in CR calls
 *
 * Input:
 *	cr_ptr			CR handle
 *	event_data		data provided by the provider callback function
 *	cr_pp			Pointer for private data
 *
 * Output:
 * 	cr_pp			Area 
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *
 */
STATIC _INLINE_ DAT_RETURN
dapls_ib_optional_prv_dat (
	IN  DAPL_CR		*cr_ptr,
	IN  const void		*event_data,
	OUT DAPL_CR             **cr_pp)
{
#ifndef CM_BUSTED

    /* Remember we have to free up these memory after done */
    cr_ptr->param.private_data = dapl_os_alloc (sizeof (struct ib_u_cm_rep_received_param *));

    if (cr_ptr->param.private_data == NULL)
    {
	return DAT_INSUFFICIENT_RESOURCES;
    }

    dapl_os_memcpy (cr_ptr->param.private_data, (void *) event_data,
                    sizeof (struct ib_u_cm_req_received_param *));

    cr_ptr->param.private_data_size = sizeof (struct ib_u_cm_rep_received_param *);
    *cr_pp = cr_ptr;

#endif

    return DAT_SUCCESS;
}


STATIC _INLINE_ int
dapls_cqe_opcode_convert (ib_work_completion_t *cqe_p)
{
    dapl_dbg_log (DAPL_DBG_TYPE_EVD,
	    	  " completion opcode = %d\n", 
		  cqe_p->opcode);
    switch (((ib_work_completion_t *)cqe_p)->opcode)
    {
	case VAPI_CQE_SQ_SEND_DATA:
	{
	    return (OP_SEND);
	}
	case VAPI_CQE_SQ_RDMA_WRITE:
	{
	    return (OP_RDMA_WRITE);
	}
	case VAPI_CQE_SQ_RDMA_READ:
	{
	    return (OP_RDMA_READ);
	}
	case VAPI_CQE_SQ_COMP_SWAP:
	{
	    return (OP_COMP_AND_SWAP);
	}
	case VAPI_CQE_SQ_FETCH_ADD:
	{
	    return (OP_FETCH_AND_ADD);
	}
	case VAPI_CQE_SQ_BIND_MRW:
	{
	    return (OP_BIND_MW);
	}
	case VAPI_CQE_RQ_SEND_DATA:
	{
	    return (OP_RECEIVE);
	}
	case VAPI_CQE_RQ_RDMA_WITH_IMM:
	{
	    return (VAPI_CQE_RQ_RDMA_WITH_IMM);
	}
        default:
        {
            dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
		          " ERROR wc opcode = %d\n", 
			  cqe_p->opcode);
            dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
		          " ERROR wc status = %d\n", 
			  cqe_p->status);
            if (cqe_p->status != VAPI_SUCCESS)
            {
                /*
                 * Error reported from driver without the opcode of WR which 
                 * associates with this cqe. We just assign some default value
                 */
                 return (OP_RDMA_WRITE);
            } 
            else
            {
                 return (VAPI_EINVAL_OP);
            }
        }
    }
}

#define DAPL_GET_CQE_WRID(cqe_p) ((ib_work_completion_t *)cqe_p)->id
#define DAPL_GET_CQE_OPTYPE(cqe_p) dapls_cqe_opcode_convert(cqe_p)
#define DAPL_GET_CQE_BYTESNUM(cqe_p) ((ib_work_completion_t *)cqe_p)->byte_len
#define DAPL_GET_CQE_STATUS(cqe_p) ((ib_work_completion_t *)cqe_p)->status
#define DAPL_GET_CQE_VND_SYND(cqe_p) ((ib_work_completion_t *)cqe_ptr)->vendor_err_syndrome
#endif /* _DAPL_TAVOR_DTO_H */
