/*
 * Copyright (c) 1999-2002, Mellanox Technologies, Inc. All rights reserved. 
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * This Software is licensed under either one of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/**********************************************************************
 * 
 * MODULE: dapl_vapi_qp.c
 *
 * PURPOSE: IB QP routines  for access to MELLANOX Tavor Verbs
 *
 * $Id: dapl_openib_qp.c,v 1.2 2004/07/28 08:19:47 rabin Exp $
 *
 **********************************************************************/


#ifdef __KERNEL__
#include "linux/types.h"
#endif

#include "dapl_openib_cm.h"
#include "dapl.h"
#include "dapl_adapter_util.h"
#include "dapl_evd_util.h"
#include "dapl_openib_util.h"

ib_cq_handle_t dapli_get_dto_cq (
	IN  DAPL_IA		*ia_ptr,
	IN  DAPL_EVD		*recv_evd_ptr );


/*
 * dapl_ib_qp_alloc
 *
 * Alloc a QP
 *
 * Input:
 *        *ep_ptr                pointer to EP INFO
 *        ib_hca_handle        provider HCA handle
 *        ib_pd_handle        provider protection domain handle
 *        cq_recv                provider recv CQ handle
 *        cq_send                provider send CQ handle
 *
 * Output:
 *         none
 *
 * Returns:
 *         DAT_SUCCESS
 *        DAT_INSUFFICIENT_RESOURCES
 *
 */
DAT_RETURN
dapls_ib_qp_alloc (
        IN  DAPL_IA                *ia_ptr,
        IN  DAPL_EP                *ep_ptr,
        IN  DAPL_EP                *ep_ctx_ptr )
{
    DAT_EP_ATTR           *attr;
    VAPI_ret_t            ib_status;
    VAPI_qp_init_attr_t   qp_attr;
    VAPI_qp_prop_t        qp_prop;
    VAPI_pd_hndl_t        ib_pd_handle;
    VAPI_hca_hndl_t       ib_hca_handle;
    VAPI_cq_hndl_t        cq_recv;
    VAPI_cq_hndl_t        cq_send;

    attr = &ep_ptr->param.ep_attr;

    dapl_os_assert ( ep_ptr->param.pz_handle != NULL );
    ib_pd_handle  = ((DAPL_PZ *)ep_ptr->param.pz_handle)->pd_handle;
    ib_hca_handle = ia_ptr->hca_ptr->ib_hca_handle;

    cq_recv = 0;
    cq_send = 0;
    cq_recv = dapli_get_dto_cq (ia_ptr, 
				(DAPL_EVD *)ep_ptr->param.recv_evd_handle);
    cq_send = dapli_get_dto_cq (ia_ptr, 
				(DAPL_EVD *)ep_ptr->param.request_evd_handle);

    qp_attr.cap.max_oust_wr_sq        = attr->max_request_dtos + attr->max_rdma_read_out;
    qp_attr.cap.max_oust_wr_rq        = attr->max_recv_dtos;
    qp_attr.cap.max_sg_size_sq        = attr->max_request_iov;
    qp_attr.cap.max_sg_size_rq        = attr->max_recv_iov;
    qp_attr.pd_hndl                   = ib_pd_handle;
    qp_attr.rdd_hndl                  = 0;
    qp_attr.rq_cq_hndl                = cq_recv;
    qp_attr.rq_sig_type               = VAPI_SIGNAL_ALL_WR;
    qp_attr.sq_cq_hndl                = cq_send;
    qp_attr.sq_sig_type               = VAPI_SIGNAL_REQ_WR;
    qp_attr.ts_type                   = VAPI_TS_RC;

    ib_status = VAPI_create_qp ( ib_hca_handle,
                                 &qp_attr,
                                 &ep_ptr->qp_handle,
                                 &qp_prop);
    if (ib_status != VAPI_OK)
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, " failed code = %d\n", ib_status);
	return dapl_ib_status_convert (ib_status);
    }

    ep_ptr->qpn = qp_prop.qp_num;

    dapl_dbg_log (DAPL_DBG_TYPE_EP, 
	    	  " created qp_hndl = %x, qpn = %x\n", 
                  (uint32_t) ep_ptr->qp_handle, 
		  ep_ptr->qpn);

    ep_ptr->qp_state = VAPI_RESET;

    dapl_dbg_log (DAPL_DBG_TYPE_EP, " using port = %d\n", ia_ptr->hca_ptr->port_num);
    ib_status = dapls_modify_qp_state_to_init ( 
                          ia_ptr->hca_ptr->ib_hca_handle,
                          ia_ptr->hca_ptr->port_num,
                          ep_ptr->qp_handle);

    if ( ib_status != VAPI_OK )
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, " failed code = %d\n", ib_status);
	return dapl_ib_status_convert (ib_status);
    }

    ep_ptr->qp_state = VAPI_INIT;

    return DAT_SUCCESS;
}


/*
 * dapl_ib_qp_free
 *
 * Free a QP
 *
 * Input:
 *        *ep_ptr                pointer to EP INFO
 *        ib_hca_handle        provider HCA handle
 *
 * Output:
 *         none
 *
 * Returns:
 *         Status of destroy operation.
 *
 */
DAT_RETURN
dapls_ib_qp_free (
        IN  DAPL_IA                *ia_ptr,
        IN  DAPL_EP                *ep_ptr )
{
    VAPI_hca_hndl_t        ib_hca_handle;
    VAPI_ret_t          ib_status;

    ib_hca_handle = ia_ptr->hca_ptr->ib_hca_handle;

    if (ep_ptr->qp_handle != IB_INVALID_HANDLE )
    {
	ib_status = VAPI_destroy_qp ( ib_hca_handle, ep_ptr->qp_handle);
	if (ib_status != VAPI_OK)
	{
	    return dapl_ib_status_convert (ib_status);
	}
    }

    return DAT_SUCCESS;
}




/*
 * dapli_get_recv_cq
 *
 * Obtain the cq_handle for a DTO EVD. If the EVD is NULL, use the
 * null_ib_cq_handle. If it hasn't been created yet, create it now in
 * the HCA structure. It will be cleaned up in dapls_ib_cqd_destroy().
 *
 * This is strictly internal to IB. DAPL allows a NULL DTO EVD handle,
 * but IB does not. So we create a CQ under the hood and make sure
 * an error is generated if the user every tries to post, by
 * setting the WQ length to 0 in ep_create and/or ep_modify.
 *
 * Returns
 *	A valid CQ handle
 */
ib_cq_handle_t
dapli_get_dto_cq (
	IN  DAPL_IA		*ia_ptr,
	IN  DAPL_EVD		*evd_ptr )
{
    ib_cq_handle_t	cq_handle;
    int			cq_entries;
    VAPI_ret_t          ib_status;

    if ( evd_ptr != DAT_HANDLE_NULL )
    {
	cq_handle = evd_ptr->ib_cq_handle;
    }
    else if ( ia_ptr->hca_ptr->ib_trans.null_ib_cq_handle != IB_INVALID_HANDLE )
    {
	cq_handle = ia_ptr->hca_ptr->ib_trans.null_ib_cq_handle;
    }
    else
    {
	cq_entries = 0;
	ib_status = VAPI_create_cq ( ia_ptr->hca_ptr->ib_hca_handle,
				     cq_entries,
				     &ia_ptr->hca_ptr->ib_trans.null_ib_cq_handle,
				     &cq_entries );
	if (ib_status != VAPI_OK)
	{
	    ia_ptr->hca_ptr->ib_trans.null_ib_cq_handle = IB_INVALID_HANDLE;
	}
	cq_handle = ia_ptr->hca_ptr->ib_trans.null_ib_cq_handle;
    }

    return cq_handle;
}


/*
 * dapl_ib_qp_modify
 *
 * Set the QP to the parameters specified in an EP_PARAM
 *
 * We can't be sure what state the QP is in so we first obtain the state
 * from the driver. The EP_PARAM structure that is provided has been
 * sanitized such that only non-zero values are valid.
 *
 * Input:
 *        ib_hca_handle                HCA handle
 *        qp_handle                QP handle
 *        ep_attr                Sanitized EP Params
 *
 * Output:
 *         none
 *
 * Returns:
 *         DAT_SUCCESS
 *        DAT_INSUFFICIENT_RESOURCES
 *        DAT_INVALID_PARAMETER
 *
 */
DAT_RETURN
dapls_ib_qp_modify (
        IN  DAPL_IA                        *ia_ptr,
        IN  DAPL_EP                        *ep_ptr,
        IN  DAT_EP_ATTR                    *ep_attr )
{
    VAPI_qp_attr_t                *qp_attr = NULL;
    VAPI_ret_t                    ib_status;
    VAPI_qp_cap_t                 *qp_cap = NULL;
    VAPI_hca_hndl_t               ib_hca_handle;
    VAPI_qp_hndl_t                qp_handle;
    VAPI_qp_attr_mask_t           qp_attr_mask;
    VAPI_qp_state_t               qp_state;
    VAPI_qp_init_attr_t           *qp_init_attr = NULL;
    DAT_RETURN                    dat_status = DAT_SUCCESS;

    ib_hca_handle = ia_ptr->hca_ptr->ib_hca_handle;
    qp_handle     = ep_ptr->qp_handle;

    qp_attr  = dapl_os_alloc (sizeof(VAPI_qp_attr_t));
    if ( qp_attr == NULL )
    {
    dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
    goto bail;
    }
    qp_init_attr  = dapl_os_alloc (sizeof(VAPI_qp_init_attr_t));
    if ( qp_init_attr == NULL )
    {
    dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
    goto bail;
    }
    qp_cap  = dapl_os_alloc (sizeof(VAPI_qp_cap_t));
    if ( qp_cap == NULL )
    {
    dat_status = DAT_ERROR (DAT_INSUFFICIENT_RESOURCES, DAT_RESOURCE_MEMORY);
    goto bail;
    }
    /* get the state */
    ib_status = VAPI_query_qp ( ib_hca_handle,
                                qp_handle,
                                qp_attr,
                                &qp_attr_mask,
                                qp_init_attr);

    if ( ib_status != VAPI_OK )
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, " failed code = %d\n", ib_status);
	dat_status = dapl_ib_status_convert (ib_status);
    goto bail;
    }

    qp_state = qp_attr->qp_state;

    /* set the new parameter values */
    QP_ATTR_MASK_CLR_ALL (qp_attr_mask);

    /*
     * NOTE: ep_attr->max_mtu_size  indicates the maximum message
     *	     size, which is always 2GB for IB. Nothing to
     *	     do with the IB implementation, nothing to set up.
     */

    if ( ep_attr->max_rdma_size > 0 )
    {
	qp_attr->cap.max_sg_size_sq= (unsigned int) ep_attr->max_rdma_size;
	qp_attr->cap.max_sg_size_rq= (unsigned int) ep_attr->max_rdma_size;
	QP_ATTR_MASK_SET (qp_attr_mask, QP_ATTR_CAP);
    }

    if ( ep_attr->max_recv_dtos > 0 )
    {
	qp_attr->cap.max_oust_wr_rq = ep_attr->max_recv_dtos;
	QP_ATTR_MASK_SET (qp_attr_mask, QP_ATTR_CAP);
    }

    if ( ep_attr->max_request_dtos > 0 )
    {
	qp_attr->cap.max_oust_wr_sq = ep_attr->max_request_dtos;
	QP_ATTR_MASK_SET (qp_attr_mask, QP_ATTR_CAP);
    }

    if ( ep_attr->max_recv_iov > 0 )
    {
	/* MELLANOX
	 *   We do not have MW on the qp attribute
	 */
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, " max_recv_iov > 0\n");
    }

    if ( ep_attr->max_request_iov > 0 )
    {
	/* MELLANOX
	 *   We do not have MW on the qp attribute
	 */
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, " max_request_iov > 0\n");
    }

    qp_attr->qp_state = qp_state;
    QP_ATTR_MASK_SET (qp_attr_mask, QP_ATTR_QP_STATE);

    ib_status = VAPI_modify_qp ( ib_hca_handle,
                                 qp_handle,
                                 qp_attr,
                                 &qp_attr_mask,
                                 qp_cap);

    if ( ib_status != VAPI_OK)
    {
	dapl_dbg_log (DAPL_DBG_TYPE_ERR, " failed code = %d\n", ib_status);
	dat_status = dapl_ib_status_convert (ib_status);
    }
bail:
    if (qp_attr != NULL) {
        dapl_os_free(qp_attr, 0);
    }
    if (qp_init_attr != NULL) {
        dapl_os_free(qp_init_attr, 0);
    }
    if (qp_cap != NULL) {
        dapl_os_free(qp_cap, 0);
    }
    return dat_status;
}


VAPI_ret_t 
dapls_modify_qp_state_to_init (
        VAPI_hca_hndl_t                hca_handle,
        IB_port_t                      ib_port,
        VAPI_qp_hndl_t                 qp )
{
    VAPI_qp_attr_t                *qp_attr = NULL;
    VAPI_qp_attr_mask_t           qp_attr_mask;
    VAPI_qp_cap_t                 *qp_cap = NULL;
    VAPI_ret_t                    ib_status;
    qp_attr = dapl_os_alloc (sizeof(VAPI_qp_attr_t));
    if ( qp_attr == NULL ) {
        ib_status = VAPI_ENOMEM;
        goto bail;
    }
    qp_cap = dapl_os_alloc (sizeof(VAPI_qp_cap_t));
    if ( qp_cap == NULL ){
        ib_status = VAPI_ENOMEM;
        goto bail;
    }
    QP_ATTR_MASK_CLR_ALL (qp_attr_mask);
    qp_attr->qp_state = VAPI_INIT;
    QP_ATTR_MASK_SET (qp_attr_mask, QP_ATTR_QP_STATE);
    qp_attr->remote_atomic_flags = VAPI_EN_REM_WRITE | 
                                  VAPI_EN_REM_READ;
    QP_ATTR_MASK_SET (qp_attr_mask, QP_ATTR_REMOTE_ATOMIC_FLAGS);
    qp_attr->pkey_ix = 0;
    QP_ATTR_MASK_SET (qp_attr_mask, QP_ATTR_PKEY_IX);
    qp_attr->port = ib_port;
    QP_ATTR_MASK_SET (qp_attr_mask, QP_ATTR_PORT);

    ib_status = VAPI_modify_qp ( hca_handle,
                                 qp,
                                 qp_attr,
                                 &qp_attr_mask,
                                 qp_cap);

    dapl_dbg_log (DAPL_DBG_TYPE_EP, " ib_status = %d\n", ib_status);
bail:
    if (qp_attr != NULL) {
        dapl_os_free(qp_attr, 0);
    }
    if (qp_cap != NULL) {
        dapl_os_free(qp_cap, 0);
    }
    return ib_status;
}



VAPI_ret_t 
dapls_modify_qp_state_to_error (
        VAPI_hca_hndl_t               hca_handle,
        VAPI_qp_hndl_t                qp )
{
    VAPI_qp_attr_t                *qp_attr = NULL;
    VAPI_qp_attr_mask_t           qp_attr_mask;
    VAPI_qp_cap_t                 *qp_cap = NULL;
    VAPI_ret_t                    ib_status;
    qp_attr = dapl_os_alloc (sizeof(VAPI_qp_attr_t));
    if ( qp_attr == NULL ) {
        ib_status = VAPI_ENOMEM;
        goto bail;
    }
    qp_cap = dapl_os_alloc (sizeof(VAPI_qp_cap_t));
    if ( qp_cap == NULL ){
        ib_status = VAPI_ENOMEM;
        goto bail;
    }
    QP_ATTR_MASK_CLR_ALL (qp_attr_mask);
    qp_attr->qp_state = VAPI_ERR;
    QP_ATTR_MASK_SET (qp_attr_mask, QP_ATTR_QP_STATE);

    ib_status = VAPI_modify_qp ( hca_handle,
                            qp,
                            qp_attr,
                            &qp_attr_mask,
                            qp_cap);
bail:
    if (qp_attr != NULL) {
        dapl_os_free(qp_attr, 0);
    }
    if (qp_cap != NULL) {
        dapl_os_free(qp_cap, 0);
    }
    return ib_status;
}

VAPI_ret_t 
dapls_modify_qp_state_to_reset (
        VAPI_hca_hndl_t                hca_handle,
        VAPI_qp_hndl_t                qp )
{
    VAPI_qp_attr_t               *qp_attr = NULL;
    VAPI_qp_attr_mask_t          qp_attr_mask;
    VAPI_qp_cap_t                *qp_cap = NULL;
    VAPI_ret_t                    ib_status;
    qp_attr = dapl_os_alloc (sizeof(VAPI_qp_attr_t));
    if ( qp_attr == NULL ) {
        ib_status = VAPI_ENOMEM;
        goto bail;
    }
    qp_cap = dapl_os_alloc (sizeof(VAPI_qp_cap_t));
    if ( qp_cap == NULL ){
        ib_status = VAPI_ENOMEM;
        goto bail;
    }
    QP_ATTR_MASK_CLR_ALL (qp_attr_mask);
    qp_attr->qp_state = VAPI_RESET;
    QP_ATTR_MASK_SET (qp_attr_mask, QP_ATTR_QP_STATE);

    ib_status = VAPI_modify_qp ( hca_handle,
                            qp,
                            qp_attr,
                            &qp_attr_mask,
                            qp_cap);
bail:
    if (qp_attr != NULL) {
        dapl_os_free(qp_attr, 0);
    }
    if (qp_cap != NULL) {
        dapl_os_free(qp_cap, 0);
    }
    return ib_status;
}



