/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.
 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

#include "dapl_proto.h"
#include "kdapl_ioctl.h"
#include "kdapl_tdep.h"
#include <linux/init.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <asm/uaccess.h>

MODULE_LICENSE ("BSD");
MODULE_DESCRIPTION("kdapltest module");
MODULE_AUTHOR("Network Appliance, Inc.");

#define EXPORT_SYMTAB

/* static data */
static int		kdapltest_major;/* device major number */
static DT_Mdep_LockType	kdapltest_lock;	/* lock global test structs */
static int		kdapltest_num;	/* unique test invocation number */
static Transaction_Stats2_t Client_Stats_T;
static Performance_Stats_t Client_Stats_P;
static void
DT_Tdep_Test_Thread (Params_t *user_params);

static int kdapltest_open (struct inode *inode, struct file *file)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    MOD_INC_USE_COUNT;
#endif
    return 0;
}

static int kdapltest_release (struct inode *inode, struct file *file)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    MOD_DEC_USE_COUNT;
#endif
    return 0;
}

static int kdapltest_ioctl (struct inode *inode,
			    struct file *file,
			    unsigned int ioctl,
			    unsigned long param)
{
    char      		buffer[PRINT_MAX];
    DT_get_printf_ioctl	print_ioctl;
    Params_t		*local_params;
    int       		rval;
    int			instance;
    
    rval = 0;

    switch (ioctl)
    {
	case KDAPL_IOCTL_START_TEST:
	{

	    local_params = DT_Mdep_Malloc (sizeof (Params_t));
	    if (!local_params)
	    {
		rval = -ENOMEM;
		break;
	    }
        memset((void *)local_params, 0, sizeof (Params_t));
	    if (copy_from_user (local_params,
				(void *)param,
				sizeof (Params_t)))
	    {
	        DT_Mdep_Free (local_params);
		rval = -EFAULT;
		break;
	    }

	    DT_Mdep_Lock (&kdapltest_lock);
	    kdapltest_num++;
	    instance = kdapltest_num;
	    DT_Mdep_Unlock (&kdapltest_lock);

	    local_params->phead = KDT_Print_Alloc (instance);
	    if (!local_params->phead)
		{
	        DT_Mdep_Free (local_params);
		rval = -ENOMEM;
		break;
		}
        g_status = 0;
	    kernel_thread ((void*)DT_Tdep_Test_Thread,
			   (void*)local_params,
			   0);
	    rval = instance;
	    break;
	}
    case KDAPL_IOCTL_END_TEST:
    {
        int status = g_status;
        if (copy_to_user((int *) param,
                 &status, sizeof status)) {
            rval = -EFAULT;
        }
        break;
    }
    case KDAPL_IOCTL_GET_STAT_T:
    {
        Transaction_Stats2_t Client_Stats_local = Client_Stats_T;
        if (copy_to_user((Transaction_Stats2_t *) param,
                         &Client_Stats_local, sizeof(Transaction_Stats2_t))) {
            rval = -EFAULT;
        }
        break;
    }
    case KDAPL_IOCTL_GET_STAT_P:
    {
        Performance_Stats_t Client_Stats_local = Client_Stats_P;
        if (copy_to_user((Performance_Stats_t *) param,
                         &Client_Stats_local, sizeof(Performance_Stats_t))) {
            rval = -EFAULT;
        }
        break;
    }
	case KDAPL_IOCTL_GET_PRINTF:
	{
	    if (copy_from_user (&print_ioctl,
				(void *)param,
				sizeof (DT_get_printf_ioctl)))
	    {
		rval = -EFAULT;
		break;
	    }
	    rval = KDT_Get_Print_Line (print_ioctl.cookie, &buffer[0]);
	    if (copy_to_user (print_ioctl.buffer,
			      &buffer[0],
			      strlen (buffer)+1))
	    {
		rval = -EFAULT;
		break;
	    }
	    break;
	}
	default:
	{
	    printk ("<1> kdapltest IOCTL %x UNKNOWN\n", ioctl);
	    rval = -EINVAL;
	    break;
	}
    }
    return rval;
}

/* fops */
static struct file_operations kdapltest_fops =
{
    open:    kdapltest_open,
    release: kdapltest_release,
    ioctl:   kdapltest_ioctl,
};

static void
DT_Tdep_Test_Thread (Params_t *user_params)
{
    /* don't want the module to disappear while threads exist */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    MOD_INC_USE_COUNT;
#endif
    DT_Execute_Test (user_params);
    if (user_params->test_type == TRANSACTION_TEST) {
        Client_Stats_T = user_params->Client_Stats_T;
    }
    if (user_params->test_type == PERFORMANCE_TEST) {
        Client_Stats_P = user_params->Client_Stats_P;
    }
    DT_Tdep_PT_Printf (user_params->phead, "");  /* indicate done */
    DT_Mdep_Free (user_params);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    MOD_DEC_USE_COUNT;
#endif
}

__init
int kdapltest_init (void)
{
    
    DT_Mdep_LockInit (&g_PerfTestLock); /* For uDAPL, this is done in Tdep_init() */
    kdapltest_major = register_chrdev (0, "kdapltest", &kdapltest_fops);
    DT_Mdep_LockInit (&kdapltest_lock);
    kdapltest_num = 0;
    KDT_Print_Init ();
    KDT_Evd_Init ();
    printk ("<1>Kdapltest Module Loaded - major %d\n", kdapltest_major);
    return (0);
}

__exit
void kdapltest_exit (void)
{
    unregister_chrdev (kdapltest_major, "kdapltest");
    printk ("<1>Kdapltest Module UnLoaded - major %d\n", kdapltest_major);
    KDT_Print_Destroy ();
    KDT_Evd_Destroy ();
    DT_Mdep_LockDestroy (&kdapltest_lock);
    DT_Mdep_LockDestroy (&g_PerfTestLock); /* For uDAPL, this is done in Tdep_end() */
}

module_init (kdapltest_init);
module_exit (kdapltest_exit);
