/* * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.  *
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

#include "dapl_proto.h"
#include "kdapl_ioctl.h"
#include "kdapl_tdep.h"
#include <linux/sched.h>
#include <linux/version.h>
typedef struct Tdep_Event_Tag
{
    DAT_EVENT		   dat_event;
    struct Tdep_Event_Tag *event_next;
} Tdep_Event;

typedef struct Tdep_Evd_Tag
{
    struct Tdep_Evd_Tag		*evd_next;
    struct Tdep_Event_Tag	*event_next;
    struct Tdep_Event_Tag	*event_last;
    DAT_EVD_HANDLE     		evd_handle;
    DT_WAIT_OBJECT     		wait_object;
} Tdep_Evd;

// static function prototypes
static void DT_Tdep_Event_Callback (
	DAT_PVOID		instance_data,
	const DAT_EVENT *	dat_event,
	DAT_BOOLEAN		bool_arg);

// static data

static Tdep_Evd		*DT_Evd_List;
static Tdep_Event	*DT_Event_Free_List;
static DT_Mdep_LockType	 DT_Evd_Lock; /* all lists protected by this lock */

void
KDT_Evd_Init(void)
{
    DT_Evd_List = NULL;
    DT_Event_Free_List = NULL;
    DT_Mdep_LockInit (&DT_Evd_Lock);
}

void
KDT_Evd_Destroy(void)
{
    Tdep_Evd	*evd_ptr;
    Tdep_Event	*event;

    /* clean up the evd list */
    while (DT_Evd_List)
	{
	evd_ptr = DT_Evd_List;
	DT_Evd_List = evd_ptr->evd_next;

	while (evd_ptr->event_next)
	    {
	    event = evd_ptr->event_next;
	    evd_ptr->event_next = event->event_next;
	    DT_Mdep_Free (event);
	    }
	DT_Mdep_wait_object_destroy (&evd_ptr->wait_object);
	DT_Mdep_Free (evd_ptr);
	}
    /* clean up the event free list */
    while (DT_Event_Free_List)
	{
	event = DT_Event_Free_List;
	DT_Event_Free_List = event->event_next;
	DT_Mdep_Free (event);
	}
    DT_Mdep_LockDestroy (&DT_Evd_Lock);
}

DAT_RETURN
DT_Tdep_evd_create (DAT_IA_HANDLE		ia_handle,
		    DAT_COUNT			evd_min_qlen,
		    DAT_CNO_HANDLE		cno_handle,
		    DAT_EVD_FLAGS		evd_flags,
		    DAT_EVD_HANDLE		*evd_handle_ptr)

{
    DAT_RETURN 		dat_status;
    DAT_UPCALL_OBJECT	upcall;
    Tdep_Evd		*evd_ptr;

    evd_ptr = NULL;
    dat_status = DAT_SUCCESS;

    /* allocate tdep evd and fill it in */
    evd_ptr = DT_Mdep_Malloc (sizeof (Tdep_Evd));
    if (!evd_ptr)
    {
	dat_status =  DAT_CLASS_ERROR |
		      DAT_INSUFFICIENT_RESOURCES |
		      DAT_RESOURCE_MEMORY;
	goto error;
    }
    upcall.upcall_func = DT_Tdep_Event_Callback;
    upcall.instance_data = evd_ptr;

    dat_status = dat_evd_kcreate (ia_handle,
				  evd_min_qlen,
				  DAT_UPCALL_SINGLE_INSTANCE,
				  &upcall,
				  evd_flags,
				  evd_handle_ptr);
    if (dat_status != DAT_SUCCESS)
    {
	goto error;
    }

    /* initialize evd_ptr */
    evd_ptr->event_next = NULL;
    evd_ptr->event_last = NULL;
    evd_ptr->evd_handle = *evd_handle_ptr;
    DT_Mdep_wait_object_init (&evd_ptr->wait_object);

    /* add evd_ptr to front of evd list */
    DT_Mdep_Lock (&DT_Evd_Lock);
    evd_ptr->evd_next = DT_Evd_List;
    DT_Evd_List = evd_ptr;
    DT_Mdep_Unlock (&DT_Evd_Lock);

error:
    if (dat_status != DAT_SUCCESS)
    {
	if (evd_ptr)
	{
	    DT_Mdep_Free (evd_ptr);
	}
    }
    return (dat_status);
}
DAT_RETURN
DT_Tdep_evd_dequeue (DAT_EVD_HANDLE evd_handle,
		     DAT_EVENT      *dat_event)
{
    DAT_RETURN dat_status;
    Tdep_Evd   *evd_ptr;    
    Tdep_Event *event;

    dat_status = DAT_SUCCESS;

    /* find the evd_ptr associated with evd_handle */
    evd_ptr = DT_Evd_List;
    while (evd_ptr)
    {
        if (evd_ptr->evd_handle == evd_handle)
        {
            break;
        }
        evd_ptr = evd_ptr->evd_next;
    }

    if (!evd_ptr)
    {
        return DAT_CLASS_ERROR |
               DAT_INTERNAL_ERROR;
    }
    /* Get event */
    DT_Mdep_Lock (&DT_Evd_Lock);
    event = evd_ptr->event_next; /* event present in evd_ptr corresponding to the given evd_handle */
    if (event)
    {
        evd_ptr->event_next = event->event_next; /* point to next event */
        if (evd_ptr->event_next == NULL)
        {
            evd_ptr->event_last = NULL;
        }
        *dat_event = event->dat_event; /* extract event */
        /* put event on free list */
        if (DT_Event_Free_List == NULL)
        {
            DT_Event_Free_List = event;
            event->event_next = NULL;
        }
        else
        {
            event->event_next = DT_Event_Free_List;
            DT_Event_Free_List = event;
        }
    }
    else /* no event found */
    {
        dat_status = DAT_QUEUE_EMPTY;
    }
    DT_Mdep_Unlock (&DT_Evd_Lock);

    return dat_status;
}

DAT_RETURN
DT_Tdep_evd_wait (DAT_EVD_HANDLE evd_handle,
		  DAT_TIMEOUT    timeout,
		  DAT_EVENT      *dat_event)
{
    DAT_RETURN  dat_status;
    Tdep_Evd	*evd_ptr;
    Tdep_Event	*event;
    int rc;

    dat_status = DAT_CLASS_ERROR | DAT_INTERNAL_ERROR;

    /* find the evd_ptr associated with evd_handle */
    evd_ptr = DT_Evd_List;
    while (evd_ptr)
    {
	if (evd_ptr->evd_handle == evd_handle)
	{
	    break;
	}
	evd_ptr = evd_ptr->evd_next;
    }

    if (!evd_ptr)
    {
	return dat_status;
    }
    /* wait for an event */
    if (timeout == DAT_TIMEOUT_INFINITE) 
    {
	rc = wait_event_interruptible (evd_ptr->wait_object, 
				       evd_ptr->event_next != NULL);
    }
    else
    {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
    rc = interruptible_sleep_on_timeout (&evd_ptr->wait_object, timeout*HZ/1000); /* time out in msec */
#else
	rc = wait_event_interruptible_timeout (evd_ptr->wait_object,evd_ptr->event_next != NULL, timeout*HZ/1000); /* time out in msec */
#endif
	if (rc == 0)
	{
	    // assume timeout, unless an event is found below.
	    dat_status =  DAT_TIMEOUT_EXPIRED;
	}
    }
    if (rc == -ERESTARTSYS)
    {
        /* user hit ^C, just exit */
	return DAT_CLASS_ERROR | DAT_INTERRUPTED_CALL;
    }

    /* Got event */
    DT_Mdep_Lock (&DT_Evd_Lock);
    event = evd_ptr->event_next;
    if (event)
    {
	dat_status = DAT_SUCCESS;
	evd_ptr->event_next = event->event_next;
	if (evd_ptr->event_next == NULL)
	{
	    evd_ptr->event_last = NULL;
	}
	*dat_event = event->dat_event;

	/* put event on free list */
	if (DT_Event_Free_List == NULL)
	{
	    DT_Event_Free_List = event;
	    event->event_next = NULL;
	}
	else
	{
	    event->event_next = DT_Event_Free_List;
	    DT_Event_Free_List = event;
	}
    }

    DT_Mdep_Unlock (&DT_Evd_Lock);
    return dat_status;
}

DAT_RETURN
DT_Tdep_evd_free (DAT_EVD_HANDLE evd_handle)
{
    Tdep_Evd	*next;
    Tdep_Evd	*last;

    last = NULL;
    DT_Mdep_Lock (&DT_Evd_Lock);
    next = DT_Evd_List;
    if (next->evd_handle == evd_handle)
    {
	DT_Evd_List = next->evd_next;
    }
    else
    {
	while (next && next->evd_handle != evd_handle)
	{
	    last = next;
	    next = next->evd_next;
	}
	if (next)
	{
	    last->evd_next = next->evd_next;
	}
    }
    DT_Mdep_Unlock (&DT_Evd_Lock);

    DT_Mdep_Free (next);
    return dat_evd_free(evd_handle);
}

static void DT_Tdep_Event_Callback (
	DAT_PVOID		instance_data,
	const DAT_EVENT *	dat_event,
	DAT_BOOLEAN		bool_arg)
{
    Tdep_Event	*event;
    Tdep_Evd	*evd_ptr;
    unsigned long flags;

    evd_ptr = (Tdep_Evd *) instance_data;
  //DT_Mdep_Lock (&DT_Evd_Lock);
  //need irq locking/unlocking here as the upcall might be called in tasklet context  
    spin_lock_irqsave(&DT_Evd_Lock, flags);
    if (DT_Event_Free_List)
    {
	event = DT_Event_Free_List;
	DT_Event_Free_List = event->event_next;
    }
    else
    {
	event = DT_Mdep_Malloc (sizeof (Tdep_Event));
    }
    if (event)
    {
	event->dat_event = *dat_event;
	event->event_next = NULL;
	if (evd_ptr->event_next == NULL)
	{
	    evd_ptr->event_next = event;
	    evd_ptr->event_last = event;
	}
	else
	{
	    evd_ptr->event_last->event_next = event;
	    evd_ptr->event_last = event;
	}
    }
  //DT_Mdep_Unlock (&DT_Evd_Lock);
    spin_unlock_irqrestore(&DT_Evd_Lock, flags);
    DT_Mdep_wait_object_wakeup (&evd_ptr->wait_object);
}
