/*
 * Copyright (c) 2002-2003, Network Appliance, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.
 *
 * This Software is licensed under both of the following two licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 * OR
 *
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * Licensee has the right to choose either one of the above two licenses.
 *
 * Redistributions of source code must retain both the above copyright
 * notice and either one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, either one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

#include "dapl_proto.h"

static void
DT_Performance_Cmd_Usage (void)
{
    DT_Mdep_printf ("USAGE: ---- PERFORMANCE TEST ----\n");
    DT_Mdep_printf ("USAGE:     dapltest -T P\n");
    DT_Mdep_printf ("USAGE:              -s <server Name>\n");
    DT_Mdep_printf ("USAGE:              [-m b|p]\n");
    DT_Mdep_printf ("USAGE:              [-D <device Name>]\n");
    DT_Mdep_printf ("USAGE:              [-d] : debug (zero)\n");
    DT_Mdep_printf ("USAGE:              [-i <num iterations>]     : (1, 000)\n");
    DT_Mdep_printf ("USAGE:              [-p <pipline>]\n");
    DT_Mdep_printf ("USAGE:              [-R <service reliability>]\n");
    DT_Mdep_printf ("USAGE:                  (BE == QOS_BEST_EFFORT - Default)\n");
    DT_Mdep_printf ("USAGE:                  (HT == QOS_HIGH_THROUGHPUT)\n");
    DT_Mdep_printf ("USAGE:                  (LL == QOS_LOW_LATENCY)\n");
    DT_Mdep_printf ("USAGE:                  (EC == QOS_ECONOMY)\n");
    DT_Mdep_printf ("USAGE:                  (PM == QOS_PREMIUM)\n");
    DT_Mdep_printf ("USAGE:           <OP>\n");
    DT_Mdep_printf ("USAGE:\n");
    DT_Mdep_printf ("USAGE: Each OP consists of:\n");
    DT_Mdep_printf ("USAGE:     <transfer_type> : \"RR\" (RDMA READ)\n");
    DT_Mdep_printf ("USAGE:                     : \"RW\" (RDMA WRITE)\n");
    DT_Mdep_printf ("USAGE:     [seg_size [num_segs] ]      : (4096, 1)\n");
}

static bool
DT_Performance_Cmd_Parse_Op (
    Performance_Cmd_t * cmd,
    int 		index,
    int 		my_argc,
    char 		**my_argv)
{
    int 		i;

    /*
     * Op Format: <RR/RW> [seg_size] [num_segs]
     */

    if ( index == my_argc )
    {
	DT_Mdep_printf ("Operation Missing Transfer Type\n");
	return (false);
    }

    for ( i = 0; index < my_argc; i++, index++ )
    {
	switch ( i )
	{
	    case 0:
	    {
		if ( 0 == strncmp (my_argv[index], "RR", strlen ("RR")) )
		{
		    cmd->op.transfer_type = RDMA_READ;
		}
		else if ( 0 == strncmp (my_argv[index], "RW", strlen ("RW")) )
		{
		    cmd->op.transfer_type = RDMA_WRITE;
		}
		else
		{
		    DT_Mdep_printf ("OP type must be <RR/RW/SR>\n");
		    return (false);
		}
		break;
	    }
	    case 1:
	    {
		cmd->op.seg_size = atoi (my_argv[index]);
		break;
	    }
	    case 2:
	    {
		cmd->op.num_segs = atoi (my_argv[index]);
		break;
	    }
	    default:
	    {
		DT_Mdep_printf ("Too many OP args\n");
		return (false);
	    }
	}
    }

    return (true);
}


static bool
DT_Performance_Cmd_Validate (
    Performance_Cmd_t 		*cmd)
{
    if ( '\0' == cmd->server_name[0] )
    {
	DT_Mdep_printf ("Must specify server_name in command line or scriptfile\n");
	return (false);
    }

    if ( '\0' == cmd->dapl_name[0] )
    {
	DT_Mdep_printf ("Must specify device_name in command line or scriptfile\n");
	return (false);
    }

    if ( 0 == cmd->pipeline_len )
    {
	DT_Mdep_printf ("Pipeline size must not be 0\n");
	return (false);
    }

    if ( cmd->debug )
    {
	DT_Performance_Cmd_Print (cmd);
    }

    return true;
}

bool
DT_Performance_Cmd_Parse (
    Performance_Cmd_t 	*cmd,
    int 		my_argc,
    char 		**my_argv,
    mygetopt_t 		*opts)
{
    int             c;
    unsigned int    len;

    for (;;)
    {
	c = DT_mygetopt_r (my_argc, my_argv, "D:dm:i:p:R:s:M:", opts);

	if ( EOF == c )
	{
	    break;
	}

	switch ( c )
	{
	    case 'D':		    /* device name */
	    {
		strncpy (cmd->dapl_name, opts->optarg, NAME_SZ);
		break;
	    }
	    case 'd':		    /* print debug messages */
	    {
		DT_dapltest_debug++;
		cmd->debug = true;
		break;
	    }
	    case 'm':		/* mode */
	    {
		if ( !strncmp (opts->optarg, "b", strlen ("b")) )
		{
		    cmd->mode = BLOCKING_MODE;
		}
		else if ( !strncmp (opts->optarg, "p", strlen ("p")) )
		{
		    cmd->mode = POLLING_MODE;
		}
		else
		{
		    DT_Mdep_printf ("Syntax Error -m <mode> option\n");
		    DT_Performance_Cmd_Usage ();
		    return (false);
		}

		break;
	    }
	    case 'i':		    /* num iterations */
	    {
		len = strspn (opts->optarg, "0123456789");
		if (len == 0 || len != strlen (opts->optarg))
		{
		    DT_Mdep_printf ("Syntax Error -i <iterations> option\n");
		    DT_Performance_Cmd_Usage ();
		    return (false);
		}
		cmd->num_iterations = atol (opts->optarg);
		break;
	    }
	    case 'p':		    /* pipline size */
	    {
		len = strspn (opts->optarg, "0123456789");
		if (len == 0 || len != strlen (opts->optarg))
		{
		    DT_Mdep_printf ("Syntax Error -p <piplein> option\n");
		    DT_Performance_Cmd_Usage ();
		    return (false);
		}
		cmd->pipeline_len = atol (opts->optarg);
		break;
	    }
	    case 'R':		    /* Service Reliability Level */
	    {
		cmd->qos = DT_ParseQoS (opts->optarg);
		break;
	    }
        case 'M':
        {
        cmd->DT_mem_type = DT_ParseMemType (opts->optarg);
        break;
        } 
	    case 's':		    /* server name */
	    {
		if ((opts->optarg == 0) ||
		    strlen (opts->optarg) == 0 ||
		    *opts->optarg == '-')
		{
		    DT_Mdep_printf ("must specify server name\n");
		    DT_Performance_Cmd_Usage ();
		    return (false);
		}

		strncpy (cmd->server_name, opts->optarg, NAME_SZ);
		break;
	    }
	    default:
	    {
		DT_Mdep_printf ("Invalid Performance Test Parameter: %c\n", c);
		DT_Performance_Cmd_Usage ();
		return (false);
	    }
	}
    }

    /*
     * now parse the op
     */
    if ( !DT_Performance_Cmd_Parse_Op (cmd, opts->optind, my_argc, my_argv) )
    {
	DT_Performance_Cmd_Usage ();
	return (false);
    }

    if ( !DT_Performance_Cmd_Validate (cmd) )
    {
	DT_Performance_Cmd_Usage ();
	return (false);
    }

    return (true);
}


bool
DT_Performance_Cmd_Init (Performance_Cmd_t * cmd)
{
    memset (cmd, 0, sizeof (Performance_Cmd_t));
    cmd->dapltest_version = DAPLTEST_VERSION;
    cmd->client_is_little_endian = DT_local_is_little_endian;
    cmd->qos = DAT_QOS_BEST_EFFORT;
    cmd->debug = false;
    cmd->num_iterations = 1000;
    cmd->pipeline_len = ~0;

    cmd->DT_mem_type = DAT_MEM_TYPE_VIRTUAL;
    cmd->op.transfer_type = RDMA_WRITE;
    cmd->op.seg_size = 4096;
    cmd->op.num_segs = 1;

    if ( !DT_Mdep_GetDefaultDeviceName (cmd->dapl_name) )
    {
	DT_Mdep_printf ("can't get default device name\n");
	return (false);
    }

    return (true);
}

void
DT_Performance_Cmd_Print (
    Performance_Cmd_t   *cmd)
{
    DT_Mdep_printf ("-------------------------------------\n");
    DT_Mdep_printf ("PerfCmd.server_name              : %s\n",
                   cmd->server_name);
    DT_Mdep_printf ("PerfCmd.dapl_name                : %s\n",
                   cmd->dapl_name);
    DT_Mdep_printf ("PerfCmd.mode                     : %s\n",
                   (cmd->mode == BLOCKING_MODE) ? "BLOCKING" : "POLLING");
    DT_Mdep_printf ("PerfCmd.num_iterations           : %d\n",
                   cmd->num_iterations);
    DT_Mdep_printf ("PerfCmd.pipeline_len             : %d\n",
                   cmd->pipeline_len);
    DT_Mdep_printf ("PerfCmd.op.transfer_type         : %s\n",
                   cmd->op.transfer_type == RDMA_READ ? "RDMA_READ" :
                   cmd->op.transfer_type == RDMA_WRITE ? "RDMA_WRITE" :
                   "SEND_RECV");
    DT_Mdep_printf ("PerfCmd.op.num_segs              : %d\n",
                   cmd->op.num_segs);
    DT_Mdep_printf ("PerfCmd.op.seg_size              : %d\n",
                   cmd->op.seg_size);
}

const char *
DT_TransferTypeToString (DT_Transfer_Type type)
{
    static char *DT_Type[] =
    {
	"RR",
	"RW",
	"SR"
    };

    if ( (0 <= type) && (type <= 2) )
    {
	return DT_Type[type];
    }
    else
    {
	return "Error: Unkown Transfer Type";
    }
}

double
DT_performance_stats_data_print (Performance_Stats_Data_t 		*data,
                                 double cpu_mhz)
{
    double 				average;

    average = (int64_t)data->total_ts / (data->num * cpu_mhz);

    printf("    Arithmetic mean      : %d.%d us\n"
		   "    maximum              : %d.%d us\n"
		   "    minimum              : %d.%d us\n",
		   DT_whole(average),DT_hundredths(average),
		   DT_whole((int64_t)data->max_ts / cpu_mhz),DT_hundredths((int64_t)data->max_ts / cpu_mhz),
		   DT_whole((int64_t)data->min_ts / cpu_mhz),DT_hundredths((int64_t)data->min_ts / cpu_mhz));
    return average;
}


void
DT_performance_stats_print (
    Performance_Stats_t 		*stats,
    Performance_Cmd_t          	*cmd)
{
    double             	cpu_mhz;
    double          	time_s;
    double          	mbytes;
    double 				ops_per_sec;
    double 				bandwidth;
    double 				latency;
    double 				time_per_post;
    double 				time_per_reap;

    /* calculate CPU utilization */
    unsigned long int system;
	unsigned long int user;
	unsigned long int idle;
	unsigned long int total;
    double			cpu_utilization;
	system = stats->post_cpu_stat.system - stats->pre_cpu_stat.system;
	user = stats->post_cpu_stat.user - stats->pre_cpu_stat.user;
	idle = stats->post_cpu_stat.idle - stats->pre_cpu_stat.idle;

	total = system + user + idle;

	if ( 0 == total )
	{
	    cpu_utilization =  0.0;
	}
	else
	{
	    cpu_utilization = (double)1.0 - ((double) idle / (double) total );
	    cpu_utilization *= 100.0;
	}
    
    cpu_mhz = DT_Mdep_GetCpuMhz ();
    latency = 0;
#if defined(WIN32)
    /*
     * The Microsoft compiler is unable to do a 64 bit conversion when
     * working with double. time_ts is a 64 bit value, so we
     * potentially lose precision, so limit it to the Windows
     * platform. Trying to do the operation below without casting
     * a 64 bit value to an unsigned int results in the error when
     * using Visual C 6.0:
     *
     *    Compiler Error C2520: conversion from unsigned __int64 to
     *    double not implemented, use signed __int64.
     *
     * Note that signed __int64 doesn't work either!
     */
    time_s = (double) ((unsigned int)stats->time_ts / (1000000.0 * cpu_mhz));
#else
    time_s = (double) (stats->time_ts / (1000000.0 * cpu_mhz));
#endif
    mbytes = (double) (1.0 * stats->bytes) / 1024 / 1024;

    if ( 0.0 == time_s )
    {
	printf( "Error determining time\n");
	return;
    }
    else if ( 0 ==  stats->num_ops )
    {
	printf( "Error determining number of operations\n");
	return;
    }
    else if ( 0.0 == cpu_mhz )
    {
	printf( "Error determining CPU speed\n");
	return;
    }

    ops_per_sec = stats->num_ops / time_s;
    bandwidth = mbytes / time_s;
    printf( "------------------------- Statistics -------------------------\n");
    printf(
		   "    Mode                 : %s\n"
		   "    Operation Type       : %s\n"
		   "    Number of Operations : %u\n"
		   "    Segment Size         : %u\n"
		   "    Number of Segments   : %u bytes\n"
		   "    Pipeline Length      : %u\n\n",
		   DT_PerformanceModeToString (cmd->mode),
		   DT_TransferTypeToString (cmd->op.transfer_type),
		   cmd->num_iterations,
		   cmd->op.seg_size,
		   cmd->op.num_segs,
		   stats->pipeline_len);

    printf(
		    "    Total Time           : %d.%d sec\n"
		    "    Total Data Exchanged : %d.%d MB\n"
		    "    CPU Utilization      : %d.%d\n"
		    "    Operation Throughput : %d.%d ops/sec\n"
		    "    Bandwidth            : %d.%d MB/sec\n", 
		    DT_whole(time_s),DT_hundredths(time_s),
		    DT_whole(mbytes),DT_hundredths(mbytes),
		    DT_whole(cpu_utilization),DT_hundredths(cpu_utilization),
		    DT_whole(ops_per_sec),DT_hundredths(ops_per_sec),
		    DT_whole(bandwidth),DT_hundredths(bandwidth));

    printf( "\nLatency\n");

    if ( stats->latency.num )
    {
	latency = DT_performance_stats_data_print (&stats->latency, 
						   cpu_mhz);
    }
    printf( "\n"
		   "Time Per Post\n"
		   "    %u posts without context switches\n",
		   stats->posts_sans_ctxt.num);

    if ( stats->posts_sans_ctxt.num )
    {
	DT_performance_stats_data_print (&stats->posts_sans_ctxt, 
					 cpu_mhz);
    }

    printf( "\n"
		   "    %u posts with context switches\n",
		   stats->posts_with_ctxt.num);

    if ( stats->posts_with_ctxt.num )
    {
	printf( "    %u number of context switches\n",
		       stats->post_ctxt_switch_num);
	DT_performance_stats_data_print (&stats->posts_with_ctxt, 
					 cpu_mhz);
    }

    printf( "\n"
		   "Time Per Reap\n"
		   "    %u reaps without context switches\n",
		   stats->reaps_sans_ctxt.num);

    if ( stats->reaps_sans_ctxt.num )
    {
	DT_performance_stats_data_print (&stats->reaps_sans_ctxt, 
					 cpu_mhz);
    }


    printf( "\n"
		   "    %u reaps with context switches\n",
		   stats->reaps_with_ctxt.num);

    if ( stats->reaps_with_ctxt.num )
    {
	printf( "\n"
		       "    %u number of context switches\n",
		       stats->reap_ctxt_switch_num);

	DT_performance_stats_data_print (&stats->reaps_with_ctxt, 
					 cpu_mhz);
    }

    time_per_post =
	(int64_t) (stats->posts_sans_ctxt.total_ts + stats->posts_with_ctxt.total_ts) /
	(cpu_mhz * (stats->posts_sans_ctxt.num + stats->posts_with_ctxt.num));

    time_per_reap =
	(int64_t) (stats->reaps_sans_ctxt.total_ts + stats->reaps_with_ctxt.total_ts) /
	(cpu_mhz * (stats->reaps_sans_ctxt.num + stats->reaps_with_ctxt.num));

    printf( 
	               "\nNOTE: 1 MB = 1024 KB = 1048576 B \n");
    printf( 
     "---------------------------------------------------------------------\n");
    printf( 
        "raw: %s, %u, %u, %u, %u, %d.%d, %d.%d, %d.%d, %d.%d, %d.%d, %d.%d \n",
		   DT_TransferTypeToString (cmd->op.transfer_type),
		   cmd->num_iterations,
		   cmd->op.seg_size,
		   cmd->op.num_segs,
		   stats->pipeline_len,
		   DT_whole      (cpu_utilization),
		   DT_hundredths (cpu_utilization),
		   DT_whole      (ops_per_sec),
		   DT_hundredths (ops_per_sec),
		   DT_whole      (bandwidth),
		   DT_hundredths (bandwidth),
		   DT_whole      (latency),
		   DT_hundredths (latency),
		   DT_whole      (time_per_post),
		   DT_hundredths (time_per_post),
		   DT_whole      (time_per_reap),
		   DT_hundredths (time_per_reap));
    printf( 
     "---------------------------------------------------------------------\n");
}

