/*
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available at
 * <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
 * license, available in the LICENSE.TXT file accompanying this
 * software.  These details are also available at
 * <http://openib.org/license.html>.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * Copyright (c) 2004 Topspin Communications.  All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

 *
 * $Id: ipoib_arp.c 432 2004-07-12 17:27:37Z roland $
 */

#include "ipoib.h"

#include "ts_kernel_trace.h"
#include "ts_kernel_services.h"

#include "ts_ib_sa_client.h"

#include <linux/slab.h>
#include <linux/if_arp.h>

enum {
	IPOIB_ADDRESS_HASH_BITS = IPOIB_ADDRESS_HASH_BYTES * 8,
};

struct ipoib_sarp_cache {
	struct list_head table[256];
};

struct ipoib_sarp {
	struct list_head cache_list;

	atomic_t refcnt;

	uint8_t hash[IPOIB_ADDRESS_HASH_BYTES];

	tTS_IB_GID gid;
	u32        qpn;
	u16        lid;
	tTS_IB_SL  sl;
	struct ib_address *address_handle;
	tTS_IB_CLIENT_QUERY_TID tid;

	unsigned long created;
	unsigned long last_verify;
	unsigned long last_directed_query;
	unsigned long first_directed_reply;

	unsigned char require_verify:1;
	unsigned char directed_query:1;
	unsigned char directed_reply:4;

	unsigned char logcount;

	struct sk_buff_head pkt_queue;
	u8          sa_query_flag;
	struct net_device *dev;
};

struct ipoib_sarp_iter {
	struct net_device *dev;
	uint8_t hash;
	struct list_head *cur;
};

struct ipoib_arp_payload {
	uint8_t src_hw_addr[IPOIB_HW_ADDR_LEN];
	uint8_t src_ip_addr[4];
	uint8_t dst_hw_addr[IPOIB_HW_ADDR_LEN];
	uint8_t dst_ip_addr[4];
};

/* =============================================================== */
/*.._ipoib_sarp_hash -- hash GID/QPN to 6 bytes                    */
static void _ipoib_sarp_hash(tTS_IB_GID gid, u32 qpn, uint8_t *hash)
{
	/* We use the FNV hash (http://www.isthe.com/chongo/tech/comp/fnv/) */
#define TS_FNV_64_PRIME 0x100000001b3ULL
#define TS_FNV_64_INIT  0xcbf29ce484222325ULL

	int i;
	uint64_t h = TS_FNV_64_INIT;

	/* make qpn big-endian so we know where digits are */
	qpn = cpu_to_be32(qpn);

	for (i = 0; i < sizeof(tTS_IB_GID) + 3; ++i) {
		h *= TS_FNV_64_PRIME;
		h ^= (i < sizeof(tTS_IB_GID)
		      ? gid[i]
		      : ((uint8_t *)&qpn)[i - sizeof(tTS_IB_GID) + 1]);
	}

	/* xor fold down to 6 bytes and make big-endian */
	h = cpu_to_be64((h >> IPOIB_ADDRESS_HASH_BITS)
			^ (h & ((1ULL << IPOIB_ADDRESS_HASH_BITS) - 1)));

	memcpy(hash, ((uint8_t *)&h) + 2, IPOIB_ADDRESS_HASH_BYTES);
}

/* =============================================================== */
/*..ipoib_sarp_get -- increment reference count for ARP entry      */
void ipoib_sarp_get(struct ipoib_sarp *entry)
{
	atomic_inc(&entry->refcnt);
}

/* =============================================================== */
/*.._ipoib_sarp_alloc -- allocate shadow ARP entry                 */
static struct ipoib_sarp *_ipoib_sarp_alloc(struct net_device *dev)
{
	struct ipoib_sarp *entry;

	entry = kmalloc(sizeof(*entry), GFP_ATOMIC);
	if (!entry)
		return NULL;

	atomic_set(&entry->refcnt, 2);	/* The calling function needs to put */

	entry->dev = dev;

	entry->require_verify = 0;
	entry->directed_query = 1;
	entry->directed_reply = 0;

	entry->created = jiffies;
	entry->last_verify = jiffies;
	entry->last_directed_query = jiffies;
	entry->first_directed_reply = jiffies;

	entry->logcount = 0;

	INIT_LIST_HEAD(&entry->cache_list);

	skb_queue_head_init(&entry->pkt_queue);

	entry->address_handle = NULL;
	entry->sa_query_flag = 0;
    
	/* Will force a trigger on the first packet we need to send */
	entry->tid = TS_IB_CLIENT_QUERY_TID_INVALID;

	return entry;
}

/* =============================================================== */
/*..ipoib_sarp_put -- decrement reference count for ARP entry      */
void ipoib_sarp_put(struct ipoib_sarp *entry)
{
	struct net_device *dev = entry->dev;

	if (atomic_dec_and_test(&entry->refcnt)) {
		TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_ARP,
			 "%s: deleting ARP shadow cache entry "
			 "%02x:%02x:%02x:%02x:%02x:%02x",
			 dev->name,
			 entry->hash[0], entry->hash[1], entry->hash[2],
			 entry->hash[3], entry->hash[4], entry->hash[5]);

		if (entry->address_handle != NULL) {
			int ret = ib_address_destroy(entry->address_handle);
			if (ret < 0)
				TS_REPORT_WARN(MOD_IB_NET,
					       "ib_address_destroy failed (ret = %d)",
					       ret);
		}

		while (!skb_queue_empty(&entry->pkt_queue)) {
			struct sk_buff *skb = skb_dequeue(&entry->pkt_queue);

			skb->dev = dev;
			dev_kfree_skb(skb);
		}

		kfree(entry);
	}
}

/* =============================================================== */
/*..__ipoib_sarp_find -- find ARP entry (unlocked)                 */
static struct ipoib_sarp *__ipoib_sarp_find(struct net_device *dev,
					    const uint8_t *hash)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct ipoib_sarp *entry;

	list_for_each_entry(entry, &priv->sarp_cache->table[hash[0]],
			    cache_list) {
		TS_REPORT_DATA(MOD_IB_NET,
			       "%s: matching %02x:%02x:%02x:%02x:%02x:%02x",
			       dev->name,
			       entry->hash[0], entry->hash[1], entry->hash[2],
			       entry->hash[3], entry->hash[4], entry->hash[5]);

		if (!memcmp(hash, entry->hash, IPOIB_ADDRESS_HASH_BYTES)) {
			ipoib_sarp_get(entry);
			return entry;
		}
	}

	return NULL;
}

/* =============================================================== */
/*.._ipoib_sarp_find -- find ARP entry                             */
static struct ipoib_sarp *_ipoib_sarp_find(struct net_device *dev,
					   const uint8_t *hash)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct ipoib_sarp *entry;
	unsigned long flags;

	spin_lock_irqsave(&priv->lock, flags);
	entry = __ipoib_sarp_find(dev, hash);
	spin_unlock_irqrestore(&priv->lock, flags);

	return entry;
}

/* =============================================================== */
/*..ipoib_sarp_iter_init -- create new ARP iterator                */
struct ipoib_sarp_iter *ipoib_sarp_iter_init(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct ipoib_sarp_iter *iter;

	iter = kmalloc(sizeof(*iter), GFP_KERNEL);
	if (!iter)
		return NULL;

	iter->dev = dev;
	iter->hash = 0;
	iter->cur = priv->sarp_cache->table[0].next;

	while (iter->cur == &priv->sarp_cache->table[iter->hash]) {
		++iter->hash;
		if (iter->hash == 0) {
			/* ARP table is empty */
			kfree(iter);
			return NULL;
		}
		iter->cur = priv->sarp_cache->table[iter->hash].next;
	}

	return iter;
}

/* =============================================================== */
/*..ipoib_sarp_iter_free -- free ARP iterator                      */
void ipoib_sarp_iter_free(struct ipoib_sarp_iter *iter)
{
	kfree(iter);
}

/* =============================================================== */
/*..ipoib_sarp_iter_next -- incr. iter. -- return non-zero at end  */
int ipoib_sarp_iter_next(struct ipoib_sarp_iter *iter)
{
	struct ipoib_dev_priv *priv = iter->dev->priv;

	while (1) {
		iter->cur = iter->cur->next;

		if (iter->cur == &priv->sarp_cache->table[iter->hash]) {
			++iter->hash;
			if (!iter->hash)
				return 1;

			iter->cur = &priv->sarp_cache->table[iter->hash];
		} else
			return 0;
	}
}

/* =============================================================== */
/*..ipoib_sarp_iter_read -- get data pointed to by ARP iterator    */
void ipoib_sarp_iter_read(struct ipoib_sarp_iter *iter, uint8_t *hash,
			  tTS_IB_GID gid, u32 *qpn,
			  unsigned long *created, unsigned long *last_verify,
			  unsigned int *queuelen, unsigned int *complete)
{
	struct ipoib_sarp *entry;

	entry = list_entry(iter->cur, struct ipoib_sarp, cache_list);

	memcpy(hash, entry->hash, IPOIB_ADDRESS_HASH_BYTES);
	memcpy(gid, entry->gid, sizeof(tTS_IB_GID));
	*qpn = entry->qpn;
	*created = entry->created;
	*last_verify = entry->last_verify;
	*queuelen = skb_queue_len(&entry->pkt_queue);
	*complete = entry->address_handle != NULL;
}

/* =============================================================== */
/*..ipoib_sarp_add -- add ARP entry                                */
struct ipoib_sarp *ipoib_sarp_add(struct net_device *dev, tTS_IB_GID gid,
				  u32 qpn)
{
	struct ipoib_dev_priv *priv = dev->priv;
	uint8_t hash[IPOIB_ADDRESS_HASH_BYTES];
	struct ipoib_sarp *entry;
	unsigned long flags;

	_ipoib_sarp_hash(gid, qpn, hash);

	entry = _ipoib_sarp_find(dev, hash);
	if (entry) {
		if (entry->qpn != qpn
		    || memcmp(entry->gid, gid, sizeof(tTS_IB_GID))) {
			TS_REPORT_WARN(MOD_IB_NET,
				       "%s: hash collision", dev->name);
			ipoib_sarp_put(entry);	/* for _find() */
			return NULL;
		} else
			return entry;
	}

	entry = _ipoib_sarp_alloc(dev);
	if (!entry) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: out of memory for ARP entry", dev->name);
		return NULL;
	}

	memcpy(entry->hash, hash, sizeof(entry->hash));
	memcpy(entry->gid, gid, sizeof(tTS_IB_GID));

	entry->qpn = qpn;

	entry->require_verify = 1;

	spin_lock_irqsave(&priv->lock, flags);
	list_add_tail(&entry->cache_list, &priv->sarp_cache->table[hash[0]]);
	spin_unlock_irqrestore(&priv->lock, flags);

	return entry;
}

/* =============================================================== */
/*..ipoib_sarp_local_add -- add ARP hash for local node            */
struct ipoib_sarp *ipoib_sarp_local_add(struct net_device *dev,
					tTS_IB_GID gid, u32 qpn)
{
	_ipoib_sarp_hash(gid, qpn, dev->dev_addr);
	return ipoib_sarp_add(dev, gid, qpn);
}

/* =============================================================== */
/*..ipoib_sarp_delete -- delete shadow ARP cache entry             */
int ipoib_sarp_delete(struct net_device *dev, const uint8_t *hash)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct ipoib_sarp *entry;
	unsigned long flags;

	spin_lock_irqsave(&priv->lock, flags);

	entry = __ipoib_sarp_find(dev, hash);
	if (!entry) {
		spin_unlock_irqrestore(&priv->lock, flags);

		return 0;
	}

	list_del_init(&entry->cache_list);

	spin_unlock_irqrestore(&priv->lock, flags);

	ipoib_sarp_put(entry);	/* for _find() */
	ipoib_sarp_put(entry);	/* for original reference */

	return 1;
}

/* =============================================================== */
/*.._ipoib_sarp_path_record_completion -- path record comp func    */
static int _ipoib_sarp_path_record_completion(tTS_IB_CLIENT_QUERY_TID tid,
					      int status,
					      struct ib_path_record *path,
					      int remaining, void *entry_ptr)
{
	struct ipoib_sarp *entry = entry_ptr;
	struct net_device *dev = entry->dev;
	struct ipoib_dev_priv *priv = dev->priv;

	TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_ARP,
		 "%s: path record lookup done, status %d", dev->name, status);

	entry->tid = TS_IB_CLIENT_QUERY_TID_INVALID;
	entry->sa_query_flag=0;

	if (!status) {
		
		struct ib_address_vector av = {
			.dlid = path->dlid,
			.service_level = path->sl,
			.port = priv->port,
			.source_path_bits = 0,
			.use_grh = 0,
			.static_rate = 0
		};
		
		if(entry->address_handle) 
			if (ib_address_destroy(entry->address_handle)) {
				TS_REPORT_WARN(MOD_IB_NET,
					       "%s: ib_address_destroy failed",
					       dev->name);
			}
		
		
		if (ib_address_create(priv->pd, &av, &entry->address_handle)) {
			entry->address_handle = NULL;
			TS_REPORT_WARN(MOD_IB_NET,
				       "%s: ib_address_create failed",
				       dev->name);
		} else {
			TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_ARP,
				 "%s: created address handle 0x%08x for LID 0x%04x, SL %d",
				 dev->name, entry->address_handle, path->dlid,
				 path->sl);
			
			entry->lid = path->dlid;
			entry->sl = path->sl;
		}
		
	} else {
		if (status != -ETIMEDOUT && entry->logcount < 20) {
			TS_REPORT_WARN(MOD_IB_NET,
				       "%s: tsIbPathRecordRequest completion failed for %02x:%02x:%02x:%02x:%02x:%02x, status = %d",
				       dev->name,
				       entry->hash[0], entry->hash[1],
				       entry->hash[2], entry->hash[3],
				       entry->hash[4], entry->hash[5], status);
			entry->logcount++;
		}
		
		
		if (status == -ETIMEDOUT) {
			TS_REPORT_DATA(MOD_IB_NET,
				       "%s: tsIbPathRecordRequest completion timeout. failed for %02x:%02x:%02x:%02x:%02x:%02x, status = %d",
				       dev->name,
				       entry->hash[0], entry->hash[1],
				       entry->hash[2], entry->hash[3],
				       entry->hash[4], entry->hash[5], status);
		}
	
		
	}


	if(entry->address_handle) {
		/* actually send any queued packets */
		while (!skb_queue_empty(&entry->pkt_queue)) {
			struct sk_buff *skb =
				skb_dequeue(&entry->pkt_queue);
			skb->dev = dev;
			if (dev_queue_xmit(skb))
				TS_REPORT_WARN(MOD_IB_NET,
					       "%s: dev_queue_xmit failed to requeue packet",
					       dev->name);
		}
		
	} else {
		
		/* Flush out any queued packets */
		while (!skb_queue_empty(&entry->pkt_queue)) {
			struct sk_buff *skb = skb_dequeue(&entry->pkt_queue);
			
			skb->dev = dev;
			dev_kfree_skb(skb);
		}
	}
	
	

	
	ipoib_sarp_put(entry);	/* for _get() in original call */
	
	/* nonzero return means no more callbacks (we have our path) */
	return 1;
}

/* =============================================================== */
/*.._ipoib_sarp_path_lookup - start path lookup                    */
static int _ipoib_sarp_path_lookup(struct ipoib_sarp *entry)
{
	struct net_device *dev = entry->dev;
	struct ipoib_dev_priv *priv = dev->priv;
	tTS_IB_CLIENT_QUERY_TID tid;

	ipoib_sarp_get(entry);
	if (tsIbPathRecordRequest(priv->ca, priv->port, priv->local_gid,
				  entry->gid, priv->pkey, 0, HZ, 3600 * HZ,	/* XXX cache jiffies */
				  _ipoib_sarp_path_record_completion,
				  entry, &tid)) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: tsIbPathRecordRequest failed", dev->name);
		ipoib_sarp_put(entry);	/* for _get() */
	} else {
		TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_ARP,
			 "%s: no address vector, starting path record lookup",
			 dev->name);
		entry->tid = tid;
	}

	return 0;
}

/* =============================================================== */
/*.._ipoib_sarp_lookup -- return address and qpn for entry         */
static int _ipoib_sarp_lookup(struct ipoib_sarp *entry)
{
	struct net_device *dev = entry->dev;
	
	if (entry->address_handle != NULL && entry->sa_query_flag == 0)               
		return 0;
    


	/*
	 * Found an entry, but without an address handle.
	 * Check to see if we have a path record lookup executing and if not,
	 * start one up.
	 */
	
	if (entry->tid == TS_IB_CLIENT_QUERY_TID_INVALID)
		_ipoib_sarp_path_lookup(entry);
	else {
		TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_ARP,
			 "%s: no address vector, but path record lookup already started",
			 dev->name);
	}

	return -EAGAIN;
}

/* =============================================================== */
/*..ipoib_sarp_lookup -- lookup a hash in shadow ARP cache         */
int ipoib_sarp_lookup(struct net_device *dev, uint8_t *hash,
		      struct ipoib_sarp **entry)
{
	struct ipoib_sarp *tentry;

	tentry = _ipoib_sarp_find(dev, hash);
	if (!tentry)
		return -ENOENT;

	*entry = tentry;

	return _ipoib_sarp_lookup(tentry);
}

/* =============================================================== */
/*.._ipoib_sarp_tx_callback -- put reference to entry after TX     */
static void _ipoib_sarp_tx_callback(void *ptr)
{
	ipoib_sarp_put((struct ipoib_sarp *)ptr); /* for _get() in orig call */
}

/* =============================================================== */
/*..ipoib_sarp_send -- send packet to dest                         */
int ipoib_sarp_send(struct net_device *dev, struct ipoib_sarp *entry,
		    struct sk_buff *skb)
{
	return ipoib_dev_send(dev, skb, _ipoib_sarp_tx_callback,
			      entry, entry->address_handle, entry->qpn);
}

/* =============================================================== */
/*..ipoib_sarp_queue_packet -- queue packet during path rec lookup */
int ipoib_sarp_queue_packet(struct ipoib_sarp *entry, struct sk_buff *skb)
{
	skb_queue_tail(&entry->pkt_queue, skb);

	return 0;
}

/* =============================================================== */
/*..ipoib_sarp_rewrite_receive -- rewrite ARP packet for Linux     */
int ipoib_sarp_rewrite_receive(struct net_device *dev, struct sk_buff *skb)
{
	struct arphdr *arp;
	struct ipoib_arp_payload *payload;
	struct arphdr *new_arp;
	struct ethhdr *header;
	uint8_t *new_payload;
	struct sk_buff *new_skb;
	struct ipoib_sarp *entry;
	uint8_t hash[IPOIB_ADDRESS_HASH_BYTES];
	int ret = 0;

	arp = (struct arphdr *)skb->data;
	payload = (struct ipoib_arp_payload *)skb_pull(skb, sizeof(*arp));

	TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_ARP,
		 "%s: ARP receive: hwtype=0x%04x proto=0x%04x hwlen=%d prlen=%d op=0x%04x "
		 "sip=%d.%d.%d.%d dip=%d.%d.%d.%d",
		 dev->name,
		 ntohs(arp->ar_hrd),
		 ntohs(arp->ar_pro),
		 arp->ar_hln,
		 arp->ar_pln,
		 ntohs(arp->ar_op),
		 payload->src_ip_addr[0], payload->src_ip_addr[1],
		 payload->src_ip_addr[2], payload->src_ip_addr[3],
		 payload->dst_ip_addr[0], payload->dst_ip_addr[1],
		 payload->dst_ip_addr[2], payload->dst_ip_addr[3]);

	new_skb = dev_alloc_skb(dev->hard_header_len
				+ sizeof(*new_arp)
				+ 2 * (IPOIB_ADDRESS_HASH_BYTES + 4));
	if (!new_skb) {
		ret = -ENOMEM;
		goto out;
	}

	new_skb->mac.raw = new_skb->data;
	header = (struct ethhdr *)new_skb->mac.raw;
	skb_reserve(new_skb, dev->hard_header_len);

	new_arp = (struct arphdr *)skb_put(new_skb, sizeof(*new_arp));
	new_payload = (uint8_t *)skb_put(new_skb,
					 2 * (IPOIB_ADDRESS_HASH_BYTES + 4));

	header->h_proto = htons(ETH_P_ARP);

	new_skb->dev = dev;
	new_skb->pkt_type = PACKET_HOST;
	new_skb->protocol = htons(ETH_P_ARP);

	/* copy ARP header */
	*new_arp = *arp;
	new_arp->ar_hrd = htons(ARPHRD_ETHER);
	new_arp->ar_hln = IPOIB_ADDRESS_HASH_BYTES;

	/* copy IP addresses */
	memcpy(new_payload + IPOIB_ADDRESS_HASH_BYTES,
	       payload->src_ip_addr, 4);
	memcpy(new_payload + 2 * IPOIB_ADDRESS_HASH_BYTES + 4,
	       payload->dst_ip_addr, 4);

	/* rewrite IPoIB hw address to hashes */
	if (be32_to_cpu(*(uint32_t *)payload->src_hw_addr) & 0xffffff) {
		_ipoib_sarp_hash(payload->src_hw_addr + 4,
				be32_to_cpu(*(uint32_t *)payload->src_hw_addr) & 0xffffff, hash);

		/* add shadow ARP entries if necessary */
		if (ARPOP_REPLY == ntohs(arp->ar_op)) {
			entry = _ipoib_sarp_find(dev, hash);
			if (entry) {
				if (entry->directed_query &&
				    time_before(jiffies,
						entry->last_directed_query +
						HZ)) {
					/* Directed query, everything's good */
					entry->last_verify = jiffies;
					entry->directed_query = 0;
				} else {
					/*
					 * If we receive another ARP packet in that's not directed and
					 * we already have a path record outstanding, don't drop it yet
					 */
					if (entry->tid ==
					    TS_IB_CLIENT_QUERY_TID_INVALID) {
						/* Delete old one and create a new one */
						TS_TRACE(MOD_IB_NET, T_VERBOSE,
							 TRACE_IB_NET_ARP,
							 "%s: LID change inferred on query for "
							 "%02x:%02x:%02x:%02x:%02x:%02x",
							 dev->name, hash[0],
							 hash[1], hash[2],
							 hash[3], hash[4],
							 hash[5]);

						entry->sa_query_flag = 1;
					} else
						TS_TRACE(MOD_IB_NET, T_VERBOSE,
							 TRACE_IB_NET_ARP,
							 "%s: lookup in progress, skipping destroying entry "
							 "%02x:%02x:%02x:%02x:%02x:%02x",
							 dev->name, hash[0],
							 hash[1], hash[2],
							 hash[3], hash[4],
							 hash[5]);
				}
			}
		} else
			entry = NULL;

		/* Small optimization, if we already found it once, don't search again */
		if (!entry)
			entry = ipoib_sarp_add(dev, payload->src_hw_addr + 4,
					       be32_to_cpu(*(uint32_t *)
							   payload->src_hw_addr) &
							   0xffffff);

		if (ARPOP_REQUEST == ntohs(arp->ar_op)) {
			if (entry && !entry->directed_reply)
				/* Record when this window started */
				entry->first_directed_reply = jiffies;
		}

		if (entry)
			ipoib_sarp_put(entry);	/* for _find() */
	} else
		memset(hash, 0, sizeof(hash));

	memcpy(new_payload, hash, IPOIB_ADDRESS_HASH_BYTES);
	memcpy(header->h_source, hash, sizeof(header->h_source));

	if (be32_to_cpu(*(uint32_t *)payload->dst_hw_addr) & 0xffffff) {
		_ipoib_sarp_hash(payload->dst_hw_addr + 4,
				be32_to_cpu(*(uint32_t *)payload->dst_hw_addr) & 0xffffff, hash);

		entry = ipoib_sarp_add(dev, payload->dst_hw_addr + 4,
				       be32_to_cpu(*(uint32_t *)payload->dst_hw_addr) &
				       0xffffff);
		if (entry)
			ipoib_sarp_put(entry);	/* for _add() */

		memcpy(new_payload + IPOIB_ADDRESS_HASH_BYTES + 4,
		       hash, IPOIB_ADDRESS_HASH_BYTES);
		memcpy(header->h_dest, hash, sizeof(header->h_dest));
	} else {
		memset(new_payload + IPOIB_ADDRESS_HASH_BYTES + 4,
		       0, IPOIB_ADDRESS_HASH_BYTES);
		memset(header->h_dest, 0xff, sizeof(header->h_dest));
	}

	netif_rx_ni(new_skb);

out:
	dev_kfree_skb(skb);
	return ret;
}

/* =============================================================== */
/*..ipoib_sarp_rewrite_send -- rewrite and send ARP packet         */
int ipoib_sarp_rewrite_send(struct net_device *dev, struct sk_buff *skb)
{
	struct ipoib_dev_priv *priv = dev->priv;
	unsigned char broadcast_mac_addr[] =
	    { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
	struct sk_buff *new_skb;
	struct arphdr *arp = (struct arphdr *)(skb->data + ETH_HLEN);
	uint8_t *payload = ((uint8_t *)arp) + sizeof(*arp);
	struct arphdr *new_arp;
	struct ipoib_arp_payload *new_payload;
	struct ipoib_sarp *dentry = NULL, *entry;
	struct ipoib_mcast *dmcast = NULL;
	int ret;

	TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_ARP,
		 "%s: ARP send: hwtype=0x%04x proto=0x%04x hwlen=%d prlen=%d op=0x%04x "
		 "sip=%d.%d.%d.%d dip=%d.%d.%d.%d",
		 dev->name,
		 ntohs(arp->ar_hrd),
		 ntohs(arp->ar_pro),
		 arp->ar_hln,
		 arp->ar_pln,
		 ntohs(arp->ar_op),
		 payload[arp->ar_hln], payload[arp->ar_hln + 1],
		 payload[arp->ar_hln + 2], payload[arp->ar_hln + 3],
		 payload[2 * arp->ar_hln + 4], payload[2 * arp->ar_hln + 5],
		 payload[2 * arp->ar_hln + 6], payload[2 * arp->ar_hln + 7]);

	if (memcmp(broadcast_mac_addr, skb->data, ETH_ALEN) == 0) {
		/* Broadcast gets handled differently */
		ret = ipoib_mcast_lookup(dev, priv->bcast_gid, &dmcast);

		/* mcast is only valid if we get a return code of 0 or -EAGAIN */
		switch (ret) {
		case 0:
			break;
		case -EAGAIN:
			ipoib_mcast_queue_packet(dmcast, skb);
			ipoib_mcast_put(dmcast);	/* for _lookup() */
			return 0;
		default:
			TS_REPORT_WARN(MOD_IB_NET,
				       "%s: dropping ARP packet with unknown dest "
				       "%02x:%02x:%02x:%02x:%02x:%02x",
				       dev->name,
				       skb->data[0], skb->data[1],
				       skb->data[2], skb->data[3],
				       skb->data[4], skb->data[5]);
			return 1;
		}
	} else {
		dentry = _ipoib_sarp_find(dev, skb->data);
		if (!dentry) {
			TS_REPORT_WARN(MOD_IB_NET,
				       "%s: dropping ARP packet with unknown dest "
				       "%02x:%02x:%02x:%02x:%02x:%02x",
				       dev->name,
				       skb->data[0], skb->data[1],
				       skb->data[2], skb->data[3],
				       skb->data[4], skb->data[5]);
			return 1;
		}

		/* Make sure we catch any LID changes */

		/* Update the entry to mark that we last sent a directed ARP query */
		if (dentry->require_verify
		    && dentry->address_handle != NULL) {
			if (ARPOP_REQUEST == ntohs(arp->ar_op)) {
				dentry->directed_query = 1;
				dentry->last_directed_query = jiffies;
			}

			/*
			 * Catch a LID change on the remote end. If we reply to 3 or more
			 * ARP queries without a reply, then ditch the entry we have and
			 * requery
			 */
			if (ARPOP_REPLY == ntohs(arp->ar_op)) {
				dentry->directed_reply++;

				if (!time_before(jiffies,
				     dentry->first_directed_reply + 4 * HZ)) {
					/* We're outside of the time window, so restart the counter */
					dentry->directed_reply = 0;
				} else if (dentry->directed_reply > 3) {
					if (dentry->tid ==
					    TS_IB_CLIENT_QUERY_TID_INVALID) {
						/* Delete old one and create a new one */
						TS_TRACE(MOD_IB_NET, T_VERBOSE,
							 TRACE_IB_NET_ARP,
							 "LID change inferred on reply for "
							 "%02x:%02x:%02x:%02x:%02x:%02x",
							 dentry->hash[0],
							 dentry->hash[1],
							 dentry->hash[2],
							 dentry->hash[3],
							 dentry->hash[4],
							 dentry->hash[5]);

						ipoib_sarp_delete(dev,
								  dentry->hash);
						entry = ipoib_sarp_add(dev,
								       dentry->gid,
								       dentry->qpn);
						if (NULL == entry) {
							TS_TRACE(MOD_IB_NET,
								 T_VERBOSE,
								 TRACE_IB_NET_ARP,
								 "could not allocate new entry for %02x:%02x:%02x:%02x:%02x:%02x",
								 dentry->hash[0],
								 dentry->hash[1],
								 dentry->hash[2],
								 dentry->hash[3],
								 dentry->hash[4],
								 dentry->hash[5]);
							ipoib_sarp_put(dentry); /* for _find() */
							return 1;
						}

						ipoib_sarp_put(dentry);	/* for _find() */

						dentry = entry;
					} else
						TS_TRACE(MOD_IB_NET, T_VERBOSE,
							 TRACE_IB_NET_ARP,
							 "lookup in progress, skipping destroying entry %02x:%02x:%02x:%02x:%02x:%02x",
							 dentry->hash[0],
							 dentry->hash[1],
							 dentry->hash[2],
							 dentry->hash[3],
							 dentry->hash[4],
							 dentry->hash[5]);
				}
			}
		}

		ret = _ipoib_sarp_lookup(dentry);
		if (ret == -EAGAIN) {
			ipoib_sarp_queue_packet(dentry, skb);
			ipoib_sarp_put(dentry);		/* for _find() */
			return 0;
		}
	}

	new_skb = dev_alloc_skb(dev->hard_header_len
				+ sizeof(*new_arp) + sizeof(*new_payload));
	if (!new_skb) {
		if (dentry)
			ipoib_sarp_put(dentry);		/* for _find() */
		if (dmcast)
			ipoib_mcast_put(dmcast);	/* for _lookup() */

		return 1;
	}
	skb_reserve(new_skb, dev->hard_header_len);

	new_arp = (struct arphdr *)skb_put(new_skb, sizeof(*new_arp));
	new_payload = (struct ipoib_arp_payload *)skb_put(new_skb,
							  sizeof(*new_payload));

	/* build what we need for the header */
	{
		uint16_t *t;

		/* ethertype */
		t = (uint16_t *)skb_push(new_skb, 2);
		*t = htons(ETH_P_ARP);

		/* leave space so send funct can skip ethernet addrs */
		skb_push(new_skb, IPOIB_ADDRESS_HASH_BYTES * 2);
	}

	/* copy ARP header */
	*new_arp = *arp;
	new_arp->ar_hrd = htons(ARPHRD_INFINIBAND);
	new_arp->ar_hln = IPOIB_HW_ADDR_LEN;

	/* copy IP addresses */
	memcpy(&new_payload->src_ip_addr, payload + arp->ar_hln, 4);
	memcpy(&new_payload->dst_ip_addr, payload + 2 * arp->ar_hln + 4, 4);

	/* rewrite hash to IPoIB hw address */
	entry = _ipoib_sarp_find(dev, payload);
	if (!entry) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: can't find hw address for hash "
			       "%02x:%02x:%02x:%02x:%02x:%02x",
			       dev->name,
			       payload[0], payload[1], payload[2],
			       payload[3], payload[4], payload[5]);
		memset(new_payload->src_hw_addr, 0, IPOIB_HW_ADDR_LEN);
	} else {
		*((uint32_t *)new_payload->src_hw_addr) =
		    cpu_to_be32(entry->qpn);
		memcpy(&new_payload->src_hw_addr[4], entry->gid,
		       sizeof(tTS_IB_GID));
		ipoib_sarp_put(entry);	/* for _find() */
	}

	if (memcmp(broadcast_mac_addr, payload + IPOIB_ADDRESS_HASH_BYTES + 4,
		   ETH_ALEN) == 0) {
		*((uint32_t *)new_payload->dst_hw_addr) =
		    cpu_to_be32(IB_MULTICAST_QPN);
		memcpy(&new_payload->dst_hw_addr[4], priv->bcast_gid,
		       sizeof(tTS_IB_GID));
	} else {
		entry = _ipoib_sarp_find(dev, payload +
					 IPOIB_ADDRESS_HASH_BYTES + 4);
		if (!entry)
			memset(new_payload->dst_hw_addr, 0, IPOIB_HW_ADDR_LEN);
		else {
			*((uint32_t *)new_payload->dst_hw_addr) =
			    cpu_to_be32(entry->qpn);
			memcpy(&new_payload->dst_hw_addr[4], entry->gid,
			       sizeof(tTS_IB_GID));
			ipoib_sarp_put(entry);	/* for _find() */
		}
	}

	dev_kfree_skb(skb);

	if (dmcast)
		ipoib_mcast_send(dev, dmcast, new_skb);
	else
		ipoib_sarp_send(dev, dentry, new_skb);

	return 0;
}

/* =============================================================== */
/*..ipoib_sarp_dev_init -- initialize ARP cache                    */
int ipoib_sarp_dev_init(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	int i;

	priv->sarp_cache = kmalloc(sizeof(*priv->sarp_cache), GFP_KERNEL);
	if (!priv->sarp_cache)
		return -ENOMEM;

	for (i = 0; i < 256; ++i)
		INIT_LIST_HEAD(&priv->sarp_cache->table[i]);

	return 0;
}

/* =============================================================== */
/*..ipoib_sarp_dev_flush -- flush ARP cache                        */
void ipoib_sarp_dev_flush(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct ipoib_sarp *entry, *tentry;
	LIST_HEAD(delete_list);
	unsigned long flags;
	int i;

	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_ARP,
		 "%s: flushing shadow ARP cache", dev->name);

	/*
	 * We move to delete_list first because putting the reference could
	 * eventually end up blocking and we're in a spinlock
	 */

	/*
	 * Instead of destroying the address vector, we destroy the entire
	 * entry, but we create a new empty entry before that. This way we
	 * don't have any races with freeing a used address vector.
	 */
	spin_lock_irqsave(&priv->lock, flags);
	for (i = 0; i < 256; ++i) {
		list_for_each_entry_safe(entry, tentry,
					 &priv->sarp_cache->table[i],
					 cache_list) {
			struct ipoib_sarp *nentry;

			/*
			 * Allocation failure isn't fatal, just drop the entry.
			 * If it's important, a new one will be generated later
			 * automatically.
			 */
			nentry = _ipoib_sarp_alloc(entry->dev);
			if (nentry) {
				memcpy(nentry->hash, entry->hash,
				       sizeof(nentry->hash));
				memcpy(nentry->gid, entry->gid,
				       sizeof(tTS_IB_GID));

				nentry->require_verify = entry->require_verify;
				nentry->qpn = entry->qpn;

				/* Add it before the current entry */
				list_add_tail(&nentry->cache_list,
					      &entry->cache_list);

				ipoib_sarp_put(nentry);	/* for _alloc() */
			}

			list_del(&entry->cache_list);
			list_add_tail(&entry->cache_list, &delete_list);
		}
	}
	spin_unlock_irqrestore(&priv->lock, flags);

	list_for_each_entry_safe(entry, tentry, &delete_list, cache_list) {
		list_del_init(&entry->cache_list);
		ipoib_sarp_put(entry);	/* for original reference */
	}
}

/* =============================================================== */
/*..ipoib_sarp_dev_destroy -- destroy ARP cache                    */
static void ipoib_sarp_dev_destroy(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct ipoib_sarp *entry, *tentry;
	LIST_HEAD(delete_list);
	unsigned long flags;
	int i;

	/*
	 * We move to delete_list first because putting the reference could
	 * eventually end up blocking and we're in a spinlock
	 */
	spin_lock_irqsave(&priv->lock, flags);
	for (i = 0; i < 256; ++i) {
		list_for_each_entry_safe(entry, tentry,
					 &priv->sarp_cache->table[i],
					 cache_list) {
			list_del(&entry->cache_list);
			list_add_tail(&entry->cache_list, &delete_list);
		}
	}
	spin_unlock_irqrestore(&priv->lock, flags);

	list_for_each_entry_safe(entry, tentry, &delete_list, cache_list) {
		list_del_init(&entry->cache_list);
		ipoib_sarp_put(entry);	/* for original reference */
	}
}

/* =============================================================== */
/*..ipoib_sarp_dev_cleanup -- clean up ARP cache                   */
void ipoib_sarp_dev_cleanup(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	TS_REPORT_CLEANUP(MOD_IB_NET, "%s: cleaning up ARP table", dev->name);

	ipoib_sarp_dev_destroy(dev);

	kfree(priv->sarp_cache);
}

/* =============================================================== */
/*..ipoib_get_gid -- find a hash in shadow ARP cache               */
int ipoib_get_gid(struct net_device *dev, uint8_t *hash, tTS_IB_GID gid)
{
	struct ipoib_sarp *entry = _ipoib_sarp_find(dev, hash);

	if (!entry)
		return -EINVAL;

	memcpy(gid, entry->gid, sizeof(tTS_IB_GID));

	ipoib_sarp_put(entry);	/* for _find() */

	return 0;
}

/*
 * Local Variables:
 * c-file-style: "linux"
 * indent-tabs-mode: t
 * End:
 */

