/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: ipoib_main.c 428 2004-07-12 04:54:43Z roland $
*/

#include "ipoib.h"

#include "ts_kernel_services.h"
#include "ts_kernel_trace.h"
#include "openib_version.h"

#include <linux/version.h>
#include <linux/module.h>

#include <linux/init.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>

#include <linux/if_arp.h>	/* For ARPHRD_xxx */

#include <linux/ip.h>
#include <linux/in.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/moduleparam.h>
#endif

MODULE_AUTHOR("Roland Dreier");
MODULE_DESCRIPTION("IP-over-InfiniBand net driver" OPENIB_REVISION);
MODULE_LICENSE("Dual BSD/GPL");

DECLARE_MUTEX(ipoib_device_mutex);
LIST_HEAD(ipoib_device_list);

extern tTS_IB_GID broadcast_mgid;

/////Module Param/////////////
int ipoib_max_address_vector = 2048;
int ipoib_tx_ring_size = 1024;
int ipoib_use_interrupt_context = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
module_param(ipoib_max_address_vector, int, 0444);
module_param(ipoib_tx_ring_size, int, 0444);
module_param(ipoib_use_interrupt_context, int, 0444);
#else
MODULE_PARM(ipoib_max_address_vector,"i");
MODULE_PARM(ipoib_tx_ring_size,"i");
MODULE_PARM(ipoib_use_interrupt_context,"i");
#endif
MODULE_PARM_DESC(ipoib_max_address_vector,"The max address vector that will be allowed by the vapi");
MODULE_PARM_DESC(ipoib_tx_ring_size,"The Number of tx ring");
MODULE_PARM_DESC(ipoib_use_interrupt_context,"When non-zero, use interrupt context");
/////End of Module Param//////

static const uint8_t broadcast_mac_addr[] = {
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};

/* This is based on the dev_change_flags from later kernels */
int ipoib_change_flags(struct net_device *dev, unsigned flags)
{
	int ret;
	int old_flags = dev->flags;

	/* Set the flags on our device */
	dev->flags = (flags & (IFF_DEBUG | IFF_NOTRAILERS | IFF_NOARP |
                               IFF_DYNAMIC | IFF_MULTICAST | IFF_PORTSEL |
                               IFF_AUTOMEDIA))
		   | (dev->flags & (IFF_UP | IFF_VOLATILE | IFF_PROMISC |
				    IFF_ALLMULTI));

	/* Load in the correct multicast list now the flags have changed */
	dev_mc_upload(dev);

	/*
	 * Have we downed the interface. We handle IFF_UP ourselves according
	 * to user attempts to set it, rather than blindly setting it.
	 */
	ret = 0;
	if ((old_flags ^ flags) & IFF_UP) {	/* Bit is different  ? */
		ret = ((old_flags & IFF_UP) ? dev_close : dev_open) (dev);

		if (ret == 0)
			dev_mc_upload(dev);
	}

#if 0
	/* We don't have access to netdev_chain from this module */
	if (dev->flags & IFF_UP &&
	    ((old_flags ^ dev->flags) &
	     ~(IFF_UP | IFF_PROMISC | IFF_ALLMULTI | IFF_VOLATILE)))
		notifier_call_chain(&netdev_chain, NETDEV_CHANGE, dev);
#endif

	if ((flags ^ dev->gflags) & IFF_PROMISC) {
		int inc = (flags & IFF_PROMISC) ? +1 : -1;

		dev->gflags ^= IFF_PROMISC;
		dev_set_promiscuity(dev, inc);
	}

	/*
	 * NOTE: order of synchronization of IFF_PROMISC and IFF_ALLMULTI
	 * is important. Some (broken) drivers set IFF_PROMISC, when
	 * IFF_ALLMULTI is requested not asking us and not reporting.
	 */
	if ((flags ^ dev->gflags) & IFF_ALLMULTI) {
		int inc = (flags & IFF_ALLMULTI) ? +1 : -1;

		dev->gflags ^= IFF_ALLMULTI;
		dev_set_allmulti(dev, inc);
	}
#if 0
	/* rtmsg_ifinfo isn't exported on all kernels we need to support */
	if (old_flags ^ dev->flags)
		rtmsg_ifinfo(RTM_NEWLINK, dev, old_flags ^ dev->flags);
#endif

	return ret;
}

int ipoib_device_handle(struct net_device *dev, struct ib_device **ca,
			tTS_IB_PORT *port, tTS_IB_GID gid, u16 *pkey)
{
	struct ipoib_dev_priv *priv = dev->priv;

    if (!priv) {
        return -1;
    }
	*ca = priv->ca;
	*port = priv->port;
	memcpy(gid, priv->local_gid, sizeof(tTS_IB_GID));
	*pkey = priv->pkey;

	return 0;
}

int ipoib_dev_open(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_FLOW_CONFIG,
		 "%s: bringing up interface", dev->name);

	set_bit(IPOIB_FLAG_ADMIN_UP, &priv->flags);

	if (ipoib_pkey_dev_delay_open(dev))
		return 0;

	if (ipoib_ib_dev_open(dev))
		return -EINVAL;

	if (ipoib_ib_dev_up(dev))
		return -EINVAL;

	if (!test_bit(IPOIB_FLAG_SUBINTERFACE, &priv->flags)) {
		struct ipoib_dev_priv *cpriv;

		/* Bring up any child interfaces too */
		down(&ipoib_device_mutex);
		list_for_each_entry(cpriv, &priv->child_intfs, list) {
			int flags;

			flags = cpriv->dev.flags;
			if (flags & IFF_UP)
				continue;

			ipoib_change_flags(&cpriv->dev, flags | IFF_UP);
		}
		up(&ipoib_device_mutex);
	}

	netif_start_queue(dev);

	return 0;
}

static int _ipoib_dev_stop(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_FLOW_CONFIG,
		 "%s: stopping interface", dev->name);

	clear_bit(IPOIB_FLAG_ADMIN_UP, &priv->flags);

	netif_stop_queue(dev);

	ipoib_ib_dev_down(dev);
	ipoib_ib_dev_stop(dev);

	if (!test_bit(IPOIB_FLAG_SUBINTERFACE, &priv->flags)) {
		struct ipoib_dev_priv *cpriv;

		/* Bring down any child interfaces too */
		down(&ipoib_device_mutex);
		list_for_each_entry(cpriv, &priv->child_intfs, list) {
			int flags;

			flags = cpriv->dev.flags;
			if (!(flags & IFF_UP))
				continue;

			ipoib_change_flags(&cpriv->dev, flags & ~IFF_UP);
		}
		up(&ipoib_device_mutex);
	}

	return 0;
}

static int _ipoib_dev_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	TS_REPORT_INOUT(MOD_IB_NET, "%s: ioctl 0x%04x", dev->name, cmd);

	/* Start with SIOCDEVPRIVATE + 3 so we don't conflict with any of
	   the standard MII ioctls */

	switch (cmd) {
	default:
		return -ENOTSUPP;
	}
}

static int _ipoib_dev_change_mtu(struct net_device *dev, int new_mtu)
{
	struct ipoib_dev_priv *priv = dev->priv;

	if (new_mtu > IPOIB_PACKET_SIZE - IPOIB_ENCAP_LEN)
		return -EINVAL;

	priv->admin_mtu = new_mtu;

	dev->mtu = min(priv->mcast_mtu, priv->admin_mtu);

	return 0;
}

static int _ipoib_dev_set_config(struct net_device *dev, struct ifmap *map)
{
	return -EOPNOTSUPP;
}

static int _ipoib_dev_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	uint16_t ethertype;
	int ret;

	ethertype = ntohs(((struct ethhdr *)skb->data)->h_proto);

	TS_REPORT_DATA(MOD_IB_NET,
		       "%s: packet to transmit, length=%d ethertype=0x%04x",
		       dev->name, skb->len, ethertype);

	if (!netif_carrier_ok(dev)) {
		TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_GEN,
			 "%s: dropping packet since fabric is not up",
			 dev->name);

		dev->trans_start = jiffies;
		++priv->stats.tx_packets;
		priv->stats.tx_bytes += skb->len;
		dev_kfree_skb(skb);
		return 0;
	}

	switch (ethertype) {
	case ETH_P_ARP:
		if (ipoib_sarp_rewrite_send(dev, skb)) {
			++priv->stats.tx_dropped;
			dev_kfree_skb(skb);
		}
		return 0;

	case ETH_P_IP:
		if (skb->data[0] == 0x01 && skb->data[1] == 0x00
		    && skb->data[2] == 0x5e
		    && (skb->data[3] & 0x80) == 0x00) {
			/* Multicast MAC addr */
			struct ipoib_mcast *mcast = NULL;
			tTS_IB_GID mgid;
			struct iphdr *iph =
			    (struct iphdr *)(skb->data + ETH_HLEN);
			u32 multiaddr = ntohl(iph->daddr);


			memcpy(mgid, ipoib_broadcast_mgid,
			       sizeof(tTS_IB_GID));

			/* Add in the P_Key */
			mgid[4] = (priv->pkey >> 8) & 0xff;
			mgid[5] = priv->pkey & 0xff;

			/* Fixup the group mapping */
			mgid[12] = (multiaddr >> 24) & 0x0f;
			mgid[13] = (multiaddr >> 16) & 0xff;
			mgid[14] = (multiaddr >> 8) & 0xff;
			mgid[15] = multiaddr & 0xff;


            ret = ipoib_mcast_lookup(dev, mgid, &mcast);
			switch (ret) {
			case 0:
				return ipoib_mcast_send(dev, mcast, skb);
			case -EAGAIN:
				ipoib_mcast_queue_packet(mcast, skb);
				ipoib_mcast_put(mcast);
				return 0;
			}
		} else
		    if (memcmp(broadcast_mac_addr, skb->data, ETH_ALEN) == 0) {
			struct ipoib_mcast *mcast = NULL;


			ret = ipoib_mcast_lookup(dev, priv->bcast_gid, &mcast);
			switch (ret) {
			case 0:
				return ipoib_mcast_send(dev, mcast, skb);
			case -EAGAIN:
				ipoib_mcast_queue_packet(mcast, skb);
				ipoib_mcast_put(mcast);
				return 0;
			}
		} else {
			struct ipoib_sarp *entry = NULL;

			ret = ipoib_sarp_lookup(dev, skb->data, &entry);
			switch (ret) {
			case 0:
				return ipoib_sarp_send(dev, entry, skb);
			case -EAGAIN:
				ipoib_sarp_queue_packet(entry, skb);
				ipoib_sarp_put(entry);
				return 0;
			}
		}

		switch (ret) {
		case 0:
		case -EAGAIN:
			/* Shouldn't get here anyway */
			break;
		case -ENOENT:
			TS_REPORT_WARN(MOD_IB_NET,
				       "%s: dropping packet with unknown dest "
				       "%02x:%02x:%02x:%02x:%02x:%02x",
				       dev->name,
				       skb->data[0], skb->data[1],
				       skb->data[2], skb->data[3],
				       skb->data[4], skb->data[5]);
			++priv->stats.tx_dropped;
			dev_kfree_skb(skb);
			return 0;
		default:
			TS_REPORT_WARN(MOD_IB_NET,
				       "%s: sending to %02x:%02x:%02x:%02x:%02x:%02x "
				       "failed (ret = %d)",
				       dev->name,
				       skb->data[0], skb->data[1],
				       skb->data[2], skb->data[3],
				       skb->data[4], skb->data[5], ret);
			++priv->stats.tx_dropped;
			dev_kfree_skb(skb);
			return 0;
		}
		return 0;

	case ETH_P_IPV6:
		TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_GEN,
			 "%s: dropping IPv6 packet", dev->name);
		++priv->stats.tx_dropped;
		dev_kfree_skb(skb);
		return 0;

	default:
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: dropping packet with unknown ethertype 0x%04x",
			       dev->name, ethertype);
		++priv->stats.tx_dropped;
		dev_kfree_skb(skb);
		return 0;
	}

	return 0;
}

struct net_device_stats *_ipoib_dev_get_stats(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	return &priv->stats;
}

static void _ipoib_dev_timeout(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	if (priv->tx_free && !test_bit(IPOIB_FLAG_TIMEOUT, &priv->flags)) {
		char ring[IPOIB_TX_RING_SIZE + 1];
		int i;

		for (i = 0; i < IPOIB_TX_RING_SIZE; ++i)
			ring[i] = priv->tx_ring[i].skb ? 'X' : '.';

		ring[i] = 0;

		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: transmit timeout: latency %ld, tx_free %d, tx_ring [%s]",
			       dev->name, jiffies - dev->trans_start,
			       priv->tx_free, ring);

		set_bit(IPOIB_FLAG_TIMEOUT, &priv->flags);
	} else
		TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_GEN,
			 "%s: transmit timeout: latency %ld",
			 dev->name, jiffies - dev->trans_start);
}

/*
 * Setup the packet to look like ethernet here, we'll fix it later when
 * we actually send it to look like an IPoIB packet
 */
static int _ipoib_dev_hard_header(struct sk_buff *skb,
				  struct net_device *dev,
				  unsigned short type,
				  void *daddr, void *saddr, unsigned len)
{
	struct ethhdr *header = (struct ethhdr *)skb_push(skb, ETH_HLEN);

	TS_REPORT_DATA(MOD_IB_NET,
		       "%s: building header, ethertype=0x%04x",
		       dev->name, type);

	if (daddr)
		memcpy(header->h_dest, daddr, IPOIB_ADDRESS_HASH_BYTES);

	if (saddr)
		memcpy(header->h_source, saddr, IPOIB_ADDRESS_HASH_BYTES);
	else
		memcpy(header->h_source, dev->dev_addr,
		       IPOIB_ADDRESS_HASH_BYTES);

	header->h_proto = htons(type);

	return 0;
}

static void _ipoib_dev_set_mcast_list(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	schedule_task(&priv->restart_task);
}

int ipoib_dev_init(struct net_device *dev, struct ib_device *ca, int port)
{
	struct ipoib_dev_priv *priv = dev->priv;

	TS_REPORT_INOUT(MOD_IB_NET, "%s: initializing device", dev->name);

	dev->open = ipoib_dev_open;
	dev->stop = _ipoib_dev_stop;
	dev->do_ioctl = _ipoib_dev_ioctl;
	dev->change_mtu = _ipoib_dev_change_mtu;
	dev->set_config = _ipoib_dev_set_config;
	dev->hard_start_xmit = _ipoib_dev_xmit;
	dev->get_stats = _ipoib_dev_get_stats;
	dev->tx_timeout = _ipoib_dev_timeout;
	dev->hard_header = _ipoib_dev_hard_header;
	dev->set_multicast_list = _ipoib_dev_set_mcast_list;
	dev->watchdog_timeo = HZ;

	dev->rebuild_header = 0;
	dev->set_mac_address = 0;
	dev->header_cache_update = 0;

	dev->flags |= IFF_BROADCAST | IFF_MULTICAST;

	dev->hard_header_len = ETH_HLEN;
	dev->addr_len = IPOIB_ADDRESS_HASH_BYTES;
	dev->type = ARPHRD_ETHER;
	dev->tx_queue_len = IPOIB_TX_RING_SIZE * 2;
	/* MTU will be reset when mcast join happens */
	dev->mtu = IPOIB_PACKET_SIZE - IPOIB_ENCAP_LEN;
	priv->mcast_mtu = priv->admin_mtu = dev->mtu;

	memset(dev->broadcast, 0xff, dev->addr_len);

	netif_carrier_off(dev);

	SET_MODULE_OWNER(dev);

	spin_lock_init(&priv->lock);
	spin_lock_init(&priv->rq_lock);

	if (ipoib_sarp_dev_init(dev))
		goto out;

	/* Allocate RX/TX "rings" to hold queued skbs */

	priv->rx_ring =
		kmalloc(IPOIB_RX_RING_SIZE * sizeof (struct ipoib_rx_buf),
			GFP_KERNEL);
	if (!priv->rx_ring) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: failed to allocate RX ring (%d entries)",
				dev->name, IPOIB_RX_RING_SIZE);
		goto out_arp_cleanup;
	}
	memset(priv->rx_ring, 0,
	       IPOIB_RX_RING_SIZE * sizeof (struct ipoib_rx_buf));

	priv->tx_ring =
	    kmalloc(IPOIB_TX_RING_SIZE * sizeof(struct ipoib_tx_buf),
		    GFP_KERNEL);
	if (!priv->tx_ring) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: failed to allocate TX ring (%d entries)",
				dev->name, IPOIB_TX_RING_SIZE);
		goto out_rx_ring_cleanup;
	}
	memset(priv->tx_ring, 0,
	       IPOIB_TX_RING_SIZE * sizeof(struct ipoib_tx_buf));

	/* set up the rest of our private data */
	priv->tx_last_wri_handel = IPOIB_TX_RING_SIZE - 1;
	priv->tx_hanging = 0;
	priv->rx_last_post_receive = IPOIB_RX_RING_SIZE - 1;
	priv->rx_last_used = IPOIB_RX_RING_SIZE - 1;


	/* priv->tx_head is already 0 */
	priv->tx_free = IPOIB_TX_RING_SIZE;
        
	if (ipoib_ib_dev_init(dev, ca, port))
		goto out_tx_ring_cleanup;

	if (ipoib_proc_dev_init(dev))
		goto out_ib_cleanup;

	return 0;

out_ib_cleanup:
	ipoib_ib_dev_cleanup(dev);

out_tx_ring_cleanup:
	kfree(priv->tx_ring);

out_rx_ring_cleanup:
	kfree(priv->rx_ring);

out_arp_cleanup:
	ipoib_sarp_dev_cleanup(dev);

out:
	return -ENOMEM;
}

void ipoib_dev_cleanup(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv, *cpriv, *tcpriv;
	int i;

	/* Delete any child interfaces first */
	/* Safe since it's either protected by ipoib_device_mutex or empty */
	list_for_each_entry_safe(cpriv, tcpriv, &priv->child_intfs, list) {
		ipoib_dev_cleanup(&cpriv->dev);
		unregister_netdev(&cpriv->dev);

		list_del(&cpriv->list);

		kfree(cpriv);
	}

	ipoib_proc_dev_cleanup(dev);
	ipoib_ib_dev_cleanup(dev);
	ipoib_sarp_dev_cleanup(dev);

	if (priv->rx_ring) {
		TS_REPORT_CLEANUP(MOD_IB_NET,
				  "%s: cleaning up RX ring", dev->name);
		
		for (i = 0; i < IPOIB_RX_RING_SIZE; ++i) {
			if (priv->rx_ring[i].skb && (priv->rx_ring[i].skb_not_alloc == 0)){
				dev_kfree_skb(priv->rx_ring[i].skb);
				priv->rx_ring[i].skb_not_alloc = 1;
			}
		}
		
		kfree(priv->rx_ring);
		priv->rx_ring = NULL;
	}

	if (priv->tx_ring) {
		TS_REPORT_CLEANUP(MOD_IB_NET,
				  "%s: cleaning up TX ring", dev->name);

		for (i = 0; i < IPOIB_TX_RING_SIZE; ++i) {
			if (priv->tx_ring[i].skb)
				dev_kfree_skb(priv->tx_ring[i].skb);
		}

		kfree(priv->tx_ring);
		priv->tx_ring = NULL;
	}
}

struct ipoib_dev_priv *ipoib_intf_alloc(void)
{
	struct ipoib_dev_priv *priv;

	priv = kmalloc(sizeof(*priv), GFP_KERNEL);
	if (!priv) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"failed to allocate private struct");
		return NULL;
	}

	memset(priv, 0, sizeof(*priv));

	sema_init(&priv->mcast_mutex, 1);

	atomic_set(&priv->mcast_joins, 0);

	INIT_LIST_HEAD(&priv->child_intfs);
	INIT_LIST_HEAD(&priv->multicast_list);

	INIT_TQUEUE(&priv->restart_task, ipoib_mcast_restart_task, &priv->dev);

	priv->dev.priv = priv;

	return priv;
}

int ipoib_add_port(const char *format, struct ib_device *hca, tTS_IB_PORT port)
{
	struct ipoib_dev_priv *priv;
	int result = -ENOMEM;

	priv = ipoib_intf_alloc();
	if (!priv)
		goto alloc_mem_failed;

	priv->pkey = 0xffff;

#if 0
	/* We'll probably use something like this in the future */
	result = ib_pkey_entry_get(hca, port, 0, &priv->pkey);
	if (result) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: ib_pkey_entry_get failed (ret = %d)",
				priv->dev.name, result);
		goto dev_pkey_get_failed;
	}
#endif

	result = dev_alloc_name(&priv->dev, format);
	if (result < 0) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"failed to get device name (ret = %d)", result);
		goto dev_alloc_failed;
	}

	result = ipoib_dev_init(&priv->dev, hca, port);
	if (result < 0) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"failed to initialize net device %d, port %d (ret = %d)",
				hca, port, result);
		goto device_init_failed;
	}

	result = ipoib_port_monitor_dev_start(&priv->dev);
	if (result < 0) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"failed to setup port monitor for device %d, "
				"port %d (ret = %d)",
				hca, port, result);
		goto port_monitor_failed;
	}

	result = register_netdev(&priv->dev);
	if (result) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: failed to initialize; error %i",
				priv->dev.name, result);
		goto register_failed;
	}

	down(&ipoib_device_mutex);
	list_add_tail(&priv->list, &ipoib_device_list);
	up(&ipoib_device_mutex);

	return 0;

register_failed:
	ipoib_port_monitor_dev_stop(&priv->dev);

port_monitor_failed:
	ipoib_dev_cleanup(&priv->dev);

device_init_failed:
	/*
	 * Nothing to do since the device name only gets finally added
	 * to the linked list in register_netdev
	 */

dev_alloc_failed:
	kfree(priv);

alloc_mem_failed:
	return result;
}

static int __init ipoib_init_module(void)
{
	int ret;

	/* add ipoib module to the list of modules
	   that should be notified when a new ib device
	   is created */
	ret = ipoib_transport_create_devices();
	if (ret)
		return ret;

	down(&ipoib_device_mutex);
	if (list_empty(&ipoib_device_list)) {
		up(&ipoib_device_mutex);
		ipoib_transport_cleanup();
		return -ENODEV;
	}
	up(&ipoib_device_mutex);

	ipoib_vlan_init();

	return 0;
}

static void __exit ipoib_cleanup_module(void)
{
	struct ipoib_dev_priv *priv, *tpriv;

	ipoib_vlan_cleanup();
	ipoib_transport_cleanup();

	down(&ipoib_device_mutex);
	list_for_each_entry_safe(priv, tpriv, &ipoib_device_list, list) {
		ipoib_ib_finish_completion(&priv->dev);
		ipoib_port_monitor_dev_stop(&priv->dev);
		ipoib_dev_cleanup(&priv->dev);
		unregister_netdev(&priv->dev);
		ipoib_mcast_dev_exit(&priv->dev);
		
		list_del(&priv->list);
		
		kfree(priv);
	}
	up(&ipoib_device_mutex);
}

module_init(ipoib_init_module);
module_exit(ipoib_cleanup_module);

/*
  Local Variables:
  c-file-style: "linux"
  indent-tabs-mode: t
  End:
*/

