/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: ipoib_main.c 428 2004-07-12 04:54:43Z roland $
*/

#include "ipoib.h"

#include "ts_kernel_services.h"
#include "ts_kernel_trace.h"
#include "ip2pr_export.h"
#include "openib_version.h"

#include <linux/version.h>
#include <linux/module.h>

#include <linux/init.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>

#include <linux/if_arp.h>	/* For ARPHRD_xxx */

#include <linux/ip.h>
#include <linux/in.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/moduleparam.h>
#endif
MODULE_AUTHOR("Roland Dreier");
MODULE_DESCRIPTION("IP-over-InfiniBand net driver" OPENIB_REVISION);
MODULE_LICENSE("Dual BSD/GPL");

DECLARE_MUTEX(ipoib_device_mutex);
LIST_HEAD(ipoib_device_list);

extern tTS_IB_GID broadcast_mgid;

/////Module Param////////////
int ipoib_max_address_vector = 2048;
int ipoib_tx_ring_size = 1024;
int ipoib_use_interrupt_context = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
module_param(ipoib_max_address_vector, int, 0444);
module_param(ipoib_tx_ring_size, int, 0444);
module_param(ipoib_use_interrupt_context, int, 0444);
#else
MODULE_PARM(ipoib_max_address_vector,"i");
MODULE_PARM(ipoib_tx_ring_size,"i");
MODULE_PARM(ipoib_use_interrupt_context,"i");
#endif
MODULE_PARM_DESC(ipoib_max_address_vector,"The max address vector that will be allowed by the vapi");
MODULE_PARM_DESC(ipoib_tx_ring_size,"The Number of tx ring");
MODULE_PARM_DESC(ipoib_use_interrupt_context,"When non-zero, use interrupt context");
/////End of Module Param/////


static const u8 ipv4_bcast_addr[] = {
	0x00, 0xff, 0xff, 0xff,
	0xff, 0x12, 0x40, 0x1b,	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,	0xff, 0xff, 0xff, 0xff
};


/* This is based on the dev_change_flags from later kernels */
int ipoib_change_flags(struct net_device *dev, unsigned flags)
{
	int ret;
	int old_flags = dev->flags;

	/* Set the flags on our device */
	dev->flags = (flags & (IFF_DEBUG | IFF_NOTRAILERS | IFF_NOARP |
                               IFF_DYNAMIC | IFF_MULTICAST | IFF_PORTSEL |
                               IFF_AUTOMEDIA))
                                | (dev->flags & (IFF_UP | IFF_VOLATILE | IFF_PROMISC |
                                                 IFF_ALLMULTI));

	/* Load in the correct multicast list now the flags have changed */
	dev_mc_upload(dev);

	/*
	 * Have we downed the interface. We handle IFF_UP ourselves according
	 * to user attempts to set it, rather than blindly setting it.
	 */
	ret = 0;
	if ((old_flags ^ flags) & IFF_UP) {	/* Bit is different  ? */
		ret = ((old_flags & IFF_UP) ? dev_close : dev_open) (dev);

		if (ret == 0)
			dev_mc_upload(dev);
	}

#if 0
	/* We don't have access to netdev_chain from this module */
	if (dev->flags & IFF_UP &&
	    ((old_flags ^ dev->flags) &
	     ~(IFF_UP | IFF_PROMISC | IFF_ALLMULTI | IFF_VOLATILE)))
		notifier_call_chain(&netdev_chain, NETDEV_CHANGE, dev);
#endif

	if ((flags ^ dev->gflags) & IFF_PROMISC) {
		int inc = (flags & IFF_PROMISC) ? +1 : -1;

		dev->gflags ^= IFF_PROMISC;
		dev_set_promiscuity(dev, inc);
	}

	/*
	 * NOTE: order of synchronization of IFF_PROMISC and IFF_ALLMULTI
	 * is important. Some (broken) drivers set IFF_PROMISC, when
	 * IFF_ALLMULTI is requested not asking us and not reporting.
	 */
	if ((flags ^ dev->gflags) & IFF_ALLMULTI) {
		int inc = (flags & IFF_ALLMULTI) ? +1 : -1;

		dev->gflags ^= IFF_ALLMULTI;
		dev_set_allmulti(dev, inc);
	}
#if 0
	/* rtmsg_ifinfo isn't exported on all kernels we need to support */
	if (old_flags ^ dev->flags)
		rtmsg_ifinfo(RTM_NEWLINK, dev, old_flags ^ dev->flags);
#endif

	return ret;
}

int ipoib_device_handle(struct net_device *dev, struct ib_device **ca,
			tTS_IB_PORT *port, tTS_IB_GID gid, u16 *pkey)
{
	struct ipoib_dev_priv *priv = dev->priv;

    if (!priv) {
        return -1;
    }
	*ca = priv->ca;
	*port = priv->port;

#ifdef MAC_20_BYTES
    memcpy(gid, priv->local_gid.raw, sizeof (union ib_gid));
#else
    memcpy(gid, priv->local_gid, sizeof(tTS_IB_GID));
#endif

    *pkey = priv->pkey;

	return 0;
}

int ipoib_dev_open(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_FLOW_CONFIG,
		 "%s: bringing up interface", dev->name);

	set_bit(IPOIB_FLAG_ADMIN_UP, &priv->flags);

	if (ipoib_pkey_dev_delay_open(dev))
		return 0;

	if (ipoib_ib_dev_open(dev))
		return -EINVAL;

	if (ipoib_ib_dev_up(dev))
		return -EINVAL;

	if (!test_bit(IPOIB_FLAG_SUBINTERFACE, &priv->flags)) {
		struct ipoib_dev_priv *cpriv;

		/* Bring up any child interfaces too */
		down(&ipoib_device_mutex);
		list_for_each_entry(cpriv, &priv->child_intfs, list) {
			int flags;

			flags = cpriv->dev.flags;
			if (flags & IFF_UP)
				continue;

			ipoib_change_flags(&cpriv->dev, flags | IFF_UP);
		}
		up(&ipoib_device_mutex);
	}

	netif_start_queue(dev);

	return 0;
}

static int _ipoib_dev_stop(struct net_device *dev)
{
	struct ipoib_dev_priv *priv;
	
	if (!dev)
	{    
		TS_REPORT_WARN(MOD_IB_NET, 
			       "_ipoib_dev_stop was called with NULL pointer"); 
		
		return -EINVAL;
	}   
	priv = dev->priv;
	
	if (!priv)
	{    
		TS_REPORT_WARN(MOD_IB_NET, 
			       "_ipoib_dev_stop was called while priv is not declared"); 
		
		return -EINVAL;
	}   

	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_FLOW_CONFIG,
		 "%s: stopping interface", dev->name);

	clear_bit(IPOIB_FLAG_ADMIN_UP, &priv->flags);

	netif_stop_queue(dev);
 
	ipoib_ib_dev_down(dev);
	ipoib_ib_dev_stop(dev);
   

	if (!test_bit(IPOIB_FLAG_SUBINTERFACE, &priv->flags)) {
		struct ipoib_dev_priv *cpriv;

		/* Bring down any child interfaces too */
		down(&ipoib_device_mutex);
		list_for_each_entry(cpriv, &priv->child_intfs, list) {
			int flags;

			flags = cpriv->dev.flags;
			if (!(flags & IFF_UP))
				continue;

			ipoib_change_flags(&cpriv->dev, flags & ~IFF_UP);
		}
		up(&ipoib_device_mutex);
	}

	return 0;
}

static int _ipoib_dev_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	TS_REPORT_INOUT(MOD_IB_NET, "%s: ioctl 0x%04x", dev->name, cmd);

	/* Start with SIOCDEVPRIVATE + 3 so we don't conflict with any of
	   the standard MII ioctls */

	switch (cmd) {
	default:
		return -ENOTSUPP;
	}
}

static int _ipoib_dev_change_mtu(struct net_device *dev, int new_mtu)
{
	struct ipoib_dev_priv *priv = dev->priv;

	if (new_mtu > IPOIB_PACKET_SIZE - IPOIB_ENCAP_LEN)
		return -EINVAL;

	priv->admin_mtu = new_mtu;

	dev->mtu = min(priv->mcast_mtu, priv->admin_mtu);

	return 0;
}
// TBD - if still needed
static int _ipoib_dev_set_config(struct net_device *dev, struct ifmap *map)
{
	return -EOPNOTSUPP;
}

////////Destroy Address Handle that we don't need any more///////////////////
static void ipoib_neigh_destructor(struct neighbour *neigh)
{
	
	struct ipoib_path     *dpath = IPOIB_PATH(neigh);
	
	
	if (dpath)
	{
		struct ipoib_dev_priv *priv = dpath->dev->priv;
		struct net_device *dev = dpath->dev;
		if (!priv || list_empty(&priv->created_ahs))
		{
			TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_ARP,
				 "Created ahs List is empty return");
			return;
		}
		
		TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_GEN,
			 "Neigh_destructor for %06x the Gid is %x" IPOIB_GID_FMT "\n",
			 be32_to_cpup((u32 *) neigh->ha),
                 IPOIB_GID_ARG(*((union ib_gid *) (neigh->ha + 4))));
		
		//Remove from the created ahs list
		TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_ARP,
			 "ipoib_neigh_destructor the ah amout is %x the max is %x",priv->ah_amount , ipoib_max_address_vector); 
		if (dpath->ah && (priv->ah_amount > ipoib_max_address_vector/2)) {
			ipoib_free_ah(dpath);
			kfree(dpath);
			ipoib_delete_unused_ah(dev);
		} else {
			dpath->unused_ah = 1;
		}
	} else {   
		TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_ARP,
			 "ipoib_neigh_destructor is called with unvalid path");
	}
}

//////////////////////////////////////////////////////////////////////
////////////////////ARP METHODS///////////////////////////////////////
//////////////////////////////////////////////////////////////////////

static int path_rec_completion(tTS_IB_CLIENT_QUERY_TID tid,
			       int status,
			       struct ib_path_record *pathrec,
			       int remaining, void *path_ptr)
{
	struct ipoib_path *path = path_ptr;
	struct ipoib_path *old_path;
	struct net_device *dev = path->dev;
	struct ipoib_dev_priv *priv = dev->priv;
	unsigned long flags;
	struct ib_address_vector av;
	u8 hwaddr[IPOIB_HW_ADDR_LEN];
	memcpy(hwaddr + 4, path->gid.raw, sizeof (union ib_gid));
	hwaddr[3] = (u8) (path->qpn      ) & 0xff;  
	hwaddr[2] = (u8) (path->qpn >> 8 ) & 0xff;
	hwaddr[1] = (u8) (path->qpn >> 16) & 0xff;  
	hwaddr[0] = (u8) (path->qpn >> 24) & 0xff;

	old_path = ipoib_get_path(dev, hwaddr, path->ipaddr);
  
	if (status)
	{	
		if (!old_path || ipoib_reproduce_ah(path, old_path))
			goto err;          
		
	} else {
		if (old_path) {
			ipoib_free_ah(old_path);  
			kfree(old_path);
		}
		av.dlid = pathrec->dlid;
		av.service_level = pathrec->sl;
		av.port = priv->port; 
		av.source_path_bits = 0; 
		av.use_grh = 0; 
		av.static_rate = 0; 
		if (ib_address_create(priv->pd, &av, &path->ah)) { 
			TS_REPORT_WARN(MOD_IB_NET, 
				       "%s: ib_address_create failed", dev->name); 
			goto err; 
		} else { 
			spin_lock_irqsave(&priv->lock, flags);
			++priv->ah_amount; 
			spin_unlock_irqrestore(&priv->lock, flags); 
			TS_TRACE(MOD_IB_NET,   T_VERY_VERBOSE, TRACE_IB_NET_ARP, 
				 "%s: created address handle 0x%08x for LID 0x%04x, SL %d", 
				 dev->name, path->ah, pathrec->dlid, pathrec->sl);
		}
	}
	if (path->ah)
	{
		spin_lock_irqsave(&priv->lock, flags); 
		list_add_tail(&path->list, &priv->created_ahs); 
		spin_unlock_irqrestore(&priv->lock, flags); 
		TS_TRACE(MOD_IB_NET,   T_VERY_VERBOSE, TRACE_IB_NET_ARP, 
                 "%s: path_rec_completion finished (succeed or reproduce) all the packets will be sent", 
                 dev->name); 
		
		/* actually send any queued packets */
		while (!skb_queue_empty(&path->queue)) { 
			struct sk_buff *skb = skb_dequeue(&path->queue); 
			skb->dev = dev; 
			if (dev_queue_xmit(skb))
				TS_REPORT_WARN(MOD_IB_NET, 
					       "%s: dev_queue_xmit failed to requeue packet", 
					       dev->name);  
		}
		return 1; 
	}

    
err:
    TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_ARP, 
             "%s: path_rec_completion fail and no old AH were found packets will be dropped", 
             dev->name); 

	spin_lock(&priv->lock); 	
	path->tid =  TS_IB_CLIENT_QUERY_TID_INVALID;   
	spin_unlock(&priv->lock); 
	/* Flush out any queued packets */
	while (!skb_queue_empty(&path->queue)) {
		struct sk_buff *skb = skb_dequeue(&path->queue);
		
		skb->dev = dev;
		dev_kfree_skb(skb);
	}
	
	return 1;
}
 
static void path_rec_start_with_path(struct sk_buff *skb, struct net_device *dev, struct ipoib_path *path)
 {
	 struct ipoib_dev_priv *priv = dev->priv;
	 tTS_IB_CLIENT_QUERY_TID tid;
	 
	 spin_lock(&priv->lock); 
	 if (path->tid != TS_IB_CLIENT_QUERY_TID_INVALID)
	 {
		 spin_unlock(&priv->lock); 
		 return;
	 }
	 spin_unlock(&priv->lock); 
	 
	 /*
	  * XXX there's a race here if path record completion runs
	  * before we get to finish up.  Add a lock to path struct?
	  */
	 
	 if (tsIbPathRecordRequest(priv->ca, priv->port,
				   priv->local_gid.raw,
				   skb->dst->neighbour->ha + 4,
				   priv->pkey, 0, HZ, 0,
				   path_rec_completion,
				   path, &tid)) {
		 TS_REPORT_WARN(MOD_IB_NET,
				"%s: tsIbPathRecordRequest failed", dev->name);
	 }
	 path->tid = tid;
	 
	 return;
	 
 }

static int path_rec_start(struct sk_buff *skb, struct net_device *dev)
 {
	 struct ipoib_dev_priv *priv = dev->priv;
	 struct ipoib_path *path = kmalloc(sizeof *path, GFP_ATOMIC);
	 tTS_IB_CLIENT_QUERY_TID tid;
     
     struct iphdr *iph =
         (struct iphdr *)(skb->data + IPOIB_ENCAP_LEN);
            
  	 if (!path)
		 goto err;
	 
	 path->ah  = NULL;
	 path->qpn = be32_to_cpup((u32 *) skb->dst->neighbour->ha);
	 path->dev = dev;
	 path->unused_ah = 0;
	 path->get_counter = 0;
	 path->ipaddr = ntohl(iph->daddr);;
	 path->tid = TS_IB_CLIENT_QUERY_TID_INVALID;
     
	 TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_ARP,
		  "path_rec_start to ip %x", path->ipaddr);
 
	 skb_queue_head_init(&path->queue);
	 __skb_queue_tail(&path->queue, skb);
	 path->neighbour = NULL;
	 
	 
	 memcpy(path->gid.raw, skb->dst->neighbour->ha + 4, sizeof (union ib_gid));
   	 /*
	  * XXX there's a race here if path record completion runs
	  * before we get to finish up.  Add a lock to path struct?
	  */
	 
	 if (tsIbPathRecordRequest(priv->ca, priv->port,
				   priv->local_gid.raw,
				   skb->dst->neighbour->ha + 4,
				   priv->pkey, 0, HZ, 0,
				   path_rec_completion,
				   path, &tid)) {
		 TS_REPORT_WARN(MOD_IB_NET,
				"%s: tsIbPathRecordRequest failed", dev->name);
		 goto err;
	 }
	 path->tid = tid;
	 
	 path->neighbour = skb->dst->neighbour;
	 
	 IPOIB_PATH(skb->dst->neighbour) = path;
	 return 0;
	 
 err:
	 kfree(path);
	 ++priv->stats.tx_dropped;
	dev_kfree_skb_any(skb);
	
	return 0;
 }
static int unicast_arp_completion(tTS_IB_CLIENT_QUERY_TID tid,
				  int status,
				  struct ib_path_record *pathrec,
				  int remaining, void *skb_ptr)
{
	struct sk_buff *skb = skb_ptr;
        struct ib_address   *ah;
        struct net_device *dev = skb->dev;
        struct ipoib_dev_priv *priv = dev->priv;
	
        struct ib_address_vector av;
        if (status)
		goto err;
	
	
        av.dlid = pathrec->dlid;
        av.service_level = pathrec->sl;
        av.port = priv->port;
        av.source_path_bits = 0;
        av.use_grh = 0;
        av.static_rate = 0;
        
	
        if (ib_address_create(priv->pd, &av, &ah)) {
            TS_REPORT_WARN(MOD_IB_NET,
                           "%s: ib_address_create failed",
                           dev->name);
            goto err;
        } else {
            spin_lock_irq(&priv->lock);
            ++priv->ah_amount; 
            spin_unlock_irq(&priv->lock);

        }
	
        *(struct ib_address **) skb->cb = ah;
        TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_ARP,
                 "%s: created address handle 0x%08x for LID 0x%04x, SL %d",
                 dev->name, ah, pathrec->dlid,
                 pathrec->sl);       
	
        if (dev_queue_xmit(skb))
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: dev_queue_xmit failed to requeue ARP packet",
			       dev->name);
	
    
        return 1;
	
 err:
	dev_kfree_skb(skb);
	return 1;
}
static void unicast_arp_finish(struct sk_buff *skb)
{
    struct ib_address *ah = *(struct ib_address **) skb->cb;
    struct net_device *dev = skb->dev;
    struct ipoib_dev_priv *priv = dev->priv;

    TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_ARP,
             "unicast_arp_finish called delete ah %p",
             ah);

	if (ah) {
		if (ib_address_destroy(ah) < 0)   {
			TS_REPORT_WARN(MOD_IB_NET,
				       "%s: ib_address_destroy failed",
				       dev->name);
		}
		else {
			spin_lock(&priv->lock);
			--priv->ah_amount;
			spin_unlock(&priv->lock);
		}
    }
	
}
/*
 * For unicast packets with no skb->dst->neighbour (unicast ARPs are
 * the main example), we fire off a path record query for each packet.
 * This is pretty bad for scalability (since this is going to hammer
 * the SM on a big fabric) but it's the best I can think of for now.
 *
 * Also we might have a problem if a path changes, because ARPs will
 * still go through (since we'll get the new path from the SM for
 * these queries) so we'll never update the neighbour.
 */
static int unicast_arp_start(struct sk_buff *skb, struct net_device *dev,
			     struct ipoib_pseudoheader *phdr)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct sk_buff *tmp_skb;
	tTS_IB_CLIENT_QUERY_TID tid;

	if (skb->destructor) {
		tmp_skb = skb;
		skb = skb_clone(tmp_skb, GFP_ATOMIC);
		dev_kfree_skb_any(tmp_skb);
		if (!skb) {
 			++priv->stats.tx_dropped;
 		}
	}
 
	skb->dev        = dev;
	skb->destructor = unicast_arp_finish;
	memset(skb->cb, 0, sizeof skb->cb);
 
    
	/*
	 * XXX We need to keep a record of the skb and TID somewhere
	 * so that we can cancel the request if the device goes down
	 * before it finishes.
	 */
	if (tsIbPathRecordRequest(priv->ca, priv->port,
				  priv->local_gid.raw,
				  phdr->hwaddr + 4,
				  priv->pkey, 0, HZ, 0,
				  unicast_arp_completion,
				  skb, &tid)) {
        
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: tsIbPathRecordRequest failed", dev->name);
		++priv->stats.tx_dropped;
 		dev_kfree_skb_any(skb);
		return 0;
 	}
 
 	return 0;
}

static int ipoib_neigh_setup(struct neighbour *neigh)
{
	/*
	 * Is this kosher?  I can't find anybody in the kernel that
	 * sets neigh->destructor, so we should be able to set it here
	 * without trouble.
	 */
	neigh->ops->destructor = ipoib_neigh_destructor;

	return 0;
}

static int ipoib_neigh_setup_dev(struct net_device *dev, struct neigh_parms *parms)
{
	parms->neigh_setup = ipoib_neigh_setup;
    
	return 0;
}



//////////////////////////////////////////////////////////////////////
////////////////////END ARP METHODS///////////////////////////////////////
//////////////////////////////////////////////////////////////////////
 
static int _ipoib_dev_xmit(struct sk_buff *skb, struct net_device *dev)
{
    //struct ipoib_dev_priv *priv = netdev_priv(dev);
    struct ipoib_dev_priv *priv = dev->priv;
    struct ipoib_path *path;

    if (skb->dst && skb->dst->neighbour) {
	if (skb->dst->neighbour->ops && !skb->dst->neighbour->ops->destructor) {
		struct ipoib_neigh_ops *ipoib_ops;
		unsigned long flags;
	       	ipoib_ops = kmalloc(sizeof *ipoib_ops, GFP_ATOMIC);
		if (!ipoib_ops) {
			TS_REPORT_WARN(MOD_IB_NET,
				       "Cant alloc ipoib_neigh_ipoib_ops\n");
			goto err;
		}
		ipoib_ops->ops = skb->dst->neighbour->ops;
		spin_lock_irqsave(&priv->lock, flags);
		ipoib_ops->ops->destructor = ipoib_neigh_destructor;
		list_add_tail(&ipoib_ops->list, &priv->ops_list);
		spin_unlock_irqrestore(&priv->lock, flags);
	}
        if (unlikely(skb->dst->neighbour->ha[4] == 0xff)) {
            struct iphdr *iph =
		    (struct iphdr *)(skb->data + IPOIB_ENCAP_LEN);
            
            u32 multiaddr = ntohl(iph->daddr);
            
            /* Add in the P_Key */
            skb->dst->neighbour->ha[8] = (priv->pkey >> 8) & 0xff;
            skb->dst->neighbour->ha[9] = priv->pkey & 0xff;
            
            /* Add in the multicast group */  
            /* Fixup the group mapping */
            if (multiaddr != 0xffffffff) {
		    skb->dst->neighbour->ha[19] = multiaddr & 0xff;
		    skb->dst->neighbour->ha[18] = (multiaddr >> 8) & 0xff;
		    skb->dst->neighbour->ha[17] = (multiaddr >> 16) & 0xff;
		    skb->dst->neighbour->ha[16] = (multiaddr >> 24) & 0x0f;
	    }
                       
            ipoib_mcast_send(dev, (union ib_gid *) (skb->dst->neighbour->ha + 4),skb);
            
            if ((multiaddr == 0xe0000016) && (priv->mcast_update_thread))  {

               TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_ARP,
                               "A join packet was transmitted");
               priv->mcast_rejoin = 1;
               wake_up_interruptible(&priv->mcast_update_wait);
            }

            return 0;
        }

        if (unlikely(!IPOIB_PATH(skb->dst->neighbour))) {
            struct iphdr *iph =
                (struct iphdr *)(skb->data + IPOIB_ENCAP_LEN);
            u32 ipaddr = ntohl(iph->daddr);
            path = ipoib_path_look_up(dev, (u8 *) skb->dst->neighbour->ha, ipaddr);
            
            if (path) {
                // Probably after new multicast ARP - new entry in the linux cache
                path->get_counter +=1;
                path->neighbour = skb->dst->neighbour;
                IPOIB_PATH(skb->dst->neighbour) = path;
            }
            else 
                return path_rec_start(skb, dev);
		
        }
	
        path = IPOIB_PATH(skb->dst->neighbour);

	if (memcmp(path->gid.raw , skb->dst->neighbour->ha + 4,
		   IPOIB_HW_ADDR_LEN - 4)) {
		ipoib_free_ah(path);
		kfree(path);
		IPOIB_PATH(skb->dst->neighbour) = NULL;

		return path_rec_start(skb, dev);
	}

	path->qpn = be32_to_cpup((u32 *) skb->dst->neighbour->ha);

        if (likely(path->ah)) {
            ipoib_send(dev, skb, path->ah, path->qpn);
            return 0;
        }  else {
            path_rec_start_with_path(skb ,dev ,path);
        }

        if (skb_queue_len(&path->queue) < IPOIB_MAX_PATH_REC_QUEUE)
            __skb_queue_tail(&path->queue, skb);
        else
            goto err;
    } else {
            struct ipoib_pseudoheader *phdr =
                (struct ipoib_pseudoheader *) skb->data;
	    // This is not an ARP packet
	    if ((phdr->hwaddr[5] != 0x12) &&
            (phdr->hwaddr[6] != 0x40) &&
            (phdr->hwaddr[4] == 0xff)) {
		    
            u8  hwaddr[IPOIB_HW_ADDR_LEN - 4];
            struct ipoib_dummy_header *poolbuf =
               (struct ipoib_dummy_header *) skb->data;
            phdr->hwaddr[10] = 8; //IPV4 packet
            phdr->hwaddr[12] = 0;
            skb_pull(skb, sizeof *poolbuf);
		    memcpy(hwaddr, broadcast_mgid, IPOIB_HW_ADDR_LEN - 4);
		    hwaddr[4] = (priv->pkey >> 8) & 0xff;
		    hwaddr[5] = priv->pkey & 0xff;
		    ipoib_mcast_send(dev, (union ib_gid *) hwaddr, skb); 
		    return 0;
	    }
	    
	    skb_pull(skb, sizeof *phdr);
	    if (phdr->hwaddr[4] == 0xff) {
		    /* Add in the P_Key */
		    phdr->hwaddr[8] = (priv->pkey >> 8) & 0xff;
		    phdr->hwaddr[9] = priv->pkey & 0xff;
		    
		    ipoib_mcast_send(dev, (union ib_gid *) (phdr->hwaddr + 4), skb);
            }  else  {
		    /* unicast GID -- ARP reply?? */
		    
		    /*
		     * If destructor is unicast_arp_finish, we've
		     * already been through the path lookup and
		     * now we can just send the packet.
		     */
		    if (skb->destructor == unicast_arp_finish) {
			    ipoib_send(dev, skb, *(struct ib_address **) skb->cb,
				       be32_to_cpup((u32 *) phdr->hwaddr));
			    
			    return 0;
		    }
		    
		    if (be16_to_cpup((u16 *) skb->data) != ETH_P_ARP)
			    TS_REPORT_WARN(MOD_IB_NET,"Unicast, no %s: type %04x, QPN %06x "
					   IPOIB_GID_FMT "\n",
					   skb->dst ? "neigh" : "dst",
					   be16_to_cpup((u16 *) skb->data),
					   be32_to_cpup((u32 *) phdr->hwaddr),
					   phdr->hwaddr[ 4], phdr->hwaddr[ 5],
					   phdr->hwaddr[ 6], phdr->hwaddr[ 7],
					   phdr->hwaddr[ 8], phdr->hwaddr[ 9],
					   phdr->hwaddr[10], phdr->hwaddr[11],
					   phdr->hwaddr[12], phdr->hwaddr[13],
					   phdr->hwaddr[14], phdr->hwaddr[15],
					   phdr->hwaddr[16], phdr->hwaddr[17],
					   phdr->hwaddr[18], phdr->hwaddr[19]);
		    
           			  
		    path = ipoib_get_path(dev, phdr->hwaddr, 0);
		    if  (path && likely(path->ah))
		    {// Sending unicast arp
			    TS_TRACE(MOD_IB_NET, T_VERBOSE, MOD_IB_NET,
				     "%s: Send Unicast ARP",
				     dev->name);
			    ipoib_send(dev, skb, path->ah, path->qpn);
			    return 0;
		    } else 
		    {
			    /* put the pseudoheader back on */
			    skb_push(skb, sizeof *phdr);
			    return unicast_arp_start(skb, dev, phdr);
		    }
	    }
    }

    return 0;

err:
    ++priv->stats.tx_dropped;
    dev_kfree_skb_any(skb);

    return 0;
}

struct net_device_stats *_ipoib_dev_get_stats(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	return &priv->stats;
}

static void _ipoib_dev_timeout(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	if (priv->tx_free && !test_bit(IPOIB_FLAG_TIMEOUT, &priv->flags)) {
		char ring[IPOIB_TX_RING_SIZE + 1];
		int i;

		for (i = 0; i < IPOIB_TX_RING_SIZE; ++i)
			ring[i] = priv->tx_ring[i].skb ? 'X' : '.';

		ring[i] = 0;

		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: transmit timeout: latency %ld, tx_free %d, tx_ring [%s]",
			       dev->name, jiffies - dev->trans_start,
			       priv->tx_free, ring);

		set_bit(IPOIB_FLAG_TIMEOUT, &priv->flags);
	} else
		TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_GEN,
			 "%s: transmit timeout: latency %ld",
			 dev->name, jiffies - dev->trans_start);
}

static int _ipoib_dev_hard_header(struct sk_buff *skb,
				  struct net_device *dev,
				  unsigned short type,
				  void *daddr, void *saddr, unsigned len)
{
   	struct ipoib_header *header;
 
	header = (struct ipoib_header *) skb_push(skb, sizeof *header);
 
	header->proto = htons(type);
	header->reserved = 0;

	/*
	 * If we don't have a neighbour structure, stuff the
	 * destination address onto the front of the skb so we can
	 * figure out where to send the packet later.
	 */
	if (!skb->dst || !skb->dst->neighbour) {
		struct ipoib_pseudoheader *phdr =
			(struct ipoib_pseudoheader *) skb_push(skb, sizeof *phdr);
		memcpy(phdr->hwaddr, daddr, IPOIB_HW_ADDR_LEN);
	}

 	return 0;
}

static void _ipoib_dev_set_mcast_list(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	schedule_task(&priv->restart_task);
}

int ipoib_dev_init(struct net_device *dev, struct ib_device *ca, int port)
{
	struct ipoib_dev_priv *priv = dev->priv;

	TS_REPORT_INOUT(MOD_IB_NET, "%s: initializing device", dev->name);

	dev->open = ipoib_dev_open;
	dev->stop = _ipoib_dev_stop;
	dev->do_ioctl = _ipoib_dev_ioctl;
	dev->change_mtu = _ipoib_dev_change_mtu;
	dev->set_config = _ipoib_dev_set_config;
	dev->hard_start_xmit = _ipoib_dev_xmit;
	dev->get_stats = _ipoib_dev_get_stats;
	dev->tx_timeout = _ipoib_dev_timeout;
	dev->hard_header = _ipoib_dev_hard_header;
	dev->set_multicast_list = _ipoib_dev_set_mcast_list;
	dev->watchdog_timeo = HZ;
	// To destroy unused address vector
	if ( 0 ) {
		dev->neigh_setup         = ipoib_neigh_setup_dev;
	}
	else
		dev->neigh_setup = NULL;

	dev->rebuild_header = NULL;
	dev->set_mac_address = NULL;
	dev->header_cache_update = NULL;

	dev->flags |= IFF_BROADCAST | IFF_MULTICAST;

	dev->hard_header_len = IPOIB_ENCAP_LEN + IPOIB_HW_ADDR_LEN;
	dev->addr_len = IPOIB_HW_ADDR_LEN;
	dev->type = ARPHRD_INFINIBAND;
	dev->tx_queue_len = IPOIB_TX_RING_SIZE * 2;

	/* MTU will be reset when mcast join happens */
	dev->mtu = IPOIB_PACKET_SIZE - IPOIB_ENCAP_LEN;
	priv->mcast_mtu = priv->admin_mtu = dev->mtu;

	memcpy(dev->broadcast, ipv4_bcast_addr, IPOIB_HW_ADDR_LEN);
    
	dev->broadcast[8] = priv->pkey >> 8;
	dev->broadcast[9] = priv->pkey & 0xff;

	netif_carrier_off(dev);

	SET_MODULE_OWNER(dev);

	spin_lock_init(&priv->lock);
	spin_lock_init(&priv->rq_lock);


	/* Allocate RX/TX "rings" to hold queued skbs */

	priv->rx_ring =
		kmalloc(IPOIB_RX_RING_SIZE * sizeof (struct ipoib_buf),
			GFP_KERNEL);
	if (!priv->rx_ring) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: failed to allocate RX ring (%d entries)",
				dev->name, IPOIB_RX_RING_SIZE);
		goto out;
	}
	memset(priv->rx_ring, 0,
	       IPOIB_RX_RING_SIZE * sizeof (struct ipoib_buf));

	priv->tx_ring =
	    kmalloc(IPOIB_TX_RING_SIZE * sizeof(struct ipoib_buf),
		    GFP_KERNEL);
	
	if (!priv->tx_ring) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: failed to allocate TX ring (%d entries)",
				dev->name, IPOIB_TX_RING_SIZE);
		goto out_rx_ring_cleanup;
	}
	memset(priv->tx_ring, 0,
	       IPOIB_TX_RING_SIZE * sizeof(struct ipoib_buf));

	/* set up the rest of our private data */
 	priv->tx_last_wri_handel = IPOIB_TX_RING_SIZE - 1;
	priv->tx_hanging = 0;
	priv->rx_last_post_receive = IPOIB_RX_RING_SIZE - 1;
	priv->rx_last_used = IPOIB_RX_RING_SIZE - 1;
	INIT_LIST_HEAD(&priv->ops_list);
	init_waitqueue_head(&priv->mcast_update_wait);
	priv->mcast_update_thread = NULL;
	priv->ah_amount = 0;

	/* priv->tx_head is already 0 */
	priv->tx_free = IPOIB_TX_RING_SIZE;
        
	if (ipoib_ib_dev_init(dev, ca, port))
		goto out_tx_ring_cleanup;

	if (ipoib_proc_dev_init(dev))
		goto out_ib_cleanup;

	return 0;

out_ib_cleanup:
	ipoib_ib_dev_cleanup(dev);

out_tx_ring_cleanup:
	kfree(priv->tx_ring);

out_rx_ring_cleanup:
	kfree(priv->rx_ring);

out:
	return -ENOMEM;
}

void ipoib_ah_delete(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct ipoib_neigh_ops *ipoib_ops, *tops;
	struct ipoib_path *path, *tpath;
   	LIST_HEAD(remove_list);
 
	spin_lock_irq(&priv->lock);

	list_for_each_entry_safe(ipoib_ops, tops, &priv->ops_list, list) {
		list_del(&ipoib_ops->list);
		ipoib_ops->ops->destructor = NULL;
		kfree(ipoib_ops);
	}

	if (list_empty(&priv->created_ahs))
	{
		
		TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_ARP,
			 "%s: Created ahs List is empty return", dev->name);
   		
        spin_unlock_irq(&priv->lock);
		return;
	}
	spin_unlock_irq(&priv->lock);
	
	while (TRUE)
	{
		spin_lock_irq(&priv->lock);
		// Clone all list
		if (list_empty(&priv->created_ahs))
		{
			spin_unlock_irq(&priv->lock);
			break;
		} else {
			
			list_for_each_entry_safe(path, tpath, &priv->created_ahs, list) {
				if (path->unused_ah != 1)
                {
                    path->neighbour = neigh_clone(path->neighbour);
                    if (atomic_read(&path->neighbour->refcnt) > 1){
                        TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_ARP,
                                 "%s: AH %x refcnt is more then 1 => delete it",dev->name, path->ah);
                        list_del(&path->list);
                        list_add_tail(&path->list, &remove_list);
			memset(path->neighbour->ha + 24, 0, 8);
					
                    } else {
    					TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_ARP,
                                 "%s: AH %x refcnt is 1 => release it",dev->name, path->ah);
                        neigh_release(path->neighbour); 
    				}
                }  else {
                    list_del(&path->list);
                    list_add_tail(&path->list, &remove_list);

                }

			}
			spin_unlock_irq(&priv->lock);
            yield();
        }
	}
	      
	// Destroy the AHs
	list_for_each_entry_safe(path, tpath, &remove_list, list) {
		TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_ARP,
			 "%s: Reaping ah %x",dev->name, path->ah);        
		ib_address_destroy(path->ah);
		--priv->ah_amount;
		if (path->unused_ah != 1) {
			if ( path->neighbour ) {
				neigh_release(path->neighbour); 
			}
			else {
				printk("ipoib: path->neighbour already released\n");
			}
		}
		
		kfree(path);
	}
	if (priv->ah_amount != 0)
	{
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: The AH amount is not ZERO after cleaning up ah_amount = %x",
			       dev->name,priv->ah_amount);
	}
	
}

void ipoib_dev_cleanup(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv, *cpriv, *tcpriv;
	int i;

	/* Delete any child interfaces first */
	/* Safe since it's either protected by ipoib_device_mutex or empty */
	list_for_each_entry_safe(cpriv, tcpriv, &priv->child_intfs, list) {
		ipoib_dev_cleanup(&cpriv->dev);
		unregister_netdev(&cpriv->dev);

		list_del(&cpriv->list);

		kfree(cpriv);
	}

	ipoib_proc_dev_cleanup(dev);
	ipoib_ib_dev_cleanup(dev);
  
	if (priv->rx_ring) {
		TS_REPORT_CLEANUP(MOD_IB_NET,
				  "%s: cleaning up RX ring", dev->name);

		for (i = 0; i < IPOIB_RX_RING_SIZE; ++i) {
			if (priv->rx_ring[i].skb && (priv->rx_ring[i].skb_not_alloc == 0)){
				dev_kfree_skb(priv->rx_ring[i].skb);
				priv->rx_ring[i].skb_not_alloc = 1;
			}
		}

		kfree(priv->rx_ring);
		priv->rx_ring = NULL;
	}

	if (priv->tx_ring) {
		TS_REPORT_CLEANUP(MOD_IB_NET,
				  "%s: cleaning up TX ring", dev->name);

		for (i = 0; i < IPOIB_TX_RING_SIZE; ++i) {
			if (priv->tx_ring[i].skb)
				dev_kfree_skb(priv->tx_ring[i].skb);
		}

		kfree(priv->tx_ring);
		priv->tx_ring = NULL;
	}
}

struct ipoib_dev_priv *ipoib_intf_alloc(void)
{
	struct ipoib_dev_priv *priv;

	priv = kmalloc(sizeof(*priv), GFP_KERNEL);
	if (!priv) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"failed to allocate private struct");
		return NULL;
	}

	memset(priv, 0, sizeof(*priv));

	sema_init(&priv->mcast_mutex, 1);

	atomic_set(&priv->mcast_joins, 0);

	INIT_LIST_HEAD(&priv->child_intfs);
    INIT_LIST_HEAD(&priv->created_ahs);
	INIT_LIST_HEAD(&priv->multicast_list);

	INIT_TQUEUE(&priv->restart_task, ipoib_mcast_restart_task, &priv->dev);

	priv->dev.priv = priv;

	return priv;
}

int ipoib_add_port(const char *format, struct ib_device *hca, tTS_IB_PORT port)
{
	struct ipoib_dev_priv *priv;
	int result = -ENOMEM;

	priv = ipoib_intf_alloc();
	if (!priv)
		goto alloc_mem_failed;

	priv->pkey = 0xffff;

#if 0
	/* We'll probably use something like this in the future */
	result = ib_pkey_entry_get(hca, port, 0, &priv->pkey);
	if (result) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: ib_pkey_entry_get failed (ret = %d)",
				priv->dev.name, result);
		return result;
	}
#endif

	result = dev_alloc_name(&priv->dev, format);
	if (result < 0) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"failed to get device name (ret = %d)", result);
		goto dev_alloc_failed;
	}
    result = ib_gid_entry_get(hca, port, 0, priv->local_gid.raw);
    if (result) {
        TS_REPORT_FATAL(MOD_IB_NET, 
          "ib_query_gid port %d failed (ret = %d)\n",
		       hca->name, port, result);
		goto dev_alloc_failed;
	} else
        memcpy(priv->dev.dev_addr + 4, priv->local_gid.raw, sizeof (union ib_gid));

	result = ipoib_dev_init(&priv->dev, hca, port);
	if (result < 0) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"failed to initialize net device %d, port %d (ret = %d)",
				hca, port, result);
		goto device_init_failed;
	}

	result = ipoib_port_monitor_dev_start(&priv->dev);
	if (result < 0) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"failed to setup port monitor for device %d, "
				"port %d (ret = %d)",
				hca, port, result);
		goto port_monitor_failed;
	}

	result = register_netdev(&priv->dev);
	if (result) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: failed to initialize; error %i",
				priv->dev.name, result);
		goto register_failed;
	}

	down(&ipoib_device_mutex);
	list_add_tail(&priv->list, &ipoib_device_list);
	up(&ipoib_device_mutex);

	return 0;

register_failed:
	ipoib_port_monitor_dev_stop(&priv->dev);

port_monitor_failed:
	ipoib_dev_cleanup(&priv->dev);

device_init_failed:
	/*
	 * Nothing to do since the device name only gets finally added
	 * to the linked list in register_netdev
	 */

dev_alloc_failed:
	kfree(priv);

alloc_mem_failed:
	return result;
}

static int __init ipoib_init_module(void)
{
	int ret;

	ret = ipoib_transport_create_devices();
	if (ret)
		return ret;

	down(&ipoib_device_mutex);
	if (list_empty(&ipoib_device_list)) {
		up(&ipoib_device_mutex);
		ipoib_transport_cleanup();
		return -ENODEV;
	}
	up(&ipoib_device_mutex);

	ipoib_vlan_init();

	return 0;
}

static void __exit ipoib_cleanup_module(void)
{
	struct ipoib_dev_priv *priv, *tpriv;
    
	ipoib_vlan_cleanup();
	ipoib_transport_cleanup();
	
	list_for_each_entry_safe(priv, tpriv, &ipoib_device_list, list) {
		_ipoib_dev_stop(&priv->dev);
	}
	
	list_for_each_entry_safe(priv, tpriv, &ipoib_device_list, list) {
		/*Clean the AH + finish cleaning the buffer*/
		ipoib_ib_finish_completion(&priv->dev);
		ipoib_ah_delete(&priv->dev);        
		ipoib_port_monitor_dev_stop(&priv->dev);
		ipoib_dev_cleanup(&priv->dev);
		unregister_netdev(&priv->dev);
		ipoib_mcast_dev_exit(&priv->dev);
		
		list_del(&priv->list);
		
		kfree(priv);
	}
	
}

module_init(ipoib_init_module);
module_exit(ipoib_cleanup_module);

/*
  Local Variables:
  c-file-style: "linux"
  indent-tabs-mode: t
  End:
*/

