/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: ipoib_multicast.c 476 2004-07-16 16:30:27Z roland $
*/

#include "ipoib.h"

#include "ts_ib_sa_client.h"

#include "ts_kernel_services.h"
#include "ts_kernel_trace.h"

#include <linux/rtnetlink.h>
#include <linux/ip.h>
#include <linux/in.h>
#include <linux/igmp.h>
#include <linux/inetdevice.h>

/* Used for all multicast joins (broadcast, IPv4 mcast and IPv6 mcast) */
struct ipoib_mcast {
	rb_node_t rb_node;
	struct list_head list;

	atomic_t refcnt;
	unsigned long created;

	struct ib_multicast_member mcast_member;
	struct ib_address *address_handle;
	tTS_IB_CLIENT_QUERY_TID tid;

#ifndef MAC_20_BYTES
 	tTS_IB_GID mgid;
#else
	union ib_gid mgid;
#endif
	unsigned long flags;
	unsigned char logcount;

	struct sk_buff_head pkt_queue;

	struct net_device *dev;
};

struct ipoib_mcast_iter {
	struct net_device *dev;
	rb_node_t *rb_node;
};


tTS_IB_GID ipoib_broadcast_mgid = {
	0xff, 0x12, 0x40, 0x1b, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff
};


/* =============================================================== */
/*..ipoib_mcast_get - get reference to multicast group             */
void ipoib_mcast_get(struct ipoib_mcast *mcast)
{
	atomic_inc(&mcast->refcnt);
}

/* =============================================================== */
/*..ipoib_mcast_put - put reference to multicast group             */
void ipoib_mcast_put(struct ipoib_mcast *mcast)
{
	struct net_device *dev = mcast->dev;
    struct ipoib_dev_priv *priv = dev->priv;

	if (atomic_dec_and_test(&mcast->refcnt)) {
         TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_MULTICAST,
                 "%s: deleting multicast group " IPOIB_GID_FMT,
                 dev->name, IPOIB_GID_ARG(mcast->mgid));
		if (mcast->address_handle != NULL) {
			int ret = ib_address_destroy(mcast->address_handle);
			if (ret < 0)  {
				TS_REPORT_WARN(MOD_IB_NET,
					       "%s: ib_address_destroy failed (ret = %d)",
					       dev->name, ret);
            }
#ifdef MAC_20_BYTES
            else {
                 spin_lock(&priv->lock);
                 --priv->ah_amount;
                 spin_unlock(&priv->lock);
            }
#endif
		}

		while (!skb_queue_empty(&mcast->pkt_queue)) {
			struct sk_buff *skb = skb_dequeue(&mcast->pkt_queue);

			skb->dev = dev;

			dev_kfree_skb(skb);
		}

		kfree(mcast);
	}
}

/* =============================================================== */
/*..ipoib_mcast_alloc - allocate mcast group entry                 */
struct ipoib_mcast *ipoib_mcast_alloc(struct net_device *dev)
{
	struct ipoib_mcast *mcast;

	mcast = kmalloc(sizeof(*mcast), GFP_ATOMIC);
	if (!mcast)
		return NULL;

	memset(mcast, 0, sizeof(*mcast));

	atomic_set(&mcast->refcnt, 2);	/* Calling function needs to put */

	mcast->dev = dev;
	mcast->created = jiffies;
	mcast->logcount = 0;

	INIT_LIST_HEAD(&mcast->list);
	skb_queue_head_init(&mcast->pkt_queue);

	mcast->address_handle = NULL;

	/* Will force a trigger on the first packet we need to send */
	mcast->tid = TS_IB_CLIENT_QUERY_TID_INVALID;

	return mcast;
}

/* =============================================================== */
/*..__ipoib_mcast_find - find multicast group            */
#ifndef MAC_20_BYTES
struct ipoib_mcast *__ipoib_mcast_find(struct net_device *dev, tTS_IB_GID mgid)    
#else
struct ipoib_mcast *__ipoib_mcast_find(struct net_device *dev, union ib_gid *mgid)
#endif
{
	struct ipoib_dev_priv *priv = dev->priv;
	rb_node_t *n = priv->multicast_tree.rb_node;

	while (n) {
		struct ipoib_mcast *mcast;
		int ret;
		mcast = rb_entry(n, struct ipoib_mcast, rb_node);

#ifndef MAC_20_BYTES
		ret = memcmp(mgid, mcast->mgid, sizeof(tTS_IB_GID));
#else
		ret = memcmp(mgid->raw, mcast->mgid.raw, sizeof(union ib_gid));
#endif
		if (ret < 0)
			n = n->rb_left;
		else if (ret > 0)
			n = n->rb_right;
		else {
			ipoib_mcast_get(mcast);
			return mcast;
		}
	}

	return NULL;
}

/* =============================================================== */
/*.._ipoib_mcast_find - find multicast group                       */
#ifndef MAC_20_BYTES
struct ipoib_mcast *_ipoib_mcast_find(struct net_device *dev, tTS_IB_GID mgid)
#else
struct ipoib_mcast *_ipoib_mcast_find(struct net_device *dev, union ib_gid *mgid)
#endif
{
	struct ipoib_mcast *mcast;
	struct ipoib_dev_priv *priv = dev->priv;
	unsigned long flags;

	spin_lock_irqsave(&priv->lock, flags);
	mcast = __ipoib_mcast_find(dev, mgid);
	spin_unlock_irqrestore(&priv->lock, flags);

	return mcast;
}

/* =============================================================== */
/*..__ipoib_mcast_add -- add multicast group to rbtree             */
static int __ipoib_mcast_add(struct net_device *dev, struct ipoib_mcast *mcast)
{
	struct ipoib_dev_priv *priv = dev->priv;
	rb_node_t **n = &priv->multicast_tree.rb_node, *pn = NULL;

	while (*n) {
		struct ipoib_mcast *tmcast;
		int ret;

		pn = *n;
		tmcast = rb_entry(pn, struct ipoib_mcast, rb_node);
#ifndef MAC_20_BYTES 
		ret = memcmp(mcast->mgid, tmcast->mgid, sizeof(tTS_IB_GID));
#else
		ret = memcmp(mcast->mgid.raw, tmcast->mgid.raw, sizeof(union ib_gid));
#endif
		if (ret < 0)
		  n = &pn->rb_left;
		else if (ret > 0)
		  n = &pn->rb_right;
		else
		  return -EEXIST;
	}

	rb_link_node(&mcast->rb_node, pn, n);
	rb_insert_color(&mcast->rb_node, &priv->multicast_tree);

	return 0;
}

/* =============================================================== */
/*.._ipoib_mcast_join_finish - finish joining mcast group entry    */
static int _ipoib_mcast_join_finish(struct ipoib_mcast *mcast,
				    struct ib_multicast_member *member_ptr)
{
	struct net_device *dev = mcast->dev;
	struct ipoib_dev_priv *priv = dev->priv;
	int ret;
	
	if (!priv || !priv->broadcast)
        return 0;
    
	mcast->mcast_member = *member_ptr;
	priv->qkey = priv->broadcast->mcast_member.qkey;
	
	if (test_and_set_bit(IPOIB_MCAST_FLAG_ATTACHED, &mcast->flags)) {
	  if (test_and_clear_bit(IPOIB_MCAST_FLAG_REJOIN, &mcast->flags)) {
	    TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE,
		     TRACE_IB_NET_MULTICAST,
		     "%s: multicast rejoin finished " IPOIB_GID_FMT, dev->name,
		     IPOIB_GID_ARG(mcast->mgid));
	    
	  } else {
	    TS_REPORT_WARN(MOD_IB_NET,
			   "%s: multicast group " IPOIB_GID_FMT
			   " already attached?", dev->name,
			   IPOIB_GID_ARG(mcast->mgid));
	  }
	  return 0;
	}
	
	/* Set the cached Q_Key before we attach if it's the broadcast group */
#ifdef MAC_20_BYTES	
	if (!memcmp(mcast->mgid.raw , priv->dev.broadcast + 4, sizeof(union ib_gid)))
          priv->qkey = priv->broadcast->mcast_member.qkey;
	
	ret = ipoib_mcast_attach(dev, mcast->mcast_member.mlid, &mcast->mgid);
	
#else	
	if (!memcmp(mcast->mgid, priv->bcast_gid, sizeof(tTS_IB_GID)))
	  priv->qkey = priv->broadcast->mcast_member.qkey;
	
	ret = ipoib_mcast_attach(dev, mcast->mcast_member.mlid, mcast->mgid);
#endif 
	
	if (ret < 0) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: couldn't attach QP to multicast group "
				IPOIB_GID_FMT, dev->name,
				IPOIB_GID_ARG(mcast->mgid));
		
		clear_bit(IPOIB_MCAST_FLAG_ATTACHED, &mcast->flags);
		return ret;
	}
	
	{
		struct ib_address_vector av = {
			.dlid = mcast->mcast_member.mlid,
			.port = priv->port,
			.service_level = mcast->mcast_member.sl,
			.source_path_bits = 0,
			.use_grh = 1,
			.flow_label = mcast->mcast_member.flowlabel,
			.hop_limit = mcast->mcast_member.hoplmt,
			.source_gid_index = 0,
			.static_rate = 0,
			.traffic_class = mcast->mcast_member.tclass,
		};

		memcpy(av.dgid, mcast->mcast_member.mgid, sizeof(av.dgid));
		
		if (ib_address_create(priv->pd, &av, &mcast->address_handle)) {
		  TS_REPORT_WARN(MOD_IB_NET,
				 "%s: ib_address_create failed",
				 dev->name);
		} else {
 
#ifdef MAC_20_BYTES           
            spin_lock_irq(&priv->lock);
            ++priv->ah_amount; 
            spin_unlock_irq(&priv->lock);
#endif
           TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE,
			   TRACE_IB_NET_MULTICAST,
			   "%s: MGID " IPOIB_GID_FMT
			   " AV 0x%08x, LID 0x%04x, SL %d", dev->name,
			   IPOIB_GID_ARG(mcast->mgid),
			   mcast->address_handle,
			   mcast->mcast_member.mlid,
			   mcast->mcast_member.sl);
		}
	}
	
	/* actually send any queued packets */
	while (!skb_queue_empty(&mcast->pkt_queue)) {
	  struct sk_buff *skb = skb_dequeue(&mcast->pkt_queue);
	  
	  skb->dev = dev;
	  
	  if (dev_queue_xmit(skb))
	    TS_REPORT_WARN(MOD_IB_NET,
			   "%s: dev_queue_xmit failed to requeue packet",
			   dev->name);
	}
	
	return 0;
}

/* =============================================================== */
/*.._ipoib_mcast_sendonly_join_complete -- handler for multicast join */
static void
_ipoib_mcast_sendonly_join_complete(tTS_IB_CLIENT_QUERY_TID tid,
			 	    int status,
				    struct ib_multicast_member *member_ptr,
				    void *mcast_ptr)
{
	struct ipoib_mcast *mcast = mcast_ptr;
	struct net_device *dev = mcast->dev;
	struct ipoib_dev_priv *priv = dev->priv;

	mcast->tid = TS_IB_CLIENT_QUERY_TID_INVALID;

	if (!status)
		_ipoib_mcast_join_finish(mcast, member_ptr);
	else {
		if (mcast->logcount++ < 20)
			TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE,
				 TRACE_IB_NET_MULTICAST,
				 "%s: multicast join failed for " IPOIB_GID_FMT
				 ", status %d", dev->name,
                     IPOIB_GID_ARG(mcast->mgid), status);

		/* Flush out any queued packets */
		while (!skb_queue_empty(&mcast->pkt_queue)) {
			struct sk_buff *skb = skb_dequeue(&mcast->pkt_queue);

			skb->dev = dev;

			dev_kfree_skb(skb);
		}

		/* Clear the busy flag so we try again */
		clear_bit(IPOIB_MCAST_FLAG_BUSY, &mcast->flags);
	}

	ipoib_mcast_put(mcast);

	atomic_dec(&priv->mcast_joins);
}

/* =============================================================== */
/*.._ipoib_mcast_sendonly_join -- start send only join             */
static int _ipoib_mcast_sendonly_join(struct ipoib_mcast *mcast)
{
	struct net_device *dev = mcast->dev;
	struct ipoib_dev_priv *priv = dev->priv;
	tTS_IB_CLIENT_QUERY_TID tid;
	int ret = 0;

	atomic_inc(&priv->mcast_joins);

	if (!test_bit(IPOIB_FLAG_OPER_UP, &priv->flags)) {
		TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_MULTICAST,
			 "%s: device shutting down, no multicast joins",
			 dev->name);
		ret = -ENODEV;
		goto out;
	}

	if (test_and_set_bit(IPOIB_MCAST_FLAG_BUSY, &mcast->flags)) {
		TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_MULTICAST,
			 "%s: multicast entry busy, skipping", dev->name);
		ret = -EBUSY;
		goto out;
	}

	ipoib_mcast_get(mcast);
	ret = tsIbMulticastGroupJoin(priv->ca,
#ifdef MAC_20_BYTES			     
				     priv->port, mcast->mgid.raw, priv->pkey,
#else
				     priv->port, mcast->mgid, priv->pkey,            
#endif
/* ib_sm doesn't support send only yet
				     TS_IB_MULTICAST_JOIN_SEND_ONLY_NON_MEMBER,
 */
				     TS_IB_MULTICAST_JOIN_FULL_MEMBER,
				     HZ,
				     _ipoib_mcast_sendonly_join_complete,				     
                     mcast, &tid,priv->qkey,1);
	
    if (ret) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: tsIbMulticastGroupJoin failed (ret = %d)",
			       dev->name, ret);
		ipoib_mcast_put(mcast);
	} else {
		TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_MULTICAST,
			 "%s: no multicast record for " IPOIB_GID_FMT
			 ", starting join", dev->name,
             IPOIB_GID_ARG(mcast->mgid));

		mcast->tid = tid;
	}

out:
	if (ret)
		atomic_dec(&priv->mcast_joins);

	return ret;
}

/* =============================================================== */
/*.._ipoib_mcast_join_complete - handle comp of mcast join         */
static void _ipoib_mcast_join_complete(tTS_IB_CLIENT_QUERY_TID tid,
				       int status,
				       struct ib_multicast_member *member_ptr,
				       void *mcast_ptr)
{
    struct ipoib_mcast *mcast = mcast_ptr;
	struct net_device *dev = mcast->dev;
	struct ipoib_dev_priv *priv = dev->priv;

	priv->mcast_tid = TS_IB_CLIENT_QUERY_TID_INVALID;

	if (!status)
		priv->mcast_status = _ipoib_mcast_join_finish(mcast, member_ptr);
	else
		priv->mcast_status = status;

	complete(&priv->mcast_complete);
}

/* =============================================================== */
/*.._ipoib_mcast_leave_complete - handle comp of mcast leave  -   currenty not called by any one     */
/*
static void _ipoib_mcast_leave_complete(tTS_IB_CLIENT_QUERY_TID tid,
					int status,
					void *mcast_ptr)
{
        struct ipoib_mcast *mcast = mcast_ptr;
	struct net_device *dev = mcast->dev;
	struct ipoib_dev_priv *priv = dev->priv;

	priv->mcast_tid = TS_IB_CLIENT_QUERY_TID_INVALID;

	if (!status)
	  TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		   "%s: tsIbMulticastGroupLeave leave succeed for MGID " IPOIB_GID_FMT
		   , dev->name,
		   IPOIB_GID_ARG(mcast->mgid));
	
	else
	  TS_REPORT_FATAL(MOD_IB_NET,
			  "%s: tsIbMulticastGroupLeave failed, status %d for MGID",
			  dev->name, status,IPOIB_GID_FMT,IPOIB_GID_ARG(mcast->mgid));
}
*/
/* =============================================================== */
/*..__ipoib_mcast_join - join multicast group for iface            */
static int __ipoib_mcast_join(struct net_device *dev, struct ipoib_mcast *mcast ,int print_warn)
{
        struct ipoib_dev_priv *priv = dev->priv;
        int status;
	int create = 1;
	struct ib_port_properties port_prop;
	
	TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_MULTICAST,
		 "%s: joining MGID " IPOIB_GID_FMT,
		 dev->name, IPOIB_GID_ARG(mcast->mgid));

	priv->mcast_status = -ERESTARTSYS;
	
	init_completion(&priv->mcast_complete);
	
#ifdef MAC_20_BYTES    
	if (mcast->mgid.raw == priv->broadcast->mgid.raw) {
	  create = 0;
	}
#else
	if (mcast->mgid == priv->broadcast->mgid) {
	  create = 0;
	}   
#endif
	
	status = tsIbMulticastGroupJoin(priv->ca,
					priv->port,
#ifdef MAC_20_BYTES					
					mcast->mgid.raw,
#else
					mcast->mgid,               
#endif
					priv->pkey,
					TS_IB_MULTICAST_JOIN_FULL_MEMBER,
					HZ,
					_ipoib_mcast_join_complete,
					mcast, &priv->mcast_tid,priv->qkey,create);
	
	if (status) {
	  TS_REPORT_FATAL(MOD_IB_NET,
			  "%s: tsIbMulticastGroupJoin failed, status %d",
			  dev->name, status);
	  priv->mcast_status = -EINVAL;
	  return status;
	}
	
	wait_for_completion(&priv->mcast_complete);
	
	if (priv->mcast_status) {
	  
	  if (priv->mcast_status == -ETIMEDOUT) {
	    if (ib_port_properties_get(priv->ca, priv->port, &port_prop)) {
	      TS_REPORT_WARN(MOD_IB_NET,
			     "ib_port_properties_get() failed for device %s: port %d",
			     dev->name, priv->port);
	      
	      TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		       "%s: multicast join failed for " IPOIB_GID_FMT
		       ", status %d", dev->name,
		       IPOIB_GID_ARG(mcast->mgid),
		       priv->mcast_status);
	    } else {
	      if ((port_prop.port_state > IB_PORT_STATE_DOWN) && (print_warn == 1)) {
                TS_REPORT_WARN(MOD_IB_NET,
                               "%s: SM/SA is not available, multicast join failed for " IPOIB_GID_FMT
                               ", status %d", dev->name,
                               IPOIB_GID_ARG(mcast->mgid),
                               priv->mcast_status);
	      } else {
              TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_MULTICAST,
                       "%s: Link is down, multicast join failed for " IPOIB_GID_FMT
                       ", status %d", dev->name,
                       IPOIB_GID_ARG(mcast->mgid),
                       priv->mcast_status);
	      }
	    }
	  } else {
	    if (priv->mcast_status == -EINTR
		|| priv->mcast_status == -EINVAL) {
	      TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		       "%s: multicast join failed for " IPOIB_GID_FMT
		       ", status %d", dev->name,
		       IPOIB_GID_ARG(mcast->mgid),
		       priv->mcast_status);
	    } else {
	      TS_REPORT_WARN(MOD_IB_NET,
			     "%s: multicast join failed for "
			     IPOIB_GID_FMT ", status %d", dev->name,
			     IPOIB_GID_ARG(mcast->mgid),
			     priv->mcast_status);
	    }
      }
      return priv->mcast_status;
	}
	
	return status;
}

/* =============================================================== */
/*.._ipoib_mcast_join - join multicast group                       */
static int _ipoib_mcast_join(struct net_device *dev, struct ipoib_mcast *mcast)
{
	unsigned int backoff = 1;
	int ret ,print_warn = 0;
	
	if (test_bit(IPOIB_MCAST_FLAG_ATTACHED, &mcast->flags) && 
	    (!test_bit(IPOIB_MCAST_FLAG_REJOIN, &mcast->flags)))
		return 0;
	
	do {
	  ret = __ipoib_mcast_join(dev, mcast, print_warn);
	  if (ret && ret != -ETIMEDOUT && ret != -EINVAL)
	    return ret;
	  
	  if (ret) {
	    TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE,
		     TRACE_IB_NET_MULTICAST,
		     "%s: couldn't join multicast group, backing off. retrying in %d seconds",
		     dev->name, backoff);
	    
	    set_current_state(TASK_INTERRUPTIBLE);
	    schedule_timeout(backoff * HZ);
	    set_current_state(TASK_RUNNING);
	    
	    if (signal_pending(current))
	      return -EINTR;
	    
	    backoff *= 2;
	    if (backoff > IPOIB_MAX_BACKOFF_SECONDS)
	      {
		backoff = IPOIB_MAX_BACKOFF_SECONDS;
		print_warn = 1;
	      }
	  }
	} while (ret);
	
	return 0;
}

/* =============================================================== */
/*.._ipoib_mcast_join_thread -- poll multicast join                */
static void _ipoib_mcast_join_thread(void *dev_ptr)
{
  	struct net_device *dev = dev_ptr;
	struct ipoib_dev_priv *priv = dev->priv;
#ifndef MAC_20_BYTES	
	struct ipoib_sarp *entry;
#endif	
	unsigned long flags;
	int ret = 0;
	struct ib_async_event_record async_record;
	
 	if (!test_bit(IPOIB_FLAG_OPER_UP, &priv->flags)) {
		TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_MULTICAST,
			 "%s: device shutting down, no multicast joins",
			 dev->name);
		return -ENODEV;
	}

	atomic_inc(&priv->mcast_joins);

	if (!priv->broadcast) {
	  priv->broadcast = ipoib_mcast_alloc(dev);
	  if (!priv->broadcast) {
			TS_REPORT_FATAL(MOD_IB_NET,
					"%s: failed to allocate broadcast group",
					dev->name);
			ret = -ENOMEM;
			goto out;
	  }
	  
#ifdef MAC_20_BYTES
	  memcpy(priv->broadcast->mgid.raw, priv->dev.broadcast + 4,
               sizeof (union ib_gid));
#else
	  memcpy(priv->bcast_gid, ipoib_broadcast_mgid,
		 sizeof(tTS_IB_GID));
	  priv->bcast_gid[4] = (priv->pkey >> 8) & 0xff;
	  priv->bcast_gid[5] = priv->pkey & 0xff;
	  
	  memcpy(priv->broadcast->mgid, priv->bcast_gid,
		 sizeof(tTS_IB_GID));
	  
#endif
	  spin_lock_irqsave(&priv->lock, flags);
	  __ipoib_mcast_add(dev, priv->broadcast);
	  spin_unlock_irqrestore(&priv->lock, flags);

	  ipoib_mcast_put(priv->broadcast);	/* for _alloc() */
	}
	
	ret = _ipoib_mcast_join(dev, priv->broadcast);
	if (ret < 0)
	  goto out;
	
	while (1) {
	  struct ipoib_mcast *mcast = NULL;
	  
	  if (signal_pending(current)) {
	    ret = -EINTR;
	    goto out;
	  }
	  
	  spin_lock_irqsave(&priv->lock, flags);
	  list_for_each_entry(mcast, &priv->multicast_list, list) {
	    if ((!test_bit(IPOIB_MCAST_FLAG_SENDONLY, &mcast->flags)
		 && !test_bit(IPOIB_MCAST_FLAG_BUSY, &mcast->flags)
		 && !test_bit(IPOIB_MCAST_FLAG_ATTACHED, &mcast->flags)) 
		|| (test_bit(IPOIB_MCAST_FLAG_REJOIN, &mcast->flags))) {
	      /* Found the next unjoined group */
		ipoib_mcast_get(mcast);
	      break;
	    }
	  }
	  spin_unlock_irqrestore(&priv->lock, flags);
	  
	  if (&mcast->list == &priv->multicast_list) {
	    /* All done */
	    break;
	  }
	  
	  ret = _ipoib_mcast_join(dev, mcast);
	  ipoib_mcast_put(mcast);
	  if (ret < 0)
	    goto out;
	}
	
	{
	  struct ib_port_lid port_lid;
	  
	  ib_cached_lid_get(priv->ca, priv->port, &port_lid);
	  
	  priv->local_lid = port_lid.lid;
	}
#ifdef MAC_20_BYTES
	if (ib_gid_entry_get(priv->ca, priv->port, 0, priv->local_gid.raw))
	  TS_REPORT_WARN(MOD_IB_NET,
			 "%s: ib_gid_entry_get() failed", dev->name);  
	else
	  memcpy(priv->dev.dev_addr + 4, priv->local_gid.raw, sizeof (union ib_gid));
	
	
#else
	if (ib_gid_entry_get(priv->ca, priv->port, 0, priv->local_gid))
	  TS_REPORT_WARN(MOD_IB_NET,
			 "%s: ib_gid_entry_get() failed", dev->name);  
#endif
	
	priv->mcast_mtu = ib_mtu_enum_to_int(priv->broadcast->mcast_member.mtu)
		- IPOIB_ENCAP_LEN;
	
	dev->mtu = min(priv->mcast_mtu, priv->admin_mtu);

#ifndef MAC_20_BYTES
	entry = ipoib_sarp_local_add(dev, priv->local_gid, priv->local_qpn);
	if (entry)
	  ipoib_sarp_put(entry);
#endif
	
	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		 "%s: successfully joined all multicast groups", dev->name);

	netif_carrier_on(dev);
	
	/*
	 * Register to SM change event
	 */
	if (!priv->mcast_rejoin_handler)
	  {
	    TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		     "%s:Registering to SM change event", dev->name);
	    
	    async_record.device = priv->qp->device;
	    async_record.event = IB_SM_CHANGE;
	    
	    ret = ib_async_event_handler_register( &async_record,
						   ipoib_mcast_restart_thread,
						   dev_ptr,
						   &priv->mcast_rejoin_handler);
	    
	    if ( ret ) {
	      TS_REPORT_FATAL(MOD_IB_NET,
			      "Registration of async event SM change status = %d",
			      ret );
	    }
	  }
	
#ifdef MAC_20_BYTES	
	ipoib_mcast_update_start_thread(dev);
#endif
	
 out:
	if (ret && ret != -EINTR)
	  TS_REPORT_FATAL(MOD_IB_NET,
				"%s: multicast join failed, status %d",
			  dev->name, ret);
	
	atomic_dec(&priv->mcast_joins);
}
/* 
   The next Mutex defend on priv->mcast_thread
*/
DECLARE_MUTEX(thread_mutex);
/* 
   The next Mutex defend on :
            priv->mcast_update_stop
	    priv->mcast_rejoin
	    priv->mcast_update_thread
*/

DECLARE_MUTEX(thread_update_mutex);

/* =============================================================== */
/*..ipoib_mcast_start_thread -- start multicast thread             */
int ipoib_mcast_start_thread(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	char thread_name[64];
	int ret = 0;
	
	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		 "%s: starting multicast thread", dev->name);
	
	down(&thread_mutex);
	
  	if (priv->mcast_thread)
	  goto out;
	
	snprintf(thread_name, sizeof(thread_name) - 1, "%s_join_mcast",
		 dev->name);
	
	if (tsKernelThreadStart(thread_name,
				_ipoib_mcast_join_thread,
				dev, &priv->mcast_thread)) {
	  TS_REPORT_FATAL(MOD_IB_NET,
			  "%s: failed to start multicast join thread",
			  dev->name);
	  ret = -1;
	}
	
 out:
	up(&thread_mutex);
	
	return ret;
}

/* =============================================================== */
/*.._ipoib_mcast_cancel -- cancel multicast group join             */
static int _ipoib_mcast_cancel(struct ipoib_mcast *mcast)
{
	if (mcast->tid != TS_IB_CLIENT_QUERY_TID_INVALID) {
		/* Kill the lookup */
		if (!ib_client_query_cancel(mcast->tid)) {
			if (test_bit(IPOIB_MCAST_FLAG_SENDONLY, &mcast->flags))
			  _ipoib_mcast_sendonly_join_complete(mcast->tid,
							      -EINTR,
							      NULL,
							      mcast);
			else
			  _ipoib_mcast_join_complete(mcast->tid,
						     -EINTR, NULL,
						     mcast);
		} else
		  TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE,
			   TRACE_IB_NET_MULTICAST,
			   "%s: multicast join TID 0x%016llx no longer exists",
			   mcast->dev->name, mcast->tid);
	}
	
	return 0;
}

/*..ipoib_mcast_restart_thread -- restart multicast thread due  to new SM */
void ipoib_mcast_restart_thread(struct ib_async_event_record * event, void *arg)
{
  	struct net_device *dev = arg;
  	struct ipoib_dev_priv *priv = NULL;
	unsigned long flags;
	struct ipoib_mcast *tmcast, *mcast = NULL;
	LIST_HEAD(remove_list);
	int result;
	
	if (dev == NULL) {
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s:WARN -  device is NULL (Please restart)");
		return;
	}
	
	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		 "%s: Detected SM change, rejoining to multicast groups", dev->name);

	priv = dev->priv;

	if (priv == NULL || priv->broadcast == NULL)
	  {
	    TS_REPORT_WARN(MOD_IB_NET,
			   "%s: WARN - Not all the args are defined (priv=%p)"
			   " (Please restart)",
			   dev->name, priv);
	    return;
	  }
	
	spin_lock_irqsave(&priv->lock, flags); 
	
	set_bit(IPOIB_MCAST_FLAG_REJOIN, &priv->broadcast->flags);
	/* Remove all of the entries which should be rejoin */
	list_for_each_entry_safe(mcast, tmcast, &priv->multicast_list, list) {
	  TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		   "%s: deleting multicast group " IPOIB_GID_FMT,
		   dev->name, IPOIB_GID_ARG(mcast->mgid));	    
	  rb_erase(&mcast->rb_node, &priv->multicast_tree);
	  /* Move to the remove list */
	  list_del(&mcast->list);
	  list_add_tail(&mcast->list, &remove_list);             
	}
	spin_unlock_irqrestore(&priv->lock, flags);
	
	/* We have to cancel outside of the spinlock */
	list_for_each_entry_safe(mcast, tmcast, &remove_list, list) {
	  list_del_init(&mcast->list);
	  _ipoib_mcast_cancel(mcast);
	  if (!test_and_clear_bit(IPOIB_MCAST_FLAG_ATTACHED, &mcast->flags)) {
		  ipoib_mcast_put(mcast);
		  continue;
	  }
	  /* Remove ourselves from the multicast group */
#ifdef MAC_20_BYTES
	  result = ipoib_mcast_detach(dev, mcast->mcast_member.mlid, &mcast->mgid);
#else	
	  result = ipoib_mcast_detach(dev, mcast->mcast_member.mlid, mcast->mgid);
#endif	
	  if (result)
	    TS_REPORT_WARN(MOD_IB_NET,
			"%s: ipoib_mcast_detach failed (result = %d)."
			" GID="IPOIB_GID_FMT,
			dev->name, result,IPOIB_GID_ARG(mcast->mgid));
	  
	  ipoib_mcast_put(mcast);
	}
   	
#ifdef MAC_20_BYTES
    //We will update the chache about the new opensm
	ipoib_path_cache_old(dev);
	down(&thread_update_mutex);
	if (priv->mcast_update_thread){
	  priv->mcast_rejoin = 1;
	  up(&thread_update_mutex);
	  wake_up_interruptible(&priv->mcast_update_wait);
	}
	else {
	  up(&thread_update_mutex);
	  ipoib_mcast_restart_task(dev);
	}
#else
	ipoib_mcast_restart_task(dev);
#endif
	
}

/* =============================================================== */
/*..ipoib_mcast_stop_thread -- stop multicast join                 */
int ipoib_mcast_stop_thread(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	int ret = 0;

	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		 "%s: stopping multicast thread", dev->name);

	down(&thread_mutex);
		
	if (priv->mcast_thread) {
	  tsKernelThreadStop(priv->mcast_thread);
	  priv->mcast_thread = NULL;
	  
	  ret = 1;
	}
    

	if (priv->mcast_tid != TS_IB_CLIENT_QUERY_TID_INVALID) {
	  /* Kill the query and thread */
	  if (!ib_client_query_cancel(priv->mcast_tid)) {
			priv->mcast_status = -EINTR;
			complete(&priv->mcast_complete);
	  } else {
			TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE,
				 TRACE_IB_NET_MULTICAST,
				 "%s: multicast Join TID 0x%016llx no longer exists",
				 dev->name, priv->mcast_tid);
	  }
	  
	  priv->mcast_tid = TS_IB_CLIENT_QUERY_TID_INVALID;
	}

	up(&thread_mutex);

	return ret;
}

#ifdef MAC_20_BYTES
/* =============================================================== */
/*..ipoib_mcast_start_thread -- start multicast thread             */
int ipoib_mcast_update_start_thread(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	char thread_name[64];
	int ret = 0;

	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		 "%s: starting multicast update thread", dev->name);
	
	down(&thread_update_mutex);
	priv->mcast_update_stop = 0;
	priv->mcast_rejoin = 0;
	
	if (priv->mcast_update_thread)
	  goto out;
        
	snprintf(thread_name, sizeof(thread_name) - 1, "%s_update_mcast",
		 dev->name);
	
	
	if (tsKernelThreadStart(thread_name,
				ipoib_mcast_update_thread,
				dev, &priv->mcast_update_thread)) {
	  TS_REPORT_FATAL(MOD_IB_NET,
			  "%s: failed to start multicast update thread",
			  dev->name);
          ret = -1;
	}
	
 out:
	up(&thread_update_mutex);
	
	return ret;
}

/* =============================================================== */
/*..ipoib_mcast_update_stop_thread -- stop multicast update                 */
int ipoib_mcast_update_stop_thread(struct net_device *dev)
{
 	struct ipoib_dev_priv *priv = dev->priv;
	int ret = 0;
	
	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		 "%s: stopping multicast update thread", dev->name);
	
	
   	down(&thread_update_mutex);
	if (priv->mcast_update_thread) {
	  priv->mcast_update_stop = 1;
	  wake_up_interruptible(&priv->mcast_update_wait);
	  tsKernelThreadStop(priv->mcast_update_thread);
	  priv->mcast_update_thread = NULL;
	  
	  ret = 1;
	}
	up(&thread_update_mutex);
	
	return ret;
}
#endif
/* =============================================================== */
/*..ipoib_mcast_leave -- leave multicast group                     */
int ipoib_mcast_leave(struct net_device *dev, struct ipoib_mcast *mcast)
{
  	struct ipoib_dev_priv *priv = dev->priv;
	int result;
	
	if (!test_and_clear_bit(IPOIB_MCAST_FLAG_ATTACHED, &mcast->flags))
	  return 0;

	/* Remove ourselves from the multicast group */
#ifdef MAC_20_BYTES
	result = ipoib_mcast_detach(dev, mcast->mcast_member.mlid, &mcast->mgid);
#else	
	result = ipoib_mcast_detach(dev, mcast->mcast_member.mlid, mcast->mgid);
#endif	
	if (result)
	  TS_REPORT_WARN(MOD_IB_NET,
			 "%s: ipoib_mcast_detach failed (result = %d)",
			 dev->name, result);
	
#ifdef MAC_20_BYTES		 
	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		 "%s: ipoib_mcast leave start for MGID " IPOIB_GID_FMT, 
		 dev->name ,IPOIB_GID_ARG_OLD(mcast->mcast_member.mgid));
#else 
   	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		 "%s: ipoib_mcast leave start for MGID " IPOIB_GID_FMT, 
		 dev->name ,IPOIB_GID_ARG(mcast->mcast_member.mgid));
#endif
	
	result = tsIbMulticastGroupLeave(priv->ca, 
					 priv->port,
					 mcast->mcast_member.mgid,                   
					 priv->pkey,
					 priv->qkey,
					 TS_IB_MULTICAST_JOIN_FULL_MEMBER,
					 HZ,
					 /* The mcast is going to be deleted so no call back for now*/
					 /* _ipoib_mcast_leave_complete,    
					    mcast,
					    &priv->mcast_tid  */
					 NULL,
					 NULL,
					 NULL
					 );    
	if (result)
	  TS_REPORT_WARN(MOD_IB_NET,
			 "%s: tsIbMulticastGroupLeave failed (result = %d)",
			 dev->name, result);
	
	return 0;
}


/* =============================================================== */
/*.._ipoib_mcast_delete -- delete multicast group join             */
#ifndef MAC_20_BYTES
static int _ipoib_mcast_delete(struct net_device *dev, tTS_IB_GID mgid)
#else
static int _ipoib_mcast_delete(struct net_device *dev, union ib_gid *mgid)
#endif
{
	struct ipoib_mcast *mcast;
	struct ipoib_dev_priv *priv = dev->priv;
	unsigned long flags;

	spin_lock_irqsave(&priv->lock, flags);
	mcast = __ipoib_mcast_find(dev, mgid);
	if (!mcast) {
		spin_unlock_irqrestore(&priv->lock, flags);
		return -ENOENT;
	}

	rb_erase(&mcast->rb_node, &priv->multicast_tree);
	spin_unlock_irqrestore(&priv->lock, flags);

	ipoib_mcast_leave(mcast->dev, mcast);
	ipoib_mcast_put(mcast);	/* for _find() */
	ipoib_mcast_put(mcast);	/* for call */

	return 0;
}

#ifndef MAC_20_BYTES
/* =============================================================== */
/*..ipoib_mcast_lookup -- return reference to multicast            */
int ipoib_mcast_lookup(struct net_device *dev,
				 tTS_IB_GID mgid,
				 struct ipoib_mcast **mmcast)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct ipoib_mcast *mcast;
	unsigned long flags;
	int ret = 0;

	spin_lock_irqsave(&priv->lock, flags);
	mcast = __ipoib_mcast_find(dev, mgid);
	if (!mcast) {
		/* Let's create a new send only group now */
		TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_MULTICAST,
			 "%s: setting up send only multicast group for "
			 IPOIB_GID_FMT, dev->name, IPOIB_GID_ARG(mgid));

		mcast = ipoib_mcast_alloc(dev);
		if (!mcast) {
			TS_REPORT_WARN(MOD_IB_NET,
				       "%s: unable to allocate memory for multicast structure",
				       dev->name);
			ret = -ENOMEM;
			goto out;
		}

		set_bit(IPOIB_MCAST_FLAG_SENDONLY, &mcast->flags);

		memcpy(mcast->mgid, mgid, sizeof(tTS_IB_GID));

		__ipoib_mcast_add(dev, mcast);

		list_add_tail(&mcast->list, &priv->multicast_list);

		/* Leave references for the calling application */
	}

	if (mcast->address_handle == NULL) {
		if (mcast->tid != TS_IB_CLIENT_QUERY_TID_INVALID)
			TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE,
				 TRACE_IB_NET_MULTICAST,
				 "%s: no address vector, but multicast join already started",
				 dev->name);
		else if (test_bit(IPOIB_MCAST_FLAG_SENDONLY, &mcast->flags))
			_ipoib_mcast_sendonly_join(mcast);

		ret = -EAGAIN;
	}

	*mmcast = mcast;

out:
	spin_unlock_irqrestore(&priv->lock, flags);

	return ret;
}

/* =============================================================== */
/*.._ipoib_mcast_tx_callback -- put reference to group after TX    */
static void _ipoib_mcast_tx_callback(void *ptr)
{
	ipoib_mcast_put((struct ipoib_mcast *)ptr);
}
#endif
/* =============================================================== */
/*..ipoib_mcast_send -- send skb to multicast group                */
#ifdef MAC_20_BYTES	
 /* =============================================================== */
/*..ipoib_mcast_send -- handle sending mcast packet                */

void ipoib_mcast_send(struct net_device *dev, union ib_gid *mgid,
		      struct sk_buff *skb)
{
  	struct ipoib_dev_priv *priv = dev->priv;
    struct ipoib_mcast *mcast;
	unsigned long flags;
    
	spin_lock_irqsave(&priv->lock, flags);
	mcast = __ipoib_mcast_find(dev, mgid);
  
    if (!mcast) {
        /* Let's create a new send only group now */
    	TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE, TRACE_IB_NET_MULTICAST,
			 "%s: setting up send only multicast group for "
			 IPOIB_GID_FMT, dev->name, IPOIB_GID_ARG(*mgid));

		mcast = ipoib_mcast_alloc(dev);
		if (!mcast) {
			TS_REPORT_WARN(MOD_IB_NET,
				       "%s: unable to allocate memory for multicast structure",
				       dev->name);

            dev_kfree_skb_any(skb);
            goto out;
		}

		set_bit(IPOIB_MCAST_FLAG_SENDONLY, &mcast->flags);
        mcast->mgid = *mgid;
        __ipoib_mcast_add(dev, mcast);
        
		list_add_tail(&mcast->list, &priv->multicast_list);

		/* Leave references for the calling application */
    }
  
	if (mcast->address_handle == NULL) {
		if (mcast->tid != TS_IB_CLIENT_QUERY_TID_INVALID)
			TS_TRACE(MOD_IB_NET, T_VERY_VERBOSE,
				 TRACE_IB_NET_MULTICAST,
				 "%s: no address vector, but multicast join already started",
				 dev->name);
		else if (test_bit(IPOIB_MCAST_FLAG_SENDONLY, &mcast->flags))
			_ipoib_mcast_sendonly_join(mcast);
  
        if (skb_queue_len(&mcast->pkt_queue) < IPOIB_MAX_MCAST_QUEUE)
			skb_queue_tail(&mcast->pkt_queue, skb);
		else
			dev_kfree_skb_any(skb);
	}
    
    
out:
	spin_unlock_irqrestore(&priv->lock, flags);
	if (mcast) {
		if (mcast->address_handle)
		  ipoib_send(dev, skb, mcast->address_handle, IB_MULTICAST_QPN);
		ipoib_mcast_put(mcast);
	}  
}


#else    
int ipoib_mcast_send(struct net_device *dev, struct ipoib_mcast *mcast,
		     struct sk_buff *skb)
{

    return ipoib_dev_send(dev, skb, _ipoib_mcast_tx_callback, mcast,
			      mcast->address_handle, IB_MULTICAST_QPN);
}
#endif


/* =============================================================== */
/*..ipoib_mcast_queue_packet -- queue skb pending join             */
int ipoib_mcast_queue_packet(struct ipoib_mcast *mcast, struct sk_buff *skb)
{
	skb_queue_tail(&mcast->pkt_queue, skb);

	return 0;
}

/* =============================================================== */
/*..ipoib_mcast_dev_flush -- flush joins and address vectors       */
void ipoib_mcast_dev_flush(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	LIST_HEAD(remove_list);
	struct ipoib_mcast *mcast, *tmcast, *nmcast;
	unsigned long flags;

	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		 "%s: flushing multicast list", dev->name);

	spin_lock_irqsave(&priv->lock, flags);
	list_for_each_entry_safe(mcast, tmcast, &priv->multicast_list, list) {
		nmcast = ipoib_mcast_alloc(dev);
		if (nmcast) {
			nmcast->flags =
			    mcast->flags & (1 << IPOIB_MCAST_FLAG_SENDONLY);

#ifdef MAC_20_BYTES
			nmcast->mgid = mcast->mgid;
#else
			memcpy(nmcast->mgid, mcast->mgid, sizeof(tTS_IB_GID));
#endif
			/* Add the new group in before the to-be-destroyed group */
			list_add_tail(&nmcast->list, &mcast->list);
			list_del_init(&mcast->list);

			rb_replace_node(&mcast->rb_node, &nmcast->rb_node,
					&priv->multicast_tree);

			ipoib_mcast_put(nmcast);	/* for _alloc() */

			list_add_tail(&mcast->list, &remove_list);
		} else {
			TS_REPORT_WARN(MOD_IB_NET,
				       "%s: could not reallocate multicast group "
				       IPOIB_GID_FMT, dev->name,
				       IPOIB_GID_ARG(mcast->mgid));
		}
	}

	if (priv->broadcast) {
		nmcast = ipoib_mcast_alloc(dev);
		if (nmcast) {

#ifdef MAC_20_BYTES
		  nmcast->mgid = priv->broadcast->mgid;
#else            
		  memcpy(nmcast->mgid, priv->broadcast->mgid,
			 sizeof(tTS_IB_GID));
#endif            
            
		  rb_replace_node(&priv->broadcast->rb_node,
				  &nmcast->rb_node,
				  &priv->multicast_tree);
		  
		  ipoib_mcast_put(nmcast);	/* for _alloc() */
		  
		  list_add_tail(&priv->broadcast->list, &remove_list);
		}

		priv->broadcast = nmcast;
	}

	spin_unlock_irqrestore(&priv->lock, flags);

	list_for_each_entry_safe(mcast, tmcast, &remove_list, list) {
		list_del_init(&mcast->list);

		_ipoib_mcast_cancel(mcast);
		ipoib_mcast_leave(dev, mcast);
		ipoib_mcast_put(mcast);
	}
}
/* ================================================================== */
/*..ipoib_mcast_dev_exit -- Exit from device => remove multicast tree */
void ipoib_mcast_dev_exit(struct net_device *dev)
{
       struct ipoib_dev_priv *priv = dev->priv;
       struct ipoib_mcast *mcast, *tmcast;

       TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		"%s: Removing multi cast list", dev->name);
       
       list_for_each_entry_safe(mcast, tmcast, &priv->multicast_list, list) {
           ipoib_mcast_put(mcast);
       }
}

/* =============================================================== */
/*..ipoib_mcast_dev_down -- delete broadcast group                 */
void ipoib_mcast_dev_down(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	/* Delete broadcast since it will be recreated */
	if (priv->broadcast) {
#ifdef MAC_20_BYTES		
        _ipoib_mcast_delete(dev, &priv->broadcast->mgid);
#else
        _ipoib_mcast_delete(dev, priv->broadcast->mgid);
#endif		
        priv->broadcast = NULL;
	}
}

/* =============================================================== */
/*..ipoib_mcast_restart_task -- rescan multicast lists             */
void ipoib_mcast_restart_task(void *dev_ptr)
{
	struct net_device *dev = dev_ptr;
	struct ipoib_dev_priv *priv = dev->priv;
   	struct ip_mc_list *im;
	struct ipoib_mcast *mcast, *tmcast;
	LIST_HEAD(remove_list);
	unsigned long flags;

	struct in_device *in_dev;
	struct ib_async_event_record async_record;
	int ret;

	if ((in_dev = in_dev_get(dev)) == NULL)
	  return;

	/*
	 * This function is a bit "evil" in some respects. We expose ourself as
	 * an ethernet device to the networking core. Unfortunately, ethernet
	 * drops the upper 5 bits of the multicast group ID when it builds the
	 * hardware address. The IB multicast spec doesn't. As a result, we
	 * need to dig into the IP layers to pull out the IP address so we can
	 * build the IB multicast GID correctly.
	 */

	ipoib_mcast_stop_thread(dev);

	spin_lock_irqsave(&priv->lock, flags);

	/*
	 * Unfortunately, the networking core only gives us a list of all of
	 * the multicast hardware addresses. We need to figure out which ones
	 * are new and which ones have been removed
	 */

	/* Clear out the found flag */
	list_for_each_entry(mcast, &priv->multicast_list, list)
	  clear_bit(IPOIB_MCAST_FLAG_FOUND, &mcast->flags);
	
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,9)
	read_lock(&in_dev->mc_list_lock);
#else
	read_lock(&in_dev->lock);
#endif

	/* Mark all of the entries that are found or don't exist */
	for (im = in_dev->mc_list; im; im = im->next) {
	  u32 multiaddr = ntohl(im->multiaddr);


#ifdef MAC_20_BYTES	
        
	  union ib_gid mgid;
	  memcpy(mgid.raw, ipoib_broadcast_mgid, sizeof mgid);
	  
	  /* Add in the P_Key */
	  mgid.raw[4] = (priv->pkey >> 8) & 0xff;
	  mgid.raw[5] = priv->pkey & 0xff;
	  
	  /* Fixup the group mapping */
	  mgid.raw[12] = (multiaddr >> 24) & 0x0f;
	  mgid.raw[13] = (multiaddr >> 16) & 0xff;
	  mgid.raw[14] = (multiaddr >> 8) & 0xff;
	  mgid.raw[15] = multiaddr & 0xff;
	  
	  mcast = __ipoib_mcast_find(dev, &mgid);
      
#else		
	  
	  tTS_IB_GID mgid;
	  memcpy(mgid, ipoib_broadcast_mgid, sizeof(tTS_IB_GID));
	  
	  /* Add in the P_Key */
	  mgid[4] = (priv->pkey >> 8) & 0xff;
	  mgid[5] = priv->pkey & 0xff;
	  
	  /* Fixup the group mapping */
	  mgid[12] = (multiaddr >> 24) & 0x0f;
	  mgid[13] = (multiaddr >> 16) & 0xff;
	  mgid[14] = (multiaddr >> 8) & 0xff;
	  mgid[15] = multiaddr & 0xff;
	  mcast = __ipoib_mcast_find(dev, mgid);

#endif		
        
	  if (!mcast
	      || test_bit(IPOIB_MCAST_FLAG_SENDONLY, &mcast->flags)) {
	    struct ipoib_mcast *nmcast;
	    
	    /* Not found or send-only group, let's add a new entry */
	    TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		     "%s: adding multicast entry for mgid "
		     IPOIB_GID_FMT, dev->name, IPOIB_GID_ARG(mgid));
	    
	    nmcast = ipoib_mcast_alloc(dev);
	    if (!nmcast) {
	      TS_REPORT_WARN(MOD_IB_NET,
			     "%s: unable to allocate memory for multicast structure",
			     dev->name);
	      continue;
	    }
	    
	    set_bit(IPOIB_MCAST_FLAG_FOUND, &nmcast->flags);
#ifdef MAC_20_BYTES
	    nmcast->mgid = mgid;
#else
	    memcpy(nmcast->mgid, mgid, sizeof(tTS_IB_GID));
#endif
	    if (mcast) {
	      /* Destroy the send only entry */
	      list_del(&mcast->list);
	      list_add_tail(&mcast->list, &remove_list);
	      
	      rb_replace_node(&mcast->rb_node,
			      &nmcast->rb_node,
			      &priv->multicast_tree);
	    } else
	      __ipoib_mcast_add(dev, nmcast);
	    
	    list_add_tail(&nmcast->list, &priv->multicast_list);
	    
	    ipoib_mcast_put(nmcast);	/* for _alloc() */
	  }
	  
	  if (mcast) {
	    set_bit(IPOIB_MCAST_FLAG_FOUND, &mcast->flags);
	    
	    /* Drop the reference for the group we found */
	    ipoib_mcast_put(mcast);	/* for _find() */
	  }
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,9)
	read_unlock(&in_dev->mc_list_lock);
#else
	read_unlock(&in_dev->lock);
#endif

	/* Remove all of the entries don't exist anymore */
	list_for_each_entry_safe(mcast, tmcast, &priv->multicast_list, list) {
	  if (!test_bit(IPOIB_MCAST_FLAG_FOUND, &mcast->flags) &&
	      !test_bit(IPOIB_MCAST_FLAG_SENDONLY, &mcast->flags)) {
	    TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		     "%s: deleting multicast group " IPOIB_GID_FMT,
		     dev->name, IPOIB_GID_ARG(mcast->mgid));
	    
	    rb_erase(&mcast->rb_node, &priv->multicast_tree);
	    
	    /* Move to the remove list */
	    list_del(&mcast->list);
	    list_add_tail(&mcast->list, &remove_list);
	  }
	}
	spin_unlock_irqrestore(&priv->lock, flags);

	/* We have to cancel outside of the spinlock */
	list_for_each_entry_safe(mcast, tmcast, &remove_list, list) {
	  list_del_init(&mcast->list);
	  _ipoib_mcast_cancel(mcast);
	  ipoib_mcast_leave(mcast->dev, mcast);
	  ipoib_mcast_put(mcast);
	}
	
	if (test_bit(IPOIB_FLAG_ADMIN_UP, &priv->flags))
        ipoib_mcast_start_thread(dev);
	 	
    in_dev_put(in_dev);
	
}

#ifdef MAC_20_BYTES
/* =============================================================== */
/*..ipoib_mcast_restart_task -- rescan multicast lists             */
void ipoib_mcast_update_thread(void *dev_ptr)
{ 
    struct net_device *dev  = dev_ptr;
    struct ipoib_dev_priv *priv = dev->priv;
  

    while (1)
    {
      /*The next will work with intel CPUs which work by command order*/
      wait_event_interruptible_timeout(priv->mcast_update_wait, 
				       (priv->mcast_rejoin == 1)  || (priv->mcast_update_stop == 1) , 
				       IPOIB_MAX_UPDATE_MULTI*HZ);

      if (priv->mcast_update_stop == 1)      {	    
	priv->mcast_update_stop = 0;
	
	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		 "%s: ipoib_mcast_update_thread Going to be dead",
		 dev->name);
	
	// I wait until to be dead
	if (signal_pending(current))
	  return;
	
	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(HZ);
	set_current_state(TASK_RUNNING);            

	/* return here anyway, since the clear intent is to stop the thread */
	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_MULTICAST,
		"%s: ipoib_mcast_update_thread. NO SIGNAL, BUT RETURNING ANYWAY",
		dev->name);
	return;   /* return anyway */

      } else {
	priv->mcast_rejoin = 0;
      
	ipoib_mcast_restart_task(dev);
      }
      
      if (signal_pending(current))
	return;
      
    }
}
#endif

/* =============================================================== */
/*..ipoib_mcast_iter_init -- create new multicast iterator         */
struct ipoib_mcast_iter *ipoib_mcast_iter_init(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct ipoib_mcast_iter *iter;
	rb_node_t *node, *parent = NULL;

	if (!priv->multicast_tree.rb_node)
		return NULL;

	iter = kmalloc(sizeof(*iter), GFP_KERNEL);
	if (!iter)
		return NULL;

	iter->dev = dev;

	/* Start at the minimum node for the tree */
	node = priv->multicast_tree.rb_node;
	while (node) {
		parent = node;
		node = node->rb_left;
	}

	iter->rb_node = parent;

	return iter;
}

/* =============================================================== */
/*..ipoib_mcast_iter_free -- free multicast iterator               */
void ipoib_mcast_iter_free(struct ipoib_mcast_iter *iter)
{
	kfree(iter);
}

/* =============================================================== */
/*..ipoib_mcast_iter_next -- incr. iter. -- return non-zero at end */
int ipoib_mcast_iter_next(struct ipoib_mcast_iter *iter)
{
	if (iter->rb_node->rb_right) {
		iter->rb_node = iter->rb_node->rb_right;
		while (iter->rb_node->rb_left)
			iter->rb_node = iter->rb_node->rb_left;
	} else {
		if (iter->rb_node->rb_parent &&
		    iter->rb_node == iter->rb_node->rb_parent->rb_left)
			iter->rb_node = iter->rb_node->rb_parent;
		else {
			while (iter->rb_node->rb_parent &&
			       iter->rb_node ==
			       iter->rb_node->rb_parent->rb_right)
                iter->rb_node = iter->rb_node->rb_parent;
			iter->rb_node = iter->rb_node->rb_parent;
		}
	}

	return iter->rb_node == NULL;
}

/* =============================================================== */
/*..ipoib_mcast_iter_read -- get data pointed to by multicast iterator */
void ipoib_mcast_iter_read(struct ipoib_mcast_iter *iter,
#ifdef MAC_20_BYTES
			   union ib_gid *mgid,           
#else			   
			   tTS_IB_GID mgid,
#endif			   
			   unsigned long *created,
			   unsigned int *queuelen,
			   unsigned int *complete,
			   unsigned int *send_only)
{
	struct ipoib_mcast *mcast;

	mcast = rb_entry(iter->rb_node, struct ipoib_mcast, rb_node);
#ifdef MAC_20_BYTES
	*mgid = mcast->mgid;
#else
	memcpy(mgid, mcast->mgid, sizeof(tTS_IB_GID));
#endif	
	*created = mcast->created;
	*queuelen = skb_queue_len(&mcast->pkt_queue);
	*complete = mcast->address_handle != NULL;
	*send_only = (mcast->flags & (1 << IPOIB_MCAST_FLAG_SENDONLY)) ? 1 : 0;
}

