/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: ipoib_proc.c 432 2004-07-12 17:27:37Z roland $
*/

#include "ipoib.h"

#include "ts_kernel_trace.h"
#include "ts_kernel_services.h"

#include <linux/ctype.h>
#include <linux/module.h>

#include <asm/uaccess.h>
#include <asm/semaphore.h>

#ifndef MAC_20_BYTES
/*
 * ARP proc file stuff
 */

static const char ipoib_arp_proc_entry_name[] = "ipoib_arp_%s";
/*
 * we have a static variable to hold the device pointer between when
 * the /proc file is opened and the seq_file start function is
 * called.  (This is a kludge to get around the fact that we don't get
 * to pass user data to the seq_file start function)
 */
static DECLARE_MUTEX(proc_arp_mutex);
static struct net_device *proc_arp_device;

/* =============================================================== */
/*.._ipoib_sarp_seq_start -- seq file handling                     */
static void *_ipoib_sarp_seq_start(struct seq_file *file, loff_t *pos)
{
	struct ipoib_sarp_iter *iter = ipoib_sarp_iter_init(proc_arp_device);
	loff_t n = *pos;

	while (n--) {
		if (ipoib_sarp_iter_next(iter)) {
			ipoib_sarp_iter_free(iter);
			return NULL;
		}
	}

	return iter;
}

/* =============================================================== */
/*.._ipoib_sarp_seq_next -- seq file handling                      */
static void *_ipoib_sarp_seq_next(struct seq_file *file, void *iter_ptr,
				  loff_t *pos)
{
	struct ipoib_sarp_iter *iter = iter_ptr;

	(*pos)++;

	if (ipoib_sarp_iter_next(iter)) {
		ipoib_sarp_iter_free(iter);
		return NULL;
	}

	return iter;
}

/* =============================================================== */
/*.._ipoib_sarp_seq_stop -- seq file handling                      */
static void _ipoib_sarp_seq_stop(struct seq_file *file, void *iter_ptr)
{
	/* nothing for now */
}

/* =============================================================== */
/*.._ipoib_sarp_seq_show -- seq file handling                      */
static int _ipoib_sarp_seq_show(struct seq_file *file, void *iter_ptr)
{
	struct ipoib_sarp_iter *iter = iter_ptr;
	uint8_t hash[IPOIB_ADDRESS_HASH_BYTES];
	char gid_buf[sizeof("ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff")];
	tTS_IB_GID gid;
	u32 qpn;
	int i, n;
	unsigned long created, last_verify;
	unsigned int queuelen, complete;

	if (iter) {
		ipoib_sarp_iter_read(iter, hash, gid, &qpn, &created,
				     &last_verify, &queuelen, &complete);

		for (i = 0; i < IPOIB_ADDRESS_HASH_BYTES; ++i) {
			seq_printf(file, "%02x", hash[i]);
			if (i < IPOIB_ADDRESS_HASH_BYTES - 1)
				seq_putc(file, ':');
			else
				seq_printf(file, "  ");
		}

		for (n = 0, i = 0; i < sizeof(tTS_IB_GID) / 2; ++i) {
			n += sprintf(gid_buf + n, "%x",
				     be16_to_cpu(((uint16_t *)gid)[i]));
			if (i < sizeof(tTS_IB_GID) / 2 - 1)
				gid_buf[n++] = ':';
		}
	}

	seq_printf(file, "GID: %*s", -(1 + (int) sizeof(gid_buf)), gid_buf);
	seq_printf(file, "QP#: 0x%06x", qpn);

	seq_printf(file,
		   " created: %10ld last_verify: %10ld queuelen: %4d complete: %d\n",
		   created, last_verify, queuelen, complete);

	return 0;
}

static struct seq_operations ipoib_sarp_seq_operations = {
	.start = _ipoib_sarp_seq_start,
	.next = _ipoib_sarp_seq_next,
	.stop = _ipoib_sarp_seq_stop,
	.show = _ipoib_sarp_seq_show,
};

/* =============================================================== */
/*.._ipoib_sarp_proc_open -- proc file handling                    */
static int _ipoib_sarp_proc_open(struct inode *inode, struct file *file)
{
	struct proc_dir_entry *pde = PDE(inode);
	struct net_device *dev = proc_arp_device;

	if (down_interruptible(&proc_arp_mutex))
		return -ERESTARTSYS;

	proc_arp_device = pde->data;

	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_GEN,
		 "%s: opening proc file", dev->name);

	return seq_open(file, &ipoib_sarp_seq_operations);
}

/*
  _ipoib_ascii_to_gid is adapted from BSD's inet_pton6, which was
  originally written by Paul Vixie
*/

/* =============================================================== */
/*.._ipoib_ascii_to_gid -- read GID from string                     */
static int _ipoib_ascii_to_gid(const char *src, tTS_IB_GID dst)
{
	static const char xdigits[] = "0123456789abcdef";
	unsigned char *tp, *endp, *colonp;
	const char *curtok;
	int ch, saw_xdigit;
	unsigned int val;

	memset((tp = dst), 0, sizeof(tTS_IB_GID));
	endp = tp + sizeof(tTS_IB_GID);
	colonp = NULL;

	/* Leading :: requires some special handling. */
	if (*src == ':' && *++src != ':')
		return 0;

	curtok = src;
	saw_xdigit = 0;
	val = 0;

	while ((ch = *src++) != '\0') {
		const char *pch;

		pch = strchr(xdigits, tolower(ch));

		if (pch) {
			val <<= 4;
			val |= (pch - xdigits);
			if (val > 0xffff)
				return 0;

			saw_xdigit = 1;
			continue;
		}

		if (ch == ':') {
			curtok = src;

			if (!saw_xdigit) {
				if (colonp)
					return 0;

				colonp = tp;
				continue;
			} else if (*src == '\0')
				return 0;

			if (tp + 2 > endp)
				return 0;

			*tp++ = (u_char) (val >> 8) & 0xff;
			*tp++ = (u_char) val & 0xff;
			saw_xdigit = 0;
			val = 0;
			continue;
		}

		return 0;
	}

	if (saw_xdigit) {
		if (tp + 2 > endp)
			return 0;

		*tp++ = (u_char) (val >> 8) & 0xff;
		*tp++ = (u_char) val & 0xff;
	}

	if (colonp) {
		memmove(endp - (tp - colonp), colonp, tp - colonp);
		memset(colonp, 0, tp - colonp);
		tp = endp;
	}

	if (tp != endp)
		return 0;

	return 1;
}

/* =============================================================== */
/*.._ipoib_sarp_proc_write -- proc file handling                   */
static ssize_t _ipoib_sarp_proc_write(struct file *file, const char *buffer,
				      size_t count, loff_t *pos)
{
	struct ipoib_sarp *entry;
	char kernel_buf[256];
	char gid_buf[sizeof("ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff")];
	tTS_IB_GID gid;
	u32 qpn;

	count = min(count, sizeof(kernel_buf));

	if (copy_from_user(kernel_buf, buffer, count))
		return -EFAULT;

	kernel_buf[count - 1] = '\0';

	if (sscanf(kernel_buf, "%39s %i", gid_buf, &qpn) != 2)
		return -EINVAL;

	if (!_ipoib_ascii_to_gid(gid_buf, gid))
		return -EINVAL;

	if (qpn > 0xffffff)
		return -EINVAL;

	entry = ipoib_sarp_add(proc_arp_device, gid, qpn);
	if (entry)
		ipoib_sarp_put(entry);

	return count;
}

/* =============================================================== */
/*.._ipoib_sarp_proc_release -- proc file handling                 */
static int _ipoib_sarp_proc_release(struct inode *inode, struct file *file)
{
	up(&proc_arp_mutex);

	return seq_release(inode, file);
}

static struct file_operations ipoib_sarp_proc_device_operations = {
	.open = _ipoib_sarp_proc_open,
	.read = seq_read,
	.write = _ipoib_sarp_proc_write,
	.llseek = seq_lseek,
	.release = _ipoib_sarp_proc_release,
};
#endif

/*
 * Multicast proc stuff
 */

static const char ipoib_mcast_proc_entry_name[] = "ipoib_mcast_%s";
/*
 * we have a static variable to hold the device pointer between when
 * the /proc file is opened and the seq_file start function is
 * called.  (This is a kludge to get around the fact that we don't get
 * to pass user data to the seq_file start function)
 */
static DECLARE_MUTEX(proc_mcast_mutex);
static struct net_device *proc_mcast_device;

/* =============================================================== */
/*.._ipoib_mcast_seq_start -- seq file handling                    */
static void *_ipoib_mcast_seq_start(struct seq_file *file, loff_t *pos)
{
	struct ipoib_mcast_iter *iter =
		ipoib_mcast_iter_init(proc_mcast_device);
	loff_t n = *pos;

	while (n--) {
		if (ipoib_mcast_iter_next(iter)) {
			ipoib_mcast_iter_free(iter);
			return NULL;
		}
	}

	return iter;
}

/* =============================================================== */
/*.._ipoib_mcast_seq_next -- seq file handling                     */
static void *_ipoib_mcast_seq_next(struct seq_file *file, void *iter_ptr,
				   loff_t *pos)
{
	struct ipoib_mcast_iter *iter = iter_ptr;

	(*pos)++;

	if (ipoib_mcast_iter_next(iter)) {
		ipoib_mcast_iter_free(iter);
		return NULL;
	}

	return iter;
}

/* =============================================================== */
/*.._ipoib_mcast_seq_stop -- seq file handling                     */
static void _ipoib_mcast_seq_stop(struct seq_file *file, void *iter_ptr)
{
	/* nothing for now */
}

#ifndef MAC_20_BYTES
/* =============================================================== */
/*.._ipoib_mcast_seq_show -- seq file handling                     */
static int _ipoib_mcast_seq_show(struct seq_file *file, void *iter_ptr)
{
	struct ipoib_mcast_iter *iter = iter_ptr;
	char gid_buf[sizeof("ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff")];
    tTS_IB_GID mgid;
    int i, n;
	unsigned long created;
	unsigned int queuelen, complete, send_only;

	if (iter) {
        ipoib_mcast_iter_read(iter, mgid, &created, &queuelen,
				      &complete, &send_only);        
		for (n = 0, i = 0; i < sizeof(tTS_IB_GID) / 2; ++i) {
			n += sprintf(gid_buf + n, "%x",
				     be16_to_cpu(((uint16_t *)mgid)[i]));
			if (i < sizeof(tTS_IB_GID) / 2 - 1)
				gid_buf[n++] = ':';
		}
	}

	seq_printf(file, "GID: %*s", -(1 + (int) sizeof(gid_buf)), gid_buf);

	seq_printf(file,
		   " created: %10ld queuelen: %4d complete: %d send_only: %d\n",
		   created, queuelen, complete, send_only);

	return 0;
}
#else
/* =============================================================== */
/*.._ipoib_mcast_seq_show -- seq file handling                     */
static int _ipoib_mcast_seq_show(struct seq_file *file, void *iter_ptr)
{
	struct ipoib_mcast_iter *iter = iter_ptr;
	char gid_buf[sizeof("ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff")];
	union ib_gid mgid;
	int i, n;
	unsigned long created;
	unsigned int queuelen, complete, send_only;

	if (iter) {
		ipoib_mcast_iter_read(iter, &mgid, &created, &queuelen,
				      &complete, &send_only);

		for (n = 0, i = 0; i < sizeof mgid / 2; ++i) {
			n += sprintf(gid_buf + n, "%x",
				     be16_to_cpu(((u16 *)mgid.raw)[i]));
			if (i < sizeof mgid / 2 - 1)
				gid_buf[n++] = ':';
		}
	}

	seq_printf(file, "GID: %*s", -(1 + (int) sizeof(gid_buf)), gid_buf);

	seq_printf(file,
		   " created: %10ld queuelen: %4d complete: %d send_only: %d\n",
		   created, queuelen, complete, send_only);

	return 0;
}
#endif

static struct seq_operations ipoib_mcast_seq_operations = {
	.start = _ipoib_mcast_seq_start,
	.next = _ipoib_mcast_seq_next,
	.stop = _ipoib_mcast_seq_stop,
	.show = _ipoib_mcast_seq_show,
};

/* =============================================================== */
/*.._ipoib_mcast_proc_open -- proc file handling                   */
static int _ipoib_mcast_proc_open(struct inode *inode, struct file *file)
{
	struct proc_dir_entry *pde = PDE(inode);
	struct net_device *dev = proc_mcast_device;

	if (down_interruptible(&proc_mcast_mutex))
		return -ERESTARTSYS;

	proc_mcast_device = pde->data;
	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_GEN,
		 "%s: opening proc file", dev->name);

	return seq_open(file, &ipoib_mcast_seq_operations);
}

/* =============================================================== */
/*.._ipoib_mcast_proc_release -- proc file handling                */
static int _ipoib_mcast_proc_release(struct inode *inode, struct file *file)
{
	up(&proc_mcast_mutex);

	return seq_release(inode, file);
}

static struct file_operations ipoib_mcast_proc_device_operations = {
	.open = _ipoib_mcast_proc_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = _ipoib_mcast_proc_release,
};

/* =============================================================== */
/*..ipoib_proc_dev_init -- set up ipoib_arp in /proc               */
int ipoib_proc_dev_init(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
#if defined MAC_20_BYTES
    char name[sizeof(ipoib_mcast_proc_entry_name) + sizeof (dev->name)];
#else
    char name[sizeof(ipoib_arp_proc_entry_name) + sizeof (dev->name)];
    
    snprintf(name, sizeof(name) - 1, ipoib_arp_proc_entry_name, dev->name);
	priv->arp_proc_entry = create_proc_entry(name,
						 S_IRUGO | S_IWUGO,
						 tsKernelProcDirGet());

	if (!priv->arp_proc_entry) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: Can't create %s in proc directory",
				dev->name, name);
		return -ENOMEM;
	}

	priv->arp_proc_entry->proc_fops = &ipoib_sarp_proc_device_operations;
	priv->arp_proc_entry->data = dev;
#endif

	snprintf(name, sizeof(name) - 1, ipoib_mcast_proc_entry_name, dev->name);
	priv->mcast_proc_entry = create_proc_entry(name,
						   S_IRUGO,
						   tsKernelProcDirGet());
	if (!priv->mcast_proc_entry) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: Can't create %s in proc directory",
				dev->name, name);
		/* FIXME: Delete ARP proc entry */
		return -ENOMEM;
	}

	priv->mcast_proc_entry->proc_fops = &ipoib_mcast_proc_device_operations;
	priv->mcast_proc_entry->data = dev;

	return 0;
}

/* =============================================================== */
/*..ipoib_proc_dev_cleanup -- unregister /proc file                */
void ipoib_proc_dev_cleanup(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

#if defined MAC_20_BYTES
    char name[sizeof(ipoib_mcast_proc_entry_name) + sizeof (dev->name)];
#else

    char name[sizeof(ipoib_arp_proc_entry_name) + sizeof(dev->name)];

	if (priv->arp_proc_entry) {
		snprintf(name, sizeof(name) - 1, ipoib_arp_proc_entry_name,
			 dev->name);

		TS_REPORT_CLEANUP(MOD_IB_NET,
				  "%s: removing proc file %s", dev->name, name);

		remove_proc_entry(name, tsKernelProcDirGet());
	}
#endif

	if (priv->mcast_proc_entry) {
		snprintf(name, sizeof(name) - 1, ipoib_mcast_proc_entry_name,
			 dev->name);

		TS_REPORT_CLEANUP(MOD_IB_NET,
				  "%s: removing proc file %s", dev->name, name);

		remove_proc_entry(name, tsKernelProcDirGet());
	}
}

