/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: ipoib_verbs.c 426 2004-07-12 04:39:39Z roland $
*/

#include "ipoib.h"

#include "ts_kernel_trace.h"

#include <ts_ib_tavor_provider.h>

extern int ipoib_max_address_vector;
extern int ipoib_use_interrupt_context;
/* =================================================================== */
/*.. ipoib_pkey_dev_check_presence - Check for the interface P_Key presence */
void ipoib_pkey_dev_check_presence(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	int pkey_index = 0;

	if (ib_cached_pkey_find(priv->ca, priv->port, priv->pkey, &pkey_index))
		clear_bit(IPOIB_PKEY_ASSIGNED, &priv->flags);
	else
		set_bit(IPOIB_PKEY_ASSIGNED, &priv->flags);
}
#if defined(MAC_20_BYTES)
int ipoib_mcast_attach(struct net_device *dev, u16 mlid, union ib_gid *mgid)
#else
int ipoib_mcast_attach(struct net_device *dev, u16 mlid, tTS_IB_GID mgid)
#endif
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct ib_qp_attribute *qp_attr;
	int ret, pkey_index;

	ret = -ENOMEM;
	qp_attr = kmalloc(sizeof(*qp_attr), GFP_ATOMIC);
	if (!qp_attr)
		goto ret;

	if (ib_cached_pkey_find(priv->ca, priv->port, priv->pkey, &pkey_index)) {
		clear_bit(IPOIB_PKEY_ASSIGNED, &priv->flags);
		ret = -ENXIO;
		goto out;
	}
	set_bit(IPOIB_PKEY_ASSIGNED, &priv->flags);

	/* set correct QKey for QP */
	qp_attr->qkey = priv->qkey;
	qp_attr->valid_fields = IB_QP_ATTRIBUTE_QKEY;
	ret = ib_qp_modify(priv->qp, qp_attr);
	if (ret) {
		TS_REPORT_FATAL(MOD_IB_NET, "%s: failed to modify QP, ret = %d",
				dev->name, ret);
		goto out;
	}

	/* attach QP to multicast group */
	down(&priv->mcast_mutex);
#if defined(MAC_20_BYTES)   
	ret = ib_multicast_attach(mlid, mgid->raw, priv->qp);
#else
	ret = ib_multicast_attach(mlid, mgid, priv->qp);
#endif	

	up(&priv->mcast_mutex);
	if (ret)
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: failed to attach to multicast group, ret = %d",
				dev->name, ret);

out:
	kfree(qp_attr);
ret:
	return ret;
}
#if defined(MAC_20_BYTES)
int ipoib_mcast_detach(struct net_device *dev, u16 mlid, union ib_gid *mgid)
#else
int ipoib_mcast_detach(struct net_device *dev, u16 mlid, tTS_IB_GID mgid)
#endif
{
	struct ipoib_dev_priv *priv = dev->priv;
	int ret;

	down(&priv->mcast_mutex);
	
#if defined(MAC_20_BYTES)
	ret = ib_multicast_detach(mlid, mgid->raw, priv->qp);
#else
	ret = ib_multicast_detach(mlid, mgid, priv->qp);
#endif	
	up(&priv->mcast_mutex);
	if (ret)
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: ib_multicast_detach failed (result = %d)",
			       dev->name, ret);

	return ret;
}

int ipoib_qp_create(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	int ret, pkey_index;
	struct ib_qp_create_param qp_create = {
		.limit = {
			  .max_outstanding_send_request = IPOIB_TX_RING_SIZE,
			  .max_outstanding_receive_request = IPOIB_RX_RING_SIZE,
			  .max_send_gather_element = 1,
			  .max_receive_scatter_element = 1,
			  },
		.pd = priv->pd,
		.send_queue = priv->cq_tq,
		.receive_queue = priv->cq_rq,
		.transport = IB_TRANSPORT_UD,
		.send_policy = IB_WQ_SIGNAL_SELECTABLE,
	};
	struct ib_qp_attribute *qp_attr;

	qp_attr = kmalloc(sizeof *qp_attr, GFP_KERNEL);
	if ( !qp_attr ) {
		return -ENOMEM;
	}

	/*
	 * Search through the port P_Key table for the requested pkey value.
	 * The port has to be assigned to the respective IB partition in
	 * advance.
	 */
	ret = ib_cached_pkey_find(priv->ca, priv->port, priv->pkey, &pkey_index);
	if (ret) {
		clear_bit(IPOIB_PKEY_ASSIGNED, &priv->flags);
		goto exit_free;
	}
	set_bit(IPOIB_PKEY_ASSIGNED, &priv->flags);

	ret = ib_qp_create(&qp_create, &priv->qp, &priv->local_qpn);
	if (ret) {
		TS_REPORT_FATAL(MOD_IB_NET, "%s: failed to create QP",
				dev->name);
		goto exit_free;
	}

#if defined(MAC_20_BYTES)
    
	priv->dev.dev_addr[1] = (priv->local_qpn >> 16) & 0xff;
	priv->dev.dev_addr[2] = (priv->local_qpn >>  8) & 0xff;
	priv->dev.dev_addr[3] = (priv->local_qpn      ) & 0xff;

#endif


	qp_attr->state = IB_QP_STATE_INIT;
	qp_attr->qkey = 0;
	qp_attr->port = priv->port;
	qp_attr->pkey_index = pkey_index;
	qp_attr->valid_fields =
	    IB_QP_ATTRIBUTE_QKEY |
	    IB_QP_ATTRIBUTE_PORT |
	    IB_QP_ATTRIBUTE_PKEY_INDEX |
	    IB_QP_ATTRIBUTE_STATE;
	ret = ib_qp_modify(priv->qp, qp_attr);
	if (ret) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: failed to modify QP to init, ret = %d",
				dev->name, ret);
		goto out_fail;
	}

	qp_attr->state = IB_QP_STATE_RTR;
	/* Can't set this in a INIT->RTR transition */
	qp_attr->valid_fields &= ~IB_QP_ATTRIBUTE_PORT;
	ret = ib_qp_modify(priv->qp, qp_attr);
	if (ret) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: failed to modify QP to RTR, ret = %d",
				dev->name, ret);
		goto out_fail;
	}

	qp_attr->state = IB_QP_STATE_RTS;
	qp_attr->send_psn = 0;
	qp_attr->valid_fields |= IB_QP_ATTRIBUTE_SEND_PSN;
	qp_attr->valid_fields &= ~IB_QP_ATTRIBUTE_PKEY_INDEX;
	ret = ib_qp_modify(priv->qp, qp_attr);
	if (ret) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: failed to modify QP to RTS, ret = %d",
				dev->name, ret);
		goto out_fail;
	}
	goto exit_free;

out_fail:
	ib_qp_destroy(priv->qp);
	priv->qp = NULL;

exit_free:
	kfree(qp_attr);

	return ret;
}

void ipoib_qp_destroy(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	if (ib_qp_detach_destroy(priv->qp))
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: ib_qp_detach_destroy failed", dev->name);

	priv->qp = NULL;
}

int ipoib_transport_dev_init(struct net_device *dev, struct ib_device *ca)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct ib_cq_callback cq_rq_callback = {
		.context  = IB_CQ_CALLBACK_PROCESS,
		.policy   = IB_CQ_PROVIDER_REARM,
		.function = { .entry = ipoib_ib_rq_completion },
		.arg = dev,
	};
	
	struct ib_cq_callback cq_tq_callback = {
		.context  = IB_CQ_CALLBACK_PROCESS,
		.policy   = IB_CQ_CONSUMER_REARM,
		.function = { .event = ipoib_ib_tq_event_handler },
        .arg = dev,
	};
	int rq_entries = IPOIB_RX_RING_SIZE + 1;
 	int tq_entries = IPOIB_TX_RING_SIZE + 1;

	struct ib_mlx_pd_param pd_param = {
		.special_qp = 0,
		.av_num = ipoib_max_address_vector,
	};
	
	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_GEN,
		 "%s: The number of IPoIB addr Vector is 0x%x", dev->name, ipoib_max_address_vector);

	if (ipoib_use_interrupt_context) {
		cq_tq_callback.context = IB_CQ_CALLBACK_INTERRUPT;
		cq_rq_callback.context = IB_CQ_CALLBACK_INTERRUPT;
		TS_REPORT_WARN(MOD_IB_NET,"%s: Using interrupt context",
			       dev->name);
	}

	if (ib_pd_create(priv->ca, &pd_param, &priv->pd)) {
		TS_REPORT_FATAL(MOD_IB_NET, "%s: failed to allocate PD",
				dev->name);
		return -ENODEV;
	}
	
	if (ib_cq_create(priv->ca, &rq_entries, &cq_rq_callback, NULL, &priv->cq_rq)) {
		TS_REPORT_FATAL(MOD_IB_NET, "%s: failed to create CQ",
				dev->name);
		goto out_free_pd;
	}
	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_GEN,
		 "%s: RQ CQ with %d entries", dev->name, rq_entries);
	
	if (ib_cq_create(priv->ca, &tq_entries, &cq_tq_callback, NULL, &priv->cq_tq)) {
		TS_REPORT_FATAL(MOD_IB_NET, "%s: failed to create CQ",
				dev->name);
		goto out_free_cq_rq;
	}
	
	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_GEN,
		 "%s: TQ CQ with %d entries", dev->name, tq_entries);		
	{
		/* XXX we assume physical memory starts at address 0. */
		struct ib_physical_buffer buffer_list = {
			.address = 0,
			.size = 1
		};
		uint64_t dummy_iova = 0;
		unsigned long tsize = (unsigned long)high_memory - PAGE_OFFSET;
		u32 rkey;

		/*
		 * Make our region have size the size of low memory rounded
		 * up to the next power of 2 (so we use as few TPT entries
		 * as possible)
		 */
		while (tsize) {
			buffer_list.size <<= 1;
			tsize >>= 1;
		}

		if (ib_memory_register_physical(priv->pd, &buffer_list,
						1,	/* list_len */
						&dummy_iova, buffer_list.size,
						0,	/* iova_offset */
						IB_ACCESS_LOCAL_WRITE,
						&priv->mr,
						&priv->lkey, &rkey)) {
			TS_REPORT_FATAL(MOD_IB_NET,
					"%s: ib_memory_register_physical failed",
					dev->name);
			goto out_free_cq_tq;
		}
	}

	return 0;

out_free_cq_tq:
	ib_cq_destroy(priv->cq_tq);

out_free_cq_rq:
	ib_cq_destroy(priv->cq_rq);

out_free_pd:
	ib_pd_destroy(priv->pd);
	return -ENODEV;
}

void ipoib_transport_dev_cleanup(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	if (priv->qp != NULL) {
		if (ib_qp_destroy(priv->qp))
			TS_REPORT_WARN(MOD_IB_NET,
				       "%s: ib_qp_destroy failed", dev->name);

		priv->qp = NULL;
		clear_bit(IPOIB_PKEY_ASSIGNED, &priv->flags);
	}

	if (ib_memory_deregister(priv->mr))
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: ib_memory_deregister failed", dev->name);

	if (ib_cq_destroy(priv->cq_rq))
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: ib_cq_destroy failed", dev->name);
	
	if (ib_cq_destroy(priv->cq_tq))
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: ib_cq_destroy failed", dev->name);


	if (ib_pd_destroy(priv->pd))
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: ib_pd_destroy failed", dev->name);
}

static void ipoib_device_notifier(struct ib_device_notifier *self,
				  struct ib_device *device, int event)
{
	struct ib_device_properties *props;
	int port;

	props = kmalloc(sizeof *props, GFP_KERNEL);
	if ( !props ) {
		TS_REPORT_WARN(MOD_IB_NET, "kmalloc(%d) failed", sizeof *props);
		return;
	}

	if (ib_device_properties_get(device, props)) {
		TS_REPORT_WARN(MOD_IB_NET, "ib_device_properties_get failed");
		goto exit;
	}

	switch (event) {
	case IB_DEVICE_NOTIFIER_ADD:
		if (props->is_switch)
			ipoib_add_port("ib%d", device, 0);
		else {
			for (port = 1; port <= props->num_port; ++port)
				ipoib_add_port("ib%d", device, port);
		}
		break;

	case IB_DEVICE_NOTIFIER_REMOVE:
		/* Yikes! We don't support devices going away from
		   underneath us yet! */
		TS_REPORT_WARN(MOD_IB_NET,
			       "IPoIB driver can't handle removal of device %s",
			       props->name);
		break;

	default:
		TS_REPORT_WARN(MOD_IB_NET, "Unknown device notifier event %d.");
		break;
	}
exit:
	kfree(props);
}

static struct ib_device_notifier ipoib_notifier = {
	.notifier = ipoib_device_notifier
};

int ipoib_transport_create_devices(void)
{
	ib_device_notifier_register(&ipoib_notifier);
	return 0;
}

void ipoib_transport_cleanup(void)
{
	ib_device_notifier_deregister(&ipoib_notifier);
}

static void ipoib_async_event(struct ib_async_event_record *record,
			      void *priv_ptr)
{
	struct ipoib_dev_priv *priv = priv_ptr;

	if (record->event == IB_PORT_ACTIVE) {
		TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_GEN,
			 "%s: Port active Event", priv->dev.name);

		ipoib_ib_dev_flush(&priv->dev);
	} else
		TS_REPORT_WARN(MOD_IB_NET,
			       "%s: Unexpected event %d", priv->dev.name,
			       record->event);
}

int ipoib_port_monitor_dev_start(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;
	struct ib_async_event_record event_record = {
		.device = priv->ca,
		.event = IB_PORT_ACTIVE,
	};

	if (ib_async_event_handler_register(&event_record,
					    ipoib_async_event,
					    priv, &priv->active_handler)) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"ib_async_event_handler_register failed for TS_IB_PORT_ACTIVE");
		return -EINVAL;
	}

	return 0;
}

void ipoib_port_monitor_dev_stop(struct net_device *dev)
{
	struct ipoib_dev_priv *priv = dev->priv;

	ib_async_event_handler_deregister(priv->active_handler);
}

/*
  Local Variables:
  c-file-style: "linux"
  indent-tabs-mode: t
  End:
*/
