/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: ipoib_vlan.c 240 2004-06-08 03:04:17Z roland $
*/

#include "ipoib.h"

#include "ts_kernel_services.h"
#include "ts_kernel_trace.h"

#include <linux/version.h>
#include <linux/module.h>

#include <linux/init.h>
#include <linux/slab.h>

#include <asm/uaccess.h>

struct ipoib_vlan_iter {
	struct list_head *pintf_cur;
	struct list_head *intf_cur;
};

static DECLARE_MUTEX(proc_mutex);

int ipoib_vlan_add(struct net_device *pdev, char *intf_name,
		   unsigned short pkey)
{
	struct ipoib_dev_priv *ppriv, *priv;
	int result = -ENOMEM;

#if !defined(TS_KERNEL_2_6)	/* XXX ??? */
	/* Make sure this interface belongs to this driver */
	if (pdev->owner != THIS_MODULE)
		return -EINVAL;
#endif

	if (!capable(CAP_NET_ADMIN))
		return -EPERM;

	ppriv = pdev->priv;

	/*
	 * First ensure this isn't a duplicate. We check the parent device and
	 * then all of the child interfaces to make sure the Pkey doesn't match.
	 */
	if (ppriv->pkey == pkey)
		return -ENOTUNIQ;

	down(&ipoib_device_mutex);
	list_for_each_entry(priv, &ppriv->child_intfs, list) {
		if (priv->pkey == pkey) {
			up(&ipoib_device_mutex);
			return -ENOTUNIQ;
		}
	}
	up(&ipoib_device_mutex);

	priv = ipoib_intf_alloc();
	if (!priv)
		goto alloc_mem_failed;

	set_bit(IPOIB_FLAG_SUBINTERFACE, &priv->flags);

	priv->pkey = pkey;

#ifdef MAC_20_BYTES 
	priv->dev.broadcast[8] = pkey >> 8;
	priv->dev.broadcast[9] = pkey & 0xff;
#endif	
	strncpy(priv->dev.name, intf_name, sizeof(priv->dev.name));

	result = ipoib_dev_init(&priv->dev, ppriv->ca, ppriv->port);
	if (result < 0) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"failed to initialize net device %d, port %d",
				ppriv->ca, ppriv->port);
		goto device_init_failed;
	}

	result = register_netdev(&priv->dev);
	if (result) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"%s: failed to initialize; error %i",
				priv->dev.name, result);
		goto register_failed;
	}

	down(&ipoib_device_mutex);
	list_add_tail(&priv->list, &ppriv->child_intfs);
	up(&ipoib_device_mutex);

	return 0;

register_failed:
	ipoib_dev_cleanup(&priv->dev);

device_init_failed:
	kfree(priv);

alloc_mem_failed:
	return result;
}

int ipoib_vlan_delete(struct net_device *pdev, unsigned short pkey)
{
	struct ipoib_dev_priv *ppriv, *priv, *tpriv;

#if !defined(TS_KERNEL_2_6)	/* XXX ??? */
	/* Make sure this interface belongs to this driver */
	if (pdev->owner != THIS_MODULE)
		return -EINVAL;
#endif

	if (!capable(CAP_NET_ADMIN))
		return -EPERM;

	ppriv = pdev->priv;

	down(&ipoib_device_mutex);
	list_for_each_entry_safe(priv, tpriv, &ppriv->child_intfs, list) {
		if (priv->pkey == pkey) {
			if (priv->dev.flags & IFF_UP) {
				up(&ipoib_device_mutex);
				return -EBUSY;
			}

			ipoib_dev_cleanup(&priv->dev);
			unregister_netdev(&priv->dev);

			list_del(&priv->list);

			kfree(priv);
			up(&ipoib_device_mutex);

			return 0;
		}
	}
	up(&ipoib_device_mutex);

	return -ENOENT;
}

/* =============================================================== */
/*..ipoib_vlan_iter_next -- incr. iter. -- return non-zero at end  */
int ipoib_vlan_iter_next(struct ipoib_vlan_iter *iter)
{
	while (1) {
		struct ipoib_dev_priv *priv;

		priv = list_entry(iter->pintf_cur, struct ipoib_dev_priv, list);
		if (!iter->intf_cur)
			iter->intf_cur = priv->child_intfs.next;
		else
			iter->intf_cur = iter->intf_cur->next;

		if (iter->intf_cur == &priv->child_intfs) {
			iter->pintf_cur = iter->pintf_cur->next;
			if (iter->pintf_cur == &ipoib_device_list)
				return 1;

			iter->intf_cur = NULL;
			return 0;
		} else
			return 0;
	}
}

/* =============================================================== */
/*.._ipoib_vlan_seq_start -- seq file handling                     */
static void *_ipoib_vlan_seq_start(struct seq_file *file, loff_t *pos)
{
	struct ipoib_vlan_iter *iter;
	loff_t n = *pos;

	iter = kmalloc(sizeof(*iter), GFP_KERNEL);
	if (!iter)
		return NULL;

	iter->pintf_cur = ipoib_device_list.next;
	iter->intf_cur = NULL;

	while (n--) {
		if (ipoib_vlan_iter_next(iter)) {
			kfree(iter);
			return NULL;
		}
	}

	return iter;
}

/* =============================================================== */
/*.._ipoib_vlan_seq_next -- seq file handling                      */
static void *_ipoib_vlan_seq_next(struct seq_file *file, void *iter_ptr,
				  loff_t *pos)
{
	struct ipoib_vlan_iter *iter = iter_ptr;

	(*pos)++;

	if (ipoib_vlan_iter_next(iter)) {
		kfree(iter);
		return NULL;
	}

	return iter;
}

/* =============================================================== */
/*.._ipoib_vlan_seq_stop -- seq file handling                      */
static void _ipoib_vlan_seq_stop(struct seq_file *file, void *iter_ptr)
{
	struct ipoib_vlan_iter *iter = iter_ptr;

	kfree(iter);
}

/* =============================================================== */
/*.._ipoib_vlan_seq_show -- seq file handling                      */
static int _ipoib_vlan_seq_show(struct seq_file *file, void *iter_ptr)
{
	struct ipoib_vlan_iter *iter = iter_ptr;

	if (iter) {
		struct ipoib_dev_priv *ppriv;

		ppriv = list_entry(iter->pintf_cur, struct ipoib_dev_priv, list);

		if (!iter->intf_cur)
			seq_printf(file, "%s 0x%04x\n", ppriv->dev.name,
				   ppriv->pkey);
		else {
			struct ipoib_dev_priv *priv;

			priv = list_entry(iter->intf_cur, struct ipoib_dev_priv,
				          list);

			seq_printf(file, " %s %s 0x%04x\n", ppriv->dev.name,
				   priv->dev.name, priv->pkey);
		}
	}

	return 0;
}

static struct seq_operations ipoib_vlan_seq_operations = {
	.start = _ipoib_vlan_seq_start,
	.next = _ipoib_vlan_seq_next,
	.stop = _ipoib_vlan_seq_stop,
	.show = _ipoib_vlan_seq_show,
};

/* =============================================================== */
/*.._ipoib_vlan_proc_open -- proc file handling                    */
static int _ipoib_vlan_proc_open(struct inode *inode, struct file *file)
{
	if (down_interruptible(&proc_mutex))
		return -ERESTARTSYS;

	TS_TRACE(MOD_IB_NET, T_VERBOSE, TRACE_IB_NET_GEN, "opening proc file");

	return seq_open(file, &ipoib_vlan_seq_operations);
}

/* =============================================================== */
/*.._ipoib_vlan_proc_write -- proc file handling                   */
static ssize_t _ipoib_vlan_proc_write(struct file *file, const char *buffer,
				      size_t count, loff_t *pos)
{
	int result;
	char *kernel_buf=NULL;
	char *intf_parent=NULL, *intf_name=NULL;
	unsigned int pkey;
	struct net_device *pdev;
	ssize_t rc = -EINVAL; /* initialized with a value to prevent compiler warning */

	kernel_buf = kmalloc(256, GFP_ATOMIC);
	if ( !kernel_buf ) {
		rc = -ENOMEM;
		goto exit;
	}


	intf_parent = kmalloc(128, GFP_ATOMIC);
	if ( !intf_parent ) {
		rc = -ENOMEM;
		goto exit;
	}

	intf_name = kmalloc(128, GFP_ATOMIC);
	if ( !intf_name ) {
		rc = -ENOMEM;
		goto exit;
	}

	count = min(count, sizeof(kernel_buf));

	if (copy_from_user(kernel_buf, buffer, count)) {
		rc = -EFAULT;
		goto exit;
    }

	kernel_buf[count - 1] = '\0';

	if (sscanf(kernel_buf, "add %128s %128s %i", intf_parent, intf_name,
		   &pkey) == 3) {
		if (pkey > 0xffff) {
			rc = -EINVAL;
			goto exit;
		}

		pdev = dev_get_by_name(intf_parent);
		if (!pdev) {
			rc = -ENOENT;
			goto exit;
		}

		result = ipoib_vlan_add(pdev, intf_name, pkey);

		dev_put(pdev);

		if (result < 0) {
			rc = result;
			goto exit;
		}
	} else if (sscanf(kernel_buf, "del %128s %i", intf_parent,
			  &pkey) == 2) {
		if (pkey > 0xffff) {
			rc = -EINVAL;
			goto exit;
		}

		pdev = dev_get_by_name(intf_parent);
		if (!pdev) {
			rc = -ENOENT;
			goto exit;
		}

		result = ipoib_vlan_delete(pdev, pkey);

		dev_put(pdev);

		if (result < 0) {
			rc = result;
			goto exit;
		}
	} else
		rc = -EINVAL;
		goto exit;

	rc = count;
exit:
	if ( kernel_buf ) kfree(kernel_buf);
	if ( intf_parent ) kfree (intf_parent);
	if ( intf_name ) kfree(intf_name);
	return rc;
}

/* =============================================================== */
/*.._ipoib_vlan_proc_release -- proc file handling                 */
static int _ipoib_vlan_proc_release(struct inode *inode, struct file *file)
{
	up(&proc_mutex);

	return seq_release(inode, file);
}

static struct file_operations ipoib_vlan_proc_operations = {
	.owner = THIS_MODULE,
	.open = _ipoib_vlan_proc_open,
	.read = seq_read,
	.write = _ipoib_vlan_proc_write,
	.llseek = seq_lseek,
	.release = _ipoib_vlan_proc_release,
};

struct proc_dir_entry *vlan_proc_entry;

int ipoib_vlan_init(void)
{
	vlan_proc_entry = create_proc_entry("ipoib_vlan",
					    S_IRUGO | S_IWUGO,
					    tsKernelProcDirGet());

	if (!vlan_proc_entry) {
		TS_REPORT_FATAL(MOD_IB_NET,
				"Can't create ipoib_vlan in proc directory\n");
		return -ENOMEM;
	}

	vlan_proc_entry->proc_fops = &ipoib_vlan_proc_operations;

	return 0;
}

void ipoib_vlan_cleanup(void)
{
	if (vlan_proc_entry) {
		TS_REPORT_CLEANUP(MOD_IB_NET, "removing proc file ipoib_vlan");
		remove_proc_entry("ipoib_vlan", tsKernelProcDirGet());
	}
}
