/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: sdp_advt.c 159 2004-05-14 00:16:54Z libor $
*/

#include "sdp_main.h"

static kmem_cache_t *_tsSdpConnAdvtCache = NULL;
static kmem_cache_t *_tsSdpConnAdvtTable = NULL;

/* --------------------------------------------------------------------- */
/*                                                                       */
/* module specific functions                                             */
/*                                                                       */
/* --------------------------------------------------------------------- */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* public advertisment object functions for FIFO object table            */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_advt_create - create an advertisment object */
struct sdpc_advt *sdp_advt_create(void)
{
	struct sdpc_advt *advt;

	TS_CHECK_NULL(_tsSdpConnAdvtCache, NULL);

	advt = kmem_cache_alloc(_tsSdpConnAdvtCache, SLAB_ATOMIC);

	if (NULL != advt) {

		advt->next = NULL;
		advt->prev = NULL;
		advt->size = 0;
		advt->post = 0;
		advt->addr = 0;
		advt->rkey = 0;

		advt->type = TS_SDP_GENERIC_TYPE_ADVT;
		advt->release =
		    (tSDP_GENERIC_DESTRUCT_FUNC) sdp_advt_destroy;
	}

	return advt;
} /* sdp_advt_create */

/* ========================================================================= */
/*..sdp_advt_destroy - destroy an advertisment object */
int sdp_advt_destroy(struct sdpc_advt *advt)
{
	TS_CHECK_NULL(advt, -EINVAL);

	if (NULL != advt->next || NULL != advt->prev) {

		return -EACCES;
	}
	/*
	 * return the object to its cache
	 */
	kmem_cache_free(_tsSdpConnAdvtCache, advt);

	return 0;
} /* sdp_advt_destroy */

/* ========================================================================= */
/*..sdp_advt_q_get - get, and remove, the object at the tables head */
struct sdpc_advt *sdp_advt_q_get(struct sdpc_advt_q *table)
{
	struct sdpc_advt *advt;
	struct sdpc_advt *next;
	struct sdpc_advt *prev;

	TS_CHECK_NULL(table, NULL);

	advt = table->head;
	if (NULL == advt) {

		return NULL;
	}

	if (advt->next == advt && advt->prev == advt) {

		table->head = NULL;
	}
	else {

		next = advt->next;
		prev = advt->prev;
		next->prev = prev;
		prev->next = next;

		table->head = next;
	}

	table->size--;

	advt->next = NULL;
	advt->prev = NULL;

	return advt;
} /* sdp_advt_q_get */

/* ========================================================================= */
/*..sdp_advt_q_look - get, without removing, the object at the head */
struct sdpc_advt *sdp_advt_q_look(struct sdpc_advt_q *table)
{
	TS_CHECK_NULL(table, NULL);

	return table->head;
} /* sdp_advt_q_look */

/* ========================================================================= */
/*..sdp_advt_q_put - put the advertisment object at the tables tail */
int sdp_advt_q_put(struct sdpc_advt_q *table, struct sdpc_advt *advt)
{
	struct sdpc_advt *next;
	struct sdpc_advt *prev;

	TS_CHECK_NULL(table, -EINVAL);
	TS_CHECK_NULL(advt, -EINVAL);

	if (NULL == table->head) {

		advt->next = advt;
		advt->prev = advt;
		table->head = advt;
	}
	else {

		next = table->head;
		prev = next->prev;

		prev->next = advt;
		advt->prev = prev;
		advt->next = next;
		next->prev = advt;
	}

	table->size++;

	return 0;
} /* sdp_advt_q_put */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* public table functions                                                */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_advt_q_create - create an advertisment table */
struct sdpc_advt_q *sdp_advt_q_create(tINT32 * result)
{
	struct sdpc_advt_q *table = NULL;

	TS_CHECK_NULL(result, NULL);
	*result = -EINVAL;
	TS_CHECK_NULL(_tsSdpConnAdvtTable, NULL);

	table = kmem_cache_alloc(_tsSdpConnAdvtTable, SLAB_ATOMIC);
	if (NULL == table) {

		*result = -ENOMEM;
		return NULL;
	}

	table->head = NULL;
	table->size = 0;

	*result = 0;
	return table;
} /* sdp_advt_q_create */

/* ========================================================================= */
/*..sdp_advt_q_init - initialize a new empty advertisment table */
int sdp_advt_q_init(struct sdpc_advt_q *table)
{
	TS_CHECK_NULL(_tsSdpConnAdvtCache, -EINVAL);
	TS_CHECK_NULL(table, -EINVAL);

	table->head = NULL;
	table->size = 0;

	return 0;
} /* sdp_advt_q_init */

/* ========================================================================= */
/*..sdp_advt_q_clear - clear the contents of an advertisment table */
int sdp_advt_q_clear(struct sdpc_advt_q *table)
{
	struct sdpc_advt *advt;
	int result;

	TS_CHECK_NULL(_tsSdpConnAdvtCache, -EINVAL);
	TS_CHECK_NULL(table, -EINVAL);
	/*
	 * drain the table of any objects
	 */
	while (NULL != (advt = sdp_advt_q_get(table))) {

		result = sdp_advt_destroy(advt);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	}

	return 0;
} /* sdp_advt_q_clear */

/* ========================================================================= */
/*..sdp_advt_q_destroy - destroy an advertisment table */
int sdp_advt_q_destroy(struct sdpc_advt_q *table)
{
	struct sdpc_advt *advt;
	int result;

	TS_CHECK_NULL(_tsSdpConnAdvtTable, -EINVAL);
	TS_CHECK_NULL(table, -EINVAL);
	/*
	 * drain the table of any objects
	 */
	while (NULL != (advt = sdp_advt_q_get(table))) {

		result = sdp_advt_destroy(advt);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	}
	/*
	 * return the table to the cache
	 */
	kmem_cache_free(_tsSdpConnAdvtTable, table);

	return 0;
} /* sdp_advt_q_destroy */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* primary initialization/cleanup functions                              */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_main_advt_init -- initialize the advertisment caches. */
tINT32 sdp_main_advt_init(void)
{
	int result;

	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "INIT: Advertisment cache initialization.");
	/*
	 * initialize the caches only once.
	 */
	if (NULL != _tsSdpConnAdvtCache || NULL != _tsSdpConnAdvtTable) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "INIT: Advertisment caches already initialized.");
		return -EINVAL;
	}

	_tsSdpConnAdvtCache = kmem_cache_create("SdpAdvtCache",
						sizeof(struct sdpc_advt),
						0, SLAB_HWCACHE_ALIGN, NULL,
						NULL);
	if (NULL == _tsSdpConnAdvtCache) {

		result = -ENOMEM;
		goto error_advt_c;
	}

	_tsSdpConnAdvtTable = kmem_cache_create("SdpAdvtTable",
						sizeof(struct sdpc_advt_q),
						0, SLAB_HWCACHE_ALIGN, NULL,
						NULL);
	if (NULL == _tsSdpConnAdvtTable) {

		result = -ENOMEM;
		goto error_advt_t;
	}

	return 0;
error_advt_t:
	kmem_cache_destroy(_tsSdpConnAdvtCache);
	_tsSdpConnAdvtCache = NULL;
error_advt_c:
	return 0;
} /* sdp_main_advt_init */

/* ========================================================================= */
/*..sdp_main_advt_cleanup -- cleanup the advertisment caches. */
tINT32 sdp_main_advt_cleanup(void)
{
	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "INIT: Advertisment cache cleanup.");
	/*
	 * cleanup the caches
	 */
	kmem_cache_destroy(_tsSdpConnAdvtCache);
	kmem_cache_destroy(_tsSdpConnAdvtTable);
	/*
	 * null out entries.
	 */
	_tsSdpConnAdvtCache = NULL;
	_tsSdpConnAdvtTable = NULL;

	return 0;
} /* sdp_main_advt_cleanup */
