/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sdp_buff.c 218 2004-06-02 22:02:33Z libor $
*/

#include "sdp_main.h"

static char _main_pool_name[] = TS_SDP_MAIN_POOL_NAME;
static tSDP_MAIN_POOL main_pool = NULL;
/*
 * data buffers managment API
 */
/* ========================================================================= */
/*.._sdp_buff_q_get - Get a buffer from a specific pool */
static __inline__ struct sdpc_buff *_sdp_buff_q_get(struct sdpc_buff_q *pool,
						 int fifo,
						 tSDP_BUFF_TEST_FUNC test_func,
						 void *usr_arg)
{
	struct sdpc_buff *buff;

	TS_CHECK_NULL(pool, NULL);

	if (NULL == pool->head) {

		return NULL;
	}

	if (TRUE == fifo) {

		buff = pool->head;
	}
	else {

		buff = pool->head->prev;
	}

	if (NULL == test_func || 0 == test_func(buff, usr_arg)) {

		if (buff->next == buff && buff->prev == buff) {

			pool->head = NULL;
		}
		else {

			buff->next->prev = buff->prev;
			buff->prev->next = buff->next;

			pool->head = buff->next;
		}

		pool->size--;

		buff->next = NULL;
		buff->prev = NULL;
		buff->pool = NULL;
	}
	else {

		buff = NULL;
	}

	return buff;
} /* _sdp_buff_q_get */

/* ========================================================================= */
/*.._sdp_buff_q_put - Place a buffer into a specific pool */
static __inline__ int _sdp_buff_q_put(struct sdpc_buff_q *pool,
				      struct sdpc_buff *buff,
				      int fifo)
{
	/* fifo: false == tail, true == head */
	TS_CHECK_NULL(pool, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	if (NULL != buff->pool) {

		return -EINVAL;
	}

	if (NULL == pool->head) {

		buff->next = buff;
		buff->prev = buff;
		pool->head = buff;
	}
	else {

		buff->next = pool->head;
		buff->prev = pool->head->prev;

		buff->next->prev = buff;
		buff->prev->next = buff;

		if (TRUE == fifo) {
			pool->head = buff;
		}
	}

	pool->size++;
	buff->pool = pool;

	return 0;
} /* _sdp_buff_q_put */

/* ========================================================================= */
/*.._sdp_buff_q_look - look at a buffer from a specific pool */
static __inline__ struct sdpc_buff *_sdp_buff_q_look(struct sdpc_buff_q *pool,
						  int fifo)
{
	TS_CHECK_NULL(pool, NULL);

	if (NULL == pool->head || TRUE == fifo) {

		return pool->head;
	}
	else {

		return pool->head->prev;
	}
} /* _sdp_buff_q_look */

/* ========================================================================= */
/*.._sdp_buff_q_remove - remove a specific buffer from a specific pool */
static __inline__ int _sdp_buff_q_remove(struct sdpc_buff_q *pool, 
					 struct sdpc_buff *buff)
{
	struct sdpc_buff *prev;
	struct sdpc_buff *next;

	TS_CHECK_NULL(pool, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);
	TS_CHECK_NULL(buff->next, -EINVAL);
	TS_CHECK_NULL(buff->prev, -EINVAL);
	TS_CHECK_NULL(buff->pool, -EINVAL);

	if (pool != buff->pool) {

		return -EINVAL;
	}

	if (buff->next == buff && buff->prev == buff) {

		pool->head = NULL;
	}
	else {

		next = buff->next;
		prev = buff->prev;
		next->prev = prev;
		prev->next = next;

		if (pool->head == buff) {

			pool->head = next;
		}
	}

	pool->size--;

	buff->pool = NULL;
	buff->next = NULL;
	buff->prev = NULL;

	return 0;
} /* _sdp_buff_q_remove */

/* ========================================================================= */
/*..sdp_buff_q_init - Init a pool drawing its buffers from the main pool */
int sdp_buff_q_init(struct sdpc_buff_q *pool, char *name, u32 size)
{
	int counter;
	int result;

	TS_CHECK_NULL(pool, -EINVAL);
	TS_CHECK_NULL(main_pool, -EINVAL);

	if (TS_SDP_POOL_NAME_MAX < strlen(name)) {

		return -ENAMETOOLONG;
	}

#ifdef _TS_SDP_DEBUG_POOL_NAME
	pool->name = name;
#endif
	pool->head = NULL;
	pool->size = 0;

	for (counter = 0; counter < size; counter++) {

		result = sdp_buff_q_put(pool, sdp_buff_pool_get());
		if (0 > result) {

			result = -ENOMEM;
			goto error;
		}
	}

	return 0;
error:
	(void)sdp_buff_q_clear(pool);
	return result;
} /* sdp_buff_q_init */

/* ========================================================================= */
/*..sdp_buff_q_remove - remove a specific buffer from a specific pool */
int sdp_buff_q_remove(struct sdpc_buff *buff)
{
	struct sdpc_buff_q *pool;
	int result;

	TS_CHECK_NULL(buff, -EINVAL);
	TS_CHECK_NULL(buff->pool, -EINVAL);

	pool = buff->pool;

	result = _sdp_buff_q_remove(pool, buff);

	return result;
} /* sdp_buff_q_remove */

/* ========================================================================= */
/*..sdp_buff_q_get - Get a buffer from a specific pool */
struct sdpc_buff *sdp_buff_q_get(struct sdpc_buff_q *pool)
{
	struct sdpc_buff *buff;

	buff = _sdp_buff_q_get(pool, TRUE, NULL, NULL);

	return buff;
} /* sdp_buff_q_get */

/* ========================================================================= */
/*..sdp_buff_q_get_head - Get the buffer at the front of the pool */
struct sdpc_buff *sdp_buff_q_get_head(struct sdpc_buff_q *pool)
{
	struct sdpc_buff *buff;

	buff = _sdp_buff_q_get(pool, TRUE, NULL, NULL);

	return buff;
} /* sdp_buff_q_get_head */

/* ========================================================================= */
/*..sdp_buff_q_get_tail - Get the buffer at the end of the pool */
struct sdpc_buff *sdp_buff_q_get_tail(struct sdpc_buff_q *pool)
{
	struct sdpc_buff *buff;

	buff = _sdp_buff_q_get(pool, FALSE, NULL, NULL);

	return buff;
} /* sdp_buff_q_get_tail */

/* ========================================================================= */
/*..sdp_buff_q_look_head - look at the buffer at the front of the pool */
struct sdpc_buff *sdp_buff_q_look_head(struct sdpc_buff_q *pool)
{
	struct sdpc_buff *buff;

	buff = _sdp_buff_q_look(pool, TRUE);

	return buff;
} /* sdp_buff_q_look_head */

/* ========================================================================= */
/*..sdp_buff_q_look_tail - look at the buffer at the end of the pool */
struct sdpc_buff *sdp_buff_q_look_tail(struct sdpc_buff_q *pool)
{
	struct sdpc_buff *buff;

	buff = _sdp_buff_q_look(pool, FALSE);

	return buff;
} /* sdp_buff_q_look_tail */

/* ========================================================================= */
/*..sdp_buff_q_fetch_head - Get the pools first buffer, if the test passes */
struct sdpc_buff *sdp_buff_q_fetch_head(struct sdpc_buff_q *pool,
				     tSDP_BUFF_TEST_FUNC test_func, 
				     void *usr_arg)
{
	struct sdpc_buff *buff;

	buff = _sdp_buff_q_get(pool, TRUE, test_func, usr_arg);

	return buff;
} /* sdp_buff_q_fetch_head */

/* ========================================================================= */
/*..sdp_buff_q_fetch_tail - Get the pools last buffer, if the test passes */
struct sdpc_buff *sdp_buff_q_fetch_tail(struct sdpc_buff_q *pool,
				     tSDP_BUFF_TEST_FUNC test_func,
				     void *usr_arg)
{
	struct sdpc_buff *buff;

	buff = _sdp_buff_q_get(pool, FALSE, test_func, usr_arg);

	return buff;
} /* sdp_buff_q_fetch_tail */

/* ========================================================================= */
/*..sdp_buff_q_fetch - Get the first matching buffer from the pool */
struct sdpc_buff *sdp_buff_q_fetch(struct sdpc_buff_q *pool,
				tSDP_BUFF_TEST_FUNC test_func,
				void *usr_arg)
{
	struct sdpc_buff *buff;
	int result = 0;
	int counter;

	TS_CHECK_NULL(pool, NULL);
	TS_CHECK_NULL(test_func, NULL);
	/*
	 * check to see if there is anything to traverse.
	 */
	if (NULL != pool->head) {
		/*
		 * lock to prevent corruption of table
		 */
		for (counter = 0, buff = pool->head;
		     counter < pool->size; counter++, buff = buff->next) {

			result = test_func(buff, usr_arg);
			if (0 < result) {

				result = _sdp_buff_q_remove(pool, buff);
				TS_EXPECT(MOD_LNX_SDP, !(0 > result));

				return buff;
			}

			if (0 > result) {

				break;
			}
		}
	}

	return NULL;
} /* sdp_buff_q_fetch_head */

/* ========================================================================= */
/*..sdp_buff_q_trav_tail - traverse buffers in pool,from the tail */
int sdp_buff_q_trav_tail(struct sdpc_buff_q *pool,
			 tSDP_BUFF_TRAV_FUNC trav_func,
			 void *usr_arg)
{
	struct sdpc_buff *buff;
	int result = 0;
	int counter;

	TS_CHECK_NULL(pool, -EINVAL);
	TS_CHECK_NULL(trav_func, -EINVAL);
	/*
	 * check to see if there is anything to traverse.
	 */
	if (NULL != pool->head) {
		/*
		 * lock to prevent corruption of table
		 */
		for (counter = 0, buff = pool->head->prev;
		     counter < pool->size; counter++, buff = buff->prev) {

			result = trav_func(buff, usr_arg);
			if (0 > result) {

				break;
			}
		}
	}

	return result;
} /* sdp_buff_q_trav_tail */

/* ========================================================================= */
/*..sdp_buff_q_trav_head - traverse buffers in pool, from the head */
int sdp_buff_q_trav_head(struct sdpc_buff_q *pool,
			 tSDP_BUFF_TRAV_FUNC trav_func,
			 void *usr_arg)
{
	struct sdpc_buff *buff;
	int result = 0;
	int counter;

	TS_CHECK_NULL(pool, -EINVAL);
	TS_CHECK_NULL(trav_func, -EINVAL);
	/*
	 * check to see if there is anything to traverse.
	 */
	if (NULL != pool->head) {
		/*
		 * lock to prevent corruption of table
		 */
		for (counter = 0, buff = pool->head;
		     counter < pool->size; counter++, buff = buff->next) {

			result = trav_func(buff, usr_arg);
			if (0 > result) {

				break;
			}
		}
	}

	return result;
} /* sdp_buff_q_trav_head */

/* ========================================================================= */
/*..sdp_buff_q_put - Place a buffer into a specific pool */
int sdp_buff_q_put(struct sdpc_buff_q *pool,
		   struct sdpc_buff *buff)
{
	int result;

	result = _sdp_buff_q_put(pool, buff, TRUE);

	return result;
} /* sdp_buff_q_put */

/* ========================================================================= */
/*..sdp_buff_q_put_head - Place a buffer into the head of a specific pool */
int sdp_buff_q_put_head(struct sdpc_buff_q *pool,
			struct sdpc_buff *buff)
{
	int result;

	result = _sdp_buff_q_put(pool, buff, TRUE);

	return result;
} /* sdp_buff_q_put_head */

/* ========================================================================= */
/*..sdp_buff_q_put_tail - Place a buffer into the tail of a specific pool */
int sdp_buff_q_put_tail(struct sdpc_buff_q *pool,
			struct sdpc_buff *buff)
{
	int result;

	result = _sdp_buff_q_put(pool, buff, FALSE);

	return result;
} /* sdp_buff_q_put_tail */

/* ========================================================================= */
/*..sdp_buff_q_clear - clear the buffers out of a specific buffer pool */
int sdp_buff_q_clear(struct sdpc_buff_q *pool)
{
	int result;
	struct sdpc_buff *buff;

	TS_CHECK_NULL(pool, -EINVAL);

	while (NULL != (buff = _sdp_buff_q_get(pool, FALSE, NULL, NULL))) {

		result = sdp_buff_pool_put(buff);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
				 "BUFF: Error <%d> returning buffer to main. <%d>",
				 result, pool->size);
		}
	}

	return 0;
} /* sdp_buff_q_clear */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* internal data buffer pool manager                                     */
/*                                                                       */
/* --------------------------------------------------------------------- */

/* ========================================================================= */
/*.._sdp_buff_pool_seg_release -- release buffers from the segment */
static int _sdp_buff_pool_seg_release(tSDP_MEMORY_SEGMENT mem_seg)
{
	TS_CHECK_NULL(mem_seg, -EINVAL);
	/*
	 * loop through pages.
	 */
	while (0 < mem_seg->head.size) {

		mem_seg->head.size--;
		free_page((unsigned long)mem_seg->list[mem_seg->head.size].
			  head);
	}
	/*
	 * free  descriptor page
	 */
	free_page((unsigned long)mem_seg);

	return 0;
} /* _sdp_buff_pool_seg_release */

/* ========================================================================= */
/*.._sdp_buff_pool_seg_release_all -- release buffers from the segment */
static int _sdp_buff_pool_seg_release_all(tSDP_MAIN_POOL m_pool)
{
	tSDP_MEMORY_SEGMENT mem_seg;
	int result;

	TS_CHECK_NULL(m_pool, -EINVAL);
	/*
	 * loop through pages.
	 */
	while (NULL != m_pool->segs) {

		mem_seg = m_pool->segs;
		m_pool->segs = mem_seg->head.next;

		m_pool->buff_cur -= mem_seg->head.size;

		result = _sdp_buff_pool_seg_release(mem_seg);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	}

	return 0;
} /* _sdp_buff_pool_seg_release_all */

/* ========================================================================= */
/*.._sdp_buff_pool_seg_alloc -- allocate more buffers for the main pool */
static tSDP_MEMORY_SEGMENT _sdp_buff_pool_seg_alloc(void)
{
	tSDP_MEMORY_SEGMENT mem_seg;
        int counter;
	int result;
	/*
	 * get descriptor page
	 */
	mem_seg = (tSDP_MEMORY_SEGMENT) __get_free_page(GFP_ATOMIC);
	if (NULL == mem_seg) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "BUFFER: Failed to allocate descriptor page.");

		goto error;
	}
	/*
	 * loop
	 */
	for (counter = 0, mem_seg->head.size = 0;
	     counter < TS_SDP_BUFF_COUNT; counter++, mem_seg->head.size++) {

        mem_seg->list[counter].head =  (void *) __get_free_page(GFP_ATOMIC);
     		if (NULL == mem_seg->list[counter].head) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
				 "BUFFER: Failed to allocate buffer page. <%d>",
				 counter);

			goto error_free;
		}

		mem_seg->list[counter].end =
		    mem_seg->list[counter].head + PAGE_SIZE;
		mem_seg->list[counter].data = mem_seg->list[counter].head;
		mem_seg->list[counter].tail = mem_seg->list[counter].head;
		mem_seg->list[counter].lkey = 0;
		mem_seg->list[counter].real = 0;
		mem_seg->list[counter].size = 0;
		mem_seg->list[counter].u_id = 0;                                           
		mem_seg->list[counter].pool = NULL;
		mem_seg->list[counter].type = TS_SDP_GENERIC_TYPE_BUFF;
		mem_seg->list[counter].release = ((tSDP_GENERIC_DESTRUCT_FUNC)
						  sdp_buff_pool_put);
	}
	/*
	 * return segment
	 */
	return mem_seg;
error_free:
	result = _sdp_buff_pool_seg_release(mem_seg);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
error:
	return NULL;
} /* _sdp_buff_pool_seg_alloc */

/* ========================================================================= */
/*.._sdp_buff_pool_alloc -- allocate more buffers for the main pool */

static int _sdp_buff_pool_alloc(tSDP_MAIN_POOL m_pool,unsigned long *flags, tUINT32 size)
{
	tSDP_MEMORY_SEGMENT head_seg = NULL;
	tSDP_MEMORY_SEGMENT mem_seg;
	tUINT32 counter = 0;
	tUINT32 total = 0;
	int result;

	TS_CHECK_NULL(m_pool, -EINVAL);
	/*
	 * check pool limits.
	 */
	if (m_pool->buff_max < (m_pool->buff_cur + size)) {

		goto error;
	}
	/*
	 * first allocate the requested number of buffers. Once complete
	 * place them all into the main pool.
	 */

    if (flags!=NULL)  {
        spin_unlock_irqrestore(&m_pool->lock, *flags);   
    }
	while (total < size) {

		mem_seg = _sdp_buff_pool_seg_alloc();
		if (NULL == mem_seg) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
				 "BUFFER: Failed to allocate segment.");

            if (flags!=NULL)  {
                spin_lock_irqsave(&m_pool->lock,*flags);
            }
			goto error;
		}

		mem_seg->head.next = head_seg;
		head_seg = mem_seg;

		total += mem_seg->head.size;
	}
    if (flags!=NULL)  {
        spin_lock_irqsave(&m_pool->lock,*flags);
    }

	/*
	 * insert each segment into the list, and insert each buffer into
	 * the main pool
	 */
	while (NULL != head_seg) {

		mem_seg = head_seg;
		head_seg = mem_seg->head.next;

		mem_seg->head.next = m_pool->segs;
		m_pool->segs = mem_seg;

		for (counter = 0; counter < mem_seg->head.size; counter++) {

			mem_seg->list[counter].u_id = m_pool->buff_cur++;

			result =
			    sdp_buff_q_put(&main_pool->pool,
					     &mem_seg->list[counter]);
			TS_EXPECT(MOD_LNX_SDP, !(0 > result));
		}
	}

	return total;
error:

	while (NULL != head_seg) {

		mem_seg = head_seg;
		head_seg = mem_seg->head.next;

		result = _sdp_buff_pool_seg_release(mem_seg);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	}

	return -ENOMEM;
} /* _sdp_buff_pool_alloc */

/* ========================================================================= */
/*..sdp_buff_pool_init - Initialize the main buffer pool of memory */
int sdp_buff_pool_init(tUINT32 buff_min, tUINT32 buff_max)
{
	int result;

	if (NULL != main_pool) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "BUFFER: Main pool already initialized!");
		return -EEXIST;
	}

	if (!(0 < buff_min) || buff_max < buff_min) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "BUFFER: Pool allocation count error. <%d:%d>",
			 buff_min, buff_max);
		return -ERANGE;
	}
	/*
	 * allocate the main pool structures
	 */
	main_pool = kmalloc(sizeof(tSDP_MAIN_POOL_STRUCT), GFP_KERNEL);
	if (NULL == main_pool) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "BUFFER: Main pool initialization failed.");
		result = -ENOMEM;
		goto done;
	}

	memset(main_pool, 0, sizeof(tSDP_MAIN_POOL_STRUCT));

	main_pool->buff_size = PAGE_SIZE;
	main_pool->buff_min = buff_min;
	main_pool->buff_max = buff_max;

	spin_lock_init(&main_pool->lock);

	result = sdp_buff_q_init(&main_pool->pool, _main_pool_name, 0);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	main_pool->pool_cache = kmem_cache_create("SdpBuffPool",
						  sizeof(struct sdpc_buff_q),
						  0, SLAB_HWCACHE_ALIGN,
						  NULL, NULL);
	if (NULL == main_pool->pool_cache) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "BUFFER: Failed to allocate pool cache.");
		result = -ENOMEM;
		goto error_pool;
	}
	/*
	 * allocate the minimum number of buffers.
	 */
	result = _sdp_buff_pool_alloc(main_pool,NULL, buff_min);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "BUFFER: Error <%d> allocating buffers. <%d>", result,
			 buff_min);
		goto error_alloc;
	}
	/*
	 * done
	 */
	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INIT,
		 "BUFFER: Main pool initialized with min:max <%d:%d> buffers.",
		 buff_min, buff_max);

	return 0;		/* success */
error_alloc:
	kmem_cache_destroy(main_pool->pool_cache);
error_pool:
	kfree(main_pool);
done:
	main_pool = NULL;
	return result;
} /* sdp_buff_pool_init */

/* ========================================================================= */
/*..sdp_buff_pool_destroy - Destroy the main buffer pool and free its memory */
void sdp_buff_pool_destroy(void)
{
	if (NULL == main_pool) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "BUFFER: Main pool dosn't exist.");
		return;
	}
	/*
	 * Free all the memory regions
	 */
	(void)_sdp_buff_pool_seg_release_all(main_pool);
	/*
	 * free pool cache
	 */

	kmem_cache_destroy(main_pool->pool_cache);
   
	/*
	 * free main
	 */
	kfree(main_pool);
	main_pool = NULL;
	/*
	 * done
	 */
	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_CLEANUP,
		 "BUFFER: Main pool destroyed.");

	return;
} /* sdp_buff_pool_destroy */

/* ========================================================================= */
/*..sdp_buff_pool_get - Get a buffer from the main buffer pool */
struct sdpc_buff *sdp_buff_pool_get(void)
{
	struct sdpc_buff *buff;
	int result;
    unsigned long flags;

	TS_CHECK_NULL(main_pool, NULL);
	/*
	 * get buffer
	 */
	spin_lock_irqsave(&main_pool->lock, flags);
    if (NULL == main_pool->pool.head) {

		result =
		    _sdp_buff_pool_alloc(main_pool, &flags,TS_SDP_BUFFER_COUNT_INC);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
				 "BUFFER: Error <%d> allocating buffers.",
				 result);

			spin_unlock_irqrestore(&main_pool->lock, flags);
			return NULL;
		}
	}

	buff = main_pool->pool.head;

	if (buff->next == buff) {

		main_pool->pool.head = NULL;
	}
	else {

		buff->next->prev = buff->prev;
		buff->prev->next = buff->next;

		main_pool->pool.head = buff->next;
	}

	main_pool->pool.size--;

	spin_unlock_irqrestore(&main_pool->lock, flags);

	buff->next = NULL;
	buff->prev = NULL;
	buff->pool = NULL;
	/*
	 * main pool specific reset
	 */
	buff->bsdh_hdr = NULL;
	buff->flags = 0;
	buff->lkey = 0;

	buff->data_size = 0;
	buff->ib_wrid = 0;

	return buff;
} /* sdp_buff_pool_get */

/* ========================================================================= */
/*..sdp_buff_pool_put - Return a buffer to the main buffer pool */
int sdp_buff_pool_put(struct sdpc_buff *buff)
{
	unsigned long flags;

	TS_CHECK_NULL(main_pool, -EINVAL);

	if (NULL == buff || NULL != buff->pool) {

		return -EINVAL;
	}

	if (NULL != buff->next || NULL != buff->prev) {

		return -ETOOMANYREFS;
	}
	/*
	 * reset pointers
	 */
	buff->data = buff->head;
	buff->tail = buff->head;
	buff->pool = &main_pool->pool;

	spin_lock_irqsave(&main_pool->lock, flags);

	if (NULL == main_pool->pool.head) {

		buff->next = buff;
		buff->prev = buff;
		main_pool->pool.head = buff;
	}
	else {

		buff->next = main_pool->pool.head;
		buff->prev = main_pool->pool.head->prev;

		buff->next->prev = buff;
		buff->prev->next = buff;
	}

	main_pool->pool.size++;

	spin_unlock_irqrestore(&main_pool->lock, flags);

	return 0;
} /* sdp_buff_pool_put */

/* ========================================================================= */
/*..sdp_buff_pool_chain_link - create a chain of buffers which can be returned */
int sdp_buff_pool_chain_link(struct sdpc_buff *head, struct sdpc_buff *buff)
{
	TS_CHECK_NULL(main_pool, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	buff->data = buff->head;
	buff->tail = buff->head;
	buff->pool = &main_pool->pool;

	if (NULL == head) {

		buff->next = buff;
		buff->prev = buff;
	}
	else {

		buff->next = head;
		buff->prev = head->prev;

		buff->next->prev = buff;
		buff->prev->next = buff;
	}

	return 0;
} /* sdp_buff_pool_chain_link */

/* ========================================================================= */
/*..sdp_buff_pool_chain_put - Return a buffer to the main buffer pool */
int sdp_buff_pool_chain_put(struct sdpc_buff *buff, u32 count)
{
	unsigned long flags;
	struct sdpc_buff *next;
	struct sdpc_buff *prev;
	/*
	 * return an entire Link of buffers to the queue, this save on
	 * lock contention for the buffer pool, for code paths where
	 * a number of buffers are processed in a loop, before being
	 * returned. (e.g. send completions, recv to userspace.
	 */
	TS_CHECK_NULL(main_pool, -EINVAL);

	if (NULL == buff || !(0 < count)) {

		return -EINVAL;
	}

	spin_lock_irqsave(&main_pool->lock, flags);

	if (NULL == main_pool->pool.head) {

		main_pool->pool.head = buff;
	}
	else {

		prev = buff->prev;
		next = main_pool->pool.head->next;

		buff->prev = main_pool->pool.head;
		main_pool->pool.head->next = buff;

		prev->next = next;
		next->prev = prev;
	}

	main_pool->pool.size += count;

	spin_unlock_irqrestore(&main_pool->lock, flags);

	return 0;
} /* sdp_buff_pool_chain_put */

/* ========================================================================= */
/*..sdp_buff_pool_size - return the number of elements in the main buffer pool */
int sdp_buff_pool_size(void)
{
	unsigned long flags;
	int size;

	if (NULL == main_pool) {

		return -1;
	}

	spin_lock_irqsave(&main_pool->lock, flags);
	size = sdp_buff_q_size(&main_pool->pool);
	spin_unlock_irqrestore(&main_pool->lock, flags);

	return size;
} /* sdp_buff_pool_size */

/* ========================================================================= */
/*..sdp_buff_pool_buff_size - return the size of buffers in the main pool */
int sdp_buff_pool_buff_size(void)
{
	int result;

	if (NULL == main_pool) {

		result = -1;
	}
	else {

		result = main_pool->buff_size;
	}

	return result;
} /* sdp_buff_pool_buff_size */

/* ========================================================================= */
/*..sdp_proc_dump_buff_pool -- write the buffer pool stats to a file (/proc) */
int sdp_proc_dump_buff_pool(char *buffer,
			    int   max_size,
			    off_t start_index,
			    long *end_index)
{
	tSDP_MEMORY_SEGMENT mem_seg;
	int buff_count;
	int offset = 0;
	int counter;
	unsigned long flags;

	TS_CHECK_NULL(buffer, -EINVAL);
	/*
	 * simple table read, without page boundry handling.
	 */
	*end_index = 0;
	/*
	 * lock the table
	 */
	spin_lock_irqsave(&main_pool->lock, flags);

	if (0 == start_index) {

		offset += sprintf((buffer + offset), "Totals:\n");
		offset += sprintf((buffer + offset), "-------\n");

		offset += sprintf((buffer + offset), 
				  "  buffer size:         %8d\n",
				  main_pool->buff_size);
		offset += sprintf((buffer + offset), 
				  "  buffers maximum:     %8d\n",
				  main_pool->buff_max);
		offset += sprintf((buffer + offset), 
				  "  buffers minimum:     %8d\n",
				  main_pool->buff_min);
		offset += sprintf((buffer + offset),
				  "  buffers allocated:   %8d\n",
				  main_pool->buff_cur);
		offset += sprintf((buffer + offset),
				  "  buffers available:   %8d\n",
				  main_pool->pool.size);
		offset += sprintf((buffer + offset),
				  "  buffers outstanding: %8d\n",
				  main_pool->buff_cur - main_pool->pool.size);
		offset += sprintf((buffer + offset), "\nBuffers:\n");
		offset += sprintf((buffer + offset), "--------\n");
		offset += sprintf((buffer + offset),
				  "     id    size pool name\n");
		offset += sprintf((buffer + offset),
				  "  -------- ---- ----------------\n");
	}
	/*
	 * buffers
	 */
	if (!(start_index < main_pool->buff_cur)) {

		goto done;
	}

	for (counter = 0, buff_count = 0, mem_seg = main_pool->segs;
	     NULL != mem_seg && TS_SDP_BUFF_OUT_LEN < (max_size - offset);
	     mem_seg = mem_seg->head.next) {

		for (counter = 0;
		     counter < mem_seg->head.size &&
			     TS_SDP_BUFF_OUT_LEN < (max_size - offset);
		     counter++, buff_count++) {

			if (start_index > buff_count) {

				continue;
			}

			offset += sprintf((buffer + offset),
					  "  %08x %04x %-16s\n",
					  mem_seg->list[counter].u_id,
					  (int)(mem_seg->list[counter].tail -
						mem_seg->list[counter].data),
#ifdef _TS_SDP_DEBUG_POOL_NAME
					  ((NULL != mem_seg->list[counter].
					    pool) ? 
					   mem_seg->list[counter].pool->name :
					   "<none>")
#else
					  "<off>"
#endif
				);
		}
	}

	*end_index = buff_count - start_index;
done:
	spin_unlock_irqrestore(&main_pool->lock, flags);

	return offset;
} /* sdp_proc_dump_buff_pool */
