/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sdp_conn.c 379 2004-07-08 02:24:32Z roland $
*/

#include "sdp_main.h"

static char _send_post_name[] = TS_SDP_CONN_SEND_POST_NAME;
static char _recv_post_name[] = TS_SDP_CONN_RECV_POST_NAME;
static char _recv_pool_name[] = TS_SDP_SOCK_RECV_DATA_NAME;

static struct sdev_root _dev_root_s;
/* --------------------------------------------------------------------- */
/*                                                                       */
/* module specific functions                                             */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_inet_accept_q_put -- put a conn into a listen conn's accept Q. */
tINT32 sdp_inet_accept_q_put(struct sdp_opt *listen_conn, struct sdp_opt *accept_conn)
{
	struct sdp_opt *next_conn;

	TS_CHECK_NULL(listen_conn, -EINVAL);
	TS_CHECK_NULL(accept_conn, -EINVAL);

	if (NULL != listen_conn->parent ||
	    NULL != accept_conn->parent ||
	    NULL == listen_conn->accept_next ||
	    NULL == listen_conn->accept_prev) {

		return -EFAULT;
	}

	next_conn = listen_conn->accept_next;

	accept_conn->accept_next = listen_conn->accept_next;
	listen_conn->accept_next = accept_conn;
	accept_conn->accept_prev = listen_conn;
	next_conn->accept_prev = accept_conn;

	accept_conn->parent = listen_conn;
	listen_conn->backlog_cnt++;
	/*
	 * up ref until we release. One ref for GW and one for INET.
	 */
	TS_SDP_CONN_HOLD(accept_conn);	/* INET reference */

	return 0;
} /* sdp_inet_accept_q_put */

/* ========================================================================= */
/*..sdp_inet_accept_q_get -- get a conn from a listen conn's accept Q. */
struct sdp_opt *sdp_inet_accept_q_get(struct sdp_opt *listen_conn)
{
	struct sdp_opt *prev_conn;
	struct sdp_opt *accept_conn;

	TS_CHECK_NULL(listen_conn, NULL);

	if (NULL != listen_conn->parent ||
	    NULL == listen_conn->accept_next ||
	    NULL == listen_conn->accept_prev ||
	    listen_conn == listen_conn->accept_next ||
	    listen_conn == listen_conn->accept_prev) {

		return NULL;
	}
	/*
	 * Return the next connection in the listening sockets accept
	 * queue. The new connections lock is acquired, the caller must
	 * unlock the connection before it is done with the connection.
	 * Also the process context lock is used, so the function may
	 * not be called from the CQ interrupt.
	 */
	accept_conn = listen_conn->accept_prev;

	TS_SDP_CONN_LOCK(accept_conn);

	prev_conn = accept_conn->accept_prev;

	listen_conn->accept_prev = accept_conn->accept_prev;
	prev_conn->accept_next = listen_conn;

	accept_conn->accept_next = NULL;
	accept_conn->accept_prev = NULL;
	accept_conn->parent = NULL;

	listen_conn->backlog_cnt--;

	return accept_conn;
} /* sdp_inet_accept_q_get */

/* ========================================================================= */
/*..sdp_inet_accept_q_remove -- remove a conn from a conn's accept Q. */
tINT32 sdp_inet_accept_q_remove(struct sdp_opt *accept_conn)
{
	struct sdp_opt *next_conn;
	struct sdp_opt *prev_conn;

	TS_CHECK_NULL(accept_conn, -EINVAL);

	if (NULL == accept_conn->parent) {

		return -EFAULT;
	}
	/*
	 * Removes the connection from the listening sockets accept queue.
	 * The listning connections lock must be acquired to access the
	 * list. The process context lock is used, so the function may
	 * not be called from the CQ interrupt.
	 */
	TS_SDP_CONN_LOCK(accept_conn->parent);

	next_conn = accept_conn->accept_next;
	prev_conn = accept_conn->accept_prev;

	prev_conn->accept_next = accept_conn->accept_next;
	next_conn->accept_prev = accept_conn->accept_prev;

	accept_conn->parent->backlog_cnt--;

	TS_SDP_CONN_UNLOCK(accept_conn->parent);

	accept_conn->accept_next = NULL;
	accept_conn->accept_prev = NULL;
	accept_conn->parent = NULL;

	return 0;
} /* sdp_inet_accept_q_remove */

/* ========================================================================= */
/*..sdp_inet_listen_start -- start listening for new connections on a socket */
tINT32 sdp_inet_listen_start(struct sdp_opt *conn)
{
	unsigned long flags;

	TS_CHECK_NULL(conn, -EINVAL);

	if (TS_SDP_SOCK_ST_CLOSED != conn->istate) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: Incorrect socket state <%04x:%04x> to start listening.",
			 conn->istate, conn->state);
		return -EBADFD;
	}

	conn->istate = TS_SDP_SOCK_ST_LISTEN;
	conn->accept_next = conn;
	conn->accept_prev = conn;
	/*
	 * table lock
	 */
	spin_lock_irqsave(&_dev_root_s.listen_lock, flags);
	/*
	 * insert into listening list.
	 */
	conn->lstn_next = _dev_root_s.listen_list;
	_dev_root_s.listen_list = conn;
	conn->lstn_p_next = &_dev_root_s.listen_list;

	if (NULL != conn->lstn_next) {

		conn->lstn_next->lstn_p_next = &conn->lstn_next;
	}

	spin_unlock_irqrestore(&_dev_root_s.listen_lock, flags);
	return 0;
} /* sdp_inet_listen_start */

/* ========================================================================= */
/*..sdp_inet_listen_stop -- stop listening for new connections on a socket */
tINT32 sdp_inet_listen_stop(struct sdp_opt *listen_conn)
{
	struct sdp_opt *accept_conn;
	tINT32 result;
	unsigned long flags;

	TS_CHECK_NULL(listen_conn, -EINVAL);

	if (TS_SDP_SOCK_ST_LISTEN != listen_conn->istate) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: Incorrect socket state <%04x:%04x> to stop listening.",
			 listen_conn->istate, listen_conn->state);
		return -EBADFD;
	}

	listen_conn->istate = TS_SDP_SOCK_ST_CLOSED;
	/*
	 * table lock
	 */
	spin_lock_irqsave(&_dev_root_s.listen_lock, flags);
	/*
	 * remove from listening list.
	 */
	if (NULL != listen_conn->lstn_next) {
		listen_conn->lstn_next->lstn_p_next = listen_conn->lstn_p_next;
	}

	*(listen_conn->lstn_p_next) = listen_conn->lstn_next;

	listen_conn->lstn_p_next = NULL;
	listen_conn->lstn_next = NULL;

	spin_unlock_irqrestore(&_dev_root_s.listen_lock, flags);
	/*
	 * reject and delete all pending connections
	 */
	while (NULL != (accept_conn = sdp_inet_accept_q_get(listen_conn))) {
		/*
		 * The connection is going to be dropped now, mark the
		 * state as such in case of conntension for this conn.
		 * Remember to unlock since the Get function will acquire
		 * the lock.
		 */
		accept_conn->istate = TS_SDP_SOCK_ST_CLOSED;

		result = sdp_wall_send_abort(accept_conn);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "STRM: Error <%d> rejecting connection <%08x:%04x><%08x:%04x>",
				 result, accept_conn->src_addr,
				 accept_conn->src_port, accept_conn->dst_addr,
				 accept_conn->dst_port);
		}

		TS_SDP_CONN_UNLOCK(accept_conn);	/* AcceptQueueGet */
		TS_SDP_CONN_PUT(accept_conn);	/* INET reference (AcceptQueuePut). */
	}

	listen_conn->accept_next = NULL;
	listen_conn->accept_prev = NULL;

	return 0;
} /* sdp_inet_listen_stop */

/* ========================================================================= */
/*..sdp_inet_listen_lookup -- lookup a connection in the listen list */
struct sdp_opt *sdp_inet_listen_lookup(tUINT32 addr, tUINT16 port)
{
	struct sdp_opt *conn;
	unsigned long flags;
	/*
	 * table lock
	 */
	spin_lock_irqsave(&_dev_root_s.listen_lock, flags);
	/*
	 * first find a listening connection
	 */
	for (conn = _dev_root_s.listen_list; NULL != conn;
	     conn = conn->lstn_next) {

		if (port == conn->src_port &&
		    (INADDR_ANY == conn->src_addr || addr == conn->src_addr)) {

			TS_SDP_CONN_HOLD(conn);
			break;
		}
	}

	spin_unlock_irqrestore(&_dev_root_s.listen_lock, flags);
	return conn;
} /* sdp_inet_listen_lookup */

/* ========================================================================= */
/*..sdp_inet_port_get -- bind a socket to a port. */
tINT32 sdp_inet_port_get(struct sdp_opt *conn, tUINT16 port)
{
	struct sock *sk;
	struct sock *srch;
	struct sdp_opt *look;
	tINT32 counter;
	tINT32 low_port;
	tINT32 top_port;
	tINT32 port_ok;
	tINT32 result;
	static tINT32 rover = -1;
	unsigned long flags;

	TS_CHECK_NULL(conn, -EINVAL);

	sk = conn->sk;
	/*
	 * lock table
	 */
	spin_lock_irqsave(&_dev_root_s.bind_lock, flags);
	/*
	 * simple linked list of sockets ordered on local port number.
	 */
	if (0 < port) {

		for (look = _dev_root_s.bind_list, port_ok = 1;
		     NULL != look; look = look->bind_next) {

			srch = look->sk;
			/*
			 * 1) same port
			 * 2) linux force reuse is off.
			 * 3) same bound interface, or neither has a bound interface
			 */
			if (look->src_port == port &&
			    !(1 < TS_SDP_OS_SK_REUSE(sk)) &&
			    !(1 < TS_SDP_OS_SK_REUSE(srch)) &&
			    TS_SDP_OS_SK_BOUND_IF(sk) ==
			    TS_SDP_OS_SK_BOUND_IF(srch)) {
				/*
				 * 3) either socket has reuse turned off
				 * 4) socket already listening on this port
				 */
				if (0 == TS_SDP_OS_SK_REUSE(sk) ||
				    0 == TS_SDP_OS_SK_REUSE(srch) ||
				    TS_SDP_SOCK_ST_LISTEN == look->istate) {
					/*
					 * 5) neither socket is using a specific address
					 * 6) both sockets are trying for the same interface.
					 */
					if (INADDR_ANY == conn->src_addr ||
					    INADDR_ANY == look->src_addr ||
					    conn->src_addr == look->src_addr) {

						TS_TRACE(MOD_LNX_SDP,
							 T_VERY_VERBOSE,
							 TRACE_FLOW_INOUT,
							 "SOCK: port rejected. <%04x><%d:%d><%d:%d><%04x><%u:%u>",
							 port,
							 TS_SDP_OS_SK_BOUND_IF
							 (sk),
							 TS_SDP_OS_SK_BOUND_IF
							 (srch),
							 TS_SDP_OS_SK_REUSE(sk),
							 TS_SDP_OS_SK_REUSE
							 (srch), look->state,
							 conn->src_addr,
							 look->src_addr);
						port_ok = 0;
						break;
					}
				}
			}
		}

		if (0 == port_ok) {

			result = -EADDRINUSE;
			goto done;
		}
	}
	else {

		low_port = TS_SDP_PORT_RANGE_LOW;
		top_port = TS_SDP_PORT_RANGE_HIGH;
		rover = (0 > rover) ? low_port : rover;

		for (counter = (top_port - low_port) + 1; counter > 0;
		     counter--) {

			rover++;
			if (rover < low_port || rover > top_port) {

				rover = low_port;
			}

			for (look = _dev_root_s.bind_list;
			     NULL != look && look->src_port != port;
			     look = look->bind_next) {
				/*
				 * pass
				 */
			}

			if (NULL == look) {

				port = rover;
				break;
			}
		}

		if (0 == port) {

			result = -EADDRINUSE;
			goto done;
		}
	}

	conn->src_port = port;
	/*
	 * insert into listening list.
	 */
	conn->bind_next = _dev_root_s.bind_list;
	_dev_root_s.bind_list = conn;
	conn->bind_p_next = &_dev_root_s.bind_list;

	if (NULL != conn->bind_next) {

		conn->bind_next->bind_p_next = &conn->bind_next;
	}

	result = 0;
done:
	spin_unlock_irqrestore(&_dev_root_s.bind_lock, flags);
	return result;
} /* sdp_inet_port_get */

/* ========================================================================= */
/*..sdp_inet_port_put -- unbind a socket from a port. */
tINT32 sdp_inet_port_put(struct sdp_opt *conn)
{
	unsigned long flags;

	TS_CHECK_NULL(conn, -EINVAL);

	if (NULL == conn->bind_p_next) {

		return -EADDRNOTAVAIL;
	}
	/*
	 * lock table
	 */
	spin_lock_irqsave(&_dev_root_s.bind_lock, flags);
	/*
	 * remove from bind list.
	 */
	if (NULL != conn->bind_next) {
		conn->bind_next->bind_p_next = conn->bind_p_next;
	}

	*(conn->bind_p_next) = conn->bind_next;

	conn->bind_p_next = NULL;
	conn->bind_next = NULL;
	conn->src_port = 0;

	spin_unlock_irqrestore(&_dev_root_s.bind_lock, flags);
	return 0;
} /* sdp_inet_port_put */

/* ========================================================================= */
/*..sdp_inet_port_inherit -- inherit a port from another socket (accept) */
tINT32 sdp_inet_port_inherit(struct sdp_opt *parent, struct sdp_opt *child)
{
	tINT32 result;
	unsigned long flags;

	TS_CHECK_NULL(child, -EINVAL);
	TS_CHECK_NULL(parent, -EINVAL);
	TS_CHECK_NULL(parent->bind_p_next, -EINVAL);
	/*
	 * lock table
	 */
	spin_lock_irqsave(&_dev_root_s.bind_lock, flags);

	if (NULL != child->bind_p_next || child->src_port != parent->src_port) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: Parents <%d> child already bound. <%d>",
			 parent->src_port, child->src_port);
		result = -EADDRNOTAVAIL;
		goto done;
	}
	/*
	 * insert into listening list.
	 */
	child->bind_next = parent->bind_next;
	parent->bind_next = child;
	child->bind_p_next = &parent->bind_next;

	if (NULL != child->bind_next) {

		child->bind_next->bind_p_next = &child->bind_next;
	}

	result = 0;
done:
	spin_unlock_irqrestore(&_dev_root_s.bind_lock, flags);
	return result;
} /* sdp_inet_port_inherit */
     

/* ========================================================================= */
/*..sdp_conn_table_insert -- insert a connection into the connection table */


tINT32 sdp_conn_table_insert(struct sdp_opt *conn)
{
	tINT32 counter;
	tINT32 result = -ENOMEM;
	unsigned long flags;
	TS_CHECK_NULL(conn, -EINVAL);

	if (TS_SDP_DEV_SK_INVALID != conn->hashent) {

		return -ERANGE;
	}
	/*
	 * lock table
	 */
	spin_lock_irqsave(&_dev_root_s.sock_lock, flags);
	/*
	 * find an empty slot.
	 */
	for (counter = 0;
	     counter < _dev_root_s.sk_size; counter++, _dev_root_s.sk_rover++) {

		if (!(_dev_root_s.sk_rover < _dev_root_s.sk_size)) {

			_dev_root_s.sk_rover = 0;
		}

		if (NULL == _dev_root_s.sk_array[_dev_root_s.sk_rover]) {

			_dev_root_s.sk_array[_dev_root_s.sk_rover] = conn;
			_dev_root_s.sk_entry++;
			conn->hashent = _dev_root_s.sk_rover;

			result = 0;
			break;
		}
	}

#if 0				/* set for reproducibility from run-run. */
	_dev_root_s.sk_rover = 0;
#endif
	/*
	 * unlock table
	 */
	spin_unlock_irqrestore(&_dev_root_s.sock_lock, flags);
	return result;
} /* sdp_conn_table_insert */

/* ========================================================================= */
/*..sdp_conn_table_remove -- remove a connection from the connection table */
static tINT32 _sdp_conn_table_remove(struct sdp_opt *conn)
{
	tINT32 result = 0;
    
	TS_CHECK_NULL(conn, -EINVAL);
	/*
	 * validate entry
	 */
	if (TS_SDP_DEV_SK_INVALID == conn->hashent) {

		goto done;
	}

	if (0 > conn->hashent || conn != _dev_root_s.sk_array[conn->hashent]) {

		result = -ERANGE;
		goto done;
	}
	/*
	 * drop entry
	 */
	_dev_root_s.sk_array[conn->hashent] = NULL;
	_dev_root_s.sk_entry--;
	conn->hashent = TS_SDP_DEV_SK_INVALID;

done:
	return result;
} /* sdp_conn_table_remove */

/* ========================================================================= */
/*..sdp_conn_table_lookup -- look up connection in the connection table */
struct sdp_opt *sdp_conn_table_lookup(tINT32 entry)
{
	struct sdp_opt *conn;
	unsigned long flags;
	/*
	 * lock table
	 */
	spin_lock_irqsave(&_dev_root_s.sock_lock, flags);
#if 0
	/*
	 * validate range
	 */
	if (0 > entry || !(_dev_root_s.sk_size > entry)) {

		conn = NULL;
		goto done;
	}
#endif

	conn = _dev_root_s.sk_array[entry];
	if (NULL == conn) {

		goto done;
	}

	TS_SDP_CONN_HOLD(conn);
done:
	spin_unlock_irqrestore(&_dev_root_s.sock_lock, flags);
	return conn;
} /* sdp_conn_table_lookup */


/* --------------------------------------------------------------------- */
/*                                                                       */
/* connection allocation/deallocation                                    */
/*                                                                      */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_conn_destroy -- final destructor for connection. */

static tINT32 _sdp_conn_destruct(struct sdp_opt *conn)
{
	tINT32 result = 0;
	tINT32 dump = 0;

	if (NULL == conn) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: sk destruct, no connection!");
		result = -EINVAL;
		goto done;
	}
   
	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "SOCK: <%d> destruct cb <%08x:%04x> <%08x:%04x> <%04x> <%u:%u>",
		 conn->hashent, conn->src_addr, conn->src_port, conn->dst_addr,
		 conn->dst_port, conn->state, conn->src_serv, conn->snk_serv);

	/*
	 * If the socket is bound, return the port
	 */
	(void)sdp_inet_port_put(conn);
	/*
	 * in case CM/IB are still tracking this connection.
	 */
	(void)ib_cm_callback_modify(conn->comm_id, NULL, NULL);


	result = __sdp_conn_stat_dump(conn);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	/*
	 * really there shouldn't be anything in these tables, but it's
	 * really bad if we leave a dangling reference here.
	 */
	(void)sdp_iocb_q_cancel_all(conn, -ECANCELED);
	(void)sdp_iocb_q_clear(&conn->r_pend);
   	(void)sdp_iocb_q_clear(&conn->r_snk);
	(void)sdp_iocb_q_clear(&conn->w_src);
    (void)sdp_iocb_q_clear(&conn->r_comp);
    
    (void)sdp_desc_q_clear(&conn->w_snk);

    (void)sdp_desc_q_clear(&conn->r_src);

	/*
	 * clear the buffer pools
	 */
	(void)sdp_buff_q_clear(&conn->recv_pool);
	(void)sdp_buff_q_clear(&conn->send_post);
	(void)sdp_buff_q_clear(&conn->recv_post);
	/*
	 * clear advertisment tables
	 */
	(void)sdp_advt_q_clear(&conn->src_pend);
	(void)sdp_advt_q_clear(&conn->src_actv);
	(void)sdp_advt_q_clear(&conn->snk_pend);
	/*
	 * generic table clear
	 */
	(void)sdp_desc_q_clear(&conn->send_ctrl);
	(void)sdp_desc_q_clear(&conn->send_queue);
	/*
	 * If the QP owner is not the CM, then destroy.
	 */
	if (conn->qp) {

		result = ib_qp_destroy(conn->qp);
		if (0 > result && -EINVAL != result) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
				 "CONN: Error <%d> detroying QP", result);
			dump++;
		}
	}
	/*
	 * destroy CQs
	 */
	if (conn->recv_cq) {

		result = ib_cq_destroy(conn->recv_cq);
		if (0 > result && -EINVAL != result) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
				 "CONN: Error <%d> detroying recv CQ", result);
			dump++;
		}
	}

	if (conn->send_cq) {

		result = ib_cq_destroy(conn->send_cq);
		if (0 > result && -EINVAL != result) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
				 "CONN: Error <%d> detroying send CQ", result);
			dump++;
		}
	}
	/*
	 * check consistancy
	 */
	if (0 > atomic_read(&conn->refcnt)) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
			 "SOCK: destruct <%08x:%04x> <%08x:%04x> low ref <%04x>",
			 conn->src_addr, conn->src_port, conn->dst_addr,
			 conn->dst_port, atomic_read(&conn->refcnt));
	}
	/*
	 * free the OS socket structure
	 */
	if (NULL == conn->sk) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: sk destruct, no connection! continuing.");
	}
	else {

		sk_free(conn->sk);
		conn->sk = NULL;
	}
	kmem_cache_free(_dev_root_s.conn_cache, conn);

#ifndef TS_KERNEL_2_6

TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,"count dec");
 MOD_DEC_USE_COUNT;
#endif

	result = 0;
done:

	if (0 != dump) {

		result = __sdp_conn_state_dump(conn);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	}

	return result;
} /* _sdp_conn_destruct */

int sdp_conn_put(struct sdp_opt *conn)
{
	unsigned long flags;
    int result = 0;
	/*
	 * lock table to prevent race against sdp_conn_table_lookup.
	 */
	spin_lock_irqsave(&_dev_root_s.sock_lock, flags);

	if (likely(!atomic_dec_and_test(&conn->refcnt))) {
		spin_unlock_irqrestore(&_dev_root_s.sock_lock, flags);
		return 1;
	}

	result = _sdp_conn_table_remove(conn);
	spin_unlock_irqrestore(&_dev_root_s.sock_lock, flags);

	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "SOCK: Error <%d> removing connection <%d> from table. <%d:%d>",
			 result, conn->hashent, _dev_root_s.sk_entry,
			 _dev_root_s.sk_size);
		return 0;
	}
	(void)_sdp_conn_destruct(conn);
	return 0;
}

/* ========================================================================= */
/*..sdp_conn_internal_lock -- lock the connection (use only from macro) */
void sdp_conn_internal_lock(struct sdp_opt *conn
#if defined(TS_KERNEL_2_6)
/* Work around the temporary workaround for THCA locking bugs :) */
			       , unsigned long *flags
#endif
    )
{
	DECLARE_WAITQUEUE(wait, current);
#if defined(TS_KERNEL_2_6)
	unsigned long f = *flags;
#endif

	add_wait_queue_exclusive(&(conn->lock.waitq), &wait);
	for (;;) {
        set_current_state(TASK_UNINTERRUPTIBLE);
	//	current->state = TASK_UNINTERRUPTIBLE;
#if !defined(TS_KERNEL_2_6)
		spin_unlock_bh(&(conn->lock.slock));
#else
		spin_unlock_irqrestore(&(conn->lock.slock), f);
#endif
		schedule();
#if !defined(TS_KERNEL_2_6)
		spin_lock_bh(&(conn->lock.slock));
#else
		spin_lock_irqsave(&(conn->lock.slock), f);
		*flags = f;
#endif
		if (0 == conn->lock.users) {

			break;
		}
	}
    set_current_state(TASK_RUNNING);
	//current->state = TASK_RUNNING;
	remove_wait_queue(&(conn->lock.waitq), &wait);

	return;
} /* sdp_conn_internal_lock */



/* ========================================================================= */
/*..sdp_conn_internal_relock -- test the connection (use only from macro) */
void sdp_conn_internal_relock(struct sdp_opt *conn)
{
	struct ib_cq_entry entry;
	tINT32 result_r;
	tINT32 result_s;
	tINT32 result;
	tINT32 rearm = 1;
    int i=0;

	while ( i++ < 5000 ) {

		result_r = ib_cq_poll(conn->recv_cq, &entry);
		if (0 == result_r) {
			/* got completion */
			result = sdp_cq_event_locked(&entry, conn);
			if (0 > result) {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_WARN,
					 "EVENT: Error <%d> from locked event handler.",
					 result);
			}

			rearm = 1;
		}

		result_s = ib_cq_poll(conn->send_cq, &entry);
		if (0 == result_s) {
			/* got completion */
			result = sdp_cq_event_locked(&entry, conn);
			if (0 > result) {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_WARN,
					 "EVENT: Error <%d> from locked event handler.",
					 result);
			}

			rearm = 1;
		}

		if (-EAGAIN == result_r && -EAGAIN == result_s) {

			if (0 < rearm) {

				result =
				    ib_cq_request_notification(conn->recv_cq, 0);
				if (0 > result) {

					TS_TRACE(MOD_LNX_SDP, T_TERSE,
						 TRACE_FLOW_WARN,
						 "EVENT: error <%d> rearming CQ. <%d:%d>",
						 result, conn->recv_cq,
						 conn->hashent);
				}

				result =
				    ib_cq_request_notification(conn->send_cq, 0);
				if (0 > result) {

					TS_TRACE(MOD_LNX_SDP, T_TERSE,
						 TRACE_FLOW_WARN,
						 "EVENT: error <%d> rearming CQ. <%d:%d>",
						 result, conn->send_cq,
						 conn->hashent);
				}

				rearm = 0;
			}
			else {
				/*
				 * exit CQ handler routine.
				 */
				break;
			}
		}

		/* need to break in case of error in one of the CQs */
		if ( (result_s != 0) && (result_s != -EAGAIN) ) {
			printk("error in send CQ (%d), aborting...\n", result_s);
			break;
		}


		if ( (result_r != 0) && (result_r != -EAGAIN) ) {
			printk("error in receive CQ (%d), aborting...\n", result_r);
			break;
		}

	}

    if ( i >= 5000 ) {
        printk("aborted: more then 5000 retries\n");
    }

	conn->flags &= ~TS_SDP_CONN_F_MASK_EVENT;

	return;
} /* sdp_conn_internal_relock */
        /* ========================================================================= */
/*..sdp_conn_cq_drain -- drain one of the the connection's CQs */
tINT32 sdp_conn_cq_drain(struct ib_cq *cq, struct sdp_opt *conn)
{
	struct ib_cq_entry entry;
	tINT32 result;
	tINT32 rearm = 1;
	tINT32 calls = 0;
	/*
	 * the function should only be called under the connection locks
	 * spinlock to ensure the call is serialized to avoid races.
	 */
	for (;;) {
		/*
		 * poll for a completion
		 */
		result = ib_cq_poll(cq, &entry);
		if (0 == result) {
			/*
			 * dispatch completion, and mark that the CQ needs 
			 * to be armed.
			 */
			result = sdp_cq_event_locked(&entry, conn);
			if (0 > result) {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_WARN,
					 "EVENT: Error <%d> from locked event handler.",
					 result);
			}

			rearm = 1;
			calls++;
		}
		else {

			if (-EAGAIN == result) {

				if (0 < rearm) {

					result =
					    ib_cq_request_notification(cq, 0);
					if (0 > result) {

						TS_TRACE(MOD_LNX_SDP, T_TERSE,
							 TRACE_FLOW_WARN,
							 "EVENT: error <%d> rearming CQ. <%d:%d>",
							 result, cq,
							 conn->hashent);
					}

					rearm = 0;
				}
				else {
					/*
					 * exit CQ handler routine.
					 */
					break;
				}
			}
			else {
				/*
				 * unexpected CQ error
				 */
				TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
					 "EVENT: CQ <%d:%d> poll error. <%d>",
					 cq, conn->hashent, result);
			}
		}
	}

	return calls;
} /* sdp_conn_cq_drain */

/* ========================================================================= */
/*..sdp_conn_internal_unlock -- lock the connection (use only from macro) */
void sdp_conn_internal_unlock(struct sdp_opt *conn)
{
	tINT32 calls = 0;
	/*
	 * poll CQs for events.
	 */
	if (NULL != conn) {

		if (0 < (TS_SDP_CONN_F_RECV_CQ_PEND & conn->flags)) {

			calls += sdp_conn_cq_drain(conn->recv_cq, conn);
		}

		if (0 < (TS_SDP_CONN_F_SEND_CQ_PEND & conn->flags)) {

			calls += sdp_conn_cq_drain(conn->send_cq, conn);
		}

		conn->flags &= ~TS_SDP_CONN_F_MASK_EVENT;
	}

	return;
} /* sdp_conn_internal_unlock */

/* ========================================================================= */
/*.._sdp_conn_lock_init -- initialize connection lock */
static tINT32 _sdp_conn_lock_init(struct sdp_opt *conn)
{
	TS_CHECK_NULL(conn, -EINVAL);

	spin_lock_init(&(conn->lock.slock));
	conn->lock.users = 0;
	init_waitqueue_head(&(conn->lock.waitq));

	return 0;
} /* _sdp_conn_lock_init */

/* ========================================================================= */
/*..sdp_conn_alloc_ib -- allocate IB structures for a new connection. */
int sdp_conn_alloc_ib(struct sdp_opt *conn,
		      struct ib_device *device,
		      tTS_IB_PORT hw_port)
{
	struct ib_qp_create_param *qp_param;
	struct ib_cq_callback     *callback;
	struct ib_qp_attribute    *qp_attr;
	struct sdev_hca_port  *port = NULL;
	struct sdev_hca       *hca = NULL;
	int                    result;

	TS_CHECK_NULL(conn, -EINVAL);
	/*
	 * look up correct HCA and port
	 */
	for (hca = _dev_root_s.hca_list; NULL != hca; hca = hca->next) {

		if (device == hca->ca) {

			for (port = hca->port_list; NULL != port;
			     port = port->next) {

				if (hw_port == port->index) {

					break;
				}
			}

			break;
		}
	}

	if (NULL == hca || NULL == port) {

		return -ERANGE;
	}
	/*
	 * allocate creation parameters
	 */
	qp_attr = kmalloc(sizeof(struct ib_qp_attribute),  GFP_KERNEL);
	if (NULL == qp_attr) {
		
		result = -ENOMEM;
		goto error_attr;
	} /* if */
  
	qp_param = kmalloc(sizeof(struct ib_qp_create_param),  GFP_KERNEL);
	if (NULL == qp_param) {
		
		result = -ENOMEM;
		goto error_param;
	} /* if */
	
	callback = kmalloc(sizeof(struct ib_cq_callback),  GFP_KERNEL);
	if (NULL == callback) {
		
		result = -ENOMEM;
		goto error_cb;
	} /* if */
	
	memset(qp_attr, 0, sizeof(struct ib_qp_attribute));
	memset(qp_param, 0, sizeof(struct ib_qp_create_param));
	memset(callback, 0, sizeof(struct ib_cq_callback));
	/*
	 * set port specific connection parameters.
	 */
	conn->ca = hca->ca;
	conn->pd = hca->pd;
	conn->hw_port = port->index;
	conn->l_key = hca->l_key;
	conn->fmr_pool = hca->fmr_pool;

	memcpy(conn->s_gid, port->gid, sizeof(tTS_IB_GID));
	/*
	 * allocate IB CQ's and QP
	 */
	if (!conn->send_cq) {

		callback->context        = IB_CQ_CALLBACK_INTERRUPT;
		callback->policy         = IB_CQ_CONSUMER_REARM;
		callback->function.event = sdp_cq_event_handler;
		callback->arg = (void *)(unsigned long)conn->hashent;

		result = ib_cq_create(conn->ca,
				      &conn->send_cq_size,
				      callback, 
				      NULL, 
				      &conn->send_cq);
		if (0 != result) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
				 "INIT: Error <%d> creating send CQ size <%d>",
				 result, conn->send_cq_size);
			goto error_scq;
		}

		result = ib_cq_request_notification(conn->send_cq, 0);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
				 "EVENT: error <%d> arming send CQ. <%d:%d>",
				 result, conn->send_cq, conn->hashent);
			goto error_rcq;
		}
	}

	if (!conn->recv_cq) {

		callback->context        = IB_CQ_CALLBACK_INTERRUPT;
		callback->policy         = IB_CQ_CONSUMER_REARM;
		callback->function.event = sdp_cq_event_handler;
		callback->arg = (void *)(unsigned long)conn->hashent;

		result = ib_cq_create(conn->ca,
				      &conn->recv_cq_size,
				      callback, 
				      NULL, 
				      &conn->recv_cq);
		if (0 != result) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
				 "INIT: Error <%d> creating recv CQ size <%d>",
				 result, conn->recv_cq_size);
			goto error_rcq;
		}

		result = ib_cq_request_notification(conn->recv_cq, 0);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
				 "EVENT: error <%d> arming recv CQ. <%d:%d>",
				 result, conn->recv_cq, conn->hashent);
			goto error_qp;
		}
	}

	if (!conn->qp) {

		qp_param->limit.max_outstanding_send_request = 
			conn->send_cq_size;
		qp_param->limit.max_outstanding_receive_request = 
			conn->recv_cq_size;
		qp_param->limit.max_send_gather_element =
		    TS_SDP_QP_LIMIT_SG_SEND;
		qp_param->limit.max_receive_scatter_element =
		    TS_SDP_QP_LIMIT_SG_RECV;

		qp_param->pd             = conn->pd;
		qp_param->send_queue     = conn->send_cq;
		qp_param->receive_queue  = conn->recv_cq;
		qp_param->send_policy    = IB_WQ_SIGNAL_SELECTABLE;
		qp_param->receive_policy = IB_WQ_SIGNAL_ALL;
		qp_param->transport      = IB_TRANSPORT_RC;

		result = ib_qp_create(qp_param, &conn->qp, &conn->s_qpn);
		if (0 != result) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
				 "INIT: Error <%d> creating queue pair.",
				 result);
			goto error_qp;
		}
		/*
		 * modify QP to INIT
		 */
		memset(qp_attr, 0, sizeof(struct ib_qp_attribute));

		qp_attr->valid_fields |= IB_QP_ATTRIBUTE_STATE;
		qp_attr->state = IB_QP_STATE_INIT;

		qp_attr->valid_fields |= IB_QP_ATTRIBUTE_RDMA_ATOMIC_ENABLE;
		qp_attr->enable_rdma_read = 1;
		qp_attr->enable_rdma_write = 1;

		qp_attr->valid_fields |= IB_QP_ATTRIBUTE_PORT;
		qp_attr->port = conn->hw_port;

		qp_attr->valid_fields |= IB_QP_ATTRIBUTE_PKEY_INDEX;
		qp_attr->pkey_index = 0;

		result = ib_qp_modify(conn->qp, qp_attr);

		if (0 != result) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
				 "INIT: Error <%d> modifying queue pair.",
				 result);
			goto error_mod;
		}
	}

	result = 0;
	goto done;

error_mod:
	(void)ib_qp_destroy(conn->qp);
error_qp:
	(void)ib_cq_destroy(conn->recv_cq);
error_rcq:
	(void)ib_cq_destroy(conn->send_cq);
error_scq:
	conn->send_cq = NULL;
	conn->recv_cq = NULL;
	conn->qp      = NULL;                     
done:
	kfree(callback);
error_cb:
	kfree(qp_param);
error_param:
	kfree(qp_attr);
error_attr:
	return result;
} /* sdp_conn_alloc_ib  */

/* ========================================================================= */
/*..sdp_conn_alloc -- allocate a new socket, and init. */
struct sdp_opt *sdp_conn_alloc(tINT32 priority, tTS_IB_CM_COMM_ID comm_id)
{
	struct sdp_opt *conn;
	struct sock *sk;
	tINT32 result;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	sk = sk_alloc(_dev_root_s.proto, priority, 1, _dev_root_s.sock_cache);
#else
	sk = sk_alloc(_dev_root_s.proto, priority, 1);
#endif
	if (NULL == sk) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "CREATE: failed to create socket for protocol. <%d:%d>",
			 _dev_root_s.proto, priority);
		return NULL;
	}
	/*
	 * initialize base socket
	 */
	sock_init_data(NULL, sk);	/* refcnt set to 1 */
	/*
	 * other non-zero sock initialization.
	 */
	TS_SDP_OS_SK_FAMILY(sk) = _dev_root_s.proto;
	TS_SDP_OS_SK_PROTOCOL(sk) = IPPROTO_TCP;
	/*
	 * some callbacks. higher levels of linux (e.g. setsockopt) call
	 * these functions so they should be shim's to our functions...
	 */
	TS_SDP_OS_SK_DESTRUCT(sk) = NULL;
	TS_SDP_OS_SK_WRITE_SPACE(sk) = sdp_inet_wake_send;
	TS_SDP_OS_SK_STATE_CHANGE(sk) = sdp_inet_wake_generic;
	TS_SDP_OS_SK_DATA_READY(sk) = sdp_inet_wake_recv;
	TS_SDP_OS_SK_ERROR_REPORT(sk) = sdp_inet_wake_error;
	/*
	 * Allocate must be called from process context, since QP
	 * create/modifies must be in that context.
	 */
	conn = kmem_cache_alloc(_dev_root_s.conn_cache, priority);
	if (NULL == conn) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "CREATE: Error allocating connection. <%d>",
			 sizeof(struct sdp_opt));
		result = -ENOMEM;
		goto error;
	}

	memset(conn, 0, sizeof(struct sdp_opt));
	/*
	 * The STRM interface specific data is map/cast over the TCP specific
	 * area of the sock.
	 */
	TS_SDP_SET_CONN(sk, conn);
	TS_SDP_CONN_ST_INIT(conn);

	conn->oob_offset  = -1;
	conn->rcv_urg_cnt = 0;

	conn->nodelay     = 0;
   // conn->src_zthresh = SDP_ZCOPY_THRSH_SRC_DEFAULT;
//	conn->snk_zthresh = SDP_ZCOPY_THRSH_SNK_DEFAULT;

   
	conn->accept_next = NULL;
	conn->accept_prev = NULL;
	conn->parent      = NULL;

	conn->pid       = 0;
	conn->sk        = sk;
	conn->hashent   = TS_SDP_DEV_SK_INVALID;
	conn->istate    = TS_SDP_SOCK_ST_CLOSED;
	conn->flags     = 0;
	conn->shutdown  = TS_SDP_SHUTDOWN_NONE;
	conn->recv_mode = TS_SDP_MODE_COMB;
	conn->send_mode = TS_SDP_MODE_COMB;

	TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_CLOSED);

	conn->comm_id   = comm_id;
	conn->send_seq  = 0;
	conn->recv_seq  = 0;
	conn->advt_seq  = 0;

	conn->nond_recv = 0;
	conn->nond_send = 0;

	conn->recv_max  = _dev_root_s.recv_buff_max;
	conn->send_max  = _dev_root_s.send_buff_max;
	conn->rwin_max  = TS_SDP_DEV_RECV_WIN_MAX;
	conn->s_wq_size = 0;

	conn->send_buf  = 0;
	conn->send_qud  = 0;
	conn->send_pipe = 0;

	conn->recv_size = sdp_buff_pool_buff_size();
	conn->send_size = 0;

	conn->src_serv  = 0;
	conn->snk_serv  = 0;
	conn->s_cur_adv = 1;
	conn->s_par_adv = 0;

	conn->src_recv  = 0;
	conn->snk_recv  = 0;
	conn->src_sent  = 0;
	conn->snk_sent  = 0;

	conn->send_bytes   = 0;
	conn->recv_bytes   = 0;
	conn->write_bytes  = 0;
	conn->read_bytes   = 0;
	conn->write_queued = 0;
	conn->read_queued  = 0;

	conn->send_usig = 0;
	conn->send_cons = 0;
	conn->usig_max  = _dev_root_s.send_usig_max;

	conn->s_wq_par = 0;

	conn->plid = TS_IP2PR_PATH_LOOKUP_INVALID;

	conn->send_cq_size = _dev_root_s.recv_post_max;
	conn->recv_cq_size = _dev_root_s.send_post_max;
	conn->s_wq_cur     = _dev_root_s.send_post_max;

	conn->send_cq  = NULL;
	conn->recv_cq  = NULL;
	conn->qp       = NULL;
	conn->ca       = NULL;
	conn->pd       = NULL;
	conn->l_key    = 0;
	conn->fmr_pool = NULL;
	conn->hw_port  = 0;
	/*
	 * generic send queue
	 */
	result = sdp_desc_q_init(&conn->send_queue);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	result = sdp_desc_q_init(&conn->send_ctrl);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	/*
	 * create buffer pools for posted events
	 */
	result = sdp_buff_q_init(&conn->recv_post, _recv_post_name, 0);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	result = sdp_buff_q_init(&conn->recv_pool, _recv_pool_name, 0);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	result = sdp_buff_q_init(&conn->send_post, _send_post_name, 0);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	/*
	 * initialize zcopy advertisment tables
	 */
	result = sdp_advt_q_init(&conn->src_pend);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	result = sdp_advt_q_init(&conn->src_actv);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	result = sdp_advt_q_init(&conn->snk_pend);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	/*
	 * initialize zcopy iocb tables
	 */
	result = sdp_iocb_q_init(&conn->r_pend);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    result = sdp_iocb_q_init(&conn->r_snk);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	result = sdp_iocb_q_init(&conn->w_src);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    result = sdp_iocb_q_init(&conn->r_comp);
       TS_EXPECT(MOD_LNX_SDP, !(0 > result));

 
	result = sdp_desc_q_init(&conn->r_src);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
   
	result = sdp_desc_q_init(&conn->w_snk);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	/*
	 * defered CM execution timer
	 */
	tsKernelTimerInit(&conn->cm_exec);
	/*
	 * connection lock
	 */
	result = _sdp_conn_lock_init(conn);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	/*
	 * insert connection into lookup table
	 */
	result = sdp_conn_table_insert(conn);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "CREATE: Error <%d> inserting conn <%d> into table. <%d:%d>",
			 result, conn->hashent, _dev_root_s.sk_entry,
			 _dev_root_s.sk_size);
		goto error_conn;
	}
	/*
	 * set reference
	 */
	atomic_set(&conn->refcnt, 1);
	/*
	 * hold disconnect messages till established state has been reached.
	 */
	conn->flags |= TS_SDP_CONN_F_DIS_HOLD;
	/*
	 * done
	 */
	return conn;
error_conn:
	kmem_cache_free(_dev_root_s.conn_cache, conn);
error:
	sk_free(sk);
	return NULL;
} /* sdp_conn_alloc  */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* module public functions                                               */
/*                                                                       */
/* --------------------------------------------------------------------- */

#define SDP_PROC_CONN_MAIN_HEAD \
        "dst address:port src address:port  ID  comm_id  pid  " \
        "    dst guid         src guid     dlid slid dqpn   "   \
	"sqpn   data sent buff'd data rcvd_buff'd "             \
	"  data written      data read     src_serv snk_serv\n" 
#define SDP_PROC_CONN_MAIN_SEP  \
	"---------------- ---------------- ---- -------- ---- " \
	"---------------- ---------------- ---- ---- ------ "   \
	"------ ---------------- ---------------- "             \
	"---------------- ---------------- -------- --------\n" 
#define SDP_PROC_CONN_MAIN_FORM \
	"%02x.%02x.%02x.%02x:%04x %02x.%02x.%02x.%02x:%04x "  \
	"%04x %08x %04x %08x%08x %08x%08x %04x %04x "         \
	"%06x %06x %08x%08x %08x%08x %08x%08x %08x%08x %08x %08x\n"

/* ========================================================================= */
/*..sdp_proc_dump_conn_main - dump the connection table to /proc */
int sdp_proc_dump_conn_main(char *buffer, 
			    int   max_size, 
			    off_t start_index, 
			    long *end_index)
{
	struct sdp_opt *conn;
	off_t counter = 0;
	int   offset = 0;
	u64   s_guid;
	u64   d_guid;
	unsigned long flags;

	TS_CHECK_NULL(buffer, -EINVAL);
	TS_CHECK_NULL(end_index, -EINVAL);

	*end_index = 0;
	/*
	 * header should only be printed once
	 */
	if (0 == start_index) {

		offset += sprintf((buffer + offset), SDP_PROC_CONN_MAIN_HEAD);
		offset += sprintf((buffer + offset), SDP_PROC_CONN_MAIN_SEP);
	}
	/*
	 * lock table
	 */
	spin_lock_irqsave(&_dev_root_s.sock_lock, flags);
	/*
	 * if the entire table has been walked, exit.
	 */
	if (!(start_index < _dev_root_s.sk_size)) {

		goto done;
	}
	/*
	 * loop across connections.
	 */
	for (counter = start_index; 
	     counter < _dev_root_s.sk_size &&
		     !(TS_SDP_CONN_PROC_MAIN_SIZE > (max_size - offset));
	     counter++) {

		if (NULL == _dev_root_s.sk_array[counter]) {

			continue;
		}

		conn = _dev_root_s.sk_array[counter];

		d_guid = cpu_to_be64(*(u64 *)(conn->d_gid + sizeof(u64)));
		s_guid = cpu_to_be64(*(u64 *)(conn->s_gid + sizeof(u64)));

		offset += sprintf((buffer + offset), SDP_PROC_CONN_MAIN_FORM,
				  (conn->dst_addr & 0xff),
				  ((conn->dst_addr >> 8) & 0xff),
				  ((conn->dst_addr >> 16) & 0xff),
				  ((conn->dst_addr >> 24) & 0xff),
				  conn->dst_port,
				  (conn->src_addr & 0xff),
				  ((conn->src_addr >> 8) & 0xff),
				  ((conn->src_addr >> 16) & 0xff),
				  ((conn->src_addr >> 24) & 0xff),
				  conn->src_port, 
				  conn->hashent,
				  conn->comm_id,
				  conn->pid,
				  (u32)((d_guid >> 32) & 0xffffffff),
				  (u32)(d_guid & 0xffffffff),
				  (u32)((s_guid >> 32) & 0xffffffff),
				  (u32)(s_guid & 0xffffffff),
				  conn->d_lid,
				  conn->s_lid,
				  conn->d_qpn,
				  conn->s_qpn,
				  (u32)((conn->send_bytes >> 32) & 0xffffffff),
				  (u32)(conn->send_bytes & 0xffffffff),
				  (u32)((conn->recv_bytes >> 32) & 0xffffffff),
				  (u32)(conn->recv_bytes & 0xffffffff),
				  (u32)((conn->write_bytes >> 32)& 0xffffffff),
				  (u32)(conn->write_bytes & 0xffffffff),
				  (u32)((conn->read_bytes >> 32) & 0xffffffff),
				  (u32)(conn->read_bytes & 0xffffffff),
				  conn->src_serv,
				  conn->snk_serv);

	}

	*end_index = counter - start_index;

done:
	spin_unlock_irqrestore(&_dev_root_s.sock_lock, flags);
	return offset;
} /* sdp_proc_dump_conn_main */

#define SDP_PROC_CONN_DATA_HEAD \
	" ID  conn inet r s fl sh send_buf recv_buf send q'd recv q'd " \
        "send_seq recv_seq advt_seq smax rmax recv_max lrcv " \
	"lavt rrcv sd sc sp rd rp swqs rbuf sbuf " \
        "us cu send_oob rurg back maxb\n"
#define SDP_PROC_CONN_DATA_SEP  \
	"---- ---- ---- - - -- -- -------- -------- -------- -------- " \
	"-------- -------- -------- ---- ---- -------- ---- " \
	"---- ---- -- -- -- -- -- ---- ---- ---- " \
	"-- -- -------- ---- ---- ----\n"
#define SDP_PROC_CONN_DATA_FORM \
	"%04x %04x %04x %01x %01x %02x %02x " \
	"%08x %08x %08x %08x %08x %08x " \
	"%08x %04x %04x %08x %04x %04x %04x %02x %02x " \
	"%02x %02x %02x %04x %04x %04x %02x %02x %08x %04x %04x %04x\n"

/* ========================================================================= */
/*..sdp_proc_dump_conn_data - dump the connection table to /proc */
int sdp_proc_dump_conn_data(char *buffer,
			    int   max_size,
			    off_t start_index,
			    long *end_index)
{
	struct sock *sk;
	struct sdp_opt *conn;
	off_t counter = 0;
	int   offset = 0;
	unsigned long flags;

	TS_CHECK_NULL(buffer, -EINVAL);
	TS_CHECK_NULL(end_index, -EINVAL);

	*end_index = 0;
	/*
	 * header should only be printed once
	 */
	if (0 == start_index) {

		offset += sprintf((buffer + offset), SDP_PROC_CONN_DATA_HEAD);
		offset += sprintf((buffer + offset), SDP_PROC_CONN_DATA_SEP);
	}
	/*
	 * lock table
	 */
	spin_lock_irqsave(&_dev_root_s.sock_lock, flags);
	/*
	 * if the entire table has been walked, exit.
	 */
	if (!(start_index < _dev_root_s.sk_size)) {

		goto done;
	}
	/*
	 * loop across connections.
	 */
	for (counter = start_index; counter < _dev_root_s.sk_size &&
		     !(TS_SDP_CONN_PROC_DATA_SIZE > (max_size - offset));
	     counter++) {

		if (NULL == _dev_root_s.sk_array[counter]) {

			continue;
		}

		conn = _dev_root_s.sk_array[counter];
		sk = conn->sk;

		offset += sprintf((buffer + offset), SDP_PROC_CONN_DATA_FORM,
				  conn->hashent,
				  conn->state,
				  conn->istate,
				  conn->recv_mode,
				  conn->send_mode,
				  conn->flags,
				  conn->shutdown,
				  conn->send_buf,
				  TS_SDP_OS_SK_RCVBUF(sk),
				  conn->send_qud,
				  conn->byte_strm,
				  conn->send_seq,
				  conn->recv_seq,
				  conn->advt_seq,
				  conn->send_max,
				  conn->recv_max,
				  conn->rwin_max,
				  conn->l_recv_bf,
				  conn->l_advt_bf,
				  conn->r_recv_bf,
				  sdp_desc_q_size(&conn->send_queue),
				  sdp_desc_q_size(&conn->send_ctrl),
				  sdp_buff_q_size(&conn->send_post),
				  sdp_buff_q_size(&conn->recv_pool),
				  sdp_buff_q_size(&conn->recv_post),
				  conn->s_wq_size,
				  conn->recv_size,
				  conn->send_size,
				  conn->send_usig,
				  conn->send_cons,
				  conn->oob_offset,
				  conn->rcv_urg_cnt,
				  conn->backlog_cnt,
				  conn->backlog_max);
	}

	*end_index = counter - start_index;
done:
	spin_unlock_irqrestore(&_dev_root_s.sock_lock, flags);
	return offset;
} /* sdp_proc_dump_conn_data */

#define SDP_PROC_CONN_RDMA_HEAD \
	" ID  rr rw lr lw sa ap aa as rpnd rsnk wsrc rsrc wsnk ra la sc sp " \
        "non_recv non_send  readq    writeq \n"
#define SDP_PROC_CONN_RDMA_SEP  \
	"---- -- -- -- -- -- -- -- -- ---- ---- ---- ---- ---- -- -- -- -- " \
	"-------- -------- -------- --------\n"
#define SDP_PROC_CONN_RDMA_FORM \
	"%04x %02x %02x %02x %02x %02x %02x %02x %02x " \
	"%04x %04x %04x %04x %04x %02x %02x %02x %02x %08x %08x %08x %08x\n"

/* ========================================================================= */
/*..sdp_proc_dump_conn_rdma - dump the connection table to /proc */
int sdp_proc_dump_conn_rdma(char *buffer, 
			    int   max_size,
			    off_t start_index,
			    long *end_index)
{
	struct sdp_opt *conn;
	off_t counter = 0;
	int   offset = 0;
	unsigned long flags;

	TS_CHECK_NULL(buffer, -EINVAL);
	TS_CHECK_NULL(end_index, -EINVAL);

	*end_index = 0;
	/*
	 * header should only be printed once
	 */
	if (0 == start_index) {

		offset += sprintf((buffer + offset), SDP_PROC_CONN_RDMA_HEAD);
		offset += sprintf((buffer + offset), SDP_PROC_CONN_RDMA_SEP);
	}
	/*
	 * lock table
	 */
	spin_lock_irqsave(&_dev_root_s.sock_lock, flags);
	/*
	 * if the entire table has been walked, exit.
	 */
	if (!(start_index < _dev_root_s.sk_size)) {

		goto done;
	}
	/*
	 * loop across connections.
	 */
	for (counter = start_index; counter < _dev_root_s.sk_size &&
		     !(TS_SDP_CONN_PROC_RDMA_SIZE > (max_size - offset));
	     counter++) {

		if (NULL == _dev_root_s.sk_array[counter]) {

			continue;
		}

		conn = _dev_root_s.sk_array[counter];

		offset += sprintf((buffer + offset),SDP_PROC_CONN_RDMA_FORM,
				  conn->hashent,
				  conn->src_recv,
	 			  conn->snk_recv,
				  conn->src_sent,
				  conn->snk_sent,
				  conn->sink_actv,
				  sdp_advt_q_size(&conn->src_pend),
				  sdp_advt_q_size(&conn->src_actv),
				  sdp_advt_q_size(&conn->snk_pend),
				  sdp_iocb_q_size(&conn->r_pend),
                  sdp_iocb_q_size(&conn->r_snk),
				  sdp_iocb_q_size(&conn->w_src),
				  sdp_desc_q_size(&conn->r_src),
				  sdp_desc_q_size(&conn->w_snk),
				  conn->r_max_adv,
				  conn->l_max_adv,
				  conn->s_cur_adv,
				  conn->s_par_adv,
				  conn->nond_recv,
				  conn->nond_send,
				  conn->read_queued,
				  conn->write_queued);
	}

	*end_index = counter - start_index;
done:
	spin_unlock_irqrestore(&_dev_root_s.sock_lock, flags);
	return offset;
} /* sdp_proc_dump_conn_rdma */

#define SDP_PROC_CONN_SOPT_HEAD \
	"dst address:port src address:port src zcpy snk zcpy nd\n"
#define SDP_PROC_CONN_SOPT_SEP  \
	"---------------- ---------------- -------- -------- --\n"
#define SDP_PROC_CONN_SOPT_FORM \
	"%02x.%02x.%02x.%02x:%04x %02x.%02x.%02x.%02x:%04x %08x %08x %04x\n"

/* ========================================================================= */
/*..sdp_proc_dump_conn_sopt - dump the options portion of each conn to /proc */
int sdp_proc_dump_conn_sopt(char *buffer,
			    int   max_size,
			    off_t start_index,
			    long *end_index)
{
	struct sdp_opt *conn;
	off_t counter = 0;
	int   offset = 0;
	unsigned long flags;

	TS_CHECK_NULL(buffer, -EINVAL);
	TS_CHECK_NULL(end_index, -EINVAL);

	*end_index = 0;
	/*
	 * header should only be printed once
	 */
	if (0 == start_index) {

		offset += sprintf((buffer + offset), SDP_PROC_CONN_SOPT_HEAD);
		offset += sprintf((buffer + offset), SDP_PROC_CONN_SOPT_SEP);
	}
	/*
	 * lock table
	 */
	spin_lock_irqsave(&_dev_root_s.sock_lock, flags);
	/*
	 * if the entire table has been walked, exit.
	 */
	if (!(start_index < _dev_root_s.sk_size)) {

		goto done;
	}
	/*
	 * loop across connections.
	 */
	for (counter = start_index; counter < _dev_root_s.sk_size &&
		     !(TS_SDP_SOPT_PROC_DUMP_SIZE > (max_size - offset));
	     counter++) {

		if (NULL == _dev_root_s.sk_array[counter]) {

			continue;
		}

		conn = _dev_root_s.sk_array[counter];

		offset += sprintf((buffer + offset), SDP_PROC_CONN_SOPT_FORM,
				  (conn->dst_addr & 0xff),
				  ((conn->dst_addr >> 8) & 0xff),
				  ((conn->dst_addr >> 16) & 0xff),
				  ((conn->dst_addr >> 24) & 0xff),
				  conn->dst_port,
				  (conn->src_addr & 0xff),
				  ((conn->src_addr >> 8) & 0xff),
				  ((conn->src_addr >> 16) & 0xff),
				  ((conn->src_addr >> 24) & 0xff),
				  conn->src_port,
				  conn->src_zthresh,
				  conn->snk_zthresh,
				  conn->nodelay);
	}

	*end_index = counter - start_index;
done:
	spin_unlock_irqrestore(&_dev_root_s.sock_lock, flags);
	return offset;
} /* sdp_proc_dump_conn_sopt */

/* ========================================================================= */
/*..sdp_proc_dump_device - dump the primary device table to /proc */
int sdp_proc_dump_device(char *buffer,
			 int   max_size,
			 off_t start_index,
			 long *end_index)
{
	struct sdev_hca_port *port;
	struct sdev_hca *hca;
	tUINT64 subnet_prefix;
	tUINT64 guid;
	int hca_count;
	int port_count;
	int offset = 0;

	TS_CHECK_NULL(buffer, -EINVAL);
	TS_CHECK_NULL(end_index, -EINVAL);

	*end_index = 0;
	/*
	 * header should only be printed once
	 */
	if (0 == start_index) {

		offset += sprintf((buffer + offset),
				  "connection table maximum: <%d>\n",
				  _dev_root_s.sk_size);
		offset += sprintf((buffer + offset),
				  "connection table entries: <%d>\n",
				  _dev_root_s.sk_entry);
		offset += sprintf((buffer + offset),
				  "connection table   rover:  <%d>\n",
				  _dev_root_s.sk_rover);

		offset += sprintf((buffer + offset),
				  "max send posted:          <%d>\n",
				  _dev_root_s.send_post_max);
		offset += sprintf((buffer + offset),
				  "max send buffered:        <%d>\n",
				  _dev_root_s.send_buff_max);
		offset += sprintf((buffer + offset),
		      "max send unsignalled:     <%d>\n",
				  _dev_root_s.send_usig_max);
		offset += sprintf((buffer + offset),
				  "max receive posted:       <%d>\n",
				  _dev_root_s.recv_post_max);
		offset += sprintf((buffer + offset),
				  "max receive buffered:     <%d>\n",
				  _dev_root_s.recv_buff_max);

		offset += sprintf((buffer + offset), "HCAs:\n");
	}
	/*
	 * HCA loop
	 */
	for (hca = _dev_root_s.hca_list, hca_count = 0;
	     NULL != hca; hca = hca->next, hca_count++) {

		offset += sprintf((buffer + offset),
				  "  hca %02x: ca <%p> pd <%p> mem <%p> "
				  "l_key <%08x>\n",
				  hca_count, hca->ca, hca->pd, 
				  hca->mem_h, hca->l_key);

		for (port = hca->port_list, port_count = 0;
		     NULL != port; port = port->next, port_count++) {

			subnet_prefix = cpu_to_be64(*(u64 *) (port->gid));
			guid = cpu_to_be64(*(u64 *)(port->gid + sizeof(u64)));

			offset += sprintf((buffer + offset),
					  "    port %02x: index <%d> gid "
					  "<%08x%08x:%08x%08x>\n",
					  port_count,
					  port->index,
					  (u32)((subnet_prefix >> 32) &
						0xffffffff),
					  (u32)(subnet_prefix & 0xffffffff),
					  (u32)((guid >> 32) & 0xffffffff),
					  (u32)(guid & 0xffffffff));
		}
	}

	return offset;
} /* sdp_proc_dump_device */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* initialization/cleanup functions                                      */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*.._sdp_device_table_init -- create hca list */
static tINT32 _sdp_device_table_init(struct sdev_root *dev_root)
{
  //#ifdef _TS_SDP_AIO_SUPPORT
        struct ib_fmr_pool_param fmr_param_s;
	//#endif
	struct ib_physical_buffer buffer_list;
	struct ib_device_properties node_info;
	struct ib_device *hca_handle;
	struct sdev_hca_port *port;
	struct sdev_hca *hca;
	tINT32 result;
	tINT32 hca_count;
	tINT32 port_count;
	tINT32 fmr_size;

	TS_CHECK_NULL(dev_root, -EINVAL);

	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "INIT: Probing HCA/Port list.");
	/*
	 * first count number of HCA's
	 */
	hca_count = 0;
   // node_info = kmalloc(sizeof(struct ib_device_properties),GFP_ATOMIC);

	while (ib_device_get_by_index(hca_count)) {

		hca_count++;
	}

	fmr_size = TS_SDP_FMR_POOL_SIZE / hca_count;

	for (hca_count = 0;
	     (hca_handle = ib_device_get_by_index(hca_count));
	     hca_count++) {

		result = ib_device_properties_get(hca_handle, &node_info);
		if (0 != result) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_FATAL,
				 "INIT: Error <%d> fetching HCA <%x:%d> type.",
				 result, hca_handle, hca_count);
			goto error;
		}
		/*
		 * allocate per-HCA structure
		 */
		hca = kmalloc(sizeof(struct sdev_hca), GFP_KERNEL);
		if (NULL == hca) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_FATAL,
				 "INIT: Error allocating HCA <%x:%d> memory.",
				 hca_handle, hca_count);
			result = -ENOMEM;
			goto error;
		}
		/*
		 * init and insert into list.
		 */
		memset(hca, 0, sizeof(struct sdev_hca));

		hca->next = dev_root->hca_list;
		dev_root->hca_list = hca;

		hca->fmr_pool = NULL;
		hca->mem_h    = NULL;
		hca->pd       = NULL;
		hca->ca = hca_handle;
		/*
		 * protection domain
		 */
		result = ib_pd_create(hca->ca, NULL, &hca->pd);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
				 "INIT: Error <%d> creating HCA <%x:%d> protection domain.",
				 result, hca_handle, hca_count);
			goto error;
		}
		/*
		 * memory registration
		 */
		buffer_list.address = 0;
		buffer_list.size = (unsigned long)high_memory - PAGE_OFFSET;

		hca->iova = 0;

		result = ib_memory_register_physical(hca->pd, 
						     &buffer_list, 
						     1,	/* list_len */
						    &hca->iova, 
						     (unsigned long)(high_memory - PAGE_OFFSET), 
						     0,	/* iova_offset */
						     IB_ACCESS_LOCAL_WRITE,
						     &hca->mem_h,
						     &hca->l_key, &hca->r_key);
		if (0 != result) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
				 "INIT: Error <%d> registering HCA <%x:%d> memory.",
				 result, hca_handle, hca_count);
			goto error;
		}
		//#ifdef _TS_SDP_AIO_SUPPORT
		/*
		 * FMR allocation
		 */
		fmr_param_s.pool_size = fmr_size;
		fmr_param_s.dirty_watermark = TS_SDP_FMR_DIRTY_SIZE;
		fmr_param_s.cache = 1;
		fmr_param_s.max_pages_per_fmr = TS_SDP_IOCB_PAGES_MAX;
		fmr_param_s.access = (IB_ACCESS_LOCAL_WRITE  |
				      IB_ACCESS_REMOTE_WRITE |
				      IB_ACCESS_REMOTE_READ);

		fmr_param_s.flush_function = NULL;
		/*
		 * create SDP memory pool
		 */
		result = ib_fmr_pool_create(hca->pd,
					    &fmr_param_s, 
					    &hca->fmr_pool);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
				 "INIT: Error <%d> creating HCA <%d:%d> fast memory pool.",
				 result, hca->ca, hca_count);
			goto error;
		}
		//#endif /* _TS_SDP_AIO_SUPPORT */
		/*
		 * port allocation
		 */
		for (port_count = 0; port_count < node_info.num_port;
		     port_count++) {

			port = kmalloc(sizeof(struct sdev_hca_port), 
				       GFP_KERNEL);
			if (NULL == port) {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_FATAL,
					 "INIT: Error allocating HCA <%d:%d> port <%x:%d> memory.",
					 hca_handle, hca_count, port_count,
					 node_info.num_port);

				result = -ENOMEM;
				goto error;
			}

			memset(port, 0, sizeof(struct sdev_hca_port));

			port->index    = port_count + 1;
			port->next     = hca->port_list;
			hca->port_list = port;

			result = ib_gid_entry_get(hca->ca, 
						  port->index, 
						  0,	/* index */
						  port->gid);
			if (0 != result) {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_FATAL,
					 "INIT: Error <%d> getting GID for HCA <%d:%d> port <%d:%d>",
					 result, hca->ca, hca_count,
					 port->index, node_info.num_port);
				goto error;
			}
		}
	}
 //   kfree(node_info);
	return 0;
error:
 //kfree(node_info);

	return result;
} /* _sdp_device_table_init */

/* ========================================================================= */
/*.._sdp_device_table_cleanup -- delete hca list */
static tINT32 _sdp_device_table_cleanup(struct sdev_root *dev_root)
{
	struct sdev_hca_port *port;
	struct sdev_hca *hca;

	TS_CHECK_NULL(dev_root, -EINVAL);
	/*
	 * free all hca/ports
	 */
	for (hca = dev_root->hca_list; NULL != hca; hca = dev_root->hca_list) {

		dev_root->hca_list = hca->next;
		hca->next = NULL;

		for (port = hca->port_list; NULL != port; port = hca->port_list) {

			hca->port_list = port->next;
			port->next = NULL;

			kfree(port);
		}

		if (NULL != hca->fmr_pool) {

			(void)ib_fmr_pool_destroy(hca->fmr_pool);
		}

		if (hca->mem_h) {

			(void)ib_memory_deregister(hca->mem_h);
		}

		if (hca->pd) {

			(void)ib_pd_destroy(hca->pd);
		}

		kfree(hca);
	}

	return 0;
} /* _sdp_device_table_cleanup */

/* ========================================================================= */
/*..sdp_conn_table_init -- create a sdp connection table */
int sdp_conn_table_init(int proto_family,
			int conn_size,
			int buff_min,
			int buff_max,
			int recv_post_max,
			int recv_buff_max,
			int send_post_max,
			int send_buff_max,
			int send_usig_max)
{
	tINT32 result;
	tINT32 byte_size;
	tINT32 page_size;

	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "INIT: creating connection table.");

	memset(&_dev_root_s, 0, sizeof(struct sdev_root));
	/*
	 * list
	 */
	_dev_root_s.listen_list = NULL;
	_dev_root_s.bind_list = NULL;

	spin_lock_init(&_dev_root_s.sock_lock);
	spin_lock_init(&_dev_root_s.bind_lock);
	spin_lock_init(&_dev_root_s.listen_lock);
	/*
	 * Initialize IB
	 */
	_dev_root_s.proto = proto_family;

	_dev_root_s.recv_post_max = recv_post_max;
	_dev_root_s.recv_buff_max = recv_buff_max;
	_dev_root_s.send_post_max = send_post_max;
	_dev_root_s.send_buff_max = send_buff_max;
	
	_dev_root_s.send_usig_max = send_usig_max;
	/*
	 * Get HCA/port list
	 */
	result = _sdp_device_table_init(&_dev_root_s);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "INIT: Error <%d> building HCA/port list.", result);
		goto error_hca;
	}
	/*
	 * create socket table
	 */
	if (!(0 < conn_size)) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "INIT: Invalid connection table size. <%d>",
			 conn_size);
		result = -EINVAL;
		goto error_size;
	}

	byte_size = conn_size * sizeof(struct sdp_opt *);
	page_size = (byte_size >> 12) + ((0xfff & byte_size) > 0 ? 1 : 0);
	for (_dev_root_s.sk_ordr = 0;
	     (1 << _dev_root_s.sk_ordr) < page_size; _dev_root_s.sk_ordr++) ;

	_dev_root_s.sk_array = (void *) __get_free_pages(GFP_KERNEL,
						       _dev_root_s.sk_ordr);
	if (NULL == _dev_root_s.sk_array) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "INIT: Failed to create connection table. <%d:%d:%d>",
			 byte_size, page_size, _dev_root_s.sk_ordr);
		result = -ENOMEM;
		goto error_array;
	}

	memset(_dev_root_s.sk_array, 0, byte_size);
	_dev_root_s.sk_size = conn_size - 1; /* top is reserved for invalid */
	_dev_root_s.sk_array++; /* bump table forward so negative -1 index
				   is always a null entry to improve invalid
				   entry processing. */
	/*
	 * IOCB table
	 */
	result = sdp_main_iocb_init();
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "INIT: Error <%d> initializing SDP IOCB table <%d>",
			 result);
		goto error_iocb;
	}
	/*
	 * buffer memory
	 */
	result = sdp_buff_pool_init(buff_min, buff_max);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "INIT: Error <%d> initializing buffer pool.", result);
		goto error_buff;
	}

	_dev_root_s.conn_cache = kmem_cache_create("SdpConnCache",
						   sizeof(struct sdp_opt),
						   0, SLAB_HWCACHE_ALIGN,
						   NULL, NULL);
    
	if (NULL == _dev_root_s.conn_cache) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "INIT: Failed to initialize connection cache.");
		result = -ENOMEM;
		goto error_conn;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        _dev_root_s.sock_cache = kmem_cache_create("sdp_sock",
						   sizeof(struct inet_sock), 
						   0, SLAB_HWCACHE_ALIGN,
						   0, 0);
        if (NULL == _dev_root_s.sock_cache) {
		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "INIT: Error creating slab cache.");
		result = -ENOMEM;
		goto error_sock;
        }
#endif
	/*
	 * start listening
	 */
     
	result = sdp_cm_listen_start(&_dev_root_s);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "INIET: failed to listen for connections on HCA. <%d>",
			 result);
		goto error_listen;
	}

	return 0;
error_listen:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	kmem_cache_destroy(_dev_root_s.sock_cache);
error_sock:
#endif
	kmem_cache_destroy(_dev_root_s.conn_cache);
error_conn:
	(void)sdp_buff_pool_destroy();
error_buff:
	(void)sdp_main_iocb_cleanup();
error_iocb:
	_dev_root_s.sk_array--;
	free_pages((unsigned long)_dev_root_s.sk_array, _dev_root_s.sk_ordr);
error_array:
error_size:
error_hca:
	(void)_sdp_device_table_cleanup(&_dev_root_s);
	return result;
} /* sdp_conn_table_init */

/* ========================================================================= */
/*..sdp_conn_table_clear -- destroy connection managment and tables */
tINT32 sdp_conn_table_clear(void)
{
	struct sdp_opt *conn;
	tINT32 counter;
#if 0
	struct sdp_opt *conn;
	int result=0;
	/*
	 * drain all the connections
	 */

	while (NULL != (conn = _dev_root_s.conn_list)) {

	}
#endif
	while (_dev_root_s.sk_entry)
		for (counter = 0; counter < _dev_root_s.sk_size; counter++) {
			int cnt = 0;
			conn = _dev_root_s.sk_array[counter];
			if (!conn)
				continue;

			printk("SDP: Killing a stray connection %p\n", conn);

			(void)ib_cm_kill(conn->comm_id);
			while(TS_SDP_CONN_PUT(conn))
				++cnt;
			printk("SDP: Killed a connection refcnt=%d\n", cnt);
		}

	(void)sdp_cm_listen_stop(&_dev_root_s);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	kmem_cache_destroy(_dev_root_s.sock_cache);

#endif
	kmem_cache_destroy(_dev_root_s.conn_cache);

	(void)sdp_buff_pool_destroy();

	(void)sdp_main_iocb_cleanup();

	_dev_root_s.sk_array--;
	free_pages((unsigned long)_dev_root_s.sk_array, _dev_root_s.sk_ordr);

	(void)_sdp_device_table_cleanup(&_dev_root_s);

	return 0;
} /* sdp_conn_table_clear */
