/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  This source code may incorporate intellectual property owned by
  Microsoft Corporation. Our provision of this source code does not
  include any licenses or any other rights to you under any Microsoft
  intellectual property. If you would like a license from Microsoft
  (e.g., to rebrand, redistribute), you need to contact Microsoft
  directly.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sdp_event.c 379 2004-07-08 02:24:32Z roland $
*/

#include "sdp_main.h"

/* --------------------------------------------------------------------- */
/*                                                                       */
/* Primary QP Event Handler                                              */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_cq_event_locked -- main per QP event handler */
tINT32 sdp_cq_event_locked(struct ib_cq_entry *comp, struct sdp_opt *conn)
{
    tINT32 result = 0;

    TS_CHECK_NULL(comp, -EINVAL);
    TS_CHECK_NULL(conn, -EINVAL);

    if (0 < (TS_SDP_ST_MASK_ERROR & conn->state))
    {
        /*
         * Ignore events in error state, connection is being 
         * terminated, connection cleanup will take care of freeing
         * posted buffers.
         */
        TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: Completions <%d:%llx:%u:%d:%d> ignored in state. <%04x>",
                 comp->status,
                 comp->work_request_id,
                 comp->op,
                 comp->bytes_transferred,
                 comp->immediate_data, conn->state);
        result = 0;
        goto done;
    }



    /*
     * event demultiplexing
     */


    switch (comp->op)
    {
    case IB_COMPLETION_OP_RECEIVE:

        result = sdp_event_recv(conn, comp);
        break;
    case IB_COMPLETION_OP_SEND:

        result = sdp_event_send(conn, comp);
        break;
    case IB_COMPLETION_OP_RDMA_READ:

        result = sdp_event_read(conn, comp);
        break;
    case IB_COMPLETION_OP_RDMA_WRITE:

        result = sdp_event_write(conn, comp);
        break;
    default:
        /*
         * sometimes the errors come from the CQ but without an
         * operation
         */
        result = 0;

        switch (comp->status)
        {
        case IB_COMPLETION_STATUS_WORK_REQUEST_FLUSHED_ERROR:
            break;
        case IB_COMPLETION_STATUS_SUCCESS:

            TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
                     " EVENT: unknown IB event. <%d>", comp->op);
            result = -ERANGE;
            break;
        default:
            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "EVENT: unhandled error status <%d> for unknown event. <%d>",
                     comp->status, comp->op);
            result = -EIO;
            break;
        }

        break;
    }
    /*
     * release socket before error processing.
     */
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
                 "EVENT: Connection <%04x> ABORT. <%d:%d> <%u:%d:%u>",
                 conn->hashent,
                 result,
                 comp->status,
                 comp->work_request_id,
                 comp->op, comp->bytes_transferred);
        /*
         * abort.
         */
        TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_ERROR_CQ);

        result = sdp_wall_abort(conn);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
                     "EVENT: error during abort. <%d>", result);
        }

        return -EFAULT;
    }

    done:
    return result;
} /* sdp_cq_event_locked */

/* ========================================================================= */
/*..sdp_cq_event_handler -- main per QP event handler, and demuxer */
void sdp_cq_event_handler(struct ib_cq *cq, void *arg)
{
    tINT32 hashent = (unsigned long)arg;
    struct sdp_opt *conn;
    tINT32 result;
#ifdef TS_KERNEL_2_6
    unsigned long flags;
#endif

#ifdef _TS_SDP_DATA_PATH_DEBUG
    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
             "EVENT: Connection <%d> event on cq <%d>", hashent, cq);
#endif
    /*
     * get socket
     */
    conn = sdp_conn_table_lookup(hashent);
    if (NULL == conn)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
                 "EVENT: unknown socket <%d> for cq <%d> event.",
                 hashent, cq);
        goto done;
    }
    /*
     * lock the bottom half of the socket. If the connection is in use,
     * then queue the event, otherwise process this event.
     */
    TS_SDP_CONN_LOCK_BH(conn);
    /*
     * Check for event completions before CM has transitioned to
     * the established state. The CQ will not be polled or rearmed
     * until the CM makes the transition. Once the CM transition
     * has been made, the act of unlocking the connection will
     * drain the CQ.
     */
    if (0 == (TS_SDP_ST_MASK_EVENTS & conn->state))
    {

        if (TS_SDP_CONN_ST_REP_SENT == conn->state ||   /* passive connect */
            TS_SDP_CONN_ST_RTU_SENT == conn->state)
        {   /* active connect */

            result = ib_cm_establish(conn->comm_id, 0);
            TS_EXPECT(MOD_LNX_SDP, !(0 > result));
        } else
        {

            TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_INOUT,
                     "EVENT: socket <%d> unexpected state. <%04x>",
                     hashent, conn->state);
        }

        conn->flags |= TS_SDP_CONN_F_MASK_EVENT;
        goto unlock;
    }

    if (0 == conn->lock.users)
    {
        /*
         * dispatch CQ completions.
         */
        (void)sdp_conn_cq_drain(cq, conn);
    } else
    {
        /*
         * Mark the event which was received, for the unlock code to
         * process at a later time.
         */
        conn->flags |= ((cq == conn->recv_cq) ?
                        TS_SDP_CONN_F_RECV_CQ_PEND :
                        TS_SDP_CONN_F_SEND_CQ_PEND);
    }

    unlock:
    TS_SDP_CONN_UNLOCK_BH(conn);
    TS_SDP_CONN_PUT(conn);
    done:
    return;
} /* sdp_cq_event_handler */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* Connection establishment IB/CM callback functions                     */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*.._sdp_cm_hello_check -- validate the hello header */
static tINT32 _sdp_cm_hello_check(struct msg_hello *msg_hello, tINT32 size)
{
    tINT32 result;

    TS_CHECK_NULL(msg_hello, -EINVAL);

    if (sizeof(struct msg_hello) != size)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: hello message size mismatch. <%d:%d>",
                 size, sizeof(struct msg_hello));
        return -EINVAL;
    }
    /*
     * endian swap
     */
    result = _sdp_msg_net_to_cpu_bsdh(&msg_hello->bsdh);
    TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    result = _sdp_msg_net_to_cpu_hh(&msg_hello->hh);
    TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    /*
     * validation and consistency checks
     */
    if (msg_hello->bsdh.size != sizeof(struct msg_hello))
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: hello message size mismatch. (2) <%d:%d>",
                 msg_hello->bsdh.size, sizeof(struct msg_hello));
        return -EINVAL;
    }

    if (TS_SDP_MSG_MID_HELLO != msg_hello->bsdh.mid)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: hello message unexpected ID. <%d>",
                 msg_hello->bsdh.mid);
        return -EINVAL;
    }

    if (!(0 < msg_hello->hh.max_adv))
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: hellomessage, bad zcopy advertisment. <%d>",
                 msg_hello->hh.max_adv);
        return -EINVAL;
    }

    if ((0xF0 & msg_hello->hh.version) != (0xF0 & TS_SDP_MSG_VERSION))
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: hello message, major version mismatch. <%d:%d>",
                 ((0xF0 & msg_hello->hh.version) >> 4),
                 ((0xF0 & TS_SDP_MSG_VERSION) >> 4));
        return -EINVAL;
    }
#ifdef _TS_SDP_MS_APRIL_ERROR_COMPAT
    if ((TS_SDP_MSG_IPVER & 0x0F) != (msg_hello->hh.ip_ver & 0x0F))
    {
#else
    if ((TS_SDP_MSG_IPVER & 0xF0) != (msg_hello->hh.ip_ver & 0xF0))
    {
#endif

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: hello message, ip version mismatch. <%d:%d>",
                 msg_hello->hh.ip_ver, TS_SDP_MSG_IPVER);
        return -EINVAL;
    }
#ifdef _TS_SDP_DATA_PATH_DEBUG
    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
             "EVENT: BSDH <%04x:%02x:%02x:%08x:%08x:%08x>",
             msg_hello->bsdh.recv_bufs,
             msg_hello->bsdh.flags,
             msg_hello->bsdh.mid,
             msg_hello->bsdh.size,
             msg_hello->bsdh.seq_num, msg_hello->bsdh.seq_ack);
    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
             "EVENT: HH  <%02x:%02x:%02x:%08x:%08x:%04x:%08x:%08x>",
             msg_hello->hh.max_adv,
             msg_hello->hh.ip_ver,
             msg_hello->hh.version,
             msg_hello->hh.r_rcv_size,
             msg_hello->hh.l_rcv_size,
             msg_hello->hh.port,
             msg_hello->hh.src.ipv4.addr, msg_hello->hh.dst.ipv4.addr);
#endif

    return 0; /* success */
} /* _sdp_cm_hello_check */

/* ========================================================================= */
/*.._sdp_cm_hello_ack_check -- validate the hello ack header */
static tINT32 _sdp_cm_hello_ack_check(struct msg_hello_ack *hello_ack,
                                      tINT32 size)
{
    tINT32 result;

    TS_CHECK_NULL(hello_ack, -EINVAL);

    if (sizeof(struct msg_hello_ack) != size)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: hello message size mismatch. <%d:%d>",
                 size, sizeof(struct msg_hello_ack));
        return -EINVAL;
    }

    /*
     * endian swap
     */
    result = _sdp_msg_net_to_cpu_bsdh(&hello_ack->bsdh);
    TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    result = _sdp_msg_net_to_cpu_hah(&hello_ack->hah);
    TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    /*
     * validation and consistency checks
     */
    if (hello_ack->bsdh.size != sizeof(struct msg_hello_ack))
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: hello ack, size mismatch. (2) <%d:%d>",
                 hello_ack->bsdh.size,
                 sizeof(struct msg_hello_ack));
        return -EINVAL;
    }

    if (TS_SDP_MSG_MID_HELLO_ACK != hello_ack->bsdh.mid)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: hello ack, unexpected message. <%d>",
                 hello_ack->bsdh.mid);
        return -EINVAL;
    }

    if (!(0 < hello_ack->hah.max_adv))
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: hello ack, bad zcopy advertisment. <%d>",
                 hello_ack->hah.max_adv);
        return -EINVAL;
    }

    if ((0xF0 & hello_ack->hah.version) != (0xF0 & TS_SDP_MSG_VERSION))
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: hello ack, major version mismatch. <%d:%d>",
                 ((0xF0 & hello_ack->hah.version) >> 4),
                 ((0xF0 & TS_SDP_MSG_VERSION) >> 4));
        return -EINVAL;
    }

#ifdef _TS_SDP_DATA_PATH_DEBUG
    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
             "EVENT: BSDH <%04x:%02x:%02x:%08x:%08x:%08x>",
             hello_ack->bsdh.recv_bufs,
             hello_ack->bsdh.flags,
             hello_ack->bsdh.mid,
             hello_ack->bsdh.size,
             hello_ack->bsdh.seq_num, hello_ack->bsdh.seq_ack);
    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
             "EVENT: HAH <%02x:%02x:%08x>",
             hello_ack->hah.max_adv,
             hello_ack->hah.version, hello_ack->hah.l_rcv_size);
#endif

    return 0; /* success */
} /* _sdp_cm_hello_ack_check */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* State specific Connection Managment callback functions                */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*.._sdp_cm_req -- handler for passive connection open completion */
static tINT32 _sdp_cm_req(tTS_IB_CM_COMM_ID comm_id,
                          struct ib_cm_req_received_param *param,
                          void *arg)
{
    struct msg_hello *msg_hello;
    struct sdp_opt *conn;
    tINT32 result;

    TS_CHECK_NULL(param, -EINVAL);
    TS_CHECK_NULL(param->remote_private_data, -EINVAL);

    msg_hello = (struct msg_hello *) param->remote_private_data;

    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
             "EVENT: REQ. commID <%08x> service ID <%08x> ca <%p> port <%08x>",
             (tUINT32) comm_id, (tUINT32) param->service_id,
             param->device, (tUINT32) param->port);
    /*
     * check Hello Header, to determine if we want the connection.
     */
    result =
    _sdp_cm_hello_check(msg_hello, param->remote_private_data_len);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: REQ recveive, hello message <%d> failed. <%d>",
                 comm_id, result);
        goto done;
    }
    /*
     * Create a connection for this request.
     */
    conn = sdp_conn_alloc(GFP_ATOMIC, comm_id); /* CM sk reference */

    if (NULL == conn)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: REQ receive, failed to create connection. <%d>",
                 comm_id);
        result = -ENOMEM;
        goto done;
    }


    conn->buffered_bytes=0;
    conn->wait_buff_before_src=0;
    conn->inside_read=0;

    conn->src_zthresh = sdp_zcopy_threshold();   //module_param
    conn->use_zcopy = sdp_zcopy_use();

#ifndef TS_KERNEL_2_6
    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,"count inc");
    MOD_INC_USE_COUNT ;
#endif

    /*
     * Lock the new connection before insterting it into any tables.
     */
    TS_SDP_CONN_LOCK(conn);
    /*
     * associate connection with a hca/port, and allocate IB.
     */
    result = sdp_conn_alloc_ib(conn, param->device, param->port);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: REQ receive, Error <%d> binding connection to HCA/port.",
                 result);
        goto error;
    }

    conn->src_addr = msg_hello->hh.dst.ipv4.addr;
    conn->src_port = TS_SDP_MSG_SID_TO_PORT(param->service_id);
    conn->dst_addr = msg_hello->hh.src.ipv4.addr;
    conn->dst_port = msg_hello->hh.port;

    TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_REQ_RECV);
    memcpy(conn->d_gid, param->dgid, sizeof(conn->d_gid));
    conn->d_lid = param->dlid;
    conn->s_lid = param->slid;
    conn->d_qpn = param->remote_qpn;
    conn->s_qpn = param->local_qpn;
    /*
     * read remote information
     */
    conn->send_size = msg_hello->hh.l_rcv_size;
    conn->r_max_adv = msg_hello->hh.max_adv;
    conn->r_recv_bf = msg_hello->bsdh.recv_bufs;
    conn->recv_seq = msg_hello->bsdh.seq_num;
    conn->advt_seq = msg_hello->bsdh.seq_num;
    /*
     * the size we advertise to the stream peer cannot be larger then our
     * internal buffer size.
     */
    conn->send_size = min((tUINT16) sdp_buff_pool_buff_size(),
                          (tUINT16) (conn->send_size -
                                     TS_SDP_MSG_HDR_SIZE));

    result = ib_cm_callback_modify(conn->comm_id,
                                   sdp_cm_event_handler,
                                   (void *) (unsigned long)conn->hashent);
    if (0 != result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "CREATE: CM callback update failed. <%d>", result);
        goto error;
    }
    /*
     * initiate a connection to the stream interface.
     */
    result = sdp_wall_recv_connect(conn);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: initiated connection failed. <%d>", result);
        goto error;
    }
    /*
     * unlock
     */
    TS_SDP_CONN_UNLOCK(conn);
    return TS_IB_CM_CALLBACK_DEFER; /* async response request */

    error:
    conn->comm_id = TS_IB_CM_COMM_ID_INVALID;
    TS_SDP_CONN_UNLOCK(conn);
    TS_SDP_CONN_PUT(conn);  /* CM sk reference */
    done:
    return result;
} /* _sdp_cm_req */

/* ========================================================================= */
/*.._sdp_cm_rep -- handler for active connection open completion */
static tINT32 _sdp_cm_rep(tTS_IB_CM_COMM_ID comm_id,
                          struct ib_cm_rep_received_param *param,
                          struct sdp_opt *conn)
{
    struct msg_hello_ack *hello_ack;
    struct sdpc_buff *buff;
    tINT32 result;
    tINT32 error;

    TS_CHECK_NULL(param, -EINVAL);
    TS_CHECK_NULL(param->remote_private_data, -EINVAL);

    if (NULL == conn)
    {
        return -EINVAL;
    }

    hello_ack = (struct msg_hello_ack *) param->remote_private_data;

    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
             "EVENT: <%d> REP receive. comm ID <%08x> qpn <%06x:%06x>",
             conn->hashent, (tINT32) comm_id,
             param->local_qpn, param->remote_qpn);
    /*
     * lock the connection
     */
    switch (conn->state)
    {
    case TS_SDP_CONN_ST_ERROR_STRM:

        result = sdp_cm_reject(conn);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                     "STRM: Connection manager reject request failed. <%d>",
                     result);
            error = result;
            goto done;
        }
        if (atomic_read(&conn->refcnt)==1)
        {
            TS_TRACE(MOD_LNX_SDP, T_VERY_TERSE, TRACE_FLOW_WARN,
                     "connection reference count is one where must be at least 2( about to be decremented inside a lock)");
        }
        TS_SDP_CONN_PUT(conn);

        break;
    case TS_SDP_CONN_ST_REQ_SENT:

        if (comm_id != conn->comm_id)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "EVENT: REP received. comm ID mismatch. <%08x:%08x>",
                     conn->comm_id, comm_id);
        }
        /*
         * check Hello Header Ack, to determine if we want
         * the connection.
         */
        result = _sdp_cm_hello_ack_check(hello_ack,
                                         param->
                                         remote_private_data_len);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                     "EVENT: REP receive <%d> hello ack error. <%d>",
                     result, comm_id);
            error = result;
            goto reject;
        }

        TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_REP_RECV);
        /*
         * read remote information
         */
        conn->send_size = hello_ack->hah.l_rcv_size;
        conn->r_max_adv = hello_ack->hah.max_adv;
        conn->r_recv_bf = hello_ack->bsdh.recv_bufs;
        conn->recv_seq = hello_ack->bsdh.seq_num;
        conn->advt_seq = hello_ack->bsdh.seq_num;
        conn->d_qpn = param->remote_qpn;
        conn->s_qpn = param->local_qpn;
        /*
         * Pop the hello message that was sent
         */
        buff = sdp_buff_q_get_head(&conn->send_post);
        if (NULL == buff)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                     "EVENT: REP receive <%d> hello msg missing in send pool.",
                     comm_id);
        } else
        {

            result = sdp_buff_pool_put(buff);
            TS_EXPECT(MOD_LNX_SDP, !(0 > result));
        }
        /*
         * The maximum amount of data that can be sent to the remote
         * peer is the smaller of the local and remote buffer sizes, 
         * minus the size of the message header.
         */
        conn->send_size = (min((u16)sdp_buff_pool_buff_size(),
                               (u16)conn->send_size) - 
                           TS_SDP_MSG_HDR_SIZE);

        result = sdp_wall_recv_accept(conn);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                     "EVENT: REP receive <%d>. accept failed. <%d>",
                     comm_id, result);
            error = result;
            goto done;
        }

        break;
    default:
        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: REP received in an unknown state. <%d>",
                 conn->state);
        /*
         * drop CM reference
         */
        result = sdp_wall_recv_drop(conn);
        TS_EXPECT(MOD_LNX_SDP, !(0 > result));

        error = -EPROTO;
        goto done;
    }

    error = TS_IB_CM_CALLBACK_DEFER;    /* async response request */
    goto out;

    reject:
    result = sdp_wall_recv_reject(conn, EPROTO);
    TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    done:
    conn->comm_id = TS_IB_CM_COMM_ID_INVALID;
    out:
    return error;
} /* _sdp_cm_rep */

/* ========================================================================= */
/*.._sdp_cm_idle -- handler for connection idle completion */
static tINT32 _sdp_cm_idle(tTS_IB_CM_COMM_ID comm_id,
                           struct ib_cm_idle_param *param,
                           struct sdp_opt *conn)
{
    tINT32 result = 0;
    tINT32 expect;

    TS_CHECK_NULL(param, -EINVAL);

    if (NULL == conn)
    {
        return -EINVAL;
    }
    /*
     * IDLE should only be called after some other action on the comm_id,
     * which means the callback argument will be a SDP conn, since the
     * only time it is not a SDP conn is the first callback in a passive
     * open.
     */
    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
             "EVENT: <%d> IDLE, comm ID <%08x> reason <%d> state <%04x>",
             conn->hashent, (tINT32) comm_id, param->reason, conn->state);
    /*
     * last comm reference
     */
    conn->comm_id = TS_IB_CM_COMM_ID_INVALID;
    /*
     * check state
     */
    switch (conn->state)
    {
    case TS_SDP_CONN_ST_REQ_PATH:
        /*
         * cancel address resolution
         */
        result = tsIp2prPathRecordCancel(conn->plid);
        TS_EXPECT(MOD_LNX_SDP, !(0 > result));
        /*
         * fall through
         */
    case TS_SDP_CONN_ST_REQ_SENT:   /* active open, Hello msg sent */

        result = sdp_wall_recv_reject(conn, ECONNREFUSED);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "EVENT: error while rejecting conn to gateway. <%d>",
                     result);
            goto error;
        }

        break;
    case TS_SDP_CONN_ST_REP_SENT:   /* passive open, Hello ack msg sent */

        result = sdp_wall_recv_failed(conn, ECONNREFUSED);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "EVENT: error while failing conn to gateway. <%d>",
                     result);
            goto error;
        }

        break;
    case TS_SDP_CONN_ST_REP_RECV:   /* active open, Hello ack msg recv'd */
    case TS_SDP_CONN_ST_REQ_RECV:   /* passive open, Hello msg recv'd */
        /*
         * connection state is outstanding to the gateway, so the
         * connection cannot be destroyed, until the gateway responds.
         * The connection is marked for destruction. No return error,
         * since this is the last CM callback.
         */
        TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_ERROR_CM);
        break;
    case TS_SDP_CONN_ST_TIME_WAIT_1:
    case TS_SDP_CONN_ST_ESTABLISHED:

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: IDLE, unexpected connection state. <%04x>",
                 conn->state);
        /*
         * fall through
         */
    case TS_SDP_CONN_ST_ERROR_STRM:
    case TS_SDP_CONN_ST_TIME_WAIT_2:
        /*
         * Connection is finally dead. Drop the CM reference
         */
        result = sdp_wall_recv_drop(conn);
        TS_EXPECT(MOD_LNX_SDP, !(0 > result));

        break;
    default:

        TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
                 "EVENT: IDLE, unknown connection state. <%04x>",
                 conn->state);
        result = -EINVAL;
        goto error;
        break;
    }

    return 0;
    error:
    /*
     * failed to notify INET of SDP termination of this connection,
     * last attempt to drop the CM reference.
     */
    expect = sdp_wall_recv_drop(conn);
    TS_EXPECT(MOD_LNX_SDP, !(0 > expect));

    return result;
} /* _sdp_cm_idle */

/* ========================================================================= */
/*.._sdp_cm_established -- handler for connection established completion */
static tINT32 _sdp_cm_established(tTS_IB_CM_COMM_ID comm_id,
                                  struct ib_cm_established_param *param,
                                  struct sdp_opt *conn)
{
    tINT32 result = 0;
    tINT32 expect;
    struct sdpc_buff *buff;

    TS_CHECK_NULL(param, -EINVAL);

    if (NULL == conn)
    {
        return -EINVAL;
    }

    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
             "EVENT: <%d> ESTABLISHED, comm ID <%08x> state <%04x>",
             conn->hashent, (tINT32) comm_id, conn->state);
    /*
     * release disconnects.
     */
    conn->flags &= ~TS_SDP_CONN_F_DIS_HOLD;
    /*
     * check state
     */
    switch (conn->state)
    {
    
    case TS_SDP_CONN_ST_REP_SENT:   /* passive open, Hello ack msg sent */

        buff = sdp_buff_q_get_head(&conn->send_post);
        if (NULL == buff)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                     "EVENT: ESTABLISHED, hello ack missing in send pool.");
        } else
        {

            result = sdp_buff_pool_put(buff);
            TS_EXPECT(MOD_LNX_SDP, !(0 > result));
        }

        TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_ESTABLISHED);

        result = sdp_wall_recv_confirm(conn);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "EVENT: error while confirming conn to gateway. <%d>",
                     result);
            goto done;  /* CM sk reference released by Confirm error processing  */
        }

        result = sdp_send_flush(conn);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "EVENT: ESTABLISHED, Failure during send queue flush. <%d>",
                     result);
            goto error;
        }

        result = sdp_recv_flush(conn);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "EVENT: Receive buffer post failed on RDMA read. <%d>",
                     result);
            goto error;
        }

        break;
    case TS_SDP_CONN_ST_RTU_SENT:   /* active open, Hello ack ack'd */
        /*
         * Set state.
         */
        TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_ESTABLISHED);
        /*
         * fall through
         */
    case TS_SDP_CONN_ST_DIS_PEND_1: /* active open, and active close */
    case TS_SDP_CONN_ST_DIS_PEND_R: /* active open, and active close, confirm */
    case TS_SDP_CONN_ST_DIS_PEND_2:

        result = sdp_send_flush(conn);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "EVENT: ESTABLISHED, Failure during send queue flush. <%d>",
                     result);
            goto error;
        }

        break;
    case TS_SDP_CONN_ST_ERROR_STRM:
        /*
         * Sockets has released reference, begin abortive disconnect.
         * Leave state unchanged, time_wait and idle will handle the
         * existing state correctly.
         */
        result = sdp_cm_disconnect(conn);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
                     "EVENT: Error posting CM disconnect. <%d>",
                     result);
        }

        break;
    default:

        TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
                 "EVENT: ESTABLISHED, unexpected connection state. <%04x>",
                 conn->state);
        result = -EINVAL;
        goto error;
        break;
    }

    return 0;
    error:
    expect = sdp_wall_recv_drop(conn);
    TS_EXPECT(MOD_LNX_SDP, !(0 > expect));
    done:
    conn->comm_id = TS_IB_CM_COMM_ID_INVALID;
    return result;
} /* _sdp_cm_established */

/* ========================================================================= */
/*.._sdp_cm_timewait -- handler for connection Time Wait completion */
static tINT32 _sdp_cm_timewait(tTS_IB_CM_COMM_ID comm_id,
                               struct ib_cm_disconnected_param *param, 
                               struct sdp_opt *conn)
{
    tINT32 result = 0;
    tINT32 expect;

    TS_CHECK_NULL(param, -EINVAL);

    if (NULL == conn)
    {

        return -EINVAL;
    }

    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
             "EVENT: <%d> TIME WAIT, ID <%08x> reason <%d> state <%04x>",
             conn->hashent, (tINT32) comm_id, param->reason, conn->state);
    /*
     * Clear out posted receives now, vs after IDLE timeout, which consumes
     * too many buffers when lots of connections are being established and
     * torn down. Here is a good spot since we know that the QP has gone to
     * reset, and pretty much all take downs end up here.
     */
    (void)sdp_buff_q_clear(&conn->recv_post);
    /*
     * check state
     */
    switch (conn->state)
    {
    case TS_SDP_CONN_ST_ERROR_STRM:
        /*
         * error on stream interface, no more call to/from those
         * interfaces.
         */
        break;
    case TS_SDP_CONN_ST_DIS_RECV_R:
    case TS_SDP_CONN_ST_DIS_SEND_2:
    case TS_SDP_CONN_ST_DIS_SENT_2:
    case TS_SDP_CONN_ST_TIME_WAIT_1:
        /*
         * SDP disconnect messages have been exchanged, and
         * DREQ/DREP received, wait for idle timer.
         */
        TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_TIME_WAIT_2);
        break;
    case TS_SDP_CONN_ST_DIS_PEND_1:
    case TS_SDP_CONN_ST_DIS_SEND_1:
    case TS_SDP_CONN_ST_DIS_SENT_1:
    case TS_SDP_CONN_ST_DIS_RECV_1:
        /*
         * connection is being closed without a disconnect message,
         * abortive close.
         */
    case TS_SDP_CONN_ST_ESTABLISHED:
        /*
         * Change state, so we only need to wait for the abort 
         * callback, and idle. Call the abort callback.
         */
        TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_TIME_WAIT_1);

        result = sdp_wall_abort(conn);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
                     "EVENT: error during abort. <%d>", result);
            goto error;
        }

        break;
    default:

        TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
                 "EVENT: TIME_WAIT, unexpected state <%04x> reason. <%d>",
                 conn->state, param->reason);
        result = -EINVAL;
        goto error;
        break;
    }

    return 0;
    error:
    expect = sdp_wall_recv_drop(conn);
    TS_EXPECT(MOD_LNX_SDP, !(0 > expect));

    conn->comm_id = TS_IB_CM_COMM_ID_INVALID;
    return result;
} /* _sdp_cm_timewait */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* Primary Connection Managment callback function                        */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_cm_event_handler -- handler for CM state transitions request */
tTS_IB_CM_CALLBACK_RETURN sdp_cm_event_handler(tTS_IB_CM_EVENT event,
                                               tTS_IB_CM_COMM_ID comm_id,
                                               void *params,
                                               void *arg)
{
    tINT32 hashent = (unsigned long)arg;
    struct sdp_opt *conn = NULL;
    tINT32 result = 0;

    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
             "EVENT: CM state transition <%d> for comm ID <%08x> conn <%d>",
             event, comm_id, hashent);
    /*
     * lookup the connection, on a REQ_RECV the sk will be empty.
     */
    conn = sdp_conn_table_lookup(hashent);
    if (NULL != conn)
    {

        TS_SDP_CONN_LOCK(conn);
    } else
    {

        if (TS_IB_CM_REQ_RECEIVED != event)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
                     "EVENT: No socket for CM state <%d> comm ID <%08x> sock <%d>",
                     event, comm_id, hashent);
        }
    }
    switch (event)
    {
    case TS_IB_CM_REQ_RECEIVED:

        result = _sdp_cm_req(comm_id, params, arg);
        break;
    case TS_IB_CM_IDLE:

        result = _sdp_cm_idle(comm_id, params, conn);
        break;
    case TS_IB_CM_REP_RECEIVED:

        result = _sdp_cm_rep(comm_id, params, conn);
        break;
    case TS_IB_CM_ESTABLISHED:

        result = _sdp_cm_established(comm_id, params, conn);
        break;
    case TS_IB_CM_DISCONNECTED:

        result = _sdp_cm_timewait(comm_id, params, conn);
        break;
    case TS_IB_CM_LAP_RECEIVED:
    case TS_IB_CM_APR_RECEIVED:
        /*
         * null transitions
         */
        result = TS_IB_CM_CALLBACK_PROCEED;
        break;
    default:

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: Unexpected CM state transition! <%d>", event);
        result = -EINVAL;
    }
    /*
     * if a socket was found, release the lock, and put the reference.
     */
    if (NULL != conn)
    {

        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
                     "EVENT: CM state error <%d> <%d:%04x>",
                     result, event, conn->state);
            /*
             * dump connection state if it is being recorded.
             */
            (void)__sdp_conn_state_dump(conn);
        }

        TS_SDP_CONN_UNLOCK(conn);
        TS_SDP_CONN_PUT(conn);
    }

    if (0 > result)
    {
        /*
         * Error in processing needs to be communicated back to the CM
         */
        result = TS_IB_CM_CALLBACK_ABORT;
    }

    return result;
} /* sdp_cm_event_handler */
