/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sdp_inet.c 225 2004-06-03 20:29:09Z libor $
*/

#include "sdp_main.h"
#include "openib_version.h"
#include <linux/tcp.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/moduleparam.h>
#endif
#include <asm/ioctls.h>    
/*
 * list of connections waiting for an incomming connection
 */
static int _proto_family = TS_SDP_DEV_PROTO;
static int _buff_min     = TS_SDP_BUFFER_COUNT_MIN;
static int _buff_max     = TS_SDP_BUFFER_COUNT_MAX;
static int _conn_size    = TS_SDP_DEV_SK_LIST_SIZE;

static int _recv_post_max   = TS_SDP_DEV_RECV_POST_MAX;
static int _recv_buff_max   = TS_SDP_RECV_BUFFERS_MAX;
static int _send_post_max   = TS_SDP_DEV_SEND_POST_MAX;
static int _send_buff_max   = TS_SDP_SEND_BUFFERS_MAX;
static int _send_usig_max   = TS_SDP_CONN_UNSIG_SEND_MAX;
static int _use_zcopy = 0;
static int _src_zthresh = SDP_ZCOPY_THRSH_SRC_DEFAULT;
                
MODULE_AUTHOR("Libor Michalek");
MODULE_DESCRIPTION("InfiniBand SDP module" OPENIB_REVISION);
MODULE_LICENSE("Dual BSD/GPL");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
module_param(_proto_family, int, 0444);
module_param(_buff_min, int, 0444);
module_param(_buff_max, int, 0444);
module_param(_conn_size, int, 0444);
module_param(_recv_post_max, int, 0444);
module_param(_recv_buff_max, int, 0444);
module_param(_send_post_max, int, 0444);
module_param(_send_buff_max, int, 0444);
module_param(_send_usig_max, int, 0444);
module_param(_use_zcopy, int, 0444);
module_param(_src_zthresh, int, 0444);
#else
MODULE_PARM(_proto_family, "i");
MODULE_PARM(_buff_min, "i");
MODULE_PARM(_buff_max, "i");
MODULE_PARM(_conn_size, "i");
MODULE_PARM(_recv_post_max, "i");
MODULE_PARM(_recv_buff_max, "i");
MODULE_PARM(_send_post_max, "i");
MODULE_PARM(_send_buff_max, "i");
MODULE_PARM(_send_usig_max, "i");
MODULE_PARM(_use_zcopy, "i");
MODULE_PARM(_src_zthresh, "i");
#endif

/*
 * socket structure relevant fields:
 *
 * struct sock {
 *    unsigned short     num;         (src port, host    byte order)
 *    __u16              sport;       (src port, network byte order)
 *    __u32              rcv_saddr;   (src addr, network byte order)
 *    __u32              saddr;       (src addr, network byte order)
 *    __u32              daddr;       (dst addr, network byte order)
 *    __u16              dport;       (dst port, network byte order)
 *    unsigned char      shutdown;    (mask of directional close)
 *    wait_queue_head_t *sleep;       (wait for event queue)
 *    int                wmem_queued; (send bytes outstanding)
 *    int                sndbuf;      (possible send bytes outstanding)
 *    unsigned long      lingertime;  (close linger time)
 *    volatile char      linger;      (close linger time valid)
 *    union {}           tp_info;     (cast for STRM/LNX CONN specific data)
 *    int                err;         (error propogation from GW to socket if)
 *    unsigned short     ack_backlog;     (current accept backlog)
 *    unsigned short     max_ack_backlog; (accept max backlog)
 * };
 */
/* --------------------------------------------------------------------- */
/*                                                                       */
/* Notification of significant events.                                   */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_inet_wake_send -- test, set, and notify socket of write space */
void sdp_inet_wake_send(struct sock *sk)
{
	struct sdp_opt *conn;



	
	if (NULL == sk || 
	    NULL == (conn = TS_SDP_GET_CONN(sk))) {

		return;
	} 

	if (NULL != TS_SDP_OS_SK_SOCKET(sk) &&
	    0 < test_bit(SOCK_NOSPACE, &(TS_SDP_OS_SK_SOCKET(sk))->flags) &&
	    0 < __sdp_inet_writable(conn)) {

		read_lock(&TS_SDP_OS_SK_CB_LOCK(sk));
		clear_bit(SOCK_NOSPACE, &(TS_SDP_OS_SK_SOCKET(sk))->flags);

		if (NULL != TS_SDP_OS_SK_SLEEP(sk) &&
		    0 < waitqueue_active(TS_SDP_OS_SK_SLEEP(sk))) {

			wake_up_interruptible(TS_SDP_OS_SK_SLEEP(sk));
		}
		/*
		 * test, clear, and notify. SOCK_ASYNC_NOSPACE
		 */
		sk_wake_async(sk, 2, POLL_OUT);
		read_unlock(&TS_SDP_OS_SK_CB_LOCK(sk));
	}

	return;
} /* sdp_inet_wake_send */

/* ========================================================================= */
/*..sdp_inet_wake_generic -- wake up a socket */
void sdp_inet_wake_generic(struct sock *sk)
{
	if (NULL != sk) {

		read_lock(&TS_SDP_OS_SK_CB_LOCK(sk));

		if (NULL != TS_SDP_OS_SK_SLEEP(sk) &&
		    waitqueue_active(TS_SDP_OS_SK_SLEEP(sk))) {

			wake_up_interruptible_all(TS_SDP_OS_SK_SLEEP(sk));
		}

		read_unlock(&TS_SDP_OS_SK_CB_LOCK(sk));
	}

	return;
} /* sdp_inet_wake_generic */

/* ========================================================================= */
/*..sdp_inet_wake_recv -- wake up a socket for read */
void sdp_inet_wake_recv(struct sock *sk, int len)
{
	if (NULL != sk) {

		read_lock(&TS_SDP_OS_SK_CB_LOCK(sk));
		if (NULL != TS_SDP_OS_SK_SLEEP(sk)) {
			
			wake_up_interruptible(TS_SDP_OS_SK_SLEEP(sk));
		}

		sk_wake_async(sk, 1, POLL_IN);
		read_unlock(&TS_SDP_OS_SK_CB_LOCK(sk));
	}

	return;
} /* sdp_inet_wake_recv */

/* ========================================================================= */
/*..sdp_inet_wake_error -- wake up a socket for error */
void sdp_inet_wake_error(struct sock *sk)
{
	if (NULL != sk) {

		read_lock(&TS_SDP_OS_SK_CB_LOCK(sk));
		if (NULL != TS_SDP_OS_SK_SLEEP(sk)) {
		
			wake_up_interruptible(TS_SDP_OS_SK_SLEEP(sk));
		}

		sk_wake_async(sk, 0, POLL_ERR);
		read_unlock(&TS_SDP_OS_SK_CB_LOCK(sk));
	}

	return;
} /* sdp_inet_wake_error */

/* ========================================================================= */
/*..sdp_inet_wake_urg -- wake up a socket for urgent data */
void sdp_inet_wake_urg(struct sock *sk)
{
	/*
	 * pid for SIGURG/SIGIO has been set. On positive send signal to
	 * process, on negative send signal to processes group.
	 */
	if (NULL != sk) {
#ifdef TS_KERNEL_2_6
		sk_send_sigurg(sk);
#else				/* TS_KERNEL_2_6 */

		if (0 != sk->proc) {

			if (0 < sk->proc) {

				kill_proc(sk->proc, SIGURG, 1);
			}
			else {

				kill_pg(-sk->proc, SIGURG, 1);
			}

			sk_wake_async(sk, 3, POLL_PRI);
		}
#endif				/* TS_KERNEL_2_6 */
	}

	return;
} /* sdp_inet_wake_urg */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* internal socket/handle managment functions                            */
/*                                                                       */
/* --------------------------------------------------------------------- */
  
/* ========================================================================= */
/*.._sdp_inet_abort -- abort an existing connection. */

static int _sdp_inet_abort(struct sdp_opt *conn)
{
	int result;

	TS_CHECK_NULL(conn, -EINVAL);

	conn->send_buf = 0;

	switch (conn->istate) {

	case TS_SDP_SOCK_ST_CONNECT:
	case TS_SDP_SOCK_ST_ACCEPTING:
	case TS_SDP_SOCK_ST_ACCEPTED:

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: Unexpected abort state <%d> for socket. <%d>",
			 conn->istate, conn->pid);

	case TS_SDP_SOCK_ST_ESTABLISHED:
	case TS_SDP_SOCK_ST_CLOSE:
	case TS_SDP_SOCK_ST_DISCONNECT:
	case TS_SDP_SOCK_ST_CLOSING:

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN, "abort 2");
		result = sdp_wall_abort(conn);
		if (0 > result) {

			result = -ECONNABORTED;
			TS_SDP_OS_CONN_SET_ERR(conn, ECONNABORTED);
			conn->istate = TS_SDP_SOCK_ST_ERROR;
		}

		break;
	case TS_SDP_SOCK_ST_LISTEN:
	case TS_SDP_SOCK_ST_CLOSED:
	case TS_SDP_SOCK_ST_ERROR:

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: Unhandled abort state <%d> for socket. <%d>",
			 conn->istate, conn->pid);

		conn->istate = TS_SDP_SOCK_ST_ERROR;
		result = -EINVAL;
		break;
	default:

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: Unknown abort state <%d> for socket. <%d>",
			 conn->istate, conn->pid);

		conn->istate = TS_SDP_SOCK_ST_ERROR;
		result = -EINVAL;
		break;
	}

	return result;
} /* _sdp_inet_abort */


int sdp_signal_abort(struct sdp_opt *conn)
{
    return _sdp_inet_abort(conn);
}



/* ========================================================================= */
/*.._sdp_inet_disconnect -- disconnect a connection. */
static int _sdp_inet_disconnect(struct sdp_opt *conn)
{
	int result = 0;

	TS_CHECK_NULL(conn, -EINVAL);

	switch (conn->istate) {

	case TS_SDP_SOCK_ST_CONNECT:

		result = sdp_wall_abort(conn);
		if (0 > result) {

			result = -ECONNABORTED;
			TS_SDP_OS_CONN_SET_ERR(conn, ECONNABORTED);
			conn->istate = TS_SDP_SOCK_ST_ERROR;
		}

		break;
	case TS_SDP_SOCK_ST_ESTABLISHED:
	case TS_SDP_SOCK_ST_ACCEPTED:

		conn->istate = TS_SDP_SOCK_ST_DISCONNECT;
		result = sdp_wall_send_close(conn);
        if (0 > result) {

			result = -ECONNABORTED;
			TS_SDP_OS_CONN_SET_ERR(conn, ECONNABORTED);
			conn->istate = TS_SDP_SOCK_ST_ERROR;
		}

		break;
	case TS_SDP_SOCK_ST_CLOSE:

		conn->istate = TS_SDP_SOCK_ST_CLOSING;
		result = sdp_wall_send_closing(conn);
		if (0 > result) {

			result = -ECONNABORTED;
			TS_SDP_OS_CONN_SET_ERR(conn, ECONNABORTED);
			conn->istate = TS_SDP_SOCK_ST_ERROR;
		}

		break;
	case TS_SDP_SOCK_ST_ACCEPTING:
	case TS_SDP_SOCK_ST_DISCONNECT:
	case TS_SDP_SOCK_ST_CLOSING:
		/*
		 * nothing to do, and somewhat unexpected state
		 */
		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: Unexpected disconnect state <%04x> for socket. <%d>",
			 conn->istate, conn->pid);
		break;
	case TS_SDP_SOCK_ST_LISTEN:
	case TS_SDP_SOCK_ST_CLOSED:
	case TS_SDP_SOCK_ST_ERROR:

		break;
	default:

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: Unknown disconnect state <%04x> for socket. <%d>",
			 conn->istate, conn->pid);

		conn->istate = TS_SDP_SOCK_ST_ERROR;
		result = -EINVAL;
		break;
	}

	return result;
} /* _sdp_inet_disconnect */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* Linux SOCKET interface, module specific functions                     */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*.._sdp_inet_release -- release/close a socket */
static int _sdp_inet_release(struct socket *sock)
{
	struct sdp_opt *conn;
	struct sock *sk;
	int  result;
	long timeout;
	u32  flags;

	TS_CHECK_NULL(sock, -EINVAL);
	TS_CHECK_NULL(sock->file, -EINVAL);

	if (NULL != sock->sk) {

		sk = sock->sk;
		conn = TS_SDP_GET_CONN(sk);

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
			 "SOCK: <%d> release <%08x:%04x> <%08x:%04x> <%04x> <%d:%d:%d:%d>",
			 conn->hashent, conn->src_addr, conn->src_port,
			 conn->dst_addr, conn->dst_port, conn->istate,
			 TS_SDP_OS_SK_LINGER(sk), TS_SDP_OS_SK_LINGERTIME(sk),
			 conn->byte_strm, conn->src_recv);
		/*
		 * clear out sock, so we only do this once.
		 */
		sock->sk = NULL;
	
		TS_SDP_CONN_LOCK(conn);
		conn->shutdown = TS_SDP_SHUTDOWN_MASK;

		if (TS_SDP_SOCK_ST_LISTEN == conn->istate) {
			/*
			 * stop listening
			 */
			result = sdp_inet_listen_stop(conn);
			if (0 > result) {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_WARN,
					 "SOCK: Error <%d> while releasing listen socket <%d>",
					 result, conn->pid);
			}
		}
		else {
			/*
			 * get blocking nature of the socket.
			 */
			if (sock->file) {
				flags =
				    (0 <
				     (sock->file->
				      f_flags & O_NONBLOCK)) ? MSG_DONTWAIT : 0;
			}
			else {
				flags = 0;
			}
			/*
			 * If there is data in the receive queue, flush it,
			 * and consider this an abort. Otherwise consider
			 * this a gracefull close.
			 */
			if (0 < sdp_buff_q_size(&conn->recv_pool) ||
			    0 < conn->src_recv ||
			    (0 < TS_SDP_OS_SK_LINGER(sk) &&
			     0 == TS_SDP_OS_SK_LINGERTIME(sk))) {
				/*
				 * abort.
				 */
				result = _sdp_inet_abort(conn);
				if (0 > result) {

					TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
						 TRACE_FLOW_WARN,
						 "SOCK: Error <%d> while aborting socket <%d> on release.",
						 result, conn->pid);
				}
			}
			else {
				/*
				 * disconnect. (state dependant)
				 */
				result = _sdp_inet_disconnect(conn);
                if (0 > result) {

					TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
						 TRACE_FLOW_WARN,
						 "SOCK: Error <%d> while disconnecting <%d> on release.",
						 result, conn->pid);
				}
				else {
               					/*
					 * Skip lingering/canceling if 
					 * non-blocking and not exiting.
					 */
					if (0 == (MSG_DONTWAIT & flags) ||
					    0 < (PF_EXITING & current->flags)) {
						/*
						 * Wait if linger is set and
						 * process is not exiting.
						 */
						if (0 < TS_SDP_OS_SK_LINGER(sk)
						    && !(PF_EXITING & current->
							 flags)) {


                           
							DECLARE_WAITQUEUE(wait,
									  current);
							timeout =
							    TS_SDP_OS_SK_LINGERTIME
							    (sk);

							add_wait_queue
							    (TS_SDP_OS_SK_SLEEP
							     (sk), &wait);
							set_current_state
							    (TASK_INTERRUPTIBLE);

							while (0 < timeout &&
							       0 ==
							       (TS_SDP_ST_MASK_CLOSED
								& conn->
								istate)) {

								TS_SDP_CONN_UNLOCK
								    (conn);
								timeout =
								    schedule_timeout
								    (timeout);
								TS_SDP_CONN_LOCK
								    (conn);

								if (signal_pending(current)) {

									break;
								}
							}

							set_current_state
							    (TASK_RUNNING);
							remove_wait_queue
							    (TS_SDP_OS_SK_SLEEP
							     (sk), &wait);
						} /* if (linger) */
#if 0
						/*
						 * On a blocking socket, if still draining after linger,
						 * Cancel write and close again to force closing the
						 * connection.
						 */
						if (0 <
						    (TS_SDP_ST_MASK_DRAIN &
						     conn->istate)) {

							result =
							    sdp_iocb_q_cancel_all_write
							    (conn, -ECANCELED);
							TS_EXPECT(MOD_LNX_SDP,
								  !(0 >
								    result));

							result =
							    _sdp_inet_disconnect
							    (conn);
							TS_EXPECT(MOD_LNX_SDP,
								  !(0 >
								    result));
						}
#endif
					} /* if (blocking) */
				} /* else (disconnect error) */
			} /* else (abortive) */
		} /* else (listen) */
      
		if (0 < (TS_SDP_ST_MASK_CLOSED & conn->istate)) {

         
			/*
			 * pass
			 */
		}
		else {

		}
		/*
		 * finally drop socket reference. (socket API reference)
		 */
		sock_orphan(sk);
		TS_SDP_CONN_UNLOCK(conn);
		TS_SDP_CONN_PUT(conn);
	}
	else {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
			 "SOCK: empty release <%d:%d> <%08x>", sock->type,
			 sock->state, sock->flags);
	}

	return 0;
} /* _sdp_inet_release */

/* ========================================================================= */
/*.._sdp_inet_bind -- bind a socket to an address/interface */
static int _sdp_inet_bind(struct socket *sock,
			     struct sockaddr *uaddr,
			     int size)
{
	struct sockaddr_in *addr = (struct sockaddr_in *)uaddr;
	struct sock *sk;
	struct sdp_opt *conn;
	unsigned int addr_result = RTN_UNSPEC;
	u16 bind_port;
	int result;

	TS_CHECK_NULL(sock, -EINVAL);
	TS_CHECK_NULL(sock->sk, -EINVAL);
	TS_CHECK_NULL(uaddr, -EINVAL);

	sk = sock->sk;
	conn = TS_SDP_GET_CONN(sk);

	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "SOCK: <%d> bind operation <%d> <%d> <%08x:%04x>",
		 conn->hashent, conn->pid, addr->sin_family,
		 addr->sin_addr.s_addr, addr->sin_port);

	if (size < sizeof(struct sockaddr_in)) {

		return -EINVAL;
	}

	if (_proto_family != addr->sin_family &&
	    AF_INET != addr->sin_family && AF_UNSPEC != addr->sin_family) {

		return -EAFNOSUPPORT;
	}
	/*
	 * Basically we're OK with INADDR_ANY or a local interface
	 * (TODO: loopback)
	 */
	if (INADDR_ANY != addr->sin_addr.s_addr) {
		/*
		 * make sure we have a valid binding address
		 */
		addr_result = inet_addr_type(addr->sin_addr.s_addr);

		if (TS_SDP_OS_SK_INET_FREEBIND(sk) == 0 &&
		    RTN_LOCAL != addr_result &&
		    RTN_MULTICAST != addr_result &&
		    RTN_BROADCAST != addr_result) {

			return -EADDRNOTAVAIL;
		}
	}
	/*
	 * check bind permission for low ports.
	 */
	bind_port = ntohs(addr->sin_port);
	if (0 < bind_port &&
	    bind_port < PROT_SOCK && 0 == capable(CAP_NET_BIND_SERVICE)) {

		return -EACCES;
	}
	/*
	 * socket checks.
	 */
	TS_SDP_CONN_LOCK(conn);

	if (TS_SDP_SOCK_ST_CLOSED != conn->istate || 0 < conn->src_port) {

		result = -EINVAL;
		goto done;
	}

	conn->src_addr = ntohl(addr->sin_addr.s_addr);

	if (RTN_MULTICAST == addr_result || RTN_BROADCAST == addr_result) {

		conn->src_addr = 0;
	}

	result = sdp_inet_port_get(conn, bind_port);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "SOCK: Error getting port during bind. <%d>", result);

		conn->src_addr = 0;

		goto done;
	}

	if (INADDR_ANY != conn->src_addr) {

		TS_SDP_OS_SK_USERLOCKS_SET(sk, SOCK_BINDADDR_LOCK);
	}

	if (0 < bind_port) {

		TS_SDP_OS_SK_USERLOCKS_SET(sk, SOCK_BINDADDR_LOCK);
	}

	TS_SDP_OS_SK_INET_RCV_SADDR(sk) = htonl(conn->src_addr);
	TS_SDP_OS_SK_INET_SADDR(sk) = htonl(conn->src_addr);

	TS_SDP_OS_SK_INET_NUM(sk) = conn->src_port;
	TS_SDP_OS_SK_INET_SPORT(sk) = htons(conn->src_port);
	TS_SDP_OS_SK_INET_DADDR(sk) = 0;
	TS_SDP_OS_SK_INET_DPORT(sk) = 0;

	result = 0;
done:
	TS_SDP_CONN_UNLOCK(conn);
	return result;
} /* _sdp_inet_bind */

/* ========================================================================= */
/*.._sdp_inet_connect -- connect a socket to a remote address */
static int _sdp_inet_connect(struct socket *sock,
				struct sockaddr *uaddr,
				int size,
				int flags)
{
	struct sockaddr_in *addr = (struct sockaddr_in *)uaddr;
	struct sock *sk;
	struct sdp_opt *conn;
	long timeout;
	int result,rc;

	TS_CHECK_NULL(sock, -EINVAL);
	TS_CHECK_NULL(sock->sk, -EINVAL);
	TS_CHECK_NULL(uaddr, -EINVAL);

	sk = sock->sk;
	conn = TS_SDP_GET_CONN(sk);

    rc=0;
	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "SOCK: <%d> connect operation <%d> <%d> <%08x:%04x>",
		 conn->hashent, conn->pid, addr->sin_family,
		 addr->sin_addr.s_addr, addr->sin_port);

	if (size < sizeof(struct sockaddr_in)) {

		return -EINVAL;
	}

	if (_proto_family != addr->sin_family &&
	    AF_INET != addr->sin_family && AF_UNSPEC != addr->sin_family) {

		return -EAFNOSUPPORT;
	}

	if (MULTICAST(addr->sin_addr.s_addr) ||
	    BADCLASS(addr->sin_addr.s_addr) ||
	    ZERONET(addr->sin_addr.s_addr) ||
	    LOCAL_MCAST(addr->sin_addr.s_addr) ||
	    INADDR_ANY == addr->sin_addr.s_addr) {

		return -EINVAL;
	}
	/*
	 * lock socket
	 */
	TS_SDP_CONN_LOCK(conn);

	switch (sock->state) {
	case SS_UNCONNECTED:

		if (0 == (TS_SDP_ST_MASK_CLOSED & conn->istate)) {

			result = -EISCONN;
			goto done;
		}

		if (0 == conn->src_port) {

			result = sdp_inet_port_get(conn, 0);
			if (0 > result) {

				TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE,
					 TRACE_FLOW_WARN,
					 "SOCK: Error getting port during connect. <%d>",
					 result);
				goto done;
			}

			TS_SDP_OS_SK_INET_NUM(sk) = conn->src_port;
			TS_SDP_OS_SK_INET_SPORT(sk) = htons(conn->src_port);
		}

		TS_SDP_OS_CONN_SET_ERR(conn, 0);

		sock->state = SS_CONNECTING;
		conn->istate = TS_SDP_SOCK_ST_CONNECT;

		conn->dst_addr = ntohl(addr->sin_addr.s_addr);
		conn->dst_port = ntohs(addr->sin_port);

		TS_SDP_CONN_HOLD(conn);	/* CM reference */
		/*
		 * close, allow connection completion notification.
		 */
		set_bit(SOCK_NOSPACE, &sock->flags);

		result = sdp_wall_send_connect(conn);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
				 "SOCK: Error connecting to gateway interface. <%d>",
				 result);

			conn->dst_addr = 0;
			conn->dst_port = 0;

			sock->state = SS_UNCONNECTED;
			conn->istate = TS_SDP_SOCK_ST_CLOSED;
            TS_SDP_CONN_UNLOCK(conn);
			TS_SDP_CONN_PUT(conn);	/* CM reference */
            return result;

           
           
		   
		}

		TS_SDP_OS_SK_INET_DADDR(sk) = htonl(conn->dst_addr);
		TS_SDP_OS_SK_INET_DPORT(sk) = htons(conn->dst_port);

		result = -EINPROGRESS;
		break;
	case SS_CONNECTING:

		result = -EALREADY;
		break;
	case SS_CONNECTED:

		result = -EISCONN;
		goto done;
	default:

		result = -EINVAL;
		goto done;
	}
	/*
	 * wait for connection to complete.
	 */
	timeout = sock_sndtimeo(sk, (O_NONBLOCK & flags));
	if (0 < timeout) {

		DECLARE_WAITQUEUE(wait, current);
		add_wait_queue(TS_SDP_OS_SK_SLEEP(sk), &wait);
		set_current_state(TASK_INTERRUPTIBLE);

		while (0 < timeout && TS_SDP_SOCK_ST_CONNECT == conn->istate) {

			TS_SDP_CONN_UNLOCK(conn);
			timeout = schedule_timeout(timeout);
			TS_SDP_CONN_LOCK(conn);

			if (signal_pending(current)) {

				break;
			}
		}

		set_current_state(TASK_RUNNING);
		remove_wait_queue(TS_SDP_OS_SK_SLEEP(sk), &wait);

		if (TS_SDP_SOCK_ST_CONNECT == conn->istate) {

			if (0 < timeout) {

				result = sock_intr_errno(timeout);
			}

			goto done;
		}
	}
	/*
	 * check state before exiting. It's possible the that connection
	 * error'd or is being closed after reaching ESTABLISHED at this
	 * point. In this case connect should return normally and allow
	 * the normal mechnaism for detecting these states.
	 */
	switch (conn->istate) {
	case TS_SDP_SOCK_ST_CONNECT:
		break;
	case TS_SDP_SOCK_ST_ESTABLISHED:
	case TS_SDP_SOCK_ST_CLOSE:

		sock->state = SS_CONNECTED;
		result = 0;
		break;
	case TS_SDP_SOCK_ST_CLOSED:
	case TS_SDP_SOCK_ST_ERROR:

		result = TS_SDP_CONN_ERROR(conn) ? : -ECONNABORTED;
		sock->state = SS_UNCONNECTED;
		break;
	default:

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "SOCK: Unexpected connection state leaving connect. <%d:%d>",
			 conn->istate, sock->state);
		break;
	}
   	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "SOCK: <%d> connect complete <%08x:%04x> state <%04x>",
		 conn->hashent, addr->sin_addr.s_addr, addr->sin_port,
		 conn->istate);

done:
	TS_SDP_CONN_UNLOCK(conn);
	return result;
} /* _sdp_inet_connect */

/* ========================================================================= */
/*.._sdp_inet_listen -- listen on a socket for incomming addresses. */
static int _sdp_inet_listen(struct socket *sock, int backlog)
{
	struct sock *sk;
	struct sdp_opt *conn;
	int result;

	TS_CHECK_NULL(sock, -EINVAL);
	TS_CHECK_NULL(sock->sk, -EINVAL);

	sk = sock->sk;
	conn = TS_SDP_GET_CONN(sk);

	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "SOCK: <%d> listen operation <%d> <%08x:%04x> <%04x>",
		 conn->hashent, conn->pid, conn->src_addr, conn->src_port,
		 conn->istate);

	TS_SDP_CONN_LOCK(conn);

	if (SS_UNCONNECTED != sock->state ||
	    (TS_SDP_SOCK_ST_CLOSED != conn->istate &&
	     TS_SDP_SOCK_ST_LISTEN != conn->istate)) {

		result = -EINVAL;
		goto done;
	}

	if (TS_SDP_SOCK_ST_LISTEN != conn->istate) {

		result = sdp_inet_listen_start(conn);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "SOCK: Failed to start listening. <%d>",
				 result);
			goto done;
		}

		if (0 == conn->src_port) {

			result = sdp_inet_port_get(conn, 0);
			if (0 > result) {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_WARN,
					 "SOCK: Error getting port during listen. <%d>",
					 result);
				goto done;
			}

			TS_SDP_OS_SK_INET_NUM(sk) = conn->src_port;
			TS_SDP_OS_SK_INET_SPORT(sk) = htons(conn->src_port);
		}
	}

#if 0				/* BUG 2034 workaround. */
	conn->backlog_max = backlog;
#else
    conn->backlog_max = min(backlog,1024);
	//conn->backlog_max = 1024;
#endif
	result = 0;

done:
	TS_SDP_CONN_UNLOCK(conn);
	return result;
} /* _sdp_inet_listen */

/* ========================================================================= */
/*.._sdp_inet_accept -- accept a new socket from a listen socket. */
static int _sdp_inet_accept(struct socket *listen_sock,
			       struct socket *accept_sock, 
			       int flags)
{
	struct sock *listen_sk;
	struct sock *accept_sk = NULL;
	struct sdp_opt *listen_conn;
	struct sdp_opt *accept_conn = NULL;
	int result;
	long timeout;

	TS_CHECK_NULL(listen_sock, -EINVAL);
	TS_CHECK_NULL(accept_sock, -EINVAL);
	TS_CHECK_NULL(listen_sock->sk, -EINVAL);

	listen_sk = listen_sock->sk;
	listen_conn = TS_SDP_GET_CONN(listen_sk);

	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "SOCK: <%d> accept operation <%d> <%08x:%04x> <%04x>",
		 listen_conn->hashent, listen_conn->pid, listen_conn->src_addr,
		 listen_conn->src_port, listen_conn->istate);

	TS_SDP_CONN_LOCK(listen_conn);

	if (TS_SDP_SOCK_ST_LISTEN != listen_conn->istate) {

		result = -EINVAL;
		goto listen_done;
	}

	timeout = sock_rcvtimeo(listen_sk, (O_NONBLOCK & flags));
	/*
	 * if there is no socket on the queue, wait for one. It' done in a
	 * loop in case there is a problem with the first socket we hit.
	 */
	while (NULL == accept_conn) {
		/*
		 * No pending socket wait.
		 */
		accept_conn = sdp_inet_accept_q_get(listen_conn);
		if (NULL == accept_conn) {

			DECLARE_WAITQUEUE(wait, current);
			add_wait_queue(TS_SDP_OS_SK_SLEEP(listen_sk), &wait);
			set_current_state(TASK_INTERRUPTIBLE);

			while (0 < timeout &&
			       TS_SDP_SOCK_ST_LISTEN == listen_conn->istate &&
			       0 == listen_conn->backlog_cnt) {

				TS_SDP_CONN_UNLOCK(listen_conn);
				timeout = schedule_timeout(timeout);
				TS_SDP_CONN_LOCK(listen_conn);

				if (signal_pending(current)) {

					break;
				}
			}

			set_current_state(TASK_RUNNING);
			remove_wait_queue(TS_SDP_OS_SK_SLEEP(listen_sk), &wait);
			/*
			 * process result
			 */
			if (0 == listen_conn->backlog_cnt) {
				result = 0;

				if (TS_SDP_SOCK_ST_LISTEN !=
				    listen_conn->istate) {

					result = -EINVAL;
				}
				if (signal_pending(current)) {

					result = sock_intr_errno(timeout);
				}
				if (0 == timeout) {

					result = -EAGAIN;
				}

				goto listen_done;
			}
		}
		else {

			accept_sk = accept_conn->sk;
			switch (accept_conn->istate) {
			case TS_SDP_SOCK_ST_ACCEPTED:

				sock_graft(accept_sk, accept_sock);

				accept_conn->pid = TS_SDP_OS_GET_PID();
				accept_sock->state = SS_CONNECTED;

				accept_conn->istate =
				    TS_SDP_SOCK_ST_ESTABLISHED;
				sdp_inet_wake_send(accept_sk);

				break;
			case TS_SDP_SOCK_ST_ACCEPTING:

				sock_graft(accept_sk, accept_sock);

				accept_conn->pid = TS_SDP_OS_GET_PID();
				accept_sock->state = SS_CONNECTED;

				accept_conn->istate = TS_SDP_SOCK_ST_ACCEPTED;
				/*
				 * connection completion/establishment will
				 * open this up
				 */
				set_bit(SOCK_NOSPACE, &accept_sock->flags);

				break;
			case TS_SDP_SOCK_ST_CLOSE:

				sock_graft(accept_sk, accept_sock);

				accept_conn->pid = TS_SDP_OS_GET_PID();
				accept_sock->state = SS_CONNECTED;

				sdp_inet_wake_send(accept_sk);

				break;
			default:

				TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE,
					 TRACE_FLOW_INOUT,
					 "SOCK: Serious accept error. bad state <%04x:%04x>",
					 accept_conn->istate,
					 accept_conn->state);

			case TS_SDP_SOCK_ST_CLOSED:
			case TS_SDP_SOCK_ST_CLOSING:
			case TS_SDP_SOCK_ST_ERROR:
				/*
				 * this accept socket has problems, keep
				 * trying.
				 */
				TS_SDP_CONN_UNLOCK(accept_conn);	/* AcceptQueueGet */
				TS_SDP_CONN_PUT(accept_conn);	/* INET reference (AcceptQueue ref) */

				accept_sk = NULL;
				accept_sock->sk = NULL;
				accept_conn = NULL;

				break;
			}

			if (NULL != accept_conn) {
				/*
				 * Connections returned from the AcceptQueue
				 * are holding their lock, before returning
				 * the connection to the user, release the
				 * lock
				 */
				TS_SDP_CONN_UNLOCK(accept_conn);	/* AcceptQueueGet */
			}
		}
	}

	result = 0;
listen_done:
	TS_SDP_CONN_UNLOCK(listen_conn);

	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "SOCK: <%d:%d> accept complete <%08x:%04x><%08x:%04x> state <%04x>",
		 listen_conn->hashent,
		 (NULL == accept_conn ?
		  TS_SDP_DEV_SK_INVALID : accept_conn->hashent),
		 (NULL == accept_sk ? 0 : accept_conn->src_addr),
		 (NULL == accept_sk ? 0 : accept_conn->src_port),
		 (NULL == accept_sk ? 0 : accept_conn->dst_addr),
		 (NULL == accept_sk ? 0 : accept_conn->dst_port),
		 (NULL == accept_sk ? 0 : accept_conn->istate),
		 listen_conn->backlog_cnt);

	return result;
} /* _sdp_inet_accept */

/* ========================================================================= */
/*.._sdp_inet_getname -- return a sockets address information */
static int _sdp_inet_getname(struct socket *sock,
				struct sockaddr *uaddr,
				int *size, 
				int peer)
{
	struct sockaddr_in *addr = (struct sockaddr_in *)uaddr;
	struct sock *sk;
	struct sdp_opt *conn;

	TS_CHECK_NULL(sock, -EINVAL);
	TS_CHECK_NULL(sock->sk, -EINVAL);
	TS_CHECK_NULL(addr, -EINVAL);
	TS_CHECK_NULL(size, -EINVAL);

	sk = sock->sk;
	conn = TS_SDP_GET_CONN(sk);

	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "SOCK: <%d> get name operation <%08x:%04x> <%08x:%04x> <%04x>",
		 conn->hashent, conn->src_addr, conn->src_port, conn->dst_addr,
		 conn->dst_port, conn->istate);

	addr->sin_family = _proto_family;
	if (0 < peer) {

		if (0 < htons(conn->dst_port) &&
		    0 == (TS_SDP_ST_MASK_CLOSED & conn->istate)) {

			addr->sin_port = htons(conn->dst_port);
			addr->sin_addr.s_addr = htonl(conn->dst_addr);
		}
		else {

			return -ENOTCONN;
		}
	}
	else {

		addr->sin_port = htons(conn->src_port);
		addr->sin_addr.s_addr = htonl(conn->src_addr);
	}

	*size = sizeof(struct sockaddr_in);

	return 0;
} /* _sdp_inet_getname */

/* ========================================================================= */
/*.._sdp_inet_poll -- poll a socket for activity */
static unsigned int _sdp_inet_poll(struct file *file,
				   struct socket *sock,
				   poll_table * wait)
{
	struct sock *sk;
	struct sdp_opt *conn;
	unsigned int mask = 0;

	TS_CHECK_NULL(sock, 0);
	TS_CHECK_NULL(sock->sk, 0);
	/*
	 * file and/or wait can be NULL, once poll is asleep and needs to
	 * recheck the falgs on being woken.
	 */
	sk = sock->sk;
	conn = TS_SDP_GET_CONN(sk);

#ifdef  _TS_SDP_DATA_PATH_DEBUG
	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "SOCK: <%d> poll <%08x:%04x> <%08x:%04x> state <%04x> flags <%08x>",
		 conn->hashent, conn->src_addr, conn->src_port, conn->dst_addr,
		 conn->dst_port, conn->istate, sock->flags);
#endif

	poll_wait(file, TS_SDP_OS_SK_SLEEP(sk), wait);
	/*
	 * general poll philosophy: too many mask bits are better then too
	 * few. POLLHUP is not direction maskable, and the recv path in more
	 * interesting for hang up. However, after receiving an EOF we want
	 * to be able to still select on write, if POLLHUP was set, this
	 * would not be possible.
	 */
	/*
	 * no locking, should be safe as is.
	 */
	switch (conn->istate) {

	case TS_SDP_SOCK_ST_LISTEN:

		mask |= (0 < conn->backlog_cnt) ? (POLLIN | POLLRDNORM) : 0;
		break;
	case TS_SDP_SOCK_ST_ERROR:

		mask |= POLLERR;
		break;
	case TS_SDP_SOCK_ST_CLOSED:

		mask |= POLLHUP;
		break;
	case TS_SDP_SOCK_ST_ESTABLISHED:
		/*
		 * fall through
		 */
	default:
		/*
		 * recv EOF _and_ recv data
		 */
		if (!(conn->byte_strm < TS_SDP_OS_SK_RCVLOWAT(sk)) ||
		    0 < (TS_SDP_SHUTDOWN_RECV & conn->shutdown)) {

			mask |= POLLIN | POLLRDNORM;
		}
		/*
		 * send EOF _or_ send data space. 
		 * (Some poll() Linux documentation says that POLLHUP is 
		 *  incompatible with the POLLOUT/POLLWR flags)
		 */
		if (0 < (TS_SDP_SHUTDOWN_SEND & conn->shutdown)) {

			mask |= POLLHUP;
		}
		else {
			/*
			 * avoid race by setting flags, and only clearing
			 * them if the test is passed. Setting after the 
			 * test, we can end up with them set and a passing
			 * test.
			 */
			set_bit(SOCK_ASYNC_NOSPACE, &sock->flags);
			set_bit(SOCK_NOSPACE, &sock->flags);

			if (0 < __sdp_inet_writable(conn)) {

				mask |= POLLOUT | POLLWRNORM;

				clear_bit(SOCK_ASYNC_NOSPACE, &sock->flags);
				clear_bit(SOCK_NOSPACE, &sock->flags);
			}
		}

		if (0 < conn->rcv_urg_cnt) {

			mask |= POLLPRI;
		}
	}

#ifdef  _TS_SDP_DATA_PATH_DEBUG
	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "SOCK: poll mask <%08x> flags <%08x> <%d:%d:%d>", mask,
		 sock->flags, conn->send_buf, conn->send_qud,
		 __sdp_inet_writable(conn));
#endif

	return mask;
} /* _sdp_inet_poll */

/* ========================================================================= */
/*.._sdp_inet_ioctl -- serivce an ioctl request on a socket */
static int _sdp_inet_ioctl(struct socket *sock,
			      unsigned int cmd,
			      unsigned long arg)
{
	struct sock *sk;
	struct sdp_opt *conn;
	struct sdpc_buff *buff;
	int result = 0;
	int value;

	TS_CHECK_NULL(sock, -EINVAL);
	TS_CHECK_NULL(sock->sk, -EINVAL);

	sk = sock->sk;
	conn = TS_SDP_GET_CONN(sk);

	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "SOCK: <%d> ioctl <%d> <%08x:%04x> <%08x:%04x> <%04x>", cmd,
		 conn->hashent, conn->src_addr, conn->src_port, conn->dst_addr,
		 conn->dst_port, conn->istate);
	/*
	 * check IOCTLs
	 */
	switch (cmd) {
		/*
		 * standard INET IOCTLs
		 */
	case SIOCGSTAMP:
		if (0 == TS_SDP_OS_SK_STAMP(sk).tv_sec) {

			result = -ENOENT;
		}
		else {

			result = copy_to_user((void *)arg,
					      &TS_SDP_OS_SK_STAMP(sk),
					      sizeof(struct timeval));
			if (0 != result) {

				result = -EFAULT;
			}
		}

		break;
		/*
		 * Standard routing IOCTLs
		 */
	case SIOCADDRT:
	case SIOCDELRT:
	case SIOCRTMSG:

		result = ip_rt_ioctl(cmd, (void *)arg);
		break;
		/*
		 * Standard ARP IOCTLs
		 */
	case SIOCDARP:
	case SIOCGARP:
	case SIOCSARP:
#if 0				/* currently not exported by the kernel :( */
		result = arp_ioctl(cmd, (void *)arg);
#else
		result = -ENOIOCTLCMD;
#endif
		break;
		/*
		 * standard INET device IOCTLs
		 */
	case SIOCGIFADDR:
	case SIOCSIFADDR:
	case SIOCGIFBRDADDR:
	case SIOCSIFBRDADDR:
	case SIOCGIFNETMASK:
	case SIOCSIFNETMASK:
	case SIOCGIFDSTADDR:
	case SIOCSIFDSTADDR:
	case SIOCSIFPFLAGS:
	case SIOCGIFPFLAGS:
	case SIOCSIFFLAGS:

		result = devinet_ioctl(cmd, (void *)arg);
		break;
		/*
		 * stadard INET STREAM IOCTLs
		 */
	case SIOCINQ:

		TS_SDP_CONN_LOCK(conn);

		if (TS_SDP_SOCK_ST_LISTEN != conn->istate) {
			/*
			 * TODO need to subtract/add URG (inline vs. OOB)
			 */
			value = conn->byte_strm;
			result = put_user(value, (int *) arg);
		}
		else {

			result = -EINVAL;
		}

		TS_SDP_CONN_UNLOCK(conn);
		break;
	case SIOCOUTQ:

		TS_SDP_CONN_LOCK(conn);

		if (TS_SDP_SOCK_ST_LISTEN != conn->istate) {

			value = conn->send_qud;
			result = put_user(value, (int *) arg);
		}
		else {

			result = -EINVAL;
		}

		TS_SDP_CONN_UNLOCK(conn);
		break;
	case SIOCATMARK:

		TS_SDP_CONN_LOCK(conn);

		value = 0;

		if (0 < conn->rcv_urg_cnt) {

			buff = sdp_buff_q_look_head(&conn->recv_pool);
			if (NULL != buff &&
			    0 < (TS_SDP_BUFF_F_OOB_PRES & buff->flags) &&
			    1 == (buff->tail - buff->data)) {

				value = 1;
			}
		}

		result = put_user(value, (int *) arg);

		TS_SDP_CONN_UNLOCK(conn);
		break;
		/*
		 * In the 2.6 kernel a number of IOCTLs were moved to the 
		 * socket.c level, and are no longer passed to the protocol
		 * family provider
		 */
#ifndef TS_KERNEL_2_6
	case FIOSETOWN:
	case SIOCSPGRP:
		result = get_user(value, (int *)arg);
		if (0 == result) {

			if (current->pid != value &&
			    current->pgrp != -value &&
			    !capable(CAP_NET_ADMIN)) {

				result = -EPERM;
			}
			else {

				sk->proc = value;
			}
		}

		break;
	case FIOGETOWN:
	case SIOCGPGRP:

		result = put_user(sk->proc, (int *)arg);
		break;
#endif
	default:

		result = dev_ioctl(cmd, (void *)arg);
		break;
	}

	return result;
} /* _sdp_inet_ioctl */

/* ========================================================================= */
/*.._sdp_inet_setopt -- set a socket option. */
static int _sdp_inet_setopt(struct socket *sock,
			       int level, 
			       int optname, 
			       char *optval, 
			       int optlen)
{
	struct sock *sk;
	struct sdp_opt *conn;
	int value;
	int result = 0;

	TS_CHECK_NULL(sock, -EINVAL);
	TS_CHECK_NULL(sock->sk, -EINVAL);
	TS_CHECK_NULL(optval, -EINVAL);

	sk = sock->sk;
	conn = TS_SDP_GET_CONN(sk);

	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "SOCK: <%d> set option <%d:%d> <%08x:%04x> <%08x:%04x> <%04x>",
		 conn->hashent, optname, level, conn->src_addr, conn->src_port,
		 conn->dst_addr, conn->dst_port, conn->istate);

	if (SOL_TCP != level && SOL_SDP != level) {
		
		 return 0;
	}

	if (optlen < sizeof(int)) {

		return -EINVAL;
	}

	if (get_user(value, (int *)optval)) {

		return -EFAULT;
	}

	TS_SDP_CONN_LOCK(conn);

	switch (optname) {
	case TCP_NODELAY:

		conn->nodelay = (0 == value) ? 0 : 1;

		if (0 < conn->nodelay) {

			result = sdp_send_flush(conn);
			TS_EXPECT(MOD_LNX_SDP, !(0 > result));
		}

		break;
	case SDP_ZCOPY_THRSH:

		conn->src_zthresh = value;
		conn->snk_zthresh =
		    ((value >
		      (conn->recv_size -
		       TS_SDP_MSG_HDR_SIZE)) ? value : (conn->recv_size -
							TS_SDP_MSG_HDR_SIZE));
		break;
	case SDP_ZCOPY_THRSH_SRC:

		conn->src_zthresh = value;
		break;
	case SDP_ZCOPY_THRSH_SNK:

		conn->snk_zthresh =
		    ((value >
		      (conn->recv_size -
		       TS_SDP_MSG_HDR_SIZE)) ? value : (conn->recv_size -
							TS_SDP_MSG_HDR_SIZE));
		break;
	case SDP_UNBIND:

		result = sdp_inet_port_put(conn);
		break;
	default:

		TS_TRACE(MOD_LNX_SDP, T_VERY_TERSE, TRACE_FLOW_WARN,
			 "SOCK: SETSOCKOPT unimplemented option <%d>.",
			 optname);
		break;
	}

	TS_SDP_CONN_UNLOCK(conn);
	return result;
} /* _sdp_inet_setopt */

/* ========================================================================= */
/*.._sdp_inet_getopt -- get a socket option. */
static int _sdp_inet_getopt(struct socket *sock,
			       int level,
			       int optname,
			       char *optval,
			       int *optlen)
{
	struct sock *sk;
	struct sdp_opt *conn;
	int value;
	int len;

	TS_CHECK_NULL(sock, -EINVAL);
	TS_CHECK_NULL(sock->sk, -EINVAL);
	TS_CHECK_NULL(optval, -EINVAL);
	TS_CHECK_NULL(optlen, -EINVAL);

	sk = sock->sk;
	conn = TS_SDP_GET_CONN(sk);

	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "SOCK: <%d> get option <%d:%d> <%08x:%04x> <%08x:%04x> <%04x>",
		 conn->hashent, optname, level, conn->src_addr, conn->src_port,
		 conn->dst_addr, conn->dst_port, conn->istate);

	if (SOL_TCP != level && SOL_SDP != level) {
		 len=0;
		 if (put_user(len, optlen)) {
		   return -EFAULT;
		 }
		 
		 return 0;
	}

	if (get_user(len, optlen)) {

		return -EFAULT;
	}

	len = min(len, (int)sizeof(int));
	if (len < sizeof(int)) {

		return -EINVAL;
	}

	TS_SDP_CONN_LOCK(conn);

	switch (optname) {
	case TCP_NODELAY:

		value = (1 == conn->nodelay);
		break;
	case SDP_ZCOPY_THRSH:

		value = ((conn->src_zthresh == conn->snk_zthresh) ?
			 conn->snk_zthresh : -EPROTO);
		break;
	case SDP_ZCOPY_THRSH_SRC:

		value = conn->src_zthresh;
		break;
	case SDP_ZCOPY_THRSH_SNK:

		value = conn->snk_zthresh;
		break;
	default:

		TS_TRACE(MOD_LNX_SDP, T_VERY_TERSE, TRACE_FLOW_WARN,
			 "SOCK: GETSOCKOPT unimplemented option <%d>.",
			 optname);
		break;
	}

	TS_SDP_CONN_UNLOCK(conn);

	if (put_user(len, optlen)) {

		return -EFAULT;
	}

	if (copy_to_user(optval, &value, len)) {

		return -EFAULT;
	}

	return 0;
} /* _sdp_inet_getopt */

/* ========================================================================= */
/*.._sdp_inet_shutdown -- shutdown a socket. */
static int _sdp_inet_shutdown(struct socket *sock, int flag)
{
	int result = 0;
	struct sdp_opt *conn;

	TS_CHECK_NULL(sock, -EINVAL);
	TS_CHECK_NULL(sock->sk, -EINVAL);
	conn = TS_SDP_GET_CONN(sock->sk);

	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "SOCK: <%d> shutdown <%d> <%08x:%04x> <%08x:%04x> <%04x>",
		 conn->hashent, flag, conn->src_addr, conn->src_port,
		 conn->dst_addr, conn->dst_port, conn->istate);
	/*
	 * flag: 0 - recv shutdown
	 *       1 - send shutdown
	 *       2 - send/recv shutdown.
	 */
	if (0 > flag || 2 < flag) {

		return -EINVAL;
	}
	else {

		flag++;		/* match shutdown mask. */
	}

	TS_SDP_CONN_LOCK(conn);

	conn->shutdown |= flag;

	switch (conn->istate) {
	case TS_SDP_SOCK_ST_CLOSED:

		result = -ENOTCONN;
		break;
	case TS_SDP_SOCK_ST_LISTEN:
		/*
		 * Send shutdown is benign.
		 */
		if (0 < (TS_SDP_SHUTDOWN_RECV & flag)) {

			result = sdp_inet_listen_stop(conn);
			if (0 > result) {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_WARN,
					 "SOCK: Error <%d> while releasing listen socket <%d>",
					 result, conn->pid);
			}
		}

		break;
	case TS_SDP_SOCK_ST_ERROR:

		result = TS_SDP_CONN_ERROR(conn);
		result = (result < 0) ? result : -ECONNABORTED;
		break;
	case TS_SDP_SOCK_ST_ACCEPTED:
	case TS_SDP_SOCK_ST_CONNECT:

		result = _sdp_inet_abort(conn);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "SOCK: Error <%d> aborting from peer.",
				 result);
		}

		break;
	case TS_SDP_SOCK_ST_ESTABLISHED:
	case TS_SDP_SOCK_ST_CLOSE:
        result = _sdp_inet_disconnect(conn);
        if (0 > result) {
            
            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                         "SOCK: Error <%d> disconnecting from peer.",
				 result);       
         }
         break;
    case TS_SDP_SOCK_ST_DISCONNECT:
	case TS_SDP_SOCK_ST_CLOSING:
		/*
		 * nothing.
		 */
		break;
	case TS_SDP_SOCK_ST_ACCEPTING:

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "SOCK: User shouldn't have a socket in this state. <%d>",
			 conn->istate);

		conn->istate = TS_SDP_SOCK_ST_ERROR;
		TS_SDP_OS_CONN_SET_ERR(conn, EPROTO);
		result = -EFAULT;

		break;
	default:

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
			 "SOCK: Unknown socket state. <%d>", conn->istate);

		conn->istate = TS_SDP_SOCK_ST_ERROR;
		TS_SDP_OS_CONN_SET_ERR(conn, EPROTO);
		result = -EFAULT;
	}

	if (0 > result) {
		sdp_inet_wake_generic(sock->sk);
	}
	else {
		sdp_inet_wake_error(sock->sk);
	}

	TS_SDP_CONN_UNLOCK(conn);
	return result;
} /* _sdp_inet_shutdown */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* Primary socket initialization                                         */
/*                                                                       */
/* --------------------------------------------------------------------- */
struct proto_ops _lnx_stream_ops = {
      family:      TS_SDP_DEV_PROTO,
      release:    _sdp_inet_release,
      bind:       _sdp_inet_bind,
      connect:    _sdp_inet_connect,
      listen:     _sdp_inet_listen,
      accept:     _sdp_inet_accept,
      sendmsg:     sdp_inet_send,
      recvmsg:     sdp_inet_recv,
      getname:    _sdp_inet_getname,
      poll:       _sdp_inet_poll,
      setsockopt: _sdp_inet_setopt,
      getsockopt: _sdp_inet_getopt,
      shutdown:   _sdp_inet_shutdown,
      ioctl:      _sdp_inet_ioctl,
#ifdef _TS_SDP_AIO_SUPPORT
      kvec_read:   sdp_inet_read,
      kvec_write:  sdp_inet_write,
#endif
      /*#ifdef _TS_SDP_SENDPAGE_SUPPORT*/
      sendpage:    sock_no_sendpage,
      /*#endif */
      socketpair:  sock_no_socketpair,
      mmap:        sock_no_mmap,
#ifdef TS_KERNEL_2_6
      owner:       THIS_MODULE,
#endif
}; /* _lnx_stream_ops */


/* ========================================================================= */
/*.._sdp_inet_create -- create a socket */
static int _sdp_inet_create(struct socket *sock, int protocol)
{
	struct sdp_opt *conn;

	TS_CHECK_NULL(sock, -EINVAL);

	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "STRM: create <%d:%d> <%d:%08x>",
		 sock->type, protocol, sock->state, sock->flags);

	if (SOCK_STREAM != sock->type ||
	    (IPPROTO_IP != protocol && IPPROTO_TCP != protocol)) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "CREATE: unsupported socket type/protocol. <%d:%d>",
			 sock->type, protocol);
		return -EPROTONOSUPPORT;
	}

	conn = sdp_conn_alloc(GFP_KERNEL, TS_SDP_CONN_COMM_ID_NULL);
  	if (NULL == conn) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "CREATE: failed to create socket for protocol. <%d:%d>",
			 sock->type, protocol);
		return -ENOMEM;
	}
    conn->use_zcopy = _use_zcopy;
    conn->buffered_bytes=0;
    conn->wait_buff_before_src=0;
    conn->inside_read=0;

   // conn->src_zthresh = SDP_ZCOPY_THRSH_SRC_DEFAULT;
    conn->src_zthresh = _src_zthresh;   //module_param
    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,"zcopy threshold is %d use_zcopy is %d",conn->src_zthresh,conn->use_zcopy);

    conn->snk_zthresh = SDP_ZCOPY_THRSH_SNK_DEFAULT;

 	sock->ops = &_lnx_stream_ops;
	sock->state = SS_UNCONNECTED;
#ifndef TS_KERNEL_2_6
TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,"count inc");
 MOD_INC_USE_COUNT ;
#endif


	sock_graft(conn->sk, sock);

	conn->pid = TS_SDP_OS_GET_PID();
	
#if 0				/* CPU affinity testing... */
#if 1
	current->cpus_allowed = (1 << 1);
#else
	current->cpus_allowed = (1 << 2) | (1 << 3);
#endif
#endif

	return 0;
} /* _sdp_inet_create */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* INET module initialization functions                                  */
/*                                                                       */
/* --------------------------------------------------------------------- */
static struct net_proto_family _sdp_proto = {
      family:   TS_SDP_DEV_PROTO,
      create:  _sdp_inet_create,
#ifdef TS_KERNEL_2_6
      owner:    THIS_MODULE,
#endif
}; /* _pf_family */


/* --------------------------------------------------------------------- */
/*                                                                       */
/* SDP zcopy threshold                                                   */
/*                                                                       */
/* --------------------------------------------------------------------- */
int sdp_zcopy_threshold(void) {
     return _src_zthresh;
 }
int sdp_zcopy_use(void) {
     return _use_zcopy;
 }


/* --------------------------------------------------------------------- */
/*                                                                       */
/* SDP host module load/unload functions                                 */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_init -- initialize the sdp host module */
int __init sdp_init(void)
{
	int result = 0;

	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "INIT: SDP module load.");


   	/*
	 * workaround for some RHEL 3.0 AIO kernels. (see sdp_iocb.h)
	 */
#if defined(_TS_SDP_AIO_SUPPORT) && defined(_TS_FILE_OP_TABLE_ADDR)
	{
		struct file_operations *socket_file_ops;
		socket_file_ops =
		    (struct file_operations *)_TS_FILE_OP_TABLE_ADDR;

		if (NULL != socket_file_ops) {

			socket_file_ops->aio_read = generic_file_aio_read;
			socket_file_ops->aio_write = generic_file_aio_write;
		}
	}
#endif
   	/*
	 * proc entries
	 */
	result = sdp_main_proc_init();
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "INIT: Error <%d> creating proc entries.", result);
		goto error_proc;
	}
	/*
	 * generic table
	 */
	result = sdp_main_desc_init();
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "INIT: Error <%d> initializing generic table <%d>",
			 result);
		goto error_generic;
	}
	/*
	 * advertisment table
	 */
	result = sdp_main_advt_init();
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "INIT: Error <%d> initializing SDP advertisments <%d>",
			 result);
		goto error_advt;
	}
	/*
	 * connection table
	 */
	result = sdp_conn_table_init(_proto_family,
				     _conn_size, 
				     _buff_min,
				     _buff_max,
				     _recv_post_max,
				     _recv_buff_max,
				     _send_post_max,
				     _send_buff_max,
				     _send_usig_max);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "INIT: Error <%d> initializing connection table.",
			 result);
		goto error_conn;
	}
	/*
	 * register
	 */
	_sdp_proto.family = _proto_family;

	result = sock_register(&_sdp_proto);
	if (result < 0) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "INIT: Error <%d> registering protocol family <%d>",
			 result, _sdp_proto.family);
		goto error_sock;
	}

	return 0;
error_sock:
	(void)sdp_conn_table_clear();
error_conn:
	(void)sdp_main_advt_cleanup();
error_advt:
	(void)sdp_main_desc_cleanup();                                        
error_generic:
	(void)sdp_main_proc_cleanup();
error_proc:
	return result;		/* success */
} /* sdp_init */

/* ========================================================================= */
/*..sdp_exit -- cleanup the sdp host module */
void sdp_exit(void)
{
  
	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "INIT: SDP module unload.");
	/*
	 * unregister
	 */
	sock_unregister(_sdp_proto.family);
	/*
	 * connection table
	 */
       	
     (void)sdp_conn_table_clear();
	/*
	 * delete advertisment table
	 */
	(void)sdp_main_advt_cleanup();
	/*
	 * delete generic table
	 */
	(void)sdp_main_desc_cleanup();
	/*
	 * proc tables
	 */
	(void)sdp_main_proc_cleanup();

	return;
} /* sdp_exit */




module_init(sdp_init);
module_exit(sdp_exit);
