
/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sdp_iocb.c 159 2004-05-14 00:16:54Z libor $
*/

#include "sdp_main.h"
#include <linux/pagemap.h>
static kmem_cache_t *_tsSdpConnIocbCache = NULL;
static kmem_cache_t *_tsSdpConnIocbTable = NULL;
static kmem_cache_t *iocb_pages_cache = NULL;





/* --------------------------------------------------------------------- */
/*                                                                       */
/* Zcopy memory registration and pinning                                 */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */

/* ========================================================================= */
/*..sdp_zcopy_release -- unregister an IOCBs memory  */
int sdp_zcopy_release(struct sdpc_iocb *iocb)
{
    int result,counter;



    result=0;
    TS_CHECK_NULL(iocb, -EINVAL);




    if (NULL != iocb->page_array)
    {

        result = ib_fmr_deregister(iocb->mem);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
                     "EVENT: Error <%d> releasing IOCB <%d> memory <%u>",
                     result, iocb->key, iocb->mem);
        }

        kfree(iocb->page_array);
    }
   
    for (counter = 0;counter < iocb->page_count; counter++)
    {
        /* 
	 * Cant do it here since lock_page is blocking.
	 * We shall be fine since zcopy_complete sets
	 * the dirty bit on the actual pages.
	 * sdp_mark_dirty( iocb->pages->pages[counter]);
	 * */
        page_cache_release(iocb->pages->pages[counter]);
    }

    kmem_cache_free(iocb_pages_cache,iocb->pages);
    iocb->page_array = NULL;
    iocb->io_addr = 0;


    if (result<0 )
    {
        TS_TRACE(MOD_LNX_SDP, T_VERY_TERSE, TRACE_FLOW_WARN,
                 "EVENT: Error fmr deregister");
    }

    return 0;
} /* sdp_zcopy_release */

/* ========================================================================= */
/*..sdp_zcopy_create - create an IOCB object */
struct sdpc_iocb *sdp_zcopy_create(void)
{
    struct sdpc_iocb *iocb;

    TS_CHECK_NULL(_tsSdpConnIocbCache, NULL);

    iocb = kmem_cache_alloc(_tsSdpConnIocbCache, SLAB_ATOMIC);
    if (NULL != iocb)
    {                                                                             

        memset(iocb, 0, sizeof(struct sdpc_iocb));
        /*
         * non-zero initialization
         */
        iocb->key = TS_SDP_IOCB_KEY_INVALID;

        iocb->type = TS_SDP_GENERIC_TYPE_ZCOPY;
        iocb->release =
        (tSDP_GENERIC_DESTRUCT_FUNC) sdp_iocb_destroy;
        iocb->aio=0;
        iocb->check=0;
        iocb->page_array=NULL;
        iocb->serial_num=0;
    }

    return iocb;
} /* sdp_iocb_create */








/*..sdp_zcopy_register -- register an IOCBs memory for advertisment */
int sdp_zcopy_register(struct sdpc_iocb *iocb, struct sdp_opt *conn)
{
    int counter;


    int len,result;
    struct ib_fmr_pool *pool;
    TS_CHECK_NULL(iocb, -EINVAL);
    TS_CHECK_NULL(conn, -EINVAL);



    pool=conn->fmr_pool;
    len=iocb->size;


    iocb->pages = (struct iocb_physical_pages*)kmem_cache_alloc(iocb_pages_cache, GFP_KERNEL);
    iocb->page_count= (tUINT32) ((( iocb->base+len-1) >> PAGE_SHIFT) - (iocb->base >>  PAGE_SHIFT) + 1);
   // iocb->page_array = kmalloc((sizeof(tUINT64) * iocb->page_count), GFP_ATOMIC);
    iocb->page_array = kmalloc((sizeof(tUINT64) * iocb->page_count), GFP_KERNEL);
    down_read(&current->mm->mmap_sem);
    if (iocb->dir==1)
    {
        result=get_user_pages(current, current->mm, (unsigned long)iocb->base, iocb->page_count , 0, 0, iocb->pages->pages, NULL);
    } else
    {
        result=get_user_pages(current, current->mm, (unsigned long)iocb->base, iocb->page_count , 1, 0, iocb->pages->pages, NULL);
    }
    up_read(&current->mm->mmap_sem);

    if (result!=    iocb->page_count)
    {
        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,"page count should be  %d but is %d",iocb->page_count,result);
        iocb->page_count=result;
        goto error_register;
    }

    for (counter = 0;counter < iocb->page_count; counter++)
    {
        iocb->page_array[counter] =  page_to_phys(iocb->pages->pages[counter]);
    }
    iocb->page_offset= iocb->base & ~PAGE_MASK;            
    iocb->io_addr = iocb->page_array[0] + iocb->page_offset;


    result = ib_fmr_register_physical(conn->fmr_pool,
                                      (uint64_t *) iocb->page_array,
                                      iocb->page_count,
                                      (uint64_t *) & iocb->io_addr, 
                                      iocb->page_offset,
                                      &iocb->mem,
                                      &iocb->l_key, &iocb->r_key);  
    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
             "POST: fmr_register: pg_count %d page_offset %x ioaddress %llx r_key %x ",
             iocb->page_count, iocb->page_offset, iocb->io_addr, iocb->r_key);


    if (0 != result)
    {
        if (-EAGAIN != result)
        {

            TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
                     "POST: Error <%d> registering physical memory. <%d:%d:%d>",
                     result, iocb->len, iocb->page_count,
                     iocb->page_offset);
        }

        goto error_register;
    }
    /*
     * some data may have already been consumed, adjust the io address
     * to take this into account
     */
    iocb->io_addr += iocb->post;

    return 0;
    error_register:
    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,"error register");
    down_read(&current->mm->mmap_sem);

    for (counter = 0;counter < iocb->page_count; counter++)
    {
        sdp_mark_dirty(iocb->pages->pages[counter]) ;
        page_cache_release(iocb->pages->pages[counter]);
    }       
    up_read(&current->mm->mmap_sem);

    kmem_cache_free(iocb_pages_cache,iocb->pages);
    iocb->io_addr = 0;
    kfree(iocb->page_array);
    iocb->page_array = NULL;

    return result;

} /* sdp_zcopy_register */


/* ========================================================================= */
/*..sdp_zcopy_complete -- complete an IOCB */
int sdp_zcopy_complete(struct sdpc_iocb *iocb, ssize_t status)
{

    int counter,copy=0;
    ssize_t result;
    int page_size=0x1000;
    struct iocb_physical_pages *pages;
    void* addr;
    TS_CHECK_NULL(iocb, -EINVAL);
    caddr_t addr_to_copy;




    if (iocb->check==1)
    {


        down_read(&current->mm->mmap_sem);

        pages = (struct iocb_physical_pages*)kmem_cache_alloc(iocb_pages_cache, GFP_KERNEL);
        result=get_user_pages(current, current->mm, (unsigned long)iocb->base, iocb->page_count , 1, 0, pages->pages, NULL);
        if (result!=iocb->page_count)
        {
            TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,"error while validating pages");
            copy=1;
        }
        for (counter = 0;counter < iocb->page_count; counter++)
        {
            if (iocb->pages->pages[counter] != pages->pages[counter])
            {
                TS_TRACE(MOD_LNX_SDP,T_TERSE, TRACE_FLOW_WARN,"error2 while validating pages");
                copy=1;
            }
        }

        for (counter = 0;counter < iocb->page_count; counter++)
        {
            sdp_mark_dirty(pages->pages[counter]) ;
            page_cache_release(pages->pages[counter]);
        }
        up_read(&current->mm->mmap_sem);

        kmem_cache_free(iocb_pages_cache,pages);


    }
    counter=0;
    if (copy==1)
    {

        while (iocb->post>0)
        {
            if (counter==0)  { //first page
                copy=min(iocb->post,page_size - iocb->page_offset);
            }  else {
                copy=min(iocb->post,page_size);
            }
            addr=kmap_atomic(iocb->pages->pages[counter], KM_USER0);

            // address calculation : base page (base address - offset) + former blocks addresses + former pages
            addr_to_copy=iocb->msg->msg_iov->iov_base - iocb->page_offset + 
                (iocb->serial_num * TS_SDP_IOCB_SIZE_MAX) + (PAGE_SIZE * counter);
            if (counter==0)
            {
               addr_to_copy+= iocb->page_offset;
            }
            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN, "base_addr is %p  copy is %d", addr_to_copy,copy);

            copy_to_user(addr_to_copy,addr,
                         copy);
            kunmap_atomic(addr, KM_USER0);

            iocb->post-=copy;
            counter+=1;

        }

    }



    /*
     * release memory
     */





    result = sdp_zcopy_release(iocb);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: Error <%d> releasing IOCB <%d> memory <%u>",
                 result, iocb->key, iocb->mem);
    }
    /*
* callback to complete IOCB
*/
    result = (0 < iocb->post) ? iocb->post : status;

    /*
     * delete IOCB
     */
    result = sdp_iocb_destroy(iocb);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: Error <%d> deleting IOCB <%d> of status <%d>",
                 result, iocb->key, status);
        goto error;
    }

    return 0;
    error:
    return result;
} /* sdp_zcopy_complete */


/* --------------------------------------------------------------------- */
/*                                                                       */
/* IOCB memory registration functions                                    */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_iocb_register -- register an IOCBs memory for advertisment */
int sdp_iocb_register(struct sdpc_iocb *iocb, struct sdp_opt *conn)
{
#ifdef _TS_SDP_AIO_SUPPORT
    struct kveclet *let;
    int result;
    tINT32 counter;

    TS_CHECK_NULL(iocb, -EINVAL);
    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * create page list.
     */
    iocb->page_count = iocb->cb.vec->nr;
    iocb->page_array =
    kmalloc((sizeof(tUINT64) * iocb->page_count), GFP_ATOMIC);

    if (NULL == iocb->page_array)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to allocate IOCB page array. <%d:%d>",
                 sizeof(tUINT64) * iocb->page_count, iocb->page_count);

        result = -ENOMEM;
        goto error;
    }
    /*
     * iterate of physical page list.
     */
    iocb->page_offset = iocb->cb.vec->veclet->offset;

    for (counter = 0, let = iocb->cb.vec->veclet;
        counter < iocb->page_count; counter++, let++)
    {

        iocb->page_array[counter] =
        TS_SDP_IOCB_PAGE_TO_PHYSICAL(let->page);
    }
    /*
     * prime io address with physical address of first byte?
     */
    iocb->io_addr = iocb->page_array[0] + iocb->page_offset;
    /*
     * register IOCBs physical memory
     */

    result = ib_fmr_register_physical(conn->fmr_pool,
                                      (uint64_t *) iocb->page_array,
                                      iocb->page_count,
                                      (uint64_t *) & iocb->io_addr,
                                      iocb->page_offset,
                                      &iocb->mem,
                                      &iocb->l_key, &iocb->r_key);
    if (0 != result)
    {
        if (-EAGAIN != result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                     "POST: Error <%d> registering physical memory. <%d:%d:%d>",
                     result, iocb->len, iocb->page_count,
                     iocb->page_offset);
        }

        goto error_register;
    }
    /*
     * some data may have already been consumed, adjust the io address
     * to take this into account
     */
    iocb->io_addr += iocb->post;

    return 0;
    error_register:
    kfree(iocb->page_array);
    iocb->page_array = NULL;
    error:
    return result;
#else
    return -EOPNOTSUPP;
#endif
} /* sdp_iocb_register */

/* ========================================================================= */
/*..sdp_iocb_release -- unregister an IOCBs memory  */
int sdp_iocb_release(struct sdpc_iocb *iocb)
{
    int result;

    TS_CHECK_NULL(iocb, -EINVAL);

    if (NULL != iocb->page_array)
    {

        result = ib_fmr_deregister(iocb->mem);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "EVENT: Error <%d> releasing IOCB <%d> memory <%u>",
                     result, iocb->key, iocb->mem);
        }

        kfree(iocb->page_array);
    }

    iocb->page_array = NULL;
    iocb->io_addr = 0;

    return 0;
} /* sdp_iocb_release */

/* ========================================================================= */
/*..sdp_iocb_complete -- complete an IOCB */
int sdp_iocb_complete(struct sdpc_iocb *iocb, ssize_t status)
{
    ssize_t result;

    TS_CHECK_NULL(iocb, -EINVAL);
    /*
     * release memory
     */
    result = sdp_iocb_release(iocb);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: Error <%d> releasing IOCB <%d> memory <%u>",
                 result, iocb->key, iocb->mem);
    }
    /*
     * callback to complete IOCB
     */
    result = (0 < iocb->post) ? iocb->post : status;

#ifdef _TS_SDP_AIO_SUPPORT
    iocb->cb.fn(iocb->cb.data, iocb->cb.vec, result);
#endif
    /*
     * delete IOCB
     */
    result = sdp_iocb_destroy(iocb);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: Error <%d> deleting IOCB <%d> of status <%d>",
                 result, iocb->key, status);
        goto error;
    }

    return 0;
    error:
    return result;
} /* sdp_iocb_complete */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* IOCB object managment                                                 */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_iocb_q_remove - remove the object from the table */
int sdp_iocb_q_remove(struct sdpc_iocb *iocb)
{
    struct sdpc_iocb_q *table;
    struct sdpc_iocb *next;
    struct sdpc_iocb *prev;

    TS_CHECK_NULL(iocb, -EINVAL);
    TS_CHECK_NULL(iocb->next, -EINVAL);
    TS_CHECK_NULL(iocb->prev, -EINVAL);
    TS_CHECK_NULL(iocb->table, -EINVAL);
    TS_CHECK_NULL(iocb->table->head, -EINVAL);

    table = iocb->table;

    if (iocb->next == iocb && iocb->prev == iocb)
    {

        table->head = NULL;
    } else
    {

        next = iocb->next;
        prev = iocb->prev;
        next->prev = prev;
        prev->next = next;

        if (table->head == iocb)
        {

            table->head = next;
        }
    }

    table->size--;

    iocb->table = NULL;
    iocb->next = NULL;
    iocb->prev = NULL;

    return 0;
} /* sdp_iocb_q_remove */

/* ========================================================================= */
/*..sdp_iocb_q_lookup - find an iocb based on key, without removing */
struct sdpc_iocb *sdp_iocb_q_lookup(struct sdpc_iocb_q *table, tUINT32 key)
{
    struct sdpc_iocb *iocb = NULL;
    tINT32 counter;

    TS_CHECK_NULL(table, NULL);

    for (counter = 0, iocb = table->head;
        counter < table->size; counter++, iocb = iocb->next)
    {

        if (iocb->key == key)
        {

            return iocb;
        }
    }

    return NULL;
} /* sdp_iocb_q_lookup */

/* ========================================================================= */
/*..sdp_iocb_create - create an IOCB object */
struct sdpc_iocb *sdp_iocb_create(void)
{
    struct sdpc_iocb *iocb;

    TS_CHECK_NULL(_tsSdpConnIocbCache, NULL);

    iocb = kmem_cache_alloc(_tsSdpConnIocbCache, SLAB_ATOMIC);
    if (NULL != iocb)
    {

        memset(iocb, 0, sizeof(struct sdpc_iocb));
        /*
         * non-zero initialization
         */
        iocb->key = TS_SDP_IOCB_KEY_INVALID;

        iocb->type = TS_SDP_GENERIC_TYPE_IOCB;
        iocb->release =
        (tSDP_GENERIC_DESTRUCT_FUNC) sdp_iocb_destroy;
        iocb->aio=1;
        iocb->check=0;
    }

    return iocb;
} /* sdp_iocb_create */

/* ========================================================================= */
/*..sdp_iocb_destroy - destroy an IOCB object */
int sdp_iocb_destroy(struct sdpc_iocb *iocb)
{
    TS_CHECK_NULL(iocb, -EINVAL);

    if (NULL != iocb->next || NULL != iocb->prev)
    {

        return -EACCES;
    }
    /*
     * page array
     */
    if (NULL != iocb->page_array)
    {

        kfree(iocb->page_array);
    }
    /*
     * clear IOCB to check for usage after free...
     */
#if 0
    memset(iocb, 0, sizeof(struct sdpc_iocb));
#endif
    /*
     * return the object to its cache
     */
    kmem_cache_free(_tsSdpConnIocbCache, iocb);

    return 0;
} /* sdp_iocb_destroy */

/* ========================================================================= */
/*..sdp_iocb_q_look - get, without removing, the object at the head */
struct sdpc_iocb *sdp_iocb_q_look(struct sdpc_iocb_q *table)
{
    TS_CHECK_NULL(table, NULL);

    return table->head;
} /* sdp_iocb_q_look */

/* ========================================================================= */
/*..sdp_iocb_q_get_key - find an iocb based on key, and remove it */
struct sdpc_iocb *sdp_iocb_q_get_key(struct sdpc_iocb_q *table, tUINT32 key)
{
    struct sdpc_iocb *iocb;
    int result;

    iocb = sdp_iocb_q_lookup(table, key);
    if (NULL == iocb)
    {

        goto done;
    }

    result = sdp_iocb_q_remove(iocb);
    if (0 > result)
    {

        iocb = NULL;
    }

    done:
    return iocb;
} /* sdp_iocb_q_get_key */

/* ========================================================================= */
/*.._sdp_iocb_q_get - get, and remove, the object at the tables head */
static struct sdpc_iocb *_sdp_iocb_q_get(struct sdpc_iocb_q *table, tBOOLEAN head)
{
    struct sdpc_iocb *iocb;
    struct sdpc_iocb *next;
    struct sdpc_iocb *prev;

    TS_CHECK_NULL(table, NULL);

    if (NULL == table->head)
    {

        return NULL;
    }

    if (TRUE == head)
    {

        iocb = table->head;
    } else
    {

        iocb = table->head->prev;
    }

    if (iocb->next == iocb && iocb->prev == iocb)
    {

        table->head = NULL;
    } else
    {

        next = iocb->next;
        prev = iocb->prev;
        next->prev = prev;
        prev->next = next;

        table->head = next;
    }

    table->size--;

    iocb->table = NULL;
    iocb->next = NULL;
    iocb->prev = NULL;

    return iocb;
} /* _sdp_iocb_q_get */

/* ========================================================================= */
/*.._sdp_iocb_q_put - put the IOCB object at the tables tail */
int _sdp_iocb_q_put(struct sdpc_iocb_q *table,
                    struct sdpc_iocb *iocb,
                    tBOOLEAN head)
{
    struct sdpc_iocb *next;
    struct sdpc_iocb *prev;

    TS_CHECK_NULL(table, -EINVAL);
    TS_CHECK_NULL(iocb, -EINVAL);

    if (NULL == table->head)
    {

        iocb->next = iocb;
        iocb->prev = iocb;
        table->head = iocb;
    } else
    {

        next = table->head;
        prev = next->prev;

        prev->next = iocb;
        iocb->prev = prev;
        iocb->next = next;
        next->prev = iocb;

        if (TRUE == head)
        {
            table->head = iocb;
        }
    }

    table->size++;

    iocb->table = table;

    return 0;
} /* _sdp_iocb_q_put */

/* ========================================================================= */
/*..sdp_iocb_q_get_tail - get an IOCB object from the tables tail */
struct sdpc_iocb *sdp_iocb_q_get_tail(struct sdpc_iocb_q *table)
{
    return _sdp_iocb_q_get(table, FALSE);
} /* sdp_iocb_q_get_tail */

/* ========================================================================= */
/*..sdp_iocb_q_get_head - get an IOCB object from the tables head */
struct sdpc_iocb *sdp_iocb_q_get_head(struct sdpc_iocb_q *table)
{
    return _sdp_iocb_q_get(table, TRUE);
} /* sdp_iocb_q_get_head */

/* ========================================================================= */
/*..sdp_iocb_q_put_tail - put the IOCB object at the tables tail */
int sdp_iocb_q_put_tail(struct sdpc_iocb_q *table, struct sdpc_iocb *iocb)
{
    return _sdp_iocb_q_put(table, iocb, FALSE);
} /* sdp_iocb_q_put_tail */

/* ========================================================================= */
/*..sdp_iocb_q_put_head - put the IOCB object at the tables head */
int sdp_iocb_q_put_head(struct sdpc_iocb_q *table, struct sdpc_iocb *iocb)
{
    return _sdp_iocb_q_put(table, iocb, TRUE);
} /* sdp_iocb_q_put_head */

/* ========================================================================= */
/*..sdp_iocb_q_cancel -- cancel all outstanding AIOs in a queue */
int sdp_iocb_q_cancel(struct sdpc_iocb_q *table,
                      u32     mask,
                      ssize_t comp)
{
    struct sdpc_iocb *iocb;
    struct sdpc_iocb *next;
    tINT32 counter;
    int result;
    tINT32 total;

    TS_CHECK_NULL(table, -EINVAL);
    /*
     * loop through IOCBs, completing each one with either a partial data
     * result, or a cancelled error.
     */
    for (counter = 0, iocb = table->head, total = table->size;
        counter < total; counter++)
    {
        next = iocb->next;

        if (0 < (mask & iocb->flags) || TS_SDP_IOCB_F_ALL == mask)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                     "IOCB: Cancel <%d> iocb <%d> flags <%04x> of size <%d:%d:%d>",
                     comp, iocb->key, iocb->flags, iocb->size,
                     iocb->post, iocb->len);

            result = sdp_iocb_q_remove(iocb);
            TS_EXPECT(MOD_LNX_SDP, !(0 > result));
            if (iocb->aio==1)
            {
                result = sdp_iocb_complete(iocb, comp);
            } else
            {
                result = sdp_zcopy_complete(iocb, comp);
            };
            TS_EXPECT(MOD_LNX_SDP, !(0 > result));
        }

        iocb = next;
    }

    return 0;
} /* sdp_iocb_q_cancel */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* public table functions                                                */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_iocb_q_create - create an IOCB table */
struct sdpc_iocb_q *sdp_iocb_q_create(tINT32 * result)
{
    struct sdpc_iocb_q *table = NULL;

    TS_CHECK_NULL(result, NULL);
    *result = -EINVAL;
    TS_CHECK_NULL(_tsSdpConnIocbTable, NULL);

    table = kmem_cache_alloc(_tsSdpConnIocbTable, SLAB_ATOMIC);
    if (NULL == table)
    {

        *result = -ENOMEM;
        return NULL;
    }

    table->head = NULL;
    table->size = 0;

    *result = 0;
    return table;
} /* sdp_iocb_q_create */

/* ========================================================================= */
/*..sdp_iocb_q_init - initialize a new empty IOCB table */
int sdp_iocb_q_init(struct sdpc_iocb_q *table)
{
    TS_CHECK_NULL(_tsSdpConnIocbCache, -EINVAL);
    TS_CHECK_NULL(table, -EINVAL);

    table->head = NULL;
    table->size = 0;

    return 0;
} /* sdp_iocb_q_init */

/* ========================================================================= */
/*..sdp_iocb_q_clear - clear the contents of an IOCB table */
int sdp_iocb_q_clear(struct sdpc_iocb_q *table)
{
    struct sdpc_iocb *iocb;
    int result;

    TS_CHECK_NULL(_tsSdpConnIocbCache, -EINVAL);
    TS_CHECK_NULL(table, -EINVAL);
    /*
     * drain the table of any objects
     */
    while (NULL != (iocb = sdp_iocb_q_get_head(table)))
    {

        result = sdp_iocb_destroy(iocb);
        TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    }

    return 0;
} /* sdp_iocb_q_clear */

/* ========================================================================= */
/*..sdp_iocb_q_destroy - destroy an IOCB table */
int sdp_iocb_q_destroy(struct sdpc_iocb_q *table)
{
    int result;

    TS_CHECK_NULL(_tsSdpConnIocbTable, -EINVAL);
    TS_CHECK_NULL(table, -EINVAL);
    /*
     * drain the table of any objects
     */
    result = sdp_iocb_q_clear(table);
    TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    /*
     * return the table to the cache
     */
    kmem_cache_free(_tsSdpConnIocbTable, table);

    return 0;
} /* sdp_iocb_q_destroy */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* primary initialization/cleanup functions                              */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_main_iocb_init -- initialize the advertisment caches. */
tINT32 sdp_main_iocb_init(void)
{
    int result;

    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
             "INIT: IOCB cache initialization.");
    /*
     * initialize kernel mapping space for IRQ mappings.
     */
    __sdp_kmap_init();
    /*
     * initialize the caches only once.
     */
    if (NULL != _tsSdpConnIocbCache || NULL != _tsSdpConnIocbTable)
    {

        TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
                 "INIT: IOCB caches already initialized.");
        return -EINVAL;
    }

    _tsSdpConnIocbCache = kmem_cache_create("SdpIocbCache",
                                            sizeof(struct sdpc_iocb),
                                            0, SLAB_HWCACHE_ALIGN, NULL,
                                            NULL);
    if (NULL == _tsSdpConnIocbCache)
    {

        result = -ENOMEM;
        goto error_iocb_c;
    }

    _tsSdpConnIocbTable = kmem_cache_create("SdpIocbTable",
                                            sizeof(struct sdpc_iocb_q),
                                            0, SLAB_HWCACHE_ALIGN, NULL,
                                            NULL);
    if (NULL == _tsSdpConnIocbTable)
    {

        result = -ENOMEM;
        goto error_iocb_t;
    }
    iocb_pages_cache = kmem_cache_create("SdpIocbPagesCache",
                                         sizeof(struct iocb_physical_pages),
                                         0, SLAB_HWCACHE_ALIGN, NULL,
                                         NULL);
    if (NULL == iocb_pages_cache )
    {

        result = -ENOMEM;
        goto error_iocb_pages_cache;
    }

    return 0;
    error_iocb_t:
    kmem_cache_destroy(_tsSdpConnIocbCache);
    _tsSdpConnIocbCache = NULL;
    error_iocb_c:
    return result;
    error_iocb_pages_cache:
    kmem_cache_destroy(iocb_pages_cache);
    iocb_pages_cache=NULL;
    return result;
} /* sdp_main_iocb_init */

/* ========================================================================= */
/*..sdp_main_iocb_cleanup -- cleanup the advertisment caches. */
tINT32 sdp_main_iocb_cleanup(void)
{
    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
             "INIT: IOCB cache cleanup.");
    /*
     * cleanup the caches
     */
    kmem_cache_destroy(_tsSdpConnIocbCache);
    kmem_cache_destroy(_tsSdpConnIocbTable);
    kmem_cache_destroy(iocb_pages_cache);
    /*
     * null out entries.
     */
    _tsSdpConnIocbCache = NULL;
    _tsSdpConnIocbTable = NULL;
    iocb_pages_cache=NULL;

    return 0;
} /* sdp_main_iocb_cleanup */


void sdp_mark_dirty(struct page *page)
{

    lock_page(page);
    set_page_dirty(page);
    unlock_page(page);
}

