/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  This source code may incorporate intellectual property owned by
  Microsoft Corporation. Our provision of this source code does not
  include any licenses or any other rights to you under any Microsoft
  intellectual property. If you would like a license from Microsoft
  (e.g., to rebrand, redistribute), you need to contact Microsoft
  directly.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sdp_post.c 379 2004-07-08 02:24:32Z roland $
*/

#include "sdp_main.h"

/* --------------------------------------------------------------------- */
/*                                                                       */
/* Connection establishment functions                                    */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*.._sdp_ip2pr_path_complete-- path lookup complete, initiate SDP connection */
static int _sdp_ip2pr_path_complete(tIP2PR_PATH_LOOKUP_ID plid,
				    int status,
				    u32 src_addr,
				    u32 dst_addr,
				    tTS_IB_PORT hw_port,
				    struct ib_device *ca,
				    struct ib_path_record *path,
				    void *arg)
{
	struct ib_cm_active_param active_param;
	struct msg_hello *hello_msg;
	struct sdp_opt *conn = (struct sdp_opt *) arg;
	int result = 0;
	struct sdpc_buff *buff;
	int expect;

	TS_CHECK_NULL(conn, -EINVAL);
	/*
	 * lock the socket
	 */
	TS_SDP_CONN_LOCK(conn);
	/*
	 * path lookup is complete
	 */
	if (plid != conn->plid) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "POST: Path record ID mismatch. <%016llx:%016llx>",
			 plid, conn->plid);
		goto done;
	}

	if (TS_SDP_CONN_ST_REQ_PATH != conn->state) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "POST: Incorrect state <%d> for path record completion.",
			 conn->state);
		goto done;
	}

	conn->plid = TS_IP2PR_PATH_LOOKUP_INVALID;
	TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_REQ_SENT);
	/*
	 * update addresses.
	 */
	conn->src_addr = ntohl(src_addr);
	conn->dst_addr = ntohl(dst_addr);
	/*
	 * create address handle
	 */
	if (0 != status) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: Path record completion with error <%d> status",
			 status);

		goto failed;
	}

	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "POST: <%d> Path record lookup complete <%016llx:%016llx:%d>",
		 conn->hashent, cpu_to_be64(*(tUINT64 *) path->dgid),
		 cpu_to_be64(*(tUINT64 *) (path->dgid + sizeof(tUINT64))),
		 path->dlid);
	/*
	 * allocate IB resources.
	 */
	result = sdp_conn_alloc_ib(conn, ca, hw_port);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "SOCK: Error <%d> allocating IB conn for socket <%d>",
			 result, conn->hashent);
		goto failed;
	}
	/*
	 * create the hello message . (don't need to worry about header
	 * space reservation)
	 */
	buff = sdp_buff_pool_get();
	if (NULL == buff) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "POST: Failed to allocate buffer for Hello Message.");
		goto failed;
	}

	hello_msg = (struct msg_hello *) buff->data;
	buff->tail = buff->data + sizeof(struct msg_hello);

	memset(hello_msg, 0, sizeof(struct msg_hello));

	conn->l_advt_bf = conn->recv_cq_size;
	conn->l_max_adv = TS_SDP_MSG_MAX_ADVS;

	hello_msg->bsdh.recv_bufs = conn->l_advt_bf;
	hello_msg->bsdh.flags     = TS_SDP_MSG_FLAG_NON_FLAG;
	hello_msg->bsdh.mid       = TS_SDP_MSG_MID_HELLO;
	hello_msg->bsdh.size      = sizeof(struct msg_hello);
	hello_msg->bsdh.seq_num   = conn->send_seq;
	hello_msg->bsdh.seq_ack   = conn->advt_seq;

	hello_msg->hh.max_adv       = conn->l_max_adv;
	hello_msg->hh.ip_ver        = TS_SDP_MSG_IPVER;
	hello_msg->hh.version       = TS_SDP_MSG_VERSION;
	hello_msg->hh.r_rcv_size    = conn->recv_size;
	hello_msg->hh.l_rcv_size    = conn->recv_size;
	hello_msg->hh.port          = conn->src_port;
	hello_msg->hh.src.ipv4.addr = conn->src_addr;
	hello_msg->hh.dst.ipv4.addr = conn->dst_addr;

	memcpy(conn->d_gid, path->dgid, sizeof(conn->d_gid));
	conn->d_lid = path->dlid;
	conn->s_lid = path->slid;
	/*
	 * endian swap
	 */
	result = _sdp_msg_cpu_to_net_bsdh(&hello_msg->bsdh);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	result = _sdp_msg_cpu_to_net_hh(&hello_msg->hh);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	/*
	 * save message
	 */
	result = sdp_buff_q_put(&conn->send_post, buff);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: Error <%d> buffering hello packet.", result);

		expect = sdp_buff_pool_put(buff);
		TS_EXPECT(MOD_LNX_SDP, !(0 > expect));

		goto failed;
	}
#if 1
	/*
	 * Mellanox performance bug workaround.
	 */
	if (IB_MTU_1024 < path->mtu) {

		path->mtu = IB_MTU_1024;
	}
#endif
	/*
	 * set QP/CM parameters.
	 */
	active_param.qp                   = conn->qp;
	active_param.req_private_data     = (void *)hello_msg;
	active_param.req_private_data_len = (int)(buff->tail - buff->data);
	active_param.responder_resources  = 4;
	active_param.initiator_depth      = 4;
	active_param.retry_count          = TS_SDP_CM_PARAM_RETRY;
	active_param.rnr_retry_count      = TS_SDP_CM_PARAM_RNR_RETRY;
	active_param.cm_response_timeout  = 20;	/* 4 seconds */
	active_param.max_cm_retries       = 15;
	active_param.flow_control         = 1;

	/* XXX set timeout to default value of 14 */
	path->packet_life = 13;

	/*
	 * initiate connection
	 */
	result = ib_cm_connect(&active_param,
			       path,
			       NULL,
			       TS_SDP_MSG_PORT_TO_SID(conn->dst_port),
			       0,
			       sdp_cm_event_handler,
			       (void *) (unsigned long)conn->hashent,
			       &conn->comm_id);
	if (0 != result) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: Connection manager connect attempt failed. <%d>",
			 result);
		goto failed;
	}

	result = 0;
	goto done;
failed:

	result = sdp_wall_recv_reject(conn, (0 - status));
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: error while failing conn to gateway. <%d>",
			 result);

		expect = sdp_wall_recv_drop(conn);
		TS_EXPECT(MOD_LNX_SDP, !(0 > expect));
	}
	/* if */
done:
	TS_SDP_CONN_UNLOCK(conn);
	TS_SDP_CONN_PUT(conn);

	return 0;
} /* _sdp_sa_path_complete */

/* ========================================================================= */
/*..sdp_cm_connect -- initiate a SDP connection with a hello message. */
int sdp_cm_connect(struct sdp_opt *conn)
{
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	/*
	 * get the buffer size we'll use for this connection. (and all others)
	 */
	if (sizeof(struct msg_hello) > conn->recv_size) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: buffer size <%d> too small. <%d>",
			 conn->recv_size, sizeof(struct msg_hello));
		result = -ENOBUFS;
		goto error;
	}

	TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_REQ_PATH);
	/*
	 * lookup the remote address
	 */
	TS_SDP_CONN_HOLD(conn);
	TS_SDP_CONN_UNLOCK(conn);

	result = tsIp2prPathRecordLookup(htonl(conn->dst_addr),
					 htonl(conn->src_addr),
					 TS_SDP_OS_SK_LOCALROUTE(conn->sk),
					 TS_SDP_OS_SK_BOUND_IF(conn->sk),
					 _sdp_ip2pr_path_complete,
					 conn, &conn->plid);
	TS_SDP_CONN_LOCK(conn);

	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "POST: Error <%d> getting link <%08x:%08x> address.",
			 result, htonl(conn->dst_addr), htonl(conn->src_addr));
		/*
		 * callback dosn't have this socket.
		 */
        if (atomic_read(&conn->refcnt)==1)
        {
            TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
                     "connection reference count is one where must be at least 2 about to be decremented inside a lock");
        }

		TS_SDP_CONN_PUT(conn);

		result = -EDESTADDRREQ;
		goto error;
	}

	return 0;
error:
	return result;
} /* sdp_cm_connect */

/* ========================================================================= */
/*..sdp_cm_accept -- respond to a connection attempt with an ack */
int sdp_cm_accept(struct sdp_opt *conn)
{
	struct ib_cm_passive_param passive_param;
	struct msg_hello_ack *hello_ack;
	struct sdpc_buff *buff;
	int result;
	int expect;

	TS_CHECK_NULL(conn, -EINVAL);
	/*
	 * build listen response headers
	 */
	if (sizeof(struct msg_hello_ack) > conn->recv_size) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: buffer size <%d> too small. <%d>",
			 conn->recv_size, sizeof(struct msg_hello_ack));
		result = -ENOBUFS;
		goto error;
	}
	/*
	 * get a buffer, in which we will create the hello header ack.
	 * (don't need to worry about header space reservation on sends)
	 */
	buff = sdp_buff_pool_get();
	if (NULL == buff) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: Failed to allocaate buffer for Hello Ack Header.");
		result = -ENOMEM;
		goto error;
	}

	hello_ack = (struct msg_hello_ack *)buff->data;
	buff->tail = buff->data + sizeof(struct msg_hello_ack);
	/*
	 * create the message
	 */
	memset(hello_ack, 0, sizeof(struct msg_hello_ack));

	conn->l_advt_bf = conn->recv_cq_size;
	conn->l_max_adv = TS_SDP_MSG_MAX_ADVS;

	hello_ack->bsdh.recv_bufs = conn->l_advt_bf;
	hello_ack->bsdh.flags     = TS_SDP_MSG_FLAG_NON_FLAG;
	hello_ack->bsdh.mid       = TS_SDP_MSG_MID_HELLO_ACK;
	hello_ack->bsdh.size      = sizeof(struct msg_hello_ack);
	hello_ack->bsdh.seq_num   = conn->send_seq;
	hello_ack->bsdh.seq_ack   = conn->advt_seq;

	hello_ack->hah.max_adv    = conn->l_max_adv;
	hello_ack->hah.version    = TS_SDP_MSG_VERSION;
	hello_ack->hah.l_rcv_size = conn->recv_size;
	/*
	 * endian swap
	 */
	result = _sdp_msg_cpu_to_net_bsdh(&hello_ack->bsdh);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	result = _sdp_msg_cpu_to_net_hah(&hello_ack->hah);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	/*
	 * save message
	 */
	result = sdp_buff_q_put(&conn->send_post, buff);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: Error <%d> buffering hello ack packet.",
			 result);

		expect = sdp_buff_pool_put(buff);
		TS_EXPECT(MOD_LNX_SDP, !(0 > expect));

		goto error;
	}
	/*
	 * Post receive buffers for this connection
	 */
	result = sdp_recv_flush(conn);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: Error posting <%d> receive buffers. <%d>",
			 result, conn->recv_cq_size);
		goto error;
	}
	/*
	 * send REP message to remote CM to continue connection.
	 */
	passive_param.qp                     = conn->qp;
	passive_param.reply_private_data     = hello_ack;
	passive_param.reply_private_data_len = buff->tail - buff->data;
	passive_param.responder_resources    = 4;
	passive_param.initiator_depth        = 4;
	passive_param.rnr_retry_count        = TS_SDP_CM_PARAM_RNR_RETRY;
	passive_param.flow_control           = 1;
	passive_param.failover_accepted      = 0;

	result = ib_cm_accept(conn->comm_id, &passive_param);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: Connection manager accept request failed <%d>",
			 result);

		goto error;
	}

	return 0;
error:
	return result;
} /* sdp_cm_accept */

/* ========================================================================= */
/*..sdp_cm_listen_start --  start listening on all possible socket ports */
int sdp_cm_listen_start(struct sdev_root *dev_root)
{
	int result = 0;

	TS_CHECK_NULL(dev_root, -EINVAL);
	/*
	 * start listening
	 */
	result = ib_cm_listen(TS_SDP_MSG_SERVICE_ID_VALUE,
			      TS_SDP_MSG_SERVICE_ID_MASK,
			      sdp_cm_event_handler,
			      (void *)TS_SDP_DEV_SK_INVALID,
			      &dev_root->listen_handle);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "POST: Failed to start listening for SDP connections. <%d>",
			 result);
	}
	else {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "POST: Started listening for incomming SDP connections");
	}

	return result;
} /* sdp_cm_listen_start */

/* ========================================================================= */
/*..sdp_cm_listen_stop --  stop listening on all possible socket ports */
int sdp_cm_listen_stop(struct sdev_root *dev_root)
{
	int result = 0;

	TS_CHECK_NULL(dev_root, -EINVAL);

	result = ib_cm_listen_stop(dev_root->listen_handle);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "POST: Failed to stop listening for SDP connections. <%d>",
			 result);
	}
	else {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "POST: Stopped listening for incomming SDP connections");
	}

	return result;
} /* sdp_cm_listen_stop */

/* ========================================================================= */
/*.._sdp_cm_disconnect -- initiate a disconnect request using the CM */
static void _sdp_cm_disconnect(void *arg)
{
	struct sdp_opt *conn = (struct sdp_opt *)arg;
	int result;

	if (NULL == conn) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "POST: Error, posting disconnect for NULL conn");
		return;
	}

	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "POST: <%d> Executing defered disconnect for conn. <%08x> <%04x>",
		 conn->hashent, conn->comm_id, conn->state);
	/*
	 * send a disconnect request using the connection manager
	 */
	result = ib_cm_disconnect(conn->comm_id);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "POST: Error <%d> posting disconnect conn <%d> state <%04x> ",
			 result, conn->comm_id, conn->state);
	}
    if (atomic_read(&conn->refcnt)==1)
    {
        TS_TRACE(MOD_LNX_SDP, T_VERY_TERSE, TRACE_FLOW_WARN,
                 "connection reference count is one where must be at least 2( about to be decremented inside a lock)");
    }        

	TS_SDP_CONN_PUT(conn);

	return;
} /* _sdp_cm_disconnect */

/* ========================================================================= */
/*.._sdp_cm_reject -- initiate a reject request using the CM */
static void _sdp_cm_reject(void *arg)
{
	struct sdp_opt *conn = (struct sdp_opt *)arg;
	int result;

	if (NULL == conn) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "POST: Error, posting reject for NULL conn");
	}

	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "POST: <%d> Executing defered reject for conn. <%08x> <%04x>",
		 conn->hashent, conn->comm_id, conn->state);
	/*
	 * send a reject request using the connection manager
	 */
	result = ib_cm_reject(conn->comm_id, NULL, 0);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "POST: Error <%d> posting reject for conn <%d>",
			 result, conn->comm_id);
	}
    if (atomic_read(&conn->refcnt)==1)
           {
               TS_TRACE(MOD_LNX_SDP, T_VERY_TERSE, TRACE_FLOW_WARN,
                        "connection reference count is one where must be at least 2( about to be decremented inside a lock)");
           }

	TS_SDP_CONN_PUT(conn);

	return;
} /* _sdp_cm_reject */

/* ========================================================================= */
/*.._sdp_cm_confirm -- initiate a confirm request using the CM */
static void _sdp_cm_confirm(void *arg)
{
	struct sdp_opt *conn = (struct sdp_opt *)arg;
	int result;

	if (NULL == conn) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "POST: Error, posting confirm for NULL conn");
	}

	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "POST: <%d> Executing defered confirm for conn. <%08x> <%04x>",
		 conn->hashent, conn->comm_id, conn->state);
	/*
	 * send a confirm request using the connection manager
	 */
	result = ib_cm_confirm(conn->comm_id, NULL, 0);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "POST: Error <%d> posting confirm for conn <%d>",
			 result, conn->comm_id);
	}
    if (atomic_read(&conn->refcnt)==1)
           {
               TS_TRACE(MOD_LNX_SDP, T_VERY_TERSE, TRACE_FLOW_WARN,
                        "connection reference count is one where must be at least 2( about to be decremented inside a lock)");
           }

	TS_SDP_CONN_PUT(conn);

	return;
} /* _sdp_cm_confirm */

/* ========================================================================= */
/*.._sdp_cm_failed -- initiate a Failed request using the CM */
static void _sdp_cm_failed(void *arg)
{
	struct sdp_opt *conn = (struct sdp_opt *)arg;
	int result;

	if (NULL == conn) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "POST: Error, posting failed for NULL conn");
	}

	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "POST: <%d> Executing defered failed for conn. <%08x> <%04x>",
		 conn->hashent, conn->comm_id, conn->state);
	/*
	 * send a failed request using the connection manager
	 */
	result = ib_cm_reject(conn->comm_id, NULL, 0);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "POST: Error <%d> posting failed for conn <%d>",
			 result, conn->comm_id);
	}
    if (atomic_read(&conn->refcnt)==1)
           {
               TS_TRACE(MOD_LNX_SDP, T_VERY_TERSE, TRACE_FLOW_WARN,
                        "connection reference count is one where must be at least 2( about to be decremented inside a lock)");
           }

	TS_SDP_CONN_PUT(conn);

	return;
} /* _sdp_cm_failed */

/* ========================================================================= */
/*.._sdp_cm_deferred_generic -- initiate a defered request using the*/
static int _sdp_cm_deferred_generic(struct sdp_opt *conn,
				    tTS_KERNEL_TIMER_FUNCTION func)
{
	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(func, -EINVAL);
	/*
	 * send a potentially defered failed request.
	 */
	TS_SDP_CONN_HOLD(conn);

	if (in_atomic() || irqs_disabled()) {
			conn->cm_exec.function = func;
			conn->cm_exec.run_time = jiffies;
			conn->cm_exec.arg = conn;
			tsKernelTimerAdd(&conn->cm_exec);
	}
	else {

		func(conn);
	}

	return 0;
} /* _sdp_cm_deferred_generic */

/* ========================================================================= */
/*..sdp_cm_disconnect -- initiate a disconnect request using the CM */
int sdp_cm_disconnect(struct sdp_opt *conn)
{
	return _sdp_cm_deferred_generic(conn, _sdp_cm_disconnect);
} /* sdp_cm_disconnect */

/* ========================================================================= */
/*..sdp_cm_reject -- initiate a reject request using the CM */
int sdp_cm_reject(struct sdp_opt *conn)
{
	return _sdp_cm_deferred_generic(conn, _sdp_cm_reject);
} /* sdp_cm_reject */

/* ========================================================================= */
/*..sdp_cm_confirm -- initiate a confirm request using the CM */
int sdp_cm_confirm(struct sdp_opt *conn)
{
	return _sdp_cm_deferred_generic(conn, _sdp_cm_confirm);
} /* sdp_cm_confirm */

/* ========================================================================= */
/*..sdp_cm_failed -- initiate a failed request using the CM */
int sdp_cm_failed(struct sdp_opt *conn)
{
	return _sdp_cm_deferred_generic(conn, _sdp_cm_failed);
} /* sdp_cm_failed */
