/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sdp_queue.c 159 2004-05-14 00:16:54Z libor $
*/

#include "sdp_main.h"

static kmem_cache_t *_tsSdpGenericTable = NULL;

/* --------------------------------------------------------------------- */
/*                                                                       */
/* module specific functions                                             */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*.._sdp_desc_q_get - Get an element from a specific table */
static struct sdpc_desc *_sdp_desc_q_get(struct sdpc_desc_q *table,
					  tBOOLEAN fifo)
{
	struct sdpc_desc *element;


	TS_CHECK_NULL(table, NULL);

	if (NULL == table->head) {

		return NULL;
	}

	if (TRUE == fifo) {

		element = table->head;
	}
	else {

		element = table->head->prev;
	}

	if (element->next == element && element->prev == element) {

		table->head = NULL;
	}
	else {
      
		element->next->prev = element->prev;
		element->prev->next = element->next;

		table->head = element->next;
	}

	table->size--;
	table->count[element->type] -=
	    ((TS_SDP_GENERIC_TYPE_NONE > element->type) ? 1 : 0);

	element->next = NULL;
	element->prev = NULL;
	element->table = NULL;

	return element;
} /* _sdp_desc_q_get */

/* ========================================================================= */
/*.._sdp_desc_q_put - Place an element into a specific table */
static __inline__ int _sdp_desc_q_put(struct sdpc_desc_q *table,
					       struct sdpc_desc *element, 
					       tBOOLEAN fifo)
{
	/*
	 * fifo: false == tail, true == head
	 */
	TS_CHECK_NULL(table, -EINVAL);
	TS_CHECK_NULL(element, -EINVAL);

	if (NULL != element->table) {

		return -EINVAL;
	}

	if (NULL == table->head) {

		element->next = element;
		element->prev = element;
		table->head = element;
      	}
	else {

		element->next = table->head;
		element->prev = table->head->prev;

		element->next->prev = element;
		element->prev->next = element;

		if (TRUE == fifo) {
			table->head = element;
		}
	}

	table->size++;
	table->count[element->type] +=
	    ((TS_SDP_GENERIC_TYPE_NONE > element->type) ? 1 : 0);
	element->table = table;

	return 0;
} /* _sdp_desc_q_put */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* public advertisment object functions for FIFO object table            */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_desc_q_remove - remove a specific element from a table */
int sdp_desc_q_remove(struct sdpc_desc *element)
{
	struct sdpc_desc_q *table;
	struct sdpc_desc *prev;
	struct sdpc_desc *next;

	TS_CHECK_NULL(element, -EINVAL);
	TS_CHECK_NULL(element->table, -EINVAL);
	TS_CHECK_NULL(element->next, -EINVAL);
	TS_CHECK_NULL(element->prev, -EINVAL);

	table = element->table;

	if (element->next == element && element->prev == element) {

		table->head = NULL;
	}
	else {

		next = element->next;
		prev = element->prev;
		next->prev = prev;
		prev->next = next;

		if (table->head == element) {

			table->head = next;
		}
	}

	table->size--;
	table->count[element->type] -=
	    ((TS_SDP_GENERIC_TYPE_NONE > element->type) ? 1 : 0);

	element->table = NULL;
	element->next = NULL;
	element->prev = NULL;

	return 0;
} /* sdp_desc_q_remove */

/* ========================================================================= */
/*..sdp_desc_q_lookup - search and return an element from the table */
struct sdpc_desc *sdp_desc_q_lookup(struct sdpc_desc_q *table, 
				     tSDP_GENERIC_LOOKUP_FUNC lookup_func,
				     void *arg)
{
	struct sdpc_desc *element;
	tINT32 counter;

	TS_CHECK_NULL(table, NULL);
	TS_CHECK_NULL(lookup_func, NULL);

	for (counter = 0, element = table->head;
	     counter < table->size; counter++, element = element->next) {

		if (0 == lookup_func(element, arg)) {

			return element;
		}
	}

	return NULL;
} /* sdp_desc_q_lookup */

/* ========================================================================= */
/*..sdp_desc_q_get_all - Get the element at the front of the table */
struct sdpc_desc *sdp_desc_q_get_all(struct sdpc_desc_q *table)
{
	struct sdpc_desc *head;

	head = table->head;

	table->head = NULL;
	table->size = 0;

	memset(table->count, 0, sizeof(table->count));

	return head;
} /* sdp_desc_q_get_all */

/* ========================================================================= */
/*..sdp_desc_q_get_head - Get the element at the front of the table */
struct sdpc_desc *sdp_desc_q_get_head(struct sdpc_desc_q *table)
{
 

	return _sdp_desc_q_get(table, TRUE);
} /* sdp_desc_q_get_head */

/* ========================================================================= */
/*..sdp_desc_q_get_tail - Get the element at the end of the table */
struct sdpc_desc *sdp_desc_q_get_tail(struct sdpc_desc_q *table)
{
	return _sdp_desc_q_get(table, FALSE);
} /* sdp_desc_q_get_tail */

/* ========================================================================= */
/*..sdp_desc_q_put_head - Place an element into the head of a table */
int sdp_desc_q_put_head(struct sdpc_desc_q *table,
				struct sdpc_desc *element)
{
	return _sdp_desc_q_put(table, element, TRUE);
} /* sdp_desc_q_put_head */

/* ========================================================================= */
/*..sdp_desc_q_put_tail - Place an element into the tail of a table */
int sdp_desc_q_put_tail(struct sdpc_desc_q *table,
				struct sdpc_desc *element)
{
	return _sdp_desc_q_put(table, element, FALSE);
} /* sdp_desc_q_put_tail */

/* ========================================================================= */
/*..sdp_desc_q_look_head - look at the front of the table */
struct sdpc_desc *sdp_desc_q_look_head(struct sdpc_desc_q *table)
{
	TS_CHECK_NULL(table, NULL);

	return table->head;
} /* sdp_desc_q_look_head */

/* ========================================================================= */
/*..sdp_desc_q_look_tail - look at the end of the table */
struct sdpc_desc *sdp_desc_q_look_tail(struct sdpc_desc_q *table)
{
	TS_CHECK_NULL(table, NULL);

	return ((NULL == table->head) ? NULL : table->head->prev);
} /* sdp_desc_q_look_tail */

/* ========================================================================= */
/*..sdp_desc_q_type_head - look at the type at the front of the table */
int sdp_desc_q_type_head(struct sdpc_desc_q *table)
{
	TS_CHECK_NULL(table, -EINVAL);

	if (NULL == table->head) {

		return TS_SDP_GENERIC_TYPE_NONE;
	}
	else {

		return table->head->type;
	}
} /* sdp_desc_q_type_head */

/* ========================================================================= */
/*..sdp_desc_q_type_tail - look at the type at the end of the table */
int sdp_desc_q_type_tail(struct sdpc_desc_q *table)
{
	TS_CHECK_NULL(table, -EINVAL);

	if (NULL == table->head) {

		return TS_SDP_GENERIC_TYPE_NONE;
	}
	else {

		return table->head->prev->type;
	}
} /* sdp_desc_q_type_tail */

/* ========================================================================= */
/*..sdp_desc_q_look_type_head - look at a specific object */
struct sdpc_desc *sdp_desc_q_look_type_head(struct sdpc_desc_q *table, 
					   tSDP_GENERIC_TYPE type)
{
	TS_CHECK_NULL(table, NULL);

	if (NULL == table->head) {

		return NULL;
	}
	else {

		return ((type == table->head->type) ? table->head : NULL);
	}
} /* sdp_desc_q_look_type_head */

/* ========================================================================= */
/*..sdp_desc_q_look_type_tail - look at the type at the end of the table */
struct sdpc_desc *sdp_desc_q_look_type_tail(struct sdpc_desc_q *table,
					   tSDP_GENERIC_TYPE type)
{
	TS_CHECK_NULL(table, NULL);

	if (NULL == table->head) {

		return NULL;
	}
	else {

		return ((type ==
			 table->head->prev->type) ? table->head->prev : NULL);
	}
} /* sdp_desc_q_look_type_tail */

/* ========================================================================= */
/*..sdp_desc_q_types_size - return the number of elements in the table */
int sdp_desc_q_types_size(struct sdpc_desc_q *table,
				 tSDP_GENERIC_TYPE type)
{
	TS_CHECK_NULL(table, -EINVAL);

	return ((TS_SDP_GENERIC_TYPE_NONE >
		 type) ? table->count[type] : -ERANGE);
} /* sdp_desc_q_types_size */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* public table functions                                                */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_desc_q_create - create/allocate a generic table */
struct sdpc_desc_q *sdp_desc_q_create(tINT32 *result)
{
	struct sdpc_desc_q *table = NULL;

	TS_CHECK_NULL(result, NULL);
	*result = -EINVAL;
	TS_CHECK_NULL(_tsSdpGenericTable, NULL);

	table = kmem_cache_alloc(_tsSdpGenericTable, SLAB_ATOMIC);
	if (NULL == table) {

		*result = -ENOMEM;
		return NULL;
	}

	table->head = NULL;
	table->size = 0;

	memset(table, 0, sizeof(struct sdpc_desc_q));

	*result = 0;
	return table;
} /* sdp_desc_q_create */

/* ========================================================================= */
/*..sdp_desc_q_init - initialize a new empty generic table */
int sdp_desc_q_init(struct sdpc_desc_q *table)
{
	TS_CHECK_NULL(table, -EINVAL);

	table->head = NULL;
	table->size = 0;

	memset(table, 0, sizeof(struct sdpc_desc_q));

	return 0;
} /* sdp_desc_q_init */

/* ========================================================================= */
/*..sdp_desc_q_clear - clear the contents of a generic table */
int sdp_desc_q_clear(struct sdpc_desc_q *table)
{
	tSDP_GENERIC_DESTRUCT_FUNC release_func;
	struct sdpc_desc *element;
	int result;

	TS_CHECK_NULL(table, -EINVAL);
	/*
	 * drain the table of any objects
	 */
	while (NULL != (element = sdp_desc_q_get_head(table))) {

		release_func = element->release;
        if (element->type== TS_SDP_GENERIC_TYPE_ZCOPY) {
                sdp_zcopy_complete ((struct sdpc_iocb *) element,0);
                   
        }
		if (NULL != release_func) {

			result = release_func(element);
			TS_EXPECT(MOD_LNX_SDP, !(0 > result));
		}
	}

	return 0;
} /* sdp_desc_q_clear */

/* ========================================================================= */
/*..sdp_desc_q_destroy - destroy a generic table */
int sdp_desc_q_destroy(struct sdpc_desc_q *table)
{
	int result;

	TS_CHECK_NULL(table, -EINVAL);
	/*
	 * drain the table of any objects
	 */
	result = sdp_desc_q_clear(table);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));
	/*
	 * return the table to the cache
	 */
	kmem_cache_free(_tsSdpGenericTable, table);

	return 0;
} /* sdp_desc_q_destroy */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* primary initialization/cleanup functions                              */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_main_desc_init -- initialize the generic table caches. */
tINT32 sdp_main_desc_init(void)
{
	int result;

	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "INIT: Generic table cache initialization.");
	/*
	 * initialize the caches only once.
	 */
	if (NULL != _tsSdpGenericTable) {

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_FATAL,
			 "INIT: Generic table caches already initialized.");
		return -EINVAL;
	}

	_tsSdpGenericTable = kmem_cache_create("SdpGenericTable",
					       sizeof
					       (struct sdpc_desc_q), 0,
					       SLAB_HWCACHE_ALIGN, NULL, NULL);
	if (NULL == _tsSdpGenericTable) {

		result = -ENOMEM;
		goto error_advt_t;
	}

	return 0;
error_advt_t:
	return 0;
} /* sdp_main_desc_init */

/* ========================================================================= */
/*..sdp_main_desc_cleanup -- cleanup the generic table caches. */
tINT32 sdp_main_desc_cleanup(void)
{
	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
		 "INIT: Generic table cache cleanup.");
	/*
	 * cleanup the caches
	 */
	kmem_cache_destroy(_tsSdpGenericTable);
	/*
	 * null out entries.
	 */
	_tsSdpGenericTable = NULL;

	return 0;
} /* sdp_main_desc_cleanup */
