/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  This source code may incorporate intellectual property owned by
  Microsoft Corporation. Our provision of this source code does not
  include any licenses or any other rights to you under any Microsoft
  intellectual property. If you would like a license from Microsoft
  (e.g., to rebrand, redistribute), you need to contact Microsoft
  directly.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sdp_read.c 379 2004-07-08 02:24:32Z roland $
*/

#include "sdp_main.h"

/* --------------------------------------------------------------------- */
/*                                                                       */
/* RDMA read processing functions                                        */
/*                                                                       */
/* --------------------------------------------------------------------- */

/* ========================================================================= */
/*.._sdp_event_read_advt -- RDMA read event handler for source advertisments. */
static tINT32
_sdp_event_read_advt(struct sdp_opt *conn, struct ib_cq_entry *comp)
{
	struct sdpc_advt *advt;
	tINT32 result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(comp, -EINVAL);
	/*
	 * if this was the last RDMA read for an advertisment, post a notice.
	 * Might want to post multiple RDMA read completion messages per
	 * advertisment, to open up a sending window? Have to test to see
	 * what MS does... (Either choice is correct)
	 */
	advt = sdp_advt_q_look(&conn->src_actv);
	if (NULL != advt && advt->wrid == comp->work_request_id) {

		advt = sdp_advt_q_get(&conn->src_actv);
		TS_EXPECT(MOD_LNX_SDP, (NULL != advt));

		conn->src_recv--;
		result = sdp_send_ctrl_rdma_rd(conn, advt->post);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));


        
		result = sdp_advt_destroy(advt);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));
		/*
		 * If a SrcAvailCancel was received, and all RDMA reads
		 * have been flushed, perform tail processing
		 */
		if (0 < (TS_SDP_CONN_F_SRC_CANCEL_R & conn->flags) &&
		    0 == conn->src_recv) {

			conn->flags &= ~TS_SDP_CONN_F_SRC_CANCEL_R;
			conn->advt_seq = conn->recv_seq;
			/*
			 * If any data was canceled, post a SendSm, also
			 */
			if (0 < (TS_SDP_CONN_F_SRC_CANCEL_C & conn->flags)) {

				result = sdp_send_ctrl_send_sm(conn);
				if (0 > result) {

					TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
						 TRACE_FLOW_WARN,
						 "EVENT: Error posting SendSm in RDMA read. <%d>",
						 result);
					goto error;
				}

				conn->flags &= ~TS_SDP_CONN_F_SRC_CANCEL_C;
			}
		}
	}
	else {

		advt = sdp_advt_q_look(&conn->src_pend);
		if (NULL != advt && advt->wrid == comp->work_request_id) {

			advt->flag &= ~TS_SDP_ADVT_F_READ;
		}
	}

	return 0;
error:
	return result;
} /* _sdp_event_read_advt */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* RDMA read QP Event Handler                                            */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_event_read -- RDMA read event handler. */
tINT32 sdp_event_read(struct sdp_opt *conn, struct ib_cq_entry *comp)
{
	struct sdpc_iocb *iocb;
	struct sdpc_buff *buff;
	tINT32 result=0;
	tINT32 type;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(comp, -EINVAL);
	/*
	 * error handling
	 */
	if (IB_COMPLETION_STATUS_SUCCESS != comp->status) {

		switch (comp->status) {
		case IB_COMPLETION_STATUS_WORK_REQUEST_FLUSHED_ERROR:
			/*
			 * clear posted buffers from error'd queue
			 */
			result = sdp_desc_q_clear(&conn->r_src);
			TS_EXPECT(MOD_LNX_SDP, !(0 > result));

			result = 0;
			break;
		default:
			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: unhandled READ error status <%d>.",
				 comp->status);
			result = -EIO;
		}

		goto done;
	}

#ifdef _TS_SDP_DATA_PATH_DEBUG
	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "EVENT: RDMA read <%08x> of <%u> bytes.",
		 comp->work_request_id, comp->bytes_transferred);
#endif
	/*
	 * update queue depth
	 */
	conn->s_wq_size--;
	/*
	 * Four basic scenarios:
	 *
	 * 1) BUFF at the head of the active read table is completed by this
	 *    read event completion
	 * 2) IOCB at the head of the active read table is completed by this
	 *    read event completion
	 * 3) IOCB at the head of the active read table is not associated
	 *    with this event, meaning a later event in flight will complete
	 *    it, no IOCB is completed by this event.
	 * 4) No IOCBs are in the active table, the head of the read pending
	 *    table, matches the work request ID of the event and the recv
	 *    low water mark has been satisfied.
	 */
	/*
	 * check type at head of queue
	 */
	type = sdp_desc_q_type_head(&conn->r_src);
    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN, "completion type is %d",type);

	switch (type) {
	case TS_SDP_GENERIC_TYPE_BUFF:

		buff = (struct sdpc_buff *) sdp_desc_q_get_head(&conn->r_src);
		TS_EXPECT(MOD_LNX_SDP, (NULL != buff));

		if (comp->work_request_id != buff->ib_wrid) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: work request ID mismatch <%08x:%08x> on BUFF READ.",
				 comp->work_request_id, buff->ib_wrid);

			(void)sdp_buff_pool_put(buff);
			result = -EPROTO;
			goto done;
		}
		/*
		 * post data to the stream interface
		 */
		result = sdp_recv_buff(conn, buff);
		if (0 < result) {
			/*
			 * count number of bytes buffered by the connection,
			 * zero byte buffers can be returned. If data was
			 * consumed by the protocol, signal the user.
			 */
			conn->byte_strm += result;

			sdp_inet_wake_recv(conn->sk, conn->byte_strm);
		}
		else {

			if (0 > result) {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_WARN,
					 "EVENT: gateway buffer recv error on RDMA read. <%d>",
					 result);
			}

			result = sdp_buff_pool_put(buff);
			TS_EXPECT(MOD_LNX_SDP, !(0 > result));
		}

		break;
   
	 case TS_SDP_GENERIC_TYPE_IOCB:

		iocb = (struct sdpc_iocb *) sdp_desc_q_look_head(&conn->r_src);
		if (NULL != iocb && iocb->wrid == comp->work_request_id) {

			iocb =
			    (struct sdpc_iocb *) sdp_desc_q_get_head(&conn->r_src);
			TS_EXPECT(MOD_LNX_SDP, (NULL != iocb));

			iocb->flags &=
			    ~(TS_SDP_IOCB_F_ACTIVE | TS_SDP_IOCB_F_RDMA_R);

			TS_SDP_CONN_STAT_READ_INC(conn, iocb->post);
			TS_SDP_CONN_STAT_RQ_DEC(conn, iocb->size);

			result = sdp_iocb_complete(iocb, 0);
			if (0 > result) {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_WARN,
					 "EVENT: Error <%d> completing iocb. <%d>",
					 result, iocb->key);
				(void)sdp_iocb_destroy(iocb);
				goto done;
			}
		}

		break;
	case TS_SDP_GENERIC_TYPE_NONE:

		iocb = sdp_iocb_q_look(&conn->r_pend);
		if (NULL == iocb) {

			result = -EPROTO;
			goto done;
		}

		if (iocb->wrid == comp->work_request_id) {

			iocb->flags &=
			    ~(TS_SDP_IOCB_F_ACTIVE | TS_SDP_IOCB_F_RDMA_R);

			if (!(TS_SDP_OS_SK_RCVLOWAT(conn->sk) > iocb->post)) {

				iocb = sdp_iocb_q_get_head(&conn->r_pend);
				TS_EXPECT(MOD_LNX_SDP, (NULL != iocb));

				TS_SDP_CONN_STAT_READ_INC(conn, iocb->post);
				TS_SDP_CONN_STAT_RQ_DEC(conn, iocb->size);
                if (iocb->aio==1) {
				  result = sdp_iocb_complete(iocb, 0);
				} else {
				  conn->r_pend_copied+=iocb->post;
                  sdp_iocb_q_put_tail(&conn->r_comp, iocb);
                  conn->flying_bytes-=iocb->post;

                  sdp_inet_wake_recv(conn->sk, conn->byte_strm);

                };
                if (0 > result) {

					TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
						 TRACE_FLOW_WARN,
						 "EVENT: Error <%d> completing iocb. <%d>",
						 result, iocb->key);
					(void)sdp_iocb_destroy(iocb);
					goto done;
				}
			}
		}

		break;
    case TS_SDP_GENERIC_TYPE_ZCOPY:
           iocb = (struct sdpc_iocb *) sdp_desc_q_look_head(&conn->r_src);
               if (NULL != iocb && iocb->wrid == comp->work_request_id) {

               iocb =      
                   (struct sdpc_iocb *) sdp_desc_q_get_head(&conn->r_src);
               TS_EXPECT(MOD_LNX_SDP, (NULL != iocb));

               iocb->flags &=
                   ~(TS_SDP_IOCB_F_ACTIVE | TS_SDP_IOCB_F_RDMA_R);

               TS_SDP_CONN_STAT_READ_INC(conn, iocb->post);
               TS_SDP_CONN_STAT_RQ_DEC(conn, iocb->size);
               conn->r_pend_copied+=iocb->post;

               sdp_iocb_q_put_tail(&conn->r_comp, iocb);
               conn->flying_bytes-=iocb->post;

               //result = sdp_zcopy_complete(iocb, 0);//TODO :goto user's context before completing
               TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,"completion received size %d flying is %d",iocb->post,conn->flying_bytes);
               sdp_inet_wake_recv(conn->sk, conn->byte_strm);
               }
               break;

	default:

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: Unknown type <%d> at head of READ SOURCE queue. <%d>",
			 type, sdp_desc_q_size(&conn->r_src));
		result = -EPROTO;
		goto done;
	}
	/*
	 * The advertisment which generated this READ needs to be checked.
	 */
	result = _sdp_event_read_advt(conn, comp);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: Error <%d> handling READ advertisment. <%08x>",
			 result, comp->work_request_id);
		goto done;
	}
	/*
	 * It's possible that the "send" queue was opened up by the completion
	 * of some RDMAs
	 */
	result = sdp_send_flush(conn);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: Flush failure during receive completion. <%d>",
			 result);
		goto done;
	}
	/*
	 * The completion of the RDMA read may allow us to post additional RDMA
	 * reads.
	 */
	result = sdp_recv_flush(conn);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: Receive buffer post failed on RDMA read. <%d>",
			 result);
		goto done;
	}

done:
	return result;
} /* sdp_event_read */
