/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  This source code may incorporate intellectual property owned by
  Microsoft Corporation. Our provision of this source code does not
  include any licenses or any other rights to you under any Microsoft
  intellectual property. If you would like a license from Microsoft
  (e.g., to rebrand, redistribute), you need to contact Microsoft
  directly.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sdp_recv.c 379 2004-07-08 02:24:32Z roland $
*/

#include "sdp_main.h"

static char _peek_queue_name[] = TS_SDP_INET_PEEK_DATA_NAME;
/* --------------------------------------------------------------------- */
/*                                                                       */
/* Receive posting function(s)                                           */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*.._sdp_post_recv_buff -- post a single buffers for data recv */
static tINT32 _sdp_post_recv_buff(struct sdp_opt *conn)
{
    struct ib_receive_param receive_param = { 0};
    tINT32 result;
    struct sdpc_buff *buff;

    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * get a buffer
     */
    buff = sdp_buff_pool_get();
    if (NULL == buff)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: failed to allocate buffer for receive queue.");
        result = -ENOMEM;
        goto error;
    }
    /*
     * The data pointer is backed up based on what the stream interface
     * peer advertised to us plus the required header. This way the
     * data we end up passing to the interface will always be within
     * the correct range.
     */
    buff->tail = buff->end;
    buff->data = buff->tail - conn->recv_size;
    buff->lkey = conn->l_key;
    buff->ib_wrid = TS_SDP_WRID_RECV_MASK & conn->recv_wrid++;

    conn->l_recv_bf++;
    /*
     * save the buffer for the event handler. Make sure it's before
     * actually posting the thing. Completion event can happen before
     * post function returns.
     */
    result = sdp_buff_q_put_tail(&conn->recv_post, buff);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Error <%d> queuing receive buffer.", result);
        goto drop;
    }
#ifdef _TS_SDP_DATA_PATH_DEBUG
    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
             "POST: RECV BUFF wrid <%08x> of <%u> bytes.",
             buff->ib_wrid, (buff->tail - buff->data));
#endif
    /*
     * post recv
     */
    receive_param.work_request_id = buff->ib_wrid;
    receive_param.scatter_list = TS_SDP_BUFF_GAT_SCAT(buff);
    receive_param.num_scatter_entries = 1;
    receive_param.signaled = 1;

    result = ib_receive(conn->qp, &receive_param, 1,NULL);
    if (0 != result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: receive failure. <%d>", result);

        (void)sdp_buff_q_get_tail(&conn->recv_post);
        goto drop;
    }

    return 0;
    drop:
    (void)sdp_buff_pool_put(buff);
    conn->l_recv_bf--;
    error:
    return result;
} /* _sdp_post_recv_buff */

/* ========================================================================= */
/*.._sdp_post_rdma_buff -- post a single buffers for rdma read on a conn */
static tINT32 _sdp_post_rdma_buff(struct sdp_opt *conn)
{
    struct ib_send_param send_param = { 0};
    struct sdpc_advt *advt;
    tINT32 result;
    struct sdpc_buff *buff;

    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * check queue depth
     */
    if (!(conn->send_cq_size > conn->s_wq_size))
    {

        result = ENODEV;
        goto done;
    }
    /*
     * get a reference to the first SrcAvail advertisment.
     */
    advt = sdp_advt_q_look(&conn->src_pend);
    if (NULL == advt)
    {

        result = ENODEV;
        goto done;
    }
    /*
     * get a buffer
     */
    buff = sdp_buff_pool_get();
    if (NULL == buff)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: failed to allocate buffer for receive queue.");

        result = -ENOMEM;
        goto error;
    }
    /*
     * The data pointer is backed up based on what the stream interface
     * peer advertised to us plus the required header. This way the
     * data we end up passing to the interface will always be within
     * the correct range.
     */
    buff->tail = buff->end;
    buff->data = buff->tail - min((tINT32) conn->recv_size, advt->size);
    buff->lkey = conn->l_key;

    buff->ib_wrid = TS_SDP_WRID_READ_FLAG | conn->recv_wrid++;

    send_param.op = IB_OP_RDMA_READ;
    send_param.remote_address = advt->addr;
    send_param.rkey = advt->rkey;
    send_param.signaled = 1;

    advt->wrid = buff->ib_wrid;
    advt->size -= (buff->tail - buff->data);
    advt->addr += (buff->tail - buff->data);
    advt->post += (buff->tail - buff->data);
    advt->flag |= TS_SDP_ADVT_F_READ;
    /*
     * If there is no more advertised space move the advertisment to the
     * active list, and match the WRID.
     */
    if (!(0 < advt->size))
    {

        advt = sdp_advt_q_get(&conn->src_pend);
        if (NULL == advt)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: src available advertisment disappeared.");
            result = -ENODEV;
            goto drop;
        }

        result = sdp_advt_q_put(&conn->src_actv, advt);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: Error <%d> queuing active advertisment.",
                     result);

            (void)sdp_advt_destroy(advt);
            goto drop;
        }
    }
    /*
     * save the buffer for the event handler. Make sure it's before
     * actually posting the thing. Completion event can happen before
     * post function returns.
     */

    result = sdp_desc_q_put_tail(&conn->r_src, (struct sdpc_desc *) buff);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Error <%d> queuing rdma read.", result);
        goto drop;
    }
#ifdef _TS_SDP_DATA_PATH_DEBUG
    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
             "POST: READ BUFF wrid <%08x> of <%u> bytes.",
             buff->ib_wrid, (buff->tail - buff->data));
#endif
    /*
     * update send queue depth
     */
    conn->s_wq_size++;
    /*
     * post rdma
     */
    send_param.work_request_id = buff->ib_wrid;
    send_param.gather_list = TS_SDP_BUFF_GAT_SCAT(buff);
    send_param.num_gather_entries = 1;

    result = ib_send(conn->qp, &send_param, 1,NULL);
    if (0 != result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: send (rdma read) failure. <%d>", result);

        (void)sdp_desc_q_get_tail(&conn->r_src);
        conn->s_wq_size--;

        goto drop;
    }

    return 0;
    drop:
    (void)sdp_buff_pool_put(buff);
    error:
    done:
    return result;
} /* _sdp_post_rdma_buff */

/* ========================================================================= */
/*.._sdp_post_rdma_iocb_src -- post a iocb for rdma read on a conn */
static tINT32 _sdp_post_rdma_iocb_src(struct sdp_opt *conn)
{
    struct ib_send_param send_param = { 0};
    struct ib_gather_scatter sg_val;
    struct sdpc_iocb *iocb;

    struct sdpc_advt *advt;
    tINT32 result;
    tINT32 zcopy;

    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * check queue depth
     */
    if (!(conn->send_cq_size > conn->s_wq_size))
    {

        result = ENODEV;
        goto done;
    }
    /*
     * get a reference to the first SrcAvail advertisment.
     */
    // TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
    //        "POST: RDMA READ REQ src_pend %d r_pend %d ",sdp_advt_q_size(&conn->src_pend), sdp_iocb_q_size(&conn->r_pend));
    advt = sdp_advt_q_look(&conn->src_pend);
    if (NULL == advt)
    {

        result = ENODEV;
        goto done;
    }
    /*
     * get a reference to the first IOCB pending.
     *
     * check if the IOCB is in cancel processing.
     * (final complete RDMA will clear it out.)
     */
    iocb = sdp_iocb_q_look(&conn->r_pend);
    if (NULL == iocb)
    {

        result = ENODEV;
        goto done;
    }
    /*
     * register IOCBs physical memory if this is the first time for
     * this IOCB.
     */
    if (NULL == iocb->page_array)
    {
        if (iocb->aio==0)
        {
            result=0;
            //   result = sdp_zcopy_register(iocb, conn);
        } else
        {
            result = sdp_iocb_register(iocb, conn);
        }
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                     "POST: Error <%d> registering IOCB. <%d:%d>",
                     result, iocb->key, iocb->len);
            goto error;
        }
    }
    /*
     * amount of data to zcopy.
     */

    iocb->len-=conn->buffered_bytes;
    iocb->post+=conn->buffered_bytes;
    iocb->io_addr+=conn->buffered_bytes;
    conn->r_zcopy_called=1;


    zcopy = min(advt->size, iocb->len);
    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,"advt size %d iocb len %d",advt->size, iocb->len);
    sg_val.address = iocb->io_addr;
    sg_val.key = iocb->l_key;
    sg_val.length = zcopy;

    send_param.op = IB_OP_RDMA_READ;
    send_param.remote_address = advt->addr;
    send_param.rkey = advt->rkey;
    send_param.signaled = 1;

    iocb->wrid = TS_SDP_WRID_READ_FLAG | conn->recv_wrid++;
    iocb->len -= zcopy;
    iocb->post += zcopy;

    iocb->io_addr += zcopy;
    iocb->flags |= TS_SDP_IOCB_F_ACTIVE;
    iocb->flags |= TS_SDP_IOCB_F_RDMA_R;

    advt->wrid = iocb->wrid;
    advt->size -= zcopy;
    advt->addr += zcopy;
    advt->post += zcopy;
    advt->flag |= TS_SDP_ADVT_F_READ;
    /*
     * if there is no more advertised space,  queue the
     * advertisment for completion
     */
    if (!(0 < advt->size))
    {

        advt = sdp_advt_q_get(&conn->src_pend);
        if (NULL == advt)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: src available advertisment disappeared.");
            result = -ENODEV;
            goto error;
        }

        result = sdp_advt_q_put(&conn->src_actv, advt);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: Error <%d> queuing active advertisment.",
                     result);

            (void)sdp_advt_destroy(advt);
            goto error;
        }
    }
    /*
     * if there is no more iocb space queue the it for completion
     */
    if (!(0 < iocb->len))
    {


        iocb = sdp_iocb_q_get_head(&conn->r_pend);

        if (NULL == iocb)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: read IOCB has disappeared from table.");
            result = -ENODEV;
            goto error;
        }
        result =
        sdp_desc_q_put_tail(&conn->r_src, (struct sdpc_desc *) iocb);
        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,   "r_src is increased");

        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: Error <%d> queuing active read IOCB.",
                     result);

            (void)sdp_iocb_destroy(iocb);
            goto error;
        }
    }
#ifdef _TS_SDP_DATA_PATH_DEBUG
    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
             "POST: READ IOCB wrid <%08x> of <%u> bytes <%d:%d>.",
             iocb->wrid, zcopy, iocb->len, advt->size);
#endif
    /*
     * update send queue depth
     */
    conn->s_wq_size++;
    /*
     * post RDMA
     */

    conn->flying_bytes+=zcopy;




    send_param.work_request_id = iocb->wrid;
    send_param.gather_list = &sg_val;

    send_param.num_gather_entries = 1;
    result = ib_send(conn->qp, &send_param, 1,NULL);
    if (0 != result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: send (rdma read) failure. <%d>", result);

        conn->s_wq_size--;
        goto error;
    }

    return 0;
    error:
    done:
    return result;
} /* _sdp_post_rdma_iocb_src */

/* ========================================================================= */
/*.._sdp_post_rdma_iocb_snk -- post a iocb for rdma read on a conn */
static tINT32 _sdp_post_rdma_iocb_snk(struct sdp_opt *conn)
{
    tINT32 result = 0;
    struct sdpc_iocb *iocb;

    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * check if sink cancel is pending
     */
    if (0 < (TS_SDP_CONN_F_SNK_CANCEL & conn->flags))
    {

        result = ENODEV;
        goto error;
    }
    /*
     * get the pending iocb
     */
    iocb = sdp_iocb_q_look(&conn->r_pend);
    if (NULL == iocb)
    {

        result = ENODEV;
        goto error;
    }
    /*
     * check zcopy threshold
     */
    if (conn->snk_zthresh > iocb->len)
    {

        result = ENODEV;
        goto error;
    }
    /*
     * check number of outstanding sink advertisments
     */
    if (!(conn->r_max_adv > conn->snk_sent))
    {

        result = ENODEV;
        goto error;
    }
    /*
     * registration
     */
    if (NULL == iocb->page_array)
    {

        result = sdp_iocb_register(iocb, conn);
        if (0 > result)
        {
            if (-EAGAIN == result)
            {

                result = EAGAIN;
            } else
            {

                TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE,
                         TRACE_FLOW_WARN,
                         "POST: Error <%d> registering IOCB. <%d:%d>",
                         result, iocb->key, iocb->len);
            }

            goto error;
        }
    }
    /*
     * IOCB
     */
    iocb->flags |= TS_SDP_IOCB_F_ACTIVE;
    iocb->flags |= TS_SDP_IOCB_F_RDMA_W;
    /*
     * queue IOCB
     */
    iocb = sdp_iocb_q_get_head(&conn->r_pend);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "WRITE: Failed to get queued write IOCB <%d> from pending table",
                 result, iocb->key,
                 sdp_iocb_q_size(&conn->r_pend));
        goto release;
    }

    result = sdp_iocb_q_put_tail(&conn->r_snk, iocb);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "WRITE: Error <%d> queueing write IOCB <%d> to active table",
                 result, iocb->key,
                 sdp_iocb_q_size(&conn->r_snk));

        goto re_q;
    }
    /*
     * Either post a send, or buffer the packet in the tx queue
     */
    result =
    sdp_send_ctrl_snk_avail(conn, iocb->len, iocb->r_key, iocb->io_addr);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to post buffered control message. <%d>",
                 result);
        goto de_q;
    }

    conn->snk_sent++;

    return 0;
    de_q:
    iocb = sdp_iocb_q_get_tail(&conn->r_snk);
    re_q:
    (void)sdp_iocb_q_put_head(&conn->r_pend, iocb);
    release:
    iocb->flags &= ~TS_SDP_IOCB_F_ACTIVE;
    iocb->flags &= ~TS_SDP_IOCB_F_RDMA_W;

    (void)sdp_iocb_release(iocb);
    error:
    return result;
} /* _sdp_post_rdma_iocb_snk */

/* ========================================================================= */
/*.._sdp_post_rdma -- post a rdma based requests for a connection */
static tINT32 _sdp_post_rdma(struct sdp_opt *conn)
{
    tINT32 result=0;
    struct sdpc_iocb *iocb;
    struct sdpc_advt *advt;


    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * Since RDMA Reads rely on posting to the Send WQ, stop if
     * we're not in an appropriate state. It's possible to queue
     * the sink advertisment, something to explore, but SrcAvail
     * slow start might make that unneccessart?
     */
    if (0 == (TS_SDP_ST_MASK_SEND_OK & conn->state))
    {

        return 0;
    }
    /*
     * loop flushing IOCB RDMAs. Read sources, otherwise post sinks.
     */
    if (0 < sdp_advt_q_size(&conn->src_pend))
    {

        if (0 == sdp_desc_q_types_size(&conn->r_src,
                                       TS_SDP_GENERIC_TYPE_BUFF))
        {
            if (0 == sdp_buff_q_size(&conn->recv_pool))
            {

                while (0 == (result = _sdp_post_rdma_iocb_src(conn)))
                {
                    /*
                     * pass, nothing to do in loop.
                       */
                } 
            }
            /*
             * check non-zero result
             */
            if (0 > result)
            {

                TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                         TRACE_FLOW_WARN,
                         "POST: failure posting RDMA IOCB read. <%d>",
                         result);
                goto done;
            }
        }
        /*
         * loop posting RDMA reads, if there is room.
         */
        if (0 == sdp_iocb_q_size(&conn->r_pend))
        {

            while (0 < sdp_advt_q_size(&conn->src_pend) &&
                   conn->recv_max >
                   sdp_buff_q_size(&conn->recv_pool) &&
                   conn->rwin_max > conn->byte_strm)
            {

                advt = sdp_advt_q_look(&conn->src_pend);
                if (advt->size > TS_SDP_BCOPY_SIZE_MAX && conn->inside_read==0)
                {
                    break;
                }


                if (0 !=
                    (result = _sdp_post_rdma_buff(conn)))
                {

                    if (0 > result)
                    {

                        TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                                 TRACE_FLOW_WARN,
                                 "POST: failure posting RDMA read. <%d>",
                                 result);
                        goto done;
                    } else
                    {
                        /*
                         * No more posts allowed.
                         */
                        break;
                    }
                }
            }
        }
    } else
    {
        iocb= sdp_iocb_q_look(&conn->r_pend);


        if (iocb!=NULL &&    TS_SDP_MODE_PIPE == conn->recv_mode &&
            0 == sdp_advt_q_size(&conn->src_actv))
        {
            if (iocb->aio==1)
            {
                while (0 == (result = _sdp_post_rdma_iocb_snk(conn)))
                {
                    /*
                     * pass
                     */
                }

                if (0 > result)
                {

                    TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                             TRACE_FLOW_WARN,
                             "POST: Error <%d> sending sink advertisment.",
                             result);
                    goto done;
                }
            }
        }
    }

    result = 0;
    done:
    return result;
} /* _sdp_post_rdma */

/* ========================================================================= */
/*..sdp_recv_flush -- post a certain number of buffers on a connection */
tINT32 sdp_recv_flush(struct sdp_opt *conn)
{
    tINT32 result;
    int counter;

    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * verify that the connection is in a posting state
     */
    if (0 == (TS_SDP_ST_MASK_RCV_POST & conn->state))
    {

        return 0;
    }
    /*
     * loop posting receive buffers onto the queue
     */
    /*
     * 1) Calculate available space in the receive queue. Take the
     *    smallest between bytes available for buffering and maximum
     *    number of buffers allowed in the queue. (this prevents a
     *    flood of small buffers)
     * 2) Subtract buffers already posted
     * 3) Take the smallest buffer count between those needed to fill
     *    the buffered receive/receive posted queue, and the maximum
     *    number which are allowed to be posted at a given time.
     */
    counter = min((s32)((conn->rwin_max - conn->byte_strm)/
                        conn->recv_size),
                  (s32) (conn->recv_max -
                         sdp_buff_q_size(&conn->recv_pool)));
    counter -= conn->l_recv_bf;

    counter = min(counter, 
                  ((s32)conn->recv_cq_size - (s32)conn->l_recv_bf));

    while (0 < counter--)
    {

        if (0 != (result = _sdp_post_recv_buff(conn)))
        {

            if (0 > result)
            {

                TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                         TRACE_FLOW_WARN,
                         "POST: failure posting receive. <%d>",
                         result);
                goto done;
            } else
            {
                /*
                 * No more recv buffers allowed.
                 */
                break;
            }
        }
    }
    /*
     * If we are in Sink Cancel processing, and the active sink queue has
     * been consumed, we can come out of sink processing.
     */
    if (0 < (TS_SDP_CONN_F_SNK_CANCEL & conn->flags) &&
        0 == sdp_iocb_q_size(&conn->r_snk))
    {

        conn->flags &= ~TS_SDP_CONN_F_SNK_CANCEL;
    }
    /*
     * Next the connection should consume RDMA Source advertisments or
     * create RDMA Sink advertisments, either way setup for RDMA's for
     * data flowing from the remote connection peer to the local
     * connection peer.
     */
    result = _sdp_post_rdma(conn);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Error <%d> while posting RDMA. <%04x>",
                 result, conn->state);
        goto done;
    }
    /*
     * Gratuitous increase of remote send credits. Independant of posting
     * recveive buffers, it may be neccessary to notify the remote client
     * of how many buffers are available. For small numbers advertise more
     * often, then for large numbers. Always advertise when we add the
     * first two buffers.
     *
     * 1) Fewer advertised buffers then actual posted buffers.
     * 2) Less then three buffers advertised. (OR'd with the next two
     *    because we can have lots of sinks, but still need to send
     *    since those sinks may never get used. (EOF))
     * 3) The discrepency between posted and advertised is greater then
     *    three
     * 4) The peer has no source or sink advertisments pending. In process
     *    advertisments generate completions, that's why no ack.
     */
    if ((3 > conn->l_advt_bf &&
         conn->l_recv_bf > conn->l_advt_bf) ||
        (TS_SDP_CONN_RECV_POST_ACK < (conn->l_recv_bf - conn->l_advt_bf) &&
         0 == ((tUINT32) conn->snk_recv + (tUINT32) conn->src_recv)))
    {

        result = sdp_send_ctrl_ack(conn);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: failed to post ACK while growing recveive pool. <%d>",
                     result);
            goto done;
        }
    }

    result = 0;
    done:
    return result;
} /* sdp_recv_flush */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* Receive incoming data function(s)                                     */
/*                                                                       */
/* --------------------------------------------------------------------- */
#ifdef _TS_SDP_AIO_SUPPORT
/* ========================================================================= */
/*.._sdp_read_buff_iocb -- read a SDP buffer into a kvec */
static tINT32 _sdp_read_buff_iocb(struct kvec_dst *dst,
                                  struct sdpc_buff *buff,
                                  tINT32 len)
{
    void *data;
    void *tail;
    tINT32 part;
    tINT32 left;
    tINT32 copy;

    TS_CHECK_NULL(dst, -EINVAL);
    TS_CHECK_NULL(buff, -EINVAL);
    /*
     * OOB buffer adjustment. We basically throw away OOB data
     * when writing into AIO buffer. We can't split it into it's
     * own AIO buffer, because that would violate outstanding
     * advertisment calculations.
     */
    data = buff->data;
    tail = buff->tail;
    buff->tail -= (0 < (TS_SDP_BUFF_F_OOB_PRES & buff->flags)) ? 1 : 0;
    /*
     * activate destination memory kmap
     */
    TS_SDP_KVEC_DST_MAP(dst);
    /*
     * copy buffer to dst
     */
    copy = min(len, (tINT32) (buff->tail - buff->data));

    for (left = copy; 0 < left;)
    {

        part = min(left, dst->space);
#ifndef _TS_SDP_DATA_PATH_NULL
        memcpy(dst->dst, buff->data, part);
#endif

        buff->data += part;
        dst->space -= part;
        dst->dst += part;
        left -= part;

        if (0 < left && 0 == dst->space)
        {

            TS_SDP_KVEC_DST_UNMAP(dst);
            dst->let++;
            dst->offset = 0;
            TS_SDP_KVEC_DST_MAP(dst);

            if (!(0 < dst->space))
            {    /* sanity check */

                buff->data = data;
                copy = -EFAULT;
                goto error;
            }
        }
    }

    error:
    /*
     * release destination memory kmap
     */
    TS_SDP_KVEC_DST_UNMAP(dst);
    /*
     * restore tail from OOB offset.
     */
    buff->tail = tail;

    return copy;
} /* _sdp_read_buff_iocb */

/* ========================================================================= */
/*.._sdp_read_buff_iocb_flush -- read multiple SDP buffers into a kvec */
tINT32 _sdp_read_buff_iocb_flush(struct sock *sk,
                                 struct kvec_dst *dst,
                                 tINT32 len)
{
    struct sdp_opt *conn;
    struct sdpc_buff *buff;
    tINT32 copied = 0;
    tINT32 partial;
    tINT32 result;
    tINT32 expect;

    TS_CHECK_NULL(sk, -EINVAL);
    TS_CHECK_NULL(dst, -EINVAL);
    conn = TS_SDP_GET_CONN(sk);
    /*
     * loop through queued buffers and copy them to the destination
     */
    while (copied < len &&
           NULL != (buff = sdp_buff_q_get_head(&conn->recv_pool)))
    {
        /*
         * OOB data is trown away.
         */
        if (0 < (TS_SDP_BUFF_F_OOB_PRES & buff->flags) &&
            1 == (buff->tail - buff->data))
        {

            conn->rcv_urg_cnt -= 1;
            conn->byte_strm -= 1;
            buff->data += 1;
        } else
        {

            partial = _sdp_read_buff_iocb(dst, buff, (len - copied));
            if (0 > partial)
            {

                TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                         TRACE_FLOW_WARN,
                         "READ: error <%d> copying data <%d:%d> to vector.",
                         partial, copied, len);

                expect =
                sdp_buff_q_put_head(&conn->recv_pool,
                                    buff);
                TS_EXPECT(MOD_LNX_SDP, !(0 > expect));

                result = partial;
                goto done;
            }

            copied += partial;
            conn->byte_strm -= partial;
        }
        /*
         * consume buffer
         */
        if (buff->tail > buff->data)
        {

            expect = sdp_buff_q_put_head(&conn->recv_pool, buff);
            TS_EXPECT(MOD_LNX_SDP, !(0 > expect));
        } else
        {

            expect = sdp_buff_pool_put(buff);
            TS_EXPECT(MOD_LNX_SDP, !(0 > expect));
        }
    }

    result = 0;
    done:

    return((0 < copied) ? copied : result);
} /* _sdp_read_buff_iocb_flush */
#endif

/* ========================================================================= */
/*.._sdp_recv_buff_iocb_active -- Ease AIO read pending pressure */
static tINT32 _sdp_recv_buff_iocb_active(struct sdp_opt *conn, struct sdpc_buff *buff)
{
#ifdef _TS_SDP_AIO_SUPPORT
    struct sdpc_iocb *iocb;
    tINT32 result;

    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(buff, -EINVAL);
    /*
     * Get the IOCB, We'll fill with exactly one
     */
    iocb = sdp_iocb_q_get_head(&conn->r_snk);
    if (NULL == iocb)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "STRM: Empty active IOCB queue. <%d>",
                 sdp_iocb_q_size(&conn->r_snk));
        return -EPROTO;
    }

    TS_EXPECT(MOD_LNX_SDP, (0 < (TS_SDP_IOCB_F_RDMA_W & iocb->flags)));
    /*
     * TODO: need to be checking OOB here.
     */
    result = _sdp_read_buff_iocb(&iocb->kvec.dst, buff, iocb->len);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "READ: error <%d> copying data <%d:%d> to vector.",
                 result, iocb->len, (buff->tail - buff->data));

        (void)sdp_iocb_q_put_head(&conn->r_snk, iocb);
        goto error;
    }

    iocb->len -= result;
    iocb->post += result;

    TS_SDP_CONN_STAT_READ_INC(conn, iocb->post);
    TS_SDP_CONN_STAT_RQ_DEC(conn, iocb->size);

    conn->nond_recv--;
    conn->snk_sent--;
    /*
     * callback to complete IOCB
     */
    result = sdp_iocb_complete(iocb, 0);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: Error <%d> completing iocb. <%d>", result,
                 iocb->key);

        (void)sdp_iocb_destroy(iocb);
    }

    return(buff->tail - buff->data);
    error:
    return result;
#else
    return 0;
#endif
} /* _sdp_recv_buff_iocb_active */
/* ========================================================================= */
/*.._sdp_recv_buff_iocb_pending -- Ease AIO read pending pressure */
static tINT32 _sdp_recv_buff_iocb_pending(struct sdp_opt *conn, struct sdpc_buff *buff)
{
#ifdef _TS_SDP_AIO_SUPPORT
    struct sdpc_iocb *iocb;
    tINT32 copied;
    tINT32 result;

    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(buff, -EINVAL);
    /*
     * check the IOCB
     */
    iocb = sdp_iocb_q_look(&conn->r_pend);
    if (NULL == iocb)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "STRM: Empty pending IOCB queue. <%d>",
                 sdp_iocb_q_size(&conn->r_pend));
        return -EPROTO;
    }
    /*
     * TODO: need to be checking OOB here.
     */
    copied = _sdp_read_buff_iocb(&iocb->kvec.dst, buff, iocb->len);
    if (0 > copied)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "READ: error <%d> copying data <%d:%d> to vector.",
                 copied, iocb->len, (buff->tail - buff->data));
        goto error;
    }

    iocb->len -= copied;
    iocb->post += copied;
    /*
     * If the amount of data moved into the IOCB is greater then the
     * socket recv low water mark, then complete the IOCB, otherwise
     * this loop is done.
     */
    if (!(conn->sk->rcvlowat > iocb->post))
    {
        /*
         * complete IOCB
         */
        iocb = sdp_iocb_q_get_head(&conn->r_pend);
        TS_EXPECT(MOD_LNX_SDP, (NULL != iocb));

        TS_SDP_CONN_STAT_READ_INC(conn, iocb->post);
        TS_SDP_CONN_STAT_RQ_DEC(conn, iocb->size);
        /*
         * callback to complete IOCB
         */
        result = sdp_iocb_complete(iocb, 0);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "EVENT: Error <%d> completing iocb. <%d>",
                     result, iocb->key);

            (void)sdp_iocb_destroy(iocb);
        }
    }

    return(buff->tail - buff->data);
    error:
    return copied;
#else
    return 0;
#endif
} /* _sdp_recv_buff_iocb_pending */

/* ========================================================================= */
/*..sdp_recv_buff -- Process a new buffer based on queue type. */
tINT32 sdp_recv_buff(struct sdp_opt *conn, struct sdpc_buff *buff)
{
    tINT32 result;
    tINT32 buffered;
    struct sdpc_iocb *iocb;


    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(buff, -EINVAL);

#ifdef _TS_SDP_DATA_PATH_DEBUG
    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
             "SOCK: <%08x:%04x><%08x:%04x>  <%04x> recv "
             "id <%08x> bytes <%d>",
             conn->src_addr, conn->src_port, conn->dst_addr, 
             conn->dst_port, conn->state, buff->u_id, 
             (buff->tail - buff->data));
#endif
    /*
     * To emulate RFC 1122 (page 88) a connection should be reset/aborted
     * if data is received and the receive half of the connection has been
     * closed. This notifies the peer that the data was not received.
     */
    if (0 < (TS_SDP_SHUTDOWN_RECV & conn->shutdown))
    {

        TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
                 "EVENT: Received Data, receive path closed. <%02x:%04x>",
                 conn->shutdown, conn->state);
        /*
         * abort connection (send reset)
         */
        result = sdp_wall_abort(conn);
        TS_EXPECT(MOD_LNX_SDP, !(0 > result));
        /*
         * drop packet
         */
        result = 0;
        goto done;
    }
    /*
     * oob notification.
     */
    if (0 < (TS_SDP_BUFF_F_OOB_PEND & buff->flags))
    {

        conn->rcv_urg_cnt++;
        sdp_inet_wake_urg(conn->sk);
    }
    /*
     * loop while there are available IOCB's, break if there is no
     * more data to read
     */

    while (0 < (sdp_iocb_q_size(&conn->r_pend) +
                sdp_iocb_q_size(&conn->r_snk)))
    {
        /*
         * if there is OOB data in a buffer, the two functions below
         * will leave the byte in the buffer, and potentially loop
         * to here. In which case we are done and the buffer is queued.
         * this allows POLL notification to work, and the OOB byte(s)
         * will not be consumed until the next AIO buffer is posted,
         * or a socket recv (regular or OOB) is called.
         */
        if (0 < (TS_SDP_BUFF_F_OOB_PRES & buff->flags) &&
            1 == (buff->tail - buff->data))
        {

            break;
        }
        /*
         * process either a sink available IOCB which needs to be
         * discarded with exactly one buffer, or process a pending
         * IOCB.
         */
        if (0 < conn->snk_sent)
        {

            result = _sdp_recv_buff_iocb_active(conn, buff);
        } else
        {
            iocb = sdp_iocb_q_look(&conn->r_pend);
            if (NULL == iocb)
            {

                TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                         "STRM: Empty pending IOCB queue. <%d>",
                         sdp_iocb_q_size(&conn->r_pend));
                return -EPROTO;
            }
            if (iocb->aio)
            {
                result = _sdp_recv_buff_iocb_pending(conn, buff);
            } else
            {
                //result = _sdp_recv_buff_zcopy_pending(conn, buff);
                //iocb->io_addr+=(buff->tail - buff->data);
                // iocb->len-=(buff->tail - buff->data);
                //iocb->post+=(buff->tail - buff->data);
                goto buffered;
            }
        }
        /*
         * Check result. Postitive result is data left in the buffer
         */
        if (0 == result)
        {

            break;
        } else
        {

            if (0 > result)
            {

                TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                         TRACE_FLOW_WARN,
                         "EVENT: Error <%d> processing buffered IOCB. <%d:%d:%d>",
                         result, conn->snk_sent,
                         sdp_iocb_q_size(&conn->r_pend),
                         sdp_iocb_q_size(&conn->r_snk));
                goto done;
            }
        }
    }
    /*
     * If there is still data in the buffer then queue it for later.
     */
    buffered:

    buffered = buff->tail - buff->data;
    conn->buffered_bytes+=buffered;
    TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
             TRACE_FLOW_WARN,"buffered is %d total is %d",buffered, conn->buffered_bytes);

    //	TS_TRACE(MOD_LNX_SDP, T_VERBOSE,TRACE_FLOW_WARN,"buffered data %d",buffered);
    if (0 < buffered)
    {
        if (conn->r_zcopy_called==0)
        {
            conn->wait_buff_before_src+=1;
        }
        result = sdp_buff_q_put_tail(&conn->recv_pool, buff);

        TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    }

    return buffered;
    done:
    return result;
} /* sdp_recv_buff */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* User initiated receive data function(s)                               */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*.._sdp_inet_recv_urg_test_func -- recv queue urgent data cleanup function */
static tINT32 _sdp_inet_recv_urg_test_func(struct sdpc_buff *buff, void *arg)
{
    TS_CHECK_NULL(buff, -EINVAL);

    return((buff->tail == buff->head) ? 1 : 0);
} /* _sdp_inet_recv_urg_test_func */

/* ========================================================================= */
/*.._sdp_inet_recv_urg_trav_func -- recv queue urg data retreival function */
static tINT32 _sdp_inet_recv_urg_trav_func(struct sdpc_buff *buff, void *arg)
{
    tUINT8 *value = (tUINT8 *) arg;
    tUINT8 update;

    TS_CHECK_NULL(buff, -EINVAL);
    TS_CHECK_NULL(value, -EINVAL);

    if (0 < (TS_SDP_BUFF_F_OOB_PRES & buff->flags))
    {

        TS_EXPECT(MOD_LNX_SDP, (buff->tail > buff->data));

        update = *value;
        *value = *(tUINT8 *) (buff->tail - 1);

        if (0 < update)
        {

            buff->tail--;
            buff->flags &= ~TS_SDP_BUFF_F_OOB_PRES;
        }

        return -ERANGE;
    }

    return 0;
} /* _sdp_inet_recv_urg_trav_func */

/* ========================================================================= */
/*.._sdp_inet_recv_urg -- recv urgent data from the network to user space */
static tINT32 _sdp_inet_recv_urg(struct sock *sk,
                                 struct msghdr *msg,
                                 int size,
                                 int flags)
{
    struct sdp_opt *conn;
    struct sdpc_buff *buff;
    tINT32 result = 0;
    tUINT8 value;

    TS_CHECK_NULL(sk, -EINVAL);
    TS_CHECK_NULL(msg, -EINVAL);
    conn = TS_SDP_GET_CONN(sk);

    if (0 < TS_SDP_OS_SK_URGINLINE(sk) || 0 == conn->rcv_urg_cnt)
    {

        return -EINVAL;
    }
    /*
     * don't cosume data on PEEK, but do consume data on TRUNC
     */
#if 0
    value = (0 < (MSG_PEEK & flags)) || (0 == size) ? 0 : 1;
#else
    value = (0 < (MSG_PEEK & flags)) ? 0 : 1;
#endif

    result = sdp_buff_q_trav_head(&conn->recv_pool,
                                  _sdp_inet_recv_urg_trav_func,
                                  (void *) & value);
    if (-ERANGE == result)
    {

        msg->msg_flags |= MSG_OOB;
        if (0 < size)
        {

            result = memcpy_toiovec(msg->msg_iov, &value, 1);
            conn->buffered_bytes-=1;

            if (0 != result)
            {

                goto done;
            }
            /*
             * clear urgent pointer on consumption
             */
            if (0 == (MSG_PEEK & flags))
            {

                conn->rcv_urg_cnt -= 1;
                conn->byte_strm -= 1;

                TS_SDP_CONN_STAT_RECV_INC(conn, 1);
                /*
                 * we've potentially emptied a buffer, if
                 * so find and dispose of it, and repost
                 * if appropriate.
                 */
                buff = sdp_buff_q_fetch(&conn->recv_pool,
                                        _sdp_inet_recv_urg_test_func,
                                        (void *) 0);
                if (NULL != buff)
                {

                    result = sdp_buff_pool_put(buff);
                    TS_EXPECT(MOD_LNX_SDP, !(0 > result));

                    result = sdp_recv_flush(conn);
                    TS_EXPECT(MOD_LNX_SDP, !(0 > result));
                }

                result = 1;
            }
        } else
        {

            msg->msg_flags |= MSG_TRUNC;
            result = 0;
        }
    } else
    {

        result = (0 != result) ? result : -EAGAIN;
    }
    done:
    return result;
} /* _sdp_inet_recv_urg */

/* ========================================================================= */
/*..sdp_inet_recv -- recv data from the network to user space. */
tINT32 sdp_inet_recv(
#ifdef TS_KERNEL_2_6
                    struct kiocb *kiocb,
                    struct socket *sock,
                    struct msghdr *msg, 
                    size_t size, 
                    int flags
#else
                    struct socket *sock,
                    struct msghdr *msg,
                    int size, 
                    int flags,
                    struct scm_cookie *scm
#endif
                    )
{
    struct sock *sk;
    struct sdp_opt *conn;
    struct sdpc_buff *buff;
    struct sdpc_buff *head = NULL;
    struct sdpc_iocb *iocb = NULL;
    long timeout;
    size_t length;
    tINT32 result = 0;
    tINT32 offset,serial = 0;
    tINT32 expect;
    tINT32 low_water;
    tINT32 copied = 0;
    tINT32 copy,zcopy,left;
    tINT32 update;
    tINT32 free_count = 0;
    tINT8 oob = 0;
    tINT8 ack = 0;
    tUINT64 base=0;
    struct sdpc_buff_q peek_queue;

    TS_CHECK_NULL(sock, -EINVAL);
    TS_CHECK_NULL(sock->sk, -EINVAL);
    TS_CHECK_NULL(msg, -EINVAL);

    sk = sock->sk;
    conn = TS_SDP_GET_CONN(sk);
    conn->flying_bytes=0;
    // size=msg->msg_iov[0].iov_len;
    zcopy=size;
    left=size;
    serial=0;

//#ifdef _TS_SDP_DATA_PATH_DEBUG
    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
             "SOCK: recv <%08x:%04x> <%08x:%04x> <%04x> flags <%08x> "
             "size <%d> pending <%d> msg addr %x",
             conn->src_addr, conn->src_port, conn->dst_addr, conn->dst_port,
             conn->state, flags, size, conn->byte_strm, msg->msg_iov->iov_base);
//#endif

    if (0 < (MSG_TRUNC & flags))
    {
        /*
         * TODO: unhandled, but need to be handled.
         */
        return -EOPNOTSUPP;
    }

    if (0 < (MSG_PEEK & flags))
    {

        (void)sdp_buff_q_init(&peek_queue, _peek_queue_name, 0);
        msg->msg_flags |= MSG_PEEK;
    }

    TS_SDP_CONN_LOCK(conn);

    if (TS_SDP_SOCK_ST_LISTEN == conn->istate ||
        TS_SDP_SOCK_ST_CLOSED == conn->istate)
    {

        result = -ENOTCONN
                 ;
        goto done;
    }
    /*
     * process urgent data
     */
    if (0 < (MSG_OOB & flags))
    {

        result = _sdp_inet_recv_urg(sk, msg, size, flags);
        copied = (0 < result) ? result : 0;
        result = (0 < result) ? 0 : result;
        goto done;
    }
    /*
     * get socket values we'll need.
     */
    timeout = sock_rcvtimeo(sk, (MSG_DONTWAIT & flags));
    low_water = sock_rcvlowat(sk, (MSG_WAITALL & flags), size);
    /*
     * process data first, and then check condition, then wait
     */
    conn->r_zcopy_called=0;


    conn->r_pend_copied=0;
    while (copied < size)
    {
        /*
         * first copy any data that might be present
         */
        while (copied < size &&  conn->r_zcopy_called==0 &&
               NULL != (buff =
                        sdp_buff_q_get_head(&conn->recv_pool)))
        {

            if (conn->wait_buff_before_src>0)
            {
                conn->wait_buff_before_src-=1;
            }; 

            length = buff->tail - buff->data;
            update = 0;

            if (0 < (TS_SDP_BUFF_F_OOB_PRES & buff->flags))
            {
                /*
                 * if data has already been read, and the
                 * next byte is the urgent byte, reading
                 * needs to terminate, taking precidence
                 * over the low water mark. There needs to
                 * be a break in the the read stream around
                 * the OOB byte regarless if it is inline
                 * or not, to ensure that the user has a
                 * chance to read the byte.
                 */
                if (1 < length)
                {

                    length--;
                } else
                {

                    if (0 < copied)
                    {
                        /*
                         * update such that we pass
                         * through the copy phase,
                         * return the buffer, and
                         * break.
                         */
                        length = 0;
                        update = 0;
                        oob = 1;    /* break on oob */
                    } else
                    {

                        if (0 ==
                            TS_SDP_OS_SK_URGINLINE(sk))
                        {
                            /*
                             * skip this byte, but
                             * make sure it's
                             * counted.
                             */
                            length = 0;
                            update =
                            (0 <
                             (MSG_PEEK & flags))
                            ? 0 : 1;
                        } /* if (OOB) */
                    } /* else (copied) */
                } /* else (sitting on URG byte) */
            } /* if (urgent present) */

            copy = min((size_t) (size - copied), length);

            if (0 < copy)
            {
#ifndef _TS_SDP_DATA_PATH_NULL
                result =
                memcpy_toiovec(msg->msg_iov, buff->data,
                               copy);
                conn->buffered_bytes-=copy;
                TS_TRACE(MOD_LNX_SDP,T_VERBOSE,TRACE_FLOW_WARN,"RECV: copied to user. %d buffered %d",copy,conn->buffered_bytes );

                if (0 > result)
                {

                    expect =
                    sdp_buff_q_put_head(&conn->
                                        recv_pool,
                                        buff);
                    TS_EXPECT(MOD_LNX_SDP, !(0 > expect));

                    goto done;
                }
#endif
                update = (0 < (MSG_PEEK & flags)) ? 0 : copy;
            }

            TS_SDP_CONN_STAT_RECV_INC(conn, update);

            conn->byte_strm -= update;
            buff->data += update;
            copied += copy;

            if (0 < (buff->tail - buff->data))
            {

                expect =
                sdp_buff_q_put_head(&conn->recv_pool,
                                    buff);
                TS_EXPECT(MOD_LNX_SDP, !(0 > expect));
                /*
                 * always break, PEEK and OOB together could
                 * throw us into a loop without a forced
                 * break here, since the buffer data pointer
                 * wasn't really updated. OOB data at the
                 * head of stream, after data has already
                 * been copied relies on this break as well.
                 */
                break;
            } else
            {

                if (0 < (MSG_PEEK & flags))
                {

                    expect =
                    sdp_buff_q_put_head(&peek_queue,
                                        buff);
                    TS_EXPECT(MOD_LNX_SDP, !(0 > expect));
                } else
                {

                    if (0 <
                        (TS_SDP_BUFF_F_OOB_PRES & buff->
                         flags))
                    {

                        conn->rcv_urg_cnt -= 1;
                    }
                    /*
                     * create a link of buffers which
                     * will be returned to the free pool
                     * in one group.
                     */
                    expect =
                    sdp_buff_pool_chain_link(head, buff);
                    TS_EXPECT(MOD_LNX_SDP, !(0 > expect));

                    head = buff;
                    free_count++;
                    /*
                     * post additional recv buffers if
                     * needed, but check only every N
                     * buffers...
                     */
                    if (TS_SDP_CONN_RECV_POST_FREQ < ++ack)
                    {

                        result = sdp_recv_flush(conn);
                        if (0 > result)
                        {

                            TS_TRACE(MOD_LNX_SDP,
                                     T_VERBOSE,
                                     TRACE_FLOW_WARN,
                                     "RECV: device buffer ack error. <%d>",
                                     result);
                            goto done;
                        }

                        ack = 0;
                    }
                }
            }
        }
        /*
         * urgent data needs to break up the data stream, regardless 
         * of low water mark, or whether there is room in the buffer.
         */
        if (0 < oob)
        {

            result = 0;
            break;  
        }
        /*
         * If there is more room for data, cycle the connection lock to
         * potentially flush events into the recv queue. This is done
         * before the low water mark is checked to optimize the number
         * of syscalls a read process needs to make for a given amount
         * of data.
         */
        if (copied < size)
        {


            if (sdp_iocb_q_size(&conn->r_pend)==0 && sdp_iocb_q_size(&conn->r_src)==0 && copied <  low_water && size > conn->src_zthresh 
                && conn->istate==TS_SDP_SOCK_ST_ESTABLISHED  &&  conn->use_zcopy==1   )
            {
                //while (left >0)
                 while (left >  conn->src_zthresh)
               
                {

                    iocb = sdp_zcopy_create();
                    iocb->serial_num=serial;
                    serial+=1;
                    if (NULL == iocb)
                    {

                        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                                 "READ: Failed to allocate iocb to save cb. <%d:%d>",
                                 size, copied);
                        result = -ENOMEM;
                        goto done;
                        //return result; /* synchronous error if no data copied */
                    }

                    base = (tUINT64)(unsigned long) msg->msg_iov->iov_base + size - left ;

                    offset = base & ~PAGE_MASK;
                    zcopy = min(TS_SDP_IOCB_SIZE_MAX - offset, left);
                    left-=zcopy;

                    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN, "base is %x",base);

                    iocb->len = zcopy;
                    iocb->post = copied;
                    iocb->size = zcopy;
                    //iocb->key = req->key;
                    iocb->msg=msg;
                    iocb->dir=0;
                    iocb->base = base;

                    TS_SDP_CONN_STAT_RQ_INC(conn, iocb->size);

                    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,"r_pend is increased %d %p",zcopy,iocb->base);

                    result = sdp_zcopy_register(iocb, conn);
                    if (0 != result)
                    {
                        (void)sdp_iocb_destroy(iocb);
                        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                                 "POST: Error <%d> registering IOCB",
                                 result);

                        goto bcopy;
                    }
                    result = sdp_iocb_q_put_tail(&conn->r_pend, iocb);
                    if (0 > result)
                    {

                        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                                 "READ: Error <%d> while queueing IOCB. <%d:%d>",
                                 result, zcopy, copied);

                        (void)sdp_iocb_destroy(iocb);
                        goto done;
                        //return result;
                    }
                }
            }
            /*
             * process backlog
             */
            conn->r_pend_copied=0;


            TS_SDP_CONN_RELOCK(conn);
            copied+=conn->r_pend_copied;
            //msg->iov->iov_len-=conn->r_pend_copied;
            //msg->iov->iov_base+=conn->r_pend_copied;


            conn->r_pend_copied=0;

            if (0 < sdp_buff_q_size(&conn->recv_pool)  &&  conn->r_zcopy_called==0)
            {


                continue;
            }
        }


        /*
         * If not enough data has been copied to userspace, check for
         * connection errors, and then wait for more data.
         */

        bcopy:    if (copied < low_water || conn->flying_bytes>0)
        {


            /*
             * check status. POSIX 1003.1g order.
             */
            if (0 != TS_SDP_OS_CONN_GET_ERR(conn))
            {
                result =
                (0 < copied) ? 0 : TS_SDP_CONN_ERROR(conn);
                break;
            }

            if (0 < (TS_SDP_SHUTDOWN_RECV & conn->shutdown))
            {
                result = 0;
                break;
            }

            if (TS_SDP_SOCK_ST_ERROR == conn->istate)
            {
                result = -EPROTO; /* error should always be
                         set, but just in case */
                break;
            }

            if (0 == timeout)
            {
                result = -EAGAIN;
                break;
            }
            if ((size < conn->src_zthresh  || sdp_iocb_q_look(&conn->r_pend)==NULL) && conn->flying_bytes==0)
            {
                conn->inside_read=1;
                TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,"inside read is 1");

            }
            expect = sdp_recv_flush(conn);
            if (0 > expect)
            {

                TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                         "RECV: device buffer ack <%d> error. <%d>", copied,
                         expect);
            }




            /*
             * wait
             */
            {
                TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,"going to sleep flying is %d",conn->flying_bytes);

                DECLARE_WAITQUEUE(wait, current);
                add_wait_queue(TS_SDP_OS_SK_SLEEP(sk), &wait);
                set_current_state(TASK_INTERRUPTIBLE);

                set_bit(SOCK_ASYNC_WAITDATA,
                        &TS_SDP_OS_SK_SOCKET(sk)->flags);

                if ((0 == sdp_buff_q_size(&conn->recv_pool) && conn->r_pend_copied==0) || conn->flying_bytes>0 )
                {

                    TS_SDP_CONN_UNLOCK(conn);
                    timeout = schedule_timeout(timeout);
                    TS_SDP_CONN_LOCK(conn);
                }

                clear_bit(SOCK_ASYNC_WAITDATA,
                          &TS_SDP_OS_SK_SOCKET(sk)->flags);
                remove_wait_queue(TS_SDP_OS_SK_SLEEP(sk),
                                  &wait);
                set_current_state(TASK_RUNNING);
                TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,"waking up");
            }
            /*
             * check signal pending
             */

            if (signal_pending(current))
            {
                TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,"signal received");
                result =
                (0 <
                 timeout) ? sock_intr_errno(timeout) :
                -EAGAIN;
                if (conn->flying_bytes==0)
                {
                  goto done;
                } else      {
                    if (conn->istate!=TS_SDP_SOCK_ST_ESTABLISHED)     {
                        goto done;
                    }
                    
                }

            }
            conn->inside_read=0;


        } else
        {

            break;
        }

        copied+=conn->r_pend_copied;

    }

    done:
    while (1)
    {

        iocb=sdp_iocb_q_get_head(&conn->r_comp);
        if (iocb!=NULL)
        {
            iocb->check=1;
            result=sdp_zcopy_complete(iocb,0);
        } else
        {
            break;
        }
    }
    while (1)
    {

        iocb=sdp_iocb_q_get_head(&conn->r_pend);
        if (iocb!=NULL)
        {
            result=sdp_zcopy_complete(iocb,0);
        } else
        {
            break;
        }
    }
    conn->r_pend_copied=0;
    /*
     * acknowledge moved data
     */
    if (0 < ack)
    {

        expect = sdp_recv_flush(conn);
        if (0 > expect)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "RECV: device buffer ack error. <%d>", 
                     expect);
        }
    }

    (void)sdp_buff_pool_chain_put(head, free_count);
    /*
     * return any peeked buffers to the recv queue, in the correct order.
     */
    if (0 < (MSG_PEEK & flags))
    {

        while (NULL != (buff = sdp_buff_q_get_tail(&peek_queue)))
        {

            expect = sdp_buff_q_put_head(&conn->recv_pool, buff);
            TS_EXPECT(MOD_LNX_SDP, !(0 > expect));
        }
    }

    TS_SDP_CONN_UNLOCK(conn);
    //conn->buffered_bytes=0;
    conn->r_zcopy_called=0;
    conn->inside_read=0;
    conn->wait_buff_before_src=sdp_buff_q_size(&conn->recv_pool);

    return((0 < copied) ? copied : result);
} /* sdp_inet_recv */

#ifdef _TS_SDP_AIO_SUPPORT
/* ========================================================================= */
/*.._sdp_inet_read_cancel_func -- lookup function for cancelation */
static tINT32 _sdp_inet_read_cancel_func(struct sdpc_desc *element, void *arg)
{
    struct sdpc_iocb *iocb = (struct sdpc_iocb *) element;
    tINT32 value = (tINT32) (unsigned long)arg;

    TS_CHECK_NULL(element, -EINVAL);

    if (TS_SDP_GENERIC_TYPE_IOCB == element->type && iocb->key == value)
    {

        return 0;
    } else
    {

        return -ERANGE;
    }
} /* _sdp_inet_read_cancel_func */

/* ========================================================================= */
/*.._sdp_inet_read_cancel -- cancel an IO operation */
static tINT32 _sdp_inet_read_cancel(struct kiocb *kiocb
                                    _TS_AIO_UNUSED_CANCEL_PARAM)
{
    struct sock *sk;
    struct sdp_opt *conn;
    struct sdpc_iocb *iocb;
    tINT32 result = 0;

    TS_CHECK_NULL(kiocb, -ERANGE);

    sk = (struct sock *)xchg(&kiocb->data, NULL);
    ts_aio_put_req(kiocb);

    if (NULL == sk)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
                 "SOCK: Cancel empty Kvec kiocb <%d:%d>",
                 kiocb->key, kiocb->users);

        result = -EAGAIN;
        goto done;
    }
    /*
     * lock the socket while we operate.
     */
    conn = TS_SDP_GET_CONN(sk);
    TS_SDP_CONN_LOCK(conn);

    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
             "SOCK: Cancel Kvec <%08x:%04x> <%08x:%04x> <%04x> kiocb <%d:%d>",
             conn->src_addr, conn->src_port, conn->dst_addr,
             conn->dst_port, conn->state, kiocb->key, kiocb->users);
    /*
     * attempt to find the IOCB for this key. we don't have an indication
     * whether this is a read or write.
     */
    iocb = sdp_iocb_q_lookup(&conn->r_pend, kiocb->key);
    if (NULL != iocb)
    {
        /*
         * always remove the IOCB.
         * If active, then place it into the correct active queue
         */
        result = sdp_iocb_q_remove(iocb);
        TS_EXPECT(MOD_LNX_SDP, !(0 > result));

        if (0 < (TS_SDP_IOCB_F_ACTIVE & iocb->flags))
        {

            if (0 < (TS_SDP_IOCB_F_RDMA_W & iocb->flags))
            {

                result =
                sdp_iocb_q_put_tail(&conn->r_snk,
                                    iocb);
                TS_EXPECT(MOD_LNX_SDP, !(0 > result));
            } else
            {

                TS_EXPECT(MOD_LNX_SDP,
                          (0 <
                           (TS_SDP_IOCB_F_RDMA_R & iocb->
                            flags)));

                result =
                sdp_desc_q_put_tail(&conn->r_src,
                                    (struct sdpc_desc *)
                                    iocb);
                TS_EXPECT(MOD_LNX_SDP, !(0 > result));
            }
        } else
        {
            /*
             * callback to complete IOCB, or drop reference
             */
            result = sdp_iocb_complete(iocb, 0);
            TS_EXPECT(MOD_LNX_SDP, !(0 > result));

            result = -EAGAIN;
            goto unlock;
        }
    }
    /*
     * check the source queue, not much to do, since the operation is
     * already in flight.
     */
    iocb = (struct sdpc_iocb *) sdp_desc_q_lookup(&conn->r_src,
                                                  _sdp_inet_read_cancel_func,
                                                  (void *) (unsigned long)kiocb->
                                                  key);
    if (NULL != iocb)
    {

        iocb->flags |= TS_SDP_IOCB_F_CANCEL;
        result = -EAGAIN;

        goto unlock;
    }
    /*
     * check sink queue. If we're in the sink queue, then a cancel
     * needs to be issued.      */
    iocb = sdp_iocb_q_lookup(&conn->r_snk, kiocb->key);
    if (NULL != iocb)
    {
        /*
         * Unfortunetly there is only a course grain cancel in SDP, so
         * we have to cancel everything. This is OKish since it usually
         * only happens at connection termination, and the remaining
         * source probably will get cancel requests as well.
         */
        if (0 == (TS_SDP_CONN_F_SNK_CANCEL & conn->flags))
        {

            result = sdp_send_ctrl_snk_cancel(conn);
            TS_EXPECT(MOD_LNX_SDP, !(0 > result));

            conn->flags |= TS_SDP_CONN_F_SNK_CANCEL;
        }

        iocb->flags |= TS_SDP_IOCB_F_CANCEL;
        result = -EAGAIN;

        goto unlock;
    }
    /*
     * no IOCB found.
     */
    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
             "SOCK: Cancelled kvec with no IOCB. <%d:%d>",
             kiocb->key, kiocb->users);

    if (0 < kiocb->users)
    {
        /*
         * shouldn't occur, but just incase, we don't want to hang
         * the user.
         */
        ts_aio_put_req(kiocb);
        result = 0;
    } else
    {
        /*
         * IOCB should be in completion processing at this point.
         */
        result = -EAGAIN;
    }

    unlock:
    TS_SDP_CONN_UNLOCK(conn);
    done:

    if (0 == result)
    {

        TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_INOUT,
                 "AIO: Kernel bug. Canceling IOCB <%d:%d> leaking <%d> bytes.",
                 kiocb->key, kiocb->users, kiocb->size);
    }

    return result;
} /* _sdp_inet_read_cancel */

/* ========================================================================= */
/*..sdp_inet_read -- recv AIO data from the network to userspace. */
int sdp_inet_read(struct socket * sock, kvec_cb_t cb, size_t size)
{
    struct kiocb *req;
    struct kvec_dst dst;
    struct sock *sk;
    struct sdp_opt *conn;
    struct sdpc_iocb *iocb;
    tINT32 copied = 0;
    tINT32 result = 0;
    tINT32 expect;

    TS_CHECK_NULL(sock, -EINVAL);
    TS_CHECK_NULL(sock->sk, -EINVAL);
    TS_CHECK_NULL(cb.fn, -EINVAL);
    TS_CHECK_NULL(cb.data, -EINVAL);
    TS_CHECK_NULL(cb.vec, -EINVAL);
    TS_CHECK_EXPR((0 < size), -EINVAL);

    sk = sock->sk;
    conn = TS_SDP_GET_CONN(sk);
    req = (struct kiocb *)cb.data;

#ifdef _TS_SDP_DATA_PATH_DEBUG
    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
             "SOCK: Read <%08x:%04x> <%08x:%04x> <%04x> size <%d>",
             conn->src_addr, conn->src_port, conn->dst_addr, conn->dst_port,
             conn->istate, size);
    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
             "SOCK: Read Kvec <%d:%d> veclet <%d:%d> fn <%08x> kiocb "
             "<%d:%d:%08x>",
             cb.vec->max_nr, cb.vec->nr,
             cb.vec->veclet->offset, cb.vec->veclet->length,
             (tUINT32) cb.fn, req->key, req->users, (tUINT32) req->data);
#endif
    /*
     * initialize memory destination
     */
    kvec_dst_init(&dst, KM_USER0);
    /*
     * set target destination address to cb vector
     */
    kvec_dst_set(&dst, cb.vec->veclet);
    /*
     * lock socket
     */
    TS_SDP_CONN_LOCK(conn);
    /*
     * sanity check the connection
     */
    if (TS_SDP_SOCK_ST_LISTEN == conn->istate ||
        TS_SDP_SOCK_ST_CLOSED == conn->istate)
    {

        result = -ENOTCONN;
        goto error;
    }
    /*
     * copy data that's already buffered.
     */
    copied = _sdp_read_buff_iocb_flush(sk, &dst, size);
    if (0 > copied)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "READ: error <%d> flushing data <%d> to dst <%d:%d:%d>",
                 copied, size, dst.space, dst.offset, dst.type);
        result = copied;
        goto error;
    }
    /*
     * check errors, and in those cases fire CB even if the region
     * is not full
     */
    if (0 != TS_SDP_OS_CONN_GET_ERR(conn))
    {

        result = TS_SDP_CONN_ERROR(conn);
        goto error;
    }

    if (0 < (TS_SDP_SHUTDOWN_RECV & conn->shutdown))
    {

        result = 0;
        goto error;
    }

    if (TS_SDP_SOCK_ST_ERROR == conn->istate)
    {
        /*
         * conn->error should always be set, but just in case
         */
        result = -ENOTCONN;
        goto error;
    }
    /*
     * if there was no error and all less data then the recv low water
     * mark was moved, save the kvec in an IOCB and pass it to the
     * SDP Conn.
     */
    if (sk->rcvlowat > copied)
    {
        /*
         * more data, save the cb for processing as data arrives.
         */
        iocb = sdp_iocb_create();
        if (NULL == iocb)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "READ: Failed to allocate iocb to save cb. <%d:%d>",
                     size, copied);
            result = -ENOMEM;
            goto error; /* synchronous error if no data copied */
        }

        iocb->req = req;
        iocb->kvec.dst = dst;
        iocb->cb = cb;
        iocb->len = size - copied;
        iocb->post = copied;
        iocb->size = size;
        iocb->key = req->key;

        TS_SDP_CONN_STAT_RQ_INC(conn, iocb->size);

        result = sdp_iocb_q_put_tail(&conn->r_pend, iocb);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "READ: Error <%d> while queueing IOCB. <%d:%d>",
                     result, size, copied);

            (void)sdp_iocb_destroy(iocb);
            goto error; /* synchronous error if no data copied */
        }

        req->data = sk;
        req->cancel = _sdp_inet_read_cancel;
        result = 0;
    } else
    {

#if 1 /* performance cheat. LFM */
        if (!(conn->snk_zthresh > size))
        {

            conn->nond_recv--;

            result = sdp_send_ctrl_snk_avail(conn, 0, 0, 0);
            if (0 > result)
            {
                /*
                 * since the message did not go out, back
                 * out the non_discard counter
                 */
                conn->nond_recv++;
            }
        }
#endif

        TS_SDP_CONN_STAT_READ_INC(conn, copied);
        result = copied;
    }
    /*
     * refresh consumed buffers, and/or consume zcopy advertisment
     */
    expect = sdp_recv_flush(conn);
    if (0 > expect)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "RECV: device buffer ack <%d> error. <%d>", copied,
                 expect);
    }

    TS_SDP_CONN_UNLOCK(conn);
    return result;
    error:
    TS_SDP_CONN_UNLOCK(conn);
    /*
     * zero error needs to generate a completion so it's not
     * mistaken as an AIO in progress.
     */
    result = (0 < copied) ? copied : result;

    if (0 == result)
    {

        cb.fn(cb.data, cb.vec, 0);
    }

    return result;
} /* sdp_inet_read */
#endif
