/*         
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  This source code may incorporate intellectual property owned by
  Microsoft Corporation. Our provision of this source code does not
  include any licenses or any other rights to you under any Microsoft
  intellectual property. If you would like a license from Microsoft
  (e.g., to rebrand, redistribute), you need to contact Microsoft
  directly.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sdp_send.c 379 2004-07-08 02:24:32Z roland $
*/

#include "sdp_main.h"


/* --------------------------------------------------------------------- */
/*                                                                       */
/*                          COMMON functions                             */
/*                                                                       */
/* --------------------------------------------------------------------- */
#ifdef _TS_SDP_AIO_SUPPORT
pte_t *__sdp_pte;
pgprot_t __sdp_pgprot;

/* ========================================================================= */
/*.._sdp_inet_write_cancel_func -- lookup function for cancelation */
static tINT32 _sdp_inet_write_cancel_func(struct sdpc_desc *element, void *arg)
{
    struct sdpc_iocb *iocb = (struct sdpc_iocb *) element;
    tINT32 value = (tINT32) (unsigned long)arg;

    TS_CHECK_NULL(element, -EINVAL);

    if (TS_SDP_GENERIC_TYPE_IOCB == element->type && iocb->key == value)
    {

        return 0;
    } else
    {

        return -ERANGE;
    }
} /* _sdp_inet_write_cancel_func */

/* ========================================================================= */
/*.._sdp_inet_write_cancel -- cancel an IO operation */
static tINT32 _sdp_inet_write_cancel(struct kiocb *kiocb
                                     _TS_AIO_UNUSED_CANCEL_PARAM)
{
    struct sock *sk;
    struct sdp_opt *conn;
    struct sdpc_iocb *iocb;
    tINT32 result = 0;

    TS_CHECK_NULL(kiocb, -ERANGE);

    sk = (struct sock *)xchg(&kiocb->data, NULL);
    ts_aio_put_req(kiocb);

    if (NULL == sk)
    {

        TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_INOUT,
                 "SOCK: Cancel empty Kvec kiocb <%d:%d>",
                 kiocb->key, kiocb->users);

        result = -EAGAIN;
        goto done;
    }
    /*
     * lock the socket while we operate.
     */
    conn = TS_SDP_GET_CONN(sk);
    TS_SDP_CONN_LOCK(conn);

    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
             "SOCK: Cancel Kvec <%08x:%04x> <%08x:%04x> <%04x> kiocb <%d:%d>",
             conn->src_addr, conn->src_port, conn->dst_addr,
             conn->dst_port, conn->state, kiocb->key, kiocb->users);
    /*
     * attempt to find the IOCB for this key. we don't have an indication
     * whether this is a read or write.
     */
    iocb = (struct sdpc_iocb *) sdp_desc_q_lookup(&conn->send_queue,
                                                  _sdp_inet_write_cancel_func,
                                                  (void *) (unsigned long)kiocb->
                                                  key);
    if (NULL != iocb)
    {
        /*
         * always remove the IOCB.
         * If active, then place it into the correct active queue
         */
        result = sdp_desc_q_remove((struct sdpc_desc *) iocb);
        TS_EXPECT(MOD_LNX_SDP, !(0 > result));

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
                 "SOCK: Cancel IOCB <%d:%d> flags <%08x>",
                 kiocb->key, iocb->post, iocb->flags);

        if (0 < (TS_SDP_IOCB_F_ACTIVE & iocb->flags))
        {

            if (0 < (TS_SDP_IOCB_F_RDMA_W & iocb->flags))
            {

                result =
                sdp_desc_q_put_tail(&conn->w_snk,
                                    (struct sdpc_desc *)
                                    iocb);
                TS_EXPECT(MOD_LNX_SDP, !(0 > result));
            } else
            {

                TS_EXPECT(MOD_LNX_SDP,
                          (0 <
                           (TS_SDP_IOCB_F_RDMA_R & iocb->
                            flags)));

                result =
                sdp_iocb_q_put_tail(&conn->w_src,
                                    iocb);
                TS_EXPECT(MOD_LNX_SDP, !(0 > result));
            }
        } else
        {
            /*
             * callback to complete IOCB.
             */
            result = sdp_iocb_complete(iocb, 0);
            TS_EXPECT(MOD_LNX_SDP, !(0 > result));

            result = -EAGAIN;
            goto unlock;
        }
    }
    /*
     * check the sink queue, not much to do, since the operation is
     * already in flight.
     */
    iocb = (struct sdpc_iocb *) sdp_desc_q_lookup(&conn->w_snk,
                                                  _sdp_inet_write_cancel_func,
                                                  (void *) (unsigned long)kiocb->
                                                  key);
    if (NULL != iocb)
    {

        iocb->flags |= TS_SDP_IOCB_F_CANCEL;
        result = -EAGAIN;

        goto unlock;
    }
    /*
     * check source queue. If we're ing the source queue, then a cancel
     * needs to be issued.
     */
    iocb = sdp_iocb_q_lookup(&conn->w_src, kiocb->key);
    if (NULL != iocb)
    {
        /*
         * Unfortunetly there is only a course grain cancel in SDP,
         * so we have to cancel everything. This is OKish since it
         * usually only happens at connection termination, and the
         * remaining source probably will get cancel requests as well.
         * The main complexity is to take all the fine grain cancels
         * from AIO and once all out standing Src messages have been
         * cancelled we can issue the course grain SDP cancel. The
         * connection is marked as being in cancel processing so no
         * other writes get into the outbound pipe.
         */
        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "SOCK: Cancellation  <%d:%d> of active Source. <%d:%d>",
                 kiocb->key, kiocb->users,
                 sdp_iocb_q_size(&conn->w_src), conn->src_cncl);

        if (0 == (TS_SDP_CONN_F_SRC_CANCEL_L & conn->flags) &&
            0 == (TS_SDP_IOCB_F_CANCEL & iocb->flags))
        {

            conn->src_cncl++;
            iocb->flags |= TS_SDP_IOCB_F_CANCEL;

            if (conn->src_cncl ==
                sdp_iocb_q_size(&conn->w_src))
            {

                result = sdp_send_ctrl_src_cancel(conn);
                TS_EXPECT(MOD_LNX_SDP, !(0 > result));

                conn->flags |= TS_SDP_CONN_F_SRC_CANCEL_L;
            }
        }

        result = -EAGAIN;
        goto unlock;
    }
    /*
     * no IOCB found.
     */
    TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_INOUT,
             "SOCK: Cancelled kvec with no IOCB. <%d:%d>",
             kiocb->key, kiocb->users);

    if (0 < kiocb->users)
    {
        /*
         * shouldn't occur, but just incase, we don't want to hang
         * the user.
         */
        ts_aio_put_req(kiocb);
        result = 0;
    } else
    {
        /*
         * IOCB should be in completion processing at this point.
         */
        result = -EAGAIN;
    }

    unlock:
    TS_SDP_CONN_UNLOCK(conn);
    done:

    if (0 == result)
    {

        TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_INOUT,
                 "AIO: Kernel bug. Canceling IOCB <%d:%d> leaking <%d> bytes.",
                 kiocb->key, kiocb->users, kiocb->size);
    }

    return result;
} /* _sdp_inet_write_cancel */
#endif

/* ========================================================================= */
/*.._sdp_send_buff_post -- Post a buffer send on a SDP connection. */
static tINT32 _sdp_send_buff_post(struct sdp_opt *conn, struct sdpc_buff *buff)
{
    struct ib_send_param send_param = { 0};
    tINT32 result;

    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(buff, -EINVAL);
    TS_CHECK_NULL(buff->bsdh_hdr, -EINVAL);
    /*
     * write header send buffer.
     */
    conn->r_recv_bf--;
    conn->s_wq_size++;
    conn->l_advt_bf = conn->l_recv_bf;
    conn->send_pipe -= buff->data_size;
    conn->oob_offset -= (0 < conn->oob_offset) ? buff->data_size : 0;

    buff->ib_wrid = conn->send_wrid++;
    buff->lkey = conn->l_key;
    buff->bsdh_hdr->recv_bufs = conn->l_advt_bf;
    buff->bsdh_hdr->size = buff->tail - buff->data;
    buff->bsdh_hdr->seq_num = ++conn->send_seq;
    buff->bsdh_hdr->seq_ack = conn->advt_seq;
    /*
     * endian swap
     */
    result = _sdp_msg_cpu_to_net_bsdh(buff->bsdh_hdr);
    TS_EXPECT(MOD_LNX_SDP, !(0 > result));

    send_param.op = IB_OP_SEND;
    /*
     * OOB processing. If there is a single OOB byte in flight then the
     * pending flag is set as early as possible. IF a second OOB byte
     * becomes queued then the pending flag for that byte will be in the
     * buffer which contains the data. Multiple outstanding OOB messages
     * is not well defined, this way we won't loose any, we'll get early
     * notification in the normal case, we adhear to the protocol, and
     * we don't need to track every message seperatly which would be
     * expensive.
     *
     * If the connections OOB flag is set and the oob
     * counter falls below 64K we set the pending flag, and clear the
     * the flag. This allows for at least one pending urgent message
     * to send early notification.
     */
    if (0 < (TS_SDP_CONN_F_OOB_SEND & conn->flags) &&
        !(0xFFFF < conn->oob_offset))
    {

        TS_SDP_MSG_HDR_SET_OOB_PEND(buff->bsdh_hdr);
        TS_SDP_BUFF_F_SET_SE(buff);

        conn->flags &= ~(TS_SDP_CONN_F_OOB_SEND);
    }
    /*
     * The buffer flag is checked to see if the OOB data is in the buffer,
     * and present flag is set, potentially OOB offset is cleared. pending
     * is set if this buffer has never had pending set.
     */
    if (0 < (TS_SDP_BUFF_F_OOB_PRES & buff->flags))
    {

        if (0 < conn->oob_offset)
        {

            TS_SDP_MSG_HDR_SET_OOB_PEND(buff->bsdh_hdr);
        } else
        {

            TS_EXPECT(MOD_LNX_SDP, !(0 > conn->oob_offset));
            conn->oob_offset = -1;
        }

        TS_SDP_MSG_HDR_SET_OOB_PRES(buff->bsdh_hdr);
        TS_SDP_BUFF_F_SET_SE(buff);
    }
    /*
     * solicite event bit.
     */
    if (0 < TS_SDP_BUFF_F_GET_SE(buff))
    {

        send_param.solicited_event = 1;
    }
    /*
     * unsignalled event
     */
    if (0 < TS_SDP_BUFF_F_GET_UNSIG(buff) &&
        conn->usig_max > conn->send_cons)
    {

        conn->send_usig++;
        conn->send_cons++;
    } else
    {

        TS_SDP_BUFF_F_CLR_UNSIG(buff);
        send_param.signaled = 1;
        conn->send_cons = 0;
    }
    /*
     * check queue membership. (first send attempt vs. flush)
     */
    if (0 < sdp_desc_q_member((struct sdpc_desc *) buff))
    {

        result = sdp_desc_q_remove((struct sdpc_desc *) buff);
        TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    }
    /*
     * save the buffer for the event handler.
     */
    result = sdp_buff_q_put_tail(&conn->send_post, buff);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Error <%d> to queueing send buffer.", result);
        goto done;
    }
    /*
     * post send
     */
    send_param.work_request_id = buff->ib_wrid;
    send_param.gather_list = TS_SDP_BUFF_GAT_SCAT(buff);
    send_param.num_gather_entries = 1;
    result = ib_send(conn->qp, &send_param, 1,NULL);
    if (0 != result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: <%04x> Error <%d> posting send. <%d:%d> <%d:%d:%d>",
                 conn->hashent, result, conn->s_wq_cur, conn->s_wq_size,
                 sdp_buff_q_size(&conn->send_post),
                 sdp_desc_q_size(&conn->r_src),
                 sdp_desc_q_size(&conn->w_snk));

        (void)sdp_buff_q_get_tail(&conn->send_post);
        goto done;
    }
    /*
     * source cancels require us to save the sequence number
     * for validation of the cancel's completion.
     */
    if (0 < (TS_SDP_CONN_F_SRC_CANCEL_L & conn->flags))
    {

        conn->src_cseq =
        ((TS_SDP_MSG_MID_SRC_CANCEL ==
          buff->bsdh_hdr->mid) ? buff->bsdh_hdr->seq_num : conn->
         src_cseq);
    }

    return 0;
    done:
    conn->r_recv_bf++;
    conn->send_seq--;
    conn->s_wq_size--;
    return result;
} /* _sdp_send_buff_post */

/* --------------------------------------------------------------------- */
/*                                                                       */
/*                          DATA functions                               */
/*                                                                       */
/* --------------------------------------------------------------------- */

/* ========================================================================= */
/*.._sdp_send_data_buff_post -- Post data for buffered transmission */
static tINT32 _sdp_send_data_buff_post(struct sdp_opt *conn, struct sdpc_buff *buff)
{
    struct sdpc_advt *advt;
    tINT32 result;

    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(buff, -EINVAL);
    /*
     * check state to determine OK to send:
     *
     * 1) sufficient remote buffer advertisments for data transmission
     * 2) outstanding source advertisments, data must be held.
     * 3) buffer from head of queue or as parameter
     * 4) nodelay check.
     */
    if (3 > conn->r_recv_bf || 0 < conn->src_sent)
    {

        return ENOBUFS;
    }
    /*
     * The rest of the checks can proceed if there is a signalled event
     * in the pipe, otherwise we could stall...
     */
    if (conn->send_usig < sdp_buff_q_size(&conn->send_post) ||
        0 < sdp_desc_q_size(&conn->w_snk))
    {

        if (buff->tail < buff->end &&
            0 == (TS_SDP_BUFF_F_OOB_PRES & buff->flags) &&
            0 == conn->nodelay)
        {
            /*
             * If the buffer is not full, and there is already
             * data in the SDP pipe, then hold on to the buffer
             * to fill it up with more data. If SDP acks clear
             * the pipe they'll grab this buffer, or send will
             * flush once it's full, which ever comes first.
             */
            return ENOBUFS;
        }
        /*
         * slow start to give sink advertisments a chance for
         * asymmetric connections. This is desirable to offload
         * the remote host.
         */
        if (!(conn->s_wq_cur > conn->s_wq_size))
        {
            /*
             * slow down the up take in the send data path to
             * give the remote side some time to post available
             * sink advertisments.
             */
            if (conn->send_cq_size > conn->s_wq_cur)
            {

                if (TS_SDP_DEV_SEND_POST_COUNT > conn->s_wq_par)
                {

                    conn->s_wq_par++;
                } else
                {

                    conn->s_wq_cur++;
                    conn->s_wq_par = 0;
                }
            }

            return ENOBUFS;
        }
    }
    /*
     * setup header.
     */
    buff->data -= sizeof(struct sdp_msg_hdr_bsdh);
    buff->bsdh_hdr = (struct sdp_msg_hdr_bsdh *) buff->data;
    buff->bsdh_hdr->mid = TS_SDP_MSG_MID_DATA;
    buff->bsdh_hdr->flags = TS_SDP_MSG_FLAG_NON_FLAG;
    /*
     * signalled? With no delay turned off, data transmission may be
     * waiting for a send completion.
     */
    TS_SDP_BUFF_F_SET_UNSIG(buff);
    /*
     * update non-discard counter.
     * Make consideration for a pending sink. (can be forced by OOB)
     */
    if (0 < sdp_advt_q_size(&conn->snk_pend))
    {
        /*
         * As sink advertisment needs to be discarded. We always
         * complete an advertisment if there is not enough room
         * for an entire buffers worth of data, this allows us to
         * not need to check how much room is going to be consumed
         * by this buffer, and only one discard is needed.
         * (remember the spec makes sure that the sink is bigger
         * then the buffer.)
         */
        advt = sdp_advt_q_get(&conn->snk_pend);
        TS_EXPECT(MOD_LNX_SDP, (NULL != advt));

        result = sdp_advt_destroy(advt);
        TS_EXPECT(MOD_LNX_SDP, !(0 > result));
        /*
         * update sink advertisments.
         */
        conn->snk_recv--;
    } else
    {

        conn->nond_send++;
    }
    /*
     * transmision time
     */
    result = _sdp_send_buff_post(conn, buff);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Error posting buffered data to SEND queue. <%d>",
                 result);
        goto error;
    }

    return 0;
    error:
    return result;
} /* _sdp_send_data_buff_post */

/* ========================================================================= */
/*.._sdp_send_data_buff_snk -- Post data for buffered transmission */
static tINT32 _sdp_send_data_buff_snk(struct sdp_opt *conn, struct sdpc_buff *buff)
{
    struct ib_send_param send_param = { 0};
    struct sdpc_advt *advt;
    tINT32 result;
    tINT32 zcopy;

    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(buff, -EINVAL);
    /*
     * check state to determine OK to send:
     *
     * 1) sufficient send resources.
     */
    if (!(conn->send_cq_size > conn->s_wq_size))
    {

        return ENOBUFS;
    }
    /*
     * confirm type
     */
    if (TS_SDP_GENERIC_TYPE_BUFF != buff->type)
    {

        return -ENOBUFS;
    }
    /*
     * nodelay buffering
     */
#if 0
    if (buff->tail < buff->end &&
        0 == conn->nodelay &&
        conn->send_usig < sdp_buff_q_size(&conn->send_post))
    {
        /*
         * If the buffer is not full, and there is already data in the
         * SDP pipe, then hold on to the buffer to fill it up with more
         * data. If SDP acks clear the pipe they'll grab this buffer,
         * or send will flush once it's full, which ever comes first.
         */
        return ENOBUFS;
    }
#endif
    /*
     * get advertisment.
     */
    advt = sdp_advt_q_look(&conn->snk_pend);
    if (NULL == advt)
    {

        return ENOBUFS;
    }
    /*
     * signalled? With no delay turned off, data transmission may be
     * waiting for a send completion.
     */
#if 0
    TS_SDP_BUFF_F_SET_UNSIG(buff);
#endif
    /*
     * setup RDMA write
     */
    send_param.op = IB_OP_RDMA_WRITE;
    send_param.remote_address = advt->addr;
    send_param.rkey = advt->rkey;
    send_param.signaled = 1;

    buff->ib_wrid = conn->send_wrid++;
    buff->lkey = conn->l_key;

    advt->wrid = buff->ib_wrid;
    advt->size -= (buff->tail - buff->data);
    advt->addr += (buff->tail - buff->data);
    advt->post += (buff->tail - buff->data);

#ifdef _TS_SDP_DATA_PATH_DEBUG
    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
             "POST: Write BUFF wrid <%08x> of <%u> bytes <%d>.",
             buff->ib_wrid, (buff->tail - buff->data), advt->size);
#endif
    /*
     * dequeue if needed and the queue buffer
     */
    if (0 < sdp_desc_q_member((struct sdpc_desc *) buff))
    {

        result = sdp_desc_q_remove((struct sdpc_desc *) buff);
        TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    }

    result = sdp_desc_q_put_tail(&conn->w_snk, (struct sdpc_desc *) buff);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Error <%d> queueing generic WRITE BUFF. <%d>",
                 result, sdp_desc_q_size(&conn->w_snk));
        goto error;
    }
    /*
     * update send queue depth
     */
    conn->s_wq_size++;
    conn->send_pipe -= buff->data_size;
    conn->oob_offset -= (0 < conn->oob_offset) ? buff->data_size : 0;
    /*
     * post RDMA
     */
    send_param.work_request_id = buff->ib_wrid;
    send_param.gather_list = TS_SDP_BUFF_GAT_SCAT(buff);
    send_param.num_gather_entries = 1;

    result = ib_send(conn->qp, &send_param, 1,NULL);
    if (0 != result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: send (rdma write) failure. <%d>", result);

        conn->s_wq_size--;
        goto error;
    }
    /*
     * If the available space is smaller then send size, complete the
     * advertisment.
     */
    if (conn->send_size > advt->size)
    {

        advt = sdp_advt_q_get(&conn->snk_pend);
        TS_EXPECT(MOD_LNX_SDP, (NULL != advt));

        zcopy = advt->post;

        result = sdp_advt_destroy(advt);
        TS_EXPECT(MOD_LNX_SDP, !(0 > result));

        result = sdp_send_ctrl_rdma_wr(conn, zcopy);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: Error <%d> completing sink advertisment. <%d>",
                     result, zcopy);
            result = -ENODEV;
            goto error;
        }
        /*
         * update sink advertisments.
         */
        conn->snk_recv--;
    }

    return 0;
    error:
    return result;
} /* _sdp_send_data_buff_snk */


/* ========================================================================= */
/*.._sdp_send_data_iocb_src -- send a zcopy read advertisment in the data path */
tINT32 _sdp_send_data_iocb_src(struct sdp_opt *conn, struct sdpc_iocb *iocb)
{
    struct sdp_msg_hdr_srcah *src_ah;
    struct sdpc_buff *buff;
    tINT32 result;

    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(iocb, -EINVAL);
    /*
     * 1) local source cancel is pending
     * 2) sufficient send credits for buffered transmission.
     */
    if (0 < (TS_SDP_CONN_F_SRC_CANCEL_L & conn->flags) ||
        3 > conn->r_recv_bf)
    {

        return ENOBUFS;
    }

    switch (conn->send_mode)
    {
    case TS_SDP_MODE_PIPE:

        if (!(conn->s_cur_adv > conn->src_sent))
        {

            return ENOBUFS;
        }

        if (conn->s_cur_adv < conn->r_max_adv)
        {

            if (!(TS_SDP_SRC_AVAIL_FRACTION > conn->s_par_adv))
            {

                conn->s_cur_adv++;
                conn->s_par_adv = 0;
            } else
            {

                conn->s_par_adv++;
            }
        } else
        {

            conn->s_cur_adv = conn->r_max_adv;
            conn->s_par_adv = 0;
        }
#if 0

        conn->s_cur_adv = ((conn->s_cur_adv < conn->r_max_adv) ?
                           conn->s_cur_adv + 1 : conn->r_max_adv);
#endif
        break;
    case TS_SDP_MODE_COMB:

        if (0 < conn->src_sent)
        {

            return ENOBUFS;
        }

        break;
    default:

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Unexpected SrcAvail mode. <%d>",
                 conn->send_mode);
        return -EPROTO;
    }
    /*
     * get buffer
     */
    buff = sdp_buff_pool_get();
    if (NULL == buff)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to allocate SrcAvail buffer. <%d>",
                 iocb->key);
        return -ENOMEM;
    }
    /*
     * register IOCBs physical memory, we check for previous registration,
     * since SrcAvail revocations can get us to this point multiple times
     * for the same IOCB.
     */
    if (NULL == iocb->page_array)
    {

        result = sdp_iocb_register(iocb, conn);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                     "POST: Error <%d> registering IOCB. <%d:%d>",
                     result, iocb->key, iocb->len);
            goto drop;
        }
    }
    /*
     * format SrcAvail
     */
    buff->tail = buff->data;
    buff->bsdh_hdr = (struct sdp_msg_hdr_bsdh *) buff->data;
    buff->bsdh_hdr->mid = TS_SDP_MSG_MID_SRC_AVAIL;
    buff->bsdh_hdr->flags = TS_SDP_MSG_FLAG_NON_FLAG;
    buff->tail += sizeof(struct sdp_msg_hdr_bsdh);
    src_ah = (struct sdp_msg_hdr_srcah *) buff->tail;
    src_ah->size = iocb->len;
    src_ah->r_key = iocb->r_key;
    src_ah->addr = iocb->io_addr;
    buff->tail += sizeof(struct sdp_msg_hdr_srcah);
    buff->data_size = 0;
    iocb->flags |= TS_SDP_IOCB_F_ACTIVE;
    iocb->flags |= TS_SDP_IOCB_F_RDMA_R;

    TS_SDP_BUFF_F_CLR_SE(buff);
    TS_SDP_BUFF_F_CLR_UNSIG(buff);

    if (TS_SDP_MODE_COMB == conn->send_mode)
    {
#ifdef _TS_SDP_AIO_SUPPORT
        void *vaddr;
        tINT32 offset;
        /*
         * In combined mode, it's a protocol requirment to send at
         * least a byte of data in the SrcAvail.
         */
        if (TS_SDP_SRC_AVAIL_GRATUITOUS > iocb->kvec.src.let->length)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: Error copying data <%d> from IOCB to SRC. <%d:%d>",
                     TS_SDP_SRC_AVAIL_GRATUITOUS,
                     iocb->kvec.src.let->length,
                     iocb->kvec.src.let->offset);

            result = -EFAULT;
            goto error;
        }
        /*
         * map, copy, unmap.
         */
        vaddr = __sdp_kmap(iocb->kvec.src.let->page);
        offset = iocb->kvec.src.let->offset;

        memcpy(buff->tail, (vaddr + offset),
               TS_SDP_SRC_AVAIL_GRATUITOUS);

        __sdp_kunmap(iocb->kvec.src.let->page);
#endif
        /*
         * update pointers
         */
        buff->data_size = TS_SDP_SRC_AVAIL_GRATUITOUS;
        buff->tail += TS_SDP_SRC_AVAIL_GRATUITOUS;
        iocb->len -= TS_SDP_SRC_AVAIL_GRATUITOUS;

        conn->nond_send++;
    }

    conn->src_sent++;
    /*
     * endian swap of extended header
     */
    result = _sdp_msg_cpu_to_net_srcah(src_ah);
    TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    /*
     * queue/send SrcAvail message
     */
    result = _sdp_send_buff_post(conn, buff);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Error <%d> posting Src Avail for IOCB. <%d>",
                 result, iocb->key);
        goto release;
    }

    return 0;
    release:
    conn->nond_send -= (TS_SDP_MODE_COMB == conn->send_mode) ? 1 : 0;
    conn->src_sent--;
#ifdef _TS_SDP_AIO_SUPPORT
    error:
#endif
    iocb->flags &= ~(TS_SDP_IOCB_F_RDMA_R | TS_SDP_IOCB_F_ACTIVE);
    iocb->len += ((TS_SDP_MODE_COMB == conn->send_mode) ?
                  TS_SDP_SRC_AVAIL_GRATUITOUS : 0);
    drop:
    (void)sdp_buff_pool_put(buff);
    return result;
} /* _sdp_send_data_iocb_src */

/* ========================================================================= */
/*.._sdp_send_data_iocb_snk -- process a zcopy write advert in the data path */
tINT32 _sdp_send_data_iocb_snk(struct sdp_opt *conn, struct sdpc_iocb *iocb)
{
    struct ib_send_param send_param = { 0};
    struct ib_gather_scatter sg_val;
    struct sdpc_advt *advt;
    tINT32 result;
    tINT32 zcopy;

    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(iocb, -EINVAL);
    /*
     * register IOCBs physical memory, we check for previous
     * registration, since multiple writes may have been required
     * to fill the advertisement
     */
    if (NULL == iocb->page_array)
    {

        result = sdp_iocb_register(iocb, conn);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                     "POST: Error <%d> registering IOCB. <%d:%d>",
                     result, iocb->key, iocb->len);
            goto error;
        }
    }
    /*
     * check queue depth
     */
    while (0 < iocb->len && conn->send_cq_size > conn->s_wq_size)
    {
        /*
         * get the pending sink advertisment.
         */
        advt = sdp_advt_q_look(&conn->snk_pend);
        if (NULL == advt)
        {

            break;
        }
        /*
         * amount of data to zcopy.
         */
        zcopy = min(advt->size, iocb->len);

        sg_val.address = iocb->io_addr;
        sg_val.key = iocb->l_key;
        sg_val.length = zcopy;

        send_param.op = IB_OP_RDMA_WRITE;
        send_param.remote_address = advt->addr;
        send_param.rkey = advt->rkey;
        send_param.signaled = 1;

        iocb->wrid = conn->send_wrid++;
        iocb->len -= zcopy;
        iocb->post += zcopy;
        iocb->io_addr += zcopy;
        iocb->flags |= TS_SDP_IOCB_F_ACTIVE;
        iocb->flags |= TS_SDP_IOCB_F_RDMA_W;

        advt->wrid = iocb->wrid;
        advt->size -= zcopy;
        advt->addr += zcopy;
        advt->post += zcopy;

#ifdef _TS_SDP_DATA_PATH_DEBUG
        TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
                 "POST: Write IOCB wrid <%08x> of <%u> bytes <%d:%d>.",
                 iocb->wrid, zcopy, iocb->len, advt->size);
#endif
        /*
         * update send queue depth
         */
        conn->s_wq_size++;
        conn->send_pipe -= zcopy;
        conn->oob_offset -= (0 < conn->oob_offset) ? zcopy : 0;
        /*
         * post RDMA
         */
        send_param.work_request_id = iocb->wrid;
        send_param.gather_list = &sg_val;
        send_param.num_gather_entries = 1;

        result = ib_send(conn->qp, &send_param, 1,NULL);
        if (0 != result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: send (rdma write) failure. <%d>",
                     result);

            conn->s_wq_size--;
            goto error;
        }
        /*
         * if there is no more advertised space,  remove the
         * advertisment from the queue, and get it ready for
         * completion. (see note in buffered send during
         * outstanding sink advertisment to see how the advt
         * size remaining is picked.)
         */
        if (conn->send_size > advt->size)
        {

            advt = sdp_advt_q_get(&conn->snk_pend);
            if (NULL == advt)
            {

                TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                         TRACE_FLOW_WARN,
                         "POST: sinkc available advertisment disappeared.");
                result = -ENODEV;
                goto error;
            }

            zcopy = advt->post;

            result = sdp_advt_destroy(advt);
            TS_EXPECT(MOD_LNX_SDP, !(0 > result));

            result = sdp_send_ctrl_rdma_wr(conn, zcopy);
            if (0 > result)
            {

                TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                         TRACE_FLOW_WARN,
                         "POST: Error <%d> completing sink advertisment. <%d>",
                         result, zcopy);
                result = -ENODEV;
                goto error;
            }
            /*
             * update sink advertisments.
             */
            conn->snk_recv--;
        }
    }

    return iocb->len;
    error:
    return result;
} /* _sdp_send_data_iocb_snk */

/* ========================================================================= */
/*.._sdp_send_data_zcopy_src -- send a zcopy read advertisment in the data path */
tINT32 _sdp_send_data_zcopy_src(struct sdp_opt *conn, struct sdpc_iocb *iocb)
{
    struct sdp_msg_hdr_srcah *src_ah;
    struct sdpc_buff *buff;
    tINT32 result;

    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(iocb, -EINVAL);
    /*
     * 1) local source cancel is pending
     * 2) sufficient send credits for buffered transmission.
     */




    if (0 < (TS_SDP_CONN_F_SRC_CANCEL_L & conn->flags) ||
        3 > conn->r_recv_bf)
    {
        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN, "no buffers");

        return ENOBUFS;
    }


    switch (conn->send_mode)
    {
    case TS_SDP_MODE_PIPE:

        if (!(conn->s_cur_adv > conn->src_sent))
        {

            return ENOBUFS;
        }

        if (conn->s_cur_adv < conn->r_max_adv)
        {

            if (!(TS_SDP_SRC_AVAIL_FRACTION > conn->s_par_adv))
            {

                conn->s_cur_adv++;
                conn->s_par_adv = 0;
            } else
            {

                conn->s_par_adv++;
            }
        } else
        {

            conn->s_cur_adv = conn->r_max_adv;
            conn->s_par_adv = 0;
        }
#if 0

        conn->s_cur_adv = ((conn->s_cur_adv < conn->r_max_adv) ?
                           conn->s_cur_adv + 1 : conn->r_max_adv);
#endif
        break;
    case TS_SDP_MODE_COMB:

        if (0 < conn->src_sent)
        {
            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,"src_sent is more than 1");

            return ENOBUFS;
        }

        break;
    default:

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Unexpected SrcAvail mode. <%d>",
                 conn->send_mode);
        return -EPROTO;
    }
    /*
     * get buffer
     */

    buff = sdp_buff_pool_get();
    if (NULL == buff)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to allocate SrcAvail buffer. <%d>",
                 iocb->key);
        return -ENOMEM;
    }
    /*
  * format SrcAvail
  */


    buff->tail = buff->data;
    buff->bsdh_hdr = (struct sdp_msg_hdr_bsdh *) buff->data;
    buff->bsdh_hdr->mid = TS_SDP_MSG_MID_SRC_AVAIL;
    buff->bsdh_hdr->flags = TS_SDP_MSG_FLAG_NON_FLAG;
    buff->tail += sizeof(struct sdp_msg_hdr_bsdh);
    src_ah = (struct sdp_msg_hdr_srcah *) buff->tail;
    src_ah->size = iocb->len;
    src_ah->r_key = iocb->r_key;
    src_ah->addr = iocb->io_addr;
  TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,"src_avail size is %d",src_ah->size);
    buff->tail += sizeof(struct sdp_msg_hdr_srcah);
    buff->data_size = 0;
    iocb->flags |= TS_SDP_IOCB_F_ACTIVE;
    iocb->flags |= TS_SDP_IOCB_F_RDMA_R;

    TS_SDP_BUFF_F_CLR_SE(buff);
    TS_SDP_BUFF_F_CLR_UNSIG(buff);

    if (TS_SDP_MODE_COMB == conn->send_mode)
    {

        /*
         * In combined mode, it's a protocol requirment to send at
         * least a byte of data in the SrcAvail.
         */
        if (TS_SDP_SRC_AVAIL_GRATUITOUS > iocb->msg->msg_iov->iov_len)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: Error copying data <%d> from IOCB to SRC. <%d>",
                     TS_SDP_SRC_AVAIL_GRATUITOUS,
                     iocb->msg->msg_iov->iov_len);

            result = -EFAULT;
            goto error;
        }
        memcpy(buff->tail, iocb->msg->msg_iov,TS_SDP_SRC_AVAIL_GRATUITOUS );

        /*
         * update pointers
         */ 
        buff->data_size = TS_SDP_SRC_AVAIL_GRATUITOUS;
        buff->tail += TS_SDP_SRC_AVAIL_GRATUITOUS;
        iocb->len -= TS_SDP_SRC_AVAIL_GRATUITOUS;

        conn->nond_send++;
    }

    conn->src_sent++;
    /*
     * endian swap of extended header
     */
    result = _sdp_msg_cpu_to_net_srcah(src_ah);
    TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    /*
     * queue/send SrcAvail message
     */

    result = _sdp_send_buff_post(conn, buff); 
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Error <%d> posting Src Avail for IOCB. <%d>",
                 result, iocb->key);
        goto release;
    }

    if (0 == result)
    {


        /*
         * queue IOCB
         */



        if (0 <
            sdp_desc_q_member((struct sdpc_desc *) iocb))
        {


            result =
            sdp_desc_q_remove((struct sdpc_desc *) iocb);
            TS_EXPECT(MOD_LNX_SDP,
                      !(0 > result));
        }


        result =
        sdp_iocb_q_put_tail(&conn->
                            w_src,
                            iocb);


        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                     TRACE_FLOW_WARN,
                     "WRITE: Error <%d> queueing write <%d> to active table",
                     result, iocb->key,
                     sdp_iocb_q_size
                     (&conn->w_src));
        }


    }



    return 0;
    release:
    conn->nond_send -= (TS_SDP_MODE_COMB == conn->send_mode) ? 1 : 0;
    conn->src_sent--;
    error:
    iocb->flags &= ~(TS_SDP_IOCB_F_RDMA_R | TS_SDP_IOCB_F_ACTIVE);
    iocb->len += ((TS_SDP_MODE_COMB == conn->send_mode) ?
                  TS_SDP_SRC_AVAIL_GRATUITOUS : 0);
    (void)sdp_buff_pool_put(buff);
    return result;
} /* _sdp_send_data_zcopy_src */

#ifdef _TS_SDP_AIO_SUPPORT
/* ========================================================================= */
/*.._sdp_send_data_iocb_buff_kvec -- write into a SDP buffer from a kvec */
static tINT32 _sdp_send_data_iocb_buff_kvec(struct kvec_dst *src, 
                                            struct sdpc_buff *buff,
                                            tINT32 len)
{
    void *tail;
    tINT32 part;
    tINT32 left;
    tINT32 copy;

    TS_CHECK_NULL(src, -EINVAL);
    TS_CHECK_NULL(buff, -EINVAL);
    /*
     * activate source memory kmap
     */
    TS_SDP_KVEC_DST_MAP(src);
    /*
     * copy from source to buffer
     */
    copy = min(len, (tINT32) (buff->end - buff->tail));
    tail = buff->tail;

    for (left = copy; 0 < left;)
    {

        part = min(left, src->space);
#ifndef _TS_SDP_DATA_PATH_NULL
        memcpy(buff->tail, src->dst, part);
#endif

        buff->tail += part;
        src->space -= part;
        src->dst += part;
        left -= part;

        if (0 < left && 0 == src->space)
        {

            TS_SDP_KVEC_DST_UNMAP(src);
            src->let++;
            src->offset = 0;
            TS_SDP_KVEC_DST_MAP(src);

            if (!(0 < src->space))
            {    /* sanity check */

                TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                         TRACE_FLOW_WARN,
                         "SEND: Kvec to IOCB sanity check failed. <%d:%d:%d:%d>",
                         len, left, copy, src->space);

                buff->tail = tail;
                copy = -EFAULT;
                goto error;
            }
        }
    }

    error:
    /*
     * release destination memory kmap
     */
    TS_SDP_KVEC_DST_UNMAP(src);

    return copy;
} /* _sdp_send_data_iocb_buff_kvec */
#endif

/* ========================================================================= */
/*.._sdp_send_data_iocb_buff -- write multiple SDP buffers from an ioc */
static tINT32 _sdp_send_data_iocb_buff(struct sdp_opt *conn, struct sdpc_iocb *iocb)
{
    struct sdpc_buff *buff;
    tINT32 copy;
    tINT32 partial = 0;
    tINT32 result;
    tINT32 w_space;

    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(iocb, -EINVAL);

    if (0 < conn->src_sent)
    {

        return ENOBUFS;
    }
    /*
     * loop through queued buffers and copy them to the destination
     */
    w_space = __sdp_inet_write_space(conn, 0);

    while (0 < w_space &&
           0 < iocb->len &&
           2 < conn->r_recv_bf &&
           conn->send_cq_size > conn->s_wq_size)
    {
        /*
         * get a buffer for posting.
         */
        buff = sdp_buff_pool_get();
        if (NULL == buff)
        {

            result = -ENOMEM;
            goto error;
        }
        /*
         * setup header.
         */
        buff->tail = buff->end - conn->send_size;
        buff->data = buff->tail;

        buff->data -= sizeof(struct sdp_msg_hdr_bsdh);
        buff->bsdh_hdr = (struct sdp_msg_hdr_bsdh *) buff->data;
        buff->bsdh_hdr->mid = TS_SDP_MSG_MID_DATA;
        buff->bsdh_hdr->flags = TS_SDP_MSG_FLAG_NON_FLAG;

        TS_SDP_BUFF_F_CLR_SE(buff);
        TS_SDP_BUFF_F_CLR_UNSIG(buff);

        copy = min(iocb->len, w_space);
        /*
         * TODO: need to be checking OOB here.
         */
#ifdef _TS_SDP_AIO_SUPPORT
        partial =
        _sdp_send_data_iocb_buff_kvec(&iocb->kvec.src, buff, copy);
        if (0 > partial)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "WRITE: error <%d> copying data <%d> from vector.",
                     partial, copy);

            result = partial;
            goto drop;
        }
#endif

        buff->data_size += partial;
        conn->send_qud += partial;
        iocb->len -= partial;
        iocb->post += partial;
        w_space -= partial;

        conn->nond_send++;
        /*
         * transmision time. An update of send_pipe is not needed,
         * since the IOCB queue took care of the increment.
         */
        result = _sdp_send_buff_post(conn, buff);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: Error posting buffered data to SEND queue. <%d>",
                     result);
            goto drop;
        }
    }

    return iocb->len;
    drop:
    (void)sdp_buff_pool_put(buff);
    error:
    return result;
} /* _sdp_send_data_iocb_buff */

/* ========================================================================= */
/*.._sdp_send_data_iocb -- Post IOCB data for  transmission */
static tINT32 _sdp_send_data_iocb(struct sdp_opt *conn, struct sdpc_iocb *iocb)
{
    tINT32 result;

    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(iocb, -EINVAL);

    if (!(conn->send_cq_size > conn->s_wq_size))
    {

        return ENOBUFS;
    }
    /*
     * confirm IOCB usage.
     */
    if (TS_SDP_GENERIC_TYPE_IOCB != iocb->type)
    {

        return -ENOBUFS;
    }
    /*
     * determin if we are sending Buffered, Source or Sink.
     */
    if (0 < sdp_advt_q_size(&conn->snk_pend))
    {

        result = _sdp_send_data_iocb_snk(conn, iocb);
        if (0 == result)
        {
            /*
             * IOCB completely processed. Otherwise we allow the
             * callers to determine the fate of the IOCB on
             * failure or partial processing.
             */
            if (0 < sdp_desc_q_member((struct sdpc_desc *) iocb))
            {

                result =
                sdp_desc_q_remove((struct sdpc_desc *)
                                  iocb);
                TS_EXPECT(MOD_LNX_SDP, !(0 > result));
            }

            result =
            sdp_desc_q_put_tail(&conn->w_snk,
                                (struct sdpc_desc *) iocb);
            if (0 > result)
            {

                TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                         TRACE_FLOW_WARN,
                         "POST: Error <%d> queuing active write IOCB.",
                         result);
            }
        }
    } else
    {
        /*
         * If there are active sink IOCBs we want to stall, in the
         * hope that a new sink advertisment will arrive, because 
         * sinks are more efficient.
         */
        if (0 == conn->sink_actv)
        {

            if (conn->src_zthresh > iocb->len ||
                TS_SDP_MODE_BUFF == conn->send_mode ||
                0 < (TS_SDP_IOCB_F_BUFF & iocb->flags))
            {

                result = _sdp_send_data_iocb_buff(conn, iocb);
                if (0 == result)
                {
                    /*
                     * complete this IOCB
                     */
                    if (0 <
                        sdp_desc_q_member((struct sdpc_desc *) iocb))
                    {

                        result =
                        sdp_desc_q_remove((struct sdpc_desc *) iocb);
                        TS_EXPECT(MOD_LNX_SDP,
                                  !(0 > result));
                    }

                    TS_SDP_CONN_STAT_WRITE_INC(conn,
                                               iocb->post);
                    TS_SDP_CONN_STAT_WQ_DEC(conn,
                                            iocb->size);

                    result = sdp_iocb_complete(iocb, 0);
                    if (0 > result)
                    {

                        TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                                 TRACE_FLOW_WARN,
                                 "WRITE: error <%d> completing iocb. <%d>",
                                 result, iocb->key);
                    }
                }
            } else
            {

                result = _sdp_send_data_iocb_src(conn, iocb);
                if (0 == result)
                {
                    /*
                     * queue IOCB
                     */
                    if (0 <
                        sdp_desc_q_member((struct sdpc_desc *) iocb))
                    {

                        result =
                        sdp_desc_q_remove((struct sdpc_desc *) iocb);
                        TS_EXPECT(MOD_LNX_SDP,
                                  !(0 > result));
                    }

                    result =
                    sdp_iocb_q_put_tail(&conn->
                                        w_src,
                                        iocb);
                    if (0 > result)
                    {

                        TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                                 TRACE_FLOW_WARN,
                                 "WRITE: Error <%d> queueing write <%d> to active table",
                                 result, iocb->key,
                                 sdp_iocb_q_size
                                 (&conn->w_src));
                    }
                }
            }
        } else
        {

            result = ENOBUFS;
        }
    }

    return result;
} /* _sdp_send_data_iocb */

/* ========================================================================= */
/*.._sdp_send_data_queue_test -- send data buffer if conditions are met */
static tINT32 _sdp_send_data_queue_test(struct sdp_opt *conn, struct sdpc_desc *element)
{
    tINT32 result;

    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(element, -EINVAL);
    /*
     * Notify caller to buffer data:
     * 1) Invalid state for transmission
     * 2) source advertisment cancel in progress.
     */



    if (0 == (TS_SDP_ST_MASK_SEND_OK & conn->state) ||
        0 < (TS_SDP_CONN_F_SRC_CANCEL_L & conn->flags))
    {

        return ENOBUFS;
    }
    if (TS_SDP_GENERIC_TYPE_BUFF == element->type)
    {

        if (0 == sdp_advt_q_look(&conn->snk_pend) ||
            0 <
            (TS_SDP_BUFF_F_OOB_PRES & ((struct sdpc_buff *) element)->flags))
        {

            result =
            _sdp_send_data_buff_post(conn, (struct sdpc_buff *) element);
        } else
        {

            result =
            _sdp_send_data_buff_snk(conn, (struct sdpc_buff *) element);
        }
    } else
    {
        if (TS_SDP_GENERIC_TYPE_IOCB == element->type)
        {
            result = _sdp_send_data_iocb(conn, (struct sdpc_iocb *) element);
        } else
        {
            TS_EXPECT(MOD_LNX_SDP,
                      (TS_SDP_GENERIC_TYPE_ZCOPY == element->type));
            if (!(conn->send_cq_size > conn->s_wq_size))
            {

                return ENOBUFS;
            }
            result = _sdp_send_data_zcopy_src(conn,  (struct sdpc_iocb *) element);  
        }
    }
    return result;
} /* _sdp_send_data_queue_test */

/* ========================================================================= */
/*.._sdp_send_data_queue_flush -- Flush data from send queue, to send post. */
static tINT32 _sdp_send_data_queue_flush(struct sdp_opt *conn)
{
    struct sdpc_desc *element;
    tINT32 result = 0;

    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * As long as there is data, try to post buffered data, until a
     * non-zero result is generated.
     * (positive: no space; negative: error)
     */
    while (0 < sdp_desc_q_size(&conn->send_queue))
    {

        element = sdp_desc_q_look_head(&conn->send_queue);
        TS_EXPECT(MOD_LNX_SDP, (NULL != element));

        result = _sdp_send_data_queue_test(conn, element);
        if (0 != result)
        {
            /*
             * error
             */
            if (0 > result)
            {

                TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                         TRACE_FLOW_WARN,
                         "POST: failed to post data message during flush. <%d>",
                         result);
                /*
                 * check for dangling element reference,
                 * since called functions can dequeue the
                 * element, and not know how to requeue it.
                 */
                if (0 == sdp_desc_q_member(element))
                {

                    result =
                    sdp_desc_q_put_head(&conn->
                                        send_queue,
                                        element);
                    TS_EXPECT(MOD_LNX_SDP, !(0 > result));
                }
            }

            break;
        }
    }

    return result;
} /* _sdp_send_data_queue_flush */

/* ========================================================================= */
/*.._sdp_send_data_queue -- send using the data queue if necessary. */

static tINT32 _sdp_send_data_queue(struct sdp_opt *conn, struct sdpc_desc *element)
{
    tINT32 result = 0;

    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(element, -EINVAL);
    /*
     * If data is being buffered, save and return, send/recv completions
     * will flush the queue.
     * If data is not being buffered, attempt to send, a positive result
     * requires us to buffer, a negative result is an error, a return
     * value of zero is a successful transmission
     */
    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN, "send queue size is %d",sdp_desc_q_size(&conn->send_queue));

    if (0 < sdp_desc_q_size(&conn->send_queue) ||
        0 < (result = _sdp_send_data_queue_test(conn, element)))
    {

        result = sdp_desc_q_put_tail(&conn->send_queue, element);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: Error <%d> queueing data buffer for later send post.",
                     result);
            goto done;
        }
        /*
         * Potentially request a switch to pipelined mode.
         */
        if (TS_SDP_MODE_COMB == conn->send_mode &&
            !(TS_SDP_DEV_SEND_BACKLOG >
              sdp_desc_q_size(&conn->send_queue)))
        {

            result =
            sdp_send_ctrl_mode_ch(conn,
                                  TS_SDP_MSG_MCH_PIPE_RECV);
            if (0 > result)
            {

                TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                         TRACE_FLOW_WARN,
                         "POST: Error <%d> posting mode change. <%d>",
                         result, conn->send_mode);
                goto done;
            }
        }
    }

    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failure during data send posting. <%d>",
                 result);
        goto done;
    }

    return 0;
    done:
    return result;
} /* _sdp_send_data_queue */

/* ========================================================================= */
/*.._sdp_send_data_buff_get -- get an appropriate write buffer for send */
static __inline__ struct sdpc_buff *_sdp_send_data_buff_get(struct sdp_opt *conn)
{
    struct sdpc_buff *buff;

    TS_CHECK_NULL(conn, NULL);
    /*
     * If there is no available buffer get a new one.
     */
    buff = (struct sdpc_buff *) sdp_desc_q_look_type_tail(&conn->send_queue,
                                                          TS_SDP_GENERIC_TYPE_BUFF);
    if (NULL == buff ||
        buff->tail == buff->end ||
        0 < (TS_SDP_BUFF_F_OOB_PRES & buff->flags))
    {

        buff = sdp_buff_pool_get();
        if (NULL != buff)
        {

            buff->tail = buff->end - conn->send_size;
            buff->data = buff->tail;
#if 0
            /*
             * validate that there is room for the header.
             * sanity check.
             */
            if (TS_SDP_MSG_HDR_SIZE > (buff->data - buff->head))
            {

                TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                         TRACE_FLOW_WARN,
                         "POST: send, not enough room in buffer for headers. <%d:%d>",
                         TS_SDP_MSG_HDR_SIZE,
                         (buff->data - buff->head));

                (void)sdp_buff_pool_put(buff);
                buff = NULL;
            }
#endif
        }
    }

    return buff;
} /* _sdp_send_data_buff_get */

/* ========================================================================= */
/*.._sdp_send_data_buff_put -- place a buffer into the send queue */
static __inline__ tINT32 _sdp_send_data_buff_put(struct sdp_opt *conn,
                                                 struct sdpc_buff *buff,
                                                 tINT32 size,
                                                 tINT32 urg)
{
    tINT32 result = 0;
    tINT32 expect;

    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(buff, -EINVAL);
    /*
     * See note on send OOB implementation in SendBuffPost.
     */
    if (0 < urg)
    {

        buff->flags |= TS_SDP_BUFF_F_OOB_PRES;
        /*
         * The OOB PEND and PRES flags need to match up as pairs.
         */
        if (0 > conn->oob_offset)
        {

            conn->oob_offset = conn->send_pipe + size;
            conn->flags |= TS_SDP_CONN_F_OOB_SEND;
        }
    }
    /*
     * if the buffer is already queue, then this was a fill of a partial
     * buffer and dosn't need to be queued now.
     */
    if (0 < (TS_SDP_BUFF_F_QUEUED & buff->flags))
    {

        buff->data_size += size;
        conn->send_qud += size;
        conn->send_pipe += size;
    } else
    {

        buff->data_size = buff->tail - buff->data;
        conn->send_qud += buff->data_size;
        conn->send_pipe += buff->data_size;

        buff->flags |= TS_SDP_BUFF_F_QUEUED;
        /*
         * finally send the data buffer
         */
        result = _sdp_send_data_queue(conn, (struct sdpc_desc *) buff);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: Error posting buffered data to SEND queue. <%d>",
                     result);

            expect = sdp_buff_pool_put(buff);
            TS_EXPECT(MOD_LNX_SDP, !(0 > expect));
        }
    }

    return result;
} /* _sdp_send_data_buff_put */

/* --------------------------------------------------------------------- */
/*                                                                       */
/*                          CONTROL functions                            */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*.._sdp_send_ctrl_buff_test -- determine if it's OK to post a control msg */
static tINT32 _sdp_send_ctrl_buff_test(struct sdp_opt *conn, struct sdpc_buff *buff)
{
    tINT32 result = 0;

    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(buff, -EINVAL);

    if (0 == (TS_SDP_ST_MASK_CTRL_OK & conn->state) ||
        !(conn->send_cq_size > conn->s_wq_size) ||
        !(0 < conn->r_recv_bf) ||
        (conn->l_recv_bf == conn->l_advt_bf && 1 == conn->r_recv_bf))
    {

        return ENOBUFS;
    }
    /*
     * post the control buffer
     */
    result = _sdp_send_buff_post(conn, buff);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: failed to post control send. <%d>", result);
        goto error;
    }

    return 0;
    error:
    return result;
} /* _sdp_send_ctrl_buff_test */

/* ========================================================================= */
/*.._sdp_send_ctrl_buff_flush -- Flush control buffers, to send post. */
static tINT32 _sdp_send_ctrl_buff_flush(struct sdp_opt *conn)
{
    struct sdpc_desc *element;
    tINT32 result = 0;

    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * As long as there are buffers, try to post  until a non-zero
     * result is generated. (positive: no space; negative: error)
     */
    while (0 < sdp_desc_q_size(&conn->send_ctrl))
    {

        element = sdp_desc_q_look_head(&conn->send_ctrl);
        TS_EXPECT(MOD_LNX_SDP, (NULL != element));

        result = _sdp_send_ctrl_buff_test(conn, (struct sdpc_buff *) element);
        if (0 != result)
        {

            if (0 > result)
            {

                TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                         TRACE_FLOW_WARN,
                         "POST: failed to post control message on flush. <%d>",
                         result);

                if (0 == sdp_desc_q_member(element))
                {

                    result =
                    sdp_desc_q_put_head(&conn->
                                        send_ctrl,
                                        element);
                    TS_EXPECT(MOD_LNX_SDP, !(0 > result));
                }
            }

            break;
        }
    }

    return result;
} /* _sdp_send_ctrl_buff_flush */

/* ========================================================================= */
/*.._sdp_send_ctrl_buff_buffered -- Send a buffered control message. */
static tINT32 _sdp_send_ctrl_buff_buffered(struct sdp_opt *conn, struct sdpc_buff *buff)
{
    tINT32 result = 0;

    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(buff, -EINVAL);
    /*
     * Either post a send, or buffer the packet in the tx queue
     */
    if (0 < sdp_desc_q_size(&conn->send_ctrl) ||
        0 < (result = _sdp_send_ctrl_buff_test(conn, buff)))
    {
        /*
         * save the buffer for later flushing into the post queue.
         */
        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,  "saving completion");

        result =
        sdp_desc_q_put_tail(&conn->send_ctrl,
                            (struct sdpc_desc *) buff);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: Error <%d> queueing control buffer for later send post.",
                     result);
            goto error;
        }
    }

    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failure during control send posting. <%d>",
                 result);
        goto error;
    }

    return 0;
    error:
    return result;
} /* _sdp_send_ctrl_buff_buffered */

/* ========================================================================= */
/*.._sdp_send_ctrl_buff -- Create and Send a buffered control message. */
static tINT32 _sdp_send_ctrl_buff(struct sdp_opt *conn,
                                  tUINT8 mid,
                                  tBOOLEAN se,
                                  tBOOLEAN sig)
{
    tINT32 result = 0;
    struct sdpc_buff *buff;

    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * create the message, which contains just the bsdh header.
     * (don't need to worry about header space reservation)
     */
    buff = sdp_buff_pool_get();
    if (NULL == buff)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to allocaate buffer for Disconnect.");
        result = -ENOMEM;
        goto error;
    }
    /*
     * setup header.
     */
    buff->bsdh_hdr = (struct sdp_msg_hdr_bsdh *) buff->data;
    buff->bsdh_hdr->mid = mid;
    buff->bsdh_hdr->flags = TS_SDP_MSG_FLAG_NON_FLAG;
    buff->tail = buff->data + sizeof(struct sdp_msg_hdr_bsdh);
    buff->data_size = 0;
    /*
     * solicite event flag for IB sends.
     */
    if (TRUE == se)
    {

        TS_SDP_BUFF_F_SET_SE(buff);
    } else
    {

        TS_SDP_BUFF_F_CLR_SE(buff);
    }
    /*
     * try for unsignalled?
     */
    if (TRUE == sig)
    {

        TS_SDP_BUFF_F_CLR_UNSIG(buff);
    } else
    {

        TS_SDP_BUFF_F_SET_UNSIG(buff);
    }
    /*
     * Either post a send, or buffer the packet in the tx queue
     */
    result = _sdp_send_ctrl_buff_buffered(conn, buff);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to post buffered control message. <%d>",
                 result);
        (void)sdp_buff_pool_put(buff);
    }

    error:
    return result;
} /* _sdp_send_ctrl_buff */

/* ========================================================================= */
/*.._sdp_send_ctrl_disconnect -- Send a disconnect request. */
static tINT32 _sdp_send_ctrl_disconnect(struct sdp_opt *conn)
{
    tINT32 result = 0;
    struct sdpc_buff *buff;

    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * create the disconnect message, which contains just the bsdh header.
     * (don't need to worry about header space reservation)
     */
    buff = sdp_buff_pool_get();
    if (NULL == buff)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to allocaate buffer for Disconnect.");
        result = -ENOMEM;
        goto error;
    }
    /*
     * setup header.
     */
    buff->bsdh_hdr = (struct sdp_msg_hdr_bsdh *) buff->data;
    buff->bsdh_hdr->mid = TS_SDP_MSG_MID_DISCONNECT;
    buff->bsdh_hdr->flags = TS_SDP_MSG_FLAG_NON_FLAG;
    buff->tail = buff->data + sizeof(struct sdp_msg_hdr_bsdh);
    buff->data_size = 0;

    TS_SDP_BUFF_F_CLR_SE(buff);
    TS_SDP_BUFF_F_CLR_UNSIG(buff);
    /*
     * change state to reflect disconnect is queued.
     * DIS_PEND_X to DIS_SEND_X
     */
    TS_SDP_ST_PEND_2_SEND(conn);
    /*
     * send
     */
    result = _sdp_send_ctrl_buff_buffered(conn, buff);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: failed to post data send for disconnect. <%d>",
                 result);
        goto error;
    }

    return 0;
    error:
    return result;
} /* _sdp_send_ctrl_disconnect */

/* ========================================================================= */
/*..sdp_send_ctrl_disconnect -- potentially send a disconnect request. */
tINT32 sdp_send_ctrl_disconnect(struct sdp_opt *conn)
{
    tINT32 result;

    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * Only create/post the message if there is no data in the data queue,
     * otherwise ignore the call. The flush queue will see to it, that a
     * Disconnect message gets queued/sent once the data queue is flushed
     * clean. The state is now in a disconnect send, the message will be
     * sent once data is flushed.
     */
    if (0 < (TS_SDP_ST_MASK_DIS_PEND & conn->state))
    {

        if (0 == (TS_SDP_CONN_F_DIS_HOLD & conn->flags) &&
            0 == sdp_desc_q_size(&conn->send_queue) &&
            0 == conn->src_sent)
        {

            result = _sdp_send_ctrl_disconnect(conn);

            TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,"send cntrl dissconnect return %d",result);
        } else
        {

            result = 0;
        }
    } else
    {

        result = -EPROTO;
    }
    return result;
} /* sdp_send_ctrl_disconnect */

/* ========================================================================= */
/*..sdp_send_ctrl_ack -- Send a gratuitous Ack. */
tINT32 sdp_send_ctrl_ack(struct sdp_opt *conn)
{
    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * The gratuitous ack is not really and ack, but an update of the
     * number of buffers posted for receive. Important when traffic is
     * only moving in one direction. We check to see if the buffer
     * credits are going to be sent in an already scheduled message
     * before posting. The send queue has different constraints/stalling
     * conditions then the control queue, so there is more checking to
     * be done, then whether there is data in the queue.
     */
    if (0 < sdp_desc_q_size(&conn->send_ctrl) ||
        (0 < sdp_desc_q_size(&conn->send_queue) &&
         2 < conn->l_advt_bf))
    {

        return 0;
    }

    return _sdp_send_ctrl_buff(conn, TS_SDP_MSG_MID_DATA, FALSE, FALSE);
} /* sdp_send_ctrl_ack */

/* ========================================================================= */
/*..sdp_send_ctrl_send_sm -- Send a request for buffered mode. */
tINT32 sdp_send_ctrl_send_sm(struct sdp_opt *conn)
{
    return _sdp_send_ctrl_buff(conn, TS_SDP_MSG_MID_SEND_SM, TRUE, TRUE);
} /* sdp_send_ctrl_send_sm */

/* ========================================================================= */
/*..sdp_send_ctrl_src_cancel -- Send a source cancel */
tINT32 sdp_send_ctrl_src_cancel(struct sdp_opt *conn)
{
    return _sdp_send_ctrl_buff(conn, TS_SDP_MSG_MID_SRC_CANCEL, TRUE, TRUE);
} /* sdp_send_ctrl_src_cancel */

/* ========================================================================= */
/*..sdp_send_ctrl_snk_cancel -- Send a sink cancel */
tINT32 sdp_send_ctrl_snk_cancel(struct sdp_opt *conn)
{
    return _sdp_send_ctrl_buff(conn, TS_SDP_MSG_MID_SNK_CANCEL, TRUE, TRUE);
} /* sdp_send_ctrl_snk_cancel */

/* ========================================================================= */
/*..sdp_send_ctrl_snk_cancel_ack -- Send an ack for a sink cancel */
tINT32 sdp_send_ctrl_snk_cancel_ack(struct sdp_opt *conn)
{
    return _sdp_send_ctrl_buff(conn, TS_SDP_MSG_MID_SNK_CANCEL_ACK, TRUE,
                               TRUE);
} /* sdp_send_ctrl_snk_cancel_ack */

/* ========================================================================= */
/*..sdp_send_ctrl_abort -- Send an abort message. */
tINT32 sdp_send_ctrl_abort(struct sdp_opt *conn)
{
    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * send
     */
    return _sdp_send_ctrl_buff(conn, TS_SDP_MSG_MID_ABORT_CONN, TRUE, TRUE);
} /* sdp_send_ctrl_abort */

/* ========================================================================= */
/*..sdp_send_ctrl_resize_buff_ack -- Send an ack for a buffer size change */
tINT32 sdp_send_ctrl_resize_buff_ack(struct sdp_opt *conn, tUINT32 size)
{
    struct sdp_msg_hdr_crbah *crbah;
    tINT32 result = 0;
    struct sdpc_buff *buff;

    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * create the message, which contains just the bsdh header.
     * (don't need to worry about header space reservation)
     */
    buff = sdp_buff_pool_get();
    if (NULL == buff)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to allocaate buffer for Disconnect.");
        result = -ENOMEM;
        goto error;
    }
    /*
     * setup header.
     */
    buff->tail = buff->data;
    buff->bsdh_hdr = (struct sdp_msg_hdr_bsdh *) buff->tail;
    buff->bsdh_hdr->mid = TS_SDP_MSG_MID_CH_RECV_BUF_ACK;
    buff->bsdh_hdr->flags = TS_SDP_MSG_FLAG_NON_FLAG;
    buff->tail += sizeof(struct sdp_msg_hdr_bsdh);
    crbah = (struct sdp_msg_hdr_crbah *) buff->tail;
    crbah->size = size;
    buff->tail += sizeof(struct sdp_msg_hdr_crbah);

    TS_SDP_BUFF_F_CLR_SE(buff);
    TS_SDP_BUFF_F_CLR_UNSIG(buff);
    /*
     * endian swap of extended header
     */
    result = _sdp_msg_cpu_to_net_crbah(crbah);
    TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    /*
     * Either post a send, or buffer the packet in the tx queue
     */
    result = _sdp_send_ctrl_buff_buffered(conn, buff);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to post buffered control message. <%d>",
                 result);
        (void)sdp_buff_pool_put(buff);
    }

    error:
    return result;
} /* sdp_send_ctrl_resize_buff_ack */

/* ========================================================================= */
/*..sdp_send_ctrl_rdma_rd -- Send an rdma read completion */
tINT32 sdp_send_ctrl_rdma_rd(struct sdp_opt *conn, tINT32 size)
{
    struct sdp_msg_hdr_rrch *rrch;
    tINT32 result = 0;
    struct sdpc_buff *buff;

    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * check size
     */
    if (0 > size)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: RDMA read completion size <%d> too small.",
                 size);
        return -ERANGE;
    }
    /*
     * create the message, which contains just the bsdh header.
     * (don't need to worry about header space reservation)
     */
    buff = sdp_buff_pool_get();
    if (NULL == buff)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to allocate buffer for RDMA read completion.");
        result = -ENOMEM;
        goto error;
    }
    /*
     * setup header.
     */
    buff->tail = buff->data;
    buff->bsdh_hdr = (struct sdp_msg_hdr_bsdh *) buff->tail;
    buff->bsdh_hdr->mid = TS_SDP_MSG_MID_RDMA_RD_COMP;
    buff->bsdh_hdr->flags = TS_SDP_MSG_FLAG_NON_FLAG;
    buff->tail += sizeof(struct sdp_msg_hdr_bsdh);
    rrch = (struct sdp_msg_hdr_rrch *) buff->tail;
    rrch->size = (tUINT32) size;
    buff->tail += sizeof(struct sdp_msg_hdr_rrch);
    /*
     * solicit event
     */
#ifdef _TS_SDP_SE_UNSIG_BUG_WORKAROUND
    TS_SDP_BUFF_F_CLR_SE(buff);
#else
    TS_SDP_BUFF_F_SET_SE(buff);
#endif
    TS_SDP_BUFF_F_SET_UNSIG(buff);
    /*
     * set PIPE bit to request switch into pipeline mode.
     */
    TS_SDP_MSG_HDR_SET_REQ_PIPE(buff->bsdh_hdr);
    /*
     * endian swap of extended header
     */
    result = _sdp_msg_cpu_to_net_rrch(rrch);
    TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    /*
     * Either post a send, or buffer the packet in the tx queue
     */
    result = _sdp_send_ctrl_buff_buffered(conn, buff);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to post buffered control message. <%d>",
                 result);
        (void)sdp_buff_pool_put(buff);
    }

    error:
    return result;
} /* sdp_send_ctrl_rdma_rd */

/* ========================================================================= */
/*..sdp_send_ctrl_rdma_wr -- Send an rdma write completion */
tINT32 sdp_send_ctrl_rdma_wr(struct sdp_opt *conn, tUINT32 size)
{
    struct sdp_msg_hdr_rwch *rwch;
    tINT32 result = 0;
    struct sdpc_buff *buff;

    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * create the message, which contains just the bsdh header.
     * (don't need to worry about header space reservation)
     */
    buff = sdp_buff_pool_get();
    if (NULL == buff)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to allocate buffer for RDMA write completion.");
        result = -ENOMEM;
        goto error;
    }
    /*
     * setup header.
     */
    buff->tail = buff->data;
    buff->bsdh_hdr = (struct sdp_msg_hdr_bsdh *) buff->tail;
    buff->bsdh_hdr->mid = TS_SDP_MSG_MID_RDMA_WR_COMP;
    buff->bsdh_hdr->flags = TS_SDP_MSG_FLAG_NON_FLAG;
    buff->tail += sizeof(struct sdp_msg_hdr_bsdh);
    rwch = (struct sdp_msg_hdr_rwch *) buff->tail;
    rwch->size = size;
    buff->tail += sizeof(struct sdp_msg_hdr_rwch);
    /*
     * solicit event
     */
#ifdef _TS_SDP_SE_UNSIG_BUG_WORKAROUND
    TS_SDP_BUFF_F_CLR_SE(buff);
#else
    TS_SDP_BUFF_F_SET_SE(buff);
#endif
    TS_SDP_BUFF_F_SET_UNSIG(buff);
    /*
     * endian swap of extended header
     */
    result = _sdp_msg_cpu_to_net_rwch(rwch);
    TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    /*
     * Either post a send, or buffer the packet in the tx queue
     */
    result = _sdp_send_ctrl_buff_buffered(conn, buff);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to post buffered control message. <%d>",
                 result);
        (void)sdp_buff_pool_put(buff);
    }

    error:
    return result;
} /* sdp_send_ctrl_rdma_wr */

/* ========================================================================= */
/*..sdp_send_ctrl_snk_avail -- Send a sink available message */
tINT32 sdp_send_ctrl_snk_avail(struct sdp_opt *conn,
                               tUINT32 size,
                               tUINT32 rkey,
                               tUINT64 addr)
{
    struct sdp_msg_hdr_snkah *snkah;
    tINT32 result = 0;
    struct sdpc_buff *buff;

    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * check mode
     */
    if (TS_SDP_MODE_PIPE != conn->recv_mode)
    {

        result = -EPROTO;
        goto error;
    }
    /*
     * create the message, which contains just the bsdh header.
     * (don't need to worry about header space reservation)
     */
    buff = sdp_buff_pool_get();
    if (NULL == buff)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to allocate buffer for RDMA read completion.");
        result = -ENOMEM;
        goto error;
    }
    /*
     * setup header.
     */
    buff->tail = buff->data;
    buff->bsdh_hdr = (struct sdp_msg_hdr_bsdh *) buff->tail;
    buff->bsdh_hdr->mid = TS_SDP_MSG_MID_SNK_AVAIL;
    buff->bsdh_hdr->flags = TS_SDP_MSG_FLAG_NON_FLAG;
    buff->tail += sizeof(struct sdp_msg_hdr_bsdh);
    snkah = (struct sdp_msg_hdr_snkah *) buff->tail;
    snkah->size = size;
    snkah->r_key = rkey;
    snkah->addr = addr;
    snkah->non_disc = conn->nond_recv;
    buff->tail += sizeof(struct sdp_msg_hdr_snkah);
    buff->data_size = 0;

    TS_SDP_BUFF_F_CLR_SE(buff);
    TS_SDP_BUFF_F_SET_UNSIG(buff);
    /*
     * endian swap of extended header
     */
    result = _sdp_msg_cpu_to_net_snkah(snkah);
    TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    /*
     * Either post a send, or buffer the packet in the tx queue
     */
    result = _sdp_send_ctrl_buff_buffered(conn, buff);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to post buffered control message. <%d>",
                 result);
        goto error;
    }

    result = 0;
    error:
    return result;
} /* sdp_send_ctrl_snk_avail */

/* ========================================================================= */
/*..sdp_send_ctrl_mode_ch -- Send a mode change command */
tINT32 sdp_send_ctrl_mode_ch(struct sdp_opt *conn, tUINT8 mode)
{
    struct sdp_msg_hdr_mch *mch;
    tINT32 result = 0;
    struct sdpc_buff *buff;

    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * validate that the requested mode transition is OK.
     */
    switch (mode)
    {
    
    case TS_SDP_MSG_MCH_BUFF_RECV:  /* source to sink */

        conn->send_mode = ((TS_SDP_MODE_COMB == conn->send_mode) ?
                           TS_SDP_MODE_BUFF : TS_SDP_MODE_ERROR);
        break;
    case TS_SDP_MSG_MCH_COMB_SEND:  /* sink to source */

        conn->recv_mode = ((TS_SDP_MODE_BUFF == conn->recv_mode) ?
                           TS_SDP_MODE_COMB : TS_SDP_MODE_ERROR);
        break;
    case TS_SDP_MSG_MCH_PIPE_RECV:  /* source to sink */

        conn->send_mode = ((TS_SDP_MODE_COMB == conn->send_mode) ?
                           TS_SDP_MODE_PIPE : TS_SDP_MODE_ERROR);
        break;
    case TS_SDP_MSG_MCH_COMB_RECV:  /* source to sink */

        conn->send_mode = ((TS_SDP_MODE_PIPE == conn->send_mode) ?
                           TS_SDP_MODE_COMB : TS_SDP_MODE_ERROR);
        break;
    default:

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Invalid mode transition <%d> requested.", mode);
        result = -EPROTO;
        goto error;
    }

    if (TS_SDP_MODE_ERROR == conn->send_mode ||
        TS_SDP_MODE_ERROR == conn->recv_mode)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: mode <%d> transition error. <%d:%d>", mode,
                 conn->send_mode, conn->recv_mode);
        result = -EPROTO;
        goto error;
    }
    /*
     * create the message, which contains just the bsdh header.
     * (don't need to worry about header space reservation)
     */
    buff = sdp_buff_pool_get();
    if (NULL == buff)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to allocate buffer for ModeChange.");
        result = -ENOMEM;
        goto error;
    }
    /*
     * setup header.
     */
    buff->tail = buff->data;
    buff->bsdh_hdr = (struct sdp_msg_hdr_bsdh *) buff->tail;
    buff->bsdh_hdr->mid = TS_SDP_MSG_MID_MODE_CHANGE;
    buff->bsdh_hdr->flags = TS_SDP_MSG_FLAG_NON_FLAG;
    buff->tail += sizeof(struct sdp_msg_hdr_bsdh);
    mch = (struct sdp_msg_hdr_mch *) buff->tail;
    buff->tail += sizeof(struct sdp_msg_hdr_mch);

    TS_SDP_BUFF_F_SET_SE(buff);
    TS_SDP_BUFF_F_CLR_UNSIG(buff);
    TS_SDP_MSG_MCH_SET_MODE(mch, mode);
    /*
     * endian swap of extended header
     */
    result = _sdp_msg_cpu_to_net_mch(mch);
    TS_EXPECT(MOD_LNX_SDP, !(0 > result));
    /*
     * Either post a send, or buffer the packet in the tx queue
     */
    result = _sdp_send_ctrl_buff_buffered(conn, buff);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,
                 "POST: Failed to post buffered control message. <%d>",
                 result);
        (void)sdp_buff_pool_put(buff);
    }

    error:
    return result;
} /* sdp_send_ctrl_mode_ch */

/* ========================================================================= */
/*.._sdp_send_flush_advt -- Flush passive sink advertisments */
static tINT32 _sdp_send_flush_advt(struct sdp_opt *conn)
{
    struct sdpc_advt *advt;
    tINT32 result;
    /*
     * If there is no data in the pending or active send pipes, and a
     * partially complete sink advertisment is pending, then it needs
     * to be completed. It might be some time until more data is ready
     * for transmission, and the remote host needs to be notified of
     * present data. (rdma ping-pong letency test...)
     */
    if (0 == sdp_desc_q_size(&conn->send_queue))
    {
        /*
         * might be more aggressive then we want it to be. maybe
         * check if the active sink queue is empty as well?
         */
        advt = sdp_advt_q_look(&conn->snk_pend);
        if (NULL != advt && 0 < advt->post)
        {

            advt = sdp_advt_q_get(&conn->snk_pend);
            TS_EXPECT(MOD_LNX_SDP, (NULL != advt));

            result = sdp_send_ctrl_rdma_wr(conn, advt->post);
            TS_EXPECT(MOD_LNX_SDP, !(0 > result));

            result = sdp_advt_destroy(advt);
            TS_EXPECT(MOD_LNX_SDP, !(0 > result));
            /*
             * update sink advertisments.
             */
            conn->snk_recv--;
        }
    }

    return 0;
} /* _sdp_send_flush_advt */

/* --------------------------------------------------------------------- */
/*                                                                       */
/*                          GENERAL functions                            */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..sdp_send_flush -- Flush buffers from send queue, in to send post. */
tINT32 sdp_send_flush(struct sdp_opt *conn)
{
    tINT32 result = 0;

    TS_CHECK_NULL(conn, -EINVAL);
    /*
     * keep posting sends as long as there is room for an SDP post.
     * Priority goes to control messages, and we need to follow the
     * send credit utilization rules.
     */
    result = _sdp_send_ctrl_buff_flush(conn);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: failed to flush control. <%d>", result);
        goto done;
    }
    /*
     * data flush
     */
    result = _sdp_send_data_queue_flush(conn);
    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "POST: failed to flush data. <%d>", result);
        goto done;
    }
    /*
     * Sink advertisment flush.
     */
    if (0 < sdp_advt_q_size(&conn->snk_pend))
    {

        result = _sdp_send_flush_advt(conn);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: failed to flush sink advertisments. <%d>",
                     result);
            goto done;
        }
    }
    /*
     * disconnect flush
     */
    if (0 < (TS_SDP_ST_MASK_DIS_PEND & conn->state))
    {

        result = sdp_send_ctrl_disconnect(conn);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "POST: failed to post disconnect during flush. <%d>",
                     result);
            goto done;
        }
    }
    /*
     * see if there is enough buffer to wake/notify writers
     */
    sdp_inet_wake_send(conn->sk); /*  conn->sk->write_space(conn->sk); */

    return 0;
    done:
    return result;
} /* sdp_send_flush */

/* ========================================================================= */
/*..sdp_inet_send -- send data from user space to the network. */
int sdp_inet_send(
#ifdef TS_KERNEL_2_6
                 struct kiocb *_kiocb,
                 struct socket *sock,
                 struct msghdr *msg,
                 size_t size
#else
                 struct socket *sock,
                 struct msghdr *msg,
                 int    size,
                 struct scm_cookie *scm
#endif
                 )
{
    struct sock *sk;
    struct sdp_opt *conn;
    struct sdpc_iocb *iocb = NULL;
    struct sdpc_buff *buff;
    tINT32 result = 0;
    tINT32 copied = 0;
    struct sdpc_desc *element=NULL; 
    tINT32 copy,left,current_wr,scheduled,completed;
    tUINT64 base=0;
    tINT32 oob,max_zcopy,offset,i,current_size;
    long timeout = -1;

    TS_CHECK_NULL(sock, -EINVAL);
    TS_CHECK_NULL(sock->sk, -EINVAL);
    TS_CHECK_NULL(msg, -EINVAL);

    sk = sock->sk;
    conn = TS_SDP_GET_CONN(sk);
    conn->s_pend_copied=0;
    left=size;
    current_wr=0;

#ifdef _TS_SDP_DATA_PATH_DEBUG
    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
             "SOCK: send <%08x:%04x> <%08x:%04x> <%04x> size <%d> flags <%08x>",
             conn->src_addr, conn->src_port, conn->dst_addr, conn->dst_port,
             conn->istate, size, msg->msg_flags);
#endif
  

    TS_SDP_CONN_LOCK(conn);
    /*
     * ESTABLISED and CLOSE can send, while CONNECT and ACCEPTED can
     * continue being processed, it'll wait below until the send window
     * is opened on sucessful connect, or error on an unsucessful attempt.
     */
    if (0 < (TS_SDP_ST_MASK_CLOSED & conn->istate))
    {

        result = -EPIPE;
        goto done;
    }
    /*
     * clear ASYN space bit, it'll be reset if there is no space.
     */
    clear_bit(SOCK_ASYNC_NOSPACE, &TS_SDP_OS_SK_SOCKET(sk)->flags);
    oob = (MSG_OOB & msg->msg_flags);
    /*
     * process data first if window is open, next check conditions, then
     * wait if there is more work to be done. The absolute window size is
     * used to 'block' the caller if the connection is still connecting.
     */

    if (0 > timeout)
    {

        timeout =  sock_sndtimeo(sk, (MSG_DONTWAIT & msg->msg_flags));
    }

    for (i=0;i<msg->msg_iovlen;i++)
    {
        scheduled=0;
        completed=0;
        left = msg->msg_iov[i].iov_len;

        current_size=left;

        while ( left > conn->src_zthresh  &&  conn->use_zcopy==1 
                && 0 == (MSG_DONTWAIT & msg->msg_flags) &&  TS_SDP_MODE_BUFF != conn->send_mode)
        {
            base = (tUINT64)(unsigned long) msg->msg_iov[i].iov_base + scheduled;
            offset = base & ~PAGE_MASK;
            max_zcopy = TS_SDP_IOCB_SIZE_MAX - offset;
            current_wr = min(max_zcopy, left);


            iocb = sdp_zcopy_create();
            if (NULL == iocb)
            {
                TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                         "WRITE: Failed to allocate iocb to save cb. <%d>",
                         size);
                goto sleep;
            }
            iocb->len = current_wr ;
            iocb->post = 0;
            iocb->size = current_wr;
            iocb->msg=msg;
            iocb->dir=1;
            iocb->base=base;

            TS_SDP_CONN_STAT_WQ_INC(conn, iocb->size);


            if (NULL == iocb->page_array)
            {

                result = sdp_zcopy_register(iocb, conn); 
                if (0 > result)
                {

                    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                             "POST: Error <%d> registering IOCB. <%d:%d>",
                             result, iocb->key, iocb->len);
                    //    goto drop;
                    result = sdp_iocb_destroy(iocb);
                    goto     sleep;
                }
            }  else {

            }

            conn->send_pipe += iocb->len;
            result = _sdp_send_data_queue (conn, (struct sdpc_desc *) iocb);


            if (0 > result)
            {

                TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                         "WRITE: Error <%d> queueing write IOCB. <%d>",
                         result, iocb->key,
                         sdp_desc_q_size(&conn->send_queue));

                (void)sdp_iocb_destroy(iocb);

                result = (0 < (scheduled+copied)) ? copied+scheduled : result;
                goto sleep;    
            }
            scheduled+=current_wr;  
            left-=current_wr;
            result = 0;

        }

        sleep:     
        if (scheduled>0)
        {

 //sleep
            while (1)
            {
                DECLARE_WAITQUEUE(wait, current);

                set_bit(SOCK_ASYNC_NOSPACE, &TS_SDP_OS_SK_SOCKET(sk)->flags);
                set_bit(SOCK_NOSPACE, &TS_SDP_OS_SK_SOCKET(sk)->flags);
                //TODO: sleep until someone wakes you up
                iocb = sdp_iocb_q_look(&conn->w_src);

                if (signal_pending(current) )
                {
                    if (iocb==NULL)
                    {
                        result =
                        (0 < timeout) ? sock_intr_errno(timeout) : -EAGAIN;

                        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                                 "Got signal pending, exiting ");     
                        copied+=conn->s_pend_copied;
                        goto done;

                    }
                    if ((conn->flags & TS_SDP_CONN_F_SRC_CANCEL_L) == 0)
                    {
                        while ( 1)    //clean send_queue of src_avail
                        {
                            element = sdp_desc_q_look_tail(&conn->send_queue);
                            if (element==NULL)
                            {
                                break;
                            } else
                            {
                                if (TS_SDP_GENERIC_TYPE_ZCOPY == element->type)
                                {
                                    element = sdp_desc_q_get_tail(&conn->send_queue);

                                    sdp_zcopy_complete((struct sdpc_iocb *) element,0);                                                     //complete

                                } else
                                {
                                    break;
                                }
                            }


                        }

                        result = sdp_send_ctrl_src_cancel(conn);
                        TS_EXPECT(MOD_LNX_SDP, !(0 > result));
                        conn->flags |= TS_SDP_CONN_F_SRC_CANCEL_L;

                    }
                    if  (conn->flags & TS_SDP_CONN_F_SRC_CANCEL_L)
                    {
                        if (timeout> (HZ/2)) {
                            timeout=(HZ/2);
                        }
                    }
                }






                if (conn->istate != TS_SDP_SOCK_ST_ESTABLISHED &&
		    conn->istate != TS_SDP_SOCK_ST_ACCEPTED) {
                    result=-ECONNRESET;
                    copied+=conn->s_pend_copied;
                    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN," conn state 0x%x link is down ,exiting", conn->istate);
                    goto done;
                }


                add_wait_queue(TS_SDP_OS_SK_SLEEP(sk), &wait);
                set_current_state(TASK_INTERRUPTIBLE);
                clear_bit(SOCK_ASYNC_NOSPACE,
                          &TS_SDP_OS_SK_SOCKET(sk)->flags);
                TS_SDP_CONN_UNLOCK(conn);
                timeout = schedule_timeout(timeout);
                TS_SDP_CONN_LOCK(conn);
                remove_wait_queue(TS_SDP_OS_SK_SLEEP(sk), &wait);
                set_current_state(TASK_RUNNING);
                if (timeout==0)
                {//expired
                    conn->src_cncl = 0;
                    conn->flags &= ~(TS_SDP_CONN_F_SRC_CANCEL_L);
                    copied+=conn->s_pend_copied;
                    conn->s_pend_copied=0;
                    result = sdp_signal_abort(conn);
                    if (0 > result) {
                        
                            TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                                     TRACE_FLOW_WARN,
                                     "SOCK: Error <%d> while aborting socket <%d> on release.",
                                        result, conn->pid);
                    }                
                                                
                    goto done;
                }
                if (conn->s_pend_copied==scheduled)
                {
                    copied+=scheduled;
                    TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_WARN,"write is done copied is %d",copied);

                    conn->s_pend_copied=0;
                    goto bcopy;
		    
                }
            }


        }

        bcopy:  
        msg->msg_iov[i].iov_base+=scheduled;
        msg->msg_iov[i].iov_len-=scheduled;   
        completed+=scheduled;
        while (current_size> completed)
        {
            //bcopy
            while (0 < __sdp_inet_write_space(conn, oob))
            {

                buff = _sdp_send_data_buff_get(conn);
                if (NULL == buff)
                {

                    result = -ENOMEM;
                    goto done;
                }

                copy =
                min((size_t) (buff->end - buff->tail),
                    (size_t) (current_size - completed));
                copy = min(copy, __sdp_inet_write_space(conn, oob));

#ifndef _TS_SDP_DATA_PATH_NULL
                result =   memcpy_fromiovec(buff->tail, &(msg->msg_iov[i]), copy);
                if (0 > result)
                {

                    (void)sdp_buff_pool_put(buff);
                    goto done;
                }
#endif
                buff->tail += copy;
                completed += copy;
                left-=copy;
                copied+=copy;
                TS_SDP_CONN_STAT_SEND_INC(conn, copy);
                result = _sdp_send_data_buff_put(conn, buff, copy,
                                                 ((copied ==
                                                   size) ? oob : 0));
                if (0 > result)
                {

                    goto done;
                }

                if (completed == current_size)
                {

                    goto finish;
                }
            }
            /*
             * onetime setup of timeout, but only if it's needed.
             */
            if (0 > timeout)
            {

                timeout =
                sock_sndtimeo(sk, (MSG_DONTWAIT & msg->msg_flags));
            }
            /*
             * check status.
             */
            set_bit(SOCK_ASYNC_NOSPACE, &TS_SDP_OS_SK_SOCKET(sk)->flags);
            set_bit(SOCK_NOSPACE, &TS_SDP_OS_SK_SOCKET(sk)->flags);

            if (0 != TS_SDP_OS_CONN_GET_ERR(conn))
            {

                result = (0 < copied) ? 0 : TS_SDP_CONN_ERROR(conn);
                goto done;
            }

            if (0 < (TS_SDP_SHUTDOWN_SEND & conn->shutdown))
            {

                result = -EPIPE;
                goto done;
            }

            if (TS_SDP_SOCK_ST_ERROR == conn->istate)
            {

                result = -EPROTO; /* error should always be set, but
                         just in case */
                goto done;
            }

            if (0 == timeout)
            {

                result = -EAGAIN;
                goto done;
            }

            if (signal_pending(current))
            {

                result =
                (0 < timeout) ? sock_intr_errno(timeout) : -EAGAIN;
                goto done;
            }
            /*
             * wait
             */
            {
                DECLARE_WAITQUEUE(wait, current);
                add_wait_queue(TS_SDP_OS_SK_SLEEP(sk), &wait);
                set_current_state(TASK_INTERRUPTIBLE);
                /*
                 * ASYNC_NOSPACE is only set if we're not sleeping,
                 * while NOSPACE is set whenever there is no space,
                 * and is only cleared once space opens up, in
                 * DevConnAck()
                 */
                clear_bit(SOCK_ASYNC_NOSPACE,
                          &TS_SDP_OS_SK_SOCKET(sk)->flags);

                TS_SDP_CONN_UNLOCK(conn);
                if (!(0 < __sdp_inet_write_space(conn, oob)))
                {
                    timeout = schedule_timeout(timeout);
                }
                TS_SDP_CONN_LOCK(conn);

                remove_wait_queue(TS_SDP_OS_SK_SLEEP(sk), &wait);
                set_current_state(TASK_RUNNING);
            }
        }
        finish: 
        continue;
    }

    done:
    while ( 1)
    {
        element = sdp_desc_q_look_head(&conn->send_queue);
        if (element==NULL)
        {
            break;
        } else
        {
            if (TS_SDP_GENERIC_TYPE_ZCOPY == element->type)
            {
                element = sdp_desc_q_get_head(&conn->send_queue);

                sdp_zcopy_complete((struct sdpc_iocb *) element,0);                                                     //complete

            }  else {
                break;
            }
        }


    }
   //conn->src_cncl = 0;
   // conn->flags &= ~(TS_SDP_CONN_F_SRC_CANCEL_L);

    TS_SDP_CONN_UNLOCK(conn);
    result = ((0 < copied) ? copied : result);

    if (-EPIPE == result && 0 == (MSG_NOSIGNAL & msg->msg_flags))
    {

        send_sig(SIGPIPE, current, 0);
    }

    return result;
} /* sdp_inet_send */

#ifdef _TS_SDP_AIO_SUPPORT
/* ========================================================================= */
/*.._sdp_inet_write_fast -- write multiple SDP buffers from an iocb */
static tINT32 _sdp_inet_write_fast(struct sdp_opt *conn,
                                   struct kvec_dst *src,
                                   tINT32 len)
{
    struct sdpc_buff *buff;
    tINT32 copied = 0;
    tINT32 expect;
    tINT32 result;
    tINT32 copy;

    TS_CHECK_NULL(conn, -EINVAL);
    TS_CHECK_NULL(src, -EINVAL);
    /*
     * loop through queued buffers and copy them to the destination
     */
    while (copied < len && 0 < __sdp_inet_write_space(conn, 0))
    {
        /*
         * get a buffer for posting.
         */
        buff = _sdp_send_data_buff_get(conn);
        if (NULL == buff)
        {

            result = -ENOMEM;
            goto done;
        }

        copy = min((len - copied), __sdp_inet_write_space(conn, 0));

#ifdef _TS_SDP_AIO_SUPPORT
        result = _sdp_send_data_iocb_buff_kvec(src, buff, copy);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "WRITE: error <%d> copying data <%d> from vector.",
                     result, copy);

            expect = sdp_buff_pool_put(buff);
            TS_EXPECT(MOD_LNX_SDP, !(0 > expect));

            goto done;
        }
#endif

        copied += result;

        result = _sdp_send_data_buff_put(conn, buff, result, 0);
        if (0 > result)
        {

            goto done;
        }
    }

    result = 0;
    done:
    return(0 < copied) ? copied : result;
} /* _sdp_inet_write_fast */

/* ========================================================================= */
/*..sdp_inet_write -- send AIO data from user space to the network. */
int sdp_inet_write(struct socket *sock, kvec_cb_t cb, size_t size)
{
    struct kiocb *req;
    struct kvec_dst src;
    struct sock *sk;
    struct sdp_opt *conn;
    struct sdpc_iocb *iocb;
    tINT32 copied = 0;
    tINT32 result = 0;

    TS_CHECK_NULL(sock, -EINVAL);
    TS_CHECK_NULL(sock->sk, -EINVAL);
    TS_CHECK_NULL(cb.fn, -EINVAL);
    TS_CHECK_NULL(cb.data, -EINVAL);
    TS_CHECK_NULL(cb.vec, -EINVAL);

    sk = sock->sk;
    conn = TS_SDP_GET_CONN(sk);
    req = (struct kiocb *)cb.data;

#ifdef _TS_SDP_DATA_PATH_DEBUG
    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
             "SOCK: Write <%08x:%04x> <%08x:%04x> <%04x> size <%d>",
             conn->src_addr, conn->src_port, conn->dst_addr, conn->dst_port,
             conn->istate, size);
    TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_INOUT,
             "SOCK: Write Kvec <%d:%d> veclet <%d:%d> args <%08x> kiocb "
             "<%d:%d:%08x>",
             cb.vec->max_nr, cb.vec->nr,
             cb.vec->veclet->offset, cb.vec->veclet->length,
             (tUINT32) cb.fn, req->key, req->users, (tUINT32) req->data);
#endif
    /*
     * initialize memory destination
     */
    kvec_dst_init(&src, KM_USER0);
    /*
     * set data source address to cb vectors address
     */
    kvec_dst_set(&src, cb.vec->veclet);

    TS_SDP_CONN_LOCK(conn);
    /*
     * check error conditions before allowing writes...
     * (send_mask and shutdown_mask should be redundant)
     */
    if (0 < (TS_SDP_ST_MASK_CLOSED & conn->istate) ||
        0 < (TS_SDP_SHUTDOWN_SEND & conn->shutdown))
    {

        result = ((0 < TS_SDP_OS_CONN_GET_ERR(conn)) ?
                  (0 - TS_SDP_OS_CONN_GET_ERR(conn)) : -EPIPE);
        goto done;
    }
    /*
     * If the IOCB queue is empty and this is a candidate for BUFFERED
     * transmission, attempt to send.
     */
    if ((conn->src_zthresh > size ||
         TS_SDP_MODE_BUFF == conn->send_mode) &&
        0 == sdp_desc_q_types_size(&conn->send_queue,
                                   TS_SDP_GENERIC_TYPE_IOCB))
    {
        /*
         * write data onto SDP if there is sufficient window space.
         */
        copied = _sdp_inet_write_fast(conn, &src, size);
        if (0 > copied)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "WRITE: error <%d> flushing data <%d> from src <%d:%d:%d>",
                     copied, size, src.space, src.offset, src.type);
            result = copied;
            goto done;
        }
    }
    /*
     * If there is room left in the source save it in an IOCB, otherwise
     * report success in the callback.
     */
    if (size > copied)
    {
        /*
         * more data, save the cb for processing as data arrives.
         */
        iocb = sdp_iocb_create();
        if (NULL == iocb)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "WRITE: Failed to allocate iocb to save cb. <%d:%d>",
                     size, copied);
            result = (0 < copied) ? copied : -ENOMEM;
            goto done;
        }

        iocb->req = req;
        iocb->kvec.src = src;
        iocb->cb = cb;
        iocb->len = size - copied;
        iocb->post = copied;
        iocb->size = size;
        iocb->key = req->key;

        TS_SDP_CONN_STAT_WQ_INC(conn, iocb->size);

        req->data = sk;
        req->cancel = _sdp_inet_write_cancel;

        conn->send_pipe += iocb->len;

        result = _sdp_send_data_queue(conn, (struct sdpc_desc *) iocb);
        if (0 > result)
        {

            TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                     "WRITE: Error <%d> queueing write IOCB. <%d>",
                     result, iocb->key,
                     sdp_desc_q_size(&conn->send_queue));

            (void)sdp_iocb_destroy(iocb);

            result = (0 < copied) ? copied : result;
            goto done;  /* synchronous error */
        }

        result = 0;
    } else
    {

        TS_SDP_CONN_STAT_WRITE_INC(conn, copied);
        result = copied;
    }

    done:
    TS_SDP_CONN_UNLOCK(conn);

    if (0 > result)
    {

        TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
                 "EVENT: Error <%d> writing IOCB <%d>.",
                 result, req->key);
    }
    /*
     * If no data copied, return result, otherwise only complete event
     * if there was an error.
     */
    return result;
} /* sdp_inet_write */
#endif

