/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  This source code may incorporate intellectual property owned by
  Microsoft Corporation. Our provision of this source code does not
  include any licenses or any other rights to you under any Microsoft
  intellectual property. If you would like a license from Microsoft
  (e.g., to rebrand, redistribute), you need to contact Microsoft
  directly.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sdp_sent.c 379 2004-07-08 02:24:32Z roland $
*/

#include "sdp_main.h"

/* --------------------------------------------------------------------- */
/*                                                                       */
/* Specific MID handler functions. (SEND)                                */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*.._sdp_sent_disconnect --  */
static int _sdp_sent_disconnect(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
		 "STRM: <%d> <%08x:%04x><%08x:%04x> disconnect sent in state <%04x>",
		 conn->hashent, conn->src_addr, conn->src_port,
		 conn->dst_addr, conn->dst_port, conn->state);

	switch (conn->state) {
	case TS_SDP_CONN_ST_TIME_WAIT_2:
		/*
		 * Nothing to do, CM disconnects have been exchanged.
		 */
		break;
	case TS_SDP_CONN_ST_DIS_SEND_1:
		/*
		 * Active disconnect message send completed
		 */
		TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_DIS_SENT_1);

		break;
	case TS_SDP_CONN_ST_DIS_SEND_2:
		/*
		 * passive disconnect message send
		 */
		TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_DIS_SENT_2);
		/*
		 * Begin IB/CM disconnect
		 */
		result = sdp_cm_disconnect(conn);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "STRM: Error posting CM disconnect. <%d>",
				 result);
			goto error;
		}

		break;
	case TS_SDP_CONN_ST_DIS_RECV_R:
		/*
		 * simultaneous disconnect. Received a disconnect, after we
		 * initiated one. This needs to be handled as the active
		 * stream interface close that it is.
		 */
		TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_TIME_WAIT_1);
		/*
		 * acknowledge disconnect to framework
		 */
		result = sdp_wall_recv_closing(conn);
		if (0 > result) {

			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "STRM: Error confirming close to stream interface. <%d>",
				 result);
			goto error;
		}
		/*
		 * Begin IB/CM disconnect
		 */
		result = sdp_cm_disconnect(conn);
		/*
		 * if the remote DREQ was already received, but unprocessed, do
		 * not treat it as an error
		 */
		if (0 > result) {

			if (-EPROTO != result) {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_WARN,
					 "STRM: Error posting CM disconnect. <%d>",
					 result);
				goto error;
			}
			else {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_WARN,
					 "STRM: CM disconnect of disconnected conn. OK");
			}
		}

		break;
	default:

		TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
			 "STRM: Disconnect send completion in an unexpected state. <%04x>",
			 conn->state);
		result = -EFAULT;
		goto error;

		break;
	}

	return 0;
error:
	return result;
} /* _sdp_sent_disconnect */

/* ========================================================================= */
/*.._sdp_sent_abort --  */
static int _sdp_sent_abort(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);
	/*
	 * The gateway interface should be in error state, initiate CM
	 * disconnect.
	 */
	TS_EXPECT(MOD_LNX_SDP, (TS_SDP_CONN_ST_ERROR_STRM == conn->state));
	TS_SDP_CONN_ST_SET(conn, TS_SDP_CONN_ST_ERROR_STRM);

	result = sdp_cm_disconnect(conn);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "STRM: Error posting CM disconnect. <%d>", result);
	}

	return result;
} /* _sdp_sent_abort */

/* ========================================================================= */
/*.._sdp_sent_send_sm --  */
static int _sdp_sent_send_sm(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	return 0;
} /* _sdp_sent_send_sm */

/* ========================================================================= */
/*.._sdp_sent_rdma_wr --  */
static int _sdp_sent_rdma_wr(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdp_msg_hdr_rwch *rwch;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	rwch = (struct sdp_msg_hdr_rwch *) buff->data;
	buff->data = buff->data + sizeof(struct sdp_msg_hdr_rwch);

	result = _sdp_msg_net_to_cpu_rwch(rwch);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	return 0;
} /* _sdp_sent_rdma_wr */

/* ========================================================================= */
/*.._sdp_sent_rdma_rd --  */
static int _sdp_sent_rdma_rd(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdp_msg_hdr_rrch *rrch;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	rrch = (struct sdp_msg_hdr_rrch *) buff->data;
	buff->data = buff->data + sizeof(struct sdp_msg_hdr_rrch);

	result = _sdp_msg_net_to_cpu_rrch(rrch);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	return 0;
} /* _sdp_sent_rdma_rd */

/* ========================================================================= */
/*.._sdp_sent_mode_change --  */
static int _sdp_sent_mode_change(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdp_msg_hdr_mch *mch;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	mch = (struct sdp_msg_hdr_mch *) buff->data;
	buff->data = buff->data + sizeof(struct sdp_msg_hdr_mch);

	result = _sdp_msg_net_to_cpu_mch(mch);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	return 0;
} /* _sdp_sent_mode_change */

/* ========================================================================= */
/*.._sdp_sent_src_cancel --  */
static int _sdp_sent_src_cancel(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	return 0;
} /* _sdp_sent_src_cancel */

/* ========================================================================= */
/*.._sdp_sent_snk_cancel --  */
static int _sdp_sent_snk_cancel(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	return 0;
} /* _sdp_sent_snk_cancel */

/* ========================================================================= */
/*.._sdp_sent_snk_cancel_ack --  */
static int _sdp_sent_snk_cancel_ack(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	return 0;
} /* _sdp_sent_snk_cancel_ack */

/* ========================================================================= */
/*.._sdp_sent_resize_buff_ack --  */
static int _sdp_sent_resize_buff_ack(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdp_msg_hdr_crbah *crbah;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	crbah = (struct sdp_msg_hdr_crbah *) buff->data;
	buff->data = buff->data + sizeof(struct sdp_msg_hdr_crbah);

	result = _sdp_msg_net_to_cpu_crbah(crbah);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	return 0;
} /* _sdp_sent_resize_buff_ack */

/* ========================================================================= */
/*.._sdp_sent_suspend --  */
static int _sdp_sent_suspend(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdp_msg_hdr_sch *sch;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	sch = (struct sdp_msg_hdr_sch *) buff->data;
	buff->data = buff->data + sizeof(struct sdp_msg_hdr_sch);

	result = _sdp_msg_net_to_cpu_sch(sch);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	return 0;
} /* _sdp_sent_suspend */

/* ========================================================================= */
/*.._sdp_sent_suspend_ack --  */
static int _sdp_sent_suspend_ack(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	return 0;
} /* _sdp_sent_suspend_ack */

/* ========================================================================= */
/*.._sdp_sent_snk_avail --  */
static int _sdp_sent_snk_avail(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdp_msg_hdr_snkah *snkah;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	snkah = (struct sdp_msg_hdr_snkah *) buff->data;
	buff->data = buff->data + sizeof(struct sdp_msg_hdr_snkah);

	result = _sdp_msg_net_to_cpu_snkah(snkah);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	return 0;
} /* _sdp_sent_snk_avail */

/* ========================================================================= */
/*.._sdp_sent_src_avail --  */
static int _sdp_sent_src_avail(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	struct sdp_msg_hdr_srcah *srcah;
	int result;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	srcah = (struct sdp_msg_hdr_srcah *) buff->data;
	buff->data = buff->data + sizeof(struct sdp_msg_hdr_srcah);

	result = _sdp_msg_net_to_cpu_srcah(srcah);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	return 0;
} /* _sdp_sent_src_avail */

/* ========================================================================= */
/*.._sdp_sent_data -- SDP data message event received */
static int _sdp_sent_data(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	int result = 0;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(conn->sk, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);

	conn->send_qud -= buff->data_size;

	return result;
} /* _sdp_sent_data */

/* ========================================================================= */
/*.._sdp_sent_unsupported -- Valid messages we're not sending */
static int _sdp_sent_unsupported(struct sdp_opt *conn, struct sdpc_buff *buff)
{
	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(buff, -EINVAL);
	/*
	 * Since the gateway only initates RDMA's but is never a target, and
	 * for a few other reasons, there are certain valid SDP messages
	 * which we never send.
	 */
	TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
		 "EVENT: Unexpected SDP message <%02x> sent!",
		 buff->bsdh_hdr->mid);

	return 0;
} /* _sdp_sent_unsupported */

/*
 * Event Dispatch table. For performance a dispatch table is used to avoid
 * a giant case statment for every single SDP event. This is a bit more
 * confusing, relies on the layout of the Message IDs, and is less
 * flexable. However, it is faster.
 *
 * Sparse table, the full table would be 16x16, where the low 4 bits, of
 * the MID byte, are one dimension, and the high 4 bits are the other
 * dimension. Since all rows, except for the first and last, are empty,
 * only those are represented in the table.
 */
static tGW_SDP_EVENT_CB_FUNC send_event_funcs[TS_SDP_MSG_EVENT_TABLE_SIZE] = {
	NULL,                      /* TS_SDP_MID_HELLO            0x00 */
	NULL,                      /* TS_SDP_MID_HELLO_ACK        0x01 */
	_sdp_sent_disconnect,      /* TS_SDP_MID_DISCONNECT       0x02 */
	_sdp_sent_abort,           /* TS_SDP_MID_ABORT_CONN       0x03 */
	_sdp_sent_send_sm,         /* TS_SDP_MID_SEND_SM          0x04 */
	_sdp_sent_rdma_wr,         /* TS_SDP_MID_RDMA_WR_COMP     0x05 */
	_sdp_sent_rdma_rd,         /* TS_SDP_MID_RDMA_RD_COMP     0x06 */
	_sdp_sent_mode_change,     /* TS_SDP_MID_MODE_CHANGE      0x07 */
	_sdp_sent_src_cancel,      /* TS_SDP_MID_SRC_CANCEL       0x08 */
	_sdp_sent_snk_cancel,      /* TS_SDP_MID_SNK_CANCEL       0x09 */
	_sdp_sent_snk_cancel_ack,  /* TS_SDP_MID_SNK_CANCEL_ACK   0x0A */
	_sdp_sent_unsupported,     /* TS_SDP_MID_CH_RECV_BUF      0x0B */
	_sdp_sent_resize_buff_ack, /* TS_SDP_MID_CH_RECV_BUF_ACK  0x0C */
	_sdp_sent_suspend,         /* TS_SDP_MID_SUSPEND          0x0D */
	_sdp_sent_suspend_ack,     /* TS_SDP_MID_SUSPEND_ACK      0x0E */
	NULL,                      /* reserved                    0x0F */
	NULL,                      /* reserved                    0xF0 */
	NULL,                      /* reserved                    0xF1 */
	NULL,                      /* reserved                    0xF2 */
	NULL,                      /* reserved                    0xF3 */
	NULL,                      /* reserved                    0xF4 */
	NULL,                      /* reserved                    0xF5 */
	NULL,                      /* reserved                    0xF6 */
	NULL,                      /* reserved                    0xF7 */
	NULL,                      /* reserved                    0xF8 */
	NULL,                      /* reserved                    0xF9 */
	NULL,                      /* reserved                    0xFA */
	NULL,                      /* reserved                    0xFB */
	NULL,                      /* reserved                    0xFC */
	_sdp_sent_snk_avail,       /* TS_SDP_MID_SNK_AVAIL        0xFD */
	_sdp_sent_src_avail,       /* TS_SDP_MID_SRC_AVAIL        0xFE */
	_sdp_sent_data             /* TS_SDP_MID_DATA             0xFF */
}; /* send_event_funcs */

/* ========================================================================= */
/*..sdp_event_send -- send event handler. */
tINT32 sdp_event_send(struct sdp_opt *conn, struct ib_cq_entry *comp)
{
	tGW_SDP_EVENT_CB_FUNC dispatch_func;
	u32 free_count = 0;
	u32 current_wrid = 0;
	u32 offset;
	int result;
	struct sdpc_buff *buff;
	struct sdpc_buff *head = NULL;

	TS_CHECK_NULL(conn, -EINVAL);
	TS_CHECK_NULL(comp, -EINVAL);
	/*
	 * error handling
	 */
	if (IB_COMPLETION_STATUS_SUCCESS != comp->status) {

		switch (comp->status) {
		case IB_COMPLETION_STATUS_WORK_REQUEST_FLUSHED_ERROR:
			/*
			 * clear posted buffers from error'd queue
			 */
			result = sdp_buff_q_clear(&conn->send_post);
			TS_EXPECT(MOD_LNX_SDP, !(0 > result));

			result = 0;
			break;
		default:
			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: <%04x> unhandled SEND error status <%d>.",
				 conn->hashent, comp->status);
			result = -EIO;
		}

		goto done;
	}
	/*
	 * get buffer.
	 */
	while (NULL != (buff = sdp_buff_q_get_head(&conn->send_post))) {
		/*
		 * sanity checks
		 */
		if (NULL == buff->bsdh_hdr) {

			TS_TRACE(MOD_LNX_SDP, T_TERSE, TRACE_FLOW_WARN,
				 "EVENT: send header is missing?!");
			result = -ENODATA;
			goto drop;
		}

		if (TS_SDP_WRID_LT(comp->work_request_id, buff->ib_wrid)) {
			/*
			 * error
			 */
			TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
				 "EVENT: send complete, wrid mismatch. <%u:%u:%d> mid <%02x>",
				 comp->work_request_id, buff->ib_wrid,
				 conn->send_usig, buff->bsdh_hdr->mid);
			result = -EINVAL;
			goto drop;
		}
		/*
		 * execute the send dispatch function, for specific actions.
		 */
#ifdef _TS_SDP_DATA_PATH_DEBUG
		TS_TRACE(MOD_LNX_SDP, T_VERY_VERBOSE, TRACE_FLOW_INOUT,
			 "EVENT: sent BSDH <%04x:%02x:%02x:%08x:%08x:%08x> <%08x> <%u>",
			 buff->bsdh_hdr->recv_bufs,
			 buff->bsdh_hdr->flags,
			 buff->bsdh_hdr->mid,
			 buff->bsdh_hdr->size,
			 buff->bsdh_hdr->seq_num,
			 buff->bsdh_hdr->seq_ack, buff->flags, buff->ib_wrid);
#endif
		/*
		 * data fast path we collapse the next level dispatch function.
		 * For all other buffers we go the slow path.
		 */
		if (TS_SDP_MSG_MID_DATA == buff->bsdh_hdr->mid) {

			conn->send_qud -= buff->data_size;
		}
		else {

			offset = buff->bsdh_hdr->mid & 0x1F;

			if (!(offset < TS_SDP_MSG_EVENT_TABLE_SIZE) ||
			    NULL == send_event_funcs[offset]) {

				TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
					 TRACE_FLOW_WARN,
					 "EVENT: send completion, unknown message ID <%d>",
					 buff->bsdh_hdr->mid);
				result = -EINVAL;
				goto drop;
			}
			else {

				TS_SDP_CONN_STAT_SEND_MID_INC(conn, offset);
                TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
                TRACE_FLOW_WARN,   "sent offset is %d",offset);
           

				dispatch_func = send_event_funcs[offset];
				result = dispatch_func(conn, buff);
				if (0 > result) {

					TS_TRACE(MOD_LNX_SDP, T_VERBOSE,
						 TRACE_FLOW_WARN,
						 "EVENT: receive completion, dispatch error. <%d>",
						 result);
					goto drop;
				}
			}
		}

		current_wrid = buff->ib_wrid;
		/*
		 * send queue size reduced by one.
		 */
		conn->s_wq_size--;

		if (0 < TS_SDP_BUFF_F_GET_UNSIG(buff)) {

			conn->send_usig--;
		}
		/*
		 * create a link of buffers which will be returned to
		 * the free pool in one group.
		 */
		result = sdp_buff_pool_chain_link(head, buff);
		TS_EXPECT(MOD_LNX_SDP, !(0 > result));

		head = buff;
		free_count++;

		if (comp->work_request_id == current_wrid) {

			break;
		}
	}

	result = sdp_buff_pool_chain_put(head, free_count);
	TS_EXPECT(MOD_LNX_SDP, !(0 > result));

	if (!(0 < free_count) || 0 > conn->send_usig) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: send processing error. mismatch. <%u:%u:%d:%d>",
			 comp->work_request_id, current_wrid, free_count,
			 conn->send_usig);
		result = -EINVAL;
		goto done;
	}
	/*
	 * Flush queued send data into the post queue if there is room.
	 */
	result = sdp_send_flush(conn);
	if (0 > result) {

		TS_TRACE(MOD_LNX_SDP, T_VERBOSE, TRACE_FLOW_WARN,
			 "EVENT: Failure during send queue flush. <%d>",
			 result);
		goto done;
	}

	return 0;
drop:
	(void)sdp_buff_pool_put(buff);
	(void)sdp_buff_pool_chain_put(head, free_count);
done:
	return result;
} /* sdp_event_send */
