/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

  $Id: srp_dm.c 479 2004-07-16 18:01:50Z roland $
*/

#include "srp_host.h"

static struct ib_path_record *srp_path_records;
static int last_used_path_record;
static int max_path_record_cache;

int srp_find_target(srp_host_port_params_t *port,
		    u64 service_name, tTS_IB_GID dgid, u16 dlid, int *tidx)
{
	int i;
	srp_target_t *target;
	u64 *p;

	p = (u64 *)dgid;

	for (i = 0; i < max_srp_targets; i++) {
		target = &port->hca->srp_targets[i];

		down(&target->sema);
		if ((target->service_name == service_name) &&
		    (target->port == port) &&
		    ((target->ioc.iou_path_record.dlid == dlid) ||
		     ((*p) && (*(p+1)) &&
		      (memcmp(target->ioc.iou_path_record.dgid, dgid, 16) == 0)))) {
			*tidx = i;
			up(&target->sema);
			return (TS_SUCCESS);
		}
		up(&target->sema);
	}
	return (TS_FAIL);
}

int srp_new_target(srp_host_port_params_t *port, tTS_IB_GUID guid,
		   u64 service_name, int *tidx)
{
	int i;
	srp_target_t *target;

	for (i = 0; i < max_srp_targets; i++) {
		target = &port->hca->srp_targets[i];

		if ((target->valid == FALSE) && (target->service_name == 0)) {
			down(&target->sema);
			TS_REPORT_STAGE(MOD_SRPTP,
					"Creating target Entry %d for 0x%llx", i,
					be64_to_cpu(*(u64 *) guid));

			memcpy(target->ioc.guid, guid, sizeof(tTS_IB_GUID));
			target->ioc.valid = TRUE;
			target->port = port;
			target->state = TARGET_POTENTIAL_CONNECTION;
			target->service_name = service_name;
			target->target_index = i;
			*tidx = i;
			up(&target->sema);
			return (TS_SUCCESS);
		}
	}

	return (TS_FAIL);
}

void srp_port_query_cancel(srp_host_port_params_t * port)
{
	struct list_head *temp_entry, *query_entry;
	srp_query_entry_t *query;
	tTS_IB_DM_CLIENT_HOST_TID dm_xid = TS_IB_DM_CLIENT_HOST_TID_INVALID;

	down(&driver_params.sema);

	list_for_each_safe(query_entry, temp_entry, &driver_params.query_list) {
		query = list_entry(query_entry, srp_query_entry_t, list);

		if (query->port == port) {
			int status = 0;

			TS_REPORT_STAGE(MOD_SRPTP,
					"Cancelling port query on hca %d, "
					"port %d",
					port->hca->hca_index + 1,
					port->local_port);

			list_del(&query->list);

			up(&driver_params.sema);
			switch (query->state) {
			case QUERY_PATH_RECORD_LOOKUP:
				status = ib_client_query_cancel(query->xid);
				if (status) {
					TS_REPORT_WARN(MOD_SRPTP,
						       "Query %d, cancel "
						       "failed %d",
						       query->id, status);
				}
				break;

			case QUERY_PORT_INFO:
				ib_host_io_query_cancel(query->dm_xid);
				break;

			default:
				TS_REPORT_FATAL(MOD_SRPTP,
						"Unknown query type");
				break;
			}
			down(&driver_params.sema);

			kfree(query);
		}
	}

	if (port->dm_query_in_progress == TRUE) {
		port->dm_query_in_progress = FALSE;
		port->dm_need_query = FALSE;

		TS_REPORT_STAGE(MOD_SRP, "Canceling DM Query on hca %d port %d",
				port->hca->hca_index + 1, port->local_port);

		dm_xid = port->dm_xid;
	}

	up(&driver_params.sema);

	if (dm_xid != TS_IB_DM_CLIENT_HOST_TID_INVALID) {
		ib_host_io_query_cancel(dm_xid);
	}
}

int srp_find_query_id(int id, srp_query_entry_t ** query_entry)
{
	struct list_head *temp_entry;

	list_for_each(temp_entry, &driver_params.query_list) {
		*query_entry = list_entry(temp_entry, srp_query_entry_t, list);

		/*
		 * If we already have a query in-progress, set the need_retry
		 * flag
		 */
		if ((*query_entry)->id == id) {
			return (TS_SUCCESS);
		}
	}

	*query_entry = NULL;
	return (TS_FAIL);
}

int srp_find_query(srp_host_port_params_t * port, u8 * gid)
{
	unsigned long cpu_flags;
	struct list_head *temp_entry;
	srp_query_entry_t *query;

	spin_lock_irqsave(&driver_params.spin_lock, cpu_flags);

	list_for_each(temp_entry, &driver_params.query_list) {
		query = list_entry(temp_entry, srp_query_entry_t, list);

		/*
		 * If we already have a query in-progress, set the need_retry
		 * flag
		 */
		if ((query->port == port) &&
		    (memcmp(query->remote_gid, gid, sizeof(tTS_IB_GID)) == 0)) {
			query->need_retry = TRUE;
			spin_unlock_irqrestore(&driver_params.spin_lock,
					       cpu_flags);
			return (TS_SUCCESS);
		}
	}

	spin_unlock_irqrestore(&driver_params.spin_lock, cpu_flags);

	return (TS_FAIL);
}

void srp_update_target_ioc(srp_target_t *target,
			   srp_host_port_params_t * port,
			   u16 dlid, u64 service_id,
			   struct ib_path_record *path_record)
{
	struct ib_path_record *svc_path_record;
	ioc_entry_t *ioc_entry = &target->ioc;

	target->port = port;
	svc_path_record = &ioc_entry->iou_path_record;
				
	ioc_entry->service_id = service_id;

	if (path_record)
		memcpy(svc_path_record, path_record, sizeof(*svc_path_record));

	if (svc_path_record->slid == 0)
		svc_path_record->slid = port->slid;
	if (svc_path_record->dlid == 0)
		svc_path_record->dlid = dlid;
	svc_path_record->pkey = 0xffff;
	memcpy(svc_path_record->sgid, port->local_gid, sizeof(tTS_IB_GID));

	ioc_entry->path_valid = TRUE;
}

int
srp_path_query_wakeup(tTS_IB_CLIENT_QUERY_TID tid,
		      int status,
		      struct ib_path_record *path_record,
		      int remaining, void *context)
{
	if ((status == 0) && (path_record != NULL)) {
		u64 *s = (u64 *)path_record->sgid;
		u64 *d = (u64 *)path_record->dgid;

		TS_REPORT_WARN(MOD_SRPTP,
			       "Path Record: mtu=%d slid=%x dlid=%x"
			       " sgid=0x%lx%lx dgid=0x%lx%lx",
			       path_record->mtu, path_record->slid,
			       path_record->dlid,
			       (unsigned long)cpu_to_be64(*s),
			       (unsigned long)cpu_to_be64(*(s+1)),
			       (unsigned long)cpu_to_be64(*d),
			       (unsigned long)cpu_to_be64(*(d+1)));
		memcpy(context, path_record, sizeof *path_record);
	} else
		TS_REPORT_WARN(MOD_SRPTP," path_query_wake error %d",status);

	up(&driver_params.path_query_sema);
	return 0;
}


int srp_process_io_service(struct ib_dm_client_host_io_svc *io_svc,
			   struct ib_path_record *path_record,
			   srp_host_port_params_t * port)
{
	char *service_name_str;
	u64 service_id;
	u64 service_name_cpu_endian, service_name;
	int tgt_index = 0;
	int status;
	srp_target_t *target;

	service_name_str = strchr(io_svc->svc_entry.service_name, ':');
	if (service_name_str == NULL) {
		TS_REPORT_WARN(MOD_SRPTP,
			       "unexpected service name %s",
			       io_svc->svc_entry.service_name);
		status = -EINVAL;
		goto exit;
	}
	TS_REPORT_STAGE(MOD_SRPTP,
			"Found service name %s on hca %d port %d",
			io_svc->svc_entry.service_name,
			port->hca->hca_index + 1, port->local_port);

	service_name_str++;
	StringToHex64(service_name_str, &service_name_cpu_endian);

	service_name = cpu_to_be64(service_name_cpu_endian);

	service_id = be64_to_cpu(*(u64 *) io_svc->svc_entry.service_id);
		    

	/*
	 * before we initialize the target,
	 * update the pathing information for the IOC
	 */
	status = srp_find_target(port, service_name, path_record->dgid,
				 io_svc->port_lid, &tgt_index);
	if (status == TS_FAIL) {
		status = srp_new_target(port, io_svc->controller_guid,
					service_name, &tgt_index);

		TS_REPORT_STAGE(MOD_SRPTP,
				"Target %llx has no entry yet, creating new "
				"entry index %d",
				be64_to_cpu(*(u64 *) io_svc->controller_guid),
					tgt_index);
						    
		if (status == TS_FAIL) {
			TS_REPORT_STAGE(MOD_SRPTP,
					"Target entry creation failed, "
					"too many targets");
			status = -ENOMEM;
			goto exit;
		}
	}

	target = &port->hca->srp_targets[tgt_index];

	down(&target->sema);
	srp_update_target_ioc(target, port, io_svc->port_lid,
			      service_id, path_record);

	if (target->valid == FALSE) {
		status = srp_host_alloc_pkts(target);
		if (status) {
			TS_REPORT_FATAL(MOD_SRPTP,
					"Could not allocate resource target %d",
					target->target_index);
				
		} else
			target->valid = TRUE;
	}
	port->new_tgt_insvc++;
	target->rescan_scsi = TRUE;
	up(&target->sema);

      exit:
	return status;
	
}

int
srp_process_io_svc_wakeup(tTS_IB_CLIENT_QUERY_TID tid,
			  int status,
			  struct ib_path_record *path_record,
			  int remaining, void *context)
{
	struct srp_io_service *s_io_svc = (struct srp_io_service *)context;

	if (!s_io_svc)
		return -EINVAL;

	if ((status == 0) && (path_record != NULL)) {
		u64 *s = (u64 *)path_record->sgid;
		u64 *d = (u64 *)path_record->dgid;

		TS_REPORT_WARN(MOD_SRPTP,
			       "Path Record: mtu=%d slid=%x dlid=%x"
			       " sgid=0x%lx%lx dgid=0x%lx%lx",
			       path_record->mtu, path_record->slid,
			       path_record->dlid,
			       (unsigned long)cpu_to_be64(*s),
			       (unsigned long)cpu_to_be64(*(s+1)),
			       (unsigned long)cpu_to_be64(*d),
			       (unsigned long)cpu_to_be64(*(d+1)));
		srp_process_io_service(&s_io_svc->io_svc, path_record,
				       s_io_svc->port);
				
	} else
		TS_REPORT_WARN(MOD_SRPTP," path_query_wake error %d",status);

	s_io_svc->port->responsed_targets++;

	TS_REPORT_WARN(MOD_SRPTP,
		       "potential_tgt=%d responsed_tgt=%d on [%d:%d]",
		       s_io_svc->port->potential_targets,
		       s_io_svc->port->responsed_targets,
		       s_io_svc->port->local_port,
		       s_io_svc->port->hca->hca_index + 1);
	if (s_io_svc->port->potential_targets ==
	    s_io_svc->port->responsed_targets) {
		s_io_svc->port->potential_targets = 0;
		s_io_svc->port->responsed_targets = 0;
		s_io_svc->port->dm_query_in_progress = FALSE;
		s_io_svc->port->dm_need_query = FALSE;
	}

	kfree(s_io_svc);
	return status;
}


void
srp_host_dm_completion(int status, struct ib_dm_client_host_io_list *io_list,
		       void *context)
{
	srp_host_port_params_t *port = (srp_host_port_params_t *) context;

	/*
	 * If we are shuting down, throw away the query
	 */
	if (driver_params.dm_shutdown) {
		if ((status == TS_SUCCESS) && (io_list)) {
			ib_host_io_list_free(io_list);
		}
		return;
	}

	down(&driver_params.sema);

	if ((status == TS_SUCCESS) && (io_list == NULL)) {
		TS_REPORT_STAGE(MOD_SRPTP,
				"DM Client Query complete hca %d port %d"
				" need_retry %d retry %d",
				port->hca->hca_index + 1, port->local_port,
				port->dm_need_retry, port->dm_retry_count);

		if (port->dm_need_retry == TRUE) {

			if (port->dm_retry_count++ < MAX_DM_RETRIES) {
				TS_REPORT_WARN(MOD_SRPTP,
					       "Restarting DM Query on hca %d "
					       "port %d timeout, retry "
					       "count %d",
					       port->hca->hca_index + 1,
					       port->local_port,
					       port->dm_retry_count);

				port->dm_need_query = TRUE;
			} else {
				TS_REPORT_WARN(MOD_SRPTP,
					       "DM Client timeout on hca %d "
					       "port %d, retry count %d "
					       "exceeded",
					       port->hca->hca_index + 1,
					       port->local_port,
					       port->dm_retry_count);
			}
		}

	} else if (status == -ETIMEDOUT) {
		TS_REPORT_WARN(MOD_SRPTP, "DM Client timeout on hca %d port %d",
			       port->hca->hca_index + 1, port->local_port);

		port->dm_need_retry = TRUE;

	} else if (status) {
		/*
		 * All other error cases
		 */
		TS_REPORT_WARN(MOD_SRPTP,
			       "DM Client error status %d from hca %d port %d",
			       status, port->hca->hca_index + 1,
			       port->local_port);
	} else {
		struct list_head *cur;
		struct ib_dm_client_host_io_svc *io_svc;
		struct srp_io_service *s_io_svc;
		u64 tid;

		TS_REPORT_STAGE(MOD_SRPTP,
				"DM Client Query complete hca %d port %d"
				" need_retry %d retry %d",
				port->hca->hca_index + 1, port->local_port,
				port->dm_need_retry, port->dm_retry_count);

		list_for_each(cur, &io_list->io_svc_list) {
			io_svc =
			    list_entry(cur, struct ib_dm_client_host_io_svc, list);
			s_io_svc = kmalloc(sizeof *s_io_svc, GFP_KERNEL);

			if (!s_io_svc) {
				ib_host_io_list_free(io_list);
				goto exit;
			}

			memcpy(&s_io_svc->io_svc, io_svc, sizeof *io_svc);
			s_io_svc->port = port;
			
			ib_get_path_record_from_lids(port->hca->ca_hndl,
						     port->local_port,
						     port->slid,
						     io_svc->port_lid,
						     20 * HZ,
						     srp_process_io_svc_wakeup,
						     (void *)s_io_svc,
						     &tid);
			port->potential_targets++;
		}

		ib_host_io_list_free(io_list);
	}

       exit:
	up(&driver_params.sema);
}

void srp_host_port_query_completion(int status,
				    struct ib_dm_client_host_io_list *io_list,
				    void *context)
{
	srp_query_entry_t *query_entry;
	srp_host_port_params_t *port;
	int error;

	/*
	 * If we are shuting down, throw away the query
	 */
	if (driver_params.dm_shutdown) {
		if ((status == TS_SUCCESS) && (io_list)) {
			ib_host_io_list_free(io_list);
		}
		return;
	}

	down(&driver_params.sema);

	error = srp_find_query_id((unsigned long)context, &query_entry);
	if (error) {
		TS_REPORT_WARN(MOD_SRPTP,
			       "Cannot find query id %d status %d, list %p",
			       context, status, io_list);
		goto normal_exit;
	}

	port = query_entry->port;

	if ((status == TS_SUCCESS) && (io_list == NULL)) {
		TS_REPORT_STAGE(MOD_SRPTP,
				"Port Query %d complete hca %d port %d",
				query_entry->id, port->hca->hca_index + 1,
				port->local_port);

		if (query_entry->need_retry == TRUE) {
			query_entry->need_retry = FALSE;

			driver_params.port_query = TRUE;

			ib_host_io_port_query(query_entry->port->hca->ca_hndl,
					      query_entry->port->local_port,
					      query_entry->remote_lid,
					      PORT_QUERY_TIMEOUT,
					      srp_host_port_query_completion,
					      (void *)(unsigned long)
					      query_entry->id,
					      &query_entry->dm_xid);
			goto normal_exit;
		} else {
			goto kill_query;
		}
	} else if ((status == -ETIMEDOUT) || (status == -EINVAL)) {
		TS_REPORT_WARN(MOD_SRPTP,
			       "Port Query %d timeout on hca %d port %d",
			       query_entry->id, port->hca->hca_index + 1,
			       port->local_port);

		if (++query_entry->retry < MAX_QUERY_RETRIES) {
			query_entry->need_retry = TRUE;
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(query_entry->retry * 5 * HZ);
			goto normal_exit;
		} else {
			TS_REPORT_WARN(MOD_SRPTP,
				       "Retries exceeded on hca %d port %d",
				       port->hca->hca_index + 1,
				       port->local_port);
			query_entry->need_retry = FALSE;
			goto kill_query;
		}
	} else if (status) {
		/*
		 * All other error cases
		 */
		TS_REPORT_WARN(MOD_SRPTP,
			       "Port Query %d error status %d from hca %d "
			       "port %d",
			       query_entry->id, status,
			       port->hca->hca_index + 1, port->local_port);
		goto kill_query;
	} else {
		struct list_head *cur;
		struct ib_dm_client_host_io_svc *io_svc;
		struct srp_io_service *s_io_svc;
		u64 tid;

		TS_REPORT_STAGE(MOD_SRPTP,
				"DM Client Query complete hca %d port %d"
				" need_retry %d retry %d",
				port->hca->hca_index + 1, port->local_port,
				port->dm_need_retry, port->dm_retry_count);

		list_for_each(cur, &io_list->io_svc_list) {
			io_svc =
			    list_entry(cur, struct ib_dm_client_host_io_svc, list);
			s_io_svc = kmalloc(sizeof *s_io_svc, GFP_KERNEL);

			if (!s_io_svc) {
				ib_host_io_list_free(io_list);
				goto normal_exit;
			}

			memcpy(&s_io_svc->io_svc, io_svc, sizeof *io_svc);
			s_io_svc->port = port;
			
			ib_get_path_record_from_lids(port->hca->ca_hndl,
						     port->local_port,
						     port->slid,
						     io_svc->port_lid,
						     20 * HZ,
						     srp_process_io_svc_wakeup,
						     (void *)s_io_svc,
						     &tid);
			port->potential_targets++;
		}
		ib_host_io_list_free(io_list);
	}

      kill_query:
	list_del(&query_entry->list);
	kfree(query_entry);
      normal_exit:
	up(&driver_params.sema);
	return;
}

static struct ib_path_record *srp_empty_path_record_cache(void)
{

	u32 index;

	struct ib_path_record unused_path_record;

	memset(&unused_path_record, 0, sizeof(struct ib_path_record));

	for (index = 0; index < max_path_record_cache; index++) {
		if (memcmp(&srp_path_records[index],
			   &unused_path_record,
			   sizeof(struct ib_path_record)) == 0) {
			if (index > last_used_path_record)
				last_used_path_record = index;

			return &srp_path_records[index];
		}
	}

	TS_REPORT_FATAL(MOD_SRPTP, "Couldnt find free path record entry");

	return NULL;
}

static void srp_flush_path_record_cache(srp_host_port_params_t * port,
					u8 * gid)
{
	u32 index;
	unsigned long cpu_flags;

	TS_REPORT_STAGE(MOD_SRPTP, "Flushing path record cache");

	spin_lock_irqsave(&driver_params.spin_lock, cpu_flags);

	for (index = 0; index < max_path_record_cache; index++) {
		if (memcmp(srp_path_records[index].sgid,
			   port->local_gid, sizeof(tTS_IB_GID)) == 0) {
			if (gid) {
				if (memcmp(srp_path_records[index].dgid,
					   gid, sizeof(tTS_IB_GID)) == 0) {
					memset(&srp_path_records[index], 0,
					       sizeof(struct ib_path_record));
					TS_REPORT_STAGE(MOD_SRPTP,
							" Flushing path "
							"record index %d",
							index);
				}
			} else {
				memset(&srp_path_records[index], 0,
				       sizeof(struct ib_path_record));
				TS_REPORT_STAGE(MOD_SRPTP,
						" Flushing path record "
						"index %d", index);
			}
		}
	}

	spin_unlock_irqrestore(&driver_params.spin_lock, cpu_flags);
}

static int srp_find_path_record(tTS_IB_GID find_gid,
				srp_host_port_params_t * port,
				struct ib_path_record **path_record)
{
	int index;

	for (index = 0; index <= last_used_path_record; index++) {
		if ((memcmp(srp_path_records[index].dgid,
			    find_gid,
			    sizeof(tTS_IB_GID)) == 0) &&
		    (srp_path_records[index].slid == port->slid)) {
			*path_record = &srp_path_records[index];
			return (TS_SUCCESS);
		}
	}

	return (TS_FAIL);
}

void srp_update_cache(struct ib_path_record *path_record,
		      srp_host_port_params_t * port)
{
	struct ib_path_record *cache_path_record;
	int status;

	status =
	    srp_find_path_record(path_record->dgid, port, &cache_path_record);

	if (status) {
		TS_REPORT_STAGE(MOD_SRPTP, "Updating path record cache");

		cache_path_record = srp_empty_path_record_cache();

		if (cache_path_record) {
			memcpy(cache_path_record, path_record,
			       sizeof(struct ib_path_record));
		}
	}
}

int srp_get_path_record(srp_host_conn_t * s,
			srp_host_port_params_t * port,
			tTS_IB_PATH_RECORD_COMPLETION_FUNC completion_function,
			void *completion_arg)
{
	int status;
	struct ib_path_record *path_record;
	int retry_count;
	u64 *rpg;

	retry_count = s->path_record_retry_count;

	rpg = (u64 *) s->redirected_port_gid;

	if ((*rpg) || (*(rpg + 1))) {
		status =
		    srp_find_path_record(s->redirected_port_gid, port,
					 &path_record);

		if (status == TS_SUCCESS) {
			TS_REPORT_STAGE(MOD_SRPTP,
					"Found Path Record in cache");

			completion_function(TS_IB_CLIENT_QUERY_TID_INVALID, 0,	/* status */
					    path_record, 0,	/* remaining */
					    completion_arg);
			goto exit;
		}
		status = tsIbPathRecordRequest(port->hca->ca_hndl,
					       port->local_port,
					       port->local_gid,
					       s->redirected_port_gid,
					       TS_IB_SA_INVALID_PKEY,
					       TS_IB_PATH_RECORD_ALLOW_DUPLICATE,
					       (retry_count + 1) * 2 * HZ,
					       (retry_count + 1) * 2 * HZ,
					       completion_function,
					       completion_arg,
					       &s->path_record_tid);
	} else {
		status = ib_get_path_record_from_lids(port->hca->ca_hndl,
						      port->local_port,
						      port->slid,
						      s->path_record.dlid,
						      (retry_count +
						       1) * 2 * HZ,
						      completion_function,
						      completion_arg,
						      &s->path_record_tid);
	}

	if (status) {
		TS_REPORT_WARN(MOD_SRPTP,
			       "tsIbPathRecordRequest failed %d "
			       "hca %d port %d",
			       status, port->hca->hca_index + 1,
			       port->local_port);
	}
      exit:
	return status;
}

int
srp_trap_path_record_completion(tTS_IB_CLIENT_QUERY_TID tid,
				int status,
				struct ib_path_record *path_record,
				int remaining, void *context)
{
	srp_query_entry_t *query_entry;
	srp_host_port_params_t *port;
	int error;

	down(&driver_params.sema);

	error = srp_find_query_id((unsigned long)context, &query_entry);
	if (error) {
		TS_REPORT_WARN(MOD_SRPTP, "Cannot find query id %d", context);
		up(&driver_params.sema);
		return (0);
	}

	port = query_entry->port;

	if (status) {
		if (status == -ETIMEDOUT) {
			if (++query_entry->retry < MAX_PATH_RECORD_RETRIES) {
				TS_REPORT_WARN(MOD_SRPTP,
					       "Query %d, Path Record timeout "
					       "on hca %d port %d",
					       query_entry->id,
					       port->hca->hca_index + 1,
					       port->local_port);

				status = tsIbPathRecordRequest(port->hca->ca_hndl,
							       port->local_port, 
							       port->local_gid,
							       query_entry->remote_gid, 
							       TS_IB_SA_INVALID_PKEY, 
							       TS_IB_PATH_RECORD_ALLOW_DUPLICATE,	/* flags */
							       PATH_RECORD_TIMEOUT,	/* timeout */
							       PATH_RECORD_TIMEOUT,	/* XXX cache jiffies */
							       srp_trap_path_record_completion,
							       (void *)(unsigned
									long)
							       query_entry->id,
							       &query_entry->
							       xid);
				if (status) {
					TS_REPORT_WARN(MOD_SRPTP,
						       " tsIbPathRecordRequest "
						       "failed status %d for "
						       "hca %d port %d",
						       status,
						       port->hca->hca_index + 1,
						       port->local_port);
					goto kill_query;
				}
			} else {
				TS_REPORT_WARN(MOD_SRPTP,
					       "Path Record exceeded retries "
					       "status %d, killing query on "
					       "hca %d port %d",
					       status, port->hca->hca_index + 1,
					       port->local_port);
				goto kill_query;
			}
		} else {
			TS_REPORT_WARN(MOD_SRPTP,
				       "Query %d, Path Record status %d, "
				       "killing query on hca %d port %d",
				       query_entry->id, status,
				       port->hca->hca_index + 1,
				       port->local_port);
			goto kill_query;
		}
	} else {
		u8 *gid = path_record->dgid;

		TS_REPORT_STAGE(MOD_SRPTP, "New path_rec DLID=0x%x DGID= "
				"%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x"
				"%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x",
				path_record->dlid, gid[0], gid[1], gid[2],
				gid[3], gid[4], gid[5], gid[6], gid[7],
				gid[8], gid[9], gid[10], gid[11],
				gid[12], gid[13], gid[14], gid[15]);
			      
		srp_update_cache(path_record, port);
 
		query_entry->remote_lid = path_record->dlid;
		query_entry->retry = 0;

		query_entry->state = QUERY_PORT_INFO;

		driver_params.port_query = TRUE;

		ib_host_io_port_query(port->hca->ca_hndl,
				      port->local_port,
				      path_record->dlid,
				      PORT_QUERY_TIMEOUT,
				      srp_host_port_query_completion,
				      (void *)(unsigned long)query_entry->id,
				      &query_entry->dm_xid);
	}

	up(&driver_params.sema);
	return (0);

      kill_query:
	list_del(&query_entry->list);
	kfree(query_entry);

	up(&driver_params.sema);
	return (0);
}

void srp_dm_notice_handler(struct ib_common_attrib_notice *notice,
			   struct ib_device *dev_hndl,
			   tTS_IB_PORT local_port, u16 io_port_lid, void *arg)
{
	u8 *gid = notice->detail.sm_trap.gid;
	int hca_index;
	srp_host_hca_params_t *hca;
	srp_host_port_params_t *port;
	srp_query_entry_t *query_entry;
	int id;
	struct ib_path_record *path_record;

	for (hca_index = 0; hca_index < MAX_HCAS; hca_index++) {
		if (hca_params[hca_index].ca_hndl == dev_hndl)
			break;
	}

	if (hca_index == MAX_HCAS) {
		TS_REPORT_FATAL(MOD_SRPTP, "Async event for unknown HCA 0x%x",
				dev_hndl);
		return;
	}

	hca = &hca_params[hca_index];
	port = &hca->port[local_port - 1];

	TS_REPORT_WARN(MOD_SRPTP, "DM async notification HCA %d port %d",
		       hca_index + 1, local_port);

	down(&driver_params.sema);

	/*
	 * Query already outstanding, do nothing
	 */
	if (srp_find_query(port, gid) == TS_SUCCESS) {
		up(&driver_params.sema);
		return;
	}

	TS_REPORT_WARN(MOD_SRPTP,
		       "New GID %02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x LID %x",
		       notice->detail.sm_trap.gid[0],
		       notice->detail.sm_trap.gid[1],
		       notice->detail.sm_trap.gid[2],
		       notice->detail.sm_trap.gid[3],
		       notice->detail.sm_trap.gid[4],
		       notice->detail.sm_trap.gid[5],
		       notice->detail.sm_trap.gid[6],
		       notice->detail.sm_trap.gid[7],
		       notice->detail.sm_trap.gid[8],
		       notice->detail.sm_trap.gid[9],
		       notice->detail.sm_trap.gid[10],
		       notice->detail.sm_trap.gid[11],
		       notice->detail.sm_trap.gid[12],
		       notice->detail.sm_trap.gid[13],
		       notice->detail.sm_trap.gid[14],
		       notice->detail.sm_trap.gid[15], io_port_lid);

	query_entry = kmalloc(sizeof(srp_query_entry_t), GFP_KERNEL);
	if (query_entry == NULL) {
		TS_REPORT_FATAL(MOD_SRPTP, "kmalloc failed in "
				"srp_dm_notice_handler");
		return;
	}

	memset(query_entry, 0, sizeof(srp_query_entry_t));

	memcpy(query_entry->remote_gid, gid, sizeof(tTS_IB_GID));

	query_entry->port = port;

	query_entry->state = QUERY_PATH_RECORD_LOOKUP;

	path_record = kmalloc(sizeof(*path_record), GFP_KERNEL);
	if (!path_record) {
		TS_REPORT_FATAL(MOD_SRPTP, "failed kmalloc path_record");
		kfree(query_entry);
		return;
	}

	memset(path_record, 0, sizeof(*path_record));
	path_record->dlid = io_port_lid;

	list_add_tail(&query_entry->list, &driver_params.query_list);
	id = query_entry->id = driver_params.query_count++;

	up(&driver_params.sema);

	srp_trap_path_record_completion(0,
					0,
					path_record,
					0, (void *)(unsigned long)id);

	kfree(path_record);
}

/*
 * In Service Notice Handler.
 */
void
srp_in_service_handler(struct ib_common_attrib_notice *notice,
		       tTS_IB_PORT port, void *arg)
{
	srp_host_port_params_t *srp_port = (srp_host_port_params_t *) arg;
	u8 *gid = notice->detail.sm_trap.gid;

	TS_REPORT_WARN(MOD_SRPTP,
		       "New IN svc GID %02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x"
		       "%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x on hca %d, port %d",
		       gid[0], gid[1], gid[2], gid[3], gid[4], gid[5],
		       gid[6], gid[7], gid[8], gid[9], gid[10], gid[11],
		       gid[12], gid[13], gid[14], gid[15],
		       srp_port->hca->hca_index + 1, srp_port->local_port);

	srp_port->dm_need_query = TRUE;
}

/*
 * Out Of Service Notice Handler.
 */
void
srp_out_of_service_handler(struct ib_common_attrib_notice *notice,
			   tTS_IB_PORT local_port, void *arg)
{

	u8 *gid = notice->detail.sm_trap.gid;
	srp_host_port_params_t *port = (srp_host_port_params_t *) arg;
	srp_target_t *target;

	TS_REPORT_WARN(MOD_SRPTP,
		       "New OUT_of_svc GID %02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x"
		       "%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x dlid %x" 
		       " on hca %d, port %d",
		       gid[0], gid[1], gid[2], gid[3], gid[4], gid[5],
		       gid[6], gid[7], gid[8], gid[9], gid[10], gid[11],
		       gid[12], gid[13], gid[14], gid[15],
		       notice->define.generic.issuer_lid,
		       port->hca->hca_index + 1, port->local_port);

	srp_flush_path_record_cache(port, gid);

	/*
	 * Right now, we get notified about any IB port going down on the IB
	 * fabric. Scan through our list of all paths for all targets and see
	 * if their redirected path matches the GID that went down.
	 */

	if (!port->hca->srp_targets)
		return;

	for (target = &port->hca->srp_targets[0];
	     target < &port->hca->srp_targets[max_srp_targets]; target++) {

		if (target->valid == FALSE)
			continue;

		down(&target->sema);

		if ((target->state != TARGET_ACTIVE_CONNECTION) ||
		    (!target->active_conn)) {
			up(&target->sema);
			continue;
		}

		if ((target->port == port) &&
		    ((memcmp(target->active_conn->path_record.dgid,
			     gid, sizeof(tTS_IB_GID)) == 0) ||
		     (target->active_conn->path_record.dlid ==
		      notice->define.generic.issuer_lid))) {
			   
			TS_REPORT_WARN(MOD_SRPTP,
				       "Target [%d:%d:%d], ...out_of_svc",
				       target->target_index, port->local_port,
				       port->hca->hca_index + 1);

			target->need_disconnect = TRUE;
		}

		up(&target->sema);
	}
}

void srp_out_of_service_completion(tTS_IB_CLIENT_QUERY_TID transaction_id,
				   int status,
				   struct ib_common_attrib_inform *inform,
				   void *arg)
{
	srp_host_port_params_t *port = (srp_host_port_params_t *) arg;

	if (transaction_id == port->out_of_service_xid) {
		if (status) {
			
			/* trap failed to register */
			TS_REPORT_WARN(MOD_SRPTP,
				       "Out of service error on hca %d port %d "
				       "status %d",
				       port->local_port,
				       port->hca->hca_index + 1, status);

			if ((status == -ETIMEDOUT) &&
			    (++port->outof_svc_count < MAX_DM_RETRIES))
				srp_register_out_of_service(port, TRUE);
			else
				TS_REPORT_WARN(MOD_SRPTP, "Unhandled error");
			
		} else {
			TS_REPORT_STAGE(MOD_SRPTP,
					"Out of service trap for hca %d "
					"port %d complete",
					port->hca->hca_index + 1,
					port->local_port);
			port->outof_svc_count = 0;
		}
	}
}

void srp_in_service_completion(tTS_IB_CLIENT_QUERY_TID transaction_id,
			       int status,
			       struct ib_common_attrib_inform *inform,
			       void *arg)
{
	srp_host_port_params_t *port = (srp_host_port_params_t *) arg;

	if (transaction_id == port->in_service_xid) {
		if (status) {
			
			/* trap failed to register */
			TS_REPORT_WARN(MOD_SRPTP,
				       "In service failed on hca %d port %d "
				       "status %d",
				       port->local_port,
				       port->hca->hca_index + 1, status);

			if ((status == -ETIMEDOUT) &&
			    (++port->in_svc_count < MAX_DM_RETRIES))
				srp_register_in_service(port, TRUE);
			else
				TS_REPORT_WARN(MOD_SRPTP,
					       "Unhandled error for in-service "
					       "trap handler");
			
		} else {
			TS_REPORT_STAGE(MOD_SRPTP,
					"In service trap for hca %d port %d "
					"complete",
					port->hca->hca_index + 1,
					port->local_port);
			port->in_svc_count = 0;
		}
	}
}

void srp_register_in_service(srp_host_port_params_t * port, int flag)
{
	int err_code;
	tTS_IB_GID zero_port_gid;
	tTS_IB_SA_NOTICE_HANDLER_FUNC handler;
	tTS_IB_INFORM_INFO_SET_COMPLETION_FUNC completion_handler;

	if (port->valid == FALSE)
		return;

	if (flag) {
		handler = srp_in_service_handler;
		completion_handler = srp_in_service_completion;

		if (port->port_state != IB_PORT_STATE_ACTIVE)
			return;

		TS_REPORT_STAGE(MOD_SRPTP,
				"Registering hca %d local port %d for IB "
				"in service traps",
				port->hca->hca_index + 1, port->local_port);
	} else {
		handler = NULL;
		completion_handler = NULL;

		TS_REPORT_STAGE(MOD_SRPTP,
				"Deregistering hca %d local port %d for IB "
				"in service traps",
				port->hca->hca_index + 1, port->local_port);
	}

	memset(zero_port_gid, 0, sizeof(tTS_IB_GID));
	err_code = tsIbSetInServiceNoticeHandler(port->hca->ca_hndl,	/* hca */
						 port->local_port,	/* local port */
						 zero_port_gid,	/* zero port GID */
						 0xffff,	/* LID begin, get traps for all ports */
						 0,	/* LID end, ignored */
						 handler,	/* out of service handler */
						 port,	/* argument */
						 5 * HZ,	/* timeout */
						 completion_handler,	/* completion function */
						 port,	/* completion arg */
						 &port->in_service_xid);
	if (err_code)
		
		TS_REPORT_FATAL(MOD_SRPTP,
				"InServiceNoticeHandler failed (%d) ",
				err_code);
	
}

void srp_register_out_of_service(srp_host_port_params_t * port, int flag)
{

	int err_code;
	tTS_IB_GID zero_port_gid;
	tTS_IB_SA_NOTICE_HANDLER_FUNC handler;
	tTS_IB_INFORM_INFO_SET_COMPLETION_FUNC completion_handler;

	if (port->valid == FALSE)
		return;

	if (flag) {
		handler = srp_out_of_service_handler;
		completion_handler = srp_out_of_service_completion;

		if (port->port_state != IB_PORT_STATE_ACTIVE)
			return;

		TS_REPORT_STAGE(MOD_SRPTP,
				"Registering hca %d local port %d for IB out "
				"of service traps",
				port->hca->hca_index + 1, port->local_port);
	} else {
		handler = NULL;
		completion_handler = NULL;

		TS_REPORT_STAGE(MOD_SRPTP,
				"Deregistering hca %d local port %d for IB "
				"out of service traps",
				port->hca->hca_index + 1, port->local_port);
	}

	memset(zero_port_gid, 0, sizeof(tTS_IB_GID));

	err_code = tsIbSetOutofServiceNoticeHandler(port->hca->ca_hndl,	/* hca */
						    port->local_port,	/* local port */
						    zero_port_gid,	/* zero port GID */
						    0xffff,	/* LID begin, get traps for all ports */
						    0,	/* LID end, ignored */
						    handler,	/* out of service handler */
						    port,	/* argument */
						    5 * HZ,	/* timeout */
						    completion_handler,	/* completion function */
						    port,	/* completion arg */
						    &port->out_of_service_xid);
	if (err_code)
		
		TS_REPORT_FATAL(MOD_SRPTP,
				"OutOfServiceNoticeHandler failed (%d) ",
				err_code);
	
}

int srp_dm_query(srp_host_port_params_t * port)
{
	int status = TS_FAIL;

	if (port->port_state == IB_PORT_STATE_ACTIVE) {

		port->dm_query_in_progress = TRUE;
		port->dm_need_retry = FALSE;

		TS_REPORT_STAGE(MOD_SRPTP,
				"DM Query Initiated on hca %d local port %d"
				" need_retry %d retry %d",
				port->hca->hca_index + 1, port->local_port,
				port->dm_need_retry, port->dm_retry_count);

		if ((status = ib_host_io_query(port->hca->ca_hndl,
					       port->local_port,
					       20 * HZ,
					       srp_host_dm_completion,
					       (void *)port, &port->dm_xid))) {

			
			TS_REPORT_WARN(MOD_SRPTP,
				       "tsIbHostIoQuery failed status 0x%x",
				       status);

			port->dm_query_in_progress = FALSE;
			
		}
	}

	return (status);
}

void srp_hca_async_event_handler(struct ib_async_event_record *event, void *arg)
{
	int hca_index, port_index;
	srp_host_port_params_t *port;
	srp_host_hca_params_t *hca;
	srp_target_t *target;

	for (hca_index = 0; hca_index < MAX_HCAS; hca_index++) {
		if (hca_params[hca_index].ca_hndl == event->device)
			break;
	}

	if (hca_index == MAX_HCAS) {
		
		TS_REPORT_FATAL(MOD_SRPTP, "Async event for unknown HCA %d",
				event->device);
		return;
		
	}

	hca = &hca_params[hca_index];
	port = &hca->port[event->modifier.port - 1];

	switch (event->event) {

	case IB_LID_CHANGE:
		TS_REPORT_WARN(MOD_SRPTP,
			       "LID_CHANGE event for hca %d port %d",
			       hca_index + 1, event->modifier.port);
		if (port->valid == FALSE)
			break;

		srptp_refresh_port_info(port);

		if (!hca->srp_targets)
			break;

		for (target = &hca->srp_targets[0];
		     target < &hca->srp_targets[max_srp_targets]; target++) {

			if ((target->valid == FALSE) ||
			    (target->port != port))
				continue;

			srp_move_to_pending(target);
			srp_pending_to_scsi(target, DID_SOFT_ERROR);
							    
			if (target->state == TARGET_ACTIVE_CONNECTION) {
				srp_host_conn_t *conn =
				    target->active_conn;

				TS_REPORT_STAGE(MOD_SRPTP,
						" Target %d active "
						"conn thru port %d "
						"active_count %d",
						target->target_index,
						conn->port->local_port,
						target->active_count);

				memset(&conn->path_record, 0,
				       sizeof(struct ib_path_record));

				remove_connection(conn,
						  TARGET_POTENTIAL_CONNECTION);
			}
		}

		down(&driver_params.sema);
		port->dm_need_query = TRUE;
		up(&driver_params.sema);

		break;

	case IB_PKEY_CHANGE:
	case IB_PORT_ACTIVE:
		/*
		 * Wake up that DM thread, so that it will start
		 * a fresh scan and initiate connections to discovered
		 * services
		 */
		TS_REPORT_WARN(MOD_SRPTP,
			       "Port active event for hca %d port %d",
			       hca_index + 1, event->modifier.port);

		if (port->valid == FALSE)
			break;

		srptp_refresh_port_info(port);

		port->outof_svc_count = 0;
		if (port->out_of_service_xid) {
			srp_register_out_of_service(port, FALSE);
			port->out_of_service_xid = 0;
		}
		srp_register_out_of_service(port, TRUE);

		port->in_svc_count = 0;
		if (port->in_service_xid) {
			srp_register_in_service(port, FALSE);
			port->in_service_xid = 0;
		}
		srp_register_in_service(port, TRUE);
		
		down(&driver_params.sema);
		port->dm_need_query = TRUE;
		up(&driver_params.sema);
		break;

	case IB_LOCAL_CATASTROPHIC_ERROR:

		TS_REPORT_WARN(MOD_SRPTP,
			       "Got \"local catastrophic error\" from "
			       "the HCA %d", hca_index);
		TS_REPORT_WARN(MOD_SRPTP,
			       "Moving all targets to offline state");
		/*
		 * As far as we are concerned,
		 * a local catastrophic error is equivalent to
		 * all the local ports going into error state
		 */
		for (port_index = 0;
		     port_index < MAX_LOCAL_PORTS_PER_HCA;
		     port_index++) {
			if (hca->port[port_index].valid == FALSE)
				continue;

			event->event = IB_PORT_ERROR;

			event->modifier.port =
			    hca->port[port_index].local_port;

			srp_hca_async_event_handler(event, NULL);
		}
		break;

	case IB_PORT_ERROR:
		TS_REPORT_WARN(MOD_SRPTP,
			       "Port error event for hca %d port %d",
			       hca_index + 1, event->modifier.port);

		if (port->valid == FALSE)
			break;

		/*
		 * Cannot call refresh hca info as the HCA may be hung,
		 * simply mark the port as being down
		 */
		port->port_state = IB_PORT_STATE_DOWN;

		/*
		 * We now have a updated path table... Disconnect all
		 * connections for this given hca and port and try
		 * reconnecting.
		 */
		if (!hca->srp_targets)
			break;

		for (target = &hca->srp_targets[0];
		     target < &hca->srp_targets[max_srp_targets]; target++) {
				
			if ((target->valid == FALSE) ||
			    (target->port != port))
				continue;

			srp_move_to_pending(target);
			srp_pending_to_scsi(target, DID_SOFT_ERROR);
							    
			down(&target->sema);

			if (target->state == TARGET_ACTIVE_CONNECTION) {
				srp_host_conn_t *conn =
				    target->active_conn;

				TS_REPORT_STAGE(MOD_SRPTP,
						" Target %d active "
						"conn thru port %d "
						"active_count %d",
						target->target_index,
						conn->port->local_port,
						target->active_count);

				memset(&conn->path_record, 0,
				       sizeof(struct ib_path_record));

				remove_connection(conn,
						  TARGET_POTENTIAL_CONNECTION);
			}

			up(&target->sema);
		}
	
		port->outof_svc_count = 0;
		if (port->out_of_service_xid) {
			srp_register_out_of_service(port, FALSE);
			port->out_of_service_xid = 0;
		}
			
		port->in_svc_count = 0;
		if (port->in_service_xid) {
			srp_register_in_service(port, FALSE);
			port->in_service_xid = 0;
		}

		srp_port_query_cancel(port);
		srp_flush_path_record_cache(port, NULL);
		break;

	default:
		TS_REPORT_FATAL(MOD_SRPTP, "Unsupported event type %d",
				event->event);
		break;
	}
}

int srp_dm_init()
{
	int i, async_event_index, hca_index, status;
	srp_host_hca_params_t *hca;
	struct ib_async_event_record async_record;

	max_path_record_cache = max_srp_targets;

	/*
	 * Initialize path record cache
	 * Allow for one path record entry for each per target
	 */
	srp_path_records =
	    (struct ib_path_record *)kmalloc(sizeof(struct ib_path_record) *
					     max_path_record_cache, GFP_KERNEL);

	if (srp_path_records == NULL) {
		TS_REPORT_FATAL(MOD_SRPTP, "kmalloc failed srp_targets");
		return -ENOMEM;
	}

	/*
	 * Setup path record cache
	 */
	last_used_path_record = 0;

	memset(srp_path_records, 0,
	       sizeof(struct ib_path_record) * max_path_record_cache);

	/*
	 * Register for async events on the HCA
	 */
	for (hca_index = 0; hca_index < MAX_HCAS; hca_index++) {
		hca = &hca_params[hca_index];

		if (hca->valid == FALSE)
			break;

		TS_REPORT_STAGE(MOD_SRPTP,
				"Registering async events handler for HCA %d",
				hca->hca_index);

		async_record.device = hca->ca_hndl;

		async_event_index = IB_LOCAL_CATASTROPHIC_ERROR;
		for (i = 0; i < MAX_ASYNC_EVENT_HANDLES; i++) {
			async_record.event = async_event_index;
			status = ib_async_event_handler_register(&async_record,
								 srp_hca_async_event_handler,
								 hca,
								 &hca->
								 async_handles
								 [i]);

			if (status) {
				TS_REPORT_FATAL(MOD_SRPTP,
						"Registration of async event "
						"%d on hca %d failed",
						i, hca->hca_index + 1, status);
				return (-EINVAL);
			}

			async_event_index++;
		}
	}

	sema_init(&driver_params.sema, 1);
	sema_init(&driver_params.path_query_sema, 0);

	/* Register with DM to register for async notification */
	ib_dm_async_notify_register(TS_IB_DM_NOTICE_READY_TO_TEST_TRAP_NUM,
				    srp_dm_notice_handler, NULL);

	return (0);
}

void srp_dm_unload()
{
	srp_host_hca_params_t *hca;
	int i, hca_index;

	/*
	 * Unegister for async events on the HCA
	 */
	for (hca_index = 0; hca_index < MAX_HCAS; hca_index++) {
		hca = &hca_params[hca_index];

		if (hca->valid == FALSE)
			break;

		TS_REPORT_STAGE(MOD_SRPTP,
				"DeRegistering async event handlers for HCA %d",
				hca_index);

		/*
		 * Loop through the async handles for the HCA and
		 * deregister them.
		 */
		for (i = 0; i < MAX_ASYNC_EVENT_HANDLES; i++) {
			ib_async_event_handler_deregister(hca->
							  async_handles[i]);
		}
	}

	/* Register with DM to register for async notification */
	ib_dm_async_notify_register(TS_IB_DM_NOTICE_READY_TO_TEST_TRAP_NUM,
				    NULL, NULL);

	kfree(srp_path_records);
}
