/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.

  $Id$
*/

#include "srp_host.h"
#include <linux/moduleparam.h>

int max_srp_targets = MAX_SRP_TARGETS;
module_param(max_srp_targets, int, 0444);
MODULE_PARM_DESC(max_srp_targets,
		 "max_srp_targets - number of targets that will be supported");

int max_cmds_per_lun = MAX_CMDS_PER_LUN;
module_param(max_cmds_per_lun, int, 0444);
MODULE_PARM_DESC(max_cmds_per_lun,
		 "max_cmds_per_lun - number of ios per lun supported");

int max_xfer_sectors_per_io = MAX_SECTORS_PER_IO;
module_param(max_xfer_sectors_per_io, int, 0444);
MODULE_PARM_DESC(max_xfer_sectors_per_io,
		 "max_xfer_sectors_per_io - maximum number of sectors (512 bytes each)"
		 "which can be transferred per IO (default is 128 (64KB)");

u32 srp_discovery_timeout = IB_DISCOVERY_TIMEOUT;	/* 60 seconds */
module_param(srp_discovery_timeout, int, 0444);
MODULE_PARM_DESC(srp_discovery_timeout,
		 "timeout (in seconds) for SRP discovery (of SRP targets) to complete");

int srp_tracelevel = 2;
module_param(srp_tracelevel, int, 0444);
MODULE_PARM_DESC(srp_tracelevel, "Debug tracelevel");

#define INVALID_CONN_HANDLE (srp_host_conn_t *)0xdeadbeef

/*
 * Local prototypes
 *
 */
const char *srp_host_info(struct Scsi_Host *);

int srp_host_login(srp_host_conn_t *);

static int srp_host_login_resp(srp_host_conn_t *, char *);

static int srp_host_login_rej(char *);

static int srp_process_resp(srp_target_t *, srp_pkt_t *);

void srp_host_iodone(struct scsi_cmnd *, srp_resp_t *, ioq_t *);

int srp_build_command(srp_pkt_t *, struct scsi_cmnd *, srp_host_conn_t *,
		      ioq_t * ioq);

void srp_select_queue_depths(struct Scsi_Host *, struct scsi_device *);

int srp_host_detect(struct scsi_host_template *template);

int srp_host_qcommand(struct scsi_cmnd *SCpnt,
		      void (*done) (struct scsi_cmnd *));

int srp_host_abort_eh(struct scsi_cmnd *SCpnt);

int srp_host_device_reset_eh(struct scsi_cmnd *SCpnt);

int srp_host_reset_eh(struct scsi_cmnd *SCpnt);

int srp_build_rdma(srp_pkt_t *, struct scsi_cmnd *, srp_host_conn_t *,
		   ioq_t * ioq, srp_cmd_t *);

int srp_proc_info(struct Scsi_Host *host, char *buffer, char **start,
		  off_t offset, int length, int inout);

static void srp_host_module_cleanup(void);

void srp_host_resource_timer(void *context);

int srp_host_send_command(srp_target_t * target, ioq_t * ioq,
			  struct scsi_cmnd *SCpnt);

void srp_host_internal_qcommand(ioq_t * ioq);

int srp_path_record_completion(tTS_IB_CLIENT_QUERY_TID tid,
			       int status,
			       struct ib_path_record *path_record,
			       int remaining, void *context);
void srp_add_target_host(srp_target_t *target);

srp_host_driver_params_t driver_params = {0};
srp_host_hca_params_t hca_params[MAX_HCAS];

struct scsi_host_template driver_template = {
	.module = THIS_MODULE,
	.queuecommand = srp_host_qcommand,
	.eh_abort_handler = srp_host_abort_eh,
	.eh_device_reset_handler = srp_host_device_reset_eh,
	.eh_host_reset_handler = srp_host_reset_eh,
	.proc_info = srp_proc_info,
	.proc_name = "srp",
	.can_queue = MAX_IO_QUEUE,
	.this_id = -1,
	.sg_tablesize = MAX_SG_TABLESIZE,
	.max_sectors = MAX_SECTORS_PER_IO,
	.cmd_per_lun = MAX_CMDS_PER_LUN,
	.use_clustering = ENABLE_CLUSTERING,
};

static int scsi_unload_in_progress = FALSE;
static int start_scsi_scan_host = FALSE;

DECLARE_WAIT_QUEUE_HEAD(dm_poll_thread_wait_queue);

static inline srp_target_t *host_to_target(struct Scsi_Host *host)
{
	return (srp_target_t *)(unsigned long)(*(unsigned long *)host->hostdata);
}

void srp_free_srlist(srp_host_hca_params_t *hca, ioq_t * ioq,
		     struct scsi_cmnd *cmnd)
{

#if 0
	if (cmnd->use_sg)
		pci_unmap_sg(hca->hca_pdev,
			     (struct scatterlist *)cmnd->request_buffer,
			     cmnd->use_sg,
			     cmnd->sc_data_direction);
	else {
		struct scatterlist scatter_list;

		memset(&scatter_list, 0, sizeof(struct scatterlist));
		scatter_list.page = virt_to_page(cmnd->request_buffer);
		scatter_list.offset =
		    ((unsigned long)cmnd->request_buffer & ~PAGE_MASK);
		scatter_list.length = cmnd->request_bufflen;
		pci_unmap_sg(hca->hca_pdev, &scatter_list, 1,
			     cmnd->sc_data_direction);
	}
#endif

	if (ioq->sr_list) {
		u32 sr_list_index = 0;
		for (; sr_list_index < ioq->sr_list_length; sr_list_index++) {
			srptp_dereg_phys_host_buf(ioq->sr_list + sr_list_index);
		}
		kfree(ioq->sr_list);
		ioq->sr_list = NULL;
	}
}

void tm_timedout(unsigned long ptr)
{
	ioq_t *ioq = (ioq_t *) ptr;
	ioq->tm_resp_code = TSK_MGT_TIMEOUT;
	up(&ioq->sema);		/* awaken the process */
}

__inline void srp_host_scsi_complete(struct scsi_cmnd *scsi_cmnd)
{
	unsigned long cpu_flags;

	/* complete back to scsi with DID_ERROR */
	spin_lock_irqsave(&scsi_cmnd->device->host->default_lock, cpu_flags);

	scsi_cmnd->SCp.buffer = NULL;
	(*scsi_cmnd->scsi_done) (scsi_cmnd);

	spin_unlock_irqrestore(&scsi_cmnd->device->host->default_lock, cpu_flags);
}


void srp_host_cmd_timeout(unsigned long ptr)
{
	ioq_t *ioq = (ioq_t *) ptr;
	srp_target_t *target = ioq->target;
	unsigned long cpu_flags;

	del_timer(&ioq->timer);
	ioq->timer.function = NULL;

	if (ioq->req && target && (target->valid == TRUE) &&
	    target->active_conn &&
	    (target->state == TARGET_ACTIVE_CONNECTION)) {
		spin_lock_irqsave(&target->spin_lock, cpu_flags);
		target->active_count--;
		list_del(&ioq->active_list);

		ioq->req_temp = ioq->req;
		ioq->req = NULL;

		if (ioq->req_temp->eh_timeout.function) {
			scsi_add_timer(ioq->req_temp, IOQ_CMD_EXTEND_TIMEOUT,
				       ioq->req_temp->eh_timeout.function);
			TS_REPORT_WARN(MOD_SRPTP,
				       "extended timeout for SCpnt %p ioq %p",
				       ioq->req_temp, ioq);
		}

		ioq->req_temp->result = DID_SOFT_ERROR << 16;

		if (ioq->req_temp->request_bufflen)
			srp_free_srlist(target->port->hca, ioq, ioq->req_temp);
		ioq->req_temp->SCp.buffer = NULL;
		srp_host_scsi_complete(ioq->req_temp);
		kfree(ioq);

		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);
	}			
}

srp_host_conn_t *srp_host_find_conn(srp_target_t * target,
				    tTS_IB_CM_COMM_ID comm_id)
{
	struct list_head *conn_entry, *temp_entry;
	srp_host_conn_t *conn;

	list_for_each_safe(conn_entry, temp_entry, &target->conn_list) {

		conn = list_entry(conn_entry, srp_host_conn_t, conn_list);

		if (conn->comm_id == comm_id)
			return (conn);

	}

	return (NULL);
}

void srp_fmr_flush_function(struct ib_fmr_pool *fmr_pool, void *flush_arg)
{
	srp_host_hca_params_t *hca = (srp_host_hca_params_t *) flush_arg;
	ioq_t *ioq = NULL;
	struct list_head *ioq_entry, *temp_entry;
	unsigned long cpu_flags;
	struct list_head temp_list;

	INIT_LIST_HEAD(&temp_list);

	spin_lock_irqsave(&hca->spin_lock, cpu_flags);

	list_for_each_safe(ioq_entry, temp_entry, &hca->resource_list) {

		ioq = list_entry(ioq_entry, ioq_t, hca_resource_list);

		list_del(&ioq->hca_resource_list);
		hca->resource_count--;

		ioq->queue_type = QUEUE_TEMP;
		list_add_tail(&ioq->temp_list, &temp_list);
	}

	spin_unlock_irqrestore(&hca->spin_lock, cpu_flags);

	TS_REPORT_STAGE(MOD_SRPTP, "FMR flush completed for hca %d",
			hca->hca_index + 1);

	/*
	 * De-registered FMR entries have now been flushed out of the hardware
	 * Retry any I/Os that got queued as a result of no FMR entries being
	 * available
	 */
	list_for_each_safe(ioq_entry, temp_entry, &temp_list) {

		ioq = list_entry(ioq_entry, ioq_t, temp_list);

		list_del(&ioq->temp_list);

		TS_REPORT_STAGE(MOD_SRPTP,
				" Retrying ioq %p, from hca resource queue",
				ioq);

		srp_host_internal_qcommand(ioq);
	}
}

/*
 * Path must be specified in the target
 */
void initialize_connection(srp_target_t * target)
{
	ioc_entry_t *ioc_entry;
	srp_host_port_params_t *port;
	srp_host_conn_t *s;
	unsigned long cpu_flags;
	u64 *dgid;

	spin_lock_irqsave(&target->spin_lock, cpu_flags);

	port = target->port;
	ioc_entry = &target->ioc;
	s = target->active_conn;

	if (s == NULL) {

		s = kmalloc(sizeof(srp_host_conn_t), GFP_ATOMIC);
		if (s == NULL) {
			TS_REPORT_FATAL(MOD_SRPTP,
					"Couldnt allocate memory for "
					"connection to 0x%llx\n",
					be64_to_cpu(target->service_name));
			spin_unlock_irqrestore(&target->spin_lock, cpu_flags);
			return;
		}

		TS_REPORT_WARN(MOD_SRPTP,
			       "Initializing connection %p for target %d:%d:%d",
			       s, target->target_index, port->local_port,
			       port->hca->hca_index + 1);

		memset(s, 0x00, sizeof(srp_host_conn_t));

		/* attach the target to the connection */
		s->target = target;
		s->cqs_hndl = target->cqs_hndl;
		s->cqr_hndl = target->cqr_hndl;

		driver_params.num_connections++;
		port->hca->num_connections++;
		target->hard_reject = FALSE;
		target->hard_reject_count = 0;

		target->active_conn = s;
		target->conn_count++;
		list_add_tail(&s->conn_list, &target->conn_list);
	}

	target->timeout = jiffies + TARGET_POTENTIAL_TIMEOUT;
	target->state = TARGET_POTENTIAL_CONNECTION;

	s->port = port;
	s->service_id = ioc_entry->service_id;

	s->state = SRP_HOST_LOGIN_INPROGRESS;
	s->redirected = FALSE;
	s->path_record_tid = 0;

	memcpy(&s->path_record, &ioc_entry->iou_path_record,
	       sizeof(struct ib_path_record));

	srp_host_login(s);

	spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

	port->hca->num_pending_connections++;
	driver_params.num_pending_connections++;

	dgid = (u64 *)s->path_record.dgid;

	if ((*dgid == 0) || (*(dgid + 1) == 0)) {
		memset(s->redirected_port_gid, 0, 16);
		s->path_record_retry_count = 0;
		s->state = SRP_HOST_GET_PATH_RECORD;
		s->path_record_tid = 0;

		if (srp_get_path_record(s, s->port,
					srp_path_record_completion, s))
			remove_connection(s, TARGET_POTENTIAL_CONNECTION);
	} else
		srptp_connect(s,
			      &s->path_record,
			      (__u8 *) s->login_buff, s->login_buff_len);
}

void connection_reset(srp_target_t * target)
{
	unsigned long cpu_flags;

	spin_lock_irqsave(&target->spin_lock, cpu_flags);

	driver_params.num_active_connections--;
	target->port->hca->num_active_connections--;
	target->active_conn->state = SRP_HOST_DISCONNECT_NEEDED;
	target->active_conn = NULL;
	target->timeout = jiffies + TARGET_POTENTIAL_TIMEOUT;
	target->state = TARGET_POTENTIAL_CONNECTION;
	target->need_device_reset = TRUE;

	spin_unlock_irqrestore(&target->spin_lock, cpu_flags);
}

void remove_connection(srp_host_conn_t * s, srp_target_state_t target_state)
{
	srp_target_t *target = s->target;
	unsigned long cpu_flags;
	int force_close = FALSE;

	TS_REPORT_WARN(MOD_SRPTP,
		       "Remove connection %p for target %d:%d:%d",
		       s, target->target_index, s->port->local_port,
		       s->port->hca->hca_index + 1);

	spin_lock_irqsave(&target->spin_lock, cpu_flags);

	if (s->state == SRP_HOST_LOGIN_INPROGRESS) {
		s->port->hca->num_pending_connections--;
		driver_params.num_pending_connections--;
	} else if (s->state == SRP_HOST_GET_PATH_RECORD) {
		force_close = TRUE;
		s->port->hca->num_pending_connections--;
		driver_params.num_pending_connections--;
	} else if (s->state == SRP_UP) {
		driver_params.num_active_connections--;
		s->port->hca->num_active_connections--;
	}

	s->port->hca->num_connections--;

	target->state = target_state;
	if (target->state == TARGET_POTENTIAL_CONNECTION)
		target->timeout = jiffies + TARGET_POTENTIAL_TIMEOUT;
	target->active_conn = NULL;
	target->need_disconnect = FALSE;
	target->hard_reject = FALSE;

	s->state = SRP_HOST_LOGOUT_INPROGRESS;

	spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

	if ((s->path_record_tid) &&
	    (s->path_record_tid != TS_IB_CLIENT_QUERY_TID_INVALID)) {
		ib_client_query_cancel(s->path_record_tid);
	}

	if (force_close) {
		srp_host_close_conn(s);
	} else {
		srptp_disconnect(s);
	}
}

static srp_pkt_t *srp_host_pkt_index_to_ptr(int index, srp_target_t * target)
{
	if (!(index > target->max_num_pkts)) {
		return (&target->srp_pkt_hdr_area[index]);
	} else {
		TS_REPORT_FATAL(MOD_SRPTP, "Out of range index %d, target %d",
				index, target->target_index);
		return (NULL);
	}
}

/*
 * Free an SRP packet.
 * IN: SRP packet
 */
static void srp_host_free_pkt(srp_pkt_t * srp_pkt)
{
	srp_target_t *target;
	unsigned long cpu_flags;

	if (srp_pkt == NULL) {
		TS_REPORT_WARN(MOD_SRPTP, "srp_pkt is NULL");
	} else {
		target = (srp_target_t *) srp_pkt->target;

		spin_lock_irqsave(&target->spin_lock, cpu_flags);

		if (srp_pkt->in_use == FALSE) {

			TS_REPORT_STAGE(MOD_SRPTP, "srp_pkt already free %d",
					srp_pkt->pkt_index);

		} else if (srp_pkt->in_use == TRUE) {

			srp_pkt->scatter_gather_list.address =
			    (u64) (unsigned long)srp_pkt->data;
			srp_pkt->scatter_gather_list.length = srp_cmd_pkt_size;
			srp_pkt->in_use = FALSE;
			srp_pkt->conn = INVALID_CONN_HANDLE;

			srp_pkt->next = target->srp_pkt_free_list;
			target->srp_pkt_free_list = srp_pkt;

			atomic_inc(&target->free_pkt_counter);
		}
		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);
	}
}

static void srp_host_free_pkt_no_lock(srp_pkt_t * srp_pkt)
{
	srp_target_t *target;

	if (srp_pkt == NULL) {
		TS_REPORT_WARN(MOD_SRPTP, "srp_pkt is NULL");
	} else {
		target = (srp_target_t *) srp_pkt->target;

		if (srp_pkt->in_use == FALSE) {

			TS_REPORT_STAGE(MOD_SRPTP, "srp_pkt already free %d",
					srp_pkt->pkt_index);

		} else if (srp_pkt->in_use == TRUE) {

			srp_pkt->scatter_gather_list.address =
			    (u64) (unsigned long)srp_pkt->data;
			srp_pkt->scatter_gather_list.length = srp_cmd_pkt_size;
			srp_pkt->in_use = FALSE;
			srp_pkt->conn = INVALID_CONN_HANDLE;

			srp_pkt->next = target->srp_pkt_free_list;
			target->srp_pkt_free_list = srp_pkt;

			atomic_inc(&target->free_pkt_counter);
		}
	}
}

void srp_host_free_pkt_index(int index, srp_target_t * target)
{
	srp_pkt_t *pkt;

	pkt = srp_host_pkt_index_to_ptr(index, target);

	srp_host_free_pkt(pkt);
}

/*
 * Get a free SRP packet structure from the free list
 */
static srp_pkt_t *srp_host_get_pkt(srp_target_t * target)
{
	unsigned long cpu_flags;
	srp_pkt_t *srp_pkt = NULL;

	spin_lock_irqsave(&target->spin_lock, cpu_flags);
	if (target->srp_pkt_free_list == NULL) {
		TS_REPORT_STAGE(MOD_SRPTP, "Packet pool exhausted on target %d",
				target->target_index);
	} else {

		srp_pkt = target->srp_pkt_free_list;
		target->srp_pkt_free_list = target->srp_pkt_free_list->next;

		atomic_dec(&target->free_pkt_counter);

		srp_pkt->next = NULL;
		srp_pkt->in_use = TRUE;
	}

	spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

	return (srp_pkt);
}

static int srp_host_pre_post(srp_target_t * target, srp_host_conn_t * s)
{
	int i;
	int count;
	srp_pkt_t *recv_pkt = NULL;

	count = atomic_read(&target->free_pkt_counter);

	for (i = 0; i < MAX_PREPOST_RECVS; i++) {

		recv_pkt = srp_host_get_pkt(target);

		if (recv_pkt) {
			recv_pkt->conn = s;
			recv_pkt->scatter_gather_list.key =
			    target->l_key;
			recv_pkt->scatter_gather_list.length = srp_cmd_pkt_size;

			if (atomic_dec_and_test(&s->rwqes_avail)) {
				TS_REPORT_FATAL(MOD_SRPTP,
						"Out of recv_wqes(0)");
				atomic_inc(&s->rwqes_avail);
			} else if (recv_pkt) {
				if (srptp_post_recv(recv_pkt)) {
					TS_REPORT_FATAL(MOD_SRPTP,
							"post_rr(0) failed rwqes_a %d",
							atomic_read(&s->
								    rwqes_avail));
				}
			} else {
				TS_REPORT_FATAL(MOD_SRPTP,
						"Out of recv_wqes(1)");
				atomic_inc(&s->rwqes_avail);
			}
		} else {
			
			TS_REPORT_WARN(MOD_SRPTP, "No buffers to pre-post %d",
				       target->target_index);
			return (-ENOMEM);
			
		}
	}

	return (TS_SUCCESS);
}

u8 *srp_host_get_pages(u32 reqS)
{
	int order, aSize;

	if (reqS <= 0)
		return NULL;

	for (order = 0, aSize = PAGE_SIZE; aSize < reqS; order++, aSize <<= 1) ;

	return ((u8 *) __get_free_pages(GFP_ATOMIC, order));
}

void srp_host_free_pages(u8 * buf, u32 size)
{
	int order, aSize;

	if (!buf)
		return;

	for (order = 0, aSize = PAGE_SIZE; aSize < size; order++, aSize <<= 1) ;

	free_pages((unsigned long)buf, order);
	return;
}

/*
 * Allocate SRP packet structures and packet data
 */
int srp_host_alloc_pkts(srp_target_t * target)
{
	int pkt_num;
	srp_pkt_t *srp_pkt;
	void *srp_pkt_data;
	int status;
	int max_num_pkts;
	srp_host_hca_params_t *hca;
	int cq_entries;
	int requested_cq_entries;
	struct ib_cq_callback cq_callback;

	/* allocate twice as many packets, send and receive */
	max_num_pkts = MAX_SEND_WQES + MAX_RECV_WQES;
	target->max_num_pkts = max_num_pkts;

	target->srp_pkt_hdr_area = (srp_pkt_t *)
	    kmalloc((max_num_pkts * sizeof(srp_pkt_t)), GFP_KERNEL);

	if (target->srp_pkt_hdr_area == NULL) {
		TS_REPORT_FATAL(MOD_SRPTP,
				"kmalloc for packet data structures failed");
		return (-ENOMEM);
	}

	target->srp_pkt_free_list = target->srp_pkt_hdr_area;
	target->srp_pkt_data_area =
	    srp_host_get_pages(max_num_pkts * srp_cmd_pkt_size);

	if (target->srp_pkt_data_area == NULL) {
		TS_REPORT_FATAL(MOD_SRPTP,
				"kmalloc for packet data area failed #pkts %d p_sz %d",
				max_num_pkts, srp_cmd_pkt_size);
		goto alloc_fail;
	}

	hca = target->port->hca;

	cq_callback.context = IB_CQ_CALLBACK_INTERRUPT;
	cq_callback.policy = IB_CQ_PROVIDER_REARM;
	cq_callback.function.entry = cq_send_handler;
	cq_callback.arg = (void *)target;

	cq_entries = MAX_SEND_WQES;
	requested_cq_entries = cq_entries;

	status = ib_cq_create(hca->ca_hndl,
			      &cq_entries,
			      &cq_callback,
			      NULL, &target->cqs_hndl);

	if (status != 0) {
		
		if (cq_entries < requested_cq_entries) {
			TS_REPORT_FATAL(MOD_SRPTP,
					"Send completion queue "
					"creation failed: %d asked "
					"for %d entries",
					cq_entries,
					requested_cq_entries);
			goto cqs_err;
		}
		
	}

	cq_callback.function.entry = cq_recv_handler;
	cq_entries = MAX_RECV_WQES;
	requested_cq_entries = cq_entries;
	status = ib_cq_create(hca->ca_hndl,
			      &cq_entries,
			      &cq_callback,
			      NULL, &target->cqr_hndl);
	if (status != 0) {
		
		if (cq_entries < requested_cq_entries) {
			TS_REPORT_FATAL(MOD_SRPTP,
					"Recv completeion queue "
					"creation failed");
			goto cqr_err;
		}
		
	}

	status = ib_memory_register(hca->pd_hndl,
				    target->srp_pkt_data_area,
				    max_num_pkts * srp_cmd_pkt_size,
				    IB_ACCESS_LOCAL_WRITE |
				    IB_ACCESS_REMOTE_READ,
				    &target->srp_pkt_data_mhndl,
				    &target->l_key, &target->r_key);

	if (status != 0) {
		
		TS_REPORT_FATAL(MOD_SRPTP,
				"Memory registration failed: %d",
				status);
		goto mr_fail;
		  
	}

	srp_pkt = target->srp_pkt_hdr_area;
	srp_pkt_data = target->srp_pkt_data_area;

	for (pkt_num = 0; pkt_num < max_num_pkts; pkt_num++) {

		memset(srp_pkt, 0x00, sizeof(srp_pkt_t));
		srp_pkt->conn = INVALID_CONN_HANDLE;
		srp_pkt->target = target;
		srp_pkt->data = srp_pkt_data;
		srp_pkt->scatter_gather_list.address =
		    (u64) (unsigned long)srp_pkt_data;
		srp_pkt->scatter_gather_list.length = srp_cmd_pkt_size;

		if (pkt_num == max_num_pkts - 1)
			srp_pkt->next = NULL;
		else
			srp_pkt->next = srp_pkt + 1;

		srp_pkt->pkt_index = pkt_num;

		srp_pkt++;
		srp_pkt_data += srp_cmd_pkt_size;

	}

	atomic_set(&target->free_pkt_counter, max_num_pkts);

	return (0);

      mr_fail:
	ib_cq_destroy(target->cqr_hndl);
      cqr_err:
	ib_cq_destroy(target->cqs_hndl);
      cqs_err:
	srp_host_free_pages(target->srp_pkt_data_area,
			    target->max_num_pkts * srp_cmd_pkt_size);

      alloc_fail:
	kfree(target->srp_pkt_hdr_area);
	return (-ENOMEM);
}


void srp_host_disconnect_targets(void)
{
	int tidx;
	int hca_index;
	srp_host_hca_params_t *hca;
	srp_target_t *target;

	for (hca_index = 0; hca_index < MAX_HCAS; hca_index++) {
		hca = &hca_params[hca_index];

		if ((hca->valid == FALSE) || !hca->srp_targets)
			continue;

		for (tidx = 0; tidx < max_srp_targets; tidx++) {
			target = &hca->srp_targets[tidx];

			if ((target->valid == TRUE) &&
			    (target->active_conn)) {
				TS_REPORT_WARN(MOD_SRPTP,
					       "disconnect target %d:%d:%d",
					       tidx, target->port->local_port,
					       hca_index + 1);
				remove_connection(target->active_conn, 
						  TARGET_NO_CONNECTIONS);
			}
		}
	}
}


void srp_host_cleanup_target(srp_target_t *target)
{
	if (target->host) {
		if (target->block_requests == TRUE)
			target->block_requests = FALSE;

		scsi_remove_host(target->host);
		scsi_host_put(target->host);
		target->host = NULL;
	}

	if (target->valid == TRUE) {
		struct ib_cq_entry cq_entry;
		int status;

                while (1) {
			if (target->cqr_hndl == NULL)
				break;
				
                        status = ib_cq_poll(target->cqr_hndl, &cq_entry);
                        if (status == -EAGAIN) {
				
                                break;
                        } else if  ((status != 0) && (status != -EAGAIN)){
                                TS_REPORT_WARN(MOD_SRPTP, "cq poll failed");
                                break;
				
                        } else {
                                cq_recv_handler(target->cqr_hndl, &cq_entry,
						target);
			}
		}
 

		while (1) {
			if (target->cqs_hndl == NULL)
				break;
				
			status = ib_cq_poll(target->cqs_hndl, &cq_entry);
			if (status == -EAGAIN) {
				
				break;
			} else if ((status != 0) && (status != -EAGAIN)) {
				TS_REPORT_WARN(MOD_SRPTP, "cq poll failed");
				break;
				
			} else {
				cq_send_handler(target->cqs_hndl,
						&cq_entry, target);
			}
		}

		if (target->srp_pkt_data_mhndl) {
			ib_memory_deregister(target->srp_pkt_data_mhndl);
			target->srp_pkt_data_mhndl = NULL;
		}

		if (target->cqr_hndl) {
			ib_cq_destroy(target->cqr_hndl);
			target->cqr_hndl = NULL;
		}

		if (target->cqs_hndl) {
			ib_cq_destroy(target->cqs_hndl);
			target->cqs_hndl = NULL;
		}

		if (target->srp_pkt_data_area) {
			srp_host_free_pages(target->srp_pkt_data_area,
					    target->max_num_pkts *
					    srp_cmd_pkt_size);
			kfree(target->srp_pkt_hdr_area);
			target->srp_pkt_data_area = NULL;
		}
	}
}
	

void srp_host_cleanup_targets(void)
{
	int tidx;
	int hca_index;
	srp_host_hca_params_t *hca;
	srp_target_t *target;

	for (hca_index = 0; hca_index < MAX_HCAS; hca_index++) {
		hca = &hca_params[hca_index];

		if ((hca->valid == FALSE) || !hca->srp_targets)
			continue;

		for (tidx = 0; tidx < max_srp_targets; tidx++) {
			target = &hca->srp_targets[tidx];
			srp_host_cleanup_target(target);
		}
		kfree(hca->srp_targets);
	}
}

void initialize_target(srp_target_t * target)
{
	spin_lock_init(&target->spin_lock);
	sema_init(&target->sema, 1);

	target->pending_count =
	    target->active_count = target->resource_count = 0;
	target->state = TARGET_NOT_INITIALIZED;
	target->timeout = jiffies + TARGET_POTENTIAL_TIMEOUT;
	target->valid = FALSE;
	target->need_disconnect = FALSE;
	target->need_device_reset = FALSE;
	target->hard_reject = FALSE;
	target->hard_reject_count = 0;
	target->active_conn = NULL;
	target->rescan_scsi = FALSE;
	target->scan_host = FALSE;
	target->service_name = 0;
	target->block_requests = FALSE;
	target->no_connection_io = 0;
	target->host = NULL;

	INIT_LIST_HEAD(&target->conn_list);
	INIT_LIST_HEAD(&target->active_list);
	INIT_LIST_HEAD(&target->pending_list);
	INIT_LIST_HEAD(&target->resource_list);
}

void srp_pending_to_scsi(srp_target_t * target, unsigned long status)
{
	struct list_head *ioq_entry, *temp_entry;
	unsigned long cpu_flags;
	ioq_t *ioq;
	struct list_head temp_list;

	TS_REPORT_STAGE(MOD_SRPTP,
			"Flushing the pending queue back to scsi for target %d",
			target->target_index);

	INIT_LIST_HEAD(&temp_list);

	spin_lock_irqsave(&target->spin_lock, cpu_flags);

	list_for_each_safe(ioq_entry, temp_entry, &target->pending_list) {

		ioq = list_entry(ioq_entry, ioq_t, pending_list);

		target->pending_count--;
		list_del(&ioq->pending_list);

		ioq->req_temp = ioq->req;
		ioq->req = NULL;

		if (ioq->type == IOQ_COMMAND) {
			ioq->req_temp->result = status << 16;

			TS_REPORT_STAGE(MOD_SRPTP,
					" Sending IO back to scsi from "
					"pending list ioq %p", ioq);

			ioq->queue_type = QUEUE_TEMP;
			list_add_tail(&ioq->temp_list, &temp_list);
		}
	}

	spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

	list_for_each_safe(ioq_entry, temp_entry, &temp_list) {
		ioq = list_entry(ioq_entry, ioq_t, temp_list);
		list_del(&ioq->temp_list);
		srp_host_scsi_complete(ioq->req_temp);
		kfree(ioq);
	}
}

void srp_lun_pending_to_scsi(srp_target_t * target, unsigned long status,
			     struct scsi_cmnd *scsi_cmd)
{
	struct list_head *ioq_entry, *temp_entry;
	unsigned long cpu_flags;
	ioq_t *ioq;
	struct list_head temp_list;
	struct scsi_cmnd *scsi_tmp;

	INIT_LIST_HEAD(&temp_list);

	spin_lock_irqsave(&target->spin_lock, cpu_flags);

	TS_REPORT_WARN(MOD_SRPTP,
			"Flush pending to scsi for LUN %d on target %d:%d:%d"
			" p_cnt %d",
			scsi_cmd->device->lun, target->target_index,
			target->port->local_port, target->port->hca->hca_index + 1,
			target->pending_count);
			

	list_for_each_safe(ioq_entry, temp_entry, &target->pending_list) {

		ioq = list_entry(ioq_entry, ioq_t, pending_list);
		scsi_tmp = ioq->req;
		if (scsi_tmp->device->lun == scsi_cmd->device->lun) {
			target->pending_count--;
			list_del(&ioq->pending_list);
			ioq->req_temp = ioq->req;
			ioq->req = NULL;
			if (ioq->type == IOQ_COMMAND) {
				ioq->req_temp->result = status << 16;

				TS_REPORT_WARN(MOD_SRPTP,
						"Sending IO back to scsi from "
						" pending list ioq %p", ioq);

				ioq->queue_type = QUEUE_TEMP;
				list_add_tail(&ioq->temp_list, &temp_list);
			}
		}
	}

	spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

	list_for_each_safe(ioq_entry, temp_entry, &temp_list) {
		ioq = list_entry(ioq_entry, ioq_t, temp_list);

		list_del(&ioq->temp_list);

		srp_host_scsi_complete(ioq->req_temp);

		kfree(ioq);
	}
}

void srp_move_to_pending(srp_target_t * target)
{
	srp_host_hca_params_t *hca = target->port->hca;
	ioq_t *ioq;
	struct list_head *ioq_entry, *temp_entry;
	unsigned long cpu_flags;

	TS_REPORT_WARN(MOD_SRPTP,
			"Moving all I/Os to pending for target %d:%d:%d"
			" a_cnt %d r_cnt %d",
			target->target_index, target->port->local_port,
			hca->hca_index + 1, target->active_count,
			target->resource_count);

	spin_lock_irqsave(&target->spin_lock, cpu_flags);

	/*
	 * First, walk the active list and move them to the pending list
	 */
	list_for_each_safe(ioq_entry, temp_entry, &target->active_list) {

		ioq = list_entry(ioq_entry, ioq_t, active_list);

		TS_REPORT_WARN(MOD_SRPTP,
				" Moving ioq %p, SCpnt %p from active to "
				"pending", ioq, ioq->req);

		target->active_count--;
		list_del(&ioq->active_list);

		if (ioq->timer.function) {
			del_timer(&ioq->timer);
			ioq->timer.function = NULL;
		}

		if (ioq->req->eh_timeout.function) {
			scsi_add_timer(ioq->req, IOQ_CMD_EXTEND_TIMEOUT,
				       ioq->req->eh_timeout.function);
			TS_REPORT_WARN(MOD_SRPTP,
				       "extended timeout for SCpnt %p ioq %p",
				       ioq->req, ioq);
		}

		if (ioq->req->request_bufflen)
			srp_free_srlist(hca, ioq, ioq->req);

		/*
		 * We are going to be given a new conn, clear the packet pointer
		 */
		ioq->pkt = NULL;
		ioq->recv_pkt = NULL;

		if (ioq->type == IOQ_ABORT) {
			TS_REPORT_STAGE(MOD_SRPTP,
					" Dumping pending flush ioq %p from "
					"pending", ioq);
			kfree(ioq);
		} else {
			ioq->queue_type = QUEUE_PENDING;
			target->pending_count++;
			list_add_tail(&ioq->pending_list,
				      &target->pending_list);
		}
	}

	/*
	 * Walk the resource list for the target and move it to the
	 * pending list
	 */
	list_for_each_safe(ioq_entry, temp_entry, &target->resource_list) {

		ioq = list_entry(ioq_entry, ioq_t, target_resource_list);
		target->resource_count--;
		list_del(&ioq->target_resource_list);

		TS_REPORT_STAGE(MOD_SRPTP,
				" Moving ioq %p from target resource to "
				"pending", ioq);
		if (ioq->req->eh_timeout.function) {
			scsi_add_timer(ioq->req, IOQ_CMD_EXTEND_TIMEOUT,
				       ioq->req->eh_timeout.function);
			TS_REPORT_WARN(MOD_SRPTP,
				       "extended timeout for SCpnt %p ioq %p",
				       ioq->req, ioq);
		}


		ioq->queue_type = QUEUE_PENDING;
		target->pending_count++;
		list_add_tail(&ioq->pending_list, &target->pending_list);
	}

	spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

	spin_lock_irqsave(&hca->spin_lock, cpu_flags);

	/*
	 * Walk the resource list for the HCA this target is using
	 */
	list_for_each_safe(ioq_entry, temp_entry, &hca->resource_list) {

		ioq = list_entry(ioq_entry, ioq_t, hca_resource_list);

		if (ioq->target == target) {
			hca->resource_count--;
			list_del(&ioq->hca_resource_list);

			TS_REPORT_STAGE(MOD_SRPTP,
					" Moving ioq %p from hca resource to "
					"pending", ioq);

			spin_lock(&target->spin_lock);

			if (ioq->req->eh_timeout.function) {
				scsi_add_timer(ioq->req, IOQ_CMD_EXTEND_TIMEOUT,
					       ioq->req->eh_timeout.function);
				TS_REPORT_WARN(MOD_SRPTP,
					       "extended timeout for SCpnt %p ioq %p",
					       ioq->req, ioq);
			}

			/*
			 * reclaim send and receive frames
			 */
			srp_host_free_pkt_no_lock(ioq->pkt);
			ioq->pkt = NULL;
			srp_host_free_pkt_no_lock(ioq->recv_pkt);
			ioq->recv_pkt = NULL;

			ioq->queue_type = QUEUE_PENDING;
			target->pending_count++;
			list_add_tail(&ioq->pending_list,
				      &target->pending_list);

			spin_unlock(&target->spin_lock);
		}
	}

	spin_unlock_irqrestore(&hca->spin_lock, cpu_flags);
}

void srp_move_lun_resource_to_pending(srp_target_t * target,
				      struct scsi_cmnd *scsi_cmd)
{
	srp_host_hca_params_t *hca = target->port->hca;
	ioq_t *ioq;
	struct list_head *ioq_entry, *temp_entry;
	struct scsi_cmnd *scsi_tmp;
	unsigned long cpu_flags;

	TS_REPORT_WARN(MOD_SRPTP,
			"Moving all I/Os in resources list to pending list"
			" for LUN %d on target %d:%d:%d resource_count %d",
			scsi_cmd->device->lun, target->target_index,
			target->port->local_port, target->port->hca->hca_index + 1,
			target->resource_count);

	spin_lock_irqsave(&target->spin_lock, cpu_flags);

	/*
	 * Walk the resource list for the target and move it to the
	 * pending list
	 */
	list_for_each_safe(ioq_entry, temp_entry, &target->resource_list) {

		ioq = list_entry(ioq_entry, ioq_t, target_resource_list);
		scsi_tmp = ioq->req;
		if (scsi_tmp->device->lun == scsi_cmd->device->lun) {
			target->resource_count--;
			list_del(&ioq->target_resource_list);

			TS_REPORT_WARN(MOD_SRPTP,
					"Moving ioq %p from target resource to "
					" pending", ioq);

			if (ioq->req->eh_timeout.function) {
				scsi_add_timer(ioq->req, IOQ_CMD_EXTEND_TIMEOUT,
					       ioq->req->eh_timeout.function);
				TS_REPORT_WARN(MOD_SRPTP,
					       "extended timeout for SCpnt %p ioq %p",
					       ioq->req, ioq);
			}

			ioq->queue_type = QUEUE_PENDING;
			target->pending_count++;
			list_add_tail(&ioq->pending_list,
				      &target->pending_list);
		}
	}

	spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

	spin_lock_irqsave(&hca->spin_lock, cpu_flags);

	/*
	 * Walk the resource list for the HCA this target is using
	 */
	list_for_each_safe(ioq_entry, temp_entry, &hca->resource_list) {
		ioq = list_entry(ioq_entry, ioq_t, hca_resource_list);
		scsi_tmp = ioq->req;
		if ((ioq->target == target) &&
		    (scsi_tmp->device->lun == scsi_cmd->device->lun)) {
			hca->resource_count--;
			list_del(&ioq->hca_resource_list);

			TS_REPORT_WARN(MOD_SRPTP,
					"Moving ioq %p from hca resource to "
					" pending", ioq);

			spin_lock(&target->spin_lock);

			if (ioq->req->eh_timeout.function) {
				scsi_add_timer(ioq->req, IOQ_CMD_EXTEND_TIMEOUT,
					       ioq->req->eh_timeout.function);
				TS_REPORT_WARN(MOD_SRPTP,
					       "extended timeout for SCpnt %p ioq %p",
					       ioq->req, ioq);
			}

			/*
			 * reclaim send and receive frames
			 */
			srp_host_free_pkt_no_lock(ioq->pkt);
			ioq->pkt = NULL;
			srp_host_free_pkt_no_lock(ioq->recv_pkt);
			ioq->recv_pkt = NULL;

			ioq->queue_type = QUEUE_PENDING;
			target->pending_count++;
			list_add_tail(&ioq->pending_list,
				      &target->pending_list);

			spin_unlock(&target->spin_lock);
		}
	}

	spin_unlock_irqrestore(&hca->spin_lock, cpu_flags);
}

void srp_move_lun_active_to_pending(srp_target_t * target,
				    struct scsi_cmnd *scsi_cmd)
{
	ioq_t *ioq;
	struct list_head *ioq_entry, *temp_entry;
	unsigned long cpu_flags;
	struct scsi_cmnd *scsi_tmp = NULL;

	TS_REPORT_WARN(MOD_SRPTP,
			"Moving all I/Os from active list to pending list"
			" for LUN %d on target %d:%d:%d a_cnt %d p_cnt %d",
			scsi_cmd->device->lun, target->target_index,
			target->port->local_port, target->port->hca->hca_index + 1,
			target->active_count, target->pending_count);

	spin_lock_irqsave(&target->spin_lock, cpu_flags);

	/*
	 * walk the active list and move them to the pending list
	 */
	list_for_each_safe(ioq_entry, temp_entry, &target->active_list) {
		ioq = list_entry(ioq_entry, ioq_t, active_list);
		scsi_tmp = ioq->req;

		if (scsi_tmp && (scsi_tmp->device->lun == scsi_cmd->device->lun)) {
			TS_REPORT_WARN(MOD_SRPTP,
					"Moving ioq %p, SCpnt %p from active to "
					" pending a_cnt %d p_cnt %d",
					ioq, ioq->req, target->active_count,
					target->pending_count);

			target->active_count--;
			list_del(&ioq->active_list);

			if (ioq->req->eh_timeout.function) {
				scsi_add_timer(ioq->req, IOQ_CMD_EXTEND_TIMEOUT,
					       ioq->req->eh_timeout.function);
				TS_REPORT_WARN(MOD_SRPTP,
					       "extended timeout for SCpnt %p ioq %p",
					       ioq->req, ioq);
			}

			if (ioq->req->request_bufflen)
				srp_free_srlist(target->port->hca, ioq, ioq->req);


			/*
			 * We are going to be given a new conn, clear the packet pointer
			 */
			ioq->pkt = NULL;
			ioq->recv_pkt = NULL;

			if (ioq->type == IOQ_ABORT) {
				TS_REPORT_WARN(MOD_SRPTP,
						"Active to pending ioq %p IOQ_ABORT"
						, ioq);
				kfree(ioq);
			} else {
				ioq->queue_type = QUEUE_PENDING;
				target->pending_count++;
				list_add_tail(&ioq->pending_list,
					      &target->pending_list);
			}
		}

	}
	spin_unlock_irqrestore(&target->spin_lock, cpu_flags);
}


void srp_flush_pending_to_active(srp_target_t * target)
{
	ioq_t *ioq;
	struct list_head *ioq_entry, *temp_entry;
	unsigned long cpu_flags;
	struct scsi_cmnd *SCpnt;
	struct list_head temp_list;

	INIT_LIST_HEAD(&temp_list);

	spin_lock_irqsave(&target->spin_lock, cpu_flags);

	list_for_each_safe(ioq_entry, temp_entry, &target->pending_list) {

		ioq = list_entry(ioq_entry, ioq_t, pending_list);

		target->pending_count--;
		list_del(&ioq->pending_list);

		ioq->queue_type = QUEUE_TEMP;

		list_add_tail(&ioq->temp_list, &temp_list);
	}

	spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

	TS_REPORT_STAGE(MOD_SRPTP, "Flush pending to active for target %d",
			target->target_index);

	list_for_each_safe(ioq_entry, temp_entry, &temp_list) {

		ioq = list_entry(ioq_entry, ioq_t, temp_list);
		list_del(&ioq->temp_list);

		SCpnt = ioq->req;

		TS_REPORT_STAGE(MOD_SRPTP, " Retrying SCpnt %p ioq %p", SCpnt,
				ioq);

		if (ioq->type == IOQ_COMMAND) {

			srp_host_internal_qcommand(ioq);

		} else {
			/*
			 * we should never have any other I/O type on the
			 * pending queue
			 */
			TS_REPORT_FATAL(MOD_SRPTP,
					"Pending queue error ioq %p type 0x%d "
					"found", ioq, ioq->type);
		}
	}
}

void sweep_targets(void)
{
	int i;
	srp_target_t *target;
	srp_host_hca_params_t *hca;

	/*
	 * Sweep all targets for connections that are down, and try to
	 * find a connection
	 */
	for (hca = &hca_params[0]; hca < &hca_params[MAX_HCAS]; hca++) {
		if ((hca->valid == FALSE) || !hca->srp_targets)
			break;

		for (i = 0; i < max_srp_targets; ++i) {
			target = &hca->srp_targets[i];

			if (target->valid == FALSE)
				continue;

			down(&target->sema);

			if (((target->state > TARGET_INITIALIZED) &&
			     (target->state < TARGET_ACTIVE_CONNECTION)) &&
			    (target->active_conn == NULL) &&
			    (target->ioc.path_valid == TRUE) &&
			    (target->port->port_state == IB_PORT_STATE_ACTIVE)) {

				TS_REPORT_FATAL(MOD_SRPTP,
						"Connecting target %d:%d:%d",
						i, target->port->local_port,
					        hca->hca_index + 1);
	
				initialize_connection(target);
			}
			up(&target->sema);
		}
	}
}


void srp_host_rediscover(void)
{
	int hca_index, port_index;
	srp_host_port_params_t *port;

	for (hca_index = 0; hca_index < MAX_HCAS; hca_index++) {

		for (port_index = 0; port_index < MAX_LOCAL_PORTS_PER_HCA;
		     port_index++) {

			port = &hca_params[hca_index].port[port_index];

			if (port->valid == FALSE)
				continue;

			port->dm_need_query = TRUE;
		}
	}
}

void srp_host_check_timeouts(srp_host_hca_params_t *hca)
{
	srp_target_t *target;
	struct list_head temp_list, *ioq_entry, *temp_entry;
	ioq_t *ioq;
	int retry_count;
	unsigned long cpu_flags;

	for (target = &hca->srp_targets[0];
	     target < &hca->srp_targets[max_srp_targets]; target++) {

		INIT_LIST_HEAD(&temp_list);

		if (target->valid == FALSE)
			continue;

		down(&target->sema);

		/*
		 * TODO:
		 * Cleanup various disconnect/reconnect methods into
		 * one method
		 */
		if (target->need_disconnect == TRUE) {
			srp_move_to_pending(target);
			srp_pending_to_scsi(target, DID_SOFT_ERROR);

			remove_connection(target->active_conn,
					  TARGET_POTENTIAL_CONNECTION);

			initialize_connection(target);

			target->need_device_reset = FALSE;
			target->hard_reject = FALSE;
		}

		if (target->need_device_reset == TRUE) {
			struct list_head *conn_entry;
			srp_host_conn_t *conn;

			target->need_device_reset = FALSE;
			target->need_disconnect = FALSE;
			target->hard_reject = FALSE;

			list_for_each(conn_entry, &target->conn_list) {
				conn = list_entry(conn_entry,
						  srp_host_conn_t,
						  conn_list);

				if (conn->state == SRP_HOST_DISCONNECT_NEEDED) {
					srp_move_to_pending(target);
					srp_pending_to_scsi(target,
							    DID_SOFT_ERROR);
					remove_connection(conn,
							  TARGET_POTENTIAL_CONNECTION);
					initialize_connection(target);
				}
			}
		}

		if ((target->hard_reject == TRUE) &&
		    target->active_conn) {
	
			target->need_device_reset = FALSE;
			target->need_disconnect = FALSE;
			target->hard_reject = FALSE;

			if (target->hard_reject_count++ <
			    MAX_HARD_REJECT_COUNT) {

				TS_REPORT_WARN(MOD_SRPTP,
					       "Retrying reject on "
					       "Target %d retry %d",
					       target->target_index,
					       target->hard_reject_count);
					       
				hca->num_pending_connections--;
				driver_params.num_pending_connections--;

				initialize_connection(target);

			} else {
				srp_host_conn_t *conn = target->active_conn;

				/*
				 * Exceeded the retries to make this
				 * connection to the redirected path.
				 * Try a new path/IOC, HARD rejects are
				 * always considered a target specific
				 * failure, so set the flag to indicate
				 * target failure only.
				 */
				remove_connection(conn,
						  TARGET_POTENTIAL_CONNECTION);

				srp_host_close_conn(conn);
			}
		}

		if (((target->state == TARGET_POTENTIAL_CONNECTION) || 
		     (target->state == TARGET_NO_PATHS)) &&
		     (time_after(jiffies, target->timeout))) {

			TS_REPORT_STAGE(MOD_SRPTP,
					"Target %d, no connection timeout",
					target->target_index);

			target->state = TARGET_NO_CONNECTIONS;

			srp_pending_to_scsi(target, DID_IMM_RETRY);
		}

		up(&target->sema);

		spin_lock_irqsave(&target->spin_lock, cpu_flags);

		retry_count = target->resource_count;

		list_for_each_safe(ioq_entry, temp_entry,
				   &target->resource_list) {

			ioq =
			    list_entry(ioq_entry, ioq_t,
				       target_resource_list);

			if (time_after(jiffies, ioq->timeout) ||
			    !target->active_conn ||
			    (target->state != TARGET_ACTIVE_CONNECTION)) {

				target->resource_count--;
				list_del(&ioq->target_resource_list);

				ioq->req->result = DID_SOFT_ERROR << 16;
				ioq->req_temp = ioq->req;
				ioq->req = NULL;

				if (ioq->req_temp->eh_timeout.function) {
					scsi_add_timer(ioq->req_temp, IOQ_CMD_EXTEND_TIMEOUT,
						       ioq->req_temp->eh_timeout.function);
				}
				TS_REPORT_FATAL(MOD_SRPTP,
						"ioq %p scsi_cmd %p timeout on" 
						" target %d:%d:%d resource list",
						ioq, ioq->req_temp,
						target->target_index,
						target->port->local_port,
						target->port->hca->hca_index + 1);

				ioq->queue_type = QUEUE_TEMP;
				list_add_tail(&ioq->temp_list,
					      &temp_list);
			}
	
			if (--retry_count == 0)
				break;
		}

		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

		/*
		 * Now actually, return the I/Os as the result of timeout
		 */
		list_for_each_safe(ioq_entry, temp_entry, &temp_list) {
	
			ioq = list_entry(ioq_entry, ioq_t, temp_list);
			list_del(&ioq->temp_list);
			srp_host_scsi_complete(ioq->req_temp);
			kfree(ioq);
		}

		if (!target->host && 
		    target->active_conn && target->rescan_scsi &&
		    (target->state == TARGET_ACTIVE_CONNECTION) ) {
			TS_REPORT_WARN(MOD_SRPTP,
				       "Add new scsi_host for tgt [%d:%d:%d]",
				       target->target_index,
				       target->port->local_port,
				       hca->hca_index + 1);
			srp_add_target_host(target);
		}

	}		/* end loop on targets */

	INIT_LIST_HEAD(&temp_list);

	/*
	 * Walk the resource list for the HCA, for timeouts...
	 * If we could not get an FMR, then return the I/O back
	 * to SCSI
	 */
	spin_lock_irqsave(&hca->spin_lock, cpu_flags);
	retry_count = hca->resource_count;

	list_for_each_safe(ioq_entry, temp_entry,
			   &hca->resource_list) {

		ioq = list_entry(ioq_entry, ioq_t, hca_resource_list);

		if (time_after(jiffies, ioq->timeout)) {

			hca->resource_count--;
			list_del(&ioq->hca_resource_list);

			ioq->req->result = DID_SOFT_ERROR << 16;
			ioq->req_temp = ioq->req;
			ioq->req = NULL;

			if (ioq->req_temp->eh_timeout.function) {
				scsi_add_timer(ioq->req_temp,
					       IOQ_CMD_EXTEND_TIMEOUT,
					       ioq->req_temp->eh_timeout.function);
			}

			TS_REPORT_FATAL(MOD_SRPTP,
					"HCA resource timeout "
					"on target %d",
					ioq->target->target_index);

			/*
			 * Return the send and receive buffers
			 */
			spin_lock(&ioq->target->spin_lock);
			srp_host_free_pkt_no_lock(ioq->pkt);
			srp_host_free_pkt_no_lock(ioq->recv_pkt);
						  
			atomic_inc(&ioq->target->active_conn->request_limit);
				   
			atomic_inc(&ioq->target->active_conn->local_request_limit);
						   
			ioq->pkt = NULL;
			ioq->recv_pkt = NULL;
			spin_unlock(&ioq->target->spin_lock);

			ioq->queue_type = QUEUE_TEMP;
			list_add_tail(&ioq->temp_list,
				      &temp_list);
		}

		if (--retry_count == 0)
			break;
	}

	spin_unlock_irqrestore(&hca->spin_lock, cpu_flags);

	/*
	 * Now actually, return the I/Os as the result of timeout
	 */
	list_for_each_safe(ioq_entry, temp_entry, &temp_list) {
		ioq = list_entry(ioq_entry, ioq_t, temp_list);
		list_del(&ioq->temp_list);
		srp_host_scsi_complete(ioq->req_temp);
		kfree(ioq);
	}

}

/*
 * The thread periodically polls the DM client for a list of SRP
 * services.
 * If it finds new services, it initiates connections to them.
 * It does not check for services which have disappeared.
 */
void srp_dm_poll_thread(void *arg)
{
	int dm_query_sum = 0;
	srp_host_hca_params_t *hca;
	srp_host_port_params_t *port;
	unsigned long dm_timeout = 0;
	int sweep_target = FALSE;


	strncpy(current->comm, "ts_srp_dm", sizeof(current->comm));
	current->comm[sizeof(current->comm) - 1] = 0;

	ib_host_io_query_init();

	while (1) {
		/*
		 * Driver is exiting, need to kill this thread
		 */
		if (driver_params.dm_shutdown) {

			for (hca = &hca_params[0]; hca < &hca_params[MAX_HCAS];
			     hca++) {
				if (hca->valid == FALSE)
					break;

				for (port = &hca->port[0];
				     port < &hca->port[MAX_LOCAL_PORTS_PER_HCA];
				     port++) {

					if (port->valid == FALSE)
						continue;

					srp_port_query_cancel(port);

					if (port->out_of_service_xid != 0) {
						port->outof_svc_count = 0;
						srp_register_out_of_service
						    (port, FALSE);
					}

					if (port->in_service_xid != 0) {
						port->in_svc_count = 0;
						srp_register_in_service(port,
									FALSE);
					}

					if (port->dm_query_in_progress == TRUE) {
						TS_REPORT_STAGE(MOD_SRP,
								"Canceling DM "
								"Query on hca "
								"%d port %d",
								hca->hca_index +
								1,
								port->
								local_port);
						ib_host_io_query_cancel(port->
									dm_xid);
					}
				}
			}
			break;
		}

		/*
		 * Fill up the srp_target entries using the DM Client.
		 * Start the connection state machine for new services.
		 */
		down(&driver_params.sema);
		dm_query_sum = 0;
		for (hca = &hca_params[0]; hca < &hca_params[MAX_HCAS]; hca++) {
			if (hca->valid == FALSE)
				break;

			for (port = &(hca->port[0]);
			     port < &(hca->port[MAX_LOCAL_PORTS_PER_HCA]);
			     port++) {
				if (port->valid == FALSE)
					continue;

				if (port->dm_query_in_progress == TRUE) {
					dm_query_sum++;
					continue;
				}

				if (port->dm_need_query == FALSE)
					continue;

				if (port->out_of_service_xid == 0) {
					port->outof_svc_count = 0;
					srp_register_out_of_service(port, TRUE);
				}

				if (port->in_service_xid == 0) {
					port->in_svc_count = 0;
					srp_register_in_service(port, TRUE);
				}

				if (srp_dm_query(port) == 0)
					dm_query_sum++;

				if (port->new_tgt_insvc) {
					dm_query_sum++;
					port->new_tgt_insvc--;
				}
			}
		}
		up(&driver_params.sema);

		if (dm_query_sum || driver_params.port_query || 
		    (!list_empty(&driver_params.query_list))) {

			sweep_target = TRUE;
			driver_params.port_query = FALSE;

			if (dm_timeout  == 0)
				dm_timeout = jiffies + DM_FILTER_TIMEOUT;

			if (time_after(jiffies, dm_timeout)) {
				dm_timeout = 0;
				sweep_targets();
			}
			driver_params.dm_active = TRUE;

		} else if (sweep_target == TRUE) {
			sweep_targets();
			sweep_target = FALSE;
			driver_params.dm_active = FALSE;
			dm_timeout = 0;
			dm_query_sum = 0;
		}

		/*
		 * Check timeouts
		 */
		for (hca = &hca_params[0]; hca < &hca_params[MAX_HCAS]; hca++) {
			if ((hca->valid == FALSE) || !hca->srp_targets)
				continue;

			srp_host_check_timeouts(hca);

			if (start_scsi_scan_host) {
				srp_target_t *target;

				for (target = &hca->srp_targets[0];
				     target < &hca->srp_targets[max_srp_targets];
				     target++) {

					if (target->valid == FALSE)
						continue;
			
					if (target->host && !target->scan_host) {
						target->scan_host = TRUE;
						scsi_scan_host(target->host);
					}
				}
			}
		}

		/*
		 * Wait for a signal to wake up, or the timeout
		 *
		 * BUG 28605 not supported any more in K2.6.10
		 *interruptible_sleep_on_timeout(&dm_poll_thread_wait_queue,
		 SRP_DM_THREAD_TIMEOUT);
		 */
		wait_event_interruptible_timeout(dm_poll_thread_wait_queue,
						 0, SRP_DM_THREAD_TIMEOUT);
	}

	ib_host_io_query_cleanup();

	return;

}

int srp_host_init(void)
{
	int err, srtp_initted = 0, srp_dm_initted = 0;
	unsigned long connections_timeout = 0;

	TS_REPORT_WARN(MOD_SRPTP,
		       "SRP Host alternately using direct/indirect addressing");

	if ((err = srptp_init_module()) < 0)
		goto fail;

	srtp_initted = 1;

	spin_lock_init(&driver_params.spin_lock);
	INIT_LIST_HEAD(&driver_params.query_list);

	if ((err = srp_dm_init()) < 0)
		goto fail;

	srp_dm_initted = 1;

	/*
	 * Initialize target structure from target_bindings string
	 */
	if (get_target_bindings())
		TS_REPORT_FATAL(MOD_SRPTP,
				"Problems parsing target_bindings string");

	err = tsKernelThreadStart("ts_srp_dm",
				  srp_dm_poll_thread,
				  NULL, &driver_params.thread);

	if (err < 0) {
		TS_REPORT_FATAL(MOD_SRPTP, "Could not start thread");
		goto fail;
	}

	/* let thread kick in */
	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(2*HZ);

	wake_up_interruptible(&dm_poll_thread_wait_queue);

	/*
	 * Following loop, is to allow time for the IB port to come up,
	 * complete DM queries and complete their connections.
	 */
	do {
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(5 * HZ);
		connections_timeout += 5 * HZ;
	
		TS_REPORT_WARN(MOD_SRPTP,
			       "connections_timeout %ld, dm_active %d,"
			       "  active_conn %d, pending_conn %d",
			       connections_timeout / HZ,
			       driver_params.dm_active,
			       driver_params.num_active_connections,
			       driver_params.num_pending_connections);
	} while (((driver_params.dm_active == TRUE) ||
		  (driver_params.num_active_connections == 0) ||
		  (driver_params.num_pending_connections != 0)) &&
		 (connections_timeout < srp_discovery_timeout));

	start_scsi_scan_host = TRUE;

	/*
	 * Figure out why we exited the loop
	 */
	TS_REPORT_WARN(MOD_SRPTP,
		       "connections_timeout %ld, dm_active %d, active_conn %d,"
		       " pending_conn %d",
		       connections_timeout / HZ,
		       driver_params.dm_active,
		       driver_params.num_active_connections,
		       driver_params.num_pending_connections);

	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(20 * HZ);

	print_target_bindings();

	return 0;
	
      fail:
	if (srp_dm_initted)
		srp_dm_unload();

	srp_host_cleanup_targets();

	if (srtp_initted)
		srptp_cleanup_module();

	return err;
	
}

int srp_proc_info(struct Scsi_Host *host, char *buffer, char **start,
		  off_t offset, int length, int inout)
{
	int len = 0;
	srp_target_t *target;
	int not_first_entry = FALSE;
	u8 *gid;
	srp_host_port_params_t *port;
	srp_host_hca_params_t *hca;

	target = host_to_target(host);
	port = target->port;
	hca = port->hca;

	if (inout == TRUE) {
		
		/* write to proc interface, redistribute connections */
		if (!buffer || length >= PAGE_SIZE) {
			return (-EINVAL);
		}

		if (strncmp("force_rediscover", buffer, length - 1) == 0) {
			srp_host_rediscover();
		}

		return (length);
		     
	}

	len = sprintf(&buffer[len], "\n\t\t\tMellanox SRP Driver\n\n");

	if (target->active_conn) {
		len += sprintf(&buffer[len],
			       "Physical Path on this scsi_host\n");
		len += sprintf(&buffer[len], "Path");
		len += sprintf(&buffer[len], "\t\tService Name");
		len +=
		    sprintf(&buffer[len],
			    "\t\tService ID");
		len +=
		    sprintf(&buffer[len], "\tInitiator_port_GUID\n");

		gid = target->port->local_gid;

		len += sprintf(&buffer[len], "[%d:%d:%d]",
			       target->target_index, port->local_port,
			       hca->hca_index + 1);

		len += sprintf(&buffer[len], "\tSRP.T10:%lx",
			       (unsigned long)cpu_to_be64(target->
							  service_name));
		len += sprintf(&buffer[len], "\t0x%lx",
			       (unsigned long)cpu_to_be64(target->ioc.service_id));

		len += sprintf(&buffer[len],
			       "\t%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x",
			       gid[8], gid[9], gid[10], gid[11],
			       gid[12], gid[13], gid[14], gid[15]);

		len +=
		    sprintf(&buffer[len], "\t\t%d %d %d\n",
			    atomic_read(&target->active_conn->local_request_limit) - 1,
			    atomic_read(&target->active_conn->request_limit) - 1,
			    atomic_read(&target->free_pkt_counter) - 1);
	} else
		len += sprintf(&buffer[len],
			       "No physical path on this scsi_host\n");

	/*
	 * Connection counters
	 */
	len += sprintf(&buffer[len],
		       "Summary of connections on HCA [%d]\n",
		       hca->hca_index + 1);
	len +=
	    sprintf(&buffer[len], "Pending Connections: %d\n",
		    hca->num_pending_connections);
	len +=
	    sprintf(&buffer[len], "Active Connections: %d\n",
		    hca->num_active_connections);
	len +=
	    sprintf(&buffer[len], "Total Connections: %d\n\n",
		    hca->num_connections);

	/*
	 * Print out target bindings string
	 */
	len += sprintf(&buffer[len], "\ttarget_bindings=");
	for (target = &hca->srp_targets[0];
	     target < &hca->srp_targets[max_srp_targets]; target++) {

		if ((target->valid == TRUE) && target->active_conn) {

			/* don't print colon on first guy */
			if (not_first_entry == TRUE) {
				len += sprintf(&buffer[len], ":");
			} else {
				not_first_entry = TRUE;
			}

			len += sprintf(&buffer[len], "%lx.%lx.%x.%d.%d.%d",
				       (unsigned long)cpu_to_be64(target->
								  service_name),
				       (unsigned long)cpu_to_be64(target->ioc.
								  service_id),
				       target->ioc.iou_path_record.dlid,
				       target->target_index,
				       target->port->local_port,
				       hca->hca_index + 1);
		}
	}
	len += sprintf(&buffer[len], "\n\n");

	return (len);
}


void srp_slave_configure(struct scsi_device *device)
{
	int default_depth = max_cmds_per_lun;

	scsi_adjust_queue_depth(device, 0, default_depth);
}


void srp_add_target_host(srp_target_t *target)
{
	struct Scsi_Host *host;
	int status;
	srp_host_hca_params_t *hca = target->port->hca;
	unsigned long cpu_flags;

	host = scsi_host_alloc(&driver_template,
			       sizeof(unsigned long));

	if (host == NULL) {
		TS_REPORT_FATAL(MOD_SRPTP,
				"Can not allocate scsi host"
				" for %d:%d:%d",
				target->target_index,
				target->port->local_port,
				hca->hca_index + 1);
		return;
	}

	*(unsigned long *) host->hostdata = (unsigned long) target;
			
	host->max_channel = 0;
	host->max_id = 1;
	host->max_lun = MAX_SRP_LUN;

	status = scsi_add_host(host, &hca->hca_pdev->dev);
	if (status) {
		TS_REPORT_FATAL(MOD_SRPTP,
				"Probe scsi_host for target %d:%d:%d failed",
				target->target_index,
				target->port->local_port,
				hca->hca_index + 1);
		scsi_host_put(host);
		return;
	}

	spin_lock_irqsave(&target->spin_lock, cpu_flags);
	target->host = host;
	target->rescan_scsi = FALSE;
	spin_unlock_irqrestore(&target->spin_lock, cpu_flags);
}


void srp_recv(int pkt_index, srp_target_t * target)
{
	srp_pkt_t *recv_pkt;
	unsigned char opcode;
	unsigned long cpu_flags;

	recv_pkt = srp_host_pkt_index_to_ptr(pkt_index, target);

	if (recv_pkt == NULL) {
		
		TS_REPORT_FATAL(MOD_SRPTP, "Bad pkt_index %d", pkt_index);
		return;
		
	}

	opcode = recv_pkt->data[0];

	switch (opcode) {
	case SRP_RSP:
		srp_process_resp(target, recv_pkt);
		break;

	case SRP_T_LOGOUT:
		TS_REPORT_WARN(MOD_SRPTP,
			       "Received Target Logout for target %d, try "
			       "reconnecting...", target->target_index);

		/*
		 * Force new I/Os to be queued...
		 * The actual disconnect will occur when we get the
		 * CM:disconnect message from the target.
		 */
		spin_lock_irqsave(&target->spin_lock, cpu_flags);
		if (target->active_conn == recv_pkt->conn) {
			target->timeout = jiffies + TARGET_POTENTIAL_TIMEOUT;
			target->state = TARGET_POTENTIAL_CONNECTION;
		}
		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);
		break;

	case SRP_TPAR_REQ:
	case SRP_AER_REQ:
	default:
		TS_REPORT_WARN(MOD_SRPTP, "SRP IU with unknown %x", opcode);
		break;
	}

	srp_host_free_pkt(recv_pkt);
}

static int srp_host_login_resp(srp_host_conn_t * s, char *data)
{
	srp_login_resp_t *resphdr = (srp_login_resp_t *) data;
	int remote_request_limit;

	remote_request_limit = be32_to_cpu(resphdr->request_limit_delta);
	atomic_set(&s->request_limit, remote_request_limit);
	atomic_set(&s->local_request_limit, MAX_IOS_PER_TARGET);

	if (atomic_read(&s->request_limit) != 0) {

		atomic_inc(&s->request_limit);

		atomic_inc(&s->local_request_limit);

		atomic_set(&s->rwqes_avail, MAX_RECV_WQES);

	} else {
		TS_REPORT_WARN(MOD_SRPTP,
			       "request limit 0 for 0x%llx",
			       be64_to_cpu(s->target->service_name));
	}

	TS_REPORT_WARN(MOD_SRPTP,
		       "Login successful for target %d:%d:%d "
		       "conn %p request_limit %d local_request_limit %d",
		       s->target->target_index,
		       s->port->local_port, s->port->hca->hca_index + 1, 
		       s, atomic_read(&s->request_limit),
		       atomic_read(&s->local_request_limit));

	return 0;
}

static int srp_host_login_rej(char *data)
{
	srp_login_rej_t *resphdr = (srp_login_rej_t *) data;

	TS_REPORT_WARN(MOD_SRPTP, "Login rejected reason code %d",
		       be32_to_cpu(resphdr->reason_code));
	return 0;
}

/*
 * Process a SRP_RSP...
 * Responeses from TASK Management and Commands will come through here
 */
static int srp_process_resp(srp_target_t * target, srp_pkt_t * pkt)
{
	srp_host_conn_t *s;
	srp_resp_t *resphdr;
	ioq_t *ioq = NULL;
	ioq_t *next_ioq = NULL;
	struct scsi_cmnd *scsi_cmnd;
	struct list_head *ioq_entry, *temp_entry;
	unsigned long cpu_flags;

	/* get some variables to get our context */
	resphdr = (srp_resp_t *) pkt->data;
	resphdr->tag = be64_to_cpu(resphdr->tag);

	spin_lock_irqsave(&target->spin_lock, cpu_flags);

	if (target->active_conn != pkt->conn)
		/* Do not process anything, if the connection is down */
		goto unlock_out;

	s = pkt->conn;

	/* set our req_limit_data from the response */
	atomic_add(be32_to_cpu(resphdr->request_limit_delta),
		   &s->request_limit);
	atomic_inc(&s->rwqes_avail);
	atomic_inc(&s->local_request_limit);

	ioq = (ioq_t *) (unsigned long)resphdr->tag;
	TS_REPORT_STAGE(MOD_SRPTP, "Host Processing an SRP Response for ioq %p",
			ioq);

	if (ioq->req == NULL) {
		/*
		 * This could happen if the I/O completes much later after an
		 * abort or lun reset.  SRP Target should ensure that this case
		 * does not occur.
		 */
		TS_REPORT_WARN(MOD_SRPTP, "NULL request queue, for tag %x",
			       (u32) resphdr->tag);
		goto unlock_out;
	} else {
		if (ioq->type == IOQ_INTERNAL_ABORT) {
			TS_REPORT_FATAL(MOD_SRPTP,
					"This ioq %p s_cmd %p marked abort",
					ioq, ioq->req);
			goto unlock_out;
		}

		if (ioq->queue_type == QUEUE_ACTIVE) {
			target->active_count--;
			list_del(&ioq->active_list);
		} else if (ioq->queue_type == QUEUE_PENDING) {
			TS_REPORT_FATAL(MOD_SRPTP,
					"taking completion on pending queue %p",
					ioq);
			goto unlock_out;
		} else {
			TS_REPORT_STAGE(MOD_SRPTP,
					"Processing a response on wrong queue %p ioq");
			goto unlock_out;
		}

		/*
		 * while we have spinlocks lets get the next I/O, if there is
		 * one
		 */
		list_for_each_safe(ioq_entry, temp_entry,
				   &target->resource_list) {

			next_ioq =
			    list_entry(ioq_entry, ioq_t, target_resource_list);

			target->resource_count--;
			list_del(&next_ioq->target_resource_list);
			break;
		}

		scsi_cmnd = (struct scsi_cmnd *)ioq->req;

		ioq->req = NULL;

		spin_unlock_irqrestore(&s->target->spin_lock, cpu_flags);

		/*
		 * Response codes only make sense if task management flags had
		 * been set.
		 */
		if ((ioq->type == IOQ_ABORT) ||
		    (ioq->type == IOQ_LUN_RESET)) {

			TS_REPORT_WARN(MOD_SRPTP,
				       "status word %08x, %02x, rsp %x, rsp_code %x",
				       resphdr->status.word,
				       resphdr->status.bit.status,
				       resphdr->status.bit.rspvalid,
				       ((resp_data_t *)(resphdr + 1))->response_code);

			if (resphdr->status.bit.rspvalid) {

				/* response data */
				ioq->tm_resp_code = 
				    ((resp_data_t *)(resphdr + 1))->response_code;
			} else {
				ioq->tm_resp_code = -1;
			}
			up(&ioq->sema);
		} else {
			if (ioq->timer.function) {
				del_timer(&ioq->timer);
				ioq->timer.function = NULL;
			}

			/* process the response for a command */
			srp_host_iodone(scsi_cmnd, resphdr, ioq);

			if (scsi_cmnd->request_bufflen)
				srp_free_srlist(s->port->hca, ioq, scsi_cmnd);

			srp_host_scsi_complete(scsi_cmnd);

			kfree(ioq);
		}

		if (next_ioq) {
			TS_REPORT_STAGE(MOD_SRPTP,
					"Retrying ioq %p for target %d from "
					"target resource q",
					next_ioq, target->target_index);

			srp_host_internal_qcommand(next_ioq);
		}

	}

	return 0;

     unlock_out:
	spin_unlock_irqrestore(&target->spin_lock, cpu_flags);
	return 0;
}

int
srp_build_taskmgt(srp_pkt_t * pkt,
		  struct scsi_cmnd *Cmnd, srp_host_conn_t * s, ioq_t * ioq)
{
	srp_tsk_mgt_t *mgt_cmd = (srp_tsk_mgt_t *) pkt->data;
	__u8 *p = (__u8 *) (mgt_cmd->lun);

	TS_REPORT_STAGE(MOD_SRP, "Building TM packet for ioq %p", ioq);

	/* build an SRP task management command */
	memset(mgt_cmd, 0, sizeof(srp_tsk_mgt_t));

	/* set indicate in the SRP packet if there is any data movement */
	/* NO DATA MOVEMENT */
	pkt->scatter_gather_list.length = sizeof(srp_tsk_mgt_t);

	/* clear SRP_CMD;     LUN field and set it */
	memset(p, 0, 8);
	p[0] = Cmnd->device->id | 0x80;
	p[1] = Cmnd->device->lun;

	/* set the SRP opcode */
	mgt_cmd->opcode = SRP_TSK_MGT;

	/* Abort command tag  */
	mgt_cmd->tag = cpu_to_be64((unsigned long)ioq);
	/* Tag  of command to abort */
	mgt_cmd->tag_task = cpu_to_be64((unsigned long)Cmnd->SCp.buffer);
	if (ioq->type == IOQ_ABORT) {
		mgt_cmd->function = SRP_ABORT_TASK;
		TS_REPORT_STAGE(MOD_SRPTP, "Abort  for target %d lun %d",
				p[0] & 0x3f, p[1]);
	} else if (ioq->type == IOQ_LUN_RESET) {
		mgt_cmd->function = SRP_LUN_RESET;

		TS_REPORT_STAGE(MOD_SRPTP, "LUN Reset  for target %d lun %d",
				p[0] & 0x3f, p[1]);
	}
	return 0;
}

int srp_host_send_taskmgt(srp_target_t * target, ioq_t * ioq,
			  struct scsi_cmnd *SCpnt)
{
	int status;
	unsigned long cpu_flags;
	srp_pkt_t *send_pkt = NULL;
	srp_pkt_t *recv_pkt = NULL;
	srp_host_conn_t *s;
	unsigned long host_flags;

	s = target->active_conn;

	/*
	 * Check if we have enough credits for one more request on that
	 * connection. If we do not, queue in target resource list
	 */
	if (atomic_dec_and_test(&s->request_limit)) {
		TS_REPORT_STAGE(MOD_SRP,
				"Queuing exceeding request limit for target "
				"%d. Queuing TM on target resource list",
				target->target_index);

		atomic_inc(&s->request_limit);
		goto SEND_FAIL;
	}

	/*
	 * Check if we have enough local credits for one more request 
	 * on that connection. If we do not, queue in target resource list
	 */
	if (atomic_dec_and_test(&s->local_request_limit)) {

		TS_REPORT_STAGE(MOD_SRP,
				"Queuing exceeding local request limit for "
				"target %d. Queuing TM on target resource list",
				target->target_index);

		atomic_inc(&s->request_limit);
		atomic_inc(&s->local_request_limit);
		goto SEND_FAIL;
	}

	ioq->sr_list = NULL;

	/* 
	 * get an srp packet and format it for srptp 
	 * If we cannot, queue in target resource
	 */
	send_pkt = srp_host_get_pkt(target);
	recv_pkt = srp_host_get_pkt(target);
	if ((send_pkt == NULL) || (recv_pkt == NULL)) {
		if (send_pkt) {
			srp_host_free_pkt(send_pkt);
		}
		if (recv_pkt) {
			srp_host_free_pkt(recv_pkt);
		}

		ioq->pkt = NULL;
		ioq->recv_pkt = NULL;

		atomic_inc(&s->request_limit);
		atomic_inc(&s->local_request_limit);
		goto SEND_FAIL;
	}

	ioq->pkt = send_pkt;
	ioq->recv_pkt = recv_pkt;
	send_pkt->conn = s;
	send_pkt->scatter_gather_list.key = target->l_key;
	recv_pkt->conn = s;
	recv_pkt->scatter_gather_list.key = target->l_key;
		    
	send_pkt->scatter_gather_list.length = sizeof(srp_tsk_mgt_t);
	send_pkt->scatter_gather_list.address =
	    (u64) (unsigned long)send_pkt->data;

	/* 
	 * Build the Task management packet
	 */
	srp_build_taskmgt(send_pkt, SCpnt, s, ioq);
	/* 
	 * Put IOQ on active list
	 */
	ioq->queue_type = QUEUE_ACTIVE;

	spin_lock_irqsave(&target->spin_lock, cpu_flags);
	target->active_count++;
	list_add_tail(&ioq->active_list, &target->active_list);
	spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

	/* 
	 * send the srp packet 
	 * we have a problem posting, disconnect, never should happen 
	 */
	local_save_flags(host_flags);
	add_timer(&ioq->timer);
	local_irq_enable();

	status = srptp_post_recv(recv_pkt);
	if (!status)
		status = srptp_post_send(send_pkt);

	if (status) {
		del_timer(&ioq->timer);
		local_irq_restore(host_flags);

		if (send_pkt) {
			srp_host_free_pkt(send_pkt);
		}
		if (recv_pkt) {
			srp_host_free_pkt(recv_pkt);
		}

		ioq->pkt = NULL;
		ioq->recv_pkt = NULL;

		atomic_inc(&s->request_limit);
		atomic_inc(&s->local_request_limit);

		spin_lock_irqsave(&target->spin_lock, cpu_flags);

		TS_REPORT_FATAL(MOD_SRP, "Send post error %d status %d",
				target->target_index, status);
		TS_REPORT_FATAL(MOD_SRP,
				"req_lim %d lreq_lim %d rwqes_avail %d",
				atomic_read(&s->request_limit),
				atomic_read(&s->local_request_limit),
				atomic_read(&s->rwqes_avail));

		target->active_count--;
		list_del(&ioq->active_list);
		target->state = TARGET_POTENTIAL_CONNECTION;
		target->need_disconnect = TRUE;

		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);
		goto SEND_FAIL;
	}

	down(&ioq->sema);
	del_timer(&ioq->timer);
	local_irq_restore(host_flags);

	return (TS_SUCCESS);

      SEND_FAIL:
	return (TS_FAIL);
}

void srp_host_internal_qcommand(ioq_t * ioq)
{
	int status;

	init_timer(&ioq->timer);
	ioq->timer.function = srp_host_cmd_timeout;
	ioq->timer.expires = ioq->timeout;
	ioq->timer.data = (unsigned long)ioq;

	status = srp_host_send_command(ioq->target, ioq, ioq->req);

	if (status) {
		TS_REPORT_STAGE(MOD_SRPTP,
				"Internal qcommand failed for ioq %p on "
				"target %d", ioq,
				ioq->target->target_index);
		srp_host_scsi_complete(ioq->req);
		ioq->req = NULL;
		kfree(ioq);
	}
}

/* NOTE:  currently a scsi target AND a scsi host map to an srp_session
 * a scsi channal maps to an srp_connection.  We need to check
 * this and figure out what we should do if we want multiple
 * hosts and multiple targets
 *
 * We probably need to have a hosts data structure encapsulating
 * the session data structure.  Who cares for now.
 */
/* We use the Scsi_Pointer structure that's included with each command
 * SCSI_Cmnd as a scratchpad for our I/O.
 *
 * SCp will always point to the SRB structure (defined in qla1280.h).
 * It is define as follows:
 *  - SCp.ptr  -- > NULL
 *  - SCp.buffer -->  pointer to ioq
 *  - SCp.buffers_residual --> NULL
 *  - SCp.have_data_in --> not used
 *  - SCp.sent_command --> not used
 *  - SCp.phase --> not used
 */
int srp_host_qcommand(struct scsi_cmnd *SCpnt,
		      void (*done) (struct scsi_cmnd *))
{
	ioq_t *ioq = NULL;
	srp_target_t *target;
	int status;
	srp_host_port_params_t *port;

	target = host_to_target(SCpnt->device->host);
	port = target->port;

	spin_unlock(&target->host->default_lock);

	/* pre-format the SCSI response */
	SCpnt->scsi_done = done;
	SCpnt->sense_buffer[0] = 0;

	if (!target || !target->valid || scsi_unload_in_progress ||
	    (target->state == TARGET_NOT_INITIALIZED)) {
		SCpnt->result = DID_BAD_TARGET << 16;
		goto fail_command;
	}

	if (!target->active_conn ||
	    (target->state != TARGET_ACTIVE_CONNECTION)) {
		
		if (target->state == TARGET_NO_CONNECTIONS)
			SCpnt->result = DID_NO_CONNECT << 16;
		else
			SCpnt->result = DID_IMM_RETRY << 16;

		if (!scsi_unload_in_progress &&
		    (target->block_requests == FALSE) &&
		    (target->state == TARGET_NO_CONNECTIONS)) {
			target->no_connection_io++;
			target->block_requests = TRUE;
		     
			if (target->no_connection_io > MAX_CMD_RETRY_COUNT)
				goto host_busy;
		}

		goto fail_command;
	}

	ioq = kmalloc(sizeof(ioq_t), GFP_ATOMIC);
	if (ioq == NULL) {
		TS_REPORT_FATAL(MOD_SRPTP,
				"kmalloc failed for ioq alloc on tgt %d",
				SCpnt->device->id);
		goto host_busy;
	}

	memset(ioq, 0, sizeof(ioq_t));
	INIT_LIST_HEAD(&ioq->temp_list);
	INIT_LIST_HEAD(&ioq->active_list);
	INIT_LIST_HEAD(&ioq->pending_list);
	INIT_LIST_HEAD(&ioq->target_resource_list);
	INIT_LIST_HEAD(&ioq->hca_resource_list);
	ioq->req = SCpnt;
	ioq->target = target;

	if (SCpnt->timeout_per_command > IOQ_CMD_DELTA_TIMEOUT)
		ioq->timeout =
		    jiffies + SCpnt->timeout_per_command - IOQ_CMD_DELTA_TIMEOUT;
	else
		ioq->timeout = jiffies + SCpnt->timeout_per_command;

	init_timer(&ioq->timer);
	ioq->timer.function = srp_host_cmd_timeout;
	ioq->timer.expires = ioq->timeout;
	ioq->timer.data = (unsigned long)ioq;

	/*
	 * fill out some back pointers for SCSI, so we have this info for
	 * aborting if needed
	 */
	SCpnt->SCp.buffers_residual = 0;
	SCpnt->SCp.ptr = NULL;
	SCpnt->SCp.buffer = (void *)ioq;

	if ((SCpnt->device->type != TYPE_TAPE) &&
	    (SCpnt->allowed < MAX_CMD_RETRY_COUNT))
		SCpnt->allowed = MAX_CMD_RETRY_COUNT;

	status = srp_host_send_command(target, ioq, SCpnt);

	if (status) {
		TS_REPORT_STAGE(MOD_SRPTP,
				"Send Command failed ioq %p on target %d:%d:%d"
				"sending back to scsi mid layer",
				ioq, target->target_index, port->local_port,
				port->hca->hca_index + 1);
		ioq->req = NULL;
		kfree(ioq);
		goto fail_command;
	} else
		goto norm_exit;

      fail_command:
	srp_host_scsi_complete(SCpnt);
      norm_exit:
	spin_lock(&target->host->default_lock);
	return 0;

      host_busy:
	spin_lock(&target->host->default_lock);
	return SCSI_MLQUEUE_HOST_BUSY;
}

int srp_host_send_command(srp_target_t * target, ioq_t * ioq,
			  struct scsi_cmnd *SCpnt)
{
	int status;
	unsigned long cpu_flags;
	srp_pkt_t *send_pkt = NULL;
	srp_pkt_t *recv_pkt = NULL;
	srp_host_conn_t *s;

	s = target->active_conn;

	/*
	 * Check if we have already done allocation
	 */
	if ((ioq->pkt == NULL) || (ioq->recv_pkt == NULL)) {

		spin_lock_irqsave(&target->spin_lock, cpu_flags);
		/* first check if we have remote credits */
		if (atomic_dec_and_test(&s->request_limit)) {

			TS_REPORT_STAGE(MOD_SRPTP,
					"Queuing exceeding request limit for "
					"target %d", target->target_index);

			atomic_inc(&s->request_limit);

			/* queue to the target */
			ioq->queue_type = QUEUE_TARGET;

			target->resource_count++;
			list_add_tail(&ioq->target_resource_list,
				      &target->resource_list);
			spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

			goto SEND_SUCCESS;
		}

		/* next check if we have local credits */
		if (atomic_dec_and_test(&s->local_request_limit)) {

			TS_REPORT_STAGE(MOD_SRPTP,
					"Queuing exceeding local request limit "
					"for target %d", target->target_index);

			atomic_inc(&s->request_limit);
			atomic_inc(&s->local_request_limit);

			if (atomic_read(&s->local_request_limit) < 0)
				TS_REPORT_FATAL(MOD_SRPTP, "less than zero ");

			/* queue to the target */
			ioq->queue_type = QUEUE_TARGET;

			target->resource_count++;
			list_add_tail(&ioq->target_resource_list,
				      &target->resource_list);
			spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

			goto SEND_SUCCESS;
		}
		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

		ioq->type = IOQ_COMMAND;
		ioq->sr_list = NULL;

		/* get an srp packet and format it for srptp */
		send_pkt = srp_host_get_pkt(target);
		recv_pkt = srp_host_get_pkt(target);
		if ((send_pkt == NULL) || (recv_pkt == NULL)) {
			if (send_pkt) {
				srp_host_free_pkt(send_pkt);
			}
			if (recv_pkt) {
				srp_host_free_pkt(recv_pkt);
			}

			ioq->pkt = NULL;
			ioq->recv_pkt = NULL;

			atomic_inc(&s->request_limit);
			atomic_inc(&s->local_request_limit);

			/* queue to the target */
			ioq->queue_type = QUEUE_TARGET;

			TS_REPORT_STAGE(MOD_SRPTP,
					"Queuing no packets target %d",
					target->target_index);

			spin_lock_irqsave(&target->spin_lock, cpu_flags);
			target->resource_count++;
			list_add_tail(&ioq->target_resource_list,
				      &target->resource_list);
			spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

			goto SEND_SUCCESS;
		}

		ioq->pkt = send_pkt;
		ioq->recv_pkt = recv_pkt;

		send_pkt->conn = s;
		send_pkt->scatter_gather_list.key = target->l_key;

		recv_pkt->conn = s;
		recv_pkt->scatter_gather_list.key = target->l_key;
	} else {
		send_pkt = ioq->pkt;
		recv_pkt = ioq->recv_pkt;
	}

	status = srp_build_command(send_pkt, SCpnt, s, ioq);
	if ((status == -EAGAIN) || (status == -ENOMEM)) {
		
		srp_host_hca_params_t *hca = target->port->hca;

		if (SCpnt->request_bufflen)
			srp_free_srlist(hca, ioq, SCpnt);

		TS_REPORT_STAGE(MOD_SRPTP,
				"Build command failed ioq %p, target %d, "
				"queueing to hca %d",
				ioq, target->target_index, hca->hca_index + 1);

		if (SCpnt->device->type == TYPE_TAPE) {
			/*
			 * this is a TAPE device, do not queue otherwise we
			 * could get out of order, instead, fail the command
			 */
			if (send_pkt) {
				srp_host_free_pkt(send_pkt);
			}
			if (recv_pkt) {
				srp_host_free_pkt(recv_pkt);
			}

			ioq->pkt = NULL;
			ioq->recv_pkt = NULL;

			atomic_inc(&s->request_limit);
			atomic_inc(&s->local_request_limit);

			goto SEND_FAIL;
		} else {
			/* queue to the hca */
			ioq->queue_type = QUEUE_HCA;

			spin_lock_irqsave(&hca->spin_lock, cpu_flags);
			hca->resource_count++;
			list_add_tail(&ioq->hca_resource_list,
				      &hca->resource_list);
			spin_unlock_irqrestore(&hca->spin_lock, cpu_flags);

			goto SEND_SUCCESS;
		}

	} else if (status == -EINVAL) {
		if (SCpnt->request_bufflen)
			srp_free_srlist(target->port->hca, ioq, SCpnt);

		srp_host_free_pkt(send_pkt);
		srp_host_free_pkt(recv_pkt);

		ioq->pkt = NULL;
		ioq->recv_pkt = NULL;

		atomic_inc(&s->request_limit);
		atomic_inc(&s->local_request_limit);

		TS_REPORT_STAGE(MOD_SRPTP,
				"ioq %p, target %p, build command failed", ioq,
				target);

		SCpnt->result = DID_BUS_BUSY << 16;

		goto SEND_FAIL;
	}
	
	ioq->queue_type = QUEUE_ACTIVE;

	spin_lock_irqsave(&target->spin_lock, cpu_flags);
	target->active_count++;
	list_add_tail(&ioq->active_list, &target->active_list);
	spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

	add_timer(&ioq->timer);

	/* send the srp packet */
	status = srptp_post_recv(recv_pkt); 
	if (!status)
		status = srptp_post_send(send_pkt);

	if (status) {
		del_timer(&ioq->timer);
		ioq->timer.function = NULL;
	
		if (SCpnt->request_bufflen)
			srp_free_srlist(target->port->hca, ioq, SCpnt);

		if (send_pkt) {
			srp_host_free_pkt(send_pkt);
		}
		if (recv_pkt) {
			srp_host_free_pkt(recv_pkt);
		}

		ioq->pkt = NULL;
		ioq->recv_pkt = NULL;

		atomic_inc(&s->request_limit);
		atomic_inc(&s->local_request_limit);

		spin_lock_irqsave(&target->spin_lock, cpu_flags);

		TS_REPORT_FATAL(MOD_SRPTP, "Send post error %d status %d",
				target->target_index, status);
		TS_REPORT_FATAL(MOD_SRPTP,
				"req_lim %d lreq_lim %d rwqes_avail %d",
				atomic_read(&s->request_limit),
				atomic_read(&s->local_request_limit),
				atomic_read(&s->rwqes_avail));

		target->active_count--;
		list_del(&ioq->active_list);
		target->state = TARGET_POTENTIAL_CONNECTION;
		target->need_disconnect = TRUE;
		SCpnt->result = DID_BUS_BUSY << 16;

		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

		goto SEND_FAIL;
		     
	}

      SEND_SUCCESS:
	return (TS_SUCCESS);

      SEND_FAIL:
	return (TS_FAIL);
}


/* generic function to send task mgt 
 * In case of abort , SCpnt is scsi_cmd to abort. IOQ is IOQ of 
 * task management command
 */

int srp_host_abort_eh(struct scsi_cmnd *SCpnt)
{
	unsigned long cpu_flags;
	ioq_t *ioq = NULL;
	ioq_t *cmd_ioq = NULL;
	int status;
	srp_target_t *target;
	srp_host_port_params_t *port;

	if (scsi_unload_in_progress == TRUE)
		return FAILED;

	target = host_to_target(SCpnt->device->host);
	port = target->port;

	cmd_ioq = (ioq_t *) SCpnt->SCp.buffer;

	if (cmd_ioq == NULL) {
		TS_REPORT_WARN(MOD_SRP,
			       "SCpnt %p on target %d:%d:%d completed",
			       SCpnt, SCpnt->device->id,
			       target->port->local_port,
			       target->port->hca->hca_index + 1);
		return SUCCESS;
	}

	spin_unlock(&target->host->default_lock);

	spin_lock_irqsave(&target->spin_lock, cpu_flags);

	TS_REPORT_WARN(MOD_SRP,
			"Aborting TAG %p SCpnt %p on target %d:%d:%d" 
			" lun %d tgt_state %d a_cnt %d p_cnt %d",
			cmd_ioq, SCpnt, SCpnt->device->id, port->local_port,
			port->hca->hca_index + 1, SCpnt->device->lun, target->state,
			target->active_count, target->pending_count);

	/* 
	 * marked the abort_ioq as internal_abort to avoid free
	 * it in active list twice (in process_resp and in 
	 * abort_eh - now free it from active_list in abort_eh
	 */
	cmd_ioq->type = IOQ_INTERNAL_ABORT;

	if (cmd_ioq->queue_type == QUEUE_TARGET) {
		TS_REPORT_WARN(MOD_SRP,
				"Abort ioq %p removing from target "
				"resource queue r_cnt %d",
				cmd_ioq, target->resource_count);
		target->resource_count--;
		list_del(&cmd_ioq->target_resource_list);

		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);
		goto out;
	} else if (cmd_ioq->queue_type == QUEUE_HCA) {
		srp_host_hca_params_t *hca = target->port->hca;

		TS_REPORT_WARN(MOD_SRP,
				"Abort ioq %p removing from hca "
				"resource queue", cmd_ioq);

		spin_lock(&hca->spin_lock);
		hca->resource_count--;
		list_del(&cmd_ioq->hca_resource_list);
		spin_unlock(&hca->spin_lock);

		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

		if (cmd_ioq->pkt) {
			srp_host_free_pkt(cmd_ioq->pkt);
		}
		if (cmd_ioq->recv_pkt) {
			srp_host_free_pkt(cmd_ioq->recv_pkt);
		}

		cmd_ioq->pkt = NULL;
		cmd_ioq->recv_pkt = NULL;

		if (target->state == TARGET_ACTIVE_CONNECTION) {
			atomic_inc(&target->active_conn->request_limit);
			atomic_inc(&target->active_conn->local_request_limit);
		}
		goto out;
	} else if (cmd_ioq->queue_type == QUEUE_PENDING) {
		TS_REPORT_WARN(MOD_SRP,
				"Abort ioq %p removing from pending "
				"queue p_cnt %d", cmd_ioq, target->pending_count);
		target->pending_count--;
		list_del(&cmd_ioq->pending_list);
		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);
		goto out;
	} else if (cmd_ioq->queue_type == QUEUE_ACTIVE) {

		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

		if ((target->state == TARGET_ACTIVE_CONNECTION) &&
		    target->active_conn) {
			ioq = kmalloc(sizeof(ioq_t), GFP_ATOMIC);
			if (ioq == NULL) {
				TS_REPORT_FATAL(MOD_SRP,
						"failed kmalloc ioq on tgt %d",
						SCpnt->device->id);
				spin_lock_irqsave(&target->spin_lock, cpu_flags);
				target->active_count--;
				list_del(&cmd_ioq->active_list);
				spin_unlock_irqrestore(&target->spin_lock, cpu_flags);
				goto out;
			}

			TS_REPORT_WARN(MOD_SRP,
					"Abort ioq %p on active queue."
					" Sending TM_ioq %p a_cnt %d",
					cmd_ioq, ioq, target->active_count);

			memset(ioq, 0, sizeof(ioq_t));

			sema_init(&ioq->sema, 0);
			init_timer(&ioq->timer);
			ioq->timer.function = tm_timedout;
			ioq->timer.expires = jiffies + HZ;
			ioq->timer.data = (unsigned long)ioq;

			ioq->type = IOQ_ABORT;
			ioq->req = SCpnt;
			ioq->target = target;

			status = srp_host_send_taskmgt(target, ioq, SCpnt);

			if (status) {
				TS_REPORT_WARN(MOD_SRP,
						"Send_TM_ioq %p to abort cmd_ioq %p"
						" failed on target %d",
						ioq, target->target_index);
				spin_lock_irqsave(&target->spin_lock, cpu_flags);
				target->active_count--;
				list_del(&cmd_ioq->active_list);
				spin_unlock_irqrestore(&target->spin_lock, cpu_flags);
				kfree(ioq);
				goto out;
			}
			
			if (ioq->tm_resp_code == TSK_MGT_TIMEOUT) {
				TS_REPORT_WARN(MOD_SRP,
						"TM_ioq %p resp. Timeout", ioq);
				atomic_inc(&target->active_conn->request_limit);
				atomic_inc(&target->active_conn->local_request_limit);
				/* remove the TM ioq out of active_list */
				spin_lock_irqsave(&target->spin_lock, cpu_flags);
				target->active_count--;
				list_del(&ioq->active_list);
				spin_unlock_irqrestore(&target->spin_lock, cpu_flags);
			} else if (ioq->tm_resp_code == -1) {
				TS_REPORT_WARN(MOD_SRPTP,
					       "No valid response TM_ioq %p", ioq);
			} else if (ioq->tm_resp_code == NO_FAILURE) {
				TS_REPORT_WARN(MOD_SRP,
						"NO FAILURE TM_ioq %p", ioq);
			} else if (ioq->tm_resp_code == TSK_MGT_FAILED) {
				TS_REPORT_WARN(MOD_SRP,
						"TSK_MGT_FAILED TM_ioq %p", ioq);
			} else if (ioq->tm_resp_code == REQUEST_FIELD_INVLD) {
				TS_REPORT_WARN(MOD_SRPTP,
					       "Invalid field TM_ioq %p", ioq);
			} else if (ioq->tm_resp_code == TSK_MGT_FUN_NOT_SUP) {
				TS_REPORT_WARN(MOD_SRPTP,
					       "Unsupported TM. TM_ioq %p", ioq);
			} else {
				TS_REPORT_WARN(MOD_SRPTP,
					       "Invalid response 0x%x. TM_ioq %p",
					       ioq->tm_resp_code, ioq);
			}

			kfree(ioq);
		}

		spin_lock_irqsave(&target->spin_lock, cpu_flags);
		target->active_count--;
		list_del(&cmd_ioq->active_list);
		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

	} else {
		BUG();
	}

      out:
	if (SCpnt->request_bufflen)
		srp_free_srlist(port->hca, cmd_ioq, SCpnt);
	SCpnt->SCp.buffer = NULL;
	if (cmd_ioq->timer.function) {
		del_timer(&cmd_ioq->timer);
		cmd_ioq->timer.function = NULL;
	}
	kfree(cmd_ioq);
	spin_lock(&target->host->default_lock);
	return SUCCESS;
}

int srp_host_device_reset_eh(struct scsi_cmnd *SCpnt)
{
	int status;
	unsigned long cpu_flags;
	srp_target_t *target;
	srp_host_port_params_t *port;

	if (scsi_unload_in_progress == TRUE)
		return FAILED;

	target = host_to_target(SCpnt->device->host);
	port = target->port;

	spin_unlock(&target->host->default_lock);

	TS_REPORT_WARN(MOD_SRPTP,
		       "Reset LUN %d of target %d:%d:%d tgt_state %d"
		       " a_cnt %d p_cnt %d r_cnt %d",
		       SCpnt->device->lun, SCpnt->device->id, port->local_port,
		       port->hca->hca_index + 1, target->state, target->active_count,
		       target->pending_count, target->resource_count);

	spin_lock_irqsave(&target->spin_lock, cpu_flags);

	if ((target->state == TARGET_NOT_INITIALIZED) ||
	    (target->state == TARGET_NO_CONNECTIONS)) {

		TS_REPORT_WARN(MOD_SRP,
			       "Aborting LUN  with no connection , target %d",
			       SCpnt->device->id);
		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);
		goto out;
	} else if ((target->state == TARGET_POTENTIAL_CONNECTION) ||
		   (target->state == TARGET_NO_PATHS)) {

		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

		srp_lun_pending_to_scsi(target, DID_SOFT_ERROR, SCpnt);
		goto out;
	} else if (target->state == TARGET_ACTIVE_CONNECTION) {
		ioq_t *ioq = NULL;

		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

		if (target->active_conn) {
			ioq = kmalloc(sizeof(ioq_t), GFP_ATOMIC);

			if (ioq == NULL) {
				TS_REPORT_FATAL(MOD_SRPTP,
						"kmalloc failed for ioq alloc on tgt %d",
						SCpnt->device->id);
				goto out;
			}
	
			memset(ioq, 0, sizeof(ioq_t));

			sema_init(&ioq->sema, 0);
			init_timer(&ioq->timer);
			ioq->timer.function = tm_timedout;
			ioq->timer.expires = jiffies + HZ;
			ioq->timer.data = (unsigned long)ioq;

			ioq->type = IOQ_LUN_RESET;
			ioq->req = SCpnt;
			ioq->target = target;

			status = srp_host_send_taskmgt(target, ioq, SCpnt);

			if (status) {
				TS_REPORT_STAGE(MOD_SRPTP,
						"LUN Reset failed ioq %p on target %d, "
						"sending back to scsi mid layer",
						ioq, target->target_index);
				srp_lun_pending_to_scsi(target, DID_BUS_BUSY, SCpnt);
				kfree(ioq);
				goto out;
			}

			TS_REPORT_WARN(MOD_SRP,
					"LUN Reset TM_ioq %p on target %d:%d:%d",
					ioq, target->target_index, port->local_port,
					port->hca->hca_index + 1);

			if (ioq->tm_resp_code == TSK_MGT_TIMEOUT) {
				TS_REPORT_WARN(MOD_SRP,
						"TM_ioq %p resp. Timeout", ioq);
				atomic_inc(&target->active_conn->request_limit);
				atomic_inc(&target->active_conn->local_request_limit);
				/* remove the TM ioq out of active_list */
				spin_lock_irqsave(&target->spin_lock, cpu_flags);
				target->active_count--;
				list_del(&ioq->active_list);
				spin_unlock_irqrestore(&target->spin_lock, cpu_flags);
			} else if (ioq->tm_resp_code == -1) {
				TS_REPORT_WARN(MOD_SRPTP,
					       "No valid response TM_ioq %p", ioq);
			} else if (ioq->tm_resp_code == NO_FAILURE) {
				TS_REPORT_WARN(MOD_SRP,
						"NO FAILURE TM_ioq %p", ioq);
			} else if (ioq->tm_resp_code == TSK_MGT_FAILED) {
				TS_REPORT_WARN(MOD_SRP,
						"TSK_MGT_FAILED TM_ioq %p", ioq);
			} else if (ioq->tm_resp_code == REQUEST_FIELD_INVLD) {
				TS_REPORT_WARN(MOD_SRPTP,
					       "Invalid field TM_ioq %p", ioq);
			} else if (ioq->tm_resp_code == TSK_MGT_FUN_NOT_SUP) {
				TS_REPORT_WARN(MOD_SRPTP,
					       "Unsupported TM. TM_ioq %p", ioq);
			} else {
				TS_REPORT_WARN(MOD_SRPTP,
					       "Invalid response 0x%x. TM_ioq %p",
					       ioq->tm_resp_code, ioq);
			}

			kfree(ioq);
		}

		srp_move_lun_resource_to_pending(target, SCpnt);
		srp_move_lun_active_to_pending(target, SCpnt);
		srp_lun_pending_to_scsi(target, DID_SOFT_ERROR, SCpnt);

	} else {
		BUG();
	}

      out:
	kfree(SCpnt->SCp.buffer);
	SCpnt->SCp.buffer = NULL;
	spin_lock(&target->host->default_lock);
	return (SUCCESS);
}

int srp_host_reset_eh(struct scsi_cmnd *SCpnt)
{
	srp_target_t *target;
	srp_host_port_params_t *port;

	if (scsi_unload_in_progress == TRUE)
		return FAILED;

	target = host_to_target(SCpnt->device->host);
	port = target->port;

	spin_unlock(&target->host->default_lock);

	TS_REPORT_WARN(MOD_SRPTP, "Host reset target %d:%d:%d",
			SCpnt->device->id, port->local_port,
			port->hca->hca_index + 1);

	srp_move_to_pending(target);
	srp_pending_to_scsi(target, DID_SOFT_ERROR);
	
	if ((target->state == TARGET_ACTIVE_CONNECTION) &&
	    target->active_conn) {
		target->state = TARGET_POTENTIAL_CONNECTION;
		target->need_disconnect = TRUE;
	}

	spin_lock(&target->host->default_lock);
	return (SUCCESS);
}

/*
 * end of Error Handling entry points
 */

int srp_host_login(srp_host_conn_t * s)
{
	srp_login_req_t *header;

	TS_REPORT_STAGE(MOD_SRPTP, "Sending Login request for conn %p", s);
	TS_REPORT_STAGE(MOD_SRPTP, "SRP Initiator GUID:      %llx",
			be64_to_cpu(*(u64 *) & s->port->local_gid[8]));

	s->login_buff_len = sizeof(srp_login_req_t);
	header = (srp_login_req_t *) s->login_buff;

	/*
	 * zero out header
	 */
	memset(header, 0, sizeof(srp_login_req_t));
	header->opcode = SRP_LOGIN_REQ;
	header->tag = 0;
	header->request_I_T_IU = cpu_to_be32(srp_cmd_pkt_size);
	header->initiator_port_id[0] = (s->path_record.dlid >> 8);
	header->initiator_port_id[1] = s->path_record.dlid & 0xff;
	header->initiator_port_id[7] = s->port->local_port;
	memcpy(&header->initiator_port_id[8], &s->port->local_gid[8],
	       sizeof(tTS_IB_GUID));
	memcpy(&header->target_port_id[0], &s->target->service_name, 8);
	memcpy(&header->target_port_id[8], s->target->ioc.guid,
	       sizeof(tTS_IB_GUID));

	header->req_buf_format.idbd = 1;
	header->req_buf_format.ddbd = 1;

	return 0;
}

int
srp_build_command(srp_pkt_t * pkt,
		  struct scsi_cmnd *Cmnd, srp_host_conn_t * s, ioq_t * ioq)
{
	srp_cmd_t *cmd = (srp_cmd_t *) pkt->data;
	__u8 *p = (__u8 *) (cmd->lun);
	int status;

	/* build an SRP command */
	memset(cmd, 0, sizeof(srp_cmd_t));

	/* set indicate in the SRP packet if there is any data movement */
	if (Cmnd->request_bufflen) {
		if (Cmnd->sc_data_direction == SCSI_DATA_READ) {
			cmd->difmt = 2;	/* indirect table */
			/*
			 * cmd->dicount will be updated in the
			 * build_routine
			 */
		} else if (Cmnd->sc_data_direction == SCSI_DATA_WRITE) {
			cmd->dofmt = 2;	/* indirect table */
			/*
			 * cmd->docount will be updated in the
			 * build_routine
			 */
		}
		status = srp_build_rdma(pkt, Cmnd, s, ioq, cmd);
		if (status) {
			
			return (status);
			
		}
	} else {
		/* NO DATA MOVEMENT */
		cmd->dofmt = cmd->difmt = 0;
		pkt->scatter_gather_list.length = sizeof(srp_cmd_t);
		ioq->sr_list = NULL;
	}

	/* copy the SCSI CDB into the SRP packet */
	memcpy(cmd->cdb, Cmnd->cmnd, Cmnd->cmd_len);

	/* clear SRP_CMD LUN field and set it */
	memset(p, 0, 8);
	p[0] = Cmnd->device->id | 0x80;
	p[1] = Cmnd->device->lun;
	TS_REPORT_STAGE(MOD_SRPTP, "command for target %d lun %d",
			p[0] & 0x3f, p[1]);

	/* set the SRP opcode */
	cmd->opcode = SRP_CMD;

	/*
	 * set the tag and increment it, make it modulo of
	 * max io queue size-2 ??
	 */
	cmd->tag = cpu_to_be64((unsigned long)ioq);
	return 0;
}

void srp_host_iodone(struct scsi_cmnd *cmnd, srp_resp_t * hdr, ioq_t * ioq)
{
	TS_REPORT_STAGE(MOD_SRPTP, "io %p complete for target %d", ioq,
			cmnd->device->id);

	cmnd->result = (DID_OK << 16) | (hdr->status.bit.status & 0xff);

	if ((hdr->status.bit.diunder) || (hdr->status.bit.diover)) {
		cmnd->resid = be32_to_cpu(hdr->data_in_residual_count);
	} else if ((hdr->status.bit.dounder) || (hdr->status.bit.doover)) {
		cmnd->resid = be32_to_cpu(hdr->data_out_residual_count);
	}

	if (hdr->status.bit.status) {
		if (status_byte(hdr->status.bit.status) == CHECK_CONDITION) {
			TS_REPORT_STAGE(MOD_SRPTP, "io %p scsi_cmd %p for lun %d" 
					" target %d check condition",
					ioq, cmnd, cmnd->device->lun, cmnd->device->id);

			memset(cmnd->sense_buffer, 0,
			       sizeof(cmnd->sense_buffer));

			if (hdr->status.bit.snsvalid) {
				int srp_sense_size =
				    be32_to_cpu(hdr->sense_len);
				char *sense_buffer;
				int sense_size;
				int i;

				if (srp_sense_size < sizeof(cmnd->sense_buffer))
					sense_size = srp_sense_size;
				else
					sense_size =
					    sizeof(cmnd->sense_buffer) - 1;

				if (hdr->status.bit.rspvalid)
					sense_buffer =
					    (char *)hdr + sizeof(srp_resp_t) +
					    be32_to_cpu(hdr->response_len);
				else
					sense_buffer =
					    (char *)hdr + sizeof(srp_resp_t);

				for (i = 0; i < sense_size; i += 4) {
					TS_REPORT_STAGE(MOD_SRPTP,
							"sense_buffer[%d] %x",
							i,
							*((u32 *) &
							  sense_buffer[i]));
				}
				TS_REPORT_STAGE(MOD_SRPTP,
						"io %p sense size %d", ioq,
						sense_size);

				memcpy(cmnd->sense_buffer, sense_buffer,
				       sense_size);
				cmnd->SCp.Message = sense_size;
			}

		} else {
			cmnd->result =
			    (DID_OK << 16) | (hdr->status.bit.status & 0xff);

			TS_REPORT_WARN(MOD_SRPTP, "io %p scsi_cmd %p lun %d"
					"  target %d status 0x%x",
					ioq, cmnd, cmnd->device->lun,
					cmnd->device->id,
					cmnd->result);
		}
	}

}

int
srp_build_indirect_data_buffer_descriptor(srp_pkt_t * srp_pkt,
					  struct scatterlist *sg_elements,
					  int sg_cnt,
					  /*
					   * output parameters
					   */
					  srp_host_buf_t ** sr_list_ptr,
					  int *sr_list_length)
{

	srp_cmd_t *srp_cmd_frame = (srp_cmd_t *) srp_pkt->data;

	struct scatterlist *sg_element;

	srp_cmd_indirect_data_buffer_descriptor_t
	    * header = (srp_cmd_indirect_data_buffer_descriptor_t *)
	    (srp_cmd_frame + 1);

	srp_remote_buf_t *curr_buff_descriptor =
	    &header->partial_memory_descriptor_list[0];

	u32 total_length = 0;

	dma_addr_t curr_dma_addr, base_dma_addr;
	u32 curr_registration_length = 0, curr_dma_length = 0;

	u64 *dma_addr_list;
	u32 dma_addr_index = 0;
	int status;

	srp_host_buf_t *sr_list;

	TS_REPORT_STAGE(MOD_SRPTP, "cmd 0x%p header 0x%p curr buff 0x%p",
			srp_cmd_frame, header, curr_buff_descriptor);

	/*
	 * allocate, assuming worst case, 1 registration per sg element
	 */
	sr_list = *sr_list_ptr = (srp_host_buf_t *)
	    kmalloc(sizeof(srp_host_buf_t) * sg_cnt, GFP_ATOMIC);
	if (!sr_list) {
		return (-ENOMEM);
	}

	sr_list->data = NULL;

	dma_addr_list = (u64 *)
	    kmalloc(sizeof(u64) *
		    ((max_xfer_sectors_per_io * 512 / PAGE_SIZE) + 2),
		    GFP_ATOMIC);
	if (dma_addr_list == NULL) {
		kfree(sr_list);
		*sr_list_ptr = NULL;
		return -ENOMEM;
	}

	dma_addr_index = 0;

	*sr_list_length = 0;

	/*
	 * register the sg elements.
	 * Try to minimize the number of registrations
	 */

	curr_dma_addr = sg_dma_address(&sg_elements[0]);

	for (sg_element = &sg_elements[0];
	     sg_element < &sg_elements[sg_cnt]; sg_element++) {

		TS_REPORT_STAGE(MOD_SRPTP, "(indb) addr 0x%lx length 0x%x",
				(unsigned long)sg_dma_address(sg_element),
				sg_dma_len(sg_element));
		total_length += sg_dma_len(sg_element);
		curr_registration_length += sg_dma_len(sg_element);
		curr_dma_length += sg_dma_len(sg_element);

		/*
		 * This field is used as the suggested IOVA to the phys memory
		 * registration verb
		 */
		if (sr_list->data == NULL)
			sr_list->data =
			    (char *)(unsigned long)sg_dma_address(sg_element);

		/*
		 * Fill in the DMA addresses of this (DMA) contiguous region
		 * for physical memory registration later.
		 */
		base_dma_addr = curr_dma_addr;
		for (;
		     ((curr_dma_addr & PAGE_MASK) <=
		      ((base_dma_addr + curr_dma_length - 1) & PAGE_MASK));
		     curr_dma_addr += PAGE_SIZE) {

			dma_addr_list[dma_addr_index++] =
			    curr_dma_addr & PAGE_MASK;

		}

		/*
		 * Register the current region
		 */
		sr_list->size = curr_registration_length;
		sr_list->r_addr = (u64) (unsigned long)sr_list->data;

		status = srptp_register_memory(srp_pkt->conn,
					       sr_list,
					       ((unsigned long)sr_list->
						data & (PAGE_SIZE - 1)),
					       dma_addr_list, dma_addr_index);

		if (status == -EAGAIN) {

			
			for (sr_list = &((*sr_list_ptr)[0]);
			     sr_list < &((*sr_list_ptr)[*sr_list_length]);
			     sr_list++) {
				srptp_dereg_phys_host_buf(sr_list);
			}

			kfree(*sr_list_ptr);
			kfree(dma_addr_list);
			*sr_list_ptr = NULL;
			*sr_list_length = 0;
			return -ENOMEM;
			

		} else if (status) {

			
			TS_REPORT_FATAL(MOD_SRPTP, "FMR failure %d", status);
			kfree(*sr_list_ptr);
			kfree(dma_addr_list);
			*sr_list_ptr = NULL;
			*sr_list_length = 0;
			return -ENOMEM;
			  
		}

		memset(curr_buff_descriptor, 0, sizeof(srp_remote_buf_t));
		curr_buff_descriptor->r_data = cpu_to_be64(sr_list->r_addr);
		curr_buff_descriptor->r_size = cpu_to_be32(sr_list->size);
		curr_buff_descriptor->r_key = cpu_to_be32((int)sr_list->r_key);

		/*
		 * Reset the dma_addr_list
		 */
		dma_addr_index = 0;
		curr_registration_length = 0;

		/*
		 * Update the sr_list
		 */
		sr_list++;
		curr_buff_descriptor++;
		(*sr_list_length)++;
		if (*sr_list_length < sg_cnt)
			sr_list->data = NULL;

		/*
		 * Reset our current DMA region to the next (DMA) address region
		 * (if there is a next region)
		 */
		if ((sg_element + 1) != &sg_elements[sg_cnt]) {
			curr_dma_addr = sg_dma_address(sg_element + 1);
			curr_dma_length = 0;
		}
	}

	kfree(dma_addr_list);

	/*
	 * Fill up information about the indirect table.
	 * This could be used by the target to RDMA read in
	 * the indirect table.
	 */
	memset(header, 0, sizeof(srp_cmd_indirect_data_buffer_descriptor_t));
	header->indirect_table_descriptor.r_data = cpu_to_be64((unsigned long)
							       &header->
							       partial_memory_descriptor_list
							       [0]);
	header->indirect_table_descriptor.r_key = cpu_to_be32(srp_pkt->r_key);
	header->indirect_table_descriptor.r_size =
	    cpu_to_be32((*sr_list_length) * sizeof(srp_remote_buf_t));

	header->total_length = cpu_to_be32(total_length);

	if (srp_cmd_frame->dofmt)
		srp_cmd_frame->docount = (u8) (*sr_list_length);
	else if (srp_cmd_frame->difmt)
		srp_cmd_frame->dicount = (u8) (*sr_list_length);
	else
		srp_cmd_frame->dicount = srp_cmd_frame->docount = 0;

	srp_pkt->scatter_gather_list.length = sizeof(srp_cmd_t) +
	    sizeof(srp_cmd_indirect_data_buffer_descriptor_t) +
	    (*sr_list_length) * sizeof(srp_remote_buf_t);

	TS_REPORT_STAGE(MOD_SRPTP,
			"total indirect buf length 0x%x dicount 0x%x docount 0x%x",
			total_length, srp_cmd_frame->dicount,
			srp_cmd_frame->docount);

	return 0;
}

int
srp_build_rdma(srp_pkt_t * srp_pkt,
	       struct scsi_cmnd *cmnd, srp_host_conn_t * s, ioq_t * ioq,
	       srp_cmd_t * srp_cmd)
{
	srp_remote_buf_t *srp_remote_buf;
	srp_host_buf_t *sr_list;
	struct scatterlist scatter_list;
	struct scatterlist *st_buffer;
	int sg_cnt;
	int num_sg_elements, unaligned;
	int offset, max_phys_pages, page_offset;
	u64 *phys_buffer_list;
	dma_addr_t new_phys_page, old_phys_page;
	int status, buf_len, num_phys_pages, old_buf_len;

	TS_REPORT_STAGE(MOD_SRPTP, "use_sg %d Requested buffer %p phys %lx",
			cmnd->use_sg, cmnd->request_buffer,
			virt_to_bus(cmnd->request_buffer));

	if (cmnd->use_sg) {
		st_buffer = (struct scatterlist *)cmnd->request_buffer;
		num_sg_elements = pci_map_sg(s->port->hca->hca_pdev,
					     st_buffer,
					     cmnd->use_sg,
					     scsi_to_pci_dma_dir(cmnd->
								 sc_data_direction));

	} else {
		memset(&scatter_list, 0, sizeof(struct scatterlist));
		scatter_list.page = virt_to_page(cmnd->request_buffer);
		scatter_list.offset =
		    ((unsigned long)cmnd->request_buffer & ~PAGE_MASK);
		scatter_list.length = cmnd->request_bufflen;
		st_buffer = &scatter_list;
		num_sg_elements =	/* 1 */
		    pci_map_sg(s->port->hca->hca_pdev, st_buffer, 1,
			       scsi_to_pci_dma_dir(cmnd->sc_data_direction));
	}

	sr_list =
	    (srp_host_buf_t *) kmalloc(sizeof(srp_host_buf_t), GFP_ATOMIC);
	if (!sr_list) {
		TS_REPORT_WARN(MOD_SRPTP, "sr_list allocation failed");
		return (-ENOMEM);
	}

	sr_list->data = (char *)(unsigned long)sg_dma_address(&st_buffer[0]);
	sr_list->r_addr = (unsigned long)sr_list->data;
	sr_list->size = (u32) cmnd->request_bufflen;

	/*
	 * compute the number of physical pages
	 * 1) compute the number of whole pages
	 * 2) if there is an offset in the first page, add page, then
	 *    check if we need to add a page for the remainder for the last page.
	 * 3) even if there isn't an offset. the buffer may still have a
	 *    remainder, in that case add a page for the remainder in the last
	 *    page.
	 */
	max_phys_pages = cmnd->request_bufflen / PAGE_SIZE;
	page_offset =
	    (unsigned long)sg_dma_address(&st_buffer[0]) & (PAGE_SIZE - 1);
	if (page_offset) {
		max_phys_pages++;
		if ((PAGE_SIZE - page_offset) <
		    (cmnd->request_bufflen % PAGE_SIZE))
			max_phys_pages++;
	} else if (cmnd->request_bufflen % PAGE_SIZE) {
		max_phys_pages++;
	}

	phys_buffer_list =
	    (u64 *) kmalloc(sizeof(u64) * max_phys_pages, GFP_ATOMIC);
	if (phys_buffer_list == NULL) {
		TS_REPORT_WARN(MOD_SRPTP, "phys buffer list allocation failed");
		kfree(sr_list);
		return (-ENOMEM);
	}

	old_phys_page = 0;
	old_buf_len = 0;
	num_phys_pages = 0;
	unaligned = 0;

	/*
	 * Next part of the code involves converting the scatter list given
	 * by the scsi stack into a list of physical pages to give to ts_api
	 * for fast memory registration.  There are 4 cases we are concerned
	 * with:
	 *  1. The scsi stack will give sequential buffers that are contigous
	 *     and are less than a single page.  These type of buffers need to
	 *     be coalesced into a single page to be given to ts_api.  This
	 *     is typical on doing reads and scsi gives a scatter list with
	 *     buffers that are less than a PAGE.
	 *  2. The scsi statck will give sequential buffers that are the same.
	 *     They have the same va and pa.  In this case we do not want to
	 *     coalesce the buffers.  This is the case when you are doing
	 *     writes from /dev/zero to your drive.
	 *  3. The scsi stack gives you sequential buffers that are less than
	 *     a page and are not contigous.  THIS WOULD BE A PROBLEM.  We
	 *     need to trap for it.
	 *  4. If the buffer is greater than a page then we want to assume
	 *     the buffer is physically contigous and add those pages as
	 *     well.
	 */
	for (sg_cnt = 0; sg_cnt < num_sg_elements; sg_cnt++) {

		new_phys_page =
		    (unsigned long)sg_dma_address(&st_buffer[sg_cnt]);
		buf_len = sg_dma_len(&st_buffer[sg_cnt]);

		TS_REPORT_STAGE(MOD_SRPTP, "(1) virtual[%x] %llx len %x",
				sg_cnt, new_phys_page, buf_len);
		if (new_phys_page & (PAGE_SIZE - 1)) {
			if (unaligned) {
				TS_REPORT_STAGE(MOD_SRPTP,
						"buffer has offset not first - build_indirect");
				kfree(phys_buffer_list);
				kfree(sr_list);
				goto build_indirect;
			}
			unaligned = 1;
		}
		if ((new_phys_page + buf_len) & (PAGE_SIZE - 1)) {
			if (unaligned) {
				TS_REPORT_STAGE(MOD_SRPTP,
						"buffer + len not page boundary not last - build_indirect");
				kfree(phys_buffer_list);
				kfree(sr_list);
				goto build_indirect;
			}
			unaligned = 1;
		}
	}

	for (sg_cnt = 0; sg_cnt < num_sg_elements; sg_cnt++) {

		new_phys_page =
		    (unsigned long)sg_dma_address(&st_buffer[sg_cnt]);
		buf_len = sg_dma_len(&st_buffer[sg_cnt]);

		TS_REPORT_STAGE(MOD_SRPTP, "(2) virtual[%x] %llx len %x",
				sg_cnt, new_phys_page, buf_len);

		if (buf_len > PAGE_SIZE) {
			int pages_in_buf;
			int pg_cnt;

			/*
			 * The following are checks to see if the buffer is
			 * aligned at the start and the end.  We only allow
			 * exceptions if the buffer is the first or the last.
			 */
			if ((new_phys_page & (PAGE_SIZE - 1)) && (sg_cnt != 0)) {
				TS_REPORT_STAGE(MOD_SRPTP,
						"buffer larger than 4k and has "
						"offset and not first");
				kfree(phys_buffer_list);
				kfree(sr_list);
				goto build_indirect;
			}
			if ((((new_phys_page + buf_len) & (PAGE_SIZE - 1)) != 0)
			    && (sg_cnt != (num_sg_elements - 1))) {
				TS_REPORT_STAGE(MOD_SRPTP,
						"buf len 0x%x "
						"new_phys_page 0x%x",
						buf_len, new_phys_page);
				TS_REPORT_STAGE(MOD_SRPTP,
						"sg cnt %d num_sg_elements %d",
						sg_cnt, num_sg_elements);
				TS_REPORT_STAGE(MOD_SRPTP, "total length 0x%x",
						cmnd->request_bufflen);
				TS_REPORT_STAGE(MOD_SRPTP,
						"buffer larger than 4k and not "
						"aligned at end and not last");
				TS_REPORT_STAGE(MOD_SRPTP,
						"next addr 0x%x len 0x%x",
						(unsigned long)(sg_dma_address
								(&st_buffer
								 [sg_cnt + 1])),
						sg_dma_len(&st_buffer
							   [sg_cnt + 1]));
				kfree(phys_buffer_list);
				kfree(sr_list);
				goto build_indirect;
			}

			pages_in_buf = buf_len / PAGE_SIZE;
			if (buf_len % PAGE_SIZE)
				pages_in_buf++;

			for (pg_cnt = 0; pg_cnt < pages_in_buf; pg_cnt++) {
				phys_buffer_list[num_phys_pages] =
				    new_phys_page & PAGE_MASK;
				TS_REPORT_STAGE(MOD_SRPTP,
						"physical[%x], phys page %llx",
						num_phys_pages,
						phys_buffer_list
						[num_phys_pages]);
				new_phys_page += PAGE_SIZE;
				num_phys_pages++;
			}

			/*
			 * back up the new_phys_page from the loop for the
			 * last entry
			 */
			new_phys_page -= PAGE_SIZE;

		} else {

			/* we have no offset, add the page */
			phys_buffer_list[num_phys_pages] =
			    new_phys_page & PAGE_MASK;
			TS_REPORT_STAGE(MOD_SRPTP,
					"physical[%x], phys page %llx",
					num_phys_pages,
					phys_buffer_list[num_phys_pages]);
			num_phys_pages++;
		}

		/* the old phys page needs to be the page with the offset for
		 * doing the contingous calculation above */
		old_phys_page = new_phys_page;
		old_buf_len = buf_len;
	}

	/* get the offset of the virtual address */
	offset = (unsigned long)(sr_list->data) & (PAGE_SIZE - 1);

	if (num_phys_pages > max_phys_pages) {
		TS_REPORT_STAGE(MOD_SRPTP,
				"physical pages (%d) > number allocated (%d)",
				max_phys_pages, num_phys_pages);
		kfree(phys_buffer_list);
		kfree(sr_list);
		goto build_indirect;
	}

	/* do fast memory registration */
	status = srptp_register_memory(s,
				       sr_list,
				       offset,
				       phys_buffer_list, num_phys_pages);

	kfree(phys_buffer_list);

	if (status) {
		
		kfree(sr_list);
		TS_REPORT_WARN(MOD_SRPTP, "srptp_register_memory failed");
		return (status);
		
	}

	ioq->sr_list = sr_list;
	ioq->sr_list_length = 1;

	srp_pkt->scatter_gather_list.length =
	    sizeof(srp_cmd_t) + sizeof(srp_remote_buf_t);
	srp_remote_buf = (void *)srp_pkt->data + sizeof(srp_cmd_t);
	memset(srp_remote_buf, 0x00, sizeof(srp_remote_buf_t));

	srp_remote_buf->r_data = cpu_to_be64(sr_list->r_addr);
	srp_remote_buf->r_size = cpu_to_be32(sr_list->size);
	srp_remote_buf->r_key = cpu_to_be32((int)sr_list->r_key);

	/* since we successfully build direct buffer, reset the srp_cmd to direct */
	if (cmnd->sc_data_direction == SCSI_DATA_READ) {
		srp_cmd->difmt = 1;
		srp_cmd->dicount = 1;
	} else if (cmnd->sc_data_direction == SCSI_DATA_WRITE) {
		srp_cmd->dofmt = 1;
		srp_cmd->docount = 1;
	}

	return (0);

      build_indirect:
	return srp_build_indirect_data_buffer_descriptor(srp_pkt,
							 st_buffer,
							 num_sg_elements,
							 &ioq->sr_list,
							 &ioq->sr_list_length);

}

const char *srp_host_info(struct Scsi_Host *SChost)
{
	return ("SRP HOST  Adapter: $Id$");
}

void srp_send_done(int pkt_index, srp_target_t * target)
{
	srp_pkt_t *send_pkt;

	send_pkt = srp_host_pkt_index_to_ptr(pkt_index, target);

	if (send_pkt == NULL) {
		TS_REPORT_FATAL(MOD_SRPTP, "Bad pkt_index %d", pkt_index);
		return;
	}

	srp_host_free_pkt(send_pkt);
}

int
srp_path_record_completion(tTS_IB_CLIENT_QUERY_TID tid,
			   int status,
			   struct ib_path_record *path_record,
			   int remaining, void *context)
{
	struct ib_path_record *redirected_path_record;
	srp_host_conn_t *s = (srp_host_conn_t *) context;
	srp_host_port_params_t *port = s->port;

	s->path_record_tid = 0;

	if ((status == 0) && (path_record != NULL)) {

		TS_REPORT_WARN(MOD_SRPTP,
			       "New path record for conn %p of target %d:%d:%d",
			       s, s->target->target_index, port->local_port,
			       port->hca->hca_index + 1);

		redirected_path_record = &s->path_record;

		memcpy(redirected_path_record,
		       path_record, sizeof(struct ib_path_record));

		memcpy(&s->target->ioc.iou_path_record,
		       path_record, sizeof(struct ib_path_record));

		redirected_path_record->slid = port->slid;

		if (tid != TS_IB_CLIENT_QUERY_TID_INVALID) {
			/* inidicates this did not come out of the cache,
			 * so go ahead and update the cache */
			srp_update_cache(path_record, port);
		}

		s->state = SRP_HOST_LOGIN_INPROGRESS;
		s->redirected = TRUE;
		srptp_connect(s,
			      redirected_path_record,
			      (__u8 *) s->login_buff, s->login_buff_len);
	} else {
		TS_REPORT_WARN(MOD_SRPTP,
			       "Path record err code %d for target %d "
			       "hca %d port %d",
			       status, s->target->target_index,
			       port->hca->hca_index + 1, port->local_port);

		s->state = SRP_HOST_GET_PATH_RECORD;

		s->path_record_retry_count++;

		if ((s->path_record_retry_count > SRP_HOST_PATH_RECORD_RETRIES)
		    || (status =
			srp_get_path_record(s, port,
					    srp_path_record_completion, s))) {

			TS_REPORT_WARN(MOD_SRPTP,
				       "Target %d, Path Record retries "
				       "exceeded %d",
				       s->target->target_index,
				       s->path_record_retry_count);

			remove_connection(s, TARGET_POTENTIAL_CONNECTION);
		}
	}

	return 0;
}

int srp_host_connect_done(srp_host_conn_t * s, int status)
{
	srp_host_port_params_t *port = s->port;
	srp_target_t *target = s->target;
	int err_code;

	switch (status) {

	case SRPTP_SUCCESS:

		if (s->state == SRP_HOST_LOGIN_INPROGRESS) {
			unsigned char opcode;

			opcode = s->login_resp_data[0];

			if (opcode == SRP_LOGIN_RESP) {

				s->state = SRP_UP;
				port->hca->num_active_connections++;
				driver_params.num_active_connections++;
				port->hca->num_pending_connections--;
				driver_params.num_pending_connections--;

				srp_host_login_resp(s, s->login_resp_data);

				srp_host_pre_post(target, s);
				target->state = TARGET_ACTIVE_CONNECTION;

				if (target->block_requests == TRUE) {
					target->block_requests = FALSE;
					target->no_connection_io = 0;
				}
				srp_flush_pending_to_active(target);

			} else if (opcode == SRP_LOGIN_RJT) {

				srp_host_login_rej(s->login_resp_data);

				remove_connection(s,
						  TARGET_POTENTIAL_CONNECTION);

				srp_host_close_conn(s);
			} else {
				TS_REPORT_WARN(MOD_SRPTP,
					       "Illegal SRP opcode 0x%x for "
					       "login response", opcode);
			}

		} else {
			TS_REPORT_WARN(MOD_SRPTP,
				       "connection %p is in wrong state %x", s,
				       s->state);

			remove_connection(s, TARGET_NO_CONNECTIONS);
		}

		break;

	case SRPTP_RETRY_REDIRECT_CONNECTION:
		memcpy(s->redirected_port_gid, s->login_resp_data, 16);
	case SRPTP_BAD_DGID_REJECT:
		if (status == SRPTP_BAD_DGID_REJECT)
			memset(s->redirected_port_gid, 0, 16);
		s->path_record_retry_count = 0;
		s->state = SRP_HOST_GET_PATH_RECORD;
		s->path_record_tid = 0;

		err_code = srp_get_path_record(s,
					       s->port,
					       srp_path_record_completion, s);

		if (err_code) {
			remove_connection(s, TARGET_POTENTIAL_CONNECTION);
		}
		break;

	case SRPTP_HARD_REJECT:
		TS_REPORT_WARN(MOD_SRPTP,
			       "SRP Target rejected for target %d, redirect %d",
			       target->target_index, s->redirected);

		target->hard_reject = TRUE;

		break;

	case SRPTP_RETRY_STALE_CONNECTION:
	default:
		if (++(s->retry_count) >= MAX_CONNECTION_RETRIES) {

			TS_REPORT_WARN(MOD_SRPTP,
				       "Conn retry count exceeded for "
				       "target %d on hca %d, port %d",
				       s->target->target_index,
				       port->hca->hca_index + 1,
				       port->local_port);

			remove_connection(s, TARGET_POTENTIAL_CONNECTION);

			srp_host_close_conn(s);
		} else {
			TS_REPORT_WARN(MOD_SRPTP,
				       "Conn problem target %d hca %d port %d, "
				       "retrying",
				       s->target->target_index,
				       port->hca->hca_index + 1,
				       port->local_port);

                        set_current_state(TASK_INTERRUPTIBLE);
                        schedule_timeout(s->retry_count * 5 * HZ);

			/*
			 * Re-try connection to the same path record
			 */
			s->state = SRP_HOST_LOGIN_INPROGRESS;

			srptp_connect(s,
				      &s->path_record,
				      (__u8 *) s->login_buff,
				      s->login_buff_len);
		}
		break;
	}

	return (0);
}

int srp_host_close_conn(srp_host_conn_t * s)
{
	srp_target_t *target = s->target;
	unsigned long cpu_flags;

	if (s->state == SRP_HOST_LOGOUT_INPROGRESS) {

		TS_REPORT_WARN(MOD_SRPTP, "Connection %p on target %d:%d:%d closed",
			       s, target->target_index, target->port->local_port,
			       target->port->hca->hca_index + 1);

		spin_lock_irqsave(&target->spin_lock, cpu_flags);

		target->conn_count--;
		list_del(&s->conn_list);

		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

		kfree(s);

		driver_params.num_connections--;

	} else if (s->state == SRP_HOST_LOGIN_INPROGRESS) {
		srp_host_port_params_t *port = s->port;

		TS_REPORT_WARN(MOD_SRPTP,
			       "Login failed for target %d through hca %d "
			       "port %d",
			       target->target_index, port->hca->hca_index + 1,
			       port->local_port);

		spin_lock_irqsave(&target->spin_lock, cpu_flags);

		port->hca->num_connections--;

		target->state = TARGET_POTENTIAL_CONNECTION;
		target->timeout = jiffies + TARGET_POTENTIAL_TIMEOUT;
		target->active_conn = NULL;

		target->conn_count--;
		list_del(&s->conn_list);

		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

		kfree(s);

		port->hca->num_pending_connections--;
		driver_params.num_pending_connections--;
		driver_params.num_connections--;

		srp_move_to_pending(target);

		initialize_connection(target);

	} else {
		TS_REPORT_WARN(MOD_SRPTP,
			       "Close conn for target %d in state %x",
			       target->target_index, s->state);
	}

	return (0);
}

/*
 * Callback from the completion handle to indicate we need
 * to kill a connection.  We maybe in interrupt context when
 * we get this so, set a flag and wakeup the thread
 */
void srp_host_completion_error(int pkt_index, srp_target_t * target)
{
	unsigned long cpu_flags;
	srp_pkt_t *pkt;
	srp_host_conn_t *s;

	pkt = srp_host_pkt_index_to_ptr(pkt_index, target);
	s = pkt->conn;

	if ((target->state == TARGET_ACTIVE_CONNECTION) &&
	    (s == target->active_conn)) {

		TS_REPORT_STAGE(MOD_SRPTP,
				"Completion error...start disconnect conn %p of "
				"target %d:%d:%d", 
			        s, target->target_index, s->port->local_port,
			        s->port->hca->hca_index + 1);

		spin_lock_irqsave(&target->spin_lock, cpu_flags);
		target->state = TARGET_POTENTIAL_CONNECTION;
		target->need_disconnect = TRUE;
		spin_unlock_irqrestore(&target->spin_lock, cpu_flags);

		srp_move_to_pending(target);
		srp_pending_to_scsi(target, DID_SOFT_ERROR);
	}

	srp_host_free_pkt(pkt);

	wake_up_interruptible(&dm_poll_thread_wait_queue);
	return;
}

/*
 * Callback in "TIMEWAIT" state. This is the first stage of the disconnect
 * done callbacks. If we did not initiate the disconnect, simply retry
 * the connection. If we initiated the disconnect, wait for the second
 * callback i.e srp_host_close_conn to continue processing.
 */
int srp_host_disconnect_done(srp_host_conn_t * s, int status)
{
	if (s->state != SRP_HOST_LOGOUT_INPROGRESS) {
		TS_REPORT_WARN(MOD_SRPTP,
			       "Target %d:%d:%d disconnected us, conn %p in state %d",
			       s->target->target_index, s->port->local_port,
			       s->port->hca->hca_index + 1, s, s->state);

		remove_connection(s, TARGET_POTENTIAL_CONNECTION);

		srp_move_to_pending(s->target);
	} else {
		TS_REPORT_WARN(MOD_SRPTP,
			       "Ignoring conn %p disconnect done for"
			       " target %d:%d:%d in state %d",
			       s, s->target->target_index,
			       s->port->local_port,
			       s->port->hca->hca_index + 1, s->state);
	}
	return (0);
}

static void srp_host_module_cleanup(void)
{
	unsigned long shutdown_timeout;

	if (scsi_unload_in_progress == TRUE) {
		TS_REPORT_WARN(MOD_SRPTP,
			       "Invoked module_cleanup more than once");
		return;
	} else
		TS_REPORT_WARN(MOD_SRPTP, "Invoked module_cleanup first time");

	/* First unregister, the scsi driver, set a flag to indicate
	 * to the abort code to complete aborts immediately */
	scsi_unload_in_progress = TRUE;

	srp_dm_unload();

	driver_params.dm_shutdown = 1;
	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(2 * HZ);

	tsKernelThreadStop(driver_params.thread);

	srp_host_disconnect_targets();

	/*
	 * Poll until all the connections are zero
	 * or the timeout has occurred
	 */
	shutdown_timeout = jiffies + SRP_SHUTDOWN_TIMEOUT;
	while ((driver_params.num_connections != 0) &&
	       (time_after(shutdown_timeout, jiffies))) {
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ);
	}

	if (driver_params.num_connections) {
		TS_REPORT_WARN(MOD_SRPTP, "Hanging connections %d",
			       driver_params.num_connections);
	}

	srp_host_cleanup_targets();

	srptp_cleanup_module();

	TS_REPORT_WARN(MOD_SRPTP, "SRP Driver unloaded");
}

/*
 *  Function defination
 */
module_init(srp_host_init);
module_exit(srp_host_module_cleanup);
