/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_CLEANUP_PROT_H
#define H_CLEANUP_PROT_H

#include "ts_ib_core.h"

#include <linux/spinlock.h>
#include <linux/wait.h>

typedef struct ib_cleanup_prot_st {
	spinlock_t              cleanup_lock;
	int                     in_cleanup;	 /* started module cleanup */
	int                     entry_exit_count; /* non-zero indicates 
	                                          callback in progress */
} ib_cleanup_prot_t;

static inline void ib_in_cleanup_init(ib_cleanup_prot_t *cl_prot)
{
	spin_lock_init(&cl_prot->cleanup_lock);
	cl_prot->entry_exit_count = 0;
	cl_prot->in_cleanup = 0;
}


/*  ib_set_in_cleanup_wait is called in module cleanup, to indicate
 *  to callbacks that cleanup has started, and to wait for all 
 *  in-progress callbacks to complete before continuing
 */
static inline void ib_set_in_cleanup_wait(ib_cleanup_prot_t *cl_prot)
{
	unsigned long flags;
	int ready = 0;

	spin_lock_irqsave(&cl_prot->cleanup_lock,flags);
	cl_prot->in_cleanup = 1;
	spin_unlock_irqrestore(&cl_prot->cleanup_lock,flags);
	/* wait until all in-progress callbacks have terminated.
	 * Note that each iteration involves giving up time slice.
	 */
	while (!ready) {
		spin_lock_irqsave(&cl_prot->cleanup_lock,flags);
		if (cl_prot->entry_exit_count == 0) {
			ready = 1;
		}
		spin_unlock_irqrestore(&cl_prot->cleanup_lock,flags);
		if (!ready) {schedule();}
	}
}

/* ib_check_in_cleanup_incr is called at the entry to each callback. If this
 * function returns 1, the callback should return immediately.  If not, it
 * is safe for the callback to continue.
 */
static inline int ib_check_in_cleanup_incr(ib_cleanup_prot_t *cl_prot)
{
	unsigned long flags;
	
	spin_lock_irqsave(&cl_prot->cleanup_lock,flags);
	if (cl_prot->in_cleanup == 1) {
		spin_unlock_irqrestore(&cl_prot->cleanup_lock,flags);
		return 1;
	}
	(cl_prot->entry_exit_count)++;
	spin_unlock_irqrestore(&cl_prot->cleanup_lock,flags);
	return 0;
}

/* ib_in_cleanup_decr should be called at the end of each protected callback,
 * to indicate to module-cleanup that the current callback has completed.
 */
static inline void ib_in_cleanup_decr(ib_cleanup_prot_t *cl_prot)
{
	unsigned long flags;
	
	spin_lock_irqsave(&cl_prot->cleanup_lock,flags);
	--(cl_prot->entry_exit_count);
	BUG_ON(cl_prot->entry_exit_count < 0);
	spin_unlock_irqrestore(&cl_prot->cleanup_lock,flags);
}

#endif /* H_CLEANUP_PROT_H */

/*
  Local Variables:
  c-file-style: "linux"
  indent-tabs-mode: t
  End:
*/
