/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: client_query.c 477 2004-07-16 16:40:34Z roland $
*/

#include "client_query.h"

#include "ts_ib_core.h"
#include "ts_ib_mad.h"
#include "ts_ib_client_query.h"
#include "ts_ib_rmpp_mad_types.h"

#include "ts_kernel_trace.h"
#include "ts_kernel_hash.h"
#include "ts_kernel_timer.h"

#include <linux/list.h>
#include <linux/slab.h>
#include <linux/random.h>
#include <linux/spinlock.h>
#include <linux/module.h>
#include "cleanup_prot.h"

enum {
	TS_IB_CLIENT_QUERY_HASH_BITS = 10,
};

#define  TS_IB_CLIENT_RMPP_RESPONSE_TIMEOUT   (1*HZ)

/*
  type definitions
*/

struct ib_client_query {
	u64 transaction_id;

	u8 r_method;
	int timeout_jiffies;

	union {
		tTS_IB_CLIENT_RESPONSE_FUNCTION function;
		tTS_IB_CLIENT_RMPP_RESPONSE_FUNCTION rmpp_function;
	} callback;
	int callback_running;
	struct list_head  throttle_list;
        struct ib_mad *packet;

	void *arg;
	tTS_KERNEL_TIMER_STRUCT timer;
	struct ib_client_rmpp_rcv *rmpp_rcv;
	tTS_LOCKED_HASH_ENTRY_STRUCT entry;
};

struct ib_client_rmpp_rcv {
	u8 *header;
	u32 header_length;
	u8 *data;
	u32 data_length;
	tTS_KERNEL_TIMER_STRUCT response_timer;
};

static ib_cleanup_prot_t cl_query_cleanup_protect;

static LIST_HEAD(throttle_queue);
static	spinlock_t        throttle_lock = SPIN_LOCK_UNLOCKED;

static int throttle_waitbuf_size = 0;
static atomic_t ib_query_num_outstanding = ATOMIC_INIT(0);
extern int ib_query_throttle_threshhold;
extern int ib_query_throttle_waitqueue_limit;

static int  ib_client_query_send_from_throttle(void);

static inline struct ib_client_query * ib_client_throttle_get(void)
{	
	struct ib_client_query *query = NULL;
	unsigned long flags;

	spin_lock_irqsave(&throttle_lock,flags);
	if (list_empty(&throttle_queue)) {
	      spin_unlock_irqrestore(&throttle_lock, flags);
	      return NULL;
	} else {
	      query = list_entry(throttle_queue.next, 
				 struct ib_client_query, throttle_list);
	      list_del(&query->throttle_list);
	      throttle_waitbuf_size--;
	      spin_unlock_irqrestore(&throttle_lock, flags);
	      TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
		         "get top entry. query = %p", query);
	      return(query);
	}
}

static inline void ib_client_throttle_cleanup(void)
{
	struct list_head *ptr;
	struct ib_client_query *iter_query;
	unsigned long flags;
	
	spin_lock_irqsave(&throttle_lock,flags);
	list_for_each(ptr, &throttle_queue) {
		iter_query = list_entry(ptr, struct ib_client_query, 
					throttle_list);
	        list_del(&iter_query->throttle_list);
		kfree(iter_query);
	}
	spin_unlock_irqrestore(&throttle_lock,flags);
}

static inline void ib_client_throttle_add(struct ib_client_query *query)
{
	unsigned long flags;
	
	spin_lock_irqsave(&throttle_lock,flags);
	list_add_tail(&query->throttle_list, &throttle_queue);
	throttle_waitbuf_size++;
	spin_unlock_irqrestore(&throttle_lock,flags);
}

/* If found, returns 1, else returns zero.  If delete_flag is non_zero, 
   deletes the found entry */
static inline int ib_client_throttle_find_delete(struct ib_client_query *query,
						  int delete_flag)
{
	int found = 0;
	struct list_head *ptr;
	struct ib_client_query *iter_query;
	unsigned long flags;

	spin_lock_irqsave(&throttle_lock,flags);
	list_for_each(ptr, &throttle_queue) {
		iter_query = list_entry(ptr, struct ib_client_query, 
					throttle_list);
		if (iter_query == query) {
			found = 1;
			if (delete_flag) {list_del(&query->throttle_list); }
			break;
		}
	}
	spin_unlock_irqrestore(&throttle_lock,flags);
	return found;
}

static inline struct ib_client_query * ib_client_throttle_find_delete_tid(
	tTS_IB_CLIENT_QUERY_TID tid, int delete_flag)
{
	int found = 0;
	struct list_head *ptr;
	struct ib_client_query *iter_query = NULL;
	unsigned long flags;

	spin_lock_irqsave(&throttle_lock,flags);
	list_for_each(ptr, &throttle_queue) {
		iter_query = list_entry(ptr, struct ib_client_query, throttle_list);
		if (iter_query->transaction_id == tid) {
			found = 1;
			if (delete_flag) 
				list_del(&iter_query->throttle_list); 
			break;
		}
	}
	spin_unlock_irqrestore(&throttle_lock,flags);
	if (found) {
		return iter_query;
	} else {
		return NULL;
	}
}

static spinlock_t alloc_tid_lock = SPIN_LOCK_UNLOCKED;

static u64 cur_transaction_id;

static tTS_HASH_TABLE query_hash;

static struct ib_client_query * ib_client_query_find(u64 transaction_id)
{
	tTS_LOCKED_HASH_ENTRY entry;

	/* check that query hash table has not been deleted */
	BUG_ON(!query_hash);

	entry = tsKernelLockedHashLookup(query_hash,
					 (u32) (transaction_id & 0xffffffff));
	return entry ? TS_KERNEL_HASH_ENTRY(entry, struct ib_client_query,
					    entry)
		: NULL;
}

/* no need for throttle involvement.  This is called only after query has been sent */
static inline void ib_client_query_put(struct ib_client_query *query)
{	
	tsKernelLockedHashRelease(&query->entry);
}

static void ib_client_query_delete(struct ib_client_query *query,
				   u16 is_delete_data)
{
	int from_throttle = 0;

	from_throttle = ib_client_throttle_find_delete(query, 1);
	TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,  "deleting query %p. from_throttle = %s",
		       query, (from_throttle ? "TRUE" : "FALSE"));
	
	if (!from_throttle) 
		tsKernelLockedHashRemove(query_hash, &query->entry);

	if (query->rmpp_rcv) {

		if (!from_throttle) 
			tsKernelTimerRemove(&query->rmpp_rcv->response_timer);

		if (is_delete_data && (query->rmpp_rcv->data)) {
			ib_rmpp_client_free(query->rmpp_rcv->data);
			query->rmpp_rcv->data = NULL;
			query->rmpp_rcv->data_length = 0;
		}

		if (query->rmpp_rcv->header) {
			ib_rmpp_client_free(query->rmpp_rcv->header);
			query->rmpp_rcv->header = NULL;
			query->rmpp_rcv->header_length = 0;
		}

		kfree(query->rmpp_rcv);
	}

	if (!from_throttle) {
		tsKernelTimerRemove(&query->timer);
		atomic_dec(&ib_query_num_outstanding);
		if ((ib_query_throttle_threshhold) &&
		    ((int) atomic_read(&ib_query_num_outstanding) < 
					      ib_query_throttle_threshhold)) {
			ib_client_query_send_from_throttle();
		}

	}

	kfree(query);
}

/*
  Free query (and drop lock) so that we don't deadlock if consumer
  calls back into our code.
*/
/* this can only be invoked once the query has been sent, so we
 * do not need to check if the query is still in the throttle queue. */
static void ib_client_query_callback(struct ib_client_query *query,
				     tTS_IB_CLIENT_RESPONSE_STATUS status,
				     struct ib_mad *packet)
{
	tTS_IB_CLIENT_QUERY_TID transaction_id = query->transaction_id;
	tTS_IB_CLIENT_RESPONSE_FUNCTION function = query->callback.function;
	void *arg = query->arg;

	tsKernelTimerRemove(&query->timer);

	if (query->callback_running) {
		ib_client_query_put(query);
		return;
	}

	query->callback_running = 1;
	ib_client_query_put(query);

	function(status, packet, arg);

	query = ib_client_query_find(transaction_id);
	if (query) {
		query->callback_running = 0;

		ib_client_query_delete(query, 0);
	}
}

static void ib_client_query_rmpp_callback(struct ib_client_query *query,
					  tTS_IB_CLIENT_RESPONSE_STATUS status,
					  u8 * header,
					  u8 * data,
					  u32 data_size,
					  u16 is_delete_data)
{
        tTS_IB_CLIENT_QUERY_TID              transaction_id = query->transaction_id;
        tTS_IB_CLIENT_RMPP_RESPONSE_FUNCTION function = query->callback.rmpp_function;
        void                                *arg      = query->arg;

        tsKernelTimerRemove(&query->timer);

        if (query->callback_running) {
		ib_client_query_put(query);
		return;
        }

        query->callback_running = 1;
        ib_client_query_put(query);

        function(status, header, data, data_size, arg);

	query = ib_client_query_find(transaction_id);
	if (query) {
		query->callback_running = 0;

		ib_client_query_delete(query, is_delete_data);
	}
}

static void ib_client_query_timeout(void *arg)
{
	u64 transaction_id = (unsigned long)arg;
	struct ib_client_query *query;
	if (ib_check_in_cleanup_incr(&cl_query_cleanup_protect)) 
		return;
	query = ib_client_query_find(transaction_id);

	if (query) {
		if (!query->rmpp_rcv) {
			ib_client_query_callback(query,
						 TS_IB_CLIENT_RESPONSE_TIMEOUT,
						 NULL);
		} else {
			ib_client_query_rmpp_callback(query,
						     TS_IB_CLIENT_RESPONSE_TIMEOUT,
						     NULL, NULL, 0, 1);
		}
	}
	ib_in_cleanup_decr(&cl_query_cleanup_protect);
}

static void ib_client_rmpp_response_timeout(void *arg)
{
	u64 transaction_id = (unsigned long)arg;
	struct ib_client_query *query;
	if (ib_check_in_cleanup_incr(&cl_query_cleanup_protect)) 
		return;

	query = ib_client_query_find(transaction_id);

	if (query) {
		if (!query->rmpp_rcv) {
			/* ERROR */
			TS_REPORT_WARN(MOD_KERNEL_IB,
				       "RMPP timeout for non-RMPP query");
			ib_client_query_put(query);
		} else {
			/*
			 * rmpp response timeout - different from transaction timeout
			 * Right now, treat response timeout and
			 * transaction timeout to be the same
			 */

			ib_client_query_put(query);
			ib_client_query_timeout(arg);
		}
	}
	ib_in_cleanup_decr(&cl_query_cleanup_protect);
}

static struct ib_client_query *
ib_client_query_alloc(struct ib_mad *packet, int timeout_jiffies,
		    tTS_IB_CLIENT_RESPONSE_FUNCTION function, void *arg)
{
	struct ib_client_query *query = kmalloc(sizeof *query, GFP_ATOMIC);

	if (!query)
		return NULL;

	memset(query, 0, sizeof *query);
	query->callback_running = 0;
	query->transaction_id = packet->transaction_id;
	query->r_method = packet->r_method;
	query->timeout_jiffies = timeout_jiffies;
	query->callback.function = function;
	query->arg = arg;
        query->rmpp_rcv = NULL;
	query->entry.key = (u32) (query->transaction_id & 0xffffffff);
	query->packet = packet;
	return query;
	
}
EXPORT_SYMBOL(ib_client_query);

static struct ib_client_query *
ib_client_rmpp_query_new(struct ib_mad *packet, int timeout_jiffies,
			 int header_length,
			 tTS_IB_CLIENT_RMPP_RESPONSE_FUNCTION function,
			 void *arg)
{
	struct ib_client_query *query = kmalloc(sizeof *query, GFP_ATOMIC);

	if (!query)
		goto mem_err_query;
	

	memset(query, 0, sizeof *query);
	TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
		 "ib_client_rmpp_query_new()\n");

	query->callback_running = 0;
	query->transaction_id = packet->transaction_id;
	query->r_method = packet->r_method;
	query->timeout_jiffies = timeout_jiffies;
	query->callback.rmpp_function = function;
	query->arg = arg;
	query->packet = packet;

	query->rmpp_rcv = kmalloc(sizeof *query->rmpp_rcv, GFP_ATOMIC);
	if (!query->rmpp_rcv) {
		goto mem_err_rmpp_rcv;
	}

	if (header_length > 0) {
		query->rmpp_rcv->header = kmalloc(header_length, GFP_ATOMIC);
		if (!query->rmpp_rcv->header) {
			goto mem_err_rmpp_rcv_header;
		}
	} else {
		query->rmpp_rcv->header = NULL;
	}
	query->rmpp_rcv->header_length = header_length;
	query->rmpp_rcv->data = NULL;
	query->rmpp_rcv->data_length = 0;

	query->entry.key = (u32) (query->transaction_id & 0xffffffff);

	return query;

mem_err_rmpp_rcv_header:
	kfree(query->rmpp_rcv);

mem_err_rmpp_rcv:
	kfree(query);

mem_err_query:
	TS_TRACE(MOD_KERNEL_IB, T_TERSE, TRACE_KERNEL_IB_GEN,
		 "Error: Failed to allocate query structure\n");
	return NULL;
}

static void ib_client_query_rmpp_rcv_mad(struct ib_client_query *query,
					 struct ib_mad *mad)
{
	struct ib_client_rmpp_mad *rmpp_mad = (struct ib_client_rmpp_mad *) mad;
	u16 status;
	u16 attribute_id;
	u32 attribute_modifier;
	u8 flag;
	int header_length = query->rmpp_rcv->header_length;
	int bytes_to_copy;

	/* Convert to host format */
	status = be16_to_cpu(mad->status);
	attribute_id = be16_to_cpu(mad->attribute_id);
	attribute_modifier = be32_to_cpu(mad->attribute_modifier);
	flag = rmpp_mad->resp_time__flags & 0x0F;

	TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
		 "ib_client_query_rmpp_rcv_mad(flag= 0x%x)\n", flag);

	/* Basic checking */
	if ((flag & TS_IB_CLIENT_RMPP_FLAG_ACTIVE) == 0)
		return;

	switch (rmpp_mad->type) {
	case TS_IB_CLIENT_RMPP_TYPE_DATA:
	{
		u32 segment_number =
			be32_to_cpu(rmpp_mad->specific.data.segment_number);
		u32 payload_length =
			be32_to_cpu(rmpp_mad->specific.data.payload_length);

		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "Data - segment_number= %d, payload_length= %d\n",
			 segment_number, payload_length);

		/* if first - allocate data */
		if (flag & TS_IB_CLIENT_RMPP_FLAG_DATA_FIRST) {
			if (query->rmpp_rcv->data) {
				/* WARNING */
			} else {
				TS_TRACE(MOD_KERNEL_IB, T_VERBOSE,
					 TRACE_KERNEL_IB_GEN,
					 "Allocate data on first segment\n");
				query->rmpp_rcv->data_length = payload_length;
				query->rmpp_rcv->data =
					kmalloc(payload_length, GFP_ATOMIC);
				if (!query->rmpp_rcv->data) {
					/* Failed to allocate memory.
					 * Wait for rexmit from SA.
					 * There may be memory available then.
					 */
					ib_client_query_put(query);
					return;
				}
			}

			if (payload_length >
			    (TS_IB_CLIENT_RMPP_PAYLOAD_LENGTH - header_length)) {
				bytes_to_copy =
					TS_IB_CLIENT_RMPP_PAYLOAD_LENGTH -
					header_length;
			} else {
				bytes_to_copy = payload_length - header_length;
			}
		} else if (flag & TS_IB_CLIENT_RMPP_FLAG_DATA_LAST) {
			bytes_to_copy = payload_length - header_length;
		} else {
			bytes_to_copy = TS_IB_CLIENT_RMPP_PAYLOAD_LENGTH -
				header_length;
		}

		/* stop response timer */
		tsKernelTimerRemove(&query->rmpp_rcv->response_timer);

		/* copy header */
		if (query->rmpp_rcv->header_length > 0) {
			memcpy(query->rmpp_rcv->header,
			       rmpp_mad->payload,
			       query->rmpp_rcv->header_length);
		}

		/* copy data */
		if (bytes_to_copy > 0) {
			int query_data_offset =
				(segment_number - 1) *
				(TS_IB_CLIENT_RMPP_PAYLOAD_LENGTH - header_length);

			memcpy(query->rmpp_rcv->data + query_data_offset,
			       rmpp_mad->payload + header_length,
			       bytes_to_copy);
		} else {
			/* Invalid payload len and header len combination */
			ib_client_query_rmpp_callback(query,
						     TS_IB_CLIENT_RESPONSE_ERROR,
						     NULL, NULL, 0, 1);
			return;
		}

		/* send back ack with sliding window of 1 */
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "Send back ack, dlid= %d\n", mad->slid);
		mad->dlid = mad->slid;
		mad->completion_func = NULL;
		mad->has_grh = 0;
		rmpp_mad->type = TS_IB_CLIENT_RMPP_TYPE_ACK;
		rmpp_mad->specific.ack.segment_number =
			cpu_to_be32(segment_number);
		rmpp_mad->specific.ack.new_window_last =
			cpu_to_be32(segment_number + 1);
		tsIbMadSend(mad);

		/* if last - call user supplied completion */
		if (flag & TS_IB_CLIENT_RMPP_FLAG_DATA_LAST) {
			TS_TRACE(MOD_KERNEL_IB, T_VERBOSE,
				 TRACE_KERNEL_IB_GEN,
				 "get last segment of transaction\n");

			ib_client_query_rmpp_callback(query,
						     TS_IB_CLIENT_RESPONSE_OK,
						     query->rmpp_rcv->header,
						     query->rmpp_rcv->data,
						     query->rmpp_rcv->data_length -
						     (segment_number *
						      query->rmpp_rcv->header_length),
						     0);
		} else {
			ib_client_query_put(query);
		}

		return;
	}
	case TS_IB_CLIENT_RMPP_TYPE_ACK:
		/* Only implement receiver - don't need to handle ack */
		ib_client_query_put(query);
		return;

	case TS_IB_CLIENT_RMPP_TYPE_STOP:
	case TS_IB_CLIENT_RMPP_TYPE_ABORT:
		/* call the callback with error code and delete query */
		ib_client_query_rmpp_callback(query,
					     TS_IB_CLIENT_RESPONSE_ERROR,
					     NULL, NULL, 0, 1);
		return;

	default:
		ib_client_query_put(query);
		return;
	}
}

static void ib_client_get_response(struct ib_mad *mad)
{
	struct ib_client_query *query;
	tTS_IB_CLIENT_RESPONSE_STATUS resp_status;

	if (ib_check_in_cleanup_incr(&cl_query_cleanup_protect)) 
		return;

	query = ib_client_query_find(mad->transaction_id);
	if (!query) {
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "packet received for unknown TID 0x%016" TS_U64_FMT
			 "x", mad->transaction_id);
		goto done;
	}

	if (query->r_method == mad->r_method) {
		ib_client_query_put(query);
		goto done;
	}


	if (mad->status) {
		resp_status = TS_IB_CLIENT_RESPONSE_ERROR;
	} else {
		resp_status = TS_IB_CLIENT_RESPONSE_OK;
	}

	if (!query->rmpp_rcv) {
		ib_client_query_callback(query, resp_status, mad);
	} else {
		ib_client_query_rmpp_rcv_mad(query, mad);
	}
done:
	ib_in_cleanup_decr(&cl_query_cleanup_protect);
	return;
}

static int  ib_client_query_send_from_throttle(void) 
{
	struct ib_client_query *query;
	struct ib_mad *tmp_packet;
	int rc = 0;
	/* save waitbuf size before getting query */
	int temp_size = throttle_waitbuf_size; 
	int do_queue_reduction = 0;

	query = ib_client_throttle_get();
	if (query == NULL) {
		/* nothing to send */
		return 0;
	}

	TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_FLOW_DATA, 
		 "sending from throttle. outs = %d, waitbuf_size=%d", 
		 (int) atomic_read(&ib_query_num_outstanding), 
		             throttle_waitbuf_size);

	/* if waitbuf is too big, adjust timeout to minimum value for
	 * this query, and only arm timers -- do not send the packet
	 */
	if (ib_query_throttle_waitqueue_limit && 
	    (ib_query_throttle_waitqueue_limit <= temp_size)) {
		do_queue_reduction = 1;
		query->timeout_jiffies = ((query->rmpp_rcv != NULL) ? 2 :1);
	}

	tsKernelLockedHashStore(query_hash, &query->entry);

	tsKernelTimerInit(&query->timer);
	if (query->timeout_jiffies > 0) {
		query->timer.run_time = jiffies + query->timeout_jiffies;
		query->timer.function = ib_client_query_timeout;
		query->timer.arg = (void *)(unsigned long)query->transaction_id;
		tsKernelTimerAdd(&query->timer);
	}

	if (query->rmpp_rcv != NULL) {
		tsKernelTimerInit(&query->rmpp_rcv->response_timer);
                query->rmpp_rcv->response_timer.run_time =
                        jiffies + (do_queue_reduction ? 1 : 
			           TS_IB_CLIENT_RMPP_RESPONSE_TIMEOUT);
		query->rmpp_rcv->response_timer.function =
			ib_client_rmpp_response_timeout;
		query->rmpp_rcv->response_timer.arg =
			(void *)(unsigned long)query->transaction_id;
		tsKernelTimerAdd(&query->rmpp_rcv->response_timer);
	}

	ib_client_query_put(query);

	if (0) {
		int i;

		TS_TRACE(MOD_KERNEL_IB, T_VERY_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "Sending query packet:");

		for (i = 0; i < 256; ++i) {
			if (i % 8 == 0) {
				printk(KERN_INFO "   ");
			}
			printk(" %02x", ((u8 *) query->packet)[i]);
			if ((i + 1) % 8 == 0) {
				printk("\n");
			}
		}
	}

	/* Send out MAD if not performing waitqueue reduction */
	tmp_packet = query->packet;
	rc = (do_queue_reduction ? 0 : tsIbMadSend(query->packet));
	/* increment the outstanding in all cases , because it is decremented 
	 * at timeout */
	atomic_inc(&ib_query_num_outstanding);

	/* query may be deleted before we get here if response arrives 
	 * quickly enough. Use stack variable to store the packet address 
	 * for freeing. No need to null in query struct, since packet field
	 * is not examined anywhere in response flow.
	 */
	kfree(tmp_packet);
	return rc;
}

int ib_client_query(struct ib_mad *packet,
		    int timeout_jiffies,
		    tTS_IB_CLIENT_RESPONSE_FUNCTION function, void *arg)
{
	struct ib_client_query *query;

	/* if cleanup has started, do not allow new queries */
	if (ib_check_in_cleanup_incr(&cl_query_cleanup_protect))
		return -ENOMEM;

	/* Allocate new query */
	query = ib_client_query_alloc(packet, timeout_jiffies, function, arg);

	if (!query) {
		ib_in_cleanup_decr(&cl_query_cleanup_protect);
		return -ENOMEM;
	}

	ib_client_throttle_add(query);
	/* invoke send  from throttle if:
	 *  a. threshold is disabled    or
	 *  b. number of outstanding queries is less than threshold  or
	 *  c. number of queries in wait queue is .GE. wait queue watermark.
	 *
	 *  Option C is a safety valve so that the wait queue does not grow
	 *  without limit. When the option C state is present, 
	 *  send_from_throttle will take the next query from the throttle
	 *  wait buffer and invoke its timeout mechanism. 
	 */
	if ((!ib_query_throttle_threshhold) ||
	     ( (int)atomic_read(&ib_query_num_outstanding) < 
	       ib_query_throttle_threshhold) ||
	     ( ib_query_throttle_waitqueue_limit && 
	       ( ib_query_throttle_waitqueue_limit <= 
	         throttle_waitbuf_size))) {
		ib_client_query_send_from_throttle();
	}

	ib_in_cleanup_decr(&cl_query_cleanup_protect);
	return 0;
}

int ib_client_query_cancel(tTS_IB_CLIENT_QUERY_TID transaction_id)
{
	struct ib_client_query *query;
	
	if ((query = ib_client_throttle_find_delete_tid(transaction_id,1)) != NULL) {
		if (!query->rmpp_rcv) {
			query->callback.function(TS_IB_CLIENT_RESPONSE_CANCEL, NULL,
						 query->arg);
		} else {
			query->callback.rmpp_function(TS_IB_CLIENT_RESPONSE_CANCEL,
						      NULL, NULL, 0, query->arg);
		}
		
		if (query->rmpp_rcv) {
			if (query->rmpp_rcv->data) {
				ib_rmpp_client_free(query->rmpp_rcv->data);
				query->rmpp_rcv->data = NULL;
				query->rmpp_rcv->data_length = 0;
			}

			if (query->rmpp_rcv->header) {
				ib_rmpp_client_free(query->rmpp_rcv->header);
				query->rmpp_rcv->header = NULL;
				query->rmpp_rcv->header_length = 0;
			}

			kfree(query->rmpp_rcv);
		}
		kfree(query);
		return 0;
	}

	query = ib_client_query_find(transaction_id);

	while (query && query->callback_running) {
		ib_client_query_put(query);
		set_current_state(TASK_RUNNING);
		schedule();
		query = ib_client_query_find(transaction_id);
	}

	if (!query) {
		return -ENOENT;
	}

	if (!query->rmpp_rcv) {
		query->callback.function(TS_IB_CLIENT_RESPONSE_CANCEL, NULL,
					 query->arg);
	} else {
		query->callback.rmpp_function(TS_IB_CLIENT_RESPONSE_CANCEL,
					      NULL, NULL, 0, query->arg);
	}

	ib_client_query_delete(query, 1);

	return 0;
}
EXPORT_SYMBOL(ib_client_query_cancel);

void ib_client_mad_handler(struct ib_mad *mad, void *arg)
{
	TS_TRACE(MOD_KERNEL_IB, T_VERY_VERBOSE, TRACE_KERNEL_IB_GEN,
		 "query packet received, TID 0x%016" TS_U64_FMT "x",
		 mad->transaction_id);
	if (0) {
		int i;

		/* Dump packet */

		for (i = 0; i < 256; ++i) {
			if (i % 8 == 0) {
				printk(KERN_INFO "   ");
			}
			printk(" %02x", ((u8 *) mad)[i]);
			if ((i + 1) % 8 == 0) {
				printk("\n");
			}
		}
	}
	switch (mad->r_method) {
	case IB_MGMT_METHOD_REPORT:
		{
			/* Send back REPORT RESPONSE */
			struct ib_mad *report_resp;
			report_resp = kmalloc(sizeof *report_resp,GFP_ATOMIC);
			if (!report_resp) {
				/* no report response is generated, malloc failure */
				TS_REPORT_WARN(MOD_KERNEL_IB, "kmalloc failed");
			} else {

				memcpy(report_resp, mad, sizeof(*report_resp));
				report_resp->r_method =
					IB_MGMT_METHOD_REPORT_RESPONSE;
				report_resp->slid = mad->dlid;
				report_resp->dlid = mad->slid;
				report_resp->sqpn = mad->dqpn;
				report_resp->dqpn = mad->sqpn;
				report_resp->completion_func = NULL;
				report_resp->has_grh = 0;
	
				tsIbMadSend(report_resp);
				kfree(report_resp);
			}
			/* Continue to call the async handler */
		}
	case IB_MGMT_METHOD_TRAP:
	case IB_MGMT_METHOD_TRAP_REPRESS:
		{
			ib_mad_dispatch_func dispatch =
			    ib_client_async_mad_handler_get(mad->mgmt_class);

			if (dispatch)
				dispatch(mad,
					 ib_client_async_mad_handler_arg_get(mad->
									mgmt_class));
			return;
		}

	default:
		ib_client_get_response(mad);
		return;
	}
}

int ib_client_query_init(void)
{
	u64 invalid_tid = TS_IB_CLIENT_QUERY_TID_INVALID & 0xFFFFFFFF;

	if (tsKernelHashTableCreate(TS_IB_CLIENT_QUERY_HASH_BITS, &query_hash))
		return -ENOMEM;

	ib_in_cleanup_init(&cl_query_cleanup_protect);
	get_random_bytes(&cur_transaction_id, sizeof cur_transaction_id);

	cur_transaction_id &= 0xFFFFFFFF;
	if (cur_transaction_id == invalid_tid)
		cur_transaction_id = 0;

	return 0;
}

void ib_client_query_cleanup(void)
{
	/* cancel all queries and delete timers */

	tTS_HASH_TABLE temp_hash;
	tTS_HASH_HEAD  bucket;
	struct ib_client_query *query;
	int i;
        tTS_LOCKED_HASH_ENTRY entry, entry_next;


	ib_client_throttle_cleanup();

	/* don't allow any queries from below */
	ib_set_in_cleanup_wait(&cl_query_cleanup_protect);
	/* free any stranded queries */
	for (i = 0; i <= query_hash->hash_mask; i++) {
		bucket = &query_hash->bucket[i];

		TS_KERNEL_HASH_FOR_EACH_ENTRY_SAFE(entry, entry_next, bucket, node) {
			query = TS_KERNEL_HASH_ENTRY(entry, struct ib_client_query,
					    entry);
			if (query->rmpp_rcv) {
				if (query->rmpp_rcv->data) {
					ib_rmpp_client_free(query->rmpp_rcv->data);
				}

				if (query->rmpp_rcv->header) {
					ib_rmpp_client_free(query->rmpp_rcv->header);
				}

				kfree(query->rmpp_rcv);
			}

			tsKernelTimerRemove(&query->timer);
			kfree(query);
		}
	}

	temp_hash = query_hash;
	query_hash = NULL;
	tsKernelHashTableDestroy(temp_hash);
}

tTS_IB_CLIENT_QUERY_TID ib_client_alloc_tid(void)
{
	tTS_IB_CLIENT_QUERY_TID tid;
	u64 invalid_tid = TS_IB_CLIENT_QUERY_TID_INVALID & 0xFFFFFFFF;
	unsigned long flags;

	spin_lock_irqsave(&alloc_tid_lock, flags);
	tid = cur_transaction_id++;
	cur_transaction_id &= 0xFFFFFFFF;
	if (cur_transaction_id == invalid_tid)
		cur_transaction_id = 0;
	spin_unlock_irqrestore(&alloc_tid_lock, flags);

	return tid;
}
EXPORT_SYMBOL(ib_client_alloc_tid);

int ib_rmpp_client_query(struct ib_mad *packet,
			 int timeout_jiffies,
			 int header_length,
			 tTS_IB_CLIENT_RMPP_RESPONSE_FUNCTION function,
			 void *arg)
{
	struct ib_client_query *query;

	TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
		 "tsIbRmppClientQuery()\n");

	/* if cleanup has started, do not allow new queries */
	if (ib_check_in_cleanup_incr(&cl_query_cleanup_protect))
		return -ENOMEM;

	query = ib_client_rmpp_query_new(packet, timeout_jiffies,
					header_length, function, arg);

	if (!query) {
		ib_in_cleanup_decr(&cl_query_cleanup_protect);
		return -ENOMEM;
	}

	ib_client_throttle_add(query);
	/* invoke send  from throttle if:
	 *  a. threshold is disabled    or
	 *  b. number of outstanding queries is less than threshold  or
	 *  c. number of queries in wait queue is .GE. wait queue watermark.
	 *
	 *  Option C is a safety valve so that the wait queue does not grow
	 *  without limit. When the option C state is present, 
	 *  send_from_throttle will take the next query from the throttle
	 *  wait buffer and invoke its timeout mechanism. 
	 */
	if (  ( !ib_query_throttle_threshhold) ||
	      ( (int)atomic_read(&ib_query_num_outstanding) < 
	        ib_query_throttle_threshhold) ||
	      ( ib_query_throttle_waitqueue_limit && 
	        ( ib_query_throttle_waitqueue_limit <= 
	          throttle_waitbuf_size))) {
		ib_client_query_send_from_throttle();
	}

	ib_in_cleanup_decr(&cl_query_cleanup_protect);
	return 0;
}
EXPORT_SYMBOL(ib_rmpp_client_query);

int ib_rmpp_client_free(u8 *data)
{
	kfree(data);
	return 0;
}
EXPORT_SYMBOL(ib_rmpp_client_free);

void ib_client_query_set_in_cleanup_wait(void)
{
	ib_set_in_cleanup_wait(&cl_query_cleanup_protect);
	return;
}
EXPORT_SYMBOL(ib_client_query_set_in_cleanup_wait);

