/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: core_cq.c 475 2004-07-16 16:29:15Z roland $
*/

#include "core_priv.h"

#include "ts_kernel_trace.h"
#include "ts_kernel_services.h"

#include <linux/version.h>
#include <linux/module.h>

#include <linux/errno.h>
#include <linux/slab.h>

struct ib_completion_event {
	struct ib_cq    *cq;
	struct list_head list;
};

int __ib_cq_create(struct ib_device          *device,
		   int                       *entries,
		   struct ib_cq_callback     *callback,
		   void                      *device_specific,
		   struct ib_cq             **cq_handle)
{
	struct ib_cq *cq;
	int           ret;

	IB_CHECK_MAGIC(device, DEVICE);

	if (!device->cq_create)
		return -ENOSYS;

	cq = kmalloc(sizeof *cq, GFP_KERNEL);
	if (!cq) {
		TS_REPORT_WARN(MOD_KERNEL_IB,"%s: cq kmalloc failure",__func__);
		return -ENOMEM;
	}

	cq->device   = device;
	cq->callback = *callback;
	INIT_LIST_HEAD(&cq->async_handler_list);
	spin_lock_init(&cq->async_handler_lock);
	ib_async_obj_init(&cq->async_obj, cq);
	/* initialize completion event reference counter */
	atomic_set(&cq->num_outstanding_comp_events,0);

	ret = device->cq_create(device, entries, device_specific, cq);
	if (!ret) {
		IB_SET_MAGIC(cq, CQ);
		*cq_handle = cq;

		if (cq->callback.policy == IB_CQ_PROVIDER_REARM) {
			ret = ib_cq_request_notification(cq, 0);
			if (ret) {
				__ib_cq_destroy(cq);
				IB_CLEAR_MAGIC(cq);
				kfree(cq);
			}
		}
	} else {
		kfree(cq);
	}

	return ret;
}
EXPORT_SYMBOL(__ib_cq_create);

int __ib_cq_destroy(struct ib_cq *cq)
{
	int           ret;
	unsigned long flags;
	int           pending = 0;

	IB_CHECK_MAGIC(cq, CQ);
	if (!cq->device->cq_destroy)
		return -ENOSYS;

	if (!list_empty(&cq->async_handler_list))
		return -EBUSY;

	spin_lock_irqsave(&cq->async_obj.lock, flags);
	cq->async_obj.dead = 1;
	spin_unlock_irqrestore(&cq->async_obj.lock, flags);

	/* disable completion event callback for this cq */
	ret = cq->device->cq_stop_comp_ev(cq);

	BUG_ON(ret);

	/* wait until all events already in completion thread are handled */
	/* (recheck approximately every 7 milliseconds) */
	while(atomic_read(&cq->num_outstanding_comp_events)) {
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout((HZ)>>7);
	}

	ret = cq->device->cq_destroy(cq);
	if (!ret) {
		IB_CLEAR_MAGIC(cq);
		spin_lock_irqsave(&cq->async_obj.lock, flags);
		if (!cq->async_obj.pending)
			pending = 1;
		spin_unlock_irqrestore(&cq->async_obj.lock, flags);
		if (pending == 1)
			kfree(cq);        
    }

	return ret;
}
EXPORT_SYMBOL(__ib_cq_destroy);

int ib_cq_create(struct ib_device          *device,
                 int                       *entries,
                 struct ib_cq_callback     *callback,
                 void                      *device_specific,
                 struct ib_cq             **cq_handle)
{
	IB_CHECK_MAGIC(device, DEVICE);

	if (!try_module_get(device->owner))
		return -ENODEV;

	return __ib_cq_create(device, entries,
			      callback, device_specific,
			      cq_handle);
}
EXPORT_SYMBOL(ib_cq_create);

int ib_cq_destroy(struct ib_cq *cq)
{
	struct module *owner;
	int            ret;

	IB_CHECK_MAGIC(cq, CQ);

	owner = cq->device->owner;

	ret = __ib_cq_destroy(cq);
	if (!ret)
		module_put(owner);

	return ret;
}
EXPORT_SYMBOL(ib_cq_destroy);

int ib_cq_resize(struct ib_cq    *cq,
                 int             *entries)
{
	IB_CHECK_MAGIC(cq, CQ);
	return cq->device->cq_resize ? cq->device->cq_resize(cq, entries) : -ENOSYS;
}
EXPORT_SYMBOL(ib_cq_resize);

int ib_cq_request_notification(struct ib_cq *cq,
			       int solicited)
{
	IB_CHECK_MAGIC(cq, CQ);
	return cq->device->cq_arm ? cq->device->cq_arm(cq, solicited) : -ENOSYS;
}
EXPORT_SYMBOL(ib_cq_request_notification);

static void ib_cq_drain(struct ib_cq *cq)
{
	/* We make copies of these for efficiency; we know they won't change
	   but the compiler can't know that */
	ib_cq_poll_func                   cq_poll = cq->device->cq_poll;
	ib_cq_entry_callback_func         func    = cq->callback.function.entry;
	void                             *arg     = cq->callback.arg;
	struct ib_cq_entry                entry;

	if (!cq_poll) {
		return;
	}

	/*
	  We re-arm and then drain the CQ.  There is no way for us to miss
	  completions because a new completion will generate a new event and
	  send us back through this loop when the event is dispatched.

	  We ask for all completions (unsolicited or solicited) to generate
	  events, since this interface is for simple consumers.
	*/
	if (cq->device->cq_arm(cq, 0)) {
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "CQ rearm failed for device %s",
			       cq->device->name);
	}

	while (cq_poll(cq, 1, &entry) == 1) {
		func(cq, &entry, arg);
	}
}

void ib_completion_event_dispatch(struct ib_cq *cq)
{
	struct ib_device_private   *priv   = cq->device->core;
	struct ib_completion_event *event;

	if (!IB_TEST_MAGIC(cq, CQ)) {
		TS_REPORT_WARN(MOD_KERNEL_IB, "Bad magic 0x%lx at %p for CQ",
			       IB_GET_MAGIC(cq), cq);
		return;
	}

	if (cq->callback.context == IB_CQ_CALLBACK_PROCESS) {
		event = kmalloc(sizeof *event, GFP_ATOMIC);
		if (event) {
			event->cq = cq;
			atomic_inc(&cq->num_outstanding_comp_events);
			tsKernelQueueThreadAdd(priv->completion_thread, &event->list);
		} else {
			TS_REPORT_WARN(MOD_KERNEL_IB,
				       "Failed to allocate completion record, dropping event");
		}
	} else {
		if (cq->callback.policy == IB_CQ_PROVIDER_REARM) {
			ib_cq_drain(cq);
		} else {
			cq->callback.function.event(cq, cq->callback.arg);
		}
	}
}
EXPORT_SYMBOL(ib_completion_event_dispatch);

void ib_completion_thread(struct list_head *entry,
                          void *device_ptr)
{
	struct ib_completion_event *event;
	event = list_entry(entry, struct ib_completion_event, list);

	if (event->cq->callback.policy == IB_CQ_PROVIDER_REARM) {
		ib_cq_drain(event->cq);
	} else {
		event->cq->callback.function.event(event->cq, event->cq->callback.arg);
	}

	atomic_dec(&event->cq->num_outstanding_comp_events);
	kfree(event);
}

/*
  Local Variables:
  c-file-style: "linux"
  indent-tabs-mode: t
  End:
*/
