/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: core_fmr.c 477 2004-07-16 16:40:34Z roland $
*/

#include "core_priv.h"

#include "ts_kernel_trace.h"
#include "ts_kernel_services.h"
#include "ts_kernel_thread.h"
#include "ts_kernel_hash.h"

#include <linux/version.h>
#include <linux/module.h>

#include <linux/errno.h>
#include <linux/spinlock.h>
#include <linux/slab.h>

#if defined(IB_FMR_NODEBUG)
#  define IB_COMPILE_FMR_DEBUGGING_CODE 0
#else
#  define IB_COMPILE_FMR_DEBUGGING_CODE 1
#endif

enum {
	TS_IB_FMR_MAX_REMAPS = 32,

	TS_IB_FMR_HASH_BITS  = 8,
	TS_IB_FMR_HASH_SIZE  = 1 << TS_IB_FMR_HASH_BITS,
	TS_IB_FMR_HASH_MASK  = TS_IB_FMR_HASH_SIZE - 1
};

/*
  If an FMR is not in use, then the list member will point to either
  its pool's free_list (if the FMR can be mapped again; that is,
  remap_count < TS_IB_FMR_MAX_REMAPS) or its pool's dirty_list (if the
  FMR needs to be unmapped before being remapped).  In either of these
  cases it is a bug if the ref_count is not 0.  In other words, if
  ref_count is > 0, then the list member must not be linked into
  either free_list or dirty_list.

  The cache_node member is used to link the FMR into a cache bucket
  (if caching is enabled).  This is independent of the reference count
  of the FMR.  When a valid FMR is released, its ref_count is
  decremented, and if ref_count reaches 0, the FMR is placed in either
  free_list or dirty_list as appropriate.  However, it is not removed
  from the cache and may be "revived" if a call to
  ib_fmr_register_physical() occurs before the FMR is remapped.  In
  this case we just increment the ref_count and remove the FMR from
  free_list/dirty_list.

  Before we remap an FMR from free_list, we remove it from the cache
  (to prevent another user from obtaining a stale FMR).  When an FMR
  is released, we add it to the tail of the free list, so that our
  cache eviction policy is "least recently used."

  All manipulation of ref_count, list and cache_node is protected by
  pool_lock to maintain consistency.
*/

struct ib_fmr_pool {
	IB_DECLARE_MAGIC
	struct ib_device         *device;

	spinlock_t                pool_lock;

	int                       pool_size;
	int                       max_pages;
	int                       dirty_watermark;
	int                       dirty_len;
	struct list_head          free_list;
	struct list_head          dirty_list;
	tTS_HASH_HEAD             cache_bucket;

	tTS_KERNEL_THREAD         thread;

	ib_fmr_flush_func         flush_function;
	void                     *flush_arg;

	atomic_t                  req_ser;
	atomic_t                  flush_ser;

	wait_queue_head_t         thread_wait;
	wait_queue_head_t         force_wait;
};

static inline u32 ib_fmr_hash(u64 first_page)
{
	return tsKernelHashFunction((u32) (first_page >> PAGE_SHIFT),
				    TS_IB_FMR_HASH_MASK);
}

/* Caller must hold pool_lock */
static inline struct ib_fmr *ib_fmr_cache_lookup(struct ib_fmr_pool *pool,
						 u64 *page_list,
						 int  page_list_len,
						 u64  io_virtual_address,
						 u64  iova_offset)
{
	tTS_HASH_HEAD  bucket;
	struct ib_fmr *fmr;

	if (!pool->cache_bucket) {
		return NULL;
	}

	bucket = &pool->cache_bucket[ib_fmr_hash(*page_list)];

	TS_KERNEL_HASH_FOR_EACH_ENTRY(fmr, bucket, cache_node)
		if (io_virtual_address == fmr->io_virtual_address &&
		    iova_offset        == fmr->iova_offset        &&
		    page_list_len      == fmr->page_list_len      &&
		    !memcmp(page_list, fmr->page_list, page_list_len * sizeof *page_list))
			return fmr;

	return NULL;
}

/* Caller must hold pool_lock */
static inline void ib_fmr_cache_store(struct ib_fmr_pool *pool,
                                      struct ib_fmr      *fmr)
{
	tsKernelHashNodeAdd(&fmr->cache_node,
			    &pool->cache_bucket[ib_fmr_hash(fmr->page_list[0])]);
}

/* Caller must hold pool_lock */
static inline void ib_fmr_cache_remove(struct ib_fmr *fmr)
{
	if (!tsKernelHashNodeUnhashed(&fmr->cache_node))
		tsKernelHashNodeRemove(&fmr->cache_node);
}

static void ib_fmr_batch_release(struct ib_fmr_pool *pool)
{
	int               ret;
	struct list_head *ptr;
	struct ib_fmr    *fmr;
	LIST_HEAD(unmap_list);

	spin_lock_irq(&pool->pool_lock);

	list_for_each(ptr, &pool->dirty_list) {
		fmr = list_entry(ptr, struct ib_fmr, list);

		ib_fmr_cache_remove(fmr);
		fmr->remap_count = 0;

		if (IB_COMPILE_FMR_DEBUGGING_CODE) {
			if (fmr->ref_count !=0) {
				TS_REPORT_WARN(MOD_KERNEL_IB,
					       "Unmapping FMR 0x%08x with ref count %d",
					       fmr, fmr->ref_count);
			}
		}
	}

	list_splice(&pool->dirty_list, &unmap_list);
	INIT_LIST_HEAD(&pool->dirty_list);
	pool->dirty_len = 0;

	spin_unlock_irq(&pool->pool_lock);

	if (list_empty(&unmap_list)) {
		return;
	}

	ret = pool->device->fmr_unmap(pool->device, &unmap_list);
	if (ret) {
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "fmr_unmap for %s returns %d",
			       pool->device->name, ret);
	}

	spin_lock_irq(&pool->pool_lock);
	list_splice(&unmap_list, &pool->free_list);
	spin_unlock_irq(&pool->pool_lock);
}

static void ib_fmr_cleanup_thread(void *pool_ptr)
{
	struct ib_fmr_pool *pool = pool_ptr;
	int ret;

	while (!signal_pending(current)) {
		ret = wait_event_interruptible(pool->thread_wait,
					       (pool->dirty_len >=
						pool->dirty_watermark) ||
					       (atomic_read(&pool->flush_ser) -
						atomic_read(&pool->req_ser) < 0));

		TS_TRACE(MOD_KERNEL_IB, T_VERY_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "cleanup thread woken up, dirty len = %d",
			 pool->dirty_len);

		if (ret)
			break;

		ib_fmr_batch_release(pool);

		atomic_inc(&pool->flush_ser);
		wake_up_interruptible(&pool->force_wait);

		if (pool->flush_function)
			pool->flush_function(pool, pool->flush_arg);
	}

	TS_REPORT_CLEANUP(MOD_KERNEL_IB, "FMR cleanup thread exiting");
}

int ib_fmr_pool_create(struct ib_pd             *pd,
		       struct ib_fmr_pool_param *params,
		       struct ib_fmr_pool      **pool_handle)
{
	struct ib_device   *device;
	struct ib_fmr_pool *pool;
	int i;
	int ret;

	IB_CHECK_MAGIC(pd, PD);

	if (!params) {
		return -EINVAL;
	}

	device = pd->device;
	if (!device->fmr_create  ||
	    !device->fmr_destroy ||
	    !device->fmr_map     ||
	    !device->fmr_unmap) {
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "Device %s does not support fast memory regions",
			       device->name);
		return -ENOSYS;
	}

	pool = kmalloc(sizeof *pool, GFP_KERNEL);
	if (!pool) {
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "couldn't allocate pool struct");
		return -ENOMEM;
	}

	pool->cache_bucket   = NULL;

	pool->flush_function = params->flush_function;
	pool->flush_arg      = params->flush_arg;

	INIT_LIST_HEAD(&pool->free_list);
	INIT_LIST_HEAD(&pool->dirty_list);

	if (params->cache) {
		pool->cache_bucket =
			kmalloc(TS_IB_FMR_HASH_SIZE * sizeof *pool->cache_bucket, GFP_KERNEL);
		if (!pool->cache_bucket) {
			TS_REPORT_WARN(MOD_KERNEL_IB, "Failed to allocate cache in pool");
			ret = -ENOMEM;
			goto out_free_pool;
		}

		for (i = 0; i < TS_IB_FMR_HASH_SIZE; ++i) {
			tsKernelHashHeadInit(&pool->cache_bucket[i]);
		}
	}

	pool->device          = device;
	pool->pool_size       = 0;
	pool->max_pages       = params->max_pages_per_fmr;
	pool->dirty_watermark = params->dirty_watermark;
	pool->dirty_len       = 0;
	spin_lock_init(&pool->pool_lock);
	atomic_set(&pool->req_ser,   0);
	atomic_set(&pool->flush_ser, 0);
	init_waitqueue_head(&pool->thread_wait);
	init_waitqueue_head(&pool->force_wait);

	ret = tsKernelThreadStart("ts_fmr",
				  ib_fmr_cleanup_thread,
				  pool,
				  &pool->thread);

	if (ret) {
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "couldn't start cleanup thread");
		goto out_free_pool;
	}

	{
		struct ib_fmr *fmr;

		for (i = 0; i < params->pool_size; ++i) {
			fmr = kmalloc(sizeof *fmr + params->max_pages_per_fmr * sizeof (u64),
				      GFP_KERNEL);
			if (!fmr) {
				TS_REPORT_WARN(MOD_KERNEL_IB,
					       "failed to allocate fmr struct for FMR %d", i);
				goto out_fail;
			}

			fmr->device           = device;
			fmr->pool             = pool;
			fmr->remap_count      = 0;
			fmr->ref_count        = 0;
			fmr->cache_node.pprev = NULL;

			if (device->fmr_create(pd,
					       params->access,
					       params->max_pages_per_fmr,
					       TS_IB_FMR_MAX_REMAPS,
					       fmr)) {
				TS_REPORT_WARN(MOD_KERNEL_IB,
					       "fmr_create failed for FMR %d", i);
				kfree(fmr);
				goto out_fail;
			}

			IB_SET_MAGIC(fmr, FMR);
			list_add_tail(&fmr->list, &pool->free_list);
			++pool->pool_size;
		}
	}

	IB_SET_MAGIC(pool, FMR_POOL);
	*pool_handle = pool;
	return 0;

 out_free_pool:
	kfree(pool->cache_bucket);
	kfree(pool);

	return ret;

 out_fail:
	IB_SET_MAGIC(pool, FMR_POOL);
	ib_fmr_pool_destroy(pool);
	*pool_handle = NULL;

	return -ENOMEM;
}
EXPORT_SYMBOL(ib_fmr_pool_create);

int ib_fmr_pool_destroy(struct ib_fmr_pool *pool)
{
	struct list_head   *ptr;
	struct list_head   *tmp;
	struct ib_fmr      *fmr;
	int                 i;

	IB_CHECK_MAGIC(pool, FMR_POOL);

	i = tsKernelThreadStop(pool->thread);
	if (i) {
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "tsKernelThreadStop for fmr pool thread failed (%d)",i);
	}
	ib_fmr_batch_release(pool);

	i = 0;
	list_for_each_safe(ptr, tmp, &pool->free_list) {
		fmr = list_entry(ptr, struct ib_fmr, list);
		pool->device->fmr_destroy(fmr);

		list_del(ptr);
		kfree(fmr);
		++i;
	}

	if (i < pool->pool_size) {
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "pool still has %d regions registered",
			       pool->pool_size - i);
	}

	kfree(pool->cache_bucket);
	kfree(pool);

	return 0;
}
EXPORT_SYMBOL(ib_fmr_pool_destroy);

int ib_fmr_pool_force_flush(struct ib_fmr_pool *pool)
{
	int serial;

	atomic_inc(&pool->req_ser);
	/* It's OK if someone else bumps req_ser again here -- we'll
	   just wait a little longer. */
	serial = atomic_read(&pool->req_ser);

	wake_up_interruptible(&pool->thread_wait);

	if (wait_event_interruptible(pool->force_wait,
				     atomic_read(&pool->flush_ser) -
				     atomic_read(&pool->req_ser) >= 0)) {
		return -EINTR;
	}

	return 0;
}
EXPORT_SYMBOL(ib_fmr_pool_force_flush);

int ib_fmr_register_physical(struct ib_fmr_pool *pool_handle,
			     u64                *page_list,
			     int                 list_len,
			     u64                *io_virtual_address,
			     u64                 iova_offset,
			     struct ib_fmr     **fmr_handle,
			     u32                *lkey,
			     u32                *rkey)
{
	struct ib_fmr_pool *pool = pool_handle;
	struct ib_fmr      *fmr;
	unsigned long       flags;
	int                 result;

	IB_CHECK_MAGIC(pool, FMR_POOL);

	if (list_len < 1 || list_len > pool->max_pages) {
		return -EINVAL;
	}

	spin_lock_irqsave(&pool->pool_lock, flags);
	fmr = ib_fmr_cache_lookup(pool,
				  page_list,
				  list_len,
				  *io_virtual_address,
				  iova_offset);
	if (fmr) {
		/* found in cache */
		++fmr->ref_count;
		if (fmr->ref_count == 1) {
			list_del(&fmr->list);
		}

		spin_unlock_irqrestore(&pool->pool_lock, flags);

		*lkey       = fmr->lkey;
		*rkey       = fmr->rkey;
		*fmr_handle = fmr;

		return 0;
	}

	if (list_empty(&pool->free_list)) {
		spin_unlock_irqrestore(&pool->pool_lock, flags);
		return -EAGAIN;
	}

	fmr = list_entry(pool->free_list.next, struct ib_fmr, list);
	list_del(&fmr->list);
	ib_fmr_cache_remove(fmr);
	spin_unlock_irqrestore(&pool->pool_lock, flags);

	result = pool->device->fmr_map(fmr,
				       page_list,
				       list_len,
				       io_virtual_address,
				       iova_offset,
				       lkey,
				       rkey);

	if (result) {
		spin_lock_irqsave(&pool->pool_lock, flags);
		list_add(&fmr->list, &pool->free_list);
		spin_unlock_irqrestore(&pool->pool_lock, flags);

		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "fmr_map returns %d",
			       result);
		*fmr_handle = NULL;

		return -EINVAL;
	}

	++fmr->remap_count;
	fmr->ref_count = 1;

	*fmr_handle = fmr;

	if (pool->cache_bucket) {
		fmr->lkey               = *lkey;
		fmr->rkey               = *rkey;
		fmr->io_virtual_address = *io_virtual_address;
		fmr->iova_offset        = iova_offset;
		fmr->page_list_len      = list_len;
		memcpy(fmr->page_list, page_list, list_len * sizeof(*page_list));

		spin_lock_irqsave(&pool->pool_lock, flags);
		ib_fmr_cache_store(pool, fmr);
		spin_unlock_irqrestore(&pool->pool_lock, flags);
	}

	return 0;
}
EXPORT_SYMBOL(ib_fmr_register_physical);

int ib_fmr_deregister(struct ib_fmr *fmr_handle)
{
	struct ib_fmr      *fmr = fmr_handle;
	struct ib_fmr_pool *pool;
	unsigned long flags;

	IB_CHECK_MAGIC(fmr, FMR);

	pool = fmr->pool;

	spin_lock_irqsave(&pool->pool_lock, flags);

	--fmr->ref_count;
	if (!fmr->ref_count) {
		if (fmr->remap_count < TS_IB_FMR_MAX_REMAPS) {
			list_add_tail(&fmr->list, &pool->free_list);
		} else {
			list_add_tail(&fmr->list, &pool->dirty_list);
			++pool->dirty_len;
			wake_up_interruptible(&pool->thread_wait);
		}
	}

	if (IB_COMPILE_FMR_DEBUGGING_CODE) {
		if (fmr->ref_count < 0) {
			TS_REPORT_WARN(MOD_KERNEL_IB,
				       "FMR %p has ref count %d < 0",
				       fmr, fmr->ref_count);
		}
	}

	spin_unlock_irqrestore(&pool->pool_lock, flags);

	return 0;
}
EXPORT_SYMBOL(ib_fmr_deregister);

/*
  Local Variables:
  c-file-style: "linux"
  indent-tabs-mode: t
  End:
*/
