/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: core_mr.c 477 2004-07-16 16:40:34Z roland $
*/

#include "core_priv.h"

#include "ts_kernel_trace.h"
#include "ts_kernel_services.h"

#include <linux/version.h>
#include <linux/module.h>

#include <linux/errno.h>
#include <linux/slab.h>

int ib_memory_register(struct ib_pd         *pd,
                       void                 *start_address,
                       u64              buffer_size,
                       enum ib_memory_access access,
                       struct ib_mr        **mr_handle,
                       u32          	    *lkey,
                       u32          	    *rkey)
{
	struct ib_mr *mr;
	int           ret;

	IB_CHECK_MAGIC(pd, PD);

	if (!pd->device->mr_register) {
		return -ENOSYS;
	}

	mr = kmalloc(sizeof *mr, GFP_KERNEL);
	if (!mr) {
		TS_REPORT_WARN(MOD_KERNEL_IB,"mr struct kmalloc failure");
		return -ENOMEM;
	}

	ret = pd->device->mr_register(pd, start_address, buffer_size, access, mr);

	if (!ret) {
		IB_SET_MAGIC(mr, MR);
		mr->device = pd->device;
		*mr_handle = mr;
		*lkey      = mr->lkey;
		*rkey      = mr->rkey;
	} else {
		kfree(mr);
	}

	return ret;
}
EXPORT_SYMBOL(ib_memory_register);

int ib_memory_register_physical(struct ib_pd              *pd,
				struct ib_physical_buffer *buffer_list,
				int                    	   list_len,
				u64               	  *io_virtual_address,
				u64               	   buffer_size,
				u64               	   iova_offset,
				enum ib_memory_access      access,
				struct ib_mr             **mr_handle,
				u32            	          *lkey,
				u32            	          *rkey)
{
	struct ib_mr *mr;
	int           ret;

	IB_CHECK_MAGIC(pd, PD);

	if (!pd->device->mr_register_physical) {
		return -ENOSYS;
	}

	mr = kmalloc(sizeof *mr, GFP_KERNEL);
	if (!mr) {
		TS_REPORT_WARN(MOD_KERNEL_IB,"mr struct kmalloc failure");
		return -ENOMEM;
	}

	ret = pd->device->mr_register_physical(pd,
					       buffer_list,
					       list_len,
					       io_virtual_address,
					       buffer_size,
					       iova_offset,
					       access,
					       mr);

	if (!ret) {
		IB_SET_MAGIC(mr, MR);
		mr->device = pd->device;
		*mr_handle = mr;
		*lkey      = mr->lkey;
		*rkey      = mr->rkey;
	} else {
		kfree(mr);
	}

	return ret;
}
EXPORT_SYMBOL(ib_memory_register_physical);

int ib_memory_deregister(struct ib_mr *mr)
{
	int           ret;

	IB_CHECK_MAGIC(mr, MR);

	if (!mr->device->mr_deregister) {
		return -ENOSYS;
	}

	ret = mr->device->mr_deregister(mr);
	if (!ret) {
		IB_CLEAR_MAGIC(mr);
		kfree(mr);
	}

	return ret;
}
EXPORT_SYMBOL(ib_memory_deregister);

/*
  Local Variables:
  c-file-style: "linux"
  indent-tabs-mode: t
  End:
*/
