/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: core_qp.c 475 2004-07-16 16:29:15Z roland $
*/

#include "core_priv.h"

#include "ts_kernel_trace.h"
#include "ts_kernel_services.h"

#include <linux/version.h>
#include <linux/module.h>

#include <linux/errno.h>
#include <linux/slab.h>

int ib_qp_create(struct ib_qp_create_param *param,
		 struct ib_qp             **qp_handle,
		 u32                       *qpn)
{
	struct ib_pd *pd;
	struct ib_qp *qp;
	int           ret;

	IB_CHECK_MAGIC(param->pd, PD);
	pd = param->pd;

	if (!pd->device->qp_create) {
		return -ENOSYS;
	}

	qp = kmalloc(sizeof *qp, GFP_KERNEL);
	if (!qp) {
		TS_REPORT_WARN(MOD_KERNEL_IB,"qp struct kmalloc failure");
		return -ENOMEM;
	}

	INIT_LIST_HEAD(&qp->async_handler_list);
	spin_lock_init(&qp->async_handler_lock);
	ib_async_obj_init(&qp->async_obj, qp);

	ret = pd->device->qp_create(pd, param, qp);

	if (!ret) {
		IB_SET_MAGIC(qp, QP);
		qp->device = pd->device;
		*qp_handle = qp;
		*qpn       = qp->qpn;
	} else {
		kfree(qp);
	}

	return ret;
}
EXPORT_SYMBOL(ib_qp_create);

int ib_special_qp_create(struct ib_qp_create_param *param,
			 tTS_IB_PORT                port,
			 enum ib_special_qp_type    qp_type,
			 struct ib_qp             **qp_handle)
{
	struct ib_pd *pd;
	struct ib_qp *qp;
	int           ret;

	IB_CHECK_MAGIC(param->pd, PD);
	pd = param->pd;

	if (!pd->device->special_qp_create) {
		return -ENOSYS;
	}

	qp = kmalloc(sizeof *qp, GFP_KERNEL);
	if (!qp) {
		TS_REPORT_WARN(MOD_KERNEL_IB,"pd struct kmalloc failure");
		return -ENOMEM;
	}

	INIT_LIST_HEAD(&qp->async_handler_list);
	spin_lock_init(&qp->async_handler_lock);
	ib_async_obj_init(&qp->async_obj, qp);

	ret = pd->device->special_qp_create(pd, param, port, qp_type, qp);

	if (!ret) {
		IB_SET_MAGIC(qp, QP);
		qp->device = pd->device;
		*qp_handle = qp;
	} else {
		kfree(qp);
	}

	return ret;
}
EXPORT_SYMBOL(ib_special_qp_create);

int ib_qp_modify(struct ib_qp           *qp,
		 struct ib_qp_attribute *attr)
{
	IB_CHECK_MAGIC(qp, QP);
	return qp->device->qp_modify ? qp->device->qp_modify(qp, attr) : -ENOSYS;
}
EXPORT_SYMBOL(ib_qp_modify);

int ib_qp_query(struct ib_qp           *qp,
		struct ib_qp_attribute *attr)
{
	IB_CHECK_MAGIC(qp, QP);
	return qp->device->qp_query ? qp->device->qp_query(qp, attr) : -ENOSYS;
}
EXPORT_SYMBOL(ib_qp_query);

int ib_qp_query_qpn(struct ib_qp    *qp,
		    u32             *qpn)
{
	IB_CHECK_MAGIC(qp, QP);

	*qpn = qp->qpn;
	return 0;
}
EXPORT_SYMBOL(ib_qp_query_qpn);

int ib_qp_destroy(struct ib_qp *qp_handle)
{
	struct ib_qp *qp = qp_handle;
	int           ret;
	int           pending = 0;
	unsigned long flags;

	IB_CHECK_MAGIC(qp, QP);

	if (!qp->device->qp_destroy) {
		return -ENOSYS;
	}

	if (!list_empty(&qp->async_handler_list)) {
		return -EBUSY;
	}

	ret = qp->device->qp_destroy(qp);
	if (!ret) {
		IB_CLEAR_MAGIC(qp);
		spin_lock_irqsave(&qp->async_obj.lock, flags);
		if (!qp->async_obj.pending)
			pending = 1;            
		spin_unlock_irqrestore(&qp->async_obj.lock, flags);
		if (pending == 1)
			kfree(qp);        
	}

	return ret;
}
EXPORT_SYMBOL(ib_qp_destroy);

int ib_qp_detach_destroy(struct ib_qp *qp_handle)
{
	struct ib_qp *qp = qp_handle;
	int           ret;
	int           pending = 0;
	unsigned long flags;

	IB_CHECK_MAGIC(qp, QP);

	if (!qp->device->qp_destroy) {
		return -ENOSYS;
	}

	if (!list_empty(&qp->async_handler_list)) {
		return -EBUSY;
	}

	ret = qp->device->qp_detach_destroy(qp);
	if (!ret) {
		IB_CLEAR_MAGIC(qp);
		spin_lock_irqsave(&qp->async_obj.lock, flags);
		if (!qp->async_obj.pending)
			pending = 1;            
		spin_unlock_irqrestore(&qp->async_obj.lock, flags);
		if (pending == 1)
			kfree(qp);        
	}

	return ret;
}
EXPORT_SYMBOL(ib_qp_detach_destroy);
/*
  Local Variables:
  c-file-style: "linux"
  indent-tabs-mode: t
  End:
*/
