/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: dm_client_class_port_info.c 481 2004-07-16 18:05:10Z roland $
*/

#include "dm_client.h"

#include "ts_ib_mad.h"
#include "ts_ib_core.h"

#include "ts_kernel_trace.h"

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/module.h>

#include <asm/system.h>
#include <asm/byteorder.h>
#include "dm_client_priv.h"

struct ib_dm_class_port_info {
	u8 reserved[40];
	u8 base_version;
	u8 class_version;
	u16 capability_mask;
	u32 resp_time;	/* reserved(27), resp_time(5) */
	tTS_IB_GID redirect_gid;
	u32 redirect_tc_sl_fl;	/* TC(8), SL(4), FL(20) */
	u16 redirect_lid;
	u16 redirect_pkey;
	u32 redirect_qp;	/* reserved(8), qpn(24) */
	u32 redirect_qkey;
	tTS_IB_GID trap_gid;
	u32 trap_tc_sl_fl;	/* TC(8), SL(4), FL(20) */
	u16 trap_lid;
	u16 trap_pkey;
	u32 trap_hl_qp;	/* HopLimit(8), QP(24) */
	u32 trap_qkey;
} __attribute__ ((packed));

struct ib_dm_class_port_info_query {
	struct ib_dm_query_hdr hdr;
	u16 dlid;
	ib_dm_class_port_info_completion_func completion_func;
	void *completion_arg;
} ib_dm_class_port_info_query;

static void ib_dm_class_port_info_response(tTS_IB_CLIENT_RESPONSE_STATUS status,
					   struct ib_mad *packet,
					   void *query_ptr)
{
	struct ib_dm_class_port_info_query *query = query_ptr;

	switch (status) {
	case TS_IB_CLIENT_RESPONSE_OK:
		{
			struct ib_dm_class_port_info *cpi_ptr =
			    (struct ib_dm_class_port_info *) & packet->payload;
			struct ib_common_attrib_cpi *cpi;
			cpi = kmalloc(sizeof *cpi, GFP_KERNEL);
			if (!cpi) {
				TS_REPORT_WARN(MOD_KERNEL_IB, "kmalloc failed");
				goto no_memory;
			}

			cpi->base_version = cpi_ptr->base_version;
			cpi->class_version = cpi_ptr->class_version;
			cpi->capability_mask =
			    be16_to_cpu(cpi_ptr->capability_mask);
			cpi->resp_time = be32_to_cpu(cpi_ptr->resp_time);
			memcpy(cpi->redirect_gid, cpi_ptr->redirect_gid,
			       sizeof(tTS_IB_GID));
			cpi->redirect_tc =
			    (be32_to_cpu(cpi_ptr->redirect_tc_sl_fl) &
			     0xFF000000) >> 24;
			cpi->redirect_sl =
			    (be32_to_cpu(cpi_ptr->redirect_tc_sl_fl) &
			     0x00F00000) >> 20;
			cpi->redirect_fl =
			    (be32_to_cpu(cpi_ptr->redirect_tc_sl_fl) &
			     0x000FFFFF);
			cpi->redirect_lid = be16_to_cpu(cpi_ptr->redirect_lid);
			cpi->redirect_pkey = be16_to_cpu(cpi_ptr->redirect_pkey);
			cpi->redirect_qpn =
			    be32_to_cpu(cpi_ptr->redirect_qp) & 0x00FFFFFF;
			cpi->redirect_qkey = be32_to_cpu(cpi_ptr->redirect_qkey);
			memcpy(cpi->trap_gid, cpi_ptr->trap_gid,
			       sizeof(tTS_IB_GID));
			cpi->trap_tc =
			    (be32_to_cpu(cpi_ptr->trap_tc_sl_fl) & 0xFF000000)
			    >> 24;
			cpi->trap_sl =
			    (be32_to_cpu(cpi_ptr->trap_tc_sl_fl) & 0x00F00000)
			    >> 20;
			cpi->trap_fl =
			    (be32_to_cpu(cpi_ptr->trap_tc_sl_fl) & 0x000FFFFF);
			cpi->trap_lid = be16_to_cpu(cpi_ptr->trap_lid);
			cpi->trap_pkey = be16_to_cpu(cpi_ptr->trap_pkey);
			cpi->trap_hop_limit =
			    (be32_to_cpu(cpi_ptr->trap_hl_qp) & 0xFF000000) >>
			    24;
			cpi->trap_qp =
			    be32_to_cpu(cpi_ptr->trap_hl_qp) & 0x00FFFFFF;
			cpi->trap_qkey = be32_to_cpu(cpi_ptr->trap_qkey);

			TS_TRACE(MOD_KERNEL_DM, T_VERBOSE,
				 TRACE_KERNEL_IB_DM_GEN,
				 "DmClassPortInfo trap LID= 0x%04x, PKEY= 0x%04x\n",
				 cpi->trap_lid, cpi->trap_pkey);

			if (query->completion_func) {
				query->completion_func(query->hdr.transaction_id,
						       0,
						       packet->slid,
						       cpi,
						       query->completion_arg);
			}
			kfree(cpi);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_ERROR:
		TS_TRACE(MOD_KERNEL_DM, T_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			 "DM client Class Port Info query status 0x%04x",
			 be16_to_cpu(packet->status));
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -EINVAL,
					       packet->slid,
					       NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_TIMEOUT:
		TS_TRACE(MOD_KERNEL_DM, T_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			 "DM client Class Port Info query timed out");
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -ETIMEDOUT,
					       query->dlid,
					       NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_CANCEL:
		TS_TRACE(MOD_KERNEL_DM, T_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			 "DM client Class Port Info query canceled");
		break;

	default:
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "DM client Class Port Info query unknown status %d",
			       status);
		break;
	}
no_memory:
	ib_dm_outs_query_remove(&query->hdr.outs_query);
	kfree(query);
}

int ib_dm_class_port_info_request(struct ib_device *device,
				  tTS_IB_PORT port,
				  u16 dst_port_lid,
				  int timeout_jiffies,
				  ib_dm_class_port_info_completion_func
				  completion_func, void *completion_arg,
				  tTS_IB_CLIENT_QUERY_TID * transaction_id)
{
	/* LS : convert mad , lid_info , gid into kmalloc */
	struct ib_mad *mad = NULL;
	struct ib_dm_class_port_info *cpi;
	struct ib_port_lid *lid_info = NULL;
	tTS_IB_GID gid;
	struct ib_dm_class_port_info_query *query;
	int ret = 0;
	tTS_IB_CLIENT_QUERY_TID temp_tid;

	mad = kmalloc(sizeof *mad, GFP_KERNEL);
	if (!mad) {
		return -ENOMEM;
	}
	lid_info = kmalloc(sizeof *lid_info, GFP_KERNEL);
	if (!lid_info) {
		kfree(mad);
		return -ENOMEM;
	}
	
	query = kmalloc(sizeof *query, GFP_KERNEL);
	if (!query) {
		kfree(mad);
		kfree(lid_info);
		return -ENOMEM;
	}

	ib_dm_client_mad_init(mad, device, port, dst_port_lid, IB_GSI_QP,
			      TS_IB_DM_METHOD_SET,
			      TS_IB_DM_ATTRIBUTE_CLASS_PORTINFO, 0);

	/* Set trap LID to notify TS SRP Mgr to forward trap info */
	cpi = (struct ib_dm_class_port_info *) mad->payload;
	ib_cached_lid_get(device, port, lid_info);
	cpi->trap_lid = cpu_to_be16(lid_info->lid);
	ib_cached_gid_get(device, port, 0, gid);
	memcpy(cpi->trap_gid, gid, sizeof(tTS_IB_GID));

	temp_tid = query->hdr.transaction_id = mad->transaction_id;
	query->dlid = dst_port_lid;
	query->completion_func = completion_func;
	query->completion_arg = completion_arg;

	ib_dm_outs_query_add(&query->hdr.outs_query);
	ret = ib_client_query(mad, timeout_jiffies, ib_dm_class_port_info_response,
			query);
	if (ret) {
		ib_dm_outs_query_remove(&query->hdr.outs_query);
		kfree(mad);
		kfree(query);
		goto proc_exit;
	}

	*transaction_id = temp_tid;
proc_exit:	
	kfree(lid_info);
	return ret;
}
EXPORT_SYMBOL(ib_dm_class_port_info_request);
