/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: dm_client_host.c 481 2004-07-16 18:05:10Z roland $
*/

#include "dm_client.h"

#include "ts_ib_mad.h"
#include "ts_ib_sa_client.h"
#include "ts_ib_dm_client_host.h"

#include "ts_kernel_trace.h"

#include "ts_kernel_trace.h"
#include "ts_kernel_hash.h"
#include "ts_kernel_timer.h"

#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/slab.h>
#include <linux/random.h>
#include <linux/spinlock.h>
#include <linux/vmalloc.h>
#include <linux/module.h>
#include <asm/system.h>
#include <asm/byteorder.h>
#include "cleanup_prot.h"

#if 1			/* low memory - problem with host with 2G memory */

#define TS_IB_DM_CLIENT_HOST_QUERY_MAX_NUM_HOST_PORTS     4
#define TS_IB_DM_CLIENT_HOST_MAX_NUM_IO_PORT              16
#define TS_IB_DM_CLIENT_HOST_MIN_CONTROLLER_ID            1
#define TS_IB_DM_CLIENT_HOST_MAX_CONTROLLER_ID            255
#define TS_IB_DM_CLIENT_HOST_MIN_SVC_ENTRY_ID             0
#define TS_IB_DM_CLIENT_HOST_MAX_NUM_DM_IOC               2
#define TS_IB_DM_CLIENT_HOST_MAX_NUM_DM_SVC               64

#else

#define TS_IB_DM_CLIENT_HOST_QUERY_MAX_NUM_HOST_PORTS     4
#define TS_IB_DM_CLIENT_HOST_MAX_NUM_IO_PORT              10
#define TS_IB_DM_CLIENT_HOST_MIN_CONTROLLER_ID            1
#define TS_IB_DM_CLIENT_HOST_MAX_CONTROLLER_ID            255
#define TS_IB_DM_CLIENT_HOST_MIN_SVC_ENTRY_ID             0
#define TS_IB_DM_CLIENT_HOST_MAX_NUM_DM_IOC               256
#define TS_IB_DM_CLIENT_HOST_MAX_NUM_DM_SVC               256

#endif

#define TS_IB_DM_CLIENT_HOST_TIMEOUT 10

enum {
	TS_IB_DM_CLIENT_HOST_MAX_OUTSTANDING_QUERIES = 10
};

enum {
	TS_IB_DM_CLIENT_MAX_IO_PORT = 10
};

enum {
	TS_IB_DM_CLIENT_HOST_QUERY_HASH_BITS = 10,
};

typedef enum {
	TS_IB_DM_PORT_INFO = 1,
	TS_IB_DM_PORT_INFO_TBL
} tTS_IB_DM_QUERY_TYPE;

struct ib_dm_client_host_io_port {
	int is_valid;
	u16 io_port_lid;
	u8 total_svc_entries_queries;
	u8 responded_svc_entries_queries;
	int is_responded;
	struct ib_iou_info iou_info;
	struct ib_ioc_profile ioc_profiles[TS_IB_DM_CLIENT_HOST_MAX_NUM_DM_IOC];
	struct ib_svc_entry svc_entries[TS_IB_DM_CLIENT_HOST_MAX_NUM_DM_IOC]
	    [TS_IB_DM_CLIENT_HOST_MAX_NUM_DM_SVC];
};

struct ib_dm_client_host_query {
	u64 transaction_id;
	struct ib_device *host_device;
	tTS_IB_PORT host_port;
	struct ib_dm_client_host_io_port
	    *io_port[TS_IB_DM_CLIENT_HOST_MAX_NUM_IO_PORT];

	int timeout_jiffies;
	ib_dm_client_host_completion_func completion_function;
	void *completion_arg;

	tTS_KERNEL_TIMER_STRUCT timer;
	tTS_IB_DM_QUERY_TYPE type;

	int callback_running;
	tTS_LOCKED_HASH_ENTRY_STRUCT entry;
};

static spinlock_t alloc_tid_lock = SPIN_LOCK_UNLOCKED;

static u64 cur_transaction_id;

static tTS_HASH_TABLE query_hash;

static ib_cleanup_prot_t dm_host_cleanup_prot;
extern void ib_client_query_set_in_cleanup_wait(void);

u64 ib_host_io_alloc_tid(void)
{
	u64 tid;
	u64 invalid_tid = TS_IB_DM_CLIENT_HOST_TID_INVALID & 0xFFFFFFFF;
	unsigned long flags;

	spin_lock_irqsave(&alloc_tid_lock, flags);
	tid = cur_transaction_id++;
	cur_transaction_id &= 0xFFFFFFFF;
	if (cur_transaction_id == invalid_tid) {
		cur_transaction_id = 0;
	}
	spin_unlock_irqrestore(&alloc_tid_lock, flags);

	return tid;
}

static struct ib_dm_client_host_query *ib_dm_alloc_query(void)
{
	int i;
	struct ib_dm_client_host_query *query;

	query = kmalloc(sizeof *query, GFP_KERNEL);

	if (!query) {
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"Memory allocation for %d bytes failed",
				sizeof(struct ib_dm_client_host_query));
		return NULL;
	}
	       
	memset(query, 0, sizeof *query);

	for (i = 0; i < TS_IB_DM_CLIENT_HOST_MAX_NUM_IO_PORT; i++) {
		query->io_port[i] = kmalloc(sizeof *query->io_port[i],
					    GFP_KERNEL);
		if (!query->io_port[i])
			goto err_free;

		memset(query->io_port[i], 0, sizeof *query->io_port[i]);
	}

	return query;

err_free:
	for (i = 0; query->io_port[i]; i++)
		kfree(query->io_port[i]);
	kfree(query);
	return NULL;
}

static void ib_dm_free_query(struct ib_dm_client_host_query *query)
{
	int i;

	for (i = 0; i < TS_IB_DM_CLIENT_HOST_MAX_NUM_IO_PORT; i++) {
		kfree(query->io_port[i]);
	}

	kfree(query);
}

static struct ib_dm_client_host_io_port *
ib_dm_alloc_io_port(struct ib_dm_client_host_query *query, u16 port_lid)
{
	int i;

	for (i = 0; i < TS_IB_DM_CLIENT_HOST_MAX_NUM_IO_PORT; i++) {
		if (query->io_port[i]->is_valid == 0) {
			query->io_port[i]->is_valid = 1;
			query->io_port[i]->io_port_lid = port_lid;
			query->io_port[i]->total_svc_entries_queries = 0;
			query->io_port[i]->responded_svc_entries_queries = 0;
			query->io_port[i]->is_responded = 0;
			return query->io_port[i];
		}
	}

	return NULL;
}

static struct ib_dm_client_host_io_port *
ib_dm_find_io_port(struct ib_dm_client_host_query *query, u16 port_lid)
{
	int i;

	for (i = 0; i < TS_IB_DM_CLIENT_HOST_MAX_NUM_IO_PORT; i++) {
		if (query->io_port[i]->is_valid == 1 &&
		    query->io_port[i]->io_port_lid == port_lid) {
			return query->io_port[i];
		}
	}

	return NULL;
}

static struct ib_dm_client_host_query *
ib_dm_client_host_query_find(u64 transaction_id)
{
	tTS_LOCKED_HASH_ENTRY entry =
	    tsKernelLockedHashLookup(query_hash,
				     (u32) (transaction_id & 0xffffffff));
	return entry ? TS_KERNEL_HASH_ENTRY(entry,
					    struct ib_dm_client_host_query,
					    entry)
	    : NULL;
}

static inline void
ib_dm_client_host_query_put(struct ib_dm_client_host_query *query)
{
	tsKernelLockedHashRelease(&query->entry);
}

static void
ib_dm_client_host_query_delete(struct ib_dm_client_host_query *query)
{
	tsKernelLockedHashRemove(query_hash, &query->entry);

	tsKernelTimerRemove(&query->timer);

	ib_dm_free_query(query);
}

static void
ib_dm_client_host_query_complete(struct ib_dm_client_host_query *query,
					   int status,
					   struct ib_dm_client_host_io_list
					   *io_list, int force_done)
{
	int i;
	int is_done = 1;
	u64 transaction_id = query->transaction_id;
	ib_dm_client_host_completion_func function = query->completion_function;
	void *arg = query->completion_arg;

	function(status, io_list, arg);

	if (force_done == 0) {
		for (i = 0; i < TS_IB_DM_CLIENT_HOST_MAX_NUM_IO_PORT; i++) {
			if (query->io_port[i]->is_valid == 1 &&
			    query->io_port[i]->is_responded != 1) {
				is_done = 0;
			}
		}
	}

	if (is_done) {
		function(0, NULL, query->completion_arg);

		query = ib_dm_client_host_query_find(transaction_id);
		if (query) {

			query->callback_running = 0;

			ib_dm_client_host_query_delete(query);
		}
	}
}

static void ib_host_io_query_timeout(void *arg)
{
	u64 host_transaction_id = (unsigned long)arg;
	struct ib_dm_client_host_query *query;

	/* if cleanup has started, do not allow new queries */
	if (ib_check_in_cleanup_incr(&dm_host_cleanup_prot))
		return; 

	 query = ib_dm_client_host_query_find(host_transaction_id);
	 while (query && query->callback_running) {
		ib_dm_client_host_query_put(query);
		schedule();
		query = ib_dm_client_host_query_find(host_transaction_id);
	}

	if (!query) {
		ib_in_cleanup_decr(&dm_host_cleanup_prot);
		return;
	}

	query->callback_running = 1;

	ib_dm_client_host_query_put(query);

	ib_dm_client_host_query_complete(query, -ETIMEDOUT, NULL, 1);

	ib_in_cleanup_decr(&dm_host_cleanup_prot);
	return;
}

static void ib_host_io_notify_caller(struct ib_dm_client_host_query *host_query,
				     struct ib_dm_client_host_io_port *io_port)
{
	struct ib_dm_client_host_io_list *io_list;

	TS_TRACE(MOD_KERNEL_DM, T_VERY_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
		 "ib_host_io_notify_caller()");

	/* Make the list of IO for the host to use */
	io_list =
	    kmalloc(sizeof(struct ib_dm_client_host_io_list), GFP_KERNEL);
	if (!io_list) {
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"Couldn't allocate space of io list");

		ib_dm_client_host_query_complete(host_query, -ENOMEM, NULL, 1);

		return;
	} else {
		int controller_id;
		struct ib_dm_client_host_io_svc *io_svc;

		memset(io_list, 0, sizeof(struct ib_dm_client_host_io_list));

		INIT_LIST_HEAD(&io_list->io_svc_list);

		for (controller_id = TS_IB_DM_CLIENT_HOST_MIN_CONTROLLER_ID;
		     controller_id <= io_port->iou_info.max_controllers;
		     controller_id++) {
			int svc_id;
			int is_present =
			    ib_is_ioc_present(&io_port->iou_info, controller_id);

			if (is_present == 0) {
				continue;
			}

			for (svc_id = TS_IB_DM_CLIENT_HOST_MIN_SVC_ENTRY_ID;
			     svc_id <
			     io_port->ioc_profiles[controller_id].
			     num_svc_entries; svc_id++) {
				io_svc =
				    kmalloc(sizeof
					    (struct ib_dm_client_host_io_svc),
					    GFP_KERNEL);

				if (!io_svc) {
					ib_host_io_list_free(io_list);

					ib_dm_client_host_query_complete
					    (host_query, -ENOMEM, NULL, 1);

					return;
				} else {
					io_svc->port_lid = io_port->io_port_lid;
					memcpy(io_svc->controller_guid,
					       io_port->
					       ioc_profiles[controller_id].guid,
					       sizeof(tTS_IB_GUID));
					memcpy(&io_svc->svc_entry,
					       &io_port->
					       svc_entries[controller_id]
					       [svc_id],
					       sizeof(struct ib_svc_entry));

					list_add_tail(&io_svc->list,
						      &io_list->io_svc_list);
				}
			}	/* svc */
		}		/* contoller */
	}

	ib_dm_client_host_query_complete(host_query, 0, io_list, 0);
}

static void ib_host_io_svc_entries_completion(tTS_IB_CLIENT_QUERY_TID
					      transaction_id, int status,
					      u16 lid,
					      struct ib_svc_entries
					      *svc_entries,
					      void *arg)
{
	u64 host_transaction_id = (unsigned long)arg;
	struct ib_dm_client_host_query *host_query =
	    ib_dm_client_host_query_find(host_transaction_id);

	u32 i;
	u32 done = 1;
	int num_entries_to_query;
	tTS_IB_CLIENT_QUERY_TID tid;

	/* Do a little query house keeping */
	if (!host_query) {
		return;
	} else {

		if (host_query->callback_running) {
			ib_dm_client_host_query_put(host_query);
			return;
		}

		host_query->callback_running = 1;
		ib_dm_client_host_query_put(host_query);
	}

	if (status == 0) {
		u32 controller_id = svc_entries->controller_id;
		u32 begin_svc_entry_id = svc_entries->begin_svc_entry_id;
		u32 end_svc_entry_id = svc_entries->end_svc_entry_id;
		u32 svc_entry_id;
		struct ib_dm_client_host_io_port *io_port =
		    ib_dm_find_io_port(host_query, lid);

		TS_TRACE(MOD_KERNEL_DM, T_VERY_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			 "SVC Entries (LID= 0x%x, controller_id= %d, "
			 "begin_id= %d, end_id= %d) lookup done, status %d\n",
			 lid,
			 controller_id,
			 begin_svc_entry_id, end_svc_entry_id, status);

		if (!io_port) {
			/* WARNING */
			TS_TRACE(MOD_KERNEL_DM, T_TERSE, TRACE_KERNEL_IB_DM_GEN,
				 "Failed to find io_port on lid= 0x%x\n", lid);
			return;
		}

		/* store svc entries */
		for (i = begin_svc_entry_id; i <= end_svc_entry_id; i++) {
			svc_entry_id = i % TS_IB_SVC_ENTRIES_NUM_SVC_ENTRY;
			memcpy(&io_port->svc_entries[controller_id][i],
			       &svc_entries->service_entries[svc_entry_id],
			       sizeof(struct ib_svc_entry));
		}

		++io_port->responded_svc_entries_queries;

		/* get the next block of svc entries */
		end_svc_entry_id += 1;
		if (end_svc_entry_id + TS_IB_SVC_ENTRIES_NUM_SVC_ENTRY <=
		    io_port->ioc_profiles[controller_id].num_svc_entries) {
			num_entries_to_query = TS_IB_SVC_ENTRIES_NUM_SVC_ENTRY;
		} else {
			num_entries_to_query =
			    io_port->ioc_profiles[controller_id].
			    num_svc_entries - end_svc_entry_id;
		}

		if (io_port->responded_svc_entries_queries !=
		    io_port->total_svc_entries_queries) {
			done = 0;
		}

		if (done) {
			io_port->is_responded = 1;
			ib_host_io_notify_caller(host_query, io_port);
		} else {
			if (num_entries_to_query > 0)
				ib_svc_entries_request(
					host_query->host_device,
					host_query->host_port,
					io_port->io_port_lid,
					controller_id,
					end_svc_entry_id,
					end_svc_entry_id +
					num_entries_to_query - 1,
					TS_IB_DM_CLIENT_HOST_TIMEOUT * HZ,
					ib_host_io_svc_entries_completion,
					(void *)(unsigned long)
					host_query->transaction_id, &tid);
		}
	} else {
		struct ib_dm_client_host_io_port *io_port =
		    ib_dm_find_io_port(host_query, lid);

		if (io_port && io_port->is_responded == 0) {
			TS_TRACE(MOD_KERNEL_DM, T_VERY_VERBOSE,
				 TRACE_KERNEL_IB_DM_GEN,
				 "SVC Entries (LID= 0x%x) lookup done, "
				 "status %d\n",
				 lid, status);

			io_port->is_responded = 1;

			ib_dm_client_host_query_complete(host_query, status,
							 NULL, 0);
		}
	}

	host_query = ib_dm_client_host_query_find(host_transaction_id);
	if (host_query) {
		host_query->callback_running = 0;

		ib_dm_client_host_query_put(host_query);
	}
}

static void ib_host_io_ioc_profile_completion(tTS_IB_CLIENT_QUERY_TID
					      transaction_id, int status,
					      u16 lid,
					      struct ib_ioc_profile
					      *ioc_profile,
					      void *arg)
{
	tTS_IB_CLIENT_QUERY_TID tid;
	u64 host_transaction_id = (unsigned long)arg;
	struct ib_dm_client_host_query *host_query =
	    ib_dm_client_host_query_find(host_transaction_id);

	/* Do a little query house keeping */
	if (!host_query) {
		return;
	} else {

		if (host_query->callback_running) {
			ib_dm_client_host_query_put(host_query);
			return;
		}

		host_query->callback_running = 1;
		ib_dm_client_host_query_put(host_query);
	}

	if (!status) {
		int svc_entry_id;
		int num_entries_to_query;
		struct ib_dm_client_host_io_port *io_port =
		    ib_dm_find_io_port(host_query, lid);

		TS_TRACE(MOD_KERNEL_DM, T_VERY_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			 "IOC Profile (LID= 0x%x, controller_id %d)"
			 " lookup done, status %d\n",
			 lid, ioc_profile->controller_id, status);

		if (!io_port) {
			/* WARNING */
			TS_TRACE(MOD_KERNEL_DM, T_TERSE, TRACE_KERNEL_IB_DM_GEN,
				 "Failed to find io_port on lid= 0x%x\n", lid);
			return;
		}

		memcpy(&io_port->ioc_profiles[ioc_profile->controller_id],
		       ioc_profile, sizeof(*ioc_profile));

		for (svc_entry_id = TS_IB_DM_CLIENT_HOST_MIN_SVC_ENTRY_ID;
		     svc_entry_id <
		     io_port->ioc_profiles[ioc_profile->controller_id].
		     num_svc_entries;
		     /* svc_entry_id incremented inside the loop */ ) {
			if ((svc_entry_id + TS_IB_SVC_ENTRIES_NUM_SVC_ENTRY) <=
			    io_port->ioc_profiles[ioc_profile->controller_id].
			    num_svc_entries) {
				num_entries_to_query =
				    TS_IB_SVC_ENTRIES_NUM_SVC_ENTRY;
			} else {
				num_entries_to_query =
				    io_port->ioc_profiles[ioc_profile->
							  controller_id].
				    num_svc_entries - svc_entry_id;
			}

			svc_entry_id += num_entries_to_query;

			++io_port->total_svc_entries_queries;
		}

		if (TS_IB_SVC_ENTRIES_NUM_SVC_ENTRY <=
		    io_port->ioc_profiles[ioc_profile->controller_id].
		    num_svc_entries) {
			num_entries_to_query = TS_IB_SVC_ENTRIES_NUM_SVC_ENTRY;
		} else {
			num_entries_to_query =
			    io_port->ioc_profiles[ioc_profile->controller_id].
			    num_svc_entries -
			    TS_IB_DM_CLIENT_HOST_MIN_SVC_ENTRY_ID;
		}

		ib_svc_entries_request(host_query->host_device,
				      host_query->host_port,
				      io_port->io_port_lid,
				      ioc_profile->controller_id,
				      TS_IB_DM_CLIENT_HOST_MIN_SVC_ENTRY_ID,
				      TS_IB_DM_CLIENT_HOST_MIN_SVC_ENTRY_ID +
				      num_entries_to_query - 1,
				      TS_IB_DM_CLIENT_HOST_TIMEOUT * HZ,
				      ib_host_io_svc_entries_completion,
				      (void *)(unsigned long)host_query->
				      transaction_id, &tid);
	} else {
		struct ib_dm_client_host_io_port *io_port =
		    ib_dm_find_io_port(host_query, lid);

		if (io_port && io_port->is_responded == 0) {
			TS_TRACE(MOD_KERNEL_DM, T_VERY_VERBOSE,
				 TRACE_KERNEL_IB_DM_GEN,
				 "IOC Profile (LID= 0x%x) lookup done, "
				 "status %d\n",
				 lid, status);

			io_port->is_responded = 1;
			ib_dm_client_host_query_complete(host_query, status,
							 NULL, 0);
		}
	}

	host_query = ib_dm_client_host_query_find(host_transaction_id);
	if (host_query) {
		host_query->callback_running = 0;

		ib_dm_client_host_query_put(host_query);
	}
}

static void
ib_host_io_iou_info_completion(tTS_IB_CLIENT_QUERY_TID transaction_id,
			       int status, u16 lid,
			       struct ib_iou_info *iou_info, void *arg)
{
	tTS_IB_CLIENT_QUERY_TID tid;
	u64 host_transaction_id = (unsigned long)arg;
	struct ib_dm_client_host_query *host_query =
	    ib_dm_client_host_query_find(host_transaction_id);

	/* Do a little query house keeping */
	if (!host_query) {
		return;
	} else {

		if (host_query->callback_running) {
			ib_dm_client_host_query_put(host_query);
			return;
		}

		host_query->callback_running = 1;
		ib_dm_client_host_query_put(host_query);
	}

	TS_TRACE(MOD_KERNEL_DM, T_VERY_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
		 "IOU info (IO port LID = 0x%x) lookup done, status %d\n",
		 lid, status);

	if (!status) {
		tTS_IB_CONTROLLER_ID controller_id;
		struct ib_dm_client_host_io_port *io_port =
		    ib_dm_find_io_port(host_query, lid);

		if (!io_port) {
			/* WARNING */
			TS_TRACE(MOD_KERNEL_DM, T_TERSE, TRACE_KERNEL_IB_DM_GEN,
				 "Failed to find io_port on lid= 0x%x\n", lid);

			return;
		}

		memcpy(&io_port->iou_info, iou_info, sizeof(*iou_info));

		/* Query IOC info */
		for (controller_id = TS_IB_DM_CLIENT_HOST_MIN_CONTROLLER_ID;
		     controller_id <= io_port->iou_info.max_controllers;
		     controller_id++) {
			int is_present =
			   ib_is_ioc_present(&io_port->iou_info, controller_id);

			if (is_present) {
				TS_TRACE(MOD_KERNEL_DM, T_VERBOSE,
					TRACE_KERNEL_IB_DM_GEN,
					"Send IOC query for io_port"
					" on lid= 0x%x\n",
					io_port->io_port_lid);

				ib_ioc_profile_request(host_query->host_device,
					host_query->host_port,
					io_port->io_port_lid,
					controller_id,
					TS_IB_DM_CLIENT_HOST_TIMEOUT
					 * HZ,
					ib_host_io_ioc_profile_completion,
					(void *)(unsigned long)
					host_query->transaction_id, &tid);
			}
		}
	} else {
		struct ib_dm_client_host_io_port *io_port =
		    ib_dm_find_io_port(host_query, lid);

		if (io_port && io_port->is_responded == 0) {
			TS_TRACE(MOD_KERNEL_DM, T_VERY_VERBOSE,
				 TRACE_KERNEL_IB_DM_GEN,
				 "IOU Info (LID= 0x%x) lookup done, "
				 "status %d\n",
				 lid, status);

			io_port->is_responded = 1;
			ib_dm_client_host_query_complete(host_query, status,
							 NULL, 0);
		}
	}

	host_query = ib_dm_client_host_query_find(host_transaction_id);
	if (host_query) {
		host_query->callback_running = 0;

		ib_dm_client_host_query_put(host_query);
	}
}

static void
ib_host_io_class_port_info_completion(tTS_IB_CLIENT_QUERY_TID transaction_id,
				      int status, u16 lid,
				      struct ib_common_attrib_cpi
				      *class_port_info, void *arg)
{
	/* Nothing to do here */
	if (!status) {
		TS_TRACE(MOD_KERNEL_DM, T_VERY_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			"DM ClassPortInfoSet(LID= 0x%x)"
			 " complete successfully\n",
			lid);
	} else {
		TS_TRACE(MOD_KERNEL_DM, T_VERY_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			 "DM ClassPortInfoSet(LID= 0x%x) failed\n", lid);
	}
}

static void ib_host_io_port_info_completion(tTS_IB_CLIENT_QUERY_TID
					    transaction_id, int status,
					    struct ib_port_info *port_info,
					    void *arg)
{
	tTS_IB_CLIENT_QUERY_TID tid;
	u64 host_transaction_id = (unsigned long)arg;
	struct ib_dm_client_host_query *host_query =
	    ib_dm_client_host_query_find(host_transaction_id);

	/* Do a little query house keeping */
	if (!host_query) {
		TS_REPORT_WARN(MOD_KERNEL_DM, "no query %d",
			       (u32) host_transaction_id);
		return;
	} else {

		if (host_query->callback_running) {
			ib_dm_client_host_query_put(host_query);
			return;
		}

		host_query->callback_running = 1;
		ib_dm_client_host_query_put(host_query);
	}

	if (status == 0) {
		if (TS_IB_PORT_CAP_GET(port_info, TS_IB_PORT_CAP_DM) != 0) {
			struct ib_dm_client_host_io_port *io_port =
			    ib_dm_alloc_io_port(host_query, port_info->lid);

			if (io_port) {
				TS_TRACE(MOD_KERNEL_DM, T_VERY_VERBOSE,
					 TRACE_KERNEL_IB_DM_GEN,
					 "Port Info (LID= 0x%x) lookup "
					 "w/DM bit set, status %d\n",
					 port_info->lid, status);

				/* Send DM ClassPortInfo to register for 
				 * DM related traps */
				ib_dm_class_port_info_request(
					host_query->host_device,
					host_query->host_port,
					io_port->io_port_lid,
					TS_IB_DM_CLIENT_HOST_TIMEOUT * HZ,
					ib_host_io_class_port_info_completion,
					(void *)(unsigned long)
					   host_query->transaction_id,
					&tid);

				/* Query IOU info */
				ib_iou_info_request(
					host_query->host_device,
					host_query->host_port,
					io_port->io_port_lid,
					TS_IB_DM_CLIENT_HOST_TIMEOUT * HZ,
					ib_host_io_iou_info_completion,
					(void *)(unsigned long)
					   host_query->transaction_id,
					&tid);
			} else {
				TS_TRACE(MOD_KERNEL_DM, T_VERBOSE,
					 TRACE_KERNEL_IB_DM_GEN,
					 "Port Info complete alloc error\n",
					 status);
				status = -ENOMEM;
			}
		} else {
			if (host_query->type == TS_IB_DM_PORT_INFO) {
				TS_TRACE(MOD_KERNEL_DM, T_VERY_VERBOSE,
					 TRACE_KERNEL_IB_DM_GEN,
					 "Port Info complete not "
					 "managed Device\n",
					 status);
				status = -EINVAL;
			}
		}
	} else {
		TS_TRACE(MOD_KERNEL_DM, T_VERY_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			 "Port Info lookup done, status %d\n", status);
	}

	/* Call completion function */
	if (status)
		ib_dm_client_host_query_complete(host_query, status, NULL, 1);

	host_query = ib_dm_client_host_query_find(host_transaction_id);
	if (host_query) {
		host_query->callback_running = 0;

		ib_dm_client_host_query_put(host_query);
	}

	return;
}

void ib_host_io_query_cancel(tTS_IB_DM_CLIENT_HOST_TID host_transaction_id)
{
	struct ib_dm_client_host_query *query =
	    ib_dm_client_host_query_find(host_transaction_id);

	while (query && query->callback_running) {
		ib_dm_client_host_query_put(query);
		schedule();
		query = ib_dm_client_host_query_find(host_transaction_id);
	}

	if (!query) {
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "canceling unknown TID 0x%016" TS_U64_FMT "x",
			       host_transaction_id);
		return;
	}

	ib_dm_client_host_query_delete(query);

	return;
}
EXPORT_SYMBOL(ib_host_io_query_cancel);

int ib_host_io_query_init(void)
{
	u64 invalid_tid = TS_IB_DM_CLIENT_HOST_TID_INVALID & 0xFFFFFFFF;

	if (tsKernelHashTableCreate
	    (TS_IB_DM_CLIENT_HOST_QUERY_HASH_BITS, &query_hash)) {
		return -ENOMEM;
	}

	ib_in_cleanup_init(&dm_host_cleanup_prot);
	get_random_bytes(&cur_transaction_id, sizeof cur_transaction_id);

	cur_transaction_id &= 0xFFFFFFFF;
	if (cur_transaction_id == invalid_tid) {
		cur_transaction_id = 0;
	}

	return 0;
}
EXPORT_SYMBOL(ib_host_io_query_init);

void ib_host_io_query_cleanup(void)
{
	/* delete all outstanding queries (which also deletes timers) */
	tTS_HASH_HEAD  bucket;
	struct ib_dm_client_host_query *query;
	int i;
	tTS_LOCKED_HASH_ENTRY entry, entry_next;

	/* free any stranded queries */
	ib_set_in_cleanup_wait(&dm_host_cleanup_prot);
	ib_client_query_set_in_cleanup_wait();
	for (i = 0; i <= query_hash->hash_mask; i++) {
		bucket = &query_hash->bucket[i];
		TS_KERNEL_HASH_FOR_EACH_ENTRY_SAFE(entry, entry_next,
						   bucket, node) {
			query = TS_KERNEL_HASH_ENTRY(entry,
					struct ib_dm_client_host_query,
					entry);
			ib_dm_client_host_query_delete(query);
		}
	}
	tsKernelHashTableDestroy(query_hash);
	query_hash = NULL;

	/* outstanding dm lower layer queries are cleaned up only 
	 * when module is uninstalled */
}
EXPORT_SYMBOL(ib_host_io_query_cleanup);

int ib_host_io_port_query(struct ib_device *device,
			  tTS_IB_PORT host_port,
			  u16 io_port_lid,
			  int timeout_jiffies,
			  ib_dm_client_host_completion_func
			  completion_function, void *completion_arg,
			  tTS_IB_DM_CLIENT_HOST_TID * transaction_id)
{
	tTS_IB_CLIENT_QUERY_TID tid;
	struct ib_dm_client_host_query *query;
	int ret = 0;

	/* if cleanup has started, do not allow new queries */
	if (ib_check_in_cleanup_incr(&dm_host_cleanup_prot))
		return -ENOMEM; /* no permission because in cleanup */

	 query = ib_dm_alloc_query();
	 if (!query) {
		TS_REPORT_FATAL(MOD_KERNEL_DM, "no mem");
		ib_in_cleanup_decr(&dm_host_cleanup_prot);
		return -ENOMEM;
	}

	query->type = TS_IB_DM_PORT_INFO;
	query->host_device = device;
	query->host_port = host_port;
	query->completion_function = completion_function;
	query->completion_arg = completion_arg;
	query->transaction_id =
	    (tTS_IB_DM_CLIENT_HOST_TID) ib_host_io_alloc_tid();

	/* start timer */
	tsKernelTimerInit(&query->timer);
	if (timeout_jiffies > 0) {
		query->timer.run_time = jiffies + timeout_jiffies;
		query->timer.function = ib_host_io_query_timeout;
		query->timer.arg = (void *)(unsigned long)query->transaction_id;
		tsKernelTimerAdd(&query->timer);
	}

	query->entry.key = (u32) (query->transaction_id & 0xffffffff);
	tsKernelLockedHashStore(query_hash, &query->entry);

	*transaction_id = query->transaction_id;

	ib_dm_client_host_query_put(query);

	/* Start with port info query */
	ret = tsIbPortInfoQuery(query->host_device,
			  query->host_port,
			  io_port_lid,
			  0,
			  TS_IB_DM_CLIENT_HOST_TIMEOUT * HZ,
			  ib_host_io_port_info_completion,
			  (void *)(unsigned long)query->transaction_id, &tid);

	if (ret) 
	      ib_dm_client_host_query_delete(query);

	ib_in_cleanup_decr(&dm_host_cleanup_prot);
	return ret;
}
EXPORT_SYMBOL(ib_host_io_port_query);

int ib_host_io_query(struct ib_device *device,
		     tTS_IB_PORT port,
		     int timeout_jiffies,
		     ib_dm_client_host_completion_func completion_function,
		     void *completion_arg,
		     tTS_IB_DM_CLIENT_HOST_TID * transaction_id)
{
	tTS_IB_CLIENT_QUERY_TID tid;
	struct ib_dm_client_host_query *query;
	int ret = 0;

	/* if cleanup has started, do not allow new queries */
	if (ib_check_in_cleanup_incr(&dm_host_cleanup_prot))
		return -ENOMEM; /* no permission because in cleanup */

	query = ib_dm_alloc_query();
	if (!query) {
		TS_REPORT_FATAL(MOD_KERNEL_DM, "no mem");
		ib_in_cleanup_decr(&dm_host_cleanup_prot);
		return -ENOMEM;
	}

	query->type = TS_IB_DM_PORT_INFO_TBL;
	query->host_device = device;
	query->host_port = port;
	query->completion_function = completion_function;
	query->completion_arg = completion_arg;
	query->transaction_id =
	    (tTS_IB_DM_CLIENT_HOST_TID) ib_host_io_alloc_tid();

	/* start timer */
	tsKernelTimerInit(&query->timer);
	if (timeout_jiffies > 0) {
		query->timer.run_time = jiffies + timeout_jiffies;
		query->timer.function = ib_host_io_query_timeout;
		query->timer.arg = (void *)(unsigned long)query->transaction_id;
		tsKernelTimerAdd(&query->timer);
	}

	query->entry.key = (u32) (query->transaction_id & 0xffffffff);
	tsKernelLockedHashStore(query_hash, &query->entry);

	*transaction_id = query->transaction_id;

	ib_dm_client_host_query_put(query);

	/* Start with port info query */
	ret = tsIbPortInfoTblQuery(query->host_device, query->host_port,
			TS_IB_DM_CLIENT_HOST_TIMEOUT * HZ,
			ib_host_io_port_info_completion,
			(void *)(unsigned long)query->transaction_id, 
			&tid);

	if (ret)
	      ib_dm_client_host_query_delete(query);

	ib_in_cleanup_decr(&dm_host_cleanup_prot);
	return ret;
}
EXPORT_SYMBOL(ib_host_io_query);

int ib_host_io_list_free(struct ib_dm_client_host_io_list *io_list)
{
	struct list_head *ptr;

	if (!io_list)
		return 0;

	for (ptr = io_list->io_svc_list.next; ptr != &io_list->io_svc_list;) {
		struct ib_dm_client_host_io_svc *io_svc =
			list_entry(ptr, struct ib_dm_client_host_io_svc, list);

		ptr = ptr->next;

		kfree(io_svc);
	}

	kfree(io_list);

	return 0;
}
EXPORT_SYMBOL(ib_host_io_list_free);
