/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: dm_client_ioc_profile.c 481 2004-07-16 18:05:10Z roland $
*/

#include "dm_client.h"

#include "ts_ib_mad.h"

#include "ts_kernel_trace.h"

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <asm/system.h>
#include <asm/byteorder.h>
#include "dm_client_priv.h"

#define TS_IB_DM_IOCPROFILE_GET_CONTROLLER_ID(x)     (x & 0x000000FF)
#define TS_IB_DM_IOCPROFILE_SET_CONTROLLER_ID(x,y)   (x | (y & 0x000000FF))
#define TS_IB_DM_IOCPROFILE_GET_VENDOR_ID(x)         ((x >> 8) & 0x00FFFFFF)
#define TS_IB_DM_IOCPROFILE__GET_SUBVENDOR_ID(x)     ((x >> 8) & 0x00FFFFFF)

struct ib_dm_ioc_profile {
	u8 reserved[40];
	tTS_IB_GUID guid;	// controller guid
	u32 vendor_id__reserved;	// Vendor ID (24), reserved (8)
	u32 device_id;
	u16 device_version;
	u16 reserved2;
	u32 subsystem_vendor_id__reserved;	// Subsystem Vendor ID (24), reserved (8)
	u32 subsystem_id;
	u16 io_class;
	u16 io_subclass;
	u16 protocol;
	u16 protocol_version;
	u16 service_conn;
	u16 initiators_supported;
	u16 send_msg_depth;
	u16 rdma_read_depth;
	u32 send_msg_size;
	u32 rdma_xfer_size;
	u8 op_capability_mask;
	u8 svc_capability_mask;
	u8 num_svc_entries;
	u8 reserved1[9];
	u8 id_string[64];
} __attribute__ ((packed));

struct ib_dm_ioc_profile_query {
	struct ib_dm_query_hdr hdr;
	u16 dlid;
	ib_ioc_profile_completion_func completion_func;
	void *completion_arg;
};

static void ib_ioc_profile_response(tTS_IB_CLIENT_RESPONSE_STATUS status,
				    struct ib_mad *packet, void *query_ptr)
{
	struct ib_dm_ioc_profile_query *query = query_ptr;

	switch (status) {
	case TS_IB_CLIENT_RESPONSE_OK:
		{
			struct ib_dm_ioc_profile *ioc_profile_ptr =
			    (struct ib_dm_ioc_profile *) & packet->payload;
			struct ib_ioc_profile *ioc_profile;
			/* LS : convert static to dynamic */
			ioc_profile = kmalloc(sizeof *ioc_profile,GFP_KERNEL);
			if (!ioc_profile) {
				/* LS : should be return ENOMEM but function works with void */
				TS_REPORT_WARN(MOD_KERNEL_IB, "kmalloc failed");
				goto no_memory;
			}

			ioc_profile->controller_id =
			    TS_IB_DM_IOCPROFILE_GET_CONTROLLER_ID(be32_to_cpu
								  (packet->
								   attribute_modifier));
			memcpy(ioc_profile->guid, ioc_profile_ptr->guid,
			       sizeof(tTS_IB_GUID));
			ioc_profile->vendor_id =
			    (be32_to_cpu(ioc_profile_ptr->vendor_id__reserved))
			    >> 8;
			ioc_profile->device_id =
			    be32_to_cpu(ioc_profile_ptr->device_id);
			ioc_profile->device_version =
			    be16_to_cpu(ioc_profile_ptr->device_version);
			ioc_profile->subsystem_vendor_id =
			    (be32_to_cpu
			     (ioc_profile_ptr->
			      subsystem_vendor_id__reserved)) >> 8;
			ioc_profile->subsystem_id =
			    be32_to_cpu(ioc_profile_ptr->subsystem_id);
			ioc_profile->io_class =
			    be16_to_cpu(ioc_profile_ptr->io_class);
			ioc_profile->io_subclass =
			    be16_to_cpu(ioc_profile_ptr->io_subclass);
			ioc_profile->protocol =
			    be16_to_cpu(ioc_profile_ptr->protocol);
			ioc_profile->protocol_version =
			    be16_to_cpu(ioc_profile_ptr->protocol_version);
			ioc_profile->service_conn =
			    be16_to_cpu(ioc_profile_ptr->service_conn);
			ioc_profile->initiators_supported =
			    be16_to_cpu(ioc_profile_ptr->initiators_supported);
			ioc_profile->send_msg_depth =
			    be16_to_cpu(ioc_profile_ptr->send_msg_depth);
			ioc_profile->rdma_read_depth =
			    be16_to_cpu(ioc_profile_ptr->rdma_read_depth);
			ioc_profile->send_msg_size =
			    be32_to_cpu(ioc_profile_ptr->send_msg_size);
			ioc_profile->rdma_xfer_size =
			    be32_to_cpu(ioc_profile_ptr->rdma_xfer_size);
			ioc_profile->op_capability_mask =
			    ioc_profile_ptr->op_capability_mask;
			ioc_profile->num_svc_entries =
			    ioc_profile_ptr->num_svc_entries;
			memcpy(ioc_profile->id_string,
			       ioc_profile_ptr->id_string, 64);

			TS_TRACE(MOD_KERNEL_DM, T_VERBOSE,
				 TRACE_KERNEL_IB_DM_GEN,
				 "IOCProfile (ID=%d, guid=%02x%02x%02x%02x%02x%02x%02x%02x, numSvcEntries = %d)\n",
				 ioc_profile->controller_id, ioc_profile->guid[0],
				 ioc_profile->guid[1], ioc_profile->guid[2],
				 ioc_profile->guid[3], ioc_profile->guid[4],
				 ioc_profile->guid[5], ioc_profile->guid[6],
				 ioc_profile->guid[7],
				 ioc_profile->num_svc_entries);

			if (query->completion_func) {
				query->completion_func(query->hdr.transaction_id,
						       0,
						       packet->slid,
						       ioc_profile,
						       query->completion_arg);
			}
			kfree(ioc_profile);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_ERROR:
		TS_TRACE(MOD_KERNEL_DM, T_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			 "DM client IOU Info MAD status 0x%04x",
			 be16_to_cpu(packet->status));
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -EINVAL,
					       packet->slid,
					       NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_TIMEOUT:
		TS_TRACE(MOD_KERNEL_DM, T_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			 "DM client IOU info query timed out");
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -ETIMEDOUT,
					       query->dlid,
					       NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_CANCEL:
		TS_TRACE(MOD_KERNEL_DM, T_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			 "DM client IOU Info query canceled");
		break;

	default:
		TS_REPORT_WARN(MOD_KERNEL_IB, "Unknown status %d", status);
		break;
	}

  no_memory:
	ib_dm_outs_query_remove(&query->hdr.outs_query);
	kfree(query);
}

int ib_ioc_profile_request(struct ib_device *device,
			  tTS_IB_PORT port,
			  u16 dst_port_lid,
			  tTS_IB_CONTROLLER_ID controller_id,
			  int timeout_jiffies,
			  ib_ioc_profile_completion_func completion_func,
			  void *completion_arg,
			  tTS_IB_CLIENT_QUERY_TID * transaction_id)
{
	struct ib_mad *mad;
	struct ib_dm_ioc_profile_query *query;
	int ret = 0;
	tTS_IB_CLIENT_QUERY_TID temp_tid;

	/* LS : convert static to kmalloc */
	mad = kmalloc(sizeof *mad,GFP_KERNEL);
	if (!mad) {
		return -ENOMEM;
	}

	query = kmalloc(sizeof *query, GFP_KERNEL);
	if (!query) {
		kfree(mad);
		return -ENOMEM;
	}

	ib_dm_client_mad_init(mad, device, port, dst_port_lid, IB_GSI_QP,
			    TS_IB_DM_METHOD_GET, TS_IB_DM_ATTRIBUTE_IOC_PROFILE,
			    TS_IB_DM_IOCPROFILE_GET_CONTROLLER_ID
			    (controller_id));

	temp_tid = query->hdr.transaction_id = mad->transaction_id;
	query->dlid = dst_port_lid;
	query->completion_func = completion_func;
	query->completion_arg = completion_arg;

	ib_dm_outs_query_add(&query->hdr.outs_query);
	ret = ib_client_query(mad, timeout_jiffies, ib_ioc_profile_response, query);
	if (ret) {
		ib_dm_outs_query_remove(&query->hdr.outs_query);
		kfree(mad);
		kfree(query);
		goto proc_exit;
	}

	*transaction_id = temp_tid;
proc_exit:	
	return ret;
}
EXPORT_SYMBOL(ib_ioc_profile_request);
