/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: dm_client_iou_info.c 481 2004-07-16 18:05:10Z roland $
*/

#include "dm_client.h"

#include "ts_ib_mad.h"

#include "ts_kernel_trace.h"

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <asm/system.h>
#include <asm/byteorder.h>
#include "dm_client_priv.h"

struct ib_dm_iou_info {
	u8 reserved[40];
	u16 change_id;	/* Incremented, with rollover, to any change to
				   ControllerList */
	u8 max_controllers;	/* Number of slots in ControllerList */
	u8 rsvd_op_rom;	/* rsvd (7), Optional ROM (1) */
	u8 controller_list[128];	/* A series of 4bit nibbles with each
					   representing a slot in the IOU */
} __attribute__ ((packed));

struct ib_dm_iou_info_query {
	struct ib_dm_query_hdr hdr;
	u16 dlid;
	ib_iou_info_completion_func completion_func;
	void *completion_arg;
} ib_dm_iou_info_query;

static void ib_iou_info_response(tTS_IB_CLIENT_RESPONSE_STATUS status,
				 struct ib_mad *packet, void *query_ptr)
{
	struct ib_dm_iou_info_query *query = query_ptr;

	switch (status) {
	case TS_IB_CLIENT_RESPONSE_OK:
		{
			struct ib_dm_iou_info *iou_info_ptr =
			    (struct ib_dm_iou_info *) & packet->payload;
			/* LS : Converty static to dynamic */
			struct ib_iou_info *iou_info;
			iou_info = kmalloc(sizeof *iou_info,GFP_KERNEL);
			if (!iou_info) {
				/* LS : should be return ENOMEM but function works with void */
				TS_REPORT_WARN(MOD_KERNEL_IB, "kmalloc failed");
				goto no_memory;
			}

			iou_info->lid = packet->slid;
			iou_info->change_id =
			    be16_to_cpu(iou_info_ptr->change_id);
			iou_info->max_controllers =
			    iou_info_ptr->max_controllers;
			iou_info->op_rom = iou_info_ptr->rsvd_op_rom & 0x1;
			memcpy(iou_info->controller_list,
			       iou_info_ptr->controller_list,
			       TS_IB_IOU_INFO_MAX_NUM_CONTROLLERS_IN_BYTE);

			TS_TRACE(MOD_KERNEL_DM, T_VERBOSE,
				 TRACE_KERNEL_IB_DM_GEN,
				 "IOUNITINFO(LID= 0x%x) - Change ID=%d; MaxController= %d\n",
				 iou_info->lid, iou_info->change_id,
				 iou_info->max_controllers);

			if (query->completion_func) {
				query->completion_func(query->hdr.transaction_id,
						       0,
						       packet->slid,
						       iou_info,
						       query->completion_arg);
			}
			kfree(iou_info);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_ERROR:
		TS_TRACE(MOD_KERNEL_DM, T_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			 "DM client IOU Info MAD status 0x%04x",
			 be16_to_cpu(packet->status));
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -EINVAL,
					       packet->slid,
					       NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_TIMEOUT:
		TS_TRACE(MOD_KERNEL_DM, T_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			 "DM client IOU info query timed out");
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -ETIMEDOUT,
					       query->dlid,
					       NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_CANCEL:
		TS_TRACE(MOD_KERNEL_DM, T_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			 "DM client IOU Info query canceled");
		break;

	default:
		TS_REPORT_WARN(MOD_KERNEL_IB, "Unknown status %d", status);
		break;
	}
no_memory:
	ib_dm_outs_query_remove(&query->hdr.outs_query);
	kfree(query);
}

int ib_iou_info_request(struct ib_device *device,
		       tTS_IB_PORT port,
		       u16 dst_port_lid,
		       int timeout_jiffies,
		       ib_iou_info_completion_func completion_func,
		       void *completion_arg,
		       tTS_IB_CLIENT_QUERY_TID * transaction_id)
{
	/* LS : convert static to kmalloc */
	struct ib_mad *mad;
	struct ib_dm_iou_info_query *query;
	int ret = 0;
	tTS_IB_CLIENT_QUERY_TID temp_tid;

	mad = kmalloc(sizeof *mad,GFP_KERNEL);
	if (!mad) {
		return -ENOMEM;
	}
	query = kmalloc(sizeof *query, GFP_KERNEL);
	if (!query) {
		kfree(mad);
		return -ENOMEM;
	}

	ib_dm_client_mad_init(mad, device, port, dst_port_lid, IB_GSI_QP,
			    TS_IB_DM_METHOD_GET, TS_IB_DM_ATTRIBUTE_IOU_INFO,
			    0);

	temp_tid = query->hdr.transaction_id = mad->transaction_id;
	query->dlid = dst_port_lid;
	query->completion_func = completion_func;
	query->completion_arg = completion_arg;

	ib_dm_outs_query_add(&query->hdr.outs_query);
	ret = ib_client_query(mad, timeout_jiffies, ib_iou_info_response, query);
	if (ret) {
		ib_dm_outs_query_remove(&query->hdr.outs_query);
		kfree(mad);
		kfree(query);
		goto proc_exit;
	}

	*transaction_id = temp_tid;
proc_exit:	
	return ret;
}
EXPORT_SYMBOL(ib_iou_info_request);

int ib_is_ioc_present(struct ib_iou_info *iou_info,
		     tTS_IB_CONTROLLER_ID controller_id)
{
	s16 index;
	u8 is_present;

	if (controller_id < 1) {
		/* Minimum controller ID is 1 */
		return 0;
	}

	index = (controller_id - 1) / 2;

	if (controller_id & 0x1) {
		is_present = ((iou_info->controller_list[index] >> 4) & 0x0F);
	} else {
		is_present = (iou_info->controller_list[index] & 0x0F);
	}

	if (is_present == 0x01)
		return 1;
	else
		return 0;
}
