/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: dm_client_svc_entries.c 481 2004-07-16 18:05:10Z roland $
*/

#include "dm_client.h"

#include "ts_ib_mad.h"

#include "ts_kernel_trace.h"

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <asm/system.h>
#include <asm/byteorder.h>
#include "dm_client_priv.h"

#define TS_IB_DM_SE_GET_CONTROLLER_ID(x)       ((x >> 16) & 0xFFFF)
#define TS_IB_DM_SE_SET_CONTROLLER_ID(x,y)     (x | ((y & 0xFFFF) << 16))
#define TS_IB_DM_SE_GET_END_ENTRY_ID(x)        ((x >> 8) & 0xFF)
#define TS_IB_DM_SE_SET_END_ENTRY_ID(x,y)      (x | ((y & 0xFF) << 8))
#define TS_IB_DM_SE_GET_BEG_ENTRY_ID(x)        (x & 0xFF)
#define TS_IB_DM_SE_SET_BEG_ENTRY_ID(x,y)      (x | (y & 0xFF))

struct ib_dm_svc_entry {
	u8 service_name[TS_IB_SVC_ENTRY_NAME_SIZE];
	u8 service_id[TS_IB_SVC_ENTRY_ID_SIZE];
} __attribute__ ((packed));

struct ib_dm_svc_entries {
	u8 reserved[40];
	struct ib_dm_svc_entry service_entries[TS_IB_SVC_ENTRIES_NUM_SVC_ENTRY];
} __attribute__ ((packed));

struct ib_dm_svc_entries_query {
	struct ib_dm_query_hdr hdr;
	u16 dlid;
	ib_svc_entries_completion_func completion_func;
	void *completion_arg;
};

static void ib_svc_entries_response(tTS_IB_CLIENT_RESPONSE_STATUS status,
				    struct ib_mad *packet, void *query_ptr)
{
	struct ib_dm_svc_entries_query *query = query_ptr;

	switch (status) {
	case TS_IB_CLIENT_RESPONSE_OK:
		{
			struct ib_dm_svc_entries *svc_entries_ptr =
			    (struct ib_dm_svc_entries *) & packet->payload;
			struct ib_svc_entries *svc_entries;
                        /* LS - static memory saving */
                        svc_entries = kmalloc(sizeof *svc_entries,GFP_KERNEL);
                        if (!svc_entries) {
                           /* LS : should be return ENOMEM but function works with void */
                           TS_REPORT_WARN(MOD_KERNEL_IB, "kmalloc failed");
                           goto no_memory;
                        }
			u32 attribute_modifier =
			    be32_to_cpu(packet->attribute_modifier);
			int i;

			svc_entries->controller_id =
			    TS_IB_DM_SE_GET_CONTROLLER_ID(attribute_modifier);
			svc_entries->begin_svc_entry_id =
			    TS_IB_DM_SE_GET_BEG_ENTRY_ID(attribute_modifier);
			svc_entries->end_svc_entry_id =
			    TS_IB_DM_SE_GET_END_ENTRY_ID(attribute_modifier);
			for (i = 0; i < TS_IB_SVC_ENTRIES_NUM_SVC_ENTRY; i++) {
				memcpy(&svc_entries->service_entries[i],
				       &svc_entries_ptr->service_entries[i],
				       sizeof(struct ib_dm_svc_entry));
			}

			for (i = 0; i < TS_IB_SVC_ENTRIES_NUM_SVC_ENTRY; i++) {
				TS_TRACE(MOD_KERNEL_DM, T_VERBOSE,
					 TRACE_KERNEL_IB_DM_GEN,
					 "SvcEntries (controller %d)(id %d)..  Name: %s, ID: %02x%02x%02x%02x%02x%02x%02x%02x\n ",
					 svc_entries->controller_id,
					 svc_entries->begin_svc_entry_id + i,
					 svc_entries->service_entries[i].
					 service_name,
					 svc_entries->service_entries[i].
					 service_id[0],
					 svc_entries->service_entries[i].
					 service_id[1],
					 svc_entries->service_entries[i].
					 service_id[2],
					 svc_entries->service_entries[i].
					 service_id[3],
					 svc_entries->service_entries[i].
					 service_id[4],
					 svc_entries->service_entries[i].
					 service_id[5],
					 svc_entries->service_entries[i].
					 service_id[6],
					 svc_entries->service_entries[i].
					 service_id[7]);
			}

			if (query->completion_func) {
				query->completion_func(query->hdr.transaction_id,
						       0,
						       packet->slid,
						       svc_entries,
						       query->completion_arg);
			}
			kfree(svc_entries);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_ERROR:
		TS_TRACE(MOD_KERNEL_DM, T_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			 "DM client Svc Entries MAD slid= 0x%04x, status 0x%04x",
			 packet->slid, be16_to_cpu(packet->status));
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -EINVAL,
					       packet->slid,
					       NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_TIMEOUT:
		TS_TRACE(MOD_KERNEL_DM, T_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			 "DM client Svc Entris query timed out");
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -ETIMEDOUT,
					       query->dlid,
					       NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_CANCEL:
		TS_TRACE(MOD_KERNEL_DM, T_VERBOSE, TRACE_KERNEL_IB_DM_GEN,
			 "DM client Svc Entries query canceled");
		break;

	default:
		TS_REPORT_WARN(MOD_KERNEL_IB, "Unknown status %d", status);
		break;
	}
    no_memory:
	ib_dm_outs_query_remove(&query->hdr.outs_query);
	kfree(query);
}

int ib_svc_entries_request(struct ib_device *device,
			  tTS_IB_PORT port,
			  u16 dst_port_lid,
			  tTS_IB_CONTROLLER_ID controller_id,
			  tTS_IB_SVC_ENTRY_ID begin_svc_entry_id,
			  tTS_IB_SVC_ENTRY_ID end_svc_entry_id,
			  int timeout_jiffies,
			  ib_svc_entries_completion_func completion_func,
			  void *completion_arg,
			  tTS_IB_CLIENT_QUERY_TID * transaction_id)
{
	/* LS : convert static to kmalloc */
        struct ib_mad *mad;
	struct ib_dm_svc_entries_query *query;
	u32 attrib_mod = 0;
	int ret;
	tTS_IB_CLIENT_QUERY_TID temp_tid;

        mad = kmalloc(sizeof *mad,GFP_KERNEL);
        if (!mad) {
            return -ENOMEM;
        }

	query = kmalloc(sizeof *query, GFP_KERNEL);
	if (!query) {
		kfree(mad);
		return -ENOMEM;
	}

	attrib_mod = TS_IB_DM_SE_SET_CONTROLLER_ID(attrib_mod, controller_id);
	attrib_mod =
	    TS_IB_DM_SE_SET_BEG_ENTRY_ID(attrib_mod, begin_svc_entry_id);
	attrib_mod = TS_IB_DM_SE_SET_END_ENTRY_ID(attrib_mod, end_svc_entry_id);
	ib_dm_client_mad_init(mad, device, port, dst_port_lid, IB_GSI_QP,
			    TS_IB_DM_METHOD_GET, TS_IB_DM_ATTRIBUTE_SVC_ENTRIES,
			    attrib_mod);

	temp_tid = query->hdr.transaction_id = mad->transaction_id;
	query->dlid = dst_port_lid;
	query->completion_func = completion_func;
	query->completion_arg = completion_arg;

	ib_dm_outs_query_add(&query->hdr.outs_query);
	ret = ib_client_query(mad, timeout_jiffies, ib_svc_entries_response, query);
	if (ret) {
		ib_dm_outs_query_remove(&query->hdr.outs_query);
		kfree(mad);
		kfree(query);
		goto proc_exit;
	}

	*transaction_id = temp_tid;
proc_exit:	
	return ret;
}
EXPORT_SYMBOL(ib_svc_entries_request);
