/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: mad_main.c 432 2004-07-12 17:27:37Z roland $
*/

#include <linux/config.h>

#include "mad_priv.h"
#include "mad_mem_compat.h"

#include "ts_ib_tavor_provider.h"

#include "ts_kernel_trace.h"
#include "ts_kernel_services.h"
#include "openib_version.h"

#include <linux/version.h>
#include <linux/module.h>

#include <linux/init.h>
#include <linux/errno.h>

#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/moduleparam.h>
#endif

MODULE_AUTHOR("Roland Dreier");
MODULE_DESCRIPTION("kernel IB MAD API" OPENIB_REVISION);
MODULE_LICENSE("Dual BSD/GPL");

kmem_cache_t *mad_cache;

////Module Param/////////////
int ib_outs_mad_recv_entries = 1024;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/moduleparam.h>
module_param(ib_outs_mad_recv_entries, int, 0444);
#else
MODULE_PARM(ib_outs_mad_recv_entries,"i");
#endif
MODULE_PARM_DESC(ib_outs_mad_recv_entries,"Max number of outstanding mgt receive entries");
/////End of Module Param//////

static int ib_mad_qp_create(struct ib_device *device,
                            tTS_IB_PORT       port,
                            u32               qpn)
{
	struct ib_mad_private *priv = device->mad;
	struct ib_qp_attribute *qp_attr;
	int                    ret;

	TS_TRACE(MOD_KERNEL_IB, T_VERY_VERBOSE, TRACE_KERNEL_IB_GEN,
		 "Creating port %d QPN %d for device %s",
		 port, qpn, device->name);
	qp_attr = kmalloc (sizeof(*qp_attr),GFP_KERNEL);
	if (!qp_attr) {
		TS_REPORT_WARN(MOD_KERNEL_IB, "no memory");
		return -ENOMEM;
	}


	{
		struct ib_qp_create_param param = { { 0 } };

		param.limit.max_outstanding_send_request    = IB_MAD_SENDS_PER_QP;
		param.limit.max_outstanding_receive_request = ib_outs_mad_recv_entries;
		param.limit.max_send_gather_element         = 1;
		param.limit.max_receive_scatter_element     = 1;

		param.pd             = priv->pd;
		param.send_queue     = priv->cq;
		param.receive_queue  = priv->cq;
		param.send_policy    = IB_WQ_SIGNAL_ALL;
		param.receive_policy = IB_WQ_SIGNAL_ALL;
		param.transport      = IB_TRANSPORT_UD;

		ret = ib_special_qp_create(&param,
					   port,
					   qpn == 0 ? IB_SMI_QP : IB_GSI_QP,
					   &priv->qp[port][qpn]);
		if (ret) {
			TS_REPORT_FATAL(MOD_KERNEL_IB,
					"ib_special_qp_create failed for %s port %d QPN %d (%d)",
					device->name, port, qpn, ret);
			goto ret_err;
		}
	}

	qp_attr->state = IB_QP_STATE_INIT;
	qp_attr->qkey  = qpn == 0 ? 0 : IB_GSI_WELL_KNOWN_QKEY;
	/* P_Key index is really irrelevant for QP0/QP1, but we have to set
	   some value for RESET->INIT transition. */
	qp_attr->pkey_index = 0;
	qp_attr->valid_fields =
		IB_QP_ATTRIBUTE_STATE |
		IB_QP_ATTRIBUTE_QKEY  |
		IB_QP_ATTRIBUTE_PKEY_INDEX;
	
	/* This is not required, according to the IB spec, but do it until
	   the Tavor driver is fixed: */
	qp_attr->port          = port;
	qp_attr->valid_fields |= IB_QP_ATTRIBUTE_PORT;

	ret = ib_qp_modify(priv->qp[port][qpn], qp_attr);
	if (ret) {
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"ib_qp_modify -> INIT failed for %s port %d QPN %d (%d)",
				device->name, port, qpn, ret);
		goto ret_err;
	}

	qp_attr->state        = IB_QP_STATE_RTR;
	qp_attr->valid_fields = IB_QP_ATTRIBUTE_STATE;
	ret = ib_qp_modify(priv->qp[port][qpn], qp_attr);
	if (ret) {
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"ib_qp_modify -> RTR failed for %s port %d QPN %d (%d)",
				device->name, port, qpn, ret);
		goto ret_err;
	}

	qp_attr->state    = IB_QP_STATE_RTS;
	qp_attr->send_psn = 0;
	qp_attr->valid_fields =
		IB_QP_ATTRIBUTE_STATE |
		IB_QP_ATTRIBUTE_SEND_PSN;
	ret = ib_qp_modify(priv->qp[port][qpn], qp_attr);
	if (ret) {
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"ib_qp_modify -> RTS failed for %s port %d QPN %d (%d)",
				device->name, port, qpn, ret);
		goto ret_err;
	}
	/*OK flow -- ret = 0 here */
ret_err:
        kfree(qp_attr);
	return(ret);
}

static int ib_mad_init_one(struct ib_device *device)
{
	struct ib_mad_private      *priv;
	/* LS convert static to dynamic mem */
	struct ib_device_properties *prop;
	int                         ret;
	int i,j;
	int have_error;

	prop = kmalloc (sizeof(*prop),GFP_KERNEL);
	if (!prop) {
		TS_REPORT_WARN(MOD_IB_CM, "no memory");
		return -ENOMEM;
	}
	
	ret = ib_device_properties_get(device, prop);
	if (ret) {
		kfree(prop);
		return ret;
	}

	TS_TRACE(MOD_KERNEL_IB, T_VERY_VERBOSE, TRACE_KERNEL_IB_GEN,
		 "Setting up device %s (%s), %d ports",
		 prop->name, prop->provider, prop->num_port);

	priv = kmalloc(sizeof *priv, GFP_KERNEL);
	if (!priv) {
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "Couldn't allocate private structure for %s",
			       prop->name);
		kfree(prop);
		return -ENOMEM;
	}

	memset(priv,0, sizeof *priv);

	do {
		int ierr, jerr;
		have_error = 0;
		for (i = 0; i <=IB_MAD_MAX_PORTS_PER_DEVICE; i++ ) {
			for (j = 0; j < 2; j++ ) {
				priv->receive_buf[i][j] = 
					kmalloc(ib_outs_mad_recv_entries * sizeof(struct ib_mad_buf), GFP_KERNEL);
				if (priv->receive_buf[i][j] == NULL) {
					have_error = 1;
					break;
				}
			}
			if (have_error) { break;}
		}
		if (!have_error) {break;}   /* no error -- have a good malloc set */

		/* have an error.  First free kmallocs */
		for (ierr = 0; ierr <=IB_MAD_MAX_PORTS_PER_DEVICE; ierr++ ) {
			for (jerr = 0; jerr < 2; jerr++ ) {
				if(priv->receive_buf[ierr][jerr] != NULL)
					kfree(priv->receive_buf[ierr][jerr]);
				        priv->receive_buf[ierr][jerr] = NULL;
			}
		}
		/* now, reduce ib_outs_mad_recv_entries */
		ib_outs_mad_recv_entries = (ib_outs_mad_recv_entries/3) *2 ;
	} while (ib_outs_mad_recv_entries >= 256 );

	if (have_error) {
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "ib_outs_mad_recv_entries reduced below minimum value of 256, and still fail allocation");
		goto error_receive_malloc;
	}
	else {
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			       "using ib_outs_mad_recv_entries = %d", ib_outs_mad_recv_entries);
	}
	device->mad    = priv;
	priv->ib_dev   = device;
	priv->num_port = (prop->is_switch) ? 1 : prop->num_port;

	/*
	  Handle device-specific special cases:
	*/

	/* First special case: Tavor provider wants to know that our PD will
	   be used for special QPs. */
	if (!strcmp(prop->provider, "tavor")) {
		struct ib_mlx_pd_param pd_param = { 0 };

		pd_param.special_qp = 1;
        pd_param.av_num = AV_NUM_DEFAULT;

		ret = __ib_pd_create(device, &pd_param, &priv->pd);
	} else
		ret = __ib_pd_create(device, NULL, &priv->pd);

	if (ret) {
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"Failed to allocate PD for %s",
				prop->name);
		goto error;
	}

	{
		int entries =
			(ib_outs_mad_recv_entries + IB_MAD_SENDS_PER_QP) * priv->num_port;
		struct ib_cq_callback callback;

		callback.context        = IB_CQ_CALLBACK_PROCESS;
		callback.policy         = IB_CQ_PROVIDER_REARM;
		callback.function.entry = ib_mad_completion;
		callback.arg            = device;

		ret = __ib_cq_create(device, &entries, &callback, NULL, &priv->cq);
		if (ret) {
			TS_REPORT_FATAL(MOD_KERNEL_IB,
					"Failed to allocate CQ for %s",
					prop->name);
			goto error_free_pd;
		}
	}

	if (ib_mad_register_memory(priv->pd, &priv->mr, &priv->lkey)) {
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"Failed to allocate MR for %s",
				prop->name);
		goto error_free_cq;
	}

	ret = tsKernelQueueThreadStart("ts_ib_mad",
				       ib_mad_work_thread,
				       device,
				       &priv->work_thread);
	if (ret) {
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "Couldn't start completion thread for %s",
			       prop->name);
		goto error_free_mr;
	}

	{
		int start_port, end_port;
		int p, q, i;

		if (prop->is_switch) {
			start_port = end_port = 0;
		} else {
			start_port = 1;
			end_port   = prop->num_port;
		}

		for (p = 0; p <= IB_MAD_MAX_PORTS_PER_DEVICE; ++p) {
			for (q = 0; q <= 1; ++q) {
				priv->qp[p][q]        = NULL;
				priv->send_free[p][q] = IB_MAD_SENDS_PER_QP;
				priv->send_next[p][q] = 0;

				INIT_LIST_HEAD(&priv->send_list[p][q]);
				for (i = 0; i < ib_outs_mad_recv_entries; ++i)
					(priv->receive_buf[p][q])[i].buf = NULL;
			}
		}

		for (p = start_port; p <= end_port; ++p) {
			if (IB_ASSIGN_STATIC_LID)
				ib_mad_static_assign(device, p);

			for (q = 0; q <= 1; ++q) {
				ret = ib_mad_qp_create(device, p, q);
				if (ret)
					goto error_free_qp;

				for (i = 0; i < ib_outs_mad_recv_entries; ++i) {
					ret = ib_mad_post_receive(device, p, q, i);
					if (ret)
						goto error_free_qp;
				}
			}
		}
    }

	kfree(prop);
	return 0;

 error_free_qp:
     {
		int p, q, i;

		for (p = 0; p <= IB_MAD_MAX_PORTS_PER_DEVICE; ++p) {
			for (q = 0; q <= 1; ++q) {
				if (priv->qp[p][q]) {
					ib_qp_destroy(priv->qp[p][q]);
					for (i = 0; i < ib_outs_mad_recv_entries; ++i) {
						if ((priv->receive_buf[p][q])[i].buf)
							pci_unmap_single(priv->ib_dev->dma_device,
									 pci_unmap_addr(&((priv->receive_buf[p][q])[i]),
											mapping),
									 IB_MAD_BUFFER_SIZE,
									 PCI_DMA_FROMDEVICE);
						kfree((priv->receive_buf[p][q])[i].buf);
					}
				}
			}
		}
	}

    tsKernelQueueThreadStop(priv->work_thread);

 error_free_mr:
	ib_memory_deregister(priv->mr);

 error_free_cq:
	__ib_cq_destroy(priv->cq);

 error_free_pd:
	__ib_pd_destroy(priv->pd);
 
	 for (i = 0; i <=IB_MAD_MAX_PORTS_PER_DEVICE; i++ ) {
		 for (j = 0; j < 2; j++ ) {
			 if(priv->receive_buf[i][j] != NULL)
				 kfree(priv->receive_buf[i][j]);
		 }
	 }

 error_receive_malloc:
 error:
	kfree(prop);

	device->mad = NULL; 
	kfree(priv);
	return ret;
}

static void ib_mad_remove_one(struct ib_device *device)
{
	int p, q, i;

	if (device->mad) {
		struct ib_mad_private *priv = device->mad;

		tsKernelQueueThreadStop(priv->work_thread);

		for (p = 0; p <= IB_MAD_MAX_PORTS_PER_DEVICE; ++p) {
			for (q = 0; q <= 1; ++q) {
				if (priv->qp[p][q]) {
					ib_qp_destroy(priv->qp[p][q]);

					for (i = 0; i < ib_outs_mad_recv_entries; ++i) {
						if ((priv->receive_buf[p][q])[i].buf)
							pci_unmap_single(priv->ib_dev->dma_device,
									 pci_unmap_addr(&((priv->receive_buf[p][q])[i]),
											mapping),
									 IB_MAD_BUFFER_SIZE,
									 PCI_DMA_FROMDEVICE);
						kfree((priv->receive_buf[p][q])[i].buf);
					}
				}
				kfree(priv->receive_buf[p][q]);
			}
		}

		ib_memory_deregister(priv->mr);
		__ib_cq_destroy(priv->cq);
		__ib_pd_destroy(priv->pd);

		kfree(priv);
		device->mad = NULL;
	}
}

static void ib_mad_device_notifier(struct ib_device_notifier *self,
				   struct ib_device          *device,
				   int                        event)
{
	switch (event) {
	case IB_DEVICE_NOTIFIER_ADD:
		if (ib_mad_init_one(device))
			TS_REPORT_WARN(MOD_KERNEL_IB,
				       "Failed to initialize device.");
		break;

	case IB_DEVICE_NOTIFIER_REMOVE:
		ib_mad_remove_one(device);
		break;

	default:
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "Unknown device notifier event %d.");
		break;
	}
}

static struct ib_device_notifier mad_notifier = {
	.notifier = ib_mad_device_notifier
};

static int __init ib_mad_init(void)
{
	int ret;

	TS_REPORT_INIT(MOD_KERNEL_IB,
		       "Initializing IB MAD layer");

	ret = ib_mad_proc_setup();
	if (ret)
		return ret;

	mad_cache = kmem_cache_create("ib_mad",
				      ALIGN(sizeof (struct ib_mad),
					    sizeof (void *)),
				      0,
				      SLAB_HWCACHE_ALIGN,
				      NULL,
				      NULL);
	if (!mad_cache) {
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"Couldn't create MAD slab cache");
		ib_mad_proc_cleanup();
		return -ENOMEM;
	}

	ib_device_notifier_register(&mad_notifier);

	TS_REPORT_INIT(MOD_KERNEL_IB,
		       "IB MAD layer initialized");

	return 0;
}

static void __exit ib_mad_cleanup(void)
{
	TS_REPORT_CLEANUP(MOD_KERNEL_IB,
			  "Unloading IB MAD layer");

	ib_device_notifier_deregister(&mad_notifier);
	ib_mad_proc_cleanup();

	if (kmem_cache_destroy(mad_cache))
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "Failed to destroy MAD slab cache (memory leak?)");

	TS_REPORT_CLEANUP(MOD_KERNEL_IB,
			  "IB MAD layer unloaded");
}

module_init(ib_mad_init);
module_exit(ib_mad_cleanup);

/*
  Local Variables:
  c-file-style: "linux"
  indent-tabs-mode: t
  End:
*/
