/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: sa_client_multicast.c 477 2004-07-16 16:40:34Z roland $
*/

#include "sa_client.h"

#include "ts_ib_core.h"
#include "ts_ib_mad.h"
#include "ts_ib_rmpp_mad_types.h"

#include "ts_kernel_trace.h"

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/errno.h>

#include <asm/system.h>
#include <asm/byteorder.h>
#include "sa_client_priv.h"

#define GID_FMT		"%02x%02x%02x%02x%02x%02x%02x%02x" \
				"%02x%02x%02x%02x%02x%02x%02x%02x"

#define GID_ARG(gid)	gid[ 0], gid[ 1], gid[ 2], gid[ 3], \
				gid[ 4], gid[ 5], gid[ 6], gid[ 7], \
				gid[ 8], gid[ 9], gid[10], gid[11], \
				gid[12], gid[13], gid[14], gid[15]

struct ib_sa_multicast_member {
	tTS_IB_GID mgid;
	tTS_IB_GID port_gid;
	u32 qkey;
	u16 mlid;
	u8 mtu;
	u8 tclass;
	u16 pkey;
	u8 rate;
	u8 packet_life;
	u32 sl_flowlabel_hoplimit;
	u8 scope_joinstate;
	u8 reserved[3];
} __attribute__ ((packed));

struct ib_sa_multicast_member_query {
	struct ib_sa_query_hdr hdr;
	tTS_IB_MULTICAST_JOIN_COMPLETION_FUNC completion_func;
	void *completion_arg;
};

struct ib_sa_multicast_member_leave_query {
	struct ib_sa_query_hdr hdr;
	tTS_IB_MULTICAST_LEAVE_COMPLETION_FUNC completion_func;
	void *completion_arg;
};

struct ib_sa_multicast_group_table_query {
	struct ib_sa_query_hdr hdr;
	tTS_IB_MCAST_GROUP_QUERY_COMPLETION_FUNC completion_func;
	void *completion_arg;
};


static void _tsIbMulticastLeaveResponse(tTS_IB_CLIENT_RESPONSE_STATUS status,
				       struct ib_mad *packet, void *query_ptr)
{
        struct ib_sa_multicast_member_leave_query *query = query_ptr;

	switch (status) {
	case TS_IB_CLIENT_RESPONSE_OK:
	  {
            struct ib_sa_payload *sa_payload =
	      (struct ib_sa_payload *) & packet->payload;
            struct ib_sa_multicast_member *mc_ptr =
	      (struct ib_sa_multicast_member *)
	      sa_payload->admin_data;
	    
            TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
		     "SA client multicast Leave Succeed fo MGID " GID_FMT, GID_ARG(mc_ptr->mgid));
	    
            if (query->completion_func) {
	      query->completion_func(query->hdr.transaction_id,
				     0,
				     query->completion_arg);
            }
	  }
	  break;
	  
	case TS_IB_CLIENT_RESPONSE_ERROR:
	  TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
		   "SA client multicast leave member MAD status 0x%04x",
		   be16_to_cpu(packet->status));
	  if (query->completion_func) {
            query->completion_func(query->hdr.transaction_id,
				   -EINVAL,
				   query->completion_arg);
	  }
	  break;
	  
	case TS_IB_CLIENT_RESPONSE_TIMEOUT:
	  TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
		   "SA client multicast leave member query timed out");
	  if (query->completion_func) {
            query->completion_func(query->hdr.transaction_id,
				   -ETIMEDOUT,
                           query->completion_arg);
	  }
	  break;
	  
	case TS_IB_CLIENT_RESPONSE_CANCEL:
	  TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
             "SA client path multicast member query canceled");
	  break;
	  
	default:
	  TS_REPORT_WARN(MOD_KERNEL_IB, "Unknown status %d", status);
	  break;
	}
	ib_sa_outs_query_remove(&query->hdr.outs_query);
	kfree(query);
	
}

static void _tsIbMulticastJoinResponse(tTS_IB_CLIENT_RESPONSE_STATUS status,
				       struct ib_mad *packet, void *query_ptr)
{
	struct ib_sa_multicast_member_query *query = query_ptr;

	switch (status) {
	case TS_IB_CLIENT_RESPONSE_OK:
		{
			struct ib_sa_payload *sa_payload =
			    (struct ib_sa_payload *) & packet->payload;
			struct ib_sa_multicast_member *mc_ptr =
			    (struct ib_sa_multicast_member *)
			    sa_payload->admin_data;
			struct ib_multicast_member mc_member;

			memcpy(mc_member.mgid, mc_ptr->mgid,
			       sizeof mc_member.mgid);

			mc_member.qkey = be32_to_cpu(mc_ptr->qkey);
			mc_member.mlid = be16_to_cpu(mc_ptr->mlid);

			if (mc_member.mlid < 0xc000 || mc_member.mlid > 0xfffe) {
				TS_REPORT_WARN(MOD_KERNEL_IB,
					       "Multicast join succeeded but returned MLID 0x%04x",
					       mc_member.mlid);
			}

			mc_member.mtu = mc_ptr->mtu & 0x3f;
			mc_member.tclass = mc_ptr->tclass;
			mc_member.pkey = be16_to_cpu(mc_ptr->pkey);
			mc_member.rate = mc_ptr->rate & 0x3f;
			mc_member.packet_life = mc_ptr->packet_life & 0x3f;
			mc_member.sl =
			    be32_to_cpu(mc_ptr->sl_flowlabel_hoplimit) >> 28;
			mc_member.flowlabel =
			    cpu_to_be32((be32_to_cpu
					 (mc_ptr->
					  sl_flowlabel_hoplimit) >> 8) &
					0xfffff);
			mc_member.hoplmt =
			    be32_to_cpu(mc_ptr->sl_flowlabel_hoplimit) & 0xff;

			TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
				 "SA client multicast response: "
				 "mlid=0x%04x", mc_member.mlid);

			if (query->completion_func) {
				query->completion_func(query->hdr.transaction_id,
						       0,
						       &mc_member,
						       query->completion_arg);
			}
		}
		break;

	case TS_IB_CLIENT_RESPONSE_ERROR:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client multicast member MAD status 0x%04x",
			 be16_to_cpu(packet->status));
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -EINVAL,
					       NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_TIMEOUT:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client multicast member query timed out");
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -ETIMEDOUT,
					       NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_CANCEL:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client path multicast member query canceled");
		break;

	default:
		TS_REPORT_WARN(MOD_KERNEL_IB, "Unknown status %d", status);
		break;
	}

	ib_sa_outs_query_remove(&query->hdr.outs_query);
	kfree(query);
}

static void _tsIbMcastGroupTableResponse(tTS_IB_CLIENT_RESPONSE_STATUS status,
					 u8 * header,
					 u8 * data,
					 u32 data_size, void *query_ptr)
{
	struct ib_sa_multicast_group_table_query *query =
	    (struct ib_sa_multicast_group_table_query *) query_ptr;

	switch (status) {
	case TS_IB_CLIENT_RESPONSE_OK:
		{
			struct ib_sa_header *sa_header =
			    (struct ib_sa_header *) header;
			s16 attrib_offset =
			    be16_to_cpu(sa_header->attrib_offset) * 8;
			s32 mcast_member_size =
			    sizeof(struct ib_sa_multicast_member);
			s32 num_entries = data_size / attrib_offset;
			s32 num_remaining_entries = num_entries;
			s32 i;

			TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
				 "_tsIbMcastGroupTableResponse(data_size= %d, mcast-member-size= %d, attrib_offset= %d, num-entries= %d)\n",
				 data_size, mcast_member_size, attrib_offset,
				 num_entries);

			for (i = 0; i < num_entries; i++) {
				struct ib_sa_multicast_member *mc_ptr =
				    (struct ib_sa_multicast_member *) (data +
								  i *
								  attrib_offset);
				struct ib_multicast_member mc_member;

				memcpy(mc_member.mgid, mc_ptr->mgid,
				       sizeof mc_member.mgid);

				mc_member.qkey = be32_to_cpu(mc_ptr->qkey);
				mc_member.mlid = be16_to_cpu(mc_ptr->mlid);
				mc_member.mtu = mc_ptr->mtu & 0x3f;
				mc_member.tclass = mc_ptr->tclass;
				mc_member.pkey = be16_to_cpu(mc_ptr->pkey);
				mc_member.rate = mc_ptr->rate & 0x3f;
				mc_member.packet_life =
				    mc_ptr->packet_life & 0x3f;
				mc_member.sl =
				    be32_to_cpu(mc_ptr->
						sl_flowlabel_hoplimit) >> 28;
				mc_member.flowlabel =
				    cpu_to_be32((be32_to_cpu
						 (mc_ptr->
						  sl_flowlabel_hoplimit) >> 8) &
						0xfffff);
				mc_member.hoplmt =
				    be32_to_cpu(mc_ptr->
						sl_flowlabel_hoplimit) & 0xff;

				TS_TRACE(MOD_KERNEL_IB, T_VERBOSE,
					 TRACE_KERNEL_IB_GEN,
					 "SA client multicast group GetTable response: "
					 "mgid=%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x "
					 "mlid=0x%04x", mc_member.mgid[0],
					 mc_member.mgid[1], mc_member.mgid[2],
					 mc_member.mgid[3], mc_member.mgid[4],
					 mc_member.mgid[5], mc_member.mgid[6],
					 mc_member.mgid[7], mc_member.mgid[8],
					 mc_member.mgid[9], mc_member.mgid[10],
					 mc_member.mgid[11], mc_member.mgid[12],
					 mc_member.mgid[13], mc_member.mgid[14],
					 mc_member.mgid[15], mc_member.mlid);

				if (query->completion_func) {
					num_remaining_entries =
					    num_entries - (i + 1);
					query->completion_func(query->
							       hdr.transaction_id,
							       0,
							       num_remaining_entries,
							       &mc_member,
							       query->
							       completion_arg);
				}
			}
		}
		break;

	case TS_IB_CLIENT_RESPONSE_ERROR:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client multicast group table MAD response error");
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -EINVAL,
					       0, NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_TIMEOUT:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client multicast group table query timed out\n");
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -ETIMEDOUT,
					       0, NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_CANCEL:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client multicast group table query canceled\n");
		break;

	default:
		TS_REPORT_WARN(MOD_KERNEL_IB, "Unknown status %d", status);
		break;
	}

	ib_rmpp_client_free(data);

	ib_sa_outs_query_remove(&query->hdr.outs_query);
	kfree(query);
}

int tsIbMulticastGroupJoin(
                           struct ib_device *device,
                           tTS_IB_PORT port,
                           tTS_IB_GID mgid,
                           u16 pkey,
                           tTS_IB_MULTICAST_JOIN_STATE join_state,
                           int timeout_jiffies,
                           tTS_IB_MULTICAST_JOIN_COMPLETION_FUNC completion_func,
                           void *completion_arg,
                           tTS_IB_CLIENT_QUERY_TID *transaction_id,
                           u32 qkey,
                           int create
                           )
{
	struct ib_mad *mad = NULL;
	struct ib_sa_payload *sa_payload = NULL;
	struct ib_sa_multicast_member *mc_member = NULL;
	struct ib_sa_multicast_member_query *query;
	int ret=0;
	tTS_IB_CLIENT_QUERY_TID temp_tid;

	query = kmalloc(sizeof *query, GFP_ATOMIC);
	if (!query) {
		ret = -ENOMEM;
		goto proc_exit;
	}
	mad = kmalloc(sizeof *mad, GFP_ATOMIC);
	if (!mad) {
		kfree(query);
		ret = -ENOMEM;
		goto proc_exit;
	}
	sa_payload = (struct ib_sa_payload *) & mad->payload;
	mc_member = (struct ib_sa_multicast_member *) sa_payload->admin_data;

	tsIbSaClientMadInit(mad, device, port);
	mad->r_method = IB_MGMT_METHOD_SET;
	mad->attribute_id = cpu_to_be16(TS_IB_SA_ATTRIBUTE_MC_MEMBER_RECORD);
	mad->attribute_modifier = 0xffffffff;	/* match attributes */

	temp_tid = query->hdr.transaction_id = mad->transaction_id;
	query->completion_func = completion_func;
	query->completion_arg = completion_arg;
    if (create) {
        sa_payload->sa_header.component_mask = cpu_to_be64((0x1ULL << 0)	/* MGID      */
                                   |(0x1ULL << 1)  /* PortGID   */
                                   |(0x1ULL << 2) /* Q_Key     */
                                   |(0x1ULL << 6)	/* TClass    */
                                   |(0x1ULL << 7)	/* P_Key     */
                                   |(0x1ULL << 12)	/* SL        */
                                   |(0x1ULL << 13)	/* FlowLabel */
                                   |(0x1ULL << 16));	/* Join State */

        memcpy(mc_member->mgid, mgid, sizeof(tTS_IB_GID));
        ib_cached_gid_get(device, port, 0, mc_member->port_gid);
        mc_member->qkey = cpu_to_be32(qkey);
        mc_member->tclass = 0;
        mc_member->pkey = cpu_to_be16(pkey);
        mc_member->sl_flowlabel_hoplimit = 0;
        mc_member->scope_joinstate = join_state;        
                      
    } else {   // old code
	/* mgid, portgid, pkey, join state */
#if defined(TS_ppc440_en_sun) || defined(TS_ppc440_fc_sun)
#if defined(__SOLARIS_IBSM__)	/* old prototype Solaris IBSRM */
	/* Simulate the behavior of the Solaris IPoIB driver - > 0x1b0c7 */
	sa_payload->sa_header.component_mask = cpu_to_be64((0x1ULL << 0)	/* MGID      */
							   |(0x1ULL << 1)
							   /* PortGID   */
							   /* | (0x1ULL << 2) */
							   /* Q_Key     */
							   |(0x1ULL << 6)	/* TClass    */
							   |(0x1ULL << 7)	/* P_Key     */
							   |(0x1ULL << 12)	/* SL        */
							   |(0x1ULL << 13)	/* FlowLabel */
							   |(0x1ULL << 15)	/* Scope     */
							   |(0x1ULL << 16));	/* Join State */
#endif				/* __SOLARIS_IBSM__ */
#define __SOLARIS_IBSRM__
#if defined(__SOLARIS_IBSRM__)	/* official IBSRM */
	/* Per SUN IBSRM group requirment the component mask has to be 0x10003 */
	sa_payload->sa_header.component_mask = cpu_to_be64((0x1ULL << 0)	/* MGID       */
							   |(0x1ULL << 1)	/* PortGID    */
							   |(0x1ULL << 16));	/* Join State */
#endif				/* __SOLARIS_IBSRM__ */
#else
	sa_payload->sa_header.component_mask = cpu_to_be64((0x1ULL << 0)
							   | (0x1ULL << 1)
							   | (0x1ULL << 7)
							   | (0x1ULL << 16));
#endif
	memcpy(mc_member->mgid, mgid, sizeof(tTS_IB_GID));
	mc_member->pkey = cpu_to_be16(pkey);

	ib_cached_gid_get(device, port, 0, mc_member->port_gid);

#if defined(TS_ppc440_en_sun) || defined(TS_ppc440_fc_sun)
#if defined(__SOLARIS_IBSM__)
	mc_member->tclass = 0;
	mc_member->sl_flowlabel_hoplimit = 0;
	mc_member->scope_joinstate = 0x20 | join_state;	/* scope = Link-Local */
#endif				/* __SOLARIS_IBSM__ */
#if defined(__SOLARIS_IBSRM__)	/* official IBSRM */
	mc_member->scope_joinstate = join_state;
#endif				/* __SOLARIS_IBSRM__ */
#else
	mc_member->scope_joinstate = join_state;
#endif
    }

    ib_sa_outs_query_add(&query->hdr.outs_query);
 	ret=ib_client_query(mad, timeout_jiffies, _tsIbMulticastJoinResponse,
			query);
    if (ret) {
	ib_sa_outs_query_remove(&query->hdr.outs_query);
	kfree(mad);
	kfree(query);
        goto proc_exit;
    }
    
    *transaction_id = temp_tid;

 proc_exit:
     return ret;
}

int tsIbMulticastGroupLeave(struct ib_device *device,
                            tTS_IB_PORT port, 
                            tTS_IB_GID mgid,
                            u16 pkey,
                            u32 qkey,
                            tTS_IB_MULTICAST_JOIN_STATE join_state,
                            int timeout_jiffies,
                            tTS_IB_MULTICAST_LEAVE_COMPLETION_FUNC completion_func,
                            void *completion_arg,
                            tTS_IB_CLIENT_QUERY_TID *transaction_id
                            )
{
        struct ib_mad *mad = NULL;
	struct ib_sa_payload *sa_payload = NULL;
	struct ib_sa_multicast_member *mc_member = NULL;
	struct ib_sa_multicast_member_leave_query *query;
	int ret=0;
	tTS_IB_CLIENT_QUERY_TID temp_tid;
	
	query = kmalloc(sizeof *query, GFP_ATOMIC);

        if (!query) {
		ret = -ENOMEM;
		goto proc_exit;
	}
	
	mad = kmalloc(sizeof *mad, GFP_ATOMIC);
	if (!mad) {
		kfree(query);
		ret = -ENOMEM;
		goto proc_exit;
	}

	sa_payload =  (struct ib_sa_payload *) &mad->payload;
	mc_member =   (struct ib_sa_multicast_member *) sa_payload->admin_data;

	tsIbSaClientMadInit(mad, device, port);
	mad->r_method = IB_MGMT_METHOD_DELETE;
	mad->attribute_id = cpu_to_be16(TS_IB_SA_ATTRIBUTE_MC_MEMBER_RECORD);
	mad->attribute_modifier = 0xffffffff;	/* match attributes */
	
	temp_tid = query->hdr.transaction_id = mad->transaction_id;   
	query->completion_func = completion_func;     
	query->completion_arg = completion_arg;       
	
	
	sa_payload->sa_header.component_mask = cpu_to_be64((0x1ULL << 0)	/* MGID      */
							   |(0x1ULL << 1)  /* PortGID   */
							   |(0x1ULL << 2) /* Q_Key     */
							   |(0x1ULL << 6)	/* TClass    */
							   |(0x1ULL << 7)	/* P_Key     */
							   |(0x1ULL << 12)	/* SL        */
							   |(0x1ULL << 13)	/* FlowLabel */
							   |(0x1ULL << 16));	/* Join State */
	
	memcpy(mc_member->mgid, mgid, sizeof(tTS_IB_GID));
	ib_cached_gid_get(device, port, 0, mc_member->port_gid);
	mc_member->qkey = cpu_to_be32(qkey);      //TBD
	mc_member->tclass = 0;
	mc_member->pkey = cpu_to_be16(pkey);      //TBD
	mc_member->sl_flowlabel_hoplimit = 0;
	mc_member->scope_joinstate = join_state;        
	
	
	ib_sa_outs_query_add(&query->hdr.outs_query);
	ret=ib_client_query(mad, timeout_jiffies, _tsIbMulticastLeaveResponse,
			query);
	if (ret) {
		/* query not deleted in ib_client_query */
		ib_sa_outs_query_remove(&query->hdr.outs_query);
		kfree(mad);
		kfree(query);
		goto proc_exit;
	}
	
	if (transaction_id) 
		*transaction_id = temp_tid;

 proc_exit:
	return ret;
}


	

int tsIbMulticastGroupTableQuery(struct ib_device *device,
				 tTS_IB_PORT port,
				 int timeout_jiffies,
				 u16 partition,
				 tTS_IB_MCAST_GROUP_QUERY_COMPLETION_FUNC
				 completion_func, void *completion_arg,
				 tTS_IB_CLIENT_QUERY_TID * transaction_id)
{
	struct ib_mad *mad = NULL;
	struct ib_sa_payload *sa_payload = NULL;
	struct ib_sa_multicast_member *mc_member = NULL;
	struct ib_sa_multicast_group_table_query *query;
	struct ib_client_rmpp_mad *rmpp_mad;
	int ret = 0;
	tTS_IB_CLIENT_QUERY_TID temp_tid;

	query = kmalloc(sizeof *query, GFP_ATOMIC);
	if (!query) {
		return -ENOMEM;
	}

	mad = kmalloc(sizeof *mad, GFP_ATOMIC);
	if (!mad) {
		kfree(query);
		return -ENOMEM;
	}

	sa_payload = (struct ib_sa_payload *) & mad->payload;
	mc_member = (struct ib_sa_multicast_member *) sa_payload->admin_data;

	tsIbSaClientMadInit(mad, device, port);
	mad->r_method = TS_IB_SA_METHOD_GET_TABLE;
	mad->attribute_id = cpu_to_be16(TS_IB_SA_ATTRIBUTE_MC_MEMBER_RECORD);
	mad->attribute_modifier = 0;

	/* rmpp header init */
	rmpp_mad = (struct ib_client_rmpp_mad *) mad;
	rmpp_mad->version = 1;
	rmpp_mad->type = TS_IB_CLIENT_RMPP_TYPE_DATA;

	temp_tid = query->hdr.transaction_id = mad->transaction_id;
	query->completion_func = completion_func;
	query->completion_arg = completion_arg;

	/* PKEY */
	sa_payload->sa_header.component_mask = cpu_to_be64(0x1ULL << 7);
	mc_member->pkey = cpu_to_be16(partition);

	ib_sa_outs_query_add(&query->hdr.outs_query);
	ret = ib_rmpp_client_query(mad, timeout_jiffies, sizeof(struct ib_sa_header),
			     _tsIbMcastGroupTableResponse, query);

	if (ret) {
	    ib_sa_outs_query_remove(&query->hdr.outs_query);
	    kfree(mad);
	    kfree(query);
	    goto proc_exit;
	}
	
	*transaction_id = temp_tid;

proc_exit:	
	return ret;
}
