/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: sa_client_node_info.c 477 2004-07-16 16:40:34Z roland $
*/

#include "sa_client.h"

#include "ts_ib_mad.h"
#include "ts_ib_rmpp_mad_types.h"

#include "ts_kernel_trace.h"

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/errno.h>

#include <asm/system.h>
#include <asm/byteorder.h>
#include "sa_client_priv.h"

struct ib_sa_node_info {
	u16 port_lid;
	u16 reserved;
	u8 base_version;
	u8 class_version;
	u8 node_type;
	u8 num_ports;
	tTS_IB_GUID system_image_guid;
	tTS_IB_GUID node_guid;
	tTS_IB_GUID port_guid;
	u16 partition_cap;
	u16 dev_id;
	u32 dev_rev;
	tTS_IB_PORT local_port_num;
	u8 vendor_id[3];
} __attribute__ ((packed));

struct ib_sa_node_info_query {
	struct ib_sa_query_hdr hdr;
	tTS_IB_NODE_INFO_QUERY_COMPLETION_FUNC completion_func;
	void *completion_arg;
};

#define ORACLE_DEMO_LANE15_HACK

static void _tsIbNodeInfoResponse(tTS_IB_CLIENT_RESPONSE_STATUS status,
				  struct ib_mad *packet, void *query_ptr)
{
	struct ib_sa_node_info_query *query = query_ptr;

	switch (status) {
	case TS_IB_CLIENT_RESPONSE_OK:
		{
			struct ib_sa_payload *sa_payload =
			    (struct ib_sa_payload *) & packet->payload;
			struct ib_sa_node_info *node_info_ptr =
			    (struct ib_sa_node_info *) sa_payload->admin_data;
			struct ib_node_info node_info;

			memset(&node_info, 0, sizeof(node_info));
			node_info.base_version = node_info_ptr->base_version;
			node_info.class_version = node_info_ptr->class_version;
			node_info.node_type = node_info_ptr->node_type;
			node_info.num_ports = node_info_ptr->num_ports;
			memcpy(node_info.system_image_guid,
			       node_info_ptr->system_image_guid,
			       sizeof(tTS_IB_GUID));
			memcpy(node_info.node_guid, node_info_ptr->node_guid,
			       sizeof(tTS_IB_GUID));
			memcpy(node_info.port_guid, node_info_ptr->port_guid,
			       sizeof(tTS_IB_GUID));
			node_info.partition_cap =
			    be16_to_cpu(node_info_ptr->partition_cap);
			node_info.dev_id = be16_to_cpu(node_info_ptr->dev_id);
			node_info.dev_rev = be32_to_cpu(node_info_ptr->dev_rev);
			node_info.local_port_num =
			    node_info_ptr->local_port_num;

			TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
				 "SA client Node Info response: "
				 "Node GUID= %02x%02x%02x%02x%02x%02x%02x%02x",
				 node_info.node_guid[0], node_info.node_guid[1],
				 node_info.node_guid[2], node_info.node_guid[3],
				 node_info.node_guid[4], node_info.node_guid[5],
				 node_info.node_guid[6],
				 node_info.node_guid[7]);

			if (query->completion_func) {
				query->completion_func(query->hdr.transaction_id,
						       0,
						       &node_info,
						       query->completion_arg);
			}
		}
		break;

	case TS_IB_CLIENT_RESPONSE_ERROR:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client node info MAD status 0x%04x",
			 be16_to_cpu(packet->status));
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -EINVAL,
					       NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_TIMEOUT:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client node info query timed out");
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -ETIMEDOUT,
					       NULL, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_CANCEL:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client node info query canceled");
		break;

	default:
		TS_REPORT_WARN(MOD_KERNEL_IB, "Unknown status %d", status);
		break;
	}

	ib_sa_outs_query_remove(&query->hdr.outs_query);
	kfree(query);
}

int tsIbNodeInfoQuery(struct ib_device *device,
		      tTS_IB_PORT port,
		      u16 port_lid,
		      int timeout_jiffies,
		      tTS_IB_NODE_INFO_QUERY_COMPLETION_FUNC completion_func,
		      void *completion_arg,
		      tTS_IB_CLIENT_QUERY_TID * transaction_id)
{
	struct ib_mad *mad;
	struct ib_sa_payload *sa_payload = NULL;
	struct ib_sa_node_info *node_info = NULL;
	struct ib_sa_node_info_query *query;
	int ret = 0;
	tTS_IB_CLIENT_QUERY_TID temp_tid;

	query = kmalloc(sizeof *query, GFP_ATOMIC);
	if (!query) {
		return -ENOMEM;
	}

	mad = kmalloc(sizeof *mad, GFP_ATOMIC);
	if (!mad) {
		kfree(query);
		return -ENOMEM;
	}

	sa_payload = (struct ib_sa_payload *) & mad->payload;
	node_info = (struct ib_sa_node_info *) sa_payload->admin_data;

	tsIbSaClientMadInit(mad, device, port);
	mad->r_method = IB_MGMT_METHOD_GET;
	mad->attribute_id = cpu_to_be16(TS_IB_SA_ATTRIBUTE_NODE_RECORD);
	mad->attribute_modifier = 0;

	temp_tid = query->hdr.transaction_id = mad->transaction_id;
	query->completion_func = completion_func;
	query->completion_arg = completion_arg;

	sa_payload->sa_header.component_mask = cpu_to_be64(0x1ULL);	/* port LID */

	node_info->port_lid = cpu_to_be16(port_lid);

	ib_sa_outs_query_add(&query->hdr.outs_query);
	ret = ib_client_query(mad, timeout_jiffies, _tsIbNodeInfoResponse, query);
	if (ret) {
		ib_sa_outs_query_remove(&query->hdr.outs_query);
		kfree(mad);
		kfree(query);
		goto proc_exit;
	}
	
	*transaction_id = temp_tid;

proc_exit:	
	return ret;
}
