/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: sa_client_path_record.c 477 2004-07-16 16:40:34Z roland $
*/

#include "sa_client.h"

#include "ts_ib_mad.h"

#include "ts_kernel_trace.h"

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/errno.h>

#include <asm/system.h>
#include <asm/byteorder.h>
#include "sa_client_priv.h"

struct ib_sa_path_record {
	u32 reserved1[2];
	tTS_IB_GID dgid;
	tTS_IB_GID sgid;
	u16 dlid;
	u16 slid;
	u32 raw_flowlabel_hoplimit;
	u8 traffic_class;
	u8 numb_path;
	u16 pkey;
	u16 sl;
	u8 mtu;
	u8 rate;
	u8 packet_life;
	u8 preference;
	u8 reserved2[6];
} __attribute__ ((packed));

struct ib_sa_path_record_query {
	struct ib_sa_query_hdr hdr;
	tTS_IB_PATH_RECORD_COMPLETION_FUNC completion_func;
	void *completion_arg;
};

static void _tsIbPathRecordResponse(tTS_IB_CLIENT_RESPONSE_STATUS status,
				    struct ib_mad *packet, void *query_ptr)
{
	struct ib_sa_path_record_query *query = query_ptr;

	switch (status) {
	case TS_IB_CLIENT_RESPONSE_OK:
		{
			struct ib_sa_payload *sa_payload =
			    (struct ib_sa_payload *) & packet->payload;
			struct ib_sa_path_record *path_ptr =
			    (struct ib_sa_path_record *) sa_payload->admin_data;
			struct ib_path_record path_rec;

			memcpy(path_rec.dgid, path_ptr->dgid,
			       sizeof path_rec.dgid);
			memcpy(path_rec.sgid, path_ptr->sgid,
			       sizeof path_rec.sgid);

			path_rec.dlid = be16_to_cpu(path_ptr->dlid);
			path_rec.slid = be16_to_cpu(path_ptr->slid);
			path_rec.flowlabel = 
			   (be32_to_cpu(path_ptr->raw_flowlabel_hoplimit)>>8) &
			   0xfffff;
			path_rec.hoplmt = 
			   be32_to_cpu(path_ptr->raw_flowlabel_hoplimit) &
			   0xff;
			path_rec.tclass = path_ptr->traffic_class;
			path_rec.pkey = be16_to_cpu(path_ptr->pkey);
			path_rec.sl = be16_to_cpu(path_ptr->sl) & 0xf;
			path_rec.mtu = path_ptr->mtu & 0x3f;
			path_rec.rate = path_ptr->rate & 0x3f;
			path_rec.packet_life = path_ptr->packet_life & 0x3f;
			path_rec.preference = path_ptr->preference;

			TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
				 "SA client path record response: "
				 "dlid=0x%04x slid=0x%04x sl=%d pkey=0x%04x",
				 path_rec.dlid,
				 path_rec.slid, path_rec.sl, path_rec.pkey);

			if (query->completion_func) {
				query->completion_func(query->hdr.transaction_id,
						       0,
						       &path_rec,
						       0,
						       query->completion_arg);
			}
		}
		break;

	case TS_IB_CLIENT_RESPONSE_ERROR:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client path record MAD status 0x%04x",
			 be16_to_cpu(packet->status));
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -EINVAL,
					       NULL, 0, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_TIMEOUT:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client path record query timed out");
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -ETIMEDOUT,
					       NULL, 0, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_CANCEL:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client path record query canceled");
		break;

	default:
		TS_REPORT_WARN(MOD_KERNEL_IB, "Unknown status %d", status);
		break;
	}

	ib_sa_outs_query_remove(&query->hdr.outs_query);
	kfree(query);
}

int tsIbPathRecordRequest(struct ib_device *device,
			  tTS_IB_PORT port,
			  tTS_IB_GID sgid,
			  tTS_IB_GID dgid,
			  u16 pkey,
			  int flags,
			  int timeout_jiffies,
			  int cache_jiffies,
			  tTS_IB_PATH_RECORD_COMPLETION_FUNC completion_func,
			  void *completion_arg,
			  tTS_IB_CLIENT_QUERY_TID * transaction_id)
{
	struct ib_mad *mad;
	struct ib_sa_payload *sa_payload = NULL;
	struct ib_sa_path_record *path_rec =  NULL;
	struct ib_sa_path_record_query *query;
	int ret = 0;
	tTS_IB_CLIENT_QUERY_TID temp_tid;

	query = kmalloc(sizeof *query, GFP_ATOMIC);
	if (!query) {
		return -ENOMEM;
	}

	mad = kmalloc(sizeof *mad, GFP_ATOMIC);
	if (!mad) {
		kfree(query);
		return -ENOMEM;
	}
	
	sa_payload = (struct ib_sa_payload *) & mad->payload;
	path_rec = (struct ib_sa_path_record *) sa_payload->admin_data;

	tsIbSaClientMadInit(mad, device, port);

	mad->r_method = IB_MGMT_METHOD_GET;
	mad->attribute_id = cpu_to_be16(TS_IB_SA_ATTRIBUTE_PATH_RECORD);
	mad->attribute_modifier = 0xffffffff;	/* match attributes */

	temp_tid = query->hdr.transaction_id = mad->transaction_id;
	query->completion_func = completion_func;
	query->completion_arg = completion_arg;

	sa_payload->sa_header.component_mask = cpu_to_be64(0x100cULL);	/* num_path, source and dest gid */

	memcpy(path_rec->sgid, sgid, sizeof(tTS_IB_GID));
	memcpy(path_rec->dgid, dgid, sizeof(tTS_IB_GID));
	path_rec->numb_path = 1;
	if (pkey != TS_IB_SA_INVALID_PKEY) {
		sa_payload->sa_header.component_mask |= cpu_to_be64(0x2000ULL);	/* pkey */
		path_rec->pkey = cpu_to_be16(pkey);
	}

	ib_sa_outs_query_add(&query->hdr.outs_query);
	ret = ib_client_query(mad, timeout_jiffies, _tsIbPathRecordResponse, query);
	if (ret) {
		ib_sa_outs_query_remove(&query->hdr.outs_query);
		kfree(mad);
		kfree(query);
		goto proc_exit;
	}

	*transaction_id = temp_tid;

proc_exit:
	return ret;
}


int ib_get_path_record_from_lids(struct ib_device *device,
				 tTS_IB_PORT port,
				 u16 slid,
				 u16 dlid,
				 int timeout_jiffies,
				 tTS_IB_PATH_RECORD_COMPLETION_FUNC completion_func,
				 void *completion_arg,
				 tTS_IB_CLIENT_QUERY_TID * transaction_id)
{
	struct ib_mad *mad = NULL;
	struct ib_sa_payload *sa_payload = NULL;
	struct ib_sa_path_record *path_rec = NULL;
	struct ib_sa_path_record_query *query = NULL;
	int ret = 0;
	tTS_IB_CLIENT_QUERY_TID temp_tid;

	query = kmalloc(sizeof *query, GFP_ATOMIC);
	if (!query) {
		return -ENOMEM;
	}

	mad = kmalloc(sizeof *mad, GFP_ATOMIC);
	if (!mad) {
		kfree(query);
		return -ENOMEM;
	}
	
	TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
		 "SA client path record query from LIDs 0x%x 0x%x",
		 slid, dlid);

	sa_payload = (struct ib_sa_payload *) & mad->payload;
	path_rec = (struct ib_sa_path_record *) sa_payload->admin_data;

	tsIbSaClientMadInit(mad, device, port);

	mad->r_method = IB_MGMT_METHOD_GET;
	mad->attribute_id = cpu_to_be16(TS_IB_SA_ATTRIBUTE_PATH_RECORD);
	mad->attribute_modifier = 0xffffffff;	/* match attributes */

	temp_tid = query->hdr.transaction_id = mad->transaction_id;
	query->completion_func = completion_func;
	query->completion_arg = completion_arg;

	/* num_path, source and dest gid */
	sa_payload->sa_header.component_mask = cpu_to_be64(0x1030ULL);

	path_rec->numb_path = 1;
	path_rec->slid = cpu_to_be16(slid);
	path_rec->dlid = cpu_to_be16(dlid);

	ib_sa_outs_query_add(&query->hdr.outs_query);
	ret = ib_client_query(mad, timeout_jiffies, _tsIbPathRecordResponse, query);
	if (ret) {
		ib_sa_outs_query_remove(&query->hdr.outs_query);
		kfree(mad);
		kfree(query);
		goto proc_exit;
	}

	*transaction_id = temp_tid;

proc_exit:	
	return ret;
}
