/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: sa_client_service.c 477 2004-07-16 16:40:34Z roland $
*/

#include "sa_client.h"

#include "ts_ib_mad.h"

#include "ts_kernel_trace.h"

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <asm/system.h>
#include <asm/byteorder.h>
#include "sa_client_priv.h"

#define TS_IB_SA_SR_SET_COMPONENT_MASK      0x57
#define TS_IB_SA_ATS_SET_COMPONENT_MASK     0x57
#define TS_IB_SA_ATS_GET_IP_COMPONENT_MASK  0x43
#define TS_IB_SA_ATS_GET_GID_COMPONENT_MASK 0x7FFFC1
#define TS_IB_SA_ATS_DAPL_SERVICE_ID        0x10000CE100415453ull	/* DAPL ATS */
#define TS_IB_SA_ATS_DEFAULT_PKEY           0xFFFF
#define TS_IB_SA_ATS_DEFAULT_LEASE          0xFFFFFFFF	/* indefinite */
#define TS_IB_SA_ATS_SERVICE_NAME           "DAPL Address Translation Service"

struct ib_sa_service {
	u64 service_id;
	tTS_IB_GID service_gid;
	u16 service_pkey;

	u16 reserved;

	u32 service_lease;

	u8 service_key[16];

	u8 service_name[64];

	u8 service_data8[16];
	u16 service_data16[8];
	u32 service_data32[4];
	u64 service_data64[2];
} __attribute__ ((packed));

struct ib_sa_service_query {
	struct ib_sa_query_hdr hdr;
	ib_service_completion_func completion_func;
	void *completion_arg;
};

struct ib_sa_service_get_gid_query {
	struct ib_sa_query_hdr hdr;
	tTS_IB_SERVICE_GET_GID_COMPLETION_FUNC completion_func;
	void *completion_arg;
};

struct ib_sa_service_get_ip_query {
	struct ib_sa_query_hdr hdr;
	tTS_IB_SERVICE_GET_IP_COMPLETION_FUNC completion_func;
	void *completion_arg;
};

static void ib_service_response(tTS_IB_CLIENT_RESPONSE_STATUS status,
				struct ib_mad *packet, void *query_ptr)
{
	struct ib_sa_service_query *query = query_ptr;

	switch (status) {
	case TS_IB_CLIENT_RESPONSE_OK:
		{
			struct ib_sa_payload *sa_payload =
			    (struct ib_sa_payload *) & packet->payload;
			struct ib_sa_service *mad_service =
			    (struct ib_sa_service *) sa_payload->admin_data;
			struct ib_common_attrib_service *service;

			TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
				 "SA client set service  status OK\n");

			service = kmalloc(sizeof *service,GFP_ATOMIC);
			if (!service) {
				TS_REPORT_WARN(MOD_KERNEL_IB, "%s: kmalloc failure for service struct",__func__);
				if (query->completion_func)
					query->completion_func(query->hdr.transaction_id,
							       -ENOMEM,
							       NULL, query->completion_arg);
				break;
			}
			memcpy(service->service_gid, mad_service->service_gid,
			       sizeof(tTS_IB_GID));
			memcpy(service->service_key, mad_service->service_key,
			       16);
			memcpy(service->service_name, mad_service->service_name,
			       16);
			service->service_pkey =
			    be16_to_cpu(mad_service->service_pkey);
			service->service_id =
			    be64_to_cpu(mad_service->service_id);
			service->service_lease =
			    be32_to_cpu(mad_service->service_lease);

			if (query->completion_func)
				query->completion_func(query->hdr.transaction_id,
						       0,
						       service,
						       query->completion_arg);
			kfree(service);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_ERROR:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client service  MAD status 0x%04x",
			 be16_to_cpu(packet->status));
		if (query->completion_func)
			query->completion_func(query->hdr.transaction_id,
					       -EINVAL,
					       NULL, query->completion_arg);
		break;

	case TS_IB_CLIENT_RESPONSE_TIMEOUT:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client multicast member query timed out");
		if (query->completion_func)
			query->completion_func(query->hdr.transaction_id,
					       -ETIMEDOUT,
					       NULL, query->completion_arg);
		break;

	case TS_IB_CLIENT_RESPONSE_CANCEL:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client service  query canceled");
		break;

	default:
		TS_REPORT_WARN(MOD_KERNEL_IB, "Unknown status %d", status);
		break;
	}

	ib_sa_outs_query_remove(&query->hdr.outs_query);
	kfree(query);
}

int ib_service_op(struct ib_device *device,
		  tTS_IB_PORT port,
		  struct ib_common_attrib_service *service,
		  u64 comp_mask,
		  int timeout_jiffies,
		  ib_service_completion_func completion_func,
		  void *completion_arg,
		  tTS_IB_CLIENT_QUERY_TID * transaction_id, u8 method)
{
	struct ib_mad *mad;
	struct ib_sa_service_query *query;
	struct ib_sa_payload *sa_payload;
	struct ib_sa_service *mad_service;
	int ret = 0;

	mad = kmalloc(sizeof *mad, GFP_ATOMIC);
	if (!mad) {
		return -ENOMEM;
	}
	query = kmalloc(sizeof *query, GFP_ATOMIC);
	if (!query) {
		kfree(mad);
		return -ENOMEM;
	}

	sa_payload = (struct ib_sa_payload *) & mad->payload;
	mad_service = (struct ib_sa_service *) sa_payload->admin_data;

	/* MAD initialization */
	tsIbSaClientMadInit(mad, device, port);
	mad->r_method = method;
	mad->attribute_id = cpu_to_be16(TS_IB_SA_ATTRIBUTE_SERVICE_RECORD);
	mad->attribute_modifier = 0;

	/* SA header */
	sa_payload->sa_header.component_mask = cpu_to_be64(comp_mask);

	/* fill in data */
	memcpy(mad_service->service_gid, service->service_gid,
	       sizeof(tTS_IB_GID));
	mad_service->service_id = cpu_to_be64(service->service_id);
	mad_service->service_pkey = cpu_to_be16(service->service_pkey);
	mad_service->service_lease = cpu_to_be32(service->service_lease);
	memcpy(mad_service->service_key, service->service_key, 16);
	memcpy(mad_service->service_name, service->service_name, 64);
	memcpy(mad_service->service_data8, service->service_data8, 16);
	memcpy(mad_service->service_data16, service->service_data16, 16);
	memcpy(mad_service->service_data32, service->service_data32, 16);
	memcpy(mad_service->service_data64, service->service_data64, 16);

	/* left here just in case some user did not check return code,
	 * and used old version of procedure */
	*transaction_id = mad->transaction_id;

	query->hdr.transaction_id = mad->transaction_id;
	query->completion_func = completion_func;
	query->completion_arg = completion_arg;

	ib_sa_outs_query_add(&query->hdr.outs_query);
	ret = ib_client_query(mad, timeout_jiffies, ib_service_response, query);
	if (ret) {
		ib_sa_outs_query_remove(&query->hdr.outs_query);
		kfree(mad);
		kfree(query);
		goto proc_exit;
	}

proc_exit:	
	return ret;
}

int ib_service_get(struct ib_device *device,
		   tTS_IB_PORT port,
		   struct ib_common_attrib_service *service,
		   u64 comp_mask,
		   int timeout_jiffies,
		   ib_service_completion_func completion_func,
		   void *completion_arg,
		   tTS_IB_CLIENT_QUERY_TID * transaction_id)
{
	return ib_service_op(device, port, service, comp_mask,
			     timeout_jiffies, completion_func,
			     completion_arg, transaction_id,
			     IB_MGMT_METHOD_GET);
}

int ib_service_set(struct ib_device *device,
		   tTS_IB_PORT port,
		   struct ib_common_attrib_service *service,
		   u64 comp_mask,
		   int timeout_jiffies,
		   ib_service_completion_func completion_func,
		   void *completion_arg,
		   tTS_IB_CLIENT_QUERY_TID * transaction_id)
{
	return ib_service_op(device, port, service, comp_mask,
			     timeout_jiffies, completion_func,
			     completion_arg, transaction_id,
			     IB_MGMT_METHOD_SET);
}

int ib_service_delete(struct ib_device *device,
		      tTS_IB_PORT port,
		      struct ib_common_attrib_service *service,
		      u64 comp_mask,
		      int timeout_jiffies,
		      ib_service_completion_func completion_func,
		      void *completion_arg,
		      tTS_IB_CLIENT_QUERY_TID * transaction_id)
{
	return ib_service_op(device, port, service, comp_mask,
			     timeout_jiffies, completion_func,
			     completion_arg, transaction_id,
			     IB_SA_METHOD_DELETE);
}

static void _tsIbServiceAtsGetGidResponse(tTS_IB_CLIENT_RESPONSE_STATUS status,
					  struct ib_mad *packet,
					  void *query_ptr)
{
	struct ib_sa_service_get_gid_query *query = query_ptr;
	tTS_IB_GID gid;

	memset(gid, 0, sizeof(gid));

	switch (status) {
	case TS_IB_CLIENT_RESPONSE_OK:
		{
			struct ib_sa_payload *sa_payload =
			    (struct ib_sa_payload *) & packet->payload;
			struct ib_sa_service *mad_service =
			    (struct ib_sa_service *) sa_payload->admin_data;

			TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
				 "SA client tsIbServiceAtsGetGidResponse() status OK\n");

			memcpy(gid, mad_service->service_gid, sizeof(gid));

			if (query->completion_func) {
				query->completion_func(query->hdr.transaction_id,
						       0,
						       gid,
						       query->completion_arg);
			}
		}
		break;

	case TS_IB_CLIENT_RESPONSE_ERROR:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client tsIbServiceAtsGetGidResponse status 0x%04x",
			 be16_to_cpu(packet->status));
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -EINVAL,
					       gid, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_TIMEOUT:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client tsIbServiceAtsGetGidResponse() query timed out");
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -ETIMEDOUT,
					       gid, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_CANCEL:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client service  query canceled");
		break;

	default:
		TS_REPORT_WARN(MOD_KERNEL_IB, "Unknown status %d", status);
		break;
	}

	ib_sa_outs_query_remove(&query->hdr.outs_query);
	kfree(query);
}

static void _tsIbServiceAtsGetIpResponse(tTS_IB_CLIENT_RESPONSE_STATUS status,
					 struct ib_mad *packet, void *query_ptr)
{
	struct ib_sa_service_get_ip_query *query = query_ptr;
	tTS_IB_ATS_IP_ADDR ip_addr;

	memset(ip_addr, 0, sizeof(ip_addr));

	switch (status) {
	case TS_IB_CLIENT_RESPONSE_OK:
		{
			struct ib_sa_payload *sa_payload =
			    (struct ib_sa_payload *) & packet->payload;
			struct ib_sa_service *mad_service =
			    (struct ib_sa_service *) sa_payload->admin_data;

			TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
				 "SA client tsIbServiceAtsGetIpResponse() status OK\n");

			memcpy(ip_addr, mad_service->service_data8,
			       sizeof(ip_addr));

			if (query->completion_func) {
				query->completion_func(query->hdr.transaction_id,
						       0,
						       ip_addr,
						       query->completion_arg);
			}
		}
		break;

	case TS_IB_CLIENT_RESPONSE_ERROR:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client tsIbServiceAtsGetIpResponse status 0x%04x",
			 be16_to_cpu(packet->status));
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -EINVAL,
					       ip_addr, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_TIMEOUT:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client tsIbServiceAtsGetIpResponse() query timed out");
		if (query->completion_func) {
			query->completion_func(query->hdr.transaction_id,
					       -ETIMEDOUT,
					       ip_addr, query->completion_arg);
		}
		break;

	case TS_IB_CLIENT_RESPONSE_CANCEL:
		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "SA client service  query canceled");
		break;

	default:
		TS_REPORT_WARN(MOD_KERNEL_IB, "Unknown status %d", status);
		break;
	}

	ib_sa_outs_query_remove(&query->hdr.outs_query);
	kfree(query);
}

int tsIbAtsServiceSet(struct ib_device *device,
		      tTS_IB_PORT port,
		      tTS_IB_GID gid,
		      tTS_IB_ATS_IP_ADDR ip_addr,
		      int timeout_jiffies,
		      ib_service_completion_func completion_func,
		      void *completion_arg,
		      tTS_IB_CLIENT_QUERY_TID * transaction_id)
{
	struct ib_mad *mad;
	struct ib_sa_payload *sa_payload = NULL;
	struct ib_sa_service *service =	NULL;
	struct ib_sa_service_query *query;
	int ret = 0;
	tTS_IB_CLIENT_QUERY_TID temp_tid;

	/* split into two calls for stack reduction */
	TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
		 "SA client register ATS service record: "
		 "GID=%02x%02x%02x%02x%02x%02x%02x%02x"
		 "%02x%02x%02x%02x%02x%02x%02x%02x\n",
		 gid[0], gid[1], gid[2], gid[3], gid[4], gid[5], gid[6], gid[7],
		 gid[8], gid[9], gid[10], gid[11], gid[12], gid[13], gid[14],
		 gid[15]);
	TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
		 "SA client register ATS service record: "
		 "IP=%02x%02x%02x%02x%02x%02x%02x%02x"
		 "%02x%02x%02x%02x%02x%02x%02x%02x\n",
		 ip_addr[0], ip_addr[1], ip_addr[2], ip_addr[3],
		 ip_addr[4], ip_addr[5], ip_addr[6], ip_addr[7], ip_addr[8],
		 ip_addr[9], ip_addr[10], ip_addr[11], ip_addr[12], ip_addr[13],
		 ip_addr[14], ip_addr[15]);
	
	mad = kmalloc(sizeof *mad, GFP_ATOMIC);
	if (!mad) {
		return -ENOMEM;
	}
	
	sa_payload = (struct ib_sa_payload *) & mad->payload;
	service = (struct ib_sa_service *) sa_payload->admin_data;

        /* MAD initialization */
	tsIbSaClientMadInit(mad, device, port);
	mad->r_method = IB_MGMT_METHOD_SET;
	mad->attribute_id = cpu_to_be16(TS_IB_SA_ATTRIBUTE_SERVICE_RECORD);
	mad->attribute_modifier = 0;

	/* SA header */
	sa_payload->sa_header.component_mask =
	    cpu_to_be64(TS_IB_SA_ATS_SET_COMPONENT_MASK);

	/* Construct service */
	memset(service, 0, sizeof(struct ib_common_attrib_service));
	service->service_id = cpu_to_be64(TS_IB_SA_ATS_DAPL_SERVICE_ID);
	memcpy(service->service_gid, gid, sizeof(tTS_IB_GID));
	service->service_pkey = cpu_to_be16(TS_IB_SA_ATS_DEFAULT_PKEY);
	service->service_lease = cpu_to_be32(TS_IB_SA_ATS_DEFAULT_LEASE);
	strcpy(service->service_name, TS_IB_SA_ATS_SERVICE_NAME);
	memcpy(service->service_data8, ip_addr, sizeof(tTS_IB_ATS_IP_ADDR));

	/* construct query */
	query = kmalloc(sizeof *query, GFP_ATOMIC);
	if (!query) {
		kfree(mad);
		return -ENOMEM;
	}
	temp_tid = query->hdr.transaction_id = mad->transaction_id;
	query->completion_func = completion_func;
	query->completion_arg = completion_arg;

	/* Subscribe/unsubscribe to SA */
	ib_sa_outs_query_add(&query->hdr.outs_query);
	ret = ib_client_query(mad, timeout_jiffies, ib_service_response, query);
	if (ret) {
		ib_sa_outs_query_remove(&query->hdr.outs_query);
		kfree(mad);
		kfree(query);
		goto proc_exit;
	}
	*transaction_id = temp_tid;

proc_exit:	
	return ret;
}

int tsIbAtsServiceGetGid(struct ib_device *device,
			 tTS_IB_PORT port,
			 tTS_IB_ATS_IP_ADDR ip_addr,
			 int timeout_jiffies,
			 tTS_IB_SERVICE_GET_GID_COMPLETION_FUNC completion_func,
			 void *completion_arg,
			 tTS_IB_CLIENT_QUERY_TID * transaction_id)
{
	struct ib_mad *mad;
	struct ib_sa_payload *sa_payload = NULL;
	struct ib_sa_service *service = NULL;
	struct ib_sa_service_get_gid_query *query;
	int ret = 0;
	tTS_IB_CLIENT_QUERY_TID temp_tid;

	TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
		 "SA client get ATS GID: "
		 "IP=%02x%02x%02x%02x%02x%02x%02x%02x"
		 "%02x%02x%02x%02x%02x%02x%02x%02x\n",
		 ip_addr[0], ip_addr[1], ip_addr[2], ip_addr[3], ip_addr[4],
		 ip_addr[5], ip_addr[6], ip_addr[7], ip_addr[8], ip_addr[9],
		 ip_addr[10], ip_addr[11], ip_addr[12], ip_addr[13],
		 ip_addr[14], ip_addr[15]);

	mad = kmalloc(sizeof *mad, GFP_ATOMIC);
	if (!mad) {
		return -ENOMEM;
	}
	
	sa_payload = (struct ib_sa_payload *) &mad->payload;
	service = (struct ib_sa_service *) sa_payload->admin_data;

	/* MAD initialization */
	tsIbSaClientMadInit(mad, device, port);
	mad->r_method = IB_MGMT_METHOD_GET;
	mad->attribute_id = cpu_to_be16(TS_IB_SA_ATTRIBUTE_SERVICE_RECORD);
	mad->attribute_modifier = 0;

	/* SA header */
	sa_payload->sa_header.component_mask =
	    cpu_to_be64(TS_IB_SA_ATS_GET_GID_COMPONENT_MASK);

	/* Construct service */
	memset(service, 0, sizeof(struct ib_common_attrib_service));
	service->service_id = cpu_to_be64(TS_IB_SA_ATS_DAPL_SERVICE_ID);
	strcpy(service->service_name, TS_IB_SA_ATS_SERVICE_NAME);
	memcpy(service->service_data8, ip_addr, sizeof(tTS_IB_ATS_IP_ADDR));

	/* construct query */
	query = kmalloc(sizeof *query, GFP_ATOMIC);
	if (!query) {
		kfree(mad);
		return -ENOMEM;
	}
	temp_tid = query->hdr.transaction_id = mad->transaction_id;
	query->completion_func = completion_func;
	query->completion_arg = completion_arg;

	/* Subscribe/unsubscribe to SA */
	ib_sa_outs_query_add(&query->hdr.outs_query);
	ret = ib_client_query(mad, timeout_jiffies, _tsIbServiceAtsGetGidResponse,
			query);
	if (ret) {
		ib_sa_outs_query_remove(&query->hdr.outs_query);
		kfree(mad);
		kfree(query);
		goto proc_exit;
	}
	*transaction_id = temp_tid;

proc_exit:	
	return ret;
}

int tsIbAtsServiceGetIp(struct ib_device *device,
			tTS_IB_PORT port,
			tTS_IB_GID gid,
			int timeout_jiffies,
			tTS_IB_SERVICE_GET_IP_COMPLETION_FUNC completion_func,
			void *completion_arg,
			tTS_IB_CLIENT_QUERY_TID * transaction_id)
{
	struct ib_mad *mad;
	struct ib_sa_payload *sa_payload = NULL;
	struct ib_sa_service *service = NULL;
	struct ib_sa_service_get_ip_query *query;
	int ret = 0;
	tTS_IB_CLIENT_QUERY_TID temp_tid;

	TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
		 "SA client register ATS service record: "
		 "GID=%02x%02x%02x%02x%02x%02x%02x%02x"
		 "%02x%02x%02x%02x%02x%02x%02x%02x ",
		 gid[0], gid[1], gid[2], gid[3], gid[4], gid[5], gid[6], gid[7],
		 gid[8], gid[9], gid[10], gid[11], gid[12], gid[13], gid[14],
		 gid[15]);

	mad = kmalloc(sizeof *mad, GFP_ATOMIC);
	if (!mad) {
		return -ENOMEM;
	}
	
	sa_payload = (struct ib_sa_payload *) &mad->payload;
	service = (struct ib_sa_service *) sa_payload->admin_data;

	/* MAD initialization */
	tsIbSaClientMadInit(mad, device, port);
	mad->r_method = IB_MGMT_METHOD_GET;
	mad->attribute_id = cpu_to_be16(TS_IB_SA_ATTRIBUTE_SERVICE_RECORD);
	mad->attribute_modifier = 0;

	/* SA header */
	sa_payload->sa_header.component_mask =
	    cpu_to_be64(TS_IB_SA_ATS_GET_IP_COMPONENT_MASK);

	/* Construct service */
	memset(service, 0, sizeof(struct ib_common_attrib_service));
	service->service_id = cpu_to_be64(TS_IB_SA_ATS_DAPL_SERVICE_ID);
	strcpy(service->service_name, TS_IB_SA_ATS_SERVICE_NAME);
	memcpy(service->service_gid, gid, sizeof(tTS_IB_GID));

	/* construct query */
	query = kmalloc(sizeof *query, GFP_ATOMIC);
	if (!query) {
		kfree(mad);
		return -ENOMEM;
	}
	temp_tid = query->hdr.transaction_id = mad->transaction_id;
	query->completion_func = completion_func;
	query->completion_arg = completion_arg;

	/* Subscribe/unsubscribe to SA */
	ib_sa_outs_query_add(&query->hdr.outs_query);
	ret = ib_client_query(mad, timeout_jiffies, _tsIbServiceAtsGetIpResponse,
			query);
	if (ret) {
		ib_sa_outs_query_remove(&query->hdr.outs_query);
		kfree(mad);
		kfree(query);
		goto proc_exit;
	}
	*transaction_id = temp_tid;

proc_exit:	
	return ret;
}

/*
  Local Variables:
  c-file-style: "linux"
  indent-tabs-mode: t
  End:
*/
