/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: useraccess_main.c 476 2004-07-16 16:30:27Z roland $
*/

#include "useraccess.h"

#include "ts_ib_core.h"

#include "ts_kernel_trace.h"
#include "ts_kernel_services.h"
#include "openib_version.h"

#include <linux/version.h>
#include <linux/module.h>

#include <linux/slab.h>
#include <linux/init.h>
#include <linux/errno.h>
#if defined(TS_KERNEL_2_6)
#include <linux/cdev.h>
#endif

/* defined in mad_main.c */
extern int ib_outs_mad_recv_entries;

MODULE_AUTHOR("Roland Dreier");
MODULE_DESCRIPTION("kernel IB userspace access" OPENIB_REVISION);
MODULE_LICENSE("Dual BSD/GPL");


enum {
	USERACCESS_NUM_DEVICE = 16
};

#if defined(TS_KERNEL_2_6)
static dev_t useraccess_devnum;
static struct cdev useraccess_cdev;
#else
static int useraccess_major;
static devfs_handle_t useraccess_devfs_dir;
#endif

static struct ib_useraccess_device useraccess_dev_list[USERACCESS_NUM_DEVICE];
static const char USERACCESS_NAME[] = "ib_useraccess";

/* =============================================================== */
/*..user_open                                                      */
static int user_open(struct inode *inode, struct file  *filp)
{
	struct ib_useraccess_device *dev;
	struct ib_useraccess_private *priv;

	int dev_num;

	if (filp->private_data) {
		dev = (struct ib_useraccess_device *) filp->private_data;
	} else {
		dev_num = MINOR(inode->i_rdev);

		TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
			 "device number %d", dev_num);

		/* not using devfs, use minor number */
		dev = &useraccess_dev_list[dev_num];
	}

	if (!dev->ib_device) {
		return -ENODEV;
	}

	priv = kmalloc(sizeof *priv, GFP_KERNEL);
	if (!priv) {
		return -ENOMEM;
	}

	priv->device = dev;
    if (!strcmp(current->comm,"opensm")) {
        priv->is_sm_process = 1;
    } else {
        priv->is_sm_process = 0;
    }
	memset(priv->port_cap_count, 0,
	       (TS_USERACCESS_MAX_PORTS_PER_DEVICE +
		1) * IB_PORT_CAP_NUM * sizeof(int));
	priv->max_mad_queue_length = ib_outs_mad_recv_entries * 2;
	TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
		       "default priv->max_mad_queue_length = %d", priv->max_mad_queue_length);
	priv->mad_queue_length = 0;
	INIT_LIST_HEAD(&priv->mad_list);
	init_waitqueue_head(&priv->mad_wait);
	sema_init(&priv->mad_sem, 1);

	filp->private_data = priv;

	return 0;
}

/* =============================================================== */
/*..user_close -                                                   */
static int user_close(struct inode *inode, struct file  *filp)
{
	struct ib_useraccess_private *priv = TS_IB_USER_PRIV_FROM_FILE(filp);
	struct ib_port_changes prop = { 0 };
	int port;
	struct ib_mad *mad = NULL;
	
	for (port = 0; port <= TS_USERACCESS_MAX_PORTS_PER_DEVICE; ++port) {
		/* Undo any port capability changes from this process */
        TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
                 "port number : %d", port);

		prop.valid_fields = 0;

		if (priv->port_cap_count[port][IB_PORT_CAP_SM] || priv->is_sm_process) {
            // Only if sm process clear IS_SM bit value , otherwise clear it .
            TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
                     "current process is SM %s",current->comm);
            prop.valid_fields |= IB_PORT_IS_SM;
        } else {
            TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
                     "current process is not SM %s - revert to  is_sm value bit\n",
                     current->comm);
        }

		if (priv->port_cap_count[port][IB_PORT_CAP_SNMP_TUN]) {
			prop.valid_fields |=
			    IB_PORT_IS_SNMP_TUNNELING_SUPPORTED;
		}

		if (priv->port_cap_count[port][IB_PORT_CAP_DEV_MGMT]) {
			prop.valid_fields |=
			    IB_PORT_IS_DEVICE_MANAGEMENT_SUPPORTED;
		}

		if (priv->port_cap_count[port][IB_PORT_CAP_VEND_CLASS]) {
			prop.valid_fields |=
			    IB_PORT_IS_VENDOR_CLASS_SUPPORTED;
		}

		if (prop.valid_fields) {
			ib_port_properties_set(priv->device->ib_device, port,
					      &prop);
		}
	
        TS_TRACE(MOD_KERNEL_IB,T_VERBOSE , TRACE_KERNEL_IB_GEN,
                 "Port %d valid field is 0x%x\n",port,prop.valid_fields);
    }

	tsIbUserFilterClear(priv);

	/* delete any outstanding MADs waiting for user-layer 'read' operation */
	if (priv) {
		while (!list_empty(&priv->mad_list)) {
			mad = list_entry(priv->mad_list.next,struct ib_mad, list);
			list_del(&mad->list);
			kfree(mad);
		}
		kfree(priv);
	}

	return 0;
}

static struct file_operations useraccess_fops = {
	.owner 	 = THIS_MODULE,
	.read 	 = tsIbUserRead,
	.write 	 = tsIbUserWrite,
	.poll 	 = tsIbUserPoll,
	.ioctl 	 = tsIbUserIoctl,
	.open 	 = user_open,
	.release = user_close
};

#if defined(TS_KERNEL_2_6)

/* Create devices for kernel 2.6 */
static int __init create_user_devices(void)
{
	int ret;

	ret = alloc_chrdev_region(&useraccess_devnum,
				  0,
				  USERACCESS_NUM_DEVICE,
				  (char *) USERACCESS_NAME);
	if (ret) {
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"Couldn't allocate device numbers for useraccess module");
		return ret;
	}

	cdev_init(&useraccess_cdev, &useraccess_fops);
	useraccess_cdev.owner = THIS_MODULE;
	ret = cdev_add(&useraccess_cdev,
		       useraccess_devnum, USERACCESS_NUM_DEVICE);
	if (ret) {
		cdev_del(&useraccess_cdev);
		unregister_chrdev_region(useraccess_devnum,
					 USERACCESS_NUM_DEVICE);
	}

	return ret;
}

#else				/* TS_KERNEL_2_6 */

/* Create devices for kernel 2.4 */
static int __init create_user_devices(void)
{
	useraccess_major = devfs_register_chrdev(0,
						 USERACCESS_NAME,
						 &useraccess_fops);
	if (useraccess_major < 0) {
		TS_REPORT_FATAL(MOD_KERNEL_IB, "Failed to register device");
		return useraccess_major;
	}

	TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
		 "TS IB user access major %d", useraccess_major);

	useraccess_devfs_dir = devfs_mk_dir(NULL, "ts_ua", NULL);

	{
		int i;
		char name[4];

		for (i = 0; i < USERACCESS_NUM_DEVICE; ++i) {
			snprintf(name, sizeof name, "%02d", i);

			useraccess_dev_list[i].devfs_handle =
			    devfs_register(useraccess_devfs_dir,
					   name,
					   DEVFS_FL_DEFAULT,
					   useraccess_major,
					   i,
					   S_IFCHR | S_IRUSR | S_IWUSR,
					   &useraccess_fops,
					   &useraccess_dev_list[i]);

			TS_TRACE(MOD_KERNEL_IB, T_VERBOSE, TRACE_KERNEL_IB_GEN,
				 "TS IB add using major %d, minor %d",
				 useraccess_major, i);

		}
	}

	return 0;
}

#endif				/* TS_KERNEL_2_6 */

static int __init user_init_module(void)
{
	int i;

	for (i = 0; i < USERACCESS_NUM_DEVICE; ++i) {
		useraccess_dev_list[i].ib_device = ib_device_get_by_index(i);
	}

	return create_user_devices();
}

#if defined(TS_KERNEL_2_6)

/* Delete devices for kernel 2.6 */
static void __exit delete_user_devices(void)
{
	unregister_chrdev_region(useraccess_devnum, USERACCESS_NUM_DEVICE);
	cdev_del(&useraccess_cdev);
}

#else				/* TS_KERNEL_2_6 */

/* Delete devices for kernel 2.4 */
static void __exit delete_user_devices(void)
{
	devfs_unregister_chrdev(useraccess_major, USERACCESS_NAME);

	{
		int i;

		for (i = 0; i < USERACCESS_NUM_DEVICE; ++i) {
			if (useraccess_dev_list[i].devfs_handle) {
				devfs_unregister(useraccess_dev_list[i].
						 devfs_handle);
			}
		}
	}

	devfs_unregister(useraccess_devfs_dir);
}

#endif				/* TS_KERNEL_2_6 */

static void __exit user_cleanup_module(void)
{
	TS_REPORT_CLEANUP(MOD_KERNEL_IB, "Unloading IB userspace access");

	delete_user_devices();

	TS_REPORT_CLEANUP(MOD_KERNEL_IB, "IB userspace access unloaded");
}

module_init(user_init_module);
module_exit(user_cleanup_module);
