/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef __cmdif_h
#define __cmdif_h

#include <mtl_types.h>
#include <hh.h>
#include <cmd_types.h>
#include <dev_cmdif.h>
#include <cmdif_priv.h>

/* #define SIMULATE_HALT_HCA  1 */

#define XHH_CMDIF_INVALID_HANDLE 0xFFFFFFFF



/******************************************************************************
 *  Function: XHH_cmd_create
 *
 *  Description: Create a XHH_cmd object
 *
 *  Parameters:
 *    hw_ver(IN) hardware version
 *    cr_base(IN) physical address of the cr-space
 *    uar0_base(IN) base address of uar0 for posting commands
 *    cmd_if_p(OUT) handle to create XHH_cmd_t object
 *    inf_timeout(IN) when TRUE the cmdif will wait infinitely for the completion of a command
 *    num_cmds_outs(IN) number of commands in the air requested by the user (the real
 *                      value may be less depending on the value returned by FW)
 *
 *  Returns:
 *
 *****************************************************************************/
HH_ret_t XHH_cmd_create(XHH_hob_t hob, u_int32_t hw_ver, MT_phys_addr_t cr_base, MT_phys_addr_t uar0_base, XHH_cmd_t *cmd_if_p,
                        MT_bool inf_timeout, u_int32_t num_cmds_outs);


/******************************************************************************
 *  Function: XHH_cmd_destroy
 *
 *  Description: Destroy the XHH_cmd object
 *
 *  Parameters:
 *    cmd_if(IN) handle of the XHH_cmd_t object
 *
 *  Returns:
 *
 *****************************************************************************/
HH_ret_t XHH_cmd_destroy(XHH_cmd_t cmd_if);



/******************************************************************************
 *  Function: XHH_cmd_set_fw_props
 *
 *  Description: set params set by QUERY_FW command
 *
 *  Parameters:
 *    cmd_if(IN) handle of the XHH_cmd_t object
 *    fw_props_p(IN) pointer to queried fw props
 *
 *  Returns:
 *
 *****************************************************************************/
XHH_cmd_status_t XHH_cmd_set_fw_props(XHH_cmd_t cmd_if, XHH_fw_props_t *fw_props);

/******************************************************************************
 *  Function: XHH_cmd_set_eq
 *
 *  Description: Enable the command interfcae object to work with events
 *
 *  Parameters:
 *    cmd_if(IN) handle of the XHH_cmd_t object
 *
 *  Returns:
 *
 *****************************************************************************/
HH_ret_t XHH_cmd_set_eq(XHH_cmd_t cmd_if);


/******************************************************************************
 *  Function: XHH_cmd_clr_eq
 *
 *  Description: inform the object to stop reporting completions to the EQ
 *
 *  Parameters:
 *    cmd_if(IN) handle of the XHH_cmd_t object
 *
 *  Returns:
 *
 *****************************************************************************/
HH_ret_t XHH_cmd_clr_eq(XHH_cmd_t cmd_if);


/******************************************************************************
 *  Function: XHH_cmd_eventh
 *
 *  Description: This function is called whenever a command was completed (when
 *               working with events)
 *
 *  Parameters:
 *    cmd_if(IN) handle of the XHH_cmd_t object
 *    result_hcr_image_p(IN) - the (big-endian) HCR image of the command completion EQE
 *
 *  Returns:
 *
 *****************************************************************************/
void XHH_cmd_eventh(XHH_cmd_t cmd_if, u_int32_t *result_hcr_image_p);


/******************************************************************************
 *  Function: XHH_cmd_assign_lamm
 *
 *  Description: Assign memory manager to be used by this object for allocating
 *               memory frm DDR
 *
 *  Parameters:
 *    cmd_if(IN) handle of the XHH_cmd_t object
 *    lamm(IN) handle of assigned ddr memory manager
 *
 *  Returns:
 *
 *****************************************************************************/
HH_ret_t XHH_cmd_assign_lamm(XHH_cmd_t cmd_if, XHH_lamm_t lamm);


/******************************************************************************
 *  Function: XHH_cmd_revoke_lamm
 *
 *  Description: revoke the associated memory manager
 *
 *  Parameters:
 *    cmd_if(IN) handle of the XHH_cmd_t object
 *
 *  Returns:
 *
 *****************************************************************************/
HH_ret_t XHH_cmd_revoke_lamm(XHH_cmd_t cmd_if);


/*****************************************************************************
*                                                                            *
*                           COMMAND FUNCTIONS                                *
*                                                                            *
******************************************************************************/

#ifdef IN
#undef IN
#endif
#define IN


#ifdef OUT
#undef OUT
#endif
#define OUT

/* System commands */

/* General queries */
XHH_cmd_status_t XHH_cmd_QUERY_DEV_LIM(IN XHH_cmd_t cmd_if, OUT XHH_dev_lim_t *dev_lim);
XHH_cmd_status_t XHH_cmd_QUERY_FW(IN XHH_cmd_t cmd_if, OUT XHH_fw_props_t *fw_props);
XHH_cmd_status_t XHH_cmd_QUERY_ADAPTER(IN XHH_cmd_t cmd_if, OUT XHH_adapter_props_t *adapter_props);

/* HCA initialization and maintenance commands */
XHH_cmd_status_t XHH_cmd_INIT_HCA(IN XHH_cmd_t cmd_if, IN XHH_hca_props_t *hca_props);
XHH_cmd_status_t XHH_cmd_INIT_IB(IN XHH_cmd_t cmd_if, IN IB_port_t port,
                                 IN XHH_port_init_props_t *port_init_props);
XHH_cmd_status_t XHH_cmd_QUERY_HCA(IN XHH_cmd_t cmd_if, OUT XHH_hca_props_t *hca_props);
XHH_cmd_status_t XHH_cmd_SET_IB(IN XHH_cmd_t cmd_if, IN IB_port_t port,
                                    IN XHH_set_ib_props_t *port_props);
XHH_cmd_status_t XHH_cmd_CLOSE_IB(IN XHH_cmd_t cmd_if, IN IB_port_t port);
#ifdef SIMULATE_HALT_HCA
XHH_cmd_status_t XHH_cmd_CLOSE_HCA(IN XHH_cmd_t cmd_if);
#else
XHH_cmd_status_t XHH_cmd_CLOSE_HCA(IN XHH_cmd_t cmd_if, MT_bool do_halt);
#endif
/* TPT commands */
XHH_cmd_status_t XHH_cmd_SW2HW_MPT(IN XHH_cmd_t cmd_if, IN XHH_mpt_index_t mpt_index,
                                   IN XHH_mpt_entry_t *mpt_entry);
XHH_cmd_status_t XHH_cmd_MODIFY_MPT(IN XHH_cmd_t cmd_if, IN XHH_mpt_index_t mpt_index,
                                   IN XHH_mpt_entry_t *mpt_entry, MT_bool modify_entire_entry);
XHH_cmd_status_t XHH_cmd_QUERY_MPT(IN XHH_cmd_t cmd_if, IN XHH_mpt_index_t mpt_index,
                                   OUT XHH_mpt_entry_t *mpt_entry);
XHH_cmd_status_t XHH_cmd_HW2SW_MPT(IN XHH_cmd_t cmd_if, IN XHH_mpt_index_t mpt_index,
                                   OUT XHH_mpt_entry_t *mpt_entry);
XHH_cmd_status_t XHH_cmd_READ_MTT(IN XHH_cmd_t cmd_if, IN u_int64_t mtt_pa, IN MT_size_t num_elems,
                                  OUT XHH_mtt_entry_t *mtt_entry);
XHH_cmd_status_t XHH_cmd_WRITE_MTT(IN XHH_cmd_t cmd_if, IN u_int64_t mtt_pa,  IN MT_size_t num_elems,
                                   IN XHH_mtt_entry_t *mtt_entry);
XHH_cmd_status_t XHH_cmd_SYNC_TPT(IN XHH_cmd_t cmd_if);

/* EQ commands */
XHH_cmd_status_t XHH_cmd_MAP_EQ(IN XHH_cmd_t cmd_if, IN XHH_eqn_t eqn, IN u_int64_t event_mask);
XHH_cmd_status_t XHH_cmd_SW2HW_EQ(IN XHH_cmd_t cmd_if, IN XHH_eqn_t eqn, IN XHH_eqc_t *eq_context);
XHH_cmd_status_t XHH_cmd_HW2SW_EQ(IN XHH_cmd_t cmd_if, IN XHH_eqn_t eqn, OUT XHH_eqc_t *eq_context);
XHH_cmd_status_t XHH_cmd_QUERY_EQ(IN XHH_cmd_t cmd_if, IN XHH_eqn_t eqn, OUT XHH_eqc_t *eq_context);


/* CQ commands */
XHH_cmd_status_t XHH_cmd_SW2HW_CQ(IN XHH_cmd_t cmd_if, IN HH_cq_hndl_t cqn,
                                  IN XHH_cqc_t *cq_context);
XHH_cmd_status_t XHH_cmd_HW2SW_CQ(IN XHH_cmd_t cmd_if, IN HH_cq_hndl_t cqn,
                                  OUT XHH_cqc_t *cq_context);
XHH_cmd_status_t XHH_cmd_QUERY_CQ(IN XHH_cmd_t cmd_if, IN HH_cq_hndl_t cqn,
                                  IN XHH_cqc_t *cq_context);
XHH_cmd_status_t XHH_cmd_RESIZE_CQ(XHH_cmd_t cmd_if, HH_cq_hndl_t cqn, 
                                   u_int64_t start_address, u_int32_t l_key, u_int8_t log_cq_size,
                                   u_int32_t *new_producer_index_p);
/* if given new_producer_index_p==NULL then opcode_modifier=1 (fixed resize CQ - FM issue #17002) */

/* QP/EE commands */
XHH_cmd_status_t XHH_cmd_MODIFY_QP(IN XHH_cmd_t cmd_if, IN IB_wqpn_t qpn,
                                   IN XHH_qpee_transition_t trans,
                                   IN XHH_qpee_context_t *qp_context,
                                   IN u_int32_t           optparammask);
XHH_cmd_status_t XHH_cmd_MODIFY_EE(IN XHH_cmd_t cmd_if, IN IB_eecn_t eecn,
                                   IN XHH_qpee_transition_t trans,
                                   IN XHH_qpee_context_t *ee_context,
                                   IN u_int32_t           optparammask);
XHH_cmd_status_t XHH_cmd_QUERY_QP(IN XHH_cmd_t cmd_if, IN IB_wqpn_t qpn,
                                  OUT XHH_qpee_context_t *qp_context);
XHH_cmd_status_t XHH_cmd_QUERY_EE(IN XHH_cmd_t cmd_if, IN IB_eecn_t eecn,
                                  OUT XHH_qpee_context_t *ee_context);
#if defined(MT_SUSPEND_QP)
XHH_cmd_status_t XHH_cmd_SUSPEND_QP(XHH_cmd_t cmd_if,  u_int32_t qpn, MT_bool suspend_flag);
#endif

/* Special QP commands */
XHH_cmd_status_t XHH_cmd_CONF_SPECIAL_QP(IN XHH_cmd_t cmd_if, IN VAPI_special_qp_t qp_type,
                                         IN IB_wqpn_t base_qpn);
XHH_cmd_status_t XHH_cmd_MAD_IFC(IN XHH_cmd_t cmd_if, 
                                 IN MT_bool mkey_validate, 
                                 IN IB_lid_t slid, /* SLID is ignored if mkey_validate is false */
                                 IN IB_port_t port,
                                 IN void *mad_in, 
                                 OUT void *mad_out);

/* SRQ commands */

XHH_cmd_status_t XHH_cmd_SW2HW_SRQ(IN XHH_cmd_t cmd_if,
                                   IN u_int32_t srqn,         /* SRQ number/index */
                                   IN XHH_srq_context_t *srqc_p);/* SRQ context   */

XHH_cmd_status_t XHH_cmd_HW2SW_SRQ(IN XHH_cmd_t cmd_if,
                                   IN u_int32_t srqn,          /* SRQ number/index */
                               OUT XHH_srq_context_t *srqc_p);/* SRQ context (NULL for no output)*/

XHH_cmd_status_t XHH_cmd_QUERY_SRQ(IN XHH_cmd_t cmd_if,
                                   IN u_int32_t srqn,          /* SRQ number/index */
                                   OUT XHH_srq_context_t *srqc_p);/* SRQ context   */
XHH_cmd_status_t XHH_cmd_ARM_SRQ(IN XHH_cmd_t cmd_if,
                                 IN u_int32_t srqn,
								 IN u_int16_t lwm); 

/* Multicast groups commands */
XHH_cmd_status_t XHH_cmd_READ_MGM(IN XHH_cmd_t cmd_if, IN u_int32_t mcg_index,
                                  IN MT_size_t  max_qp_per_mcg, OUT XHH_mcg_entry_t *mcg_entry);
XHH_cmd_status_t XHH_cmd_WRITE_MGM(IN XHH_cmd_t cmd_if, IN u_int32_t mcg_index,
                                   IN MT_size_t  max_qp_per_mcg, IN XHH_mcg_entry_t *mcg_entry);
XHH_cmd_status_t XHH_cmd_MGID_HASH(IN XHH_cmd_t cmd_if, IN IB_gid_t mgid, OUT XHH_mcg_hash_t *hash_val);

/* fatal error notification */
XHH_cmd_status_t XHH_cmd_notify_fatal(IN XHH_cmd_t cmd_if, IN XHH_fatal_err_t fatal_err);
XHH_cmd_status_t XHH_cmd_handle_fatal(IN XHH_cmd_t cmd_if);
XHH_cmd_status_t XHH_cmd_INJECT_HCAMEM_ERR(XHH_cmd_t cmd_if, 
				    XHH_acc_hcamem_inj_err_inp_param_t *acc_hcamem_inj_err_param_p, 
				    XHH_acc_hcamem_inj_err_inp_mod_t *acc_hcamem_inj_err_mod_p);
const char *str_XHH_cmd_status_t(XHH_cmd_status_t status);


HH_ret_t alloc_prm_ctx(struct cmd_if_context_st *entry,
                              prms_buf_t *prm_p,
                              MT_size_t buf_sz);
HH_ret_t de_alloc_prm_ctx(struct cmd_if_context_st *entry,
                                 prms_buf_t *prms_p);


#undef IN
#undef OUT

#endif /* __cmdif_h */
