/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include "cqm.h"
#if defined(USE_STD_MEMORY)
# include <memory.h>
#endif
#include <mtl_common.h>
#include <epool.h>
#include <log2.h>
#include <cmdif.h>
#include <hob.h>
#include <hh_common.h>
#include <tptm.h>
#include <cr_types.h>

#include <cqm_priv.h>
#include <dev_cqm_inline.h>

/************************************************************************/
/************************************************************************/
/*                         interface functions                            */


/************************************************************************/
HH_ret_t  XHH_cqm_create(
  XHH_hob_t   hob,          /* IN  */
  XHH_cqm_init_t* p_init_params,
  XHH_cqm_t*  cqm_p         /* OUT */
)
{
  XHH_cmd_t       cmd_if;
  XHH_ver_info_t version;
  HH_ret_t        rc;
  TCQM_t*         cqm = 0;
  CQCM_entry_ut*  entries = 0;
  unsigned long   ncq = 1ul << p_init_params->log2_max_cq;
  unsigned long   device_num_reserved_cqs = 1ul << p_init_params->log2_rsvd_cqs;
  MTL_TRACE1("{XHH_cqm_create: hob=%p, log2_max_cq=%d, rsvd_cqs=%lu\n", 
             hob, p_init_params->log2_max_cq, device_num_reserved_cqs);

#ifdef NO_CQ_CI_DBELL
  MTL_ERROR4(MT_FLFMT("WARNING: HCA driver is in CQ-Overrun-Ignore mode !"));
#endif
  
  rc = XHH_hob_get_cmd_if(hob, &cmd_if);
  if (rc != HH_OK) {
    return rc;
  }
  
  rc= XHH_hob_get_ver_info(hob, &version);
  if (rc != HH_OK) {
    return rc;
  }

  cqm = TMALLOC(TCQM_t);
  if (!cqm ) {
    MT_COV_MEM_FLTR_START
    return HH_EAGAIN;
    MT_COV_MEM_FLTR_END
  }

  entries = ((p_init_params->log2_max_cq < 24) && (ncq > device_num_reserved_cqs)
               ? TNVMALLOC(CQCM_entry_ut, ncq)
               : NULL);
  if (!entries) {
    MT_COV_MEM_FLTR_START
    FREE(cqm);
    return HH_EAGAIN;
    MT_COV_MEM_FLTR_END
  }
      
  /* clearing is needed, but for the sake of consistency */
  memset(cqm, 0, sizeof(TCQM_t));
  memset(entries, 0, ncq * sizeof(CQCM_entry_ut));
  
  rc = XHH_hob_get_mrwm(hob, &cqm->mrwm_internal);
  if (rc != HH_OK){
      FREE(cqm);
      VFREE(entries);
      return HH_EAGAIN;
  }
      
  XHH_DEV_SPECIFIC_CQM_INIT
  
  cqm->hob           = hob;
  cqm->cmd_if        = cmd_if;
  cqm->log2_max_cq   = p_init_params->log2_max_cq;
  cqm->max_cq        = ncq;
  cqm->entries       = entries;
  cqm->flist.entries = entries;
  cqm->flist.size    = ncq;
  cqm->flist.meta    = &fl_meta;
  epool_init(&cqm->flist);
    /* reserve is simpler than using an offset */
  epool_reserve(&cqm->flist, 0, device_num_reserved_cqs); 

  cqm->cq_resize_fixed= (version.fw_ver_major >= 3);
  *cqm_p = cqm;
  
  MTL_TRACE1("}XHH_cqm_create: cqm=%p\n", cqm);
  logIfErr("XHH_cqm_create");
  return  rc;
} /* XHH_cqm_create */


/************************************************************************/
HH_ret_t  XHH_cqm_destroy(
  XHH_cqm_t  cqm,         /* IN */
  MT_bool    hca_failure  /* IN */
)
{
  HH_ret_t        rc = HH_OK;
  MTL_TRACE1("{XHH_cqm_destroy: cqm=%p, hca_failure=%d\n", cqm, hca_failure);
  if (!hca_failure)
  {
    CQCM_entry_ut*  e = cqm->entries;
    CQCM_entry_ut*  e_end = e + cqm->max_cq;
    XHH_mrwm_t      mrwm_internal = cqm->mrwm_internal;
    int             any_busy = 0;
    for (;  (e != e_end) && (rc == HH_OK);  ++e)
    {
      if (e->in_use)
      {
        HH_ret_t  mrrc = XHH_mrwm_deregister_mr(mrwm_internal, e->u.used.lkey);
        switch (mrrc)
        {
          case HH_OK:
            break;
          case HH_EINVAL:
            rc = HH_EINVAL; MTL_ERROR2(MT_FLFMT(""));
            break;
          case HH_EBUSY:
            any_busy = 1;   MTL_ERROR2(MT_FLFMT(""));  /* Cannot happen! */
            break;
          default:          MTL_ERROR2(MT_FLFMT(""));
        }
      }
    }
    if (any_busy) { rc = HH_EINVAL; } /* again... should not happen */
  }
  VFREE(cqm->entries);
  XHH_DEV_SPECIFIC_CQM_DONE
  
  FREE(cqm);
  MTL_TRACE1("}XHH_cqm_destroy\n");
  logIfErr("XHH_cqm_destroy");
  return  rc;
} /* XHH_cqm_destroy */


/************************************************************************/
HH_ret_t  XHH_cqm_create_cq(
  XHH_cqm_t               cqm,                     /* IN  */
  MOSAL_protection_ctx_t  user_protection_context, /* IN  */
  XHH_eqn_t               comp_eqn,                /* IN  */
  XHH_eqn_t               error_eqn,               /* IN  */
  XHH_cq_ul_resources_t*  cq_ul_resources_p,       /* IO  */
  HH_cq_hndl_t*           cq_p                     /* OUT */
)
{
  MT_virt_addr_t    cqe_buf = cq_ul_resources_p->cqe_buf;
  MT_virt_addr_t    unalligned_bits = cqe_buf & CQE_size_mask;
  MT_size_t  buf_sz = cq_ul_resources_p->cqe_buf_sz;
  MT_size_t  residue = buf_sz % CQE_size;
  VAPI_lkey_t    lkey;
  HH_ret_t       mr_rc = HH_ERR;
  unsigned long  n_cq_entries = (unsigned long)(buf_sz / CQE_size);
  u_int32_t      cqc_index;
  XHH_internal_mr_t  *mr_internal_p = NULL;
  HH_ret_t       rc = ((unalligned_bits == 0) && (residue == 0)
                       ? HH_OK : HH_EINVAL);
  u_int32_t new_cqn= 0xFFFFFFFF; /* Initialize to invalid CQN */


#if MAX_TRACE >= 1
  char  ulr_tbuf[256], *ulr_buf = &ulr_tbuf[0];
  MTL_TRACE1("{XHH_cqm_create_cq: cqm=%p, ctx=0x%x, Ceqn=0x%x, Eeqn=0x%x\n"
             "    %s\n", cqm, user_protection_context, comp_eqn, error_eqn,
             ulr_print(ulr_buf, cq_ul_resources_p));
#endif
  
  if (rc!= HH_OK) {
    return rc;
  }
   
  mr_internal_p = (XHH_internal_mr_t *)MALLOC(sizeof(XHH_internal_mr_t));
  if (mr_internal_p == NULL) {
       MT_COV_MEM_FLTR_START
       MTL_ERROR2(MT_FLFMT("%s: Could not allocate memory for mr_internal_p."), __func__);
       return HH_EAGAIN;
       MT_COV_MEM_FLTR_END
  }
  
  cqc_index = epool_alloc(&cqm->flist);
  if (cqc_index == EPOOL_NULL) {
	  FREE(mr_internal_p);
      MTL_ERROR2(MT_FLFMT("CQ pool is drained.\n"));
      return HH_EAGAIN;
  }
  
  XHH_DEV_SPECIFIC_CQM_CREATE_CQ

  memset(mr_internal_p, 0, sizeof(XHH_internal_mr_t));
  mr_internal_p->start        = cq_ul_resources_p->cqe_buf;
  mr_internal_p->size         = buf_sz;
  mr_internal_p->pd           = XHH_RESERVED_PD(cqm->hob);
  mr_internal_p->vm_ctx       = user_protection_context;
  mr_internal_p->force_memkey = FALSE;
  mr_internal_p->memkey       = (VAPI_lkey_t)0;

  
  mr_rc = XHH_mrwm_register_internal(cqm->mrwm_internal, mr_internal_p, &lkey);
  new_cqn= ( ++(cqm->entries[cqc_index].cqn_prefix) << cqm->log2_max_cq ) | cqc_index;
  rc = ((mr_rc == HH_OK)? sw2hw_cq(cqm->cmd_if, cq_ul_resources_p, new_cqn, n_cq_entries,
                       lkey, comp_eqn, error_eqn,cqm,cqc_index): mr_rc);
  MTL_DEBUG4(MT_FLFMT("mr=%d=%s, rc=%d=%s"), 
                 mr_rc, HH_strerror_sym(mr_rc), rc, HH_strerror_sym(rc));
  if (rc != HH_OK) {
    if (mr_rc == HH_OK){
         (void)XHH_mrwm_deregister_mr(cqm->mrwm_internal, lkey);
    }
    
    XHH_DEV_SPECIFIC_CQM_DESTROY_CQ

    epool_free(&cqm->flist,cqc_index);
    logIfErr("XHH_cqm_create_cq");
    FREE(mr_internal_p);
    return rc;
  }
  FREE(mr_internal_p);
  
  
  /* cqm->entries[cq].u.used.props = *cq_props_p; */
  cqm->entries[cqc_index].u.used.n_cq = n_cq_entries;
  cqm->entries[cqc_index].u.used.lkey = lkey;
  /* Save protection context for CQ-resize */
  cqm->entries[cqc_index].u.used.user_protection_context= user_protection_context;
  cqm->entries[cqc_index].in_use = 1;
#if defined(MT_SUSPEND_QP)
  cqm->entries[cqc_index].u.used.is_suspended = FALSE;
#endif
  *cq_p = new_cqn;

      
  MTL_TRACE1("}XHH_cqm_create_cq, cq=0x%x\n", *cq_p);
  
  return  HH_OK;
} /* XHH_cqm_create_cq */

/************************************************************************/
HH_ret_t  XHH_cqm_resize_cq(
  XHH_cqm_t               cqm,                     /* IN */
  HH_cq_hndl_t            cq,                      /* IN */
  XHH_cq_ul_resources_t*  cq_ul_resources_p        /* IO */
)
{
  MT_virt_addr_t    cqe_buf = cq_ul_resources_p->cqe_buf;
  MT_virt_addr_t    unalligned_bits = cqe_buf & CQE_size_mask;
  MT_size_t  buf_sz = cq_ul_resources_p->cqe_buf_sz;
  MT_size_t  residue = buf_sz % CQE_size;
  VAPI_lkey_t    lkey;
  unsigned long  n_cq_entries = (unsigned long)(buf_sz / CQE_size);
  CQCM_entry_ut*  sw_cqc_p;
  XHH_internal_mr_t  *mr_internal_p = NULL;
  XHH_cmd_status_t   cmd_rc;
  HH_ret_t           rc;
  
  /* Validate parameters */
  if (!in_use_cq(cqm, cq)) {
    MTL_ERROR1(MT_FLFMT("Invalid CQ handle (0x%X)"),cq);
    return HH_EINVAL_CQ_HNDL;
  }
  sw_cqc_p = &cqm->entries[cq & MASK32(cqm->log2_max_cq)];

  if ((unalligned_bits != 0) || (residue != 0) || (buf_sz == 0)) {
    MTL_ERROR1(MT_FLFMT("%s: Invalid CQEs buffer (va="VIRT_ADDR_FMT" , size="SIZE_T_FMT")"),
               __func__,cqe_buf,buf_sz);
    return HH_EINVAL;
  }

   
  mr_internal_p = (XHH_internal_mr_t *)MALLOC(sizeof(XHH_internal_mr_t));
  if (mr_internal_p == NULL) {
       MT_COV_MEM_FLTR_START
       MTL_ERROR2(MT_FLFMT("%s: Could not allocate memory for mr_internal_p."), __func__);
       return HH_EAGAIN;
       MT_COV_MEM_FLTR_END
  }
  
  /* Register new CQEs buffer */
  memset(mr_internal_p, 0, sizeof(XHH_internal_mr_t));
  mr_internal_p->start        = cq_ul_resources_p->cqe_buf;
  mr_internal_p->size         = buf_sz;
  mr_internal_p->pd           = XHH_RESERVED_PD(cqm->hob);
  mr_internal_p->vm_ctx       = sw_cqc_p->u.used.user_protection_context;
  mr_internal_p->force_memkey = FALSE;
  mr_internal_p->memkey       = (VAPI_lkey_t)0;

  rc= XHH_mrwm_register_internal(cqm->mrwm_internal, mr_internal_p, &lkey);
  if (rc != HH_OK) {
    MTL_ERROR2(MT_FLFMT("%s: Failed registering new CQEs buffer (%s)"),
               __func__,HH_strerror_sym(rc));
    FREE(mr_internal_p);
    return rc;
  }
  FREE(mr_internal_p);
  
  cmd_rc= XHH_cmd_RESIZE_CQ(cqm->cmd_if, cq, cqe_buf, lkey, (u_int8_t)floor_log2(n_cq_entries),
                            cqm->cq_resize_fixed ? NULL : &cq_ul_resources_p->new_producer_index);
  if (cmd_rc != XHH_CMD_STAT_OK) {
    MTL_ERROR2(MT_FLFMT("%s: Failed command RESIZE_CQ (%s)"),
               __func__,str_XHH_cmd_status_t(cmd_rc));
    switch (cmd_rc) {
      case XHH_CMD_STAT_BAD_SIZE:
        rc= HH_E2BIG_CQE_NUM;  /* Retry after polling some CQEs */
        break;
      case XHH_CMD_STAT_BAD_RES_STATE:  /* CQ in error state or does not exist anymore */
      case XHH_CMD_STAT_BAD_INDEX:      /* Wrong CQ number */
        rc= HH_EINVAL_CQ_HNDL;  
        break;
      case XHH_CMD_STAT_BAD_OP:
        rc= HH_ENOSYS;  /* Probably old firmware */
        break;
      case  XHH_CMD_STAT_EINTR:
        rc = HH_EINTR;
        break;
      default:
        rc= HH_EFATAL; /* Unexpected error */
        break;
    }
    (void)XHH_mrwm_deregister_mr(cqm->mrwm_internal, lkey); /* deregister new buffer */
    return rc;
  }
   
  rc= XHH_mrwm_deregister_mr(cqm->mrwm_internal, sw_cqc_p->u.used.lkey);
  if (rc != HH_OK) {
    MTL_ERROR1(MT_FLFMT("%s: Failed deregistration of old CQEs buffer (%d = %s) !!"),
               __func__,rc,HH_strerror_sym(rc));
    /* Nothing we can do about old CQEs region but anyway nobody uses it for any other resource */
  }
  /* Save new parameters of the CQ */
  sw_cqc_p->u.used.n_cq = n_cq_entries;
  sw_cqc_p->u.used.lkey = lkey;
  
  return  HH_OK;
}

/************************************************************************/
HH_ret_t  XHH_cqm_destroy_cq(
  XHH_cqm_t     cqm /* IN */,
  HH_cq_hndl_t  cq  /* IN */
)
{
  u_int32_t cqc_index= cq & MASK32(cqm->log2_max_cq);
  HH_ret_t  rc = HH_EINVAL_CQ_HNDL;
  MTL_TRACE1("{XHH_cqm_destroy_cq, cqm=%p, cq=0x%x\n", cqm, cq);
  if (in_use_cq(cqm, cq))
  {
    rc = CMDRC2HH_ND(XHH_cmd_HW2SW_CQ(cqm->cmd_if, cq, NULL));
    if ((rc == HH_OK) || (rc == HH_EFATAL))
    {
      CQCM_entry_ut*  e = &cqm->entries[cqc_index];
      XHH_cmd_status_t  mrrc = XHH_mrwm_deregister_mr(cqm->mrwm_internal,
                                                    e->u.used.lkey);
      if (mrrc != XHH_CMD_STAT_OK)
      {
        MTL_ERROR1(MT_FLFMT("%s: Failed deregistration of CQEs buffer (%d = %s) !!"),
                   __func__,mrrc, str_XHH_cmd_status_t(mrrc));
        rc = CMDRC2HH_BUSY(mrrc);
      }
      else
      {
         /* If we are in a fatal error, return OK for destruction */
        if (rc == HH_EFATAL){
            MTL_DEBUG1(MT_FLFMT("%s: in fatal error"), __func__);
            rc = HH_OK;
        }
        
        XHH_DEV_SPECIFIC_CQM_DESTROY_CQ
        
        e->in_use = 0;
        epool_free(&cqm->flist, cqc_index);

      }
    }
  }
  
  
  MTL_TRACE1("}XHH_cqm_destroy_cq\n");
  logIfErr("XHH_cqm_destroy_cq");
  return  rc;
} /* XHH_cqm_destroy_cq */


/************************************************************************/
/* Note: we actually not validating that given 'cq' is indeed in use. */
HH_ret_t  XHH_cqm_query_cq(
  XHH_cqm_t        cqm,           /* IN  */
  HH_cq_hndl_t     cq,            /* IN  */
  VAPI_cqe_num_t*  num_o_cqes_p   /* IN  */
)
{
  u_int32_t cqc_index= cq & MASK32(cqm->log2_max_cq);
  HH_ret_t  rc = (in_use_cq(cqm, cq) ? HH_OK : HH_EINVAL_CQ_HNDL);
  MTL_TRACE1("{XHH_cqm_query_cq: cqm=%p, cq=0x%x\n", cqm, cq);
  if (rc == HH_OK)  rc= (TCQM_CQN(cqm,cqc_index) == cq) ? HH_OK : HH_EINVAL_CQ_HNDL;
  if (rc == HH_OK)
  {
    *num_o_cqes_p = cqm->entries[cqc_index].u.used.n_cq;
  }
  MTL_TRACE1("}XHH_cqm_query_cq\n");
  logIfErr("XHH_cqm_query_cq");
  return  rc;
} /* XHH_cqm_query_cq */



/************************************************************************/
/* Assumed to be the first called in this module, single thread.        */
void  XHH_cqm_init(void)
{
  MTL_TRACE1("XHH_cqm_init\n");
} /* XHH_cqm_init */


/************************************************************************/
HH_ret_t  XHH_cqm_get_num_cqs(
  XHH_cqm_t  cqm,         /* IN */
  u_int32_t  *num_cqs_p   /* OUT*/
)
{
  CQCM_entry_ut*  e;
  CQCM_entry_ut*  e_end;
  int             num_alloc_cqs = 0;

  if (cqm == NULL) {
      return HH_EINVAL;
  }

  e = cqm->entries;
  e_end = e + cqm->max_cq;
  for (;  (e != e_end) ;  ++e){
      if (e->in_use){ num_alloc_cqs++; }
  }
  return  HH_OK;
} /* XHH_cqm_destroy */

#if defined(MT_SUSPEND_QP)
HH_ret_t  XHH_cqm_suspend_cq(
  XHH_cqm_t        cqm,           /* IN  */
  HH_cq_hndl_t     cq,            /* IN  */ 
  MT_bool          do_suspend     /* IN  */)
{
  u_int32_t cqc_index= cq & MASK32(cqm->log2_max_cq);
  HH_ret_t  rc = HH_EINVAL_CQ_HNDL;
  MTL_TRACE1("{XHH_cqm_suspend_cq, cqm=%p, cq=0x%x, do_suspend=%s\n",
             cqm, cq,(do_suspend==FALSE)?"FALSE":"TRUE");
  if (in_use_cq(cqm, cq))
  {
      CQCM_entry_ut*  e = &cqm->entries[cqc_index];
      MT_bool         is_suspended = e->u.used.is_suspended;
      if (do_suspend == is_suspended) {
          /* cq suspension already in desired state */
          MTL_DEBUG2(MT_FLFMT("%s: CQ 0x%x already %s"),
                     __func__,cq, (is_suspended == TRUE)?"suspended" : "unsuspended");
          return HH_OK;
      }
      rc = XHH_mrwm_suspend_internal(cqm->mrwm_internal,e->u.used.lkey, do_suspend);
      if (rc != HH_OK)
      {
        MTL_ERROR1(MT_FLFMT("%s: Failed XHH_mrwm_suspend_mr of CQEs buffer region(%d: %s) !!"),
                   __func__,rc, HH_strerror_sym(rc));
      }
      else
      {
          MTL_DEBUG2(MT_FLFMT("%s: CQ 0x%x is %s"),
                     __func__,cq, (do_suspend == TRUE)?"suspended" : "unsuspended");
          e->u.used.is_suspended = do_suspend;
      }
  }
  MTL_TRACE1("}XHH_cqm_suspend_cq\n");
  logIfErr("XHH_cqm_suspend_cq");
  return  rc;
} /* XHH_cqm_suspend_cq */
#endif
