/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <cmdif.h>
#include <mosal.h>
#include <eventp_priv.h>
#include <eventp_dev_funcs.h>
#include <hob.h>
#include <vapi_common.h>

#include <cr_types.h>

#ifdef XHH_CMD_TIME_TRACK
u_int64_t XHH_eventp_last_cmdif_interrupt;
#endif

/*================ macro definitions ===============================================*/
#define EQE_IN_SW_OWNER(eqe_buff_p)      \
  ((((u_int8_t*)(eqe_buff_p))[EQE_OWNER_BYTE_OFFSET] & EQE_HW_OWNER) == EQE_SW_OWNER)

#define CONS_CNTR_2_INDX(cons_cntr, mask) ((cons_cntr) & (mask))

/*================ type definitions ================================================*/


typedef struct eq_entry_st {
  u_int8_t event_type;
  ADD_DEV_PREFIX_LC(if_port_event_subtype_t) event_sub_type;
  u_int32_t event_data[6];      /* Delivers auxiliary data to handle event. - 24 bytes */
}eq_entry_t;


/*================ static functions prototypes =====================================*/

static void handle_eqe_ib_events(EQP_eq_entry_t *eqe_p,
                                 eq_entry_t *eq_data_p);
/*================ global functions definitions ====================================*/



/************************************************************************
 *  Function: xhh_intr_handler
 *  
 
    Arguments:
    eventp - handler to eventp object
    isr_ctx1 & isr_ctx2 - ignored
    
    
    Description:
    
    Clear_INT(); // De-assert INT pin (write to Clr_INT register) 
    foreach (EQ) { 
      if (head of EQ in SW ownership)
      then Schedule_EQ_handler(EQN);      // schedule appropriate handler 
      else arm(EQ)
    }
    
    
    
************************************************************************/


irqreturn_t xhh_intr_handler(MT_ulong_ptr_t eventp_dpc, void* isr_ctx1, void* isr_ctx2)
{
  XHH_eventp_t eventp = (XHH_eventp_t)eventp_dpc;
  int eq_num;
  EQP_eq_entry_t *event_entry;
  u_int32_t  *eqe_buff_p;
  u_int32_t mask_val=0;
    
  /* suppress warning */
  if ( 0 ) mask_val += 0;

  MTL_DEBUG1(MT_STDFMT("xhh_intr_handler called"));

  /* De-assert INT pin (write to Clr_INT register) */
  MOSAL_MMAP_IO_WRITE_DWORD(eventp->intr_clr_reg, eventp->intr_clr_mask);

  for(eq_num=0; eq_num<(int)eventp->free_eq_num; eq_num++) {      /* foreach active EQ */
    event_entry = &(eventp->eq_table[eq_num]);
	eqe_buff_p = ((u_int32_t*)event_entry->eq_buff)+ 
			     EQE_DWORD_SIZE * CONS_CNTR_2_INDX(event_entry->cons_cntr,event_entry->cons_cntr_mask);
    /* test if EQE is in SW ownership */
    if ((EQE_IN_SW_OWNER(eqe_buff_p))){
	  MTL_DEBUG1("%s: *** eq_num=%d in SW ownership  cons_cntr=%d\n", __func__, eq_num, event_entry->cons_cntr);
      if (MOSAL_DPC_schedule(&event_entry->polling_dpc)){  /* schedule DPC to poll EQ */
        atomic_inc(&event_entry->dpc_cntr); /* Monitor number of outstanding DPCs */
        #ifdef XHH_CMD_TIME_TRACK
        if (eq_num == EQP_CMD_IF_EQN) {
          XHH_eventp_last_cmdif_interrupt= MOSAL_get_time_counter();
        }
        #endif
      }
    } 
    else {
      MTL_DEBUG1("%s: eq_num=%d in HW ownership\n", __func__, eq_num);
      RE_ARM_EQ(eventp, eq_num, mask_val);
    }
        
  } /* foreach EQ */
  ARM_EQS(eventp, mask_val); /* in Arbel we can re-arm all EQs at once at the end of interrupt handler */

  return IRQ_HANDLED;
} /* xhh_intr_handler() */

/************************************************************************************************/

irqreturn_t xhh_msix_intr_handler(MT_ulong_ptr_t eventp_entry, void* isr_ctx1, void* isr_ctx2)
{
  EQP_eq_entry_t *entry = (EQP_eq_entry_t *)eventp_entry;

  if (MOSAL_DPC_schedule(&entry->polling_dpc)){  /* schedule DPC to poll EQ */
    atomic_inc(&entry->dpc_cntr); /* Monitor number of outstanding DPCs */
  }

  return IRQ_HANDLED;
}






/* inline functions must be placed BEFORE their call */
/************************************************************************************************/
inline static MT_bool read_eq_entry(u_int32_t  *eqe_buff_p,
                          EQ_type_t  eq_type,
                          eq_entry_t *eqe_p)
{
  u_int32_t tmp_eq_data[8] = {0,0,0,0,0,0,0,0}; /* entry size is 32 bytes */
  unsigned int i;


  MTL_DEBUG1("%s: eqe_buff= %x %x %x %x %x %x %x %x \n", __func__, *eqe_buff_p,
	  *(eqe_buff_p+1), *(eqe_buff_p+2), *(eqe_buff_p+3),*(eqe_buff_p+4),
	  *(eqe_buff_p+5), *(eqe_buff_p+6), *(eqe_buff_p+7));
  /* first extract the owner & event_type only*/
  
  /* if EQE not in SW ownership - end of EQEs to handle */
  if (!(EQE_IN_SW_OWNER(eqe_buff_p))){
    MTL_DEBUG1("%s: EQE_HW_OWNER\n", __func__);
    return FALSE;
  }
  rmb();

  /* event_type & event_sub_type */
  tmp_eq_data[EQE_EVENT_TYPE_OFFSET] = MOSAL_be32_to_cpu(*(eqe_buff_p+EQE_EVENT_TYPE_OFFSET));
  eqe_p->event_type = MT_EXTRACT_ARRAY32(tmp_eq_data, MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(event_queue_entry_st), event_type),
                                      MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(event_queue_entry_st), event_type));

  eqe_p->event_sub_type = (ADD_DEV_PREFIX_LC(if_port_event_subtype_t))
        MT_EXTRACT_ARRAY32(tmp_eq_data, MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(event_queue_entry_st), event_sub_type),
                                      MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(event_queue_entry_st), event_sub_type));

  /* extract of data will be done according to the event type  */
  
  if(eq_type != EQP_CMD_IF_EVENT) {
    for(i=EQE_DATA_OFFSET; i< EQE_OWNER_OFFSET; i++) {
      eqe_p->event_data[i-EQE_DATA_OFFSET] = MOSAL_be32_to_cpu(*(eqe_buff_p + i));
    }
  }
  else {   /* need to leave as big-endien for the cmd_if callback */
    memcpy(eqe_p->event_data, eqe_buff_p+EQE_DATA_OFFSET, EQE_DATA_BYTE_SIZE);
  }
  FUNC_OUT;
  return TRUE;
}

/************************************************************************************************/

inline static void move_eqe_to_hw(EQP_eq_entry_t *eqe_p, 
								  u_int32_t      *eqe_buff_p)     /* pointer to start of EQE buffer */
{
  /* move EQE to HW ownership */
  *(eqe_buff_p+EQE_OWNER_OFFSET) = 0xffffffff; /* no need for cpu to be32 since all is ff */
  
  /* update consumer counter of EQ */
  eqe_p->cons_cntr++;

  MTL_DEBUG1("%s: cons_cntr=%d \n", __func__, eqe_p->cons_cntr);\
  /* update consumer index the FW take care to the cyclic buffer update */
  EQ_UPDATE_CONS_CNTR(eqe_p->eventp_p, eqe_p);
}



/************************************************************************
 *  Function: eq_polling_dpc
 *  
 
    Arguments:
    
    Returns:
    HH_OK 
    HH_EINVAL -Invalid parameters 
    
    Description: 
    
    While (EQE[Consumer_indx].Owner == SW) { 
      consume_entry - call EQ handler; // remove entry from the queue 
      EQE[Consumer_indx++].Owner = HW; // mark entry to the Hardware ownership for next time around 
      consumer_indx &= MASK; // wrap around index 
    } 
    update_consumer_index(EQ,consumer_indx); // update consumer index in HCA via UAR 
    subscribe_for_event(EQ); // subscribe for event for next time
    unlock EQ spinlock
    
************************************************************************/

void eq_polling_dpc(DPC_CONTEXT_t *dpc_obj_p)
{
  EQP_eq_entry_t *eqe_p = (EQP_eq_entry_t *)dpc_obj_p->func_ctx; /* EQ context entry (not the EQE) */ 
  u_int32_t   cons_indx;
  eq_entry_t  eqe;
  u_int32_t   *eqe_buff_p;
  void* cyc_buff;
  
  
  cyc_buff = eqe_p->eq_buff;
  /* EQEs polling loop */
  /* need to protect only in case simultanuous DPC can be called at the same time 
   currently will be set only for windows */
#ifdef SIMULTANUOUS_DPC 
  MOSAL_spinlock_dpc_lock(&eqe_p->dpc_lock); /* protect consumer index access */
#endif

  for(cons_indx = CONS_CNTR_2_INDX(eqe_p->cons_cntr,eqe_p->cons_cntr_mask); /* break done inside the loop*/ ;
                  cons_indx = CONS_CNTR_2_INDX(cons_indx+1,eqe_p->cons_cntr_mask)) {

    eqe_buff_p = ((u_int32_t*)cyc_buff) + EQE_DWORD_SIZE*cons_indx ;
    /* read the EQE and change to machine endianess */
    if (read_eq_entry(eqe_buff_p, eqe_p->eq_type, &eqe) == FALSE){
      MTL_DEBUG1("%s: entry not in SW ownership, cons_indx=%d cyc_buff=%p\n", 
				 __func__,cons_indx, cyc_buff);
      break; /* no more EQEs to poll */
    }

    /* if we are here then we have EQE in SW ownership */

    /* first return EQE to HW ownership and update consumer counter */
    move_eqe_to_hw(eqe_p, eqe_buff_p);
    
    /* now check that we don't have overrun - valid for Tavor only (parameter - label) */
	OVERRUN_CHECK(dpc_done)
    
    MTL_DEBUG1("%s: EQ type =%d\n", __func__, eqe_p->eq_type);
    switch (eqe_p->eq_type) {
      /* handle CMD_IF EQ */
      case EQP_CMD_IF_EVENT:
        /* notify cmd_if module */
        XHH_cmd_eventh(eqe_p->eventp_p->cmd_if, (u_int32_t*)eqe.event_data);
        break;
      /* IB events */
      case EQP_IB_EVENT:
        handle_eqe_ib_events(eqe_p,&eqe);
        break;
      /* Completion Events */     
      case EQP_CQ_COMP_EVENT: 
        {
          u_int32_t   cqnum=0;
          cqnum = MT_EXTRACT_ARRAY32(eqe.event_data, 
									 MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(completion_event_st), cqn),
                                     MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(completion_event_st), cqn));
          MTL_DEBUG2("%s: Got completion event on CQN=%d. Going to DIS-ARM CQ\n", 
					 __func__,cqnum);
          /* disarm CQ */
		  DISARM_CQ(eqe_p->eventp_p, eqe_p->eqn, cqnum);
          /* call handler */
          eqe_p->handler.comp_event_h(eqe_p->eventp_p->hh_hca_hndl, 
									  (HH_cq_hndl_t)cqnum, eqe_p->priv_context);

		  MTL_DEBUG1("%s: got completion event cq_num=%d\n", __func__, cqnum);
        }
        break;
	  case EQP_MLX_EVENT:
	  case EQP_CATAS_ERR_EVENT:
	  default:
        MTL_ERROR1("%s: internal ERROR got event from wrong type\n", __func__);
        break;
    }
  } /* polling EQE for loop */

/* 	valid for Tavor only (parameter - label) */								
  OVERRUN_CHECK_FAILURE(dpc_done)
  
/* re-arm EQ for all EQs */
  ARM_EQ_AT_END(eqe_p->eventp_p, eqe_p->eqn);

#ifdef SIMULTANUOUS_DPC
  MOSAL_spinlock_unlock(&eqe_p->dpc_lock);
#endif
  atomic_dec(&eqe_p->dpc_cntr); /* signal DPC done (for cleanup) */
  
}

/************************************************************************************************/


static inline VAPI_event_record_type_t tavor2vapi_qp_error_type(u_int8_t event_type, 
																MT_bool is_qp)
{
  
  VAPI_event_record_type_t vapi_event_type;
  switch(event_type) {
    case ADD_DEV_PREFIX(IF_EV_TYPE_PATH_MIG): 
      vapi_event_type = is_qp ? VAPI_QP_PATH_MIGRATED : VAPI_EEC_PATH_MIGRATED;
      break;
    case ADD_DEV_PREFIX(IF_EV_TYPE_SEND_Q_DRAINED): 
      vapi_event_type = VAPI_SEND_QUEUE_DRAINED;
      break;
    case ADD_DEV_PREFIX(IF_EV_TYPE_PATH_MIG_FAIL): 
      vapi_event_type = VAPI_PATH_MIG_REQ_ERROR;
      break;
    case ADD_DEV_PREFIX(IF_EV_TYPE_COMM_EST):
      vapi_event_type =  is_qp ? VAPI_QP_COMM_ESTABLISHED : VAPI_EEC_COMM_ESTABLISHED;
      break;
    case ADD_DEV_PREFIX(IF_EV_TYPE_LOCAL_WQ_CATAS_ERR):
      vapi_event_type = is_qp ? VAPI_LOCAL_WQ_CATASTROPHIC_ERROR : VAPI_LOCAL_EEC_CATASTROPHIC_ERROR;
      break;
    case ADD_DEV_PREFIX(IF_EV_TYPE_LOCAL_WQ_INVALID_REQ_ERR): 
      vapi_event_type = VAPI_LOCAL_WQ_INV_REQUEST_ERROR;
      break;
    case ADD_DEV_PREFIX(IF_EV_TYPE_LOCAL_WQ_ACCESS_VIOL_ERR): 
      vapi_event_type = VAPI_LOCAL_WQ_ACCESS_VIOL_ERROR;
      break;
    case ADD_DEV_PREFIX(IF_EV_TYPE_SRQ_QP_LAST_WQE_REACHED):
      vapi_event_type = VAPI_RECEIVE_QUEUE_DRAINED;
      break;
    case ADD_DEV_PREFIX(IF_EV_TYPE_LOCAL_SRQ_CATAS_ERR):
      vapi_event_type = VAPI_SRQ_CATASTROPHIC_ERROR;
      break;
    case ADD_DEV_PREFIX(IF_EV_TYPE_PATH_MIG_ARMED):
  	  vapi_event_type = VAPI_PATH_MIG_ARMED;
  	  break;
    case ADD_DEV_PREFIX(IF_EV_TYPE_SRQ_LIMIT_REACHED):
      vapi_event_type = VAPI_SRQ_LIMIT_REACHED;
      break;
    default:
      MTL_ERROR1("%s: Unknown event type = %d\n", __func__,event_type);
      vapi_event_type = VAPI_LOCAL_CATASTROPHIC_ERROR;
  }
  return vapi_event_type;
}
         

  
/************************************************************************************************/
static inline void handle_qp_event(eq_entry_t     *eq_data_p,
								   HH_event_record_t *event_record_p)

{
  u_int32_t qpn=0;
  u_int32_t is_qp=1;
  u_int8_t fw_syndrome=0;
  
  fw_syndrome = ((u_int8_t*)eq_data_p->event_data)[14];	/* coordination with FW */
  qpn = MT_EXTRACT_ARRAY32(eq_data_p->event_data, MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_event_st), qpn_een),
                        MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_event_st), qpn_een));
  
  is_qp = !(MT_EXTRACT_ARRAY32(eq_data_p->event_data, MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_event_st), e_q),
                         MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_event_st), e_q)));

  if (is_qp) {
    event_record_p->event_modifier.qpn = qpn;
    /* need to translate to VAPI_event_record_type_t */ 
    event_record_p->etype = tavor2vapi_qp_error_type(eq_data_p->event_type,(MT_bool)is_qp);
    if (ADD_DEV_PREFIX(IF_EV_TYPE_COMM_EST) != eq_data_p->event_type) {
      MTL_ERROR1("%s: QP event on qpn= 0x%X event type is %s (%d), syndrome=0x%x\n", __func__,qpn,
  			 VAPI_event_record_sym(event_record_p->etype), event_record_p->etype,fw_syndrome);
    }
  }
  else { /* EEC is not supported now */
    MTL_ERROR1("%s: Internal error: is_qp = 0 but EEC not supported. qpn=%d\n", 
  			 __func__, qpn);
  }
}


static inline void handle_srq_event(eq_entry_t     *eq_data_p,
								   HH_event_record_t *event_record_p)
{
  event_record_p->event_modifier.srq= 
  MT_EXTRACT_ARRAY32(eq_data_p->event_data, 
                     MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(qp_ee_event_st), qpn_een),
                     MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(qp_ee_event_st), qpn_een));
  event_record_p->etype = tavor2vapi_qp_error_type(eq_data_p->event_type,FALSE);
  MTL_ERROR1("%s: SRQ event on srqn= 0x%X event type is %s (%d)\n", __func__,
     event_record_p->event_modifier.srq,
     VAPI_event_record_sym(event_record_p->etype), event_record_p->etype);
}

static inline void handle_cq_event(eq_entry_t     *eq_data_p,
								   HH_event_record_t *event_record_p)
{

  u_int32_t   cqnum=0;
  u_int8_t syndrome=0;
  
  cqnum = MT_EXTRACT_ARRAY32(eq_data_p->event_data, MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_error_st), cqn),
                          MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_error_st), cqn));
  syndrome = MT_EXTRACT_ARRAY32(eq_data_p->event_data, MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_error_st), syndrome),
                             MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_error_st), syndrome));
  event_record_p->etype = VAPI_CQ_ERROR;
  event_record_p->event_modifier.cq = cqnum;
  event_record_p->syndrome = (syndrome == ADD_DEV_PREFIX(IF_CQ_OVERRUN)) ? VAPI_CQ_ERR_OVERRUN : 
                 ((syndrome == ADD_DEV_PREFIX(IF_CQ_ACCSS_VIOL_ERR)) ? VAPI_CQ_ERR_ACCESS_VIOL : 
                   VAPI_EV_SYNDROME_NONE);
  MTL_ERROR1("%s: CQ error on CQ number= 0x%x syndrome is %s (%d)\n", __func__,cqnum,
             VAPI_event_syndrome_sym(event_record_p->syndrome), event_record_p->syndrome);
}


static inline void handle_port_event(eq_entry_t     *eq_data_p,
								   HH_event_record_t *event_record_p)
{
  IB_port_t port;
  port = MT_EXTRACT_ARRAY32(eq_data_p->event_data,MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(port_state_change_st), p),
                         MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(port_state_change_st), p));

  if (eq_data_p->event_sub_type == ADD_DEV_PREFIX(IF_SUB_EV_PORT_DOWN)) {
    event_record_p->etype = VAPI_PORT_ERROR;
  }
  else if (eq_data_p->event_sub_type == ADD_DEV_PREFIX(IF_SUB_EV_PORT_UP)) {
    event_record_p->etype = VAPI_PORT_ACTIVE;
  }
  else {
    MTL_ERROR1("%s: Wrong sub-type for Port event on port=%d sub_type=%d\n", __func__,port,eq_data_p->event_sub_type);
  }

  event_record_p->event_modifier.port = port;
  MTL_DEBUG1("%s: Port change event on port=%d sub_type=%d\n", __func__,port,eq_data_p->event_sub_type);
}


static inline void handle_gpio_event(eq_entry_t     *eq_data_p,
								   HH_event_record_t *event_record_p)
{ 
  /* gpio_lo is in the syndrom, gpio_hi is in event_modifier.qpn */
  event_record_p->syndrome = MT_EXTRACT_ARRAY32(eq_data_p->event_data, 
							MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(gpio_event_data_st), gpio_event_lo),
							MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(gpio_event_data_st), gpio_event_lo));
  event_record_p->event_modifier.qpn = MT_EXTRACT_ARRAY32(eq_data_p->event_data, 
							MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(gpio_event_data_st), gpio_event_hi),
							MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(gpio_event_data_st), gpio_event_hi));

  event_record_p->etype = VAPI_GPIO;
  MTL_ERROR1("%s: Got GPIO event gpio_low=0x%X gpio_high=0x%X\n", __func__, 
			 event_record_p->syndrome, event_record_p->event_modifier.qpn);
}

static inline void handle_ecc_event(eq_entry_t     *eq_data_p,
								   HH_event_record_t *event_record_p)
{
  /* there are 6 words in event_data 
     only the last 2 are relevant for this event
	 word 4 is in the syndrom, word 5 is in event_modifier.qpn */
  event_record_p->syndrome = eq_data_p->event_data[4];
  event_record_p->event_modifier.qpn = eq_data_p->event_data[5];
  event_record_p->etype = VAPI_ECC_DETECT;
  MTL_ERROR1("%s: Got ECC_DETECT event \n", __func__);
}

static inline void handle_client_reregister_event(eq_entry_t *eq_data_p,
												  HH_event_record_t *event_record_p)
{

  event_record_p->etype = VAPI_CLIENT_REREGISTER;
  event_record_p->event_modifier.port = 
			  MT_EXTRACT_ARRAY32(eq_data_p->event_data,MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(port_state_change_st), p),
								 MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(port_state_change_st), p));
  MTL_DEBUG1("%s: client reregister event on port=%d\n", __func__,event_record_p->event_modifier.port);
}


static void handle_eqe_ib_events(EQP_eq_entry_t *eqe_p,
                                 eq_entry_t     *eq_data_p)
{
  HH_event_record_t event_record;
  VAPI_event_syndrome_t vapi_syndrome = VAPI_EV_SYNDROME_NONE;
  event_record.syndrome = VAPI_EV_SYNDROME_NONE;
  
  switch(eq_data_p->event_type) {
#if 0    /* TK: currently not supported by Tavor */
    case ADD_DEV_PREFIX(IF_EV_TYPE_LOCAL_EE_CATAS_ERR):                  
#endif
    /* QP/EEC errors */
    case ADD_DEV_PREFIX(IF_EV_TYPE_PATH_MIG_FAIL):
    case ADD_DEV_PREFIX(IF_EV_TYPE_PATH_MIG_ARMED):
	case ADD_DEV_PREFIX(IF_EV_TYPE_PATH_MIG):                 
    case ADD_DEV_PREFIX(IF_EV_TYPE_COMM_EST):
    case ADD_DEV_PREFIX(IF_EV_TYPE_LOCAL_WQ_CATAS_ERR):
    case ADD_DEV_PREFIX(IF_EV_TYPE_LOCAL_WQ_INVALID_REQ_ERR):
    case ADD_DEV_PREFIX(IF_EV_TYPE_LOCAL_WQ_ACCESS_VIOL_ERR):
    case ADD_DEV_PREFIX(IF_EV_TYPE_SEND_Q_DRAINED): 
    case ADD_DEV_PREFIX(IF_EV_TYPE_SRQ_QP_LAST_WQE_REACHED):  /* this is QP event */
    {
        handle_qp_event(eq_data_p, &event_record);
		break;
    }
    case ADD_DEV_PREFIX(IF_EV_TYPE_SRQ_LIMIT_REACHED):
    case ADD_DEV_PREFIX(IF_EV_TYPE_LOCAL_SRQ_CATAS_ERR):
    {
        handle_srq_event(eq_data_p, &event_record);
	    break;
    }
    /* IB - affiliated errors CQ  */
    case ADD_DEV_PREFIX(IF_EV_TYPE_CQ_ERR):                           
    {
		handle_cq_event(eq_data_p, &event_record);
        break;
    }
    /* Unaffiliated errors */
    case ADD_DEV_PREFIX(IF_EV_TYPE_LOCAL_CATAS_ERR):
    {
        MTL_ERROR1("%s: CATASTROPHIC ERROR - should not be in this EQ: \n", __func__);
        MTL_ERROR1("CATASTROPHIC ERROR: data: %x %x %x %x %x %x \n", *(eq_data_p->event_data),
                   *(eq_data_p->event_data+1), *(eq_data_p->event_data+2), *(eq_data_p->event_data+3),
                   *(eq_data_p->event_data+4), *(eq_data_p->event_data+5));
        break;                 
    }
    case ADD_DEV_PREFIX(IF_EV_TYPE_PORT_ERR):                        
    { 
		handle_port_event(eq_data_p, &event_record);
        break;
    }
    case ADD_DEV_PREFIX(IF_EV_TYPE_GPIO):                        
    { 
		handle_gpio_event(eq_data_p, &event_record);
        break;
    }
    case ADD_DEV_PREFIX(IF_EV_TYPE_ECC_DETECT):                        
    { 
  	  handle_ecc_event(eq_data_p, &event_record);
  	  break;
    }
    case ADD_DEV_PREFIX(IF_EV_TYPE_CLIENT_REREGISTER):                        
    { 
	    handle_client_reregister_event(eq_data_p, &event_record);
	    break;
    }
	
    default:
        MTL_ERROR1("%s: Unsupported event type = %d\n", __func__,eq_data_p->event_type);
        /* in case of catastrophic error - no call to upper layer but notify HOB to handle */
        MTL_ERROR1("CATASTROPHIC ERROR: data: %x %x %x %x %x %x \n", *(eq_data_p->event_data),
                 *(eq_data_p->event_data+1), *(eq_data_p->event_data+2), *(eq_data_p->event_data+3),
                 *(eq_data_p->event_data+4), *(eq_data_p->event_data+5));
        XHH_hob_fatal_error(eqe_p->eventp_p->hob, XHH_FATAL_EVENT, vapi_syndrome);
        /**** !!!! FUNCTION returns HERE !!!! */
        return;
  } /* switch(eq_data_p->event_type)  */
  
  /* call the event callback */
  eqe_p->handler.ib_comp_event_h(eqe_p->eventp_p->hh_hca_hndl, &event_record, eqe_p->priv_context);

  
}


/************************************************************************************************/
/* Catastrophic Error EQ handling */

void catas_error_dpc(DPC_CONTEXT_t *dpc_obj_p)
{
  EQP_eq_entry_t *eqe_p = (EQP_eq_entry_t *)dpc_obj_p->func_ctx; /* EQ context entry */ 

  MTL_ERROR1("%s: Got CATASTROPHIC ERROR event from HCA\n", __func__);

  XHH_hob_fatal_error(eqe_p->eventp_p->hob, XHH_FATAL_EVENT, VAPI_CATAS_ERR_MSI);

  atomic_dec(&eqe_p->dpc_cntr); /* signal DPC done (for cleanup) */
  
}
