/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/


#include <mosal.h>
#include <xhh.h>
#include <eventp_priv.h>
#include <eventp_dev_funcs.h>
#include <cmdif.h>
#include <hob.h>
#include <log2.h>
#include <tptm.h>
//#define MTPERF
#include <mtperf.h>


/*================ macro definitions ===============================================*/


#define NOT_VALID_EQ_NUM(eq_num) (((int)(eq_num) < 0) || ((int)(eq_num) >= EQP_MAX_EQS))


/*================ type definitions ================================================*/

#define PHYS_EQ_MAX_SIZE 0x4000

/*================ global variables definitions ====================================*/


MTPERF_EXTERN_SEGMENT(interupt_segment);       
MTPERF_EXTERN_SEGMENT(dpc_segment);
MTPERF_EXTERN_SEGMENT(inter2dpc_segment);
MTPERF_EXTERN_SEGMENT(part_of_DPC_segment);


/*================ static functions prototypes =====================================*/

#ifdef MAX_DEBUG
static char * const eq_type_str(EQ_type_t eq_type)
{
  switch ( eq_type ) {
    case EQP_CQ_COMP_EVENT:
      return "EQP_CQ_COMP_EVENT";
    case EQP_IB_EVENT:
      return "EQP_IB_EVENT";
    case EQP_CMD_IF_EVENT:
      return "EQP_CMD_IF_EVENT";
    case EQP_CATAS_ERR_EVENT:
	  return "EQP_CATAS_ERR_EVENT";
    case EQP_MLX_EVENT:
      return "EQP_MLX_EVENT";
    default:
      return "***UNKNOWN***";
  }
}
#else
	/* an empty macro for Linux */
	EQ_TYPE_STR		
#endif


static XHH_eqn_t insert_new_eq(XHH_eventp_t     eventp,
                               void*            eventh, 
                               void             *priv_context, 
                               MT_size_t        max_outs_eqe, 
                               EQ_type_t        eq_type);

static HH_ret_t map_eq(XHH_eventp_t        eventp,
                        XHH_eqn_t           eqn,
                        XHH_eventp_mtmask_t hca_mask);

static void remove_eq(XHH_eventp_t        eventp,
                 XHH_eqn_t           eqn);

static HH_ret_t teardown_eq(/*IN */ XHH_eventp_t eventp, 
                     /*IN */ XHH_eqn_t eqn );

static HH_ret_t prepare_intr_resources(XHH_eventp_t eventp, XHH_eventp_res_t *eventp_res);

static HH_ret_t remove_intr_resources(XHH_eventp_t eventp);

static HH_ret_t add_catast_err_eq(XHH_eventp_t eventp);


/*================ external functions definitions ====================================*/



extern void eq_polling_dpc(DPC_CONTEXT_t *func_ctx);
extern void catas_error_dpc(DPC_CONTEXT_t *func_ctx);
extern irqreturn_t xhh_intr_handler(MT_ulong_ptr_t eventp, void* isr_ctx1, void* isr_ctx2);
extern irqreturn_t xhh_msix_intr_handler(MT_ulong_ptr_t eventp_entry, void* isr_ctx1, void* isr_ctx2);



/************************************************************************
 *  Function: XHH_eventp_create
 *  
 
    Arguments:
    version_p - Version information 
    event_res_p - See 7.2.1 XHH_eventp_res_t - Event processing resources on page 63 
    cmd_if - Command interface object to use for EQ setup commands 
    eventp_p - Returned object handle 
    
    Returns:
    HH_OK 
    HH_EINVAL -Invalid parameters 
    HH_EAGAIN -Not enough resources to create object 
    HH_ERR - internal error
    
    Description: 
    Create XHH_eventp object context. No EQs are set up until an event consumer registers 
    using one of the functions below. 
    
    
************************************************************************/
 
HH_ret_t XHH_eventp_create ( /*IN */ XHH_hob_t hob,
                             /*IN */ XHH_eventp_res_t *event_res_p, 
                             /*OUT*/ XHH_eventp_t *eventp_p )
{

  XHH_eventp_t new_eventp_p = NULL;
  u_int32_t i;
  HH_ret_t ret;

  FUNC_IN;
                         
  
  /* allocation of object structure */
  new_eventp_p = (XHH_eventp_t)VMALLOC(sizeof(XHH_eventp_internal_t));
  if (!new_eventp_p) {
    MT_COV_MEM_FLTR_START
    MTL_ERROR4("%s: Cannot allocate EVENTP object.\n", __func__);
    MT_RETURN(HH_EAGAIN);
    MT_COV_MEM_FLTR_END
  }
  
  memset(new_eventp_p,0,sizeof(XHH_eventp_internal_t));
  MOSAL_mutex_init(&new_eventp_p->mtx);
  for (i= 0; i < EQP_MAX_EQS; i++) {
    new_eventp_p->eq_table[i].eq_buff_entry_num= 0;
    new_eventp_p->eq_table[i].alloc_mem_addr_p= 0;
#ifdef SIMULTANUOUS_DPC
    if (MOSAL_spinlock_init(&(new_eventp_p->eq_table[i].dpc_lock)) != MT_OK){
      MTL_ERROR4("%s: Failed to initializing dpc_lock.\n", __func__);
      ret= HH_ERR;
      goto err_free_mem;
    } 
#endif    
    
  }

  /* filling eventp structure */
  memcpy(&new_eventp_p->event_resources, event_res_p, sizeof(XHH_eventp_res_t));
  if (event_res_p->intr_props->num_allocated_vectors == XHH_get_num_req_vectors() ) {
    new_eventp_p->max_eq_num_used = EQP_MAX_EQS;
  }
  else {
	new_eventp_p->max_eq_num_used = EQP_MAX_EQS-1; /* no CATAS error EQ */
  }
  new_eventp_p->have_fatal = FALSE;
  new_eventp_p->hob = hob;
  if (XHH_hob_get_cmd_if (hob, &new_eventp_p->cmd_if) != HH_OK){
    MTL_ERROR4("%s: Cannot get cmd_if object handle.\n", __func__);
    ret = HH_ERR;
    goto err_free_mem;
  }
  if ((ret = init_eventp_dev(new_eventp_p, event_res_p)) != HH_OK){ 
    MTL_ERROR4("%s: Failed to init_eventp_device_resources. ret=%d.\n", __func__,ret);
    ret = HH_ERR;
    goto err_free_mem;
  }
  if (XHH_hob_get_ver_info(hob, &new_eventp_p->version) != HH_OK){
    MTL_ERROR4("%s: Cannot get version.\n", __func__);
    ret = HH_ERR;
    goto err_free_mem;
  }
  if (XHH_hob_get_mrwm(hob, &new_eventp_p->mrwm_internal) != HH_OK){
    MTL_ERROR4("%s: Cannot get mrwm_internal.\n", __func__);
    ret = HH_ERR;
    goto err_free_mem;
  }
  if (XHH_hob_get_hca_hndl(hob, &new_eventp_p->hh_hca_hndl) != HH_OK){
    MTL_ERROR4("%s: Cannot get HH_HCA_hndl.\n", __func__);
    ret = HH_ERR;
    goto err_free_mem;
  }

  new_eventp_p->ctx_internal = MOSAL_get_kernel_prot_ctx();
#ifndef NO_EVENTS
  if (prepare_intr_resources(new_eventp_p, event_res_p)  != HH_OK){
    MTL_ERROR4("%s: Cannot set interrupt resources.\n", __func__);
    ret = HH_ERR;
    goto err_free_mem;
  }
  MTL_DEBUG1(MT_STDFMT("prepare_intr_resources success"));
#endif

  /* open the FATAL EQ if MSIX is supported and we have > 2 vectors 
     eventp create will not fail if this EQ is not opened */
  if (event_res_p->intr_props->num_allocated_vectors > 2) {
    MTL_DEBUG1(MT_STDFMT("going to add_catast_err_eq"));
	/* setup the catastrophic error EQ - must be a separete EQ initialzed at the begining */
    if (HH_OK != add_catast_err_eq(new_eventp_p)) {
      MTL_DEBUG1(MT_STDFMT("add_catast_err_eq failed"));
	}
  }
  
  *eventp_p = new_eventp_p;
  MT_RETURN(HH_OK);

  /* error handling cleanup */
err_free_mem:
  VFREE(new_eventp_p);
  MT_RETURN(ret);

}

/************************************************************************
 *  Function: XHH_eventp_destroy
 *  
 
   Arguments:
   eventp -The XHH_eventp object to destroy 
   
   Returns:
   HH_OK 
   HH_EINVAL -Invalid event object handle 
   
   Description: 
   Destroy object context. If any EQs are still set they are torn-down when this 
   function is called Those EQs should generate an HCA catastrophic error ((i.e.callbacks 
   for IB compliant, proprietary and debug events are invoked) since this call implies 
   abnormal HCA closure (in a normal HCA closure all EQs should be torn-down before a 
   call to this function). 
   
   
 ************************************************************************/

HH_ret_t XHH_eventp_destroy( /*IN */ XHH_eventp_t eventp)
{
#ifndef NO_EVENTS
  XHH_eqn_t eqn;
  
  FUNC_IN;
  
  if (eventp == NULL) {
    MTL_ERROR4("%s: eventp is NULL.\n", __func__);
    MT_RETURN( HH_EINVAL);
  }

  /* Teardown all EQs
   * - must first close all EQs that are NOT CMDIF
   * - then close CMDIF EQ in polling mode - this is the LAST EQ (#2)
   * - the data structure of the EQ will be removed only 
   *   after we remove the interrupts resources to ensure 
   *   we will not fail in the interrupt handler if an event is still in the air
  */
  
  MOSAL_mutex_acq_ui(&eventp->mtx);
  /* Teardown any EQ still up */
  for (eqn=0; eqn<eventp->free_eq_num; eqn++) {
    teardown_eq(eventp, eqn);
  }
  
  remove_intr_resources(eventp);
  
  for (eqn=0; eqn<eventp->free_eq_num; eqn++) {
    remove_eq(eventp, eqn);
  }
  
  MOSAL_mutex_rel(&eventp->mtx);
  MOSAL_mutex_free(&eventp->mtx);
#endif
  VFREE(eventp);
  MT_RETURN( HH_OK);
  

}
/************************************************************************
 *  Function: XHH_eventp_setup_comp_eq
 *  
   Arguments:
   eventp -The XHH_eventp object handle 
   eventh -The callback handle for events over this EQ 
   priv_context -Private context to be used in callback invocation 
   max_outs_eqe -Maximum outstanding EQEs in EQ created 
   eqn_p -Allocated EQ index 
   
   Returns:
   HH_OK
   HH_EINVAL -Invalid handle 
   HH_EAGAIN -Not enough resources available (e.g.EQC,mem- ory,etc.). 
   
   Description: 
   Set up an EQ for completion events and register given handler as a callback for 
   such events. 
   Note that the created EQ is not mapped to any event at this stage since completion 
   events are mapped using the CQC set up on CQ creation.
   
 
 ************************************************************************/

         
HH_ret_t XHH_eventp_setup_comp_eq(/*IN */ XHH_eventp_t eventp, 
                                  /*IN */ HH_comp_eventh_t eventh, 
                                  /*IN */ void *priv_context, 
                                  /*IN */ MT_size_t max_outs_eqe, 
                                  /*OUT*/ XHH_eqn_t *eqn_p )

{
  XHH_eqn_t new_eq;
  
  FUNC_IN;

  if (eventp == NULL || eventh == NULL || max_outs_eqe == 0) {
    MTL_ERROR4("%s: NULL parameter. eventp=%p, eventh=%p, max_outs_eqe=%d\n", __func__,
               eventp, eventh, (u_int32_t)max_outs_eqe);
    MT_RETURN(HH_EINVAL);
  }
  if (MOSAL_mutex_acq(&eventp->mtx, TRUE) != MT_OK){

    MTL_ERROR4("%s: MOSAL_mutex_acq failed\n", __func__);
    MT_RETURN(HH_EINTR);
  }


  new_eq = insert_new_eq(eventp, (void*)eventh, priv_context, max_outs_eqe, 
                         EQP_CQ_COMP_EVENT);
  
  if (new_eq == EQP_MAX_EQS) { /* All EQs are occupied */
    MTL_ERROR4("%s: Fail in adding new EQ.\n", __func__);
    MOSAL_mutex_rel(&eventp->mtx);
    MT_RETURN( HH_EAGAIN);
  }
  MTL_DEBUG1("%s success: eqn=%d\n", __func__, new_eq);

  *eqn_p = new_eq;
  MOSAL_mutex_rel(&eventp->mtx);

  MT_RETURN( HH_OK);
}

/************************************************************************
 *  Function: 
 *  
    Arguments:
    eventp -The XHH_eventp object handle 
    eventh -The callback handle for events over this EQ 
    priv_context -Private context to be used in callback invocation 
    max_outs_eqe -Maximum outstanding EQEs in EQ created 
    eqn_p -Allocated EQ index 
    
    Returns:
    HH_OK 
    HH_EINVAL -Invalid handle 
    HH_EAGAIN -Not enough resources available (e.g.EQC,mem- ory,etc.). 
    HH_ERR    - Internal error
    
    Description: 
    Set up an EQ and map events given in mask to it. Events over new EQ are reported to 
    given handler. 
    
    
 ************************************************************************/

HH_ret_t XHH_eventp_setup_ib_eq(/*IN */ XHH_eventp_t eventp, 
                                /*IN */ HH_async_eventh_t eventh, 
                                /*IN */ void *priv_context, 
                                /*IN */ MT_size_t max_outs_eqe, 
                                /*OUT*/ XHH_eqn_t *eqn_p )

{
  
#ifndef NO_EVENTS
  XHH_eqn_t new_eq;
  XHH_eventp_mtmask_t hca_mask=0;

  FUNC_IN;

  if (eventp == NULL || eventh == NULL || max_outs_eqe == 0) {
    MTL_ERROR4("%s: NULL parameter. eventp=%p, eventh=%p, max_outs_eqe=%d\n", __func__,
               eventp, eventh, (u_int32_t)max_outs_eqe);
    MT_RETURN( HH_EINVAL);
  }
  
  if (MOSAL_mutex_acq(&eventp->mtx, TRUE) != MT_OK){

    MTL_ERROR4("%s: MOSAL_mutex_acq failed\n", __func__);
    MT_RETURN(HH_EINTR);
  }


  new_eq = insert_new_eq(eventp, (void*)eventh, priv_context, max_outs_eqe, 
                         EQP_IB_EVENT);
  
  if (new_eq == EQP_MAX_EQS) { /* All EQs are occupied */
    MTL_ERROR4("%s: Fail in adding new EQ.\n", __func__);
    MOSAL_mutex_rel(&eventp->mtx);
    MT_RETURN( HH_EAGAIN);
  }

  /* map the EQ to events */
  /* prepare mask for all IB events */
  DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_PATH_MIG));
  DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_COMM_EST));
  DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_SEND_Q_DRAINED));
  DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_CQ_ERR));
  DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_LOCAL_WQ_CATAS_ERR));
  DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_LOCAL_EE_CATAS_ERR));
  DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_PATH_MIG_FAIL));
  DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_PORT_ERR));
  DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_LOCAL_WQ_INVALID_REQ_ERR));
  DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_LOCAL_WQ_ACCESS_VIOL_ERR));
  DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_GPIO));
  DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_ECC_DETECT));
  DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_PATH_MIG_ARMED));
  DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_CLIENT_REREGISTER));
  if (eventp->event_resources.is_srq_enable) {
    DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_LOCAL_SRQ_CATAS_ERR));
    DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_SRQ_QP_LAST_WQE_REACHED));
	DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_SRQ_LIMIT_REACHED));
  }
  if (map_eq(eventp, new_eq, hca_mask) != HH_OK){
    MTL_ERROR4("%s: Failed to map EQ.\n", __func__);
    MOSAL_mutex_rel(&eventp->mtx);
    MT_RETURN( HH_ERR);
  }
  MTL_DEBUG1("%s: Succeeded to map EQ=%d. mask="U64_FMT"\n", __func__, new_eq, (u_int64_t)hca_mask);
  *eqn_p = new_eq;

  MOSAL_mutex_rel(&eventp->mtx);
#endif  
  MT_RETURN( HH_OK);
}



/************************************************************************
 *  Function: 
 *  
    Arguments:
    eventp -The XHH_eventp object handle 
    
    Arguments:
    HH_OK 
    HH_EINVAL -Invalid handle 
    HH_EAGAIN -Not enough resources available (e.g.EQC,mem- ory,etc.). 
    EE_ERR    - internal error
    
    Description: 
    This function setup an EQ and maps command interface events to it. It also takes 
    care of notifying the XHH_cmd associated with it (as de  ned on the XHH_eventp creation)of 
    this EQ availability using the XHH_cmd_set_eq() (see page 38)function.This causes the XHH_cmd 
    to set event generation to given EQ for all commands dispatched after this noti  cation. The 
    XHH_eventp automatically sets noti  cation of events from this EQ to the XHH_cmd_eventh() 
    (see page 39)callback of associated XHH_cmd. The function should be invoked by the XHH_hob 
    in order to cause the XHH_cmd associated with this eventp to execute commands using events. 
    
    
 ************************************************************************/

HH_ret_t XHH_eventp_setup_cmd_eq ( /*IN */ XHH_eventp_t eventp,
                                   /*IN */ MT_size_t max_outs_eqe)

{
  
#ifndef NO_EVENTS
  XHH_eqn_t new_eq;
  XHH_eventp_mtmask_t hca_mask = 0;
  HH_ret_t ret;
  
  
  FUNC_IN;
  
  
  if (eventp == NULL || max_outs_eqe == 0) {
    MTL_ERROR4("%s: NULL parameter. eventp=%p, max_outs_eqe=%d\n", __func__,
               eventp, (u_int32_t)max_outs_eqe);
    MT_RETURN( HH_EINVAL);
  }
  
  if (MOSAL_mutex_acq(&eventp->mtx, TRUE) != MT_OK){
    MTL_ERROR4("%s: MOSAL_mutex_acq failed\n", __func__);
    MT_RETURN(HH_EINTR);
  }
  

  new_eq = insert_new_eq(eventp, NULL, NULL, max_outs_eqe, EQP_CMD_IF_EVENT);
  
  if (new_eq == EQP_MAX_EQS) { /* All EQs are occupied */
    MTL_ERROR4("%s: Fail in adding new EQ.\n", __func__);
    MOSAL_mutex_rel(&eventp->mtx);
    MT_RETURN( HH_EAGAIN);
  }
  /* map the EQ to events */
  DEV_IF_EV_MASK_SET(hca_mask, ADD_DEV_PREFIX(IF_EV_MASK_CMD_IF_COMP));
  if (map_eq(eventp,new_eq,hca_mask) != HH_OK){
    MTL_ERROR4("%s: Failed to map EQ.\n", __func__);
    MOSAL_mutex_rel(&eventp->mtx);
    MT_RETURN( HH_ERR);
  }

  if ((ret = XHH_cmd_set_eq(eventp->cmd_if)) != HH_OK){
    MTL_ERROR4("%s: Failed to XHH_cmd_set_eq. ret=%d\n", __func__, ret);
    MOSAL_mutex_rel(&eventp->mtx);
    MT_RETURN( HH_ERR);
  }

  MOSAL_mutex_rel(&eventp->mtx);

  MTL_DEBUG1("%s success: eqn=%d\n", __func__, new_eq);
#endif  
  MT_RETURN( HH_OK);
}


/************************************************************************
 *  Function: XHH_eventp_setup_mt_eq
 *  
    Arguments:
    eventp -The XHH_eventp object handle 
    event_mask -Flags combination of events to map to this EQ 
    eventh -The callback handle for events over this EQ 
    priv_context -Private context to be used in callback invocation 
    max_outs_eqe -Maximum outstanding EQEs in EQ created eqn_p -Allocated EQ index 
    
    Returns:
    HH_OK 
    HH_EINVAL -Invalid handle 
    HH_EAGAIN -Not enough resources available (e.g.EQC,mem- ory,etc.). 
    HH_ERR    - internal error
    
    Description: 
    Set up an EQ for reporting events beyond IB-spec.(debug events and others). All events 
    given in the event_mask are mapped to the new EQ. 
    
    
    
 ************************************************************************/

HH_ret_t XHH_eventp_setup_mt_eq(/*IN */ XHH_eventp_t eventp, 
                                /*IN */ XHH_eventp_mtmask_t event_mask, 
                                /*IN */ XHH_mlx_eventh_t eventh, 
                                /*IN */ void *priv_context, 
                                /*IN */ MT_size_t max_outs_eqe, 
                                /*OUT*/ XHH_eqn_t *eqn_p)

{
  
#ifndef NO_EVENTS
  XHH_eqn_t new_eq;
  
  FUNC_IN;

  if (eventp == NULL || eventh == NULL || max_outs_eqe == 0 || event_mask == 0) {
    MTL_ERROR4("%s: NULL parameter. eventp=%p, eventh=%p, max_outs_eqe=%d, event_mask="U64_FMT"\n", __func__,
               eventp, eventh, (u_int32_t)max_outs_eqe, (u_int64_t)event_mask);
    MT_RETURN( HH_EINVAL);
  }
  
  if (MOSAL_mutex_acq(&eventp->mtx, TRUE) != MT_OK){
    MTL_ERROR4("%s: MOSAL_mutex_acq failed\n", __func__);
    MT_RETURN(HH_EINTR);
  }

  new_eq = insert_new_eq(eventp, (void*)eventh, priv_context, max_outs_eqe, EQP_MLX_EVENT);

  if (new_eq == EQP_MAX_EQS) { /* All EQs are occupied */
    MTL_ERROR4("%s: Fail in adding new EQ.\n", __func__);
    MOSAL_mutex_rel(&eventp->mtx);
    MT_RETURN( HH_EAGAIN);
  }
  /* map the EQ to events */
  if (map_eq(eventp,new_eq,event_mask) != HH_OK){
    MTL_ERROR4("%s: Failed to map EQ.\n", __func__);
    MOSAL_mutex_rel(&eventp->mtx);
    MT_RETURN( HH_ERR);
  }
  *eqn_p = new_eq;
  MOSAL_mutex_rel(&eventp->mtx);

#endif
  MT_RETURN( HH_OK);
}


/************************************************************************
 *  Function: 
 *  
    Arguments:
    eventp 
    eq -The EQ to replace handler for 
    eventh -The new handler 
    priv_context -Private context to be used with handler
    
    Returns:
    HH_OK 
    HH_EINVAL 
    HH_ENORSC -Given EQ is not set up 
    
    Description: 
    Replace the callback function of an EQ previously set up. This may be used 
    in order to change the handler without loosing events.It retains the EQ and 
    just replaces the callback function.All EQEs polled after a return from this 
    function will be reported to the new handler.
    
 ************************************************************************/

HH_ret_t XHH_eventp_replace_handler(/*IN */ XHH_eventp_t eventp, 
                                    /*IN */ XHH_eqn_t eqn, 
                                    /*IN */ XHH_eventp_handler_t eventh, 
                                    /*IN */ void *priv_context)

{
  
  HH_ret_t ret=HH_OK;
  
#ifndef NO_EVENTS
  FUNC_IN;


  if (eventp == NULL || NOT_VALID_EQ_NUM((int32_t)eqn)){
    MTL_ERROR4("%s: Invalid parameter. eventp=%p, eq_num=%d\n", __func__,
               eventp, (u_int32_t)eqn);
    MT_RETURN( HH_EINVAL);
  }

  switch (eventp->eq_table[eqn].eq_type)
  {
  case EQP_CQ_COMP_EVENT:
      if(eventh.comp_event_h == NULL) {
        MTL_ERROR4("%s: Invalid event handler is NULL\n", __func__);
        ret = HH_EINVAL;
      }
      else {
        eventp->eq_table[eqn].handler.comp_event_h= eventh.comp_event_h;
      }
      break;
    case EQP_IB_EVENT:
      if(eventh.ib_comp_event_h == NULL) {
        MTL_ERROR4("%s: Invalid event handler is NULL\n", __func__);
        ret = HH_EINVAL;
      }
      else {
        eventp->eq_table[eqn].handler.ib_comp_event_h= eventh.ib_comp_event_h;
      }
      break;
    case EQP_CMD_IF_EVENT:    /* no event handler in this case */
      break;
    case EQP_MLX_EVENT:
      if(eventh.mlx_event_h == NULL) {
        MTL_ERROR4("%s: Invalid event handler is NULL\n", __func__);
        ret = HH_EINVAL;
      }
      else {
        eventp->eq_table[eqn].handler.mlx_event_h= eventh.mlx_event_h;
      }
      break;
    default:
      MTL_ERROR4("%s: Internal error: invalid event type.\n", __func__);
      ret= HH_ERR;
  }
  eventp->eq_table[eqn].priv_context = priv_context;
#endif
  MT_RETURN(ret);

}


/************************************************************************
 *  Function: XHH_eventp_notify_fatal
 *  
    Arguments:
    eventp -The XHH_eventp object handle
    fatal_err - the error code of the fatal error 
    
    Return:
    HH_OK 
    HH_EINVAL -Invalid handle 
    
    Description: 
    This function is invoked by XHH_hob_fatal_error to notify eventp when a fatal error
    has occurred.
    
 ************************************************************************/

HH_ret_t XHH_eventp_notify_fatal ( /*IN */ XHH_eventp_t eventp,
                                   /*IN */ XHH_fatal_err_t fatal_err)
{
  
  FUNC_IN;
  
  if (eventp == NULL ) {
    MTL_ERROR4("%s: NULL parameter. eventp=%p\n",__func__,eventp);
    MT_RETURN( HH_EINVAL);
  }

  eventp->have_fatal = TRUE;
  MT_RETURN(HH_OK);
}

/************************************************************************
 *  Function: XHH_eventp_handle_fatal
 *  
    Arguments:
    eventp -The XHH_eventp object handle
    
    Return:
    HH_OK 
    HH_EINVAL -Invalid handle 
    
    Description: 
    This function is invoked by XHH_hob_fatal_error t
    
    
    
    o handle the fatal error
    has occurred.
    
 ************************************************************************/

HH_ret_t XHH_eventp_handle_fatal ( /*IN */ XHH_eventp_t eventp)
{
  
  FUNC_IN;
  
  if (eventp == NULL ) {
    MTL_ERROR4("%s: NULL parameter. eventp=%p\n", __func__, eventp);
    MT_RETURN( HH_EINVAL);
  }

  MT_RETURN(HH_OK);
}


/********************* STATIC FUNCTIONS ************************************************/

HH_ret_t register_msi_handler(XHH_eventp_t     eventp,
                              void*            eventh, 
                              void*            priv_context, 
                              EQ_type_t        eq_type,
							  unsigned int     eqn,
							  unsigned int*    vector_idx,
							  int*             requesting_irq)
{
  char type_str[12];
  char name[HCA_MAX_NAME+sizeof(type_str)];
  MOSAL_ISR_func_t handler = NULL;
  MT_ulong_ptr_t isr_ctx = 0;


  *vector_idx = 0;

  /* vector allocation:
     completion EQ allways in vector 0;
     2 vectors: IB & CMD in vector 1; CATAS ERR - not opened
	 3 vectors: IB & CMD in vector 1; CATAS ERR in vector 2
	 4 vectors: each EQ in a different vector (0-3)
   */

  /* first handle special case of COMP EQ */
  if (eq_type==EQP_CQ_COMP_EVENT) {
	*requesting_irq = 1;
    handler = xhh_msix_intr_handler;
	isr_ctx = (MT_ulong_ptr_t)(&eventp->eq_table[eqn]);
	*vector_idx = 0;
  }
  
  /* all events - case 4 will reset values of comp_eq */
  switch (eventp->event_resources.intr_props->num_allocated_vectors) {
  case 2:
    if (eq_type==EQP_CATAS_ERR_EVENT) {
  	  return HH_EBUSY; /* not allocating catas error EQ if we don't have at least 3 vectors */
    }
  case 3:
    if (eq_type==EQP_CMD_IF_EVENT) {
  	    *requesting_irq = 1;
 	    handler = xhh_intr_handler;
 	    isr_ctx = (MT_ulong_ptr_t)eventp;
 	    *vector_idx = 1;
     }
     else if (eq_type==EQP_IB_EVENT) {
      *requesting_irq = 0;
	  *vector_idx = 1;
    }
     else if (eq_type==EQP_CATAS_ERR_EVENT) {
      *requesting_irq = 1;
       handler = xhh_msix_intr_handler;
       isr_ctx = (MT_ulong_ptr_t)(&eventp->eq_table[eqn]);
       *vector_idx = 2;
     }
    break;
  case 4:
    *requesting_irq = 1;
    handler = xhh_msix_intr_handler;
    isr_ctx = (MT_ulong_ptr_t)(&eventp->eq_table[eqn]);
    *vector_idx = eqn;
    break;
  default:
    MTL_ERROR1(MT_STDFMT("invalid number of vectors allocated - %d"),
               eventp->event_resources.intr_props->num_allocated_vectors);
    return HH_ERR;
  }
  
  /* building MSIX string */
  switch (eq_type) {
  case EQP_CQ_COMP_EVENT:
    strcpy(type_str, "-comp");
    break;
  case EQP_IB_EVENT:
    strcpy(type_str, "-ib");
    break;
  case EQP_CMD_IF_EVENT:
    if (eventp->event_resources.intr_props->num_allocated_vectors == 4) {
	  strcpy(type_str, "-cmdif");
    }
    else {
      strcpy(type_str, "-cmdif+ib");
    }
    break;
  case EQP_CATAS_ERR_EVENT:
    strcpy(type_str,"-catas-err");
    break;
  case EQP_MLX_EVENT:    /* currently not supported */
  default:
    MTL_ERROR4("%s: Internal error: invalid event queue type.\n", __func__);
    return HH_ERR;
  }

  /* if we work in MSIX mode then the handler must be registered here */
  if (*requesting_irq) {
    MTL_DEBUG1(MT_STDFMT("requesting irq"));
    memset(name, 0, sizeof(name));
    strcpy(name, eventp->event_resources.name);
    strcat(name, type_str);
    if (MT_OK != MOSAL_ISR_set(&eventp->isr_obj[*vector_idx], handler,
                       eventp->event_resources.intr_props->irq[*vector_idx],
                       name, isr_ctx)) {
      MTL_ERROR1(MT_STDFMT("failed requesting interrupt %d"),
                          eventp->event_resources.intr_props->irq[eqn]);
      return HH_ERR;
    }
  }
  return HH_OK;

}

/* This function must be invoked with eventp's mutex locked */
static XHH_eqn_t insert_new_eq(XHH_eventp_t     eventp,
                               void*            eventh, 
                               void*            priv_context, 
                               MT_size_t        max_outs_eqe, 
                               EQ_type_t        eq_type)
{
  unsigned int new_eq;
  EQP_eq_entry_t *new_entry;
  MT_size_t entries_num;
  call_result_t rc;
  XHH_eqc_t eq_context;
  MT_bool virtual_eq = TRUE;
  XHH_cmd_status_t cmd_ret;
  HH_ret_t ret=HH_OK;
  VAPI_size_t alloc_mem_bytes_size;
  unsigned int vector_idx = 0;
  int requesting_irq = 0;
  

  FUNC_IN;

  new_eq = eventp->free_eq_num;

  MTL_DEBUG3("%s: in params: max_outs_eqe = "SIZE_T_FMT", eq_type=%s; got EQ num = %d\n", 
             __func__, max_outs_eqe, eq_type_str(eq_type), new_eq);

  /* register MSIX handler if we work with MSIX */
  if (eventp->event_resources.intr_props->num_allocated_vectors > 1) {
    if (register_msi_handler(eventp, eventh, priv_context, eq_type, new_eq, 
							 &vector_idx, &requesting_irq) 
		!= HH_OK){
	  MT_RETURN(EQP_MAX_EQS);
    }
  }

  new_entry = &(eventp->eq_table[new_eq]);
  /* number of outs_eqes must be power of 2 */
  entries_num = XHH_CYCLIC_BUFF_SIZE(max_outs_eqe); /* already take care for the empty entry in cyclic buff */
  
  /* EQS are in main memory */
  /* need to add 1 for the alignment */
  alloc_mem_bytes_size = (entries_num+1)*EQ_ENTRY_SIZE;
  /* if this is a small EQ then work with physically contiguous memory */
  if (alloc_mem_bytes_size <= PHYS_EQ_MAX_SIZE) {
    new_entry->alloc_mem_addr_p = (MT_virt_addr_t)MOSAL_pci_phys_alloc_consistent((MT_size_t)alloc_mem_bytes_size, floor_log2(EQ_ENTRY_SIZE));
    if (new_entry->alloc_mem_addr_p != 0) {
      virtual_eq=FALSE;
      MTL_TRACE5("%s: EQ %d is going to be with physical memory. \n", __func__,new_eq);
    }
  }

  if (virtual_eq) {
    new_entry->alloc_mem_addr_p = (MT_virt_addr_t)MOSAL_pci_virt_alloc_consistent((MT_size_t)alloc_mem_bytes_size, floor_log2(EQ_ENTRY_SIZE));
    if (!new_entry->alloc_mem_addr_p) {
      MTL_ERROR4("%s: Cannot allocate EQE buffer.\n", __func__);
	  goto err_undo_irq;
    }
  }

  if (!new_entry->alloc_mem_addr_p) {
    /* allocation failed */
    goto err_undo_irq;
  }
  
  /* EQEs cyclic buffer should be aligned to entry size */
  new_entry->eq_buff = (void *)MT_UP_ALIGNX_VIRT(new_entry->alloc_mem_addr_p, floor_log2(EQ_ENTRY_SIZE)); 
  memset((void*)new_entry->alloc_mem_addr_p,0xff,(MT_size_t)alloc_mem_bytes_size);


  MTL_DEBUG3("%s: real entries_num = "SIZE_T_FMT", table size= "SIZE_T_FMT"\n", __func__, entries_num,(entries_num * EQ_ENTRY_SIZE) );
  new_entry->pd = XHH_RESERVED_PD(eventp->hob);
  new_entry->priv_context = priv_context;
  atomic_set(&new_entry->dpc_cntr,0);
  new_entry->eventp_p = eventp;
  new_entry->virtual_eq = virtual_eq;
  new_entry->eq_type = eq_type;
  new_entry->eqn = new_eq;
  switch (eq_type)
  {
    case EQP_CQ_COMP_EVENT:
      new_entry->handler.comp_event_h = (HH_comp_eventh_t)eventh;
      break;
    case EQP_IB_EVENT:
      new_entry->handler.ib_comp_event_h = (HH_async_eventh_t)eventh;
      break;
    case EQP_CMD_IF_EVENT:    /* no event handler in this case */
	case EQP_CATAS_ERR_EVENT:
      break;
    case EQP_MLX_EVENT:
      new_entry->handler.mlx_event_h = (XHH_mlx_eventh_t)eventh;
      break;
    default:
      MTL_ERROR4("%s: Internal error: invalid event queue type.\n", __func__);
      goto err_free_mem;
  }
  
  if (virtual_eq) {	 
    XHH_internal_mr_t  *params_p;
    /* registering memory region for this buffer */
    params_p = (XHH_internal_mr_t*)MALLOC(sizeof(XHH_internal_mr_t));
	if (!params_p) {
	  MT_COV_MEM_FLTR_START
	  MTL_ERROR4("%s: Cannot allocate params_p.\n", __func__);
	  goto err_free_mem;
	  MT_COV_MEM_FLTR_END
	}
    memset(params_p, 0, sizeof(XHH_internal_mr_t));
    params_p->start        = (IB_virt_addr_t)(MT_virt_addr_t)(new_entry->eq_buff);
    params_p->size         = (VAPI_size_t)entries_num * EQ_ENTRY_SIZE;
    params_p->pd           = XHH_RESERVED_PD(eventp->hob);
    params_p->vm_ctx       = MOSAL_get_kernel_prot_ctx(); //eventp->ctx_internal;
    params_p->force_memkey = FALSE;

    MTL_DEBUG4("%s: registering mem region. start addr="U64_FMT", size="U64_FMT"\n", __func__,
               params_p->start, params_p->size);
    if((ret = XHH_mrwm_register_internal(eventp->mrwm_internal, params_p, &new_entry->mem_lkey)) != HH_OK){
      MTL_ERROR4("%s: Failed to register EQ buffer in memory. ret=%d\n", __func__, ret);
	  FREE(params_p);
      goto err_free_mem;
    }
	FREE(params_p);
    MTL_TRACE4("%s: SUCCESS to register EQ buffer in memory. \n", __func__);
  }
  else {
    new_entry->mem_lkey = 0;
  }
  
  /* prepare EQ for HW ownership */
  memset(&eq_context,0,sizeof(XHH_eqc_t));
  eq_context.st = EQ_STATE_ARMED;
  eq_context.oi = FALSE;	/* Overrun detection ignore */
  eq_context.tr = virtual_eq;	/* Translation Required. If set - EQ access undergo address translation. */
  eq_context.owner = XHH_OWNER_HW;	/* SW/HW ownership */
  eq_context.status = EQ_STATUS_OK;	/* EQ status:\;0000 - OK\;1001 - EQ overflow\;1010 - EQ write failure */
  if (virtual_eq) {
    eq_context.start_address = (u_int64_t)(MT_virt_addr_t)new_entry->eq_buff;	/* Start Address of Event Queue. Must be aligned on 32-byte boundary */
  }
  else {
    MT_phys_addr_t pa;

    rc = MOSAL_virt_to_phys(MOSAL_get_kernel_prot_ctx(), (MT_virt_addr_t)new_entry->eq_buff, &pa);
    if ( rc != MT_OK ) {
      MTL_ERROR4(MT_FLFMT("%s: failed va=%p"), __func__, new_entry->eq_buff);
      goto err_unreg_mem;
    }
    else {
      eq_context.start_address = (u_int64_t)pa;
    }
  }
  
  if ((ret = init_eq_dev(eventp, &eq_context, new_entry)) != HH_OK){
    MTL_ERROR4("%s: Failed to init_eq_dev. ret=%d.\n", __func__,ret);
    goto err_unreg_mem;
  }

  eq_context.log_eq_size = (u_int8_t)floor_log2(entries_num);	/* Log2 of the amount of entries in the EQ */
  
  /* Interrupt (message) to be generated to report event to INT layer.
                     00iiiiii - specifies GPIO pin to be asserted
                     10jjjjjj - specificies type of interrupt message to be generated (total 128 different messages supported). */
  if ( eventp->event_resources.intr_props->int_mode == MT_INT_MODE_MSI ) {
    eq_context.intr = 0x80 | vector_idx;  /* 0 here means the number of the vector to use:
                                    0 - eventp->event_resources.intr_props->num_allocated_vectors */
    MTL_DEBUG1(MT_FLFMT("using MSI"));
  }
  else {
    MTL_DEBUG1(MT_FLFMT("using PIN"));
    eq_context.intr = eventp->event_resources.intr_clr_bit;
  }
  
  eq_context.lkey = new_entry->mem_lkey;	/* Memory key (L-Key) to be used to access EQ */
  eq_context.consumer_indx = 0;	/* Contains next entry to be read upon poll for completion. Must be initialized to '0 while opening EQ */
  eq_context.producer_indx = 0;	/* Contains next entry in EQ to be written by the HCA. Must be initialized to '1 while opening EQ. */
  eq_context.pd = XHH_RESERVED_PD(eventp->hob); // TK - need to add to structure
  
  if ((cmd_ret = XHH_cmd_SW2HW_EQ(eventp->cmd_if, new_eq, &eq_context)) != XHH_CMD_STAT_OK){
    MTL_ERROR4("%s: Failed to move EQ to HW ownership. CMD_IF error:%d.\n", __func__,cmd_ret);
    goto err_unreg_mem;
  }
  
  new_entry->cons_cntr = 0;
  new_entry->cons_cntr_mask = MASK32(eq_context.log_eq_size);

  MTL_TRACE4("%s: SUCCESS to XHH_cmd_SW2HW_EQ EQ=%d. \n", __func__,new_eq);
  /* init DPC of this EQ */
  MOSAL_DPC_init(&new_entry->polling_dpc, (eq_type == EQP_CATAS_ERR_EVENT ? catas_error_dpc : eq_polling_dpc), 
				 (MT_ulong_ptr_t)new_entry,  MOSAL_SINGLE_CTX);
  eventp->free_eq_num++; /* enlarge next free only after we completed all work */

  /* this must be done last since this is the indication that the entry is valid */
  new_entry->alloc_mem_bytes_size = alloc_mem_bytes_size;
  new_entry->eq_buff_entry_num = entries_num;
  MT_RETURN( (XHH_eqn_t)new_eq);

  /* error handling */

err_unreg_mem:
  if (virtual_eq) {
    if ((ret = XHH_mrwm_deregister_mr(eventp->mrwm_internal, new_entry->mem_lkey)) != HH_OK){
      MTL_ERROR4("%s: Failed to deregister memory region. ret=%d\n", __func__,ret);
    }
  }

err_free_mem:  
  if (virtual_eq) {
    MOSAL_pci_virt_free_consistent((void *)new_entry->alloc_mem_addr_p, alloc_mem_bytes_size);
  }
  else {
    MOSAL_pci_phys_free_consistent((void *)(MT_ulong_ptr_t)new_entry->alloc_mem_addr_p, (MT_size_t)alloc_mem_bytes_size);
  }

err_undo_irq:
    if (requesting_irq && (rc=MOSAL_ISR_unset(&eventp->isr_obj[vector_idx])) != MT_OK ) {
      MTL_ERROR1(MT_STDFMT("MOSAL_ISR_unset failed - %s"), mtl_strerror_sym(rc));
    }

  MT_RETURN(EQP_MAX_EQS);

} /* insert_new_eq */
  

/************************************************************************
 *  Function: teardown_eq
 *  
    Arguments:
    eventp -The XHH_eventp object handle 
    eqn -The EQ to teardown 
    
    Returns:
    HH_OK 
    HH_EINVAL -Invalid handles (e.g.no such EQ set up) 
    
    Description: 
    This function tear down an EQ set up by one of the previous functions. Given eqn  
    which EQ to tear down. This teardown includes cleaning of any context relating to 
    the callback associated with it. 
    
    
 ************************************************************************/

HH_ret_t teardown_eq(/*IN */ XHH_eventp_t eventp, 
                     /*IN */ XHH_eqn_t eqn )

{
  
  XHH_cmd_status_t cmd_ret;
  HH_ret_t ret;
  unsigned long i=0;
  
  FUNC_IN;
  MTL_DEBUG4("%s: start. eqn=%d type=%s\n", __func__, eqn, eq_type_str(eventp->eq_table[eqn].eq_type));
  if (eventp == NULL || NOT_VALID_EQ_NUM(eqn)){
    MTL_ERROR4("%s: Invalid parameter. eventp=%p, eq_num=%d\n", __func__,
               eventp, (u_int32_t)eqn);
    MT_RETURN( HH_EINVAL);
  }
  
  /* Wait for all outstanding DPCs */
  while (atomic_read(&eventp->eq_table[eqn].dpc_cntr))
  {
    /* this must be done for Linux only since there is no preemption in Linux */
    /* (DPC/tasklet cannot run while this context hold the CPU)               */
    MT_DPC_SCHEDULE;
    i++;
	if (i==0xffffffff) {
      MTL_DEBUG4("%s: dpc_cntr was not zero after %lu iterations for eq_num=%d, dpc_cntr=%d\n",
                 __func__, i, eqn, atomic_read(&eventp->eq_table[eqn].dpc_cntr));
      i=0;
      //break;
    }
  }
  
  /* for CMD_IF event need to notify the cmd_if object */
  if (eventp->eq_table[eqn].eq_type == EQP_CMD_IF_EVENT){
    if ((ret = XHH_cmd_clr_eq(eventp->cmd_if)) != HH_OK){
      MTL_ERROR4("%s: Failed to XHH_cmd_clr_eq. ret=%d\n", __func__, ret);
    }
  }

  /* unmap events of EQ by putting special EQ number */
  /* in case of fatal error - do not call CMD_IF */
  if (eventp->have_fatal == FALSE) {
    if ((cmd_ret = XHH_cmd_MAP_EQ(eventp->cmd_if, ADD_DEV_PREFIX(IF_UNMAP_QP_BIT), eventp->eq_table[eqn].events_mask))
        != XHH_CMD_STAT_OK){
      MTL_ERROR4("%s: Failed to unmap EQ events. CMD_IF error:%d.\n", __func__,cmd_ret);
    }
  }
  MT_RETURN( HH_OK);

}


/************************************************************************/

static HH_ret_t map_eq(XHH_eventp_t         eventp,
                       XHH_eqn_t            eqn,
                       XHH_eventp_mtmask_t  hca_mask)
{
  
  XHH_cmd_status_t cmd_ret;

  
  FUNC_IN;
  /* save mask for later unmapping when EQ is teared down */
  eventp->eq_table[eqn].events_mask = hca_mask;
  MTL_TRACE4("%s:  EQ=%d mask="U64_FMT" \n", __func__, eqn, (u_int64_t)hca_mask);

  if ((cmd_ret = XHH_cmd_MAP_EQ(eventp->cmd_if, eqn, hca_mask))!= XHH_CMD_STAT_OK){
    MTL_ERROR4("%s: Failed to map EQ events. CMD_IF error:%d EQN=%d.\n", __func__,cmd_ret,eqn);
    /* due to this error need to remove EQ */
    remove_eq(eventp, eqn);
    MT_RETURN( HH_ERR);
  }
  MTL_TRACE4("%s: SUCCESS to XHH_cmd_MAP_EQ. EQ=%d mask="U64_FMT" \n", __func__, eqn, (u_int64_t)hca_mask);
  MT_RETURN( HH_OK);
}


/************************************************************************/

static void remove_eq(XHH_eventp_t        eventp,
                      XHH_eqn_t           eqn)
{
  XHH_cmd_status_t cmd_ret;
  XHH_eqc_t eq_context;
  VAPI_size_t buf_sz;
  void* buf_addr;
  MT_bool virt_buf;
  HH_ret_t ret;

  
  FUNC_IN;
  
  /* clear EQC (in SW ownership) */
  memset(&eq_context,0,sizeof(XHH_eqc_t));
  
  /* in case of fatal error - do not call CMD_IF */
  if (eventp->have_fatal == FALSE) {
    if ((cmd_ret = XHH_cmd_HW2SW_EQ(eventp->cmd_if, eqn, &eq_context)) != XHH_CMD_STAT_OK){
      MTL_ERROR4("%s: Failed to move EQ to SW ownership. CMD_IF error:%d.\n", __func__,cmd_ret);
      /* TK - maybe we need to exit in this case */
    }
  }
  /* unregister MR of buffer and free the eq_buff */
  if (eventp->eq_table[eqn].virtual_eq) {
    ret = XHH_mrwm_deregister_mr(eventp->mrwm_internal, eventp->eq_table[eqn].mem_lkey);
    if (ret != HH_OK){ 
        MTL_ERROR4("%s: Failed to deregister memory region. ret=%d\n", __func__,ret);
    }
  }
  virt_buf= eventp->eq_table[eqn].virtual_eq;
  buf_addr= (void*)eventp->eq_table[eqn].alloc_mem_addr_p;
  buf_sz= eventp->eq_table[eqn].alloc_mem_bytes_size;
  eventp->eq_table[eqn].alloc_mem_addr_p= 0;
  eventp->eq_table[eqn].eq_buff_entry_num = 0; 
  eventp->eq_table[eqn].alloc_mem_bytes_size = 0;

  /* Free buffer */
  
  if (virt_buf) {
    MOSAL_pci_virt_free_consistent(buf_addr, buf_sz);
  }
  else {
    MOSAL_pci_phys_free_consistent(buf_addr, (MT_size_t)buf_sz);
  }
  
  FUNC_OUT;
  return;
}


/************************************************************************************************/

HH_ret_t prepare_intr_resources(XHH_eventp_t eventp, XHH_eventp_res_t *eventp_res)
{
  call_result_t mt_ret;
  MT_virt_addr_t temp_addr;  /* used due to "volatile" modifier for intr_clr_reg */

  FUNC_IN;

  /* if interrupt bit < 32: we use the low word and otherwise the high word */
  
   
  mt_ret  = MOSAL_iomem_map(MOSAL_get_kernel_prot_ctx(),
			    MOSAL_MEM_FLAGS_NO_CACHE | MOSAL_MEM_FLAGS_PERM_READ | 
			                   MOSAL_MEM_FLAGS_PERM_WRITE,
			    eventp_res->clr_int_base_addr 
                                        + (eventp_res->intr_clr_bit<32 ? 4:0),
                            (MT_size_t) sizeof(u_int32_t),
			    &(eventp->intr_clr_reg_iomem),&(temp_addr) );

  if (mt_ret != MT_OK) {
     MTL_ERROR1("%s: Failed to MOSAL_iomem_map for intr_clr_reg (%d)\n", __func__, mt_ret);
     goto map_err;
  }

  eventp->intr_clr_reg = temp_addr;

  if (init_dev_regs(eventp, eventp_res) != HH_OK){
    MTL_ERROR1("%s: Failed to init_eq_regs \n", __func__);
    goto init_devregs_err;
  }

  eventp->intr_clr_mask = MOSAL_be32_to_cpu(1 << (eventp_res->intr_clr_bit % 32));
  if ( eventp_res->intr_props->num_allocated_vectors == 1 ) {
    if ((mt_ret = MOSAL_ISR_set(&eventp->isr_obj[0], xhh_intr_handler, 
                                 eventp_res->intr_props->irq[0],
                                 eventp_res->name, (MT_ulong_ptr_t)eventp)) != MT_OK){
      MTL_ERROR1("%s: Failed to MOSAL_ISR_set MOSAL ret=%d\n", __func__, mt_ret);
      goto isr_set_err;
    }
  }
  MTL_DEBUG1(MT_STDFMT("successful MOSAL_ISR_set, irq=%d"), eventp_res->intr_props->irq[0]);

  MT_RETURN( HH_OK);


isr_set_err:   
   /* init_dev_regs does own cleanup */
init_devregs_err:
    MOSAL_iomem_unmap(eventp->intr_clr_reg_iomem, MOSAL_get_kernel_prot_ctx());
    
map_err:
    eventp->intr_clr_reg = 0;
    eventp->intr_clr_reg_iomem = (MOSAL_iomem_t) NULL;

    MT_RETURN(HH_ERR);
}


/*
***********************************************************************************************/

HH_ret_t remove_intr_resources(XHH_eventp_t eventp)
{
  int i;
  call_result_t rc;

  FUNC_IN;

  if (eventp->event_resources.intr_props->num_allocated_vectors == 1) {
    if ((rc = MOSAL_ISR_unset(&eventp->isr_obj[0])) != MT_OK){
      MTL_ERROR1("%s: Failed to MOSAL_ISR_unset MOSAL ret=%s\n", __func__, mtl_strerror_sym(rc));
      MT_RETURN( HH_ERR);
    }
  }
  else if (eventp->event_resources.intr_props->num_allocated_vectors > 1) {
    for (i=0; i<eventp->event_resources.intr_props->num_allocated_vectors; ++i) {
      if ( (rc = MOSAL_ISR_unset(&eventp->isr_obj[i])) != MT_OK ) {
        MTL_ERROR1(MT_STDFMT("MOSAL_ISR_unset failed for irq%d - %s"),
    						   eventp->event_resources.intr_props->irq[i], mtl_strerror_sym(rc));
      }
    }
  }

  MOSAL_iomem_unmap(eventp->intr_clr_reg_iomem,MOSAL_get_kernel_prot_ctx());
  eventp->intr_clr_reg = 0;
  eventp->intr_clr_reg_iomem = (MOSAL_iomem_t) NULL;

  remove_dev_regs(eventp);

  MT_RETURN( HH_OK);
}

/************************************************************************************************/
static HH_ret_t add_catast_err_eq(XHH_eventp_t eventp)
{
  XHH_eventp_mtmask_t hca_mask = 0;
  HH_ret_t ret = HH_OK;
  unsigned int eqn;

  if (MOSAL_mutex_acq(&eventp->mtx, TRUE) != MT_OK){
    MTL_ERROR4("%s: MOSAL_mutex_acq failed\n", __func__);
    MT_RETURN(HH_EINTR);
  }
  
  if ((eqn = insert_new_eq(eventp, (void*)NULL, NULL, 2, EQP_CATAS_ERR_EVENT)) == EQP_MAX_EQS) {
    MTL_ERROR4("%s: Failed to add the catastrophic event EQ.\n", __func__);
    ret = HH_ERR;
	goto exit;
  }

  DEV_IF_EV_MASK_SET(hca_mask,ADD_DEV_PREFIX(IF_EV_MASK_LOCAL_CATAS_ERR));

  if (map_eq(eventp, eqn, hca_mask) != HH_OK){
    MTL_ERROR4("%s: Failed to map catastrophic error EQ.\n", __func__);
    ret = HH_ERR;
	goto exit;
  }

exit:
  MOSAL_mutex_rel(&eventp->mtx);
  MT_RETURN(ret);

}

