/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <mtl_common.h>
#include <cr_types.h>
#include <vapi.h>
#include <vapi_common.h>
#include <hh.h>
#include <hob_priv.h>
#include <ul_hob.h>
#include <sm_mad.h>
#include <log2.h>
#include <default_profile.h>
#include <srqm.h>
#include <hob_imp.h>
#include <hob_pci_caps.h>
#include <hob_comm.h>

static volatile unsigned long hotplug_support = 1;

/******************************************************************************
 *  Function:     XHH_hob_modify
 *
 *  Description:  Implements the VAPI_modify_hca verb
 *
 *  input:
 *                hca_hndl
 *                port_num - 1 or 2
 *                hca_attr_p - contains values to modify
 *                hca_attr_mask_p - mask specifying which values in hca_attr_p should
 *                                  be used for modification.
 *  output: 
 *                none
 *  returns:
 *                HH_OK
 *                HH_EINVAL
 *                HH_EINVAL_HCA_HNDL
 *                HH_EINVAL_PORT
 *                HH_ENOSYS
 *
 *  Comments:     Implements IB Spec 1.0a now.  must be modified to support IB Spec 1.1
 *                JPM
 *
 *****************************************************************************/
HH_ret_t XHH_hob_modify(
                        HH_hca_hndl_t        hca_hndl,
                        IB_port_t            port_num,
                        VAPI_hca_attr_t      *hca_attr_p,
                        VAPI_hca_attr_mask_t *hca_attr_mask_p)
{
  /* TBD, will use SET_IB command.  Problem is that can only set PKey and QKey counters to zero. */
  HH_ret_t  retn;
  XHH_cmd_status_t cmd_ret;
  VAPI_hca_port_t  port_props;
  XHH_set_ib_props_t    set_ib_props;
  IB_port_cap_mask_t    capabilities;           
  XHH_hob_t             hob;
   
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_modify: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_modify : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }

  hob = XHHOBP(hca_hndl);

  if (hob == NULL) {
      MTL_ERROR1("XHH_hob_modify : ERROR : No device registered\n");
      return HH_EINVAL;
  }
  TEST_RETURN_FATAL(hob);

  if (port_num > hob->dev_lims.num_ports || port_num < 1) {
      MTL_ERROR2( "XHH_hob_modify: ERROR : invalid port number(%d)\n", port_num);
      return HH_EINVAL_PORT;
  }

  memset(&set_ib_props, 0, sizeof(XHH_set_ib_props_t));
  
  set_ib_props.rqk = hca_attr_p->reset_qkey_counter;

  /* start with current capabilities */
  retn = XHH_hob_query_port_prop(hca_hndl, port_num, &port_props);
  if (retn != HH_OK) {
      MTL_ERROR1("XHH_hob_modify : ERROR : cannot get current capabilities (%d)\n", retn);
      return HH_EAGAIN;
  }
  capabilities = port_props.capability_mask;
  
  /* now, modify the capability mask according to the input */
  if (HCA_ATTR_IS_FLAGS_SET(*hca_attr_mask_p)) {
      /* calculate capabilities modification mask */
      if(HCA_ATTR_IS_SET(*hca_attr_mask_p, HCA_ATTR_IS_SM) ) {
          if (hca_attr_p->is_sm) {
              IB_CAP_MASK_SET(capabilities, IB_CAP_MASK_IS_SM);
          } else {
              IB_CAP_MASK_CLR(capabilities, IB_CAP_MASK_IS_SM);
          }
      }
      if(HCA_ATTR_IS_SET(*hca_attr_mask_p, HCA_ATTR_IS_SNMP_TUN_SUP) ) {
          if (hca_attr_p->is_snmp_tun_sup) {
              IB_CAP_MASK_SET(capabilities, IB_CAP_MASK_IS_SNMP_TUNN_SUP);
          } else {
              IB_CAP_MASK_CLR(capabilities, IB_CAP_MASK_IS_SNMP_TUNN_SUP);
          }
      }
      if(HCA_ATTR_IS_SET(*hca_attr_mask_p, HCA_ATTR_IS_DEV_MGT_SUP) ) {
          if (hca_attr_p->is_dev_mgt_sup) {
              IB_CAP_MASK_SET(capabilities, IB_CAP_MASK_IS_DEVICE_MGMT_SUP);
          } else {
              IB_CAP_MASK_CLR(capabilities, IB_CAP_MASK_IS_DEVICE_MGMT_SUP);
          }
      }
      if(HCA_ATTR_IS_SET(*hca_attr_mask_p, HCA_ATTR_IS_VENDOR_CLS_SUP) ) {
          if (hca_attr_p->is_vendor_cls_sup) {
              IB_CAP_MASK_SET(capabilities, IB_CAP_MASK_IS_VENDOR_CLS_SUP);
          } else {
              IB_CAP_MASK_CLR(capabilities, IB_CAP_MASK_IS_VENDOR_CLS_SUP);
          }
      }
  }

  set_ib_props.capability_mask = capabilities;
  
  /* now, perform the CMD */
  cmd_ret = XHH_cmd_SET_IB(hob->cmd , port_num, &set_ib_props);
  if (cmd_ret != XHH_CMD_STAT_OK) {
      TEST_CMD_FATAL(cmd_ret);
      MTL_ERROR1("XHH_hob_modify: CMD_error in XHH_cmd_SET_IB (%d)\n", cmd_ret);
      return HH_EINVAL;
  }

  return(HH_OK);
}
/******************************************************************************
 *  Function:     XHH_hob_query_port_prop
 *
 *  Description:  Implements the VAPI_query_hca_port_prop verb
 *
 *  input:
 *                hca_hndl
 *                port_num - 1 or 2
 *  output: 
 *                hca_port_p - port properties output structure
 *  returns:
 *                HH_OK
 *                HH_EINVAL
 *                HH_EINVAL_HCA_HNDL
 *                HH_EINVAL_PORT
 *                HH_ERR
 *
 *  Comments:     Does MAD query to get the data in real time.  Data is not pre-fetched, because
 *                the current port state is needed for the answer -- so the query must be performed
 *                anyway.
 *
 *****************************************************************************/
HH_ret_t XHH_hob_query_port_prop(HH_hca_hndl_t  hca_hndl,
                                    IB_port_t           port_num,
                                    VAPI_hca_port_t     *hca_port_p ) 
{
    SM_MAD_PortInfo_t  *port_info;
    u_int8_t       *mad_frame_in;
    u_int8_t       *mad_frame_out;
    XHH_cmd_status_t  cmd_ret;

    XHH_hob_t  hob;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1("XHH_hob_query_port_prop: NOT IN TASK CONTEXT)\n");
        return HH_ERR;
    }

    if (hca_hndl == NULL) {
        MTL_ERROR1("XHH_hob_query_port_prop : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
    }
    hob = XHHOBP(hca_hndl);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_query_port_prop : ERROR : No device registered\n");
        return HH_EINVAL;
    }
    TEST_RETURN_FATAL(hob);

    if (port_num > hob->dev_lims.num_ports || port_num < 1) {
        MTL_ERROR2( "XHH_hob_query_port_prop: ERROR : invalid port number(%d)\n", port_num);
        return HH_EINVAL_PORT;
    }

    port_info = TMALLOC(SM_MAD_PortInfo_t);
    if ( !port_info )
{
		MT_COV_MEM_FLTR_START
		return HH_EAGAIN;
		MT_COV_MEM_FLTR_END
}
    mad_frame_in = TNMALLOC(u_int8_t, IB_MAD_SIZE);
    if ( !mad_frame_in ) {
      MT_COV_MEM_FLTR_START
      FREE(port_info);
      return HH_EAGAIN;
      MT_COV_MEM_FLTR_END
    }
    mad_frame_out = TNMALLOC(u_int8_t, IB_MAD_SIZE);
    if ( !mad_frame_out ) {
      MT_COV_MEM_FLTR_START
      FREE(port_info);
      FREE(mad_frame_in);
      return HH_EAGAIN;
      MT_COV_MEM_FLTR_END
    }
    memset(mad_frame_in, 0, sizeof(mad_frame_in));
    memset(mad_frame_out, 0, sizeof(mad_frame_out));

    /* get port props using MAD commands in XHH_cmd object */
    /* First, build the MAD header */
    MADHeaderBuild(IB_CLASS_SMP, 
                      0,
                      IB_METHOD_GET,
                      IB_SMP_ATTRIB_PORTINFO,
                      (u_int32_t)   port_num,
                      &(mad_frame_in[0]));

    /* issue the query */
    cmd_ret = XHH_cmd_MAD_IFC(hob->cmd, 0, 0, port_num, &(mad_frame_in[0]), &(mad_frame_out[0]));
    if (cmd_ret != XHH_CMD_STAT_OK) {
        TEST_CMD_FATAL(cmd_ret);
        MTL_ERROR2( "XHH_hob_query_port_prop: ERROR : Get Port Info command failed (%d) for port %d\n", cmd_ret, port_num);
        FREE(port_info);
        FREE(mad_frame_out);
        FREE(mad_frame_in);
        if ( cmd_ret == XHH_CMD_STAT_EINTR ) {
          return HH_EINTR;
        }
        return HH_ERR;
    }
    /* now, translate the response to a structure */
    PortInfoMADToSt(port_info, &(mad_frame_out[0]));

	/* finally, extract the information we want */
    hca_port_p->bad_pkey_counter = port_info->wPKViolations;
    hca_port_p->capability_mask  = port_info->dwCapMask;
    hca_port_p->gid_tbl_len      = port_info->bGUIDCap;
    hca_port_p->lid              = port_info->wLID;
    hca_port_p->lmc              = port_info->cLMC;
    hca_port_p->max_msg_sz       = IB_MAX_MESSAGE_SIZE;
    hca_port_p->max_mtu          = hob->dev_lims.max_mtu;
    hca_port_p->max_vl_num       = port_info->cVLCap;
    hca_port_p->pkey_tbl_len     = hob->init_ib_props[port_num-1].max_pkey;
    hca_port_p->qkey_viol_counter = port_info->wQKViolations;
    hca_port_p->sm_lid            = port_info->wMasterSMLID;
    hca_port_p->sm_sl             = port_info->cMasterSMSL;
    hca_port_p->state             = port_info->cPortState;
    hca_port_p->subnet_timeout    = port_info->cSubnetTO;

    hca_port_p->initTypeReply     = 0;   /* not yet supported in FW */
    
  FREE(port_info);
  FREE(mad_frame_out);
  FREE(mad_frame_in);
  return(HH_OK);
}

/******************************************************************************
 *  Function:     XHH_hob_get_pkey_tbl_local
 *
 *  Description:  Gets PKEY table for a given port  
 *
 *  input:
 *                hca_hndl
 *                port_num - 1 or 2
 *                tbl_len_in - size of table provided for response (in pkeys)
 *                use_mad_query_for_pkeys - if TRUE, query device for pkeys
 *                       else, use pkey_table tracking in xhh_qpm
 *  output: 
 *                tbl_len_out - size of returned table (in pkeys)
 *                pkey_tbl_p  - pointer to table containing data (space provided by caller)
 *  returns:
 *                HH_OK
 *                HH_EINVAL
 *                HH_EINVAL_HCA_HNDL
 *                HH_ERR
 *
 *  Comments:     Does MAD query to get the data in real time. 
 *
 *****************************************************************************/
static HH_ret_t XHH_hob_get_pkey_tbl_local(HH_hca_hndl_t  hca_hndl,
                                     IB_port_t     port_num,
                                     u_int16_t     tbl_len_in,
                                     u_int16_t     *tbl_len_out,
                                     IB_pkey_t     *pkey_tbl_p,
                                     MT_bool       use_mad_query_for_pkeys)
{
  SM_MAD_Pkey_table_t*  pkey_table_p = NULL;
  u_int8_t       *mad_frame_in = NULL;
  u_int8_t       *mad_frame_out = NULL;
  int  i,j;
  int num_pkeys, pkey_index, num_pkeytable_commands;
  XHH_cmd_status_t  cmd_ret;
  XHH_hob_t  xhh_hob_p;
  XHH_qpm_t      qpm;
  HH_ret_t     hh_ret = HH_OK;
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);

  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_get_pkey_tbl: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_get_pkey_tbl : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  /* check if have valid port number */
  if (port_num > xhh_hob_p->dev_lims.num_ports || port_num < 1) {
      MTL_ERROR1("XHH_hob_get_pkey_tbl: port number (%d) not valid)\n", port_num);
      return HH_EINVAL_PORT;
  }
  
  if (tbl_len_out == NULL) {
      return HH_EINVAL;
  }
  

  /* check that pkey table has enough space */
  num_pkeys =  xhh_hob_p->init_ib_props[port_num-1].max_pkey;
  *tbl_len_out = (u_int16_t)num_pkeys;

  if (tbl_len_in < num_pkeys) {
      if (!tbl_len_in) {
          MTL_TRACE2( "XHH_hob_get_pkey_tbl: returning number of pkeys configured (%d)\n", num_pkeys);
      } else {
          MTL_ERROR2( "XHH_hob_get_pkey_tbl: ERROR : not enough space in return value table. num keys = %d\n",
                      num_pkeys);
      }
      return HH_EAGAIN;
  }

  /* check that have valid output buffer area */
  if (pkey_tbl_p == NULL) {
      return HH_EINVAL;
  }

  
  /* get KEY table using MAD command in XHH_cmd object */
  /* get PKey table using MAD commands in XHH_cmd object */
  /* First, build the MAD header */
  if (use_mad_query_for_pkeys == TRUE) {
      num_pkeytable_commands = ((num_pkeys - 1) / 32) + 1;
    
      pkey_index =  0;

      mad_frame_in = TNMALLOC(u_int8_t, IB_MAD_SIZE);
      if ( !mad_frame_in ) {
	MT_COV_MEM_FLTR_START
	return HH_EAGAIN;
	MT_COV_MEM_FLTR_END
      }
      mad_frame_out = TNMALLOC(u_int8_t, IB_MAD_SIZE);
      if ( !mad_frame_out ) {
	MT_COV_MEM_FLTR_START
	FREE(mad_frame_in);
	return HH_EAGAIN;
	MT_COV_MEM_FLTR_END
      }

      pkey_table_p = (SM_MAD_Pkey_table_t *)MALLOC(sizeof(SM_MAD_Pkey_table_t));
      if (pkey_table_p == NULL) {
	 MT_COV_MEM_FLTR_START
	 FREE(mad_frame_in);
	 FREE(mad_frame_out);
	 return HH_EAGAIN;
	 MT_COV_MEM_FLTR_END
      }
      for (i = 0; i < num_pkeytable_commands; i++) {
          memset(mad_frame_in, 0, sizeof(mad_frame_in));
          memset(mad_frame_out, 0, sizeof(mad_frame_out));
	  memset(pkey_table_p, 0, sizeof(SM_MAD_Pkey_table_t));
          MADHeaderBuild(IB_CLASS_SMP, 
                            0,
                            IB_METHOD_GET,
                            IB_SMP_ATTRIB_PARTTABLE,
                            (u_int32_t) i,
                            &(mad_frame_in[0]));
    
          cmd_ret = XHH_cmd_MAD_IFC(xhh_hob_p->cmd, 0, 0, port_num, &(mad_frame_in[0]), &(mad_frame_out[0]));
          if (cmd_ret != XHH_CMD_STAT_OK) {
              TEST_CMD_FATAL(cmd_ret);
              MTL_ERROR2( "XHH_hob_get_pkey_tbl: ERROR : Get Partition Table command failed (%d) for port %d\n", cmd_ret, port_num);
              FREE(mad_frame_out);
              FREE(mad_frame_in);
	      FREE(pkey_table_p);
              return HH_ERR;
          }
          PKeyTableMADToSt(pkey_table_p, &(mad_frame_out[0]));
    
          for (j = 0; j < 32; j++) {
              pkey_tbl_p[pkey_index++] = pkey_table_p->pkey[j];
              if (pkey_index == num_pkeys) {
                  break;
              }
          }
      }
      FREE(mad_frame_out);
      FREE(mad_frame_in);
      FREE(pkey_table_p);
  } else {
      hh_ret =  XHH_hob_get_qpm ( xhh_hob_p, &qpm );
      if (hh_ret != HH_OK) {
          MTL_ERROR2( "XHH_hob_get_qpm: invalid QPM handle (ret= %d)\n", hh_ret);
          FREE(mad_frame_out);
          FREE(mad_frame_in);
	  FREE(pkey_table_p);
          return HH_EINVAL;
      }
      hh_ret = XHH_qpm_get_all_pkeys(qpm,port_num,(u_int16_t)num_pkeys, pkey_tbl_p);
      if (hh_ret != HH_OK) {
          MTL_ERROR2( "XHH_qpm_get_all_sgids failed (ret= %d)\n", hh_ret);
          FREE(mad_frame_out);
          FREE(mad_frame_in);
	  FREE(pkey_table_p);
          return HH_EINVAL;
      }
  }
  return(HH_OK);
}
/******************************************************************************
 *  Function:     XHH_hob_get_pkey_tbl
 *
 *  Description:  Gets PKEY table for a given port  
 *
 *  input:
 *                hca_hndl
 *                port_num - 1 or 2
 *                tbl_len_in - size of table provided for response (in pkeys)
 *  output: 
 *                tbl_len_out - size of returned table (in pkeys)
 *                pkey_tbl_p  - pointer to table containing data (space provided by caller)
 *  returns:
 *                HH_OK
 *                HH_EINVAL
 *                HH_EINVAL_HCA_HNDL
 *                HH_ERR
 *
 *  Comments:     Does MAD query to get the data in real time. 
 *
 *****************************************************************************/
HH_ret_t XHH_hob_get_pkey_tbl(HH_hca_hndl_t  hca_hndl,
                                     IB_port_t     port_num,
                                     u_int16_t     tbl_len_in,
                                     u_int16_t     *tbl_len_out,
                                     IB_pkey_t     *pkey_tbl_p)
{
    MT_bool is_legacy = FALSE;
    XHH_hob_t  xhh_hob_p;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    FUNC_IN;

    MTL_DEBUG4("XHH_hob_get_pkey_tbl:  hca_hndl=0x%p, port= %d, return table len = %d\n",
                 hca_hndl, port_num, tbl_len_in);


    if (hca_hndl == NULL) {
        MTL_ERROR1("XHH_hob_get_pkey_tbl : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);
    if (xhh_hob_p == NULL) {
        MTL_ERROR1("XHH_hob_get_pkey_tbl : ERROR : No device registered\n");
        return HH_EINVAL;
    }
    TEST_RETURN_FATAL(xhh_hob_p);

    XHH_hob_get_legacy_mode(xhh_hob_p, &is_legacy);
    return(XHH_hob_get_pkey_tbl_local(hca_hndl,port_num,tbl_len_in,
                                     tbl_len_out,pkey_tbl_p,is_legacy));

}

HH_ret_t  XHH_hob_init_pkey_tbl( HH_hca_hndl_t  hca_hndl,
                                      IB_port_t      port_num,
                                      u_int16_t      tbl_len_in,
                                      u_int16_t*     tbl_len_out,
                                      IB_pkey_t     *pkey_tbl_p)
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    return(XHH_hob_get_pkey_tbl_local(hca_hndl,port_num,tbl_len_in,
                                     tbl_len_out,pkey_tbl_p,1));

}


/******************************************************************************
 *  Function:     XHH_hob_get_gid_tbl_local
 *
 *  Description:  Gets GID table for a given port  
 *
 *  input:
 *                hca_hndl
 *                port_num - 1 or 2
 *                tbl_len_in - size of table provided for response (in pkeys)
 *                use_mad_query_for_gid_prefix - if TRUE, query device for gid prefix.
 *                       else, use gid_table tracking in xhh_qpm
 *  output: 
 *                tbl_len_out - size of returned table (in pkeys)
 *                param_gid_p  - pointer to table containing data (space provided by caller)
 *  returns:
 *                HH_OK
 *                HH_EINVAL
 *                HH_EINVAL_HCA_HNDL
 *                HH_ERR
 *
 *  Comments:     Does MAD query to get the data in real time. 
 *
 *****************************************************************************/
static 
HH_ret_t  XHH_hob_get_gid_tbl_local( HH_hca_hndl_t  hca_hndl,
                                      IB_port_t      port,
                                      u_int16_t      tbl_len_in,
                                      u_int16_t*     tbl_len_out,
                                      IB_gid_t*      param_gid_p,
                                      MT_bool        use_mad_query_for_gid_prefix)
{
  SM_MAD_PortInfo_t  *port_info_p = NULL;
  SM_MAD_GUIDInfo_t  *guid_info_p = NULL;
  u_int8_t       *mad_frame_in = NULL;
  u_int8_t       *mad_frame_out = NULL;
  XHH_cmd_status_t  cmd_ret;
  int            num_guids, guid_index;
  int            num_guidinfo_commands;
  u_int8_t       *gid_p = (u_int8_t *) param_gid_p;
  int            i,j;
  HH_ret_t       hh_ret = HH_OK;
  XHH_qpm_t      qpm;
  IB_gid_t       gid;
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  FUNC_IN;

  MTL_DEBUG4("XHH_hob_get_gid_tbl_local:  hca_hndl=0x%p, port= %d, return table len = %d\n",
               hca_hndl, port, tbl_len_in);

  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_get_gid_tbl: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_get_gid_tbl : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);
  

  /* check if have valid port number */
  if (port > xhh_hob_p->dev_lims.num_ports || port < 1) {
      MTL_ERROR1("XHH_hob_get_gid_tbl: port number (%d) not valid)\n", port);
      return HH_EINVAL_PORT;
  }

  if (tbl_len_out == NULL) {
      return HH_EINVAL;
  }
  

  /* check that gid table has enough space */
  num_guids =  xhh_hob_p->init_ib_props[port-1].max_gid;
  *tbl_len_out = (u_int16_t)num_guids;

  if (tbl_len_in < num_guids) {
      if (!tbl_len_in) {
          MTL_TRACE2( "XHH_hob_get_gid_tbl: returning gid table configured size (%d)\n", num_guids);
      } else {
          MTL_ERROR2( "XHH_hob_get_gid_tbl: ERROR : not enough space in return value table.  Need %d\n",
                      num_guids);
      }
      return HH_EAGAIN;
  }

  /* check that have valid output buffer area */
  if (param_gid_p == NULL) {
      return HH_EINVAL;
  }

  /* get GID table using MAD commands in XHH_cmd object */
  if (use_mad_query_for_gid_prefix == TRUE) {
      /* First, get the GID prefix from via MAD query  */
      mad_frame_in = TNMALLOC(u_int8_t, IB_MAD_SIZE);
      if ( !mad_frame_in ) {
	MT_COV_MEM_FLTR_START
	return HH_EAGAIN;
	MT_COV_MEM_FLTR_END
      }
      mad_frame_out = TNMALLOC(u_int8_t, IB_MAD_SIZE);
      if ( !mad_frame_out ) {
	MT_COV_MEM_FLTR_START
	FREE(mad_frame_in);
	return HH_EAGAIN;
	MT_COV_MEM_FLTR_END
      }

      port_info_p = (SM_MAD_PortInfo_t *)MALLOC(sizeof(SM_MAD_PortInfo_t));
      if (port_info_p == NULL) {
	 MT_COV_MEM_FLTR_START
	 FREE(mad_frame_in);
	 FREE(mad_frame_out);
	 return HH_EAGAIN;
	 MT_COV_MEM_FLTR_END
      }

      guid_info_p = (SM_MAD_GUIDInfo_t *)MALLOC(sizeof(SM_MAD_GUIDInfo_t));
      if (guid_info_p == NULL) {
	 MT_COV_MEM_FLTR_START
	 FREE(mad_frame_in);
	 FREE(mad_frame_out);
	 FREE(port_info_p);
	 return HH_EAGAIN;
	 MT_COV_MEM_FLTR_END
      }
      
      memset(port_info_p, 0, sizeof(SM_MAD_PortInfo_t));
      memset(mad_frame_in, 0, sizeof(mad_frame_in));
      memset(mad_frame_out, 0, sizeof(mad_frame_out));
      MADHeaderBuild(IB_CLASS_SMP, 
                        0,
                        IB_METHOD_GET,
                        IB_SMP_ATTRIB_PORTINFO,
                        (u_int32_t)   port,
                        &(mad_frame_in[0]));
    
      /* issue the query */
      cmd_ret = XHH_cmd_MAD_IFC(xhh_hob_p->cmd, 0, 0, port, &(mad_frame_in[0]), &(mad_frame_out[0]));
      if (cmd_ret != XHH_CMD_STAT_OK) {
          TEST_CMD_FATAL(cmd_ret);
          MTL_ERROR2( "XHH_hob_get_gid_tbl: ERROR : Get Port Info command failed (%d) for port %d\n", cmd_ret, port);
          FREE(mad_frame_out);
          FREE(mad_frame_in);
	  FREE(port_info_p);
	  FREE(guid_info_p);
          return HH_ERR;
      }
      PortInfoMADToSt(port_info_p, &(mad_frame_out[0]));
      PortInfoPrint(port_info_p);
      memcpy(&gid, &(port_info_p->qwGIDPrefix), sizeof(port_info_p->qwGIDPrefix));

      /* Now, get the GUIDs, and build GIDS */
      num_guidinfo_commands = ((num_guids - 1) / 8) + 1;

      guid_index =  0;
      for (i = 0; i < num_guidinfo_commands; i++) {
	  memset(guid_info_p, 0, sizeof(SM_MAD_GUIDInfo_t));
          memset(mad_frame_in, 0, sizeof(mad_frame_in));
          memset(mad_frame_out, 0, sizeof(mad_frame_out));
          MADHeaderBuild(IB_CLASS_SMP, 
                            0,
                            IB_METHOD_GET,
                            IB_SMP_ATTRIB_GUIDINFO,
                            (u_int32_t) i,
                            &(mad_frame_in[0]));

          cmd_ret = XHH_cmd_MAD_IFC(xhh_hob_p->cmd, 0, 0, port, &(mad_frame_in[0]), &(mad_frame_out[0]));
          if (cmd_ret != XHH_CMD_STAT_OK) {
              TEST_CMD_FATAL(cmd_ret);
              MTL_ERROR2( "XHH_hob_get_gid_tbl: ERROR : Get GUID Info command failed (%d) for port %d\n", cmd_ret, port);
              FREE(mad_frame_out);
              FREE(mad_frame_in);
	      FREE(port_info_p);
	      FREE(guid_info_p);
              return HH_ERR;
          }
          GUIDInfoMADToSt(guid_info_p, &(mad_frame_out[0]));
          GUIDInfoPrint(guid_info_p);

          for (j = 0; j < 8; j++) {
              memcpy (gid_p, &(gid), sizeof(port_info_p->qwGIDPrefix));
              gid_p += sizeof(port_info_p->qwGIDPrefix);
              memcpy (gid_p, &(guid_info_p->guid[j]), sizeof(IB_guid_t));
              gid_p += sizeof(u_int64_t);
              guid_index++;
              if (guid_index == num_guids) {
                  break;
              }
          }
      }
      FREE(mad_frame_out);
      FREE(mad_frame_in);
      FREE(port_info_p);
      FREE(guid_info_p);
  } else {
      hh_ret =  XHH_hob_get_qpm ( xhh_hob_p, &qpm );
      if (hh_ret != HH_OK) {
          MTL_ERROR2( "XHH_hob_get_qpm: invalid QPM handle (ret= %d)\n", hh_ret);
          return HH_EINVAL;
      }
      hh_ret = XHH_qpm_get_all_sgids(qpm,port,(u_int8_t)num_guids, param_gid_p);
      if (hh_ret != HH_OK) {
          MTL_ERROR2( "XHH_qpm_get_all_sgids failed (ret= %d)\n", hh_ret);
          return HH_EINVAL;
      }
      return HH_OK;
  }
  return HH_OK;
} /* XHH_get_gid_tbl */
/******************************************************************************
 *  Function:     XHH_hob_get_gid_tbl
 *
 *  Description:  Gets GID table for a given port  
 *
 *  input:
 *                hca_hndl
 *                port_num - 1 or 2
 *                tbl_len_in - size of table provided for response (in pkeys)
 *  output: 
 *                tbl_len_out - size of returned table (in pkeys)
 *                param_gid_p  - pointer to table containing data (space provided by caller)
 *  returns:
 *                HH_OK
 *                HH_EINVAL
 *                HH_EINVAL_HCA_HNDL
 *                HH_ERR
 *
 *  Comments:     Does MAD query to get the data in real time. 
 *
 *****************************************************************************/
HH_ret_t  XHH_hob_get_gid_tbl( HH_hca_hndl_t  hca_hndl,
                                      IB_port_t      port,
                                      u_int16_t      tbl_len_in,
                                      u_int16_t*     tbl_len_out,
                                      IB_gid_t*      param_gid_p)
{
    MT_bool is_legacy = FALSE;
    XHH_hob_t  xhh_hob_p;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    FUNC_IN;

    MTL_DEBUG4("XHH_hob_get_gid_tbl:  hca_hndl=0x%p, port= %d, return table len = %d\n",
                 hca_hndl, port, tbl_len_in);


    if (hca_hndl == NULL) {
        MTL_ERROR1("XHH_hob_get_gid_tbl : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);
    if (xhh_hob_p == NULL) {
        MTL_ERROR1("XHH_hob_get_gid_tbl : ERROR : No device registered\n");
        return HH_EINVAL;
    }
    TEST_RETURN_FATAL(xhh_hob_p);

    
    XHH_hob_get_legacy_mode(xhh_hob_p, &is_legacy);
    return(XHH_hob_get_gid_tbl_local(hca_hndl,port,tbl_len_in,
                                     tbl_len_out,param_gid_p,is_legacy));

}

HH_ret_t  XHH_hob_init_gid_tbl( HH_hca_hndl_t  hca_hndl,
                                      IB_port_t      port,
                                      u_int16_t      tbl_len_in,
                                      u_int16_t*     tbl_len_out,
                                      IB_gid_t*      param_gid_p)
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    return(XHH_hob_get_gid_tbl_local(hca_hndl,port,tbl_len_in,
                                     tbl_len_out,param_gid_p,1));

}

/*****************************************************************************
******************************************************************************
************** PASS-THROUGH FUNCTIONS  ********************************** ****
******************************************************************************
*****************************************************************************/

/******************************************************************************
 *  Function:     XHH_hob_register_mr <==> XHH_mrwm_register_mr
 *****************************************************************************/
HH_ret_t XHH_hob_register_mr(HH_hca_hndl_t  hca_hndl, 
                                    HH_mr_t       *mr_props_p, 
                                    VAPI_lkey_t   *lkey_p, 
                                    IB_rkey_t     *rkey_p)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_register_mr: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_register_mr : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_register_mr: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_register_mr: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_mrwm_register_mr(xhh_hob_p->mrwm, mr_props_p, lkey_p, rkey_p);
}



/******************************************************************************
 *  Function:     XHH_hob_reregister_mr <==> XHH_mrwm_reregister_mr
 *****************************************************************************/
HH_ret_t XHH_hob_reregister_mr(HH_hca_hndl_t  hca_hndl,
                               VAPI_lkey_t    lkey,  
                                      VAPI_mr_change_t  change_mask,
                                      HH_mr_t           *mr_props_p, 
                                      VAPI_lkey_t*       lkey_p, 
                                      IB_rkey_t         *rkey_p)
{
  XHH_hob_t  xhh_hob_p;


  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
 if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_reregister_mr: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_reregister_mr : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_reregister_mr: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_reregister_mr: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_mrwm_reregister_mr(xhh_hob_p->mrwm, lkey, change_mask, mr_props_p,lkey_p, rkey_p);
}



/******************************************************************************
 *  Function:     XHH_hob_register_smr <==> XHH_mrwm_register_smr
 *****************************************************************************/
HH_ret_t XHH_hob_register_smr(HH_hca_hndl_t  hca_hndl, 
                                    HH_smr_t       *mr_props_p, 
                                    VAPI_lkey_t    *lkey_p, 
                                    IB_rkey_t      *rkey_p)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_register_smr: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_register_smr : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_register_smr: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_register_smr: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_mrwm_register_smr(xhh_hob_p->mrwm, mr_props_p, lkey_p, rkey_p);
}


/******************************************************************************
 *  Function:     XHH_hob_query_mr <==> XHH_mrwm_query_mr
 *****************************************************************************/
HH_ret_t XHH_hob_query_mr(HH_hca_hndl_t  hca_hndl, 
                                 VAPI_lkey_t      lkey, 
                                 HH_mr_info_t     *mr_info_p)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_query_mr: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_query_mr : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_query_mr: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_query_mr: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_mrwm_query_mr(xhh_hob_p->mrwm, lkey, mr_info_p);
}



/******************************************************************************
 *  Function:     XHH_hob_deregister_mr <==> XHH_mrwm_deregister_mr
 *****************************************************************************/
HH_ret_t XHH_hob_deregister_mr(HH_hca_hndl_t  hca_hndl, 
                                      VAPI_lkey_t      lkey)
{
  XHH_hob_t  xhh_hob_p;
  HH_ret_t   rc;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_deregister_mr: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_deregister_mr : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_deregister_mr: ERROR : No device registered\n");
      return HH_EAGAIN;
  }


  if (xhh_hob_p->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_deregister_mr: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  rc = XHH_mrwm_deregister_mr(xhh_hob_p->mrwm, lkey);
  return rc;
}

/******************************************************************************
 *  Function:     XHH_hob_alloc_mw <==> XHH_mrwm_alloc_mw
 *****************************************************************************/
HH_ret_t XHH_hob_alloc_mw(HH_hca_hndl_t  hca_hndl, 
                                 HH_pd_hndl_t     pd,
                                 IB_rkey_t        *initial_rkey_p)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_alloc_mw: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_alloc_mw : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }

  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_alloc_mw: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_alloc_mw: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_mrwm_alloc_mw(xhh_hob_p->mrwm, pd, initial_rkey_p);
}


/******************************************************************************
 *  Function:     XHH_hob_query_mw <==> XHH_mrwm_query_mw
 *****************************************************************************/
HH_ret_t XHH_hob_query_mw(HH_hca_hndl_t  hca_hndl,
                                 IB_rkey_t        initial_rkey,
                                 IB_rkey_t        *current_rkey_p,
                                 HH_pd_hndl_t     *pd_p)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_query_mw: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_query_mw : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_query_mw: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_query_mw: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  MTL_TRACE1("%s: -KL- called for key 0x%x", __func__,initial_rkey);
  return XHH_mrwm_query_mw(xhh_hob_p->mrwm, initial_rkey, current_rkey_p, pd_p);
}


/******************************************************************************
 *  Function:     XHH_hob_free_mw <==> XHH_mrwm_free_mw
 *****************************************************************************/
HH_ret_t XHH_hob_free_mw(HH_hca_hndl_t  hca_hndl, 
                                IB_rkey_t        initial_rkey)
{
  XHH_hob_t  xhh_hob_p;
  HH_ret_t   rc;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_free_mw: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_free_mw : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_free_mw: ERROR : No device registered\n");
      return HH_EAGAIN;
  }


  if (xhh_hob_p->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_free_mw: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  rc = XHH_mrwm_free_mw(xhh_hob_p->mrwm, initial_rkey);
  return rc;
}

/******************************************************************************
 *  Function:     XHH_hob_create_cq <==> XHH_cqm_create_cq
 *****************************************************************************/
HH_ret_t XHH_hob_create_cq(HH_hca_hndl_t  hca_hndl, 
                                  MOSAL_protection_ctx_t  user_prot_context, 
                                  void                    *cq_ul_resources_p,
                                  HH_cq_hndl_t            *cq_p)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_create_cq: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_create_cq : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_create_cq: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);

  if (xhh_hob_p->cqm == (XHH_cqm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_create_cq: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_cqm_create_cq(xhh_hob_p->cqm, user_prot_context, xhh_hob_p->compl_eq,
                           xhh_hob_p->ib_eq, 
                           (XHH_cq_ul_resources_t*)cq_ul_resources_p, cq_p);
}

/******************************************************************************
 *  Function:     XHH_hob_resize_cq <==> XHH_cqm_modify_cq
 *****************************************************************************/
HH_ret_t XHH_hob_resize_cq(HH_hca_hndl_t  hca_hndl, 
                           HH_cq_hndl_t            cq,
                           void                    *cq_ul_resources_p)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_resize_cq: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_resize_cq : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_resize_cq: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->cqm == (XHH_cqm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_resize_cq: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_cqm_resize_cq(xhh_hob_p->cqm, cq, (XHH_cq_ul_resources_t*)cq_ul_resources_p);
}

/******************************************************************************
 *  Function:     XHH_hob_query_cq <==> XHH_cqm_query_cq
 *****************************************************************************/
HH_ret_t XHH_hob_query_cq(HH_hca_hndl_t  hca_hndl, 
                                 HH_cq_hndl_t            cq,
                                 VAPI_cqe_num_t          *num_o_cqes_p)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_query_cq: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_query_cq : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_query_cq: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->cqm == (XHH_cqm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_query_cq: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_cqm_query_cq(xhh_hob_p->cqm, cq, num_o_cqes_p);
}


/******************************************************************************
 *  Function:     XHH_hob_destroy_cq <==> XHH_cqm_destroy_cq
 *****************************************************************************/
HH_ret_t XHH_hob_destroy_cq(HH_hca_hndl_t  hca_hndl, 
                                 HH_cq_hndl_t    cq)
{
  XHH_hob_t  xhh_hob_p;
  HH_ret_t   rc = HH_OK;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_destroy_cq: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_destroy_cq : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_destroy_cq: ERROR : No device registered\n");
      return HH_EAGAIN;
  }


  if (xhh_hob_p->cqm == (XHH_cqm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_destroy_cq: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  rc = XHH_cqm_destroy_cq(xhh_hob_p->cqm, cq);
  return rc;
}


/******************************************************************************
 *  Function:     XHH_hob_create_qp <==> XHH_qpm_create_qp
 *****************************************************************************/
HH_ret_t XHH_hob_create_qp(HH_hca_hndl_t  hca_hndl, 
                           HH_qp_init_attr_t  *init_attr_p, 
                           void               *qp_ul_resources_p, 
                           IB_wqpn_t          *qpn_p)
{
  XHH_hob_t  xhh_hob_p;
  HH_ret_t   rc = HH_OK;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_create_qp: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_create_qp : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_create_qp: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->qpm == (XHH_qpm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_create_qp: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  if ((rc=XHH_hob_check_qp_init_attrs(xhh_hob_p,init_attr_p,FALSE)) != HH_OK) {
      MTL_ERROR1("XHH_hob_create_qp: ERROR : requested capabilities exceed HCA limits\n");
      return rc;
  }

  return XHH_qpm_create_qp(xhh_hob_p->qpm, init_attr_p, 0, (XHH_qp_ul_resources_t*)qp_ul_resources_p, qpn_p);
}

/******************************************************************************
 *  Function:     XHH_hob_get_special_qp <==> XHH_qpm_get_special_qp
 *****************************************************************************/
HH_ret_t XHH_hob_get_special_qp(HH_hca_hndl_t  hca_hndl,
                                VAPI_special_qp_t  qp_type,
                                IB_port_t          port, 
                                HH_qp_init_attr_t  *init_attr_p, 
                                void               *qp_ul_resources_p, 
                                IB_wqpn_t          *sqp_hndl_p)
{
  XHH_hob_t  xhh_hob_p;
  HH_ret_t   rc = HH_OK;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_get_special_qp: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_get_special_qp : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_get_special_qp: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);

  if (xhh_hob_p->qpm == (XHH_qpm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_get_special_qp: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  if ((rc=XHH_hob_check_qp_init_attrs(xhh_hob_p,init_attr_p,TRUE)) != HH_OK) {
      MTL_ERROR1("XHH_hob_get_special_qp: ERROR : requested capabilities exceed HCA limits\n");
      return rc;
  }

  return XHH_qpm_get_special_qp(xhh_hob_p->qpm, qp_type, port, init_attr_p, (XHH_qp_ul_resources_t*)qp_ul_resources_p, sqp_hndl_p);
}

/******************************************************************************
 *  Function:     XHH_hob_modify_qp <==> XHH_qpm_modify_qp
 *****************************************************************************/
HH_ret_t XHH_hob_modify_qp(HH_hca_hndl_t  hca_hndl, 
                                  IB_wqpn_t            qpn, 
                                  VAPI_qp_state_t      cur_qp_state,
                                  VAPI_qp_attr_t       *qp_attr_p,
                                  VAPI_qp_attr_mask_t  *qp_attr_mask_p)
{
  HH_ret_t	 ret;
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_modify_qp: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_modify_qp : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_modify_qp: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);

  if (xhh_hob_p->qpm == (XHH_qpm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_modify_qp: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }
  
  ret = XHH_qpm_modify_qp(xhh_hob_p->qpm, qpn, cur_qp_state, qp_attr_p, qp_attr_mask_p);

  return ret;
}


/******************************************************************************
 *  Function:     XHH_hob_query_qp <==> XHH_qpm_query_qp
 *****************************************************************************/
HH_ret_t XHH_hob_query_qp(HH_hca_hndl_t  hca_hndl, 
                                 IB_wqpn_t          qpn, 
                                 VAPI_qp_attr_t     *qp_attr_p)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_query_qp: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_query_qp : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_query_qp: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->qpm == (XHH_qpm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_query_qp: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_qpm_query_qp(xhh_hob_p->qpm, qpn, qp_attr_p);
}


/******************************************************************************
 *  Function:     XHH_hob_destroy_qp <==> XHH_qpm_destroy_qp
 *****************************************************************************/
HH_ret_t XHH_hob_destroy_qp(HH_hca_hndl_t  hca_hndl, 
                                   IB_wqpn_t          qpn)
{
  XHH_hob_t  xhh_hob_p;
  HH_ret_t   rc = HH_OK;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_destroy_qp: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_destroy_qp : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_destroy_qp: ERROR : No device registered\n");
      return HH_EAGAIN;
  }


  if (xhh_hob_p->qpm == (XHH_qpm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_destroy_qp: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  rc = XHH_qpm_destroy_qp(xhh_hob_p->qpm, qpn);
  return rc;
}

/* HH_create_srq */
HH_ret_t XHH_hob_create_srq(HH_hca_hndl_t hca_hndl, HH_pd_hndl_t pd, void *srq_ul_resources_p, 
                            HH_srq_hndl_t     *srq_p)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("%s: NOT IN TASK CONTEXT)\n", __func__);
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("%s : ERROR : Invalid HCA handle\n", __func__);
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("%s: ERROR : No device registered\n", __func__);
      return HH_EINVAL_HCA_HNDL;
  }
  TEST_RETURN_FATAL(xhh_hob_p);

  if (xhh_hob_p->srqm == (XHH_srqm_t)XHH_INVALID_HNDL) {
      /* check QPM to make sure that is not just unsupported configuration */
      if (xhh_hob_p->qpm == (XHH_qpm_t)XHH_INVALID_HNDL) {
	  MTL_ERROR1("%s: ERROR : HCA device has not yet been opened\n", __func__);
	  return HH_EAGAIN;
      }	else {
	  MTL_ERROR1("%s: SRQs are not supported in this HCA configuration\n", __func__);
	  return HH_ENOSYS;
      }
  }

  return XHH_srqm_create_srq(xhh_hob_p->srqm, pd, srq_ul_resources_p, srq_p);
}

HH_ret_t XHH_hob_query_srq(HH_hca_hndl_t hca_hndl, HH_srq_hndl_t srq, u_int32_t *limit_p)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("%s: NOT IN TASK CONTEXT)\n", __func__);
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("%s : ERROR : Invalid HCA handle\n", __func__);
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("%s: ERROR : No device registered\n", __func__);
      return HH_EINVAL_HCA_HNDL;
  }
  TEST_RETURN_FATAL(xhh_hob_p);

  if (xhh_hob_p->srqm == (XHH_srqm_t)XHH_INVALID_HNDL) {
      /* check QPM to make sure that is not just unsupported configuration */
      if (xhh_hob_p->qpm == (XHH_qpm_t)XHH_INVALID_HNDL) {
	  MTL_ERROR1("%s: ERROR : HCA device has not yet been opened\n", __func__);
	  return HH_EAGAIN;
      }	else {
	  MTL_ERROR1("%s: SRQs are not supported in this HCA configuration\n", __func__);
	  return HH_ENOSYS;
      }
  }

  return XHH_srqm_query_srq(xhh_hob_p->srqm, srq, limit_p);
}

HH_ret_t XHH_hob_modify_srq(HH_hca_hndl_t hca_hndl, HH_srq_hndl_t srq, void *srq_ul_resources_p)
{
  XHH_hob_t  xhh_hob_p;

  MTL_TRACE4("%s: srq=0x%X)\n", __func__, srq);

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("%s: NOT IN TASK CONTEXT)\n", __func__);
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("%s : ERROR : Invalid HCA handle\n", __func__);
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("%s: ERROR : No device registered\n", __func__);
      return HH_EINVAL_HCA_HNDL;
  }
  TEST_RETURN_FATAL(xhh_hob_p);

  if (xhh_hob_p->srqm == (XHH_srqm_t)XHH_INVALID_HNDL) {
      /* check QPM to make sure that is not just unsupported configuration */
      if (xhh_hob_p->qpm == (XHH_qpm_t)XHH_INVALID_HNDL) {
	  MTL_ERROR1("%s: ERROR : HCA device has not yet been opened\n", __func__);
	  return HH_EAGAIN;
      }	else {
	  MTL_ERROR1("%s: SRQs are not supported in this HCA configuration\n", __func__);
	  return HH_ENOSYS;
      }
  }

  return XHH_srqm_modify_srq(xhh_hob_p->srqm, srq, srq_ul_resources_p);
}


HH_ret_t XHH_hob_destroy_srq(HH_hca_hndl_t hca_hndl, HH_srq_hndl_t srq)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("%s: NOT IN TASK CONTEXT)\n", __func__);
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("%s : ERROR : Invalid HCA handle\n", __func__);
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("%s: ERROR : No device registered\n", __func__);
      return HH_EINVAL_HCA_HNDL;
  }
  TEST_RETURN_FATAL(xhh_hob_p);

  if (xhh_hob_p->srqm == (XHH_srqm_t)XHH_INVALID_HNDL) {
      /* check QPM to make sure that is not just unsupported configuration */
      if (xhh_hob_p->qpm == (XHH_qpm_t)XHH_INVALID_HNDL) {
	  MTL_ERROR1("%s: ERROR : HCA device has not yet been opened\n", __func__);
	  return HH_EAGAIN;
      }	else {
	  MTL_ERROR1("%s: SRQs are not supported in this HCA configuration\n", __func__);
	  return HH_ENOSYS;
      }
  }

  return XHH_srqm_destroy_srq(xhh_hob_p->srqm, srq);
}


/******************************************************************************
 *  Function:     XHH_hob_process_local_mad <==> XHH_qpm_process_local_mad
 *****************************************************************************/
HH_ret_t XHH_hob_process_local_mad(
                          HH_hca_hndl_t        hca_hndl,
                          IB_port_t            port,
                          IB_lid_t             slid, /* For Mkey violation trap */
                          EVAPI_proc_mad_opt_t proc_mad_opts,
                          void *               mad_in_p,
                          void *               mad_out_p)
{
    XHH_hob_t  xhh_hob_p;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1("XHH_hob_process_local_mad: NOT IN TASK CONTEXT)\n");
        return HH_ERR;
    }

    if (hca_hndl == NULL) {
        MTL_ERROR1("XHH_hob_process_local_mad : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);

    if (xhh_hob_p == NULL) {
        MTL_ERROR1("XHH_hob_process_local_mad: ERROR : No device registered\n");
        return HH_EAGAIN;
    }
    TEST_RETURN_FATAL(xhh_hob_p);


    if (xhh_hob_p->qpm == (XHH_qpm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_hob_destroy_qp: ERROR : HCA device has not yet been opened\n");
        return HH_EAGAIN;
    }

    return XHH_qpm_process_local_mad(xhh_hob_p->qpm,port,slid,proc_mad_opts,mad_in_p,mad_out_p);
}
    
/******************************************************************************
 *  Function:     XHH_hob_get_qp1_pkey <==> XHH_qpm_get_qp1_pkey
 *****************************************************************************/
HH_ret_t XHH_hob_get_qp1_pkey(
                          HH_hca_hndl_t  hca_hndl,
                          IB_port_t  port,/*IN */
                          VAPI_pkey_t* pkey_p/*OUT*/)

{
    XHH_hob_t  xhh_hob_p;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hca_hndl == NULL) {
      MTL_ERROR1(MT_FLFMT("Invalid HCA handle"));
      return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);

    if (xhh_hob_p == NULL) {
      MTL_ERROR1(MT_FLFMT("Invalid HCA handle"));
      return HH_EAGAIN;
    }
    TEST_RETURN_FATAL(xhh_hob_p);

    if (xhh_hob_p->qpm == (XHH_qpm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1(MT_FLFMT("HCA %s has not yet been opened"),xhh_hob_p->dev_name);
      return HH_EAGAIN;
    }

    return XHH_qpm_get_qp1_pkey(xhh_hob_p->qpm,port,pkey_p);
}

HH_ret_t XHH_hob_get_pkey(
                          HH_hca_hndl_t  hca_hndl,
                          IB_port_t  port,/*IN */
                          VAPI_pkey_ix_t pkey_index, /*IN*/
                          VAPI_pkey_t* pkey_p/*OUT*/)
{
    XHH_hob_t  xhh_hob_p;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hca_hndl == NULL) {
      MTL_ERROR1(MT_FLFMT("Invalid HCA handle"));
      return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);

    if (xhh_hob_p == NULL) {
      MTL_ERROR1(MT_FLFMT("Invalid HCA handle"));
      return HH_EAGAIN;
    }
    TEST_RETURN_FATAL(xhh_hob_p);

    if (xhh_hob_p->qpm == (XHH_qpm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1(MT_FLFMT("HCA %s has not yet been opened"),xhh_hob_p->dev_name);
      return HH_EAGAIN;
    }

    return XHH_qpm_get_pkey(xhh_hob_p->qpm,port,pkey_index,pkey_p);
}

/******************************************************************************
 *  Function:     XHH_hob_get_sgid <==> XHH_qpm_get_sgid
 *****************************************************************************/
HH_ret_t XHH_hob_get_sgid(
                          HH_hca_hndl_t  hca_hndl,
                          IB_port_t  port,/*IN */
                          u_int8_t index,/*IN*/
                          IB_gid_t* gid_p/*OUT*/)

{
    XHH_hob_t  xhh_hob_p;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hca_hndl == NULL) {
      MTL_ERROR1(MT_FLFMT("Invalid HCA handle"));
      return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);

    if (xhh_hob_p == NULL) {
      MTL_ERROR1(MT_FLFMT("Invalid HCA handle"));
      return HH_EAGAIN;
    }
    TEST_RETURN_FATAL(xhh_hob_p);

    if (xhh_hob_p->qpm == (XHH_qpm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1(MT_FLFMT("HCA %s has not yet been opened"),xhh_hob_p->dev_name);
      return HH_EAGAIN;
    }

    return XHH_qpm_get_sgid(xhh_hob_p->qpm,port,index,gid_p);
}

HH_ret_t XHH_hob_get_legacy_mode(XHH_hob_t xhh_hob_p,MT_bool *p_mode)
{
    if (xhh_hob_p == NULL) {
      MTL_ERROR1(MT_FLFMT("Invalid HCA handle"));
      return HH_EAGAIN;
    }

	*p_mode = xhh_hob_p->module_flags.legacy_sqp;

	return HH_OK;
}


MT_bool XHH_is_sqp_sched(XHH_hob_t xhh_hob_p)
{
    if (xhh_hob_p == NULL) {
      MTL_ERROR1(MT_FLFMT("Invalid HCA handle"));
      return FALSE;
    }

	return xhh_hob_p->module_flags.eneble_sqp_sched;
    
}


 /******************************************************************************
 *  Function:     XHH_hob_create_eec <==> XHH_eecm_create_eec
 *****************************************************************************/
HH_ret_t XHH_hob_create_eec(HH_hca_hndl_t  hca_hndl, 
                                   HH_rdd_hndl_t  rdd, 
                                   IB_eecn_t      *eecn_p)
{
#if 0
  XHH_hob_t  xhh_hob_p;

  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_create_eec: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_create_eec : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);
  TEST_RETURN_FATAL(xhh_hob_p);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_create_eec: ERROR : No device registered\n");
      return HH_EAGAIN;
  }


  if (xhh_hob_p->eecm == (XHH_eecm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_create_eec: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_eecm_create_eec(xhh_hob_p->eecm, rdd, eecn_p);
#else
  return HH_ENOSYS;
#endif
}

/******************************************************************************
 *  Function:     XHH_hob_modify_eec <==> XHH_eecm_modify_eec
 *****************************************************************************/
HH_ret_t XHH_hob_modify_eec(HH_hca_hndl_t  hca_hndl,
                                   IB_eecn_t           eecn, 
                                   VAPI_qp_state_t     cur_ee_state, 
                                   VAPI_qp_attr_t      *ee_attr_p, 
                                   VAPI_qp_attr_mask_t *ee_attr_mask_p)
{
#if 0
  XHH_hob_t  xhh_hob_p;

  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_modify_eec: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_modify_eec : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_modify_eec: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->eecm == (XHH_eecm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_modify_eec: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_eecm_modify_eec(xhh_hob_p->eecm, eecn, cur_ee_state, ee_attr_p, ee_attr_mask_p);
#else
  return HH_ENOSYS;
#endif
}



/******************************************************************************
 *  Function:     XHH_hob_query_eec <==> XHH_eecm_query_eec
 *****************************************************************************/
HH_ret_t XHH_hob_query_eec(HH_hca_hndl_t  hca_hndl,
                                   IB_eecn_t        eecn, 
                                   VAPI_qp_attr_t   *ee_attr_p)
{
#if 0
  XHH_hob_t  xhh_hob_p;

  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_query_eec: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_query_eec : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_query_eec: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->eecm == (XHH_eecm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_query_eec: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_eecm_query_eec(xhh_hob_p->eecm, eecn, ee_attr_p);
#else
  return HH_ENOSYS;
#endif
}



/******************************************************************************
 *  Function:     XHH_hob_destroy_eec <==> XHH_eecm_destroy_eec
 *****************************************************************************/
HH_ret_t XHH_hob_destroy_eec(HH_hca_hndl_t  hca_hndl,
                                    IB_eecn_t      eecn)
{
#if 0
  XHH_hob_t  xhh_hob_p;
  HH_ret_t   rc = HH_OK;

  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_destroy_eec: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_destroy_eec : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_destroy_eec: ERROR : No device registered\n");
      return HH_EAGAIN;
  }


  if (xhh_hob_p->eecm == (XHH_eecm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_destroy_eec: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  rc = XHH_eecm_destroy_eec(xhh_hob_p->eecm, eecn);
  return rc;
#else
  return HH_ENOSYS;
#endif
}

/******************************************************************************
 *  Function:     XHH_hob_attach_to_multicast <==> XHH_mcgm_attach_qp
 *****************************************************************************/
HH_ret_t XHH_hob_attach_to_multicast(
                          HH_hca_hndl_t  hca_hndl, 
                          IB_wqpn_t      qpn,
                          IB_gid_t       dgid)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_attach_to_multicast: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_attach_to_multicast : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_attach_to_multicast: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);

  if (xhh_hob_p->mcgm == (XHH_mcgm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_attach_to_multicast: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_mcgm_attach_qp(xhh_hob_p->mcgm, qpn, dgid);
}


/******************************************************************************
 *  Function:     XHH_hob_detach_from_multicast <==> XHH_mcgm_detach_qp
 *****************************************************************************/
HH_ret_t XHH_hob_detach_from_multicast(
                          HH_hca_hndl_t  hca_hndl, 
                          IB_wqpn_t      qpn,
                          IB_gid_t       dgid)
{
  XHH_hob_t  xhh_hob_p;
  HH_ret_t   rc = HH_OK;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_detach_from_multicast: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_detach_from_multicast : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_detach_from_multicast: ERROR : No device registered\n");
      return HH_EAGAIN;
  }

  if (xhh_hob_p->mcgm == (XHH_mcgm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1( "XHH_hob_detach_from_multicast: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  rc = XHH_mcgm_detach_qp(xhh_hob_p->mcgm, qpn, dgid);
  return rc;
}

VIP_delay_unlock_t XHH_hob_get_delay_unlock(XHH_hob_t hob)
{ 
    if (hob == NULL) {
        return NULL;
    } else {
        return (hob->delay_unlocks);
    }
}

/******************************************************************************
 *  Function:     XHH_hob_get_num_ports
 *
 *  Description:  Gets number of physical ports configured for HCA  
 *
 *  input:
 *                hca_hndl
 *  output: 
 *                num_ports_p - 1 or 2
 *  returns:
 *                HH_OK
 *                HH_EINVAL
 *                HH_EINVAL_HCA_HNDL
 *                HH_ERR
 *
 *  Comments:     Does MAD query to get the data in real time. 
 *
 *****************************************************************************/
HH_ret_t  XHH_hob_get_num_ports( HH_hca_hndl_t  hca_hndl,
                               IB_port_t      *num_ports_p)
{
    XHH_hob_t  xhh_hob_p;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hca_hndl == NULL) {
      MTL_ERROR1(MT_FLFMT("Invalid HCA handle"));
      return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);
    
    if (xhh_hob_p == NULL) {
      MTL_ERROR1(MT_FLFMT("Invalid HCA handle"));
      return HH_EAGAIN;
    }
    TEST_RETURN_FATAL(xhh_hob_p);
    *num_ports_p = (IB_port_t)(xhh_hob_p->dev_lims.num_ports);
    return HH_OK;
}

/******************************************************************************
 *  Function:     XHH_hob_external_fatal
 *
 *  Description:  Artificially generates the fatal error flow from an external 
 *                call  
 *
 *  input:
 *                hca_hndl
 *  output: 
 *                
 *  returns:
 *                HH_OK
 *                HH_EINVAL_HCA_HNDL
 *                HH_EFATAL -- driver already in fatal state
 *
 *  Comments:     Meant to enable performing a VAPI STOP following a device reset. 
 *
 *****************************************************************************/
HH_ret_t  XHH_hob_external_fatal( HH_hca_hndl_t  hca_hndl)
{
    XHH_hob_t  xhh_hob_p;

    if (hca_hndl == NULL) {
      MTL_ERROR1(MT_FLFMT("Invalid HCA handle"));
      return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);
    
    if (xhh_hob_p == NULL) {
      MTL_ERROR1(MT_FLFMT("Invalid HCA handle"));
      return HH_EINVAL_HCA_HNDL;
    }
    TEST_RETURN_FATAL(xhh_hob_p);
    MTL_ERROR1(MT_FLFMT("%s: Generating a fatal error for device %s"), 
               __func__, hca_hndl->dev_desc);
    XHH_hob_fatal_error(xhh_hob_p, XHH_FATAL_EXTERNAL,VAPI_EV_SYNDROME_NONE);
    return HH_OK;
}

#if defined(MT_SUSPEND_QP)
/******************************************************************************
 *  Function:     XHH_hob_suspend_qp <==> XHH_qpm_suspend_qp
 *****************************************************************************/
HH_ret_t XHH_hob_suspend_qp(HH_hca_hndl_t  hca_hndl, 
                            IB_wqpn_t      qpn, 
                            MT_bool        suspend_flag)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_suspend_qp: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_suspend_qp : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_suspend_qp: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->qpm == (XHH_qpm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_suspend_qp: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_qpm_suspend_qp(xhh_hob_p->qpm, qpn, suspend_flag);
}
/******************************************************************************
 *  Function:     XHH_hob_suspend_cq <==> XHH_qpm_suspend_cq
 *****************************************************************************/
HH_ret_t XHH_hob_suspend_cq(HH_hca_hndl_t  hca_hndl, 
                            HH_cq_hndl_t   cq, 
                            MT_bool        do_suspend)
{
  XHH_hob_t  xhh_hob_p;

  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
  if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
      MTL_ERROR1("XHH_hob_suspend_qp: NOT IN TASK CONTEXT)\n");
      return HH_ERR;
  }

  if (hca_hndl == NULL) {
      MTL_ERROR1("XHH_hob_suspend_qp : ERROR : Invalid HCA handle\n");
      return HH_EINVAL_HCA_HNDL;
  }
  xhh_hob_p = XHHOBP(hca_hndl);

  if (xhh_hob_p == NULL) {
      MTL_ERROR1("XHH_hob_suspend_qp: ERROR : No device registered\n");
      return HH_EAGAIN;
  }
  TEST_RETURN_FATAL(xhh_hob_p);


  if (xhh_hob_p->cqm == (XHH_cqm_t)XHH_INVALID_HNDL) {
      MTL_ERROR1("XHH_hob_suspend_qp: ERROR : HCA device has not yet been opened\n");
      return HH_EAGAIN;
  }

  return XHH_cqm_suspend_cq(xhh_hob_p->cqm, cq, do_suspend);
}
#endif


#define NUM_REQ_VECTORS 4
/*
 *  XHH_get_num_req_vectors
 */
unsigned int XHH_get_num_req_vectors(void)
{
  return NUM_REQ_VECTORS;
}

HH_ret_t XHH_hob_get_ver_info ( /*IN*/  XHH_hob_t        hob, 
                                /*OUT*/ XHH_ver_info_t  *version_p )
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_get_ver_info: ERROR : No device registered\n");
        return HH_EINVAL;
    }
    
    memcpy(version_p, &(hob->version_info), sizeof(XHH_ver_info_t));
    return HH_OK;
}

HH_ret_t XHH_hob_get_cmd_if ( /*IN*/  XHH_hob_t   hob, 
                              /*OUT*/ XHH_cmd_t   *cmd_if_p )
{
    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_get_cmd_if: ERROR : No device registered\n");
        return HH_EINVAL;
    }


    if (hob->cmd == XHH_CMDIF_INVALID_HANDLE) {
        MTL_ERROR1("XHH_hob_get_cmd_if: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }

    *cmd_if_p = hob->cmd;
    return HH_OK;
}


HH_ret_t XHH_hob_hca_get_cmd_if ( /*IN*/  HH_hca_hndl_t  hca_hndl, 
                              /*OUT*/ XHH_cmd_t   *cmd_if_p )
{
    XHH_hob_t  xhh_hob_p;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hca_hndl == NULL) {
      MTL_ERROR1(MT_FLFMT("Invalid HCA handle"));
      return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);
    
    if (xhh_hob_p == NULL) {
      MTL_ERROR1(MT_FLFMT("Invalid HCA handle"));
      return HH_EINVAL;
    }
    return (XHH_hob_get_cmd_if ( xhh_hob_p, cmd_if_p ));
}

HH_ret_t XHH_hob_get_mrwm ( /*IN*/ XHH_hob_t hob, 
                            /*OUT*/ XHH_mrwm_t *mrwm_p )
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_get_mrwm: ERROR : No device registered\n");
        return HH_EINVAL;
    }


    if (hob->mrwm == (XHH_mrwm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_hob_get_mrwm: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }
   
    *mrwm_p = hob->mrwm;
   return HH_OK;
}

HH_ret_t XHH_hob_get_qpm ( /*IN*/ XHH_hob_t hob, 
                           /*OUT*/ XHH_qpm_t *qpm_p )
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_get_qpm: ERROR : No device registered\n");
        return HH_EINVAL;
    }


    if (hob->qpm == (XHH_qpm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_hob_get_qpm: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }
   
    *qpm_p = hob->qpm;
   return HH_OK;
}

HH_ret_t XHH_hob_get_cqm ( /*IN*/ XHH_hob_t hob, 
                           /*OUT*/ XHH_cqm_t *cqm_p )
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_get_cqm: ERROR : No device registered\n");
        return HH_EINVAL;
    }


    if (hob->cqm == (XHH_cqm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_hob_get_cqm: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }
   
    *cqm_p = hob->cqm;
   return HH_OK;
}

HH_ret_t XHH_hob_get_eventp ( /*IN*/ XHH_hob_t hob, 
                             /*OUT*/ XHH_eventp_t *eventp_p )
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_get_eventp: ERROR : No device registered\n");
        return HH_EINVAL;
    }


    if (hob->eventp== (XHH_eventp_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_hob_get_eventp: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }
   
    *eventp_p = hob->eventp;
   return HH_OK;
}

HH_ret_t XHH_hob_get_hca_hndl ( /*IN*/  XHH_hob_t hob, 
                                /*OUT*/ HH_hca_hndl_t *hca_hndl_p )
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (hob == NULL) {
        MTL_ERROR1("XHH_hob_get_hca_hndl: ERROR : No device registered\n");
        return HH_EINVAL;
    }
   
    *hca_hndl_p = hob->hh_hca_hndl;
    return HH_OK;
}


HH_ret_t   XHH_hob_halt_hca(/*IN*/ XHH_hob_t hob)
{
    XHH_cmd_status_t  stat;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    stat = XHH_cmd_CLOSE_HCA(hob->cmd, TRUE); // HALT the HCA (panic close)
    if (stat != XHH_CMD_STAT_OK) {
	MTL_ERROR1(MT_FLFMT("%s: HALT HCA returned 0x%x"), __func__,stat);  
    }
    return HH_OK;
}

HH_ret_t XHH_hob_reset_device(XHH_hw_props_t *hw_props_p)
{
	MT_virt_addr_t     va;
	HH_ret_t	   ret;
	MOSAL_iomem_t      temp_iomem;
	call_result_t      rc;
	/* Do the device RESET */
	
	rc = MOSAL_iomem_map(MOSAL_get_kernel_prot_ctx(),
                 MOSAL_MEM_FLAGS_NO_CACHE | MOSAL_MEM_FLAGS_PERM_READ | MOSAL_MEM_FLAGS_PERM_WRITE,
	         (MT_phys_addr_t) (hw_props_p->cr_base + 0xF0010),
	         (MT_size_t) 4,
	         &temp_iomem, &va);
	if ( rc == MT_OK ) {
	     /* perform sw reset */
	      MTL_DEBUG1(MT_FLFMT("%s: PERFORMING SW RESET. pa="PHYS_ADDR_FMT" va="VIRT_ADDR_FMT),
	                      __func__, hw_props_p->cr_base + 0xF0010, va);
	      MOSAL_MMAP_IO_WRITE_DWORD(va,MOSAL_cpu_to_be32(0x00000001));
	      /* sleep for one second, per PRM */
	      MOSAL_delay_execution(1000000);
	      MOSAL_iomem_unmap(temp_iomem, MOSAL_get_kernel_prot_ctx());
	      ret = HH_OK;
	} else {
	    MTL_ERROR1(MT_FLFMT("%s: failed to map pa="PHYS_ADDR_FMT" to a virtual address. DEVICE RESET NOT PERFORMED"),
			    __func__, hw_props_p->cr_base + 0xF0010);
	    ret = HH_ERR;
	}
	return(ret);
}

static VAPI_event_syndrome_t get_fatal_err_syndrome(XHH_hob_t hob) 
{
    u_int32_t    temp;
    int          i;
    
    if ((hob->fw_error_buf_start_va != (MT_virt_addr_t)(MT_ulong_ptr_t) NULL) && 
        (hob->fw_error_buf != NULL) && 
        (hob->fw_props.error_buf_size > 0) ) {
        MOSAL_MMAP_IO_READ_BUF_DWORD(hob->fw_error_buf_start_va,
                                 hob->fw_error_buf,hob->fw_props.error_buf_size);
        /* check for non-zero data in fw catastrophic error buffer */
        temp = 0;
        for (i = 0; i < (int) hob->fw_props.error_buf_size; i++) {
            temp |= hob->fw_error_buf[i];
        }
        if (temp == 0) {
            return VAPI_EV_SYNDROME_NONE; 
        } else {
            /* Have non-zero data. print out the syndrome details, and return general category */
            for (i = 0; i < (int) hob->fw_props.error_buf_size; i++) {
                MTL_ERROR1(MT_FLFMT("get_fatal_err_syndrome: FW CATASTR ERRBUF[%d] = 0x%x"),
                           i, MOSAL_be32_to_cpu(hob->fw_error_buf[i]));
            }
            switch( (MOSAL_be32_to_cpu(hob->fw_error_buf[0])>>24) & 0xFF) {
            case ADD_DEV_PREFIX(IF_EV_CATAS_ERR_FW_INTERNAL_ERR):
                return VAPI_CATAS_ERR_FW_INTERNAL;
            case ADD_DEV_PREFIX(IF_EV_CATAS_ERR_MISBEHAVED_UAR_PAGE):
                return VAPI_CATAS_ERR_MISBEHAVED_UAR_PAGE;
            case ADD_DEV_PREFIX(IF_EV_CATAS_ERR_UPLINK_BUS_ERR):
                return VAPI_CATAS_ERR_UPLINK_BUS_ERR;
            case ADD_DEV_PREFIX(IF_EV_CATAS_ERR_HCA_DDR_DATA_ERR):
                return VAPI_CATAS_ERR_HCA_DDR_DATA_ERR;
            case ADD_DEV_PREFIX(IF_EV_CATAS_ERR_INTERNAL_PARITY_ERR):
                return VAPI_CATAS_ERR_INTERNAL_PARITY_ERR;
            case ADD_DEV_PREFIX(IF_EV_CATAS_ERR_EQ_ERR):
                return VAPI_CATAS_ERR_EQ_OVERFLOW;
            default:
                return VAPI_CATAS_ERR_GENERAL; 
            }
        }
    } else {
        /* no access to the fw cat error buffer */
        return VAPI_CATAS_ERR_GENERAL; 
    }
}

#define CHECK_FATAL_INTERVAL 1000000

int XHH_hob_fatal_err_thread(void  *arg)
{
   XHH_hob_t                hob_p;
   XHH_hob_cat_err_thread_t  *fatal_thread_obj_p = (XHH_hob_cat_err_thread_t  *)arg;
   HH_event_record_t fatal_ev_rec;
   XHH_cmd_t    cmd_if;
   XHH_eventp_t eventp;
   call_result_t     mosal_ret;
   int loop =1;
   MT_size_t  check_fatal_interval = CHECK_FATAL_INTERVAL;
   
   MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
   MTL_TRACE2("%s: Initializing\n", __func__);

   hob_p  = (XHH_hob_t)(fatal_thread_obj_p->hob);
   MOSAL_thread_set_name(&fatal_thread_obj_p->mto, "cleanup_thread");


   /* signal that thread is up */
   fatal_thread_obj_p->have_fatal = FALSE;
   MOSAL_syncobj_signal(&fatal_thread_obj_p->start_sync);

   while (loop) {
   
     MTL_TRACE3("%s: about to wait on fatal error signal\n", __func__);
     mosal_ret=MOSAL_syncobj_waiton(&hob_p->fatal_thread_obj.fatal_err_sync,
                                    check_fatal_interval);
      switch (mosal_ret) {
      case MT_EINTR:
      case MT_OK: /* this is a real wake-up */
        if (mosal_ret == MT_EINTR || hob_p->fatal_thread_obj.have_fatal == FALSE) {
            MTL_DEBUG1(MT_FLFMT("%s: GOT termination request"), __func__);
            /* if no fatal error, just return */
            MOSAL_syncobj_signal(&fatal_thread_obj_p->stop_sync);
            return 1;
        }
        MTL_ERROR1(MT_FLFMT("%s: RECEIVED FATAL ERROR WAKEUP"), __func__);
        loop=0;
        break;
      case MT_ETIMEDOUT:
        if (get_fatal_err_syndrome(hob_p) != VAPI_EV_SYNDROME_NONE){
          MTL_ERROR1(MT_FLFMT("%s: FATAL ERROR DETECTED IN WAKE-UP"), __func__);
          XHH_hob_fatal_error(hob_p, XHH_FATAL_EVENT, VAPI_CATAS_ERR_GENERAL);
		  /* eliminate timeout for fatal checking, since already have our fatal */
		  check_fatal_interval = MOSAL_SYNC_TIMEOUT_INFINITE;
        }
        break;
      default:
        MTL_ERROR1(MT_FLFMT("%s: Internal error"), __func__);
        break;
      }

   } /* while 1 */

   /* fatal error processing */
   if (XHH_hob_get_cmd_if(hob_p, &cmd_if) == HH_OK) {
       XHH_cmd_handle_fatal(cmd_if);
   }
   if (XHH_hob_get_eventp(hob_p,&eventp) == HH_OK) {
       XHH_eventp_handle_fatal(eventp);
   }

   /* Halt HCA here */
   if (hob_p->module_flags.fatal_delay_halt == 0) {
       MTL_DEBUG1(MT_FLFMT("%s: halting the HCA"), __func__);
       XHH_hob_halt_hca(hob_p);
   }

   MOSAL_spinlock_dpc_lock(&hob_p->fatal_spl);
   /* turn off STARTED bit, and turn on HALTED bit */
   hob_p->xhh_state &= ~(XHH_STATE_FATAL_START);
   hob_p->xhh_state |= XHH_STATE_FATAL_HCA_HALTED;
   MOSAL_syncobj_signal(&hob_p->xhh_fatal_complete_syncobj);
   MOSAL_spinlock_unlock(&hob_p->fatal_spl);

   /* INVOKE THE async event callback with fatal error */
   if (hob_p->hh_hca_hndl != NULL && hob_p->async_eventh != NULL) {
       MTL_TRACE1(MT_FLFMT("%s: INVOKE ASYNC CALLBACK"), __func__);  
       memset(&fatal_ev_rec, 0, sizeof(HH_event_record_t));
       fatal_ev_rec.etype = VAPI_LOCAL_CATASTROPHIC_ERROR;
       fatal_ev_rec.syndrome = hob_p->fatal_syndrome;
	   /* way for the verification to know that fatal event got using MSIX */
       fatal_ev_rec.event_modifier.port = hob_p->from_msix;
       (*(hob_p->async_eventh))(hob_p->hh_hca_hndl, &fatal_ev_rec, hob_p->async_ev_private_context);
   }
   MOSAL_syncobj_signal(&fatal_thread_obj_p->stop_sync);
   return 0;
}

/****************************************************************************************
 * name:      XHH_hob_fatal_error
 * function:  Called by cmdif or event manager when a fatal error is detected 
 * args:
 * returns:
 * descr:     This procedure initiates fatal error handling. It signals the fatal error
 *            thread to start its fatal processing.  This separation is necessary because
 *            this procedure may be called in a DPC context, but fatal processing cannot
 *            take place in this context (needs command interface).
 ****************************************************************************************/

HH_ret_t   XHH_hob_fatal_error(/*IN*/ XHH_hob_t hob,
                               /*IN*/ XHH_fatal_err_t  fatal_err_type,
                               /*IN*/ VAPI_event_syndrome_t  syndrome)
{
    XHH_cmd_t    cmd_if;
    XHH_eventp_t eventp;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    FUNC_IN;
    MTL_DEBUG1(MT_FLFMT("%s: device=%s, err_type=%d, syndrome=%d"), __func__,  
               hob->dev_name, fatal_err_type, syndrome);

    if (hob == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: ERROR : No device registered"), __func__);  
        MT_RETURN(HH_EAGAIN);
    }
    
    /* make sure that only one invocation is allowed */
    MOSAL_spinlock_dpc_lock(&hob->fatal_spl);
    if ((hob->xhh_state & XHH_STATE_HAVE_ANY_FATAL) != 0) {
        /* already in FATAL state */
        MTL_DEBUG4(MT_FLFMT("%s: already in FATAL state"), __func__);  
        MOSAL_spinlock_unlock(&hob->fatal_spl);
        MT_RETURN(HH_OK);
    }

    MTL_ERROR1(MT_FLFMT("%s: device=%s, err_type=%d, syndrome=%d"), __func__,  
               hob->dev_name, fatal_err_type, syndrome);
    
    switch(fatal_err_type) {
    /* get syndrome from iomapped firmware memory */
    case XHH_FATAL_MASTER_ABORT:  /* detected master abort */
        hob->fatal_syndrome = VAPI_CATAS_ERR_MASTER_ABORT;
        break;
    case XHH_FATAL_GOBIT:         /* GO bit of HCR remains set (i.e., stuck) */
        hob->fatal_syndrome = VAPI_CATAS_ERR_GO_BIT;
        break;
    case XHH_FATAL_CMD_TIMEOUT:   /* timeout on a command execution */
        hob->fatal_syndrome = VAPI_CATAS_ERR_CMD_TIMEOUT;
        break;
    case XHH_FATAL_EQ_OVF:        /* an EQ has overflowed */
        hob->fatal_syndrome = VAPI_CATAS_ERR_EQ_OVERFLOW;
        break;
    case XHH_FATAL_EVENT:    /* firmware has generated a LOCAL CATASTROPHIC ERR event */
        hob->fatal_syndrome = get_fatal_err_syndrome(hob);
        break;
    case XHH_FATAL_CR:            /* unexpected read from CR-space */
        hob->fatal_syndrome = VAPI_CATAS_ERR_EQ_OVERFLOW;
        break;
    case XHH_FATAL_TOKEN:         /* invalid token on command completion */
        hob->fatal_syndrome = VAPI_CATAS_ERR_FATAL_TOKEN;
        break;
    case XHH_FATAL_EXTERNAL:         /* invalid token on command completion */
        hob->fatal_syndrome = VAPI_CATAS_ERR_FATAL_EXTERNAL;
        break;
    case XHH_FATAL_NONE: 
    default:
        hob->fatal_syndrome = VAPI_CATAS_ERR_GENERAL;
    }

    MTL_ERROR1(MT_FLFMT("%s: Fatal Event Syndrome = %s (%d)"), 
               __func__,VAPI_event_syndrome_sym(hob->fatal_syndrome), hob->fatal_syndrome);

    if (hob->xhh_state == XHH_STATE_RUNNING) {
        /* make use of thread to perform HALT and signal user apps */
        hob->xhh_state |= XHH_STATE_FATAL_START;
    } else  {
        /* creating, opening, closing, or destroying HCA.
         * Indicate HCA_HALTED directly.
         */
        hob->xhh_state |= XHH_STATE_FATAL_HCA_HALTED;
    }
    MOSAL_spinlock_unlock(&hob->fatal_spl);

    /* notify cmd and eventp objects, if they exist */
    if (XHH_hob_get_cmd_if(hob, &cmd_if) == HH_OK) {
        XHH_cmd_notify_fatal(cmd_if, fatal_err_type);
    }
    
    if (XHH_hob_get_eventp(hob,&eventp) == HH_OK) {
       XHH_eventp_notify_fatal(eventp, fatal_err_type);
    }

	/* remember if we came from MSIX */
	hob->from_msix = (syndrome == VAPI_CATAS_ERR_MSI) ? 1:0;
	
    /* now, signal the fatal error thread, ONLY IF WE WERE IN RUNNING STATE */ 
    if ((hob->xhh_state & XHH_STATE_RUNNING) != 0) {
        hob->fatal_thread_obj.have_fatal = TRUE;
        MTL_TRACE1(MT_FLFMT("%s: signalling fatal thread"), __func__);  
        MOSAL_syncobj_signal(&hob->fatal_thread_obj.fatal_err_sync);
    }

    MT_RETURN(HH_OK);
}

HH_ret_t  XHH_hob_get_fatal_syncobj(XHH_hob_t xhh_hob_p, MOSAL_syncobj_t *syncobj)
{
  MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (xhh_hob_p == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: ERROR : No device registered"), __func__);  
        return HH_EAGAIN;
    }
    
    if (syncobj == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: ERROR : NULL syncobj return parameter"), __func__);  
        return HH_EINVAL;
    }

    *syncobj = xhh_hob_p->xhh_fatal_complete_syncobj;
    return HH_OK;
}

HH_ret_t XHH_hob_wait_if_fatal(XHH_hob_t xhh_hob_p, MT_bool *had_fatal)
{
    XHH_hob_state_t   state;

    FUNC_IN;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (xhh_hob_p == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: Received NULL HOB pointer"), __func__);  
        *had_fatal = FALSE;
        MT_RETURN(HH_OK);
    }

    /*get fatal state value */
    MOSAL_spinlock_dpc_lock(&(xhh_hob_p->fatal_spl));
    state = xhh_hob_p->xhh_state;
    MOSAL_spinlock_unlock(&(xhh_hob_p->fatal_spl));
    
    MTL_DEBUG4(MT_FLFMT("%s: FATAL STATE=%d"), __func__, state);  

    if ((state & XHH_STATE_HAVE_ANY_FATAL) == 0) {
        *had_fatal = FALSE;
        MT_RETURN(HH_OK);
    }
    
    /* We were in running state.  Wait for fatal thread to complete HCA HALT */
    if ((state & XHH_STATE_FATAL_START) != 0) {
        MOSAL_syncobj_waiton_ui(&xhh_hob_p->xhh_fatal_complete_syncobj, 10000000);
    }
    
    /* We are in the FATAL_HCA_HALTED compound state */
    *had_fatal = TRUE;
    MT_RETURN(HH_OK);

}

HH_ret_t  XHH_hob_restart(/*IN*/ HH_hca_hndl_t  hca_hndl)
{
    XHH_hw_props_t     *hw_props_p = NULL;
    u_int32_t          hca_seq_num;
    XHH_module_flags_t mod_flags;
    XHH_hob_t          xhh_hob_p;
    HH_hca_hndl_t      new_hh_hca_hndl; 
    HH_ret_t           rc;
    

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    FUNC_IN;
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1(MT_FLFMT("%s: NOT IN TASK CONTEXT"), __func__);  
        return HH_ERR;
    }

    if (hca_hndl == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: ERROR : Invalid HCA handle"), __func__);  
        return HH_EINVAL_HCA_HNDL;
    }
    xhh_hob_p = XHHOBP(hca_hndl);

    if (xhh_hob_p == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: ERROR : no device registered"), __func__);  
        return HH_EAGAIN;
    }

    if ((xhh_hob_p->xhh_state & XHH_STATE_FATAL_HCA_HALTED) == 0) {
        MTL_ERROR1(MT_FLFMT("%s: HCA is not halted (state 0x%x)"), __func__, xhh_hob_p->xhh_state);  
        return HH_ERR;
    }

    hw_props_p = (XHH_hw_props_t *)MALLOC(sizeof(XHH_hw_props_t));
    if (hw_props_p == NULL) {
       MT_COV_MEM_FLTR_START
       return HH_EAGAIN;
       MT_COV_MEM_FLTR_END
    }
    memset(hw_props_p, 0, sizeof(XHH_hw_props_t));

    XHH_hob_get_init_params(xhh_hob_p, hw_props_p, &hca_seq_num, &mod_flags);

    /* PCI reset is done in destroy, if have catastrophic error */
    MTL_ERROR1(MT_FLFMT("%s: destroying old HOB"), __func__);  
    rc = XHH_hob_destroy_internal(hca_hndl, TRUE);
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("%s: cannot destroy old HOB (ret=%d)"), __func__, rc);  
    }
    
    MTL_ERROR1(MT_FLFMT("%s: creating new HOB"), __func__);  
    rc = XHH_hob_create_internal(hw_props_p,hca_seq_num,&mod_flags,hca_hndl,&new_hh_hca_hndl);
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("%s: cannot create new HOB (ret=%d)"), __func__, rc);  
    }
    FREE(hw_props_p);
    return rc;
}

HH_ret_t XHH_hob_waitfor_hca_closed(HH_hca_hndl_t  hca_hndl)
{
    XHH_hob_t  hob_p;

    MT_RETURN_IF_LOW_STACK(XHH_WATERMARK);
    if (MOSAL_get_exec_ctx() != MOSAL_IN_TASK) {
        MTL_ERROR1("XHH_hob_waitfor_hca_closed: NOT IN TASK CONTEXT)\n");
        return HH_ERR;
    }

    if (hca_hndl == NULL) {
        MTL_ERROR1("XHH_hob_waitfor_hca_closed : ERROR : Invalid HCA handle\n");
        return HH_EINVAL_HCA_HNDL;
    }
    hob_p = XHHOBP(hca_hndl);
    
    while (hotplug_support && (hca_hndl->status != HH_HCA_STATUS_CLOSED)) {
	    MOSAL_delay_execution(5000000);
    }

    return HH_OK;
}

unsigned long XHH_show_hotplug_support(void) {
	return hotplug_support;
}

void XHH_set_hotplug_support(unsigned long hotp) {
	hotplug_support = (hotp == 0) ? 0 : 1;
}
