/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_XHH_HOB_COMMON_IMP_H
#define H_XHH_HOB_COMMON_IMP_H

#define TEST_RETURN_FATAL(hob)  if ((hob->xhh_state & XHH_STATE_HAVE_ANY_FATAL) != 0) {          \
                           MTL_ERROR1(MT_FLFMT("%s: Device in FATAL state"), __func__);  \
                           return HH_EFATAL;                               \
                           }
#define TEST_CMD_FATAL(ret)  if (ret == XHH_CMD_STAT_EFATAL) {                          \
                           MTL_ERROR1(MT_FLFMT("%s: cmdif returned FATAL"), __func__);  \
                           return HH_EFATAL;                                            \
                           }
#define FREE_RSRC_RET(hob)    rc = ((have_fatal == TRUE) ? HH_OK : rc) ; return rc
#define DECLARE_FATAL_VARS    MT_bool  have_fatal=FALSE

#define XHH_WATERMARK        350 /* our bigest function now takes 160 */
#define IB_MAX_MESSAGE_SIZE  (1 << 31)
			   
/* the memory key used for the UDAVM in XHH (privileged mode) */
#define XHH_UDAVM_PRIV_RESERVED_LKEY      0

#define XHH_INVALID_HNDL       ((MT_ulong_ptr_t) (-1L))

#define MAX_INT_VECTORS 32 /* max number of vectors supported by the driver */

typedef enum {
  MT_INT_MODE_PIN,
  MT_INT_MODE_MSI
}
MT_interrupt_mode_t;

typedef struct XHH_intr_props_st {
    MOSAL_IRQ_ID_t    irq[MAX_INT_VECTORS];
    u_int8_t    intr_pin;
    MT_interrupt_mode_t  int_mode;
    unsigned int num_allocated_vectors; /* number of vectors allocated */
} XHH_intr_props_t;

typedef struct XHH_hw_props_st {
    MOSAL_pci_dev_t   pci_dev;
    u_int8_t          bus;
    u_int8_t          dev_func;
    u_int16_t         device_id;
    u_int16_t         pci_vendor_id;
    u_int32_t         hw_ver;
    MT_phys_addr_t       cr_base;
    MT_phys_addr_t       uar_base;
    MT_phys_addr_t       ddr_base;
    XHH_intr_props_t  interrupt_props;
} XHH_hw_props_t;


#define MSIX_CAPID 0x11
#define MSIX_DWSZ 3
#define MSI_CAPID 0x5
#define MSI_DWSZ 4
#define PCIX_CAPID  0x7
#define PCIX_DWSZ   4
#define VPD_CAPID		0x03
#define VPD_CAPDWSZ   2
#define PM_CAPID		0x1
#define PM_CAPDWSZ   2
#define PCIEXPRESS_CAPID  0x10
#define PCIEXPRESS_DWSZ   9

#define MAX_PCI_CAP 8
#define MAX_CAP_DW 9

typedef struct XHH_cap_container_st {
  u_int8_t cap_ptr;
  int cap_size_dw;
  u_int32_t cap_data[MAX_CAP_DW];
}
XHH_cap_container_t;

typedef struct XHH_pci_config_space_container_st {
  u_int32_t seg1[16];
  XHH_cap_container_t clist[MAX_PCI_CAP];
  int num_cap;
}
XHH_pci_config_space_container_t;



/******************************************************************************
 *  Function:     XHH_hob_create 
 *
 *  Arguments:
 *        hw_props_p - pointer to HW properties.
 *        hca_seq_num  - a sequence number assigned to this HCA to differentiate it 
 *                        from other HCAs on this host
 *        mod_flags - ptr to struct containing flags passed during module initialization
 *                    (e.g. insmod)
 *        hh_hndl_p - Returned HH context handle.  HOB object is accessed via device field 
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameter
 *
 *  Description:
 *    Allocate XHH_hob object and register the device in HH.
 */
HH_ret_t    XHH_hob_create(/*IN*/  XHH_hw_props_t    *hw_props_p,
                           /*IN*/  u_int32_t         hca_seq_num,
                           /*IN*/  XHH_module_flags_t *mod_flags,
                           /*OUT*/ HH_hca_hndl_t     *hh_hndl_p );

/******************************************************************************
 *  Function:     XHH_hob_destroy 
 *
 *  Arguments:
 *        hca_hndl - The HCA handle allocated on device registration 
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL_HCA_HNDL - The HCA either is not opened or is unknown
 *
 *  Description:
 *    Deregister given device from HH and free all associated resources.
 *    If the HCA is still open, perform XHH_hob_close_hca() before freeing the XHH_hob.
 */
HH_ret_t    XHH_hob_destroy(HH_hca_hndl_t hca_hndl);

/******************************************************************************
 *  Function:     XHH_hob_create_internal
 *
 *  Description:  Creates the HOB object for an HCA, and registers it in HH  
 *
 *  input:
 *                hw_props_p -- PCI properties (BARs, etc)
 *                hca_seq_num  - a sequence number assigned to this HCA to differentiate it 
 *                               from other HCAs on this host
 *                mod_flags   - flags passed in at module initialization (e.g., insmod)
 *                curr_hh_hndl  - current HH handle - valid for fatal error restart. NULL otherwise
 *                                (NULL value indicates new creation)
 *  output: 
 *                hh_hndl_p  - returned HH handle
 *  returns:
 *                HH_OK
 *                HH_EAGAIN
 *                HH_ERR -- other errors
 *
 *  Comments:     This function involves the following steps: 
 *                    1.Allocate XHH_hob data context. 
 *                    2.Register HCA in HH (i.e.call HH_add_hca_dev()) for a new registration, or set
 *                      the HH object state to HH_HCA_STATUS_CLOSED for a re-activation in fatal error
 *                      restart
 *
 *                Also initializes the HOB mutex for controlling Open HCA and Close HCA 
 *
 *****************************************************************************/
HH_ret_t    XHH_hob_create_internal(/*IN*/  XHH_hw_props_t    *hw_props_p,
                           /*IN*/  u_int32_t         hca_seq_num,
                           /*IN*/  XHH_module_flags_t *mod_flags,
			   /*IN*/  HH_hca_hndl_t     curr_hh_hndl,
                           /*OUT*/ HH_hca_hndl_t     *hh_hndl_p );


/******************************************************************************
 *  Function:     XHH_hob_destroy_internal
 *
 *  Description:  Deregister given device from HH and free the HH object.
 *
 *  input:
 *                hca_hndl
 *                restarting -- TRUE if called as part of fatal error restart.
 *  output: 
 *                none
 *  returns:
 *                HH_OK
 *                HH_EINVAL_HCA_HNDL
 *
 *  Comments:     If HCA is still open,XHH_hob_close_hca() is 
 *                invoked before freeing the XHH_hob.
 *                If called as part of restarting, the HH device is NOT de-registered.
 *
 *                Returns HH_EINVAL_HCA_HNDL if any function called internally fails  
 *
 *****************************************************************************/
HH_ret_t    XHH_hob_destroy_internal (HH_hca_hndl_t hca_hndl, MT_bool restarting);

/******************************************************************************
 *  Function:     XHH_hob_get_ver_info 
 *
 *  Arguments:
 *        hob
 *        version_p - Returned version information 
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameters
 *
 *  Description:
 *    Get version information of device associated with given hob.
 */
HH_ret_t XHH_hob_get_ver_info ( /*IN*/ XHH_hob_t        hob, 
                                /*OUT*/ XHH_ver_info_t  *version_p );

/******************************************************************************
 *  Function:     XHH_hob_get_cmd_if 
 *
 *  Arguments:
 *        hob
 *        cmd_if_p - Included command interface object 
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameters
 *
 *  Description:
 *    Get a handle to associated command interface object.
 */
HH_ret_t XHH_hob_get_cmd_if ( /*IN*/   XHH_hob_t   hob, 
                              /*OUT*/ XHH_cmd_t    *cmd_if_p );

/******************************************************************************
 *  Function:     XHH_hob_hca_get_cmd_if
 *
 *  Description:  Returns handle of cmdif object given HH_hca_hndl  
 *
 *  Arguments:
 *        hca_hndl - hh hca_handle
 *        cmd_if_p - Included command interface object 
 *
 *  returns:
 *                HH_OK
 *                HH_EINVAL
 *                HH_EINVAL_HCA_HNDL
 *
 *  Description:
 *    Get a handle to associated command interface object.
 *
 *****************************************************************************/
HH_ret_t XHH_hob_hca_get_cmd_if ( /*IN*/  HH_hca_hndl_t  hca_hndl, 
                              /*OUT*/ XHH_cmd_t   *cmd_if_p );

/******************************************************************************
 *  Function:     XHH_hob_get_mrwm 
 *
 *  Arguments:
 *        hob
 *        mrwm_p - Included XHH_mrwm object 
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameters
 *
 *  Description:
 *    Get a handle to associated memory regions/windows management object.
 */
HH_ret_t XHH_hob_get_mrwm ( /*IN*/ XHH_hob_t hob, 
                            /*OUT*/ XHH_mrwm_t *mrwm_p );

/******************************************************************************
 *  Function:     XHH_hob_get_hca_hndl 
 *
 *  Arguments:
 *        hob
 *        hca_hndl_p - Included XHH_mrwm object 
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameters
 *
 *  Description:
 *    Get a handle to the HH hca object.
 */
HH_ret_t XHH_hob_get_hca_hndl ( /*IN*/  XHH_hob_t hob, 
                                /*OUT*/ HH_hca_hndl_t *hca_hndl_p );
/******************************************************************************
 *  Function:     XHH_hob_get_lamm 
 *
 *  Arguments:
 *        hob
 *        uldm_p - Included XHH_lamm object 
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameters
 *
 *  Description:
 *    Get a handle to DDR memory management object.
 */
HH_ret_t XHH_hob_get_lamm ( /*IN*/ XHH_hob_t hob, 
                            /*OUT*/ XHH_lamm_t *lamm_p );

/******************************************************************************
 *  Function:     XHH_hob_get_qpm 
 *
 *  Arguments:
 *        hob
 *        eventp_p - Included XHH_qpm object 
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameters
 *
 *  Description:
 *    Get a handle to qp management object.
 */
HH_ret_t XHH_hob_get_qpm ( /*IN*/ XHH_hob_t hob, 
                          /*OUT*/ XHH_qpm_t *qpm_p );

/******************************************************************************
 *  Function:     XHH_hob_get_cqm 
 *
 *  Arguments:
 *        hob
 *        eventp_p - Included XHH_cqm object 
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameters
 *
 *  Description:
 *    Get a handle to cq management object.
 */
HH_ret_t XHH_hob_get_cqm ( /*IN*/ XHH_hob_t hob, 
                          /*OUT*/ XHH_cqm_t *cqm_p );

/******************************************************************************
 *  Function:     XHH_hob_get_eventp 
 *
 *  Arguments:
 *        hob
 *        eventp_p - Included XHH_eventp object 
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameters
 *
 *  Description:
 *    Get a handle to event management object.
 */
HH_ret_t XHH_hob_get_eventp ( /*IN*/ XHH_hob_t hob, 
                              /*OUT*/ XHH_eventp_t *eventp_p );

/******************************************************************************
 *  Function:     XHH_hob_set_comp_eventh
 *
 *  Description:  Sets completion event handler for VIP layer (below vapi).  Used internally
 *                by VAPI  
 *
 *  input:
 *                hca_hndl
 *                event -  pointer to handler function
 *                private_data - pointer to context data provided to handler
 *  returns:
 *                HH_OK
 *                HH_EAGAIN
 *                HH_EINVAL_HCA_HNDL
 *                HH_ERR
 *
 *  Comments:     Initial (dummy) handler is provided during open_hca.  Therefore,
 *                the function XHH_eventp_replace_handler is used to register the handler.
 *
 *****************************************************************************/
HH_ret_t XHH_hob_set_comp_eventh(HH_hca_hndl_t      hca_hndl,
                                        HH_comp_eventh_t   event,
                                        void*              private_data);

/******************************************************************************
 *  Function:     XHH_hob_set_async_eventh
 *
 *  Description:  Sets async handler for VIP layer (below vapi).  Used internally
 *                by VAPI  
 *
 *  input:
 *                hca_hndl
 *                event -  pointer to handler function
 *                private_data - pointer to context data provided to handler
 *  returns:
 *                HH_OK
 *                HH_EAGAIN
 *                HH_EINVAL_HCA_HNDL
 *                HH_ERR
 *
 *  Comments:     Initial (dummy) handler is provided during open_hca.  Therefore,
 *                the function XHH_eventp_replace_handler is used to register the handler.
 *
 *****************************************************************************/
HH_ret_t XHH_hob_set_async_eventh( HH_hca_hndl_t      hca_hndl,
                                          HH_async_eventh_t  event,
                                          void*              private_data);

/******************************************************************************
 *  Function:     XHH_hob_fatal_error 
 *
 *  Arguments:
 *        hob
 *        fatal_err_type - type of fatal error which occurred 
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameters
 *
 *  Description:
 *    Initiates centralized fatal error handling when a fatal error is detected
 */
HH_ret_t   XHH_hob_fatal_error(/*IN*/ XHH_hob_t hob,
                               /*IN*/ XHH_fatal_err_t  fatal_err_type,
                               /*IN*/ VAPI_event_syndrome_t  syndrome);

HH_ret_t   XHH_hob_get_init_params(/*IN*/ XHH_hob_t  thh_hob_p, 
                   /*OUT*/  XHH_hw_props_t     *hw_props_p,
                   /*OUT*/  u_int32_t          *hca_seq_num,
                   /*OUT*/  XHH_module_flags_t *mod_flags);
HH_ret_t XHH_hob_restart(HH_hca_hndl_t  hca_hndl);

HH_ret_t XHH_hob_get_state(XHH_hob_t thh_hob_p, XHH_hob_state_t *fatal_state);
HH_ret_t XHH_hob_get_fatal_syncobj(XHH_hob_t thh_hob_p, MOSAL_syncobj_t *syncobj);
HH_ret_t XHH_hob_wait_if_fatal(XHH_hob_t thh_hob_p, MT_bool *had_fatal);

HH_ret_t XHH_hob_modify(
                        HH_hca_hndl_t        hca_hndl,
                        IB_port_t            port_num,
                        VAPI_hca_attr_t      *hca_attr_p,
                        VAPI_hca_attr_mask_t *hca_attr_mask_p);

HH_ret_t XHH_hob_query_port_prop(HH_hca_hndl_t  hca_hndl,
                                    IB_port_t           port_num,
                                    VAPI_hca_port_t     *hca_port_p );

HH_ret_t XHH_hob_get_pkey_tbl(HH_hca_hndl_t  hca_hndl,
                                     IB_port_t     port_num,
                                     u_int16_t     tbl_len_in,
                                     u_int16_t     *tbl_len_out,
                                     IB_pkey_t     *pkey_tbl_p);

HH_ret_t  XHH_hob_init_pkey_tbl( HH_hca_hndl_t  hca_hndl,
                                      IB_port_t      port_num,
                                      u_int16_t      tbl_len_in,
                                      u_int16_t*     tbl_len_out,
                                      IB_pkey_t     *pkey_tbl_p);

HH_ret_t  XHH_hob_get_gid_tbl( HH_hca_hndl_t  hca_hndl,
                                      IB_port_t      port,
                                      u_int16_t      tbl_len_in,
                                      u_int16_t*     tbl_len_out,
                                      IB_gid_t*      param_gid_p);

HH_ret_t  XHH_hob_init_gid_tbl( HH_hca_hndl_t  hca_hndl,
                                      IB_port_t      port,
                                      u_int16_t      tbl_len_in,
                                      u_int16_t*     tbl_len_out,
                                      IB_gid_t*      param_gid_p);


HH_ret_t XHH_hob_query_port_prop(HH_hca_hndl_t  hca_hndl,IB_port_t port_num,VAPI_hca_port_t *hca_port_p );
HH_ret_t XHH_hob_alloc_ul_res(HH_hca_hndl_t hca_hndl,MOSAL_protection_ctx_t prot_ctx,void *hca_ul_resources_p);
HH_ret_t XHH_hob_free_ul_res(HH_hca_hndl_t hca_hndl,void *hca_ul_resources_p);
HH_ret_t XHH_hob_alloc_pd(HH_hca_hndl_t hca_hndl, MOSAL_protection_ctx_t prot_ctx, void * pd_ul_resources_p,HH_pd_hndl_t *pd_num_p);
HH_ret_t XHH_hob_free_pd(HH_hca_hndl_t hca_hndl, HH_pd_hndl_t pd_num);
HH_ret_t XHH_hob_alloc_rdd(HH_hca_dev_t *hh_dev_p, HH_rdd_hndl_t *rdd_p);
HH_ret_t XHH_hob_free_rdd(HH_hca_dev_t *hh_dev_p, HH_rdd_hndl_t rdd);
HH_ret_t XHH_hob_register_mr(HH_hca_hndl_t  hca_hndl, 
                                    HH_mr_t       *mr_props_p, 
                                    VAPI_lkey_t   *lkey_p, 
                                    IB_rkey_t     *rkey_p);

HH_ret_t XHH_hob_reregister_mr(HH_hca_hndl_t  hca_hndl,
                               VAPI_lkey_t    lkey,  
                                      VAPI_mr_change_t  change_mask,
                                      HH_mr_t           *mr_props_p, 
                                      VAPI_lkey_t*       lkey_p, 
                                      IB_rkey_t         *rkey_p);

HH_ret_t XHH_hob_register_smr(HH_hca_hndl_t  hca_hndl, 
                                    HH_smr_t       *mr_props_p, 
                                    VAPI_lkey_t    *lkey_p, 
                                    IB_rkey_t      *rkey_p);

HH_ret_t XHH_hob_query_mr(HH_hca_hndl_t  hca_hndl, 
                                 VAPI_lkey_t      lkey, 
                                 HH_mr_info_t     *mr_info_p);

HH_ret_t XHH_hob_deregister_mr(HH_hca_hndl_t  hca_hndl, 
                                      VAPI_lkey_t      lkey);

HH_ret_t XHH_hob_alloc_mw(HH_hca_hndl_t  hca_hndl, 
                                 HH_pd_hndl_t     pd,
                                 IB_rkey_t        *initial_rkey_p);

HH_ret_t XHH_hob_query_mw(HH_hca_hndl_t  hca_hndl,
                                 IB_rkey_t        initial_rkey,
                                 IB_rkey_t        *current_rkey_p,
                                 HH_pd_hndl_t     *pd_p);

HH_ret_t XHH_hob_free_mw(HH_hca_hndl_t  hca_hndl, 
                                IB_rkey_t        initial_rkey);

HH_ret_t XHH_hob_create_cq(HH_hca_hndl_t  hca_hndl, 
                                  MOSAL_protection_ctx_t  user_prot_context, 
                                  void                    *cq_ul_resources_p,
                                  HH_cq_hndl_t            *cq_p);

HH_ret_t XHH_hob_resize_cq(HH_hca_hndl_t  hca_hndl, 
                           HH_cq_hndl_t            cq,
                           void                    *cq_ul_resources_p);

HH_ret_t XHH_hob_query_cq(HH_hca_hndl_t  hca_hndl, 
                                 HH_cq_hndl_t            cq,
                                 VAPI_cqe_num_t          *num_o_cqes_p);

HH_ret_t XHH_hob_destroy_cq(HH_hca_hndl_t  hca_hndl, 
                                 HH_cq_hndl_t    cq);

HH_ret_t XHH_hob_create_qp(HH_hca_hndl_t  hca_hndl, 
                           HH_qp_init_attr_t  *init_attr_p, 
                           void               *qp_ul_resources_p, 
                           IB_wqpn_t          *qpn_p);

HH_ret_t XHH_hob_get_special_qp(HH_hca_hndl_t  hca_hndl,
                                VAPI_special_qp_t  qp_type,
                                IB_port_t          port, 
                                HH_qp_init_attr_t  *init_attr_p, 
                                void               *qp_ul_resources_p, 
                                IB_wqpn_t          *sqp_hndl_p);

HH_ret_t XHH_hob_modify_qp(HH_hca_hndl_t  hca_hndl, 
                                  IB_wqpn_t            qpn, 
                                  VAPI_qp_state_t      cur_qp_state,
                                  VAPI_qp_attr_t       *qp_attr_p,
                                  VAPI_qp_attr_mask_t  *qp_attr_mask_p);

HH_ret_t XHH_hob_query_qp(HH_hca_hndl_t  hca_hndl, 
                                 IB_wqpn_t          qpn, 
                                 VAPI_qp_attr_t     *qp_attr_p);

HH_ret_t XHH_hob_destroy_qp(HH_hca_hndl_t  hca_hndl, 
                                   IB_wqpn_t          qpn);

HH_ret_t XHH_hob_check_qp_init_attrs (XHH_hob_t hob, HH_qp_init_attr_t * init_attr_p,
                                           MT_bool is_special_qp );

HH_ret_t XHH_hob_create_srq(HH_hca_hndl_t hca_hndl, 
							HH_pd_hndl_t pd, 
							void *srq_ul_resources_p, 
							HH_srq_hndl_t     *srq_p);

HH_ret_t XHH_hob_query_srq(HH_hca_hndl_t hca_hndl, HH_srq_hndl_t srq, 
			   u_int32_t *limit_p);

HH_ret_t XHH_hob_modify_srq(HH_hca_hndl_t hca_hndl, HH_srq_hndl_t srq, 
			    void *srq_ul_resources_p);

HH_ret_t XHH_hob_destroy_srq(HH_hca_hndl_t hca_hndl, HH_srq_hndl_t srq);

HH_ret_t XHH_hob_process_local_mad(
                          HH_hca_hndl_t        hca_hndl,
                          IB_port_t            port,
                          IB_lid_t             slid, /* For Mkey violation trap */
                          EVAPI_proc_mad_opt_t proc_mad_opts,
                          void *               mad_in_p,
                          void *               mad_out_p);

HH_ret_t XHH_hob_get_qp1_pkey(
                          HH_hca_hndl_t  hca_hndl,
                          IB_port_t  port,/*IN */
                          VAPI_pkey_t* pkey_p/*OUT*/);

HH_ret_t XHH_hob_get_pkey(
                          HH_hca_hndl_t  hca_hndl,
                          IB_port_t  port,/*IN */
                          VAPI_pkey_ix_t pkey_index, /*IN*/
                          VAPI_pkey_t* pkey_p/*OUT*/);

HH_ret_t XHH_hob_get_sgid(
                          HH_hca_hndl_t  hca_hndl,
                          IB_port_t  port,/*IN */
                          u_int8_t index,/*IN*/
                          IB_gid_t* gid_p/*OUT*/);

HH_ret_t XHH_hob_get_legacy_mode(XHH_hob_t xhh_hob_p,MT_bool *p_mode);

MT_bool XHH_is_sqp_sched(XHH_hob_t xhh_hob_p);

HH_ret_t XHH_hob_create_eec(HH_hca_hndl_t  hca_hndl, 
                                   HH_rdd_hndl_t  rdd, 
                                   IB_eecn_t      *eecn_p);

HH_ret_t XHH_hob_modify_eec(HH_hca_hndl_t  hca_hndl,
                                   IB_eecn_t           eecn, 
                                   VAPI_qp_state_t     cur_ee_state, 
                                   VAPI_qp_attr_t      *ee_attr_p, 
                                   VAPI_qp_attr_mask_t *ee_attr_mask_p);

HH_ret_t XHH_hob_query_eec(HH_hca_hndl_t  hca_hndl,
                                   IB_eecn_t        eecn, 
                                   VAPI_qp_attr_t   *ee_attr_p);

HH_ret_t XHH_hob_destroy_eec(HH_hca_hndl_t  hca_hndl,
                                    IB_eecn_t      eecn);

HH_ret_t XHH_hob_attach_to_multicast(
                          HH_hca_hndl_t  hca_hndl, 
                          IB_wqpn_t      qpn,
                          IB_gid_t       dgid);

HH_ret_t XHH_hob_detach_from_multicast(
                          HH_hca_hndl_t  hca_hndl, 
                          IB_wqpn_t      qpn,
                          IB_gid_t       dgid);

HH_ret_t XHH_hob_alloc_fmr(HH_hca_hndl_t hca_hndl, HH_pd_hndl_t pd,
                           VAPI_mrw_acl_t acl,MT_size_t max_pages,u_int8_t log2_page_sz,VAPI_lkey_t* last_lkey_p);
HH_ret_t  XHH_hob_map_fmr(HH_hca_hndl_t hca_hndl,VAPI_lkey_t last_lkey,
                          EVAPI_fmr_map_t* map_p,VAPI_lkey_t* lkey_p,IB_rkey_t* rkey_p);
HH_ret_t  XHH_hob_unmap_fmr(HH_hca_hndl_t hca_hndl,u_int32_t num_of_fmrs_to_unmap, VAPI_lkey_t*  last_lkeys_array);
HH_ret_t  XHH_hob_free_fmr(HH_hca_hndl_t  hca_hndl,VAPI_lkey_t     last_lkey);


VIP_delay_unlock_t XHH_hob_get_delay_unlock(XHH_hob_t hob);

HH_ret_t  XHH_hob_get_num_ports( HH_hca_hndl_t  hca_hndl,
                               IB_port_t      *num_ports_p);

HH_ret_t  XHH_hob_external_fatal( HH_hca_hndl_t  hca_hndl);

#if defined(MT_SUSPEND_QP)
HH_ret_t XHH_hob_suspend_qp(HH_hca_hndl_t  hca_hndl, 
                            IB_wqpn_t      qpn, 
                            MT_bool        suspend_flag);

HH_ret_t XHH_hob_suspend_cq(HH_hca_hndl_t  hca_hndl, 
                            HH_cq_hndl_t   cq, 
                            MT_bool        do_suspend);
#endif

/*
 *  XHH_get_num_req_vectors
 *
 *  return te number of vectors that the hob wants for its operation.
 *  It may though get along with fewer
 */
unsigned int XHH_get_num_req_vectors(void);

HH_ret_t XHH_hob_reset_device(XHH_hw_props_t *hw_props_p);


unsigned long XHH_show_hotplug_support(void);
void XHH_set_hotplug_support(unsigned long hotp);

#endif

