/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/
 
#include <hob_pci_caps.h>

/*
 *  mosal_find_capbility_ptr
 */
static call_result_t mosal_find_capbility_ptr(MOSAL_pci_dev_t   pci_dev,
                                              u_int8_t cap_id,
                                              u_int8_t *cap_ptr_p)
{
  call_result_t rc;
  u_int8_t cap_ptr;
  u_int32_t cap_val_dw;
  int c=0;

  /* read cap pointer */
  rc = MOSAL_PCI_read_config(pci_dev, 0x34, 1, &cap_ptr);
  if ( rc != MT_OK ) {
    MTL_ERROR1(MT_FLFMT("%s: failed reading cap pointer - %s"), __func__, mtl_strerror(rc));
    return rc; 
  }

  while ( c < 64  ) {
    rc = MOSAL_PCI_read_config(pci_dev, cap_ptr, 4, &cap_val_dw);
    if ( rc != MT_OK ) {
      MTL_ERROR1(MT_FLFMT("%s: failed reading dword at address 0x%x - %s"), __func__, cap_ptr, mtl_strerror(rc));
      return rc; 
    }
    
    if ( (cap_val_dw&0xff) == cap_id ) {
      *cap_ptr_p = cap_ptr;
      return MT_OK;
    }
    cap_ptr = (u_int8_t)(cap_val_dw>>8) & 0xfc; /* mask 2 lsbs */
    if ( cap_ptr == 0 ) break;
    c++;
 }
  return MT_ENORSC;
}


/*
 *  XHH_set_max_read_request_size
 *
 *  set Max Read Request Size to 5 => 4096 bytes
 */
static call_result_t XHH_set_max_read_request_size(XHH_hw_props_t *hw_props_p,
												   u_int8_t mrrs_val)
{
  call_result_t rc;
  u_int8_t cap_ptr;
  u_int16_t cap_val;
  const u_int8_t rrs_cap_id = 16; /* Max Read Request Size capability ID */


  if (mrrs_val > 5) {
	  mrrs_val= 5;
  }

  rc = mosal_find_capbility_ptr(hw_props_p->pci_dev, rrs_cap_id, &cap_ptr);
  if ( rc != MT_OK ) {
    MTL_DEBUG1(MT_FLFMT("%s: failed to find MRRS capability - %s"), __func__, mtl_strerror_sym(rc));
    return rc;
  }
  rc = MOSAL_PCI_read_config(hw_props_p->pci_dev, cap_ptr+8, 2, &cap_val);
  if ( rc != MT_OK ) {
    MTL_ERROR1(MT_FLFMT("%s: failed to read rbc - %s"), __func__, mtl_strerror_sym(rc));
    return rc;
  }
  MTL_TRACE1(MT_STDFMT("rrs required=%d, orig val=0x%02x, new val=%02x"), mrrs_val, cap_val, (cap_val&0x8fff) | (mrrs_val<<12));
  cap_val &= 0x8fff;
  cap_val |= (mrrs_val<<12);
  rc = MOSAL_PCI_write_config(hw_props_p->pci_dev, cap_ptr+8, 2, cap_val);
  if ( rc != MT_OK ) {
    MTL_ERROR1(MT_FLFMT("%s: failed to write rbc - %s"), __func__, mtl_strerror_sym(rc));
  }
  return rc;
}

/*
 *  XHH_set_rbc
 *
 *  set the default Read Byte to 3 ==> 4096 bytes
 */
static call_result_t XHH_set_rbc(XHH_hw_props_t *hw_props_p, u_int8_t rbc_val)
{
  call_result_t rc;
  u_int8_t cap_ptr, cap_val;
  const u_int8_t rbc_cap_id = 7; /* Read Byte Count capability ID */

  if ( !hw_props_p ) {
    MTL_ERROR1(MT_FLFMT("hw_props_p=null"));
    return MT_EINVAL;
  }

  if (rbc_val > 3) {
	  rbc_val= 3;
  }

  rc = mosal_find_capbility_ptr(hw_props_p->pci_dev, rbc_cap_id, &cap_ptr);
  if ( rc != MT_OK ) {
    MTL_DEBUG1(MT_FLFMT("%s: failed to find RBC capability - %s"), __func__, mtl_strerror_sym(rc));
    return rc;
  }

  rc = MOSAL_PCI_read_config(hw_props_p->pci_dev, cap_ptr+2, 1, &cap_val);
  if ( rc != MT_OK ) {
    MTL_ERROR1(MT_FLFMT("%s: failed to read rbc - %s"), __func__, mtl_strerror_sym(rc));
    return rc;
  }
  MTL_TRACE1(MT_STDFMT("rbc required=%d, orig val=0x%02x, new val=%02x"), rbc_val, cap_val, (cap_val&0x3) | (rbc_val<<2));
  cap_val &= 0xf3;
  cap_val |= (rbc_val<<2);
  rc = MOSAL_PCI_write_config(hw_props_p->pci_dev, cap_ptr+2, 1, cap_val);
  if ( rc != MT_OK ) {
    MTL_ERROR1(MT_FLFMT("%s: failed to write rbc - %s"), __func__, mtl_strerror_sym(rc));
  }
  return rc;
}



/*
 *  XHH_set_capabilities
 */
void XHH_set_capabilities(XHH_hw_props_t *hw_props_p,
						  XHH_module_flags_t *mod_flags)
{
	/* set the default Read Byte Count for Tavor */
	XHH_set_rbc(hw_props_p, mod_flags->pci_cap_read_byte_count);
	
	/* set max read request size in capabilty structure of PCI express */
	XHH_set_max_read_request_size(hw_props_p, mod_flags->pci_cap_max_read_request_size);
}


#define dw2capid(dw) ((dw)&0xff)
#define dw2cap_ptr(dw) ((u_int8_t)((dw)>>8) & 0xfc)

static int cap2dwsize(int capid)
{
  switch ( capid ) {
  case MSIX_CAPID:
    return MSIX_DWSZ;

  case MSI_CAPID:
    return MSI_DWSZ;

  case PCIX_CAPID:
    return PCIX_DWSZ;

  case VPD_CAPID:
    return VPD_CAPDWSZ;

  case PM_CAPID:
    return PM_CAPDWSZ;

  case PCIEXPRESS_CAPID:
    return PCIEXPRESS_DWSZ;
  }

  return -1;
}

/*
 *  XHH_get_capabilities
 *  get the PCI capabilities settings
 */
HH_ret_t XHH_get_pci_capabilities(MOSAL_pci_dev_t pci_dev, XHH_pci_config_space_container_t *config)
{
  call_result_t rc;
  u_int8_t cap_ptr;
  u_int32_t cap_val_dw;
  int dwsize, capid, i;
  HH_ret_t hhrc = HH_OK;


  config->num_cap = 0;

  /* read cap pointer */
  rc = MOSAL_PCI_read_config(pci_dev, 0x34, 1, &cap_ptr);
  if ( rc != MT_OK ) {
    MTL_ERROR1(MT_STDFMT("failed reading cap pointer - %s"), mtl_strerror(rc));
    return rc; 
  }

  while ( cap_ptr ) {

    if ( config->num_cap >= MAX_PCI_CAP ) {
      MTL_ERROR1(MT_STDFMT("no allocation for more than %d capabilities"), MAX_PCI_CAP);
      return HH_ERR;
    }

    rc = MOSAL_PCI_read_config(pci_dev, cap_ptr, 4, &cap_val_dw);
    if ( rc != MT_OK ) {
      MTL_ERROR1(MT_STDFMT("failed reading dword at address 0x%x - %s"), cap_ptr, mtl_strerror(rc));
      return HH_ERR;
    }

    capid = dw2capid(cap_val_dw);
    dwsize = cap2dwsize(capid);

    MTL_DEBUG1(MT_STDFMT("cap_ptr=0x%x, capid=0x%x, dw_size=%d"), cap_ptr, capid, dwsize);

    if ( dwsize == -1  ) {
      MTL_ERROR1(MT_STDFMT("unsupported capid: 0x%x"), capid);
      return HH_ERR;
    }


    config->clist[config->num_cap].cap_ptr = cap_ptr;
    config->clist[config->num_cap].cap_size_dw = dwsize;

    if ( dwsize > MAX_CAP_DW ) {
      MTL_ERROR1(MT_STDFMT("no allocation for more than %d dwords"), MAX_CAP_DW);
      return HH_ERR;
    }

    for ( i=0; i<dwsize; ++i ) {
      rc = MOSAL_PCI_read_config(pci_dev, cap_ptr +i*4, 4, &config->clist[config->num_cap].cap_data[i]);
      if ( rc != MT_OK ) {
        MTL_ERROR1(MT_STDFMT("failed reading dword at address 0x%x - %s"), cap_ptr+i*4, mtl_strerror(rc));
        return HH_ERR;
      }
      MTL_DEBUG1(MT_STDFMT("read dw=0x%08x"), config->clist[config->num_cap].cap_data[i]);
    }

    config->num_cap++;

    cap_ptr = dw2cap_ptr(cap_val_dw);
  }

  return hhrc;
}


/*
 *  XHH_set_pci_capabilities
 *  set the PCI capabilities settings
 */
HH_ret_t XHH_set_pci_capabilities(MOSAL_pci_dev_t pci_dev, XHH_pci_config_space_container_t *config)
{
  int i, j;
  u_int8_t offset;
  call_result_t rc;
    
  for ( i=0; i<config->num_cap; ++i ) {
    offset = config->clist[i].cap_ptr;
    for ( j=0; j<config->clist[i].cap_size_dw; ++j ) {
      rc = MOSAL_PCI_write_config(pci_dev, offset+j*4, 4, config->clist[i].cap_data[j]);
      if ( rc != MT_OK ) {
        MTL_ERROR1(MT_STDFMT("failed writing dw to addr=0x%x"), offset+j*4);
      }
      else {
        MTL_DEBUG1(MT_STDFMT("restoring: cap_ptr=%x data=0x%08x"), config->clist[i].cap_ptr, config->clist[i].cap_data[j]);
      }
    }
  }
  return HH_OK;
}

/*  
 *  XHH_read_pci_config -- reads all configuration registers for given device
 *  except for skipping regs 22 and 23
 */
HH_ret_t XHH_read_pci_config(MOSAL_pci_dev_t pci_dev, XHH_pci_config_space_container_t *config) 
{
    u_int8_t offset;
    HH_ret_t  rc = HH_OK;

    for (offset = 0; offset < 64; offset += 4) {
      if (offset == 22 || offset == 23) {
          continue;
      }
      rc = MOSAL_PCI_read_config(pci_dev, offset, 4, &config->seg1[offset>>2]);
      if (rc != MT_OK) {
          return HH_ERR;
      }
    }
  
    return XHH_get_pci_capabilities(pci_dev, config);
}


/*
 *  XHH_write_pci_config
 */
HH_ret_t XHH_write_pci_config(MOSAL_pci_dev_t pci_dev, XHH_pci_config_space_container_t *config) 
{
    u_int8_t offset = 0;
    HH_ret_t  rc = HH_OK;

    for (offset = 0; offset < 64; offset += 4) {
      if (offset == 22 || offset == 23) {
          continue;
      }
      rc = MOSAL_PCI_write_config(pci_dev, offset, 4, config->seg1[offset>>2]);
      if (rc != MT_OK) {
          return HH_ERR;
      }
    }

    XHH_set_pci_capabilities(pci_dev, config);

    return HH_OK;
}

