/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#if !defined(_ICMM__H)
#define _ICMM__H

#include <mtl_common.h>
#include <mosal.h>
#include <hh.h>
#include <xhh.h>


#define XHH_ICMM_INVALID_HANDLE  ((XHH_icmm_t)0)
#define XHH_ICMM_INVALID_SZ ((MT_size_t)MAKE_LONGLONG(-1))
#define XHH_ICMM_INVALID_VA ((XHH_icm_va_t)MAKE_LONGLONG(-1))

typedef u_int32_t XHH_icmm_map_flags_t;

/************************************************************************
 *  Function: XHH_icmm_create
 *
 *  Arguments:
 *    hob   The parent of this ICM 
 *    icm_byte_sz - ICM virtual space size in bytes 
 *    lam_byte_sz - Size of LAM in bytes.0 if not available. 
 *    lam_addr - Physical address of LAM.Ignored if lam_byte_sz==0 
 *    fw_base_addr - Physical address of firmware in LAM. Ignored if lam_byte_sz==0 
 *    fw_size - size of FW area.Ignored if lam_byte_sz==0 
 *    icmm_p - Created icmm object handle
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL_HCA_HANDLE
 *    HH_EAGAIN - Not enough resources for creating this object
 *
 *  Description:
 *    Create XHH_icmm object that manages icm_byte_sz bytes of virtual space.
 *    If a LAM is present, it allocates a LAM manager using the lam_byte_sz,
 *    lam_addr, fw_base_addr, and fw_size parameters. 
 */
extern HH_ret_t  XHH_icmm_create(
  /*IN */ XHH_hob_t  hob, 
  /*IN */ u_int64_t  icm_byte_sz, 
  /*IN */ u_int64_t  lam_byte_sz, 
  /*IN */ MT_phys_addr_t  lam_addr, 
  /*IN*/  MT_phys_addr_t  fw_base_addr,
  /*IN*/  MT_size_t       fw_size,
  /*OUT*/ XHH_icmm_t *icmm_p  
);


/************************************************************************
 *  Function: XHH_icmm_destroy
 *
 *  Arguments:
 *   icmm - The handle of the object to destroy
 *
 *  Returns:
 *   HH_OK HH_EINVAL - No such object
 *
 *  Description:
 *   Free given object resources.
 *
 *   - UNMAP_ICM and free ICM physical pages (if allocated in host memory). 
 *   - UNMAP_ICM_AUX and free auxiliary ICM memory. 
 *   - Free memory management structures. 
 *
 *   All memory management context will be lost after this function is completed,
 *   so it is recommended to call this function only when all the allocated memory 
 *   chunks are revoked of their owners (i.e.destroy this object after all other 
 *   dependent objects are destroyed).
 *
 */
extern HH_ret_t  XHH_icmm_destroy( /*IN */ XHH_icmm_t icmm);


/************************************************************************
 *  Function: XHH_icmm_alloc_sz_aligned
 *
 *  Arguments:
 *    icmm - XHH_icmm context
 *    num_o_chunks - The number of chunks to allocate (size of arrays below)
 *    chunks_sizes - An array of num_o_chunks sizes (log2) array.
 *                   One for each chunk.
 *    chunks_addrs - An array of num_o_chunks addresses allocated
 *                   for the chunks based on given size/alignment
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameters (NULL...)
 *    HH_EAGAIN - No resources available to match all chunks requirements
 *
 *  Description:
 *    HCA resources which are allocated during HCA initialization are all
 *    required to be size aligned (context tables etc.). By providing all
 *    those memory requirement at once this function allows the XHH_icmm
 *    to efficiently allocate ICM memory resources under the
 *    alignment constrains.
 *
 *    Use this function during the open-HCA process to allocate ICM virtual
 *    memory for all the HCA context tables. 
 *
 *    Note: this function may not allocate/map physical pages for the returned 
 *    address space until XHH_icmm_map is invoked.                                            
 *
 **************************************************************************************/
extern HH_ret_t  XHH_icmm_alloc_sz_aligned(
  /*IN  */ XHH_icmm_t      icmm,             
  /*IN  */ MT_size_t       num_o_chunks,
  /*IN  */ MT_size_t*      chunks_log2_sizes, 
  /*OUT */ XHH_icm_va_t*   chunks_addrs       
);


/************************************************************************
 *  Function: XHH_icmm_alloc
 *
 *  Arguments:
 *    icmm -        The object handle
 *    size -        Size in bytes of memory chunk required
 *    align_shift - Alignment shift of chunk (log2 of alignment value)
 *    icm_va_p -    The returned allocated ICM address
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalud object handle
 *    HH_EAGAIN - Not enough resources for required allocation
 *
 *  Description:
 *    Allocate a virtually contiguous memory chunk in ICM memory which
 *    follows requirement of size and alignment.
 */
extern HH_ret_t  XHH_icmm_alloc(
  /*IN  */ XHH_icmm_t      icmm, 
  /*IN  */ u_int64_t       size, 
  /*IN  */ u_int8_t        align_shift,
  /*OUT */ XHH_icm_va_t*   icm_va_p 
);


/************************************************************************
 *  Function: XHH_icmm_free
 *
 *  Arguments:
 *    icmm -   The object handle
 *    icm_va - Exact address of buffer as given in allocation
 *    size -   Original size (in bytes)of buffer as given in allocation
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Given handle in unknown or given address is not
 *                an address returned by XHH_icmm_alloc
 *                (or XHH_icmm_alloc_sz_aligned)
 *
 *  Description:
 *    Free a virtual memory chunk allocated by XHH_icmm_alloc or
 *    RHH_icmm_alloc_sz_aligned. 
 *
 *    Note that the given buffer address must be the exact address returned 
 *    during allocation (and the exact size). One cannot free a fragment of 
 *    an allocated chunk.
 */
extern HH_ret_t  XHH_icmm_free(
  /*IN  */ XHH_icmm_t      icmm,
  /*IN  */ XHH_icm_va_t    icm_va,
  /*IN  */ u_int64_t       size
);

/************************************************************************
 *  Function: XHH_icmm_map
 *
 *  Arguments:
 *    icmm -     The object handle
 *    icm_va -   Exact address of buffer as given in allocation
 *    size -     Original size (in bytes) of buffer as given in allocation
 *    flags -
 *    prot_ctx - (optional)protection context to map pages to. 
 *    os_map_p - OS virtual mapping of given icm_va within given prot_ctx. 
 *               Mapping to prot_ctx performed only if this parameter is not NULL.
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL -    Given handle is unknown 
 *    HH_EINVAL_VA - given address space is not within ICM space, or 
 *                   given address is not an address returned by XHH_icmm_alloc
 *                   (or XHH_icmm_alloc_sz_aligned)
 *
 *  Description:
 *    Map given ICM virtual address space to physical pages.
 *    If os_map_pp is not NULL, the pages mapped to the ICM are also mapped 
 *    to the virtual address space in the given protection context. 
 *    
 *    Currently, if os_map_pp is not NULL (i.e.,OS mapping required) we are can  
 *    map only a physically contiguous chunk, so allocated physical chunk must be  
 *    allocated as physically contiguous. This may increase the chance for failure.
 */
extern HH_ret_t XHH_icmm_map ( 
    /*IN  */  XHH_icmm_t           icmm, 
    /*IN  */  XHH_icm_va_t         icm_va, 
    /*IN  */  u_int64_t            size, 
    /*IN  */  XHH_icmm_map_flags_t map_flags, 
    /*IN  */  MOSAL_prot_ctx_t     prot_ctx, 
    /*OUT */  MT_virt_addr_t       *os_map_p
);

/************************************************************************
 *  Function: XHH_icmm_unmap
 *
 *  Arguments:
 *    icmm -     The object handle
 *    icm_va -   Exact address of buffer as given in XHH_icmm_map
 *    size -     Original size (in bytes) of buffer as given in XHH_icmm_map
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL -    Given handle is unknown 
 *    HH_EINVAL_VA - given address space is not within ICM space, or 
 *                   given address is not an address returned by XHH_icmm_alloc
 *                   (or XHH_icmm_alloc_sz_aligned)
 *
 *  Description:
 *    Unmap given ICM virtual address. All physical pages that were allocated 
 *    by the the mapping call are freed. If an o/s virtual address was returned by
 *    the mapping call, that virtual address will no longer be valid after this call
 *    (i.e., the o/s virtual memory returned by the mapping call will be unmapped).
 */
extern HH_ret_t XHH_icmm_unmap (
     /*IN */ XHH_icmm_t   icmm, 
     /*IN */ XHH_icm_va_t icm_va, 
     /*IN */ u_int64_t    size
);

/************************************************************************
 *  Function: XHH_icmm_query
 *
 *  Arguments:
 *    icmm -                The object handle
 *    align_shift -         Alignment requirements for returned  largest_chunk
 *    total_mem -           Total byte count of memory managed by this object
 *    free_mem -            Total byte count of free memory
 *    largest_chunk -       Largest chunk possible to allocate with given
 *                          align_shift  requirements
 *    largest_free_addr_p - Address of the above largest chunk
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid handle
 *
 *  Description:
 *    Query icmm object for allocation capabilities with given alignment
 *    contraints (use 0 if alignment is not needed). 
 */
extern HH_ret_t  XHH_icmm_query(
  /*IN  */ XHH_icmm_t     icmm,              
  /*IN  */ u_int8_t       align_shift, 
  /*OUT */ u_int64_t*     total_mem,          
  /*OUT */ u_int64_t*     free_mem, 
  /*OUT */ u_int64_t*     largest_chunk,
  /*OUT */ XHH_icm_va_t*  largest_free_addr_p
);

/************************************************************************
 *  Function: XHH_icmm_kernel_map_create
 *
 *  Arguments:
 *    icmm -     The object handle
 *    icm_blk_va -  Address of block in which given icm_map_va is located
 *    icm_map_va -  Specific entry to map.
 *    map_size -    Mapping size (may be cross blocks)
 *    kernel_map_p - Returned mapping object (to be used with XHH_icmm_kernel_map_write)
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL -    Given handle is unknown 
 *    HH_EINVAL_VA - given address space is not within ICM space, or 
 *                   given address is not an address returned by XHH_icmm_alloc
 *                   (or XHH_icmm_alloc_sz_aligned)
 *
 *  Description:
 *    Create an object that allows direct access from kernel (using XHH_icmm_kernel_map_write)
 *    to given ICM address range (even if crossing blocks).
 *    Note: It is the reposibility of the object creator to retain ICM mapping 
 *          until this object is destroyed. 
 *    
 */
extern HH_ret_t XHH_icmm_kernel_map_create(
     /*IN*/ XHH_icmm_t   icmm, 
     /*IN*/ XHH_icm_va_t icm_start_blk_va, 
     /*IN*/ XHH_icm_va_t icm_map_va, 
     /*IN*/ u_int64_t    map_size,
     /*OUT*/ XHH_icmm_kernel_map_t *kernel_map_p
);

/************************************************************************
 *  Function: XHH_icmm_kernel_map_destroy
 *
 *  Arguments:
 *    icmm -     The object handle
 *    kernel_map - Kernel mapping object
 *
 *  Returns: (void)
 *
 *  Description:
 *    Free resources of mapping object..
 */
extern void XHH_icmm_kernel_map_destroy(
     /*IN*/ XHH_icmm_t   icmm, 
     /*IN*/ XHH_icmm_kernel_map_t kernel_map
);

/************************************************************************
 *  Function: XHH_icmm_kernel_map_write_buf
 *
 *  Arguments:
 *    icmm -     The object handle
 *    kernel_map - Kernel mapping object
 *    data - The data buffer to write
 *    size - Data buffer size. May be less than or equal mapping size.
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL -    Given handle is unknown 
 *
 *  Description:
 *    Write given data buffer to ICM space mapped by this object (starting at map beginning).
 */
HH_ret_t XHH_icmm_kernel_map_write_buf(
     /*IN*/ XHH_icmm_kernel_map_t kernel_map,
     /*IN*/ void* data,
     /*IN*/ MT_size_t size
);

/************************************************************************
 *  Function: XHH_icmm_kernel_map_write_qword/dword/byte
 *
 *  Arguments:
 *    icmm -     The object handle
 *    kernel_map - Kernel mapping object
 *    offset - (byte) Offset in given mapping to write.
 *    data - The data to write
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL -    Given handle is unknown 
 *
 *  Description:
 *    Write given data at given offset of ICM space mapped by this object.
 */
HH_ret_t XHH_icmm_kernel_map_write_qword(
     /*IN*/ XHH_icmm_kernel_map_t kernel_map,
     /*IN*/ XHH_icm_va_t offset,
     /*IN*/ u_int64_t data
);
HH_ret_t XHH_icmm_kernel_map_write_dword(
     /*IN*/ XHH_icmm_kernel_map_t kernel_map,
     /*IN*/ XHH_icm_va_t offset,
     /*IN*/ u_int32_t data
);
HH_ret_t XHH_icmm_kernel_map_write_byte(
     /*IN*/ XHH_icmm_kernel_map_t kernel_map,
     /*IN*/ XHH_icm_va_t offset,
     /*IN*/ u_int8_t data
);

#endif /* _ICMM__H */
