/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/linkage.h>
#include <asm/uaccess.h>
#include <linux/pci.h>

#include <hh.h>
#include <xhh.h>
#include <hob.h>
#include <xhh_init.h>
#include <card_defs.h>
#include <driver_defs.h>
#include <linux_module.h>
#include <vapi_version.h>

#if LINUX_KERNEL_2_6
#include <linux/moduleparam.h>
#endif

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Infiniband hardware driver module" VAPI_VERSION_STRING);

#ifndef DMA_64BIT_MASK
#define DMA_64BIT_MASK  0xffffffffffffffffULL
#endif

#ifndef DMA_32BIT_MASK
#define DMA_32BIT_MASK  0x00000000ffffffffULL
#endif

static struct pci_device_id hca_pci_tbl[] = HCA_TBL_INITIALIZER;

/* hca_pci_tbl is defined in card_defs.ic */
MODULE_DEVICE_TABLE(pci, hca_pci_tbl) ;

static int device_probe(struct pci_dev *dev, const struct pci_device_id *id);
static void device_remove(struct pci_dev *dev);

static struct pci_driver hca_driver = {
  .name       = DRV_NAME,
  .id_table   = hca_pci_tbl,
  .probe      = device_probe,
  .remove     = __devexit_p(device_remove),
};

/*
 * Put all Tavors in device tables
 *
 */

#if LINUX_KERNEL_2_6
#define  DEV_NAME(pci_dev_p)  pci_name(pci_dev_p)
#else
#define  DEV_NAME(pci_dev_p)  (pci_dev_p)->name
#endif

#define PCI_ENUM_CONT \
      failed_init++; \
      continue

#define XHHOBP(hca_hndl)   ((XHH_hob_t)(hca_hndl->device))

/* if non-zero, indicates legacy sqp initialization.  May be modified by insmod parameter */
static int  xhh_legacy_sqp = 0;
static int  av_in_host_mem = 0;
static int  infinite_cmd_timeout = 0; /* when 1 we use inifinite timeouts on commands completion */
static int  fatal_delay_halt = 0; /* when 1, HALT_HCA on fatal is delayed to just before the reset */
static int  num_cmds_outs = 0; /* max number of outstanding commands */
static int  async_eq_size = 0; /* size of async event queue */
static int  cmdif_use_uar0 = 1; /* when 0, cmdif posts commands to HCR */
static int  ignore_subsystem_id = 0; /* when 1, we do not check the subsystem_vendor_id & subsystem_id */
static int  legacy_name = 0;   /* use "InfiniHost" as the device prefix*/
static int  max_kernel_avs = 0;	/* Max number of UD AVs in kernel */
static int  disable_msix = 0;   /* disable usage of MSIX interrupts by the driver */
static int  eneble_sqp_sched = 0; /* Eneble higher priority scheduling of SQPs in the HCA engines */
static int  pci_cap_max_read_request_size=5; /*  0  128 bytes max read request size
												 1  256 bytes max read request size
											     2  512 bytes max read request size
											     3  1024 bytes max read request size
											     4  2048 bytes max read request size
											     5  4096 bytes max read request size */

static int pci_cap_read_byte_count=3; /* 0    512 bytes
									     1   1024 bytes
									     2   2048 bytes
									 	 3   4096 bytes */

#if LINUX_KERNEL_2_6
	static int  use_sysfs = 0;
	module_param(xhh_legacy_sqp, int, 0444);
	MODULE_PARM_DESC(xhh_legacy_sqp, "use internal SMA");
	module_param(av_in_host_mem, int, 0444);
	MODULE_PARM_DESC(av_in_host_mem, "put UD AVs in host memory");
	module_param(infinite_cmd_timeout, int, 0444);
	MODULE_PARM_DESC(infinite_cmd_timeout, "wait for commands forever");
	module_param(fatal_delay_halt, int, 0444);
	MODULE_PARM_DESC(fatal_delay_halt, "do not reset the HCA after fatal");
	module_param(num_cmds_outs, int, 0444);
	MODULE_PARM_DESC(num_cmds_outs, "max number of outstanding commands supported");
	module_param(async_eq_size, int, 0444);
	MODULE_PARM_DESC(async_eq_size, "async events EQ size");
	module_param(cmdif_use_uar0, int, 0444);
	MODULE_PARM_DESC(cmdif_use_uar0, "use UAR0 for posting commands");
	module_param(ignore_subsystem_id, int, 0444);
	MODULE_PARM_DESC(ignore_subsystem_id, "");
	module_param(legacy_name, int, 0444);
	MODULE_PARM_DESC(legacy_name, "use InfiniHost as the device prefix");
	module_param(max_kernel_avs, int, 0444);
	MODULE_PARM_DESC(max_kernel_avs, "Max number of UD AVs in kernel");
	module_param(disable_msix, int, 0444);
	MODULE_PARM_DESC(disable_msix, "disable usage of MSIX interrupts by the driver");
	module_param(use_sysfs, int, 0444);
	MODULE_PARM_DESC(use_sysfs, "show attributes in sysfs");
	module_param(pci_cap_max_read_request_size, int, 0444);
	MODULE_PARM_DESC(pci_cap_max_read_request_size, "pci capability max read request size");
	module_param(pci_cap_read_byte_count, int, 0444);
	MODULE_PARM_DESC(pci_cap_read_byte_count, "pci capability read byte count");


	

/*    module_param(eneble_sqp_sched, int, 0444);
	MODULE_PARM_DESC(eneble_sqp_sched, "Enable higher priority scheduling of SQPs in the HCA engines");
*/
#else
	MODULE_PARM(xhh_legacy_sqp,"i");
	MODULE_PARM(av_in_host_mem,"i");
	MODULE_PARM(infinite_cmd_timeout, "i");
	MODULE_PARM(fatal_delay_halt, "i");
	MODULE_PARM(num_cmds_outs, "i");
	MODULE_PARM(async_eq_size, "i");
	MODULE_PARM(cmdif_use_uar0, "i");
	MODULE_PARM(ignore_subsystem_id, "i");
	MODULE_PARM(legacy_name,"i");
	MODULE_PARM(max_kernel_avs,"i");
	MODULE_PARM(disable_msix,"i");
	MODULE_PARM(pci_cap_max_read_request_size,"i");
	MODULE_PARM(pci_cap_read_byte_count,"i");
	
//	MODULE_PARM(eneble_sqp_sched,"i");
#endif


static XHH_module_flags_t  module_flags;



/*
 *  init_module_flags
 */
static void init_module_flags(void)
{
  int i;
  /*initialize module flags structure */
  memset(&(module_flags), 0, sizeof(XHH_module_flags_t));

  module_flags.legacy_sqp = (xhh_legacy_sqp == 0 ? FALSE : TRUE);
  module_flags.av_in_host_mem = (av_in_host_mem == 0 ? FALSE : TRUE);
  module_flags.inifinite_cmd_timeout = (infinite_cmd_timeout==1 ? TRUE : FALSE);
  module_flags.fatal_delay_halt = (fatal_delay_halt==1 ? TRUE : FALSE);
  module_flags.cmdif_post_uar0 = cmdif_use_uar0==1 ? TRUE : FALSE;
  module_flags.legacy_name = (legacy_name == 0 ? FALSE : TRUE);
  module_flags.max_kernel_avs = (max_kernel_avs == 0 ? 0xFFFFFFFF : (u_int32_t) max_kernel_avs );
  module_flags.eneble_sqp_sched = (eneble_sqp_sched == 0 ? FALSE : TRUE);
  module_flags.pci_cap_max_read_request_size= pci_cap_max_read_request_size;
  module_flags.pci_cap_read_byte_count= pci_cap_read_byte_count;

  if ( num_cmds_outs == 0 ) {
    module_flags.num_cmds_outs = 0xffffffff;
  }
  else {
    module_flags.num_cmds_outs = num_cmds_outs;
  }

  module_flags.async_eq_size = async_eq_size;

  MTL_DEBUG1(MT_STDFMT("ignore_subsystem_id=%d"), ignore_subsystem_id);
  if ( ignore_subsystem_id ) {
    for ( i=0; i<(sizeof(hca_pci_tbl)/sizeof(hca_pci_tbl[0])-1); ++i ) {
      hca_pci_tbl[i].subvendor = PCI_ANY_ID;
      hca_pci_tbl[i].subdevice = PCI_ANY_ID;
    }
  }
}


/*
 *  XHH_module_init
 */
static int XHH_module_init(void)
{
  HH_ret_t  ret;
  int rc;

#if LINUX_KERNEL_2_6
  if ( use_sysfs ) {
	  rc = vapi_sysfs_setup();
	  if ( rc ) {
		  MTL_ERROR1(MT_STDFMT("Couldn't create vapi device class"));
		  return -1;
	  }
  }
#endif

  /* static context initialization */
  ret = XHH_init();
  if ( ret != HH_OK ) {
    MTL_ERROR1(MT_STDFMT("XHH_init failed (%s)"), HH_strerror_sym(ret));
    rc = -1;
	goto undo_sysfs_setup;
  }

  /* initialize module flags based on module args */
  init_module_flags();

  /* register pci devices with the driver */
  if ( pci_register_driver(&hca_driver) < 0 ) {
    printk("driver_register failed\n");
    rc = -1;
	goto undo_sysfs_setup;
  }
  rc = 0;
  printk("<1> "XHH_NAME" kernel module initialized successfully\n");
  goto exit;

undo_sysfs_setup:
#if LINUX_KERNEL_2_6
  if ( use_sysfs ) {
	  vapi_sysfs_cleanup();
  }
#endif
exit:
  return rc;
}


/*
 *  XHH_module_exit
 */
static void XHH_module_exit(void)
{
  HH_ret_t        ret;

  pci_unregister_driver(&hca_driver);

  ret = XHH_cleanup();
  if ( ret != HH_OK ) {
    MTL_ERROR1(MT_STDFMT("cleanup_module:  XHH_cleanup failed (%s)"), HH_strerror_sym(ret));
  }


#if LINUX_KERNEL_2_6
  if ( use_sysfs ) {
	  vapi_sysfs_cleanup();
  }
#endif

  if ( ret == HH_OK ) {
    printk("<1> "XHH_NAME" kernel module removed successfully\n");
  }


  return;
}

/*
 *  device_probe
 */
static int __devinit device_probe(struct pci_dev *dev, const struct pci_device_id *id)
{
  static int found_dev_index = 0;
  XHH_hw_props_t *hw_props_p = NULL;
  u_int8_t   rev_id;
  const unsigned int cr_res_index=0;
  const unsigned int uar_res_index=2;
  const unsigned int ddr_res_index=4;
  HH_ret_t ret;
  int rc=-1;
  HH_hca_hndl_t hh_hndl;
  struct hca_driver_data_st *driver_data;
#if defined(CONFIG_PCI_MSI)
  unsigned int num_req_vectors;
#endif
  int i;
  int hide_ddr;
  struct msix_entry *vect_arr=NULL;

  hw_props_p = (XHH_hw_props_t *)MALLOC(sizeof(XHH_hw_props_t));
  if (hw_props_p == NULL) {
     MT_COV_MEM_FLTR_START
     return -1;
     MT_COV_MEM_FLTR_END
  }
  memset(hw_props_p, 0, sizeof(XHH_hw_props_t));
  
  hw_props_p->pci_dev = dev;
  hw_props_p->bus = dev->bus->number;
  hw_props_p->dev_func = dev->devfn;
  hw_props_p->device_id = id->device;
  hw_props_p->pci_vendor_id = id->vendor;

  /* enable the device prior to using it */
  rc = pci_enable_device(dev);
  if ( rc != 0 ) {
    MTL_ERROR1(MT_STDFMT("pci_enable_device failed: "MT_PCI_DEV_FMT), MT_PCI_DEV_ARGS(dev));
    rc = -1;
    goto ex_device_probe;
  }


#if defined(CONFIG_PCI_MSI)

  if ( !disable_msix ) {
	  num_req_vectors = XHH_get_num_req_vectors();
	  if ( num_req_vectors > MAX_INT_VECTORS ) {
		MTL_ERROR1(MT_STDFMT("Error: hob requests %d interrupts but hard coded limit is %d."),
				   num_req_vectors, MAX_INT_VECTORS);
		rc = -1;
		goto ex_undo_pci_enable;
	  }

	revert:
	  vect_arr = TNMALLOC(struct msix_entry, num_req_vectors);
	  if ( !vect_arr ) {
		MT_COV_MEM_FLTR_START
		MTL_ERROR1(MT_STDFMT("allocation failed for %d struct msix_entry objects"),
					num_req_vectors);
		rc = -1;
		goto ex_undo_pci_enable;
		MT_COV_MEM_FLTR_END
	  }

	  /* initialize the array with unique values */
	  for ( i=0; i<num_req_vectors; ++i ) {
		  vect_arr[i].entry = i;
	  }

	  rc = pci_enable_msix(dev, vect_arr, num_req_vectors);
	  if ( rc != 0 ) {
		/* failed enabling MSIX */
		MTL_ERROR1(MT_STDFMT("failed to allocate %d vectors, got %d, device="MT_PCI_DEV_FMT),
							 num_req_vectors, rc, MT_PCI_DEV_ARGS(dev));
		if ( rc > 0 ) {
		  /* if we are allowed to get less vectors than we requested this
			 may be 1 or 2 and we can use any of these numbers of vectors */
		  FREE(vect_arr);
		  vect_arr = NULL;
		  num_req_vectors = rc;
		  goto revert;
		}
		else {
		  /* this should never happen since we should be able to receive at least
			 one vector */
		  rc = -1;
		  goto ex_free_vect_arr;
		}
	  }
	  else {
		MTL_DEBUG1(MT_STDFMT("using MSIX for device="MT_PCI_DEV_FMT),
							 MT_PCI_DEV_ARGS(dev));
		for ( i=0; i<num_req_vectors; ++i ) {
			hw_props_p->interrupt_props.irq[i] = vect_arr[i].vector;
			MTL_DEBUG1(MT_STDFMT("using MSIX vector=%d"), vect_arr[i].vector);
		}
		hw_props_p->interrupt_props.int_mode = MT_INT_MODE_MSI;
		hw_props_p->interrupt_props.num_allocated_vectors = num_req_vectors;
	  }
  }
  else
#else
	disable_msix = 1;
#endif
  {
	  hw_props_p->interrupt_props.irq[0] = dev->irq;
	  hw_props_p->interrupt_props.int_mode = MT_INT_MODE_PIN;
	  hw_props_p->interrupt_props.num_allocated_vectors = 1;
  }

  driver_data = TMALLOC(struct hca_driver_data_st);
  if ( !driver_data ) {
    MT_COV_MEM_FLTR_START
    goto ex_undo_enable_msi;
    MT_COV_MEM_FLTR_END
  }
  else {
	memset(driver_data, 0, sizeof *driver_data);
    for ( i=0; i<3; ++i ) {
      driver_data->resource_requested[i] = 0;
    }
    driver_data->vect_arr = vect_arr;
	driver_data->dev = dev;
  }

  if ( !(pci_resource_flags(dev, cr_res_index) & IORESOURCE_MEM) || 
       pci_resource_len(dev, cr_res_index) != (1<<20) ) {
       MTL_ERROR1(MT_STDFMT("Bad cr-space "MT_PCI_DEV_FMT), MT_PCI_DEV_ARGS(dev));
              goto ex_undo_ddata;
  }
  if ( !(pci_resource_flags(dev, uar_res_index) & IORESOURCE_MEM) ) {
      MTL_ERROR1(MT_STDFMT("Bad UAR space "MT_PCI_DEV_FMT), MT_PCI_DEV_ARGS(dev));
      goto ex_undo_ddata;
  }
  if ( !(pci_resource_flags(dev, ddr_res_index) & IORESOURCE_MEM) ) {
  /* hide ddr mode */
    hide_ddr = 1;
    MTL_DEBUG1(MT_STDFMT("device operates without DDR: "MT_PCI_DEV_FMT), MT_PCI_DEV_ARGS(dev));
  }
  else {
    hide_ddr = 0;
  }

#if defined(CONFIG_PCI_MSI)
  /* We do this because the msix bus drivers requests this region
     in order to configure the address/data pairs for the entries
     and we don't want it to fail on this request */
  if (  hw_props_p->interrupt_props.int_mode == MT_INT_MODE_PIN ) {
#else
  if ( 1 ) {
#endif
    rc = pci_request_region(dev, cr_res_index, DRV_NAME);
    if ( rc != 0 ) {
      MTL_ERROR1(MT_STDFMT("Failed obtaining PCI resources[%d] "MT_PCI_DEV_FMT), cr_res_index, MT_PCI_DEV_ARGS(dev));
      goto ex_undo_ddata;
    }
    else {
      driver_data->resource_requested[0] = 1;
      MTL_DEBUG1(MT_FLFMT("requested region idx=%d, start=0x%lx, size=0x%lx"), cr_res_index,
               pci_resource_start(dev, cr_res_index), pci_resource_len(dev, cr_res_index));
    }
  }

  rc = pci_request_region(dev, uar_res_index, DRV_NAME);
  if ( rc != 0 )  {
    MTL_ERROR1(MT_STDFMT("Failed obtaining PCI resources[%d] "MT_PCI_DEV_FMT), uar_res_index, MT_PCI_DEV_ARGS(dev));
    goto ex_undo_request_regions;
  }
  else {
    driver_data->resource_requested[1] = 1;
    MTL_DEBUG1(MT_STDFMT("requested region idx=%d, start=0x%lx, size=0x%lx"), uar_res_index,
               pci_resource_start(dev, uar_res_index), pci_resource_len(dev, uar_res_index));
  }


  if ( !hide_ddr ) {
    rc = pci_request_region(dev, ddr_res_index, DRV_NAME);
    if ( rc != 0 ) {
      MTL_ERROR1(MT_STDFMT("Failed obtaining PCI resources[%d] "MT_PCI_DEV_FMT), ddr_res_index, MT_PCI_DEV_ARGS(dev));
      goto ex_undo_request_regions;
    }
    else {
      driver_data->resource_requested[2] = 1;
      MTL_DEBUG1(MT_STDFMT("requested region idx=%d, start=0x%lx, size=0x%lx"), ddr_res_index,
                 pci_resource_start(dev, ddr_res_index), pci_resource_len(dev, ddr_res_index));
    }
  }

  /* set the master bit of the device */
	pci_set_master(dev);

	rc = pci_set_dma_mask(dev, DMA_64BIT_MASK);
	if ( rc ) {
    MTL_ERROR1(MT_STDFMT("could not set 64-bit PCI DMA mask "MT_PCI_DEV_FMT), MT_PCI_DEV_ARGS(dev));
		rc = pci_set_dma_mask(dev, DMA_32BIT_MASK);
		if ( rc ) {
      MTL_ERROR1(MT_STDFMT("Can't set PCI DMA mask "MT_PCI_DEV_FMT), MT_PCI_DEV_ARGS(dev));
			goto ex_undo_request_regions;
		}
	}

#if LINUX_KERNEL_2_6
	rc = pci_set_consistent_dma_mask(dev, DMA_64BIT_MASK);
	if ( rc ) {
    MTL_ERROR1(MT_STDFMT("could not set 64-bit consistent PCI DMA mask "MT_PCI_DEV_FMT), MT_PCI_DEV_ARGS(dev));
		rc = pci_set_consistent_dma_mask(dev, DMA_32BIT_MASK);
		if ( rc ) {
      MTL_ERROR1(MT_STDFMT("could not set consistent PCI DMA mask "MT_PCI_DEV_FMT), MT_PCI_DEV_ARGS(dev));
      goto ex_undo_request_regions;
		}
	}
#endif

	hw_props_p->cr_base = pci_resource_start(dev, cr_res_index);
	hw_props_p->uar_base = pci_resource_start(dev, uar_res_index);
  if ( hide_ddr ) {
    hw_props_p->ddr_base = 0;
  }
  else {
    hw_props_p->ddr_base = pci_resource_start(dev, ddr_res_index);
  }


  if (pci_read_config_byte(dev, PCI_REVISION_ID, &rev_id) != 0) {
    MTL_ERROR1(MT_STDFMT("Failed reading HW revision id"));
    goto ex_undo_request_regions;
  }
  hw_props_p->hw_ver = rev_id;

  if ( pci_read_config_byte(dev, PCI_INTERRUPT_PIN, &(hw_props_p->interrupt_props.intr_pin)) != 0 ) {
    MTL_ERROR1(MT_STDFMT("Failed reading HW revision id"));
    goto ex_undo_request_regions;
  }

  MTL_DEBUG1(MT_STDFMT("Interrupt pin %d routed to IRQ %d"),
             hw_props_p->interrupt_props.intr_pin,hw_props_p->interrupt_props.irq[0]);

  ret = XHH_hob_create(hw_props_p, found_dev_index, &module_flags, &hh_hndl);
  if ( ret != HH_OK ) {
	MTL_ERROR1(MT_STDFMT("Failed creating "XHH_NAME" hob for device %d in index %d"), 
			   hw_props_p->device_id, found_dev_index);
    rc = -1;
    goto ex_undo_request_regions;
  }
  else {
    ret = HH_copy_device_desc(hh_hndl, driver_data->dev_desc);
    if ( ret != HH_OK ) {
	    MTL_ERROR1(MT_STDFMT("Failed getting device description for device %d in index %d"), 
				   hw_props_p->device_id, found_dev_index);
		goto ex_undo_hob_create;
    }
    pci_set_drvdata(dev, driver_data);
  }


  driver_data->found_index = found_dev_index;
#if LINUX_KERNEL_2_6
  if ( use_sysfs ) {
	  driver_data->class_dev.dev = &dev->dev;
	  rc = vapi_device_register_sysfs(driver_data);
	  if ( rc ) {
		  goto ex_undo_hob_create;
	  }
  }
#endif

  rc = 0;
  found_dev_index++;
  goto ex_device_probe;

ex_undo_hob_create:
  XHH_hob_destroy(hh_hndl);
ex_undo_request_regions:
  for ( i=0; i<3; ++i ) {
    if ( driver_data->resource_requested[i] ) {
      pci_release_region(dev, i*2);
    }
  }
ex_undo_ddata:
  FREE(driver_data);
ex_undo_enable_msi:
#if defined(CONFIG_PCI_MSI)
  if ( !disable_msix ) {
	  pci_disable_msix(dev);
	ex_free_vect_arr:
	  FREE(vect_arr);
  }
ex_undo_pci_enable:
#endif
  pci_disable_device(dev);
ex_device_probe:
  FREE(hw_props_p);
  return rc;
}


static void __devexit device_remove(struct pci_dev *dev)
{
  HH_ret_t ret;
  HH_hca_hndl_t hh_hndl;
  struct hca_driver_data_st *driver_data;
  int i;

  driver_data = pci_get_drvdata(dev);
  if ( !driver_data ) {
    /* This is fatal and should not happen */
    MTL_ERROR1(MT_STDFMT("******** no private data !!!!!!!!!!!"));
    return;
  }

  ret = HH_lookup_hca(driver_data->dev_desc, &hh_hndl);
  if ( ret != HH_OK ) {
    MTL_ERROR1(MT_STDFMT("cleanup_module: Failed HH_lookup_hca failed for device = %s - %s\n"),
               driver_data->dev_desc, HH_strerror_sym(ret));
    return;
  }
  ret = XHH_hob_destroy(hh_hndl);
  if ( ret != HH_OK ) {
    MTL_ERROR1(MT_STDFMT("cleanup_module: Failed XHH_hob_destroy for handle=%p - %s\n"),
               hh_hndl, HH_strerror_sym(ret));
  }

  for ( i=0; i<3; ++i ) {
    if ( driver_data->resource_requested[i] ) {
      pci_release_region(dev, i*2);
    }
  }

#if LINUX_KERNEL_2_6
  if ( use_sysfs ) {
	  vapi_device_unregister_sysfs(driver_data);
  }
#endif

  if ( driver_data->vect_arr )
    FREE(driver_data->vect_arr);
  FREE(driver_data);
#if defined(CONFIG_PCI_MSI)
	if ( !disable_msix ) {
		pci_disable_msix(dev);
	}
#endif
  pci_disable_device(dev);
}

module_init(XHH_module_init);
module_exit(XHH_module_exit);
