/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <linux/device.h>
#include <mtl_types.h>
#include <mosal.h>
#include <linux_module.h>

					
#if LINUX_KERNEL_2_6
static ssize_t show_max_lock(struct class_device *class_dev, char * buf)
{
	unsigned long cur;

	cur = MOSAL_get_max_lock_pages();

	return sprintf(buf, "%lu\n", cur);
}

static ssize_t store_max_lock(struct class_device *class_dev, const char * buf, size_t count)
{
	int rc;
	unsigned long new_max;

	rc = sscanf(buf, "%lu", &new_max);
	if ( rc == 1 ) {
		MOSAL_set_max_lock_pages(new_max);
		return count;
	}
	else {
		return 0;
	}
}

static ssize_t show_cur_lock(struct class_device *class_dev, char * buf)
{
	unsigned long cur;

	cur = MOSAL_get_cur_lock_pages();

	return sprintf(buf, "%lu\n", cur);
}

static CLASS_DEVICE_ATTR(max_lock, S_IRWXUGO, show_max_lock, store_max_lock);
static CLASS_DEVICE_ATTR(cur_lock, S_IRUGO, show_cur_lock, NULL);

//static struct class_device_attribute *max_lock_attribute = &class_device_attr_max_lock;
//static struct class_device_attribute *cur_lock_attribute = &class_device_attr_cur_lock;


static struct class_device_attribute *driver_attributes[] = {
	&class_device_attr_max_lock,
	&class_device_attr_cur_lock
};


static void mlxhh_release(struct class_device *cdev)
{
	struct hca_driver_data_st *drv_data = container_of(cdev, struct hca_driver_data_st, class_dev);

	if ( 0 )
		printk("%s called: "MT_PCI_DEV_FMT"\n", __func__, MT_PCI_DEV_ARGS(drv_data->dev));
}

static int mlxhh_hotplug(struct class_device *cdev, char **envp,
			     int num_envp, char *buf, int size)
{
	struct hca_driver_data_st *drv_data = container_of(cdev, struct hca_driver_data_st, class_dev);

	if ( 0 )
		printk("%s called: "MT_PCI_DEV_FMT" num_envp=%d, buf=%s, size=%d\n", __func__, MT_PCI_DEV_ARGS(drv_data->dev), num_envp, buf, size);
	return 0;
}


static struct class vapi_class = {
	.name    = "mellanox-vapi",
	.release = mlxhh_release,
	.hotplug = mlxhh_hotplug,
};


int vapi_device_register_sysfs(struct hca_driver_data_st *driver_data)
{
	struct class_device *class_dev = &driver_data->class_dev;
	int ret, i;
	char *cid_name=NULL;

	cid_name = kmalloc(BUS_ID_SIZE, GFP_KERNEL);
	if ( !cid_name )
		return -ENOMEM;

	//snprintf(cid_name, BUS_ID_SIZE, "memory_attrib_%d", driver_data->found_index);
	snprintf(cid_name, BUS_ID_SIZE, "%s", driver_data->dev_desc);

	class_dev->class = &vapi_class;
	class_dev->class_data = driver_data;
	strlcpy(class_dev->class_id, cid_name, BUS_ID_SIZE);


	ret = class_device_register(class_dev);
	if ( ret )
		goto exit;

	for ( i=0; i<(sizeof driver_attributes)/(sizeof (struct class_device_attribute *)); ++i ) {
		ret = class_device_create_file(class_dev, driver_attributes[i]);
		if ( ret )
			goto undo_register;
	}

	goto exit;

undo_register:
	class_device_unregister(class_dev);
exit:
	if ( cid_name ) kfree(cid_name);
	return ret;
}


void vapi_device_unregister_sysfs(struct hca_driver_data_st *driver_data)
{
	class_device_unregister(&driver_data->class_dev);
}


static ssize_t immediate_mod_remove_show(struct class *cls, char *buf)
{
	unsigned long sup = XHH_show_hotplug_support();

	if ( sup ) {
		return sprintf(buf, "%s\n", "no");
	}
	else {
		return sprintf(buf, "%s\n", "yes");
	}
}

static ssize_t immediate_mod_remove_store(struct class *cls, const char *buf, size_t count)
{
	if ( memcmp(buf, "no", 2) == 0 ) {
		XHH_set_hotplug_support(1);
	}
	else if ( memcmp(buf, "yes", 3) == 0 ) {
		XHH_set_hotplug_support(0);
	}

	return count;
}


static CLASS_ATTR(immediate_mod_remove, S_IRWXUGO, immediate_mod_remove_show, immediate_mod_remove_store);


static struct class_attribute *vapi_attributes[] = {
	&class_attr_immediate_mod_remove,
};

int vapi_sysfs_setup(void)
{
	int rc;
	int i;

	rc = class_register(&vapi_class);
	if ( rc )
		return rc;

	for ( i=0; i<(sizeof(vapi_attributes)/sizeof(struct class_attribute *)); ++i ) {
		rc = class_create_file(&vapi_class, vapi_attributes[0]);
		if ( rc )
			goto undo_class_register;
	}

	goto exit;

undo_class_register:
	class_unregister(&vapi_class);
exit:
	return rc;
}

void vapi_sysfs_cleanup(void)
{
	class_remove_file(&vapi_class, &class_attr_immediate_mod_remove);
	class_unregister(&vapi_class);
}
#endif
