/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <lamm.h>
#include <mtl_common.h>
#include <log2.h>
#include <extbuddy.h>
#include <mosal.h>
#include <vip_hashp.h>
#include <mem_utils.h>
#include <icmm.h>
#include <hob.h>

#define ICMM_MAP_TO_LAM         1

#define ELSE_ACQ_ERROR(f) else { MTL_ERROR1("%s MOSAL_mutex_acq failed\n", f); }
#define logIfErr(f) \
  if (rc != HH_OK) { MTL_ERROR1("%s: rc=%s\n", f, HH_strerror_sym(rc)); }


typedef struct XHH_icmm_st
{
  XHH_hob_t hob;
  XHH_lamm_t  lamm;
  XHH_cmd_t   cmd;
  u_int64_t icm_byte_sz;
  u_int64_t lam_byte_sz; 
  u_int64_t lam_addr;
  Extbuddy_hndl  xb;
  MOSAL_mutex_t  mtx; /* protect xb */
  VIP_hashp_p_t  icmm_map_tbl;
} ICMM_t;


typedef struct XHH_icm_map_entry_st {
    u_int32_t           icm_va_pageaddr;   /* ICM va right-shifted by ICM page size (4K) */
    u_int32_t           num_icm_pages;     /* number of ICM pages mapped.  (ICM page size = 4K) */
    u_int16_t           num_chunks;
    u_int8_t            flags; /* lam_memory, va_mapped... */
    MOSAL_phys_ctg_t    *chunks_p;  /* if mapped, object also contains va and prot ctx  */
    MT_phys_addr_t      lam_addr;
    MT_virt_addr_t      lam_os_addr;
    MOSAL_iomem_t       iomem_obj;
    MOSAL_prot_ctx_t    prot_ctx;
} XHH_icm_map_entry_t;  

typedef struct kernel_map_entry_st {
	MT_virt_addr_t map_va;
	MT_size_t map_size;
	enum {KMAP_HOST, KMAP_LAM} mapped_object_type;
	union {
	    MOSAL_phys_ctg_t	hostmem;
	    MOSAL_iomem_t       lamm_mem;
	};
	struct kernel_map_entry_st *next;
} XHH_icmm_kernel_map_entry_t;

struct XHH_icmm_kernel_map_st {
	unsigned long num_of_chunks;
	XHH_icmm_kernel_map_entry_t *chunks; /* List of mapped chunks */
	XHH_icm_va_t start_blk_va;
	XHH_icm_va_t map_start_va;
	u_int64_t map_size;
};

static HH_ret_t icmm_map_lam ( /*IN*/ XHH_icmm_t icmm, 
                        /*IN*/ u_int64_t icm_va, 
                        /*IN*/ u_int64_t size, 
                        /*IN*/ MOSAL_prot_ctx_t prot_ctx, 
                        /*IN*/ u_int32_t set_char_flag,
                        /*IN*/ u_int8_t  set_char,
                        /*OUT*/ MT_virt_addr_t *os_map_p);

static HH_ret_t icmm_map_hostmem ( /*IN*/ XHH_icmm_t icmm, 
                        /*IN*/ u_int64_t icm_va, 
                        /*IN*/ u_int64_t size, 
                        /*IN*/ MOSAL_prot_ctx_t prot_ctx, 
                        /*IN*/ u_int32_t set_char_flag,
                        /*IN*/ u_int8_t  set_char,
                        /*OUT*/ MT_virt_addr_t *os_map_p);

static HH_ret_t XHH_icmm_free_map_entry(XHH_icmm_t icmm, 
                        XHH_icm_map_entry_t* map_p); 

static HH_ret_t icmm_map_internal ( /*IN*/ XHH_icmm_t icmm, 
                        /*IN*/ u_int64_t icm_va, 
                        /*IN*/ u_int64_t size, 
                        /*IN*/ XHH_icmm_map_flags_t map_flags, 
                        /*IN*/ MOSAL_prot_ctx_t prot_ctx, 
                        /*IN*/ u_int32_t set_char_flag,
                        /*IN*/ u_int8_t  set_char,
                        /*OUT*/ MT_virt_addr_t *os_map_p);

static HH_ret_t icmm_memset_internal (
     /*IN */ XHH_icmm_t   icmm, 
     /*IN */ XHH_icm_va_t icm_va,
     /*IN */ u_int8_t     val,
     /*IN */ u_int32_t    size,
     XHH_icm_map_entry_t* map_p);

/************************************************************************/
/************************************************************************/
/*                         private functions                            */

/************************************************************************/

static void 
XHH_icmm_remove_map_entry(VIP_hash_key_t key,
                          VIP_hashp_value_t hash_val, 
                          void * priv_data)
{
    HH_ret_t rc;
    XHH_icm_map_entry_t* map_p = (XHH_icm_map_entry_t*)hash_val;
    XHH_icmm_t icmm = (XHH_icmm_t)priv_data;
    rc= XHH_icmm_free_map_entry(icmm,map_p);
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("%s: failed destroying map entry (%d:%s"),
		   __func__, rc, HH_strerror_sym(rc));
    } else {
	FREE(map_p);
    }
    
    return;
}

/************************************************************************/
static void  lookup_sort(
  MT_size_t         n,
  const MT_size_t*  sizes,
  MT_size_t*        lut
)
{
  unsigned int   i, i1, j;
  for (i = 0;  i != n;   lut[i] = i, ++i); /* identity init */
  for (i = 0, i1 = 1; i1 != n;  i = i1++)  /* small n, so Bubble sort O(n^2) */
  {
    MT_size_t  luiMax = i;
    MT_size_t  iMax   = sizes[ lut[i] ];
    for (j = i1;  j != n;  ++j)
    {
      MT_size_t v = sizes[ lut[j] ];
      if (iMax < v)
      {
        iMax = v;
        luiMax = j;
      }
    }
    j = (unsigned int)lut[i];  lut[i] = lut[luiMax];  lut[luiMax] = j; /* swap */
  }
} /* lookup_sort */


/************************************************************************/
/************************************************************************/
/*                         interface functions                          */


/************************************************************************/
HH_ret_t  XHH_icmm_create(
  /*IN*/       XHH_hob_t         hob, 
  /*IN*/ const XHH_icmm_init_t*  init_attr_p,
  /*OUT*/      XHH_icmm_t        *icmm_p 
)
{
  HH_ret_t       rc = HH_OK;
  Extbuddy_hndl  xb = NULL;
  XHH_icmm_t     icm = NULL;
  VIP_common_ret_t  vip_rc = VIP_OK;

  /* allocate an icmm object */
  icm = (XHH_icmm_t) MALLOC(sizeof(struct XHH_icmm_st));
  if (icm == NULL) {
      MTL_ERROR1(MT_FLFMT("%s: malloc failure"),__func__);
      return HH_EAGAIN;
  }
  memset(icm, 0, sizeof(struct XHH_icmm_st));

  XHH_hob_get_lamm(hob,&(icm->lamm));
  XHH_hob_get_cmd_if (hob,&(icm->cmd));

  MTL_TRACE1(MT_FLFMT("%s: icm_byte_sz="U64_FMT), __func__,
             init_attr_p->icm_byte_sz);
  
  /* create a buddy system in units of pages (not bytes).  Using pages as the unit of allocation allows
   * us to support 2^43 = 8 terabytes of virtual memory space.  Current requirement is 0.25 terabytes */
  xb = extbuddy_create((u_int32_t) (init_attr_p->icm_byte_sz >> MOSAL_SYS_PAGE_SHIFT), 0);
  
  if (xb == NULL)
  {
    MTL_ERROR1(MT_FLFMT("%s: could not create extbuddy for lam manager (%d:%s)"),
                 __func__, rc, HH_strerror_sym(rc));
    rc = HH_EAGAIN;
    goto extbuddy_fail;
  }
  else
  {
    icm->xb       = xb;
    MOSAL_mutex_init(&icm->mtx);
    *icmm_p = icm;
  }

  /* create hash table to hold icm mapping information */
  vip_rc = VIP_hashp_create(0, &(icm->icmm_map_tbl));
  if (vip_rc != VIP_OK) {
      MTL_ERROR1(MT_FLFMT("%s: could not create vip hash for lam manager (%d)"),
                   __func__, vip_rc );
      rc = HH_EAGAIN;
      goto hash_fail;
  }

  MTL_TRACE1(MT_FLFMT("%s: icmm_p=%p"), __func__, *icmm_p);
  return HH_OK;

hash_fail:

  if (xb != NULL) {
      extbuddy_destroy(xb);
  }

extbuddy_fail:
  FREE(icm);
  MT_RETURN(rc);
} /* XHH_lamm_create */


/************************************************************************/
HH_ret_t  XHH_icmm_destroy(XHH_icmm_t icmm)
{
  HH_ret_t  rc = HH_OK;
  VIP_common_ret_t vcr= VIP_OK;

  MTL_TRACE1("{XHH_icmm_destroy: icmm=%p\n", icmm);
  vcr = VIP_hashp_destroy(icmm->icmm_map_tbl,&XHH_icmm_remove_map_entry,icmm);
  if (vcr != VIP_OK) {
    MTL_ERROR1(MT_FLFMT("failed destroying map hash table \n"));
  }
  extbuddy_destroy(icmm->xb);
  MOSAL_mutex_free(&icmm->mtx);
  FREE(icmm);
  MTL_TRACE1("}XHH_icmm_destroy\n");
  logIfErr("XHH_icmm_destroy")
  return rc;
} /* XHH_lamm_destroy */



/************************************************************************/
/* Note: For chunks in the array of size XHH_DDRMM_INVALID_SZ, no allocation is made */
HH_ret_t  XHH_icmm_alloc_sz_aligned(
  XHH_icmm_t    icmm,              /* IN  */
  MT_size_t     num_o_chunks,      /* IN  */
  MT_size_t*    chunks_log2_sizes, /* IN  */
  XHH_icm_va_t* chunks_addrs       /* OUT */
)
{
  HH_ret_t    rc = HH_OK;
  MT_size_t*  slut; /* Sorted Look-Up Table - (:politely incorrect:) */
  MT_size_t i, si;
  u_int8_t   log2sz;

  MTL_DEBUG1("{XHH_icmm_alloc_sz_aligned: icmm=%p, n="SIZE_T_FMT"\n", 
             icmm, num_o_chunks);
  slut  = TNMALLOC(MT_size_t, num_o_chunks); /* small, so not VMALLOC */
  if (!slut){
      rc= HH_EAGAIN;
      goto Exit;
  }
  
  lookup_sort(num_o_chunks, chunks_log2_sizes, slut);
  for (i = 0;  (i != num_o_chunks) && (rc == HH_OK);  ++i)
  {
      si = slut[i];
      log2sz = (u_int8_t)chunks_log2_sizes[si];
      MTL_DEBUG1("%s: i:"SIZE_T_FMT"   si:"SIZE_T_FMT"  log2 sz:%d \n",__func__,i,si,log2sz);
      if (log2sz != (u_int8_t)XHH_ICMM_INVALID_SZ)  {
        rc = XHH_icmm_alloc(icmm, MAKE_ULONGLONG(1) << log2sz, log2sz, &chunks_addrs[si]);
        MTL_DEBUG1("icm adrs: "U64_FMT"\n",chunks_addrs[si]);
      } else {
        chunks_addrs[si]= XHH_ICMM_INVALID_VA;
        /* No allocation if given log size is zero (workaround struct design in ddr_alloc_size_vec)*/
      }

  }
  
  if (rc != HH_OK)
    { /* Backwards. Note that we avoid (MT_size_t)-1 > 0 */
      MTL_ERROR1(MT_FLFMT("%s:failed icmm alloc \n"),__func__);
      while (i-- > 0)
      { /* Now i >= 0 */
         MT_size_t  si = slut[i];
         u_int8_t   log2sz = (u_int8_t)chunks_log2_sizes[si];
         if (log2sz != (u_int8_t)XHH_ICMM_INVALID_SZ) 
           XHH_icmm_free(icmm, chunks_addrs[si], MAKE_ULONGLONG(1) << log2sz);
      }
  }
    
  FREE(slut);
Exit:
  MTL_TRACE1("}XHH_icmm_alloc_sz_aligned\n");
  logIfErr("XHH_icmm_alloc_sz_aligned")
  return rc;
} /* XHH_icmm_alloc_sz_aligned */

/*********************************************************************************/

HH_ret_t  XHH_icmm_free_sz_aligned(
  XHH_icmm_t    icmm,              /* IN  */
  MT_size_t     num_o_chunks,      /* IN  */
  MT_size_t*    chunks_log2_sizes, /* IN  */
  XHH_icm_va_t* chunks_addrs       /* IN */
)
{
  HH_ret_t    rc = HH_EAGAIN;

  MTL_TRACE1("{XHH_icmm_alloc_sz_aligned: icmm=%p, n="SIZE_T_FMT"\n", 
             icmm, num_o_chunks);
  {
    MT_size_t  i;
    rc = HH_OK;
    for (i = 0;  (i != num_o_chunks) && (rc == HH_OK);  ++i)
    {
      u_int8_t   log2sz = (u_int8_t)chunks_log2_sizes[i];
      if (log2sz != (u_int8_t)XHH_ICMM_INVALID_SZ)  {
        rc = XHH_icmm_free(icmm, chunks_addrs[i], MAKE_ULONGLONG(1) << log2sz);
      }
    }
  }
  MTL_TRACE1("}XHH_icmm_free_sz_aligned\n");
  logIfErr("XHH_icmm_free_sz_aligned")
  return rc;
} /* XHH_icmm_free_sz_aligned */



/************************************************************************/
HH_ret_t  XHH_icmm_alloc(
  XHH_icmm_t    icmm,         /* IN  */
  u_int64_t     size,         /* IN  */
  u_int8_t      align_shift,  /* IN  */
  XHH_icm_va_t* icm_va_p      /* OUT */
)
{
  HH_ret_t     rc = HH_EINVAL;
  u_int32_t    log2size_pages;
  u_int32_t    align_shift_pages;
  /* internally (extbuddy) we need power-2 size */
  MT_size_t    log2sz = ceil_log2(size);

  MTL_TRACE1("{XHH_icmm_alloc: icmm=%p, sz="U64_FMT", lg2=%d, shift=%d\n", 
             icmm, size, (u_int32_t)log2sz, align_shift);
  
  /* translate size and alignment in bytes to size and alignment in pages */
  log2size_pages = (u_int32_t)((log2sz > MT_ARBEL_PAGE_SHIFT) ? (log2sz - MT_ARBEL_PAGE_SHIFT) : 0 );
  align_shift_pages = (align_shift > MT_ARBEL_PAGE_SHIFT) ? (align_shift - MT_ARBEL_PAGE_SHIFT) : 0 ;

  if (log2size_pages >= align_shift_pages)
  {
    u_int32_t  p = EXTBUDDY_NULL;
    MOSAL_mutex_acq_ui(&icmm->mtx);
    p = extbuddy_alloc(icmm->xb, (u_int8_t)log2size_pages);
    MOSAL_mutex_rel(&icmm->mtx);
    MTL_DEBUG7(MT_FLFMT("log2size_pages=%u, p=0x%x"), log2size_pages, p);
    rc = HH_EAGAIN;
    if (p != EXTBUDDY_NULL)
    {
      /* translate address in pages to address in bytes.  Assume that VA space is 0-based. */
      *icm_va_p = ( ((XHH_icm_va_t)p) << MT_ARBEL_PAGE_SHIFT);
       rc       = HH_OK;
    }
  } else {
      MTL_ERROR1(MT_FLFMT("%s: log num pages requested (%u) is less than requested alignment  (%u)"),
                 __func__, log2size_pages, align_shift_pages);
  }
  MTL_TRACE1("}XHH_icmm_alloc: buf="U64_FMT"\n", (u_int64_t)*icm_va_p);
  if ( rc != HH_OK ) {
     if (rc == HH_EAGAIN) {
         MTL_DEBUG1("%s: rc=%s\n",__func__, HH_strerror_sym(rc));
     } else {
         MTL_ERROR1("%s: rc=%s\n",__func__, HH_strerror_sym(rc));
     }
  }
  return rc;
} /* XHH_icmm_alloc */


/************************************************************************/
HH_ret_t  XHH_icmm_free(
  XHH_icmm_t      icmm,    /* IN  */
  XHH_icm_va_t    icm_va,  /* IN */
  u_int64_t       size     /* IN  */
)
{
  HH_ret_t     rc = HH_OK;
  u_int32_t    log2size_pages;
  u_int32_t    va_page = (u_int32_t) (icm_va << MT_ARBEL_PAGE_SHIFT);
  /* internally (extbuddy) we need power-2 size */
  MT_size_t    log2sz = ceil_log2(size);

  MTL_TRACE1("{XHH_icmm_free: icmm=%p, icm_va = "U64_FMT "sz="U64_FMT", lg2=%d\n", 
               icmm, icm_va, size, (u_int32_t)log2sz);

  /* translate size in bytes to size in pages */
  log2size_pages = (u_int32_t)((log2sz > MT_ARBEL_PAGE_SHIFT) ? (log2sz - MT_ARBEL_PAGE_SHIFT) : 1 );

  MOSAL_mutex_acq_ui(&icmm->mtx);
  extbuddy_free(icmm->xb, va_page, (u_int8_t)log2size_pages);
  MOSAL_mutex_rel(&icmm->mtx);
  MTL_TRACE1("}XHH_icmm_free\n");
  logIfErr("XHH_icmm_free")
  return rc;
} /* XHH_icmm_free */


/************************************************************************/
HH_ret_t  XHH_icmm_query(
  XHH_icmm_t     icmm,               /* IN  */
  u_int8_t       align_shift,        /* IN  */
  u_int64_t*     total_mem,          /* OUT */
  u_int64_t*     free_mem,           /* OUT */
  u_int64_t*     largest_chunk,      /* OUT */
  XHH_icm_va_t*  largest_free_addr_p /* OUT */
)
{
  HH_ret_t   rc = HH_OK;
  int        log2_sz_page;     /* size in page units */
  u_int32_t  align_shift_page; /* align shift in page units */

  MTL_TRACE1("{XHH_icmm_query: icmm=%p, shift=%d\n", icmm, align_shift);
  /* Note:  all icm quantities in extbuddy are in page units.  In all cases (i.e., byte counts and addresses),
   * need to translate extbuddy units (pages) to byte units.
   */
  *total_mem     = icmm->icm_byte_sz;
  *free_mem      = ((u_int64_t)(extbuddy_total_available(icmm->xb))) << MT_ARBEL_PAGE_SHIFT;
  log2_sz_page   = extbuddy_log2_max_available(icmm->xb);
  *largest_chunk = 0;
  *largest_free_addr_p = 0; /* like null */
  align_shift_page = (align_shift > MT_ARBEL_PAGE_SHIFT) ? (align_shift - MT_ARBEL_PAGE_SHIFT) : 1;
  if ((u_int32_t)log2_sz_page >= align_shift_page)
  {
    u_int32_t  p;
    *largest_chunk = MAKE_ULONGLONG(1) << (log2_sz_page + MT_ARBEL_PAGE_SHIFT);
    extbuddy_query_chunks(icmm->xb, (u_int8_t)log2_sz_page, 1, &p);
    *largest_free_addr_p = (XHH_icm_va_t)( ((XHH_icm_va_t)p) << MT_ARBEL_PAGE_SHIFT);
  }
  MTL_TRACE1("}XHH_icmm_query: total="U64_FMT", free="U64_FMT", lc="U64_FMT", p="U64_FMT"\n",
             *total_mem, *free_mem,         
             *largest_chunk, (VAPI_phy_addr_t)*largest_free_addr_p);
  logIfErr("XHH_icmm_query")
  return rc;
} /* XHH_icmm_query */

static HH_ret_t icmm_map_lam ( /*IN*/ XHH_icmm_t icmm, 
                        /*IN*/ u_int64_t icm_va, 
                        /*IN*/ u_int64_t size, 
                        /*IN*/ MOSAL_prot_ctx_t prot_ctx, 
                        /*IN*/ u_int32_t set_char_flag,
                        /*IN*/ u_int8_t  set_char,
                        /*OUT*/ MT_virt_addr_t *os_map_p)
{
    /*TBD implementation */
    HH_ret_t  rc = HH_OK;
    call_result_t  cres;
    XHH_icm_map_entry_t *new_map_p = NULL;
    VIP_common_ret_t    vip_rc = VIP_OK;
    XHH_cmd_pa_vec_t  *pa_vec_p;
    u_int32_t          pa_vec_size;
    XHH_cmd_status_t   cmd_ret;
    static const MOSAL_mem_flags_t
      mem_flags = MOSAL_MEM_FLAGS_NO_CACHE |
                  MOSAL_MEM_FLAGS_PERM_READ |
                  MOSAL_MEM_FLAGS_PERM_WRITE;


    new_map_p = TMALLOC(XHH_icm_map_entry_t);
    if (new_map_p == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: could not allocate new mapping object"),__func__);
        return HH_EAGAIN;
    }
    memset(new_map_p,0,sizeof(XHH_icm_map_entry_t));
    new_map_p->prot_ctx = prot_ctx;
    new_map_p->flags = ICMM_MAP_TO_LAM;

    rc = XHH_lamm_alloc(icmm->lamm, (MT_size_t)size,MT_ARBEL_PAGE_SHIFT, &(new_map_p->lam_addr));
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("%s: XHH_lamm_alloc failed (%d:%s)"),
                   __func__, rc, HH_strerror_sym(rc));
        goto lamm_alloc_fail;
    }

    /* do virtual addr map if requested.  If alloc call succeeded, we know that only
     * one chunk was returned
     */
    new_map_p->num_chunks = 1;
    if (os_map_p != NULL) {
        cres = MOSAL_iomem_map(prot_ctx,mem_flags, new_map_p->lam_addr, (MT_size_t)size, 
                              &(new_map_p->iomem_obj),os_map_p);
                                                     
        if (cres != MT_OK) {
            MTL_ERROR1(MT_FLFMT("%s: MOSAL_iomem_map failed. physical address =" PHYS_ADDR_FMT " (%d:%s)"), 
                        __func__,new_map_p->lam_addr, cres, mtl_strerror_sym(cres));
            new_map_p->lam_addr = 0;
            rc = HH_EAGAIN;
            goto mem_map_fail;
        }
    }

    /* initialize the lamm memory if requested */
    if (set_char_flag == TRUE) {
        rc = icmm_memset_internal(icmm,icm_va, set_char, (u_int32_t)size, new_map_p);
        if (rc != HH_OK) {
            MTL_ERROR1(MT_FLFMT("%s: icmm_memset_internal failed (%d:%s)"),
                       __func__, rc, HH_strerror_sym(rc));
            goto mem_set_fail;
        }
    }
    
    new_map_p->num_icm_pages = (u_int32_t)(size >> MT_ARBEL_PAGE_SHIFT);
    new_map_p->icm_va_pageaddr = (u_int32_t)((u_int64_t)icm_va >> MT_ARBEL_PAGE_SHIFT);

    rc = XHH_build_lamm_pa_vec(new_map_p->lam_addr, size, &pa_vec_size, &pa_vec_p);
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("%s: build_lamm_pa_vec failed (%d:%s)"),
                   __func__, rc, HH_strerror_sym(rc));
        goto pa_vec_fail;
    }

    cmd_ret = XHH_cmd_MAP_ICM(icmm->cmd,icm_va,pa_vec_size, pa_vec_p);
    if (cmd_ret != XHH_CMD_STAT_OK) {
        MTL_ERROR1(MT_FLFMT("%s: XHH_cmd_MAP_ICM failed (%d)"),__func__, cmd_ret);
	rc = (cmd_ret == XHH_CMD_STAT_EFATAL) ? HH_EFATAL : HH_ERR;
        goto map_command_err;
    }

    /* now, save the mem object in the hash table */
    vip_rc = VIP_hashp_insert(icmm->icmm_map_tbl, (u_int32_t) (icm_va >>MT_ARBEL_PAGE_SHIFT),
                              new_map_p);
    if (vip_rc != VIP_OK) {
        MTL_ERROR1(MT_FLFMT("%s: VIP_hashp_insert failed (%d:%s)"),
                   __func__, vip_rc,VIP_common_strerror_sym(vip_rc));
        rc = HH_ERR;
        goto hash_insert_fail;
    }

    if (pa_vec_p != NULL) {
        FREE(pa_vec_p);
    }
    return HH_OK;


hash_insert_fail:
    cmd_ret = XHH_cmd_UNMAP_ICM(icmm->cmd,icm_va, new_map_p->num_icm_pages);
    if (cmd_ret != XHH_CMD_STAT_OK) {
        MTL_ERROR1(MT_FLFMT("%s: XHH_cmd_UNMAP_ICM failed (%d). icm_va="U64_FMT", num_pages=0x%x"),
                   __func__, cmd_ret, icm_va, new_map_p->num_icm_pages);
    }

map_command_err:
    if (pa_vec_p != NULL) {
        FREE(pa_vec_p);
    }

pa_vec_fail:
mem_set_fail:
    if (os_map_p != NULL) {
        MOSAL_iomem_unmap(new_map_p->iomem_obj, prot_ctx);
    }

mem_map_fail:
        XHH_lamm_free(icmm->lamm, new_map_p->lam_addr, (MT_size_t)size);
    
lamm_alloc_fail:
    if (new_map_p != NULL) {
        FREE(new_map_p);
    }
    return rc;
}
static HH_ret_t icmm_map_hostmem ( /*IN*/ XHH_icmm_t icmm, 
                        /*IN*/ u_int64_t icm_va, 
                        /*IN*/ u_int64_t size, 
                        /*IN*/ MOSAL_prot_ctx_t prot_ctx, 
                        /*IN*/ u_int32_t set_char_flag,
                        /*IN*/ u_int8_t  set_char,
                        /*OUT*/ MT_virt_addr_t *os_map_p)
{
    HH_ret_t  rc = HH_OK;
    XHH_icm_map_entry_t *new_map_p = NULL;
    call_result_t  cr = MT_OK;
    VIP_common_ret_t    vip_rc = VIP_OK;
    u_int32_t         pa_vec_size;
    XHH_cmd_pa_vec_t  *pa_vec_p = NULL;
    XHH_cmd_status_t  cmd_ret;
    MOSAL_phys_ctg_props_t props;            

    static const MOSAL_mem_flags_t
      mem_flags = MOSAL_MEM_FLAGS_PERM_READ |
                  MOSAL_MEM_FLAGS_PERM_WRITE;

    vip_rc = VIP_hashp_find(icmm->icmm_map_tbl, (u_int32_t) (icm_va >>MT_ARBEL_PAGE_SHIFT),
                              NULL);
    if (vip_rc == VIP_OK) {
        MTL_ERROR1(MT_FLFMT("%s: have duplicate entry. icm_va="U64_FMT", size="U64_FMT",prot_ctx=%d,flag=%d,va_p=0x%p"),
                   __func__, icm_va,size,prot_ctx,set_char_flag,os_map_p);
        return HH_EBUSY;
    }
    
    new_map_p = TMALLOC(XHH_icm_map_entry_t);
    if (new_map_p == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: could not allocate new mapping object"),__func__);
        return HH_EAGAIN;
    }
    memset(new_map_p,0,sizeof(XHH_icm_map_entry_t));

    new_map_p->prot_ctx = prot_ctx;
    rc = XHH_phys_mem_alloc(size,
                            MOSAL_SYS_PAGE_SHIFT, 
                            MOSAL_MEM_ALLOC_FLAGS_NONE,
                            (os_map_p != NULL) ? XHH_MEM_ALLOC_FLAGS_SINGLE_BLOCK : 0,
                            &(new_map_p->chunks_p), 
                            &(new_map_p->num_chunks));
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("%s: XHH_phys_mem_alloc failed (%d:%s)"),
                   __func__, rc, HH_strerror_sym(rc));
        goto phys_mem_alloc_fail;
    }

    /* we know here that size is a multiple of O/S page size */
    new_map_p->num_icm_pages = (u_int32_t)(size >> MT_ARBEL_PAGE_SHIFT);
    new_map_p->icm_va_pageaddr = (u_int32_t)((u_int64_t)icm_va >> MT_ARBEL_PAGE_SHIFT);

    /* do virtual addr map if requested.  If alloc call succeeded, we know that only
     * one chunk was returned
     */
    if (os_map_p != NULL) {
        cr = MOSAL_mem_phys_ctg_map(*(new_map_p->chunks_p),prot_ctx,mem_flags,os_map_p);
        if (cr != MT_OK) {
            MTL_ERROR1(MT_FLFMT("%s: MOSAL_phys_ctg_map failed (%d:%s). chunks=%d, prot_ctx=%d, pages=%d, os_map_p=0x%p"),
                       __func__, cr, mtl_strerror_sym(cr), new_map_p->num_chunks, (int) prot_ctx,
                       new_map_p->num_icm_pages, os_map_p);
            cr = MOSAL_mem_phys_ctg_get_props(*(new_map_p->chunks_p),&props);
            if (cr == MT_OK) {
                MTL_ERROR1(MT_FLFMT("%s: pa="PHYS_ADDR_FMT", order=%d, u_map_count=%d, k_map_count=%d"),
                           __func__, props.pa, props.order, props.u_map_count, props.k_map_count);
            }
            rc = HH_ERR;
            goto mem_map_fail;
        }
    }
    
    /* initialize the host memory if requested */
    if (set_char_flag == TRUE) {
        rc = icmm_memset_internal(icmm,icm_va, set_char, (u_int32_t)size, new_map_p);
        if (rc != HH_OK) {
            MTL_ERROR1(MT_FLFMT("%s: icmm_memset_internal failed (%d:%s)"),
                       __func__, rc, HH_strerror_sym(rc));
            goto mem_set_fail;
        }
    }

    pa_vec_size = new_map_p->num_chunks; 
    pa_vec_p = TNMALLOC(XHH_cmd_pa_vec_t, pa_vec_size);
    if (pa_vec_p == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: TMALLOC of pa_vec failed"),__func__);
        rc = HH_ENOMEM;
        goto alloc_pa_vec_err;
    }
    rc = XHH_build_pa_vec(new_map_p->chunks_p, pa_vec_size, pa_vec_p); 
    if (rc != HH_OK) {
        MTL_ERROR1(MT_FLFMT("%s: XHH_build_pa_vec failed"),__func__);
        goto build_pa_vec_err;
    }
    
    cmd_ret = XHH_cmd_MAP_ICM(icmm->cmd,icm_va,pa_vec_size, pa_vec_p);
    if (cmd_ret != XHH_CMD_STAT_OK) {
        MTL_ERROR1(MT_FLFMT("%s: XHH_cmd_MAP_ICM failed (%d)"),__func__, cmd_ret);
	rc = (cmd_ret == XHH_CMD_STAT_EFATAL) ? HH_EFATAL : HH_ERR;
        goto map_command_err;
    }

    /* now, save the mem object in the hash table */

    vip_rc = VIP_hashp_insert(icmm->icmm_map_tbl, (u_int32_t) (icm_va >>MT_ARBEL_PAGE_SHIFT),
                              new_map_p);
    if (vip_rc != VIP_OK) {
        MTL_ERROR1(MT_FLFMT("%s: VIP_hashp_insert failed (%d:%s)"),
                   __func__, vip_rc,VIP_common_strerror_sym(vip_rc));
        rc = HH_ERR;
        goto hash_insert_fail;
    }

    FREE(pa_vec_p);
    return HH_OK;

hash_insert_fail:
    cmd_ret = XHH_cmd_UNMAP_ICM(icmm->cmd,icm_va, new_map_p->num_icm_pages);
    if (cmd_ret != XHH_CMD_STAT_OK) {
        MTL_ERROR1(MT_FLFMT("%s: XHH_cmd_UNMAP_ICM failed (%d). icm_va="U64_FMT", num_pages=0x%x"),
                   __func__, cmd_ret, icm_va, new_map_p->num_icm_pages);
    }

map_command_err:
build_pa_vec_err:
    if (pa_vec_p != NULL) {
        FREE(pa_vec_p);
    }
alloc_pa_vec_err:
mem_set_fail:
mem_map_fail:
        /* Will unmap if needed */
        XHH_phys_mem_free(new_map_p->chunks_p,new_map_p->num_chunks);

phys_mem_alloc_fail:
    if (new_map_p != NULL) {
        FREE(new_map_p);
    }
    return rc;
}

/*
ARGUMENTS:
          icmm -   The XHH_icmm context 
          icm_va - Virtual address in ICM space. Must be 4KB aligned. 
          size -   Size to map in bytes. Must be 4KB multiple. 
          map_flags - ICMM_FORCE_LAM, ICMM_FORCE_HOST_MEM.  Specifying both flags is illegal.
                      If map_flags == 0 and LAM exists, first try LAM, then Host memory.
          prot_ctx - (optional) protection context to map pages to. 
          os_map_p - OS virtual mapping of given icm_va within given prot_ctx.
                       Mapping to prot_ctx performed only if this parameter is not NULL. 
RETURNS:
        HH_OK 
        HH_EINVAL - Given handle in unknown 
        HH_EINVAL_VA - given address space is not within ICM space. 
        
DESCRIPTION: Map given ICM virtual address space to physical pages.If os_map_pp is not NULL,
        the pages mapped to the ICM are also mapped to the virtual address space in the 
        given protection context. Currently,if os_map_pp is not NULL (i.e.,OS mapping required)
        we are can map only a physically contiguous chunk, so allocated physical chunk 
        would be allocated as physically contiguous. This may increase the chance for failure. 
        
*/
/*
 * Cases for icmm_map:
 *  1. Using LAM :
 *     a. Don't map to OS virt addr
 *         -- Can use a vector of LAM chunks.  LAM "page size" is 4K
 *     b. Map to virt addr:
 *         -- Need to use a SINGLE chunk of LAM, since we are mapping to virt addr space ( and the O/S
 *            determines the virtual address mapped to, not the caller)
 *         -- Can still use 4k "page size" as alignment, since dealing with device memory (not swappable).
 *          1) kernel protection context
 *          2) user protection context
 *
 *  2. Using HOST:
 *     a. Don't map to OS virt addr
 *         -- best to use min chunk size = HOST page size
 *         -- Can use a vector of HOST phys memory chunks
 *     b. Map to virt addr:
 *         -- Need to use a SINGLE chunk of HOST phys memory, since we are mapping to virt addr space.
 *         -- MUST use HOST page size as alignment, since dealing with Host managed virtual memory.
 *          1) kernel protection context
 *          2) user protection context
 *
 */
static HH_ret_t icmm_map_internal ( /*IN*/ XHH_icmm_t icmm, 
                        /*IN*/ u_int64_t icm_va, 
                        /*IN*/ u_int64_t size, 
                        /*IN*/ XHH_icmm_map_flags_t map_flags, 
                        /*IN*/ MOSAL_prot_ctx_t prot_ctx, 
                        /*IN*/ u_int32_t set_char_flag,
                        /*IN*/ u_int8_t  set_char,
                        /*OUT*/ MT_virt_addr_t *os_map_p)
{
    HH_ret_t    rc = HH_OK;
/* sanity checks: */
    if ((map_flags & ICMM_MAP_FORCE_LAM) && (map_flags & ICMM_MAP_FORCE_HOSTMEM)) {
        MTL_ERROR1(MT_FLFMT("%s: both FORCE LAM and FORCE_HOST_MEM flags set"),__func__);
        return HH_EINVAL;
    }

    if ((map_flags & ICMM_MAP_FORCE_LAM) && (icmm->lamm == (XHH_lamm_t)XHH_INVALID_HNDL)) {
        MTL_ERROR1(MT_FLFMT("%s: FORCE LAM requested, but LAM not present"),__func__);
        return HH_EINVAL;
    }
    if (icm_va & ((MAKE_ULONGLONG(1) << MOSAL_SYS_PAGE_SHIFT) - 1)) {
        MTL_ERROR1(MT_FLFMT("%s: virtual address provided ("U64_FMT")is not aligned to sys page size"),
                   __func__,icm_va);
        MTL_DEBUG2(MT_FLFMT("%s: size="U64_FMT", map_flags=%d, prot_ctx= 0x%x, set flag=%d, osmap=0x%p"),
                      __func__, size, map_flags,prot_ctx,set_char_flag,os_map_p);
        return HH_EINVAL;
    }

    if (size & ((MAKE_ULONGLONG(1) << MOSAL_SYS_PAGE_SHIFT) - 1)) {
        MTL_ERROR1(MT_FLFMT("%s: requested size ("U64_FMT")is not a multiple of sys page size"),
                   __func__,size);
        return HH_EINVAL;
    }

    if (!(map_flags & ICMM_MAP_FORCE_HOSTMEM)) {
        /* try LAM first if present */
        if (icmm->lamm != (XHH_lamm_t)XHH_INVALID_HNDL) {
            rc = icmm_map_lam(icmm,icm_va,size,prot_ctx, set_char_flag, set_char,os_map_p);
            if (rc != HH_OK) {
                if (map_flags & ICMM_MAP_FORCE_LAM) {
                    MTL_ERROR1(MT_FLFMT("%s: FORCE LAM, and could not map LAM memory to given icm (icm_va="U64_FMT",size="U64_FMT" (%d:%s)"),
                               __func__,icm_va, size, rc, HH_strerror_sym(rc));
                    return rc;
                } else {
                    MTL_DEBUG1(MT_FLFMT("%s: Could not map LAM memory to given icm (icm_va="U64_FMT",size="U64_FMT" (%d:%s). Try host mem."),
                               __func__,icm_va, size, rc, HH_strerror_sym(rc));
                }
            } else {
                return HH_OK;  /* map to LAM succeeded */
            }
        }
        /* If get here, lam not present, or map attempt failed. try using host memory */
        MTL_DEBUG2(MT_FLFMT("%s: Mapping host mem (icm_va="U64_FMT",size="U64_FMT" prot_ctx= 0x%x, set flag=%d, osmap=0x%p"),
                      __func__,icm_va, size, prot_ctx,set_char_flag,os_map_p);
        rc = icmm_map_hostmem(icmm,icm_va,size,prot_ctx,set_char_flag, set_char,os_map_p);
        if (rc != HH_OK) {
             MTL_ERROR1(MT_FLFMT("%s: Could not map HOST memory to given icm (icm_va="U64_FMT",size="U64_FMT" (%d:%s)"),
                           __func__,icm_va, size, rc, HH_strerror_sym(rc));
             return rc;
        }
    } else {
        /* force host memory set -- try host memory immediately. */
        rc = icmm_map_hostmem(icmm, icm_va, size, prot_ctx, set_char_flag, set_char,os_map_p);
        if (rc != HH_OK) {
             MTL_ERROR1(MT_FLFMT("%s: Could not map HOST memory to given icm (icm_va="U64_FMT",size="U64_FMT" (%d:%s)"),
                           __func__,icm_va, size, rc, HH_strerror_sym(rc));
             return rc;
        }
    }

    return rc;
}

HH_ret_t XHH_icmm_map( /*IN*/ XHH_icmm_t icmm, 
                        /*IN*/ u_int64_t icm_va, 
                        /*IN*/ u_int64_t size, 
                        /*IN*/ XHH_icmm_map_flags_t map_flags, 
                        /*IN*/ MOSAL_prot_ctx_t prot_ctx, 
                        /*OUT*/ MT_virt_addr_t *os_map_p)
{
    return icmm_map_internal (icmm,icm_va, size, map_flags, prot_ctx, FALSE,0, os_map_p);
}
/***************************************************************************************/
HH_ret_t XHH_icmm_set_n_map ( 
    /*IN  */  XHH_icmm_t           icmm, 
    /*IN  */  XHH_icm_va_t         icm_va, 
    /*IN  */  u_int64_t            size, 
    /*IN  */  XHH_icmm_map_flags_t map_flags, 
    /*IN  */  MOSAL_prot_ctx_t     prot_ctx, 
    /*IN */   u_int8_t             set_ch,
    /*OUT */  MT_virt_addr_t       *os_map_p
)
{
    HH_ret_t rc;
    rc = icmm_map_internal(icmm, icm_va, size,map_flags, prot_ctx, TRUE, set_ch,os_map_p);
    return rc;
}

/******************************************************************************************/
HH_ret_t 
XHH_icmm_unmap (
     /*IN */ XHH_icmm_t   icmm, 
     /*IN */ XHH_icm_va_t icm_va, 
     /*IN */ u_int64_t    size) 
{
    XHH_icm_map_entry_t* map_p;
    VIP_common_ret_t vcr;
    VIP_hashp_value_t hash_val;
    HH_ret_t rc = HH_OK;
    
    vcr= VIP_hashp_erase(icmm->icmm_map_tbl,(u_int32_t) (icm_va >>MT_ARBEL_PAGE_SHIFT),&hash_val);
    if (vcr != VIP_OK) {
            MTL_ERROR1(MT_FLFMT("%s: Failed hashp erase (%d:%s)"), 
		       __func__, vcr, VIP_common_strerror_sym(vcr));
            rc= HH_EINVAL;
            goto Exit;
    }else map_p = (XHH_icm_map_entry_t*)hash_val;
        
    rc= XHH_icmm_free_map_entry(icmm,map_p);
    if (rc != HH_OK) {
	MTL_ERROR1(MT_FLFMT("%s: failed XHH_icmm_free_map_entry (%d:%s)"),__func__, 
		   rc, HH_strerror_sym(rc));
    } else {
	FREE(map_p);
    }
    

Exit:    
    return rc;
}


static HH_ret_t  
XHH_icmm_free_map_entry(XHH_icmm_t icmm,XHH_icm_map_entry_t* map_p)
{
    HH_ret_t  rc, rc1 = HH_OK;
    call_result_t cr= MT_OK;
    XHH_cmd_status_t  cmd_stat;
    MT_size_t sz = map_p->num_icm_pages << MT_ARBEL_PAGE_SHIFT;

    cmd_stat = XHH_cmd_UNMAP_ICM(icmm->cmd,
                     (u_int64_t)(((u_int64_t)map_p->icm_va_pageaddr) << MT_ARBEL_PAGE_SHIFT), 
                     map_p->num_icm_pages);
    if (cmd_stat != XHH_CMD_STAT_OK) {
       MTL_ERROR1( MT_FLFMT("%s: failed XHH_cmd_UNMAP_ICM(%d). page_addr shifted = 0x%x, num_pages=0x%x "),
                   __func__, cmd_stat, map_p->icm_va_pageaddr, map_p->num_icm_pages);
    }
    
    if ((map_p->flags) & ICMM_MAP_TO_LAM) {
        if (map_p->lam_os_addr != 0) {
            cr= MOSAL_iomem_unmap(map_p->iomem_obj,map_p->prot_ctx);
            if (cr != MT_OK) {
                MTL_ERROR1(MT_FLFMT("failed MOSAL_iomem_unmap \n"));
                rc1 = HH_ERR;
            }
        }
        rc= XHH_lamm_free(icmm->lamm,map_p->lam_addr,sz);
        if (rc != HH_OK) {
             MTL_ERROR1(MT_FLFMT("failed deallocating lamm free\n"));
        }
    }else {
        /* was mapped to host memory. XHH_phys_mem_free also unmaps host memory */
        rc= XHH_phys_mem_free(map_p->chunks_p,map_p->num_chunks);
        if (rc != HH_OK) {
            MTL_ERROR1(MT_FLFMT("failed XHH_phys_mem_free \n"));
        }

    }
    
    if (rc1 != HH_OK) {
        rc = rc1;
    }
    
    return rc;
}


/* Insert to the list pointed by last_kmap_entry_p a kernel_map_entry for the LAM chunk given */
static XHH_icmm_kernel_map_entry_t *
		kernel_map_lam_block(XHH_icm_map_entry_t* blk_p,
				     XHH_icmm_kernel_map_t kernel_map,
				     XHH_icmm_kernel_map_entry_t *last_kmap_entry_p)

{
	XHH_icmm_kernel_map_entry_t *kmap_entry_p;
	XHH_icm_va_t blk_map_offset, blk_map_sz;
	const XHH_icm_va_t icm_map_va= kernel_map->map_start_va;
	const u_int64_t    map_size= kernel_map->map_size;
	const XHH_icm_va_t icm_map_start_page= icm_map_va >> MT_ARBEL_PAGE_SHIFT;
	const XHH_icm_va_t icm_map_end_page= (icm_map_va+map_size-1) >> MT_ARBEL_PAGE_SHIFT;
	call_result_t mosal_rc;

	kmap_entry_p= MALLOC(sizeof(XHH_icmm_kernel_map_entry_t));
	if (kmap_entry_p == NULL) {
		MTL_ERROR1(MT_FLFMT("Failed allocating mapping object"));
		return NULL;
	}
	
	if (icm_map_start_page < blk_p->icm_va_pageaddr) {
		blk_map_offset=  0;
	} else {
		/* Compute based on offset in block if start in this block */
		blk_map_offset=  icm_map_va - (blk_p->icm_va_pageaddr<<MT_ARBEL_PAGE_SHIFT);
	}
	if (icm_map_end_page >= (blk_p->icm_va_pageaddr + blk_p->num_icm_pages)) {
		blk_map_sz= (blk_p->num_icm_pages << MT_ARBEL_PAGE_SHIFT) - blk_map_offset;
	} else {
		/* Mapping size depends on offset of end from map_start */
		blk_map_sz= icm_map_va + map_size - (blk_p->icm_va_pageaddr + blk_map_offset);
	}

	mosal_rc = MOSAL_iomem_map(MOSAL_get_kernel_prot_ctx(),
                       MOSAL_MEM_FLAGS_NO_CACHE | MOSAL_MEM_FLAGS_PERM_READ | MOSAL_MEM_FLAGS_PERM_WRITE,
		       blk_p->lam_addr + blk_map_offset,
		       (MT_size_t) blk_map_sz,
		       &(kmap_entry_p->lamm_mem), &(kmap_entry_p->map_va));

	if (mosal_rc != MT_OK) {
	    MTL_ERROR1(MT_FLFMT("failed MOSAL_iomem_map (%d)"), mosal_rc);
	    FREE(kmap_entry_p);
	    return NULL;
	}

	kmap_entry_p->mapped_object_type= KMAP_LAM;
	kmap_entry_p->map_size= blk_map_sz;
	/* Links to chunks list */
	kmap_entry_p->next= last_kmap_entry_p;
	if (last_kmap_entry_p)  last_kmap_entry_p->next= kmap_entry_p;
	else kernel_map->chunks= kmap_entry_p; /* first */
	kernel_map->num_of_chunks ++;

	return kmap_entry_p;
}

/* Insert to the list pointed by last_kmap_entry_p kernel_map_entry's for the chunks given in blk_p */
static XHH_icmm_kernel_map_entry_t *
	kernel_map_hostmem_block(XHH_icm_map_entry_t* blk_p,
				 XHH_icmm_kernel_map_t kernel_map,
				 XHH_icmm_kernel_map_entry_t *last_kmap_entry_p)
{
	int i;
	call_result_t mrc;
	XHH_icm_va_t cur_chunk_page;
	MOSAL_phys_ctg_props_t chunk_props;
	XHH_icmm_kernel_map_entry_t *kmap_entry_p;
	MOSAL_mem_flags_t map_flags= MOSAL_MEM_FLAGS_PERM_WRITE | 
					MOSAL_MEM_FLAGS_PERM_READ;
	const XHH_icm_va_t icm_map_va= kernel_map->map_start_va;
	const u_int64_t    map_size= kernel_map->map_size;
	const XHH_icm_va_t icm_map_start_page= icm_map_va >> MT_ARBEL_PAGE_SHIFT;
	XHH_icm_va_t offset_in_chunk= 0;
	const XHH_icm_va_t icm_map_end_page= (icm_map_va+map_size-1) >> MT_ARBEL_PAGE_SHIFT;

	MTL_DEBUG6(MT_FLFMT("%s: blk_p=%p  last_kmap_entry_p=%p  icm_map_va="U64_FMT"  map_size="U64_FMT), __func__,
		   blk_p, last_kmap_entry_p, icm_map_va, map_size);

	for (i= 0, cur_chunk_page= blk_p->icm_va_pageaddr; 
	      (i < blk_p->num_chunks) && (cur_chunk_page <= icm_map_end_page); 
	      i++, cur_chunk_page += (1 << (chunk_props.order - MT_ARBEL_PAGE_SHIFT))) { 
		/* Find the chunk of this offset */
		mrc = MOSAL_mem_phys_ctg_get_props(blk_p->chunks_p[i], &chunk_props);
		if (mrc != MT_OK) {
		   MTL_ERROR1(MT_FLFMT("failed MOSAL_mem_phys_ctg_get_props for chunk %d at icm_va="U64_FMT" (%s)"),  
			      i, cur_chunk_page << MT_ARBEL_PAGE_SHIFT, mtl_strerror_sym(mrc));
		   break;
		}
		if (icm_map_start_page < cur_chunk_page+(1 << (chunk_props.order - MT_ARBEL_PAGE_SHIFT)) ) {
			if (icm_map_va > (cur_chunk_page << MT_ARBEL_PAGE_SHIFT)) { /* First chunk only */
				offset_in_chunk= icm_map_va - (cur_chunk_page << MT_ARBEL_PAGE_SHIFT);
			}
			/* Found a chunk to map (overlaps, at least partially) */
			kmap_entry_p= MALLOC(sizeof(XHH_icmm_kernel_map_entry_t));
			
			if (kmap_entry_p == NULL) {
				MTL_ERROR1(MT_FLFMT("Failed allocating mapping object"));
				return NULL;
			}

			mrc = MOSAL_mem_phys_ctg_map(blk_p->chunks_p[i], blk_p->prot_ctx, map_flags, 
						     &kmap_entry_p->map_va);
			if (mrc != MT_OK) {
			   MTL_ERROR1(MT_FLFMT("failed MOSAL_mem_phys_ctg_unmap of pa="PHYS_ADDR_FMT" (%s)"),  
				      chunk_props.pa, mtl_strerror_sym(mrc));
			   FREE(kmap_entry_p);
			   return NULL;
			}
			/* Update mapping va based on offset of start_page in current chunk */
			kmap_entry_p->map_va += offset_in_chunk;
			kmap_entry_p->mapped_object_type= KMAP_HOST;
			kmap_entry_p->hostmem= blk_p->chunks_p[i];
			MTL_DEBUG6(MT_FLFMT("phys_ctg_map hostmem=%p (i=%d), kmap_entry_p=%p, last_kmap_entry_p=%p, kernel_map->chunks=%p"), 
				   kmap_entry_p->hostmem, i, kmap_entry_p, last_kmap_entry_p, kernel_map->chunks);
			kmap_entry_p->map_size= (1 << chunk_props.order) - offset_in_chunk;
			/* link entry to chunks list */
			kmap_entry_p->next= NULL; /* New entry is last in the list */
			if (last_kmap_entry_p)  last_kmap_entry_p->next= kmap_entry_p;
			else kernel_map->chunks= kmap_entry_p; /* first */
			last_kmap_entry_p= kmap_entry_p;
			kernel_map->num_of_chunks ++;
			offset_in_chunk= 0; /* For all intermediate and last chunks */
		}

	}
	
	return last_kmap_entry_p;
}

/******************************************************************************************/
HH_ret_t XHH_icmm_kernel_map_create(
     /*IN*/ XHH_icmm_t   icmm, 
     /*IN*/ XHH_icm_va_t icm_start_blk_va, 
     /*IN*/ XHH_icm_va_t icm_map_va, 
     /*IN*/ u_int64_t    map_size,
     /*OUT*/ XHH_icmm_kernel_map_t *kernel_map_p
)
{
	XHH_icm_map_entry_t* blk_p;
	XHH_icmm_kernel_map_entry_t *last_kmap_entry_p= NULL;
	VIP_common_ret_t vcr;
	VIP_hashp_value_t hash_val;
	HH_ret_t rc = HH_OK;
	//XHH_icm_va_t end_va= icm_map_va + map_size;
	XHH_icm_va_t cur_blk_page;
	const XHH_icm_va_t icm_map_end_page= (icm_map_va+map_size-1) >> MT_ARBEL_PAGE_SHIFT;
	
	if (icm_map_va < icm_start_blk_va) {
		MTL_ERROR1(MT_FLFMT("Invalid map va ("U64_FMT") for start_blk_va ("U64_FMT")"),
			   icm_map_va, icm_start_blk_va);
		return HH_EINVAL;
	}
	
	*kernel_map_p= MALLOC(sizeof(struct XHH_icmm_kernel_map_st));
	if (*kernel_map_p == NULL) {
		MTL_ERROR1(MT_FLFMT("Failed allocating mapping object"));
		return HH_EAGAIN;
	}
	(*kernel_map_p)->map_size= map_size;
	(*kernel_map_p)->map_start_va= icm_map_va;
	(*kernel_map_p)->start_blk_va= icm_start_blk_va;
	(*kernel_map_p)->num_of_chunks= 0; /* Count chunks as they are mapped (in map_*_block) */
	(*kernel_map_p)->chunks= NULL;
	
	for (cur_blk_page= icm_start_blk_va >> MT_ARBEL_PAGE_SHIFT; 
	     cur_blk_page <= icm_map_end_page;
	     cur_blk_page += blk_p->num_icm_pages) { 
		/* Scan all blocks that cover requested mapping space */
		vcr= VIP_hashp_find(icmm->icmm_map_tbl,(u_int32_t)cur_blk_page, &hash_val);
		if (vcr != VIP_OK) {
			MTL_ERROR1(MT_FLFMT("%s: Failed hashp find for icm_blk_va="U64_FMT" (%d:%s)"), 
				   __func__, cur_blk_page<<MT_ARBEL_PAGE_SHIFT, 
				   vcr, VIP_common_strerror_sym(vcr));
			rc= HH_EINVAL;
			break;

		} else blk_p = (XHH_icm_map_entry_t*)hash_val;
		
		MTL_DEBUG6(MT_FLFMT("%s: Found blk_p=%p for cur_blk_page=0x%X"), __func__, blk_p, (u_int32_t)cur_blk_page);

		if ((blk_p->flags) & ICMM_MAP_TO_LAM) { /* A block may have only one LAM chunk */
			/* Case of LAM (io) mapping */
			last_kmap_entry_p= kernel_map_lam_block(blk_p, *kernel_map_p, last_kmap_entry_p);
		}else { 
			/* was mapped to host memory */
			last_kmap_entry_p= kernel_map_hostmem_block(blk_p, *kernel_map_p, last_kmap_entry_p);
		}
		
		if (last_kmap_entry_p == NULL) {
			rc= HH_EAGAIN;
			break;
		}
		
	}

	if (rc != HH_OK) {
		XHH_icmm_kernel_map_destroy(icmm, *kernel_map_p);
	}
	return rc;

}

/******************************************************************************************/
void XHH_icmm_kernel_map_destroy(
     /*IN*/ XHH_icmm_t   icmm, 
     /*IN*/ XHH_icmm_kernel_map_t kernel_map
)
{
	XHH_icmm_kernel_map_entry_t *kmap_entry_p;

	while ((kernel_map)->chunks != NULL) { /* Free items in chunks list */
		kmap_entry_p= kernel_map->chunks;
		kernel_map->chunks= kmap_entry_p->next;
		if (kmap_entry_p->mapped_object_type == KMAP_LAM) {
			MOSAL_iomem_unmap(kmap_entry_p->lamm_mem, MOSAL_get_kernel_prot_ctx());
		} else {
			MOSAL_mem_phys_ctg_unmap(kmap_entry_p->hostmem, MOSAL_get_kernel_prot_ctx());
		}
		FREE(kmap_entry_p);
	}
	FREE(kernel_map);
}



/******************************************************************************************/
HH_ret_t XHH_icmm_kernel_map_write_buf(
     /*IN*/ XHH_icmm_kernel_map_t kernel_map,
     /*IN*/ void* data,
     /*IN*/ MT_size_t size
)
{
	XHH_icmm_kernel_map_entry_t *cur_kmap_entry_p;
	MT_size_t offset_in_data, bytes2copy= 0;

	for (cur_kmap_entry_p= kernel_map->chunks , offset_in_data= 0;
	     (cur_kmap_entry_p != NULL) && (offset_in_data < size);
	     offset_in_data += bytes2copy , cur_kmap_entry_p= cur_kmap_entry_p->next) { 

		bytes2copy= size - offset_in_data;
		if (bytes2copy > cur_kmap_entry_p->map_size) {
			bytes2copy= cur_kmap_entry_p->map_size;
		}

		if (cur_kmap_entry_p->mapped_object_type == KMAP_HOST) {
			memcpy((void*)cur_kmap_entry_p->map_va, data + offset_in_data, bytes2copy);

		} else { /* KMAP_LAM: For I/O mapped memory we need to use MOSAL's macros */
			if ((bytes2copy & 7) == 0) {
				MOSAL_MMAP_IO_WRITE_BUF_QWORD(cur_kmap_entry_p->map_va, 
							      data + offset_in_data,
							      bytes2copy >> 3);
			} else if ((bytes2copy & 4) == 0) {
				MOSAL_MMAP_IO_WRITE_BUF_DWORD(cur_kmap_entry_p->map_va, 
							      data + offset_in_data,
							      bytes2copy >> 2);
			} else {
				MOSAL_MMAP_IO_WRITE_BUF_BYTE(cur_kmap_entry_p->map_va, 
							      data + offset_in_data,
							      bytes2copy);
			}
		}
	}
	
	if (offset_in_data < size) {
		MTL_ERROR1(MT_FLFMT("Copied only "SIZE_T_FMT" bytes out of "SIZE_T_FMT" due to mapping limitation !"),
			   offset_in_data, size);
		return HH_EINVAL;
	}
	return HH_OK;
}

#define XHH_ICMM_KERNEL_MAP_WRITE_TEMPLATE(data_type, func_suffix, mosal_mmap_write_suffix)	\
HH_ret_t XHH_icmm_kernel_map_write_##func_suffix(						\
     /*IN*/ XHH_icmm_kernel_map_t kernel_map,							\
     /*IN*/ XHH_icm_va_t offset,								\
     /*IN*/ data_type data									\
)												\
{												\
	XHH_icmm_kernel_map_entry_t *cur_kmap_entry_p;						\
	XHH_icm_va_t cur_chunk_offset, next_chunk_offset;					\
												\
	for (cur_kmap_entry_p= kernel_map->chunks , cur_chunk_offset= 0;			\
	     (cur_kmap_entry_p != NULL);							\
	     cur_chunk_offset= next_chunk_offset , cur_kmap_entry_p= cur_kmap_entry_p->next) { 	\
												\
		next_chunk_offset= cur_chunk_offset + cur_kmap_entry_p->map_size;		\
		if (offset >= next_chunk_offset)  continue; /* not in this chunk */		\
		if ((next_chunk_offset - offset) < sizeof(data_type)) {				\
			MTL_ERROR1(MT_FLFMT("data at offset "U64_FMT" crosses chunk boundry !"  \
					"(chunk_offset="U64_FMT" chunk_size="SIZE_T_FMT),		\
					offset, cur_chunk_offset, cur_kmap_entry_p->map_size);	\
			return HH_EINVAL;							\
		}										\
												\
		if (cur_kmap_entry_p->mapped_object_type == KMAP_HOST) {			\
			(*(volatile data_type*)(MT_virt_addr_t)(cur_kmap_entry_p->map_va +	\
						offset - cur_chunk_offset))= data;		\
		} else { /* KMAP_LAM: For I/O mapped memory we need to use MOSAL's macros */	\
			MOSAL_MMAP_IO_WRITE_##mosal_mmap_write_suffix((MT_virt_addr_t)		\
				(cur_kmap_entry_p->map_va + offset - cur_chunk_offset), data);	\
		}										\
		break; /* Write done */								\
	}											\
												\
	if (cur_kmap_entry_p == NULL) { 							\
		/* reached end of chunks list without finding one with given offset */		\
		MTL_ERROR1(MT_FLFMT("Failed to find a chunk at offset "U64_FMT			\
					" (map_size="U64_FMT") !"), 				\
					offset, kernel_map->map_size);				\
		return HH_EINVAL;								\
	}											\
	return HH_OK;										\
}

XHH_ICMM_KERNEL_MAP_WRITE_TEMPLATE(u_int64_t, qword, QWORD)
XHH_ICMM_KERNEL_MAP_WRITE_TEMPLATE(u_int32_t, dword, DWORD)
XHH_ICMM_KERNEL_MAP_WRITE_TEMPLATE(u_int8_t,  byte,  BYTE)

/******************************************************************************************/
HH_ret_t 
static icmm_memset_internal (
     /*IN */ XHH_icmm_t   icmm, 
     /*IN */ XHH_icm_va_t icm_va,
     /*IN */ u_int8_t     val,
     /*IN */ u_int32_t    size,
     XHH_icm_map_entry_t* map_p)
{
    MOSAL_phys_ctg_props_t host_props;
    call_result_t rc;
    MT_virt_addr_t k_va, write_va;
    MOSAL_iomem_t  k_iomem_obj = (MOSAL_iomem_t)NULL;
    u_int64_t      write_offset;
    u_int32_t      mapped_kernel_hostmem = FALSE;
    u_int32_t      mapped_kernel_iomem   = FALSE;
    static const MOSAL_mem_flags_t
      mem_flags = MOSAL_MEM_FLAGS_PERM_READ |
                  MOSAL_MEM_FLAGS_PERM_WRITE;
    
    /* Check validity.  For now, we will handle ONLY case where there is ONE phys chunk */
    if (map_p == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: map object is null"),__func__);
        return HH_EINVAL;
    }

    if (map_p->num_chunks != 1) {
        MTL_ERROR1(MT_FLFMT("%s: icmm VA at this addr has %d chunks. Only map if one chunk"),
                   __func__, map_p->num_chunks);
        return HH_ENOSYS;
    }

    /* check if LAMM memory or HOST memory*/
    if ((map_p->flags) & ICMM_MAP_TO_LAM) {
        /* map LAMM memory */
        if (map_p->prot_ctx != MOSAL_get_kernel_prot_ctx()) {
            /* map temporarily */
            rc = MOSAL_iomem_map(MOSAL_get_kernel_prot_ctx(),
                                 (mem_flags | MOSAL_MEM_FLAGS_NO_CACHE), 
                                 map_p->lam_addr, size, 
                                 &(k_iomem_obj),&k_va);

            if (rc != MT_OK) {
                MTL_ERROR1(MT_FLFMT("%s: KERNEL temporary MOSAL_iomem_map failed. physical address =" PHYS_ADDR_FMT " (%d:%s)"), 
                            __func__,map_p->lam_addr, rc, mtl_strerror_sym(rc));
                return HH_EAGAIN;
            }
            mapped_kernel_iomem = TRUE;
        } else {
            /* already mapped */
            k_va = map_p->lam_os_addr;
        }
    } else {
        /* mapping HOST memory */
        /* check if need to map the memory temporarily before writing */
        rc =  MOSAL_mem_phys_ctg_get_props(*(map_p->chunks_p),&host_props);
        if (rc != MT_OK) {
            MTL_ERROR1(MT_FLFMT("%s: MOSAL_mem_phys_ctg_get_props failed (%d:%s)"),
                       __func__, rc,mtl_strerror_sym(rc));
            return HH_ERR;
        }
        if (host_props.k_map_count == 0) {
             /* map temporarily */
	    MTL_DEBUG1(MT_FLFMT("%s: MAPPED PHYS to USER: pa =" PHYS_ADDR_FMT ", u_va="VIRT_ADDR_FMT", order=%u"), 
			__func__,host_props.pa, host_props.u_va,host_props.order);
            rc = MOSAL_mem_phys_ctg_map(*(map_p->chunks_p),MOSAL_get_kernel_prot_ctx(),
                                        mem_flags,&k_va);
            if (rc != MT_OK) {
                MTL_ERROR1(MT_FLFMT("%s: MOSAL_phys_ctg_map failed (%d:%s)"),
                           __func__, rc,mtl_strerror_sym(rc));
                return HH_EAGAIN;
            }
            mapped_kernel_hostmem = TRUE;
        } else {
            /* already mapped */
            k_va = host_props.k_va;
        }
    }

    /* Write the requested data at the required offset and size*/
    write_offset = icm_va - (u_int64_t)(((u_int64_t)map_p->icm_va_pageaddr) << MT_ARBEL_PAGE_SHIFT);
    write_va = k_va + (MT_virt_addr_t)write_offset;
    memset((void *)write_va, val, size);

    /* Now, unmap if performed a temporary mapping.  If unmapping fails, print out error message, 
     * but still return OK
     */
    if (mapped_kernel_hostmem == TRUE) {
        rc= MOSAL_mem_phys_ctg_unmap(*(map_p->chunks_p),MOSAL_get_kernel_prot_ctx());
        if (rc != MT_OK) {
            MTL_ERROR1(MT_FLFMT("%s: Failed MOSAL_mem_phys_ctg_unmap (%d:%s)"),
                       __func__,rc,mtl_strerror_sym(rc));
        }
    } else if (mapped_kernel_iomem == TRUE) {
        rc= MOSAL_iomem_unmap(k_iomem_obj, MOSAL_get_kernel_prot_ctx());
        if (rc != MT_OK) {
            MTL_ERROR1(MT_FLFMT("%s: Failed KERNEL MOSAL_iomem_unmap (%d:%s)"),
                       __func__,rc,mtl_strerror_sym(rc));
        }
    }
    return HH_OK;
}

/******************************************************************************************/
HH_ret_t 
XHH_icmm_memset (
     /*IN */ XHH_icmm_t   icmm, 
     /*IN */ XHH_icm_va_t icm_va,
     /*IN */ u_int8_t     val,
     /*IN */ u_int32_t    size)
{
    VIP_common_ret_t vcr;
    XHH_icm_map_entry_t* map_p;
    VIP_hashp_value_t hash_val;

    /* first, find the icmm va map entry for the memset. */
    vcr= VIP_hashp_find(icmm->icmm_map_tbl,(u_int32_t)((u_int64_t)(icm_va >>MT_ARBEL_PAGE_SHIFT)),&hash_val);
    if (vcr != VIP_OK) {
        MTL_ERROR1(MT_FLFMT("%s: Failed hashp find of icm_va="U64_FMT" (%d:%s)"),__func__,
                   icm_va, vcr,VIP_common_strerror_sym(vcr));
        return HH_EINVAL;
    } else map_p = (XHH_icm_map_entry_t*)hash_val;

    return icmm_memset_internal(icmm, icm_va, val, size, map_p);
}

