#include <mapm.h>
#include <vip_array.h>
#include <icmm.h>

#include <hob.h>
typedef struct XHH_mapm_st {
    XHH_icm_va_t tbl_base;
    u_int8_t log2_tbl_sz;   /* number of entries in tbl */
    u_int8_t log2_blk_sz;   /* number of entries in block */
    u_int8_t log2_entry_sz; /*entry Byte size */
    XHH_icmm_t icmm;
    MOSAL_mutex_t mutex;
    VIP_array_p_t blks;
    u_int32_t last_idx_rsvd;

}XHH_mapm_obj_t;


HH_ret_t 
XHH_mapm_calc(XHH_mapm_t mapm, u_int32_t blk_idx, XHH_icm_va_t* start_p, u_int64_t* sz_p );

HH_ret_t
XHH_mapm_create(IN XHH_icm_va_t tbl_base, 
            IN u_int8_t log2_tbl_sz, 
            IN u_int8_t log2_blk_sz, 
            IN u_int8_t log2_entry_sz, 
            IN XHH_hob_t hob,
            IN u_int32_t num_reserved_rsc,
            OUT XHH_mapm_t* p_mapm
           )
{
    VIP_common_ret_t vcr;
    XHH_mapm_t mobj;
    XHH_icmm_t icmm;
    HH_ret_t ret;


    FUNC_IN;

    ret = XHH_hob_get_icmm(hob,&icmm); 
    if (ret != HH_OK) {
        MTL_ERROR1(MT_FLFMT("%s: XHH_hob_get_icmm failed (%d: %s)"),__func__, ret, HH_strerror_sym(ret));
        return ret;
    }

    mobj = MALLOC(sizeof(struct XHH_mapm_st));
    if (mobj == NULL) {
        MT_COV_MEM_FLTR_START
        MTL_ERROR1(MT_FLFMT("%s: MALLOC failed"),__func__);
        return HH_EAGAIN;
        MT_COV_MEM_FLTR_END
    }
    mobj->icmm = icmm;
    mobj->tbl_base = tbl_base;
    mobj->log2_tbl_sz = log2_tbl_sz;
    mobj->log2_blk_sz = log2_blk_sz; 
    mobj->log2_entry_sz= log2_entry_sz; 
    mobj->last_idx_rsvd = num_reserved_rsc - 1;

    MOSAL_mutex_init(&mobj->mutex);

    if (log2_tbl_sz < log2_blk_sz)  {
        MTL_ERROR1(MT_FLFMT("%s: log2 tbl sz (0x%x) is less than log2 blk sz(0x%x)"),__func__,
                   log2_tbl_sz, log2_blk_sz);
        return HH_EINVAL_PARAM;
    }
    
    vcr= VIP_array_create((u_int32_t)(1UL << (log2_tbl_sz - log2_blk_sz)),&mobj->blks);
    if (vcr != VIP_OK) {
        MOSAL_mutex_free(&mobj->mutex);
        FREE(mobj);
        return HH_EAGAIN;
    }

    *p_mapm = mobj;
    return HH_OK;
}


HH_ret_t
XHH_mapm_destroy(IN XHH_mapm_t mapm)
{
    VIP_common_ret_t vcr;
    HH_ret_t ret= HH_OK;

    MOSAL_mutex_free(&mapm->mutex);

    vcr= VIP_array_destroy(mapm->blks,NULL);
    if (vcr != VIP_OK) {
        ret= HH_ERR;
    }

    FREE(mapm);
    return ret;
}


/* Should be called with mapm mutex acquired */
static HH_ret_t unmap_blocks(XHH_mapm_t mapm, u_int32_t start_blk_idx, u_int32_t end_blk_idx)
{
	HH_ret_t ret= HH_OK;
	VIP_common_ret_t vcr;
	u_int32_t blk_idx;
	XHH_icm_va_t start_unmap= 0;
	u_int64_t unmap_sz= 0;

	for (blk_idx= start_blk_idx; 
	     blk_idx <= end_blk_idx; 
	     blk_idx++) { 
		vcr= VIP_array_find_release_erase(mapm->blks,(VIP_array_handle_t)blk_idx,NULL);
		if (vcr != VIP_OK) {
		    if (vcr == VIP_EBUSY) {
			/*  EBUSY means that the blk containing the rsc entry is mapped but cannot
			    be erased due to other rsc allocated in it, so we are only decreasing
			    the ref count
			*/
			continue;
		    }else {
			    MTL_ERROR1(MT_FLFMT("Failed finding mapped block at index 0x%X"), blk_idx);
			    ret = HH_EFATAL;
			    break;
		    }
		}
	
		/* the blk containing the resource was found */
		ret= XHH_mapm_calc(mapm,blk_idx,&start_unmap,&unmap_sz);
		if (ret != HH_OK) {
			MTL_ERROR1(MT_FLFMT("Failed XHH_mapm_calc for block at index 0x%X"), blk_idx);
			break;
		}
	
		if (start_unmap == XHH_ICMM_INVALID_VA) {
			continue; /* was not mapped by mapm - no need to unmap */
		}
	
		MTL_DEBUG4(MT_FLFMT("%s: start_unmap:"U64_FMT" unmap_sz:"U64_FMT),__func__,start_unmap,unmap_sz); 
		ret= XHH_icmm_unmap(mapm->icmm, start_unmap, unmap_sz);
		MTL_DEBUG4(MT_FLFMT("%s: After unmap"),__func__);
		if (ret!= VIP_OK) {
		    MTL_ERROR1(MT_FLFMT("%s : failed unmapping "U64_FMT" bytes at icm_va="U64_FMT),__func__, 
			       start_unmap, unmap_sz); 
		    break;
		}
	}

	return ret;
}

/* Should be called with mapm mutex acquired */
static HH_ret_t map_blocks(XHH_mapm_t mapm, u_int32_t start_blk_idx, u_int32_t end_blk_idx)
{
	HH_ret_t ret= HH_OK;
	VIP_common_ret_t vcr;
	u_int32_t blk_idx;
	XHH_icm_va_t start_map= 0;
	u_int64_t map_sz= 0;

	for (blk_idx= start_blk_idx; 
	      blk_idx <= end_blk_idx; 
	      blk_idx++) { 
		/* Make sure all blocks given are mapped */

		vcr= VIP_array_find_hold(mapm->blks,(VIP_array_handle_t)blk_idx,NULL);
		if (vcr== VIP_OK) {
		    MTL_DEBUG4(MT_FLFMT("%s: rsc is contained in a block (0x%x) already mapped to icm"),__func__,blk_idx);
		    continue;
		} 

		if (vcr != VIP_EINVAL_HNDL) {
		    MTL_ERROR1(MT_FLFMT("%s: problems occured while trying to find the block containing the resource"),
			       __func__);
		    ret= HH_EFATAL;
		    break;
		}

		/* blk - or part of it (last_rsvd_blk_idx) is not mapped yet */

		ret= XHH_mapm_calc(mapm,blk_idx,&start_map,&map_sz);
		if (ret != HH_OK) {
		    MTL_ERROR1(MT_FLFMT("%s : failed XHH_mapm_calc. blk_idx=0x%x"), 
			       __func__, blk_idx);
		    break;
		}

		if (start_map != XHH_ICMM_INVALID_VA) {
			ret= XHH_icmm_map(mapm->icmm, start_map, map_sz, ICMM_MAP_FORCE_NONE,0,NULL);
			if (ret!= VIP_OK) {
			    MTL_ERROR1(MT_FLFMT("%s : failed mapping rsc to icm. blk_idx=0x%x, start_map="U64_FMT", map_sz:"
						U64_FMT",tbl_base="U64_FMT", log_tbl_size=0x%x, log_blk_size=0x%x, log_entry_sz=0x%x, last_ix=0x%x"),
				       __func__,blk_idx,start_map,map_sz, mapm->tbl_base, mapm->log2_tbl_sz, 
				       mapm->log2_blk_sz, mapm->log2_entry_sz,mapm->last_idx_rsvd); 
			    break;
			}
		}


		vcr= VIP_array_insert2hndl(mapm->blks,NULL,(VIP_array_handle_t)blk_idx);
		if (vcr != VIP_OK) {
			MTL_ERROR1(MT_FLFMT("Failed to insert into supposed to be free entry (blk_idx=0x%X)"),
				   blk_idx);
		    XHH_icmm_unmap(mapm->icmm,start_map,map_sz);
		    ret= HH_EFATAL;
		    break;
		}

		vcr= VIP_array_find_hold(mapm->blks,(VIP_array_handle_t)blk_idx,NULL);
		if (vcr != VIP_OK) {
			MTL_ERROR1(MT_FLFMT("Failed to find_hold new entry (blk_idx=0x%X)"),
				   blk_idx);
		    VIP_array_erase(mapm->blks,(VIP_array_handle_t)blk_idx,NULL);
		    XHH_icmm_unmap(mapm->icmm,start_map,map_sz);
		    ret= HH_EFATAL;
		    break;
		}
	}

	if (ret != HH_OK) {
		/* revert "mapping" of block already mapped within this call */
                MTL_ERROR1(MT_FLFMT("Failed %s: ret=%s. Unmapping(rollback) from start_blk_idx=0x%X to 0x%X"),
                                __func__, HH_strerror_sym(ret), start_blk_idx, blk_idx - 1);
		unmap_blocks(mapm, start_blk_idx, blk_idx - 1);
	}

	return ret; /* Return last block mapped's index */
}

HH_ret_t 
XHH_mapm_alloc_rsc(XHH_mapm_t mapm,u_int32_t rsc_idx) {
	return XHH_mapm_alloc_rscs(mapm, rsc_idx, 1);
}

HH_ret_t 
XHH_mapm_alloc_rscs(XHH_mapm_t mapm,u_int32_t rsc_idx, u_int32_t num_of_rscs)
{
    HH_ret_t ret= HH_OK; 
    u_int32_t start_blk_idx, end_blk_idx;
    
    FUNC_IN;
    MTL_DEBUG3(MT_FLFMT("%s: tbl_base="U64_FMT", log2_tbl_sz=%d, log2_blk_sz=%d, log2_entry_sz=%d, last_rsvd_idx=%d"),
               __func__, mapm->tbl_base,mapm->log2_tbl_sz,mapm->log2_blk_sz,mapm->log2_entry_sz,
               mapm->last_idx_rsvd);

    MTL_DEBUG4(MT_FLFMT("%s: rsc idx: 0x%x"),__func__,rsc_idx);
    
    if (MOSAL_mutex_acq(&mapm->mutex,TRUE) != MT_OK) {
	return HH_EINTR;
    }
    
    start_blk_idx = rsc_idx >> mapm->log2_blk_sz ;
    end_blk_idx = (rsc_idx + num_of_rscs - 1) >> mapm->log2_blk_sz ;

    ret= map_blocks(mapm,start_blk_idx,end_blk_idx);
    
    MOSAL_mutex_rel(&mapm->mutex);
    MT_RETURN(ret);
}

HH_ret_t 
XHH_mapm_free_rsc(XHH_mapm_t mapm, u_int32_t rsc_idx)
{
	return XHH_mapm_free_rscs(mapm,rsc_idx,1);
}

HH_ret_t 
XHH_mapm_free_rscs(XHH_mapm_t mapm, u_int32_t rsc_idx, u_int32_t num_of_rscs)
{
	HH_ret_t ret= HH_OK; 
	u_int32_t start_blk_idx, end_blk_idx;
    
    FUNC_IN;    
    MTL_DEBUG1(MT_FLFMT("%s: tbl_base="U64_FMT", log2_tbl_sz=%d, log2_blk_sz=%d, log2_entry_sz=%d, last_rsvd_idx=%d"),
               __func__, mapm->tbl_base,mapm->log2_tbl_sz,mapm->log2_blk_sz,mapm->log2_entry_sz,
               mapm->last_idx_rsvd);

    if (MOSAL_mutex_acq(&mapm->mutex,TRUE) != MT_OK) {
        return HH_EINTR;
    }

    start_blk_idx = rsc_idx >> mapm->log2_blk_sz ;
    end_blk_idx = (rsc_idx + num_of_rscs - 1) >> mapm->log2_blk_sz ;

    ret= unmap_blocks(mapm,start_blk_idx,end_blk_idx);
    
    MOSAL_mutex_rel(&mapm->mutex);
    MT_RETURN(ret);
}


HH_ret_t
XHH_mapm_alloc_all(XHH_icm_va_t tbl_base,
                   u_int8_t log2_tbl_sz, 
                   u_int8_t log2_tbl_entry_sz,
                   XHH_hob_t hob,
                   XHH_mapm_t* p_mapm)
{
    HH_ret_t ret;
    XHH_mapm_t mobj = MALLOC(sizeof(struct XHH_mapm_st));

    FUNC_IN;
    if (mobj == NULL) {
        MT_COV_MEM_FLTR_START
        MT_RETURN(HH_EAGAIN);
        MT_COV_MEM_FLTR_END
    }

    mobj->tbl_base = tbl_base;
    mobj->log2_tbl_sz = log2_tbl_sz;
    mobj->log2_entry_sz = log2_tbl_entry_sz;
    mobj->icmm= NULL;
    XHH_hob_get_icmm(hob,&mobj->icmm);
        
    ret= XHH_icmm_map(mobj->icmm,
                      tbl_base,
                      (MAKE_ULONGLONG(1) << (mobj->log2_entry_sz + mobj->log2_tbl_sz)),
                      ICMM_MAP_FORCE_NONE,
                      0,
                      NULL);
    if (ret!= VIP_OK) {
        MTL_ERROR1(MT_FLFMT("failed unmapping \n"));
        FREE(mobj);
    }else *p_mapm = mobj;

    MT_RETURN(ret);
}


HH_ret_t
XHH_mapm_free_all(XHH_mapm_t mapm)
{
    HH_ret_t ret= HH_OK;
    
    FUNC_IN;
    ret= XHH_icmm_unmap(mapm->icmm,mapm->tbl_base,
                        (1 << (mapm->log2_entry_sz + mapm->log2_tbl_sz)));
    if (ret!= VIP_OK) {
        MTL_ERROR1(MT_FLFMT("%s: failed unmapping"),__func__);
    } else {
	FREE(mapm);
    }

    MT_RETURN(ret);
}
   
HH_ret_t 
XHH_mapm_calc(XHH_mapm_t mapm, u_int32_t blk_idx, XHH_icm_va_t* start_p, u_int64_t* sz_p )
{
    u_int32_t last_rsvd_blk_idx = mapm->last_idx_rsvd >> mapm->log2_blk_sz;
    u_int32_t ofst_last_idx_rsvd;
    u_int32_t ofst_last_rsvd_pg_idx;
    u_int32_t pgs_per_blk;

    FUNC_IN;
    *start_p = XHH_ICMM_INVALID_VA;
    *sz_p =    XHH_ICMM_INVALID_SZ;

    if (blk_idx < last_rsvd_blk_idx ) {
        MTL_ERROR1(MT_FLFMT("%s: rsc block (0x%x) inside reserved block area (0x%x)"),
                   __func__,blk_idx, last_rsvd_blk_idx);
        MT_RETURN(HH_EINVAL);
    }

    /* the blk containing the resource was found */
    if (blk_idx == last_rsvd_blk_idx) {
        MTL_DEBUG1("%s: rsc block (0x%x) contains rsvd rscs\n",__func__,blk_idx);

        /* the blk is containing also rsvd resources - cannot be totally unmapped*/
        ofst_last_idx_rsvd = mapm->last_idx_rsvd & ( (1UL << mapm->log2_blk_sz) -1 );
        ofst_last_rsvd_pg_idx = ofst_last_idx_rsvd >> (MOSAL_SYS_PAGE_SHIFT - mapm->log2_entry_sz);

        pgs_per_blk = (1UL << (mapm->log2_blk_sz + mapm->log2_entry_sz - MOSAL_SYS_PAGE_SHIFT )); 
        if ((ofst_last_rsvd_pg_idx +1) == pgs_per_blk){
            MTL_DEBUG1(MT_FLFMT("%s: the resource idx is in the last page of the block"),__func__ );
            /* all the pages of the block contain rsvd resources */
        }else
        {
            *start_p = mapm->tbl_base + (((u_int64_t)blk_idx) << (mapm->log2_blk_sz + mapm->log2_entry_sz)) 
                          + (((u_int64_t)(ofst_last_rsvd_pg_idx+1)) << MOSAL_SYS_PAGE_SHIFT);  
            *sz_p = ((u_int64_t)pgs_per_blk - ((u_int64_t)(ofst_last_rsvd_pg_idx+1) )) << MOSAL_SYS_PAGE_SHIFT;
            
        }
    }
    else {
        /* map the block - normal flow */
        *start_p = mapm->tbl_base + (((u_int64_t)blk_idx) << (mapm->log2_blk_sz + mapm->log2_entry_sz));
        *sz_p = MAKE_ULONGLONG(1) << (mapm->log2_blk_sz + mapm->log2_entry_sz); 
    }

    MT_RETURN(HH_OK);
}


HH_ret_t 
XHH_mapm_rscs_map(XHH_mapm_t mapm, u_int32_t rsc_start_idx, u_int32_t num_of_rscs,
		 XHH_icmm_kernel_map_t *kernel_map_p)
{
	VIP_common_ret_t vcr;
	HH_ret_t rc= HH_OK; 
	u_int32_t blk_idx;
	XHH_icm_va_t blk_start, rscs_start;
	u_int64_t blk_sz, rscs_sz;
	
	blk_idx = rsc_start_idx >> mapm->log2_blk_sz ;
	
	vcr= VIP_array_find_hold(mapm->blks,(VIP_array_handle_t)blk_idx,NULL);
	if (vcr != VIP_OK) {
		MTL_ERROR1(MT_FLFMT("Block (index=0x%X) of resource (index=0x%X) is not mapped to ICM"),
			   blk_idx, rsc_start_idx);
		return HH_ENOMEM;
	}

	/* the blk containing the resource was found */
	rc= XHH_mapm_calc(mapm,blk_idx,&blk_start,&blk_sz);
	if (rc != HH_OK) {
	    goto bail;
	}
	rscs_start= mapm->tbl_base + (rsc_start_idx << mapm->log2_entry_sz);
	rscs_sz= (num_of_rscs << mapm->log2_entry_sz);
	if ((blk_start == XHH_ICMM_INVALID_VA) || (blk_start > rscs_start)) {
		/* If block is the "reserved" block - give start as the table base */
		blk_start= mapm->tbl_base; 
	}

	MTL_DEBUG4(MT_FLFMT("%s: tbl_base="U64_FMT" blk_start="U64_FMT" rsc_start_idx=0x%X"
			    " rscs_start="U64_FMT"rscs_sz="U64_FMT), __func__,
		   mapm->tbl_base, blk_start, rsc_start_idx, rscs_start, rscs_sz);
	rc= XHH_icmm_kernel_map_create(mapm->icmm,blk_start,rscs_start,rscs_sz,kernel_map_p);


	bail:
		VIP_array_find_release(mapm->blks,(VIP_array_handle_t)blk_idx);

	return rc;
}

void 
XHH_mapm_rscs_unmap(XHH_mapm_t mapm, XHH_icmm_kernel_map_t kernel_map)
{
	XHH_icmm_kernel_map_destroy(mapm->icmm, kernel_map);
}


