/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/
#include <mtl_common.h>
#include <mosal.h>
#include <hh.h>
#include "mem_utils.h"

#define XHH_PHYSMEM_LOG2_MAX_CHUNK_SIZE	16
typedef struct temp_objvec_st {
   MOSAL_phys_ctg_t   mem;
   struct temp_objvec_st  *next;
   struct temp_objvec_st  *prev;
} temp_objvec_t;

/* JPM:  FOR NOW:  DEAL ONLY WITH CASE WHERE ARBEL AND SYSTEM PAGE SIZE ARE IDENTICAL*/
HH_ret_t  XHH_phys_mem_alloc(/*IN*/   u_int64_t               size,
                             /*IN*/   MT_u_int_t              log2_align,
                             /*IN*/   MOSAL_mem_alloc_flags_t flags,
                             /*IN*/   u_int32_t               xhh_flags, /*force single block */
                             /*OUT*/  MOSAL_phys_ctg_t        **mem_vec_p,
                             /*OUT*/  u_int16_t               *mem_vec_size_p)
{
    int64_t remaining, this_alloc;
    MT_u_int_t     log2_size = floor_log2(size);
    u_int8_t log_to_try;
    temp_objvec_t *linked_head = NULL, *new_obj, *iterator = NULL, *linked_end;
    MOSAL_phys_ctg_t temp_mem_obj;
    call_result_t    rc;
    u_int32_t        object_count = 0;
    HH_ret_t         ret = HH_OK;

    FUNC_IN;
    MTL_DEBUG3(MT_FLFMT("%s: size="U64_FMT", align=0x%x"),__func__,size,log2_align);

    /* sanity checks */
    if ((mem_vec_p == NULL) || (mem_vec_size_p == NULL)) {
        MTL_ERROR1(MT_FLFMT("%s: null output parameter"), __func__);
        MT_RETURN(HH_EINVAL_PARAM);
    }

    if (log2_align < MOSAL_SYS_PAGE_SHIFT) {
        MTL_ERROR1(MT_FLFMT("%s: align shift value (%d) less than log2 system page size"), 
                  __func__, log2_align);
        MT_RETURN(HH_EINVAL_PARAM);
    }
    if (size < MOSAL_SYS_PAGE_SIZE) {
        MTL_ERROR1(MT_FLFMT("%s: size requested ("U64_FMT") less than system page size"), 
                  __func__, size);
        MT_RETURN(HH_EINVAL_PARAM);
    }
    if ((size & ((u_int64_t)(MOSAL_SYS_PAGE_SIZE-1))) != 0) {
        MTL_ERROR1(MT_FLFMT("%s: size requested ("U64_FMT") is not a multiple of system page size (0x%x)"), 
                  __func__, size, MOSAL_SYS_PAGE_SIZE);
        MT_RETURN(HH_EINVAL_PARAM);
    }

    if (log2_size > 62) {
        MTL_ERROR1(MT_FLFMT("%s: log2 size requested (%d) too large"), 
                  __func__, log2_size);
        MT_RETURN(HH_EINVAL_PARAM);
    }

    if ((xhh_flags & XHH_MEM_ALLOC_FLAGS_SINGLE_BLOCK) &&
        (size != (MAKE_ULONGLONG(1) << ceil_log2(size)))) {
        MTL_ERROR1(MT_FLFMT("%s: Single alloc requested, and size("U64_FMT") is not a power of 2"), 
                  __func__, size);
        MT_RETURN(HH_EINVAL_PARAM);
    }
    
    remaining = size;

    /* set a max chunk size if not requesting single-block allocation */
    if (xhh_flags & XHH_MEM_ALLOC_FLAGS_SINGLE_BLOCK) {
        log_to_try = (u_int8_t)log2_size;
    }  else {
        log_to_try = (log2_size > XHH_PHYSMEM_LOG2_MAX_CHUNK_SIZE) ? 
	    (u_int8_t) XHH_PHYSMEM_LOG2_MAX_CHUNK_SIZE : (u_int8_t)log2_size;
    }

    while (remaining > 0) {
        rc = MOSAL_mem_phys_ctg_alloc(log_to_try, log2_align, flags, &temp_mem_obj);
        MTL_DEBUG3(MT_FLFMT("%s: Remaining="U64_FMT", log_to_try=0x%x, size2try="U64_FMT",rc=%s"),
                   __func__,(u_int64_t)remaining,log_to_try, (u_int64_t)(MAKE_ULONGLONG(1)<<log_to_try),mtl_strerror_sym(rc));
        if (rc == MT_OK) {
            /* allocation succeeded */
            MOSAL_phys_ctg_props_t props;            
            rc = MOSAL_mem_phys_ctg_get_props(temp_mem_obj,&props);
            this_alloc = MAKE_ULONGLONG(1) << log_to_try;
            new_obj = TMALLOC(temp_objvec_t);
            if (new_obj == NULL) {
                MTL_ERROR1(MT_FLFMT("%s: MALLOC failure. remaining="U64_FMT",log_to_try=%d"), 
                          __func__, (u_int64_t)remaining, log_to_try);
                ret = HH_ENOMEM;
                break;
            }
            new_obj->mem = temp_mem_obj;
            new_obj->next = linked_head;
            new_obj->prev = NULL;
            if (linked_head != NULL) {
                linked_head->prev = new_obj;
            }
            linked_head = new_obj;
            object_count++;
            remaining -= this_alloc;
            if ((remaining > 0) && (log_to_try > floor_log2((u_int64_t) remaining))) {
                log_to_try = (u_int8_t)floor_log2((u_int64_t) remaining);
            }

            /* if have remaining and single alloc requested, fail here */

        } else {
            if ((log_to_try <= MOSAL_SYS_PAGE_SHIFT) || (xhh_flags & XHH_MEM_ALLOC_FLAGS_SINGLE_BLOCK)) {
                /* have reached minimum, or have requested that the allocation be in
                 * a single phys contiguous block.  Allocation fails */
                MTL_ERROR1(MT_FLFMT("%s: MOSAL_mem_phys_ctg_alloc failed (%s), log2try=%d,log2align=%d,flags=0x%x,xhh_flags=0x%x"),
                           __func__,mtl_strerror_sym(rc),log_to_try,log2_align,flags,xhh_flags);
                ret = HH_ENOMEM;
                break;
            } else {
                log_to_try--;
            }
        }
    }

    if ((remaining > 0) || 
        ((*mem_vec_p = TNMALLOC(MOSAL_phys_ctg_t,object_count )) == NULL)) {
        /* failed the allocation. Deallocate what was allocated */
        MTL_ERROR1(MT_FLFMT("%s: Failed: remaining=%d, *mem_vec_p = 0x%p"),
                   __func__, (int) remaining, *mem_vec_p);
        iterator = linked_head;
        while (iterator != NULL) {
            MOSAL_mem_phys_ctg_free(iterator->mem);
            new_obj = iterator;
            iterator = iterator->next;
            FREE(new_obj);
        }
        ret = HH_ENOMEM;
    } else {
        /* build the return allocation vector */
        /* Find list end */
        MTL_DEBUG3(MT_FLFMT("%s: OK. Count = %d, remaining=%d, *mem_vec_p = 0x%p"),
                   __func__, object_count, (int) remaining,  *mem_vec_p );
        iterator = linked_head;
        linked_end = NULL;
        while (iterator != NULL) {
            linked_end = iterator;
            iterator = iterator->next;
            MTL_DEBUG3(MT_FLFMT("%s: Finding linked list end. iterator=0x%p, linked_end = 0x%p"),
                       __func__, iterator, linked_end );
        }

        *mem_vec_size_p = (u_int16_t)object_count;

        /* add objects to return vector */
        object_count = 0;
        while (linked_end != NULL) {
            (*mem_vec_p)[object_count] = linked_end->mem;
            MTL_DEBUG3(MT_FLFMT("%s: address of mem obj = 0x%p"),
                       __func__,&((*mem_vec_p)[object_count]) );
            new_obj = linked_end;
            linked_end = linked_end->prev;
            FREE(new_obj);
            object_count++;
        }
        MT_RETURN(HH_OK);
    }
    MT_RETURN(ret);
}

HH_ret_t  XHH_phys_mem_pattern_initialize (/*IN*/ MOSAL_phys_ctg_t    *mem_vec_p,
                             /*IN*/ u_int32_t           mem_vec_size)
{
    u_int32_t i;
    MOSAL_phys_ctg_t *iterator = mem_vec_p;
    MT_virt_addr_t os_map;
    call_result_t  cr = MT_OK;
    u_int8_t       *fw_adr_p, u8val;
    u_int32_t      ui, usize;
    MOSAL_phys_ctg_props_t phys_props;

    FUNC_IN;
    /* sanity checks */
    if ((mem_vec_p == NULL) || (mem_vec_size == 0)) {
        MTL_ERROR1(MT_FLFMT("%s: null input parameter"), __func__);
        MT_RETURN(HH_EINVAL_PARAM);
    }

    for (i = 0; i < mem_vec_size; i++) {
        cr = MOSAL_mem_phys_ctg_map(*iterator,MOSAL_get_kernel_prot_ctx(),
                        (MOSAL_MEM_FLAGS_PERM_READ |MOSAL_MEM_FLAGS_PERM_WRITE),&os_map);
        if (cr != MT_OK) {
            MTL_ERROR1(MT_FLFMT("%s: MOSAL_phys_ctg_map failed (%d:%s) at chunk %u"),
                       __func__, cr, mtl_strerror_sym(cr), i);
	    MT_RETURN(HH_ERR);
        } else {
	    cr =  MOSAL_mem_phys_ctg_get_props(*iterator, &phys_props);
	    if (cr != MT_OK) {
		MTL_ERROR1(MT_FLFMT("%s: MOSAL_mem_phys_ctg_get_props failed (%d:%s) at chunk %u"),
			   __func__, cr, mtl_strerror_sym(cr), i);
		MOSAL_mem_phys_ctg_unmap(*iterator, MOSAL_get_kernel_prot_ctx());
		MT_RETURN(HH_ERR);
	    }
            fw_adr_p = (u_int8_t *)os_map;
            usize = (((u_int32_t)1) << phys_props.order);
	    u8val = 0;
	    MTL_ERROR1(MT_FLFMT("%s: fw_adr_p = %p, size=%u, order=%u"),
		       __func__, fw_adr_p, usize, phys_props.order);
            for (ui=0; ui<usize; ui++) {
                *fw_adr_p++ = u8val++;
            }

            MOSAL_mem_phys_ctg_unmap(*iterator, MOSAL_get_kernel_prot_ctx());
        }
        iterator++;
    }
    MT_RETURN(HH_OK);
}

HH_ret_t  XHH_phys_mem_free (/*IN*/ MOSAL_phys_ctg_t    *mem_vec_p,
                             /*IN*/ u_int32_t           mem_vec_size)
{
    int i;
    MOSAL_phys_ctg_t *iterator = mem_vec_p;

    FUNC_IN;
    /* sanity checks */
    if ((mem_vec_p == NULL) || (mem_vec_size == 0)) {
        MTL_ERROR1(MT_FLFMT("%s: null input parameter"), __func__);
        MT_RETURN(HH_EINVAL_PARAM);
    }

    for (i = 0; (u_int32_t)i < mem_vec_size; i++) {
        MOSAL_mem_phys_ctg_free(*iterator);
        iterator++;
    }
    FREE(mem_vec_p);
    MT_RETURN(HH_OK);
}

HH_ret_t  XHH_build_pa_vec(/*IN*/  MOSAL_phys_ctg_t  *mem_vec_p,
                           /*IN*/  u_int32_t         pa_vec_size,
                           /*OUT*/ XHH_cmd_pa_vec_t  *pa_vec_p)
{
    int i;
    call_result_t rc;
    MOSAL_phys_ctg_props_t phys_props;

    FUNC_IN;
    //MTL_ERROR1(MT_FLFMT("%s: ENTERING"), __func__);
    for (i = 0; (u_int32_t)i < pa_vec_size; i++, mem_vec_p++, pa_vec_p++) {
      MTL_DEBUG3(MT_FLFMT("%s: loop=%d, mem_vec_p=0x%p, pa_vec_p=0x%p"),
                  __func__,i,mem_vec_p, pa_vec_p);
      rc =  MOSAL_mem_phys_ctg_get_props(*mem_vec_p, &phys_props);
      if (rc != MT_OK) {
	  MTL_ERROR1(MT_FLFMT("%s: LEAVING -- ERROR"), __func__);
          MT_RETURN(HH_ERR);
      }
      pa_vec_p->phys_addr = phys_props.pa;
      pa_vec_p->log2size  = ((u_int8_t) phys_props.order - MT_ARBEL_PAGE_SHIFT);
      MTL_DEBUG3(MT_FLFMT("%s: phys_addr = "PHYS_ADDR_FMT", log2_size=%d"), 
		 __func__, pa_vec_p->phys_addr, pa_vec_p->log2size);
     }
    //MTL_ERROR1(MT_FLFMT("%s: LEAVING"), __func__);
    MT_RETURN(HH_OK);
}

HH_ret_t XHH_build_lamm_pa_vec(MT_phys_addr_t lam_addr,
                                  u_int64_t size,
                                  u_int32_t *pa_vec_size_p, 
                                  XHH_cmd_pa_vec_t  **pa_vec_pp)
{
    u_int32_t i,ones_count;
    u_int64_t temp_size,loop_size, min_size;
    XHH_cmd_pa_vec_t *pa_vec_p, *pa_vec_p_iterator;
    MT_phys_addr_t  iterator_phys_addr = lam_addr;

    /* count number of 1's in size -- this is size of pa_vec */
    for (i=0,temp_size=size,ones_count=0; 
          i<(sizeof(u_int64_t)*8); 
          i++,temp_size>>=1) {
        if(temp_size&1) ones_count++;
    }
    MTL_ERROR1(MT_FLFMT("%s: lam_addr="PHYS_ADDR_FMT", size="U64_FMT", ones_count=%d"),
               __func__, lam_addr,size,ones_count);

    pa_vec_p_iterator = pa_vec_p = TNMALLOC(XHH_cmd_pa_vec_t, ones_count);
    if (pa_vec_p == NULL) {
        MTL_ERROR1(MT_FLFMT("%s: TMALLOC of pa_vec failed"),__func__);
        return HH_ENOMEM;
    }

    temp_size = size;
    min_size = MAKE_ULONGLONG(1) << MT_ARBEL_PAGE_SHIFT;
    loop_size = MAKE_ULONGLONG(1) << ((sizeof(u_int64_t)*8)-1);
    while (loop_size >= min_size) {
        if (temp_size & loop_size) {
            /* have a power-of-2 */
            pa_vec_p_iterator->phys_addr = iterator_phys_addr;
            pa_vec_p_iterator->log2size  = ((u_int8_t)ceil_log2(loop_size) - MT_ARBEL_PAGE_SHIFT);
            MTL_ERROR1(MT_FLFMT("%s: pwr-of-2 = "U64_FMT", phys addr="PHYS_ADDR_FMT", log2_size=0x%x"),
                       __func__, loop_size, pa_vec_p_iterator->phys_addr, pa_vec_p_iterator->log2size);
            iterator_phys_addr += (MT_phys_addr_t)loop_size;
            pa_vec_p_iterator++;
            temp_size -= loop_size;
        }
        loop_size >>= 1;
    }

    *pa_vec_pp = pa_vec_p;
    *pa_vec_size_p = ones_count;
    return HH_OK;
}


