/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_XHH_PDM_H
#define H_XHH_PDM_H

#include <mosal.h>
#include <dev_common.h>
#include <hob.h>

#include "pdm_priv.h"

/******************************************************************************
 *  Function:     XHH_pdm_create
 *
 *  Arguments:
 *        hob -  The XHH_hob object in which this object will be included
 *        max_pd -  Maximum PDs allowed to be allocated
 *        pdm_p -  Allocated XHH_pdm object
 
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameter
 *    HH_EAGAIN - Not enough resources for creating this object
 *
 *  Description:
 *    Create object context which manages the PD resources.
 */
HH_ret_t XHH_pdm_create( /*IN*/  XHH_hob_t    hob, 
                          /*IN*/ XHH_pdm_init_t* init_p,
                          /*OUT*/ XHH_pdm_t   *pdm_p );

/******************************************************************************
 *  Function:     XHH_pdm_destroy
 *
 *  Arguments:
 *        pdm -  XHH_pdm object to destroy
 
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameter
 *
 *  Description:
 *    Free resources of the XHH_pdm object.
 */
HH_ret_t XHH_pdm_destroy( /*IN*/ XHH_pdm_t pdm );


/******************************************************************************
 *  Function:     XHH_pdm_alloc_pd
 *
 *  Arguments:
 *        pdm
 *        prot_ctx -  Protection context of user asking for this PD
 *        pd_ul_resources_p - Mostly UD AV table memory to register 
 *        pd_p - Allocated PD handle
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameter
 *    HH_EAGAIN - No free PDs to allocate
 *
 *  Description:
 *    Allocate a PD for given protection context.
 */
HH_ret_t XHH_pdm_alloc_pd( /*IN*/ XHH_pdm_t                     pdm, 
                            /*IN*/ MOSAL_protection_ctx_t       prot_ctx, 
                            /*IN*/ XHH_pd_ul_resources_t        *pd_ul_resources_p, 
                            /*OUT*/ HH_pd_hndl_t                *pd_p );

/******************************************************************************
 *  Function:     XHH_pdm_free_pd 
 *
 *  Arguments:
 *        pdm
 *        pd   - PD to free
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameter
 *
 *  Description:
 *    Free the PD.
 */
HH_ret_t XHH_pdm_free_pd( /*IN*/ XHH_pdm_t    pdm, 
                           /*IN*/ HH_pd_hndl_t  pd ) ;

/******************************************************************************
 *  Function:     XHH_pdm_get_protection_ctx 
 *
 *  Arguments:
 *        pdm
 *        pd -  PD for which the protection context is required
 *        prot_ctx_p - Returned protection context 
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameter
 *
 *  Description:
 *    This function returns the protection context associated with a PD.  It is used by
 *    XHH_mrwm_register_internal() in memory locking and mapping of WQE buffers to physical
 *    pages. (the mrwm is given a PD handle, and needs to retrieve the associated protection
 *    context).
 */
HH_ret_t XHH_get_protection_ctx( /*IN*/ XHH_pdm_t                 pdm, 
                                      /*IN*/ HH_pd_hndl_t               pd, 
                                      /*OUT*/ MOSAL_protection_ctx_t    *prot_ctx_p );


#endif
