/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef _UARM_H
#define _UARM_H

#include <mtl_common.h>
#include <mosal.h>
#include <xhh.h>
#include <dev_common.h>
#include <icmm.h>
#include <vip_array.h>
#include <vip_hashp.h>
#include "uarm_priv.h"

typedef enum {
    UARM_DBRECORD_SEND = 0,
    UARM_DBRECORD_RECEIVE = 1,
    UARM_DBRECORD_SET_CI = 2,
    UARM_DBRECORD_ARM_CQ = 3
}XHH_uarm_dbrecord_type_enum_t; 

typedef u_int8_t XHH_uarm_dbrecord_type_t;


/************************************************************************
 *  Function: XHH_uarm_create
 *
 *  Arguments:
 *    hob   The parent of this ICM 
 *    uar_base - physical adrs of the start of uar table  
 *    log2_max_uar - log of maximium number of uar can be allocated in the system 
 *    log2_uar_pg_sz - log of size of uar page 
 *    uar_ctx_base -va adrs in icm of the start of uar ctx table
 *    log2_uar_ctx_sz - log of uar ctx entry size
 *    uarm_p - return XHH_uarm obj
 *
 *  Returns:
 *    HH_OK
 *    HH_EAGAIN - Not enough resources for creating this object
 *
 *  Description:
 *    Create XHH_uarm object that manages the uar table and the uar ctx table in icm
 *  ALSO
 *      manages allocation of door bell records
 */
HH_ret_t XHH_uarm_create( /*IN*/  XHH_hob_t    hob, 
                          /*IN */ XHH_uarm_init_t* init_p,
                          /*OUT*/ XHH_uarm_t   *uarm_p );


/************************************************************************
 *  Function: XHH_uarm_destroy
 *
 *  Arguments:
 *   uarm - The handle of the object to destroy
 *
 *  Returns:
 *   HH_OK HH_EINVAL - No such object
 *
 *  Description:
 *   Free given object resources.
 *
 *   - Free all door bell records and unmap uar pages from icm and from user ctx
 *   - Free memory management structures. 
 *
 */
HH_ret_t XHH_uarm_destroy( /*IN*/ XHH_uarm_t uarm );



/*************************************************************************
 * Function: XHH_uarm_alloc_ul_res
 *
 *  Arguments:
 *        uldm 
 *        prot_ctx - Protection context of user level
 *        hca_ul_resources_p - Returned user level resources
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameter
 *    HH_EAGAIN - No available resources to allocate
 *
 *  Description:
 *    Allocate user level resources (UAR).
 */
HH_ret_t XHH_uarm_alloc_ul_res( /*IN*/ XHH_uarm_t              uarm, 
                                /*IN*/ MOSAL_protection_ctx_t  prot_ctx, 
                                /*OUT*/ XHH_hca_ul_resources_t *hca_ul_resources_p );

/*************************************************************************
 * Function: XHH_uarm_free_ul_res
 *
 *  Arguments:
 *        uldm 
 *        hca_ul_resources_p - A copy of resources structure returned on
 *                             resources allocation
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameter
 *
 *  Description:
 *    Free the resources allocated using XHH_uldm_alloc_ul_res()
 */
HH_ret_t XHH_uarm_free_ul_res( /*IN*/ XHH_uarm_t             uarm, 
                               /*IN*/ XHH_hca_ul_resources_t *hca_ul_resources_p);


/*************************************************************************
 * Function: XHH_uldm_alloc_uar
 *
 *  Arguments:
 *        uarm
 *        prot_ctx -  User level protection context to map UAR to 
 *        uar_index - Returned index of allocated UAR
 *        uar_map -   Virtual address in user level context to which the
 *                    allocated UAR is mapped
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameter
 *    HH_EAGAIN - No available UAR
 *
 *  Description:
 *    Allocate an available UAR and map it to user level memory space.
 */
HH_ret_t XHH_uarm_alloc_uar( /*IN*/ XHH_uarm_t              uarm, 
                             /*IN*/ MOSAL_protection_ctx_t  prot_ctx, 
                             /*OUT*/ u_int32_t              *uar_index, 
                             /*OUT*/ MT_virt_addr_t            *uar_map );


/******************************************************************************
 *  Function:     XHH_uarm_free_uar
 *
 *  Arguments:
 *        uarm
 *        uar_index - Index of UAR to free
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameter
 *
 *  Description:
 *    Unmap given UAR from user level memory space and return it to free UARs pool.
 */
HH_ret_t XHH_uarm_free_uar( /*IN*/ XHH_uarm_t    uarm, 
                            /*IN*/ XHH_uar_index_t    uar_index );


/*************************************************************************
 * Function: XHH_uldm_alloc_db_record
 *
 *  Arguments:
 *        uarm
 *        uar_index - idx of the uar to allocate in it the door bell record 
 *        p_db_record_idx- pointer to the new allocated db record in the uar
 *        db_record_va_p - pointer to the db record va  
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameter
 *    HH_EAGAIN - No available UAR
 *
 *  Description:
 *    allocate db record in a specified uar and map its page (if needed)
 */
HH_ret_t XHH_uarm_alloc_db_record(  /*IN*/ XHH_uarm_t                   uarm, 
                                    /*IN*/ XHH_uar_index_t              uar_idx, 
                                    /*IN*/ XHH_uarm_dbrecord_type_t     dbr_type,
                                    /*OUT*/ u_int32_t*                  p_db_record_idx,
                                    /*OUT*/ void**                      db_record_va_p);
/******************************************************************************
 *  Function:     XHH_uarm_free_db_record
 *
 *  Arguments:
 *        uarm
 *        uar_index - Index of UAR where the db record lies in it
 *        db_record_idx - Index of the db record to free
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameter
 *
 *  Description:
 *    
 */
HH_ret_t XHH_uarm_free_db_record(XHH_uarm_t uarm, 
                                 XHH_uar_index_t uar_idx, 
                                 XHH_uarm_dbrecord_type_t dbr_type,
                                 u_int32_t db_record_idx);



/*************************************************************************
 * Function: XHH_uarm_alloc_cmdsn_byte
 *
 *  Arguments:
 *        uarm
 *        uar_index - idx of the uar of the CQ that is going to use this cmd_sn 
 *	  kernel_cmdsn_pp - Returned kernel mapping of the cmd_sn (for updates)
 *	  consumer_cmdsn_pp - Returned consumer mapping of the cmd_sn (to return in cq_ul_resources)
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameter
 *    HH_EAGAIN - No available memory for a new cmd_sn
 *    HH_EINTR
 *
 *  Description:
 *    Allocate a byte for the expected cmd_sn of a CQ
 */
HH_ret_t XHH_uarm_alloc_cmdsn_byte( XHH_uarm_t          uarm,		/*IN*/ 
                                    XHH_uar_index_t     uar_idx,	/*IN*/ 
				    u_int8_t**		kernel_cmdsn_pp,	/*OUT*/
				    u_int8_t**		consumer_cmdsn_pp);	/*OUT*/

/******************************************************************************
 *  Function:     XHH_uarm_free_db_record
 *
 *  Arguments:
 *        uarm
 *        uar_idx - Index of UAR where the cmd_sn (CQ) belongs
 *        kernel_cmdsn_p - Address of kernel level mapping of the cmd_sn.
 *
 *  Returns:
 *    HH_OK
 *    HH_EINVAL - Invalid parameter (uarm or uar_idx)
 *    HH_EINVAL_VA - kernel_cmdsn_p given is unknown in this UAR
 *
 *  Description: Free cmd_sn byte 
 *    
 */
HH_ret_t XHH_uarm_free_cmdsn_byte( XHH_uarm_t          uarm,		/*IN*/ 
                                    XHH_uar_index_t     uar_idx,	/*IN*/ 
				    u_int8_t*		kernel_cmdsn_p);/*IN*/


#endif
