/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <mosal.h>
#include <udavm.h>
#include <ul_pdm.h>
#include <vip_array.h>
#include <vip_common.h>
#include <udav_comm.h>

/*================ type definitions ================================================*/

/* The main UDAV-manager structure */
typedef struct XHH_udavm_st {
  VIP_array_p_t    udavs_array;       /* the array which holds the free list of AV entries */
  XHHUL_pdm_t pdm;
  int kernel_max_av;
}XHH_udavm_int_t;

typedef struct XHH_ud_av_st {
  VAPI_ud_av_t ud_av;
  HH_pd_hndl_t pd;
  HHUL_pd_hndl_t ul_pd;
  u_int32_t udav_hw_entry[UD_AV_ENTRY_DWORD_SIZE];
} XHH_ud_av_t;


#ifdef MT_KERNEL
#define AV_MALLOC INTR_MALLOC
#else
#define AV_MALLOC MALLOC
#endif


#define UDAVM_FROM_HOB(hobul_hndl, udavm)          \
  hh_ret= XHHUL_hob_get_udavm(hobul_hndl, &udavm); \
  if (hh_ret != HH_OK) {                           \
    MTL_ERROR1(MT_FLFMT("%s: Failed getting XHHUL_hob's UDAVM (err=%d)."), __func__, hh_ret);\
    MT_RETURN(HH_EINVAL);                          \
  }



/*================ global functions definitions ====================================*/


HH_ret_t XHH_udavm_create( /*IN */ XHH_ver_info_t *version_p, 
                           /*IN */ XHHUL_pdm_t pdm,
                           /*IN */ u_int32_t  max_kernel_avs,
                           /*OUT*/ XHH_udavm_t *udavm_p)

{
  XHH_udavm_t new_udavm_p = NULL;
  HH_ret_t ret;
  
  FUNC_IN;

  /* allocation of object structure */
  new_udavm_p = (XHH_udavm_t)MALLOC(sizeof(XHH_udavm_int_t));
  if (!new_udavm_p) {
	MT_COV_MEM_FLTR_START
    MTL_ERROR4("%s: Cannot allocate UDAVM object.\n", __func__);
    MT_RETURN( HH_EAGAIN);
	MT_COV_MEM_FLTR_END
  }
  memset(new_udavm_p,0,sizeof(XHH_udavm_int_t));

  /* init the free list */
#ifdef MT_KERNEL
  new_udavm_p->kernel_max_av = max_kernel_avs;
  /* must use max_size since to limit the size used in kernel */
  ret = VIP_array_create_maxsize(1024,new_udavm_p->kernel_max_av,&(new_udavm_p->udavs_array));
  if ( ret != VIP_OK ) {
	MT_COV_ERR_FLTR_START
    MTL_ERROR1("%s: VIP_array_create_maxsize failed, ret=%d \n", __func__, ret);
    goto err_free_mem;
	MT_COV_ERR_FLTR_END
  }

#else /* user level - no need for maxsize */
  ret = VIP_array_create(256,&(new_udavm_p->udavs_array));
  if ( ret != VIP_OK ) {
	MT_COV_ERR_FLTR_START
    MTL_ERROR1("%s: VIP_array_create failed, ret=%d \n", __func__, ret);
    goto err_free_mem;
	MT_COV_ERR_FLTR_END
  }

#endif

  new_udavm_p->pdm = pdm;
  /* succeeded to create object - return params: */

  *udavm_p = new_udavm_p;

  MT_RETURN(HH_OK);

  /* error handling cleanup */
err_free_mem:
  MT_COV_ERR_FLTR_START
  FREE(new_udavm_p);
  MT_RETURN(ret);
  MT_COV_ERR_FLTR_END

}


/************************************************************************************/

HH_ret_t XHH_udavm_destroy( /*IN */ XHH_udavm_t udavm)
{
  
  FUNC_IN;
  
  if (udavm == NULL) {  
	MT_COV_ERR_FLTR_START
    MTL_ERROR4(MT_FLFMT("%s: udavm is NULL."),__func__);   
    MT_RETURN(HH_EINVAL);  
	MT_COV_ERR_FLTR_END
  }                                                        
  /* destroy the handlers array */
  if (VIP_array_destroy(udavm->udavs_array, NULL) != VIP_OK) {
	MT_COV_ERR_FLTR_START
    MTL_ERROR1("%s: VIP_array_destroy failed \n", __func__);
	MT_COV_ERR_FLTR_END
  }
  FREE(udavm);
  MT_RETURN(HH_OK);
}



/************************************************************************************/
HH_ret_t XHH_udavm_create_av( /*IN */ HHUL_hca_hndl_t hobul_hndl, 
                              /*IN */ HHUL_pd_hndl_t pd, 
                              /*IN */ VAPI_ud_av_t *av_p, 
                              /*OUT*/ HHUL_ud_av_hndl_t *ah_p)
{
  XHH_ud_av_t* av_entry_p;
  VIP_common_ret_t ret;
  HH_pd_hndl_t     hh_pd;
  HH_ret_t hh_ret=HH_OK;
  XHH_udavm_t udavm;

  FUNC_IN;

  UDAVM_FROM_HOB(hobul_hndl, udavm);

  XHH_UDAVM_VALID_CHECKS(udavm, av_p);
  XHH_UDAVM_AH_NULL(ah_p);
   
  av_entry_p = (XHH_ud_av_t*)AV_MALLOC(sizeof(XHH_ud_av_t));
  if (av_entry_p == NULL) {
	MT_COV_MEM_FLTR_START
    return HH_EAGAIN;
	MT_COV_MEM_FLTR_END
  }
  memset(av_entry_p, 0, sizeof(XHH_ud_av_t));
  
  if ((hh_ret = XHHUL_pdm_get_hh_pd_hold_ul_pd(udavm->pdm,pd,&hh_pd)) != HH_OK) {
	  MTL_ERROR1("XHH_udavm_create_av: could not get pd number from ul pd handle=%lu\n",pd);
	  goto pd_err;
  }

  memcpy(&av_entry_p->ud_av, av_p, sizeof(VAPI_ud_av_t));
  /* fix static rate */
  av_entry_p->ud_av.static_rate =  (av_p->static_rate==0) ? 0:3;

  XHH_udavm_fill_udav_entry(hh_pd, TRUE, av_p, av_entry_p->udav_hw_entry,FALSE,FALSE);
  av_entry_p->pd = hh_pd;
  av_entry_p->ul_pd = pd;

  /* get a free handle for the udav */
  ret = VIP_array_insert(udavm->udavs_array, av_entry_p,(VIP_array_handle_t*)ah_p);
  if (ret != VIP_OK) {
    MTL_ERROR4("XHH_udavm_create_av: Not enough resources.\n");
	hh_ret = HH_EAGAIN;
    goto insert_err;
  }

  MT_RETURN(HH_OK);

insert_err:
  XHHUL_pdm_rel_ul_pd(udavm->pdm,pd);
pd_err:
  FREE(av_entry_p);
  MT_RETURN(hh_ret);
}

/************************************************************************************/
HH_ret_t XHH_udavm_modify_av( /*IN */ HHUL_hca_hndl_t hobul_hndl, 
                              /*IN */ HH_ud_av_hndl_t ah, 
                              /*IN */ VAPI_ud_av_t *av_p )
{
  XHH_ud_av_t* av_entry_p;
  HH_ret_t hh_ret=HH_OK;
  VIP_common_ret_t ret=VIP_OK;
  XHH_udavm_t udavm;

  FUNC_IN;
  
  UDAVM_FROM_HOB(hobul_hndl, udavm);
  XHH_UDAVM_VALID_CHECKS(udavm, av_p);

  /* check that ah is a valid (allocated) handle */
  ret = VIP_array_find_hold(udavm->udavs_array, (VIP_array_handle_t)(MT_ulong_ptr_t)ah, (VIP_array_obj_t *)&av_entry_p);
  MTL_DEBUG3(MT_FLFMT("Calling VIP_array_find_hold, ah="MT_ULONG_PTR_FMT", ret=%d\n"), ah, ret);
  if ( ret!=VIP_OK ) {
    MTL_ERROR4(MT_FLFMT("%s: handle is not valid. ah="MT_ULONG_PTR_FMT", ret=%d\n"),__func__, ah, ret);
    MT_RETURN(HH_EINVAL_AV_HNDL);
  } 
  memcpy(&av_entry_p->ud_av, av_p, sizeof(VAPI_ud_av_t));
  /* fix static rate */
  av_entry_p->ud_av.static_rate =  (av_p->static_rate==0) ? 0:3;
  /* TK: Do we need a spinlock here or maybe VIP array is enough we need this to protect from geting the uadv for the post */
  XHH_udavm_fill_udav_entry(av_entry_p->pd, TRUE, av_p, av_entry_p->udav_hw_entry,FALSE,FALSE);
  
  ret = VIP_array_find_release(udavm->udavs_array, (VIP_array_handle_t)(MT_ulong_ptr_t)ah);
  MTL_DEBUG3(MT_FLFMT("Calling VIP_array_find_release ret=%d"), ret);
  if ( ret!=VIP_OK ) {
    MTL_ERROR1("%s: Internal mismatch - ah ("MT_ULONG_PTR_FMT") is not in array\n", __func__, ah);
    hh_ret = ret;
  }
  MTL_DEBUG4(MT_FLFMT("%s: address handle = "MT_ULONG_PTR_FMT"\n"),__func__, ah);
  MT_RETURN(hh_ret);
}


HH_ret_t XHH_udavm_query_av_internal(/*IN */ XHH_udavm_t udavm, 
                                     /*IN */ HH_ud_av_hndl_t ah, 
                                     /*OUT*/ VAPI_ud_av_t *av_p)
{
	XHH_ud_av_t* av_entry_p;
	HH_ret_t hh_ret=HH_OK;
	VIP_common_ret_t ret=VIP_OK;
	
	XHH_UDAVM_BASIC_VALID_CHECKS(udavm, av_p);

	ret = VIP_array_find_hold(udavm->udavs_array, (VIP_array_handle_t)(MT_ulong_ptr_t)ah, (VIP_array_obj_t *)&av_entry_p);
	MTL_DEBUG3(MT_FLFMT("Calling VIP_array_find_hold, ah="MT_ULONG_PTR_FMT", ret=%d\n"), ah, ret);
	if ( ret!=VIP_OK ) {
      MTL_ERROR4(MT_FLFMT("%s: handle is not valid. ah="MT_ULONG_PTR_FMT", ret=%d\n"),__func__, ah, ret);
	  MT_RETURN(HH_EINVAL_AV_HNDL);
    } 
	memcpy(av_p, &av_entry_p->ud_av, sizeof(VAPI_ud_av_t));

	ret = VIP_array_find_release(udavm->udavs_array, (VIP_array_handle_t)(MT_ulong_ptr_t)ah);
	MTL_DEBUG3(MT_FLFMT("Calling VIP_array_find_release ret=%d"), ret);
	if ( ret!=VIP_OK ) {
	  MT_COV_ERR_FLTR_START
	  MTL_ERROR1("%s: Internal mismatch - ah ("MT_ULONG_PTR_FMT") is not in array\n", __func__,ah);
	  hh_ret = ret;
	  MT_COV_ERR_FLTR_END
	}
	MTL_DEBUG4(MT_FLFMT("%s: address handle ="MT_ULONG_PTR_FMT"\n"),__func__,ah);

      #if 1 <= MAX_DEBUG
	print_udav(av_p);
      #endif
	MT_RETURN(hh_ret);
}

/************************************************************************************/
HH_ret_t XHH_udavm_query_av( /*IN */ HHUL_hca_hndl_t hobul_hndl, 
                             /*IN */ HHUL_ud_av_hndl_t ah, 
                             /*OUT*/ VAPI_ud_av_t *av_p )
{
	XHH_udavm_t udavm;
	HH_ret_t hh_ret=HH_OK;

	FUNC_IN;

	UDAVM_FROM_HOB(hobul_hndl, udavm);
	return XHH_udavm_query_av_internal(udavm, ah,av_p);
}

/************************************************************************************/
HH_ret_t XHH_udavm_destroy_av( /*IN */ HHUL_hca_hndl_t hobul_hndl, 
                               /*IN */ HHUL_ud_av_hndl_t ah )
{
  
  VIP_common_ret_t ret=VIP_OK;
  HH_ret_t hh_ret=HH_OK;
  XHH_udavm_t udavm;
  VIP_array_obj_t  udav_obj;
  XHH_ud_av_t* av_entry_p;

  FUNC_IN;
  
  UDAVM_FROM_HOB(hobul_hndl, udavm);
  
  if (udavm == NULL) { 
	MT_COV_ERR_FLTR_START
    MTL_ERROR4(MT_FLFMT("%s: udavm is NULL."),__func__);   
    MT_RETURN(HH_EINVAL);
	MT_COV_ERR_FLTR_END
  }                                                        

  ret = VIP_array_erase(udavm->udavs_array, (VIP_array_handle_t)(MT_ulong_ptr_t)ah, &udav_obj);
  MTL_DEBUG3(MT_FLFMT("Calling VIP_array_erase, ah="MT_ULONG_PTR_FMT", ret=%d\n"), ah, ret);
  if (ret==VIP_EBUSY ) {
    MTL_ERROR4(MT_FLFMT("%s: handle is BUSY. ah="MT_ULONG_PTR_FMT", ret=%d\n"),__func__, ah, ret);
    hh_ret = HH_EBUSY;
  } else if (ret == VIP_EINVAL_HNDL) {
    MTL_ERROR4(MT_FLFMT("%s: handle is not valid. ah="MT_ULONG_PTR_FMT", ret=%d\n"),__func__, ah, ret);
    hh_ret = HH_EINVAL_AV_HNDL;
  }

  if (ret == VIP_OK) {
      av_entry_p = (XHH_ud_av_t*)udav_obj;
      XHHUL_pdm_rel_ul_pd(udavm->pdm,av_entry_p->ul_pd);
      FREE((void *)udav_obj);
  }

  MTL_DEBUG4(MT_FLFMT("XHH_udavm_destroy_av: entry index="MT_ULONG_PTR_FMT"\n"),ah);
  MT_RETURN(hh_ret);
}





/************************************************************************************/


HH_ret_t XHH_udavm_put_av(/*IN*/ XHH_udavm_t udavm,
                          /*IN*/ HHUL_ud_av_hndl_t ah,
                          /*OUT*/ volatile u_int32_t *wqe_ud_segment_p)
{
  HH_ret_t hh_ret=HH_OK;
  VIP_common_ret_t ret=VIP_OK;
  XHH_ud_av_t* av_entry_p;

  FUNC_IN;
  
  /* check that ah is a valid (allocated) handle */
  ret = VIP_array_find_hold(udavm->udavs_array, (VIP_array_handle_t)(MT_ulong_ptr_t)ah, (VIP_array_obj_t *)&av_entry_p);
  MTL_DEBUG3(MT_FLFMT("Calling VIP_array_find_hold, ah="MT_ULONG_PTR_FMT", ret=%d\n"), ah, ret);
  if ( ret!=VIP_OK ) {
    MTL_ERROR4(MT_FLFMT("%s: handle is not valid. ah="MT_ULONG_PTR_FMT", ret=%d\n"),__func__, ah, ret);
    MT_RETURN(HH_EINVAL_AV_HNDL);
  } 
  
  memcpy((void*)wqe_ud_segment_p, &av_entry_p->udav_hw_entry,UD_AV_ENTRY_SIZE);

  ret = VIP_array_find_release(udavm->udavs_array, (VIP_array_handle_t)(MT_ulong_ptr_t)ah);
  MTL_DEBUG3(MT_FLFMT("Calling VIP_array_find_release ret=%d"), ret);
  if ( ret!=VIP_OK ) {
	MT_COV_ERR_FLTR_START
    MTL_ERROR1("%s: Internal mismatch - ah ("MT_ULONG_PTR_FMT") is not in array\n", __func__,ah);
    hh_ret = ret;
	MT_COV_ERR_FLTR_END
  }
  MT_RETURN(hh_ret);

}
