/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include "ul_pdm_priv.h"

HH_ret_t XHHUL_pdm_create (
   XHHUL_hob_t hob, 
   XHHUL_pdm_t *pdm_p )
{
    XHHUL_pdm_t    new_pdm_obj;
    VIP_common_ret_t  vret;
	
    /* create new pdm object */
    new_pdm_obj = TMALLOC(struct XHHUL_pdm_st);
    if (!new_pdm_obj) {
        MT_COV_MEM_FLTR_START
        return HH_EINVAL;
        MT_COV_MEM_FLTR_END
    }
    memset(new_pdm_obj, 0, sizeof(struct XHHUL_pdm_st));

    vret = VIP_array_create(XHHUL_PDM_INITIAL_NUM_PDS,&(new_pdm_obj->pd_array));
    if (vret != VIP_OK) {
        MTL_ERROR1("XHHUL_pdm_create: ERROR (%d) : could not create PD array\n", vret);
        FREE(new_pdm_obj);
        return HH_EAGAIN;
    }

    *pdm_p = new_pdm_obj;
    return HH_OK;
}

HH_ret_t XHHUL_pdm_destroy ( XHHUL_pdm_t pdm)
{
    VIP_array_destroy(pdm->pd_array,NULL);
    FREE(pdm);
    return HH_OK;
}

HH_ret_t XHHUL_pdm_alloc_pd_avs_prep (
   HHUL_hca_hndl_t hca, 
   u_int32_t max_num_avs,
   HH_pdm_pd_flags_t pd_flags,
   HHUL_pd_hndl_t *pd_p, 
   void *pd_ul_resources_p )
{
    HH_ret_t              ret;
    XHHUL_pdm_t           pdm;
    XHH_pd_ul_resources_t *pd_ul_res = (XHH_pd_ul_resources_t *)pd_ul_resources_p;
    VIP_common_ret_t      vret;
    VIP_array_handle_t    local_pd_hndl;

    ret = XHHUL_hob_get_pdm(hca, &pdm);
    if (ret != HH_OK) {
        MTL_ERROR1("XHHUL_pdm_alloc_pd_avs_prep: ERROR (%d) : PDM object has not yet been created\n", ret);
        return HH_EINVAL;
    }

    memset(pd_ul_res, 0, sizeof(XHH_pd_ul_resources_t));

    /* add to array */
    if ((vret=VIP_array_insert_prep(pdm->pd_array, (VIP_array_obj_t)0, &local_pd_hndl)) != VIP_OK) {
        MTL_ERROR1("XHHUL_pdm_alloc_pd_avs_prep: ERROR (%d) : Insertion failure.\n", vret);
        return HH_EAGAIN;
    }

    *pd_p = (HHUL_pd_hndl_t)local_pd_hndl;  /* return allocated PD handle */
    return HH_OK;
}

HH_ret_t XHHUL_pdm_alloc_pd_done (
   HHUL_hca_hndl_t hca, 
   HHUL_pd_hndl_t hhul_pd, 
   HH_pd_hndl_t hh_pd, 
   void  *pd_ul_resources_p )
{
    HH_ret_t     ret;
    XHHUL_pdm_t  pdm;
    VIP_common_ret_t vret;

    MTL_DEBUG3("==> XHHUL_pdm_alloc_pd_done\n");

    ret = XHHUL_hob_get_pdm(hca, &pdm);
    if (ret != HH_OK) {
        MTL_ERROR1("XHHUL_pdm_alloc_pd_done: ERROR (%d) : PDM object has not yet been created\n", ret);
        return HH_EINVAL;
    }

   if ((vret=VIP_array_insert_done(pdm->pd_array, (VIP_array_obj_t)(MT_ulong_ptr_t)hh_pd, (VIP_array_handle_t)hhul_pd)) != VIP_OK) {
        MTL_ERROR1("XHHUL_pdm_alloc_pd_done: ERROR (%d) : Insertion failure.\n", vret);
        return HH_EAGAIN;
    }

    MTL_DEBUG3("<== XHHUL_pdm_alloc_pd_done\n");
    return HH_OK;
}


HH_ret_t XHHUL_pdm_free_pd_prep (
   HHUL_hca_hndl_t hca, 
   HHUL_pd_hndl_t hhul_pd,
   MT_bool        undo_flag )
{
    HH_ret_t     ret;
    XHHUL_pdm_t  pdm;
    VIP_common_ret_t vret;

    ret = XHHUL_hob_get_pdm(hca, &pdm);
    if (ret != HH_OK) {
        MTL_ERROR1("XHHUL_pdm_free_pd_prep: ERROR (%d) : PDM object has not yet been created\n", ret);
        return HH_EINVAL;
    }

    /* undoing a previous prep */
    if (undo_flag == TRUE) {
        if ((vret=VIP_array_erase_undo(pdm->pd_array,(VIP_array_handle_t)hhul_pd)) != VIP_OK){
                MTL_ERROR1("XHHUL_pdm_free_pd_prep: ERROR (%d) : invalid handle\n", vret);
                return HH_EINVAL_PD_HNDL;
        }
        return HH_OK;
    }

    /* preparing a PD FREE */
    /* need to find pd table entry in pd list and signal it as prepared for erase.  */
    if ((vret=VIP_array_erase_prepare(pdm->pd_array,(VIP_array_handle_t)hhul_pd, NULL)) != VIP_OK){
        if (vret == VIP_EBUSY) { 
            MTL_DEBUG4("XHHUL_pdm_free_pd_prep:  PD object is busy\n");
            return HH_EBUSY;
        } else {
            MTL_ERROR1("XHHUL_pdm_free_pd_prep: ERROR (%d) : Could not find PD object\n", vret);
            return HH_EINVAL_PD_HNDL;
        }
    }
    return HH_OK;
}


HH_ret_t XHHUL_pdm_free_pd_done (
   HHUL_hca_hndl_t hca, 
   HHUL_pd_hndl_t hhul_pd )
{
    HH_ret_t     ret;
    XHHUL_pdm_t  pdm;
    VIP_common_ret_t vret;

    ret = XHHUL_hob_get_pdm(hca, &pdm);
    if (ret != HH_OK) {
        MTL_ERROR1("XHHUL_pdm_free_pd_done: ERROR (%d) : PDM object has not yet been created\n", ret);
        return HH_EINVAL;
    }

    /* need to find pd table entry in pd list  and delete the entry from list */
    if ((vret=VIP_array_erase_done(pdm->pd_array,(VIP_array_handle_t)hhul_pd, 
                              (VIP_array_obj_t*) (NULL))) != VIP_OK){
            MTL_ERROR1("XHHUL_pdm_free_pd_done: ERROR (%d) : Could not find PD object\n", vret);
            return HH_EINVAL_PD_HNDL;
    }

    return HH_OK;
}


HH_ret_t XHHUL_pdm_get_hh_pd (
   XHHUL_pdm_t  pdm,
   HHUL_pd_hndl_t hhul_pd, 
   HH_pd_hndl_t *hh_pd)
{
    VIP_common_ret_t vret;
    VIP_array_obj_t  vip_obj;

    MTL_DEBUG3("==> XHHUL_pdm_get_hh_pd\n");


    if ((vret=VIP_array_find(pdm->pd_array, (VIP_array_handle_t)hhul_pd, &vip_obj)) != VIP_OK) {
        MTL_ERROR1("XHHUL_pdm_get_hh_pd: ERROR (%d:%s) : Insertion failure.\n",
				   vret, VIP_common_strerror_sym(vret) );
        return HH_EAGAIN;
    }

    *hh_pd = (HHUL_pd_hndl_t) (MT_virt_addr_t)(vip_obj);
    MTL_DEBUG3("<== XHHUL_pdm_get_hh_pd\n");
    return HH_OK;
}

HH_ret_t XHHUL_pdm_get_hh_pd_hold_ul_pd (
   XHHUL_pdm_t  pdm,
   HHUL_pd_hndl_t hhul_pd, 
   HH_pd_hndl_t *hh_pd)
{
    VIP_common_ret_t vret;
    VIP_array_obj_t  vip_obj;

    MTL_DEBUG3("==> XHHUL_pdm_get_hh_pd_hold_ul_pd, hhul_pd=%lu\n", hhul_pd);


    if ((vret=VIP_array_find_hold(pdm->pd_array, (VIP_array_handle_t)hhul_pd, &vip_obj)) != VIP_OK) {
        MTL_ERROR1("XHHUL_pdm_get_hh_pd_hold_ul_pd: ERROR (%d:%s) : Insertion failure.\n",
				   vret, VIP_common_strerror_sym(vret) );
        return HH_EAGAIN;
    }

    *hh_pd = (HHUL_pd_hndl_t) (MT_virt_addr_t)(vip_obj);
    MTL_DEBUG3("<== XHHUL_pdm_get_hh_pd_hold_ul_pd\n");
    return HH_OK;
}

HH_ret_t XHHUL_pdm_rel_ul_pd(
   XHHUL_pdm_t  pdm,
   HHUL_pd_hndl_t hhul_pd)
{
    VIP_common_ret_t vret;

    MTL_DEBUG3("==> XHHUL_pdm_rel_ul_pd, hhul_pd=%lu\n", hhul_pd);


    if ((vret=VIP_array_find_release(pdm->pd_array, (VIP_array_handle_t)hhul_pd)) != VIP_OK) {
        MTL_ERROR1("XHHUL_pdm_rel_ul_pd: ERROR (%d:%s) : find_release failure.\n",
				   vret, VIP_common_strerror_sym(vret) );
        return HH_EAGAIN;
    }

    MTL_DEBUG3("<== XHHUL_pdm_rel_ul_pd\n");
    return HH_OK;
}

