/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#define C_XHH_SRQM_C

#include <vapi_common.h>
#include <vip_common.h>
#include <vip_array.h>
#include <hob.h>
#include <cmdif.h>
#include <tptm.h>

#include <srqm.h>
#include <srqm_priv.h>
#include "srqm_dev_inline.h"

/************************************************************************/
/*                      Private functions                               */
/************************************************************************/

void free_srq_context(void* srq_context)
{
  MTL_ERROR1(MT_FLFMT("XHH_srqm_destroy: Garbage collection: Releasing SRQ #%u"),
             ((XHH_srq_t)srq_context)->lkey);
  /* Internal memory regions are cleaned-up on XHH_mrwm_destroy */
  FREE(srq_context);
}


/************************************************************************/
/*                      Public functions                                */
/************************************************************************/

HH_ret_t  XHH_srqm_create(
  XHH_hob_t              hob,           /* IN  */
  XHH_srqm_init_t*       p_init_params,
  XHH_srqm_t*            srqm_p         /* OUT */
)
{
  HH_ret_t ret;
  VIP_common_ret_t vret;
  u_int32_t rsvd_srq= 1 << p_init_params->log2_rsvd_srq;
  u_int32_t max_srq= (1 << p_init_params->log2_max_srq) - rsvd_srq;
  u_int32_t initial_array_sz= max_srq > 1024 ? 1024 : max_srq;
  MTL_DEBUG1(MT_FLFMT("%s: Invoked with log2_max_srq=0x%u log2_rsrv_srq=0x%u srqm_p=0x%p"), 
             __func__, p_init_params->log2_max_srq, p_init_params->log2_rsvd_srq, srqm_p);
  
  *srqm_p= MALLOC(sizeof(struct XHH_srqm_st));
  if (*srqm_p == NULL)  return HH_EAGAIN;

  XHH_DEV_SPECIFIC_SRQM_INIT

  vret= VIP_array_create_maxsize(initial_array_sz, max_srq, &(*srqm_p)->srq_array);
  if (vret != VIP_OK) {
    MTL_ERROR2(MT_FLFMT("%s: Failed VIP_array_create_maxsize (%u-%s)"), __func__, 
               vret, VAPI_strerror_sym(vret));
    ret= HH_EAGAIN;
    goto vip_array_create_failed;
  }

  ret= XHH_hob_get_cmd_if(hob, &(*srqm_p)->cmdif);
  if (ret != HH_OK) {
    MTL_ERROR2(MT_FLFMT("%s: Failed XHH_hob_get_cmd_if (%s)"), __func__, HH_strerror_sym(ret));
    goto get_failed;
  }

  ret= XHH_hob_get_mrwm(hob, &(*srqm_p)->mrwm);
  if (ret != HH_OK) {
    MTL_ERROR2(MT_FLFMT("%s: Failed XHH_hob_get_mrwm (%s)"), __func__, HH_strerror_sym(ret));
    goto get_failed;
  }

  XHH_DEV_SPECIFIC_SRQM_GET_PROT_MODULE  
  
  (*srqm_p)->hob= hob;
  (*srqm_p)->max_srq= max_srq;
  (*srqm_p)->rsvd_srq= rsvd_srq;
  
  return HH_OK;

  get_failed:
    VIP_array_destroy((*srqm_p)->srq_array, NULL);
  vip_array_create_failed:
    FREE(*srqm_p);
    return ret;
}


HH_ret_t  XHH_srqm_destroy(
  XHH_srqm_t  srqm        /* IN */
)
{
  VIP_common_ret_t vret;
  
  if (srqm == (XHH_srqm_t)XHH_INVALID_HNDL)  {
    MTL_ERROR1(MT_FLFMT("%s: Invoked for XHH_INVALID_HNDL"), __func__);
    return HH_EINVAL;
  }
  MTL_DEBUG1(MT_FLFMT("%s: Releasing SRQM handle 0x%p"), __func__, srqm);

  /* In case of abnormal HCA termination we may still have unreleased SRQ resources */
  vret= VIP_array_destroy(srqm->srq_array, free_srq_context);
  if (vret != VIP_OK) {
    MTL_ERROR2(MT_FLFMT("%s: Failed VIP_array_destroy (%u-%s) - completing SRQM destroy anyway"),
               __func__, vret, VAPI_strerror_sym(vret));
  }

  XHH_DEV_SPECIFIC_SRQM_DONE
  
  FREE(srqm);

  return HH_OK;
}


HH_ret_t  XHH_srqm_create_srq(
  XHH_srqm_t         srqm,                    /* IN */
  HH_pd_hndl_t       pd,                      /* IN */
  XHH_srq_ul_resources_t *srq_ul_resources_p, /* IO  */
  HH_srq_hndl_t     *srq_p                    /* OUT */
)
{
  HH_ret_t ret;
  XHH_cmd_status_t cmd_ret;
  VIP_common_ret_t vret;
  VIP_array_handle_t vip_hndl;
  XHH_srq_t srq;
  XHH_internal_mr_t *mr_props_p = NULL;
  MOSAL_prot_ctx_t vm_ctx;
  u_int32_t srqn;
  XHH_srq_context_t xhh_srqc;

  if (srq_ul_resources_p->wqes_buf == 0) {
    MTL_ERROR1(MT_FLFMT("%s: Got wqes_buf=NULL. WQEs in DDR-mem are not supported, yet."), 
               __func__);
    return HH_ENOSYS;
  }

  ret = XHH_get_protection_ctx(srqm->prot_module, pd, &vm_ctx);
  if (ret != HH_OK) {
    MTL_ERROR2(MT_FLFMT("%s: Failed XHH_uldm_get_protection_ctx (%s)"), __func__, 
               mtl_strerror_sym(ret));
    return ret;
  }

  srq= MALLOC(sizeof(struct XHH_srq_st));
  if (srq == NULL) {
    return HH_EAGAIN;
  }

  vret= VIP_array_insert(srqm->srq_array, srq, &vip_hndl);
  if (vret != VIP_OK) {
    MTL_ERROR2(MT_FLFMT("%s: Failed VIP_array_insert (%u)"), __func__, vret);
    ret= HH_EAGAIN;
    goto vip_array_insert_failed;
  }


  mr_props_p = (XHH_internal_mr_t *)MALLOC(sizeof(XHH_internal_mr_t));
  if (mr_props_p == NULL) {
     MT_COV_MEM_FLTR_START
     ret = HH_EAGAIN;
     goto register_internal_failed;
     MT_COV_MEM_FLTR_END
  }
  memset(mr_props_p, 0, sizeof(XHH_internal_mr_t));
  
  mr_props_p->start= srq_ul_resources_p->wqes_buf;
  mr_props_p->size= srq_ul_resources_p->wqes_buf_sz;
  mr_props_p->pd= pd;
  mr_props_p->vm_ctx= vm_ctx;
  mr_props_p->force_memkey = FALSE;
  ret= XHH_mrwm_register_internal(srqm->mrwm, mr_props_p, &srq->lkey);
  if (ret != HH_OK) {
    MTL_ERROR2(MT_FLFMT("%s: Failed XHH_mrwm_register_internal (%s)"), __func__, 
               HH_strerror_sym(ret));
    FREE(mr_props_p);
    goto register_internal_failed;
  }

  FREE(mr_props_p);
  srqn= vip_hndl + srqm->rsvd_srq;

  XHH_DEV_SPECIFIC_SRQM_CREATE_SRQ
  xhh_srqc.lwm = 0; 
  xhh_srqc.pd= pd;
  xhh_srqc.l_key= srq->lkey;
  xhh_srqc.wqe_addr_h= /* Upper 32b of WQEs buffer */
    (u_int32_t)((sizeof(MT_virt_addr_t) > 4) ? (srq_ul_resources_p->wqes_buf >> 32) : 0);
  cmd_ret= XHH_cmd_SW2HW_SRQ(srqm->cmdif, srqn, &xhh_srqc);
  if (cmd_ret != XHH_CMD_STAT_OK) {
    MTL_ERROR2(MT_FLFMT("%s: Failed XHH_cmd_SW2HW_SRQ for srqn=0x%X (%s)"), __func__, 
               srqn, str_XHH_cmd_status_t(cmd_ret));
    ret= HH_EFATAL; /* Unexpected error */
    goto sw2hw_failed;
  }

  *srq_p= srqn;
  MTL_DEBUG4(MT_FLFMT("%s: Allocated SRQn=0x%X"), __func__, srqn); 
  
  return HH_OK;

  sw2hw_failed:
    XHH_DEV_SPECIFIC_SRQM_DESTROY_SRQ

  XHH_DEV_SPECIFIC_SRQM_CREATE_SRQ_FAIL_LABEL
    XHH_mrwm_deregister_mr(srqm->mrwm, srq->lkey);
  register_internal_failed:
    VIP_array_erase(srqm->srq_array, vip_hndl, NULL);
  vip_array_insert_failed:
    FREE(srq);
    return ret;
}

HH_ret_t  XHH_srqm_destroy_srq(
  XHH_srqm_t         srqm,                    /* IN */
  HH_srq_hndl_t      srqn                     /* IN */
)
{
  HH_ret_t ret;
  XHH_cmd_status_t cmd_ret;
  VIP_common_ret_t vret;
  VIP_array_obj_t vip_obj;
  XHH_srq_t srq;
  MT_bool have_fatal= FALSE;

  vret= VIP_array_erase_prepare(srqm->srq_array, srqn - srqm->rsvd_srq, &vip_obj);
  if (vret != HH_OK) {
    MTL_ERROR2(MT_FLFMT("%s: Failed VIP_array_erase_prepare (%u)"), __func__, vret);
    return HH_EINVAL;
  }
  srq= (XHH_srq_t)vip_obj;

  cmd_ret= XHH_cmd_HW2SW_SRQ(srqm->cmdif, srqn, NULL);
  if (cmd_ret != XHH_CMD_STAT_OK) {
    MTL_ERROR1(MT_FLFMT("%s: Failed XHH_cmd_SW2HW_SRQ for srqn=0x%X (%s)"), __func__, 
               srqn, str_XHH_cmd_status_t(cmd_ret));
    have_fatal= TRUE; /* Unexpected error */
  } else {
    ret= XHH_mrwm_deregister_mr(srqm->mrwm, srq->lkey);
    if (ret != HH_OK) {
      MTL_ERROR2(MT_FLFMT("%s: Failed XHH_mrwm_deregister_mr (%s)"), __func__,
                 HH_strerror_sym(ret));
      have_fatal= TRUE;
    }
  }

  if (!have_fatal) {
    vret= VIP_array_erase_done(srqm->srq_array, srqn - srqm->rsvd_srq, NULL);
    if (vret != HH_OK) {
      MTL_ERROR2(MT_FLFMT("%s: Failed VIP_array_erase_done (%u)"), __func__, vret);
      have_fatal= TRUE;
    } else {
        XHH_DEV_SPECIFIC_SRQM_DESTROY_SRQ
        FREE(srq);
    }

  } else {
    VIP_array_erase_undo(srqm->srq_array, srqn - srqm->rsvd_srq); 
    /* Leave for srqm_destroy cleanup */
  }

  
  return HH_OK; /* resource cleanup is always OK - even if fatal */
}


HH_ret_t  XHH_srqm_query_srq(
  XHH_srqm_t         srqm,                    /* IN */
  HH_srq_hndl_t      srq,                     /* IN */
  u_int32_t          *limit_p                 /* OUT */
)
{
  VIP_common_ret_t vret;
  HH_ret_t  rc = HH_OK;
  XHH_srq_context_t xhh_srqc;
  XHH_cmd_status_t cmd_ret;
  
  vret= VIP_array_find_hold(srqm->srq_array, srq - srqm->rsvd_srq, NULL);
  if (vret != HH_OK) {
    return HH_EINVAL_SRQ_HNDL;
  }

  cmd_ret= XHH_cmd_QUERY_SRQ(srqm->cmdif, srq, &xhh_srqc); 
  if (cmd_ret != XHH_CMD_STAT_OK) {
         MTL_ERROR2(MT_FLFMT("%s: Failed XHH_cmd_QUERY_SRQ for srqn=0x%X (%s)"), __func__,
		       srq, str_XHH_cmd_status_t(cmd_ret));
	 rc = HH_ERR; /* Unexpected error */
  } else {
     *limit_p = xhh_srqc.lwm;							       \
	 if (xhh_srqc.state == ADD_DEV_PREFIX(IF_SRQ_STATE_ERROR)) {
	      rc = HH_ESRQ;
	  }
  }

  VIP_array_find_release(srqm->srq_array, srq - srqm->rsvd_srq);
  return rc;
}

HH_ret_t  XHH_srqm_modify_srq(
  XHH_srqm_t         srqm,                    /* IN */
  HH_srq_hndl_t      srqn,                     /* IN */
  XHH_srq_ul_resources_t *srq_ul_resources_p  /* IO */
)
{
    HH_ret_t ret = HH_OK;
    VIP_common_ret_t vret;
    VIP_array_obj_t vip_obj;
    XHH_srq_t srq_obj;
    XHH_internal_mr_t mr_props;
    XHH_mrwm_modify_flags_t flags = XHH_MRWM_MODIFY_FLAGS_TRANSLATION;

    MTL_TRACE1(MT_FLFMT("%s: SRQn=0x%X"), __func__, srqn);

    if (srq_ul_resources_p->srq_attr_mask == 0) {
      /* Nothing to do */
      return HH_OK;
    }

    if (srq_ul_resources_p->srq_attr_mask & VAPI_SRQ_ATTR_MAX_OUTS_WR) {
	if ((srq_ul_resources_p->wqes_buf == 0) && (srq_ul_resources_p->wqes_buf_sz != 0))  {
	  MTL_ERROR1(MT_FLFMT("%s: Got wqes_buf=NULL. WQEs in DDR-mem are not supported, yet."), 
		     __func__);
	  return HH_ENOSYS;
	}
    }
    
    vret= VIP_array_find_hold(srqm->srq_array, srqn - srqm->rsvd_srq, &vip_obj);
    if (vret != HH_OK) {
      return HH_EINVAL_SRQ_HNDL;
    }
    srq_obj= (XHH_srq_t)vip_obj;
    do {
	if (srq_ul_resources_p->srq_attr_mask & VAPI_SRQ_ATTR_MAX_OUTS_WR) {
	    if (srq_ul_resources_p->wqes_buf_sz == 0) {
	      /* "Shrinking" SRQ succeeds with NOP */
	      break;
	    }

	    memset(&mr_props, 0, sizeof(mr_props));
	    mr_props.start= srq_ul_resources_p->wqes_buf;
	    mr_props.size= srq_ul_resources_p->wqes_buf_sz;
	
	    ret= XHH_mrwm_modify_internal(srqm->mrwm, srq_obj->lkey, &mr_props, flags);
	    if (ret != HH_OK) {
	      MTL_ERROR2(MT_FLFMT("%s: Failed XHH_mrwm_modify_internal (%d: %s), start="VIRT_ADDR_FMT", size="SIZE_T_DFMT),
			  __func__, ret, HH_strerror_sym(ret), srq_ul_resources_p->wqes_buf, srq_ul_resources_p->wqes_buf_sz);
	      VIP_array_find_release(srqm->srq_array, srqn - srqm->rsvd_srq);
	      return ret;
	    } else {
	      MTL_DEBUG4(MT_FLFMT("%s: modified SRQn=0x%X. start="VIRT_ADDR_FMT", new size="SIZE_T_DFMT), 
			 __func__, srqn, srq_ul_resources_p->wqes_buf, srq_ul_resources_p->wqes_buf_sz); 
	    }
	}
    } while (0);

    if (srq_ul_resources_p->srq_attr_mask & VAPI_SRQ_ATTR_LIMIT) {
        XHH_cmd_status_t cmd_ret;
        cmd_ret = XHH_cmd_ARM_SRQ(srqm->cmdif, srqn,srq_ul_resources_p->srq_limit);
        if (cmd_ret != XHH_CMD_STAT_OK) {
           MTL_ERROR1(MT_FLFMT("%s: Failed XHH_cmd_ARM_SRQ for srqn=0x%X (%s)"), __func__,
      		       srqn, str_XHH_cmd_status_t(cmd_ret));
           ret = HH_ERR; /* Unexpected error */
        } else {
           ret= HH_OK;
        }
	}
    
    VIP_array_find_release(srqm->srq_array, srqn - srqm->rsvd_srq);
    return ret;
}



