/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <mosal.h>
#include <MT23108.h>
#include <udavm.h>
#include <vip_array.h>
#include <vip_common.h>
#include <dev_common.h>
#include <udav_comm.h>

/*================ macro definitions ===============================================*/
/* check that ud_av_p is in the av table range and that it is aligned to entry size */
#define IS_INVALID_AV_MEMBER(udavm, ud_av_p) \
  (((ud_av_p) < udavm->ud_av_table) || \
   (((MT_virt_addr_t)ud_av_p) >= ((MT_virt_addr_t)(udavm->ud_av_table))+(udavm->ud_av_table_sz))  )
#define IS_INVALID_AV_ALIGN(ud_av_p) \
   (((MT_virt_addr_t)(ud_av_p)) & UD_AV_ALIGNMENT_MASK)

#define HANDLE_2_INDEX(udavm, ah) (((MT_ulong_ptr_t)ah - ((MT_ulong_ptr_t)(udavm)->ud_av_table)) >> UD_AV_ENTRY_SIZE_LOG2)

/*================ type definitions ================================================*/


/* The main UDAV-manager structure */
struct XHH_udavm_st {
  u_int32_t*       ud_av_table; /* the actual table of UDAV*/
  u_int32_t        max_av;  /* number of entries in table */
  MT_size_t        ud_av_table_sz; /* total table size in bytes */
  VAPI_lkey_t      table_memkey; 
  VIP_array_p_t    udavs_array;       /* the array which holds the free list of AV entries */
  MOSAL_spinlock_t table_spinlock;     /* protect on the table from destroying UDAV and modifying at the same time */
}XHH_udavm_int_t;



/*================ all common static functions & macros  ============================*/

/*================ global functions definitions ====================================*/


HH_ret_t XHH_udavm_create( /*IN */ XHH_ver_info_t *version_p, 
                           /*IN */ VAPI_lkey_t ud_av_table_memkey, 
                           /*IN */ MT_virt_addr_t ud_av_table, 
                           /*IN */ MT_size_t ud_av_table_sz,
                           /*OUT*/ XHH_udavm_t *udavm_p)

{
  XHH_udavm_t new_udavm_p = NULL;
  HH_ret_t ret;
  
  FUNC_IN;

  /* allocation of object structure */
  new_udavm_p = (XHH_udavm_t)MALLOC(sizeof(XHH_udavm_int_t));
  if (!new_udavm_p) {
	MT_COV_MEM_FLTR_START
    MTL_ERROR4("%s: Cannot allocate UDAVM object.\n", __func__);
    MT_RETURN( HH_EAGAIN);
	MT_COV_MEM_FLTR_END
  }

  memset(new_udavm_p,0,sizeof(XHH_udavm_int_t));
  
  /* filling the of UDAV struct */
  if (MOSAL_spinlock_init(&(new_udavm_p->table_spinlock)) != MT_OK){
    MT_COV_ERR_FLTR_START
    MTL_ERROR4("%s: Failed to initializing spinlocks.\n", __func__);
    ret= HH_ERR;
    goto err_free_mem;
	MT_COV_ERR_FLTR_END
  } 
  new_udavm_p->ud_av_table = (u_int32_t*)ud_av_table;
  memset(new_udavm_p->ud_av_table,0,ud_av_table_sz);
  new_udavm_p->max_av = (u_int32_t)(ud_av_table_sz/UD_AV_ENTRY_SIZE);
  new_udavm_p->ud_av_table_sz = ud_av_table_sz;
  new_udavm_p->table_memkey = ud_av_table_memkey;
  /* init the free list */
  ret = VIP_array_create_maxsize(new_udavm_p->max_av,new_udavm_p->max_av,&(new_udavm_p->udavs_array));
  if ( ret != VIP_OK ) {
    MT_COV_ERR_FLTR_START
    MTL_ERROR1("%s: VIP_array_create_maxsize failed, ret=%d \n", __func__, ret);
    goto err_free_mem;
	MT_COV_ERR_FLTR_END
  }

  /* succeeded to create object - return params: */

  *udavm_p = new_udavm_p;

  MT_RETURN(HH_OK);

  /* error handling cleanup */
err_free_mem:
  MT_COV_ERR_FLTR_START
  FREE(new_udavm_p);
  MT_RETURN(ret);
  MT_COV_ERR_FLTR_END

}


/************************************************************************************/

HH_ret_t XHH_udavm_destroy( /*IN */ XHH_udavm_t udavm )
{
  
  FUNC_IN;
  if (udavm == NULL) {
    MT_COV_ERR_FLTR_START
    MTL_ERROR4("%s: udavm is NULL.\n", __func__);
    MT_RETURN(HH_EINVAL);
	MT_COV_ERR_FLTR_END
  }

  /* destroy the handlers array */
  if (VIP_array_destroy(udavm->udavs_array, NULL) != VIP_OK) {
    MT_COV_ERR_FLTR_START
    MTL_ERROR1("%s: VIP_array_destroy failed \n", __func__);
	MT_COV_ERR_FLTR_END
  }
  FREE(udavm);
  MT_RETURN(HH_OK);
}


/************************************************************************************/
HH_ret_t XHH_udavm_create_av( /*IN */ XHH_udavm_t udavm, 
                              /*IN */ HH_pd_hndl_t pd, 
                              /*IN */ VAPI_ud_av_t *av_p, 
                              /*OUT*/ HH_ud_av_hndl_t *ah_p)
{
  u_int32_t* av_entry_p;
  u_int32_t ah_index=0;
  VIP_common_ret_t ret;
  
  FUNC_IN;
  
  XHH_UDAVM_VALID_CHECKS(udavm, av_p);
  XHH_UDAVM_AH_NULL(ah_p);
   
  /* get a free index for the udav */
  ret = VIP_array_insert(udavm->udavs_array, NULL, &ah_index);
  if (ret != VIP_OK) {
    MTL_ERROR4("XHH_udavm_create_av: Not enough resources.\n");
    MT_RETURN(HH_EAGAIN);
  }

  av_entry_p = (u_int32_t*) (((MT_virt_addr_t)((udavm)->ud_av_table)) + 
                             ((ah_index)*UD_AV_ENTRY_SIZE)); 

  /* filling the entry */

  XHH_udavm_fill_udav_entry(pd, TRUE, av_p, av_entry_p,TRUE,TRUE);
  
  *ah_p = (HH_ud_av_hndl_t)av_entry_p; 
  MTL_DEBUG4(MT_FLFMT("Allocated address handle = " MT_ULONG_PTR_FMT ", entry index=%d"),*ah_p, ah_index);
  MT_RETURN(HH_OK);
}

/************************************************************************************/
HH_ret_t XHH_udavm_modify_av( /*IN */ XHH_udavm_t udavm, 
                              /*IN */ HH_ud_av_hndl_t ah, 
                              /*IN */ VAPI_ud_av_t *av_p )
{
  u_int32_t *ud_av_p= (u_int32_t*)(MT_virt_addr_t)ah;
  HH_ret_t hh_ret;
  u_int32_t ah_index;
  VIP_common_ret_t ret=VIP_OK;

  FUNC_IN;
  
  XHH_UDAVM_VALID_CHECKS(udavm, av_p);

  /* Check that ah within the table and aligned to entry beginning */
  if (IS_INVALID_AV_ALIGN(ud_av_p)) {
      MTL_ERROR4("XHH_udavm_modify_av: invalid av alignment.\n");
      MT_RETURN(HH_EINVAL);
  }
  if (IS_INVALID_AV_MEMBER(udavm, ud_av_p)) {
    MTL_DEBUG4("XHH_udavm_modify_av: invalid ah (" MT_ULONG_PTR_FMT ").\n",ah);
    MT_RETURN(HH_EINVAL_AV_HNDL);
  }
  
  /* check that ah is a valid (allocated) handle */
  ah_index = (u_int32_t)HANDLE_2_INDEX(udavm, ud_av_p); 
  ret = VIP_array_find_hold(udavm->udavs_array, ah_index, NULL);
  MTL_DEBUG3(MT_FLFMT("Calling VIP_array_find_hold, ah_index=%d, ret=%d\n"), ah_index, ret);
  if ( ret!=VIP_OK ) {
	MT_COV_ERR_FLTR_START
    MTL_ERROR4("XHH_udavm_modify_av: handle is not valid.\n");
    hh_ret = HH_EINVAL_AV_HNDL;
	MT_COV_ERR_FLTR_END
  } else {
      MOSAL_spinlock_irq_lock(&(udavm->table_spinlock));
      XHH_udavm_fill_udav_entry(0, FALSE, av_p, ud_av_p,TRUE,TRUE);
      hh_ret = HH_OK;
      MOSAL_spinlock_unlock(&(udavm->table_spinlock));
  }
  ret = VIP_array_find_release(udavm->udavs_array, ah_index);
  MTL_DEBUG3(MT_FLFMT("Calling VIP_array_find_release ret=%d"), ret);
  if ( ret!=VIP_OK ) {
	MT_COV_ERR_FLTR_START
    MTL_ERROR1("%s: Internal mismatch - hv_index (%d) is not in array\n", __func__,ah_index);
    hh_ret = ret;
	MT_COV_ERR_FLTR_END
  }
  MTL_DEBUG4(MT_FLFMT("XHH_udavm_modify_av: address handle = " MT_ULONG_PTR_FMT ", entry index=%d"),ah, ah_index);
  MT_RETURN(hh_ret);
}

HH_ret_t XHH_udavm_query_av_internal(/*IN */ XHH_udavm_t udavm, 
                                     /*IN */ HH_ud_av_hndl_t ah, 
                                     /*OUT*/ VAPI_ud_av_t *av_p)
{
	return XHH_udavm_parse_udav_entry((u_int32_t*)ah, av_p);
}



/************************************************************************************/
HH_ret_t XHH_udavm_query_av( /*IN */ XHH_udavm_t udavm, 
                             /*IN */ HH_ud_av_hndl_t ah, 
                             /*OUT*/ VAPI_ud_av_t *av_p )
{
  HH_ret_t hh_ret;
  u_int32_t ah_index;
  VIP_common_ret_t ret=VIP_OK;
  u_int32_t *ud_av_p= (u_int32_t*)(MT_virt_addr_t)ah;
 
  FUNC_IN;
  
  XHH_UDAVM_BASIC_VALID_CHECKS(udavm, av_p);

  /* Check that ah within the table and aligned to entry beginning */
  if (IS_INVALID_AV_ALIGN(ud_av_p)) {
      MTL_ERROR4("XHH_udavm_query_av: invalid av alignment.\n");
      MT_RETURN(HH_EINVAL);
  }
  if (IS_INVALID_AV_MEMBER(udavm, ud_av_p)) {
    MTL_DEBUG4("XHH_udavm_query_av: invalid ah (" MT_ULONG_PTR_FMT ").\n",ah);
    MT_RETURN(HH_EINVAL_AV_HNDL);
  }

  /* check that ah is a valid (allocated) handle */
  ah_index = (u_int32_t)HANDLE_2_INDEX(udavm, ud_av_p); 
  ret = VIP_array_find_hold(udavm->udavs_array, ah_index, NULL);
  MTL_DEBUG3(MT_FLFMT("Calling VIP_array_find_hold, hv_index=%d, ret=%d\n"), ah_index, ret);
  if ( ret!=VIP_OK ) {
	MT_COV_ERR_FLTR_START
    MTL_ERROR4("XHH_udavm_query_av: handle is not valid.\n");
    hh_ret = HH_EINVAL_AV_HNDL;
	MT_COV_ERR_FLTR_END
  } else { /* DO IT */
    MOSAL_spinlock_irq_lock(&(udavm->table_spinlock));
    hh_ret = XHH_udavm_parse_udav_entry(ud_av_p, av_p);
    MOSAL_spinlock_unlock(&(udavm->table_spinlock));
  }
  ret = VIP_array_find_release(udavm->udavs_array, ah_index);
  MTL_DEBUG3(MT_FLFMT("Calling VIP_array_find_release ret=%d"), ret);
  if ( ret!=VIP_OK ) {
	MT_COV_ERR_FLTR_START
    MTL_ERROR1("%s: Internal mismatch - hv_index (%d) is not in array\n", __func__, ah_index);
    hh_ret = ret;
	MT_COV_ERR_FLTR_END
  }

  MTL_DEBUG4(MT_FLFMT("XHH_udavm_query_av: address handle = " MT_ULONG_PTR_FMT ", entry index=%d"),ah, ah_index);
#if 1 <= MAX_DEBUG
  print_udav(av_p);
#endif
  MT_RETURN(hh_ret);
  
}

/************************************************************************************/
HH_ret_t XHH_udavm_destroy_av( /*IN */ XHH_udavm_t udavm, 
                               /*IN */ HH_ud_av_hndl_t ah )
{
  u_int32_t ah_index;
  VIP_common_ret_t ret=VIP_OK;
  u_int32_t *ud_av_p= (u_int32_t*)(MT_virt_addr_t)ah;
  HH_ret_t hh_ret = HH_OK;
  
  FUNC_IN;
  
  if (udavm == NULL) {
	MT_COV_ERR_FLTR_START
    MTL_ERROR4("XHH_udavm_destroy_av: udavm is NULL.\n");
    MT_RETURN(HH_EINVAL);
	MT_COV_ERR_FLTR_END
  }

  /* Check that given ah is within the table and aligned to entry size */
  if (IS_INVALID_AV_ALIGN(ud_av_p)) {
      MTL_ERROR4("XHH_udavm_destroy_av: invalid av alignment.\n");
      MT_RETURN(HH_EINVAL);
  }
  if (IS_INVALID_AV_MEMBER(udavm, ud_av_p)) {
    MTL_DEBUG4("XHH_udavm_destroy_av: invalid ah (" MT_ULONG_PTR_FMT ").\n",ah);
    MT_RETURN(HH_EINVAL_AV_HNDL);
  }
  
  /* check that ah is a valid (allocated) handle and release it */
  ah_index = (u_int32_t)HANDLE_2_INDEX(udavm, ud_av_p); 
  ret = VIP_array_erase_prepare(udavm->udavs_array, ah_index, NULL);
  MTL_DEBUG3(MT_FLFMT("Calling VIP_array_erase_prepare, hv_index=%d, ret=%d\n"), ah_index, ret);
  if ( ret==VIP_OK ) { /* DO IT */
    /* Use the knowledge that PD is in the first Dword of UDAV entry - so only put zeros there */
    ud_av_p[0] = 0;
    ret = VIP_array_erase_done(udavm->udavs_array, ah_index, NULL);
    if ( ret!=VIP_OK ) { 
	  MT_COV_ERR_FLTR_START
      MTL_ERROR4("XHH_udavm_destroy_av: internal error VIP_array_erase_done failed.\n");
      /* returning OK, since the prepare succeeded, making the AV handle unavailable.*/
	  MT_COV_ERR_FLTR_END
    }
  } else if ( ret==VIP_EBUSY ) {
    MTL_ERROR4("XHH_udavm_destroy_av: handle is busy (in modify or query).\n");
    hh_ret = HH_EBUSY;
  } else if (ret == VIP_EINVAL_HNDL) {
    MTL_ERROR4("XHH_udavm_destroy_av: Invalid handle.\n");
    hh_ret = HH_EINVAL_AV_HNDL;
  } 

  MTL_DEBUG4(MT_FLFMT("XHH_udavm_destroy_av: address handle = " MT_ULONG_PTR_FMT ", entry index=%d"),ah, ah_index);
  MT_RETURN(hh_ret);
}

