/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <uldm_priv.h>

/******************************************************************************
 *  Function:     XHH_uldm_create
 *****************************************************************************/
HH_ret_t XHH_uldm_create( /*IN*/  XHH_hob_t    hob, 
                          /*IN*/  MT_phys_addr_t  uar_base, 
                          /*IN*/  u_int8_t     log2_max_uar, 
                          /*IN*/  u_int8_t     log2_uar_pg_sz,
                          /*IN*/  u_int32_t    max_pd, 
                          /*OUT*/ XHH_uldm_t   *uldm_p )
{
    u_int32_t   max_uar;
    XHH_uldm_t  new_uldm_obj;

    MTL_DEBUG4("ENTERING XHH_uldm_create: uar_base = " PHYS_ADDR_FMT ", log2_max_uar = %d\nlog2_uar_pg_sz=%d, max_pd=%d\n",
               uar_base, log2_max_uar, log2_uar_pg_sz, max_pd);
    /* create new uldm object */
    new_uldm_obj = (XHH_uldm_t) MALLOC(sizeof(XHH_uldm_obj_t));
    if (!new_uldm_obj) {
        MT_COV_MEM_FLTR_START
        MTL_ERROR1("XHH_uldm_create: MALLOC of new_uldm_obj failed\n");
        return HH_ENOMEM;
        MT_COV_MEM_FLTR_END
    }

    max_uar = (1 << log2_max_uar);
    new_uldm_obj->max_uar = max_uar;


    /* create UAR free pool */
    new_uldm_obj->uldm_uar_table = (XHH_uldm_uar_entry_t *) VMALLOC(max_uar * sizeof(XHH_uldm_uar_entry_t));
    if (!new_uldm_obj->uldm_uar_table) {
        MT_COV_MEM_FLTR_START
        FREE(new_uldm_obj);
        MTL_ERROR1("XHH_uldm_create: VMALLOC of uldm_uar_table failed\n");
        return HH_ENOMEM;
        MT_COV_MEM_FLTR_END
    }
    MTL_DEBUG4("XHH_uldm_create: allocated uar table: addr = %p, size=%d\n",
                 (void *) (new_uldm_obj->uldm_uar_table), max_uar);

    memset(new_uldm_obj->uldm_uar_table, 0, max_uar * sizeof(XHH_uldm_uar_entry_t));

    MTL_TRACE4("XHH_uldm_create: creating UAR pool\n");
    new_uldm_obj->uar_list.entries = new_uldm_obj->uldm_uar_table;
    new_uldm_obj->uar_list.size = max_uar ;  /* list size is number of ENTRIES */
    new_uldm_obj->uar_list.head = 0;
    new_uldm_obj->uar_list.meta = &(new_uldm_obj->uar_meta);

    new_uldm_obj->uar_meta.entry_size =  sizeof(XHH_uldm_uar_entry_t);
    new_uldm_obj->uar_meta.prev_struct_offset = (MT_ulong_ptr_t) &(((XHH_uldm_uar_entry_t *) (NULL))->u1.epool.prev);
    new_uldm_obj->uar_meta.next_struct_offset = (MT_ulong_ptr_t) &(((XHH_uldm_uar_entry_t *) (NULL))->u1.epool.next);

    MTL_DEBUG4("XHH_uldm_create: calling epool_init: entries = %p, size=%lu, head=%lu\nentry_size=%d, prev_offs=%d, next_offs=%d\n",
                new_uldm_obj->uar_list.entries, new_uldm_obj->uar_list.size, new_uldm_obj->uar_list.head,
                new_uldm_obj->uar_meta.entry_size, new_uldm_obj->uar_meta.prev_struct_offset,
                new_uldm_obj->uar_meta.next_struct_offset);
    epool_init(&(new_uldm_obj->uar_list));

    /* set uar's 0 and 1 to unavailable */
    MTL_TRACE4("XHH_uldm_create: Reserving UARs 0 and 1\n");
    epool_reserve(&(new_uldm_obj->uar_list), 0, 2);

    /* create PD free pool */
    MTL_TRACE4("XHH_uldm_create: creating PD pool\n");
    new_uldm_obj->uldm_pd_table = (XHH_uldm_pd_entry_t *) VMALLOC(max_pd * sizeof(XHH_uldm_pd_entry_t));
    if (!new_uldm_obj->uldm_pd_table) {
        MT_COV_MEM_FLTR_START
        VFREE(new_uldm_obj->uldm_uar_table);
        FREE(new_uldm_obj);
        return HH_ENOMEM;
        MT_COV_MEM_FLTR_END
    }
    memset(new_uldm_obj->uldm_pd_table, 0, max_pd * sizeof(XHH_uldm_pd_entry_t));


    new_uldm_obj->pd_list.entries = new_uldm_obj->uldm_pd_table;
    new_uldm_obj->pd_list.size = max_pd;    /* list size is number of ENTRIES */
    new_uldm_obj->pd_list.head = 0;
    new_uldm_obj->pd_list.meta = &(new_uldm_obj->pd_meta);

    new_uldm_obj->pd_meta.entry_size =  sizeof(XHH_uldm_pd_entry_t);
    new_uldm_obj->pd_meta.prev_struct_offset = (MT_ulong_ptr_t) &(((XHH_uldm_pd_entry_t *) (NULL))->u1.epool.prev);
    new_uldm_obj->pd_meta.next_struct_offset = (MT_ulong_ptr_t) &(((XHH_uldm_pd_entry_t *) (NULL))->u1.epool.next);

    MTL_DEBUG4("XHH_uldm_create: calling epool_init: entries = %p, size=%lu, head=%lu\nentry_size=%d, prev_offs=%d, next_offs=%d\n",
                new_uldm_obj->pd_list.entries, new_uldm_obj->pd_list.size, new_uldm_obj->pd_list.head,
                new_uldm_obj->pd_meta.entry_size, new_uldm_obj->pd_meta.prev_struct_offset,
                new_uldm_obj->pd_meta.next_struct_offset);
    epool_init(&(new_uldm_obj->pd_list));

    /* set pd's 0 and 1 to unavailable */
    MTL_TRACE4("XHH_uldm_create: Reserving PDs 0 and 1\n");
    epool_reserve(&(new_uldm_obj->pd_list), 0, XHH_NUM_RSVD_PD);

    new_uldm_obj->uar_base = uar_base;
    new_uldm_obj->log2_max_uar = log2_max_uar;
    new_uldm_obj->log2_uar_pg_sz = log2_uar_pg_sz;
    new_uldm_obj->max_pd = max_pd;
  
    /* save HOB handle */
    new_uldm_obj->hob = hob;

    /* return the object handle */
    *uldm_p =  new_uldm_obj;

    MTL_DEBUG4("LEAVING XHH_uldm_create - OK\n");

    return HH_OK;
}

/******************************************************************************
 *  Function:     XHH_uldm_destroy
 *****************************************************************************/
HH_ret_t XHH_uldm_destroy( /*IN*/ XHH_uldm_t uldm )
{
    u_int32_t i;

    MTL_DEBUG4("==> XHH_uldm_destroy\n");
    for (i=0; i< uldm->max_uar; i++) {
        if (uldm->uldm_uar_table[i].valid) {
            /* is uar valid check is within this function */
            XHH_uldm_free_uar(uldm,i);
        }
    }
    VFREE(uldm->uldm_uar_table);
    VFREE(uldm->uldm_pd_table);

    FREE(uldm);
    MTL_DEBUG4("<== XHH_uldm_destroy\n");
    return HH_OK;
}

/*************************************************************************
 * Function: XHH_uldm_alloc_ul_res
 *************************************************************************/ 
HH_ret_t XHH_uldm_alloc_ul_res( /*IN*/ XHH_uldm_t              uldm, 
                                /*IN*/ MOSAL_protection_ctx_t  prot_ctx, 
                                /*OUT*/ XHH_hca_ul_resources_t *hca_ul_resources_p )
{
  u_int32_t    uar_index;
  MT_virt_addr_t  uar_map;
  HH_ret_t     ret = HH_OK;

  MTL_DEBUG4("==> XHH_uldm_alloc_ul_res. prot_ctx = 0x%x\n", prot_ctx);

  ret = XHH_uldm_alloc_uar(uldm,prot_ctx, &uar_index, &uar_map);
  if (ret != HH_OK) {
      MTL_ERROR1("%s: failed allocating UAR. ERROR = %d\n", __func__, ret);
      goto uldm_err;
  }

  hca_ul_resources_p->uar_index = uar_index;
  hca_ul_resources_p->uar_map   = uar_map;

uldm_err:
  MTL_DEBUG4("<== XHH_uldm_alloc_ul_res. ret = %d\n", ret);
  return ret;
}

/*************************************************************************
 * Function: XHH_uldm_free_ul_res
 *************************************************************************/ 
HH_ret_t XHH_uldm_free_ul_res( /*IN*/ XHH_uldm_t             uldm, 
                               /*IN*/ XHH_hca_ul_resources_t *hca_ul_resources_p)
{
    HH_ret_t     ret = HH_OK;

    MTL_DEBUG4("==> XHH_uldm_free_ul_res. resources ptr = %p\n", (void *) hca_ul_resources_p);
    if (hca_ul_resources_p->uar_index > 1) {
        ret = XHH_uldm_free_uar(uldm, hca_ul_resources_p->uar_index);
        if (ret != HH_OK) {
            MTL_ERROR1("%s: failed freeing UAR index %d. ERROR = %d\n", __func__, hca_ul_resources_p->uar_index, ret);
            goto uldm_err;
        }
    }

    hca_ul_resources_p->uar_index = 0;
    hca_ul_resources_p->uar_map   = 0;

uldm_err:
    MTL_DEBUG4("<== XHH_uldm_free_ul_res. ret = %d\n", ret);
    return ret;
}

/*************************************************************************
 * Function: XHH_uldm_alloc_uar
 *************************************************************************/
HH_ret_t XHH_uldm_alloc_uar( /*IN*/ XHH_uldm_t              uldm, 
                             /*IN*/ MOSAL_protection_ctx_t  prot_ctx, 
                             /*OUT*/ u_int32_t              *uar_index, 
                             /*OUT*/ MT_virt_addr_t            *uar_map )
{
    unsigned long i;
    HH_ret_t     ret = HH_OK;
    call_result_t  rc;
    
    MTL_DEBUG4("==> XHH_uldm_alloc_uar. prot_ctx = 0x%x\n", prot_ctx);
    
    if (uldm == (XHH_uldm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_uldm_alloc_uar: ERROR : HCA device has not yet been opened\n");
        ret = HH_EINVAL;
        goto uldm_err;
    }

    i = epool_alloc(&(uldm->uar_list)) ;
    if (i == EPOOL_NULL) {
        MTL_ERROR1("XHH_uldm_alloc_uar: ERROR : Could not allocate a UAR from pool\n");
        ret = HH_EINVAL;
        goto uldm_err;
    } else {
        *uar_index = i;
        uldm->uldm_uar_table[i].u1.uar.prot_ctx = prot_ctx;
        uldm->uldm_uar_table[i].valid = TRUE;
        rc = MOSAL_iomem_map(prot_ctx,
                              MOSAL_MEM_FLAGS_NO_CACHE | MOSAL_MEM_FLAGS_PERM_WRITE,
                              (uldm->uar_base + (i * (1 << uldm->log2_uar_pg_sz))),
                              (1 << uldm->log2_uar_pg_sz),
                              &(uldm->uldm_uar_table[i].u1.uar.iomem_obj),
                              uar_map);
                                                     
        if (rc != MT_OK) {
            MTL_ERROR1(MT_FLFMT("%s: MOSAL_iomem_map failed. prot ctx=%d, pa=" PHYS_ADDR_FMT ", size=%d (%d:%s)"), 
                        __func__,prot_ctx,
                       (MT_phys_addr_t) (uldm->uar_base + (i * (1 << uldm->log2_uar_pg_sz))),
                       (1 << uldm->log2_uar_pg_sz),
                        rc, mtl_strerror_sym(rc));
            return HH_EINVAL;
        }
        uldm->uldm_uar_table[i].u1.uar.virt_addr = *uar_map;
        MTL_DEBUG4("XHH_uldm_alloc_uar: index = %d, addr = " VIRT_ADDR_FMT "\n", *uar_index, *uar_map);
        ret = HH_OK;
    }

uldm_err:
    MTL_DEBUG4("<== XHH_uldm_alloc_uar. ret = %d\n", ret);
    return ret;
}

/******************************************************************************
 *  Function:     XHH_uldm_free_uar
 *****************************************************************************/
HH_ret_t XHH_uldm_free_uar( /*IN*/ XHH_uldm_t    uldm, 
                            /*IN*/ u_int32_t     uar_index )
{
    call_result_t  rc;
    HH_ret_t       ret = HH_OK;

    /* check that index is valid */
    MTL_DEBUG4("==> XHH_uldm_free_uar. uar_index = %d\n", uar_index);
    if (uldm == (XHH_uldm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_uldm_free_uar: ERROR : HCA device has not yet been opened\n");
        ret = HH_EINVAL;
        goto err;
    }

    if (uar_index > uldm->max_uar ) {
        MTL_ERROR1("XHH_uldm_free_uar: uar_index out of range (%d)\n", uar_index);
        ret = HH_EINVAL;
        goto err;
    }

    if (!(uldm->uldm_uar_table[uar_index].valid)) {
        MTL_ERROR1("XHH_uldm_free_uar: uar_index was not allocated (%d)\n", uar_index);
        ret = HH_EINVAL;
        goto err;
    }
    /* Unmap previously mapped physical (will be page aligned, of course) */

    rc = MOSAL_iomem_unmap(uldm->uldm_uar_table[uar_index].u1.uar.iomem_obj,
                           uldm->uldm_uar_table[uar_index].u1.uar.prot_ctx);

    if (rc != MT_OK) {
        MTL_ERROR1(MT_FLFMT("XHH_uldm_free_uar: MOSAL_iomem_unmap failed for prot ctx %d, addr %p, size %d (%d:%s)"),
                   uldm->uldm_uar_table[uar_index].u1.uar.prot_ctx,
                   (void *)uldm->uldm_uar_table[uar_index].u1.uar.virt_addr,
                   (1 << uldm->log2_uar_pg_sz),
                   rc, mtl_strerror_sym(rc));
        ret = HH_EINVAL;
        goto err;
    }
    
    uldm->uldm_uar_table[uar_index].valid = FALSE;
    epool_free(&(uldm->uar_list), uar_index);
    
err:
    MTL_DEBUG4("<== XHH_uldm_free_uar. ret = %d\n", ret);
    return ret;

}

/******************************************************************************
 *  Function:     XHH_uldm_alloc_pd
 *****************************************************************************/
HH_ret_t XHH_uldm_alloc_pd( /*IN*/ XHH_uldm_t                   uldm, 
                            /*IN*/ MOSAL_protection_ctx_t       prot_ctx, 
                            /*IN-OUT*/ XHH_pd_ul_resources_t    *pd_ul_resources_p, 
                            /*OUT*/ HH_pd_hndl_t                *pd_p )
{   
    unsigned long      i;
    HH_ret_t           ret = HH_OK;
    XHH_internal_mr_t  *mr_data_p = NULL;
    VAPI_lkey_t        lkey;
    XHH_mrwm_t         mrwm;
    XHH_lamm_t        lamm;
    XHH_udavm_t        udavm;
    MT_size_t          adj_ud_av_table_sz = 0;
    MT_virt_addr_t        udav_virt_addr = 0;
    MT_phys_addr_t        udav_phys_addr = 0;
    VAPI_size_t        udav_phys_buf_size = 0;
    MT_bool            use_priv_udav, av_in_host_mem, hide_ddr;
    u_int32_t          pd_index;
    u_int32_t          max_ah_num = 0, max_requested_avs = 0;
    unsigned int page_size;
    call_result_t rc;
    MOSAL_iomem_t      iomem_obj = (MOSAL_iomem_t)NULL;

    MTL_DEBUG4("==> XHH_uldm_alloc_pd. uldm = %p, prot_ctx = 0x%x, resources_p = %p\n", (void *) uldm, prot_ctx, (void *) pd_ul_resources_p);
    
    if (uldm == (XHH_uldm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("<==XHH_uldm_alloc_pd: ERROR : HCA device has not yet been opened\n");
        return HH_EINVAL;
    }


    ret = XHH_hob_get_udavm_info (uldm->hob, &udavm, &use_priv_udav, &av_in_host_mem, 
                                  &lkey, &max_ah_num, &hide_ddr );
    if (ret != HH_OK) {
        MTL_ERROR1("<==XHH_uldm_alloc_pd: ERROR:  could not acquire udavm information (%d)\n", ret);
        return ret;
    }

    rc = MOSAL_get_page_size(prot_ctx, pd_ul_resources_p->udavm_buf, &page_size);
    if ( rc != MT_OK ) {
      MTL_ERROR1(MT_FLFMT("%s: could not obtain page size of address="VIRT_ADDR_FMT), __func__,
                           pd_ul_resources_p->udavm_buf);
      return HH_ENOMEM;
    }
    
    if (use_priv_udav && (((pd_ul_resources_p->udavm_buf_sz) & (page_size - 1)) != 0)){
        MTL_ERROR1("<==XHH_uldm_alloc_pd: ERROR : udavm_buf_size ("SIZE_T_FMT") is not a multiple of the page size\n",
			pd_ul_resources_p->udavm_buf_sz);
        return HH_EINVAL;
     }

    if (!use_priv_udav) {
        /* check to see that we are not trying to allocate space for more AVs than max reported in HCA capabilities */
        max_requested_avs = (u_int32_t)((pd_ul_resources_p->udavm_buf_sz) / (sizeof(struct tavorprm_ud_address_vector_st) / 8));
        if (max_requested_avs > max_ah_num) {
            MTL_ERROR1("<==XHH_uldm_alloc_pd: max AVs too large: requested %d, max available=%d\n",
                         max_requested_avs, max_ah_num);
            return HH_EINVAL;
        }
    }

    i = epool_alloc(&uldm->pd_list) ;
    if (i == EPOOL_NULL) {
        MTL_ERROR1("<==XHH_uldm_alloc_pd: could not allocate a PD from pool\n");
        return HH_EAGAIN;
    } 


    pd_index = (u_int32_t) i;
    uldm->uldm_pd_table[pd_index].u1.pd.prot_ctx = prot_ctx;
    *pd_p = (HH_pd_hndl_t) i;

    if (use_priv_udav) {
	pd_ul_resources_p->udavm_buf_memkey = lkey;
    } else {
	ret = XHH_hob_get_mrwm(uldm->hob,&mrwm);
	if (ret != HH_OK) {
	    MTL_ERROR1("XHH_uldm_alloc_pd: could not acquire MRWM handle (%d)\n", ret);
	    goto udavm_get_mwrm_err;
	}
	/* Use DDR resources if appropriate flag set at module initialization,
	   or if Tavor firmware indicates that should not hide DDR memory, or if not a PD for a SQP
	 */
	if ((hide_ddr == FALSE) && (av_in_host_mem == FALSE) && (pd_ul_resources_p->pd_flags != PD_FOR_SQP)) {
	    ret = XHH_hob_get_lamm(uldm->hob,&lamm);
	    if (ret != HH_OK) {
		MTL_ERROR1("XHH_uldm_alloc_pd: could not acquire DDRMM handle (%d)\n", ret);
	    } else {

		/* get memory in DDR. If cannot, then use regular memory allocated in  */
		/* call to XHHUL_uldm_alloc_pd_prep */
		adj_ud_av_table_sz =  MT_UP_ALIGNX_SIZE((pd_ul_resources_p->udavm_buf_sz), MOSAL_SYS_PAGE_SHIFT);
		udav_phys_buf_size = adj_ud_av_table_sz;
				    #ifdef MT_DONT_USE_DDR_FOR_AV
		ret = HH_EAGAIN;	/* to cause the fallback to using of system memory for AVS */
				    #else
		ret = XHH_lamm_alloc(lamm, 
				      adj_ud_av_table_sz, 
				      MOSAL_SYS_PAGE_SHIFT,
				      &udav_phys_addr);
				    
				    #endif
		if (ret != HH_OK) {
		   MTL_DEBUG1("XHH_uldm_alloc_pd: could not allocate protected udavm area in DDR(err = %d)\n", ret);
		   udav_phys_addr = (MT_phys_addr_t) 0;
		} else {
		   MTL_DEBUG4("XHH_uldm_alloc_pd. prot_ctx = 0x%x, phys_addr = " PHYS_ADDR_FMT 
			      ", buf_size="SIZE_T_FMT", adj_size="SIZE_T_FMT"\n",
			       prot_ctx,
			       udav_phys_addr, pd_ul_resources_p->udavm_buf_sz,
			       adj_ud_av_table_sz);
		   rc = MOSAL_iomem_map(prot_ctx,
					 MOSAL_MEM_FLAGS_NO_CACHE | MOSAL_MEM_FLAGS_PERM_WRITE | 
					      MOSAL_MEM_FLAGS_PERM_READ,
					 udav_phys_addr,
					 (MT_size_t)udav_phys_buf_size,
					 &(iomem_obj),
					 &udav_virt_addr);
		   MTL_DEBUG4("XHH_uldm_alloc_pd. udav virt_addr = " VIRT_ADDR_FMT "\n", udav_virt_addr);
		   if (rc != MT_OK) {
		       MTL_ERROR1(MT_FLFMT("%s: MOSAL_iomem_map failed. pa=" PHYS_ADDR_FMT ", size="U64_FMT" (%d:%s)"), 
				  __func__,udav_phys_addr, udav_phys_buf_size,
				  rc, mtl_strerror_sym(rc));

		       ret = XHH_lamm_free(lamm,udav_phys_addr,(MT_size_t)udav_phys_buf_size);
		       if (ret != HH_OK) {
			  MTL_ERROR1("XHH_uldm_alloc_pd: could not free protected udavm area in DDR(err = %d)\n", ret);
		       }
		       udav_phys_addr = (MT_phys_addr_t) 0;
		   } else {
		       /* substitute DDR-allocated region for the one passed by pd_ul_resources_p */
		       pd_ul_resources_p->udavm_buf = udav_virt_addr;
		   }
	       }
	    }
	}
	/* check if have a buffer allocated for udav table */
	if (pd_ul_resources_p->udavm_buf == (MT_virt_addr_t) NULL) {
	    MTL_ERROR1("XHH_uldm_alloc_pd: no udavm area allocated.\n");
	    /*value of ret was set above by some failure */
	    goto no_udavm_area;
	}


	mr_data_p = (XHH_internal_mr_t *)MALLOC(sizeof(XHH_internal_mr_t));
	if (mr_data_p == NULL) {
	   MT_COV_MEM_FLTR_START
	   ret = HH_EAGAIN;
	   goto no_udavm_area;
	   MT_COV_MEM_FLTR_END
	}
	memset(mr_data_p, 0, sizeof(XHH_internal_mr_t));
	mr_data_p->force_memkey = FALSE;
	mr_data_p->memkey       = 0;
	mr_data_p->pd           = (HH_pd_hndl_t) pd_index;
	mr_data_p->size         = pd_ul_resources_p->udavm_buf_sz;
	mr_data_p->vm_ctx       = prot_ctx;
	if (udav_phys_addr) {
	    VAPI_phy_addr_t udav_phy = udav_phys_addr;
	    mr_data_p->num_bufs = 1;      /*  != 0   iff   physical buffesrs supplied */
	    mr_data_p->phys_buf_lst = &udav_phy;  /* size = num_bufs */
	    mr_data_p->buf_sz_lst = &udav_phys_buf_size;    /* [num_bufs], corresponds to phys_buf_lst */
	    mr_data_p->start        = pd_ul_resources_p->udavm_buf;
	} else {
	    /* using user-level buffer: check that buffer address is aligned  to entry size; */
	    if (pd_ul_resources_p->udavm_buf != 
		 (MT_UP_ALIGNX_VIRT((pd_ul_resources_p->udavm_buf), 
			      ceil_log2((sizeof(struct tavorprm_ud_address_vector_st) / 8))))) {
		MTL_ERROR1("XHH_uldm_alloc_pd: provided HOST MEM buffer not properly aligned.\n");
		/*value of ret was set above by some failure */
		ret = HH_EINVAL;
		FREE(mr_data_p);
		goto no_udavm_area;
	    }
	    mr_data_p->start = (IB_virt_addr_t)pd_ul_resources_p->udavm_buf;
	    MTL_DEBUG1(MT_FLFMT("%s: User level UDAV tbl = " U64_FMT), __func__, mr_data_p->start);
	}
	uldm->uldm_pd_table[pd_index].valid = TRUE; /* set to valid here, so that XHH_uldm_get_protection_ctx will work */
	ret = XHH_mrwm_register_internal(mrwm, mr_data_p, &lkey);
	if (ret != HH_OK) {
	    MTL_ERROR1("XHH_uldm_alloc_pd: could not register udavm table (%d)\n", ret);
	    uldm->uldm_pd_table[pd_index].valid = FALSE;
	    FREE(mr_data_p);
	    goto udavm_table_register_err;
	}
	FREE(mr_data_p);
	pd_ul_resources_p->udavm_buf_memkey = lkey;
	uldm->uldm_pd_table[pd_index].u1.pd.lkey   = lkey;
	uldm->uldm_pd_table[pd_index].u1.pd.udav_table_ddr_phys_addr   = udav_phys_addr;
	uldm->uldm_pd_table[pd_index].u1.pd.udav_table_ddr_virt_addr   = udav_virt_addr;
	uldm->uldm_pd_table[pd_index].u1.pd.iomem_obj   = iomem_obj;
	uldm->uldm_pd_table[pd_index].u1.pd.udav_table_size   = (MT_size_t)((udav_phys_addr ? udav_phys_buf_size : 0));
	MTL_DEBUG4("XHH_uldm_alloc_pd: PD %u: saving phys addr = " PHYS_ADDR_FMT ", virt addr = " VIRT_ADDR_FMT ", size="SIZE_T_FMT"\n",
		   pd_index, udav_phys_addr, udav_virt_addr, uldm->uldm_pd_table[pd_index].u1.pd.udav_table_size );
    }
    /* set VALID flag only if successful */
    uldm->uldm_pd_table[pd_index].valid = TRUE;
    ret =  HH_OK;
    MTL_DEBUG4("XHH_uldm_alloc_pd. PD = %u\n", pd_index);
    goto ok_retn;

udavm_table_register_err:
if (iomem_obj != (MOSAL_iomem_t) NULL) {
    MOSAL_iomem_unmap(iomem_obj,prot_ctx);
    XHH_lamm_free(lamm,udav_phys_addr,(MT_size_t)udav_phys_buf_size);
}
no_udavm_area:
udavm_get_mwrm_err:
    epool_free(&(uldm->pd_list),i);
ok_retn:
    MTL_DEBUG4("<== XHH_uldm_alloc_pd. ret = %d\n", ret);
    return ret;

}

/******************************************************************************
 *  Function:     XHH_uldm_free_pd 
 *****************************************************************************/
HH_ret_t XHH_uldm_free_pd( /*IN*/ XHH_uldm_t    uldm, 
                           /*IN*/ HH_pd_hndl_t  pd )
{

    HH_ret_t           ret = HH_OK;
    VAPI_lkey_t        lkey;
    XHH_mrwm_t         mrwm;
    XHH_udavm_t        udavm;
    XHH_lamm_t        lamm;
    MT_virt_addr_t        udav_virt_addr = 0;
    MT_phys_addr_t        udav_phys_addr = 0;
    MT_size_t          udav_size = 0;
    MT_bool            use_priv_udav, av_in_host_mem, hide_ddr;
    call_result_t      res;
    u_int32_t          max_ah_num = 0;
    
    MTL_DEBUG4("==> XHH_uldm_free_pd\n");
    if (uldm == (XHH_uldm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_uldm_free_pd: ERROR : HCA device has not yet been opened\n");
        ret = HH_EINVAL;
        goto err_retn;
    }
    
    if (pd > uldm->max_pd ) {
        MTL_ERROR1("XHH_uldm_free_pd: pd out of range (%d)\n", pd);
        ret = HH_EINVAL;
        goto err_retn;
    }

    if (!(uldm->uldm_pd_table[pd].valid)) {
        MTL_ERROR1("XHH_uldm_free_pd: pd was not allocated (%d)\n", pd);
        ret = HH_EINVAL;
        goto err_retn;
    }

    ret = XHH_hob_get_udavm_info (uldm->hob, &udavm, &use_priv_udav, &av_in_host_mem, 
                                  &lkey, &max_ah_num, &hide_ddr );
    if (ret != HH_OK) {
        MTL_ERROR1("XHH_uldm_free_pd: could not acquire udavm information (%d)\n", ret);
        goto get_udavm_info_err;
    }
    MTL_DEBUG4("XHH_uldm_free_pd: udavm=%p, use_priv_udav = %s, lkey = 0x%x\n", 
               udavm, (use_priv_udav ? "TRUE" : "FALSE"), lkey);
    if (!use_priv_udav) {

        ret = XHH_hob_get_mrwm(uldm->hob,&mrwm);
        if (ret != HH_OK) {
            MTL_ERROR1("XHH_uldm_free_pd: could not acquire MRWM handle (%d)\n", ret);
            ret = HH_OK;
            goto udavm_get_mwrm_err;
        }
        
        lkey =  uldm->uldm_pd_table[pd].u1.pd.lkey; /* get lkey saved with this PD */
        MTL_DEBUG4("XHH_uldm_free_pd: DEREGISTERING mem region with lkey = 0x%x\n", lkey); 
        ret = XHH_mrwm_deregister_mr(mrwm, lkey);
        if (ret != HH_OK) {
            MTL_ERROR1("XHH_uldm_free_pd: XHH_mrwm_deregister_mr error (%d)\n", ret);
            ret = HH_OK;
            goto  mrwm_unregister_err;
        }
        
        udav_virt_addr = uldm->uldm_pd_table[pd].u1.pd.udav_table_ddr_virt_addr;
        udav_phys_addr = uldm->uldm_pd_table[pd].u1.pd.udav_table_ddr_phys_addr;
        udav_size      = uldm->uldm_pd_table[pd].u1.pd.udav_table_size;
        
        MTL_DEBUG4("XHH_uldm_free_pd: PD %d: udav_phys_addr =" PHYS_ADDR_FMT ", udav_virt_addr = " VIRT_ADDR_FMT  ", udav_size = "SIZE_T_FMT"\n", 
                   pd, udav_phys_addr, udav_virt_addr, udav_size);
        /* If UDAV was allocated in DDR, free up the DDR memory here */
        if (udav_phys_addr != (MT_phys_addr_t) 0) {
            ret = XHH_hob_get_lamm(uldm->hob,&lamm);
            if (ret != HH_OK) {
                MTL_ERROR1("XHH_uldm_free_pd: could not acquire DDRMM handle (%d)\n", ret);
                ret = HH_OK;
                goto udavm_get_lamm_err;
            }
    
            if ((res = MOSAL_iomem_unmap(uldm->uldm_pd_table[pd].u1.pd.iomem_obj, 
                                        uldm->uldm_pd_table[pd].u1.pd.prot_ctx)) != MT_OK) {
                MTL_ERROR1(MT_FLFMT("XHH_uldm_free_pd: MOSAL_iomem_unmap error for udavm (%d:%s)"), 
                           res, mtl_strerror_sym(res));
                ret = HH_OK;
                goto unmap_phys_addr_err;
            }
            ret = XHH_lamm_free(lamm, udav_phys_addr, udav_size);
            if (ret != HH_OK) {
                MTL_ERROR1("XHH_uldm_free_pd: XHH_lamm_free error (%d)\n", ret);
                ret = HH_OK;
                goto  lamm_free_err;
            }
        }
    }
    ret = HH_OK;
    goto ok_retn;

lamm_free_err:
  unmap_phys_addr_err:
udavm_get_lamm_err:
mrwm_unregister_err:
udavm_get_mwrm_err:
ok_retn:
    /* make resource inaccessible */
    uldm->uldm_pd_table[pd].valid = FALSE;
    epool_free(&(uldm->pd_list), pd);

get_udavm_info_err:
err_retn:
    MTL_DEBUG4("<== XHH_uldm_free_pd. ret = %d\n", ret);
    return ret;
}

/******************************************************************************
 *  Function:     XHH_uldm_get_protection_ctx 
 *****************************************************************************/
HH_ret_t XHH_get_protection_ctx( /*IN*/ XHH_uldm_t                 uldm, 
                                      /*IN*/ HH_pd_hndl_t               pd, 
                                      /*OUT*/ MOSAL_protection_ctx_t    *prot_ctx_p )
{
    HH_ret_t  ret = HH_OK;

    MTL_DEBUG4("==> XHH_uldm_get_protection_ctx. uldm = %p, pd = %d, prot_ctx_p = %p\n", 
               (void *) uldm, pd, (void *)prot_ctx_p);
    if (uldm == (XHH_uldm_t)XHH_INVALID_HNDL) {
        MTL_ERROR1("XHH_uldm_get_protection_ctx: ERROR : HCA device has not yet been opened\n");
        ret = HH_EINVAL;
        goto err_retn;
    }
    
    /* protect against bad calls */
    if (uldm == (XHH_uldm_t)0) {
        MTL_ERROR1("XHH_uldm_get_protection_ctx: ERROR : uldm handle is ZERO\n");
        ret = HH_EINVAL;
    goto err_retn;
}

    if (pd > uldm->max_pd ) {
        MTL_ERROR1("XHH_uldm_get_protection_ctx: pd out of range (%d)\n", pd);
        ret = HH_EINVAL;
        goto err_retn;
    }

    if (!(uldm->uldm_pd_table[pd].valid)) {
        if (pd != XHH_RESERVED_PD(uldm->hob)) {
             MTL_ERROR1("XHH_uldm_get_protection_ctx: pd was not allocated (%d)\n", pd);
        }
        ret = HH_EINVAL;
        goto err_retn;
    }

    *prot_ctx_p = uldm->uldm_pd_table[pd].u1.pd.prot_ctx;
    MTL_DEBUG4("XHH_uldm_get_protection_ctx. ctx  = %d\n", *prot_ctx_p);

err_retn:
    MTL_DEBUG4("<== XHH_uldm_get_protection_ctx. ret = %d\n", ret);
    return ret;
}

HH_ret_t XHH_uldm_get_num_objs( /*IN*/ XHH_uldm_t uldm, u_int32_t *num_alloc_us_res_p,
                                           u_int32_t  *num_alloc_pds_p)
{
    u_int32_t i;
    u_int32_t  alloc_res = 0;

    MTL_DEBUG4("==> XHH_uldm_get_num_objs\n");
    if ((uldm == (XHH_uldm_t)XHH_INVALID_HNDL) ||
         (uldm == (XHH_uldm_t)0) ||
        (!num_alloc_us_res_p && !num_alloc_pds_p)){
        MTL_ERROR1("XHH_uldm_get_num_objs: Invalid uldm handle, or all return params are NULL\n");
        return HH_EINVAL;
    }
    
    for (i=0; i< uldm->max_uar; i++) {
        if (uldm->uldm_uar_table[i].valid) {
            /* is uar valid check is within this function */
            alloc_res++;
        }
    }
    if (num_alloc_us_res_p) {
        *num_alloc_us_res_p=alloc_res;
    }

    alloc_res = 0;
    for (i=0; i< uldm->max_pd; i++) {
        if (uldm->uldm_pd_table[i].valid) {
            /* is uar valid check is within this function */
            alloc_res++;
        }
    }
    if (num_alloc_pds_p) {
        *num_alloc_pds_p=alloc_res;
    }
    
    return HH_OK;
}

