/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#define C_XHHUL_CQM_C

#include <hh.h>
#include <hhul.h>
#include <ul.h>
#include <ul_hob.h>
#include <ul_srqm.h>
#include <ul_qpm.h>
#include <ul_cqm.h>
#include <ul_cqm_priv.h>



/**********************************************************************************************
 *                    Private inline functions 
 **********************************************************************************************/

/* Check if a CQE is in hardware ownership (over original CQE) */
inline static MT_bool is_cqe_hw_own(volatile u_int32_t* cqe_p)
{
  return (((volatile u_int8_t*)cqe_p)[CQE_OWNER_BYTE_OFFSET] >> CQE_OWNER_SHIFT);  
  /* bit is '1 for HW-own.*/
}

inline static u_int32_t get_qp_hash_index(IB_wqpn_t qpn)
{
  return (qpn & MASK32(LOG2_QP_HASH_TBL_SZ));
}

#define INVALID_QP_HNDL NULL
/* Invoke with CQ's lock held */
inline static XHHUL_qp_t find_qp(XHHUL_cq_t *cq_p, IB_wqpn_t qpn)
{
	u_int32_t hash_index= get_qp_hash_index(qpn);
	qp_hash_entry_t *cur_entry= cq_p->qpn2qp_hash_tbl[hash_index];
	
	while (cur_entry != NULL) {
		if (cur_entry->qpn == qpn)  {
#ifdef CQM_OPTIMIZE_CQE_BURSTS
/* Optimization for cases where bursts of completions from the same QP arrive 
 * and there are many QP's associated with the CQ (that may be in the same hash bin) */
			/* Make entry the first so next time will be faster for it */
			if (cur_entry->prev != NULL) {
				entry2remove_p->prev->next= entry2remove_p->next;
				if (entry2remove_p->next != NULL) {
					entry2remove_p->next->prev= entry2remove_p->prev;
				}
				cur_entry->next= cq_p->qpn2qp_hash_tbl[hash_index];
				cur_entry->prev= NULL;
				cq_p->qpn2qp_hash_tbl[hash_index]= cur_entry;
			}
#endif /*CQM_OPTIMIZE_CQE_BURSTS*/
			return cur_entry->qp;
		}
		cur_entry= cur_entry->next;
	}
	return INVALID_QP_HNDL;
}

/* include device-specific static inline functions */
#include <dev_ul_cqm_inline.h>

/* Return CQE to HW ownership (change bit directly over CQE) */
inline static void set_cqe_to_hw_own(volatile u_int32_t *cqe_p)
{
  ((volatile u_int8_t*)cqe_p)[CQE_OWNER_BYTE_OFFSET]= (1 << CQE_OWNER_SHIFT);
}

/* Set ownership of CQE to hardware ownership (over original CQE) and */
/* increment consumer index (software + hardware) */
/* This function assumes CQ lock is already acquired by this thread */
inline static void  free_cqe(XHHUL_cq_t *cq_p, volatile u_int32_t* cqe_p)
{
  /* Pass ownership to HW */
  set_cqe_to_hw_own(cqe_p);
  /* Update software consumer index */
  cq_p->cur_buf.consumer_counter++; 
  MTL_DEBUG6(MT_FLFMT("CQn=0x%X , consumer_counter=0x%X  log2_num_o_cqes=%u cur_spare_cqes=%u pending_cq_dbell=%u"),
	     cq_p->cq_num, cq_p->cur_buf.consumer_counter, 
	     cq_p->cur_buf.log2_num_o_cqes, cq_p->cur_spare_cqes, cq_p->pending_cq_dbell);
  
#ifndef NO_CQ_CI_DBELL
  cq_p->pending_cq_dbell++;
  if (cq_p->pending_cq_dbell > cq_p->cur_spare_cqes) {
    /* update consumer index if needed. If updated, sets pending_cq_dbell to zero*/
    sync_consumer_index(cq_p,"free_cqe");
  } 
#endif  /* NO_CQ_CI_DBELL */
  
}

/* Translate from Tavor's error syndrom (in CQE.ib_syn) status encoding to VAPI's */
inline static VAPI_wc_status_t decode_error_syndrome(ADD_DEV_PREFIX_LC(if_comp_status_t) tstatus)
{
  switch (tstatus) {
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_LCL_LEN): 
      return VAPI_LOC_LEN_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_LCL_QP_OP):
      return VAPI_LOC_QP_OP_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_LCL_EE_OP):
      return VAPI_LOC_EE_OP_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_LCL_PROT):
      return VAPI_LOC_PROT_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_FLUSH):
      return VAPI_WR_FLUSH_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_MWIN_BIND):
      return VAPI_MW_BIND_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_BAD_RESP):
      return VAPI_BAD_RESP_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_LCL_ACCS):
      return VAPI_LOC_ACCS_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_RMT_INVAL_REQ):
      return VAPI_REM_INV_REQ_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_RMT_ACCSS):
      return VAPI_REM_ACCESS_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_RMT_OP):
      return VAPI_REM_OP_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_TRANS_RETRY_EX):
      return VAPI_RETRY_EXC_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_RNR_RETRY_EX):
      return VAPI_RNR_RETRY_EXC_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_LCL_RDD_VIOL):
      return VAPI_LOC_RDD_VIOL_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_RMT_INVAL_REQ_RD):
      return VAPI_REM_INV_RD_REQ_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_RMT_ABORT):
      return VAPI_REM_ABORT_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_INVAL_EEC_NUM):
      return VAPI_INV_EECN_ERR;
    case ADD_DEV_PREFIX(IF_COMP_STATUS_ERR_INVAL_EEC_STT):
      return VAPI_INV_EEC_STATE_ERR;
    default:
      MTL_ERROR1(MT_FLFMT("Invalid CQE error syndrome (0x%X)"),tstatus);
      return VAPI_COMP_GENERAL_ERR;
  }
}

inline static HH_ret_t decode_opcode(MT_bool send_q, u_int8_t cqe_opcode,
  VAPI_cqe_opcode_t *vapi_cqe_opcode_p, MT_bool *immediate_valid_p)
{
  *immediate_valid_p= FALSE;  /* Innocent until proven guilty... */
  if (send_q) { /* Send queue - use "nopcode" encoding */
    switch (cqe_opcode) {
      case ADD_DEV_PREFIX(IF_NOPCODE_RDMAW):
        *vapi_cqe_opcode_p= VAPI_CQE_SQ_RDMA_WRITE;
        return HH_OK;
      case ADD_DEV_PREFIX(IF_NOPCODE_RDMAW_IMM):
        *vapi_cqe_opcode_p= VAPI_CQE_SQ_RDMA_WRITE;
        *immediate_valid_p= TRUE;
        return HH_OK;
      case ADD_DEV_PREFIX(IF_NOPCODE_SEND):
        *vapi_cqe_opcode_p= VAPI_CQE_SQ_SEND_DATA;
        return HH_OK;
      case ADD_DEV_PREFIX(IF_NOPCODE_SEND_IMM):
        *vapi_cqe_opcode_p= VAPI_CQE_SQ_SEND_DATA;
        *immediate_valid_p= TRUE;
        return HH_OK;
      case ADD_DEV_PREFIX(IF_NOPCODE_RDMAR):
        *vapi_cqe_opcode_p= VAPI_CQE_SQ_RDMA_READ;
        return HH_OK;
      case ADD_DEV_PREFIX(IF_NOPCODE_ATOM_CMPSWP):
        *vapi_cqe_opcode_p= VAPI_CQE_SQ_COMP_SWAP;
        return HH_OK;
      case ADD_DEV_PREFIX(IF_NOPCODE_ATOM_FTCHADD):
        *vapi_cqe_opcode_p= VAPI_CQE_SQ_FETCH_ADD;
        return HH_OK;
      case ADD_DEV_PREFIX(IF_NOPCODE_BIND_MEMWIN):
        *vapi_cqe_opcode_p= VAPI_CQE_SQ_BIND_MRW;
        return HH_OK;
      default:
        return HH_EINVAL; /* Invalid opcode - shouldn't happen */
    }
  
  } else {  /* receive queue - use IB encoding */
    /* bits 4:0 are of the opcode are common to all transport types */
    switch (cqe_opcode & MASK32(5)) { 
      case IB_OP_SEND_LAST:
      case IB_OP_SEND_ONLY:
        *vapi_cqe_opcode_p= VAPI_CQE_RQ_SEND_DATA;
        return HH_OK;
      case IB_OP_SEND_IMM_LAST:
      case IB_OP_SEND_IMM_ONLY:
        *vapi_cqe_opcode_p= VAPI_CQE_RQ_SEND_DATA;
        *immediate_valid_p= TRUE;
        return HH_OK;
      case IB_OP_RDMAW_IMM_LAST:
      case IB_OP_RDMAW_IMM_ONLY:
        *vapi_cqe_opcode_p= VAPI_CQE_RQ_RDMA_WITH_IMM;
        *immediate_valid_p= TRUE;
        return HH_OK;
      default:
        return HH_EINVAL;
    }
  }
}

/* Extract CQE fields but for "status", "free_res_count" and "id" (already filled in poll4cqe) */
/* This function is used only for successfull completions. */
/* Given CQE is already in CPU endianess */
inline static HH_ret_t extract_cqe(u_int32_t *cqe, VAPI_wc_desc_t *vapi_cqe_p, 
  VAPI_special_qp_t qp_type, VAPI_ts_type_t qp_ts_type) 
{
  HH_ret_t rc;
  MT_bool send_cqe= MT_EXTRACT_ARRAY32(cqe,
    MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),s),
    MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),s));
  
  rc= decode_opcode(
    send_cqe ,
    MT_EXTRACT_ARRAY32(cqe,
      MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),opcode),
      MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),opcode)   ),
    &(vapi_cqe_p->opcode),&(vapi_cqe_p->imm_data_valid));
  if (rc != HH_OK) {
    MTL_ERROR4(MT_FLFMT("Invalid %s Opcode=0x%X"),send_cqe ? "send" : "receive",
    MT_EXTRACT_ARRAY32(cqe,
      MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),opcode),
      MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),opcode)   ) );
    return rc;
  }
  if (send_cqe && 
      ((vapi_cqe_p->opcode == VAPI_CQE_SQ_COMP_SWAP) || 
       (vapi_cqe_p->opcode == VAPI_CQE_SQ_FETCH_ADD)   
       ) 
      ) { /* Atomic operations are always of length 8 */
    vapi_cqe_p->byte_len= 8;
  } else { /* Get bytes transfered from CQE (see FM issue #15659) */
    vapi_cqe_p->byte_len= cqe[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),byte_cnt)>>2];
  }
  
  /* Get data from immediate_ethertype_pkey_indx_eecredits if valid */
      
  switch (qp_ts_type) {
    case VAPI_TS_UD:
      if (!send_cqe) {   /* see IB-spec. 11.4.2.1: Output Modifiers */
        vapi_cqe_p->remote_node_addr.type= VAPI_RNA_UD;
        vapi_cqe_p->remote_node_addr.qp_ety.qp= 
          cqe[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),rqpn)>>2] & MASK32(24);
        vapi_cqe_p->remote_node_addr.ee_dlid.dst_path_bits= MT_EXTRACT_ARRAY32(cqe,
          MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),ml_path),
          MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),ml_path)   );
        vapi_cqe_p->remote_node_addr.slid= MT_EXTRACT_ARRAY32(cqe,
          MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),rlid),MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),rlid));
        vapi_cqe_p->remote_node_addr.sl= MT_EXTRACT_ARRAY32(cqe,
          MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),sl),MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),sl));
        vapi_cqe_p->grh_flag= MT_EXTRACT_ARRAY32(cqe,
          MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),g),MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),g));
      }
      break;
    case VAPI_TS_RD:
      if (!send_cqe) {  /* see IB-spec. 11.4.2.1: Output Modifiers */
        vapi_cqe_p->remote_node_addr.type= VAPI_RNA_RD;
        vapi_cqe_p->remote_node_addr.qp_ety.qp= 
          cqe[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),rqpn)>>2] & MASK32(24);
        vapi_cqe_p->remote_node_addr.ee_dlid.loc_eecn= 
          cqe[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),my_ee)>>2] & MASK32(24);
      }
      break;
    default:
      break;
  }
      

  switch (qp_type) {
  case VAPI_REGULAR_QP:
    /*if (vapi_cqe_p->imm_data_valid) */ 
    /* Copy immediate_ethertype_pkey_indx_eecredits even if no immediate data, 
     * in order to get eecredits (for SQ) */
        vapi_cqe_p->imm_data = 
            cqe[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),immediate_ethertype_pkey_indx_eecredits)>>2];
    break;
  case VAPI_RAW_ETY_QP:
    vapi_cqe_p->remote_node_addr.type= VAPI_RNA_RAW_ETY;
    vapi_cqe_p->remote_node_addr.qp_ety.ety= 
      (VAPI_ethertype_t)cqe[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),immediate_ethertype_pkey_indx_eecredits)>>2];
    break;
  
  case VAPI_GSI_QP:
    vapi_cqe_p->pkey_ix= 
      cqe[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),immediate_ethertype_pkey_indx_eecredits)>>2]
        >> 16;  /* Pkey index is on bits 31:16 */
    vapi_cqe_p->imm_data_valid= FALSE;  
    /* QP1's RQ requests complete as "send w/immediate", even though it is just a send */
    break;
  
  default:
    break;

  }

  return HH_OK;
}


/**********************************************************************************************
 *                    Private functions declarations
 **********************************************************************************************/
static HH_ret_t cqe_buf_alloc(XHHUL_cqe_buf_t *cqe_buf, VAPI_cqe_num_t num_o_cqes);

static void cqe_buf_free(XHHUL_cqe_buf_t *cqe_buf);

static const char* cq_state_str(XHHUL_cq_state_t cq_state)
{
  switch (cq_state) {
    case XHHUL_CQ_PREP: return "XHHUL_CQ_PREP";
    case XHHUL_CQ_IDLE: return "XHHUL_CQ_IDLE";
    case XHHUL_CQ_RESIZE_PREP: return "XHHUL_CQ_RESIZE_PREP";
    default: return "(unknown state)";
  }
}

static u_int32_t cqe_buf_cleanup(XHHUL_cqe_buf_t *cqe_buf,IB_wqpn_t qp,
                                 XHHUL_srqm_t srqm, HHUL_srq_hndl_t srq,
                                 u_int32_t *cur_producer_counter_p);

static void cqe_buf_cpy2resized(
  XHHUL_cqe_buf_t *cur_buf, 
  XHHUL_cqe_buf_t *resized_buf); /* New resize-CQ flow */

/**********************************************************************************************
 *                    Public API Functions (defined in xhhul_hob.h)
 **********************************************************************************************/


HH_ret_t XHHUL_cqm_create( 
  /*IN*/ XHHUL_hob_t  hob, 
  /*OUT*/ XHHUL_cqm_t *cqm_p 
) 
{ 
  XHHUL_cqm_t new_cqm;

  new_cqm= (XHHUL_cqm_t)MALLOC(sizeof(struct XHHUL_cqm_st));
  if (new_cqm == NULL) {
    MTL_ERROR1("XHHUL_cqm_create: Failed to allocate memory for a new CQM.\n");
    return HH_EAGAIN;
  }
  new_cqm->cq_list= NULL;
  MOSAL_mutex_init(&(new_cqm->cqm_lock));

  *cqm_p= new_cqm;
  return HH_OK;
}

HH_ret_t XHHUL_cqm_destroy (
  /*IN*/ XHHUL_cqm_t cqm
) 
{ 
    XHHUL_cq_t *xhhul_cq;
       
    while (cqm->cq_list) {  
        xhhul_cq = cqm->cq_list;
        cqm->cq_list= xhhul_cq->next;
        MOSAL_memory_unlock(xhhul_cq->cur_buf.cqe_buf_base,
                          (1<<xhhul_cq->cur_buf.log2_num_o_cqes)*CQE_SZ);
        cqe_buf_free(&(xhhul_cq->cur_buf));
        FREE(xhhul_cq);
    }

    MOSAL_mutex_free(&(cqm->cqm_lock));
    FREE(cqm);
    return HH_OK; 
}


HH_ret_t XHHUL_cqm_create_cq_prep(
  /*IN*/  HHUL_hca_hndl_t hca, 
  /*IN*/  VAPI_cqe_num_t  num_o_cqes, 
  /*OUT*/ HHUL_cq_hndl_t  *hhul_cq_p, 
  /*OUT*/ VAPI_cqe_num_t  *num_o_cqes_p, 
  /*OUT*/ void/*XHH_cq_ul_resources_t*/ *cq_ul_resources_p 
) 
{ 
  XHHUL_cqm_t cqm;
  XHHUL_cq_t *new_cq;
  XHH_cq_ul_resources_t *ul_res_p= (XHH_cq_ul_resources_t*)cq_ul_resources_p;
  HH_ret_t rc;
  XHH_hca_ul_resources_t *hca_ul_res_p = NULL;
  call_result_t mrc;
  int i;

  MTL_DEBUG2(MT_FLFMT("Entering %s: hca hndl=0x%p, num cqes=0x%x"),
             __func__,hca,num_o_cqes);
  rc= XHHUL_hob_get_cqm(hca,&cqm);
  if (rc != HH_OK) {
    MTL_ERROR1("XHHUL_cqm_create_cq_prep: Invalid HCA handle.\n");
    return HH_EINVAL;
  }

  rc= XHHUL_hob_get_hca_ul_res_p(hca,&hca_ul_res_p);
  if (rc != HH_OK) {
    MTL_ERROR2("XHHUL_cqm_create_cq_prep: Failed XHHUL_hob_get_hca_ul_res_p (err=%d).\n",rc);
    return rc;
  }

  MTL_DEBUG2(MT_FLFMT("%s: After XHHUL_hob_get_hca_ul_res_p"),__func__);
  if (num_o_cqes > hca_ul_res_p->max_num_ent_cq) {
      MTL_ERROR2("XHHUL_cqm_create_cq_prep: cq_num_of_entries requested exceeds hca cap\n");
      return HH_E2BIG_CQE_NUM;
  }

  new_cq= (XHHUL_cq_t*)MALLOC(sizeof(XHHUL_cq_t));
  if (new_cq == NULL) {
    MTL_ERROR1("XHHUL_cqm_create_cq_prep: Failed to allocate XHHUL_cq_t.\n");
    return HH_EAGAIN;
  }
  memset(new_cq, 0, sizeof(XHHUL_cq_t));

  rc= cqe_buf_alloc(&new_cq->cur_buf,num_o_cqes);
  if (rc != HH_OK) goto failed_cqe_buf;
  new_cq->cq_state= XHHUL_CQ_PREP; 
  new_cq->cur_spare_cqes= new_cq->cur_buf.spare_cqes;
  new_cq->pending_cq_dbell= 0;
  
  rc = init_uar_info(hca, new_cq, ul_res_p);
  if (rc != HH_OK) {
    MTL_ERROR1("XHHUL_cqm_create_cq_prep: init_uar_info error (%d).",rc);
    goto failed_uar;
  }

  rc= XHHUL_hob_get_qpm(hca,&(new_cq->qpm));
  if (rc != HH_OK) {
    MTL_ERROR1("XHHUL_cqm_create_cq_prep: Failed getting XHHUL_hob's QPM (err=%d).\n",rc);
    goto failed_qpm;
  }

  for (i= 0; i < QP_HASH_TBL_SZ; i++) { /* Init. qpn2qp hash table */
	  new_cq->qpn2qp_hash_tbl[i]= NULL;
  }

  new_cq->cq_num= INVALID_CQ_NUM;
  MOSAL_spinlock_init(&(new_cq->cq_lock));
  /* Add CQ to CQs list */
  if (MOSAL_mutex_acq(&(cqm->cqm_lock),TRUE) != 0)  {rc= HH_EINTR; goto failed_mutex;}
  new_cq->next= cqm->cq_list; /* Add before first (if any) */
  cqm->cq_list= new_cq;
  MOSAL_mutex_rel(&(cqm->cqm_lock));

  mrc = MOSAL_memory_lock(new_cq->cur_buf.cqe_buf_base, (1<<new_cq->cur_buf.log2_num_o_cqes)*CQE_SZ);
  if ( mrc != MT_OK ) {
    rc = MOSAL2HH_RET(mrc);
    goto failed_mlock;
  }

  /* Output modifiers */
  *hhul_cq_p= (HHUL_cq_hndl_t)new_cq;
  /* One CQE is always reserved (for CQ-full indication) */
  *num_o_cqes_p= (1U << new_cq->cur_buf.log2_num_o_cqes) - 1 - new_cq->cur_buf.spare_cqes; 
  ul_res_p->cqe_buf= new_cq->cur_buf.cqe_buf_base;  /* Registration is required only from here */
  ul_res_p->cqe_buf_sz= (1U << new_cq->cur_buf.log2_num_o_cqes) * CQE_SZ ;  
  /* ul_res_p->uar_index ==> already set above */
  return HH_OK;

  failed_mlock:
  failed_mutex:
  failed_qpm:
  failed_uar:
    cqe_buf_free(&(new_cq->cur_buf));
  failed_cqe_buf:
    FREE(new_cq);
    return rc;   
}


HH_ret_t XHHUL_cqm_create_cq_done(
  /*IN*/ HHUL_hca_hndl_t hca, 
  /*IN*/ HHUL_cq_hndl_t hhul_cq, 
  /*IN*/ HH_cq_hndl_t hh_cq, 
  /*IN*/ void/*XHH_cq_ul_resources_t*/ *cq_ul_resources_p 
) 
{ 
  XHHUL_cq_t *cq= (XHHUL_cq_t*)hhul_cq;
  
  if (cq == NULL) {
    MTL_ERROR1("XHHUL_cqm_create_cq_done: NULL CQ handle.\n");
    return HH_EINVAL;
  }
  if (hh_cq > MAX_CQ_NUM) {
    MTL_ERROR1("XHHUL_cqm_create_cq_done: Invalid CQ number (0x%X).\n",hh_cq);
    return HH_EINVAL;
  }
  
  MOSAL_spinlock_irq_lock(&(cq->cq_lock));
  if (cq->cq_state != XHHUL_CQ_PREP) {
    MOSAL_spinlock_unlock(&(cq->cq_lock));
    MTL_ERROR1("XHHUL_cqm_create_cq_done: Library inconsistancy ! Given CQ is not in XHHUL_CQ_PREP state.\n");
    return HH_ERR;
  }


  cq->cq_state= XHHUL_CQ_IDLE; 
  cq->cq_num= hh_cq;

  XHH_DEV_SPECIFIC_CREATE_CQ_DONE

  MOSAL_spinlock_unlock(&(cq->cq_lock));
  
  return HH_OK; 
}


HH_ret_t XHHUL_cqm_destroy_cq_done(
  /*IN*/ HHUL_hca_hndl_t hca_hndl, 
  /*IN*/ HHUL_cq_hndl_t cq 
) 
{ 
  XHHUL_cq_t *xhhul_cq= (XHHUL_cq_t*)cq;
  XHHUL_cq_t *prev_cq,*cur_cq;
  XHHUL_cqm_t cqm;
  HH_ret_t rc;
  int i;
  qp_hash_entry_t *cur_qp_entry;

  MTL_DEBUG2(MT_FLFMT("Entering %s: hca hndl=0x%p, cq hndl=0x%p"),__func__,hca_hndl,cq);
  rc= XHHUL_hob_get_cqm(hca_hndl,&cqm);
  if (rc != HH_OK) {
    MTL_ERROR1("XHHUL_cqm_destroy_cq_done: Invalid HCA handle.\n");
    return HH_EINVAL;
  }
  if (cq == NULL) {
    MTL_ERROR1("XHHUL_cqm_destroy_cq_done: NULL CQ handle.\n");
    return HH_EINVAL;
  }
  
  if (xhhul_cq->cq_state == XHHUL_CQ_RESIZE_PREP) {
    /* Someone invoked VAPI_destroy_cq while invoking VAPI_resize_cq */
    MTL_ERROR2(MT_FLFMT("%s: Invoked while in XHHUL_CQ_RESIZE_PREP (cqn=0x%X) !"),__func__,
               xhhul_cq->cq_num);
    return HH_EBUSY;
  }
  /* Remove from CQs list */
  MOSAL_mutex_acq_ui(&(cqm->cqm_lock));
  for (cur_cq= cqm->cq_list, prev_cq= NULL; cur_cq != NULL; 
       prev_cq= cur_cq, cur_cq= cur_cq->next) {
    if (cur_cq == cq) break;
  }
  if (cur_cq == NULL) { /* CQ not found */
    MOSAL_mutex_rel(&(cqm->cqm_lock));
    MTL_ERROR1("XHHUL_cqm_destroy_cq_done: invalid CQ handle (not found).\n");
    return HH_EINVAL;
  }
  if (prev_cq == NULL) {  /* First in list */
    cqm->cq_list= xhhul_cq->next; /* Make next be the first */
  } else {  
    prev_cq->next= xhhul_cq->next;  /* Link previous to next */
  }
  MOSAL_mutex_rel(&(cqm->cqm_lock));
  
  MOSAL_memory_unlock(xhhul_cq->cur_buf.cqe_buf_base,
                    (1<<xhhul_cq->cur_buf.log2_num_o_cqes)*CQE_SZ);

  /* Cleanup CQ resources */
  cqe_buf_free(&(xhhul_cq->cur_buf));
  /* Assure hash table is clean (not suppose to hold any QP after CQ is destroyed) */
  for (i= 0; i < QP_HASH_TBL_SZ; i++) { 
	  while (xhhul_cq->qpn2qp_hash_tbl[i] != NULL) {
		  cur_qp_entry= xhhul_cq->qpn2qp_hash_tbl[i];
		  MTL_ERROR1(MT_FLFMT("%s: Found QPn=0x%X in CQ's hash table after CQ destroyed (CQn=0x%X)."
				      " Freeing hash table entry."),
			     __func__, cur_qp_entry->qpn, xhhul_cq->cq_num);
		  xhhul_cq->qpn2qp_hash_tbl[i]= cur_qp_entry->next;
		  FREE(cur_qp_entry);
	  }
  }
    
  FREE(xhhul_cq);
  return HH_OK;
}


HH_ret_t XHHUL_cqm_resize_cq_prep(
  /*IN*/ HHUL_hca_hndl_t hca_hndl, 
  /*IN*/ HHUL_cq_hndl_t cq, 
  /*IN*/ VAPI_cqe_num_t num_o_cqes, 
  /*OUT*/ VAPI_cqe_num_t *num_o_cqes_p, 
  /*OUT*/ void/*XHH_cq_ul_resources_t*/ *cq_ul_resources_p 
) 
{ 
  XHHUL_cq_t *cq_p= (XHHUL_cq_t*)cq;
  XHH_cq_ul_resources_t *ul_res_p= (XHH_cq_ul_resources_t*)cq_ul_resources_p;
  HH_ret_t rc;
  XHH_hca_ul_resources_t *hca_ul_res_p = NULL;
  XHHUL_cqe_buf_t new_buf;
  call_result_t mrc;

  if (cq_p == NULL) {
    MTL_ERROR1("%s: NULL CQ handle.\n",__func__);
    return HH_EINVAL;
  }
  
  rc= XHHUL_hob_get_hca_ul_res_p(hca_hndl,&hca_ul_res_p);
  if (rc != HH_OK) {
      MTL_ERROR2("XHHUL_cqm_create_cq_prep: Failed XHHUL_hob_get_hca_ul_res_p (err=%d).\n",rc);
      return rc;
  }

  if (num_o_cqes > hca_ul_res_p->max_num_ent_cq) {
    MTL_ERROR2("XHHUL_cqm_create_cq_prep: cq_num_of_entries requested exceeds hca cap\n");
    return HH_E2BIG_CQE_NUM;
  }

  rc= cqe_buf_alloc(&new_buf,num_o_cqes);
  if (rc != HH_OK) {
    MTL_ERROR2(MT_FLFMT("%s: Allocating buffer for resized CQ 0x%X has failed"),__func__,
                cq_p->cq_num);
    return rc;
  }
  
  MOSAL_spinlock_irq_lock(&(cq_p->cq_lock));
  if (cq_p->cq_state != XHHUL_CQ_IDLE) {
    MTL_ERROR1("%s: CQ is not in IDLE state (current state=%d=%s)\n", __func__,
               cq_p->cq_state,cq_state_str(cq_p->cq_state));
    MOSAL_spinlock_unlock(&(cq_p->cq_lock));
    cqe_buf_free(&new_buf);
    return HH_EBUSY;
  }
  memcpy(&cq_p->resized_buf, &new_buf, sizeof(XHHUL_cqe_buf_t));

  /* Update CQ to real number of outstanding CQEs */
  /* (avoid failure in case of reduction in CQ size which matches real num. of outstanding CQEs)*/
  sync_consumer_index(cq_p, __func__);
  /* start working with the limitations of the new CQEs buffer 
   * (after resizing we don't want to be in "overdraft")      */
  if (cq_p->cur_buf.spare_cqes > cq_p->resized_buf.spare_cqes) { 
    cq_p->cur_spare_cqes = cq_p->resized_buf.spare_cqes;
  }
  
  cq_p->cq_state= XHHUL_CQ_RESIZE_PREP; 
  MOSAL_spinlock_unlock(&(cq_p->cq_lock));
  
  memset(ul_res_p,0,sizeof(XHH_cq_ul_resources_t));
  ul_res_p->cqe_buf= cq_p->resized_buf.cqe_buf_base;  /* Registration is required only from here */
  ul_res_p->cqe_buf_sz= (1U<<cq_p->resized_buf.log2_num_o_cqes) * CQE_SZ ;  
  /* lock the buffer */
  mrc = MOSAL_memory_lock(cq_p->resized_buf.cqe_buf_base,
					(1U<<cq_p->resized_buf.log2_num_o_cqes) * CQE_SZ);
  if ( mrc != MT_OK ) {
	  MOSAL_spinlock_irq_lock(&(cq_p->cq_lock));
	  cq_p->cq_state = XHHUL_CQ_IDLE;
	  MOSAL_spinlock_unlock(&(cq_p->cq_lock));
	  cqe_buf_free(&new_buf);
	  rc = MOSAL2HH_RET(mrc);
  }
  else {
	  /* unlock old buffer */
	  MOSAL_memory_unlock(cq_p->cur_buf.cqe_buf_base, (1<<cq_p->cur_buf.log2_num_o_cqes) * CQE_SZ);
	  rc = HH_OK;
  }

  *num_o_cqes_p= (1U<<cq_p->resized_buf.log2_num_o_cqes) - 1 - cq_p->resized_buf.spare_cqes;

  return rc; 
}


HH_ret_t XHHUL_cqm_resize_cq_done( 
  /*IN*/ HHUL_hca_hndl_t hca_hndl, 
  /*IN*/ HHUL_cq_hndl_t cq, 
  /*IN*/ void/*XHH_cq_ul_resources_t*/ *cq_ul_resources_p 
) 
{ 
  XHHUL_cq_t *cq_p= (XHHUL_cq_t*)cq;
  XHH_cq_ul_resources_t *ul_res_p= (XHH_cq_ul_resources_t*)cq_ul_resources_p;
  XHHUL_cqe_buf_t rm_buf;
  call_result_t rc;
  HH_ret_t hrc=HH_OK;

  
  if (cq_p == NULL) {
    MTL_ERROR1("XHHUL_cqm_resize_cq_done: NULL CQ handle.\n");
    return HH_EINVAL;
  }
  
  MOSAL_spinlock_irq_lock(&(cq_p->cq_lock));
  
  if (cq_p->cq_state == XHHUL_CQ_RESIZE_PREP) { 
    /* Still polling in old buffer */
    if (!ul_res_p) { /* HH_resize_cq  failed - clean up allocated buffer */
       /* save buffer info to perform cleanup outside of spinlock */
      memcpy(&rm_buf,&(cq_p->resized_buf),sizeof(XHHUL_cqe_buf_t));
      /* restore original "spare_cqes" of cur_buf (may have been reduced in resize_cq_prep) */
      cq_p->cur_spare_cqes= cq_p->cur_buf.spare_cqes;

	  /* lock original buffer */
	  rc = MOSAL_memory_lock(cq_p->cur_buf.cqe_buf_base, (1<<cq_p->cur_buf.log2_num_o_cqes) * CQE_SZ);
	  if ( rc != MT_OK ) {
		  hrc = MOSAL2HH_RET(rc);
	  }
	  else {
		  /* unlock resized buffer */
		  MOSAL_memory_unlock(rm_buf.cqe_buf_base, (1<<rm_buf.log2_num_o_cqes) * CQE_SZ);
	  }
    } else { /* Activate resized buffer */
      /* copy CQEs from old buffer to resized buffer */
      cqe_buf_cpy2resized(&(cq_p->cur_buf),&(cq_p->resized_buf)); 
      memcpy(&rm_buf,&(cq_p->cur_buf),sizeof(XHHUL_cqe_buf_t)); /* save old buffer */
      /* Make resize buffer current. New consumer index is already updated. */
      memcpy(&(cq_p->cur_buf),&(cq_p->resized_buf),sizeof(XHHUL_cqe_buf_t));
      cq_p->cur_spare_cqes= cq_p->cur_buf.spare_cqes; /* work with the spare CQEs of new buffer */
    }
  
  } else if (cq_p->cq_state == XHHUL_CQ_RESIZE_DONE) {
    /* Transition to resized buffer already done. No CQEs to copy. Just free old buffer */
    /* (transition done in cq_transition_to_resized_buf)                                */
    if (!ul_res_p) { 
      /* Sanity check - RESIZE_CQ command is not suppose to fail if new buffer was activated */
      MTL_ERROR1(MT_FLFMT("%s: Inconsistancy ! Got failure in RESIZE_CQ"
                          " after finding CQEs in the new buffer (cqn=0x%X)"),
        __func__, cq_p->cq_num);
    }
    memcpy(&rm_buf,&(cq_p->resized_buf),sizeof(XHHUL_cqe_buf_t)); /* save old buffer */
  
  } else { /* Invalid CQ state for this function call */
    MOSAL_spinlock_unlock(&(cq_p->cq_lock));
    MTL_ERROR1("XHHUL_cqm_resize_cq_done: Given CQ is not in XHHUL_CQ_RESIZE_PREP/DONE state.\n");
    return HH_ERR;
  }
  
  /* Good flow finalization */
  cq_p->cq_state= XHHUL_CQ_IDLE; /* new resize may be initiated */
  MOSAL_spinlock_unlock(&(cq_p->cq_lock));
  cqe_buf_free(&rm_buf); /* Free old CQEs buffer (must be done outside spinlock section) */
  return hrc; 
}


HH_ret_t XHHUL_cqm_attach_qp( 
  /*IN*/ HHUL_cq_hndl_t cq, 
  /*IN*/ XHHUL_qp_t qp,
  /*IN*/ IB_wqpn_t qpn
) 
{
	XHHUL_cq_t *cq_p= (XHHUL_cq_t*)cq;
	u_int32_t hash_index= get_qp_hash_index(qpn);
	qp_hash_entry_t* new_entry_p;

	/* Allocate hash table entry for the new QP */
	new_entry_p= (qp_hash_entry_t*)MALLOC(sizeof(qp_hash_entry_t));
	if (new_entry_p == NULL) {
	  MTL_ERROR2(MT_FLFMT("%s: Failed allocating hash table entry for QPn=0x%X."), __func__,
		     qpn);
	  return HH_EAGAIN;
	}
	/* Set entry key (QPN) and value (QP pointer) */
	new_entry_p->qpn= qpn;
	new_entry_p->qp= qp;

	/* Add to the hash bin */
	new_entry_p->prev= NULL; /* previous to first is NULL */
	MOSAL_spinlock_irq_lock(&(cq_p->cq_lock));
	if (cq_p->qpn2qp_hash_tbl[hash_index] == NULL) {  /* First entry in the bin */
	  new_entry_p->next= NULL;
	  cq_p->qpn2qp_hash_tbl[hash_index]= new_entry_p;
	} else {         /* Add as first before existing entries in the bin */
	  new_entry_p->next= cq_p->qpn2qp_hash_tbl[hash_index];
	  new_entry_p->next->prev= new_entry_p;
	  cq_p->qpn2qp_hash_tbl[hash_index]= new_entry_p;
	}
	MOSAL_spinlock_unlock(&(cq_p->cq_lock));

	return HH_OK;

}

HH_ret_t XHHUL_cqm_detach_qp( 
  /*IN*/ HHUL_cq_hndl_t cq, 
  /*IN*/ XHHUL_qp_t qp,
  /*IN*/ IB_wqpn_t qpn
) 
{
	XHHUL_cq_t *cq_p= (XHHUL_cq_t*)cq;
	u_int32_t hash_index= get_qp_hash_index(qpn);
	qp_hash_entry_t *entry2remove_p;

	MOSAL_spinlock_irq_lock(&(cq_p->cq_lock));

	/* Scan hash bin to find given QP's entry */
	for (entry2remove_p= cq_p->qpn2qp_hash_tbl[hash_index]; entry2remove_p != NULL;
	     entry2remove_p= entry2remove_p->next) {
	  if (entry2remove_p->qpn == qpn) break;
	}
	if (entry2remove_p == NULL) {
	  MTL_ERROR4(MT_FLFMT("%s: qpn=0x%X not found in the hash table."), 
		     __func__, qpn);
	  MOSAL_spinlock_unlock(&(cq_p->cq_lock));
	  return HH_EINVAL;
	}
	/* Remove (bypass) entry */
	if (entry2remove_p->prev != NULL) {
		entry2remove_p->prev->next= entry2remove_p->next;
	} else {
		cq_p->qpn2qp_hash_tbl[hash_index]= entry2remove_p->next;
	}
	if (entry2remove_p->next != NULL) {
		entry2remove_p->next->prev= entry2remove_p->prev;
	}

	MOSAL_spinlock_unlock(&(cq_p->cq_lock));

	FREE(entry2remove_p); 
	return HH_OK;
}


HH_ret_t XHHUL_cqm_cq_cleanup( 
  /*IN*/ HHUL_cq_hndl_t cq, 
  /*IN*/ IB_wqpn_t qp,
  /*IN*/ XHHUL_srqm_t srqm,
  /*IN*/ HHUL_srq_hndl_t srq
) 
{ 
  XHHUL_cq_t *xhhul_cq_p= (XHHUL_cq_t*)cq;
  u_int32_t removed_cqes= 0;
  u_int32_t cur_buf_pc; /* Current buffer producer counter */
  HH_ret_t rc= HH_OK;
  
  MTL_DEBUG2("XHHUL_cqm_cq_cleanup(cq_p=%p,qp=%d) {\n",xhhul_cq_p,qp);
  if (xhhul_cq_p == NULL) {
    MTL_ERROR1("XHHUL_cqm_cq_cleanup: NULL CQ handle.\n");
    return HH_EINVAL;
  }
  
  MOSAL_spinlock_irq_lock(&(xhhul_cq_p->cq_lock));

  removed_cqes+= cqe_buf_cleanup(&(xhhul_cq_p->cur_buf), qp, srqm, srq, &cur_buf_pc );
  /* In case we are resizing the CQ, new buffer may already include CQEs of that QP */
  if (xhhul_cq_p->cq_state == XHHUL_CQ_RESIZE_PREP) {
    /* Update resized's CI based on new resize-cq flow (FM issue #16969). 
     * (This flow will fail for old flow anyway - so no need to distinguish between cases) 
     */
    xhhul_cq_p->resized_buf.consumer_counter= cur_buf_pc;
    removed_cqes+= cqe_buf_cleanup(&(xhhul_cq_p->resized_buf), qp, srqm, srq, &cur_buf_pc);
  }
    
#ifndef NO_CQ_CI_DBELL
  /* Ring CQ-cmd doorbell to update consumer index (Removed CQEs + pending CQ doorbells) */
  if (removed_cqes) {
    xhhul_cq_p->pending_cq_dbell += (u_int16_t)removed_cqes;
    sync_consumer_index(xhhul_cq_p,"XHHUL_cqm_cq_cleanup");  
  }
#endif
  
  MOSAL_spinlock_unlock(&(xhhul_cq_p->cq_lock));

  return rc; 
}

#ifdef XHHUL_CQM_DEBUG_WQE_REUSE
void XHHUL_cqm_dump_cq(
  /*IN*/ HHUL_cq_hndl_t cq 
)
{
  XHHUL_cq_t *xhhul_cq_p= (XHHUL_cq_t*)cq;
  volatile u_int32_t *cur_cqe;
  int cqe_index;
  static HHUL_cq_hndl_t last_cq= NULL;
  static u_int32_t last_consumer_counter= 0xFFFFFFFF;

  /* Do not dump again for the same CQ/consumer-index */
  if ((cq == last_cq) && (last_consumer_counter == xhhul_cq_p->cur_buf.consumer_counter)  return;
  last_cq= cq; 
  last_consumer_counter= xhhul_cq_p->cur_buf.consumer_counter & MASK32(xhhul_cq_p->cur_buf.log2_num_o_cqes);

  MTL_ERROR4("XHHUL_cqm_dump_cq: cq=%d consumer_index=%d\n",
    xhhul_cq_p->cq_num,(last_consumer_counter & MASK32(xhhul_cq_p->cur_buf.log2_num_o_cqes)));
  for (cqe_index= 0; cqe_index < (1 << xhhul_cq_p->cur_buf.log2_num_o_cqes); cqe_index++) {
    cur_cqe= (volatile u_int32_t *)
      (xhhul_cq_p->cur_buf.cqe_buf_base + (cqe_index << LOG2_CQE_SZ)); 
    DUMP_CQE(xhhul_cq_p->cq_num,cqe_index,cur_cqe);
  }

}
#endif

/* Check if can transition to new (resized) CQEs buffer and make cur_buf<--resized_buf
 * - Valid for fixed resize cq FW version
 * - Return TRUE if transition and cur_cqe is updated to new location in resized buf.
 * - Must be invoked with CQ lock locked and only when CQE at current CI is invalid
 */
static MT_bool cq_transition_to_resized_buf(
  XHHUL_cq_t *cq_p,
  volatile u_int32_t **cur_cqe_p
)
{
  XHHUL_cqe_buf_t rm_buf;
  
  if ((cq_p->cq_state == XHHUL_CQ_RESIZE_PREP))                   { /* Peek into resized buffer */
    cq_p->resized_buf.consumer_counter = 
        XHH_DEV_RESIZE_COUNTER_INIT_VAL(cq_p->cur_buf.consumer_counter,MASK32(cq_p->cur_buf.log2_num_o_cqes));
    *cur_cqe_p= (volatile u_int32_t *)
      (cq_p->resized_buf.cqe_buf_base + 
       (((cq_p->resized_buf.consumer_counter)& MASK32(cq_p->resized_buf.log2_num_o_cqes)) << LOG2_CQE_SZ)); 
    if (!is_cqe_hw_own(*cur_cqe_p)) { /* Found CQE in new (resized) buffer */
      MTL_DEBUG4(MT_FLFMT("%s: transition to resized (cqn=0x%x old_pi=%u new_pi=%u cur_cqe=%p)"),
                 __func__,
                 cq_p->cq_num, cq_p->cur_buf.consumer_counter, cq_p->resized_buf.consumer_counter,
                 *cur_cqe_p);
      /* Transition to resized buffer */
      memcpy(&rm_buf,&(cq_p->cur_buf),sizeof(XHHUL_cqe_buf_t)); /* save old buffer */
      /* Make resize buffer current. New consumer index is already updated. */
      memcpy(&(cq_p->cur_buf),&(cq_p->resized_buf),sizeof(XHHUL_cqe_buf_t));
      cq_p->cur_spare_cqes= cq_p->cur_buf.spare_cqes; /* work with the spare CQEs of new buffer */
      /* Save old buffer to be freed in "resize_done" (no need to hurry...) */
      memcpy(&(cq_p->resized_buf),&rm_buf,sizeof(XHHUL_cqe_buf_t)); 
      cq_p->cq_state= XHHUL_CQ_RESIZE_DONE; /* XHHUL_cqm_resize_cq_done is still expected */
      return TRUE;
    }
  }
  return FALSE;
}

HH_ret_t XHHUL_cqm_poll4cqe( 
  /*IN*/ HHUL_hca_hndl_t hca_hndl, 
  /*IN*/ HHUL_cq_hndl_t cq, 
  /*OUT*/ VAPI_wc_desc_t *vapi_cqe_p 
) 
{ 
  XHHUL_cq_t *xhhul_cq_p= (XHHUL_cq_t*)cq;
  XHHUL_qp_t comp_qp; /* QP of given completion */
  volatile u_int32_t *cur_cqe;
  u_int32_t wqe_addr_32lsb,next_wqe_addr_32lsb = 0;
  IB_wqpn_t qpn;
  u_int8_t opcode;
  MT_bool is_SQ_completion;
  u_int8_t dbd_bit = 0;
  VAPI_special_qp_t qp_type = 0;
  VAPI_ts_type_t qp_ts_type = 0;
  u_int32_t i;
  HH_ret_t rc = HH_OK;
  u_int32_t cqe_cpy[CQE_SZ>>2]; /* CQE copy */
  /* The CQE copy is required for 2 reasons:
   * 1) Hold in CPU endianess. 
   * 2) Free real CQE as soon as possible in order to release CQ lock quickly.
   */

  if (xhhul_cq_p == NULL) {
    MTL_ERROR1("XHHUL_cqm_poll4cqe: NULL CQ handle.\n");
    return HH_EINVAL_CQ_HNDL;
  }

  MOSAL_spinlock_irq_lock(&(xhhul_cq_p->cq_lock));

  /* Check if CQE at consumer index is valid */
  cur_cqe= (volatile u_int32_t *)
    (xhhul_cq_p->cur_buf.cqe_buf_base + 
        ((xhhul_cq_p->cur_buf.consumer_counter & MASK32(xhhul_cq_p->cur_buf.log2_num_o_cqes))<< LOG2_CQE_SZ)); 
  if (is_cqe_hw_own(cur_cqe) &&                 /* CQE is still in HW ownership */
      (!cq_transition_to_resized_buf(xhhul_cq_p, &cur_cqe)) )  { 

    MOSAL_spinlock_unlock(&(xhhul_cq_p->cq_lock));    
#if 0
    XHHUL_cqm_dump_cq(cq);
#endif
    return HH_CQ_EMPTY;
  }

  /* Make CQE copy in correct endianess */
  for (i= 0; i < (CQE_SZ>>2); i++) {  
    cqe_cpy[i]= MOSAL_be32_to_cpu(cur_cqe[i]);
  }
  /* Extract QP/WQE context fields from the CQE */
  qpn= (cqe_cpy[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),my_qpn)>>2] & MASK32(24) );
  wqe_addr_32lsb= (cqe_cpy[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),wqe_adr)>>2] & 
                   (~MASK32(CQE_WQE_ADR_BIT_SZ)) );
  vapi_cqe_p->local_qp_num= qpn;
  /* new CQE: completion status is taken from "opcode" field */
  opcode= MT_EXTRACT_ARRAY32(cqe_cpy,
  	MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),opcode), 
  	MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),opcode));

  comp_qp= find_qp(xhhul_cq_p,qpn);
  if (comp_qp == INVALID_QP_HNDL) {
	  MTL_ERROR1(MT_FLFMT("%s: Got completion for unknown QPn=0x%X"
			      " (not associated with this CQn=0x%X)"), 
		     __func__, qpn, xhhul_cq_p->cq_num);
	  DUMP_CQE(xhhul_cq_p->cq_num,
		   xhhul_cq_p->cur_buf.consumer_counter & MASK32(xhhul_cq_p->cur_buf.log2_num_o_cqes),
		   cur_cqe);
	  free_cqe(xhhul_cq_p,cur_cqe); /* Free original CQE and update consumer index */
	  MOSAL_spinlock_unlock(&(xhhul_cq_p->cq_lock));    
	  return HH_EINVAL_QP_NUM;
  }

  if ((opcode & CQE_ERROR_STATUS_MASK) != CQE_ERROR_STATUS_MASK) {  /* Completed OK */	
    MTPERF_TIME_START(free_cqe);
    free_cqe(xhhul_cq_p,cur_cqe); /* Free original CQE and update consumer index */
    MTPERF_TIME_END(free_cqe);

  /* DEBUG: Sanity check that the same WQE is not used twice simultaneosly */
#ifdef XHHUL_CQM_DEBUG_WQE_REUSE
    /* Get next CQE and check if valid and NDA equals freed CQE's */
  cur_cqe= (volatile u_int32_t *)
    (xhhul_cq_p->cur_buf.cqe_buf_base + 
     ((xhhul_cq_p->cur_buf.consumer_counter & MASK32(xhhul_cq_p->cur_buf.log2_num_o_cqes))<< LOG2_CQE_SZ)); 
  if ((!is_cqe_hw_own(cur_cqe)) &&
      ( (MOSAL_be32_to_cpu(
             cur_cqe[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),wqe_adr)>>2]) & 
           (~MASK32(CQE_WQE_ADR_BIT_SZ)) ) == wqe_addr_32lsb) ){ 
    MTL_ERROR1(MT_FLFMT("%s: Duplicate NDA on next CQE (NDA=0x%X , consumer index=%u,%u)"),
           __func__, wqe_addr_32lsb, 
           (xhhul_cq_p->cur_buf.consumer_counter& MASK32(xhhul_cq_p->cur_buf.log2_num_o_cqes))-1, 
               (xhhul_cq_p->cur_buf.consumer_counter& MASK32(xhhul_cq_p->cur_buf.log2_num_o_cqes)));
    XHHUL_cqm_dump_cq(cq);
  }
#endif 
  is_SQ_completion = ((MT_EXTRACT_ARRAY32(cqe_cpy, 
                             MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),s), 
		             MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),s))==0) ? 
	               FALSE : TRUE);

    rc= XHHUL_qpm_comp_ok(xhhul_cq_p->qpm, comp_qp, wqe_addr_32lsb, is_SQ_completion,
      &qp_type,&qp_ts_type,&(vapi_cqe_p->id),&(vapi_cqe_p->free_res_count));
    MOSAL_spinlock_unlock(&(xhhul_cq_p->cq_lock));    
    
    if (rc != HH_OK) {
      MTL_ERROR1("XHHUL_cqm_poll4cqe: Failed updating associated QP (QPn=0x%X,CQn=0x%X,is_SQ_compl=%u, opc=0x%x).\n",
		 qpn, xhhul_cq_p->cq_num, is_SQ_completion, opcode);
      for (i= 0; i < (CQE_SZ>>2); i++) {  
        MTL_ERROR1(MT_FLFMT("CQ[0x%X][%u][%u]=0x%X"),xhhul_cq_p->cq_num,
                   (xhhul_cq_p->cur_buf.consumer_counter - 1) & MASK32(xhhul_cq_p->cur_buf.log2_num_o_cqes),
                   i, cqe_cpy[i]);
      }
      return HH_EFATAL; /* unexpected error */
    }
    /* Extract the rest of the CQE fields into vapi_cqe_p*/
    rc= extract_cqe(cqe_cpy,vapi_cqe_p,qp_type,qp_ts_type); 
    vapi_cqe_p->status= VAPI_SUCCESS;
  
  } else {  /* Completion with error  */
    MTL_DEBUG4("XHHUL_cqm_poll4cqe: completion with error: cq=%d consumer_counter=%d\n",
      xhhul_cq_p->cq_num,xhhul_cq_p->cur_buf.consumer_counter);
    DUMP_CQE(xhhul_cq_p->cq_num,xhhul_cq_p->cur_buf.consumer_counter,cur_cqe);
    
    is_SQ_completion = (opcode == (u_int8_t)(ADD_DEV_PREFIX(IF_CQE_ERR_OPCODE_SQ)& 0xFF)) ? 
			    TRUE : FALSE;

    rc= XHHUL_qpm_comp_err(xhhul_cq_p->qpm, comp_qp, wqe_addr_32lsb,is_SQ_completion,
      &(vapi_cqe_p->id),&(vapi_cqe_p->free_res_count),&next_wqe_addr_32lsb,&dbd_bit);
    if (rc != HH_OK) {
      MTL_ERROR1("XHHUL_cqm_poll4cqe: Failed updating associated QP (QPn=0x%X , CQn=0x%X, opcode=0x%x).\n",
                 qpn, xhhul_cq_p->cq_num, opcode);
      MOSAL_spinlock_unlock(&(xhhul_cq_p->cq_lock));    
      return HH_EFATAL; /* unexpected error */
    }
    vapi_cqe_p->status= decode_error_syndrome((ADD_DEV_PREFIX_LC(if_comp_status_t))MT_EXTRACT_ARRAY32(cqe_cpy,
      CQE_ERROR_SYNDROM_BIT_OFFSET, CQE_ERROR_SYNDROM_BIT_SIZE) );
    vapi_cqe_p->vendor_err_syndrome= MT_EXTRACT_ARRAY32(cqe_cpy,
      CQE_ERROR_VENDOR_SYNDROM_BIT_OFFSET, CQE_ERROR_VENDOR_SYNDROM_BIT_SIZE);

    XHH_DEV_FREE_CQE

    MOSAL_spinlock_unlock(&(xhhul_cq_p->cq_lock)); 
    /* Only WQE-ID, free_res_count and status are required for completion with error. 
     * No other CQE fields are extracted (see IB-spec. 11.4.2.1).                   
     * Even though, for the sake of some legacy code:
     * ...putting an opcode to distinguish completion of SQ from RQ*/
    if (opcode == CQE_ERROR_ON_SQ) { 
      vapi_cqe_p->opcode= VAPI_CQE_SQ_SEND_DATA; 
    } else { /* receive queue completion */
      vapi_cqe_p->opcode= VAPI_CQE_RQ_SEND_DATA; 
    }
  }

  return rc;   
}

HH_ret_t XHHUL_cqm_peek_cq( 
  /*IN*/ HHUL_hca_hndl_t hca_hndl, 
  /*IN*/ HHUL_cq_hndl_t cq, 
  /*IN*/ VAPI_cqe_num_t cqe_num
)
{
  XHHUL_cq_t *cq_p= (XHHUL_cq_t*)cq;
  volatile u_int32_t *cur_cqe;
  HH_ret_t ret;

  /* parameters checks */
  if (cq_p == NULL) {
    MTL_ERROR1("XHHUL_cqm_peek_cq: NULL CQ handle.\n");
    return HH_EINVAL_CQ_HNDL;
  }

  /* Find CQE and check ownership */
  MOSAL_spinlock_irq_lock(&(cq_p->cq_lock));
  /* The following check must be done with CQ-lock locked, since resize may change cur_buf 
     on the same time                                                                       */
  if ((cqe_num >= ((1U << cq_p->cur_buf.log2_num_o_cqes) - cq_p->cur_buf.spare_cqes)) || (cqe_num == 0)) { 
    /* reminder: 1 CQE is always reserved */
    MTL_ERROR2("XHHUL_cqm_peek_cq(cqn=0x%X): cqe_num=%u , max_num_o_cqes=%u .\n",
               cq_p->cq_num,cqe_num,
               ((1U << cq_p->cur_buf.log2_num_o_cqes) - cq_p->cur_buf.spare_cqes - 1));
    ret= HH_E2BIG_CQE_NUM;
  } else {
    cur_cqe= (volatile u_int32_t *)
      (cq_p->cur_buf.cqe_buf_base + 
       (((cq_p->cur_buf.consumer_counter + cqe_num - 1) & MASK32(cq_p->cur_buf.log2_num_o_cqes)) 
        << LOG2_CQE_SZ));
    ret= XHH_DEV_CHECK_CQE_STATE(cq_p, cur_cqe, cqe_num);
  }
  
  MOSAL_spinlock_unlock(&(cq_p->cq_lock));    

  return ret; 
}



HH_ret_t XHHUL_cqm_req_comp_notif( 
  /*IN*/ HHUL_hca_hndl_t hca_hndl, 
  /*IN*/ HHUL_cq_hndl_t cq, 
  /*IN*/ VAPI_cq_notif_type_t notif_type 
) 
{ 
  XHHUL_cq_t *xhhul_cq_p= (XHHUL_cq_t*)cq;
  u_int32_t last_consumer_index;
  HH_ret_t rc;

  if (xhhul_cq_p == NULL) {
    MTL_ERROR1("XHHUL_cqm_req_comp_notif: NULL CQ handle.\n");
    return HH_EINVAL_CQ_HNDL;
  }
  
  MOSAL_spinlock_irq_lock(&(xhhul_cq_p->cq_lock));
#ifdef NO_CQ_CI_DBELL
  /* In "overrun ignore" mode, last consumed must be given */
  last_consumer_index= 
    ((xhhul_cq_p->cur_buf.consumer_counter - 1) & MASK32(xhhul_cq_p->cur_buf.log2_num_o_cqes));
#else
  /* Consumer index is updated on every poll, so InfiniHost has its updated value */
  last_consumer_index= 0xFFFFFFFF ; /* Use current CI value */
#endif
  sync_consumer_index(xhhul_cq_p, __func__);/*Consumer index must be updated before req. an event*/

  switch (notif_type) {
    case VAPI_SOLIC_COMP:
      rc= uar_arm_cq_cmd(xhhul_cq_p, ADD_DEV_PREFIX(IF_UAR_CQ_NOTIF_SOLIC_COMP),last_consumer_index);
      break;
    case VAPI_NEXT_COMP:
      rc= uar_arm_cq_cmd(xhhul_cq_p, ADD_DEV_PREFIX(IF_UAR_CQ_NOTIF_NEXT_COMP),last_consumer_index);
      break;
    default:
      	/* error return code here, with side-effect of sync_consumer_index -- because this is
	 * data path, and this error should not occur.  We don't want extra checks in data path
	 */
	rc= HH_EINVAL_CQ_NOT_TYPE;  /* Invalid notification request type*/
  }
  MOSAL_spinlock_unlock(&(xhhul_cq_p->cq_lock));
  
  return rc;  
}

HH_ret_t XHHUL_cqm_req_ncomp_notif( 
  /*IN*/ HHUL_hca_hndl_t hca_hndl, 
  /*IN*/ HHUL_cq_hndl_t cq, 
  /*IN*/ VAPI_cqe_num_t cqe_num 
) 
{ 
  XHHUL_cq_t *cq_p= (XHHUL_cq_t*)cq;
  u_int32_t arm_n_param;
  HH_ret_t rc;

  if (cq_p == NULL) {
    MTL_ERROR1("XHHUL_cqm_req_ncomp_notif: NULL CQ handle.\n");
    return HH_EINVAL;
  }

  /* cqe_num fix (due to CQEs with error... external CQEs) :
   * Check if cqe_num requirement was already fulfilled with external/sw CQEs
   * If yes: generate immediate event by setting cqe_num to 1
   * Otherwise: cqe_num is set to the HW number of CQEs based on current difference from sw_cnt 
   */
  MOSAL_spinlock_irq_lock(&(cq_p->cq_lock));
  /* Check cqe_num limits (must be done with lock held to avoid a race with "resize") */
  if ((cqe_num >= ((1U << cq_p->cur_buf.log2_num_o_cqes) - cq_p->cur_buf.spare_cqes)) || 
      (cqe_num == 0)  ||
      (cqe_num > MAX_NCOMP_NOTIF)) { 
    /* reminder: 1 CQE is always reserved */
    MTL_ERROR2("XHHUL_cqm_req_ncomp_notif(cqn=%d): cqe_num=%d , max_num_o_cqes=%d .\n",
               cq_p->cq_num,cqe_num,
               ((1U << cq_p->cur_buf.log2_num_o_cqes) - cq_p->cur_buf.spare_cqes - 1));
    MOSAL_spinlock_unlock(&(cq_p->cq_lock));
    return HH_E2BIG_CQE_NUM;
  }

  arm_n_param= compute_arm_n_param(cq_p, cqe_num);

  sync_consumer_index(cq_p, __func__);/*Consumer index must be updated before req. an event*/

  rc= uar_arm_cq_cmd(cq_p, ADD_DEV_PREFIX(IF_UAR_CQ_NOTIF_NCOMP), arm_n_param);
  MOSAL_spinlock_unlock(&(cq_p->cq_lock));
  
  return rc;  
}




/**********************************************************************************************
 *                    Private Functions 
 **********************************************************************************************/


static HH_ret_t cqe_buf_alloc(XHHUL_cqe_buf_t *cqe_buf, VAPI_cqe_num_t num_o_cqes)
{
  u_int32_t i;
  volatile u_int8_t* cur_cqe_owner_byte;
  VAPI_cqe_num_t actual_num_o_cqes; /* Number of CQEs in the CQEs buffer */
  VAPI_cqe_num_t possible_spare_cqes;
  
  cqe_buf->log2_num_o_cqes= floor_log2(num_o_cqes) + 1; /* next power of 2 including extra CQE */
  actual_num_o_cqes= (1 << cqe_buf->log2_num_o_cqes);  
#ifdef MT_KERNEL
  if (((actual_num_o_cqes + 1) * CQE_SZ) <= CQ_KMALLOC_LIMIT) 
    cqe_buf->cqe_buf_orig= 
      MOSAL_pci_phys_alloc_consistent((actual_num_o_cqes + 1) * CQE_SZ, LOG2_CQE_SZ); /* one extra for alignment */
  else
#endif
#if !defined(MT_KERNEL) && defined(MT_FORK_SUPPORT)
/* Fork workaroud - cover full pages */
    cqe_buf->cqe_buf_orig= 
      MOSAL_pci_virt_alloc_consistent(
        (MOSAL_SYS_PAGE_SIZE-1)/*for page alignement*/ +
            MT_UP_ALIGNX_ULONG_PTR(actual_num_o_cqes * CQE_SZ, MOSAL_SYS_PAGE_SHIFT),
        LOG2_CQE_SZ);
#else
    cqe_buf->cqe_buf_orig= 
      MOSAL_pci_virt_alloc_consistent((actual_num_o_cqes + 1) * CQE_SZ, LOG2_CQE_SZ); /* one extra for alignment */
#endif
  if (cqe_buf->cqe_buf_orig == NULL) {
    MTL_ERROR1("%s: Failed to allocate CQEs buffer of 0x%X bytes.\n",__func__,
      (actual_num_o_cqes + 1) * CQE_SZ);
    return HH_EAGAIN;
  }
#if !defined(MT_KERNEL) && defined(MT_FORK_SUPPORT)
  cqe_buf->cqe_buf_base= MT_UP_ALIGNX_VIRT((MT_virt_addr_t)cqe_buf->cqe_buf_orig,MOSAL_SYS_PAGE_SHIFT);
#else  
  /* buffer must be aligned to CQE size */
  cqe_buf->cqe_buf_base= MT_UP_ALIGNX_VIRT((MT_virt_addr_t)cqe_buf->cqe_buf_orig,LOG2_CQE_SZ);
#endif
  /* Initialize all CQEs to HW ownership */
  cur_cqe_owner_byte= (volatile u_int8_t*)(cqe_buf->cqe_buf_base+CQE_OWNER_BYTE_OFFSET);
  for (i= 0; i < actual_num_o_cqes; i++) { 
    *cur_cqe_owner_byte= (1<<CQE_OWNER_SHIFT);
    cur_cqe_owner_byte+= CQE_SZ;
  }
  possible_spare_cqes= actual_num_o_cqes - 1/*Reserved CQE*/ - num_o_cqes; /* For CQ DB coalescing */
  cqe_buf->spare_cqes= (possible_spare_cqes > MAX_CQDB2DELAY) ? MAX_CQDB2DELAY : possible_spare_cqes;
  MTL_DEBUG4(MT_FLFMT("%s: spare_cqes=%u"),__func__,cqe_buf->spare_cqes);
  cqe_buf->consumer_counter= 0;
  return HH_OK;
}

static void cqe_buf_free(XHHUL_cqe_buf_t *cqe_buf)
{
  /* Cleanup CQ resources */
#ifdef MT_KERNEL
  if ((((1U<<cqe_buf->log2_num_o_cqes)+1)*CQE_SZ) <= CQ_KMALLOC_LIMIT) 
    MOSAL_pci_phys_free_consistent(cqe_buf->cqe_buf_orig, ((1U<<cqe_buf->log2_num_o_cqes)+1)*CQE_SZ);
  else
#endif
    MOSAL_pci_virt_free_consistent(cqe_buf->cqe_buf_orig, ((1U<<cqe_buf->log2_num_o_cqes)+1)*CQE_SZ);
}

/* Perform the "CQ cleanup" flow (removing of CQEs of a RESET QP) and return 
 * amount of removed CQEs (the change in consumer index)
 * - cur_pi_p : return PI at given buffer (to be used in XHHUL_CQ_RESIZE_PREP state)
 */
static u_int32_t cqe_buf_cleanup(XHHUL_cqe_buf_t *cqe_buf,IB_wqpn_t qp,
                                 XHHUL_srqm_t srqm, HHUL_srq_hndl_t srq,
                                 u_int32_t *cur_producer_counter_p)
{
  u_int32_t cur_tail_counter,next_consumer_counter,cur_cqe_counter;
  u_int32_t outstanding_cqes,i;
  u_int32_t removed_cqes= 0;
  volatile u_int32_t *cur_cqe_p;
  volatile u_int32_t *next_cqe_p;
  IB_wqpn_t cur_qpn;
  const u_int32_t num_o_cqes_mask= MASK32(cqe_buf->log2_num_o_cqes);
  MT_bool is_rq_cqe;
  u_int32_t wqe_addr_32lsb;
  u_int8_t opcode;
#ifndef NO_UL_SRQM
  VAPI_wr_id_t wqe_id;
#endif  

  /* Find last CQE is software ownership (cur_tail) */
  outstanding_cqes= 0;
  cur_tail_counter= cqe_buf->consumer_counter;
#ifndef NO_CQ_CI_DBELL
  while (1) { /* Break out when find a CQE in HW ownership */
    /* (there must be at least one CQE in HW ownership - the reserved "full" CQE)*/
#else
  while (outstanding_cqes <= (1<<cqe_buf->log2_num_o_cqes)) {
    /* In CQ-overrun-ignore mode, all CQEs in CQ may be in SW ownership... */
#endif
    cur_cqe_p= (volatile u_int32_t *)
      (cqe_buf->cqe_buf_base + ((cur_tail_counter & num_o_cqes_mask) << LOG2_CQE_SZ));
    if (is_cqe_hw_own(cur_cqe_p))  break; /* no more outstanding CQEs */ 

    outstanding_cqes++;
    cur_tail_counter++;
  }
  *cur_producer_counter_p= cur_tail_counter; /* To be used on resized buffer */
  /* move back to last in SW ownership */
  cur_cqe_counter = next_consumer_counter = cur_tail_counter - 1;

  MTL_DEBUG4(MT_FLFMT("Found %d outstanding CQEs. Last CQE at index %d."),
    outstanding_cqes, (cur_cqe_counter & num_o_cqes_mask));
  
  /* Scan back CQEs (cur_cqe) and move all CQEs not of given qpn back to "cur_head" */
  for (i= 0; i < outstanding_cqes; i++) {
    cur_cqe_p= (volatile u_int32_t *)
      (cqe_buf->cqe_buf_base + ((cur_cqe_counter & num_o_cqes_mask)<< LOG2_CQE_SZ));
    cur_qpn= 
      MOSAL_be32_to_cpu(cur_cqe_p[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),my_qpn)>>2])
      & MASK32(24);

    if (cur_qpn == qp) {  /* A CQE to remove */
      /* Go back only with the cur_cqe_index, leave next_consumer_index behind (for next copy) */
      cur_cqe_counter--;
      
      /* If associated with SRQ must invoke XHHUL_srqm_comp to release WQE */
      if (srq != HHUL_INVAL_SRQ_HNDL) {
	  opcode= MT_EXTRACT32(
	        MOSAL_be32_to_cpu(cur_cqe_p[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),opcode)>>2]),
		MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),opcode) & MASK32(5), 
		MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),opcode));

	  if ((opcode & CQE_ERROR_STATUS_MASK) != CQE_ERROR_STATUS_MASK) {  /* Completed OK */	
	    is_rq_cqe= (MT_EXTRACT32(
	      MOSAL_be32_to_cpu(cur_cqe_p[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),s)>>2]),
	      MT_BIT_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),s) & MASK32(5),
	      MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),s))  == 0);
	  } else {
	      is_rq_cqe = (opcode == CQE_ERROR_ON_RQ);
	  }
        if (is_rq_cqe) { /* Completion must be reported to SRQ */
          wqe_addr_32lsb= /* Mask off size bits from dword of WQE addr. */
            MOSAL_be32_to_cpu(
              cur_cqe_p[MT_BYTE_OFFSET(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),wqe_adr)>>2]
            ) & 
            (0xFFFFFFFF << MT_BIT_SIZE(ADD_DEV_PRM_PREFIX(completion_queue_entry_st),wqe_size));
#ifndef NO_UL_SRQM
          if (XHHUL_srqm_comp(srqm, srq, wqe_addr_32lsb, &wqe_id) != HH_OK) {  /* Release WQE */
	      MTL_ERROR1(MT_FLFMT("%s: opcode=0x%x, cur_cqe_counter=0x%x, cur_cqe_p=%p"),
			 __func__, opcode, cur_cqe_counter, cur_cqe_p); 
	  }
#endif
        }
      }

    } else { /* A CQE to copy (if any CQE was removed) */
      if (cur_cqe_counter != next_consumer_counter) { /* Copy required */
        MTL_DEBUG4(MT_FLFMT("Moving CQE at index %d to index %d\n"),
          (cur_cqe_counter & num_o_cqes_mask),(next_consumer_counter & num_o_cqes_mask)); 
        next_cqe_p= (volatile u_int32_t *)
          (cqe_buf->cqe_buf_base + ((next_consumer_counter & num_o_cqes_mask) << LOG2_CQE_SZ));
        memcpy((void*)next_cqe_p,(void*)cur_cqe_p,CQE_SZ);
      }
      /* Go back with both indices */
      cur_cqe_counter--;
      next_consumer_counter--;
    }
  }
  
  if (cur_cqe_counter != next_consumer_counter) { /* CQEs were removed */
    /* Return to hardware ownership CQEs at amount of removed CQEs (from consumer index side) */
    for (cur_cqe_counter = cqe_buf->consumer_counter; ; cur_cqe_counter++) {
      removed_cqes++;
      cur_cqe_p= (volatile u_int32_t *) 
        (cqe_buf->cqe_buf_base + ((cur_cqe_counter & num_o_cqes_mask)<< LOG2_CQE_SZ));
      set_cqe_to_hw_own(cur_cqe_p);
      if (cur_cqe_counter == next_consumer_counter)  break; /* Returned all including this one */
    
    }
    
    /* update consumer counter - go back to location of last copied */
    cqe_buf->consumer_counter= next_consumer_counter + 1; 
  }

  return removed_cqes;
}

/* This function copies any old CQEs left in cur_buf to resized_buf */
/* (Should be called with CQ lock held) */
static void cqe_buf_cpy2resized(
  XHHUL_cqe_buf_t *cur_buf, 
  XHHUL_cqe_buf_t *resized_buf) /* New resize-CQ flow */
{
  u_int32_t cur_cqe_counter;
  u_int32_t outs_cqes_at_ci=0,outs_cqes_at_cq_base=0;
  u_int32_t *outs_cqes_p; /* pointer to currect count (one of the above) */
  volatile u_int32_t *cur_cqe_p;
  void *cur_cpy_p;     /* Pointer to copy (from) point */
  void *resized_cpy_p; /* Pointer to copy (to) point */
  u_int32_t resized_cqes_at_top; /* CQEs available to buffer's top - when wrap around */
  u_int32_t resized_cur_pc; /* Current "producer counter" */
  const u_int32_t num_o_cqes_mask= MASK32(cur_buf->log2_num_o_cqes);
  const u_int32_t new_num_o_cqes_mask= MASK32(resized_buf->log2_num_o_cqes); /* for modulo */
  

  /* Count number of CQEs in cur_buf */
  outs_cqes_p= &outs_cqes_at_ci; /* First count the CQEs above the consumer index */
  cur_cqe_counter= cur_buf->consumer_counter;
  while (1) { /* Break out when find a CQE in HW ownership */
    /* (there must be at least one CQE in HW ownership - the reserved "full" CQE)*/
    cur_cqe_p= (volatile u_int32_t *)
      (cur_buf->cqe_buf_base + ((cur_cqe_counter & num_o_cqes_mask)<< LOG2_CQE_SZ));
    if (is_cqe_hw_own(cur_cqe_p))  break; /* no more outstanding CQEs */ 
    (*outs_cqes_p)++;
    cur_cqe_counter++;
    if ((cur_cqe_counter & num_o_cqes_mask) == 0) { /* next CQEs are at CQ base */
      outs_cqes_p= &outs_cqes_at_cq_base;
    }
  }

  resized_buf->consumer_counter= 
      XHH_DEV_RESIZE_COUNTER_INIT_VAL(cur_cqe_counter, num_o_cqes_mask); /* New fixed resize-CQ */
  MTL_DEBUG5(MT_FLFMT("%s: old_pc=%u new_pc=%u new_log2_sz=%u"), __func__,
                cur_cqe_counter, resized_buf->consumer_counter, resized_buf->log2_num_o_cqes);

  resized_buf->consumer_counter = resized_buf->consumer_counter - outs_cqes_at_ci - outs_cqes_at_cq_base;
  resized_cur_pc= resized_buf->consumer_counter; /* Where CQE copy starts at resized buffer */

  if (outs_cqes_at_ci > 0) { /* First copy CQEs above consumer index */
    cur_cpy_p= (void *)
      (cur_buf->cqe_buf_base + ((cur_buf->consumer_counter & num_o_cqes_mask)<< LOG2_CQE_SZ));
    resized_cpy_p= (void *)
      (resized_buf->cqe_buf_base + ((resized_buf->consumer_counter & new_num_o_cqes_mask)<< LOG2_CQE_SZ));
    resized_cqes_at_top= (1U<<resized_buf->log2_num_o_cqes) - (resized_cur_pc & new_num_o_cqes_mask);
    if (resized_cqes_at_top > outs_cqes_at_ci)  { /* enough room for all CQEs at CI ? */
      memcpy(resized_cpy_p, cur_cpy_p, outs_cqes_at_ci << LOG2_CQE_SZ);
    } else {
      memcpy(resized_cpy_p, cur_cpy_p, resized_cqes_at_top << LOG2_CQE_SZ);
      resized_cpy_p= (void *)(resized_buf->cqe_buf_base);
      cur_cpy_p= (char*)cur_cpy_p + (resized_cqes_at_top << LOG2_CQE_SZ);
      memcpy(resized_cpy_p, cur_cpy_p, (outs_cqes_at_ci - resized_cqes_at_top) << LOG2_CQE_SZ);
    }
    resized_cur_pc +=  outs_cqes_at_ci;
  }
  if (outs_cqes_at_cq_base > 0) { /* Next copy CQEs at CQ base (wrap around...) */
    cur_cpy_p= (void *) cur_buf->cqe_buf_base ;
    resized_cpy_p= (void *) (resized_buf->cqe_buf_base + ((resized_cur_pc & new_num_o_cqes_mask)<< LOG2_CQE_SZ)) ;
    resized_cqes_at_top= (1U<<resized_buf->log2_num_o_cqes) - (resized_cur_pc & new_num_o_cqes_mask);
    if (resized_cqes_at_top > outs_cqes_at_cq_base)  { /* enough room for all CQEs at base ? */
      memcpy(resized_cpy_p, cur_cpy_p, outs_cqes_at_cq_base << LOG2_CQE_SZ);
    } else {
      memcpy(resized_cpy_p, cur_cpy_p, resized_cqes_at_top << LOG2_CQE_SZ);
      resized_cpy_p= (void *)(resized_buf->cqe_buf_base);
      cur_cpy_p= (char*)cur_cpy_p + (resized_cqes_at_top << LOG2_CQE_SZ);
      memcpy(resized_cpy_p, cur_cpy_p, (outs_cqes_at_cq_base - resized_cqes_at_top) << LOG2_CQE_SZ);
    }
  }

  return;
}


