/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/modversions.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <mtl_types.h>

#include <mosal.h>
#include <8xx_cs_map.h>

#define NUM_OF_CS 8		/* number of cs supported by MPC860 */
#define BR_OFFSET 0x100

#define BR_ATTRIBUTES 0x000000801
#define OR_ATTRIBUTES 0x0000001f8

#define MASK_BR_ACTIVE 0x00000001
#define BR_BASE_MASK   0xffff8000

#define OR_SIZE_MASK   0xffff8000

#define _8XX_MAX_ADDR_VAL   0xffff8000
#define MIN_MAP_SIZE 0x8000

typedef struct {
	u_int32_t br;
	u_int32_t or;
} cs_regs_st;

typedef struct {
	u_int8_t reserved[BR_OFFSET];
	cs_regs_st cs_regs[NUM_OF_CS];
} ppc_regs_st;

bool cs_bit_map[NUM_OF_CS] = { FALSE };

#define GET_CS_ADDR(cs_num)        (virt_immr->cs_regs[(cs_num)].br & ~(MIN_MAP_SIZE - 1))
#define GET_CS_SIZE(cs_num)        (~(virt_immr->cs_regs[cs_num].or & ~(MIN_MAP_SIZE -1)) + 1)
#define IS_CS_ACTIVE(cs_num)       (virt_immr->cs_regs[cs_num].br & MASK_BR_ACTIVE)
#define GET_IMMR()                 ((ppc_regs_st*)ioremap_nocache((get_immr() & 0xffff0000), sizeof(ppc_regs_st)))
#define FREE_IMMR(virt_immr)       iounmap(virt_immr)

/*===============================================static functions definitions====================================*/
static u_int32_t get_immr(void);
static u_int32_t MOSAL_8xx_find_free_segment(ppc_regs_st *, u_int32_t size);

/*======================================module internals implementation==========================================*/

MT_phys_addr_t MOSAL_8xx_cs_map(short cs_num, u_int32_t size)
{
	u_int32_t segment;
	ppc_regs_st *virt_immr;

	virt_immr = GET_IMMR();

	// Check that requested chip selected if free
	if (virt_immr->cs_regs[cs_num].br & MASK_BR_ACTIVE) {
		printk("%s: Selected chip select (cs = %x) allready active.\n",
		       __func__, cs_num);
		FREE_IMMR(virt_immr);
		return (-1);
	}

	// It is possible to map only a minimum of 32K
	if (size < MIN_MAP_SIZE) {
		printk("%s: address size is too small ( 0x%08x < 0x%08x ).\n",
		       __func__, size, MIN_MAP_SIZE);
		FREE_IMMR(virt_immr);
		return (-1);
	}

	segment = MOSAL_8xx_find_free_segment(virt_immr, size);
	if (segment == 0) {
		FREE_IMMR(virt_immr);
		return (-1);
	}

	MTL_TRACE3("%s: found segment(0x%08x) for chipselect %x\n", __func__,
		   segment, cs_num);

	/* initialize chip of ppc to work */
	virt_immr->cs_regs[cs_num].br = segment | BR_ATTRIBUTES;
	virt_immr->cs_regs[cs_num].or = ~(size - 1) | OR_ATTRIBUTES;

	MTL_TRACE4("%s: br = 0x%08x : or = 0x%08x.\n", __func__,
		   virt_immr->cs_regs[cs_num].br,
		   virt_immr->cs_regs[cs_num].or);

	FREE_IMMR(virt_immr);

	cs_bit_map[cs_num] = TRUE;

	return (segment);
}

/*
 *      clear_cs
 */
call_result_t MOSAL_8xx_cs_unmap(short cs_num)
{
	ppc_regs_st *virt_immr;

	virt_immr = GET_IMMR();

	if (cs_bit_map[cs_num] != TRUE)
		MTL_ERROR3
		    ("%s: WARNING : unmapping cs %x, not mapped by MOSAL\n",
		     __func__, cs_num);

	if (!(virt_immr->cs_regs[cs_num].br & MASK_BR_ACTIVE)) {
		MTL_ERROR3("%s: wrong cs %x, cs is already disabled\n",
			   __func__, cs_num);
		FREE_IMMR(virt_immr);
		return (MT_ERROR);
	}

	virt_immr->cs_regs[cs_num].br &= ~MASK_BR_ACTIVE;
	cs_bit_map[cs_num] = FALSE;

	FREE_IMMR(virt_immr);
	return (MT_OK);

}

call_result_t MOSAL_8xx_cs_unmap_all(void)
{
	int count;
	call_result_t ret;

	for (count = 0; count < NUM_OF_CS; count++) ;
	if (cs_bit_map[count] == TRUE) {
		MTL_TRACE3("%s: unmapping cs %x\n", __func__, count);
		ret = MOSAL_8xx_cs_unmap(count);
		if (MT_OK != ret)
			return (ret);
	}

	return (MT_OK);

}

static u_int32_t get_immr(void)
{
	u_int32_t result;

	asm("mfspr 3, 638");
      asm("stw 3, %0":"=m"(result));

	return result;
}

static u_int32_t MOSAL_8xx_find_free_segment(ppc_regs_st * virt_immr,
					     u_int32_t size)
{
	int i, j;
	u_int32_t a1, s1, a2;

	for (i = 0; i < NUM_OF_CS; i++) {
		// Check if chip select is active
		if (!IS_CS_ACTIVE(i))
			continue;

		// Take address and size of current memory map
		a1 = GET_CS_ADDR(i);
		s1 = GET_CS_SIZE(i);

		// Check this region doesn't ovelap
		if ((a1 + s1) == 0)
			continue;

		a2 = _8XX_MAX_ADDR_VAL;

		// Look for next adjacent address
		for (j = 0; j < NUM_OF_CS; j++) {
			int j_addr = GET_CS_ADDR(j);

			if (!IS_CS_ACTIVE(j))
				continue;

			a2 = (j_addr > a1 && j_addr < a2) ? j_addr : a2;

		}

		if ((a2 - a1 - s1) >= size) {
			return (a1 + s1);
		}
	}

	printk("%s: Failed to found segment available (size = %x).\n", __func__,
	       size);

	return (0);

}
