/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/slab.h>
#include <asm/errno.h>
#include <asm/current.h>
#include <asm/page.h>
#include <linux/list.h>
#include <linux/securebits.h>
#include <linux/utsname.h>
#include <linux/highmem.h>
#include <linux/pagemap.h>
#include <mtl_common.h>
#include "mosal_iobuf_imp.h"
#include "mosal_priv.h"

#if !LINUX_KERNEL_2_6
#include <linux/compatmac.h>
#else
#include <asm/tlbflush.h>
#endif


#if defined(VM_HUGETLB)
  #define MT_VM_HUGETLB VM_HUGETLB
#elif defined(VM_BIGPAGE)
  #define MT_VM_HUGETLB VM_BIGPAGE
#else
  #define MT_VM_HUGETLB 0
#endif


#if defined(HPAGE_SHIFT)
  #ifdef HPAGE_SHIFT_DEFAULT
    /* for kernel 2.6.x on ia64 */
    #define MT_HPAGE_SHIFT HPAGE_SHIFT_DEFAULT
  #else
    #define MT_HPAGE_SHIFT HPAGE_SHIFT
  #endif  
#elif defined(BIGPAGE_SHIFT)
  #define MT_HPAGE_SHIFT BIGPAGE_SHIFT
#else
  #define MT_HPAGE_SHIFT 0
#endif

#if defined(HPAGE_SHIFT)
  #ifdef HPAGE_SHIFT_DEFAULT
    /* for kernel 2.6.x on ia64 */
    #define MT_KERNEL_LOWMEM_PAGE_SHIFT HPAGE_SHIFT_DEFAULT
  #else
    #define MT_KERNEL_LOWMEM_PAGE_SHIFT HPAGE_SHIFT
  #endif  
#elif defined(BIGPAGE_SHIFT)
  #define MT_KERNEL_LOWMEM_PAGE_SHIFT BIGPAGE_SHIFT
#else
  #define MT_KERNEL_LOWMEM_PAGE_SHIFT PAGE_SHIFT
#endif  


static unsigned long max_lock_pages = (unsigned long)(-1);
static spinlock_t locked_pages_spl = SPIN_LOCK_UNLOCKED;
static unsigned long num_locked_pages = 0;

unsigned long MOSAL_get_max_lock_pages(void)
{
	unsigned long cur_max_lock_pages;

	spin_lock(&locked_pages_spl);
	cur_max_lock_pages = max_lock_pages;
	spin_unlock(&locked_pages_spl);

	return cur_max_lock_pages;
}


unsigned long MOSAL_set_max_lock_pages(unsigned long new_max)
{
	unsigned long ret;

	spin_lock(&locked_pages_spl);
	if ( new_max <= num_locked_pages ) {
		ret = max_lock_pages;
	}
	else {
		max_lock_pages = new_max;
		ret = new_max;
	}
	spin_unlock(&locked_pages_spl);

	return ret;
}


unsigned long MOSAL_get_cur_lock_pages(void)
{
	unsigned long cur_lock_pages;

	spin_lock(&locked_pages_spl);
	cur_lock_pages = num_locked_pages;
	spin_unlock(&locked_pages_spl);

	return cur_lock_pages;
}


#define mk_filep(num) ((struct file *)(num))


#define STD_VMA_ERR(va, size) \
	MTL_ERROR1(MT_STDFMT( \
	"region not in process's address space. va="VIRT_ADDR_FMT \
                          " size="SIZE_T_FMT), (va), (size)); \
	rc = MT_ENOMEM;
	

#define STD_PERM_ERR(va, size, flags) \
	MTL_ERROR1(MT_STDFMT("permission not granted. va="VIRT_ADDR_FMT \
                          " size="SIZE_T_FMT", required: read(%s),"  \
                          "write(%s) avail: read(%s), write(%s)"), \
                          (va), (size), flag_set_str(iobuf->req_perm, \
			  MOSAL_PERM_READ), \
                          flag_set_str(iobuf->req_perm, MOSAL_PERM_WRITE), \
                          flag_set_str((flags), VM_READ), \
                          flag_set_str((flags), VM_WRITE)); \
	rc = MT_EPERM;



#define flag_set_str(flags, mask) (((flags)&(mask)) ? "yes" : "no")


#define IOBUF_ALLOC_THRESHOLD PAGE_SIZE



static kmem_cache_t *iobuf_cache = NULL;


/*
 *  print_iobuf
 */
static void print_iobuf(struct mosal_iobuf_st *iobuf)
{
	MTL_ERROR1(MT_FLFMT("dump iobuf at %p :"), iobuf);
	MTL_ERROR1("pid=%d\n", current->pid);
	MTL_ERROR1("va = " VIRT_ADDR_FMT "\n", iobuf->va);
	MTL_ERROR1("size = " SIZE_T_FMT "\n", iobuf->size);
	MTL_ERROR1("prot_ctx = %s\n", MOSAL_prot_ctx_str(iobuf->prot_ctx));
	MTL_ERROR1("req_perm = %s\n", MOSAL_perm_str(iobuf->req_perm));
	MTL_ERROR1("page_shift = %d\n", iobuf->page_shift);
	if ( !iobuf->phys_ctg ) {
		MTL_ERROR1("map_list = %p\n", iobuf->map_u.map_list);
	}
	else {
		if ( iobuf->prot_ctx == MOSAL_get_current_prot_ctx() )
			MTL_ERROR1("pa = "PHYS_ADDR_FMT"\n", iobuf->map_u.pa);
		else
			MTL_ERROR1("page = %p\n", iobuf->map_u.page);
	}
	MTL_ERROR1("phys_ctg = %s\n", flag_set_str(iobuf->phys_ctg, 1));
	MTL_ERROR1("need_page_secure = %s\n",
	        flag_set_str(iobuf->need_page_secure, 1));
	MTL_ERROR1("ref_iobuf = %p\n", iobuf->ref_iobuf);
	MTL_ERROR1("file_shared = %s\n", flag_set_str(iobuf->file_shared, 1));
	MTL_ERROR1("file_p = %p\n", iobuf->file_p);
	MTL_ERROR1("file_pgoff = 0x%lx\n", iobuf->file_pgoff);
	MTL_ERROR1("iomem = %s\n", flag_set_str(iobuf->iomem, 1));

	MTL_ERROR1("mlock_ctx = %p\n", iobuf->mlock_ctx);
	MTL_ERROR1("prev = %p\n", iobuf->prev);
	MTL_ERROR1("next = %p\n", iobuf->next);
}


/*
 *  get_chunk
 */
static inline struct page **get_chunk(map_item_t *item, unsigned int *num_p)
{
	if ( item ) {
		*num_p = item->elem_count;
		return item->arr;
	}
	else {
		return NULL;
	}
}



/*
 *  get_first_chunk
 */
static inline struct page **get_first_chunk(struct map_item_st **iter_p,
				     map_list_t *list,
				     unsigned int *num_p)
{
	*iter_p = list->head;

	return get_chunk(*iter_p, num_p);
}


/*
 *  get_next_chunk
 */
static inline struct page **get_next_chunk(struct map_item_st **iter_p,
				    map_list_t *list,
				    unsigned int *num_p)
{
	*iter_p = (*iter_p)->next;

	return get_chunk(*iter_p, num_p);
}




/******************************************************************************
 *  Function: mosal_iobuf_cmp_tpt
 *
 *  Description: compare the tpt of two iobufs
 *
 *  Parameters:
 *       iobuf_1(in) first iobuf object
 *       iobuf_2(in) second iobuf object
 *
 *  Returns:
 *    0    tpt's are equal
 *    != 0 tpt's differ
 *  Notes: The function compares iobufs belonging to the same protection context.
 *         If the protection context differs it returns -1; If the protection
 *         context is kernel than only the va and size of both iobus are
 *         compared. It is assumed that there can be no two vitual addresses
 *         that map to the same physical address. In user space it could be
 *         that the sizes of the buffers are not equal bu the number of pages
 *         is equal. In that case if the physical addresses of the two buffers
 *         are equal then the iobufs are considered equal
 *        
 ******************************************************************************/
static int mosal_iobuf_cmp_tpt(MOSAL_iobuf_t iobuf1, MOSAL_iobuf_t iobuf2)
{
	struct map_item_st *iter1, *iter2;
	struct page **pages1, **pages2;
	unsigned int num1, num2, i;

	/* sanity check */
	if ( !iobuf1 || !iobuf2 ) return -1;

	if ( iobuf1->prot_ctx != iobuf2->prot_ctx ) {
		/* iobufs do not belog to the same protection context i.e
		   one is user spcae and the other is kernel space */
		return -1;
	}
	else if ( iobuf1->prot_ctx == MOSAL_get_kernel_prot_ctx() ) {
		/* iobufs define kernel addresses */
		if ((iobuf1->va!=iobuf2->va) || (iobuf1->size!=iobuf2->size)) {
			return -1;
		}
		else {
			return 0;
		}
	}
	else {
		/* iobufs define user space addresses */
		if ( iobuf1->page_shift != iobuf2->page_shift ) {
			return -1;
		}
		if ( iobuf1->size != iobuf2->size ) {
			return -1;
		}

		if ( iobuf1->phys_ctg != iobuf2->phys_ctg ) {
			return -1;
		}
		else if ( iobuf1->phys_ctg ) {
			if ( iobuf1->map_u.pa != iobuf2->map_u.pa ) {
				return -1;
			}
			else
				return 0;
		}

		for (pages1 = get_first_chunk(&iter1,
		       		        iobuf1->map_u.map_list, &num1),
		     pages2 = get_first_chunk(&iter2,
		       		        iobuf2->map_u.map_list, &num2);
		     pages1 && pages2;
		     pages1 = get_next_chunk(&iter1,
		       		       iobuf1->map_u.map_list, &num1),
		     pages2 = get_next_chunk(&iter2,
		       		      iobuf2->map_u.map_list, &num2)) {
			if ( num1 != num2 ) {
				/* should not happen in ths implementation */
				return -1;
			}
			for ( i=0; i<num1; ++i ) {
				if ( pages1[i] != pages2[i] ) {
					return -1;
				}
			}
		}
		if ( !pages1 && !pages2 ) 
			return 0;
		else
			return -1;
	}
}



/*
 *  perm_ok
 */
static int inline perm_ok(MOSAL_mem_perm_t req_perm, unsigned long vm_flags)
{
	int mosal_wr = req_perm & MOSAL_PERM_WRITE;
	int mosal_rd = req_perm & MOSAL_PERM_READ;
	int vm_wr = vm_flags & VM_WRITE;
	int vm_rd = vm_flags & VM_READ;

	return (!mosal_wr && !mosal_rd) ||
	    (vm_wr && !mosal_rd) || (!mosal_wr && vm_rd) || (vm_wr && vm_rd);
}


#define BUF_PAGES(va, len, page_size)  (((va) + (len) + \
		(page_size) - 1)/(page_size) - (va)/(page_size))


/*
 *  large_kva
 */
static inline int large_kva(void)
{
	if ( sizeof(void *) < 8 ) {
		return 0;
	}
	else {
		return 1;
	}
}


/*
 *  get_map_item_size
 */
static inline unsigned int  get_map_item_size(unsigned int elem_count)
{
	return sizeof(map_item_t) + sizeof(struct page *) * elem_count;
}


/*
 *  map_item_alloc
 */
static inline map_item_t *map_item_alloc(unsigned int elem_count)
{
	unsigned int size = get_map_item_size(elem_count);
	map_item_t *buf;

	if ( size <= IOBUF_ALLOC_THRESHOLD ) {
		buf = (map_item_t *)MALLOC(size);
	}
	else {
		buf = (map_item_t *)VMALLOC(size);
	}
	return buf;
}



/*
 *  map_item_free
 */
static inline void map_item_free(map_item_t *item)
{
	unsigned int size = get_map_item_size(item->elem_count);

	if ( size <= IOBUF_ALLOC_THRESHOLD ) {
		FREE(item);
	}
	else {
		VFREE(item);
	}
}


/*
 *  free_map_list
 */
static void free_map_list(map_list_t *list)
{
	map_item_t *tmp;

	if (!list)
		return;
	if (atomic_dec_and_test(&list->ref_count)) {
		MTL_DEBUG1(MT_FLFMT("ref count reached zero. really freeing"));
		while (list->head) {
			tmp = list->head->next;
			map_item_free(list->head);
			list->head = tmp;
		}
		FREE(list);
	} else {
		MTL_DEBUG1(MT_FLFMT("just decremented ref count"));
	}
}

/*
 *  alloc_map_list
 */
static map_list_t *alloc_map_list(unsigned int num_elem)
{
	map_list_t *list;
	map_item_t *tmp;
	unsigned int max_elem_in_item;
	unsigned int elem_count;
	
	list = TMALLOC(map_list_t);
	if (!list)
		return NULL;
	list->head = NULL;
	atomic_set(&list->ref_count, 1);

	if ( large_kva() ) {
		max_elem_in_item = (unsigned int)(-1);
	}
	else {
		max_elem_in_item = ( IOBUF_ALLOC_THRESHOLD - 
				     sizeof(map_item_t) ) /
				     sizeof(struct page *);
	}

	elem_count = num_elem % max_elem_in_item;
	if ( elem_count == 0 ) elem_count=max_elem_in_item;
	
	while ( num_elem > 0 ) {
		tmp = map_item_alloc(elem_count);
		if ( !tmp ) {
			free_map_list(list);
			return NULL;
		}
		tmp->elem_count = elem_count;
		tmp->next = list->head;
		num_elem -= elem_count;
		list->head = tmp;
		elem_count = max_elem_in_item;
	}

	return list;
}



/*
 *  is_flag_set
 */
static inline int is_flag_set(unsigned long flags, unsigned long flag_bit)
{
  if ( flags & flag_bit ) return 1;
  else return 0;
}


/*
 *  validate_flag
 */
static int validate_flag(int prev, unsigned long flags, unsigned long flag_bit)
{
	if ( prev == -1 ) {
		return is_flag_set(flags, flag_bit);
	}
	else if ( prev == -2 ) {
		return -2;
	}
	else if ( prev != is_flag_set(flags, flag_bit) ) {
		return -2;
	}

	return 0;
}


/*
 *  validate_filep
 */
static struct file *validate_filep(struct file *prev,
				   struct vm_area_struct *vma,
				   unsigned long *pgoff_p)
{
	if ( prev == mk_filep(-1) ) {
		*pgoff_p = vma->vm_pgoff;
		return vma->vm_file;
	}
	else if ( prev == mk_filep(-2) ) {
		return mk_filep(-2);
	}
	else if ( prev != vma->vm_file ) {
		return mk_filep(-2);
	}

	return vma->vm_file;
}


/*
 *  is_shared_mapping
 */
static inline int is_shared_mapping(struct mosal_iobuf_st *iobuf)
{
	if ( iobuf->file_p &&
	     iobuf->ref_iobuf &&
	     iobuf->file_p == iobuf->ref_iobuf->file_p &&
	     iobuf->file_pgoff == iobuf->ref_iobuf->file_pgoff )
		return 1;
	else
		return 0;
}



/*
 *  current_populate_iobuf
 */
static call_result_t current_populate_iobuf(struct mosal_iobuf_st *iobuf)
{
	call_result_t rc=MT_OK;
	MT_virt_addr_t addr = iobuf->va, end = iobuf->va + iobuf->size;
	struct vm_area_struct *vma;
	int big_pages = -1;
	int iomem = -1;
	struct file *file_p = mk_filep(-1);
	unsigned long pgoff;

	down_read(&current->mm->mmap_sem);
	/* check permissions */
	while ( addr < end ) {
		vma = find_vma(current->mm, addr);
		if ( !vma || (addr<vma->vm_start) ) {
			STD_VMA_ERR(iobuf->va, iobuf->size);
			break;
		}
		else if ( !perm_ok(iobuf->req_perm, vma->vm_flags) ) {
			STD_PERM_ERR(iobuf->va, iobuf->size, vma->vm_flags);
			break;
		}
		else {
			/* we don't want to register a region covering
			   VMAs of different page sizes */
			big_pages =
				validate_flag(big_pages, vma->vm_flags,
					       MT_VM_HUGETLB);
			if ( big_pages == -2 ) {
				MTL_ERROR1(MT_STDFMT("mixed page sizes:"
						"va="VIRT_ADDR_FMT
						", size="SIZE_T_FMT),
					        iobuf->va, iobuf->size);
				rc = MT_ENOMEM;
				break;
			}

			iomem =
				validate_flag(iomem, vma->vm_flags, VM_IO);
			if ( iomem == -2 ) {
				MTL_ERROR1(MT_STDFMT("mixed vma flags:"
						"va="VIRT_ADDR_FMT
						", size="SIZE_T_FMT),
					        iobuf->va, iobuf->size);
				rc = MT_ENOMEM;
				break;
			}

			file_p = validate_filep(file_p, vma, &pgoff);

			addr = vma->vm_end;
		}
	} /* end of while loop */
	up_read(&current->mm->mmap_sem);

	if ( rc != MT_OK ) {
		goto ex_func;
	}

	if ( big_pages ) {
		iobuf->page_shift = MT_HPAGE_SHIFT;
	}
	else {
		iobuf->page_shift = PAGE_SHIFT;
	}

	if ( file_p == mk_filep(-2) ) {
		iobuf->file_p = NULL;
		iobuf->file_pgoff = 0;
	}
	else {
		iobuf->file_p = file_p;
		iobuf->file_pgoff = pgoff;
	}

	iobuf->file_shared = is_shared_mapping(iobuf);

	if ( iomem ) {
		iobuf->phys_ctg = 1;
	}
	else {
		iobuf->phys_ctg = 0;
	}
	iobuf->iomem = iomem;

	rc = MT_OK;
	goto ex_func;

ex_func:
	return rc;
}


/*
 *  kernel_populate_iobuf
 */
static call_result_t kernel_populate_iobuf(struct mosal_iobuf_st *iobuf)
{
	if ( vmalloced_addr(iobuf->va) ) {
		iobuf->page_shift = PAGE_SHIFT;
		iobuf->phys_ctg = 0;
	}
	else {
		iobuf->page_shift = MT_KERNEL_LOWMEM_PAGE_SHIFT;
		iobuf->phys_ctg = 1;
	}
	iobuf->file_shared = 0;
	return MT_OK;
}

/*
 *  populate_iobuf
 */
static call_result_t populate_iobuf(struct mosal_iobuf_st *new_iobuf)
{
	if ( new_iobuf->prot_ctx == MOSAL_get_current_prot_ctx() ) {
		return current_populate_iobuf(new_iobuf);
	}
	else {
		return kernel_populate_iobuf(new_iobuf);
	}
}



/*
 *  mlock_iobuf
 */
static call_result_t mlock_iobuf(struct mosal_iobuf_st *iobuf)
{
	if ( iobuf->prot_ctx == MOSAL_get_kernel_prot_ctx() ) {
		return MT_OK;
	}
	else {
		if ( !iobuf->file_shared ) {
			return MOSAL_mlock_iobuf(iobuf->va,
						 iobuf->size,
						 iobuf,
						 iobuf->page_shift);
		}
		else {
			/* for shared region we do not mlock to save
		           the time needed for paging in */
			return MT_OK;
		}
	}

}


/*
 *  munlock_iobuf
 */
static call_result_t munlock_iobuf(struct mosal_iobuf_st *iobuf)
{
	if ( iobuf->prot_ctx == MOSAL_get_kernel_prot_ctx() ) {
		return MT_OK;
	}
	else {
		if ( !iobuf->file_shared ) {
			return MOSAL_munlock_iobuf(iobuf->va,
						   iobuf->size,
						   iobuf,
						   iobuf->page_shift);
		}
		else
			return MT_OK;
	}
}



static int mosal_get_user_pages(MT_virt_addr_t va,
				                struct page **pages,
								unsigned int num,
								unsigned int page_shift,
								int write)
{
	int ret;
	MT_virt_addr_t addr;
	int i, j, ok;


	spin_lock(&locked_pages_spl);
	if ( (num_locked_pages+num) > max_lock_pages ) {
		ret = -1;
	}
	else {
		num_locked_pages += num;
		ret = 0;
	}
	spin_unlock(&locked_pages_spl);

	if ( ret ) {
		return -1;
	}

    down_read(&current->mm->mmap_sem);
	if ( page_shift == PAGE_SHIFT ) {
		ret = get_user_pages(current, current->mm, va,
					     num, write, 0, pages, NULL);
	}
	else {
		ret = 0;
		for ( addr=va, i=0; i<num; ++i, addr+=MT_HPAGE_SHIFT ) {
			ok = get_user_pages(current, current->mm, addr,
					     1, write, 0, &pages[i], NULL);
			if ( ok == 1 ) ret += ok;
			else {
				for ( j=0; j<i; ++j )
					page_cache_release(pages[j]);
				ret = -1;
				break;
			}

		}
	}
    up_read(&current->mm->mmap_sem);

	if ( ret != num ) {
		spin_lock(&locked_pages_spl);
		num_locked_pages -= num;
		spin_unlock(&locked_pages_spl);
	}

	return ret;
}


/*
 *  current_get_pages
 */
static call_result_t current_get_pages(struct mosal_iobuf_st *iobuf)
{
	struct page **pages, **failed_pages;
	unsigned int num, i;
	int ret;
	struct map_item_st *iter;
	call_result_t rc;
	MT_virt_addr_t va=iobuf->va;

	if ( iobuf->iomem ) {
		rc = MOSAL_virt_to_phys_ex(iobuf->prot_ctx,
					   iobuf->va,
					   NULL,
					   &iobuf->map_u.pa);
		if ( rc != MT_OK ) {
			return rc;
		}
		else
			return MT_OK;
	}

	for ( pages = get_first_chunk(&iter, iobuf->map_u.map_list, &num);
	      pages;
	      pages = get_next_chunk(&iter, iobuf->map_u.map_list, &num) ) {
		if ( !iobuf->file_shared ) {
			ret = mosal_get_user_pages(va, pages, num, iobuf->page_shift,
									    (iobuf->req_perm & MOSAL_PERM_WRITE) ? 1 : 0);
			if ( ret != num ) {
				MTL_ERROR1(MT_STDFMT("num=%d, page_shift=%d"),
						     num, iobuf->page_shift);
				failed_pages = pages;
				goto exit_err;
			}


			if ( !iobuf->need_page_secure ) {
				for ( i=0; i<num; ++i ) {
					page_cache_release(pages[i]);
				}
			}
			va = va + (num << iobuf->page_shift);
		}
		else {
			if ( iobuf->need_page_secure ) {
				for ( i=0; i<num; ++i ) {
					page_cache_get(pages[i]);
				}
			}
		}
	}

	return MT_OK;

exit_err:
	if ( iobuf->need_page_secure ) {
		for ( pages = get_first_chunk(&iter,
					       iobuf->map_u.map_list, &num);
		      pages;
		      pages = get_next_chunk(&iter,
					      iobuf->map_u.map_list, &num) ) {
			if ( pages == failed_pages )
				break;
			for ( i=0; i<num; ++i ) {
				page_cache_release(pages[i]);
			}
			spin_lock(&locked_pages_spl);
			num_locked_pages -= num;
			spin_unlock(&locked_pages_spl);
		}
	}
	return MT_ENOMEM;
}



/*
 *  kernel_get_pages
 */
static call_result_t kernel_get_pages(struct mosal_iobuf_st *iobuf)
{
	unsigned long addr, page_size;
	struct page **pages;
	unsigned int num, i;
	struct map_item_st *iter;

	if ( iobuf->phys_ctg ) {
		iobuf->map_u.page = virt_to_page(iobuf->va);
		if ( iobuf->map_u.page ) {
			return MT_OK;
		}
		else {
			return MT_ENOMEM;
		}
	}
	else {
		addr = iobuf->va;
		page_size = 1 << iobuf->page_shift;
		for ( pages = get_first_chunk(&iter,
					       iobuf->map_u.map_list, &num);
		      pages;
		      pages = get_next_chunk(&iter,
					      iobuf->map_u.map_list, &num) ) {
			for ( i=0; i<num; ++i, addr+=page_size ) {
				pages[i] = vmalloc_to_page((void *)addr);
				if ( !pages[i] ) {
					return MT_ENOMEM;
				}
			}
		}
	}
	return MT_OK;
}



/*
 *  current_put_pages
 */
static call_result_t current_put_pages(struct mosal_iobuf_st *iobuf)
{
	struct page **pages;
	unsigned int num, i;
	struct map_item_st *iter;


	if ( iobuf->iomem || !iobuf->need_page_secure )
		return MT_OK;

	for ( pages = get_first_chunk(&iter, iobuf->map_u.map_list, &num);
	      pages;
	      pages = get_next_chunk(&iter, iobuf->map_u.map_list, &num) ) {
		for ( i=0; i<num; ++i ) {
/*            set_page_dirty_lock(pages[i]); */
			page_cache_release(pages[i]);
		}
		spin_lock(&locked_pages_spl);
		num_locked_pages -= num;
		spin_unlock(&locked_pages_spl);
	}



	return MT_OK;
}



/*
 *  kernel_get_pages
 */
static inline call_result_t kernel_put_pages(struct mosal_iobuf_st *iobuf)
{
	return MT_OK;
}



/*
 *  mosal_get_pages
 */
static inline call_result_t mosal_get_pages(struct mosal_iobuf_st *iobuf)
{
	if ( iobuf->prot_ctx == MOSAL_get_current_prot_ctx() ) {
		return current_get_pages(iobuf);
	}
	else {
		return kernel_get_pages(iobuf);
	}
}



/*
 *  mosal_put_pages
 */
static call_result_t mosal_put_pages(struct mosal_iobuf_st *iobuf)
{
	if ( iobuf->prot_ctx == MOSAL_get_current_prot_ctx() ) {
		return current_put_pages(iobuf);
	}
	else {
		return kernel_put_pages(iobuf);
	}
}




/*
 *  determine_page_secure
 */
static void determine_page_secure(struct mosal_iobuf_st *iobuf)
{
	if ( iobuf->prot_ctx == MOSAL_get_current_prot_ctx() &&
	     iobuf->page_shift != MT_HPAGE_SHIFT ) {
		iobuf->need_page_secure = 1;
	}
	else {
		iobuf->need_page_secure = 0;
	}
}




/*
 *  share_map_list
 */
static inline void share_map_list(struct mosal_iobuf_st *iobuf)
{
	atomic_inc(&iobuf->ref_iobuf->map_u.map_list->ref_count);
	iobuf->map_u.map_list = iobuf->ref_iobuf->map_u.map_list;
}

/*
 *  verify_sharability
 */
static call_result_t verify_sharability(MOSAL_prot_ctx_t prot_ctx,
					MOSAL_iobuf_t ref_iobuf,
					MT_virt_addr_t va,
					MT_size_t size)
{
	unsigned long mask;

	if ( !ref_iobuf )
		return MT_OK;

	if ( prot_ctx == MOSAL_get_kernel_prot_ctx() ) {
		/* for kernel memory address and size must be identical */
		if ( (va!=ref_iobuf->va) || (size!=ref_iobuf->size) )
			return MT_EINVAL;
		else
			return MT_OK;
	}
	else {
		mask = ((unsigned long)1 << ref_iobuf->page_shift) - 1;

		if ( ((va&mask) != (ref_iobuf->va&mask)) || (size!=ref_iobuf->size) )
			return MT_EINVAL;
		else
			return MT_OK;
	}
}


/****** global functions ***********************************/

/*
 *  MOSAL_iobuf_register
 */
call_result_t
#ifndef MTL_TRACK_ALLOC
	MOSAL_iobuf_register
#else
	MOSAL_iobuf_register_memtrack
#endif
		(MOSAL_prot_ctx_t prot_ctx,
		 MT_virt_addr_t va,
		 MT_size_t size,
                 MOSAL_mem_perm_t req_perm,
		 MOSAL_iobuf_t ref_iobuf,
		 MOSAL_iobuf_t * iobuf_p)
{
	call_result_t rc;
	MOSAL_iobuf_t new_iobuf;
	unsigned int page_size, num_pages;

	/*** check args ***/

	/* protection context */
	if ( ((prot_ctx != MOSAL_get_current_prot_ctx()) &&
	     (prot_ctx != MOSAL_get_kernel_prot_ctx())) ) {
		return MT_EINVAL;
	}

	/* va and size */
	if (va == VA_NULL || size == 0)
		return MT_EINVAL;

	/* iobuf_p */
	if ( !iobuf_p )
		return MT_EINVAL;

	/* in case of shared registration we want to verify that the
	 * offset from the start of the page and the size of the two
	 * iobufs are identical */
	rc = verify_sharability(prot_ctx, ref_iobuf, va, size);
	if ( rc != MT_OK ) return rc;


	/* create object */
	new_iobuf = 
	    (struct mosal_iobuf_st *)kmem_cache_alloc(iobuf_cache, GFP_KERNEL);
	if ( !new_iobuf ) {
		rc = MT_EAGAIN;
		goto ex_iobuf_register;
	}
	memset(new_iobuf, 0, sizeof(*new_iobuf));


	MTL_DEBUG1(MT_STDFMT("prot_ctx=%s, va="VIRT_ADDR_FMT ", size="
			SIZE_T_FMT ", req_perm=%s, ref_iobuf=%p, iobuf=%p"),
		        MOSAL_prot_ctx_str(prot_ctx), va, size,
		        MOSAL_perm_str(req_perm), ref_iobuf, new_iobuf);

	new_iobuf->va = va;
	new_iobuf->size = size;
	new_iobuf->prot_ctx = prot_ctx;
	new_iobuf->req_perm = req_perm;
	new_iobuf->ref_iobuf = ref_iobuf;

	/* build values into the object */
	rc = populate_iobuf(new_iobuf);
	if ( rc != MT_OK ) {
		goto ex_free_iobuf;
	}

	/* allocate map list
	   physically contigous buffers and shared memory do not
	   need map list */
	if ( !new_iobuf->phys_ctg && !new_iobuf->file_shared ) {
		page_size = 1 << new_iobuf->page_shift;
		num_pages = BUF_PAGES(new_iobuf->va, new_iobuf->size, page_size);
		MTL_DEBUG1(MT_STDFMT("num_pages=%d"), num_pages);
		new_iobuf->map_u.map_list = alloc_map_list(num_pages);
		if ( !new_iobuf->map_u.map_list ) {
			MTL_ERROR1(MT_STDFMT("allocation failed %d elems"),
				    num_pages);
			rc = MT_EAGAIN;
			goto ex_free_iobuf;
		}
	}

	/* check if we should share the map list */
	if ( new_iobuf->file_shared && !ref_iobuf->phys_ctg ) {
		share_map_list(new_iobuf);
	}

	/* lock the pages */
	rc = mlock_iobuf(new_iobuf);
	if ( rc != MT_OK ) {
		MTL_ERROR1(MT_STDFMT("MOSAL_mlock_iobuf failed - %s"),
			   mtl_strerror_sym(rc));
		goto ex_free_maplist;
	}

	/* determine if the pages have to be secured
	   i.e call get/put_page for them
	   update a value in the object to reflect that */
	determine_page_secure(new_iobuf);

	/* get the translation (put values in the page descriptor pointers
	   and optionally call get_page */
	rc = mosal_get_pages(new_iobuf);
	if ( rc != MT_OK ) {
		MTL_ERROR1(MT_STDFMT("mosal_get_pages failed - %s"),
			   mtl_strerror_sym(rc));
		goto ex_munlock;
	}

	/* if the caller required shared registration but the
	   could not share a file, we still compare the pages
	   one by one to see if pages can be shared and only if
	   there is not a match the function fails */
	if ( ref_iobuf && (ref_iobuf->phys_ctg || !new_iobuf->file_shared) ) {
		/* compare the hard way */
		if ( mosal_iobuf_cmp_tpt(new_iobuf, ref_iobuf) != 0 ) {
			MTL_ERROR2(MT_STDFMT("TPT mismatch when trying to share ref_iobuf %p with new_iobuf %p:"),
				   ref_iobuf, new_iobuf);
			print_iobuf(ref_iobuf);
			rc = MT_EACCES;
			goto ex_put_pages;
		}
	}

	*iobuf_p = new_iobuf;
	rc = MT_OK;
	goto ex_iobuf_register;

ex_put_pages:
	mosal_put_pages(new_iobuf);
ex_munlock:
	munlock_iobuf(new_iobuf);
ex_free_maplist:
	print_iobuf(new_iobuf);
	if ( !new_iobuf->phys_ctg ) {
		free_map_list(new_iobuf->map_u.map_list);
	}
ex_free_iobuf:
	kmem_cache_free(iobuf_cache, new_iobuf);
ex_iobuf_register:
	return rc;
}

/*
 *  MOSAL_iobuf_deregister
 */
call_result_t
#ifndef MTL_TRACK_ALLOC
	MOSAL_iobuf_deregister
#else
	MOSAL_iobuf_deregister_memtrack
#endif
		(MOSAL_iobuf_t iobuf)
{
	if ( !iobuf ) return MT_EINVAL;

	mosal_put_pages(iobuf);
	munlock_iobuf(iobuf);
	if ( !iobuf->phys_ctg ) {
		free_map_list(iobuf->map_u.map_list);
	}
	kmem_cache_free(iobuf_cache, iobuf);
	return MT_OK;
}

/*
 *  MOSAL_iobuf_get_tpt
 */
call_result_t MOSAL_iobuf_get_tpt(MOSAL_iobuf_t iobuf,
				  u_int32_t npages,
				  MT_phys_addr_t * pa_arr,
				  u_int32_t * page_size_p,
				  u_int32_t * act_table_sz_p)
{
	return MT_ENOSYS;
}




/*
 *  MOSAL_iobuf_get_tpt_seg
 */
call_result_t MOSAL_iobuf_get_tpt_seg(MOSAL_iobuf_t iobuf,
				      MOSAL_iobuf_iter_t * iterator_p,
				      MT_size_t n_pages_in,
				      MT_size_t * n_pages_out_p,
				      MT_phys_addr_t * page_tbl_p)
{
	unsigned int total_left, i, j, n, nr_pages, page_size;
	MT_phys_addr_t pa;

	/* sanity check */
	if ( !iobuf ||
	     !iterator_p ||
	     !n_pages_out_p ||
	     !page_tbl_p ) return MT_EINVAL;


	MTL_DEBUG1(MT_STDFMT("prot_ctx=%s, va="VIRT_ADDR_FMT ", size="
			SIZE_T_FMT ", req_perm=%s, ref_iobuf=%p, iobuf=%p"
			", n_pages_in"SIZE_T_FMT),
		        MOSAL_prot_ctx_str(iobuf->prot_ctx), iobuf->va,
					   iobuf->size,
		                           MOSAL_perm_str(iobuf->req_perm),
					   iobuf->ref_iobuf,
					   iobuf, n_pages_in);

	page_size = 1 << iobuf->page_shift;
	nr_pages = BUF_PAGES(iobuf->va, iobuf->size, page_size);
	if ( iobuf->phys_ctg ) {
		 /* number of translations not yet provided */
		total_left = nr_pages - iterator_p->elem_idx;
		n = total_left <= n_pages_in ? total_left : n_pages_in;
		if ( iobuf->prot_ctx == MOSAL_get_current_prot_ctx() ) {
			pa = iobuf->map_u.pa + iterator_p->elem_idx * page_size;
		}
		else {
			pa = mosal_page_to_phys(iobuf->map_u.page, iobuf->iomem);
		}
		for ( i=0; i<n; ++i, pa += page_size ) {
			page_tbl_p[i] = pa;
		}
		iterator_p->elem_idx += n;
		*n_pages_out_p = n;
	}
	else {
	  struct page *page;
	  j = 0;
	  *n_pages_out_p = 0;
	  while ( iterator_p->item_p ) {
	     /* total left in this item */
	    total_left = iterator_p->item_p->elem_count - iterator_p->elem_idx;
	    n = total_left <= n_pages_in ? total_left : n_pages_in;
	    for ( i=0; i<n; ++i, ++iterator_p->elem_idx, ++j ) {
	      page = iterator_p->item_p->arr[iterator_p->elem_idx];
	      pa = mosal_page_to_phys(page, 0);
	      MTL_DEBUG1(MT_FLFMT("pa="PHYS_ADDR_FMT), pa);
	      page_tbl_p[j] = pa;
	    }
	    (*n_pages_out_p) += n;
	    if ( iterator_p->elem_idx < iterator_p->item_p->elem_count ) {
	      break;
	    }
	    else {
	      iterator_p->item_p = iterator_p->item_p->next;
	      iterator_p->elem_idx = 0;
	      n_pages_in -= n;
	    }
	  }
	}
	return MT_OK;
}

/*
 *  MOSAL_iobuf_iter_init
 */
call_result_t MOSAL_iobuf_iter_init(MOSAL_iobuf_t iobuf,
				    MOSAL_iobuf_iter_t * iterator_p)
{
	/* sanity check */
	if ( !iobuf ) return MT_EINVAL;

	if ( !iobuf->phys_ctg ) {
	  iterator_p->item_p = iobuf->map_u.map_list->head;
	}
	iterator_p->elem_idx = 0;
	return MT_OK;
}


/*
 *  MOSAL_iobuf_get_props
 */
call_result_t MOSAL_iobuf_get_props(MOSAL_iobuf_t iobuf,
				    MOSAL_iobuf_props_t * props_p)
{
	unsigned int nr_pages;

	if ( !iobuf ) {
		return MT_EINVAL;
	}

	nr_pages = BUF_PAGES(iobuf->va, iobuf->size, 1<<iobuf->page_shift);

	props_p->va = iobuf->va;
	props_p->size = iobuf->size;
	props_p->prot_ctx = iobuf->prot_ctx;
	props_p->nr_pages = nr_pages;
	props_p->page_shift = iobuf->page_shift;

	return MT_OK;
}


/*
 *  MOSAL_iobuf_init
 */
call_result_t MOSAL_iobuf_init(void)
{
	iobuf_cache = kmem_cache_create("MOSAL_iobuf_t",
					sizeof(struct mosal_iobuf_st),
					0,
					SLAB_HWCACHE_ALIGN,
					NULL,
					NULL);
	if ( !iobuf_cache ) {
		MTL_ERROR1(MT_STDFMT("failed to allocate cache"));
		return MT_EAGAIN;
	}

	return MT_OK;
}

void MOSAL_iobuf_cleanup(void)
{
	if ( kmem_cache_destroy(iobuf_cache) ) {
		MTL_ERROR1(MT_STDFMT
			   ("MOSAL_iobuf_t cache destroy failed -"
			    " probably a memory leak"));
	}
}
