/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <linux/kernel.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/version.h>
#include <linux/autoconf.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/linkage.h>
#include <linux/mm.h>
#include <linux/mman.h>
#if defined(__i386__) || defined(__ia64__)
#include <linux/rwsem.h>
#endif
#include <linux/pci.h>
#include <asm/pgtable.h>
#include <linux/highmem.h>

#include "mosal_priv.h"

#if !LINUX_KERNEL_2_6
#include <linux/wrapper.h>
#endif

/****************  *********************/
#define MOSAL_DO_MUNMAP(mm,virt,size) ({                                       \
								int rc = 0;                                                            \
								if (mm!=NULL) {                                                        \
																down_write(&mm->mmap_sem);                                     \
																rc = do_munmap(mm,virt,size MT_UNMAP_SUFFIX);                  \
																up_write(&mm->mmap_sem);                                       \
								}                                                                      \
								rc;                                                                    \
})


#ifdef EXTRA_REMAP_ARG
#define REMAP_PREFIX vma,
#else
#define REMAP_PREFIX
#endif

/*
 *  in_address_space - good for user space only
 */
static inline int in_address_space(MT_virt_addr_t va)
{
	struct vm_area_struct *vma;
	int rc;

	down_read(&current->mm->mmap_sem);
	vma = find_vma(current->mm, va);
	if (!vma || (va < vma->vm_start))
		rc = 0;
	else
		rc = 1;
	up_read(&current->mm->mmap_sem);
	return rc;
}



static call_result_t mt_virt_to_phys(const MT_virt_addr_t va,
									 struct page ** page_pp,
									 MT_phys_addr_t * pa_p)
{
#if LINUX_VERSION_CODE >=  KERNEL_VERSION(2,6,11)
	pgd_t *pgd;
	pud_t *pud;
	pmd_t *pmd;
	pte_t *ptep, pte;
	unsigned long pfn;
	struct page *page;
	struct mm_struct *mm=current->mm;

	pgd = pgd_offset(mm, va);
	if (pgd_none(*pgd) || unlikely(pgd_bad(*pgd)))
		goto out;

	pud = pud_offset(pgd, va);
	if (pud_none(*pud) || unlikely(pud_bad(*pud)))
		goto out;
	
	pmd = pmd_offset(pud, va);
	if (pmd_none(*pmd) || unlikely(pmd_bad(*pmd)))
		goto out;

	ptep = pte_offset_map(pmd, va);
	if (!ptep)
		goto out;

	pte = *ptep;
	pte_unmap(ptep);
	if (pte_present(pte)) {
		pfn = pte_pfn(pte);
		if (pfn_valid(pfn)) {
			page = pfn_to_page(pfn);
			*pa_p = page_to_phys(page);
			if ( page_pp )
				*page_pp = page;
			return MT_OK;
		}
	}

out:
	return MT_ENOMEM;
#else
	return MT_ENOSYS;
#endif
}

call_result_t MOSAL_virt_to_phys(MOSAL_prot_ctx_t prot_ctx,
				 const MT_virt_addr_t va, MT_phys_addr_t * pa_p)
{
	return MOSAL_virt_to_phys_ex(prot_ctx, va, NULL, pa_p);
}


call_result_t MOSAL_virt_to_phys_ex(const MOSAL_prot_ctx_t prot_ctx,
				    const MT_virt_addr_t va,
				    struct page ** page_pp,
				    MT_phys_addr_t * pa_p)
{
	call_result_t rc;
	struct vm_area_struct *vma;
	struct page *page;
	int ret;
	MT_phys_addr_t pa;

	if ( prot_ctx == MOSAL_get_current_prot_ctx() ) {
		down_read(&current->mm->mmap_sem);
		vma = find_vma(current->mm, va);
		if ( vma && (vma->vm_start <= va) ) {
			ret = get_user_pages(current, current->mm, va, 1, 0, 0, &page, NULL);
			if ( ret == 1 ) {
				pa = page_to_phys(page);
				*pa_p = pa | (va & (PAGE_SIZE-1));
				if ( page_pp )
					*page_pp = page;
				put_page(page);
				rc = MT_OK;
			}
			else if ( vma->vm_flags & VM_IO ) {
				if ( vma->vm_private_data ) {
					if ( !page_pp ) {
						*pa_p = (MT_phys_addr_t)(unsigned long)(vma->vm_private_data);
						rc = MT_OK;
					}
					else
						rc = MT_EINVAL;
				}
				else
					rc = mt_virt_to_phys(va, page_pp, pa_p);
			}
			else
				rc = MT_ENOMEM;
		}
		else
			rc = MT_ENOMEM;
		up_read(&current->mm->mmap_sem);
	}
	else if ( prot_ctx == MOSAL_get_kernel_prot_ctx() ) {
		if ( (va>=VMALLOC_START) && (va<VMALLOC_END) ) {
			page = vmalloc_to_page((void *)va);
			if ( page ) {
				pa = page_to_phys(page);
				*pa_p = pa | (va & (PAGE_SIZE-1));
				if ( page_pp )
					*page_pp = page;
				rc = MT_OK;
			}
			else
				rc = MT_ENOMEM;

		}
		else if ( (va>=PAGE_OFFSET) && ((void *)va<high_memory) ) {
			page = virt_to_page(va);
			if ( page ) {
				pa = page_to_phys(page);
				*pa_p = pa | (va & (PAGE_SIZE-1));
				if ( page_pp )
					*page_pp = page;
				rc = MT_OK;
			}
			else
				rc = MT_ENOMEM;
		}
		else
			rc = MT_ENOMEM;
	}
	else
		rc = MT_EINVAL;

	return rc;
}


/*
 *  MOSAL_is_iomem
 */
call_result_t MOSAL_is_iomem(MOSAL_protection_ctx_t prot_ctx,
							 MT_virt_addr_t va,
							 MT_size_t size,
							 int *iomem_p)
{
	call_result_t rc = MT_OK;
	struct vm_area_struct *vma;
	unsigned long start=va, sz=size;
	int tmp=-1;
	int break_while = 0;

	if ( prot_ctx == MOSAL_get_kernel_prot_ctx() ) {
		*iomem_p = 0;
		return MT_OK;
	}
	else {
		if ( current->mm ) {
			down_read(&current->mm->mmap_sem);
			while ( sz > 0 ) {
				vma = find_vma(current->mm, start);
				if ( !vma || start < vma->vm_start ) {
					MTL_DEBUG1(MT_STDFMT("start="VIRT_ADDR_FMT", iomem=%d"), start, *iomem_p);
					rc = MT_ENOMEM;
					break;
				}
				else {
					switch ( tmp ) {
						case -1:
							if ( vma->vm_flags & VM_IO )
								tmp = 1;
							else
								tmp = 0;
							break;

						case 0:
							if ( vma->vm_flags & VM_IO ) {
								rc = MT_ENOMEM;
								break_while = 1;
							}
							break;

						case 1:
							if ( !(vma->vm_flags & VM_IO) ) {
								rc = MT_ENOMEM;
								break_while = 1;
							}
							break;
					}

					if ( break_while )
						break;

					if ( (start+sz) > vma->vm_end ) {
						MTL_DEBUG1(MT_STDFMT("start="VIRT_ADDR_FMT", iomem=%d"), start, *iomem_p);
						sz -= (vma->vm_end-start);
						start = vma->vm_end;
					}
					else
						break;
				}
			}
			up_read(&current->mm->mmap_sem);
		}
		else {
			return MT_ENOMEM;
		}
	}
	if ( rc == MT_OK ) {
		*iomem_p = tmp;
	}
	return rc;
}

/*
 *  mosal_get_page_table_lock
 */
static inline spinlock_t *mosal_get_page_table_lock(const MOSAL_prot_ctx_t
						    prot_ctx)
{
	if (prot_ctx == MOSAL_get_current_prot_ctx()) {
		return &current->mm->page_table_lock;
	} else if (prot_ctx == MOSAL_get_kernel_prot_ctx()) {
		return &init_mm.page_table_lock;
	} else {
		return NULL;
	}
}

/*
 *  mosal_get_pgd
 */
static inline pgd_t *mosal_get_pgd(const MOSAL_prot_ctx_t prot_ctx,
				   const MT_virt_addr_t va)
{
	pgd_t *pgd;

	if (prot_ctx == MOSAL_get_current_prot_ctx()) {
		pgd = pgd_offset(current->mm, va);
	} else if (prot_ctx == MOSAL_get_kernel_prot_ctx()) {
		pgd = pgd_offset_k(va);
	} else {
		MTL_ERROR1(MT_FLFMT("%s: invalid prot_ctx arg(%d)"), __func__,
			   prot_ctx);
		return NULL;
	}

	if (pgd_none(*pgd) || pgd_bad(*pgd)) {
		return NULL;
	}

	return pgd;
}

/*
 *  mosal_get_pmd
 */
static inline pmd_t *mosal_get_pmd(MOSAL_prot_ctx_t prot_ctx,
				   const pgd_t * pgd_p, const MT_virt_addr_t va)
{
#if LINUX_VERSION_CODE >=  KERNEL_VERSION(2,6,11)
	return NULL;
#else
	pmd_t *pmd_p;

	pmd_p = pmd_offset((pgd_t *) pgd_p, va);

	if (pmd_none(*pmd_p)) {
		return NULL;
	}

	return pmd_p;
#endif
}

/*
 *  mosal_is_big_page
 */
static inline int mosal_is_big_page(const pmd_t * pmd)
{
#ifdef __i386__
#if !LINUX_KERNEL_2_6
	if (pmd_val(*pmd) & _PAGE_PSE)
		return 1;
	else
		return 0;
#else
	if (pmd_large(*pmd))
		return 1;
	else
		return 0;
#endif
#else
	return 0;
#endif
}

/*
 *  get_bp_page_p
 */
static inline struct page *get_bp_page_p(const pmd_t * pmd_p,
					 const MT_virt_addr_t va)
{
#ifdef __i386__
	struct page *page_p;
	unsigned long mask;

	mask = (1UL << PMD_SHIFT) - 1;
	page_p = (struct page *)pmd_page(*pmd_p) + (va & mask) / PAGE_SIZE;
#ifdef VALID_PAGE
	if (!VALID_PAGE(page_p)) {
		return NULL;
	}
#endif
	return page_p;
#else
	return NULL;
#endif
}

/*
 *  mosal_get_pte
 */
static inline call_result_t mosal_get_pte(pmd_t * pmd_p,
					  const MT_virt_addr_t va,
					  pte_t * arg_pte_p)
{
	pte_t *pte_p;

#ifdef MT_PTE_OFFSET_MAP
	pte_p = pte_offset_map(pmd_p, va);
	if (!pte_p) {
		return MT_EAGAIN;
	}
	*arg_pte_p = *pte_p;
	pte_unmap(pte_p);
#else

	pte_p = pte_offset(pmd_p, va);
	if (!pte_p) {
		return MT_EAGAIN;
	}
	*arg_pte_p = *pte_p;
#ifdef MT_SUSE
	pte_kunmap(pte_p);
#endif				/* MT_SUSE */
#endif

	return MT_OK;
}

#if defined(__ia64__) && ( defined(MT_SUSE) || defined(RH_AS_3_0) )
#define VTP_CAST(x) ((void *)x)
#else
#define VTP_CAST(x) x
#endif


/*
 *  MOSAL_get_page_shift
 */
call_result_t MOSAL_get_page_shift(MOSAL_prot_ctx_t prot_ctx, MT_virt_addr_t va,
				   unsigned int *page_shift_p)
{
#ifdef __ia64__
	if (prot_ctx == MOSAL_get_kernel_prot_ctx()) {
		*page_shift_p = MOSAL_SYS_PAGE_SHIFT;
	} else {
		if (!in_address_space(va)) {
			return MT_ENOMEM;
		} else {
			*page_shift_p = MOSAL_SYS_PAGE_SHIFT;
		}
	}
	return MT_OK;
#else
	struct vm_area_struct *vma;

	if (prot_ctx == MOSAL_get_current_prot_ctx()) {
		/* user space address */
		down_read(&current->mm->mmap_sem);
		vma = find_vma(current->mm, va);
		if (!vma || (va < vma->vm_start)) {
			up_read(&current->mm->mmap_sem);
			return MT_ENOMEM;
		}
#ifdef VM_BIGPAGE
		if (vma->vm_flags & VM_BIGPAGE) {
			*page_shift_p = PMD_SHIFT;
		} else {
			*page_shift_p = MOSAL_SYS_PAGE_SHIFT;
		}
#else
	#ifdef VM_HUGETLB
		if (vma->vm_flags & VM_HUGETLB) {
			*page_shift_p = PMD_SHIFT;
		} else {
			*page_shift_p = MOSAL_SYS_PAGE_SHIFT;
		}
	#else
		*page_shift_p = MOSAL_SYS_PAGE_SHIFT;
#endif
#endif
		up_read(&current->mm->mmap_sem);
		return MT_OK;
	} else {
		/* kernel space addresses */
		if ( va >= VMALLOC_START ) {
			*page_shift_p = PAGE_SHIFT;
		}
		else {
			*page_shift_p = PMD_SHIFT;
		}

		return MT_OK;
	}
#endif
}

/*====================================================================================*/


/*
 *  map_open
 */
static void map_open(struct vm_area_struct *area)
{
  MTL_DEBUG1(MT_STDMSG(""));
}


/*
 *  map_close
 */
static void map_close(struct vm_area_struct *area)
{
  MTL_DEBUG1(MT_STDMSG(""));
}

/*
 *  map_nopage
 */
#if LINUX_KERNEL_2_6
static struct page *map_nopage(struct vm_area_struct *area, unsigned long address, int *type)
#else
static struct page *map_nopage(struct vm_area_struct *area, unsigned long address, int unused)
#endif
{
  MOSAL_phys_ctg_t pobj = (MOSAL_phys_ctg_t)area->vm_private_data;
  unsigned int i, maxi;
  struct page *page = NOPAGE_SIGBUS;

  MTL_DEBUG1(MT_STDMSG("address=0x%lx, area->vm_start=0x%lx, area->vm_end=0x%lx, area->vm_pgoff=0x%lx"),
                       address, area->vm_start, area->vm_end, area->vm_pgoff);
  i = (address - area->vm_start) >> PAGE_SHIFT;
  maxi = 1 << (pobj->order-PAGE_SHIFT);
  if ( i < maxi ) {
    page = pobj->pages+i;
    get_page(page);
  }
  else {
    MTL_ERROR1(MT_STDMSG("address=0x%lx, area->vm_start=0x%lx, area->vm_end=0x%lx, area->vm_pgoff=0x%lx"),
                         address, area->vm_start, area->vm_end, area->vm_pgoff);
  }
  return page;
}

static struct vm_operations_struct vma_ops = {
  open: map_open,
  close: map_close,
  nopage: map_nopage,
};



/*
 *  MOSAL_mem_phys_ctg_alloc
 */
call_result_t MOSAL_mem_phys_ctg_alloc(IN MT_u_int_t log2_size,
				       IN MT_u_int_t log2_align,
				       IN MOSAL_mem_alloc_flags_t flags,
				       OUT MOSAL_phys_ctg_t * pobj_p)
{
	MT_u_int_t order;
	struct page *pages;
	dma_addr_t phys_addr, align_mask;
	call_result_t rc;
	MOSAL_phys_ctg_t newobj;

	MTL_DEBUG1(MT_STDMSG("log2_size=%u, log2_align=%u, flags=0x%x"),
		   log2_size, log2_align, flags);
	if (!pobj_p)
		return MT_EINVAL;

	if (flags != MOSAL_MEM_ALLOC_FLAGS_NONE) {
		/* no particular flags are supported in this implementation */
		return MT_EINVAL;
	}

	/* calculate order which is log2 of number of required pages */
	if (log2_size >= PAGE_SHIFT)
		order = log2_size - PAGE_SHIFT;
	else
		order = 0;

	/* allocate object */
	newobj = TMALLOC(struct MOSAL_phys_ctg_st);
	if (!newobj)
{
		MT_COV_MEM_FLTR_START
		return MT_ENOMEM;
		MT_COV_MEM_FLTR_END
}
	memset(newobj, 0, sizeof(*newobj));

	if ((rc = MOSAL_mutex_init(&newobj->mtx)) != MT_OK) {
		goto ex_freeobj;
	}

	pages = alloc_pages(GFP_KERNEL, order);
	if (!pages) {
		rc = MT_ENOMEM;
		goto ex_freeobj;
	}

	MTL_DEBUG1(MT_STDMSG("kernel address=%p"), page_address(pages));

	align_mask = (1 << log2_align) - 1;
	phys_addr = page_to_phys(pages);
	if (phys_addr & align_mask) {
		/* It may be necessary to try harder here */
		rc = MT_EALIGN;
		goto ex_no_align;
	}

	newobj->pages = pages;
	newobj->order = order + PAGE_SHIFT;
	newobj->kctx.map_count = 0;
	newobj->kctx.map_flags = 0;
	newobj->kctx.va = VA_NULL;
	newobj->uctx.map_count = 0;
	newobj->uctx.map_flags = 0;
	newobj->uctx.va = VA_NULL;


	*pobj_p = newobj;
	MTL_DEBUG1(MT_STDMSG("newobj=%p"), newobj);
	return MT_OK;

      ex_no_align:
	__free_pages(pages, order);
      ex_freeobj:
	FREE(newobj);
	return rc;
}

/*
 *  MOSAL_mem_phys_ctg_free
 */
call_result_t MOSAL_mem_phys_ctg_free(IN MOSAL_phys_ctg_t pobj)
{
	call_result_t rc = MT_OK;

	if (!pobj)
		return MT_EINVAL;

	if (MOSAL_mutex_acq(&pobj->mtx, TRUE) != MT_OK)
		return MT_EINTR;

	if (pobj->kctx.map_count > 1) {
		/* object is mapped */
		MTL_ERROR1(MT_STDMSG("object %p is mapped when freed"), pobj);	// ??
		pobj->kctx.map_count = 1;
		MOSAL_mem_phys_ctg_unmap(pobj, MOSAL_get_kernel_prot_ctx());
	}
	if (pobj->uctx.map_count > 1) {
		/* object is mapped */
		MTL_ERROR1(MT_STDMSG("object %p is mapped when freed"), pobj);	// ??
		pobj->uctx.map_count = 1;
		MOSAL_mem_phys_ctg_unmap(pobj, MOSAL_get_current_prot_ctx());
	}
	__free_pages(pobj->pages, pobj->order - PAGE_SHIFT);

	MOSAL_mutex_rel(&pobj->mtx);
	FREE(pobj);

	return rc;
}

/*
 *  MOSAL_mem_phys_ctg_map
 */
call_result_t MOSAL_mem_phys_ctg_map(IN MOSAL_phys_ctg_t pobj,
				     IN MOSAL_prot_ctx_t prot_ctx,
				     IN MOSAL_mem_flags_t flags,
				     OUT MT_virt_addr_t * va_p)
{
	unsigned long mapped_addr;
	call_result_t rc = MT_OK;
	unsigned long prot_flags = 0, vm_flags = 0;
	struct vm_area_struct *vma;

	MTL_DEBUG1(MT_STDMSG("pobj=%p"), pobj);

	/* check arguments */
	if (!pobj)
		return MT_EINVAL;
	if (prot_ctx != MOSAL_get_current_prot_ctx() &&
	    prot_ctx != MOSAL_get_kernel_prot_ctx())
		return MT_EINVAL;
	if (flags & MOSAL_MEM_FLAGS_NO_CACHE) {
		MTL_ERROR1(MT_STDMSG
			   ("NO CACHING for physical address is not supported"));
		return MT_EINVAL;
	}

	if (MOSAL_mutex_acq(&pobj->mtx, TRUE) != MT_OK)
		return MT_EINTR;

	if (prot_ctx == MOSAL_get_kernel_prot_ctx()) {
		/* map to kernel space */
		if (pobj->kctx.map_count > 0) {
			/* already mapped. just increment the reference count */
			pobj->kctx.map_count++;
			if (va_p)
				*va_p = pobj->kctx.va;
		} else {
			/* first time object is mapped */
			pobj->kctx.va =
			    (MT_virt_addr_t) page_address(pobj->pages);
			if (pobj->kctx.va == VA_NULL) {
				rc = MT_ENOMEM;
			} else {
				pobj->kctx.map_count++;
				pobj->kctx.map_flags = flags;
				if (va_p)
					*va_p = pobj->kctx.va;
			}
		}
	} else {
		/* map to user space */
		if (pobj->uctx.map_count > 0) {
			/* object already mapped to user space */
			rc = MT_EBUSY;
		} else {
			/* edit flags according to caller's args */
			if (flags & MOSAL_MEM_FLAGS_PERM_READ) {
				prot_flags = PROT_READ;
				vm_flags = MAP_PRIVATE;
			}
			if (flags & MOSAL_MEM_FLAGS_PERM_WRITE) {
				prot_flags |= PROT_WRITE;
				vm_flags = MAP_SHARED;
			}
			vm_flags |= MAP_ANONYMOUS;

			down_write(&current->mm->mmap_sem);
			mapped_addr =
			    do_mmap_pgoff(NULL, 0, 1 << pobj->order, prot_flags,
					  vm_flags, 0);
			if (IS_ERR((void *)mapped_addr)) {
				MTL_ERROR1(MT_STDMSG
					   ("failed to obtain va range"));
				rc = MT_ENOMEM;
				goto ex_unlock;
			}
			MTL_DEBUG1(MT_STDMSG("obtained va range. va=0x%lx"),
				   mapped_addr);

			vma = find_vma(current->mm, mapped_addr);
			if (!vma || mapped_addr < vma->vm_start) {
				MTL_ERROR1(MT_STDMSG("not likely to happen"));
				rc = MT_ERROR;
				goto ex_unlock;
			}

			vma->vm_ops = NULL;
			vma->vm_flags |= VM_LOCKED;
			current->mm->locked_vm += (1<<(pobj->order-PAGE_SHIFT));

			/* check that for the new vma object, vm_ops and vm_private_data are not used */
			if ( vma->vm_ops || vma->vm_private_data ) {
			  MTL_ERROR1(MT_STDMSG("vma->vm_ops=%p, vma->vm_private_data=%p"), vma->vm_ops, vma->vm_private_data);
			  rc = MT_EBUSY;
			  goto ex_unlock;
			}		 
			MTL_DEBUG1(MT_STDMSG("vma->vm_start=0x%lx, vma->vm_end=0x%lx"), vma->vm_start, vma->vm_end);

			vma->vm_ops = &vma_ops;
			vma->vm_private_data = (void *)pobj;

			pobj->uctx.mm = current->mm;
			pobj->uctx.pid = current->pid;
			pobj->uctx.va = mapped_addr;
			pobj->uctx.map_count++;
			if (va_p)
				*va_p = pobj->uctx.va;
		      ex_unlock:
			up_write(&current->mm->mmap_sem);
		}
	}
	MOSAL_mutex_rel(&pobj->mtx);

	return rc;
}

/*
 *  MOSAL_mem_phys_ctg_unmap
 */
call_result_t MOSAL_mem_phys_ctg_unmap(IN MOSAL_phys_ctg_t pobj,
				       IN MOSAL_prot_ctx_t prot_ctx)
{
	call_result_t rc = MT_OK;

	if (!pobj)
		return MT_EINVAL;
	if (prot_ctx != MOSAL_get_current_prot_ctx() &&
	    prot_ctx != MOSAL_get_kernel_prot_ctx())
		return MT_EINVAL;

	if (MOSAL_mutex_acq(&pobj->mtx, TRUE) != MT_OK)
		return MT_EINTR;

	if (prot_ctx == MOSAL_get_kernel_prot_ctx()) {
		/* unmap form kernel space */
		if (pobj->kctx.map_count > 0) {
			pobj->kctx.map_count--;
		}
                else rc = MT_ENOMEM;
	} else {
		/* unmap from user spcae */
		if (pobj->uctx.map_count > 0) {
			if (current->mm && current->mm == pobj->uctx.mm
			    && current->pid == pobj->uctx.pid) {
				MOSAL_DO_MUNMAP(current->mm,
						(unsigned long)pobj->uctx.va,
						1 << pobj->order);
			}
			pobj->uctx.map_count--;
		}
                else rc = MT_ENOMEM;
	}

	MOSAL_mutex_rel(&pobj->mtx);
	return rc;
}

/*
 *  MOSAL_mem_phys_ctg_get_props
 */
call_result_t MOSAL_mem_phys_ctg_get_props(IN MOSAL_phys_ctg_t pobj,
					   OUT MOSAL_phys_ctg_props_t * props_p)
{
	call_result_t rc = MT_OK;

	if (!pobj)
		return MT_EINVAL;
	MTL_DEBUG1(MT_STDMSG("pobj=%p"), pobj);

	if (MOSAL_mutex_acq(&pobj->mtx, TRUE) != MT_OK)
		return MT_EINTR;

	props_p->pa = page_to_phys(pobj->pages);
	props_p->order = pobj->order;

	props_p->k_map_count = pobj->kctx.map_count;
	if (pobj->kctx.map_count > 0) {
		props_p->k_va = pobj->kctx.va;
		props_p->k_map_flags = pobj->kctx.map_flags;
	} else {
		props_p->k_va = VA_NULL;
		props_p->k_map_flags = 0;
	}

	props_p->u_map_count = pobj->uctx.map_count;
	if (pobj->uctx.map_count) {
		props_p->u_va = pobj->uctx.va;
		props_p->u_map_flags = pobj->uctx.map_flags;
	} else {
		props_p->u_va = VA_NULL;
		props_p->u_map_flags = 0;
	}

	MOSAL_mutex_rel(&pobj->mtx);
	return rc;
}

/*
 *  MOSAL_mem_iomem_map
 */
call_result_t MOSAL_iomem_map(IN MOSAL_prot_ctx_t prot_ctx,
			      IN MOSAL_mem_flags_t flags,
			      IN MT_phys_addr_t pa,
			      IN MT_size_t size,
			      OUT MOSAL_iomem_t * iomemobj_p,
			      OUT MT_virt_addr_t * va_p)
{
	call_result_t rc;
	MOSAL_iomem_t newobj;
	unsigned long prot_flags = 0, vm_flags = 0;
	unsigned long mapped_addr;
	struct vm_area_struct *vma;
	pgprot_t pgprot_flags = __pgprot(0);
	int ret;

	/* check parameters */
	if (!iomemobj_p)
		return MT_EINVAL;
	if (prot_ctx != MOSAL_get_current_prot_ctx() &&
	    prot_ctx != MOSAL_get_kernel_prot_ctx())
		return MT_EINVAL;

	MTL_DEBUG1(MT_STDMSG
		   ("prot_ctx=%s, flags=0x%x, size=" SIZE_T_FMT ", pa="
		    PHYS_ADDR_FMT),
		   prot_ctx == MOSAL_get_current_prot_ctx()? "USER" : "KERNEL",
		   flags, size, pa);

	/* aloocate object memory */
	newobj = TMALLOC(struct MOSAL_iomem_st);
	if (!newobj)
{
		MT_COV_MEM_FLTR_START
		return MT_EAGAIN;
		MT_COV_MEM_FLTR_END
}

	if (prot_ctx == MOSAL_get_kernel_prot_ctx()) {
		/* kernel protection context */
		if (flags & MOSAL_MEM_FLAGS_NO_CACHE) {
			mapped_addr = (unsigned long)MOSAL_io_remap(pa, size);
		} else {
			mapped_addr = (unsigned long)ioremap(pa, size);
		}

		if (mapped_addr)
			rc = MT_OK;
		else
			rc = MT_ENOMEM;
	} else {
		/* user protection context */
		if (flags & MOSAL_MEM_FLAGS_PERM_READ) {
			vm_flags = MAP_PRIVATE;
			prot_flags = PROT_READ;
			pgprot_flags = PAGE_READONLY;
		}
		if (flags & MOSAL_MEM_FLAGS_PERM_WRITE) {
			vm_flags = MAP_SHARED;
			prot_flags |= PROT_WRITE;
			pgprot_flags = PAGE_SHARED;
		}
		vm_flags |= MAP_ANONYMOUS;

		down_write(&current->mm->mmap_sem);
		mapped_addr =
		    do_mmap_pgoff(NULL, 0, size, prot_flags, vm_flags, 0);
		if (IS_ERR((void *)mapped_addr)) {
			MTL_ERROR1(MT_STDMSG("failed to obtain va range"));
			rc = MT_ENOMEM;
			goto ex_unlock;
		}
		MTL_DEBUG1(MT_STDMSG("obtained va range. va=0x%lx"),
			   mapped_addr);

		vma = find_vma(current->mm, mapped_addr);
		if (!vma || mapped_addr < vma->vm_start) {
			MTL_ERROR1(MT_STDMSG("not likely to happen"));
			MOSAL_DO_MUNMAP(current->mm, mapped_addr, size);
			rc = MT_ERROR;
			goto ex_unlock;
		}

		/* mark vm flags as IO and NO_UNLOCK */
		vma->vm_flags |= VM_IO;
#ifdef VM_NO_UNLOCK
		vma->vm_flags |= VM_NO_UNLOCK;
#endif

		if (flags & MOSAL_MEM_FLAGS_NO_CACHE) {
#if defined(pgprot_noncached)
			pgprot_flags = pgprot_noncached(pgprot_flags);
#else
			unsigned long protval;

			protval = pgprot_val(pgprot_flags);
#if defined(__i386__) || defined(__x86_64__)
			if (boot_cpu_data.x86 > 3)
				protval |= _PAGE_PCD | _PAGE_PWT;
#elif defined(__powerpc__)
			protval |= _PAGE_NO_CACHE | _PAGE_GUARDED;
#else
#error Unsupported platform. Cannot set no-caching flags.
#endif
			pgprot_flags = __pgprot(protval);
#endif
		}
		MTL_DEBUG1(MT_STDFMT("vm_flags=0x%lx, pgprot_flags=0x%lx"),
			   vma->vm_flags, pgprot_flags);

#ifdef REMAP_VER_PFN
		ret	= remap_pfn_range
		    (vma, mapped_addr, pa>>PAGE_SHIFT, size, pgprot_flags);
#else
		ret	= remap_page_range
		    (REMAP_PREFIX mapped_addr, pa, size, pgprot_flags);
#endif
		if ( ret ) {
			MOSAL_DO_MUNMAP(current->mm, mapped_addr, size);
			rc = MT_ENOMEM;
		} else {
			newobj->mm = current->mm;
			newobj->pid = current->pid;
			if ( vma->vm_private_data ) {
				MTL_ERROR1(MT_STDFMT("vm_private_data expected to be NULL = %p"), vma->vm_private_data);
				MOSAL_DO_MUNMAP(current->mm, mapped_addr, size);
				rc = MT_ERROR;
			}
			else {
				vma->vm_private_data = (void *)(unsigned long)pa;
				rc = MT_OK;
			}
		}

		/* check that for the new vma object, vm_ops and vm_private_data are not used */
	      ex_unlock:
		up_write(&current->mm->mmap_sem);
	}

	if (rc == MT_OK) {
		newobj->va = mapped_addr;
		newobj->pa = pa;
		newobj->size = size;
		newobj->prot_ctx = prot_ctx;
		if (va_p)
			*va_p = mapped_addr;
		*iomemobj_p = newobj;
	} else {
		FREE(newobj);
	}

	return rc;
}

/*
 *  MOSAL_mem_iomem_unmap
 */
call_result_t MOSAL_iomem_unmap(IN MOSAL_iomem_t iomemobj,
				IN MOSAL_prot_ctx_t prot_ctx)
{
	if (!iomemobj)
		return MT_EINVAL;
	if (prot_ctx != MOSAL_get_current_prot_ctx() &&
	    prot_ctx != MOSAL_get_kernel_prot_ctx())
		return MT_EINVAL;

	if (prot_ctx != iomemobj->prot_ctx)
		return MT_EINVAL;

	if (prot_ctx == MOSAL_get_kernel_prot_ctx()) {
		MOSAL_io_unmap(iomemobj->va);
	} else {
		if (current->mm && current->mm == iomemobj->mm
		    && current->pid == iomemobj->pid) {
			MOSAL_DO_MUNMAP(current->mm,
					(unsigned long)iomemobj->va,
					iomemobj->size);
		}
	}
	FREE(iomemobj);
	return MT_OK;
}

/*
 *  MOSAL_mem_iomem_get_props
 */
call_result_t MOSAL_iomem_get_props(IN MOSAL_iomem_t iomemobj,
				    OUT MOSAL_iomem_props_t * props_p)
{
	if (!iomemobj)
		return MT_EINVAL;

	props_p->pa = iomemobj->pa;
	props_p->va = iomemobj->va;
	props_p->size = iomemobj->size;
	props_p->prot_ctx = iomemobj->prot_ctx;

	return MT_OK;
}
