/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <linux/kernel.h>
#include <linux/sched.h>
#include <asm/unistd.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/slab.h>
#include <asm/errno.h>
#include <asm/current.h>
#include <asm/page.h>
#include <linux/list.h>

#include <linux/securebits.h>
#include <linux/utsname.h>
#include <linux/highmem.h>

#include <mtl_common.h>
#include "mosal_priv.h"
#include <mosal_mlock.h>
#include <linux/swap.h>


/*
 *  should_lock
 */
static inline int should_lock(struct vm_area_struct *vma)
{
	if ((vma->vm_flags & VM_LOCKED) || (vma->vm_flags & VM_IO)
#ifdef VM_HUGETLB
	    || (vma->vm_flags & VM_HUGETLB)
#endif
	    ) {
		return 0;
	} else {
		return 1;
	}
}

/*
 *  should_unlock
 */
static inline int should_unlock(struct vm_area_struct *vma)
{
	if ((vma->vm_flags & VM_IO)
#ifdef VM_NO_UNLOCK
	    || (vma->vm_flags & VM_NO_UNLOCK)
#endif
#ifdef VM_HUGETLB
	    || (vma->vm_flags & VM_HUGETLB)
#endif
	    ) {
		return 0;
	} else {
		return 1;
	}
}

struct lock_item_st {
	struct lock_item_st *next;
	MT_virt_addr_t addr;
	MT_size_t size;
};

/*
 *  mark_vma_for_fork
 *    this function should be called after locking since
 *    locking the memory might split VMAs and we want to mark
 *    only those that we locked
 */
static call_result_t mark_vma_for_fork(MT_virt_addr_t va, MT_size_t size,
				       int mark)
{
	struct vm_area_struct *vma;
	call_result_t rc = MT_OK;
	unsigned long addr = va, vma_sz;
	MT_size_t sz = size;
	unsigned long mask;
	unsigned long will_undergoe_cow;

#ifdef VM_NO_COW
	if (mark)
		mask = VM_NO_COW;
	else
		mask = ~VM_NO_COW;
#else
	if (mark)
		mask = VM_DONTCOPY;
	else
		mask = ~VM_DONTCOPY;
#endif

	down_write(&current->mm->mmap_sem);
	while (sz > 0) {
		vma = find_vma(current->mm, addr);
		if (!vma || addr < vma->vm_start) {
			rc = MT_ENOMEM;
			MTL_DEBUG1(MT_STDFMT
				   ("%s: Failed finding VMA for va=0x%lX"),
				   __func__, addr);
			goto ex_mark_vma_for_fork;
		} else {
			will_undergoe_cow =
			    (((vma->vm_flags & VM_SHARED) | (vma->
							     vm_flags &
							     VM_MAYWRITE)) ==
			     VM_MAYWRITE);
			if (will_undergoe_cow) {
        MTL_DEBUG1(MT_STDFMT("start=0x%lx, end=0x%lx, mark=%d"), vma->vm_start, vma->vm_end, mark);
				if (mark)
					vma->vm_flags |= mask;
				else
					vma->vm_flags &= mask;
			}
		}
		vma_sz = vma->vm_end - vma->vm_start;
		addr += vma_sz;
                sz = sz>vma_sz ? sz-vma_sz : 0;
	}
      ex_mark_vma_for_fork:
	up_write(&current->mm->mmap_sem);
	return rc;
}


/*
 *  MOSAL_mlock_init
 */
call_result_t MOSAL_mlock_init(void)
{
	MTL_DEBUG7(MT_FLFMT("==> MOSAL_mlock_init"));

	MTL_DEBUG7(MT_FLFMT("<== MOSAL_mlock_init"));
	return MT_OK;
}

/*
 *  MOSAL_mlock_cleanup
 */
void MOSAL_mlock_cleanup(void)
{
}


call_result_t MOSAL_mlock_ctx_init(MOSAL_mlock_ctx_t * mlock_ctx_p)
{
  *mlock_ctx_p = NULL;
  return MT_OK;
}

call_result_t MOSAL_mlock_ctx_cleanup(MOSAL_mlock_ctx_t mlock_ctx)
{
	return MT_OK;
}

/*
 *  MOSAL_mlock_iobuf
 */
call_result_t MOSAL_mlock_iobuf(MT_virt_addr_t addr, MT_size_t size,
				MOSAL_iobuf_t mosal_iobuf,
				unsigned int page_shift)
{
  return mark_vma_for_fork(addr, size, 1);
}

/*
 *  MOSAL_munlock_iobuf
 */
call_result_t MOSAL_munlock_iobuf(MT_virt_addr_t addr, MT_size_t size,
				  MOSAL_iobuf_t mosal_iobuf,
				  unsigned int page_shift)
{
	return MT_OK;
}


/*
 *  MOSAL_memory_lock
 */ 
call_result_t MOSAL_memory_lock(MT_virt_addr_t va,
                              MT_size_t size)
{
  return MT_OK;
}


/*
 *  MOSAL_memory_unlock
 */ 
call_result_t MOSAL_memory_unlock(MT_virt_addr_t va,
                              MT_size_t size)
{
  return MT_OK;
}


call_result_t MOSAL_undo_register(MT_virt_addr_t va,
                                  MT_size_t size)
{
  return mark_vma_for_fork(va, size, 0);
}
