/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

/*
 * Save log at DRAM and not file system.
 */
#ifndef MT_KERNEL

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#define MAX_FILENAME 256
#define MTL_LOG_ENV  "MTL_LOG"
#define MTL_LOG_DEF  "-"

#endif				/* not MT_KERNEL */

/* Here we force this feature out */
#undef MTL_LOG_MALLOC

/* Take local header before getting it from the release. */
#include "mtl_sys_defs.h"

#include "mtl_common.h"
#include "mtl_common_priv.h"

const char *mtl_strerror(call_result_t errnum)
{
	switch (errnum) {
#define INFO(A, B, C) case A: return C;
		ERROR_LIST
#undef INFO
	default:
		return "Unknown error";
	}
}

const char *mtl_strerror_sym(call_result_t errnum)
{
	switch (errnum) {
#define INFO(A, B, C) case A: return #A;
		ERROR_LIST
#undef INFO
	default:
		return "Unknown error";
	}
}

const char *mtl_basename(const char *filename)
{
	const char *slash_p = strchr(filename, '/');
	if (slash_p) {
		filename = slash_p + 1;
	}
	return filename;
}				/* mtl_basename */

static MT_bool extract_key(const char *str, const char *key, const char *suffix,
			   char *result, const int result_max_len)
{
	char *p, *pstr = (char *)str;
	int i = 0;

	if (!pstr)
		return FALSE;

	while ((p = strstr(pstr, key))) {
		if (!strncmp(p + strlen(key), suffix, strlen(suffix))) {
			char *q = p + strlen(key) + strlen(suffix);
			while (*q && *q != ' ') {
				*result++ = *q++;
				if (++i >= result_max_len)
					break;
			}
			*result = '\0';
			return TRUE;
		} else
			pstr = p + strlen(key);
	}
	return FALSE;
}

/************** Kernel Specific Code ************************/

#ifdef MT_KERNEL

static struct log_info {
	char *name;		// Module name
	struct log_info *next;	// Pointer to next record
	struct print_info {
		char *name;
		char sevs[MAX_MTL_LOG_SEVERITIES + 1];
	} print_info[MAX_MTL_LOG_TYPES];
} *log_first = (struct log_info *)NULL, log_default = {
	NULL, NULL, {
		{
		"trace", ""}, {
		"debug", ""}, {
		"error", "1234"}
	}
};

//TODO: Bug: strlen run on user space buffer!
//static char *mtl_strdup(const char *str)
//    return s ? strcpy(s, str) : NULL;
//#define mtl_strdup(str) ({ char *s = (char*)MALLOC(strlen(str)+1); s? strcpy(s, (str)) : NULL; })
static __INLINE__ char *mtl_strdup(const char *str)
{
	char *s = (char *)MALLOC(strlen(str) + 1);
	return s ? strcpy(s, (str)) : NULL;
}

static int debug_print = 0;

void mtl_log_set(char *layer, char *info)
{
	struct log_info *p, *prev;

	printk("mtl_log_set: layer '%s', info '%s'\n", layer, info);

	if (!strcmp(layer, "mtl_log_dbg_print")) {
		debug_print ^= 1;
		return;
	}

	if (strcmp(layer, "print")) {
		/*
		 * Find necessary record
		 */
		for (prev = p = log_first; p; prev = p, p = p->next) {
			if (!strcmp(layer, p->name))
				break;
		}
		if (!p) {
			/* Not found - create */
			/* Do not use MALLOC macro to avoid infinite recursion */
			p = (struct log_info *)QMALLOC(sizeof(struct log_info));
			memcpy(p, &log_default, sizeof(struct log_info));
			p->name = mtl_strdup(layer);
			if (prev)
				prev->next = p;
			else
				log_first = p;
		}
		/*
		 * Now "p" is a pointer to corresponding record, either existed
		 * or created just now
		 */
		extract_key(info, "trace", ":",
			    p->print_info[mtl_log_trace].sevs,
			    MAX_MTL_LOG_SEVERITIES);
		extract_key(info, "debug", ":",
			    p->print_info[mtl_log_debug].sevs,
			    MAX_MTL_LOG_SEVERITIES);
		extract_key(info, "error", ":",
			    p->print_info[mtl_log_error].sevs,
			    MAX_MTL_LOG_SEVERITIES);
		/* if there is a Logger */
		MT_LOGGER_CONFIG
		
	} else {
		printk("<1>\n");
		printk("<1> Layers and severities for print\n");
		printk("<1> -------------------------------\n");
		for (p = log_first; p; p = p->next) {
			int i;

			printk("<1> Layer - \"%s\":\n", p->name);
			for (i = 0; i < MAX_MTL_LOG_TYPES; i++) {
				printk
				    ("<1>     Name=\"%s\", severities=\"%s\"\n",
				     p->print_info[i].name,
				     p->print_info[i].sevs);
			}
		}
		printk("<1>\n");
	}
}

/*
 *  mtl_common_cleanup
 */
void mtl_common_cleanup(void)
{
	struct log_info *p = log_first, *next;

	while (p) {
		if (p->name) {
			FREE(p->name);
			p->name = NULL;
		}
		next = p->next;
		FREE(p);
		p = next;
	}
}

void mtl_log(const char *layer, mtl_log_types log_type, char sev,
	     const char *fmt, ...)
{
	char *pbuff;
	struct log_info *p, *prev;
	va_list ap;
	if (debug_print)
		printk("***** mtl_log:DEBUG: layer '%s', type %d, sev '%c'\n",
		       layer, log_type, sev);

	/*
	 * Find necessary record
	 */
	for (prev = p = log_first; p; prev = p, p = p->next) {
		if (!strcmp(layer, p->name))
			break;
	}

	if (!p) {
		if (debug_print)
			printk
			    ("***** mtl_log:DEBUG: Not found layer '%s' - create\n",
			     layer);

		// Not found - create
		// Avoid call to MALLOC to avoid infinite recursion
		p = (struct log_info *)QMALLOC(sizeof(struct log_info));
		if (p == NULL) {
			/* malloc failure. just return */
			return;
		}
		memcpy(p, &log_default, sizeof(struct log_info));
		p->name = mtl_strdup(layer);
		if (p->name == NULL) {
			FREE(p);
			return;
		}
		if (prev)
			prev->next = p;
		else
			log_first = p;
	} else {
		if (debug_print)
			printk
			    ("***** mtl_log:DEBUG: Found layer '%s', Name=\"%s\", sev=\"%s\"\n",
			     p->name, p->print_info[log_type].name,
			     p->print_info[log_type].sevs);
	}

	/*
	 * Now "p" is a pointer to corresponding record, either existed
	 * or created just now
	 */

	/*
	 * Log printing
	 */
	if (strchr(p->print_info[log_type].sevs, sev)) {
		if (debug_print)
			printk("***** mtl_log:DEBUG: print string\n");
		pbuff = kmalloc(MAX_MTL_LOG_LEN, GFP_ATOMIC);
		if ( pbuff ) {
			va_start(ap, fmt);
			vsnprintf(pbuff, MAX_MTL_LOG_LEN, fmt, ap);
			va_end(ap);
			printk("<1> %s(%c): %s", layer, sev, pbuff);
			kfree(pbuff);
		}
		/* if there is a Logger */
		MT_LOGGER_WRITE
	}
}

/************************************************************************/
/*           Kernel memory allocation logging/debugging                 */

#if 0
/* Alas, we do not have such machinery for the kernel (yet?) */

/* #include <execinfo.h> */
static const char *getBackTrace(char *buf, int bufLen)
{
	enum { Depth = 5 };
	char *pBuf = buf;
	void *array[Depth];
	size_t size = backtrace(array, Depth);
	if (size > 0) {
		char *pBufLimit = buf + bufLen - 2;
		char **syms = backtrace_symbols(array, size);
		char **symsEnd = syms += size;
		char *sym = syms[0];
		char *pBufNext = pBuf + strlen(sym);
		while ((syms != symsEnd) && (pBufNext < pBufLimit)) {
			strcpy(pBuf, sym);
			pBuf = pBufNext;
			*pBuf++ = '|';
			++syms;
			if (syms != symsEnd) {
				sym = *syms;
				pBufNext += strlen(sym);
			}
		}
	}
	*pBuf = '\0';
	return buf;
}				/* getBackTrace */
#endif

static const char *mtl_malloc_modname = "MEMCHECK";
static const char *mallog_magic = "takeme4memcheck";

void *mtl_log_vmalloc(const char *fn, int ln, int bsize)
{
	/*  char   symTrace[256]; 
	 *  getBackTrace(symTrace, sizeof(symTrace));  */
	void *ptr = (void *)QVMALLOC(bsize);
	mtl_log(mtl_malloc_modname, mtl_log_debug, '1',
		"%s[%d]: 0x%p := vmalloc(%d) %s %s\n",
		fn, ln, ptr, bsize, mallog_magic, "");
	return ptr;
}

void mtl_log_vfree(const char *fn, int ln, void *ptr)
{
	mtl_log(mtl_malloc_modname, mtl_log_debug, '1',
		"%s[%d]: vfree(0x%p) %s\n", fn, ln, ptr, mallog_magic);
	QVFREE(ptr);
}

void *mtl_log_kmalloc(const char *fn, int ln, int bsize, unsigned g)
{
	/*  char   symTrace[256];
	 *  getBackTrace(symTrace, sizeof(symTrace));  */
	void *ptr = (void *)QCMALLOC(bsize, g);
	mtl_log(mtl_malloc_modname, mtl_log_debug, '1',
		"%s[%d]: 0x%p := kmalloc(%d, 0x%x) %s %s\n",
		fn, ln, ptr, bsize, g, mallog_magic, "");
	return ptr;
}

void mtl_log_kfree(const char *fn, int ln, void *ptr)
{
	mtl_log(mtl_malloc_modname, mtl_log_debug, '1',
		"%s[%d]: kfree(0x%p) %s\n", fn, ln, ptr, mallog_magic);
	QFREE(ptr);
}

/*                                                                      */
/************************************************************************/
#else				/* not defined MT_KERNEL */

/************** User Specific Code ************************/

static struct log_info {
	char *name;		// Module name
	FILE *fp;		// File for output
	int first;		// first usage == 1
	struct log_info *next;	// Pointer to next record
	struct print_info {
		char *name;
		char sevs[MAX_MTL_LOG_SEVERITIES + 1];
	} print_info[MAX_MTL_LOG_TYPES];
} *log_first = (struct log_info *)NULL, log_default = {
	NULL, NULL, 1, NULL, {
		{
		"trace", ""}, {
		"debug", ""}, {
		"error", ""}
	}
};

static FILE *open_logfile(const char *fname)
{
	FILE *rc;

	if (!strcmp(fname, "-") || !strcmp(fname, "&"))
		rc = stderr;
	else if (!strcmp(fname, ">"))
		rc = stdout;
	else if ((rc = fopen(fname, "w")) == NULL) {
		fprintf(stderr, "Can't open \"%s\" - %s\nUse stderr instead.\n",
			fname, strerror(errno));
		rc = stderr;
	}
	return rc;
}

void mtl_log_DB_print(void)
{
	struct log_info *p;
	printf("<1>\n");
	printf("<1> Layers and severities for print\n");
	printf("<1> -------------------------------\n");
	for (p = log_first; p; p = p->next) {
		int i;

		printf("<1> Layer - \"%s\":\n", p->name);
		for (i = 0; i < MAX_MTL_LOG_TYPES; i++) {
			printf("<1>     Name=\"%s\", severities=\"%s\"\n",
			       p->print_info[i].name, p->print_info[i].sevs);
		}
	}
	printf("<1>\n");
}

void mtl_log(const char *layer, mtl_log_types log_type, char sev,
	     const char *fmt, ...)
{
    static char  *mtl_log_env;
    static int   first=1;
    char            fname[MAX_FILENAME+1];
    struct log_info *p, *prev;
    va_list         ap;
	MTL_MUTEX_INIT(log_lock);	

	MTL_MUTEX_LOCK(log_lock);

	/*
	 * Initialization
	 */
	if (first) {
		first = 0;
		mtl_log_env = getenv(MTL_LOG_ENV);

		// Now fill in default record
		if (!extract_key(mtl_log_env, "file", ":", fname, MAX_FILENAME))
			strcpy(fname, MTL_LOG_DEF);
		log_default.fp = open_logfile(fname);
		extract_key(mtl_log_env, "trace", ":",
			    log_default.print_info[mtl_log_trace].sevs,
			    MAX_MTL_LOG_SEVERITIES);
		extract_key(mtl_log_env, "debug", ":",
			    log_default.print_info[mtl_log_debug].sevs,
			    MAX_MTL_LOG_SEVERITIES);
		extract_key(mtl_log_env, "error", ":",
			    log_default.print_info[mtl_log_error].sevs,
			    MAX_MTL_LOG_SEVERITIES);

	}

	/*
	 * Find necessary record
	 */
	for (prev = p = log_first; p; prev = p, p = p->next) {
		if (!strcmp(layer, p->name))
			break;
	}
	if (!p) {
		// Not found - create
		p = malloc(sizeof(struct log_info));
		if (p == NULL) {
			/* malloc failure. just return */
			MTL_MUTEX_UNLOCK(log_lock);
			return;
		}
		memcpy(p, &log_default, sizeof(struct log_info));
		p->name = strdup(layer);
		if (prev)
			prev->next = p;
		else
			log_first = p;
	}
	/*
	 * Now "p" is a pointer to corresponding record, either existed
	 * or created just now
	 */

	/*
	 * First time for this module - check overwriting info in mtl_log_env
	 */
	if (p->first) {
		p->first = 0;
		if (extract_key
		    (mtl_log_env, layer, "_file:", fname, MAX_FILENAME))
			p->fp = open_logfile(fname);
		extract_key(mtl_log_env, layer, "_trace:",
			    p->print_info[mtl_log_trace].sevs,
			    MAX_MTL_LOG_SEVERITIES);
		extract_key(mtl_log_env, layer, "_debug:",
			    p->print_info[mtl_log_debug].sevs,
			    MAX_MTL_LOG_SEVERITIES);
		extract_key(mtl_log_env, layer, "_error:",
			    p->print_info[mtl_log_error].sevs,
			    MAX_MTL_LOG_SEVERITIES);
	}

	/*
	 * Log printing
	 */
	if (strchr(p->print_info[log_type].sevs, sev)) {
		fprintf(p->fp, "%s(%c): ", layer, sev);
		va_start(ap, fmt);
		vfprintf(p->fp, fmt, ap);
		va_end(ap);
		fflush(p->fp);
	}

	MTL_MUTEX_UNLOCK(log_lock);
}

#endif				/* MT_KERNEL */
