/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <vapi_common.h>
#include <vip.h>
#include <vipkl.h>

struct semaphore mod_sem;

#define MAX_LISTED_HCAS 4
#define EXCD_ERR_MSG "too many HCAs in the system\n"
#define GET_ERR_MSG "failed to retrieve info\n"
#define OVR_ERR_MSG "passed end of file\n"
#define LINE_SZ 60

static void fixup_line(char *str, int count)
{
	memset(str + count, ' ', LINE_SZ - count - 1);
	str[LINE_SZ - 1] = '\n';
}

static int read_vip_data(char *page, char **start, off_t offset, int count,
			 int *eof, void *data)
{
	VIP_ret_t rc;
	VAPI_hca_id_t hca_id_buf[MAX_LISTED_HCAS];
	u_int32_t i, j, off = 0, cnt, off_in_str;
	VAPI_hca_hndl_t hca_hndl;
	static VIPKL_rsrc_info_t rsrc_info[MAX_LISTED_HCAS];
	static int cur_hca;
	static u_int32_t num_of_hcas;
	static int first_time = 1;
	static char buf[LINE_SZ];
	off_t ln_cnt, ln_tot, local_ln_cnt;
	struct qp_data_st *qp_lst;

	if (first_time) {
		/* first time called */
		if (down_interruptible(&mod_sem)) {
			return -ERESTARTSYS;
		}

		rc = VIPKL_list_hcas(MAX_LISTED_HCAS, &num_of_hcas, hca_id_buf);
		if (rc != VIP_OK) {
			strcpy(page, EXCD_ERR_MSG);
			*eof = 1;
			up(&mod_sem);
			return strlen(EXCD_ERR_MSG);
		}

		for (i = 0; i < num_of_hcas; ++i) {
			rc = VIPKL_get_hca_hndl(hca_id_buf[i], &hca_hndl);
			if (rc != VIP_OK) {
				continue;
			}

			rc = VIPKL_get_rsrc_info(hca_hndl, &rsrc_info[i]);
			if (rc != VIP_OK) {
				for (j = 0; j < i; ++j) {
					VIPKL_rel_qp_rsrc_info(rsrc_info[i].qp.
							       qp_list);
				}
				strcpy(page, GET_ERR_MSG);
				*eof = 1;
				up(&mod_sem);
				return strlen(GET_ERR_MSG);
			}
		}
		first_time = 0;
	}

	ln_cnt = offset / LINE_SZ;
	local_ln_cnt = ln_cnt;
	ln_tot = 0;
	for (i = 0; i < num_of_hcas; ++i) {
		ln_tot += rsrc_info[i].qp.count;
		if (ln_cnt < ln_tot) {
			break;
		}
		local_ln_cnt -= rsrc_info[i].qp.count;
	}

	if (i >= num_of_hcas) {
		strcpy(page, OVR_ERR_MSG);
		*eof = 1;
		up(&mod_sem);
		return strlen(OVR_ERR_MSG);
	}

	qp_lst = rsrc_info[i].qp.qp_list;
	for (j = 0; j < local_ln_cnt; ++j) {
		qp_lst = qp_lst->next;
	}

	cnt = VIPKL_get_qp_rsrc_str(qp_lst, buf);
	fixup_line(buf, cnt);
	off_in_str = offset % 60;
	memcpy(page, buf + off_in_str, LINE_SZ - off_in_str);
	off += LINE_SZ - off_in_str;

	qp_lst = qp_lst->next;

	for (cur_hca = i; cur_hca < num_of_hcas; cur_hca++) {
		while (1) {
			cnt = VIPKL_get_qp_rsrc_str(qp_lst, buf);
			if (cnt == 0) {
				if ((cur_hca + 1) == num_of_hcas) {
					*eof = 1;
					up(&mod_sem);
//          printk("(%d) off=%d\n", __LINE__, off);
					first_time = 1;
					for (i = 0; i < num_of_hcas; ++i) {
						VIPKL_rel_qp_rsrc_info(rsrc_info
								       [i].qp.
								       qp_list);
					}
					return off;
				} else {
					break;
				}
			}

			fixup_line(buf, cnt);
			memcpy(page + off, buf, LINE_SZ);
			off += LINE_SZ;
			if ((PAGE_SIZE - off) < (2 * LINE_SZ)) {
				goto ex_1;
				break;
			}
			qp_lst = qp_lst->next;
		}
	}

      ex_1:
	*start = page;
//  printk("(%d) off=%d\n", __LINE__, off);
	return off;
}

int init_module(void)
{

	sema_init(&mod_sem, 1);
	create_proc_read_entry("vip", 0, NULL, read_vip_data, NULL);

	return 0;
}

void cleanup_module(void)
{
	remove_proc_entry("vip", NULL);
}
