/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: tavor_main.c,v 1.9 2004/02/25 01:03:34 roland Exp $
*/

#include "tavor_priv.h"
#include "ts_ib_provider.h"

#include "ts_kernel_trace.h"
#include "ts_kernel_services.h"
#include "openib_version.h"

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>

#include <linux/init.h>
#include <linux/errno.h>
#include <linux/slab.h>

MODULE_AUTHOR("Roland Dreier");
MODULE_DESCRIPTION("IB API Tavor provider" OPENIB_REVISION);
MODULE_LICENSE("Dual BSD/GPL");

extern  VAPI_ret_t EVAPI_set_pnp_event_handler(VAPI_pnp_event_handler_t pnp_cb); 
extern  VAPI_ret_t EVAPI_clear_pnp_event_handler(void);

typedef struct hca_device {
	struct ib_device *dev;
	VAPI_hca_id_t    name;
} hca_device_t;

static hca_device_t tavor_array[TS_IB_MAX_TAVOR] = { {0} };

static struct semaphore dev_allocate_mutex;

static int allocate_device(VAPI_hca_id_t name, int *dev_index)
{
	int i = 0;

	if (down_interruptible(&dev_allocate_mutex)) {
		return -ERESTARTSYS;
	}

	/* check if this device is already registered */
	for (i = 0; i < TS_IB_MAX_TAVOR; i++) {
		if (!strcmp(name, tavor_array[i].name)) {
			up (&dev_allocate_mutex);
			return -EEXIST;
		}
	}
	/* check for non-null-string name array */
	i = 0;
	while ( i < TS_IB_MAX_TAVOR) {
		if (tavor_array[i].name[0] == '\0') {
			break;
		}
		i++;
	}

	if (i == TS_IB_MAX_TAVOR) {
		up (&dev_allocate_mutex);
		return -EAGAIN;
	}

	strncpy((char *) tavor_array[i].name, name, sizeof(VAPI_hca_id_t)-1);
	/* note:  null terminator in last character is initialized, and never
	 *        modified.
	 */
	*dev_index = i;
	up (&dev_allocate_mutex);
	return 0;
}

static int find_release_device(VAPI_hca_id_t name, struct ib_device **dev)
{
	int i = 0;

	if (down_interruptible(&dev_allocate_mutex)) {
		return -ERESTARTSYS;
	}

	/* check if this device is already registered */
	for (i = 0; i < TS_IB_MAX_TAVOR; i++) {
		if (!strcmp(name, tavor_array[i].name)) {
			*dev = tavor_array[i].dev;
			tavor_array[i].dev = NULL;
			tavor_array[i].name[0] = '\0';
			up (&dev_allocate_mutex);
			return 0;
		}
	}

	up (&dev_allocate_mutex);
	return -ENODEV;
}

static int _tsIbTavorInitOne(VAPI_hca_id_t name)
{
	VAPI_ret_t result;
	VAPI_hca_vendor_t *hca_vendor;
	VAPI_hca_cap_t *hca_cap;
	struct ib_device *device;
	int rc = 0;
	int dev_index = 0;
	tTS_IB_TAVOR_PRIVATE priv;

	device = kmalloc(sizeof *device, GFP_KERNEL);
	if (!device) {
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"%s: Failed to allocate device structure",
				name);
		return -ENOMEM;
	}

	memset(device, 0, sizeof *device);

	priv = kmalloc(sizeof *priv, GFP_KERNEL);
	if (!priv) {
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"%s: Failed to allocate private structure",
				name);
		kfree(device);
		return -ENOMEM;
	}


	rc = allocate_device(name,&dev_index);
	switch(rc) {
	case 0:
		/* OK  */
		break;
	case -EEXIST:
		/* device already registered */
		kfree(device);
		return 0;
	case -EAGAIN:
		/* no more device slots available */
		kfree(device);
		return -EAGAIN;
	default:
		kfree(device);
		return -EINVAL;
	}
	sema_init(&priv->fmr_unmap_mutex, 1);

	device->owner = THIS_MODULE;
	device->private = priv;
	device->flags = 0;

	result = VAPI_open_hca(name, &priv->vapi_handle);
	switch (result) {
	case VAPI_OK:
	case VAPI_EBUSY:
		break;

	default:
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"%s: VAPI_open_hca failed, status %d (%s)",
				name, result, VAPI_strerror(result));
		goto error;
	}

	result = EVAPI_get_hca_hndl(name, &priv->vapi_handle);
	if (result != VAPI_OK) {
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"%s: EVAPI_get_hca_hndl failed, status %d (%s)",
				name, result, VAPI_strerror(result));
		goto undo_open_hca;
	}

	hca_vendor = kmalloc(sizeof *hca_vendor, GFP_KERNEL);
	if ( !hca_vendor ) {
		goto undo_get_hca_hndl;
	}
	hca_cap = kmalloc(sizeof *hca_cap, GFP_KERNEL);
	if ( !hca_cap ) {
		goto undo_malloc_vendor;
	}

	result = VAPI_query_hca_cap(priv->vapi_handle, hca_vendor, hca_cap);
	if (result) {
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"%s: VAPI_query_hca_cap failed, status %d (%s)",
				name, result, VAPI_strerror(result));
		goto undo_malloc_cap;
	}

	device->dma_device = pci_find_slot(hca_vendor->pci_bus,
					   hca_vendor->pci_devfn);
	if (!device->dma_device) {
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"%s: Could not find pci device %02x:%02x",
				name, hca_vendor->pci_bus, hca_vendor->pci_devfn);
		goto undo_malloc_cap;
	}

	if (pci_set_dma_mask(device->dma_device, 0xffffffffffffffffull)) {
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "%s: Warning: couldn't set 64-bit PCI DMA mask.",
			       name);
		if (pci_set_dma_mask(device->dma_device, 0xffffffffull)) {
			TS_REPORT_FATAL(MOD_KERNEL_IB,
					"%s: Can't set PCI DMA mask.", name);
			goto undo_malloc_cap;
		}
	}

	result = VAPI_set_async_event_handler(priv->vapi_handle,
					      tsIbTavorAsyncHandler, device);
	if (result) {
		TS_REPORT_FATAL(MOD_KERNEL_IB,
				"%s: VAPI_set_async_event_handler failed, status %d (%s)",
				name, result, VAPI_strerror(result));
		goto undo_malloc_cap;
	}

	strncpy(device->name, name, IB_DEVICE_NAME_MAX);
	device->provider             = "tavor";
	device->device_query         = tsIbTavorDeviceQuery;
	device->port_query           = tsIbTavorPortQuery;
	device->port_modify          = tsIbTavorPortModify;
	device->pkey_query           = tsIbTavorPkeyQuery;
	device->gid_query            = tsIbTavorGidQuery;
	device->pd_create            = tsIbTavorPdCreate;
	device->pd_destroy           = tsIbTavorPdDestroy;
	device->address_create       = tsIbTavorAddressCreate;
	device->address_destroy      = tsIbTavorAddressDestroy;
	device->qp_create            = tsIbTavorQpCreate;
	device->special_qp_create    = tsIbTavorSpecialQpCreate;
	device->qp_modify            = tsIbTavorQpModify;
	device->qp_query             = tsIbTavorQpQuery;
	device->qp_destroy           = tsIbTavorQpDestroy;
	device->send_post            = tsIbTavorSendPost;
    device->receive_post         = tsIbTavorReceivePost;
	device->cq_create            = tsIbTavorCqCreate;
	device->cq_destroy           = tsIbTavorCqDestroy;
	device->cq_resize            = tsIbTavorCqResize;
	device->cq_poll              = tsIbTavorCqPoll;
	device->cq_arm               = tsIbTavorCqArm;
	device->cq_stop_comp_ev      = tsIbTavorCqStopCompEvents;
	device->mr_register          = tsIbTavorMemoryRegister;
	device->mr_register_physical = tsIbTavorMemoryRegisterPhysical;
	device->mr_deregister        = tsIbTavorMemoryDeregister;
	device->mw_create            = tavor_mw_create;
	device->mw_destroy           = tavor_mw_destroy;
	device->mw_bind              = tavor_mw_bind;
	device->fmr_create           = tsIbTavorFmrCreate;
	device->fmr_destroy          = tsIbTavorFmrDestroy;
	device->fmr_map              = tsIbTavorFmrMap;
	device->fmr_unmap            = tsIbTavorFmrUnmap;
	device->multicast_attach     = tsIbTavorMulticastAttach;
	device->multicast_detach     = tsIbTavorMulticastDetach;
	device->mad_process          = tsIbTavorMadProcess;
	device->qp_detach_destroy    = tsIbTavorQpDetachDestroy;

	tsIbTavorSetNodeDesc(priv, dev_index);

	if (tsIbDeviceRegister(device)) {
		goto undo_malloc_cap;
	}

	tavor_array[dev_index].dev = device;

	kfree(hca_cap);
	kfree(hca_vendor);

	return 0;
undo_malloc_cap:
	kfree(hca_cap);
undo_malloc_vendor:
	kfree(hca_vendor);
undo_get_hca_hndl:
    VAPI_set_async_event_handler(priv->vapi_handle, NULL, NULL);
	EVAPI_release_hca_hndl(priv->vapi_handle);
undo_open_hca:
//	VAPI_close_hca(priv->vapi_handle);
error:
	kfree(device->private);
	kfree(device);

	return -ENODEV;
}

static void _tsIbTavorCleanupOne(VAPI_hca_id_t name)
{
	struct ib_device *device = NULL;
	tTS_IB_TAVOR_PRIVATE priv;
	int rc = 0;
	VAPI_ret_t result;

	rc = find_release_device(name,&device); 
	if (rc != 0) {
		return;
	}
	if (!device)
		return;

	priv = device->private;

	result = VAPI_set_async_event_handler(priv->vapi_handle, NULL, NULL);
	if (result != VAPI_OK) {
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "%s: VAPI_set_async_event_handler (unset) failed, return code = %d (%s)",
			       (char *) name, result, VAPI_strerror(result));
	}

	tsIbDeviceDeregister(device); /* always returns 0 */

	result = EVAPI_release_hca_hndl(priv->vapi_handle);
	if (result != VAPI_OK) {
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "%s: EVAPI_release_hca_hndl failed, return code = %d (%s)",
			       (char *) name, result, VAPI_strerror(result));
	}

	kfree(device->private);
	kfree(device);
}

static void _tslbTavorPnPEventHandler(VAPI_hca_id_t hca_id, 
				      VAPI_pnp_event_t pnp_event)
{
	int rc;

	switch(pnp_event) {
	case PNP_HCA_CARD_ADDED:
		rc = _tsIbTavorInitOne(hca_id);
		TS_TRACE(MOD_KERNEL_IB,T_VERBOSE, TRACE_FLOW_CONFIG,
		       "%s: adding HCA %s",__func__, hca_id);
		if (rc != 0) {
		   TS_REPORT_WARN(MOD_KERNEL_IB, 
		          "%s: could not add HCA %s (%d)",
		          __func__,hca_id, rc);
		}
		break;
	case PNP_HCA_CARD_REMOVED:
		_tsIbTavorCleanupOne(hca_id); 
		TS_TRACE(MOD_KERNEL_IB,T_VERBOSE, TRACE_FLOW_CONFIG,
		       "%s: removing HCA %s", __func__, hca_id);
		break;
	default:
		TS_REPORT_WARN(MOD_KERNEL_IB,
			       "%s: got unsupported pnp event (%0x%x) for HCA %s",
			       __func__,pnp_event,hca_id);
		break;
	}
	return;
}

static int __init _tsIbTavorInitModule(void)
{
	VAPI_ret_t result;
	int ret;

	TS_REPORT_INIT(MOD_KERNEL_IB, "Initializing IB API Tavor provider");

	ret  = tsIbTavorInitQp();                                           /*added by itayr*/
	if(ret) {
		return ret;
	}

	/* initialize device allocator mutex */
	sema_init(&dev_allocate_mutex, 1);

	result = EVAPI_set_pnp_event_handler(_tslbTavorPnPEventHandler);
	if (result != VAPI_OK) {
	    TS_REPORT_FATAL(MOD_KERNEL_IB,
		"Could not register PnP event handler; status %d (%s)",
		result, VAPI_strerror(result));
	    tsIbTavorCleanUpQp();
	    return -EINVAL;
	}

	TS_REPORT_INIT(MOD_KERNEL_IB, "IB API Tavor provider initialized");

	return 0;
}

static void __exit _tsIbTavorCleanupModule(void)
{
	int i;
	VAPI_ret_t result;

	TS_REPORT_CLEANUP(MOD_KERNEL_IB, "Unloading IB API Tavor provider");

	result = EVAPI_clear_pnp_event_handler();
	if (result != VAPI_OK) {
	    TS_REPORT_WARN(MOD_KERNEL_IB,
		"Could not clear PnP event handler; status %d (%s)",
		result, VAPI_strerror(result));
	}
	tsIbTavorCleanUpQp();                                               /*added by itayr*/

	for (i = 0; i < TS_IB_MAX_TAVOR; ++i) {
		_tsIbTavorCleanupOne(tavor_array[i].name);
	}

	TS_REPORT_CLEANUP(MOD_KERNEL_IB, "IB API Tavor provider unloaded");
}

module_init(_tsIbTavorInitModule);
module_exit(_tsIbTavorCleanupModule);

