/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/
#define MT_MOSAL_WRAP_USER_C

#include <mosal_priv.h>


/* 
	Wrapper functions' implementation
	-----------------------------

1. Main ideas
	- all input parameters are serialiazed into an INPUT buffer; output ones -into an OUPUT buffer;
	- pointers are never relayed: if a parameter is a pointer to a structure/buffer, 
	   the structure/buffer is copied to the INPUT buffer;
	- kernel function return code is always relayed as the first field of the OUTPUT buffer;
	- the INPUT/OUTPUT buffers are relayed to/from kernel by an OS-dependent function;

2. Implementation detailes;
	- the buffers are allocated on the stack (macros W_ALLOCA_xxx);
	- if there is only one input variable, it is used as an INPUT buffer (macros W_USE_xxx_AS_IBUF);
	- wrapper functions are written with the help of several macros which use the following
	  a) conventions:
	  	-- for function XXX, 
	  		INPUT buffer has 'struct i_XXX_ops_t' type,
	  		OUTPUT buffer has 'struct i_XXX_ops_t' type;
	  	-- if a function parameter is pointer to a structure or array, it must be named YYY_p and 
	  	    the appropriate field of INPUT or OUTPUT buffer must be named YYY;
	  	-- if a function parameter is a value of an integral type, its name must be the same, as the 
	  	    name of the appropriate INPUT buffer field;
	  	    
	  b) local variables:
	  	-- pi 	- pointer to the INPUT buffer of 'struct i_<func_name>_ops_t *' type
	  	-- pi_sz	- size of the INPUT buffer;
	  	-- po 	- pointer to the OUTPUT buffer of 'struct o_<func_name>_ops_t *' type
	  	-- pi_sz	- size of the OUTPUT buffer;

*/

/***************************************************************
 * Helper macros
 **************************************************************/
 
/* allocate on stack N variables of type T */
#define TNALLOCA(T, N)      			(T*)ALLOCA((N) * sizeof(T))

/* allocate on stack a variable of type T */
#define TALLOCA(T)         				TNALLOCA(T, 1)

/* allocate a structure of type 'type' and declare a pointer 'var' to it */
#define ALLOCA_STRUCT(var,type)		type * var = TALLOCA(type)  


/***************************************************************
 * INPUT buffer allocation 
 **************************************************************/

/* no IN buffer */
#define W_NO_IBUF					u_int32_t pi_sz = 0; void * pi = NULL

/* allocate INPUT buffer of type 'type' (declares and inits 'pi' and 'pi_sz'); */
#define W_ALLOCA_IBUF(type)			ALLOCA_STRUCT(pi, struct i_##type##_ops_t); \
	u_int32_t pi_sz = sizeof(struct i_##type##_ops_t)

/* allocate INPUT buffer of type 'type' and size 'size' (declares and inits 'pi' and 'pi_sz'); */
#define W_ALLOCA_IBUF_SZ(type, size)	 u_int32_t pi_sz = size;		\
	struct i_##type##_ops_t *pi = (struct i_##type##_ops_t *)ALLOCA(size)
	
/* allocate INPUT buffer of type 'type', extended after field 'name' by 'size' bytes (declares and inits 'pi' and 'pi_sz'); */
#define W_ALLOCA_IBUF_EX(type,name,size)		\
	u_int32_t pi_sz = (u_int32_t)(FIELD_OFFSET(struct i_##type##_ops_t,name) + size);	\
	struct i_##type##_ops_t *pi = (struct i_##type##_ops_t *)ALLOCA(pi_sz)

/* use integral input parameter 'name' as an INPUT buffer of type 'type' (declares and inits 'pi' and 'pi_sz'); */
#define W_USE_VAR_AS_IBUF(name,type)		u_int32_t pi_sz = sizeof(struct i_##type##_ops_t);	\
	struct i_##type##_ops_t *pi = (struct i_##type##_ops_t *)&name
	
/* use string input parameter 'name' as an INPUT buffer of type 'type' (declares and inits 'pi' and 'pi_sz'); */
#define W_USE_STRING_AS_IBUF(name,type)	u_int32_t pi_sz = strlen(name)+1;	\
	struct i_##type##_ops_t *pi = (struct i_##type##_ops_t *)name


/***************************************************************
 * INPUT buffer filling
 **************************************************************/

/* copy an input parameter of integral type to the INPUT buffer */
#define W_PUT_VAR(name)				pi->name = name
    
/* copy an input parameter of a non-integral type to the INPUT buffer */
#define W_PUT_LONGVAR(name)    			memcpy( (void*)&pi->name, (void*)&name, sizeof(pi->name) )

/* copy an input parameter of a structure type to the INPUT buffer */
#define W_PUT_STRUCT(name)    			memcpy( (void*)&pi->name, (void*)name##_p, sizeof(pi->name) )
    
/* copy a buffer to the INPUT buffer */
#define W_PUT_BUF(name,size)			memcpy( (void*)pi->name, (void*)name##_p, (size_t)size )

/* copy an ASCIIZ string to the INPUT buffer */
#define W_PUT_STRING(name)				do {                                                  \
    strncpy((void*)pi->name,(void*)name,sizeof(pi->name)-1);                            \
    ((char*)pi->name)[sizeof(pi->name)-1]= 0;   /* Assure string in NULL terminated */  \
  } while(0)


/***************************************************************
 * OUTPUT buffer allocation 
 **************************************************************/

/* no OUT buffer */
#define W_NO_OBUF					u_int32_t po_sz = 0; void * po = NULL

/* allocate OUTPUT buffer of type 'type' (declares and inits 'po' and 'po_sz'); */
#define W_ALLOCA_OBUF(type)			ALLOCA_STRUCT(po, struct o_##type##_ops_t); \
	u_int32_t po_sz = sizeof(struct o_##type##_ops_t)

/* allocate OUTPUT buffer of type 'type' and size 'size' (declares and inits 'po' and 'po_sz'); */
#define W_ALLOCA_OBUF_SZ(type, size)	 u_int32_t po_sz = size;	\
	struct o_##type##_ops_t *po = (struct o_##type##_ops_t *)ALLOCA(size)
	
/* allocate OUTPUT buffer of type 'type', extended after field 'name' by 'size' bytes (declares and inits 'po' and 'po_sz'); */
#define W_ALLOCA_OBUF_EX(type,name,size)		\
	u_int32_t po_sz = (u_int32_t)(FIELD_OFFSET(struct o_##type##_ops_t,name) + size);	\
	struct o_##type##_ops_t *po = (struct o_##type##_ops_t *)ALLOCA(po_sz)


/***************************************************************
 * OUTPUT buffer emptying
 **************************************************************/

/* copy a field of OUTPUT bufer of integral type to the variable, pointed by 'name' */
#define W_GET_VAR(name)				*name##_p = po->name

/* copy an output parameter of a structure type from the OUTPUT buffer */
#define W_GET_LONGVAR(name)    			memcpy( (void*)&name, (void*)&po->name, sizeof(po->name) )
    
/* copy an output parameter of a structure type from the OUTPUT buffer */
#define W_GET_STRUCT(name)    			memcpy( (void*)name##_p, (void*)&po->name, sizeof(po->name) )
    
/* copy a buffer from the OUTPUT buffer */
#define W_GET_BUF(name,size)			memcpy( (void*)name##_p, (void*)&po->name, (size_t)size )

/* copy an ASCIIZ string from the OUTPUT buffer */
#define W_GET_STRING(name)				W_GET_BUF(name,strlen(name##_p)+1)


/***************************************************************
 * Various
 **************************************************************/

/* return code of the kernel function */
#define W_RET							po->ret

/* standard return from a function */
#define W_RETURN(fcn)	\
	if (W_RET) {	\
	    	MTL_DEBUG4(MT_FLFMT(#fcn ": func_rc=" U64_FMT ", func_num=%d, pi_sz=%u, po_sz=%u"), (u_int64_t)W_RET, (int)fcn, pi_sz, po_sz); \
	} \
	return W_RET

/* call the function, performing the call to kernel; on error - return  */
#define W_ISSUE_IOCTL(cmd)				\
	do { if (mosal_ioctl_wrapper( cmd,pi, pi_sz, po, po_sz )) return MT_ESYSCALL; }	\
    	while (W_RET == MT_EINTR)

/* call the function, performing the call to kernel; on error - return  */
#define W_ISSUE_IOCTL_NORC(cmd)				\
	mosal_ioctl_wrapper( cmd,pi, pi_sz, po, po_sz )

#include <mosal_wrap.h>

/* These inline static functions decode parameters
 * and delegate each to the appropriate MOSAL KL API */

/*****************************************
  struct i_MOSAL_MPC860_present_ops_t {
    IN int dummy;
  };
  struct o_MOSAL_MPC860_present_ops_t {
    OUT bool ret;
  };
******************************************/
bool MOSAL_MPC860_present( void )
{
	W_NO_IBUF;
	W_ALLOCA_OBUF(MOSAL_MPC860_present);
	W_ISSUE_IOCTL_NORC(MOSAL_MPC860_PRESENT);
	W_RETURN(MOSAL_MPC860_PRESENT);
}
    	
/*****************************************
  struct i_MOSAL_MPC860_read_ops_t {
    IN u_int32_t addr;
    IN u_int32_t size;
  };
  struct o_MOSAL_MPC860_read_ops_t {
    OUT call_result_t ret;
    OUT u_int8_t data[1];
  };
******************************************/
call_result_t MOSAL_MPC860_read( u_int32_t addr, u_int32_t size, void * data_p )
{
	W_ALLOCA_IBUF(MOSAL_MPC860_read);
	W_ALLOCA_OBUF_EX(MOSAL_MPC860_read,data,size);
	W_PUT_VAR(addr);
	W_PUT_VAR(size);
	W_ISSUE_IOCTL(MOSAL_MPC860_READ);
	if (W_RET == MT_OK) {
      		W_GET_BUF(data,size);
	}
	W_RETURN(MOSAL_MPC860_READ);
}
    	
/*****************************************
  struct i_MOSAL_MPC860_write_ops_t {
    IN u_int32_t addr;
    IN u_int32_t size;
    IN u_int8_t data[1];
  };
  struct o_MOSAL_MPC860_write_ops_t {
    OUT call_result_t ret;
  };
******************************************/
call_result_t MOSAL_MPC860_write( u_int32_t addr, u_int32_t size, void * data_p )
{
	W_ALLOCA_IBUF_EX(MOSAL_MPC860_write,data,size);
	W_ALLOCA_OBUF(MOSAL_MPC860_write);
	W_PUT_VAR(addr);
	W_PUT_VAR(size);
	W_PUT_BUF(data,size);
	W_ISSUE_IOCTL(MOSAL_MPC860_WRITE);
	W_RETURN(MOSAL_MPC860_WRITE);
}
    	
/*****************************************
  struct i_MOSAL_PCI_find_device_ops_t {
    IN u_int16_t vendor_id;
    IN u_int16_t dev_id;
    IN u_int16_t index;
  };
  struct o_MOSAL_PCI_find_device_ops_t {
    OUT call_result_t ret;
    OUT u_int8_t bus;
    OUT u_int8_t dev_func;
  };
******************************************/
/*****************************************
 struct i_MOSAL_PCI_get_cfg_hdr_ops_t {
    IN MOSAL_pci_dev_t pci_dev;
  };
  struct o_MOSAL_PCI_get_cfg_hdr_ops_t {
    OUT call_result_t ret;
    OUT MOSAL_PCI_cfg_hdr_t cfg_hdr;
  };
******************************************/
call_result_t MOSAL_PCI_get_cfg_hdr(MOSAL_pci_dev_t pci_dev, MOSAL_PCI_cfg_hdr_t * cfg_hdr_p )
{
	W_ALLOCA_IBUF(MOSAL_PCI_get_cfg_hdr);
	W_ALLOCA_OBUF(MOSAL_PCI_get_cfg_hdr);
	W_PUT_VAR(pci_dev);
	W_ISSUE_IOCTL(MOSAL_PCI_GET_CFG_HDR);
	if (W_RET == MT_OK) {
      		W_GET_STRUCT(cfg_hdr);
	}
	W_RETURN(MOSAL_PCI_GET_CFG_HDR);
}

/*****************************************
  struct i_MOSAL_PCI_read_io_byte_ops_t {
    IN u_int32_t addr;
  };
  struct o_MOSAL_PCI_read_io_byte_ops_t {
    OUT call_result_t ret;
    OUT u_int8_t data;
  };
******************************************/
call_result_t MOSAL_PCI_read_io_byte( u_int32_t addr, u_int8_t * data_p )
{
	W_ALLOCA_IBUF(MOSAL_PCI_read_io_byte);
	W_ALLOCA_OBUF(MOSAL_PCI_read_io_byte);
	W_PUT_VAR(addr);
	W_ISSUE_IOCTL(MOSAL_PCI_READ_IO_BYTE);
	if (W_RET == MT_OK) {
      		W_GET_VAR(data);
	}
	W_RETURN(MOSAL_PCI_READ_IO_BYTE);
}

/*****************************************
  struct i_MOSAL_PCI_read_io_word_ops_t {
    IN u_int32_t addr;
  };
  struct o_MOSAL_PCI_read_io_word_ops_t {
    OUT call_result_t ret;
    OUT u_int16_t data;
  };
******************************************/
call_result_t MOSAL_PCI_read_io_word( u_int32_t addr, u_int16_t * data_p )
{
	W_ALLOCA_IBUF(MOSAL_PCI_read_io_word);
	W_ALLOCA_OBUF(MOSAL_PCI_read_io_word);
	W_PUT_VAR(addr);
	W_ISSUE_IOCTL(MOSAL_PCI_READ_IO_WORD);
	if (W_RET == MT_OK) {
      		W_GET_VAR(data);
	}
	W_RETURN(MOSAL_PCI_READ_IO_WORD);
}

/*****************************************
  struct i_MOSAL_PCI_read_io_dword_ops_t {
    IN u_int32_t addr;
  };
  struct o_MOSAL_PCI_read_io_dword_ops_t {
    OUT call_result_t ret;
    OUT u_int32_t data;
  };
******************************************/
call_result_t MOSAL_PCI_read_io_dword( u_int32_t addr, u_int32_t * data_p )
{
	W_ALLOCA_IBUF(MOSAL_PCI_read_io_dword);
	W_ALLOCA_OBUF(MOSAL_PCI_read_io_dword);
	W_PUT_VAR(addr);
	W_ISSUE_IOCTL(MOSAL_PCI_READ_IO_DWORD);
	if (W_RET == MT_OK) {
      		W_GET_VAR(data);
	}
	W_RETURN(MOSAL_PCI_READ_IO_DWORD);
}

/*****************************************
  struct i_MOSAL_PCI_write_io_byte_ops_t {
    IN u_int32_t addr;
    IN u_int8_t data;
  };
  struct o_MOSAL_PCI_write_io_byte_ops_t {
    OUT call_result_t ret;
  };
******************************************/
call_result_t MOSAL_PCI_write_io_byte( u_int32_t addr, u_int8_t data )
{
	W_ALLOCA_IBUF(MOSAL_PCI_write_io_byte);
	W_ALLOCA_OBUF(MOSAL_PCI_write_io_byte );
	W_PUT_VAR(addr);
	W_PUT_VAR(data);
	W_ISSUE_IOCTL(MOSAL_PCI_WRITE_IO_BYTE);
 	W_RETURN(MOSAL_PCI_WRITE_IO_BYTE);
}

/*****************************************
  struct i_MOSAL_PCI_write_io_word_ops_t {
    IN u_int32_t addr;
    IN u_int16_t data;
  };
  struct o_MOSAL_PCI_write_io_word_ops_t {
    OUT call_result_t ret;
  };
******************************************/
call_result_t MOSAL_PCI_write_io_word( u_int32_t addr, u_int16_t data )
{
	W_ALLOCA_IBUF(MOSAL_PCI_write_io_word);
	W_ALLOCA_OBUF(MOSAL_PCI_write_io_word);
	W_PUT_VAR(addr);
	W_PUT_VAR(data);
	W_ISSUE_IOCTL(MOSAL_PCI_WRITE_IO_WORD);
	W_RETURN(MOSAL_PCI_WRITE_IO_WORD);
}

/*****************************************
  struct i_MOSAL_PCI_write_io_dword_ops_t {
    IN u_int32_t addr;
    IN u_int32_t data;
  };
  struct o_MOSAL_PCI_write_io_dword_ops_t {
    OUT call_result_t ret;
  };
******************************************/
call_result_t MOSAL_PCI_write_io_dword( u_int32_t addr, u_int32_t data )
{
	W_ALLOCA_IBUF(MOSAL_PCI_write_io_dword);
	W_ALLOCA_OBUF(MOSAL_PCI_write_io_dword);
	W_PUT_VAR(addr);
	W_PUT_VAR(data);
	W_ISSUE_IOCTL(MOSAL_PCI_WRITE_IO_DWORD);
	W_RETURN(MOSAL_PCI_WRITE_IO_DWORD);
}


/*****************************************
  struct i_MOSAL_get_counts_per_sec_ops_t {
    IN int dummy;
  };
  struct o_MOSAL_get_counts_per_sec_ops_t {
    OUT u_int64_t ret;
  };
******************************************/
u_int64_t MOSAL_get_counts_per_sec( void )
{
	W_NO_IBUF;
	W_ALLOCA_OBUF(MOSAL_get_counts_per_sec);
	W_ISSUE_IOCTL(MOSAL_GET_COUNTS_PER_SEC);
	W_RETURN(MOSAL_GET_COUNTS_PER_SEC);
}

/*****************************************
  struct i_MOSAL_get_page_shift_ops_t {
    IN MT_virt_addr_t va;
    IN MOSAL_prot_ctx_t prot_ctx;
  };
  struct o_MOSAL_get_page_shift_ops_t {
    OUT call_result_t ret;
    OUT unsigned int page_shift;
  };
******************************************/
call_result_t MOSAL_get_page_shift( MOSAL_prot_ctx_t prot_ctx, MT_virt_addr_t va, unsigned * page_shift_p )
{
	W_ALLOCA_IBUF(MOSAL_get_page_shift);
	W_ALLOCA_OBUF(MOSAL_get_page_shift);
	W_PUT_LONGVAR(va);
	W_PUT_LONGVAR(prot_ctx);
	W_ISSUE_IOCTL(MOSAL_GET_PAGE_SHIFT);
	if (W_RET == MT_OK) {
      		W_GET_VAR(page_shift);
	}
	W_RETURN(MOSAL_GET_PAGE_SHIFT);
}

/*****************************************
  struct i_MOSAL_is_privileged_ops_t {
    IN int dummy;
  };
  struct o_MOSAL_is_privileged_ops_t {
    OUT MT_bool ret;
  };
******************************************/
MT_bool MOSAL_is_privileged( void )
{
	W_NO_IBUF;
	W_ALLOCA_OBUF(MOSAL_is_privileged);
	W_ISSUE_IOCTL_NORC(MOSAL_IS_PRIVILIGED);
	W_RETURN(MOSAL_IS_PRIVILIGED);
}

/*****************************************
  struct i_MOSAL_virt_to_phys_ops_t {
    IN MT_virt_addr_t va;
    IN MOSAL_prot_ctx_t prot_ctx;
  };
  struct o_MOSAL_virt_to_phys_ops_t {
    OUT MT_phys_addr_t pa;
    OUT call_result_t ret;
  };
******************************************/
call_result_t MOSAL_virt_to_phys( MOSAL_prot_ctx_t prot_ctx, const MT_virt_addr_t va, MT_phys_addr_t * pa_p )
{
	W_ALLOCA_IBUF(MOSAL_virt_to_phys);
	W_ALLOCA_OBUF(MOSAL_virt_to_phys);
	W_PUT_LONGVAR(va);
	W_PUT_LONGVAR(prot_ctx);
	W_ISSUE_IOCTL(MOSAL_VIRT_TO_PHYS);
	if (W_RET == MT_OK) {
      		W_GET_STRUCT(pa);
	}
	W_RETURN(MOSAL_VIRT_TO_PHYS);
}


call_result_t MOSAL_undo_register(MT_virt_addr_t va, MT_size_t size)
{
	W_ALLOCA_IBUF(MOSAL_undo_register);
	W_ALLOCA_OBUF(MOSAL_undo_register);
	W_PUT_LONGVAR(va);
	W_PUT_LONGVAR(size);
	W_ISSUE_IOCTL(MOSAL_UNDO_REGISTER);
	W_RETURN(MOSAL_UNDO_REGISTER);
}


/*****************************************
  struct i_k2u_cbk_pollq_ops_t {
    IN k2u_cbk_hndl_t k2u_cbk_h;
  };
  struct o_k2u_cbk_pollq_ops_t {
    OUT MT_size_t size;
    OUT call_result_t ret;
    OUT k2u_cbk_id_t cbk_id;
    OUT u_int8_t data[MAX_CBK_DATA_SZ];
  };
******************************************/
call_result_t k2u_cbk_pollq( k2u_cbk_hndl_t k2u_cbk_h, k2u_cbk_id_t * cbk_id_p, void * data_p, MT_size_t * size_p )
{
	W_ALLOCA_IBUF(k2u_cbk_pollq);
	W_ALLOCA_OBUF(k2u_cbk_pollq);
	W_PUT_LONGVAR(k2u_cbk_h);
	W_ISSUE_IOCTL(MOSAL_K2U_CBK_POLLQ);
	if (W_RET == MT_OK) {
      		W_GET_STRUCT(cbk_id);
      		W_GET_STRUCT(size);
      		W_GET_BUF(data, MAX_CBK_DATA_SZ);
	}
	W_RETURN(MOSAL_K2U_CBK_POLLQ);
}

/*****************************************
  struct i_mtl_log_set_ops_t {
    IN char layer[MAX_MTL_LOG_LAYER];
    IN char info[MAX_MTL_LOG_INFO];
  };
  struct o_mtl_log_set_ops_t {
    IN int dummy;
  };
******************************************/
void mtl_log_set( char * layer, char * info )
{
	W_ALLOCA_IBUF(mtl_log_set);
	W_NO_OBUF;
	W_PUT_STRING(layer);
	W_PUT_STRING(info);
	W_ISSUE_IOCTL_NORC(MOSAL_MTL_LOG_SET);
}

