/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#define C_K2U_CBK_U_C

#include <syscall.h>
#include <string.h>
#include <pthread.h>
#include <sys/ioctl.h>

#include "mosal_priv.h"
#include "mosal_k2u_cbk_priv.h"

#include <mosal_wrap.h>

extern call_result_t _dev_mosal_open(  int *fd_p );

static int mosal_fd=-1;


static int update_mosal_fd(void);

/* Maximum callbacks per process */
#define MAX_CBK 32

static pthread_mutex_t k2u_lock= PTHREAD_MUTEX_INITIALIZER;  /* Protect global context define below */
static k2u_cbk_hndl_t this_proc_cbk_hndl; /* Init when registering first callback */
static k2u_cbk_t cbks[MAX_CBK]={NULL};   /* callbacks for this process (index=id) */
static unsigned int num_o_cbks= 0;

/* polling thread status */
static volatile enum {
  POLLT_INIT, /* Thread is on initialization phase */
  POLLT_RUN,  /* Thread is running */
  POLLT_EXIT_DONE  /* Thread cleanup done */
}
poll_thread_state = POLLT_EXIT_DONE;


/* handler (thread) for callbacks info from kernel level */
static void *cbk_polling_thread(void* dummy_ctx)
{
  k2u_cbk_id_t id;
  u_int8_t data[MAX_CBK_DATA_SZ];
  MT_size_t size;
  call_result_t rc;

  /* Make initialization of kernel resources from this thread to allow owner validation in kernel */
  if ( (rc=k2u_cbk_init(&this_proc_cbk_hndl)) != MT_OK ) {
    MTL_ERROR1(MT_FLFMT("%s: k2u_cbk_init failed - %s"), __func__, mtl_strerror(rc));
    poll_thread_state = POLLT_EXIT_DONE;
    return NULL;
  }
  
  poll_thread_state = POLLT_RUN;

  /* poll private message queue until MT_EAGAIN (empty) */
  while ( 1 ) {
    rc = k2u_cbk_pollq(this_proc_cbk_hndl, &id, data, &size);
    if (rc==MT_EINTR)  {
      MTL_DEBUG2(MT_FLFMT("k2u_cbk_pollq returned: %s"), mtl_strerror(rc));
      continue;
    }
    if (rc != MT_OK)  break;
    if (id == K2U_CBK_CLEANUP_CBK_ID)  break;
    if (pthread_mutex_lock(&k2u_lock) != 0)  break;
    if ((id < MAX_CBK) && (cbks[id] != NULL)) { /* found matching callback */
      cbks[id](id,data,size);  /* invoke callback */
    }
    pthread_mutex_unlock(&k2u_lock);
  }

  if ((rc == MT_ENORSC) || (rc == MT_EPERM) || (id == K2U_CBK_CLEANUP_CBK_ID) ) {
    MTL_DEBUG1(MT_FLFMT("k2u_cbk_pollq: polling de-registered(%s). k2u_cbk will be disabled."),
      mtl_strerror_sym(rc)); 
  } else {
    MTL_ERROR1(MT_FLFMT("k2u_cbk_pollq returned error (%s). k2u_cbk will be disabled."),
      mtl_strerror_sym(rc)); 
  }
  poll_thread_state = POLLT_EXIT_DONE;
  return NULL;
}

/*
 *  k2u_cbk_register
 */
call_result_t k2u_cbk_register(k2u_cbk_t cbk, k2u_cbk_hndl_t *cbk_hndl_p, k2u_cbk_id_t *cbk_id_p)
{
  k2u_cbk_id_t id;
  /* Polling thread resources */
  pthread_t polling_thread;
  pthread_attr_t thread_attr;
  MT_bool first_cbk;

  if ( !cbk ) {
    MTL_ERROR2(MT_FLFMT("%s: NULL callback"), __func__);
    return MT_EINVAL;
  }
  
  if (pthread_mutex_lock(&k2u_lock) != 0) {  
    return MT_EAGAIN;
  }

  /* Find free cbk id  */
  if ( num_o_cbks == MAX_CBK ) {
    MTL_ERROR2(MT_FLFMT("%s: All callback resources are used (%d)"), __func__, MAX_CBK);
    pthread_mutex_unlock(&k2u_lock);
    return MT_EAGAIN;
  }

  /* From this point we are sure that we will find at least one free id */
  for ( id=0; id<MAX_CBK; ++id ) {
    if ( cbks[id] == NULL ) break;
  }

#ifdef MAX_DEBUG
  /* this is not necessary since we already checked that there are free callback entries
     and we have not yet freed the lock */
  if ( id == MAX_CBK ) {
    MTL_ERROR1(MT_FLFMT("%s: No free callback entry was found - *** FATAL ***"), __func__);
    pthread_mutex_unlock(&k2u_lock);
    return MT_EAGAIN;
  }
#endif  

  num_o_cbks++;

  cbks[id]= cbk;  /* Assure assignment is done only when data is valid*/
  first_cbk = (num_o_cbks == 1);
  /* block additional allocations while setting up polling thread (with mutex unlocked) */ 
  if (first_cbk) num_o_cbks= MAX_CBK;
  pthread_mutex_unlock(&k2u_lock);

  if ( first_cbk ) {
    while ( poll_thread_state != POLLT_EXIT_DONE );  /* Wait for old thread to complete cleanup */
    poll_thread_state= POLLT_INIT;
    pthread_attr_init(&thread_attr);
    pthread_attr_setdetachstate(&thread_attr,PTHREAD_CREATE_DETACHED);
    if (pthread_create(&polling_thread,&thread_attr,cbk_polling_thread,NULL) != 0) {
      MTL_ERROR1(MT_FLFMT("%s: Failed creating polling thread"), __func__);
      goto thread_fail;
    }

    /* Wait for thread initialization */
    while ( poll_thread_state == POLLT_INIT ) {
      sched_yield();
    }
    if ( poll_thread_state != POLLT_RUN ) goto thread_fail; /* Thread exited due to error */
    num_o_cbks= 1; /* enable registering additinal callbacks on the same thread infrastructure */
  }

  *cbk_id_p= id;
  *cbk_hndl_p= this_proc_cbk_hndl;
  return MT_OK;
  
  thread_fail:
    poll_thread_state= POLLT_EXIT_DONE; /* Allow additional trials */
    cbks[id]= NULL;
    num_o_cbks= 0;
    return MT_EAGAIN;
}


call_result_t k2u_cbk_deregister(k2u_cbk_id_t cbk_id)
{

  if (pthread_mutex_lock(&k2u_lock) != 0) {  
    return MT_EAGAIN;
  }
  
  if ((cbk_id > MAX_CBK) || (cbks[cbk_id] == NULL)) {
    pthread_mutex_unlock(&k2u_lock);
    MTL_ERROR2("%s: Invalid callback ID (%d).\n", __func__, cbk_id);
    return MT_EINVAL;
  }
  /* Remove from table before freeing (avoid inconsistant data while getting signal) */
  cbks[cbk_id] = NULL;  
  num_o_cbks--;
  if (num_o_cbks == 0) { /* no more callbacks - terminate polling thread */
    k2u_cbk_cleanup(this_proc_cbk_hndl);  /* free kernel resources */
    pthread_mutex_unlock(&k2u_lock);
    MTL_DEBUG4("%s: cleaned up this process cbk resources.\n", __func__);
    while ( poll_thread_state != POLLT_EXIT_DONE ) {
      sched_yield();
    }
  }
  else {
    pthread_mutex_unlock(&k2u_lock);
  }
  
  return MT_OK;
}


call_result_t k2u_cbk_init(k2u_cbk_hndl_t *k2u_cbk_h_p)
{
  int ioctl_rc;

  if ( mosal_fd == -1 ) {
    if ( update_mosal_fd() != 0 ) {
      return MT_ERROR;
    }
  }

  ioctl_rc = ioctl(mosal_fd, K2U_CBK_CBK_INIT, k2u_cbk_h_p);

  return ioctl_rc ? MT_ERROR : MT_OK;
}

call_result_t k2u_cbk_cleanup(k2u_cbk_hndl_t k2u_cbk_h)
{
  int ioctl_rc;

  if ( mosal_fd == -1 ) {
    /* when cleanup is called there must already be a valid file handle */
    return MT_ERROR;
  }

  ioctl_rc = ioctl(mosal_fd, K2U_CBK_CBK_CLEANUP, k2u_cbk_h);
  return ioctl_rc ? MT_ERROR : MT_OK;
}

static int update_mosal_fd(void)
{
	int fd;
	int rc = _dev_mosal_open(&fd);
	if (rc == MT_OK) {
		mosal_fd = fd;
		}
	return rc;
}
