/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <pthread.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <errno.h>


#include "mosal_priv.h"

#include <sys/mman.h>
#include <mosal_wrap.h>

/* Library resources */
extern call_result_t _dev_mosal_open(  int *fd_p );
static int mosal_fd= -1; /* MOSAL device file descriptor */
static pthread_mutex_t fd_lock= PTHREAD_MUTEX_INITIALIZER;


void MOSAL_user_lib_init(void)
{
	call_result_t rc;
  int fd=-1;

  pthread_mutex_lock(&fd_lock);
  if (mosal_fd != -1) { /* already initialized ? */
    pthread_mutex_unlock(&fd_lock);
    return; 
  }
  /* Open device in order to create MOSAL_mlock context */
  rc= _dev_mosal_open(&fd);
  if ((rc != MT_OK) || (fd == -1)) {
    pthread_mutex_unlock(&fd_lock);
    MTL_ERROR1(MT_FLFMT("MOSAL_user_lib_init(): Failed _dev_mosal_open() - (%s)"),
               (rc == MT_OK) ? "fd==-1": mtl_strerror_sym(rc) );
    return;
  }
  mosal_fd = fd; 
  pthread_mutex_unlock(&fd_lock);
}

 

/*
 * Map io memory to user address space
 *
 */
MT_virt_addr_t MOSAL_io_remap(MT_phys_addr_t pa, MT_size_t size)
{
		void *bp;
		int fd;	
		if(getuid() != 0){
				MTL_ERROR1("MOSAL_io_remap: Only root can do that\n");
				return(0);
		}

		
		/* Check for alignments */
		if(size <  MIN_PAGE_SZ_ALIGN(size) || pa <  MIN_PAGE_SZ_ALIGN(pa)){
				MTL_ERROR1("MOSAL_io_remap: Physical address or size not aligned to min. page size\n");
				return(0);
		}

#ifdef	STRICT_ALIGNEMENT

		if(pa < ((pa + (size - 1)) & (~((~0UL + size))))) {
				MTL_ERROR1("MOSAL_io_remap: Physical address must be sized aligned\n");
				return(0);
		}

#endif  /* STRICT ALIGNEMENT */
		
			
		/* Open the kernel mapping file */
		fd = open("/dev/kmem", O_RDWR);
		
		
		if(fd == -1){
				MTL_ERROR('1', "MOSAL_io_remap: Fail to open file /dev/kmem\n");
				return(0);
		}

		bp = mmap(0, size, (PROT_READ | PROT_WRITE), MAP_SHARED, fd, pa);

		if(bp == MAP_FAILED) {
				MTL_ERROR('1', "MOSAL_io_remap: Failed to mmap file\n");
				return(0);
		}                                                           
		
		close(fd);
		return((MT_virt_addr_t)bp);
}

//Generate a Linux key (int) from generic key (NULL terminated char*)
static int create_linux_key(MOSAL_shmem_key_t key) 
{
    int linux_key=0;
    u_int32_t pow;

    for ( pow=1; *key != 0;pow *=10 ) {
        linux_key += (*key)*pow;
        key++;
    }

    return linux_key;
}

call_result_t MOSAL_shmget(MOSAL_shmem_key_t key, 
                           MT_size_t size, 
                           u_int32_t flags, 
                           MOSAL_shmid_t * id_p)
{
    key_t linux_key = create_linux_key(key);
    int linux_flags=0;

    if (flags & MOSAL_SHM_CREATE) {
        linux_flags |= IPC_CREAT;
    }

    if (flags & MOSAL_SHM_EXCL) {
        linux_flags |= IPC_EXCL;
    }

    //Set permition bits
    if (flags & MOSAL_SHM_READONLY) {
        linux_flags |= 0444; //ReadOnly for everybody
    } else {
        linux_flags |= 0666; //ReadWrite for everybody
    }

// add huge pages flag (if supported)
#ifdef SHM_HUGETLB
    if (flags & MOSAL_SHM_HUGEPAGE) {
	linux_flags |= SHM_HUGETLB;
    }
#endif
    *id_p = shmget(linux_key, size, linux_flags);
    if (*id_p < 0) {
        switch (errno) {
            case EINVAL:  return MT_EINVAL;
            case EACCES: return MT_EACCES;
            case ENOMEM:  return MT_EAGAIN;
            case EEXIST:  return MT_EBUSY;
            default:      return MT_EINVAL;
        }
    } else {
        return MT_OK;
    }
}


call_result_t MOSAL_shmat(MOSAL_shmid_t id, int flags, void ** addr_p)
{
    int linux_flags=0;


    if (flags & MOSAL_SHM_READONLY) {
        linux_flags |= SHM_RDONLY;
    }

    *addr_p = shmat(id, *addr_p, linux_flags);

    if (*addr_p == (void*)(-1)) {
        switch (errno) {
            case EACCES: return MT_EACCES;
            case EINVAL: return MT_EINVAL;
            case ENOMEM: return MT_EAGAIN;
            default:     return MT_ERROR;
        }
    } else {
        return MT_OK;
    }
}

call_result_t MOSAL_shmdt(void * addr)
{
    if (shmdt(addr) == -1) {
        return MT_EINVAL;
    } else {
        return MT_OK;
    }
}

call_result_t MOSAL_shmrm(MOSAL_shmid_t id)
{
    if (shmctl(id, IPC_RMID, NULL) != 0) {
        switch (errno) {
            case EPERM: return MT_EACCES;
            default:    return MT_EINVAL;
        }
    } else {
        return MT_OK;
    }
}



MOSAL_pid_t MOSAL_getpid(void)
{
  return getpid();
}

/* Get process name */
const char* MOSAL_get_proc_name(void)
{
  return ""; /* TBD: check how to get process command line from global context */
}
