/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <sys/mman.h>
#include <asm/unistd.h>

#include <mtl_types.h>
#include <mosal.h>
#include <stdio.h>
#include <pthread.h>

static inline void *VOID2MT_VIRT(MT_virt_addr_t va)
{
  return(void *)va;
}


static inline size_t MT_SIZE2SIZET(MT_size_t size)
{
  return(size_t)size;
}

static inline call_result_t LINUX2MTL_RC(int rc_errno)
{
  /* put something here */
  switch ( rc_errno ) {
    case 0: return MT_OK;
    case EAGAIN: return MT_EAGAIN;
    case EPERM: return MT_EPERM;
    case ENOMEM: return MT_ENOMEM;
    default: return MT_ERROR;
  }
}


#define PARANOID

/********************************************************************
   list_head implementation is taken from the Linux
   source code but preceded with mt_ prefix
*********************************************************************/

struct mt_list_head {
  struct mt_list_head *next, *prev;
};


/**
 * mt_list_for_each_safe	-	iterate over a list safe against removal of list entry
 * @pos:	the &struct mt_list_head to use as a loop counter.
 * @n:		another &struct mt_list_head to use as temporary storage
 * @head:	the head for your list.
 */
#define mt_list_for_each_safe(pos, n, head) \
	for (pos = (head)->next, n = pos->next; pos != (head); \
		pos = n, n = pos->next)


/**
 * mt_list_entry - get the struct for this entry
 * @ptr:	the &struct mt_list_head pointer.
 * @type:	the type of the struct this is embedded in.
 * @member:	the name of the list_struct within the struct.
 */
#define mt_list_entry(ptr, type, member) \
    (type *)((unsigned long)(ptr)-((unsigned long)(&(((type *)0)->member))))


/*
 * Delete a list entry by making the prev/next entries
 * point to each other.
 *
 * This is only for internal list manipulation where we know
 * the prev/next entries already!
 */
static inline void __mt_list_del(struct mt_list_head * prev, struct mt_list_head * next)
{
  next->prev = prev;
  prev->next = next;
}

/**
 * mt_list_del - deletes entry from list.
 * @entry: the element to delete from the list.
 * Note: list_empty on entry does not return true after this, the entry is
 * in an undefined state.
 */
static inline void mt_list_del(struct mt_list_head *entry)
{
  __mt_list_del(entry->prev, entry->next);
  entry->next = NULL;
  entry->prev = NULL;
}


/*
 * Insert a new entry between two known consecutive entries.
 *
 * This is only for internal list manipulation where we know
 * the prev/next entries already!
 */
static inline void __mt_list_add(struct mt_list_head *new,
                                 struct mt_list_head *prev,
                                 struct mt_list_head *next)
{
  next->prev = new;
  new->next = next;
  new->prev = prev;
  prev->next = new;
}

/**
 * list_add - add a new entry
 * @new: new entry to be added
 * @head: list head to add it after
 *
 * Insert a new entry after the specified head.
 * This is good for implementing stacks.
 */
static inline void mt_list_add(struct mt_list_head *new, struct mt_list_head *head)
{
  __mt_list_add(new, head, head->next);
}



static struct mt_list_head rgn_list = {
  .prev = &rgn_list,
  .next = &rgn_list,
};

inline static unsigned long my_ulmin(unsigned long x, unsigned long y)
{
  if (x < y) {
    return x;
  }
  return y;
}

typedef struct {
  struct mt_list_head list; /* used to link a list of regions */
  unsigned long pg_start; /* start page number of this region */
  unsigned long pg_end; /* last page number of this region */
  u_int32_t count;  /* reference count on the region */
}
mregion_t;

static pthread_mutex_t mtx = PTHREAD_MUTEX_INITIALIZER;


/*
 *  clear_dont_copy_flag
 *    this function should be called after locking since
 *    locking the memory might split VMAs and we want to mark
 *    only those that we locked
 */
static call_result_t clear_dont_copy_flag(MT_virt_addr_t va, MT_size_t size,
                                       int mark)
{
  return MOSAL_undo_register(va, size);
}


static void cond_unlock_region(MT_virt_addr_t addr, MT_size_t size)
{
  munlock((void *)addr, size);
}


#define NEWREG(start, end, cnt, prev_link) ({                                                                 \
                                              mregion_t *ptr;                                                 \
                                              ptr = (mregion_t *)malloc(sizeof(mregion_t));                   \
                                              if ( ptr ) {                                                    \
                                                ptr->pg_start = (start);                                      \
                                                ptr->pg_end = (end);                                          \
                                                ptr->count = (cnt);                                           \
                                                mt_list_add(&ptr->list, (prev_link));                         \
                                              }                                                               \
                                              ptr;                                                            \
                                           })




static call_result_t remove_region(struct mt_list_head *head,
                                   unsigned long pg_start, unsigned long pg_end,
                                   u_int8_t page_shift, int unlock)
{
  struct mt_list_head *pos, *next, *save;
  mregion_t *cur = NULL;
  call_result_t rc = MT_OK;

  MTL_TRACE1(MT_FLFMT("%s: called for start=0x%lx, end=0x%lx"), __func__,
             pg_start, pg_end);
  if (pg_start > pg_end) {
    /* this case may happen deliberatly when insert_region fails
       and we call rollback when nothing was really done yet */
    return MT_EINVAL;
  }

  mt_list_for_each_safe(pos, next, head) {
    cur = mt_list_entry(pos, mregion_t, list);
    if (pg_start <= cur->pg_end)
      break;
  }

  if (pos == next) {
    /* list is empty */
    MTL_ERROR1(MT_FLFMT
               ("%s: trying to unlock unavailable region: start=0x%lx, end=0x%lx"),
               __func__, pg_start << page_shift,
               pg_end << page_shift);
    return MT_ENOMEM;
  }

  if (pg_start > cur->pg_end) {
    MTL_ERROR1(MT_FLFMT
               ("%s: trying to unlock unavailable region: start=0x%lx, end=0x%lx"),
               __func__, pg_start << page_shift,
               pg_end << page_shift);
    return MT_ENOMEM;
  }

  while (1) {
    if (pg_start < cur->pg_start) {
      unsigned long tmp_end =
      my_ulmin(pg_end, cur->pg_start - 1);
      MTL_ERROR1(MT_FLFMT
                 ("%s: trying to unlock unavailable region: start=0x%lx, end=0x%lx"),
                 __func__, pg_start << page_shift,
                 tmp_end << page_shift);
      rc = MT_ENOMEM;
      if (pg_end < cur->pg_start) {
        return rc;
      }
      pg_start = cur->pg_start;
    }
    /* pg_start >= cur->pg_start */

    if ((pg_start == cur->pg_start) && (pg_end == cur->pg_end)) {
      if (cur->count > 1) {
        cur->count--;
      }
      else {
        if ( unlock ) {
          cond_unlock_region(pg_start << page_shift,
                             (pg_end - pg_start +
                              1) << page_shift);
          clear_dont_copy_flag(pg_start << page_shift,
                            (pg_end - pg_start +
                             1) << page_shift, 0);
        }
        mt_list_del(pos);
        free(cur);
      }
      return rc;
    }

    if ((pg_start == cur->pg_start) && (pg_end < cur->pg_end)) {
      cur->pg_start = pg_end + 1;
      if (cur->count == 1) {
        if ( unlock ) {
          cond_unlock_region(pg_start << page_shift,
                             (pg_end - pg_start +
                              1) << page_shift);
          clear_dont_copy_flag(pg_start << page_shift,
                            (pg_end - pg_start +
                             1) << page_shift, 0);
        }
      }
      else {
        NEWREG(pg_start, pg_end, cur->count - 1,
               pos->prev);
      }
      return rc;
    }

    if ((pg_start == cur->pg_start) && (pg_end > cur->pg_end)) {
      pg_start = cur->pg_end + 1;
      save = pos->next;
      if (cur->count > 1) {
        cur->count--;
      }
      else {
        if ( unlock ) {
          cond_unlock_region(cur->pg_start << page_shift,
                             (cur->pg_end -
                              cur->pg_start +
                              1) << page_shift);
          clear_dont_copy_flag(cur->pg_start << page_shift,
                            (cur->pg_end - cur->pg_start +
                             1) << page_shift, 0);
        }
        mt_list_del(pos);
        free(cur);
      }
      if (save == head) {
        MTL_ERROR1(MT_FLFMT
                   ("%s: trying to unlock unavailable region: start=0x%lx, end=0x%lx"),
                   __func__, pg_start << page_shift,
                   pg_end << page_shift);
        return MT_ENOMEM;
      }
      else {
        pos = save;
        cur = mt_list_entry(pos, mregion_t, list);
        continue;
      }
    }

    if (pg_start > cur->pg_start) {
#ifdef PARANOID
      if (pg_start > cur->pg_end) {
        MTL_ERROR1(MT_FLFMT
                   ("%s: algorithm bug: start=0x%lx, end=0x%lx, cur->pg_start=0x%lx, cur->pg_end=0x%lx"),
                   __func__, pg_start << page_shift,
                   pg_end << page_shift,
                   cur->pg_start << page_shift,
                   cur->pg_end << page_shift);
        return MT_ENOMEM;
      }
#endif
      NEWREG(cur->pg_start, pg_start - 1, cur->count,
             pos->prev);
      cur->pg_start = pg_start;
      continue;
    }
  }
}


/*
 *  insert_region
 *       insert a region into the list of regions to a list of regions
 *
 *       head_p(in) pointer to the head of the list
 *       pg_start(in) page number of the start of the region
 *       pg_end(in) page number of the end of the region
 */
static call_result_t insert_region(struct mt_list_head *head,
                                   unsigned long pg_start,
                                   unsigned long pg_end,
                                   u_int8_t page_shift)
{
  struct mt_list_head *pos, *next;
  mregion_t *cur = NULL;
  call_result_t rc;
  unsigned long orig_start = pg_start, rback_end = pg_start - 1;

  MTL_TRACE1(MT_FLFMT("%s: called for start=0x%lx, end=0x%lx"), __func__,
             pg_start, pg_end);
  mt_list_for_each_safe(pos, next, head) {
    cur = mt_list_entry(pos, mregion_t, list);
    if (pg_start <= cur->pg_end)
      break;
  }

  if (pos == next) {
    /* this is the case when the list was empty so I just
       add the new entry to the head of the list */
    rc = MT_OK;
    if ((rc == MT_OK) && !NEWREG(pg_start, pg_end, 1, head)) {
      return MT_EAGAIN;
    }
    return rc;
  }

  if (pg_start > cur->pg_end) {
    /* all over this function the call to lock_region is neutralized
       since the whole region gets locked before calling this function */
    rc = MT_OK;
    if ((rc == MT_OK) && !NEWREG(pg_start, pg_end, 1, pos->prev)) {
      return MT_EAGAIN;
    }
    return rc;
  }

  while (1) {

    if (pg_start > cur->pg_end) {
      rc = MT_OK;
      if (rc == MT_OK) {
        if (!NEWREG(pg_start, pg_end, 1, pos->prev)) {
          return MT_EAGAIN;
        }
      }
      return rc;
    }

    if (pg_start < cur->pg_start) {
      unsigned long tmp_end =
      my_ulmin(pg_end, cur->pg_start - 1);

      rc = MT_OK;
      if (rc == MT_OK) {
        if (((cur->pg_start - pg_end) == 1)
            && (cur->count == 1)) {
          /* cur may be expanded downward */
          cur->pg_start = pg_start;
          return rc;
        }
        else {
          if (!NEWREG
              (pg_start, tmp_end, 1, pos->prev)) {
            /* rollback */
            remove_region(head, orig_start,
                          rback_end,
                          page_shift, 0);
            return MT_EAGAIN;
          }
          if (pg_end < cur->pg_start) {
            return rc;
          }
          else {
            pg_start = cur->pg_start;
            rback_end = cur->pg_start - 1;
            continue;
          }
        }
      }
      else {
        /* lock failed - rollback */
        remove_region(head, orig_start, rback_end,
                      page_shift, 0);
        return rc;
      }
    }
    else {
      /* pg_start >= cur->pg_start */
      if ((pg_start == cur->pg_start)
          && (pg_end == cur->pg_end)) {
        cur->count++;
        return MT_OK;
      }
      if ((pg_start == cur->pg_start)
          && (pg_end < cur->pg_end)) {
        if (!NEWREG
            (pg_start, pg_end, cur->count + 1,
             pos->prev)) {
          /* rollback */
          remove_region(head, orig_start,
                        rback_end, page_shift, 0);
          return MT_EAGAIN;
        }
        cur->pg_start = pg_end + 1;
        /* merge with prev ?? */
        return MT_OK;
      }
      if ((pg_start == cur->pg_start)
          && (pg_end > cur->pg_end)) {
        cur->count++;
        pg_start = cur->pg_end + 1;
        rback_end = cur->pg_end;
        if (pos->next == head) {
          if (!NEWREG(pg_start, pg_end, 1, pos)) {
            remove_region(head, orig_start,
                          rback_end,
                          page_shift, 0);
            return MT_EAGAIN;
          }
          return MT_OK;
        }
        pos = pos->next;
        cur = mt_list_entry(pos, mregion_t, list);
        continue;
      }
#ifdef PARANOID
      if (pg_start <= cur->pg_start) {
        MTL_ERROR1(MT_FLFMT
                   ("%s: algorithm bug: start=0x%lx, end=0x%lx, cur->pg_start=0x%lx, cur->pg_end=0x%lx"),
                   __func__, pg_start << page_shift,
                   pg_end << page_shift,
                   cur->pg_start << page_shift,
                   cur->pg_end << page_shift);
      }
#endif

      if (pg_start > cur->pg_start) { // this if seems redundant !!! ???
        if (!NEWREG
            (cur->pg_start, pg_start - 1, cur->count,
             pos->prev)) {
          /* rollback */
          remove_region(head, orig_start,
                        rback_end, page_shift, 0);
          return MT_EAGAIN;
        }
        cur->pg_start = pg_start;
        rback_end = pg_start - 1;
        continue;
      }
    }
  }
}



struct mlock_op_st {
  int retval;
  unsigned long addr;
  unsigned int size;
};

int mlx_mlock(void *va, unsigned long size)
{
  long int ret;

  ret = syscall(__NR_mlock, va, size);

  return ret;
}

/*
 *  MOSAL_memory_lock
 */ 
call_result_t MOSAL_memory_lock(MT_virt_addr_t va,
                              MT_size_t size)
{
  int rc, rc_errno;
  void *vva = VOID2MT_VIRT(va);
  size_t sz = MT_SIZE2SIZET(size);
  unsigned long pg_start, pg_end; /* normalized start and end in PAGE_SIZE units */

  pg_start = (unsigned long)(va >> MOSAL_SYS_PAGE_SHIFT);
  pg_end = (unsigned long)((va + size - 1) >> MOSAL_SYS_PAGE_SHIFT);

  if ( pthread_mutex_lock(&mtx) == 0 ) {
    rc = mlock(vva, sz);
    if ( rc == 0 ) {
      insert_region(&rgn_list, pg_start, pg_end, MOSAL_SYS_PAGE_SHIFT);
	  rc_errno = 0;
    }
    else {
		rc_errno = errno;
    }
    pthread_mutex_unlock(&mtx);
  }
  else {
    rc_errno = -1;
  }

  return LINUX2MTL_RC(rc_errno);
}


/*
 *  MOSAL_memory_unlock
 */ 
call_result_t MOSAL_memory_unlock(MT_virt_addr_t va,
                                MT_size_t size)
{
  int rc;
  unsigned long pg_start, pg_end; /* normalized start and end in PAGE_SIZE units */

  pg_start = (unsigned long)(va >> MOSAL_SYS_PAGE_SHIFT);
  pg_end = (unsigned long)((va + size - 1) >> MOSAL_SYS_PAGE_SHIFT);


  if ( pthread_mutex_lock(&mtx) == 0 ) {
    rc = remove_region(&rgn_list, pg_start, pg_end, MOSAL_SYS_PAGE_SHIFT, 1);
    pthread_mutex_unlock(&mtx);
  }
  else {
    rc = MT_ERROR;
  }

  return LINUX2MTL_RC(rc);
}
