/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <pthread.h>
#include <mosal.h>
#include <time.h>
#include <sys/time.h>
#include <semaphore.h>

/*
 *  MOSAL_mutex_init
 */
call_result_t MOSAL_mutex_init(MOSAL_mutex_t *mtx_p)
{
  pthread_mutex_init(mtx_p, 0);
  return MT_OK;
}


/*
 *  MOSAL_mutex_acq
 */
call_result_t MOSAL_mutex_acq(MOSAL_mutex_t *mtx_p, MT_bool block)
{
  int res;

  if ( block ) {
    res = pthread_mutex_lock(mtx_p);
  }
  else {
    res = pthread_mutex_trylock(mtx_p);
  }
  return res==0 ? MT_OK : MT_EAGAIN;
}


/*
 *  MOSAL_mutex_acq_ui
 */
void MOSAL_mutex_acq_ui(MOSAL_mutex_t *mtx_p)
{
  int res;

  res = pthread_mutex_lock(mtx_p);
  switch ( res ) {
    case 0:
      break;
    case EINVAL:
      MTL_ERROR1(MT_FLFMT("%s: the mutex has not been properly initialized"), __func__);
      break;
    case EDEADLK:
      MTL_ERROR1(MT_FLFMT("%s: the  mutex  is  already  locked  by  the  calling  thread"), __func__);
      break;
    default:
      MTL_ERROR1(MT_FLFMT("%s: unknown error"), __func__);
      break;
  }
}


/*
 *  MOSAL_mutex_rel
 */
void MOSAL_mutex_rel(MOSAL_mutex_t *mtx_p)
{
  pthread_mutex_unlock(mtx_p);
}



/*
 *  MOSAL_delay_execution
 */
void MOSAL_delay_execution(u_int32_t time_micro)
{
  struct timespec req;
  
  if ( time_micro < 1000000 ) {
    req.tv_sec = 0;
    req.tv_nsec = time_micro*1000;
  }
  else {
    req.tv_sec = time_micro / 1000000;
    req.tv_nsec = (long)(((u_int64_t)time_micro)*1000LL - (u_int64_t)(req.tv_sec)*1000000000LL);
  }
  nanosleep(&req, NULL);
}


/*
 *  MOSAL_spinlock_init
 */
call_result_t MOSAL_spinlock_init(MOSAL_spinlock_t  *sp)
{
  ul_spin_lock_init(sp);
  return MT_OK;
}


////////////////////////////////////////////////////////////////////////////////
// Semaphores
////////////////////////////////////////////////////////////////////////////////

/*
 *  MOSAL_sem_init
 */
call_result_t MOSAL_sem_init(MOSAL_semaphore_t *sem_p, MT_size_t count)
{
  int rc;
  rc = sem_init(&(sem_p->sem),0,count);
  switch ( rc ) {
    case -EINVAL:
      return MT_EINVAL;
    case -ENOSYS:
      return MT_ENOSYS;
    case 0:
      return MT_OK;
    default:
      return MT_ERROR;
  }
}


/*
 *  MOSAL_sem_acq
 */
call_result_t MOSAL_sem_acq(MOSAL_semaphore_t *sem_p, MT_bool block)
{
  if ( block ) {
    /* sem_wait always returns 0 */
    sem_wait(&(sem_p->sem));
    return MT_OK;
  }
  else {
    int irc;
    irc = sem_trywait(&(sem_p->sem));
    return irc==0 ? MT_OK : MT_EAGAIN;
  }
}


/*
 *  MOSAL_sem_acq
 */
void MOSAL_sem_acq_ui(MOSAL_semaphore_t *sem_p)
{
  sem_wait(&(sem_p->sem));
}

/*
 *  MOSAL_sem_rel
 */
void MOSAL_sem_rel(MOSAL_semaphore_t *sem_p)
{
  sem_post(&(sem_p->sem));
}


/* sync object dummies */
call_result_t MOSAL_syncobj_init(MOSAL_syncobj_t *obj_p)
{
  int rc;

  rc = pthread_mutex_init(&obj_p->mutex, NULL); /* init mutex with default attributes */
  if ( rc != 0 ) {
    return MT_ERROR;
  }

  rc = pthread_cond_init(&obj_p->cond, NULL); /* init the conditional variable with default attributes */
  if ( rc != 0 ) {
    return MT_ERROR;
  }
  return MT_OK;
}

/*
 *  MOSAL_syncobj_waiton
 */
call_result_t MOSAL_syncobj_waiton(MOSAL_syncobj_t *obj_p, MT_size_t micro_sec)
{
  struct timeval now;
  struct timespec timeout;
  unsigned long sec, nsec;
  int rc;

  if ( pthread_mutex_lock(&obj_p->mutex) != 0 ) {
    MTL_ERROR1(MT_FLFMT("%s: failed to lock mutex"), __func__);
    return MT_ERROR;
  }

  if ( micro_sec != MOSAL_SYNC_TIMEOUT_INFINITE ) {
    gettimeofday(&now, 0);
	sec = micro_sec / 1000000;
	nsec = (micro_sec - sec * 1000000) * 1000;
    timeout.tv_sec = now.tv_sec + sec;
    timeout.tv_nsec = now.tv_usec * 1000 + nsec;
	if ( timeout.tv_nsec >= 1000000000 ) {
			timeout.tv_sec++;
			timeout.tv_nsec -= 1000000000;
	}

    do {
      rc = pthread_cond_timedwait(&obj_p->cond, &obj_p->mutex, &timeout);
    } while ( (rc!=0) && (rc!=ETIMEDOUT) );
  }
  else {
    rc = pthread_cond_wait(&obj_p->cond, &obj_p->mutex);
  }
  pthread_mutex_unlock(&obj_p->mutex);

  return rc==0 ? MT_OK : MT_ETIMEDOUT;
}


call_result_t MOSAL_syncobj_waiton_ui(MOSAL_syncobj_t *obj_p, MT_size_t micro_sec)
{
  return MOSAL_syncobj_waiton(obj_p, micro_sec);
}

/*
 *  MOSAL_syncobj_signal
 */
void MOSAL_syncobj_signal(MOSAL_syncobj_t *obj_p)
{
  if ( pthread_mutex_lock(&obj_p->mutex) != 0 ) {
    MTL_ERROR1(MT_FLFMT("failed to lock mutex"));
    return;
  }
  pthread_cond_broadcast(&obj_p->cond);
  pthread_mutex_unlock(&obj_p->mutex);
}

/*
 *  MOSAL_syncobj_clear
 */
void MOSAL_syncobj_clear(MOSAL_syncobj_t *obj_p)
{
  if ( pthread_mutex_lock(&obj_p->mutex) != 0 ) {
    MTL_ERROR1(MT_FLFMT("failed to lock mutex"));
    return;
  }
   /* init the conditional variable with default attributes */
  if (pthread_cond_init(&obj_p->cond, NULL) != 0) {
     MTL_ERROR1(MT_FLFMT("failed to init conditional variable"));
     return;
  }
  pthread_mutex_unlock(&obj_p->mutex);
}


int MOSAL_sleep(u_int32_t sec) 
{
    sleep(sec);
    return 0;
}

call_result_t MOSAL_usleep( u_int32_t usec )
{
    usleep(usec);
    return MT_OK;
}

void MOSAL_gettimeofday(MOSAL_time_t * time_p)
{
    struct timeval tv;	
    gettimeofday( &tv, NULL);

    time_p->sec = tv.tv_sec;
    time_p->msec = tv.tv_usec/1000;
}
