/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <pthread.h>
#include <mosal.h>

/******************************************************************************
 *  Function:
 *    MOSAL_thread_start 
 *
 *  Description:
 *    create a tread and run a t-function in its context
 *
 *  Parameters: 
 *		mto_p(IN) 		pointer to MOSAL thread object
 *		flags(IN)		flags for thread creation	
 *		mtf(IN) 		t-function  
 *		mtf_ctx(IN) 	t-function context
 *
 *  Returns:
 *		MT_OK		    - thread created
 *		MT_EAGAIN		- resources are temporarily un available
 *      MT_EPERM        - not enough premissions
 *      otherwise, MT_ERROR
 *
 ******************************************************************************/
call_result_t MOSAL_thread_start( 
	MOSAL_thread_t *mto_p,			// pointer to MOSAL thread object
	u_int32_t	flags,				// flags for thread creation	
	MOSAL_thread_func_t mtf,		// t-function name 
	void *mtf_ctx					// t-function context (optionally) 
 )
{	
	int ret = 0;
    pthread_attr_t  thread_attrs;

    if (mtf == NULL)
		return MT_EINVAL;

	// init thread object
	mto_p->func 		= mtf;
	mto_p->func_ctx 	= mtf_ctx;

    if (pthread_attr_init(&thread_attrs) ) {
		return MT_ERROR;
    }
    if (pthread_attr_setdetachstate(&thread_attrs, 
        flags & MOSAL_THREAD_FLAGS_DETACHED ? PTHREAD_CREATE_DETACHED : PTHREAD_CREATE_JOINABLE)) {
		return MT_ERROR;
    }
    
    // create and run the thread
  	ret = pthread_create(&(mto_p->thread),&thread_attrs,mto_p->func,mto_p->func_ctx);
	if (ret == ENOMEM || ret == EAGAIN) {
        return MT_EAGAIN;
    }
    if (ret == EPERM) {
        return MT_EPERM;
    }
    if (ret != 0)
		return MT_ERROR;
  	
	return MT_OK;

}	

/******************************************************************************
 *  Function:
 *    MOSAL_thread_kill 
 *
 *  Description:
 *    terminate the tread brutally
 *
 *  Parameters: 
 *		mto_p(IN) 		pointer to MOSAL thread object
 *
 *  Returns:
 *		MT_OK		- thread terminated
 *		MT_ERROR	- a failure on thread termination
 *
 ******************************************************************************/
call_result_t MOSAL_thread_kill( 
	MOSAL_thread_t *mto_p			// pointer to MOSAL thread object
 )
 {
	if (pthread_cancel(mto_p->thread) == 0)
        return MT_OK; 
     
     return MT_ERROR;
 }

/******************************************************************************
 *  Function:
 *    MOSAL_thread_wait_for_term 
 *
 *  Description:
 *    wait for a thread to exit (only if not detached)
 *
 *  Parameters: 
 *		mto_p(IN) 		pointer to MOSAL thread objectto exit
 *		exit_code(OUT)	return code of the thread
 *
 *  Returns:
 *		MT_OK		- thread started; for blocking mode - t-function is running;
 *		MT_EAGAIN	- thread hasn't started yet; 
 *		other		- error;
 *
 ******************************************************************************/
call_result_t MOSAL_thread_wait_for_term( 
	MOSAL_thread_t *mto_p,			// pointer to MOSAL thread object
	u_int32_t	*exit_code			// return code of the thread
	)
{
	int ret;
    
    
    ret = pthread_join(mto_p->thread,(void*)(&exit_code));
    if (ret == 0) {
        return MT_OK;
    }
    if (ret == ESRCH) {
        return MT_EAGAIN;
    }
    return MT_ERROR;
}

