/*
*-------------------------------------------------------------------
* Test that check the APM (Automatic Path Migration)
* the test has 2 sides, one of them post RDMA write all the time, and
* the second side is the listener.
* The user should remove the cable from the primary port, and there will
* be APM to the other side.
* 
* one side should be started with -d, and the other with --ip=*.*.*.*, for example:
*  apm_example -d
*  apm_example --ip=10.3.6.8
*-------------------------------------------------------------------								 
*  Version: $Id: main.c,v 1.3 2004/09/27 15:27:36 dotanb Exp $
*  Author:  Dotan Barak (dotanb@mellanox.co.il)
*-------------------------------------------------------------------
*/

#include <stdio.h>
#include <mosal.h>
#include <vapi.h>
#include <vapi_common.h>

#include "types.h"
#include "cmd_pars.h"
#include "sock.h"

#define CHECK_VERB_RETURN_VALUE(verb, result)   if (result != VAPI_OK) {                                            \
                                            printf("(E)Verb %s failed: %s, %s\n", verb, VAPI_strerror_sym(result), VAPI_strerror(result));     \
                                            return result;                                                          \
                                            }

config_t config =   {
    "InfiniHost0",      /* hca_id       */
    0,                  /* is_daemon    */
    20000,              /* tcp port     */
    "0",                /* ip           */
    1,                  /* ib_port      */
    2                   /* alt_ib_port  */
};


/* function declerations */
void async_event_handler( VAPI_hca_hndl_t hca_hndl, VAPI_event_record_t* event_record_p, void* private_data );
void print_config(void);
void resources_initialize(resources_t *resources);
VAPI_ret_t resources_create(resources_t *resources);
VAPI_ret_t modify_qp_2_rts(resources_t *resources, struct connection_data_t *remote_data);
VAPI_ret_t do_test(resources_t *resources, struct connection_data_t *remote_data);
VAPI_ret_t resources_destroy(resources_t *resources);

/* this variable mark the daemon when to get out from the loop in which he sends data to the client */
volatile int apm_occured = 0;

int main(int argc, char* argv[])
{
    VAPI_ret_t rc = VAPI_EBUSY;
    int test_failed;
    struct connection_data_t local_con_data, remote_con_data, tmp_con_data;
    resources_t resources;


    test_failed = 1;
    
    /* read the command line parameters */
    rc = parse_argv( argc, argv );
    if( rc )
    {
        return rc;
    }

    /* print test configuration */
    print_config();

    resources_initialize(&resources);
    rc = resources_create(&resources);
    if (rc != VAPI_OK)
    {
        goto cleanup;
    }

    /* prepare data to be sent to the other side */
    printf("local props: lid=%u, qp_number=0x%x, start="U64_FMT", rkey=0x%x\n",
           resources.hca_port.lid, resources.qp_prop.qp_num, resources.rep_mrw.start, resources.rep_mrw.r_key);

    local_con_data.lid = htons(resources.hca_port.lid);
    local_con_data.alt_lid = htons(resources.alt_hca_port.lid);
    local_con_data.qp_num = htonl(resources.qp_prop.qp_num);
    local_con_data.start = MOSAL_cpu_to_be64(resources.rep_mrw.start);
    local_con_data.rkey = htonl(resources.rep_mrw.r_key);


    /* exchange the date between the two sides */
    rc = sock_cm_data_sync( resources.sockfd, config.is_daemon, 
                            (char*) &local_con_data, (char*)&tmp_con_data, sizeof(local_con_data));
    if (rc)
    {
        printf( "Connection was lost when transfering connection info\n");
        goto cleanup;
    }

    /* convert the network endian to host endian */
    remote_con_data.lid = ntohs(tmp_con_data.lid);
    remote_con_data.alt_lid = ntohs(tmp_con_data.alt_lid);
    remote_con_data.qp_num = ntohl(tmp_con_data.qp_num);
    remote_con_data.start = MOSAL_be64_to_cpu(tmp_con_data.start);
    remote_con_data.rkey = ntohl(tmp_con_data.rkey);

    printf("remote props: lid=%u, qp_number=0x%x, start="U64_FMT", rkey=0x%x\n",
           remote_con_data.lid, remote_con_data.qp_num, remote_con_data.start, remote_con_data.rkey);

    rc = modify_qp_2_rts(&resources, &remote_con_data);
    if (rc != VAPI_OK)
    {
        goto cleanup;
    }

    /* sync between sides, so they both be in RTS before starting to send data */
    rc = sock_cm_ready_sync(resources.sockfd, config.is_daemon);
    if (rc)
    {
        printf( "Connection was lost when while sync\n");
        goto cleanup;
    }

    rc = do_test(&resources, &remote_con_data);
    if (rc)
    {
        goto cleanup;
    }

    /* sync between sides before the test ends */
    rc = sock_cm_ready_sync(resources.sockfd, config.is_daemon);
    if (rc)
    {
        printf( "Connection was lost when while synchronising\n");
        goto cleanup;
    }

    test_failed = 0;
cleanup:

    rc = resources_destroy(&resources);
    if (rc != VAPI_OK)
    {
        test_failed = 1;
    }

    if (test_failed) {
        printf("Test Failed\n");
        return 1;
    } else
    {
        printf("Test PASSED\n");
        return 0;
    }
}

/*****************************************************************
 async event handler: only PORT_ERROR and PATH_MIGRATED events are expected
*****************************************************************/
void async_event_handler( VAPI_hca_hndl_t hca_hndl, VAPI_event_record_t* event_record_p, void* private_data )
{
    printf("Got asynchronous event. Type: %s Syndrome: %s\n", VAPI_event_record_sym(event_record_p->type), VAPI_event_syndrome_sym(event_record_p->syndrome));

    switch (event_record_p->type)
    {
    case VAPI_PORT_ERROR: /* don't do nothing on this event */
        break;

    case VAPI_PATH_MIG_ARMED:
        printf(" [*] Alternate Path was set, and QP was ARMed\n");
        break;

    case VAPI_QP_PATH_MIGRATED:
        printf(" [*] Path was migrated\n");
        apm_occured = 1;
        break;

    default:
        printf("Error: the event was unexpected\n");
        exit(1);
    }
}

/*****************************************************************
 print the configuration of the test
*****************************************************************/
void print_config()
{
    printf(" apm_example configuration report\n" );
    printf(" ------------------------------------------------\n" );
    printf(" Current pid                  : "MT_PID_FMT"\n", MOSAL_getpid() );
    printf(" HCA ID (name)                :\"%s\"\n", config.hca_id );
    printf(" Process Mode                 : %s\n", (config.is_daemon)?"Daemon":"Client");
    printf(" TCP Port                     : %u\n", config.tcp_port );
    printf(" IP                           : %s\n", config.ip);
    printf(" Primary IB port              : %u\n", config.ib_port);
    printf(" Alternate IB port            : %u\n", config.alt_ib_port);
    printf(" ------------------------------------------------\n\n" );
}

/*****************************************************************
 invalidate all the handles
*****************************************************************/
void resources_initialize(resources_t *resources)
{
    resources->hca_hndl = VAPI_INVAL_HNDL;
    resources->pd_hndl = VAPI_INVAL_PD_HNDL;
    resources->cq_hndl = VAPI_INVAL_HNDL;
    resources->qp_hndl = VAPI_INVAL_HNDL;
    resources->async_handler_hndl = VAPI_INVAL_HNDL;
    resources->mr_hndl = VAPI_INVAL_HNDL;
    resources->buf_ptr = NULL;
    resources->sockfd = -1;
}


/*****************************************************************
 create all the resources that the test uses
*****************************************************************/
VAPI_ret_t resources_create(resources_t *resources)
{
    VAPI_ret_t rc;
    VAPI_mrw_t req_mrw;
    VAPI_qp_init_attr_t qp_init_attr;
    VAPI_hca_vendor_t hca_vendor;
    
    rc = EVAPI_get_hca_hndl (config.hca_id, &(resources->hca_hndl));
    CHECK_VERB_RETURN_VALUE("EVAPI_get_hca_hndl", rc);
    printf(" [*] Get HCA handle\n");

    rc = EVAPI_set_async_event_handler(resources->hca_hndl, async_event_handler, NULL, &(resources->async_handler_hndl));
    CHECK_VERB_RETURN_VALUE("EVAPI_set_async_event_handler", rc);
    printf(" [*] Asynchronous event handler was set\n");

    rc = VAPI_query_hca_cap (resources->hca_hndl, &hca_vendor, &(resources->hca_cap));
    CHECK_VERB_RETURN_VALUE("VAPI_query_hca_cap", rc);
    printf(" [*] Query HCA capabilities\n");

    rc = VAPI_query_hca_port_prop (resources->hca_hndl, config.ib_port, &(resources->hca_port));
    CHECK_VERB_RETURN_VALUE("VAPI_query_hca_port_prop", rc);

    rc = VAPI_query_hca_port_prop (resources->hca_hndl, config.alt_ib_port, &(resources->alt_hca_port));
    CHECK_VERB_RETURN_VALUE("VAPI_query_hca_port_prop", rc);

    rc = VAPI_alloc_pd (resources->hca_hndl, &(resources->pd_hndl));
    CHECK_VERB_RETURN_VALUE("VAPI_alloc_pd", rc);
    printf(" [*] Allocate PD\n");

    req_mrw.size = MR_SIZE;
    resources->buf_ptr = MALLOC (req_mrw.size);
    if (resources->buf_ptr == NULL)
    {
        printf("Cannot allocate memory\n");
        return VAPI_ENOMEM;
    }
    req_mrw.start = (VAPI_virt_addr_t)(unsigned long)resources->buf_ptr;
    req_mrw.type = VAPI_MR;
    req_mrw.acl = VAPI_EN_LOCAL_WRITE | VAPI_EN_REMOTE_WRITE | VAPI_EN_REMOTE_READ;
    req_mrw.pd_hndl = resources->pd_hndl;
    
    rc = VAPI_register_mr (resources->hca_hndl, &req_mrw, &(resources->mr_hndl), &(resources->rep_mrw));
    CHECK_VERB_RETURN_VALUE("VAPI_register_mr", rc);
    printf(" [*] Register MR with "U64_FMT" bytes at address %p\n", /*page_size,*/
           resources->rep_mrw.size,(void*)(MT_virt_addr_t)resources->rep_mrw.start); 

    rc = VAPI_create_cq(resources->hca_hndl, CQ_SIZE, &(resources->cq_hndl), &(resources->cq_size));
    CHECK_VERB_RETURN_VALUE("VAPI_create_cq", rc);
    printf(" [*] Create Receive CQ with %u entries\n", resources->cq_size);

    qp_init_attr.rq_cq_hndl = resources->cq_hndl;
    qp_init_attr.sq_cq_hndl = resources->cq_hndl;
    qp_init_attr.pd_hndl = resources->pd_hndl;
    qp_init_attr.cap.max_oust_wr_rq = 1;
    qp_init_attr.cap.max_oust_wr_sq = 1;
    qp_init_attr.cap.max_sg_size_rq = 1;
    qp_init_attr.cap.max_sg_size_sq = 1;
    qp_init_attr.sq_sig_type = VAPI_SIGNAL_ALL_WR;
    qp_init_attr.rq_sig_type = VAPI_SIGNAL_ALL_WR;
    qp_init_attr.ts_type = VAPI_TS_RC;

    rc = VAPI_create_qp (resources->hca_hndl, &qp_init_attr, &(resources->qp_hndl), &(resources->qp_prop));
    CHECK_VERB_RETURN_VALUE("VAPI_create_qp", rc);
    printf(" [*] QP 0x%x was created\n", resources->qp_prop.qp_num);


    /* open a socket between the 2 sides */
    if (config.is_daemon)
    {
        printf(" [*] Daemon waiting for connection: "); 
        fflush(stdout);
        resources->sockfd = sock_daemon_wait(config.tcp_port);
    } else
    {
        printf(" [*] Client connecting to the daemon: "); 
        fflush(stdout);
        resources->sockfd = sock_client_connect(config.ip, config.tcp_port);
    }
    if (resources->sockfd < 0)
    {
        return VAPI_EGEN;
    }

    printf("Connection established\n");

    return VAPI_OK;
}

/*****************************************************************
 function to modify the QP state from reset to RTS
*****************************************************************/
VAPI_ret_t modify_qp_2_rts(resources_t *resources, struct connection_data_t *remote_data)
{
    VAPI_ret_t rc;
    VAPI_qp_attr_mask_t qp_attr_mask;
    VAPI_qp_attr_t qp_attr;
    VAPI_qp_cap_t qp_cap;
    

    /*/////////////////////////////////////////////////////////////////////////////////
    // change QP state Reset -> Init
    /////////////////////////////////////////////////////////////////////////////////*/
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);

    qp_attr.qp_state = VAPI_INIT;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_QP_STATE);

    qp_attr.pkey_ix  = DEF_PKEY_IX; 
    QP_ATTR_MASK_SET( qp_attr_mask, QP_ATTR_PKEY_IX );

    qp_attr.port = config.ib_port;
    QP_ATTR_MASK_SET( qp_attr_mask, QP_ATTR_PORT );

    QP_ATTR_MASK_SET( qp_attr_mask, QP_ATTR_REMOTE_ATOMIC_FLAGS );
    qp_attr.remote_atomic_flags = VAPI_EN_REM_WRITE | VAPI_EN_REM_READ | VAPI_EN_REM_ATOMIC_OP; 

    rc = VAPI_modify_qp(resources->hca_hndl, resources->qp_hndl, &qp_attr, &qp_attr_mask, &qp_cap);
    CHECK_VERB_RETURN_VALUE("VAPI_modify_qp", rc);

    printf(" [*] QP state was changed: Reset -> Init ");

    /*/////////////////////////////////////////////////////////////////////////////////
    // change QP state Init -> RTR
    /////////////////////////////////////////////////////////////////////////////////*/
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);

    qp_attr.qp_state = VAPI_RTR;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_QP_STATE);

    qp_attr.av.sl = DEF_SL;
    qp_attr.av.grh_flag = DEF_GRH_FLAG;
    qp_attr.av.dlid = remote_data->lid;
    qp_attr.av.static_rate = DEF_STATIC_RATE;
    qp_attr.av.src_path_bits = DEF_SRC_PATH_BITS; 
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_AV);

    qp_attr.path_mtu = MTU2048;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PATH_MTU);

    qp_attr.rq_psn = DEF_PSN; 
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_RQ_PSN);

    qp_attr.dest_qp_num = remote_data->qp_num;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_DEST_QP_NUM);

    qp_attr.qp_ous_rd_atom  = DEF_QP_OUS_RD_ATOM;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_QP_OUS_RD_ATOM);

    qp_attr.min_rnr_timer = IB_RNR_NAK_TIMER_655_36; 
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_MIN_RNR_TIMER);

    /* set the alternate path */
    qp_attr.alt_av.sl = DEF_SL;
    qp_attr.alt_av.grh_flag = DEF_GRH_FLAG;
    qp_attr.alt_av.dlid = remote_data->alt_lid;
    qp_attr.alt_av.static_rate = DEF_STATIC_RATE;
    qp_attr.alt_av.src_path_bits = DEF_SRC_PATH_BITS; 
    qp_attr.alt_pkey_ix = DEF_PKEY_IX;
    qp_attr.alt_port = config.alt_ib_port;
    qp_attr.alt_timeout = DEF_QP_TIMEOUT;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_ALT_PATH);

    rc = VAPI_modify_qp(resources->hca_hndl, resources->qp_hndl, &qp_attr, &qp_attr_mask, &qp_cap);
    CHECK_VERB_RETURN_VALUE("VAPI_modify_qp", rc);

    printf("-> RTR ");



    /*/////////////////////////////////////////////////////////////////////////////////
    // change QP state RTR -> RTS
    /////////////////////////////////////////////////////////////////////////////////*/

    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);

    qp_attr.qp_state = VAPI_RTS;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_QP_STATE);

    qp_attr.timeout = DEF_QP_TIMEOUT;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_TIMEOUT);

    qp_attr.retry_count = DEF_RETRY_COUNT;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_RETRY_COUNT);

    qp_attr.rnr_retry = DEF_RNR_RETRY;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_RNR_RETRY);

    qp_attr.min_rnr_timer = DEF_MIN_RNR_TIMER;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_MIN_RNR_TIMER);

    qp_attr.ous_dst_rd_atom = DEF_QP_OUS_RD_ATOM; 
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_OUS_DST_RD_ATOM);

    /* change the APM machine state */
    qp_attr.path_mig_state = VAPI_REARM;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_PATH_MIG_STATE);

    qp_attr.sq_psn = DEF_PSN;
    QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_SQ_PSN);

    rc = VAPI_modify_qp(resources->hca_hndl, resources->qp_hndl, &qp_attr, &qp_attr_mask, &qp_cap);
    CHECK_VERB_RETURN_VALUE("VAPI_modify_qp", rc);

    printf("-> RTS\n");

    fflush(stdout);

    return VAPI_OK;
}

/*****************************************************************
 the actual test: the daemon post RDMA write operations to the client
*****************************************************************/
VAPI_ret_t do_test(resources_t *resources, struct connection_data_t *remote_data)
{
    VAPI_ret_t rc;

    /* create and send the WQEs (only by the daemon) */
    if (config.is_daemon)
    {
        VAPI_qp_attr_t query_qp_attr;
        VAPI_qp_attr_mask_t query_qp_attr_mask;
        VAPI_qp_init_attr_t query_qp_init_attr;
        VAPI_sg_lst_entry_t sg_entry;
        VAPI_sr_desc_t sr_desc;
        VAPI_wc_desc_t comp_desc;
        VAPI_wr_id_t id = 0;


        printf(" [*] Start sending data to client\n");
        
        /* prepere one wqe for the SEND operation */
        sg_entry.addr = resources->rep_mrw.start;
        sg_entry.len = resources->rep_mrw.size;
        sg_entry.lkey = resources->rep_mrw.l_key;

        sr_desc.opcode = VAPI_RDMA_WRITE;
        sr_desc.comp_type = VAPI_SIGNALED;
        sr_desc.sg_lst_len = 1;
        sr_desc.sg_lst_p = &sg_entry;
        sr_desc.fence = FALSE;
        sr_desc.set_se = FALSE;
        sr_desc.r_key = remote_data->rkey;
        sr_desc.remote_addr = remote_data->start;

        /* send data to the client untill there will be a path migration event */
        do
        {
            sr_desc.id = id;

            /* post the WR */
            rc = VAPI_post_sr(resources->hca_hndl, resources->qp_hndl, &sr_desc);
            CHECK_VERB_RETURN_VALUE("VAPI_post_sr", rc);

            /* get the completion */
            do
            {
                rc = VAPI_poll_cq(resources->hca_hndl, resources->cq_hndl, &comp_desc);

            } while (rc == VAPI_CQ_EMPTY);
            CHECK_VERB_RETURN_VALUE("VAPI_post_sr", rc);

            /* check that the completion is good */
            if (comp_desc.status != VAPI_SUCCESS)
            {
                printf("Completion was polled with bad status: %s, Syndrom = 0x%x\n", VAPI_wc_status_sym(comp_desc.status), comp_desc.vendor_err_syndrome);
                return VAPI_EGEN;
            }

            /* after the first post, both of the sides should load the alternate path, make sure
             that it is true. */
            if (id == 0)
            {

                /* check the status of the APM machine */
                rc = VAPI_query_qp (resources->hca_hndl, resources->qp_hndl, &query_qp_attr, &query_qp_attr_mask, &query_qp_init_attr);
                CHECK_VERB_RETURN_VALUE("VAPI_query_qp", rc);

                if (!(query_qp_attr_mask & QP_ATTR_ALT_PATH))
                {
                    printf("Error, the alternate path wasn't loaded\n");
                    return VAPI_EGEN;
                }

                if (!(query_qp_attr_mask & QP_ATTR_PATH_MIG_STATE))
                {
                    printf("Error, the migration state machine wasn't changed\n");
                    return VAPI_EGEN;
                }

                /* check that the mig state machine is in armed */
                if (query_qp_attr.path_mig_state != VAPI_ARMED)
                {
                    printf("Error, the migration state machine is not in ARMED, current machine state is %s\n", VAPI_mig_state_sym(query_qp_attr.path_mig_state));
                    return VAPI_EGEN;
                }

                printf("Alterante path was loaded, please remove the cable of port 1 to start APM\n");
            }

            id ++;

        } while (!apm_occured);
    } else
    {
        printf(" [*] Start receiving data from daemon\n");
    }

    return VAPI_OK;
}

/*****************************************************************
 destroy all the resources the test uses
*****************************************************************/
VAPI_ret_t resources_destroy(resources_t *resources)
{
    VAPI_ret_t rc;

    if (resources->mr_hndl != VAPI_INVAL_HNDL)
    {
        rc = VAPI_deregister_mr (resources->hca_hndl, resources->mr_hndl);
        CHECK_VERB_RETURN_VALUE("VAPI_deregister_mr", rc);
        printf(" [*] Deregistered MR\n");
    }

    if (resources->buf_ptr != NULL)
    {
        FREE(NULL);
    }

    if (resources->qp_hndl != VAPI_INVAL_HNDL)
    {
        rc = VAPI_destroy_qp (resources->hca_hndl, resources->qp_hndl);
        CHECK_VERB_RETURN_VALUE("VAPI_destroy_qp", rc);
        printf(" [*] QP was destroyed\n");
    }

    if (resources->cq_hndl != VAPI_INVAL_HNDL)
    {
        rc = VAPI_destroy_cq (resources->hca_hndl, resources->cq_hndl);
        CHECK_VERB_RETURN_VALUE("VAPI_destroy_cq", rc);
        printf(" [*] CQ was destroyed\n");
    }
    
    if (resources->pd_hndl != VAPI_INVAL_PD_HNDL)
    {
        rc = VAPI_dealloc_pd (resources->hca_hndl, resources->pd_hndl);
        CHECK_VERB_RETURN_VALUE("VAPI_dealloc_pd", rc);
        printf(" [*] PD was destroyed\n");
    }


    if (resources->async_handler_hndl != VAPI_INVAL_HNDL)
    {
        rc = EVAPI_clear_async_event_handler(resources->hca_hndl, resources->async_handler_hndl);
        CHECK_VERB_RETURN_VALUE("EVAPI_clear_async_event_handler", rc);
        printf(" [*] Asynchronous event handler was cleared\n");
    }

    if (resources->hca_hndl != VAPI_INVAL_HNDL)
    {
        rc = EVAPI_release_hca_hndl (resources->hca_hndl);
        CHECK_VERB_RETURN_VALUE("EVAPI_release_hca_hndl", rc);
        printf(" [*] HCA handle was released\n");
    }

    if (resources->sockfd >= 0)
    {
        close(resources->sockfd);
    }

    return VAPI_OK;
}

