/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/


#include <stdio.h>
#include <stdlib.h>
#include <mtl_common.h>
#include <vapi.h>
#include <evapi.h>
#include <vapi_types.h>
#ifndef __WIN__
#include <signal.h>
#endif



/*============== Macro variables ============================*/
#define MAX_HCAS 10

/*============== Global variables ============================*/

VAPI_hca_id_t *hca_id_buf;
VAPI_hca_hndl_t hca_hndl_arr[MAX_HCAS];
u_int32_t num_of_hcas;
EVAPI_async_handler_hndl_t  async_handler_hndl[MAX_HCAS];
/*============== static functions prototypes ==================*/
static VAPI_ret_t get_list_of_hcas(VAPI_hca_id_t **hca_id_buf_p, u_int32_t *num_of_hcas);
#ifdef __WIN__
BOOL WINAPI sig_handler(  DWORD dwCtrlType );
#else
void sig_handler(int signum);
#endif
void reg_signals(void);
void async_event_handler(VAPI_hca_hndl_t hca_hndl,
                         VAPI_event_record_t *event_record_p, 
                         void* private_data);


/*=============================================================*/
/*
 *  main
 */
int main(int argc, char *argv[])
{
  VAPI_ret_t rc;
  unsigned long i;
  VAPI_hca_hndl_t hca_hndl;

  rc = get_list_of_hcas(&hca_id_buf, &num_of_hcas);
  if ( rc != VAPI_OK ) {
    printf("Error: failed to retrieve the list of HCAs\n");
    return -1;
  }
  printf("%d HCA%s found:\n", num_of_hcas, num_of_hcas>1?"s":"");
  for ( i=0; i<num_of_hcas; ++i ) {
    printf("\thca_id=%s\n", hca_id_buf[i]);
    /* try to get a handle to the given hca_id */
    rc = EVAPI_get_hca_hndl(hca_id_buf[i], &hca_hndl);
    if ( rc != VAPI_OK ) {
      printf("Error: Could not retrieve hanlde to the HCA\n");
      continue;
    }
    hca_hndl_arr[i] = hca_hndl;
    /* register asynch events hanler to all HCAs*/
    rc = EVAPI_set_async_event_handler(hca_hndl_arr[i], async_event_handler, (void*)(MT_ulong_ptr_t)i,
                                       &async_handler_hndl[i]);
    if ( rc != VAPI_OK ) {
      printf("Error: EVAPI_set_async_event_handler failed ret=%d.\n", rc);
      exit(1);
    }

  }
    
  if (i > 0) {
    reg_signals();  /* register cleanup handler */
    while (1) {
      /*loop forever waiting for events to arrive */
    }
  }
  
  return 0;
}

/*============== static functions implemenmtation ==================*/

#ifdef __WIN__
BOOL WINAPI sig_handler(  DWORD dwCtrlType )
{
	int signum = (int)dwCtrlType;
#else
void sig_handler(int signum)
{
#endif
  int i;
  VAPI_ret_t rc;
  static int got_signal = 0;
  /* prevent handling any signal more than once, so that the cleanup code is executed 
   * only once.  Note that since there is the same behavior for ANY signal handled by
   * this handler (i.e., the process cleans itself up and exits), this code should be 
   * executed only once, even if it is invoked for a different signal number.
   */
  if (!got_signal) {
    got_signal++;
    printf("Got signal %d - exiting.\n",signum);

    for (i= 0; i < (int)num_of_hcas; i++) {
      if (hca_hndl_arr[i] != VAPI_INVAL_HNDL) {
        rc = EVAPI_clear_async_event_handler(hca_hndl_arr[i], async_handler_hndl[i]);
        if ( rc != VAPI_OK ) {
          printf("Error: EVAPI_clear_async_event_handler failed for %d HCA. ret=%d.\n", i, rc);
        }
        rc = EVAPI_release_hca_hndl(hca_hndl_arr[i]);
        if ( rc != VAPI_OK ) {
          printf("Error: EVAPI_release_hca_hndl failed for %d HCA. ret=%d.\n", i, rc);
        }
      }
    }
    if (hca_id_buf) {
      free(hca_id_buf);
    }
  } else {
    MTL_TRACE1("%s:got another signal (%d). signal count = %d. IGNORING\n", __func__,signum, got_signal);
    got_signal++;
  }
  exit(0);
}

/*
 *  reg_signals
 */

void reg_signals(void)
{
#ifdef __WIN__
    SetConsoleCtrlHandler(sig_handler, TRUE);
#else    
  struct sigaction sigact;
  sigact.sa_handler = sig_handler;
  sigemptyset(&sigact.sa_mask);
  sigact.sa_flags = 0;
  sigaction(SIGINT, &sigact, NULL);
#endif  
  return;
}




/*
 *  get_list_of_hcas
 */
static VAPI_ret_t get_list_of_hcas(VAPI_hca_id_t **hca_id_buf_p, u_int32_t *num_of_hcas)
{
  VAPI_hca_id_t *hca_id_buf;
  VAPI_ret_t rc;
  u_int32_t local_num_of_hcas;

  *hca_id_buf_p = NULL;
  rc = EVAPI_list_hcas(0, &local_num_of_hcas, NULL);
  switch ( rc ) {
    case VAPI_OK:
      *num_of_hcas = local_num_of_hcas;
      return VAPI_OK;

    case VAPI_EAGAIN:
      hca_id_buf = malloc(sizeof(VAPI_hca_id_t)*local_num_of_hcas);
      if ( !hca_id_buf ) {
        MT_COV_MEM_FLTR_START
        return VAPI_ERR;
        MT_COV_MEM_FLTR_END
      }
      rc = EVAPI_list_hcas(local_num_of_hcas, &local_num_of_hcas, hca_id_buf);
      if ( rc != VAPI_OK ) {
        free(hca_id_buf);
        return VAPI_ERR;
      }
      *num_of_hcas = local_num_of_hcas;
      *hca_id_buf_p = hca_id_buf;
      return VAPI_OK;

    default:
      return VAPI_ERR;
  }
}



void async_event_handler(VAPI_hca_hndl_t hca_hndl,
                         VAPI_event_record_t *event_record_p, 
                         void* private_data)
{
  unsigned long i = (unsigned long) (MT_ulong_ptr_t)private_data;
  if (i>MAX_HCAS) {
    MTL_ERROR1("%s:got non-valid index=%ld type=%d. IGNORING\n", __func__,i, event_record_p->type);
    return;
  }

  switch (event_record_p->type) {
    case VAPI_PORT_ERROR:
      printf("Got PORT_ERROR event. HCA=%s, port number=%d\n", hca_id_buf[i],
             event_record_p->modifier.port_num); 
      break;
    case VAPI_PORT_ACTIVE:
      printf("Got PORT_ACTIVE event. HCA=%s, port number=%d\n", hca_id_buf[i],
             event_record_p->modifier.port_num); 
      break;
  case  VAPI_PATH_MIG_ARMED:    /*QP*/
      printf("Got VAPI_PATH_MIG_ARMED event. HCA=%s, qp_hndl=" MT_ULONG_PTR_FMT "\n", hca_id_buf[i],
             event_record_p->modifier.qp_hndl); 
  case VAPI_QP_PATH_MIGRATED:    /*QP*/
      printf("Got VAPI_QP_PATH_MIGRATED event. HCA=%s, qp_hndl=" MT_ULONG_PTR_FMT "\n", hca_id_buf[i],
             event_record_p->modifier.qp_hndl); 

      break;
  case VAPI_EEC_PATH_MIGRATED:   /*EEC*/
      printf("Got EEC_PATH_MIGRATED event. HCA=%s, eec_hndl=%d\n", hca_id_buf[i],
             event_record_p->modifier.eec_hndl); 

      break;
  case VAPI_QP_COMM_ESTABLISHED: /*QP*/
      printf("Got QP_COMM_ESTABLISHED event. HCA=%s, qp_hndl=" MT_ULONG_PTR_FMT "\n", hca_id_buf[i],
             event_record_p->modifier.qp_hndl); 

      break;
    case VAPI_EEC_COMM_ESTABLISHED: /*EEC*/ 
      printf("Got EEC_COMM_ESTABLISHED event. HCA=%s, eec_hndl=%d\n", hca_id_buf[i],
             event_record_p->modifier.eec_hndl); 

      break;
  case VAPI_SEND_QUEUE_DRAINED:  /*QP*/
      printf("Got SEND_QUEUE_DRAINED event. HCA=%s, qp_hndl=" MT_ULONG_PTR_FMT "\n", hca_id_buf[i],
             event_record_p->modifier.qp_hndl); 

      break;
  case VAPI_CQ_ERROR:            /*CQ*/
      printf("Got CQ_ERROR event. HCA=%s, cq_hndl=" MT_ULONG_PTR_FMT "\n", hca_id_buf[i],
             event_record_p->modifier.cq_hndl); 

      break;
  case VAPI_LOCAL_WQ_INV_REQUEST_ERROR: /*QP*/
      printf("Got LOCAL_WQ_INV_REQUEST_ERROR event. HCA=%s, qp_hndl=" MT_ULONG_PTR_FMT "\n", hca_id_buf[i],
             event_record_p->modifier.qp_hndl); 

      break;
  case VAPI_LOCAL_WQ_ACCESS_VIOL_ERROR: /*QP*/
      printf("Got LOCAL_WQ_ACCESS_VIOL_ERROR event. HCA=%s, qp_hndl=" MT_ULONG_PTR_FMT "\n", hca_id_buf[i],
             event_record_p->modifier.qp_hndl); 

      break;
  case VAPI_LOCAL_WQ_CATASTROPHIC_ERROR: /*QP*/
      printf("Got LOCAL_WQ_CATASTROPHIC_ERROR event. HCA=%s, qp_hndl=" MT_ULONG_PTR_FMT "\n", hca_id_buf[i],
             event_record_p->modifier.qp_hndl); 

      break;
  case VAPI_PATH_MIG_REQ_ERROR:  /*QP*/
      printf("Got PATH_MIG_REQ_ERROR event. HCA=%s, qp_hndl=" MT_ULONG_PTR_FMT "\n", hca_id_buf[i],
             event_record_p->modifier.qp_hndl); 

      break;
  case VAPI_LOCAL_CATASTROPHIC_ERROR: /*none*/
      printf("Got LOCAL_CATASTROPHIC_ERROR event. HCA=%s", hca_id_buf[i]);
	  if (event_record_p->modifier.port_num == 1) printf(" from MSIX\n");
      else printf("\n");
      break;  

  case VAPI_GPIO: /*none*/
      printf("Got GPIO event for HCA=%s; gpio_low=0x%X gpio_high=0x%X", 
			 hca_id_buf[i], event_record_p->syndrome, event_record_p->modifier.eec_hndl);
      break; 

  case VAPI_ECC_DETECT: /*none*/
      printf("Got ECC_DETECT event for HCA=%s; err_word_4=0x%X err_word_5=0x%X;\n", 
			 hca_id_buf[i], event_record_p->syndrome, event_record_p->modifier.eec_hndl);
      break; 


  default:
      MTL_ERROR1("%s:got non-valid event type=%d. IGNORING\n", __func__,event_record_p->type);
  }
}

