/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/


#include <string.h>
#include <stdio.h>

#include <mosal.h>
#include <vapi.h>
#include <evapi.h>
#include <ib_mgt.h>

#include "mad_pcounters.h"

/* default hca device name */
#define HCA_ID  "InfiniHost0"

#ifndef _BIG_ENDIAN
/* we are on little-endian machine */

static void
cpu_to_be32_buf(wbuf, wsize)
     unsigned       *wbuf;
     unsigned        wsize;
{
    int             i;
    for (i = 0; (unsigned)i < wsize; i++) {
        unsigned        __x = wbuf[i];

        wbuf[i] = (((__x & (unsigned) 0x000000ffUL) << 24)
                   | ((__x & (unsigned) 0x0000ff00UL) << 8)
                   | ((__x & (unsigned) 0x00ff0000UL) >> 8)
                   | ((__x & (unsigned) 0xff000000UL) >> 24));
    }
}

#else 
/* we are on big-endian machine */
#define cpu_to_be32_buf(wbuf, wsize)
#endif

#define be32_to_cpu_buf cpu_to_be32_buf

struct mad_pcounters req;

/* HCA Identificator */
char *hca_id = "InfiniHost0";

/* HCA port */
int port = 1;

static void
get_pcounters(res, port)
     struct mad_pcounters *res;
     int        port;
{
    IB_MGT_ret_t    rc;
    static u_int32_t tid = 1;

    req.Method = 1;             /* 1 - Get, 2 - Set, 0x81 - GetResp */
    req.R = 0;                  /* Response bit. */
    req.ClassVersion = 1;       /* 1 */
    req.MgmtClass = 0x4;        /* 0x4 - Performance Management Class */
    req.BaseVersion = 1;        /* Version of MAD base format. This shall be 1 */
    req.TransactionID__63_32 = 0;
    req.TransactionID__31_0 = tid++;
    req.AttributeID = 0x12;     /* 0x12 - PortCounters */
    req.PortSelect = port;      /* Port Number */

    // Convert packed to BE 
    cpu_to_be32_buf((unsigned *) &req, 64);

    // Process MAD
    rc = IB_MGT_local_mad(hca_id, (IB_port_t)port, 0, IB_MGT_MAD_IGNORE_MKEY,(void *) &req, (void *) res);
    if (rc != IB_MGT_OK) {
        printf("Error, could not execute: IB_MGT_local_mad\n");
        exit(1);
    }
    // Convert result to CPU
    be32_to_cpu_buf((unsigned *) res, 64);
}

static void
usage ()
{
    fprintf(stderr, "usage: get_pcounter [-h HCA_DEVICEDLID] -p PORT\n");
    fprintf(stderr, "   -h HCA_DEVICE       HCA-id (default: InfiniHost0)\n");
    fprintf(stderr, "   -p PORT             HCA logical port (default: 1)\n");
    exit(2);
}

int
MT_API main(argc, argv)
     int             argc;
     char          **argv;
{
    struct mad_pcounters pc;
    int             i;

    if (argc < 2) {
        usage();
    }

    for (i = 1; i < argc; i++) {
        if (!strcmp("-h", argv[i])) {
            i++;
            if (i < argc)
                hca_id = argv[i];
        } else if (!strcmp("-p", argv[i])) {
            i++;
            if (i < argc)
                port = strtoul(argv[i], 0, 0);
        } else {
            printf("%s: unknown command line option \"%s\"\n", argv[0],
                   argv[i]);
            usage();
        }

    }
    printf("HCA_DEVICE=%s\n", hca_id);
    printf("PORT=%d\n", port);
    printf("\n");

    
    get_pcounters(&pc, port);      // port 1

    // -------------------------------------------------------
    // parse the packet


    printf("Status\n");
    printf("------\n");
    printf("Request Status:                0x%x\n", pc.Status);
    printf("PortSelect:                    0x%x\n", pc.PortSelect);
    printf("CounterSelect:                 0x%x\n", pc.CounterSelect);
    printf("\n");
    printf("Error counters\n");
    printf("--------------\n");
    printf("PortRcvErrors:                 %10u\n", pc.PortRcvErrors);
    printf("PortRcvRemotePhysicalErrors:   %10u\n", pc.PortRcvRemotePhysicalErrors);
    printf("PortRcvSwitchRelayErrors:      %10u\n", pc.PortRcvSwitchRelayErrors);
    printf("PortXmitDiscards:              %10u\n", pc.PortXmitDiscards);
    printf("PortXmitConstraintErrors:      %10u\n", pc.PortXmitConstraintErrors);
    printf("PortRcvConstraintErrors:       %10u\n", pc.PortRcvConstraintErrors);
    printf("LocalLinkIntegrityErrors:      %10u\n", pc.LocalLinkIntegrityErrors);
    printf("ExcessiveBufferOverrunErrors:  %10u\n", pc.ExcessiveBufferOverrunErrors);
    printf("VL15Dropped:                   %10u\n", pc.VL15Dropped);
    printf("SymbolErrorCounter:            %10u\n", pc.SymbolErrorCounter);
    printf("LinkErrorRecoveryCounter:      %10u\n", pc.LinkErrorRecoveryCounter);
    printf("LinkDownedCounter:             %10u\n", pc.LinkDownedCounter);
    printf("\n");
    printf("Packet counters\n");
    printf("--------------\n");
    printf("PortXmitData:                  %10u\n", pc.PortXmitData);
    printf("PortRcvData:                   %10u\n", pc.PortRcvData);
    printf("PortXmitPkts:                  %10u\n", pc.PortXmitPkts);
    printf("PortRcvPkts:                   %10u\n", pc.PortRcvPkts);
    printf("\n");


    return 0;
}
