#!/bin/bash 
# limit_max_rd: Limit max outstanding reads for specific PCI chipsets
# usage: ./limit_max_rd [-v]
#        -v : Verbose output 

# global definitions ("macros")
PROBLEMATIC_VENDORS="1166"
PROBLEMATIC_CHIPSET_DEVS="0008 0009"
TAVOR_HCA_VENDOR_DEVID="15b3:5a44"

# Byte offset of lower byte of PCI-X's command register (hex)
let TAVOR_MAX_OUTS_RD_CAP_OFFSET="72"
# mask for "zeroing" max. outstanding split tranaction bits in PCI-X command register
let OUTS_READ_MASK=0x8F

# Set debug environment if requested (-d option)
no_echo()
{
  # do nothing
  echo > /dev/null
}

DEBUG_ECHO=no_echo
if [ "$1" == "-v" ]; then
  DEBUG_ECHO=echo
fi


if [ $UID != 0 ]; then
  echo You must be root to access PCI configuration space
  exit 1
fi


if ( ! [ -x /sbin/setpci ] ); then
  echo "setpci utility (from PCI tools package) is required to run this script"
  exit 1
fi

# Find Host bridge (not all systems locate it at 00:00.0) and get its props.
host_bridge=`/sbin/lspci | awk '(FNR == 1) {print $1}'`
chipset_vendor=`/sbin/setpci -s $host_bridge 0.W`
if [ $? != 0 ]; then
  echo Failed invoking "setpci"
  exit 2
fi
chipset_devid=`/sbin/setpci -s $host_bridge 2.W`
if [ $? != 0 ]; then
  echo Failed invoking "setpci"
  exit 2
fi

$DEBUG_ECHO Chipset info: vendor=0x$chipset_vendor device=0x$chipset_devid

if [ `expr match "$PROBLEMATIC_VENDORS" "[0-9a-f ]\?\+$chipset_vendor"` != 0 ]; then  
  # vendor of chipset is in "black list"
  if [ `expr match  "$PROBLEMATIC_CHIPSET_DEVS" "[0-9a-f ]\?\+$chipset_devid"` != 0 ]; then 
    # chipset device ID is in "black list"
    $DEBUG_ECHO "Found a suspected chipset from vendor $chipset_vendor (dev. $chipset_devid)" 
    $DEBUG_ECHO "Setting maximum outstanding reads to 1 for all of installed InfiniHost devices:"
    a=(`/sbin/setpci -v -d $TAVOR_HCA_VENDOR_DEVID $TAVOR_MAX_OUTS_RD_CAP_OFFSET.B`)
    if [ $? != 0 ]; then
      echo Failed \"/sbin/setpci -v -d $TAVOR_HCA_VENDOR_DEVID $TAVOR_MAX_OUTS_RD_CAP_OFFSET.B\"
      exit 3
    fi
    let num_of_tavors=${#a[*]}/3
    $DEBUG_ECHO "Found $num_of_tavors InfiniHost devices:"
    for (( i=0 ; $i < $num_of_tavors; )); do
      let location_offset=$i\*3
      let value_offset=$location_offset+2
      location=`expr substr ${a[$location_offset]} 1 7`
      let original_cap_val=0x${a[$value_offset]}
      #$DEBUG_ECHO "Tavor on $location with value $original_cap_val at offset $TAVOR_MAX_OUTS_RD_CAP_OFFSET"
      # clear max. split trans. (limit to 1)
      let new_cap_val=$original_cap_val\&$OUTS_READ_MASK
      $DEBUG_ECHO `printf "(%s) == 0x%X --> 0x%X\n" ${a[$location_offset]} $original_cap_val $new_cap_val`
      new_cap_val_hex=`printf "%x" $new_cap_val`
      /sbin/setpci -s $location $TAVOR_MAX_OUTS_RD_CAP_OFFSET.B=$new_cap_val_hex
      if [ $? != 0 ]; then
        echo Failed \"/sbin/setpci -s $location $TAVOR_MAX_OUTS_RD_CAP_OFFSET.B=$new_cap_val_hex\"
        exit 4
      fi
      let i++ 
    done
  fi
fi

exit 0
