/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <vapi.h>
#include <evapi.h>
#include <vapi_common.h>
#include <stdio.h>
#include <stdlib.h>
#if      defined(linux) || defined(__LINUX__)
#include <getopt.h>
#include <readline/readline.h>
#endif

/*============== macro definitions ============================*/
#define GID_FMT "%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u.%02u"
#define GID_FMT_DATA(gid) \
      gid[0],gid[1],gid[2],gid[3],gid[4],gid[5],gid[6],gid[7],      \
      gid[8],gid[9],gid[10],gid[11],gid[12],gid[13],gid[14],gid[15]
      
/*============== static functions prototypes ==================*/
static VAPI_ret_t get_list_of_hcas(VAPI_hca_id_t **hca_id_buf_p, u_int32_t *num_of_hcas);
const char *port_state_str(IB_port_state_t pstate);
const char *mtu_str(IB_mtu_t max_mtu);
static MT_bool test_hca_ok(VAPI_hca_id_t hca_id);
static char *guid_str(IB_guid_t node_guid);
static IB_gid_t *alloc_get_all_port_gids(VAPI_hca_hndl_t hca_hndl, IB_port_t port, u_int16_t *num_p);
static void free_port_gids(IB_gid_t *gid_tbl);
static int print_hca_cap(VAPI_hca_id_t hca_id);


static inline int null_gid(IB_gid_t gid)
{
  return !(gid[8] | gid[9] | gid[10] | gid[11] | 
         gid[12] | gid[13] | gid[14] | gid[15]); 
}


static MT_bool verbose = FALSE;
static MT_bool one_device = FALSE;
static VAPI_hca_id_t device_id;

#define USAGE_STR "usage: vstat [--verbose | -v ] [--list] [--dev_name DEVICE_NAME]"


/*
 *  main
 */
int MT_API main(int argc, char *argv[])
{
  VAPI_ret_t rc;
  int ret = 0;
  VAPI_hca_id_t *hca_id_buf;
  u_int32_t num_of_hcas, i;
  
  /* parse command line options */
  #if  defined(__WIN__) 

  		/* "list" */
        if ( argc > 1 && !strcmp(argv[1], "list") ) {
          rc = get_list_of_hcas(&hca_id_buf, &num_of_hcas);
          if ( rc != VAPI_OK ) {
            printf("failed to retrieve the list of HCAs\n");
            return -1;
          }
          printf("%d HCA%s found:\n", num_of_hcas, num_of_hcas>1?"s":"");
          if (num_of_hcas > 0) 
              for ( i=0; i<num_of_hcas; ++i ) {
                  printf("\t%s : %s\n", hca_id_buf[i], 
                         test_hca_ok(hca_id_buf[i]) ? "[ OPENED ]" : "[ CLOSED ]");
              }
          printf("\n");
          if ( hca_id_buf ) free(hca_id_buf);
          return 0;
        }
        if ( argc > 1 && (!strcmp(argv[1], "-v") || !strcmp(argv[1], "--verbose"))) {
          verbose = TRUE;
        }
  #else
  int ch, option_index = 0;
  while ( 1 ) {
    static struct option long_options[] = {
      {"list", 0, 0, 0},
      {"verbose", 0, 0, 0},
      {"dev_name", required_argument, 0, 0},
      {0, 0, 0, 0}
    };

    ch = getopt_long(argc, argv, "v", long_options, &option_index);
    if ( ch == -1 ) {
      break;
    }
    switch ( ch ) {
      case 0:
        if ( strcmp(long_options[option_index].name, "list")==0 ) {
          rc = get_list_of_hcas(&hca_id_buf, &num_of_hcas);
          if ( rc != VAPI_OK ) {
            printf("failed to retrieve the list of HCAs\n");
            return -1;
          }
          printf("%d HCA%s found:\n", num_of_hcas, num_of_hcas>1?"s":"");
          if (num_of_hcas > 0)
              for ( i=0; i<num_of_hcas; ++i ) {
                  printf("\t%s :    %s\n", hca_id_buf[i], 
                         test_hca_ok(hca_id_buf[i]) ? "[ OPENED ]" : "[ CLOSED ]");
              }
          printf("\n");
          if ( hca_id_buf ) free(hca_id_buf);
        }
        else if ( strcmp(long_options[option_index].name, "verbose") == 0 ) {
          verbose = TRUE;
          break;
        } 
        else if ( strcmp(long_options[option_index].name, "dev_name") == 0 ) {
          one_device = TRUE;
          strncpy(device_id, optarg, HCA_MAXNAME);
          break;
        }

        return 0;
      case 'v':
        verbose = TRUE;
        break;
      default:
        printf(USAGE_STR"\n");
        return -1;

    }
  }
#endif

  /* if only one device query is needed */
  if (one_device)
  {
      ret |= print_hca_cap(device_id);
  } else
  {
      rc = get_list_of_hcas(&hca_id_buf, &num_of_hcas);
      if ( rc != VAPI_OK ) {
          printf("Error: failed to retrieve the list of HCAs\n");
          return -1;
      }

      printf("%d HCA%s found:\n", num_of_hcas, num_of_hcas>1?"s":"");
      for ( i=0; i<num_of_hcas; ++i ) {
          ret |= print_hca_cap(hca_id_buf[i]);
      }
      if ( hca_id_buf )free(hca_id_buf);
  }

  return ret;
}

/*
 *  test_hca_ok
 */
static MT_bool test_hca_ok(VAPI_hca_id_t hca_id){
    VAPI_ret_t rc;
    VAPI_hca_hndl_t hca_hndl;

    /* try to get a handle to the given hca_id */
    if (hca_id == NULL) {
        return FALSE;
    }
    rc = EVAPI_get_hca_hndl(hca_id, &hca_hndl);
    if ( rc != VAPI_OK ) {
        return FALSE;
    }
    EVAPI_release_hca_hndl(hca_hndl);
    return TRUE;
}

/*
 *  get_list_of_hcas
 */
static VAPI_ret_t get_list_of_hcas(VAPI_hca_id_t **hca_id_buf_p, u_int32_t *num_of_hcas)
{
  VAPI_hca_id_t *hca_id_buf;
  VAPI_ret_t rc;
  u_int32_t local_num_of_hcas;

  *hca_id_buf_p = NULL;
  rc = EVAPI_list_hcas(0, &local_num_of_hcas, NULL);
  switch ( rc ) {
    case VAPI_OK:
      *num_of_hcas = local_num_of_hcas;
      return VAPI_OK;

    case VAPI_EAGAIN:
      hca_id_buf = malloc(sizeof(VAPI_hca_id_t)*local_num_of_hcas);
      if ( !hca_id_buf ) {
        return VAPI_ERR;
      }
      rc = EVAPI_list_hcas(local_num_of_hcas, &local_num_of_hcas, hca_id_buf);
      if ( rc != VAPI_OK ) {
        free(hca_id_buf);
        return VAPI_ERR;
      }
      *num_of_hcas = local_num_of_hcas;
      *hca_id_buf_p = hca_id_buf;
      return VAPI_OK;

    default:
      return VAPI_ERR;
  }
}

/*
 *  port_state_str
 */
const char *port_state_str(IB_port_state_t pstate)
{
  switch ( pstate ) {
    case PORT_DOWN:
      return "PORT_DOWN";
    case PORT_INITIALIZE:
      return "PORT_INITIALIZE";
    case PORT_ARMED:
      return "PORT_ARMED";
    case PORT_ACTIVE:
      return "PORT_ACTIVE";
    default:
      return "Illegal state";
  }
}

/*
 *  mtu_str
 */
const char *mtu_str(IB_mtu_t max_mtu)
{
  switch ( max_mtu ) {
    case MTU256:
      return "256";
    case MTU512:
      return "512";
    case MTU1024:
      return "1024";
    case MTU2048:
      return "2048";
    case MTU4096:
      return "4096";
    default:
      return "invalid MTU";
  }
}



/*
 *  guid_str
 */
static char *guid_str(IB_guid_t node_guid)
{
  int i;
  char tmp[4], *str=malloc(2*8+7+2);
  if ( !str ) {
    return 0;
  }
  for ( str[0]=0, i=0; i<8; ++i) {
    sprintf(tmp, "%02x:", node_guid[i]);
    strcat(str, tmp);
  }
  str[strlen(str)-1]=0;
  return str;
}


/*
 *  alloc_get_all_port_gids
 */
static IB_gid_t *alloc_get_all_port_gids(VAPI_hca_hndl_t hca_hndl, IB_port_t port, u_int16_t *num_p)
{
  VAPI_ret_t rc;
  u_int16_t out_size, in_size;
  IB_gid_t *gid_tbl_p;
  int i, next_free;

  rc = VAPI_query_hca_gid_tbl(hca_hndl, port, 0, &out_size, 0);
  if ( rc == VAPI_OK ) {
    /* no entries in the gid table */
    return NULL;
  }
  else if ( rc == VAPI_EAGAIN ) {
    gid_tbl_p = malloc(out_size * sizeof(IB_gid_t));
    if ( !gid_tbl_p ) {
      printf("malloc failed allocating for gid table\n");
      return NULL;
    }
    in_size = out_size;
    rc = VAPI_query_hca_gid_tbl(hca_hndl, port, in_size, &out_size, gid_tbl_p);
    if ( rc != VAPI_OK ) {
      printf("VAPI_query_hca_gid_tbl failed - %s\n", VAPI_strerror_sym(rc));
      return NULL;
    }
    for ( i=0, next_free=0; i<out_size; ++i ) {
      if ( !null_gid(gid_tbl_p[i] ) ) {
        memcpy(gid_tbl_p+next_free, gid_tbl_p+i, sizeof(IB_gid_t));
        next_free++;
      }
    }
    *num_p = (u_int16_t)next_free;
    return gid_tbl_p;
  }
  else {
    printf("error retrieving gid table for port %d - %s\n", port, VAPI_strerror_sym(rc));
    return NULL;
  }
}

/*
 *  free_port_gids
 */
static void free_port_gids(IB_gid_t *gid_tbl)
{
  free(gid_tbl);
}

static char *fw_ver_str(u_int64_t fw_ver, char *str)
{
	u_int32_t major, minor, sub_minor;
	char tmp[15];

	major = (fw_ver >> 32) & 0xffff;
	minor = (fw_ver >> 16) & 0xffff;
	sub_minor = fw_ver & 0xffff;

	str[0] = 0;
	sprintf(tmp, "%d.", major);
	strcat(str, tmp);
	sprintf(tmp, "%d.", minor);
	strcat(str, tmp);
	sprintf(tmp, "%d", sub_minor);
	strcat(str, tmp);

	return str;
}

/*
 *  print_hca_cap
 */
static int print_hca_cap(VAPI_hca_id_t hca_id)
{
    VAPI_ret_t rc;
    VAPI_hca_vendor_t hca_vendor;
    VAPI_hca_cap_t hca_cap;
    VAPI_hca_hndl_t hca_hndl;
    VAPI_hca_port_t hca_port;
    IB_gid_t *gid_tbl;
    u_int16_t num_gids;
    char *str, buf[40];
    u_int32_t i, port;

    printf("\thca_id=%s\n", hca_id);

    /* try to get a handle to the given hca_id */
    rc = EVAPI_get_hca_hndl(hca_id, &hca_hndl);
    if ( rc != VAPI_OK ) {
      printf("Error: Could not retrieve handle to the HCA %s (%s)\n",
        hca_id,VAPI_strerror_sym(rc));
      return -1;
    }

    rc = VAPI_query_hca_cap(hca_hndl, &hca_vendor, &hca_cap);
    if ( rc != VAPI_OK) {
      EVAPI_release_hca_hndl(hca_hndl);
      printf("Error: Could not retrieve HCA capabilities\n");
      return -1;
    }
    printf("\tpci_location={BUS=0x%02X,DEV/FUNC=0x%02X}\n",hca_vendor.pci_bus, hca_vendor.pci_devfn);
    printf("\tvendor_id=0x%04X\n", hca_vendor.vendor_id);
    printf("\tvendor_part_id=0x%04X\n", hca_vendor.vendor_part_id);
    printf("\thw_ver=0x%X\n", hca_vendor.hw_ver);
    printf("\tfw_ver=%s\n", fw_ver_str(hca_vendor.fw_ver, buf));
    if ((unsigned char)(hca_vendor.psid[0]) == (unsigned char)0xff ) {
	printf("\tPSID not available -- FW not installed using fail-safe mode\n");
    } else {
	printf("\tPSID=%s\n", (char *) hca_vendor.psid);
    }
    printf("\tnum_phys_ports=%d\n", hca_cap.phys_port_num);
    if ( verbose ) {
      printf("\tmax_num_qp=0x%x\t(Maximum Number of QPs supported)\n", hca_cap.max_num_qp);
      printf("\tmax_qp_ous_wr=0x%x\t(Maximum Number of oustanding WR on any WQ)\n", hca_cap.max_qp_ous_wr);
      printf("\tflags=0x%08x\n", hca_cap.flags);
      printf("\tmax_num_sg_ent=0x%x\t(Max num of scatter/gather entries for WQE other than RD)\n", hca_cap.max_num_sg_ent);
      printf("\tmax_num_sg_ent_rd=0x%x\t(Max num of scatter/gather entries for RD WQE)\n", hca_cap.max_num_sg_ent_rd);
      printf("\tmax_num_srq=0x%x\t(Maximum Number of SRQs supported)\n", hca_cap.max_num_srq);
      if (hca_cap.max_num_srq > 0) {
        printf("\tmax_wqe_per_srq=0x%x\t(Maximum Number of oustanding WR on any SRQ)\n", hca_cap.max_wqe_per_srq);
        printf("\tmax_srq_sentries=0x%x\t(Maximum Number of scatter entries for SRQ WQE)\n", hca_cap.max_srq_sentries);
        printf("\tsrq_resize_supported=%s\t(SRQ resize supported)\n",(hca_cap.srq_resize_supported) ? "TRUE" : "FALSE");
      }
      printf("\tmax_num_cq=0x%x\t(Max num of supported CQs)\n", hca_cap.max_num_cq);
      printf("\tmax_num_ent_cq=0x%x\t(Max num of supported entries per CQ)\n", hca_cap.max_num_ent_cq);
      printf("\tmax_num_mr=0x%x\t(Maximum number of memory region supported)\n", hca_cap.max_num_mr);
      printf("\tmax_mr_size="U64_FMT"\t(Largest contigous block of memory region in bytes)\n", hca_cap.max_mr_size);
      printf("\tmax_pd_num=0x%x\t(Maximum number of protection domains supported)\n", hca_cap.max_pd_num);
      printf("\tpage_size_cap=0x%x\t(Largest page size supported by this HCA)\n", hca_cap.page_size_cap);
      printf("\tmax_pkeys=0x%x\t(Maximum number of partitions supported)\n", hca_cap.max_pkeys);
      printf("\tnode_guid=%s\t(Node GUID for this hca)\n", str=guid_str(hca_cap.node_guid));
      if ( str ) free(str);
      printf("\tlocal_ca_ack_delay=0x%x\t(Log2 4.096usec Max. RX to ACK or NAK delay)\n", hca_cap.local_ca_ack_delay);
      printf("\tmax_qp_ous_rd_atom=0x%x\t(Maximum number of oust. RDMA read/atomic as target)\n", hca_cap.max_qp_ous_rd_atom);
      printf("\tmax_ee_ous_rd_atom=0x%x\t(EE Maximum number of outs. RDMA read/atomic as target)\n", hca_cap.max_ee_ous_rd_atom);
      printf("\tmax_res_rd_atom=0x%x\t(Max. Num. of resources used for RDMA read/atomic as target)\n", hca_cap.max_res_rd_atom);
      printf("\tmax_qp_init_rd_atom=0x%x\t(Max. Num. of outs. RDMA read/atomic as initiator)\n", hca_cap.max_qp_init_rd_atom);
      printf("\tmax_ee_init_rd_atom=0x%x\t(EE Max. Num. of outs. RDMA read/atomic as initiator)\n", hca_cap.max_ee_init_rd_atom);
      printf("\tatomic_cap=%s\t(Level of Atomicity supported)\n", VAPI_atomic_cap_sym(hca_cap.atomic_cap));
      printf("\tmax_ee_num=0x%x\t(Maximum number of EEC supported)\n", hca_cap.max_ee_num);
      printf("\tmax_rdd_num=0x%x\t(Maximum number of IB_RDD supported)\n", hca_cap.max_rdd_num);
      printf("\tmax_mw_num=0x%x\t(Maximum Number of memory windows supported)\n", hca_cap.max_mw_num);
      printf("\tmax_raw_ipv6_qp=0x%x\t(Maximum number of Raw IPV6 QPs supported)\n", hca_cap.max_raw_ipv6_qp);
      printf("\tmax_raw_ethy_qp=0x%x\t(Maximum number of Raw Ethertypes QPs supported)\n", hca_cap.max_raw_ethy_qp);
      printf("\tmax_mcast_grp_num=0x%x\t(Maximum Number of multicast groups)\n", hca_cap.max_mcast_grp_num);
      printf("\tmax_mcast_qp_attach_num=0x%x\t(Maximum number of QP per multicast group)\n", hca_cap.max_mcast_qp_attach_num);
      printf("\tmax_total_mcast_qp_attach_num=0x%x\t(Maximum number of QPs which can be attached to a mcast grp)\n", hca_cap.max_total_mcast_qp_attach_num);
      printf("\tmax_ah_num=0x%x\t(Maximum number of address handles)\n", hca_cap.max_ah_num);
      printf("\tmax_num_fmr=%d\t(maximum number FMRs)\n", hca_cap.max_num_fmr);
      printf("\tmax_num_map_per_fmr=%d\t(Maximum number of (re)maps per FMR before an unmap operation in required)\n", hca_cap.max_num_map_per_fmr);
    }
    for ( port=1; port<=hca_cap.phys_port_num; ++port ) {
      printf("\t\tport=%d\n", port);
      rc = VAPI_query_hca_port_prop(hca_hndl, (IB_port_t)port, &hca_port);
      if ( rc != VAPI_OK ) {
        printf("Error: could not retrieve port properties\n");
        return -1;
      }
      printf("\t\tport_state=%s\n", port_state_str(hca_port.state));
      printf("\t\tsm_lid=0x%04x\n", hca_port.sm_lid);
      printf("\t\tport_lid=0x%04x\n", hca_port.lid);
      printf("\t\tport_lmc=0x%02x\n", hca_port.lmc);
      printf("\t\tmax_mtu=%s\n", mtu_str(hca_port.max_mtu));
      if ( verbose ) {
        printf("\t\tmax_msg_sz=0x%x\t(Max message size)\n", hca_port.max_msg_sz);
        printf("\t\tcapability_mask=0x%08x\n", hca_port.capability_mask); 
        printf("\t\tmax_vl_num=0x%x\t(Maximum number of VL supported by this port)\n", hca_port.max_vl_num);
        printf("\t\tbad_pkey_counter=0x%x\t(Bad PKey counter)\n", hca_port.bad_pkey_counter);
        printf("\t\tqkey_viol_counter=0x%x\t(QKey violation counter)\n", hca_port.qkey_viol_counter);
        printf("\t\tsm_sl=0x%x\t(IB_SL to be used in communication with subnet manager)\n", hca_port.sm_sl);
        printf("\t\tpkey_tbl_len=0x%x\t(Current size of pkey table)\n", hca_port.pkey_tbl_len);
        printf("\t\tgid_tbl_len=0x%x\t(Current size of GID table)\n", hca_port.gid_tbl_len);
        printf("\t\tsubnet_timeout=0x%x\t(Subnet Timeout for this port (see PortInfo))\n", hca_port.subnet_timeout);
        printf("\t\tinitTypeReply=0x%x\t(optional InitTypeReply value. 0 if not supported)\n", hca_port.initTypeReply);
        
        gid_tbl = alloc_get_all_port_gids(hca_hndl, (IB_port_t)port, &num_gids);
        if ( gid_tbl ) {
            for ( i=0; i<num_gids; ++i ) {
              printf("\t\tGID[%3d]= " GID_FMT "\n",i,GID_FMT_DATA(gid_tbl[i]));
            }

            free_port_gids(gid_tbl);
        }
      }
      printf("\n");
    }
    EVAPI_release_hca_hndl(hca_hndl);

    return 0;
}

