/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifdef __WIN__

//#include <sys/types.h>
#include <stdlib.h>
#include "sock.h"
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <mosal.h>

/*
 *  create_tcp_connection
 */ 
int create_tcp_connection(const char *ip, int port)
{
  struct in_addr remote_addr;
  int s, rc;
  struct sockaddr_in my_addr, server_addr;
  int optval=1;

  s = MOSAL_socket_socket(MOSAL_PF_INET, MOSAL_SOCK_STREAM, MOSAL_IPPROTO_TCP);
  if ( s == -1 ) {
    return -1;
  }

  rc = MOSAL_socket_setsockopt(s, MOSAL_SOL_SOCKET, MOSAL_SO_REUSEADDR, &optval, sizeof(int));
  if ( rc != 0 ) {
    return -1;
  }

  my_addr.sin_family = AF_INET;
  my_addr.sin_port = htons((unsigned short)port);
  my_addr.sin_addr.s_addr = INADDR_ANY;

  if ( inet_aton(ip,  &remote_addr) == 0 ) {
    return -1;
  }
  server_addr.sin_family = AF_INET;
  server_addr.sin_port = htons((unsigned short)port);
  server_addr.sin_addr = remote_addr;
  rc = MOSAL_socket_connect(s, (void *)(&server_addr), sizeof(server_addr));
  if ( rc == -1 ) {
    printf("%s\n", strerror(MOSAL_socket_get_last_error()));
    return -1;
  }

  return s;
}

int wait_for_connection(int port)
{
  int s, rc, newsock, addr_len, optval=1;
  struct sockaddr_in my_addr, remote_addr;

  s = MOSAL_socket_socket(MOSAL_PF_INET, MOSAL_SOCK_STREAM, MOSAL_IPPROTO_TCP);
  if ( s == -1 ) {
    return -1;
  }

  rc = MOSAL_socket_setsockopt(s, MOSAL_SOL_SOCKET, MOSAL_SO_REUSEADDR, &optval, sizeof(int));
  if ( rc != 0 ) {
    return -1;
  }
  
  my_addr.sin_family = AF_INET;
  my_addr.sin_port = htons((unsigned short)port);
  my_addr.sin_addr.s_addr = INADDR_ANY;
  rc = MOSAL_socket_bind(s, (void *)(&my_addr), sizeof(my_addr));
  if ( rc == -1 ) {
    printf("%s\n", strerror(MOSAL_socket_get_last_error()));
    return -1;
  }

  rc = MOSAL_socket_listen(s, 1);
  if ( rc == -1 ) {
    printf("%s\n", strerror(MOSAL_socket_get_last_error()));
    return -1;
  }

  addr_len = sizeof(remote_addr);
  newsock = MOSAL_socket_accept(s, (void *)(&remote_addr), &addr_len);
  if ( rc == -1 ) {
    printf("%s\n", strerror(MOSAL_socket_get_last_error()));
    return -1;
  }
  MOSAL_socket_close(s);
  return newsock;
}

#else

#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "sock.h"
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>


/*
 *  create_tcp_connection
 */ 
int create_tcp_connection(const char *ip, int port)
{
  struct in_addr remote_addr;
  int s, rc;
  struct sockaddr_in my_addr, server_addr;
  int optval=1;

  s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if ( s == -1 ) {
    return -1;
  }


  rc = setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(int));
  if ( rc != 0 ) {
    return -1;
  }

  my_addr.sin_family = AF_INET;
  my_addr.sin_port = htons(port);
  my_addr.sin_addr.s_addr = INADDR_ANY;

  if ( inet_aton(ip,  &remote_addr) == 0 ) {
    return -1;
  }
  server_addr.sin_family = AF_INET;
  server_addr.sin_port = htons(port);
  server_addr.sin_addr = remote_addr;
  rc = connect(s, (struct sockaddr *)(&server_addr), sizeof(server_addr));
  if ( rc == -1 ) {
#ifdef DEBUG
    printf("%s\n", strerror(errno));
#endif
    return -1;
  }

  return s;
}



int wait_for_connection(int port)
{
  int s, rc, newsock, addr_len, optval=1;
  struct sockaddr_in my_addr, remote_addr;

  s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if ( s == -1 ) {
    return -1;
  }

  rc = setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(int));
  if ( rc != 0 ) {
    return -1;
  }
  
  my_addr.sin_family = AF_INET;
  my_addr.sin_port = htons(port);
  my_addr.sin_addr.s_addr = INADDR_ANY;
  rc = bind(s, (struct sockaddr *)(&my_addr), sizeof(my_addr));
  if ( rc == -1 ) {
    printf("%s\n", strerror(errno));
    return -1;
  }

  rc = listen(s, 1);
  if ( rc == -1 ) {
    printf("%s\n", strerror(errno));
    return -1;
  }

  addr_len = sizeof(remote_addr);
  newsock = accept(s, (struct sockaddr *)(&remote_addr), &addr_len);
  if ( rc == -1 ) {
    printf("%s\n", strerror(errno));
    return -1;
  }
  close(s);
  return newsock;
}


#endif
