/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <string.h>
#include "sock.h"
//#include "utils.h"
#include <mosal.h>
#include <stdio.h>

/*
 *  create_tcp_connection
 */ 
int create_tcp_connection(const char *ip, int port)
{
    struct in_addr;
    int s, rc,on;
    struct sockaddr_in my_addr, server_addr;
    struct hostent *ent;

    s = MOSAL_socket_socket(MOSAL_PF_INET, MOSAL_SOCK_STREAM, MOSAL_IPPROTO_TCP);
    if ( s == -1 ) {
        return -1;
    }

    on = 1;
    rc = MOSAL_socket_setsockopt(s, MOSAL_SOL_SOCKET, MOSAL_SO_REUSEADDR, &on, sizeof(on));
    if (-1 == rc)
    {
        return -1;
    }

    my_addr.sin_family = AF_INET;
    my_addr.sin_port = htons(port);
    my_addr.sin_addr.s_addr = INADDR_ANY;

    if(!(ent = gethostbyname2(ip,AF_INET))||!ent->h_addr) {
        return -1;
    }

    server_addr.sin_family = AF_INET;
    server_addr.sin_port = htons(port);
    memcpy(&server_addr.sin_addr,ent->h_addr,sizeof(server_addr.sin_addr));
    rc = MOSAL_socket_connect(s, (void *)(&server_addr), sizeof(server_addr));
    if ( rc == -1 ) {
        return -1;
    }

    return s;
}


/* socket server waits for a client to connect */
int wait_for_connection(int port)
{
    int on,s, rc, newsock, addr_len;
    struct sockaddr_in my_addr, remote_addr;

  s = MOSAL_socket_socket(MOSAL_PF_INET, MOSAL_SOCK_STREAM, MOSAL_IPPROTO_TCP);
    if ( s == -1 ) {
        return -1;
    }
    
    on = 1;
    rc = MOSAL_socket_setsockopt(s, MOSAL_SOL_SOCKET, MOSAL_SO_REUSEADDR, &on, sizeof(on));
    if (-1 == rc)
    {
        return -1;
    }

    my_addr.sin_family = AF_INET;
    my_addr.sin_port = htons(port);
    my_addr.sin_addr.s_addr = INADDR_ANY;
    rc = MOSAL_socket_bind(s, (void *)(&my_addr), sizeof(my_addr));
    if ( rc == -1 ) {
        printf("%s\n", strerror(errno));
        return -1;
    }

    rc = MOSAL_socket_listen(s, 1);
    if ( rc == -1 ) {
        printf("%s\n", strerror(errno));
        return -1;
    }

    addr_len = sizeof(remote_addr);
    newsock = MOSAL_socket_accept(s, (void *)(&remote_addr), &addr_len);
    if ( rc == -1 ) {
        printf("%s\n", strerror(errno));
        return -1;
    }

    //DIEGO
    MOSAL_socket_close(s);
    
    return newsock;
}

//REV need to change so that if socket got closed the program does not hang forever waiting
//for the desired msg len
int sock_get_msg(int socket, char *msg, int len)
{
    u_int16_t msg_len =0, net_msg_len = 0;
    int length;
    for(length=0;length<2;)
      length += MOSAL_socket_recv(socket, (char *)&net_msg_len+length, 2-length, 0);

    /* turn from net work ordering to cpu */
    msg_len = MOSAL_be16_to_cpu(net_msg_len);
    
    if (msg_len > len-1) {
        fprintf(stderr,"Error: msg_len=%d but buffer is only %d\n",msg_len, len);
        return -1;
    }

    memset(msg,0,msg_len);
    for(length=0;length<msg_len;) 
      length += MOSAL_socket_recv(socket, msg+length, msg_len-length, 0);
    msg[msg_len] = '\0';
    //fprintf(stderr,"receive: %s [%d]\n", msg,msg_len);
    return length;
}



int sock_send_msg(int socket, const char *format, ...)
{
    char msg_out[MAX_MSG_LEN];
    char msg[MAX_MSG_LEN];
    int rc;
    u_int16_t len,net_len;
    va_list arglist;   

    va_start (arglist, format);   
    vsprintf (msg, format, arglist);   
    va_end (arglist);   

    len = strlen(msg);

    if (len > (MAX_MSG_LEN-3)) {
        fprintf(stderr,"Error: Send msg len=%d but buffer is only %d\n",len,MAX_MSG_LEN );
        return -1;
    }

    memset(msg_out,0,MAX_MSG_LEN);
    //turn msg to big endianess - network order
    net_len = MOSAL_cpu_to_be16(len);
    memcpy(msg_out,&net_len,2);
    memcpy(msg_out+2,msg,len);
    
    rc = MOSAL_socket_send(socket, msg_out, len+2, 0);
    if ( rc != (len+2) ) {
        fprintf(stderr,"Error: Couldn't send messags=%s len %d rlen %d error=%s\n",msg_out,len+2,rc,strerror(errno));
        return -1;
    }
    //printf("send: %s [%d]\n", msg,len+2);
    return 0;
}




void sock_msg2num(char *msg, char *label, long def, void *out, unsigned int len)
{
    char *startstr;
    u_int64_t tmp_val;

//long tmp_val;

    startstr = strstr(msg,label);
    if(NULL == startstr){
        tmp_val = def;
    }
    else {
      //        tmp_val = STR_2_ULL(startstr+strlen(label));
        tmp_val = strtoull(startstr+strlen(label),(char **) NULL,0);
        //tmp_val = atoll(startstr+strlen(label));
    }

    switch (len) {
        case 1:
            *((char *)out) = (char) tmp_val;
            break;
        case 2:
            *((short *)out) = (short ) tmp_val;
            break;
        case 4:
            *((int *)out) = (int ) tmp_val;
            break;
        case 8:
            *((u_int64_t *)out) = (u_int64_t ) tmp_val;
            break;
        default:
            *((char *)out) = (char) tmp_val;
            break;
    }
}
     
void sock_msg2str(char *msg, char *label, char *def, char  *out)
{
    char *startstr;
    
    startstr = strstr(msg,label);
    if(NULL == startstr){
        strcpy(out,def);
        return;
    }
    if(0 == sscanf(startstr+strlen(label),"%s",out)){
        strcpy(out,def);
    }
}

unsigned char sock_is_exists(char *msg, char *label)
{
    return (NULL == strstr(msg,label)) ? 0 : 1;
}


void sock_sync_msg(int socket, int is_sender)
{
    char                   sock_msg[500];
    memset(sock_msg,0,500);
    
    /* sync start test */
    if (is_sender) {
        while (1) {
            //COND_PRINTF2(("Sender: Waiting to get start........"));
            sock_get_msg(socket, sock_msg, 500);
            if (sock_is_exists(sock_msg,"start")) {
                //COND_PRINTF2(("Sender Got start\n"));
                sock_send_msg(socket, "start");
                //COND_PRINTF2(("Sender send start\n"));
                break;
            }
        }
//	sleep(2);
    }
    else {
        sock_send_msg(socket, "start");
        //COND_PRINTF2(("Receiver send start\n"));
        while (1) {
            //COND_PRINTF2(("Receiver Waiting to get start........"));
            sock_get_msg(socket, sock_msg, 500);
            if (sock_is_exists(sock_msg,"start")) {
                //COND_PRINTF2(("Receiver Got start\n"));
                break;
            }
        }
    }
}

