/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include "vping.h"
#include <mpga/mpga.h>

#include "sock.h"


#define VPING_MGR_MAXIPNODES  2048L
#define VPING_MGR_MAXLINESIZE 100

/* VPing Manager Database Struct */
typedef struct {
  int socket;
  char remote_ip[100];
  IB_port_t         port;
  IB_lid_t        dlid;
  VAPI_qp_num_t   remote_qpnum;
} vping_mgrdb_t;



//-------------------------------------------------------------------------------------------
// VPing Manager Functions
//-------------------------------------------------------------------------------------------
VAPI_ret_t vping_manager_parse_ip_list(vping_res_t* res,vping_mgrdb_t *pDB, int *pNumHosts)
{
  VAPI_ret_t ret;
  FILE *rc;
  int nHosts, rtn;
  unsigned int temp_port;
  bool bCont;
  char szLine[VPING_MGR_MAXLINESIZE];

  ret= VAPI_OK;

  
  if ((rc = fopen(res->remote_ip, "r")) == NULL){
    MTL_ERROR1("Could not open IP list file name %s.\n",res->remote_ip);
    ret= VAPI_EGEN; //REV what is the right error code?
    goto err_exit;
  }

  MTL_ERROR1("Building Hosts Database\n");
  for(nHosts=0,bCont=TRUE;(bCont) && (nHosts < VPING_MGR_MAXIPNODES);)  {
    if (fgets(szLine, VPING_MGR_MAXLINESIZE, rc) == NULL) {
      //reached end of file
      bCont=FALSE;
      break;
    }
    if (szLine[0] == '#') {
        continue;  /* ignore lines beginning with # so can use /etc/hosts */
    }
    if (((rtn=sscanf(szLine,"%s %u %*s",pDB[nHosts].remote_ip, &temp_port)) == 0) || (rtn == 1)) {
        if (sscanf(szLine,"%s %*s",pDB[nHosts].remote_ip) == 0) {
            MTL_ERROR1("\t Invalid host information line: %s\n",szLine);
            ret= VAPI_EGEN; //REV what is the right error code?
            goto err_exit;
        } else {
            temp_port = 1;
        }
    } else {
        if (temp_port == 0 || temp_port > 2) {
            MTL_ERROR1("\t Host DB file gives invalid port number (%u) for host %s\n",
                       temp_port, pDB[nHosts].remote_ip);
            ret= VAPI_EGEN; //REV what is the right error code?
            goto err_exit;
        }
    }

    //discard empty lines
    if (strlen(pDB[nHosts].remote_ip))  {
      pDB[nHosts].port= (IB_port_t) temp_port;
      MTL_ERROR1("\t Adding %s/port %u to Database\n",pDB[nHosts].remote_ip, pDB[nHosts].port);
      nHosts++;
    }
  }

 if (nHosts == VPING_MGR_MAXIPNODES) {
    MTL_ERROR1("### Warning - %s contains more than the supported max number of remote hosts\n",res->remote_ip);
    MTL_ERROR1("   Ignoring hosts beyond %s \n",pDB[nHosts-1].remote_ip);
  }

  MTL_ERROR1("%d hosts in Database\n\n\n",nHosts);

  *pNumHosts=nHosts;

  fclose(rc);

err_exit:
  return ret;
}
//-------------------------------------------------------------------------------------------
VAPI_ret_t vping_manager(vping_res_t* res)
{
  VAPI_ret_t      ret = VAPI_OK;
  char sock_msg[VPING_TCPD_MAX_MSG_LEN+1], *pMsg, szAddress[VPING_TCPD_ADDRESSSIZE+1];
  u_int8_t        cRes;
  int             n, i, nPings, h, hh, nHosts, s;
  vping_mgrdb_t   *pDB;
  MT_bool         last_sample_ok = FALSE;
  int             have_errors = 0;


 if (res->simultpingers > MAX_OUTSTANDING_PINGS) {
    MTL_ERROR1("### Warning - %d simult pingers while VPING daemon is configured to post %d outstanding receive WQEs\n",res->simultpingers,MAX_OUTSTANDING_PINGS);
    MTL_ERROR1("\tPing requests may be dropped by the daemon and false failures may be reported \n");
    MTL_ERROR1("\tIncrease MAX_OUTSTANDING_PINGS in vping.h to match number of desired simult pingers and rerun\n");
  }

  pDB=(vping_mgrdb_t *)malloc(VPING_MGR_MAXIPNODES * sizeof(vping_mgrdb_t));
  if (pDB == NULL) {
    MTL_ERROR1("Could not allocate vping manager Database (size:%luKB)\n",(VPING_MGR_MAXIPNODES * sizeof(vping_mgrdb_t))>>10);
    ret= VAPI_EGEN; //REV what is the right error code?
    goto err_exit;
  }

  ret=vping_manager_parse_ip_list(res,pDB,&nHosts);
  if (ret != VAPI_OK) goto err_exit;

  MTL_ERROR1("Connecting to remote daemons to gather LID and QP \n");
  //connect to daemons and get lid/qp information
  for (h=0;h<nHosts;h++)
  {
    pDB[h].socket = create_tcp_connection(pDB[h].remote_ip,res->socket_port+pDB[h].port);
    if (pDB[h].socket == -1) {
      MTL_ERROR1("Could not connect to Daemon Socket at %s:%d.\n",pDB[h].remote_ip,res->socket_port+1);
      if (last_sample_ok == TRUE) {
          fprintf(stderr, "\n");
      }
      fprintf(stdout, ".");
      fflush(stdout);
      last_sample_ok = FALSE;
      have_errors++;
      fprintf(stderr, "IP %s, TCP port %d : failed to create TCP connection with vping tcp daemon\n",
              pDB[h].remote_ip,res->socket_port+1);
      fflush(stderr);
      pDB[h].dlid=0;
      pDB[h].remote_qpnum=0;
    } else {
      sock_get_msg(pDB[h].socket, sock_msg, VPING_TCPD_MAX_MSG_LEN+1);
      sock_msg2num(sock_msg, "LID=", 0,&(pDB[h].dlid), sizeof(pDB[h].dlid));
      sock_msg2num(sock_msg, "QP=", 0,&(pDB[h].remote_qpnum),sizeof(pDB[h].remote_qpnum));
      MTL_ERROR1("\t%s IBPort=%d : DLID=0x%04X QP=0x%06X\n",pDB[h].remote_ip,pDB[h].port,pDB[h].dlid,pDB[h].remote_qpnum);
    }
  }
  MTL_ERROR1("\n\n");

  nPings=nHosts;

  for(hh=0;hh<nHosts;hh+=res->simultpingers)
  {
    //trigger remote daemons to start pinging
    //REV can be optimized to build all the tcp msgs once and then send to all dest
    MTL_ERROR1("Sending Ping trigger command to remote daemons \n");
    for (h=hh,s=0;(h<nHosts)&&(s<res->simultpingers);h++,s++)
    {
      if (pDB[h].socket != -1) {
        sock_send_msg(pDB[h].socket, "nCmd=%d",VPING_TCPD_REMOTE_COMMAND_PING);
        sock_send_msg(pDB[h].socket, "nPings=%d",nPings);

        //Send ping dlid/qps to host h
        for(n=0;n<nPings;) {
          sock_msg[0]=0;
          for(pMsg=sock_msg,i=0;(n<nPings) && (i<VPING_TCPD_ADDRESSPERMSG);i++,n++,pMsg+=VPING_TCPD_ADDRESSSIZE) {
            sprintf(szAddress,"L0x%04XQ0x%06X",pDB[n].dlid,pDB[n].remote_qpnum);
            strcpy(pMsg,szAddress);
          }
          sock_send_msg(pDB[h].socket, sock_msg);
        }
      }

      //progress indicator
      if ((h & 0xF) == 0xF) {
        MTL_ERROR1("\n");
      } else {
        MTL_ERROR1(".");
      }

    }
    MTL_ERROR1("\n\n");


    MTL_ERROR1("### VPING Manager Results ###\n");
    //get ping results from remote daemons and display results
    for (h=hh,s=0;(h<nHosts)&&(s<res->simultpingers);h++,s++)
    {
      if (pDB[h].socket != -1) {
        MTL_ERROR1("From %s IBPort=%d LID=0x%04X \n",pDB[h].remote_ip,pDB[h].port,pDB[h].dlid);
        //get ping results from host h
        for(n=0;n<nPings;) {
          sock_get_msg(pDB[h].socket, sock_msg, VPING_TCPD_MAX_MSG_LEN+1);
          for(pMsg=sock_msg,i=0;(n<nPings) && (i<VPING_TCPD_RESULTPERMSG);i++,n++,pMsg+=VPING_TCPD_RESULTSIZE) {
            sock_msg2num(pMsg, "R", VPING_TCPD_RESULT_NORES,&cRes, sizeof(cRes));
            switch(cRes) {

              case VPING_TCPD_RESULT_OK:
                //MTL_ERROR1("\t to %s IBPort=%d LID=0x%04X : OK\n",pDB[n].remote_ip,pDB[n].port,pDB[n].dlid);
                fprintf(stdout, ".");
                fflush(stdout);
                last_sample_ok = TRUE;
                break;

              case VPING_TCPD_RESULT_BADDEST:
                MTL_ERROR1("\t to %s IBPort=%d  LID=0x%04X : BAD Destination\n",pDB[n].remote_ip,pDB[n].port,pDB[n].dlid);
                if (last_sample_ok == TRUE) {
                    fprintf(stderr, "\n");
                }
                last_sample_ok = FALSE;
                have_errors++;
                fprintf(stderr, "From %s IBPort=%d LID=0x%04X to  %s IBPort=%d  LID=0x%04X : BAD Destination\n",
                        pDB[h].remote_ip,pDB[h].port,pDB[h].dlid,pDB[n].remote_ip,pDB[n].port,pDB[n].dlid);
                fflush(stderr);
                break;

              case VPING_TCPD_RESULT_FAILURE:
                MTL_ERROR1("\t to %s IBPort=%d  LID=0x%04X : FAILED\n",pDB[n].remote_ip,pDB[n].port,pDB[n].dlid);
                if (last_sample_ok == TRUE) {
                    fprintf(stderr, "\n");
                }
                last_sample_ok = FALSE;
                have_errors++;
                fprintf(stderr, "From %s IBPort=%d LID=0x%04X to  %s IBPort=%d  LID=0x%04X : FAILED\n",
                        pDB[h].remote_ip,pDB[h].port,pDB[h].dlid,pDB[n].remote_ip,pDB[n].port,pDB[n].dlid);
                fflush(stderr);
                break;

              case VPING_TCPD_RESULT_NORES:
              default:
                  if (last_sample_ok == TRUE) {
                      fprintf(stderr, "\n");
                  }
                  last_sample_ok = FALSE;
                  have_errors++;
                  fprintf(stderr, "From %s IBPort=%d LID=0x%04X to %s IBPort=%d  LID=0x%04X : UNEXPECTED RESULT 0x%02X\n",
                          pDB[h].remote_ip,pDB[h].port,pDB[h].dlid,pDB[n].remote_ip,pDB[n].port,pDB[n].dlid, cRes);
                  fflush(stderr);
                  MTL_ERROR1("\t to %s IBPort=%d  LID=0x%04X : ***** UNEXPECTED RESULT 0x%02X - QUIT AND TRY AGAIN\n",pDB[n].remote_ip,pDB[n].port,pDB[n].dlid,cRes);
                  break;
            }
          }
        }

        //terminate remote session
        sock_send_msg(pDB[h].socket, "nCmd=%d",VPING_TCPD_REMOTE_COMMAND_DONE);

        //close tcp connection to daemon
        MOSAL_socket_close(pDB[h].socket);
        pDB[h].socket=-1;

      }
      MTL_ERROR1("\t Node Done\n");
    }
  }

  fprintf(stdout,"\nDONE\n");
  if (have_errors > 0) {
      fprintf(stdout, "There were %d connectivity errors -- check stderr log if it was redirected\n", have_errors);
  }
  //free vping manager database
  free(pDB);

err_exit:

  return VAPI_OK;
}
//-------------------------------------------------------------------------------------------





//-------------------------------------------------------------------------------------------
// VPing TCP Daemon Functions
//-------------------------------------------------------------------------------------------
VAPI_ret_t vping_remote_tcpdaemon_ping(vping_res_t* vres)
{
VAPI_ret_t ret;
int nPings,n,i;
vping_res_t *pResults,*res;
char sock_msg[VPING_TCPD_MAX_MSG_LEN+1], *pMsg, szResult[VPING_TCPD_RESULTSIZE+1];

  ret=VAPI_OK;

  //get number of Pings requested by remote
  sock_get_msg(vres->socket, sock_msg, VPING_TCPD_MAX_MSG_LEN+1);
  sock_msg2num(sock_msg, "nPings=", -1,&nPings, sizeof(nPings));
  MTL_ERROR1("  nPings %d\n",nPings);

  if (nPings > 0) {

    //allocate buffer for nPing ctxts and results
    pResults=(vping_res_t *)malloc(nPings*sizeof(vping_res_t));

    if (pResults) {

      //get DLID and DestQP for nPing pings
      for(n=0;n<nPings;) {
        sock_get_msg(vres->socket, sock_msg, VPING_TCPD_MAX_MSG_LEN+1);
        for(pMsg=sock_msg,i=0;(n<nPings) && (i<VPING_TCPD_ADDRESSPERMSG);i++,n++,pMsg+=VPING_TCPD_ADDRESSSIZE) {
          res=pResults+n;
          memset(res, 0, sizeof(vping_res_t));
          sock_msg2num(pMsg, "L", 0,&(pResults[n].dlid), sizeof(pResults[n].dlid));
          sock_msg2num(pMsg, "Q", 0,&(pResults[n].remote_qpnum), sizeof(pResults[n].remote_qpnum));
        }
      }

      //do nPing pings
      for(n=0;n<nPings;n++) {
        MTL_ERROR1("  DLID 0x%04X RemoteQPN 0x%06X for ping %d\n",pResults[n].dlid,pResults[n].remote_qpnum,n);
        res=pResults+n;
        res->exit_flag= FALSE;
        res->opts= 0;
        res->num_of_mcg= 0;
        res->port= vres->port;
        res->qkey= VPING_Q_KEY;
        strcpy(res->hca_id, HCA_ID);
        res->payload_size= MIN_PAYLOAD_SZ;
        res->vping_max_avs = VPING_MAX_AVS;
        res->vping_prealloc_avs = VPING_PREALLOC_AVS;
        res->cluster_conn_test_fl = TRUE;
        //res->opts |= VPING_OPTS_BLOCK_POLL; //REV do we need it?

        if ((res->dlid == 0) || (res->remote_qpnum == 0)){
          res->tcpdret = VPING_TCPD_RESULT_BADDEST;
        } else  {
          //REV optimization: could send all pings in parallel and wait for responses
          ret = (VAPI_ret_t)(MT_virt_addr_t)start_vping(res); //REV signals and abnormal exit?
          if (ret == VAPI_OK) {
            res->tcpdret = VPING_TCPD_RESULT_OK;
          } else {
            res->tcpdret = VPING_TCPD_RESULT_FAILURE;
          }
        }

      }

      //Send ping results
      for(n=0;n<nPings;) {
        sock_msg[0]=0;
        for(pMsg=sock_msg,i=0;(n<nPings) && (i<VPING_TCPD_RESULTPERMSG);i++,n++,pMsg+=VPING_TCPD_RESULTSIZE) {
          res=pResults+n;
          sprintf(szResult,"R0x%02X",res->tcpdret);
          strcpy(pMsg,szResult);
        }
        sock_send_msg(vres->socket, sock_msg);
      }

    free((void *)pResults);
    } else {
      MTL_ERROR1("  Could not allocate Results struct for %d remote pings - Terminating Remote VPING session\n",nPings);
      ret=VAPI_EGEN; //REV put right error code
    }
  } else {
    MTL_ERROR1("  Invalid nPings: %d - Terminating Remote VPING session\n",nPings);
    ret=VAPI_EGEN; //REV put right error code
  }

return ret;
}
//-------------------------------------------------------------------------------------------
/* Using void* in parameter in order to match pthread start function prototype */
void* start_vpingtcpdaemon(void *void_res)
{
  VAPI_ret_t ret;
  vping_res_t *vres= void_res;
  char sock_msg[VPING_TCPD_MAX_MSG_LEN+1];
  int nCmd;
  bool bCont;

  while(1)  {
    MTL_ERROR1("Waiting for TCP connection on port %d.\n",vres->socket_port+vres->port);
    vres->socket = wait_for_connection(vres->socket_port+vres->port);
    if (vres->socket == -1) {
      MTL_ERROR1("Wait for TCP connection failed. Aborting.\n");
      ret=VAPI_EAGAIN;
      return (void*)(MT_virt_addr_t)ret;
    }
    MTL_ERROR1("  TCP connection established\n");

    MTL_ERROR1("  Sending local LID=0x%04X and QP=0x%06X over TCP connection\n",vres->slid,vres->local_qpnum);
    sock_send_msg(vres->socket, "LID=0x%04X QP=0x%06X",vres->slid,vres->local_qpnum);

    //respond to remote commands until VPING_TCPD_REMOTE_COMMAND_DONE
    bCont=TRUE;
    while(bCont) {
      sock_get_msg(vres->socket, sock_msg, VPING_TCPD_MAX_MSG_LEN+1);
      sock_msg2num(sock_msg, "nCmd=", -1,&nCmd, sizeof(nCmd));
      switch(nCmd) {
        case VPING_TCPD_REMOTE_COMMAND_PING:
          ret=vping_remote_tcpdaemon_ping(vres);
          if (ret != VAPI_OK) {
            MTL_ERROR1("  Remote VPING Failed\n");
            bCont=FALSE;
          }
          break;

        case VPING_TCPD_REMOTE_COMMAND_DONE:
          MTL_ERROR1("  Remote VPING Done\n");
          bCont=FALSE;
          break;

        case -1:
        default:
          MTL_ERROR1("  Wrong Remote VPING Command received: %d - Terminating Remote VPING session\n",nCmd);
          bCont=FALSE;
          break;
      }
    }

    MOSAL_socket_close(vres->socket);
    vres->socket=-1;
    MTL_ERROR1("  TCP connection Closed\n");
  }
  return (void*)(MT_virt_addr_t)VAPI_OK;
}
//-------------------------------------------------------------------------------------------


