/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#include <vipkl_wrap.h>
/* #include "vipkl_hpool.c" - obsolete */
#include <errno.h>
#define MTPERF
#include <mtperf.h>
MTPERF_NEW_SEGMENT(evapi_comp_handler_agent,3000);
MTPERF_NEW_SEGMENT(evapi_comp_handler_handler,3000);


/* 
	Wrapper functions' implementation
	-----------------------------

1. Main ideas
	- all input parameters are serialiazed into an INPUT buffer; output ones -into an OUPUT buffer;
	- pointers are never relayed: if a parameter is a pointer to a structure/buffer, 
	   the structure/buffer is copied to the INPUT buffer;
	- kernel function return code is always relayed as the first field of the OUTPUT buffer;
	- the INPUT/OUTPUT buffers are relayed to/from kernel by an OS-dependent function;

2. Implementation detailes;
	- the buffers are allocated on the stack (macros W_ALLOCA_xxx);
	- if there is only one input variable, it is used as an INPUT buffer (macros W_USE_xxx_AS_IBUF);
	- wrapper functions are written with the help of several macros which use the following
	  a) conventions:
	  	-- for function XXX, 
	  		INPUT buffer has 'struct i_XXX_ops_t' type,
	  		OUTPUT buffer has 'struct i_XXX_ops_t' type;
	  	-- if a function parameter is pointer to a structure or array, it must be named YYY_p and 
	  	    the appropriate field of INPUT or OUTPUT buffer must be named YYY;
	  	-- if a function parameter is a value of an integral type, its name must be the same, as the 
	  	    name of the appropriate INPUT buffer field;
	  	    
	  b) local variables:
	  	-- pi 	- pointer to the INPUT buffer of 'struct i_<func_name>_ops_t *' type
	  	-- pi_sz	- size of the INPUT buffer;
	  	-- po 	- pointer to the OUTPUT buffer of 'struct o_<func_name>_ops_t *' type
	  	-- pi_sz	- size of the OUTPUT buffer;

*/

/***************************************************************
 * Helper macros
 **************************************************************/
 
/* allocate on stack N variables of type T */
#define TNALLOCA(T, N)      			(T*)ALLOCA((N) * sizeof(T))

/* allocate on stack a variable of type T */
#define TALLOCA(T)         				TNALLOCA(T, 1)

/* allocate a structure of type 'type' and declare a pointer 'var' to it */
#define ALLOCA_STRUCT(var,type)		type * var = TALLOCA(type)  


/***************************************************************
 * INPUT buffer allocation 
 **************************************************************/

/* allocate INPUT buffer of type 'type' (declares and inits 'pi' and 'pi_sz'); */
#define W_ALLOCA_IBUF(type)			ALLOCA_STRUCT(pi, struct i_##type##_ops_t); \
	u_int32_t pi_sz = sizeof(struct i_##type##_ops_t)

/* allocate INPUT buffer of type 'type' and size 'size' (declares and inits 'pi' and 'pi_sz'); */
#define W_ALLOCA_IBUF_SZ(type, size)	 u_int32_t pi_sz = size;		\
	struct i_##type##_ops_t *pi = (struct i_##type##_ops_t *)ALLOCA(size)
	
/* allocate INPUT buffer of type 'type', extended after field 'name' by 'size' bytes (declares and inits 'pi' and 'pi_sz'); */
#define W_ALLOCA_IBUF_EX(type,name,size)		\
	u_int32_t pi_sz = (u_int32_t)(FIELD_OFFSET(struct i_##type##_ops_t,name) + size);	\
	struct i_##type##_ops_t *pi = (struct i_##type##_ops_t *)ALLOCA(pi_sz)

/* use integral input parameter 'name' as an INPUT buffer of type 'type' (declares and inits 'pi' and 'pi_sz'); */
#define W_USE_VAR_AS_IBUF(name,type)		u_int32_t pi_sz = sizeof(struct i_##type##_ops_t);	\
	struct i_##type##_ops_t *pi = (struct i_##type##_ops_t *)&name
	
/* use string input parameter 'name' as an INPUT buffer of type 'type' (declares and inits 'pi' and 'pi_sz'); */
#define W_USE_STRING_AS_IBUF(name,type)	u_int32_t pi_sz = strlen(name)+1;	\
	struct i_##type##_ops_t *pi = (struct i_##type##_ops_t *)name


/***************************************************************
 * INPUT buffer filling
 **************************************************************/

/* copy an input parameter of integral type to the INPUT buffer */
#define W_PUT_VAR(name)				pi->name = name
    
/* copy an input parameter of a non-integral type to the INPUT buffer */
#define W_PUT_LONGVAR(name)    			memcpy( (void*)&pi->name, (void*)&name, sizeof(pi->name) )

/* copy an input parameter of a structure type to the INPUT buffer */
#define W_PUT_STRUCT(name)    			memcpy( (void*)&pi->name, (void*)name##_p, sizeof(pi->name) )
    
/* copy a buffer to the INPUT buffer */
#define W_PUT_BUF(name,size)			memcpy( (void*)pi->name, (void*)name##_p, size )

/* copy an ASCIIZ string to the INPUT buffer */
#define W_PUT_STRING(name)				do {                                                  \
    strncpy((void*)pi->name,(void*)name,sizeof(pi->name)-1);                            \
    ((char*)pi->name)[sizeof(pi->name)-1]= 0;   /* Assure string in NULL terminated */  \
  } while(0)


/***************************************************************
 * OUTPUT buffer allocation 
 **************************************************************/

/* allocate OUTPUT buffer of type 'type' (declares and inits 'po' and 'po_sz'); */
#define W_ALLOCA_OBUF(type)			ALLOCA_STRUCT(po, struct o_##type##_ops_t); \
	u_int32_t po_sz = sizeof(struct o_##type##_ops_t)

/* allocate OUTPUT buffer of type 'type' and size 'size' (declares and inits 'po' and 'po_sz'); */
#define W_ALLOCA_OBUF_SZ(type, size)	 u_int32_t po_sz = size;	\
	struct o_##type##_ops_t *po = (struct o_##type##_ops_t *)ALLOCA(size)
	
/* allocate OUTPUT buffer of type 'type', extended after field 'name' by 'size' bytes (declares and inits 'po' and 'po_sz'); */
#define W_ALLOCA_OBUF_EX(type,name,size)		\
	u_int32_t po_sz = (u_int32_t)(FIELD_OFFSET(struct o_##type##_ops_t,name) + size);	\
	struct o_##type##_ops_t *po = (struct o_##type##_ops_t *)ALLOCA(po_sz)


/***************************************************************
 * OUTPUT buffer emptying
 **************************************************************/

/* copy a field of OUTPUT bufer of integral type to the variable, pointed by 'name' */
#define W_GET_VAR(name)				*name##_p = po->name

/* copy an output parameter of a structure type from the OUTPUT buffer */
#define W_GET_LONGVAR(name)    			memcpy( (void*)&name, (void*)&po->name, sizeof(po->name) )
    
/* copy an output parameter of a structure type from the OUTPUT buffer */
#define W_GET_STRUCT(name)    			memcpy( (void*)name##_p, (void*)&po->name, sizeof(po->name) )
    
/* copy a buffer from the OUTPUT buffer */
#define W_GET_BUF(name,size)			memcpy( (void*)name##_p, (void*)&po->name, size )

/* copy an ASCIIZ string from the OUTPUT buffer */
#define W_GET_STRING(name)				W_GET_BUF(name,strlen(name##_p)+1)


/***************************************************************
 * Various
 **************************************************************/

/* return code of the kernel function */
#define W_RET							po->ret
#define W_OPEN_IOCTL                                		\
	do {							\
		VIP_ret_t ret;					\
		u_int32_t lib_ver= VIPKL_VERSION;	 	\
		if (vip_ioctl_open()) return VIP_ESYSCALL;	\
		if (						\
		    vip_ioctl_wrapper(VIPKL_LIB_VER, &lib_ver, sizeof(u_int32_t), &ret, sizeof(VIP_ret_t)))	\
		    return VIP_ESYSCALL;										\
		if (ret != VIP_OK)  return ret; /* no matching in versions of library and driver */	\
	} while (0)

/* call the function, performing the call to kernel; on error - return  */
#define W_ISSUE_IOCTL(cmd)				\
	do { if (vip_ioctl_wrapper( cmd,pi, pi_sz, po, po_sz )) return VIP_ESYSCALL; }	\
    	while (W_RET == VIP_EINTR)

/* call the function, performing the call to kernel; on error - goto 'lbl'  */
#define W_ISSUE_IOCTL_LABEL(cmd,lbl)					\
	do { 												\
		if (vip_ioctl_wrapper( cmd,pi, pi_sz, po, po_sz )) {	\
			ret = VIP_ESYSCALL; goto lbl; 					\
		}												\
		else												\
			ret = W_RET;									\
	}													\
    	while (W_RET == VIP_EINTR)


/***************************************************************
 * reference counting mechanism for callbacks
 **************************************************************/

#define INC_REF_CNT(handler,proc_cbk_hndl_p,handler_id_p)                           \
    /* Increase by 1 reference count and check whether it is the first bind */      \
    while (MOSAL_mutex_acq(&cbk_ref_mutex,TRUE) != MT_OK); /* user space -> retry */\
    if (cbk_ref_cnt == 0) { /* First bind request required k2u_cbk resources init*/ \
      MTL_DEBUG4(MT_FLFMT("Invoking k2u_cbk_register"));                            \
      rc= k2u_cbk_register(handler,proc_cbk_hndl_p,handler_id_p);                   \
      if (rc != MT_OK) {                                                            \
        MOSAL_mutex_rel(&cbk_ref_mutex);                                       \
        MTL_DEBUG2("%s: failed allocating k2u_cbk resources.\n", __func__);        \
        return VIP_EAGAIN;                                                          \
      }                                                                             \
    }                                                                               \
    cbk_ref_cnt++;                                                                  \
    MTL_DEBUG4(MT_FLFMT("cbk_ref_cnt= %d -> %d"),cbk_ref_cnt-1,cbk_ref_cnt);        \
    MOSAL_mutex_rel(&cbk_ref_mutex)

#define DEC_REF_CNT(handler_id)                                                     \
    while (MOSAL_mutex_acq(&cbk_ref_mutex,TRUE) != MT_OK); /* user space -> retry */\
    if (cbk_ref_cnt == 1) { /* Last unbind - clean k2u_cbk resources */             \
      rc= k2u_cbk_deregister(handler_id);                                           \
      if (rc != MT_OK) {                                                            \
        MOSAL_mutex_rel(&cbk_ref_mutex);                                       \
        MTL_DEBUG2("%s: failed allocating k2u_cbk resources.\n", __func__);        \
        return VIP_EAGAIN;                                                          \
      }                                                                             \
    }                                                                               \
    cbk_ref_cnt--;                                                                  \
    MTL_DEBUG4(MT_FLFMT("cbk_ref_cnt= %d -> %d"),cbk_ref_cnt+1,cbk_ref_cnt);        \
    MOSAL_mutex_rel(&cbk_ref_mutex)
    


/* These inline static functions decode parameters
 * and delegate each to the appropriate VIP KL API */

/*****************************************
  struct i_VIPKL_open_hca_ops_t {
    IN VAPI_hca_id_t hca_id;
    IN MT_bool profile_is_null;
    IN MT_bool sugg_profile_is_null;
    IN EVAPI_hca_profile_t profile;
  };
  struct o_VIPKL_open_hca_ops_t {
    OUT VIP_ret_t ret;
    OUT EVAPI_hca_profile_t  sugg_profile;
    OUT VIP_hca_hndl_t hca_hndl;
  };
******************************************/
VIP_ret_t VIPKL_open_hca(  
    /*IN*/ VAPI_hca_id_t hca_id_p
  , /*IN*/ EVAPI_hca_profile_t  *profile_p
  ,  /*OUT*/ EVAPI_hca_profile_t  *sugg_profile_p
  ,  /*OUT*/ VIP_hca_hndl_t *hca_hndl_p)
{
    MT_bool profile_is_null, sugg_profile_is_null;
	W_ALLOCA_IBUF(VIPKL_open_hca);
	W_ALLOCA_OBUF(VIPKL_open_hca);
    if (profile_p == NULL) {
        profile_is_null = TRUE;
    } else {
        profile_is_null = FALSE;
        W_PUT_STRUCT(profile);
    }
    W_PUT_VAR(profile_is_null);
    
    if (sugg_profile_p == NULL) {
        sugg_profile_is_null = TRUE;
    } else {
        sugg_profile_is_null = FALSE;
    }
    W_PUT_VAR(sugg_profile_is_null);
    W_PUT_BUF(hca_id, sizeof(VAPI_hca_id_t));
    W_OPEN_IOCTL;
    W_ISSUE_IOCTL(VIPKL_OPEN_HCA);
	W_GET_VAR(hca_hndl);
    if (sugg_profile_p != NULL) {
       W_GET_STRUCT(sugg_profile);
    }
	return W_RET;
}

/*****************************************
  struct i_VIPKL_close_hca_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
  };
  struct o_VIPKL_close_hca_ops_t {
    OUT VIP_ret_t ret;
  };
******************************************/
VIP_ret_t VIPKL_close_hca(
    /*IN*/ VIP_hca_hndl_t hca_hndl)
{
	W_USE_VAR_AS_IBUF(hca_hndl, VIPKL_close_hca);
	W_ALLOCA_OBUF(VIPKL_close_hca);
	W_ISSUE_IOCTL(VIPKL_CLOSE_HCA);
	return W_RET;
}

/*****************************************
  struct i_VIPKL_get_hca_hndl_ops_t {
    IN VAPI_hca_id_t hca_id; 
  };
  struct o_VIPKL_get_hca_hndl_ops_t {
    OUT VIP_ret_t ret;
    OUT VIP_hca_hndl_t 	hca_hndl;
  };
******************************************/
VIP_ret_t VIPKL_get_hca_hndl(  
    /*IN*/ VAPI_hca_id_t hca_id
  ,  /*OUT*/ VIP_hca_hndl_t *hca_hndl_p)
{
	W_ALLOCA_IBUF(VIPKL_get_hca_hndl);
	W_ALLOCA_OBUF(VIPKL_get_hca_hndl);
	W_PUT_STRING(hca_id);
  W_OPEN_IOCTL;
	W_ISSUE_IOCTL(VIPKL_GET_HCA_HNDL);
	if (W_RET == VIP_OK) {
		if (hca_hndl_p) W_GET_VAR(hca_hndl);
	}
	return W_RET;
}

/*****************************************
struct i_VIPKL_list_hcas_ops_t {
  IN u_int32_t         hca_id_buf_sz; 
};
struct o_VIPKL_list_hcas_ops_t {
  OUT VIP_ret_t ret;
  OUT u_int32_t       num_of_hcas;
  OUT VAPI_hca_id_t   hca_id_buf[1];
};
******************************************/
VIP_ret_t VIPKL_list_hcas(  /*IN*/ u_int32_t  hca_id_buf_sz
  ,  /*OUT*/ u_int32_t*      num_of_hcas_p
  ,  /*OUT*/ VAPI_hca_id_t*  hca_id_buf_p)
{
	W_USE_VAR_AS_IBUF(hca_id_buf_sz, VIPKL_list_hcas);
	W_ALLOCA_OBUF_EX(VIPKL_list_hcas,hca_id_buf,sizeof(VAPI_hca_id_t) * hca_id_buf_sz);

	/* check parameters */
	if (num_of_hcas_p == NULL || 
		(hca_id_buf_sz > 0 && hca_id_buf_p == NULL) ||
		(hca_id_buf_sz == 0 && hca_id_buf_p != NULL) )
		return VIP_EINVAL_PARAM;
    W_OPEN_IOCTL;
	W_ISSUE_IOCTL(VIPKL_LIST_HCAS);
	
	/* store the results */	
	if 	(W_RET == VIP_OK || W_RET == VIP_EAGAIN) {
		int num_to_copy, i;
		W_GET_VAR(num_of_hcas);
		if (hca_id_buf_sz > 0) {
			VAPI_hca_id_t * hca_id_p = (VAPI_hca_id_t *)po->hca_id_buf;
    			num_to_copy = MIN(hca_id_buf_sz,*num_of_hcas_p);
    			for (i = 0; i < num_to_copy; i++) {
	    	    		memset((char *)hca_id_buf_p[i], 0, sizeof(VAPI_hca_id_t));
		        	strncpy(hca_id_buf_p[i], hca_id_p[i], HCA_MAXNAME-1); 
    			}
		}
	}
	return W_RET;
}

/*****************************************
  struct i_VIPKL_alloc_ul_resources_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
  	IN MOSAL_protection_ctx_t prot_ctx;
    IN MT_size_t hca_ul_resources_sz;
  };
  struct o_VIPKL_alloc_ul_resources_ops_t {
    OUT VIP_ret_t ret;
    OUT EM_async_ctx_hndl_t async_hndl_ctx;
    OUT unsigned long hca_ul_resources[1];
  };
******************************************/
VIP_ret_t VIPKL_alloc_ul_resources(
     /*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/  MOSAL_protection_ctx_t  prot_ctx
  ,  /*IN*/  MT_size_t hca_ul_resources_sz
  ,  /*OUT*/ void *hca_ul_resources_p
  ,  /*OUT*/ EM_async_ctx_hndl_t *async_hndl_ctx_p)
{
	W_ALLOCA_IBUF(VIPKL_alloc_ul_resources);
	W_ALLOCA_OBUF_EX(VIPKL_alloc_ul_resources, hca_ul_resources, (u_int32_t)hca_ul_resources_sz);
	if ( hca_ul_resources_sz > 0 && hca_ul_resources_p == NULL) 
		return VIP_EINVAL_PARAM;
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(prot_ctx);
	W_PUT_VAR(hca_ul_resources_sz);
	W_ISSUE_IOCTL(VIPKL_ALLOC_UL_RESOURCES);
	if (W_RET == VIP_OK) {
      W_GET_BUF(hca_ul_resources,hca_ul_resources_sz);
      W_GET_VAR(async_hndl_ctx);
	}
	return W_RET;
}

/*****************************************
  struct i_VIPKL_free_ul_resources_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN MT_size_t hca_ul_resources_sz;
    IN EM_async_ctx_hndl_t async_hndl_ctx;
    IN u_int8_t  hca_ul_resources[1];
  };
  struct o_VIPKL_free_ul_resources_ops_t {
    OUT VIP_ret_t ret;
  };
******************************************/
VIP_ret_t VIPKL_free_ul_resources(/*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ MT_size_t hca_ul_resources_sz
  ,  /*IN*/ void *hca_ul_resources_p
  ,  /*IN*/ EM_async_ctx_hndl_t async_hndl_ctx)
{
	W_ALLOCA_IBUF_EX(VIPKL_free_ul_resources, hca_ul_resources, (u_int32_t)hca_ul_resources_sz);
	W_ALLOCA_OBUF(VIPKL_free_ul_resources);
	if ( hca_ul_resources_sz > 0 && hca_ul_resources_p == NULL) 
		return VIP_EINVAL_PARAM;
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(hca_ul_resources_sz);
	W_PUT_BUF(hca_ul_resources, hca_ul_resources_sz);
    W_PUT_VAR(async_hndl_ctx);
	W_ISSUE_IOCTL(VIPKL_FREE_UL_RESOURCES);
	return W_RET;
}

/*****************************************
  struct i_VIPKL_query_hca_cap_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN MOSAL_protection_ctx_t prot_ctx;
  };
  struct o_VIPKL_query_hca_cap_ops_t {
    OUT VIP_ret_t ret;
    OUT VAPI_hca_cap_t caps;
  };
******************************************/
VIP_ret_t VIPKL_query_hca_cap(/*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ MOSAL_protection_ctx_t prot_ctx
  ,  /*OUT*/ VAPI_hca_cap_t *caps_p)
{
	W_ALLOCA_IBUF(VIPKL_query_hca_cap);
	W_ALLOCA_OBUF(VIPKL_query_hca_cap);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(prot_ctx);
	W_ISSUE_IOCTL(VIPKL_QUERY_HCA_CAP);
	if 	(W_RET == VIP_OK) {
		W_GET_STRUCT(caps);
	}
	return W_RET;
}

/*****************************************
  struct i_VIPKL_query_port_prop_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN IB_port_t port_num;
  };
  struct o_VIPKL_query_port_prop_ops_t {
    OUT VIP_ret_t ret;
    OUT VAPI_hca_port_t port_props;
  };
******************************************/
VIP_ret_t VIPKL_query_port_prop(/*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ IB_port_t port_num
  ,  /*OUT*/ VAPI_hca_port_t *port_props_p)
{
	W_ALLOCA_IBUF(VIPKL_query_port_prop);
	W_ALLOCA_OBUF(VIPKL_query_port_prop);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(port_num);
	W_ISSUE_IOCTL(VIPKL_QUERY_HCA_PORT_PROP);
	if (W_RET == VIP_OK) {
		W_GET_STRUCT(port_props);
	}
	return W_RET;
}

/*****************************************
  struct i_VIPKL_query_port_pkey_tbl_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN IB_port_t port_num;
    IN u_int16_t tbl_len_in;
  };
  struct o_VIPKL_query_port_pkey_tbl_ops_t {
    OUT VIP_ret_t ret;
    OUT u_int16_t 	tbl_len_out;
    OUT VAPI_pkey_t pkey_tbl[1];
  };
******************************************/
VIP_ret_t VIPKL_query_port_pkey_tbl(/*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ IB_port_t port_num
  ,  /*IN*/ u_int16_t tbl_len_in
  ,  /*OUT*/ u_int16_t *tbl_len_out_p
  ,  /*OUT*/ VAPI_pkey_t *pkey_tbl_p)
{
	W_ALLOCA_IBUF(VIPKL_query_port_pkey_tbl);
	W_ALLOCA_OBUF_EX(VIPKL_query_port_pkey_tbl,pkey_tbl,sizeof(VAPI_pkey_t)* tbl_len_in);
  	if (tbl_len_in > 0 && pkey_tbl_p == NULL) {
	    	*tbl_len_out_p=0;
		return VIP_EINVAL_PARAM;
	}
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(port_num);
	W_PUT_VAR(tbl_len_in);
	W_ISSUE_IOCTL(VIPKL_QUERY_HCA_PKEY_TBL);
	if 	(W_RET == VIP_OK || W_RET == VIP_EAGAIN) {
		if (tbl_len_out_p) W_GET_VAR(tbl_len_out);
		if (pkey_tbl_p != NULL) {
    			int num_to_copy = MIN(tbl_len_in,*tbl_len_out_p);
    			memcpy( pkey_tbl_p, (void *)po->pkey_tbl, sizeof(VAPI_pkey_t )*num_to_copy );
		}
	}
	return W_RET;
}

/*****************************************
  struct i_VIPKL_query_port_gid_tbl_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN IB_port_t port_num;
    IN u_int16_t tbl_len_in;
  };
  struct o_VIPKL_query_port_gid_tbl_ops_t {
    OUT VIP_ret_t ret;
    OUT u_int16_t  tbl_len_out;
    OUT VAPI_gid_t gid_tbl[1];
  };
******************************************/
VIP_ret_t VIPKL_query_port_gid_tbl(/*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ IB_port_t port_num
  ,  /*IN*/ u_int16_t tbl_len_in
  ,  /*OUT*/ u_int16_t *tbl_len_out_p
  ,  /*OUT*/ VAPI_gid_t *gid_tbl_p)
{
	W_ALLOCA_IBUF(VIPKL_query_port_gid_tbl);
	W_ALLOCA_OBUF_EX(VIPKL_query_port_gid_tbl,gid_tbl,sizeof(VAPI_gid_t)* tbl_len_in);
  	if (tbl_len_in > 0 && gid_tbl_p == NULL) {
	    	*tbl_len_out_p=0;
		return VIP_EINVAL_PARAM;
	}
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(port_num);
	W_PUT_VAR(tbl_len_in);
	W_ISSUE_IOCTL(VIPKL_QUERY_HCA_GID_TBL);
	if 	(W_RET == VIP_OK || W_RET == VIP_EAGAIN) {
		if (tbl_len_out_p) W_GET_VAR(tbl_len_out);
		if (gid_tbl_p != NULL) {
    			int num_to_copy = MIN(tbl_len_in,*tbl_len_out_p);
    			memcpy( gid_tbl_p, (void *)po->gid_tbl, sizeof(VAPI_gid_t )*num_to_copy );
		}
	}
	return W_RET;
}

/*****************************************
  struct i_VIPKL_modify_hca_attr_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN IB_port_t port_num;
    IN VAPI_hca_attr_t 		hca_attr;
    IN VAPI_hca_attr_mask_t hca_attr_mask;
  };
  struct o_VIPKL_modify_hca_attr_ops_t {
    OUT VIP_ret_t ret;
  };
******************************************/
VIP_ret_t VIPKL_modify_hca_attr(/*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ IB_port_t port_num
  ,  /*IN*/ VAPI_hca_attr_t *hca_attr_p
  ,  /*IN*/ VAPI_hca_attr_mask_t *hca_attr_mask_p)
{
	W_ALLOCA_IBUF(VIPKL_modify_hca_attr);
	W_ALLOCA_OBUF(VIPKL_modify_hca_attr);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(port_num);
	W_PUT_STRUCT(hca_attr);
	W_PUT_STRUCT(hca_attr_mask);
	W_ISSUE_IOCTL(VIPKL_MODIFY_HCA_ATTR);
	return W_RET;
}

/*****************************************
  struct i_VIPKL_get_hh_hndl_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
  };
  struct o_VIPKL_get_hh_hndl_ops_t {
    OUT VIP_ret_t ret;
    OUT HH_hca_hndl_t hh;
  };
******************************************/
VIP_ret_t VIPKL_get_hh_hndl(/*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*OUT*/ HH_hca_hndl_t *hh_p)
{
	W_USE_VAR_AS_IBUF(hca_hndl, VIPKL_get_hh_hndl);
	W_ALLOCA_OBUF(VIPKL_get_hh_hndl);
	W_ISSUE_IOCTL(VIPKL_GET_HH_HNDL);
	if (W_RET == VIP_OK) {
		W_GET_VAR(hh);
	}
	return W_RET;
}
/*****************************************
  struct i_VIPKL_get_hca_ul_info_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
  };
  struct o_VIPKL_get_hca_ul_info_ops_t {
    OUT VIP_ret_t ret;
    OUT HH_hca_dev_t hca_ul_info;
  };
******************************************/
VIP_ret_t VIPKL_get_hca_ul_info(/*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*OUT*/ HH_hca_dev_t *hca_ul_info_p)
{
	W_USE_VAR_AS_IBUF(hca_hndl, VIPKL_get_hca_ul_info);
	W_ALLOCA_OBUF(VIPKL_get_hca_ul_info);
	W_ISSUE_IOCTL(VIPKL_GET_HCA_UL_INFO);
	if (W_RET == VIP_OK) {
		W_GET_STRUCT(hca_ul_info);
	}
	return W_RET;
}

/*****************************************
  struct i_VIPKL_get_hca_id_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
  };
  struct o_VIPKL_get_hca_id_ops_t {
    OUT VIP_ret_t ret;
    OUT VAPI_hca_id_t hca_id;
  };
******************************************/
VIP_ret_t VIPKL_get_hca_id(/*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*OUT*/ VAPI_hca_id_t *hca_id_p)
{
	W_USE_VAR_AS_IBUF(hca_hndl, VIPKL_get_hca_id);
	W_ALLOCA_OBUF(VIPKL_get_hca_id);
	W_ISSUE_IOCTL(VIPKL_GET_HCA_ID);
	if (W_RET == VIP_OK) {
		W_GET_STRUCT(hca_id);
	}
	return W_RET;
}

/*****************************************
  struct i_VIPKL_create_pd_ops_t {
  	IN VIP_hca_hndl_t hca_hndl;
	IN MOSAL_protection_ctx_t prot_ctx; 
  	IN MT_size_t pd_ul_resources_sz;
  	IN OUT u_int8_t pd_ul_resources[1]; 
  };
  struct o_VIPKL_create_pd_ops_t {
       OUT VIP_ret_t ret;
  	OUT PDM_pd_hndl_t 	pd_hndl; 
  	OUT	HH_pd_hndl_t 	pd_num;
  	IN OUT u_int8_t 	pd_ul_resources[1]; 
  };
******************************************/
VIP_ret_t VIPKL_create_pd(  
     /*IN*/ VIP_RSCT_t usr_ctx
     ,  /*IN*/ VIP_hca_hndl_t hca_hndl
     , /*IN*/MOSAL_protection_ctx_t prot_ctx
     , /*IN*/MT_size_t pd_ul_resources_sz
     , /*IN-OUT*/void * pd_ul_resources_p
     , /*OUT*/PDM_pd_hndl_t *pd_hndl_p
     ,  /*OUT*/HH_pd_hndl_t *pd_num_p)
{
	W_ALLOCA_IBUF_EX(VIPKL_create_pd, pd_ul_resources, (u_int32_t)pd_ul_resources_sz);
	W_ALLOCA_OBUF_EX(VIPKL_create_pd, pd_ul_resources, (u_int32_t)pd_ul_resources_sz);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(prot_ctx);
	W_PUT_VAR(pd_ul_resources_sz);
	W_PUT_BUF(pd_ul_resources, pd_ul_resources_sz);
    W_ISSUE_IOCTL(VIPKL_ALLOC_PD);
	if 	(W_RET == VIP_OK || W_RET == VIP_EAGAIN) {
		if (pd_hndl_p) W_GET_VAR(pd_hndl);
		if (pd_num_p) W_GET_VAR(pd_num);
		if (pd_ul_resources_p != NULL) W_GET_BUF(pd_ul_resources, pd_ul_resources_sz);
	}
	return W_RET;
}

/*****************************************
  struct i_VIPKL_destroy_pd_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN PDM_pd_hndl_t pd_hndl;
  };
  struct o_VIPKL_destroy_pd_ops_t {
    OUT VIP_ret_t ret;
  };
******************************************/
VIP_ret_t VIPKL_destroy_pd( 
    VIP_RSCT_t usr_ctx 
   , /*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ PDM_pd_hndl_t pd_hndl)
{
	W_ALLOCA_IBUF(VIPKL_destroy_pd);
	W_ALLOCA_OBUF(VIPKL_destroy_pd);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(pd_hndl);
	W_ISSUE_IOCTL(VIPKL_DEALLOC_PD);
	return W_RET;
}

/*****************************************
struct i_VIPKL_create_qp_ops_t {
  	IN VIP_hca_hndl_t hca_hndl;
  	VAPI_qp_hndl_t vapi_qp_hndl;
	IN QPM_qp_init_attr_t qp_init_attr;
    IN EM_async_ctx_hndl_t async_hndl_ctx;
  	IN MT_size_t qp_ul_resources_sz;
  	IN u_int8_t	qp_ul_resources[1];
};
struct o_VIPKL_create_qp_ops_t {
	OUT VIP_ret_t ret;
  	OUT QPM_qp_hndl_t qp_hndl;
  	OUT VAPI_qp_num_t qp_num;
  	IN OUT u_int8_t   qp_ul_resources[1]; 
};
******************************************/
VIP_ret_t VIPKL_create_qp(
    /*IN*/ VIP_RSCT_t usr_ctx
  , /*IN*/ VIP_hca_hndl_t hca_hndl
  , /*IN*/  VAPI_qp_hndl_t vapi_qp_hndl
  , /*IN*/ EM_async_ctx_hndl_t async_hndl_ctx
  , /*IN*/ MT_size_t          qp_ul_resources_sz
  ,  /*IN*/ void               *qp_ul_resources_p
  ,  /*IN*/ QPM_qp_init_attr_t *qp_init_attr_p
  ,  /*OUT*/ QPM_qp_hndl_t *qp_hndl_p
  ,  /*OUT*/ VAPI_qp_num_t *qp_num_p)
{
	W_ALLOCA_IBUF_EX(VIPKL_create_qp, qp_ul_resources, (u_int32_t)qp_ul_resources_sz);
	W_ALLOCA_OBUF_EX(VIPKL_create_qp, qp_ul_resources, (u_int32_t)qp_ul_resources_sz);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(vapi_qp_hndl);
  W_PUT_VAR(async_hndl_ctx);
	W_PUT_VAR(qp_ul_resources_sz);
	W_PUT_STRUCT(qp_init_attr);
	W_PUT_BUF(qp_ul_resources, qp_ul_resources_sz);
	W_ISSUE_IOCTL(VIPKL_CREATE_QP);
	if 	(W_RET == VIP_OK) {
		if (qp_hndl_p) W_GET_VAR(qp_hndl);
		if (qp_num_p) W_GET_VAR(qp_num);
		if (qp_ul_resources_p != NULL) W_GET_BUF(qp_ul_resources, qp_ul_resources_sz);
	}
	return W_RET;
}

/*****************************************
  struct i_VIPKL_destroy_qp_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN QPM_qp_hndl_t qp_hndl;
  };
  struct o_VIPKL_destroy_qp_ops_t {
	OUT VIP_ret_t ret;
  };
******************************************/
VIP_ret_t VIPKL_destroy_qp( VIP_RSCT_t usr_ctx,
     /*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ QPM_qp_hndl_t qp_hndl
  ,  /*IN*/MT_bool in_rsct_cleanup)
{
	W_ALLOCA_IBUF(VIPKL_destroy_qp);
	W_ALLOCA_OBUF(VIPKL_destroy_qp);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(qp_hndl);
	W_ISSUE_IOCTL(VIPKL_DESTROY_QP);
	return W_RET;
}

/*****************************************
  struct i_VIPKL_modify_qp_ops_t {
    IN VIP_hca_hndl_t 		hca_hndl;
    IN QPM_qp_hndl_t 		qp_hndl;
    IN VAPI_qp_attr_mask_t 	qp_mod_mask;
    IN VAPI_qp_attr_t 		qp_mod_attr;
  };
  struct o_VIPKL_modify_qp_ops_t {
	OUT VIP_ret_t ret;
  };
******************************************/
VIP_ret_t VIPKL_modify_qp(VIP_RSCT_t usr_ctx,  /*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ QPM_qp_hndl_t qp_hndl
  ,  /*IN*/ VAPI_qp_attr_mask_t *qp_mod_mask_p
  ,  /*IN*/ VAPI_qp_attr_t *qp_mod_attr_p)
{
	W_ALLOCA_IBUF(VIPKL_modify_qp);
	W_ALLOCA_OBUF(VIPKL_modify_qp);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(qp_hndl);
	W_PUT_STRUCT(qp_mod_mask);
	W_PUT_STRUCT(qp_mod_attr);
	W_ISSUE_IOCTL(VIPKL_MODIFY_QP);
	return W_RET;
}

/*****************************************
  struct i_VIPKL_query_qp_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN QPM_qp_hndl_t qp_hndl;
  };
  struct o_VIPKL_query_qp_ops_t {
    OUT VIP_ret_t ret;
    OUT QPM_qp_query_attr_t qp_query_prop;
    OUT VAPI_qp_attr_mask_t qp_mod_mask;
  };
******************************************/
VIP_ret_t VIPKL_query_qp(VIP_RSCT_t usr_ctx,  /*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ QPM_qp_hndl_t qp_hndl
  ,  /*OUT*/ QPM_qp_query_attr_t *qp_query_prop_p
  ,  /*OUT*/ VAPI_qp_attr_mask_t *qp_mod_mask_p)
{
	W_ALLOCA_IBUF(VIPKL_query_qp);
	W_ALLOCA_OBUF(VIPKL_query_qp);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(qp_hndl);
	W_ISSUE_IOCTL(VIPKL_QUERY_QP);
	if (W_RET == VIP_OK) {
		W_GET_STRUCT(qp_query_prop);
		W_GET_VAR(qp_mod_mask);
	}
	return W_RET;
}

/*****************************************
struct i_VIPKL_create_srq_ops_t {
                  VIP_hca_hndl_t hca_hndl;
                  VAPI_srq_hndl_t   vapi_srq_hndl;
                  PDM_pd_hndl_t     pd_hndl;
                  EM_async_ctx_hndl_t async_hndl_ctx;
                  MT_size_t srq_ul_resources_sz;
                  MT_ulong_ptr_t srq_ul_resources[1];
};


struct o_VIPKL_create_srq_ops_t {
                  VIP_ret_t ret;
                  SRQM_srq_hndl_t srq_hndl;
                  MT_ulong_ptr_t srq_ul_resources[1]; 
              };
******************************************/
VIP_ret_t
VIPKL_create_srq (VIP_RSCT_t usr_ctx,
                 VIP_hca_hndl_t hca_hndl,
                  VAPI_srq_hndl_t   vapi_srq_hndl, 
                  PDM_pd_hndl_t     pd_hndl,
                  EM_async_ctx_hndl_t async_hndl_ctx,
                  MT_size_t           srq_ul_resources_sz,
                  void                *srq_ul_resources_p,
                  SRQM_srq_hndl_t     *srq_hndl_p )
{
	W_ALLOCA_IBUF_EX(VIPKL_create_srq, srq_ul_resources, srq_ul_resources_sz);
	W_ALLOCA_OBUF_EX(VIPKL_create_srq, srq_ul_resources, srq_ul_resources_sz);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(vapi_srq_hndl);
  W_PUT_VAR(pd_hndl);
  W_PUT_VAR(async_hndl_ctx);
	W_PUT_VAR(srq_ul_resources_sz);
	W_PUT_BUF(srq_ul_resources, srq_ul_resources_sz);
	W_ISSUE_IOCTL(VIPKL_CREATE_SRQ);
	if 	(W_RET == VIP_OK) {
		if (srq_hndl_p) W_GET_VAR(srq_hndl);
		if (srq_ul_resources_p != NULL) W_GET_BUF(srq_ul_resources, srq_ul_resources_sz);
	}
	return W_RET;
}


/*****************************************
struct i_VIPKL_destroy_srq_ops_t {
                  VIP_hca_hndl_t hca_hndl;
                  SRQM_srq_hndl_t srq_hndl;
};

struct o_VIPKL_destroy_srq_ops_t {
  VIP_ret_t ret;
};
******************************************/
VIP_ret_t
VIPKL_destroy_srq(VIP_RSCT_t usr_ctx,
                  VIP_hca_hndl_t hca_hndl,
                  SRQM_srq_hndl_t srq_hndl)
{
	W_ALLOCA_IBUF(VIPKL_destroy_srq);
	W_ALLOCA_OBUF(VIPKL_destroy_srq);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(srq_hndl);
	W_ISSUE_IOCTL(VIPKL_DESTROY_SRQ);
	return W_RET;
}


/*****************************************
struct i_VIPKL_query_srq_ops_t {
                VIP_hca_hndl_t hca_hndl;
                SRQM_srq_hndl_t srq_hndl;
}
                
struct o_VIPKL_query_srq_ops_t {
                VIP_ret_t ret;
                u_int32_t limit;
};
******************************************/
VIP_ret_t
VIPKL_query_srq(VIP_RSCT_t usr_ctx,
                VIP_hca_hndl_t hca_hndl,
                SRQM_srq_hndl_t srq_hndl,
                u_int32_t *limit_p)
{
	W_ALLOCA_IBUF(VIPKL_query_srq);
	W_ALLOCA_OBUF(VIPKL_query_srq);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(srq_hndl);
	W_ISSUE_IOCTL(VIPKL_QUERY_SRQ);
	if 	(W_RET == VIP_OK) {
		if (limit_p) W_GET_VAR(limit);
	}
	return W_RET;
}

/******************************************************************************
struct i_VIPKL_modify_srq_ops_t {
                VIP_hca_hndl_t hca_hndl;
                SRQM_srq_hndl_t srq_hndl;
                MT_size_t srq_ul_resources_sz;
                MT_ulong_ptr_t srq_ul_resources[1];
};
                
struct o_VIPKL_modify_srq_ops_t {
                VIP_ret_t ret;
                MT_ulong_ptr_t srq_ul_resources[1]; 
};
 *****************************************************************************/
VIP_ret_t
VIPKL_modify_srq(VIP_RSCT_t     usr_ctx,
                VIP_hca_hndl_t  hca_hndl,
                SRQM_srq_hndl_t srq_hndl,
                MT_size_t       srq_ul_resources_sz,
                void            *srq_ul_resources_p)
{
	W_ALLOCA_IBUF_EX(VIPKL_modify_srq, srq_ul_resources, srq_ul_resources_sz);
	W_ALLOCA_OBUF_EX(VIPKL_modify_srq, srq_ul_resources, srq_ul_resources_sz);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(srq_hndl);
	W_PUT_VAR(srq_ul_resources_sz);
	W_PUT_BUF(srq_ul_resources, srq_ul_resources_sz);
	W_ISSUE_IOCTL(VIPKL_MODIFY_SRQ);
	if 	(W_RET == VIP_OK) {
		if (srq_ul_resources_p != NULL) W_GET_BUF(srq_ul_resources, srq_ul_resources_sz);
	}
	return W_RET;
}

/*****************************************
  struct i_VIPKL_attach_to_multicast_ops_t {
     IN	VAPI_hca_hndl_t     hca_hndl;
     IN IB_gid_t            mcg_dgid;
     IN VAPI_qp_hndl_t      qp_hndl;
  };
  struct o_VIPKL_attach_to_multicast_ops_t {
	OUT VIP_ret_t ret;
  };
******************************************/
VIP_ret_t VIPKL_attach_to_multicast(VIP_RSCT_t usr_ctx,
     /*IN*/      VAPI_hca_hndl_t     hca_hndl,
     /*IN*/      IB_gid_t            mcg_dgid_p,
     /*IN*/      VAPI_qp_hndl_t      qp_hndl)
{
	W_ALLOCA_IBUF(VIPKL_attach_to_multicast);
	W_ALLOCA_OBUF(VIPKL_attach_to_multicast);
	W_PUT_VAR(hca_hndl);
	W_PUT_STRUCT(mcg_dgid);
	W_PUT_VAR(qp_hndl);
	W_ISSUE_IOCTL(VIPKL_ATTACH_TO_MULTICAST);
	return W_RET;
}
 
/*****************************************
  struct i_VIPKL_detach_from_multicast_ops_t 
     IN      VAPI_hca_hndl_t     hca_hndl;
     IN      IB_gid_t            mcg_dgid;
     IN      VAPI_qp_hndl_t      qp_hndl;
  };
  struct o_VIPKL_detach_from_multicast_ops_t {
	OUT VIP_ret_t ret;
  };
******************************************/
VIP_ret_t VIPKL_detach_from_multicast(VIP_RSCT_t usr_ctx,
     /*IN*/      VAPI_hca_hndl_t     hca_hndl,
     /*IN*/      IB_gid_t            mcg_dgid_p,
     /*IN*/      VAPI_qp_hndl_t      qp_hndl)
{
	W_ALLOCA_IBUF(VIPKL_detach_from_multicast);
	W_ALLOCA_OBUF(VIPKL_detach_from_multicast);
	W_PUT_VAR(hca_hndl);
	W_PUT_STRUCT(mcg_dgid);
	W_PUT_VAR(qp_hndl);
	W_ISSUE_IOCTL(VIPKL_DETACH_FROM_MULTICAST);
	return W_RET;
}

/*****************************************
struct i_VIPKL_create_cq_ops_t {
  IN VIP_hca_hndl_t hca_hndl;
  IN VAPI_cq_hndl_t vapi_cq_hndl;
  IN MOSAL_protection_ctx_t  usr_prot_ctx;
  IN MT_size_t cq_ul_resources_sz;
  IN EM_async_ctx_hndl_t async_hndl_ctx,
  IN OUT u_int8_t 	cq_ul_resources[1];
};
struct o_VIPKL_create_cq_ops_t {
  OUT VIP_ret_t ret;
  OUT CQM_cq_hndl_t cq;
  OUT HH_cq_hndl_t	cq_id;
  IN OUT u_int8_t 	cq_ul_resources[1];
};
******************************************/
VIP_ret_t VIPKL_create_cq(VIP_RSCT_t usr_ctx,
    /*IN*/ VIP_hca_hndl_t hca_hndl,
    /*IN*/ VAPI_cq_hndl_t vapi_cq_hndl,
    /*IN*/ MOSAL_protection_ctx_t  usr_prot_ctx,
    /*IN*/ EM_async_ctx_hndl_t async_hndl_ctx,
    /*IN*/ MT_size_t cq_ul_resources_sz,
    /*IN*/ void  *cq_ul_resources_p,
    /*OUT*/ CQM_cq_hndl_t *cq_p,
    /*OUT*/ HH_cq_hndl_t* cq_id_p
    )
{
	W_ALLOCA_IBUF_EX(VIPKL_create_cq, cq_ul_resources, (u_int32_t)cq_ul_resources_sz);
	W_ALLOCA_OBUF_EX(VIPKL_create_cq, cq_ul_resources, (u_int32_t)cq_ul_resources_sz);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(vapi_cq_hndl);
	W_PUT_VAR(usr_prot_ctx);
    W_PUT_VAR(async_hndl_ctx);
	W_PUT_VAR(cq_ul_resources_sz);
	W_PUT_BUF(cq_ul_resources, cq_ul_resources_sz);
	W_ISSUE_IOCTL(VIPKL_CREATE_CQ);
	if 	(W_RET == VIP_OK) {
		if (cq_p) W_GET_VAR(cq);
		if (cq_id_p) W_GET_VAR(cq_id);
		if (cq_ul_resources_p != NULL) W_GET_BUF(cq_ul_resources, cq_ul_resources_sz);
	}
	return W_RET;
}

/*****************************************
  struct i_VIPKL_destroy_cq_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN CQM_cq_hndl_t cq;
  };
  struct o_VIPKL_destroy_cq_ops_t {
    OUT VIP_ret_t ret;
  };
******************************************/
VIP_ret_t VIPKL_destroy_cq(VIP_RSCT_t usr_ctx,  /*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ CQM_cq_hndl_t cq, /*IN*/MT_bool in_rsct_cleanup)
{
	W_ALLOCA_IBUF(VIPKL_destroy_cq);
	W_ALLOCA_OBUF(VIPKL_destroy_cq);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(cq);
	W_ISSUE_IOCTL(VIPKL_DESTROY_CQ);
	return W_RET;
}

/*****************************************
  struct i_VIPKL_get_cq_props_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN CQM_cq_hndl_t cq;
  };
  struct o_VIPKL_get_cq_props_ops_t {
    OUT VIP_ret_t ret;
    OUT HH_cq_hndl_t	cq_id;
    OUT VAPI_cqe_num_t	num_o_entries;
  };
******************************************/
VIP_ret_t VIPKL_get_cq_props(VIP_RSCT_t usr_ctx,  /*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ CQM_cq_hndl_t cq
  ,  /*OUT*/ HH_cq_hndl_t* cq_id_p
  ,  /*OUT*/ VAPI_cqe_num_t *num_o_entries_p)
{
	W_ALLOCA_IBUF(VIPKL_get_cq_props);
	W_ALLOCA_OBUF(VIPKL_get_cq_props);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(cq);
	W_ISSUE_IOCTL(VIPKL_QUERY_CQ);
	if 	(W_RET == VIP_OK) {
		if (cq_id_p) W_GET_VAR(cq_id);
		if (num_o_entries_p) W_GET_VAR(num_o_entries);
	}
	return W_RET;
}


/*****************************************
  struct i_VIPKL_resize_cq_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN CQM_cq_hndl_t cq;
    IN MT_size_t cq_ul_resources_sz;
    IN OUT unsigned long cq_ul_resources[1];
  };
  struct o_VIPKL_resize_cq_ops_t {
    OUT VIP_ret_t ret;
    IN OUT unsigned long cq_ul_resources[1];
  };
******************************************/
VIP_ret_t VIPKL_resize_cq( VIP_RSCT_t usr_ctx, /*IN*/ VIP_hca_hndl_t hca_hndl,
    /*IN*/ CQM_cq_hndl_t cq,
    /*IN*/ MT_size_t cq_ul_resources_sz,
    /*IN/OUT*/ void *cq_ul_resources_p)
{
	W_ALLOCA_IBUF_EX(VIPKL_resize_cq,cq_ul_resources,(u_int32_t)cq_ul_resources_sz);
	W_ALLOCA_OBUF_EX(VIPKL_resize_cq,cq_ul_resources,(u_int32_t)cq_ul_resources_sz);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(cq);
	W_PUT_VAR(cq_ul_resources_sz);
	W_PUT_BUF(cq_ul_resources, cq_ul_resources_sz);
	W_ISSUE_IOCTL(VIPKL_RESIZE_CQ);
	if 	(W_RET == VIP_OK) {
		W_GET_BUF(cq_ul_resources, cq_ul_resources_sz);
	}
	return W_RET;
}

/*****************************************
  struct i_VIPKL_create_mr_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN VAPI_mrw_t mrw_req;
    IN PDM_pd_hndl_t pd_hndl;
    IN VAPI_phy_buf_t pbuf_list[1];
  };
  struct o_VIPKL_create_mr_ops_t {
    OUT VIP_ret_t ret;
    OUT MM_mrw_hndl_t mrw_hndl;
    OUT MM_VAPI_mro_t mr_prop;
  };
******************************************/
VIP_ret_t VIPKL_create_mr(VIP_RSCT_t usr_ctx,  
    /*IN*/ VIP_hca_hndl_t hca_hndl,
    /*IN*/ VAPI_mrw_t* mrw_req_p,
  /*IN*/ PDM_pd_hndl_t pd_hndl,
   /*OUT*/ MM_mrw_hndl_t *mrw_hndl_p,
   /*OUT*/ MM_VAPI_mro_t* mr_prop_p)
{
  u_int32_t pbuf_list_sz = (u_int32_t)((mrw_req_p->type == VAPI_MPR) ? (mrw_req_p->pbuf_list_len * sizeof(VAPI_phy_buf_t)) : 0);
	W_ALLOCA_IBUF_EX(VIPKL_create_mr,pbuf_list,pbuf_list_sz);
	W_ALLOCA_OBUF(VIPKL_create_mr);
	W_PUT_VAR(hca_hndl);
	W_PUT_STRUCT(mrw_req);
	W_PUT_VAR(pd_hndl);
	if (mrw_req_p->pbuf_list_len > 0) 
	   	memcpy( (void*)pi->pbuf_list, (void*)mrw_req_p->pbuf_list_p, pbuf_list_sz );
	W_ISSUE_IOCTL(VIPKL_CREATE_MR);
	if 	(W_RET == VIP_OK) {
		if (mrw_hndl_p) W_GET_VAR(mrw_hndl);
		if (mr_prop_p) W_GET_STRUCT(mr_prop);
	}
	return W_RET;
}


VIP_ret_t VIPKL_is_iomem(VIP_RSCT_t usr_ctx,
						 VIP_hca_hndl_t hca_hndl,
						 MOSAL_protection_ctx_t prot_ctx,
						 MT_virt_addr_t va,
						 MT_size_t size, int *iomem_p)
{
	W_ALLOCA_IBUF(VIPKL_is_iomem);
	W_ALLOCA_OBUF(VIPKL_is_iomem);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(prot_ctx);
	W_PUT_VAR(va);
	W_PUT_VAR(size);
	W_ISSUE_IOCTL(VIPKL_IS_IOMEM);
	if 	(W_RET == VIP_OK) {
		W_GET_VAR(iomem);
	}
	return W_RET;
}

/*****************************************
  struct i_VIPKL_create_smr_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN MM_mrw_hndl_t     orig_mr_hndl;
    IN VAPI_mrw_t     mrw_req;
    IN PDM_pd_hndl_t pd_hndl;
  };
  struct o_VIPKL_create_smr_ops_t {
    OUT VIP_ret_t ret;
    OUT MM_mrw_hndl_t mrw_hndl;
    OUT MM_VAPI_mro_t mr_prop;
  };
******************************************/

VIP_ret_t VIPKL_create_smr(  
    VIP_RSCT_t usr_ctx,
    /*IN*/  VIP_hca_hndl_t      hca_hndl,
    /*IN*/  MM_mrw_hndl_t       orig_mr_hndl,
    /*IN*/  VAPI_mr_t           *mrw_req_p,
    /*IN*/  PDM_pd_hndl_t       pd_hndl,
    /*OUT*/ MM_mrw_hndl_t      *mrw_hndl_p,
    /*OUT*/ MM_VAPI_mro_t      *mr_prop_p)
{
	W_ALLOCA_IBUF(VIPKL_create_smr);
	W_ALLOCA_OBUF(VIPKL_create_smr);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(orig_mr_hndl);
	W_PUT_STRUCT(mrw_req);
	W_PUT_VAR(pd_hndl);
	W_ISSUE_IOCTL(VIPKL_CREATE_SMR);
	if 	(W_RET == VIP_OK) {
		if (mrw_hndl_p) W_GET_VAR(mrw_hndl);
		if (mr_prop_p) W_GET_STRUCT(mr_prop);
	}
	return W_RET;
}


/*****************************************
  struct i_VIPKL_create_mw_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN PDM_pd_hndl_t pd_hndl;
  };
  struct o_VIPKL_create_mw_ops_t {
    OUT VIP_ret_t ret;
    OUT MM_key_t		r_key;
  };
******************************************/
VIP_ret_t VIPKL_create_mw( 
    VIP_RSCT_t usr_ctx, 
    /*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ PDM_pd_hndl_t pd_hndl
  ,  /*OUT*/ MM_key_t* r_key_p)
{
	W_ALLOCA_IBUF(VIPKL_create_mw);
	W_ALLOCA_OBUF(VIPKL_create_mw);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(pd_hndl);
	W_ISSUE_IOCTL(VIPKL_CREATE_MW);
	if 	(W_RET == VIP_OK) {
		if (r_key_p) W_GET_VAR(r_key);
	}
	return W_RET;
}

/*****************************************
struct i_VIPKL_reregister_mr_ops_t {
    IN  VAPI_hca_hndl_t       hca_hndl;
    IN  VAPI_mr_hndl_t        mr_hndl;
    IN  VAPI_mr_change_t      change_type;
    IN  VAPI_mrw_t           mrw_req;
    IN PDM_pd_hndl_t       pd_hndl;
    IN VAPI_phy_buf_t pbuf_list[1];
};
struct o_VIPKL_reregister_mr_ops_t {
    OUT VIP_ret_t ret;
    OUT  MM_mrw_hndl_t       rep_mr_hndl;
    OUT MM_VAPI_mro_t        rep_mrw;
};
******************************************/
VIP_ret_t VIPKL_reregister_mr(        VIP_RSCT_t usr_ctx,
                               VIP_hca_hndl_t hca_hndl,
                              VAPI_mr_hndl_t        mr_hndl,
                              VAPI_mr_change_t      change_type,
                              VAPI_mrw_t*           mrw_req_p,
                              PDM_pd_hndl_t pd_hndl,
                              MM_mrw_hndl_t *rep_mr_hndl_p,
                              MM_VAPI_mro_t* rep_mrw_p)
{
  	u_int32_t pbuf_list_sz = (u_int32_t)((mrw_req_p->type == VAPI_MPR) ? (mrw_req_p->pbuf_list_len * sizeof(VAPI_phy_buf_t)) : 0);
	W_ALLOCA_IBUF_EX(VIPKL_reregister_mr,pbuf_list,pbuf_list_sz);
	W_ALLOCA_OBUF(VIPKL_reregister_mr);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(mr_hndl);
	W_PUT_VAR(change_type);
	W_PUT_STRUCT(mrw_req);
	W_PUT_VAR(pd_hndl);
	if (mrw_req_p->pbuf_list_len > 0) 
	   	memcpy( (void*)pi->pbuf_list, (void*)mrw_req_p->pbuf_list_p, pbuf_list_sz );
	W_ISSUE_IOCTL(VIPKL_REREGISTER_MR);
	if 	(W_RET == VIP_OK) {
		if (rep_mr_hndl_p) W_GET_VAR(rep_mr_hndl);
		if (rep_mrw_p) W_GET_STRUCT(rep_mrw);
	}
	return W_RET;
}


#if 0
/* FMRs are allowed in kernel only */

/*****************************************
struct i_VIPKL_alloc_fmr_ops_t {
    IN  VIP_hca_hndl_t      hca_hndl;
    IN  EVAPI_fmr_t         fmr_props;
    IN  PDM_pd_hndl_t       pd_hndl;
};
struct o_VIPKL_alloc_fmr_ops_t {
    OUT VIP_ret_t ret;
    OUT MM_mrw_hndl_t      fmr_hndl;
};
******************************************/
VIP_ret_t VIPKL_alloc_fmr(VIP_RSCT_t usr_ctx,VIP_hca_hndl_t hca_hndl,EVAPI_fmr_t* fmr_props_p,PDM_pd_hndl_t pd_hndl,
                          MM_mrw_hndl_t* fmr_hndl_p)
{
	W_ALLOCA_IBUF(VIPKL_alloc_fmr);
	W_ALLOCA_OBUF(VIPKL_alloc_fmr);
	W_PUT_VAR(hca_hndl);
	W_PUT_STRUCT(fmr_props);
	W_PUT_VAR(pd_hndl);
	W_ISSUE_IOCTL(VIPKL_ALLOC_FMR);
	if 	(W_RET == VIP_OK) {
		if (fmr_hndl_p) W_GET_VAR(fmr_hndl);
	}
	return W_RET;
}


/*****************************************
struct i_VIPKL_map_fmr_ops_t {
    IN   VAPI_hca_hndl_t      hca_hndl;
    IN   EVAPI_fmr_hndl_t       fmr_hndl;
    IN   EVAPI_fmr_map_t      map;
    IN   VAPI_phy_addr_t	page_array[1];
};
struct o_VIPKL_map_fmr_ops_t {
    OUT VIP_ret_t ret;
    OUT  VAPI_lkey_t          l_key; 
    OUT  VAPI_rkey_t          r_key;
};
******************************************/
VIP_ret_t VIPKL_map_fmr(VIP_RSCT_t usr_ctx, 
                        VAPI_hca_hndl_t      hca_hndl,
                         EVAPI_fmr_hndl_t     fmr_hndl,
                         EVAPI_fmr_map_t      *map_p,
                         VAPI_lkey_t*          l_key_p,
                         VAPI_rkey_t*          r_key_p)

{
  	u_int32_t pg_array_sz = (u_int32_t)(map_p->page_array_len * sizeof(VAPI_phy_addr_t));
	W_ALLOCA_IBUF_EX(VIPKL_map_fmr,page_array,pg_array_sz);
	W_ALLOCA_OBUF(VIPKL_map_fmr);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(fmr_hndl);
	W_PUT_STRUCT(map);
	if (pg_array_sz > 0) 
	   	memcpy( (void*)pi->page_array, (void*)map_p->page_array, pg_array_sz );
	W_ISSUE_IOCTL(VIPKL_MAP_FMR);
	if 	(W_RET == VIP_OK) {
		if (l_key_p) W_GET_VAR(l_key);
		if (r_key_p) W_GET_VAR(r_key);
	}
	return W_RET;
}


/*****************************************
struct i_VIPKL_unmap_fmr_ops_t {
    IN   VAPI_hca_hndl_t      hca_hndl;
    IN   MT_size_t            num_of_fmr_to_unmap;
    IN   EVAPI_fmr_hndl_t       fmr_hndl_array[1];
};
struct o_VIPKL_unmap_fmr_ops_t {
    OUT VIP_ret_t ret;
};
******************************************/
VIP_ret_t VIPKL_unmap_fmr(VIP_RSCT_t usr_ctx,
                          VAPI_hca_hndl_t      hca_hndl,
                          MT_size_t            num_of_fmr_to_unmap,
                          EVAPI_fmr_hndl_t*    fmr_hndl_array)
{
  	u_int32_t fmr_hndl_array_sz = (u_int32_t)(num_of_fmr_to_unmap * sizeof(EVAPI_fmr_hndl_t));
	W_ALLOCA_IBUF_EX(VIPKL_unmap_fmr,fmr_hndl_array,fmr_hndl_array_sz);
	W_ALLOCA_OBUF(VIPKL_unmap_fmr);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(num_of_fmr_to_unmap);
	if (fmr_hndl_array_sz > 0) 
	   	memcpy( (void*)pi->fmr_hndl_array, (void*)fmr_hndl_array, fmr_hndl_array_sz );
	W_ISSUE_IOCTL(VIPKL_UNMAP_FMR);
	return W_RET;
}

/*****************************************
struct i_VIPKL_free_fmr_ops_t {
    IN   VAPI_hca_hndl_t      hca_hndl;
    IN   EVAPI_fmr_hndl_t    fmr_hndl;
};
struct o_VIPKL_free_fmr_ops_t {
    OUT VIP_ret_t ret;
};
******************************************/
VIP_ret_t VIPKL_free_fmr(VIP_RSCT_t usr_ctx,
                         VAPI_hca_hndl_t      hca_hndl,
                         EVAPI_fmr_hndl_t       fmr_hndl)
{
	W_ALLOCA_IBUF(VIPKL_free_fmr);
	W_ALLOCA_OBUF(VIPKL_free_fmr);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(fmr_hndl);
	W_ISSUE_IOCTL(VIPKL_FREE_FMR);
	return W_RET;
}

#endif /*FMRs*/

/*****************************************
  struct i_VIPKL_destroy_mr_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN MM_mrw_hndl_t mrw_hndl;
  };
struct o_VIPKL_destroy_mr_ops_t {
    OUT VIP_ret_t ret;
};
******************************************/
VIP_ret_t VIPKL_destroy_mr( VIP_RSCT_t usr_ctx, /*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ MM_mrw_hndl_t mrw_hndl)
{
	W_ALLOCA_IBUF(VIPKL_destroy_mr);
	W_ALLOCA_OBUF(VIPKL_destroy_mr);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(mrw_hndl);
	W_ISSUE_IOCTL(VIPKL_DESTROY_MR);
	return W_RET;
}

/*****************************************
  struct i_VIPKL_destroy_mw_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN MM_mrw_hndl_t mrw_hndl;
  };
struct o_VIPKL_destroy_mw_ops_t {
    OUT VIP_ret_t ret;
};
******************************************/
VIP_ret_t VIPKL_destroy_mw(VIP_RSCT_t usr_ctx,  /*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ MM_mrw_hndl_t mrw_hndl)
{
	W_ALLOCA_IBUF(VIPKL_destroy_mw);
	W_ALLOCA_OBUF(VIPKL_destroy_mw);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(mrw_hndl);
	W_ISSUE_IOCTL(VIPKL_DESTROY_MW);
	return W_RET;
}

/*****************************************
  struct i_VIPKL_query_mr_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN MM_mrw_hndl_t mrw_hndl;
  };
  struct o_VIPKL_query_mr_ops_t {
    OUT VIP_ret_t ret;
    OUT MM_VAPI_mro_t mr_prop;
  };
******************************************/
VIP_ret_t VIPKL_query_mr(VIP_RSCT_t usr_ctx,  /*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ MM_mrw_hndl_t mrw_hndl
  ,  /*OUT*/ MM_VAPI_mro_t *mr_prop_p)
{
	W_ALLOCA_IBUF(VIPKL_query_mr);
	W_ALLOCA_OBUF(VIPKL_query_mr);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(mrw_hndl);
	W_ISSUE_IOCTL(VIPKL_QUERY_MR);
	if 	(W_RET == VIP_OK) {
		if (mr_prop_p) W_GET_STRUCT(mr_prop);
	}
	return W_RET;
}

/*****************************************
 struct i_VIPKL_query_mw_ops_t {
    IN VIP_hca_hndl_t hca_hndl;
    IN IB_rkey_t initial_key;
  };
  struct o_VIPKL_query_mw_ops_t {
    OUT VIP_ret_t ret;
    OUT VAPI_rkey_t current_key;
  };
******************************************/
VIP_ret_t VIPKL_query_mw(   VIP_RSCT_t usr_ctx,
                            /*IN*/ VIP_hca_hndl_t hca_hndl,
                           /*IN*/ IB_rkey_t initial_key,
                           /*OUT*/ VAPI_rkey_t* current_key_p)
{
	W_ALLOCA_IBUF(VIPKL_query_mw);
	W_ALLOCA_OBUF(VIPKL_query_mw);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(initial_key);
	MTL_TRACE1("VIPKL_query_mw: -UL- called for key 0x%x",initial_key);
	W_ISSUE_IOCTL(VIPKL_QUERY_MW);
	if 	(W_RET == VIP_OK) {
		if (current_key_p) W_GET_VAR(current_key);
	}
	return W_RET;
}

/************** Callbacks wrapping to user space ***************************/

 
/******** VIPKL_EQ functions ********/

/***************************************************
struct i_VIPKL_EQ_new_ops_t {
  IN VAPI_hca_hndl_t hca_hndl;
  IN VIP_EQ_cbk_type_t cbk_type,
  IN EM_async_ctx_hndl_t async_ctx
};
struct o_VIPKL_EQ_new_ops_t {
  OUT VIP_ret_t ret;
  OUT VIPKL_EQ_hndl_t vipkl_eq_h;
};
****************************************************/
VIP_ret_t VIPKL_EQ_new(VIP_RSCT_t usr_ctx,VAPI_hca_hndl_t hca_hndl,
                       VIPKL_EQ_cbk_type_t cbk_type,EM_async_ctx_hndl_t async_ctx, 
                       VIPKL_EQ_hndl_t *vipkl_eq_h_p)
{
	W_ALLOCA_IBUF(VIPKL_EQ_new);
	W_ALLOCA_OBUF(VIPKL_EQ_new);
  W_PUT_VAR(hca_hndl);
  W_PUT_VAR(cbk_type);
  W_PUT_VAR(async_ctx);
	W_ISSUE_IOCTL(VIPKL_EQ_NEW);
	if (W_RET == VIP_OK) {
		if (vipkl_eq_h_p) W_GET_VAR(vipkl_eq_h);
	}
	return W_RET;
}

/***************************************************
struct i_VIPKL_EQ_del_proc_ops_t {
  IN  VIPKL_EQ_hndl_t vipkl_eq;
};
struct o_VIPKL_EQ_del_proc_ops_t {
  OUT VIP_ret_t ret;
};
****************************************************/
VIP_ret_t VIPKL_EQ_del(VIP_RSCT_t usr_ctx,VAPI_hca_hndl_t hca_hndl,VIPKL_EQ_hndl_t vipkl_eq)
{
	W_ALLOCA_IBUF(VIPKL_EQ_del);
  W_ALLOCA_OBUF(VIPKL_EQ_del);
  W_PUT_VAR(hca_hndl);
  W_PUT_VAR(vipkl_eq);
  W_ISSUE_IOCTL(VIPKL_EQ_DEL);
	return W_RET;
}

/***************************************************
struct i_VIPKL_EQ_evapi_set_comp_eventh_ops_t {
  IN  VIPKL_EQ_hndl_t                  vipkl_eq;
  IN  CQM_cq_hndl_t                    vipkl_cq;
  IN  VAPI_completion_event_handler_t  completion_handler;
  IN  void *                           private_data;
};
struct o_VIPKL_EQ_evapi_set_comp_eventh_ops_t {
  OUT VIP_ret_t ret;
};
****************************************************/
VIP_ret_t VIPKL_EQ_evapi_set_comp_eventh(
    VIP_RSCT_t usr_ctx,
    VIP_hca_hndl_t						hca_hndl,
   /*IN*/VIPKL_EQ_hndl_t                 vipkl_eq,
  /*IN*/CQM_cq_hndl_t                    vipkl_cq,
  /*IN*/VAPI_completion_event_handler_t  completion_handler,
  /*IN*/void *                           private_data)
{
	W_ALLOCA_IBUF(VIPKL_EQ_evapi_set_comp_eventh);
	W_ALLOCA_OBUF(VIPKL_EQ_evapi_set_comp_eventh);
    W_PUT_VAR(hca_hndl);
	W_PUT_VAR(vipkl_eq);
	W_PUT_VAR(vipkl_cq);
	W_PUT_VAR(completion_handler);
	W_PUT_VAR(private_data);
	W_ISSUE_IOCTL(VIPKL_EQ_EVAPI_SET_COMP_EVENTH);
	return W_RET;
}

/***************************************************
struct i_VIPKL_EQ_evapi_clear_comp_eventh_ops_t {
  IN  VIPKL_EQ_hndl_t                  vipkl_eq;
  IN  CQM_cq_hndl_t                    vipkl_cq;
};
struct o_VIPKL_EQ_evapi_clear_comp_eventh_ops_t {
  OUT VIP_ret_t ret;
};
****************************************************/
VIP_ret_t VIPKL_EQ_evapi_clear_comp_eventh(
	VIP_RSCT_t usr_ctx,	 
    VIP_hca_hndl_t						hca_hndl,
   /*IN*/VIPKL_EQ_hndl_t                  vipkl_eq,
  /*IN*/ CQM_cq_hndl_t                    vipkl_cq)
{
	W_ALLOCA_IBUF(VIPKL_EQ_evapi_clear_comp_eventh);
	W_ALLOCA_OBUF(VIPKL_EQ_evapi_clear_comp_eventh);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(vipkl_eq);
	W_PUT_VAR(vipkl_cq);
    W_ISSUE_IOCTL(VIPKL_EQ_EVAPI_CLEAR_COMP_EVENTH);
	return W_RET;
}

/***************************************************
struct i_VIPKL_EQ_set_async_event_handler_ops_t {
  IN VIP_RSCT_t usr_ctx;
  IN VAPI_hca_hndl_t hca_hndl;
  IN VIPKL_EQ_hndl_t                  vipkl_eq;
  IN VAPI_async_event_handler_t       async_eventh;
  IN void *                           private_data;
};
struct o_VIPKL_EQ_set_async_event_handler_ops_t {
  OUT VIP_ret_t ret;
  OUT EVAPI_async_handler_hndl_t      async_handler_hndl;
};
****************************************************/
VIP_ret_t VIPKL_EQ_set_async_event_handler(VIP_RSCT_t usr_ctx,VAPI_hca_hndl_t hca_hndl,
  /*IN*/VIPKL_EQ_hndl_t                  vipkl_eq,
  /*IN*/VAPI_async_event_handler_t       async_eventh,
  /*IN*/void *                           private_data,
  /*OUT*/EVAPI_async_handler_hndl_t      *async_handler_hndl_p)
{
	W_ALLOCA_IBUF(VIPKL_EQ_set_async_event_handler);
	W_ALLOCA_OBUF(VIPKL_EQ_set_async_event_handler);
  W_PUT_VAR(hca_hndl);
	W_PUT_VAR(vipkl_eq);
	W_PUT_VAR(async_eventh);
	W_PUT_VAR(private_data);
	W_ISSUE_IOCTL(VIPKL_EQ_SET_ASYNC_EVENT_HANDLER);
	if (W_RET == VIP_OK) {
		if (async_handler_hndl_p) W_GET_VAR(async_handler_hndl);
	}
	return W_RET;
}


/***************************************************
struct i_VIPKL_EQ_clear_async_event_handler_ops_t {
  IN VIP_RSCT_t usr_ctx;
  IN VAPI_hca_hndl_t hca_hndl;
  IN VIPKL_EQ_hndl_t                  vipkl_eq;
  IN EVAPI_async_handler_hndl_t       async_handler_hndl;
}
struct o_VIPKL_EQ_clear_async_event_handler_ops_t {
  OUT VIP_ret_t ret;
};
****************************************************/
VIP_ret_t VIPKL_EQ_clear_async_event_handler(VIP_RSCT_t usr_ctx,VAPI_hca_hndl_t hca_hndl,
  /*IN*/VIPKL_EQ_hndl_t                  vipkl_eq,
  /*IN*/EVAPI_async_handler_hndl_t       async_handler_hndl)
{
	W_ALLOCA_IBUF(VIPKL_EQ_clear_async_event_handler);
	W_ALLOCA_OBUF(VIPKL_EQ_clear_async_event_handler);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(vipkl_eq);
	W_PUT_VAR(async_handler_hndl);
    W_ISSUE_IOCTL(VIPKL_EQ_CLEAR_ASYNC_EVENT_HANDLER);
	return W_RET;
}


/***************************************************
struct i_VIPKL_EQ_poll_ops_t {
  IN  VAPI_hca_hndl_t hca_hndl;
  IN  VIPKL_EQ_hndl_t vipkl_eq;
};
struct o_VIPKL_EQ_poll_ops_t {
  OUT VIP_ret_t ret;
  OUT VIPKL_EQ_event_t eqe;
};
****************************************************/
VIP_ret_t VIPKL_EQ_poll(
    VIP_RSCT_t usr_ctx,
    /*IN*/VIP_hca_hndl_t   hca_hndl,
    /*IN*/VIPKL_EQ_hndl_t vipkl_eq,
    /*OUT*/VIPKL_EQ_event_t *eqe_p)
{
	W_ALLOCA_OBUF(VIPKL_EQ_poll);
	W_ALLOCA_IBUF(VIPKL_EQ_poll);
    W_PUT_VAR(hca_hndl);
    W_PUT_VAR(vipkl_eq);
    W_ISSUE_IOCTL(VIPKL_EQ_POLL);
	if (W_RET == VIP_OK) {
		if (eqe_p) W_GET_VAR(eqe);
	}
	return W_RET;
}


/*************************************************************************
 * VIPKL_cqblk functions 
 *************************************************************************/ 

/***************************************************
struct i_VIPKL_cqblk_alloc_ctx_ops_t {
  IN	VIP_hca_hndl_t   vip_hca;
  IN	CQM_cq_hndl_t    vipkl_cq;
};
struct o_VIPKL_cqblk_alloc_ctx_ops_t {
  OUT VIP_ret_t ret;
  OUT VIPKL_cqblk_hndl_t cqblk_hndl;
};
****************************************************/
VIP_ret_t VIPKL_cqblk_alloc_ctx(
    VIP_RSCT_t usr_ctx,
    /*IN*/VIP_hca_hndl_t   hca_hndl,
  /*IN*/CQM_cq_hndl_t    vipkl_cq,
  /*OUT*/VIPKL_cqblk_hndl_t *cqblk_hndl_p)
{
	W_ALLOCA_IBUF(VIPKL_cqblk_alloc_ctx);
	W_ALLOCA_OBUF(VIPKL_cqblk_alloc_ctx);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(vipkl_cq);
	W_ISSUE_IOCTL(VIPKL_CQBLK_ALLOC_CTX);
	if (W_RET == VIP_OK) {
		if (cqblk_hndl_p) W_GET_VAR(cqblk_hndl);
	}
	return W_RET;
}

/***************************************************
struct i_VIPKL_cqblk_free_ctx_ops_t {
  IN	VIPKL_cqblk_hndl_t cqblk_hndl;
};
struct o_VIPKL_cqblk_free_ctx_ops_t {
  OUT VIP_ret_t ret;
};
****************************************************/
VIP_ret_t VIPKL_cqblk_free_ctx(
    VIP_RSCT_t usr_ctx, VIP_hca_hndl_t hca_hndl,
    /*IN*/VIPKL_cqblk_hndl_t cqblk_hndl)
{
	W_ALLOCA_IBUF(VIPKL_cqblk_free_ctx);
    W_ALLOCA_OBUF(VIPKL_cqblk_free_ctx);
    W_PUT_VAR(hca_hndl);
    W_PUT_VAR(cqblk_hndl);
    W_ISSUE_IOCTL(VIPKL_CQBLK_FREE_CTX);
	return W_RET;
}


/***************************************************
struct i_VIPKL_cqblk_wait_ops_t {
  IN	VIPKL_cqblk_hndl_t cqblk_hndl;
  IN	MT_size_t timeout_usec;
};
struct o_VIPKL_cqblk_wait_ops_t {
  OUT VIP_ret_t ret;
};
****************************************************/
VIP_ret_t VIPKL_cqblk_wait(
    VIP_RSCT_t usr_ctx, VIP_hca_hndl_t hca_hndl,
    /*IN*/VIPKL_cqblk_hndl_t cqblk_hndl,
  /*IN*/MT_size_t timeout_usec)
{
	W_ALLOCA_IBUF(VIPKL_cqblk_wait);
	W_ALLOCA_OBUF(VIPKL_cqblk_wait);
	W_PUT_VAR(hca_hndl);
    W_PUT_VAR(cqblk_hndl);
	W_PUT_VAR(timeout_usec);
	W_ISSUE_IOCTL(VIPKL_CQBLK_WAIT);
	return W_RET;
}

/***************************************************
struct i_VIPKL_cqblk_signal_ops_t {
  IN	VIPKL_cqblk_hndl_t cqblk_hndl;
};
struct o_VIPKL_cqblk_signal_ops_t {
  OUT VIP_ret_t ret;
};
****************************************************/
VIP_ret_t VIPKL_cqblk_signal(VIP_RSCT_t usr_ctx, VIP_hca_hndl_t hca_hndl,/*IN*/VIPKL_cqblk_hndl_t cqblk_hndl)
{
    W_ALLOCA_IBUF(VIPKL_cqblk_signal);
	W_ALLOCA_OBUF(VIPKL_cqblk_signal);
    W_PUT_VAR(cqblk_hndl);
    W_PUT_VAR(hca_hndl);
    W_ISSUE_IOCTL(VIPKL_CQBLK_SIGNAL);
	return W_RET;
}



/***************************************************
struct i_VIPKL_process_local_mad_ops_t {
  IN    VIP_hca_hndl_t hca_hndl;
  IN    IB_port_t      port_num;
  IN    IB_lid_t       slid,
  IN	  EVAPI_proc_mad_opt_t	 proc_mad_opts;
  IN    u_int8_t       mad_in[IB_MAD_LEN];
};
struct o_VIPKL_process_local_mad_ops_t {
  OUT VIP_ret_t ret;
  OUT   u_int8_t       mad_out[IB_MAD_LEN];
};
****************************************************/
VIP_ret_t VIPKL_process_local_mad(/*IN*/ VIP_hca_hndl_t hca_hndl,
    /*IN*/  IB_port_t port_num,
    /*IN*/  IB_lid_t  slid,
    /*IN*/ EVAPI_proc_mad_opt_t      proc_mad_opts,
    /*IN*/  void*     mad_in_p,
    /*OUT*/ void*     mad_out_p)
{
	W_ALLOCA_IBUF(VIPKL_process_local_mad);
	W_ALLOCA_OBUF(VIPKL_process_local_mad);
	if (mad_in_p == NULL) {
	    	MTL_ERROR2("VIPKL_process_local_mad: no input packet provided. \n");
	    	return VIP_EINVAL_PARAM;
  	} 
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(port_num);
	W_PUT_VAR(slid);
	W_PUT_VAR(proc_mad_opts);
	W_PUT_STRUCT(mad_in);
	W_ISSUE_IOCTL(VIPKL_PROCESS_LOCAL_MAD);
	if (W_RET == VIP_OK) {
		if (mad_out_p) W_GET_STRUCT(mad_out);
	}
	return W_RET;
}

/*************************************************************************
struct i_VIPKL_alloc_devmem_ops_t {
  IN   VIP_hca_hndl_t 	hca_hndl;
  IN   EVAPI_devmem_type_t mem_type;
  IN   MT_size_t           bsize;
  IN   u_int8_t            align_shift;
};
struct o_VIPKL_alloc_devmem_ops_t {
  OUT VIP_ret_t ret;
  OUT  MT_phys_addr_t     buf;
};
 *************************************************************************/ 
VAPI_ret_t VIPKL_alloc_map_devmem(
					        VIP_RSCT_t usr_ctx,
                            VAPI_hca_hndl_t 	hca_hndl,
                            EVAPI_devmem_type_t mem_type,
                            VAPI_size_t           bsize,
                            u_int8_t            align_shift, 
                            VAPI_phy_addr_t*    buf_p,
                            void**              virt_addr_p,
                            VAPI_devmem_hndl_t* dm_hndl_p)
{
	W_ALLOCA_IBUF(VIPKL_alloc_map_devmem);
	W_ALLOCA_OBUF(VIPKL_alloc_map_devmem);
	if (buf_p == NULL) {
	    	MTL_ERROR2("VIPKL_alloc_devmem: no input buffer provided. \n");
	    	return VIP_EINVAL_PARAM;
  	} 
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(mem_type);
	W_PUT_VAR(bsize);
	W_PUT_VAR(align_shift);
    if (virt_addr_p == NULL) {
        pi->map = FALSE;
    }else pi->map= TRUE;
	W_ISSUE_IOCTL(VIPKL_ALLOC_MAP_DEVMEM);
	if (W_RET == VIP_OK) {
		W_GET_VAR(buf);
        if (virt_addr_p != NULL) {
            W_GET_VAR(virt_addr);
        }
        W_GET_VAR(dm_hndl);
	}
	return W_RET;
}

/*************************************************************************
struct i_VIPKL_query_devmem_ops_t {
  IN   VIP_hca_hndl_t 	hca_hndl;
  IN   EVAPI_devmem_type_t  mem_type; 	   
};
struct o_VIPKL_query_devmem_ops_t {
  OUT VIP_ret_t ret;
  OUT  EVAPI_devmem_info_t  devmem_info;
};
 *************************************************************************/ 
VAPI_ret_t VIPKL_query_devmem(
					        VIP_RSCT_t usr_ctx,
                            VAPI_hca_hndl_t 	    hca_hndl,
                            EVAPI_devmem_type_t     mem_type,
                            u_int8_t                align_shift,
                            EVAPI_devmem_info_t*    devmem_info_p
                            )
{
	W_ALLOCA_IBUF(VIPKL_query_devmem);
	W_ALLOCA_OBUF(VIPKL_query_devmem);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(mem_type);
    W_PUT_VAR(align_shift);
	W_ISSUE_IOCTL(VIPKL_QUERY_DEVMEM);
	if (W_RET == VIP_OK) {
		W_GET_VAR(devmem_info);
	}
	return W_RET;
}

/*************************************************************************
struct i_VIPKL_free_devmem_ops_t {
  IN   VIP_hca_hndl_t 	hca_hndl;
  IN   EVAPI_devmem_type_t mem_type;
  IN   MT_phys_addr_t      buf;
  IN   MT_size_t           bsize;
 };
struct o_VIPKL_free_devmem_ops_t {
  OUT VIP_ret_t ret;
 };
 *************************************************************************/ 
VAPI_ret_t VIPKL_free_unmap_devmem(
					        VIP_RSCT_t usr_ctx,
                            VAPI_hca_hndl_t 	hca_hndl,
                            VAPI_devmem_hndl_t  dm_hndl)
{
	W_ALLOCA_IBUF(VIPKL_free_unmap_devmem);
	W_ALLOCA_OBUF(VIPKL_free_unmap_devmem);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(dm_hndl);
	W_ISSUE_IOCTL(VIPKL_FREE_UNMAP_DEVMEM);
	return W_RET;
}

void VIPKL_wrap_user_init(void)
{
  /* init_handlers_pool; - Handlers pool is not used anymore */
}

#if defined(MT_SUSPEND_QP)
/*****************************************
  struct i_VIPKL_suspend_qp_ops_t {
    IN VIP_hca_hndl_t 		hca_hndl;
    IN QPM_qp_hndl_t 		qp_hndl;
    IN MT_bool              suspend_flag;
  };
  struct o_VIPKL_suspend_qp_ops_t {
	OUT VIP_ret_t ret;
  };
******************************************/
VIP_ret_t VIPKL_suspend_qp(VIP_RSCT_t usr_ctx,  /*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ QPM_qp_hndl_t qp_hndl
  ,  /*IN*/ MT_bool       suspend_flag)
{
	W_ALLOCA_IBUF(VIPKL_suspend_qp);
	W_ALLOCA_OBUF(VIPKL_suspend_qp);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(qp_hndl);
	W_PUT_VAR(suspend_flag);
	W_ISSUE_IOCTL(VIPKL_SUSPEND_QP);
	return W_RET;
}

/*****************************************
  struct i_VIPKL_suspend_cq_ops_t {
    IN VIP_hca_hndl_t 		hca_hndl;
    IN CQM_cq_hndl_t 		cq_hndl;
    IN MT_bool              do_suspend;
  };
  struct o_VIPKL_suspend_cq_ops_t {
	OUT VIP_ret_t ret;
  };
******************************************/
VIP_ret_t VIPKL_suspend_cq(VIP_RSCT_t usr_ctx,  /*IN*/ VIP_hca_hndl_t hca_hndl
  ,  /*IN*/ CQM_cq_hndl_t cq_hndl
  ,  /*IN*/ MT_bool       do_suspend)
{
	W_ALLOCA_IBUF(VIPKL_suspend_cq);
	W_ALLOCA_OBUF(VIPKL_suspend_cq);
	W_PUT_VAR(hca_hndl);
	W_PUT_VAR(cq_hndl);
	W_PUT_VAR(do_suspend);
	W_ISSUE_IOCTL(VIPKL_SUSPEND_CQ);
	return W_RET;
}
#endif
