/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: ts_ib_core.h 476 2004-07-16 16:30:27Z roland $
*/

#ifndef _TS_IB_CORE_H
#define _TS_IB_CORE_H

#include "ts_ib_core_types.h"

#if defined(__KERNEL__)

static inline int ib_mtu_enum_to_int(enum ib_mtu mtu)
{
	switch (mtu) {
	case IB_MTU_256:  return  256;
	case IB_MTU_512:  return  512;
	case IB_MTU_1024: return 1024;
	case IB_MTU_2048: return 2048;
	case IB_MTU_4096: return 4096;
	default: 	  return -1;
	}
}

struct ib_device *ib_device_get_by_name(const char *name);
struct ib_device *ib_device_get_by_index(int index);
int ib_device_notifier_register(struct ib_device_notifier *notifier);
int ib_device_notifier_deregister(struct ib_device_notifier *notifier);

int ib_device_properties_get(struct ib_device            *device,
			     struct ib_device_properties *properties);
int ib_device_properties_set(struct ib_device            *device,
			     struct ib_device_changes    *properties);
int ib_port_properties_get(struct ib_device              *device,
			   tTS_IB_PORT                    port,
			   struct ib_port_properties     *properties);
int ib_port_properties_set(struct ib_device       	 *device,
			   tTS_IB_PORT            	  port,
			   struct ib_port_changes 	 *properties);

int ib_pkey_entry_get(struct ib_device    *device,
		      tTS_IB_PORT          port,
		      int                  index,
		      u16                 *pkey);
int ib_gid_entry_get(struct ib_device    *device,
		     tTS_IB_PORT          port,
		     int                  index,
		     tTS_IB_GID           gid);

int ib_pd_create(struct ib_device    *device,
                 void                *device_specific,
                 struct ib_pd       **pd);
int ib_pd_destroy(struct ib_pd *pd);

int ib_address_create(struct ib_pd             *pd,
                      struct ib_address_vector *address,
                      struct ib_address       **address_handle);
int ib_address_query(struct ib_address        *address_handle,
		     struct ib_address_vector *address);
int ib_address_destroy(struct ib_address *address_handle);

int ib_qp_create(struct ib_qp_create_param *param,
                 struct ib_qp             **qp,
                 u32                       *qpn);
int ib_special_qp_create(struct ib_qp_create_param *param,
			 tTS_IB_PORT                port,
			 enum ib_special_qp_type    qp_type,
			 struct ib_qp             **qp);
int ib_qp_modify(struct ib_qp           *qp,
		 struct ib_qp_attribute *attr);
int ib_qp_query(struct ib_qp           *qp,
		struct ib_qp_attribute *attr);
int ib_qp_query_qpn(struct ib_qp    *qp,
		    u32             *qpn);
int ib_qp_destroy(struct ib_qp *qp);


static inline int ib_send(struct ib_qp         *qp,  /*changed by itayr*/
			  struct ib_send_param *param,
			  int                   num_work_requests,
              int                  *actual_num_wr)
{
	IB_CHECK_MAGIC(qp, QP);
	return qp->device->send_post(qp, param, num_work_requests, actual_num_wr);
}
static inline int ib_receive(struct ib_qp            *qp,
			     struct ib_receive_param *param,
			     int                      num_work_requests,
                 int                  *actual_num_wr)
{
	IB_CHECK_MAGIC(qp, QP);
	return qp->device->receive_post(qp, param, num_work_requests, actual_num_wr);
}

int ib_cq_create(struct ib_device          *device,
                 int                       *entries,
                 struct ib_cq_callback     *callback,
                 void                      *device_specific,
                 struct ib_cq             **cq);
int ib_cq_destroy(struct ib_cq *cq);
int ib_cq_resize(struct ib_cq *cq,
                 int          *entries);
/*
 * Poll a CQ for (possibly multiple) completions.  cq is the CQ to
 * poll, num_entries the maximum number of completions to return, and
 * entry is an array of CQ entries of size num_entries.
 *
 * If the return value is < 0, an error occurred.  If the return value
 * is >= 0, it is the number of completions returned.  If the return
 * value is non-negative and < num_entries, then the CQ was emptied.
 */
static inline int ib_cq_poll_multi(struct ib_cq *cq, int num_entries,
		     		   struct ib_cq_entry *entry)
{
	IB_CHECK_MAGIC(cq, CQ);
	return cq->device->cq_poll(cq, num_entries, entry);
}
static inline int ib_cq_poll(struct ib_cq *cq, struct ib_cq_entry *entry)
{
	int ret = ib_cq_poll_multi(cq, 1, entry);
	if (ret > 0)
		return 0;
	if (ret == 0)
		return -EAGAIN;
	return ret;
}

int ib_cq_request_notification(struct ib_cq *cq,
			       int solicited);

int ib_memory_register(struct ib_pd          *pd,
                       void                  *start_address,
                       uint64_t               buffer_size,
                       enum ib_memory_access  access,
                       struct ib_mr         **memory,
                       u32                   *lkey,
                       u32                   *rkey);
int ib_memory_register_physical(struct ib_pd              *pd,
				struct ib_physical_buffer *buffer_list,
				int                   	   list_len,
				uint64_t              	  *io_virtual_address,
				uint64_t              	   buffer_size,
				uint64_t              	   iova_offset,
				enum ib_memory_access  	   access,
				struct ib_mr    	 **mr,
				u32                  	  *lkey,
				u32                  	  *rkey);
int ib_memory_deregister(struct ib_mr *memory);

int ib_mw_create(struct ib_pd     *pd,
		 struct ib_mw    **mw,
		 u32              *rkey);
int ib_mw_destroy(struct ib_mw *mw);
int ib_mw_bind(struct ib_qp            *qp,
	       struct ib_mw            *mw,
	       struct ib_mw_bind_param *param,
	       u32                     *new_rkey);

int ib_fmr_pool_create(struct ib_pd             *pd,
		       struct ib_fmr_pool_param *params,
		       struct ib_fmr_pool      **pool);
int ib_fmr_pool_destroy(struct ib_fmr_pool *pool);
int ib_fmr_pool_force_flush(struct ib_fmr_pool *pool);
int ib_fmr_register_physical(struct ib_fmr_pool     *pool,
			     uint64_t               *page_list,
			     int                     list_len,
			     uint64_t               *io_virtual_address,
			     uint64_t                iova_offset,
			     struct ib_fmr         **fmr,
			     u32                    *lkey,
			     u32                    *rkey);
int ib_fmr_deregister(struct ib_fmr *fmr);

int ib_multicast_attach(u16           multicast_lid,
			tTS_IB_GID    multicast_gid,
			struct ib_qp *qp);
int ib_multicast_detach(u16           multicast_lid,
			tTS_IB_GID    multicast_gid,
			struct ib_qp *qp);

int ib_async_event_handler_register(struct ib_async_event_record   *record,
				    ib_async_event_handler_func     function,
				    void                           *arg,
				    struct ib_async_event_handler **handle);
int ib_async_event_handler_deregister(struct ib_async_event_handler *handle);

int ib_cached_node_guid_get(struct ib_device    *device,
			    tTS_IB_GUID          node_guid);
int ib_cached_port_properties_get(struct ib_device          *device,
				  tTS_IB_PORT                port,
				  struct ib_port_properties *properties);
int ib_cached_sm_path_get(struct ib_device    *device,
			  tTS_IB_PORT          port,
			  struct ib_sm_path   *sm_path);
int ib_cached_lid_get(struct ib_device    *device,
		      tTS_IB_PORT          port,
		      struct ib_port_lid  *port_lid);
int ib_cached_gid_get(struct ib_device    *device,
		      tTS_IB_PORT          port,
		      int                  index,
		      tTS_IB_GID           gid);
int ib_cached_gid_find(tTS_IB_GID            gid,
		       struct ib_device    **device,
		       tTS_IB_PORT          *port,
		       int                  *index);
int ib_cached_pkey_get(struct ib_device    *device_handle,
		       tTS_IB_PORT          port,
		       int                  index,
		       u16                 *pkey);
int ib_cached_pkey_find(struct ib_device    *device,
			tTS_IB_PORT          port,
			u16                  pkey,
			int                 *index);

/* Variants that do not modify the provider module's reference count.
   These should only be used in code that handles provider removal
   properly via ib_device_notifier. */
int __ib_pd_create(struct ib_device    *device,
		   void                *device_specific,
		   struct ib_pd       **pd);
int __ib_pd_destroy(struct ib_pd *pd);
int __ib_cq_create(struct ib_device       *device,
		   int                    *entries,
		   struct ib_cq_callback  *callback,
		   void                   *device_specific,
		   struct ib_cq          **cq);
int __ib_cq_destroy(struct ib_cq *cq);

#endif /* __KERNEL__ */

#endif /* _TS_IB_CORE_H */

/*
  Local Variables:
  c-file-style: "linux"
  indent-tabs-mode: t
  End:
*/
