/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies.   All rights reserved.


  $Id: ts_ib_core_types.h 432 2004-07-12 17:27:37Z roland $
*/

#ifndef _TS_IB_CORE_TYPES_H
#define _TS_IB_CORE_TYPES_H

#include <linux/types.h>
#if defined(__KERNEL__)
#  include <linux/list.h>
#  include <linux/spinlock.h>
#  include <ts_kernel_hash.h>
#else
#  include <stdint.h>
#endif

#include <ts_ib_magic.h>

/* basic type definitions */

enum {
	IB_NODE_DESC_SIZE = 64
};

typedef __u8     tTS_IB_GUID[8];
typedef __u8     tTS_IB_GID[16];
typedef __u8     tTS_IB_SL;
typedef __u8     tTS_IB_VL;
typedef __u8     tTS_IB_PORT;
typedef __u8     tTS_IB_MKEY[8];
typedef __u8     tTS_IB_TIME;
typedef __u32    tTS_IB_PSN;

struct ib_device;

enum ib_mtu {
	IB_MTU_256  = 1,
	IB_MTU_512  = 2,
	IB_MTU_1024 = 3,
	IB_MTU_2048 = 4,
	IB_MTU_4096 = 5
};

enum ib_port_state {
	IB_PORT_STATE_NOP    = 0,
	IB_PORT_STATE_DOWN   = 1,
	IB_PORT_STATE_INIT   = 2,
	IB_PORT_STATE_ARMED  = 3,
	IB_PORT_STATE_ACTIVE = 4
};

enum ib_mad_result {
	IB_MAD_RESULT_FAILURE      = 0,        // (!SUCCESS is the important flag)
	IB_MAD_RESULT_SUCCESS      = 1 << 0,   // MAD was successfully processed
	IB_MAD_RESULT_REPLY        = 1 << 1,   // Reply packet needs to be sent
	IB_MAD_RESULT_CONSUMED     = 1 << 2    // Packet consumed: stop processing
};

struct ib_port_properties {
	enum ib_mtu        max_mtu;
	__u32              max_message_size;
	__u16              lid;
	__u8               lmc;
	enum ib_port_state port_state;
	int                gid_table_length;
	int                pkey_table_length;
	int                max_vl;
	__u32              bad_pkey_counter;
	__u32              qkey_violation_counter;
	__u8               init_type_reply;
	__u16              sm_lid;
	tTS_IB_SL          sm_sl;
	tTS_IB_TIME        subnet_timeout;
	__u32              capability_mask;
};

enum ib_port_properties_mask {
	IB_PORT_SHUTDOWN_PORT                  = 1 << 0,
	IB_PORT_INIT_TYPE                      = 1 << 1,
	IB_PORT_QKEY_VIOLATION_COUNTER_RESET   = 1 << 2,
	IB_PORT_IS_SM                          = 1 << 3,
	IB_PORT_IS_SNMP_TUNNELING_SUPPORTED    = 1 << 4,
	IB_PORT_IS_DEVICE_MANAGEMENT_SUPPORTED = 1 << 5,
	IB_PORT_IS_VENDOR_CLASS_SUPPORTED      = 1 << 6
};

struct ib_port_changes {
	enum ib_port_properties_mask valid_fields;
	__u8                         init_type;
	int                          shutdown:1;
	int                          qkey_violation_counter_reset:1;
	int                          is_sm:1;
	int                          is_snmp_tunneling_supported:1;
	int                          is_device_management_supported:1;
	int                          is_vendor_class_supported:1;
};

enum ib_rate {
	IB_RATE_2GB5      =  2,
	IB_RATE_10GB      =  3,
	IB_RATE_30GB      =  4,
};

struct ib_path_record {
	tTS_IB_GID   dgid;
	tTS_IB_GID   sgid;
	__u16        dlid;
	__u16        slid;
	__u32        flowlabel;
	__u8         hoplmt;
	__u8         tclass;
	__u16        pkey;
	tTS_IB_SL    sl;
	enum ib_mtu  mtu;
	enum ib_rate rate;
	__u8         packet_life;
	__u8         preference;
};

#define IB_DEVICE_NAME_MAX 64

enum ib_atomic_support {
	IB_NO_ATOMIC_OPS,
	IB_ATOMIC_HCA,
	IB_ATOMIC_ALL
};

struct ib_device_properties {
	char                   name[IB_DEVICE_NAME_MAX];
	char                  *provider;
	__u32                  vendor_id;
	__u16                  device_id;
	__u32                  hw_rev;
	__u64                  fw_rev;
	int                    max_qp;
	int                    max_wr_per_qp;
	int                    max_wr_per_post;
	int                    max_sg_per_wr;
	int                    max_sg_per_wr_rd;
	int                    max_cq;
	int                    max_mr;
	__u64                  max_mr_size;
	int                    max_pd;
	int                    page_size_cap;
	int                    num_port;
	int                    max_pkey;
	tTS_IB_TIME            local_ca_ack_delay;
	int                    max_responder_per_qp;
	int                    max_responder_per_eec;
	int                    max_responder_per_hca;
	int                    max_initiator_per_qp;
	int                    max_initiator_per_eec;
	enum ib_atomic_support atomic_support;
	int                    max_eec;
	int                    max_rdd;
	int                    max_mw;
	int                    max_raw_ipv6_qp;
	int                    max_raw_ethertype_qp;
	int                    max_mcg;
	int                    max_mc_qp;
	int                    max_qp_per_mcg;
	int                    max_ah;
	int                    max_fmr;
	int                    max_map_per_fmr;
	tTS_IB_GUID            node_guid;
	int                    is_switch:1;
	int                    ah_port_num_check:1;
	int                    rnr_nak_supported:1;
	int                    port_shutdown_supported:1;
	int                    init_type_supported:1;
	int                    port_active_event_supported:1;
	int                    system_image_guid_supported:1;
	int                    bad_pkey_counter_supported:1;
	int                    qkey_violation_counter_supported;
	int                    modify_wr_num_supported:1;
	int                    raw_multicast_supported:1;
	int                    apm_supported:1;
	int                    qp_port_change_supported:1;
};


enum ib_transport {
	IB_TRANSPORT_RC          = 0,
	IB_TRANSPORT_UC          = 1,
	IB_TRANSPORT_RD          = 2,
	IB_TRANSPORT_UD          = 3,
};

enum ib_qp_state {
	IB_QP_STATE_RESET,
	IB_QP_STATE_INIT,
	IB_QP_STATE_RTR,
	IB_QP_STATE_RTS,
	IB_QP_STATE_SQD,
	IB_QP_STATE_SQE,
	IB_QP_STATE_ERROR
};





#ifdef __KERNEL__

enum ib_op {
	IB_OP_RECEIVE,
	IB_OP_SEND,
	IB_OP_SEND_IMMEDIATE,
	IB_OP_RDMA_WRITE,
	IB_OP_RDMA_WRITE_IMMEDIATE,
	IB_OP_RDMA_READ,
	IB_OP_COMPARE_SWAP,
	IB_OP_FETCH_ADD,
	IB_OP_MEMORY_WINDOW_BIND
};

enum ib_completion_op {
	IB_COMPLETION_OP_RECEIVE,
	IB_COMPLETION_OP_RDMA_WRITE_RECEIVE,
	IB_COMPLETION_OP_SEND,
	IB_COMPLETION_OP_RDMA_WRITE,
	IB_COMPLETION_OP_RDMA_READ,
	IB_COMPLETION_OP_COMPARE_SWAP,
	IB_COMPLETION_OP_FETCH_ADD,
	IB_COMPLETION_OP_MEMORY_WINDOW_BIND,
};

enum ib_completion_status {
	IB_COMPLETION_STATUS_SUCCESS,
	IB_COMPLETION_STATUS_LOCAL_LENGTH_ERROR,
	IB_COMPLETION_STATUS_LOCAL_QP_OPERATION_ERROR,
	IB_COMPLETION_STATUS_LOCAL_EEC_OPERATION_ERROR,
	IB_COMPLETION_STATUS_LOCAL_PROTECTION_ERROR,
	IB_COMPLETION_STATUS_WORK_REQUEST_FLUSHED_ERROR,
	IB_COMPLETION_STATUS_MEMORY_WINDOW_BIND_ERROR,
	IB_COMPLETION_STATUS_BAD_RESPONSE_ERROR,
	IB_COMPLETION_STATUS_LOCAL_ACCESS_ERROR,
	IB_COMPLETION_STATUS_REMOTE_INVALID_REQUEST_ERROR,
	IB_COMPLETION_STATUS_REMOTE_ACCESS_ERORR,
	IB_COMPLETION_STATUS_REMOTE_OPERATION_ERROR,
	IB_COMPLETION_STATUS_TRANSPORT_RETRY_COUNTER_EXCEEDED,
	IB_COMPLETION_STATUS_RNR_RETRY_COUNTER_EXCEEDED,
	IB_COMPLETION_STATUS_LOCAL_RDD_VIOLATION_ERROR,
	IB_COMPLETION_STATUS_REMOTE_INVALID_RD_REQUEST,
	IB_COMPLETION_STATUS_REMOTE_ABORTED_ERROR,
	IB_COMPLETION_STATUS_INVALID_EEC_NUMBER,
	IB_COMPLETION_STATUS_INVALID_EEC_STATE,
	IB_COMPLETION_STATUS_UNKNOWN_ERROR
};

enum ib_async_event {
	IB_QP_PATH_MIGRATED,
	IB_EEC_PATH_MIGRATED,
	IB_QP_COMMUNICATION_ESTABLISHED,
	IB_EEC_COMMUNICATION_ESTABLISHED,
	IB_SEND_QUEUE_DRAINED,
	IB_CQ_ERROR,
	IB_LOCAL_WQ_INVALID_REQUEST_ERROR,
	IB_LOCAL_WQ_ACCESS_VIOLATION_ERROR,
	IB_LOCAL_WQ_CATASTROPHIC_ERROR,
	IB_PATH_MIGRATION_ERROR,
	IB_LOCAL_EEC_CATASTROPHIC_ERROR,
	IB_LOCAL_CATASTROPHIC_ERROR,
	IB_PORT_ERROR,
	IB_PORT_ACTIVE,
	IB_LID_CHANGE,
	IB_PKEY_CHANGE,
	IB_SM_CHANGE,
	IB_GPIO,
	IB_ECC_DETECT,
	IB_PATH_MIG_ARMED,
	IB_CLIENT_REREGISTER
};

struct ib_pd {
	IB_DECLARE_MAGIC
	struct ib_device *device;
	void             *private;
};

struct ib_address {
	IB_DECLARE_MAGIC
	struct ib_device *device;
	void             *private;
};

struct ib_async_obj {
	void *           free_ptr;
	spinlock_t       lock;
	int              pending;
	int              dead;
};

struct ib_qp {
	IB_DECLARE_MAGIC
	struct ib_device *device;
	u32               qpn;
	struct ib_async_obj async_obj;
	struct list_head  async_handler_list;
	spinlock_t        async_handler_lock;
	void             *private;
};

struct ib_mr {
	IB_DECLARE_MAGIC
	struct ib_device *device;
	u32               lkey;
	u32               rkey;
	void             *private;
};

enum ib_cq_callback_context {
	IB_CQ_CALLBACK_INTERRUPT,
	IB_CQ_CALLBACK_PROCESS
};

enum ib_cq_rearm_policy {
	IB_CQ_PROVIDER_REARM,
	IB_CQ_CONSUMER_REARM
};

struct ib_cq_entry {
	void                   	 *device_specific;
	u64                       work_request_id;
	u64                       bytes_transferred;
	enum ib_completion_status status;
	enum ib_completion_op     op;
	int                       pkey_index;
	u32                       immediate_data;
	u32                       sqpn;
	u32                       local_eecn;
	u32               	  freed_resource_count;
	u16                       ethertype;
	u16                       slid;
	u8                        sl;
	u8                        dlid_path_bits;
	int                    	  grh_present:1;
	int                    	  immediate_data_valid:1;
};

struct ib_cq;

typedef void (*ib_cq_entry_callback_func)(struct ib_cq       *cq,
					  struct ib_cq_entry *entry,
					  void               *arg);
typedef void (*ib_cq_event_callback_func)(struct ib_cq *cq,
					  void         *arg);

struct ib_cq_callback {
	enum ib_cq_callback_context         context;
	enum ib_cq_rearm_policy             policy;
	union {
		ib_cq_entry_callback_func entry;
		ib_cq_event_callback_func event;
	} function;
	void                               *arg;
};

struct ib_cq {
	IB_DECLARE_MAGIC
	struct ib_device     *device;
	struct ib_cq_callback callback;
	struct ib_async_obj   async_obj;
	struct list_head      async_handler_list;
	spinlock_t            async_handler_lock;
	void                 *private;
	atomic_t              num_outstanding_comp_events;
};

struct ib_fmr_pool;		/* actual definition in core_fmr.c */

struct ib_fmr {
	IB_DECLARE_MAGIC
	struct ib_device      *device;
	void                  *private;
	struct ib_fmr_pool    *pool;
	u32                    lkey;
	u32                    rkey;
	int                    ref_count;
	int                    remap_count;
	struct list_head       list;
	tTS_HASH_NODE_STRUCT   cache_node;
	u64                    io_virtual_address;
	u64                    iova_offset;
	int                    page_list_len;
	u64                    page_list[0];
};

typedef void (*ib_fmr_flush_func)(struct ib_fmr_pool *pool,
				  void               *arg);

struct ib_async_event_handler;	/* actual definition in core_async.c */

struct ib_async_event_record {
	struct ib_device    *device;
	enum ib_async_event  event;
	union {
		struct ib_qp  *qp;
		struct ib_eec *eec;
		struct ib_cq  *cq;
		int            port;
	} modifier;
};

typedef void (*ib_async_event_handler_func)(struct ib_async_event_record *record,
					    void *arg);

/* enum definitions */

#define IB_MULTICAST_QPN   0xffffff

enum ib_static_rate {
	IB_STATIC_RATE_FULL      =  0,
	IB_STATIC_RATE_4X_TO_1X  =  3,
	IB_STATIC_RATE_12X_TO_4X =  2,
	IB_STATIC_RATE_12X_TO_1X = 11
};

enum ib_rnr_timeout {
	IB_RNR_TIMER_655_36 =  0,
	IB_RNR_TIMER_000_01 =  1,
	IB_RNR_TIMER_000_02 =  2,
	IB_RNR_TIMER_000_03 =  3,
	IB_RNR_TIMER_000_04 =  4,
	IB_RNR_TIMER_000_06 =  5,
	IB_RNR_TIMER_000_08 =  6,
	IB_RNR_TIMER_000_12 =  7,
	IB_RNR_TIMER_000_16 =  8,
	IB_RNR_TIMER_000_24 =  9,
	IB_RNR_TIMER_000_32 = 10,
	IB_RNR_TIMER_000_48 = 11,
	IB_RNR_TIMER_000_64 = 12,
	IB_RNR_TIMER_000_96 = 13,
	IB_RNR_TIMER_001_28 = 14,
	IB_RNR_TIMER_001_92 = 15,
	IB_RNR_TIMER_002_56 = 16,
	IB_RNR_TIMER_003_84 = 17,
	IB_RNR_TIMER_005_12 = 18,
	IB_RNR_TIMER_007_68 = 19,
	IB_RNR_TIMER_010_24 = 20,
	IB_RNR_TIMER_015_36 = 21,
	IB_RNR_TIMER_020_48 = 22,
	IB_RNR_TIMER_030_72 = 23,
	IB_RNR_TIMER_040_96 = 24,
	IB_RNR_TIMER_061_44 = 25,
	IB_RNR_TIMER_081_92 = 26,
	IB_RNR_TIMER_122_88 = 27,
	IB_RNR_TIMER_163_84 = 28,
	IB_RNR_TIMER_245_76 = 29,
	IB_RNR_TIMER_327_68 = 30,
	IB_RNR_TIMER_491_52 = 31
};

enum ib_device_properties_mask {
	IB_DEVICE_SYSTEM_IMAGE_GUID            = 1 << 0
};

enum ib_special_qp_type {
	IB_SMI_QP,
	IB_GSI_QP,
	IB_RAW_IPV6_QP,
	IB_RAW_ETHERTYPE_QP
};

enum ib_wq_signal_policy {
	IB_WQ_SIGNAL_ALL,
	IB_WQ_SIGNAL_SELECTABLE
};

enum ib_migration_state {
	IB_MIGRATED,
	IB_REARM,
	IB_ARMED
};

enum ib_qp_attribute_mask {
	IB_QP_ATTRIBUTE_STATE                  = 1 <<  0,
	IB_QP_ATTRIBUTE_SEND_PSN               = 1 <<  1,
	IB_QP_ATTRIBUTE_RECEIVE_PSN            = 1 <<  2,
	IB_QP_ATTRIBUTE_DESTINATION_QPN        = 1 <<  3,
	IB_QP_ATTRIBUTE_QKEY                   = 1 <<  4,
	IB_QP_ATTRIBUTE_PATH_MTU               = 1 <<  5,
	IB_QP_ATTRIBUTE_MIGRATION_STATE        = 1 <<  6,
	IB_QP_ATTRIBUTE_INITIATOR_DEPTH        = 1 <<  7,
	IB_QP_ATTRIBUTE_RESPONDER_RESOURCES    = 1 <<  8,
	IB_QP_ATTRIBUTE_RETRY_COUNT            = 1 <<  9,
	IB_QP_ATTRIBUTE_RNR_RETRY_COUNT        = 1 << 10,
	IB_QP_ATTRIBUTE_RNR_TIMEOUT            = 1 << 11,
	IB_QP_ATTRIBUTE_PKEY_INDEX             = 1 << 12,
	IB_QP_ATTRIBUTE_PORT                   = 1 << 13,
	IB_QP_ATTRIBUTE_ADDRESS                = 1 << 14,
	IB_QP_ATTRIBUTE_LOCAL_ACK_TIMEOUT      = 1 << 15,
	IB_QP_ATTRIBUTE_ALT_PKEY_INDEX         = 1 << 16,
	IB_QP_ATTRIBUTE_ALT_PORT               = 1 << 17,
	IB_QP_ATTRIBUTE_ALT_ADDRESS            = 1 << 18,
	IB_QP_ATTRIBUTE_ALT_LOCAL_ACK_TIMEOUT  = 1 << 19,
	IB_QP_ATTRIBUTE_RDMA_ATOMIC_ENABLE     = 1 << 20,
	IB_QP_ATTRIBUTE_SQD_ASYNC_EVENT_ENABLE = 1 << 21
};

enum ib_memory_access {
	IB_ACCESS_LOCAL_WRITE   = 1 << 0,
	IB_ACCESS_REMOTE_WRITE  = 1 << 1,
	IB_ACCESS_REMOTE_READ   = 1 << 2,
	IB_ACCESS_REMOTE_ATOMIC = 1 << 3,
	IB_ACCESS_ENABLE_WINDOW = 1 << 4
};

/* structures */

struct ib_mw {
	IB_DECLARE_MAGIC
	struct ib_device *device;
	struct ib_pd     *pd;
	u32               rkey;
	void             *private;
};

enum {
	IB_DEVICE_NOTIFIER_ADD,
	IB_DEVICE_NOTIFIER_REMOVE
};

struct ib_device_notifier {
	void (*notifier)(struct ib_device_notifier *self,
			 struct ib_device          *device,
			 int event);
	struct list_head list;
};

struct ib_device_changes {
	enum ib_device_properties_mask valid_fields;
	tTS_IB_GUID                    system_image_guid;
};

struct ib_address_vector {
	int                 service_level;
	enum ib_static_rate static_rate;
	int                 source_path_bits;
	u16                 dlid;
	tTS_IB_PORT         port;
	u32                 flow_label;
	int                 source_gid_index;
	u8                  hop_limit;
	u8                  traffic_class;
	tTS_IB_GID          dgid;
	int                 use_grh:1;
};

struct ib_qp_limit {
	int max_outstanding_send_request;
	int max_outstanding_receive_request;
	int max_send_gather_element;
	int max_receive_scatter_element;
};

struct ib_qp_create_param {
	struct ib_qp_limit     	 limit;
	struct ib_pd           	*pd;
	struct ib_cq           	*send_queue;
	struct ib_cq           	*receive_queue;
	enum ib_wq_signal_policy send_policy;
	enum ib_wq_signal_policy receive_policy;
	struct ib_rdd          	*rd_domain;
	enum ib_transport      	 transport;
	void                   	*device_specific;
};

struct ib_qp_attribute {
	enum ib_qp_attribute_mask    valid_fields;
	enum ib_qp_state             state;
	tTS_IB_PSN                   send_psn;
	tTS_IB_PSN                   receive_psn;
	u32                          destination_qpn;
	u32                          qkey;
	enum ib_mtu                  path_mtu;
	enum ib_migration_state      migration_state;
	int                          initiator_depth;
	int                          responder_resources;
	u8                           retry_count;
	u8                           rnr_retry_count;
	enum ib_rnr_timeout          rnr_timeout;
	int                          pkey_index;
	tTS_IB_PORT                  port;
	struct ib_address_vector     address;
	u8                           local_ack_timeout;
	int                          alt_pkey_index;
	tTS_IB_PORT                  alt_port;
	struct ib_address_vector     alt_address;
	u8                           alt_local_ack_timeout;
	int                          enable_atomic:1;
	int                          enable_rdma_read:1;
	int                          enable_rdma_write:1;
	int                          sqd_async_event_enable:1;
	int                          sq_drained:1;
};

struct ib_gather_scatter {
	u64         address;
	u32         length;
	u32         key;
};

struct ib_send_param {
	u64             	  work_request_id;
	enum ib_op             	  op;
	struct ib_gather_scatter *gather_list;
	int                   	  num_gather_entries;
	u64                   	  remote_address;
	u32                   	  rkey;
	u32             	  dest_qpn;
	u32                  	  dest_qkey;
	struct ib_address        *dest_address;
	u32                   	  immediate_data;
	u64                   	  compare_add;
	u64                   	  swap;
	u32	           	  eecn;
	u16                   	  ethertype;
	enum ib_static_rate   	  static_rate;
	int                   	  pkey_index;
	void                  	 *device_specific;
	int                   	  solicited_event:1;
	int                   	  signaled:1;
	int                   	  immediate_data_valid:1;
	int                   	  fence:1;
	int                   	  inline_data:1;
};

struct ib_receive_param {
	u64			  work_request_id;
	struct ib_gather_scatter *scatter_list;
	int                   	  num_scatter_entries;
	void                  	 *device_specific;
	int                   	  signaled:1;
};

struct ib_mw_bind_param {
	u64		       work_request_id;
	u32                    rkey;
	u32                    lkey;
	u64                    address;
	u64                    length;
	enum ib_memory_access  access;
	int                    signaled:1;
	int                    fence:1;
};

struct ib_physical_buffer {
	u64      address;
	u64      size;
};

struct ib_fmr_pool_param {
	int                       max_pages_per_fmr;
	enum ib_memory_access     access;
	int                       pool_size;
	int                       dirty_watermark;
	ib_fmr_flush_func         flush_function;
	void                     *flush_arg;
	int                       cache:1;
};

struct ib_sm_path {
	u16         sm_lid;
	tTS_IB_SL   sm_sl;
};

struct ib_port_lid {
	u16        lid;
	u8         lmc;
};

typedef int (*ib_device_query_func)(struct ib_device            *device,
				    struct ib_device_properties *properties);
typedef int (*ib_device_modify_func)(struct ib_device         *device,
				     struct ib_device_changes *properties);
typedef int (*ib_port_query_func)(struct ib_device          *device,
				  tTS_IB_PORT                port,
				  struct ib_port_properties *properties);
typedef int (*ib_port_modify_func)(struct ib_device       *device,
				   tTS_IB_PORT             port,
				   struct ib_port_changes *properties);
typedef int (*ib_pkey_query_func)(struct ib_device *device,
				  tTS_IB_PORT  	    port,
				  int          	    index,
				  u16         	   *pkey);
typedef int (*ib_gid_query_func)(struct ib_device *device,
				 tTS_IB_PORT       port,
				 int           	   index,
				 tTS_IB_GID    	   gid);
typedef int (*ib_pd_create_func)(struct ib_device *device,
				 void             *device_specific,
				 struct ib_pd     *pd);
typedef int (*ib_pd_destroy_func)(struct ib_pd *pd);
typedef int (*ib_address_create_func)(struct ib_pd             *pd,
				      struct ib_address_vector *address_vector,
				      struct ib_address        *address);
typedef int (*ib_address_query_func)(struct ib_address        *address,
				     struct ib_address_vector *address_vector);
typedef int (*ib_address_destroy_func)(struct ib_address *address);
typedef int (*ib_qp_create_func)(struct ib_pd              *pd,
				 struct ib_qp_create_param *param,
				 struct ib_qp              *qp);
typedef int (*ib_special_qp_create_func)(struct ib_pd              *pd,
					 struct ib_qp_create_param *param,
					 tTS_IB_PORT                port,
					 enum ib_special_qp_type    qp_type,
					 struct ib_qp              *qp);
typedef int (*ib_qp_modify_func)(struct ib_qp           *qp,
				 struct ib_qp_attribute *attr);
typedef int (*ib_qp_query_func)(struct ib_qp           *qp,
				struct ib_qp_attribute *attr);
typedef int (*ib_qp_destroy_func)(struct ib_qp *qp);
typedef int (*ib_send_post_func)(struct ib_qp         *qp,  /*changed by itayr*/
				 struct ib_send_param *param,
				 int                   num_work_requests,
                 int                  *actual_num_wq);
typedef int (*ib_receive_post_func)(struct ib_qp            *qp,
				    struct ib_receive_param *param,
				    int                      num_work_requests,
                    int                  *actual_num_wq);                
typedef int (*ib_cq_create_func)(struct ib_device *device,
				 int          	  *entries,
				 void         	  *device_specific,
				 struct ib_cq     *cq);
typedef int (*ib_cq_destroy_func)(struct ib_cq *cq);
typedef int (*ib_cq_stop_comp_ev_func)(struct ib_cq *cq);
typedef int (*ib_cq_resize_func)(struct ib_cq *cq,
				 int          *entries);
typedef int (*ib_cq_poll_func)(struct ib_cq       *cq,
			       int                 num_entries,
			       struct ib_cq_entry *entry);
typedef int (*ib_cq_arm_func)(struct ib_cq *cq,
			      int           solicited);
typedef int (*ib_mr_register_func)(struct ib_pd         *pd,
				   void                 *start_address,
				   u64                   buffer_size,
				   enum ib_memory_access access,
				   struct ib_mr         *mr);
typedef int (*ib_mr_register_physical_func)(struct ib_pd              *pd,
					    struct ib_physical_buffer *buffer_list,
					    int                        list_len,
					    u64                       *io_virtual_address,
					    u64                        buffer_size,
					    u64                        iova_offset,
					    enum ib_memory_access      access,
					    struct ib_mr              *mr);
typedef int (*ib_mr_deregister_func)(struct ib_mr *mr);
typedef int (*ib_mw_create_func)(struct ib_pd  *pd,
				 struct ib_mw **mw,
				 u32           *rkey);
typedef int (*ib_mw_destroy_func)(struct ib_mw *mw);
typedef int (*ib_mw_bind_func)(struct ib_qp            *qp,
			       struct ib_mw            *mw,
			       struct ib_mw_bind_param *param,
			       u32                     *new_rkey);
typedef int (*ib_fmr_create_func)(struct ib_pd         *pd,
				  enum ib_memory_access access,
				  int                   max_pages,
				  int                   max_remaps,
				  struct ib_fmr        *fmr);
typedef int (*ib_fmr_destroy_func)(struct ib_fmr *fmr);
typedef int (*ib_fmr_map_func)(struct ib_fmr *fmr,
			       u64           *page_list,
			       int            list_len,
			       u64           *io_virtual_address,
			       u64            iova_offset,
			       u32           *lkey,
			       u32           *rkey);
typedef int (*ib_fmr_unmap_func)(struct ib_device *device,
				 struct list_head *fmr_list);
typedef int (*ib_multicast_attach_func)(struct ib_qp *qp,
					u16        lid,
					tTS_IB_GID gid);
typedef int (*ib_multicast_detach_func)(struct ib_qp *qp,
					u16        lid,
					tTS_IB_GID gid);

struct ib_mad;

typedef enum ib_mad_result (*ib_mad_process_func)(struct ib_device *device,
						  int               ignore_mkey,
						  struct ib_mad    *in_mad,
						  struct ib_mad    *response_mad);

/* structures */

enum {
	IB_MAD_NO_HOP_POINTER_INCR = 1 << 0,
	IB_MAD_LOCAL_USE_QP        = 1 << 1
};

struct ib_device {
	IB_DECLARE_MAGIC

	struct module                *owner;
	struct pci_dev               *dma_device;

	char                          name[IB_DEVICE_NAME_MAX];
	char                         *provider;
	void                         *private;
	struct list_head              core_list;
	void                         *core;
	void                         *mad;
	u32                           flags;

	ib_device_query_func         device_query;
	ib_device_modify_func        device_modify;
	ib_port_query_func           port_query;
	ib_port_modify_func          port_modify;
	ib_pkey_query_func           pkey_query;
	ib_gid_query_func            gid_query;
	ib_pd_create_func            pd_create;
	ib_pd_destroy_func           pd_destroy;
	ib_address_create_func       address_create;
	ib_address_query_func        address_query;
	ib_address_destroy_func      address_destroy;
	ib_qp_create_func            qp_create;
	ib_special_qp_create_func    special_qp_create;
	ib_qp_modify_func            qp_modify;
	ib_qp_query_func             qp_query;
	ib_qp_destroy_func           qp_destroy;
	ib_send_post_func            send_post;
    ib_receive_post_func         receive_post;
	ib_cq_create_func            cq_create;
	ib_cq_destroy_func           cq_destroy;
	ib_cq_resize_func            cq_resize;
	ib_cq_poll_func              cq_poll;
	ib_cq_arm_func               cq_arm;
	ib_cq_stop_comp_ev_func      cq_stop_comp_ev;
	ib_mr_register_func          mr_register;
	ib_mr_register_physical_func mr_register_physical;
	ib_mr_deregister_func        mr_deregister;
	ib_mw_create_func            mw_create;
	ib_mw_destroy_func           mw_destroy;
	ib_mw_bind_func              mw_bind;
	ib_fmr_create_func           fmr_create;
	ib_fmr_destroy_func          fmr_destroy;
	ib_fmr_map_func              fmr_map;
	ib_fmr_unmap_func            fmr_unmap;
	ib_multicast_attach_func     multicast_attach;
	ib_multicast_detach_func     multicast_detach;
	ib_mad_process_func          mad_process;
	ib_qp_destroy_func           qp_detach_destroy;

	struct ib_switch_ops        *switch_ops;
};

#endif /* __KERNEL__ */

#endif /* _TS_IB_CORE_TYPES_H */

/*
  Local Variables:
  c-file-style: "linux"
  indent-tabs-mode: t
  End:
*/
