/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.

  $Id: ts_ib_dm_client_host.h 481 2004-07-16 18:05:10Z roland $
*/

#ifndef _TS_IB_DM_CLIENT_HOST_H
#define _TS_IB_DM_CLIENT_HOST_H

#include <linux/types.h>
#include <linux/list.h>

#include "ts_ib_core_types.h"
#include "ts_ib_dm_client.h"

/* enum definitions */

typedef uint32_t tTS_IB_DM_CLIENT_HOST_TID;
#define TS_IB_DM_CLIENT_HOST_TID_INVALID  0xffffffff

struct ib_dm_client_host_io_svc {
        u16                       port_lid;
        tTS_IB_GUID               controller_guid;
        struct ib_svc_entry       svc_entry;
        struct list_head          list;
};

struct ib_dm_client_host_io_list {
	struct list_head io_svc_list;
};

/**
   Type for callback passed to tsIbIocProfileRequest()

   @param status - status of the query response (0, -ENOMEM, -ETIMEOUT, -EINVAL)
   @param io_list - list contains all the io_svc assigned to the host_port. It's
   NULL if status is not 0.
   @param arg - completion argument given in tsIbHostIoQuery()

   NOTE: io_list memory is a variable list of io_svc entries.  The memory
   needs to be freed using tsIbHostIoListFree() after the called is done
   with io_list.
*/
typedef void (*ib_dm_client_host_completion_func)(int status,
						  struct ib_dm_client_host_io_list
						  *io_list,
						  void* arg);


/* function declarations */

/**
   Cancel callbacks for a pending host IO query
   @param transaction_id Transaction ID of request to cancel
*/

void ib_host_io_query_cancel(tTS_IB_DM_CLIENT_HOST_TID transaction_id);

/**
   Initialize host IO query data structure
*/
int ib_host_io_query_init(void);

/**
   Clean up internal resouce for host IO
*/
void ib_host_io_query_cleanup(void);

/**
   Initiate a host IO query to a specific IO based on io_port_lid

   @param device Device on the host machine that is used.
   @param host_port - port on the host machine that is used.
   @param io_port_lid - destination IO port LID.
   @param timeout_jiffies - Time (in jiffies) to allow before
   considering request time out.
   @param completion_function - Function to call on completion of
   the lookup.
   @param completion_arg - Argument to supply to completion_function
   @param transaction_id - Pointer used to return transaction_id which
   can be used in tsIbDmClientHostQueryCancel()
*/
int ib_host_io_port_query(struct ib_device *device, tTS_IB_PORT host_port,
			  u16 io_port_lid, int timeout_jiffies,
			  ib_dm_client_host_completion_func
			  completion_function, void *completion_arg,
			  tTS_IB_DM_CLIENT_HOST_TID *transaction_id);

/**
   Initiate a host IO query.

   @param device Device on the host machine that is used.
   @param host_port - port on the host machine that is used.
   @param timeout_jiffies - Time (in jiffies) to allow before
   considering request time out.
   @param completion_function - Function to call on completion of
   the lookup.
   @param completion_arg - Argument to supply to completion_function
   @param transaction_id - Pointer used to return transaction_id which
   can be used in tsIbDmClientHostQueryCancel()
*/
int ib_host_io_query(struct ib_device *device, tTS_IB_PORT host_port,
		     int timeout_jiffies,
		     ib_dm_client_host_completion_func completion_function,
		     void *completion_arg,
		     tTS_IB_DM_CLIENT_HOST_TID *transaction_id);

/**
   Free the io_list that is returned to the called in the completion_function

   io_list - io_list argument in  tTS_IB_DM_CLIENT_HOST_COMPLETION_FUNC.
*/
int ib_host_io_list_free(struct ib_dm_client_host_io_list *io_list);

#endif /* _TS_IB_DM_CLIENT_HOST_H */
