#!/bin/bash

# This software is available to you under a choice of one of two
# licenses.  You may choose to be licensed under the terms of the GNU
# General Public License (GPL) Version 2, available at
# <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
# license, available in the LICENSE.TXT file accompanying this
# software.  These details are also available at
# <http://openib.org/license.html>.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
#
# Description: RPM build script
#


usage()
{
cat << EOF

    Usage: `basename $0`    [--prefix PREFIX] [--batch] [--help] [--tmpdir DIR] [-kver <kernel version>] [-k|--ksrc DIR]  [--memtrack]
			    [--nosu] [--clean] [--static] [--packages package1 package2 ... -- | --all]
			    [--build_root BUILD_ROOT]

			    Available Packages: ib_verbs ib_ipoib ib_cm ib_sdp ib_dapl ib_srp


EOF
}


#################################################################################################
#                                       Default values assigment                                #
#################################################################################################

# remove any aliases
unalias -a
# Find the directory where this script resides
installdir=`cd ${0%*/*};pwd`
prog=$0
package_name=$(echo ${installdir##*/} | cut -d '-' -f 1 | tr -d '[:space:]')
tmpdir=/tmp/openib
prefix=/usr/local/ibgd
clean=no
batch=no
nosu=no
SET_ENV=1
MEMTRACK="no"
PACKAGES=""
ALL_PACKAGES="ib_verbs ib_ipoib ib_cm ib_sdp ib_dapl ib_srp"
NULL=/dev/null
IB_CONFIG_DIR=/etc/infiniband
BUILD_ROOT=""

verbs=n
ipoib=n
cm=n
sdp=n
dapl=n
srp=n
srp_target=n

# Handle the case there is no makedepend...
export PATH=${PATH}:${installdir}
# COLOR SETTINGS
SETCOLOR_SUCCESS="echo -en \\033[1;34m"
SETCOLOR_FAILURE="echo -en \\033[1;31m"
SETCOLOR_WARNING="echo -en \\033[1;35m"
SETCOLOR_NORMAL="echo -en \\033[0;39m"

#################################################################################################
#                                       Parsing command line                                    #
#################################################################################################
while [ ! -z "$1" ]; do
	case "$1" in
		-t|--tmpdir)
			tmpdir=$2
			shift 2
			;;
		-kver)
			KVERSION=$2
			shift 2
			;;
		-k|--ksrc)
			KSRC=$2
			shift 2
			;;
		-h|--help)
			usage
			shift
			exit 0
			;;
		-b|--batch)
			batch=yes
			shift
			;;
		-c|--clean)
			clean=yes
			shift
			;;
		--nosu)
			nosu=yes
			shift
			;;
		--build_root)
			BUILD_ROOT=$2
			shift 2
			;;
		--build_only)
			SET_ENV=0
			shift
			;;
		--memtrack)
			MEMTRACK="yes"
			shift
			;;
		-p|--prefix)
			prefix=$2
			shift 2
			;;
		--packages)
			shift
			while [ "$1" != "--" ]
			do
			    PACKAGES="$PACKAGES $1"
			    shift
			done    
			if [ -z "$PACKAGES" ]; then
			    usage
			    exit 1
			fi
 
			shift
			;;
		--all)
			PACKAGES=$ALL_PACKAGES
			shift
			;;
		--static)
			STATIC_LIB=yes
			shift
			;;
		*)
			echo "Wrong parameter error: $1"
			usage
			exit 1
			;;
	esac
done

if [ "$nosu" == "no" ]; then
    if [ $UID -ne 0 ]; then
	echo You must be root to install
	exit 1
    fi
fi

# Set default value for PACKAGES  
if [ -z "$PACKAGES" ]; then
    PACKAGES=$ALL_PACKAGES
fi

#################################################################################################
#                                       Global environment                                      #
#################################################################################################

HOST_INFO=${tmpdir}/host_info.log
DEBUG_INFO=${tmpdir}/debug_info.tgz
KVERSION=${KVERSION:-$(uname -r)}
KVERSION_TWO_DIGITS=2$(echo ${KVERSION} | cut -d'.' -f2)
KSRC=${KSRC:-/lib/modules/${KVERSION}/build}
KERNEL_TREE_ROOT=${KSRC%*/*}
MAKE_KERNEL_DIR=${KSRC}
BUILD_IB_KERNEL_HOME=${tmpdir}/infiniband
IB_KERNEL_HOME=${KSRC}/drivers/infiniband

# make sure we have the tmpdir
/bin/rm -rf ${tmpdir}
mkdir -p ${tmpdir}

#LOGFILE=${tmpdir}/${package_name}.kernel.make.log
#/bin/rm -f $LOGFILE

#Special case for SuSE9.1 and SlES9.0 build tree
if [ -d ${KERNEL_TREE_ROOT}/source/ ]; then
   KSRC=${KERNEL_TREE_ROOT}/source
   MAKE_KERNEL_DIR=${KERNEL_TREE_ROOT}/build
fi

MODULES_CONF_PARAMS=""
MAKE_PARAMS=""

if [ "${KVERSION_TWO_DIGITS}" == "26" ]; then
	MAKE_PARAMS="V=1"
fi

#################################################################################################
#                                       Global functions                                        #
#################################################################################################
trap 'cleanup' 2 9 15

err_echo()
{
    echo
    [ "$TT_COLOR" = "yes" ] && $SETCOLOR_FAILURE
    echo "ERROR: $@" 
    [ "$TT_COLOR" = "yes" ] && $SETCOLOR_NORMAL
    echo "ERROR: $@" >> $LOGFILE
}

pass_echo()
{
   echo
   [ "$TT_COLOR" = "yes" ] && $SETCOLOR_SUCCESS
   echo "$@"
   [ "$TT_COLOR" = "yes" ] && $SETCOLOR_NORMAL
   echo  "$@" >> $LOGFILE
}

# Create debug info
get_debug_info()
{
    /bin/rm -f $HOST_INFO
    touch $HOST_INFO
    echo "Hostname: `hostname -s`" >> $HOST_INFO
    echo "openib BUILD_ID=`cat ${installdir}/BUILD_ID`" >> $HOST_INFO
    test -e /etc/issue && echo "OS: `cat /etc/issue`" >> $HOST_INFO
    echo "Current kernel: `uname -r`" >> $HOST_INFO
    echo "Building modules for kernel: ${KVERSION}" >> $HOST_INFO
    echo "Architecture: `uname -m`" >> $HOST_INFO
    echo "GCC version: `gcc --version`"  >> $HOST_INFO
    echo "CPU: `cat /proc/cpuinfo | /bin/grep -E \"model name|arch\" | head -1`" >> $HOST_INFO
    echo "`cat /proc/meminfo | /bin/grep \"MemTotal\"`" >> $HOST_INFO
    echo "Chipset: `/sbin/lspci | head -1 | cut -d ':' -f 2-`" >> $HOST_INFO
    echo >> $HOST_INFO
    echo "############# LSPCI ##############" >> $HOST_INFO
    /sbin/lspci >> $HOST_INFO
    echo >> $HOST_INFO
    echo "############# LSPCI -N ##############" >> $HOST_INFO
    /sbin/lspci -n >> $HOST_INFO

    echo                                 
    cd ${tmpdir}
    tar czf ${DEBUG_INFO} *.log
    echo "Please send $DEBUG_INFO to Mellanox representative"
}

# Clean on failure
cleanup()
{
    trap 'echo Please wait ...' 2
    printf "\nRunning cleanup ..."
    # Restore Gen2 sources
    rm -rf ${BUILD_ROOT}/${KSRC}/drivers/infiniband
    if [ -d ${BUILD_ROOT}/${KSRC}/drivers/infiniband-gen2 ]; then
    	echo "Restored Gen2 infiniband kernel sources"
    	mv ${BUILD_ROOT}/${KSRC}/drivers/infiniband-gen2 ${BUILD_ROOT}/${KSRC}/drivers/infiniband
    fi
    
    # Restore Gen2 modules
    if [ -d ${BUILD_ROOT}/lib/modules/${KVERSION}/kernel/drivers/infiniband-gen2 ]; then
    	rm -rf  ${BUILD_ROOT}/lib/modules/${KVERSION}/kernel/drivers/infiniband
    	mv ${BUILD_ROOT}/lib/modules/${KVERSION}/kernel/drivers/infiniband-gen2 ${BUILD_ROOT}/lib/modules/${KVERSION}/kernel/drivers/infiniband
    	for mod in `find ${BUILD_ROOT}/lib/modules/${KVERSION}/kernel/drivers/infiniband -name "*.ko.gen2"`
    	do
            mv ${mod} ${mod%*.*}
    	done 
    fi

    rm -rf ${BUILD_ROOT}/${prefix}
    printf "\r\t\t\t\t\t (done)"

}

# Execute and print command. Put output into the LOGFILE. Check exit status.
ex()
{
	echo "Running: $@" >> $LOGFILE
	eval $@ >> $LOGFILE 2>&1
	if [ $? -ne 0 ]; then
	    err_echo "Failed to execute: $@"
	    echo "See $LOGFILE for more details"
	    echo
            cleanup
	    get_debug_info
	    exit 1
	fi
}


check_prev_install()
{
   # make sure we have the prefix dir is empty
   if [ -d ${prefix} ]; then
	if [ "$batch" == "no" ]; then
		echo "${prefix} already exists!"
		if ! test -x ${prefix}/uninstall.sh; then
			echo "Please remove installation from ${prefix} and rerun `basename $0`"
			exit 1
		fi
		echo -n "Remove currently installed components (${prefix}/uninstall.sh) (y/n) [y] ? "
		read force
	else
		force="yes"
	fi

	if [ "x$force" == "xy" ] \
		 || [ "x$force" == "xyes" ] \
		 || [ "x$force" == "xY" ] \
		 || [ "x$force" == "xYES" ] \
		 || [ "x$force" == "x" ]  ; then
		echo "Clean old installation at ${prefix}"
		echo "Running: yes | ${prefix}/uninstall.sh" >> $LOGFILE
		yes | ${prefix}/uninstall.sh >> $LOGFILE 2>&1
		if [ $? -ne 0 ]; then
			if [ "$batch" == "no" ]; then
				err_echo "${prefix}/uninstall.sh failed"
				exit 1
			else
				echo "Forcing uninstall (rm -rf ${prefix}):"
				ex "/bin/rm -rf ${prefix}"
			fi
		fi
	else
		echo "Aborting installation"
		exit 1
	fi

   fi

   if [ -d /usr/mellanox ]; then
	if [ "$batch" == "no" ]; then
		echo "Found /usr/mellanox!"
		if ! test -x /usr/mellanox/uninstall.sh; then
			echo "Please remove installation from /usr/mellanox and rerun `basename $0`"
			exit 1
		fi
		echo -n "Remove currently installed components (/usr/mellanox/uninstall.sh) (y/n) [y] ? "
		read force
	else
		force="yes"
	fi

	if [ "x$force" == "xy" ] \
		 || [ "x$force" == "xyes" ] \
		 || [ "x$force" == "xY" ] \
		 || [ "x$force" == "xYES" ] \
		 || [ "x$force" == "x" ]  ; then
		echo "Clean old distribution at /usr/mellanox"
		echo "Running: yes | /usr/mellanox/uninstall.sh" >> $LOGFILE
		yes | /usr/mellanox/uninstall.sh >> $LOGFILE 2>&1
		if [ $? -ne 0 ]; then
                        if [ "$batch" == "no" ]; then
                                err_echo "/usr/mellanox/uninstall.sh failed"
                                exit 1
                        else
				if [ "${prefix}" == "/usr/mellanox" ]; then
	                                echo "Forcing uninstall (rm -rf ${prefix}):"
        	                        ex "/bin/rm -rf ${prefix}"
				else
					echo "WARNING: Leaving /usr/mellanox. Continuing installation..."
				fi
                        fi
                fi
	else
		echo "Aborting installation"
		exit 1
	fi


   fi
}


set_deps()
{
    for pack in $PACKAGES
    do
	case $pack in
		ib_verbs)
		    verbs=y
		;;
		ib_ipoib)
		    verbs=y
		    ipoib=y
		;;
		ib_cm)
		    verbs=y
		    ipoib=y
		    cm=y
		;;
		ib_sdp)
		    verbs=y
		    ipoib=y
		    cm=y
		    sdp=y
		;;
		ib_dapl)
		    verbs=y
		    ipoib=y
		    cm=y
		    dapl=y
		;;
		ib_srp)
		    verbs=y
		    ipoib=y
		    cm=y
		    srp=y
		;;
		ib_srp_target)
		    verbs=y
		    ipoib=y
		    cm=y
		    srp_target=y
		;;
		*)
		    echo "Wrong package name: $pack"
		    usage
		    exit 2
		;;    
                
	esac
    done


    # Build MODULES_CONF_PARAMS

    if [ "$verbs" == "y" ]; then
	MODULES_CONF_PARAMS="$MODULES_CONF_PARAMS CONFIG_INFINIBAND=m CONFIG_INFINIBAND_MELLANOX_HCA=m"
    else
	MODULES_CONF_PARAMS="$MODULES_CONF_PARAMS CONFIG_INFINIBAND=n CONFIG_INFINIBAND_MELLANOX_HCA=n"
    fi
    
    if [ "$ipoib" == "y" ]; then
	MODULES_CONF_PARAMS="$MODULES_CONF_PARAMS CONFIG_INFINIBAND_IPOIB=m"
    else
	MODULES_CONF_PARAMS="$MODULES_CONF_PARAMS CONFIG_INFINIBAND_IPOIB=n"
    fi
    
    if [ "$cm" == "y" ]; then
	MODULES_CONF_PARAMS="$MODULES_CONF_PARAMS CONFIG_INFINIBAND_USER_CM=m"
    else
	MODULES_CONF_PARAMS="$MODULES_CONF_PARAMS CONFIG_INFINIBAND_USER_CM=n"
    fi
    
    if [ "$sdp" == "y" ]; then
	MODULES_CONF_PARAMS="$MODULES_CONF_PARAMS CONFIG_INFINIBAND_SDP=m"
    else
	MODULES_CONF_PARAMS="$MODULES_CONF_PARAMS CONFIG_INFINIBAND_SDP=n"
    fi
    
    if [ "$dapl" == "y" ]; then
	MODULES_CONF_PARAMS="$MODULES_CONF_PARAMS CONFIG_INFINIBAND_DAPL=m CONFIG_INFINIBAND_DAPL_SRV=m CONFIG_DAT=m CONFIG_INFINIBAND_KDAPL=m CONFIG_INFINIBAND_KDAPLTEST=m"
    else
	MODULES_CONF_PARAMS="$MODULES_CONF_PARAMS CONFIG_INFINIBAND_DAPL=n CONFIG_INFINIBAND_DAPL_SRV=n CONFIG_DAT=n CONFIG_INFINIBAND_KDAPL=n CONFIG_INFINIBAND_KDAPLTEST=n"
    fi
    
    if [ "$srp" == "y" ]; then
	MODULES_CONF_PARAMS="$MODULES_CONF_PARAMS CONFIG_INFINIBAND_SRP=m"
    else
	MODULES_CONF_PARAMS="$MODULES_CONF_PARAMS CONFIG_INFINIBAND_SRP=n"
    fi
    
    if [ "$srp_target" == "y" ]; then
	MODULES_CONF_PARAMS="$MODULES_CONF_PARAMS CONFIG_INFINIBAND_SRP_TARGET=m"
    else
	MODULES_CONF_PARAMS="$MODULES_CONF_PARAMS CONFIG_INFINIBAND_SRP_TARGET=n"
    fi
     
}


print_selected_packages()
{
	echo
	echo "The following packages will be installed:"
	echo
	[ "$verbs" == "y" ] 		&& echo ib_verbs
	[ "$ipoib" == "y" ] 		&& echo ib_ipoib
	[ "$cm" == "y" ] 		&& echo ib_cm
	[ "$sdp" == "y" ] 		&& echo ib_sdp
	[ "$dapl" == "y" ] 		&& echo ib_dapl
	[ "$srp" == "y" ] 		&& echo ib_srp
	[ "$srp_target" == "y" ] 	&& echo ib_srp_target
	
}

#################################################################################################
#                                       Create kernel build tree                                #
#################################################################################################
create_kernel_tree()
{
   
	echo
	echo Creating build tree under ${tmpdir}

        
        ex mkdir -p ${BUILD_IB_KERNEL_HOME}
        
	ex cp -p ${installdir}/Makefile* ${BUILD_IB_KERNEL_HOME}
        ex cp -p ${installdir}/Kconfig ${BUILD_IB_KERNEL_HOME}

	ex /bin/cp -prf ${installdir}/include ${BUILD_IB_KERNEL_HOME}

	if [ "$verbs" == "y" ]; then
            ex /bin/cp -prf ${installdir}/ib_verbs/kernel ${BUILD_IB_KERNEL_HOME}/ib_verbs
            # Update Kconfig
	    perl -i -ne 'if (m@^endmenu@) {

        		 print q@
config INFINIBAND
        tristate "InfiniBand support"
        default n
        depends on BROKEN || m
        ---help---
          Core support for InfiniBand (IB).  Make sure to also select
          any protocols you wish to use as well as drivers for your
          InfiniBand hardware.

config INFINIBAND_MELLANOX_HCA
        tristate "Mellanox HCA support"
        depends on PCI && INFINIBAND && (X86 || X86_64 || IA64)
        ---help---
          This driver supports Mellanox InfiniHost host channel
          adapters, including the MT23108 PCI-X HCA and the MT25208
          PCI Express HCA.

endmenu
@;
                 	 } else {
                     			print;
                 	 }'  ${BUILD_IB_KERNEL_HOME}/Kconfig

            # Update Makefile
            echo "obj-\$(CONFIG_INFINIBAND) += ib_verbs/" >> ${BUILD_IB_KERNEL_HOME}/Makefile.26
            perl -i -ne 'if (m@^include*@) {

        print q@
ifeq ($(CONFIG_INFINIBAND),m)
subdir-$(CONFIG_INFINIBAND)             += ib_verbs
endif
     
include $(TOPDIR)/Rules.make
@;
                 } else {
                     print;
                 }' ${BUILD_IB_KERNEL_HOME}/Makefile.24                    
	fi
        
	if [ "$ipoib" == "y" ]; then
            ex /bin/cp -prf ${installdir}/ib_ipoib/kernel ${BUILD_IB_KERNEL_HOME}/ib_ipoib
	    perl -i -ne 'if (m@^endmenu@) {

        		 print q@
config INFINIBAND_IPOIB
        tristate "IP-over-InfiniBand"
        depends on INFINIBAND && NETDEVICES && INET
        ---help---
          Support for the IP-over-InfiniBand protocol (IPoIB).  This
          transports IP packets over InfiniBand so you can use your IB
          device as a fancy NIC.  To configure interfaces on multiple
          InfiniBand partitions (P_Keys) you will need the ipoibcfg
          utility from <http://openib.org>.

endmenu
@;
                 	 } else {
                     			print;
                 	 }'  ${BUILD_IB_KERNEL_HOME}/Kconfig

            # Update Makefile
            echo "obj-\$(CONFIG_INFINIBAND_IPOIB) += ib_ipoib/" >> ${BUILD_IB_KERNEL_HOME}/Makefile.26
            perl -i -ne 'if (m@^include*@) {

        print q@
ifeq ($(CONFIG_INFINIBAND_IPOIB),m)
subdir-$(CONFIG_INFINIBAND_IPOIB)       += ib_ipoib
endif
     
include $(TOPDIR)/Rules.make
@;
                 } else {
                     print;
                 }' ${BUILD_IB_KERNEL_HOME}/Makefile.24                    

	fi
        
	if [ "$cm" == "y" ]; then
            ex /bin/cp -prf ${installdir}/ib_cm/kernel ${BUILD_IB_KERNEL_HOME}/ib_cm
	    perl -i -ne 'if (m@^endmenu@) {

        		 print q@
config INFINIBAND_USER_CM
        tristate "Userspace CM"
        depends on INFINIBAND && INFINIBAND_MELLANOX_HCA && INFINIBAND_IPOIB
        ---help---
          Userspace access to the kernel communication manager.
          This allows userspace protocols such as uDAPL or MPI to use
          the kernels services for establishing connections.

endmenu
@;
                 	 } else {
                     			print;
                 	 }'  ${BUILD_IB_KERNEL_HOME}/Kconfig

            # Update Makefile
            echo "obj-\$(CONFIG_INFINIBAND_USER_CM) += ib_cm/" >> ${BUILD_IB_KERNEL_HOME}/Makefile.26
            perl -i -ne 'if (m@^include*@) {

        print q@
ifeq ($(CONFIG_INFINIBAND_USER_CM),m)
subdir-$(CONFIG_INFINIBAND_USER_CM)     += ib_cm
endif
     
include $(TOPDIR)/Rules.make
@;
                 } else {
                     print;
                 }' ${BUILD_IB_KERNEL_HOME}/Makefile.24                    

	fi
        
	if [ "$sdp" == "y" ]; then
            ex /bin/cp -prf ${installdir}/ib_sdp/kernel ${BUILD_IB_KERNEL_HOME}/ib_sdp
	    perl -i -ne 'if (m@^endmenu@) {

        		 print q@
config INFINIBAND_SDP
        tristate "Sockets Direct Protocol"
        depends on INFINIBAND && INFINIBAND_IPOIB && INFINIBAND_USER_CM
        ---help---
          Support for Sockets Direct Protocol (SDP).  This provides
          sockets semantics over InfiniBand via address family
          AF_INET_SDP (address family 26).  You can also LD_PRELOAD the
          libsdp library from <http://openib.org> to have standard
          sockets applications use SDP.

endmenu
@;
                 	 } else {
                     			print;
                 	 }'  ${BUILD_IB_KERNEL_HOME}/Kconfig

            # Update Makefile
            echo "obj-\$(CONFIG_INFINIBAND_SDP) += ib_sdp/" >> ${BUILD_IB_KERNEL_HOME}/Makefile.26
            perl -i -ne 'if (m@^include*@) {

        print q@
ifeq ($(CONFIG_INFINIBAND_SDP),m)
subdir-$(CONFIG_INFINIBAND_SDP)         += ib_sdp
endif
     
include $(TOPDIR)/Rules.make
@;
                 } else {
                     print;
                 }' ${BUILD_IB_KERNEL_HOME}/Makefile.24                    
           
	fi
        
	if [ "$dapl" == "y" ]; then
            ex /bin/cp -prf ${installdir}/ib_dapl/kernel ${BUILD_IB_KERNEL_HOME}/ib_dapl
	    perl -i -ne 'if (m@^endmenu@) {

        		 print q@
config INFINIBAND_DAPL
        tristate "InfiniBand DAPL"
        depends on INFINIBAND && INFINIBAND_IPOIB && INFINIBAND_MELLANOX_HCA && INFINIBAND_USER_CM
        ---help---
          Kernel InfiniBand support for the DAT API.

config DAT
        tristate "DAT Registry Module"
        depends on INFINIBAND_DAPL
        ---help---
          Support for the Direct Acess Transport (DAT) API, a programming
          interface for Remote Direct Memory Access (RDMA) transports
          such as InfiniBand and iWARP.

config INFINIBAND_DAPL_SRV
        tristate "DAPL helper module"
        depends on INFINIBAND_DAPL
        ---help---
          Dapl helper module. Need for both udapl and kdapl

config INFINIBAND_KDAPL
        tristate "InfiniBand DAT Provider"
        depends on INFINIBAND_DAPL && INFINIBAND_DAPL_SRV
        ---help---
          Kernel InfiniBand support for the DAT API.

config INFINIBAND_KDAPLTEST
        tristate "kdapltest module"
        depends on INFINIBAND_KDAPL
        ---help---
          kdapltest module. In order to run need a user level part also.
            
endmenu
@;
                 	 } else {
                     			print;
                 	 }'  ${BUILD_IB_KERNEL_HOME}/Kconfig

            # Update Makefile
            echo "obj-\$(CONFIG_INFINIBAND_DAPL) += ib_dapl/" >> ${BUILD_IB_KERNEL_HOME}/Makefile.26
            perl -i -ne 'if (m@^include*@) {

        print q@
ifeq ($(CONFIG_INFINIBAND_DAPL),m)
subdir-$(CONFIG_INFINIBAND_DAPL)        += ib_dapl
endif
     
include $(TOPDIR)/Rules.make
@;
                 } else {
                     print;
                 }' ${BUILD_IB_KERNEL_HOME}/Makefile.24                    
           
	fi
        
	if [ "$srp" == "y" ]; then
            ex /bin/cp -prf ${installdir}/ib_srp/kernel/srp-$KVERSION_TWO_DIGITS ${BUILD_IB_KERNEL_HOME}/ib_srp
	    perl -i -ne 'if (m@^endmenu@) {

        		 print q@
config INFINIBAND_SRP
        tristate "SCSI RDMA Protocol"
        depends on INFINIBAND && INFINIBAND_USER_CM && SCSI
        ---help---
          Support for SCSI RDMA Protocol (SRP).  This transports SCSI
          commands over InfiniBand and allows you to access storage
          connected via IB.

endmenu
@;
                 	 } else {
                     			print;
                 	 }'  ${BUILD_IB_KERNEL_HOME}/Kconfig

            # Update Makefile
            echo "obj-\$(CONFIG_INFINIBAND_SRP) += ib_srp/" >> ${BUILD_IB_KERNEL_HOME}/Makefile.26
            perl -i -ne 'if (m@^include*@) {

        print q@
ifeq ($(CONFIG_INFINIBAND_SRP),m)
subdir-$(CONFIG_INFINIBAND_SRP)         += ib_srp
endif
     
include $(TOPDIR)/Rules.make
@;
                 } else {
                     print;
                 }' ${BUILD_IB_KERNEL_HOME}/Makefile.24                    

	fi
        
	if [ "$srp_target" == "y" ]; then
	    if [ -d ${installdir}/ib_srp_target/kernel ]; then
		
            ex /bin/cp -prf ${installdir}/ib_srp_target/kernel ${BUILD_IB_KERNEL_HOME}/ib_srp_target
	    perl -i -ne 'if (m@^endmenu@) {

        		 print q@
config INFINIBAND_SRP_TARGET
        tristate "SRP Target Emulation"
        depends on INFINIBAND && SCSI
        ---help---
          Support for SRP Target Emulation.
          This emulates an Infiniband target and receives SCSI commands
          over InfiniBand.

endmenu
@;
                 	 } else {
                     			print;
                 	 }'  ${BUILD_IB_KERNEL_HOME}/Kconfig
           
            # Update Makefile
            echo "obj-\$(CONFIG_INFINIBAND_SRP_TARGET) += ib_srp_target/" >> ${BUILD_IB_KERNEL_HOME}/Makefile.26
            perl -i -ne 'if (m@^include*@) {

        print q@
ifeq ($(CONFIG_INFINIBAND_SRP_TARGET),m)
subdir-$(CONFIG_INFINIBAND_SRP_TARGET)  += ib_srp_target
endif
     
include $(TOPDIR)/Rules.make
@;
                 } else {
                     print;
                 }' ${BUILD_IB_KERNEL_HOME}/Makefile.24                    
	
	    else
		echo "No srp_target sources found..."
	    fi
	fi

        ex cd ${BUILD_IB_KERNEL_HOME}
        find . -name Makefile -type l | xargs /bin/rm -f
    
	MF=`find . -name Makefile.$KVERSION_TWO_DIGITS`
    
	# Take Makefiles according to the kernel version
	for mf in $MF; do
	    ex mv $mf `dirname $mf`/Makefile
	done
        
	find . -name Makefile.2* -exec /bin/rm -f {} \;

	# Save Gen2 sources if exist
	if test -d ${BUILD_ROOT}/${IB_KERNEL_HOME}/hw/mthca; then
            mv ${BUILD_ROOT}/${IB_KERNEL_HOME} ${BUILD_ROOT}/${IB_KERNEL_HOME}-gen2
        else    
	    /bin/rm -rf ${BUILD_ROOT}/${IB_KERNEL_HOME}
	fi
                        
	if [ ! -d ${BUILD_ROOT}/${KSRC}/drivers ]; then
		ex mkdir -p ${BUILD_ROOT}/${KSRC}/drivers
	fi
	ex /bin/cp -prf ${BUILD_IB_KERNEL_HOME} ${BUILD_ROOT}/${KSRC}/drivers/
   


}


#################################################################################################
#                                       Build ib kernel modules                                 #
#################################################################################################
build_kernel_modules()
{
	printf "\nBuilding InfiniBand kernel modules\n"

	# Set VAPI_KERNEL_SRC_DIR to the temporary build directory
	export IB_SRC_DIR=${BUILD_IB_KERNEL_HOME}
	export VAPI_KERNEL_SRC_DIR=${BUILD_IB_KERNEL_HOME}/ib_verbs/hw/mellanox-hca

	# Define default EXTRAVERSION
	EXTRAVERSION=$(echo $KVERSION | sed 's/^[0-9]\+\.[0-9]\+\.[0-9]\+//')
        
	# Set EXTRAVERSION for old SuSE releases
	if [[ -f /etc/SuSE-release && "${KVERSION_TWO_DIGITS}" == "24" ]]; then
		if ! ( /bin/ls -l /lib/modules/${KVERSION}/build | grep include > /dev/null ); then
			EXTRAVERSION=""
		fi
	fi
        
	# Set EXTRAVERSION for Rocks
	if [ -f /etc/rocks-release ]; then
		EXTRAVERSION=$(echo $EXTRAVERSION | sed 's/smp//')
	fi

	if [ "${MEMTRACK}" == "yes" ]; then
		if [ "${KVERSION_TWO_DIGITS}" == "24" ]; then
			echo "MEMTRACK is not supported on kernels 2.4.xx"
			MEMTRACK="no"
		else
			/bin/cp -p ${BUILD_IB_KERNEL_HOME}/ib_verbs/hw/mellanox-hca/include/memtrack.h ${BUILD_IB_KERNEL_HOME}/include
		fi
	fi


	###########################################################################################################################
	###                                             Kernel 2.4 preconfig                                                    ###
	###########################################################################################################################
	if [ "${KVERSION_TWO_DIGITS}" == "24" ]; then
		 #We need to make oldconfig only if it is not a new SuSE release or if .config does not exists
		 if ! ( /bin/ls -l /lib/modules/${KVERSION}/build | grep include > /dev/null ); then
		     if ! [ -f /lib/modules/${KVERSION}/build/.config ]; then
			printf "\n Running Mrproper"
			ex make -C ${MAKE_KERNEL_DIR} mrproper
                        printf "\r\t\t\t\t\t (done)"
		     fi
		     printf "\n Making oldconfig ..."
		     ex make -C ${MAKE_KERNEL_DIR} CONFIG_INFINIBAND=n oldconfig
                     printf "\r\t\t\t\t\t (done)"
                             
                     printf "\n Making dep ..."
		     ex make -C ${MAKE_KERNEL_DIR} ${MODULES_CONF_PARAMS} KERNELRELEASE=${KVERSION} EXTRAVERSION=${EXTRAVERSION} dep
                     printf "\r\t\t\t\t\t (done)"
		 fi
        
		 # Make dep with SUBDIRS
                 printf "\n Making dep with subdirs ..."
		 ex make -C ${MAKE_KERNEL_DIR} SUBDIRS=${BUILD_IB_KERNEL_HOME} ${MODULES_CONF_PARAMS} KERNELRELEASE=${KVERSION} EXTRAVERSION=${EXTRAVERSION} dep
                 printf "\r\t\t\t\t\t (done)"
	fi

	###########################################################################################################################
	###                                     Modules Compilation and Installation                                            ###
	###########################################################################################################################
        
	if [ "${KVERSION_TWO_DIGITS}" == "24" ]; then
        
		printf "\n Making modules ..."
		ex make -C ${MAKE_KERNEL_DIR} SUBDIRS=${BUILD_IB_KERNEL_HOME} ${MODULES_CONF_PARAMS} KERNELRELEASE=${KVERSION} EXTRAVERSION=${EXTRAVERSION} ${MAKE_PARAMS} modules
                printf "\r\t\t\t\t\t (done)"
        
		printf "\n Installing modules ..."
		ex make -C ${MAKE_KERNEL_DIR} SUBDIRS=${BUILD_IB_KERNEL_HOME} ${MODULES_CONF_PARAMS} KERNELRELEASE=${KVERSION} EXTRAVERSION=${EXTRAVERSION} INSTALL_MOD_PATH=${tmpdir} ${MAKE_PARAMS} modules_install
                printf "\r\t\t\t\t\t (done)"
        
	else # Kernel 2.6
       
		if [ "${MEMTRACK}" == "yes" ]; then
		    export EXTRA_CFLAGS="-I ${BUILD_IB_KERNEL_HOME}/include -include mtrack.h"
		fi

		printf "\n Making modules ..."
		ex make -C ${MAKE_KERNEL_DIR} SUBDIRS=${BUILD_IB_KERNEL_HOME} ${MODULES_CONF_PARAMS} KERNELRELEASE=${KVERSION} EXTRAVERSION=${EXTRAVERSION} ${MAKE_PARAMS} modules
                printf "\r\t\t\t\t\t (done)"
                        
		printf "\n Installing modules ..."
		ex make -C ${MAKE_KERNEL_DIR} SUBDIRS=${BUILD_IB_KERNEL_HOME} ${MODULES_CONF_PARAMS} KERNELRELEASE=${KVERSION} EXTRAVERSION=${EXTRAVERSION} INSTALL_MOD_PATH=${tmpdir} ${MAKE_PARAMS} modules_install
                printf "\r\t\t\t\t\t (done)"

	fi

}

install_kernel_modules()
{

    if [ "${KVERSION_TWO_DIGITS}" == "24" ]; then
	    if ! test -d ${BUILD_ROOT}/lib/modules/${KVERSION}/kernel/drivers; then
		ex mkdir -p ${BUILD_ROOT}/lib/modules/${KVERSION}/kernel/drivers
	    else
		ex /bin/rm -rf ${BUILD_ROOT}/lib/modules/${KVERSION}/kernel/drivers/infiniband    
	    fi    
            
	    if test -d ${tmpdir}/lib/modules/${KVERSION}/kernel/drivers/infiniband; then
		    ex mv ${tmpdir}/lib/modules/${KVERSION}/kernel/drivers/infiniband ${BUILD_ROOT}/lib/modules/${KVERSION}/kernel/drivers/
	    elif test -d ${tmpdir}/lib/modules/${KVERSION}/kernel/Lost; then
		    ex mv ${tmpdir}/lib/modules/${KVERSION}/kernel/Lost ${BUILD_ROOT}/lib/modules/${KVERSION}/kernel/drivers/infiniband
	    else
		    ex mkdir ${BUILD_ROOT}/lib/modules/${KVERSION}/kernel/drivers/infiniband
		    find ${tmpdir}/lib/modules/${KVERSION}/kernel -name '*.o' -exec install -m 0644 {} ${BUILD_ROOT}/lib/modules/${KVERSION}/kernel/drivers/infiniband \;
	    fi
    else
            if [ $SET_ENV -eq 1 ]; then
                # Save Gen2 modules
                if [ -d /lib/modules/${KVERSION}/kernel/drivers/infiniband/hw/mthca/ib_mthca.ko ]; then
                	find /lib/modules/${KVERSION}/kernel/drivers/infiniband -name "*.ko" -exec mv {} {}.gen2 \;
                	mv /lib/modules/${KVERSION}/kernel/drivers/infiniband /lib/modules/${KVERSION}/kernel/drivers/infiniband-gen2 
                else
                	rm -rf %{LIB_MOD_DIR}
                fi
            fi    

	    ex /bin/rm -rf ${BUILD_ROOT}/lib/modules/${KVERSION}/kernel/drivers/infiniband
	    ex /bin/rm -rf ${BUILD_ROOT}/lib/modules/${KVERSION}/extra/infiniband
	    ex mkdir -p ${BUILD_ROOT}/lib/modules/${KVERSION}/kernel/drivers/infiniband
	    find ${tmpdir}/lib/modules/${KVERSION}/extra -name '*.ko' -exec install -m 0644 {} ${BUILD_ROOT}/lib/modules/${KVERSION}/kernel/drivers/infiniband \;
    fi

}

build_user_level()
{


	printf "\nBuilding InfiniBand user level\n"

	mkdir -p ${BUILD_ROOT}/${prefix}/bin
	mkdir -p ${BUILD_ROOT}/${prefix}/utils
	mkdir -p ${BUILD_ROOT}/${prefix}/src
	cp ${installdir}/uninstall.sh ${BUILD_ROOT}/${prefix}
	cp ${installdir}/BUILD_ID ${BUILD_ROOT}/${prefix}
	chmod a+x ${BUILD_ROOT}/${prefix}/uninstall.sh

	mkdir -p ${tmpdir}/user

	if [ "$verbs" == "y" ]; then
            printf "\n Making ib_verbs (user)..."
	    ex /bin/cp -prf ${installdir}/ib_verbs/user ${tmpdir}/user/ib_verbs
	    ex /bin/cp -prf ${installdir}/ib_verbs/user ${BUILD_ROOT}/${prefix}/src/ib_verbs

	    find ${BUILD_ROOT}/${prefix}/src/ib_verbs -name Rules.Make | xargs perl -ni -e "s@^(MT_INSTALL_PREFIX \?=).*@\$1 ${prefix}/driver/infinihost@; print"
            
	    if [ "$STATIC_LIB" == "yes" ]; then
	       if ( /sbin/lspci -n | grep -E "15b3:6282|15b3:5e8c|15b3:6274" > /dev/null ); then # Found Arbel in Arbel mode or Sinai
		    export MT_STATIC_LIB=1;
		    export ARBEL_LIB=1;
		    echo "MT_STATIC_LIB_MEMFREE=yes" >> ${BUILD_ROOT}/${prefix}/BUILD_ID
            
	       elif ( /sbin/lspci -n | grep -E "15b3:5a44|15b3:6278" > /dev/null ); then # Tavor mode
		    export MT_STATIC_LIB=1;
		    export TAVOR_LIB=1;
		    echo "MT_STATIC_LIB_TAVOR=yes" >> ${BUILD_ROOT}/${prefix}/BUILD_ID
	       else
		    export MT_STATIC_LIB=1;
		    export ARBEL_LIB=1;
		    echo "MT_STATIC_LIB_MEMFREE=yes" >> ${BUILD_ROOT}/${prefix}/BUILD_ID
	       fi
	    fi

	    export VAPI_KERNEL_SRC_DIR=${tmpdir}/${IB_KERNEL_HOME}/ib_verbs/hw/mellanox-hca
	    export MT_INSTALL_PREFIX=${BUILD_ROOT}/${prefix}/driver/infinihost
	    ex cd ${tmpdir}/user/ib_verbs/mellanox-hca                                                               
	    ex make VAPI_KERNEL_SRC_DIR=${BUILD_IB_KERNEL_HOME}/ib_verbs/hw/mellanox-hca MT_INSTALL_PREFIX=${BUILD_ROOT}/${prefix}/driver/infinihost PREFIX=${BUILD_ROOT}/${prefix}/driver/infinihost install
	    install -m 0755 ${installdir}/utils/create_devs ${BUILD_ROOT}/${prefix}/utils

	    # Make links from /usr/bin
	    if [ $SET_ENV -eq 1 ]; then
		ex cd /usr/bin
		for bin in ${prefix}/driver/infinihost/bin/*
		do
		    [ "$bin" != "${prefix}/driver/infinihost/bin/*" ] && ex ln -snf ${bin}
		done

		#for bin in ${prefix}/bin/*
		#do
		#    [ "$bin" != "${prefix}/bin/*" ] && ex ln -snf ${bin}
		#done

		cd -
	    fi

            printf "\r\t\t\t\t\t (done)"
	fi

	if [ "$ipoib" == "y" ]; then
	    install -m 0755 ${installdir}/utils/update_ipoib_intfs ${BUILD_ROOT}/${prefix}/utils
	fi
        
	if [ "$cm" == "y" ]; then
            printf "\n Making cmlib..."
	    ex /bin/cp -prf ${installdir}/ib_cm/user ${tmpdir}/user/ib_cm
	    ex /bin/cp -prf ${installdir}/ib_cm/user ${BUILD_ROOT}/${prefix}/src/ib_cm
            
	    ex cd ${tmpdir}/user/ib_cm/cmlib
	    ex make OPENIB_USER_DIR=${BUILD_ROOT}/$prefix KERNEL_SRC_DIR=${BUILD_ROOT}/$KSRC install
            printf "\r\t\t\t\t\t (done)"
	fi
        
	if [ "$sdp" == "y" ]; then
            printf "\n Making sdp (user)..."
	    ex /bin/cp -prf ${installdir}/ib_sdp/user ${tmpdir}/user/ib_sdp
	    ex /bin/cp -prf ${installdir}/ib_sdp/user ${BUILD_ROOT}/${prefix}/src/ib_sdp
            
	    ex cd ${tmpdir}/user/ib_sdp/libsdp

	    if [ "`uname -m`" == "x86_64"  ]; then
                  # Making IBGD libsdp in ${tmpdir} (32 bit version )
		  export CFLAGS=-m32
		  export CXXFLAGS=-m32
		  export FFLAGS=-m32
		  export LDFLAGS="-m32 -Xcompiler -m32"
		  ./configure --prefix=${BUILD_ROOT}/${prefix} >> $LOGFILE 2>&1
		  if [ $? -ne 0 ]; then
			 echo
			 echo "Configure for a 32 bit version failed. For more information look at: $LOGFILE"
			 echo "Skipping libsdp 32 bit version."
		  else
			 ex make install
		  fi
		  ex make clean
                
		  export CFLAGS=-m64
		  export CXXFLAGS=-m64
		  export FFLAGS=-m64
		  export LDFLAGS="-m64 -Xcompiler -m64"
                
		  # Making IBGD libsdp in ${tmpdir} (64 bit version )
		  ex ./configure --prefix=${BUILD_ROOT}/${prefix} --libdir=${BUILD_ROOT}/${prefix}/lib64
		  ex make install
	    else
		  # Making IBGD libsdp in ${tmpdir}
		  ex ./configure --prefix=${BUILD_ROOT}/${prefix}
		  ex make install
	    fi

            printf "\r\t\t\t\t\t (done)"
	fi
        
	if [ "$dapl" == "y" ]; then
                printf "\n Making dapl (user)..."
		ex /bin/cp -prf ${installdir}/ib_dapl/user ${tmpdir}/user/ib_dapl
		ex /bin/cp -prf ${installdir}/ib_dapl/user ${BUILD_ROOT}/${prefix}/src/ib_dapl
                
		ex cd ${tmpdir}/user/ib_dapl/dapl_new
		if [ ! -d ${BUILD_ROOT}/$prefix/etc ]; then
			ex mkdir -p ${BUILD_ROOT}/$prefix/etc
		fi
		ex make PREFIX=${BUILD_ROOT}/$prefix OPENIB_USER_DIR=${BUILD_ROOT}/$prefix DAT_CONF_TARGET=${BUILD_ROOT}/$prefix/etc/ KERNEL_SRC_DIR=${BUILD_ROOT}/$KSRC OPENIB_KERNEL_DIR=${BUILD_IB_KERNEL_HOME} install
                
		install -m 0755 ${installdir}/utils/DaplRegress.sh ${BUILD_ROOT}/${prefix}/utils

		perl -ni -e "s@/usr/local/ibgd@${prefix}@;print" ${BUILD_ROOT}/${prefix}/etc/dat.conf
                
		if [ $SET_ENV -eq 1 ]; then
			if [ -x ${prefix}/bin/dapltest ]; then
				ln -snf ${prefix}/bin/dapltest /usr/bin/dapltest
			fi

			if [ -x ${prefix}/bin/kdapltest ]; then
				ln -snf ${prefix}/bin/kdapltest /usr/bin/kdapltest
			fi
			mv -f ${prefix}/etc/dat.conf /etc/dat.conf
			if ( /sbin/lspci -n | grep "15b3:6282" > /dev/null ); then #Found Arbel memfree
			     perl -ni -e  "s@InfiniHost0@InfiniHost_III_Ex0@;print" /etc/dat.conf
			fi
			if ( /sbin/lspci -n | grep -E "15b3:5e8c|15b3:6274" > /dev/null ); then #Found Sinai
			     perl -ni -e  "s@InfiniHost0@InfiniHost_III_Lx0@;print" /etc/dat.conf
			fi
		fi
                printf "\r\t\t\t\t\t (done)"
	fi

	if [ "$srp" == "y" ]; then
                printf "\n Making scsidev..."
		ex /bin/cp -prf ${installdir}/ib_srp/user ${tmpdir}/user/ib_srp
                ex /bin/cp -prf ${installdir}/ib_srp/user ${BUILD_ROOT}/${prefix}/src/ib_srp
		ex cd ${tmpdir}/user/ib_srp/scsidev
		mkdir -p ${BUILD_ROOT}/${prefix}/man/man8
                
		# Making scsidev in ${tmpdir}
		ex ./configure --quiet --prefix=${BUILD_ROOT}/$prefix --exec-prefix=${BUILD_ROOT}/$prefix
		ex make prefix=${BUILD_ROOT}/$prefix exec_prefix=${BUILD_ROOT}/$prefix install

		install -m 0755 ${installdir}/utils/srp_persistent_bind.sh ${BUILD_ROOT}/${prefix}/utils
		install -m 0755 ${installdir}/utils/remove_srp_bind_persistent.sh  ${BUILD_ROOT}/${prefix}/utils
		install -m 0755 ${installdir}/utils/rescan-scsi-bus.sh  ${BUILD_ROOT}/${prefix}/utils
		install -m 0644 ${installdir}/utils/srp_persistent_bind.conf  ${BUILD_ROOT}/${prefix}/utils

		# Update srp configuration scripts
		perl -ni -e "s@/usr/local/ibgd@${prefix}@;print" ${BUILD_ROOT}/${prefix}/utils/srp_persistent_bind.sh
		perl -ni -e "s@/usr/local/ibgd@${prefix}@;print" ${BUILD_ROOT}/${prefix}/utils/remove_srp_bind_persistent.sh
                
                printf "\r\t\t\t\t\t (done)"
	fi

#         # No user level for srp_target
#         if [ "$srp_target" == "y" ]; then
#         fi

	if [ $SET_ENV -eq 1 ]; then

		if [ -d  ${prefix}/driver/infinihost/lib ]; then
			perl -i.orig -ne "print unless (m@${prefix}/driver/infinihost/lib@)" /etc/ld.so.conf
			echo "${prefix}/driver/infinihost/lib" >> /etc/ld.so.conf
		fi
             
		if [ -d ${prefix}/driver/infinihost/lib64 ]; then
			perl -i.orig -ne "print unless (m@${prefix}/driver/infinihost/lib64@)" /etc/ld.so.conf
			echo "${prefix}/driver/infinihost/lib64" >> /etc/ld.so.conf
		fi

		if [ -d  ${prefix}/lib ]; then
			perl -i.orig -ne "print unless (m@${prefix}/lib@)" /etc/ld.so.conf
			echo "${prefix}/lib" >> /etc/ld.so.conf
		fi        
        
		if [ -d ${prefix}/lib64 ]; then
			perl -i.orig -ne "print unless (m@${prefix}/lib64@)" /etc/ld.so.conf
			echo "${prefix}/lib64" >> /etc/ld.so.conf
		fi

		ex /sbin/ldconfig
                
		# Take care of libsdp.conf
		/bin/rm -f /etc/libsdp.conf
	 fi
}

count_ib_ports()
{
    local cnt=0
    local tmp_cnt=0
    
    tmp_cnt=$(/sbin/lspci -n | grep "15b3:6282" | wc -l | tr -d '[:space:]') # Arbel mode
    cnt=$[ $cnt + 2*${tmp_cnt} ]
    
    tmp_cnt=$(/sbin/lspci -n | grep -E "15b3:5e8c|15b3:6274" | wc -l | tr -d '[:space:]') # Sinai
    cnt=$[ $cnt + ${tmp_cnt} ]

    tmp_cnt=$(/sbin/lspci -n | grep -E "15b3:5a44|15b3:6278" | wc -l | tr -d '[:space:]') # Tavor mode
    cnt=$[ $cnt + 2*${tmp_cnt} ]
    
    return $cnt
}

ipoib_config()
{
	# Network configuration
	if [ -f /etc/redhat-release ]; then
	    DISTRIBUTION="RedHat"
	elif [ -f /etc/rocks-release ]; then
	    DISTRIBUTION="Rocks"
	elif [ -f /etc/SuSE-release ]; then
	    DISTRIBUTION="SuSE"
	else
	    DISTRIBUTION=`ls /etc/*-release | head -n 1 | xargs -iXXX basename XXX -release 2> /dev/null`
	fi

	INTERFACES=`/sbin/ifconfig | grep eth | awk '{print$1}' | sort -n | tr -s '\n' '\ '`
	for INTERFACE in $INTERFACES
	do
		ADDRS=`/sbin/ifconfig $INTERFACE | grep "inet addr"`
		if [[ -n $ADDRS ]]; then
		      LAN_INTERFACE=$INTERFACE
		      break
		fi
	done

	LAN_INTERFACE=${LAN_INTERFACE:-eth0}


	count_ib_ports
	ports_num=$?

	DEVICES=""
                           
	for (( i=0 ; i < $ports_num ; i++ ))
	do
	    DEVICES="$DEVICES ib${i}"
	done    
        
	# Set default number of devices in case that no IB device currently installed on the system
	[ "$DEVICES" == "" ] && DEVICES="ib0 ib1"

	high_IP=`/sbin/ifconfig | sed -e "/\($LAN_INTERFACE\).*/{ N; s/.*inet addr:\([0-9]*\).*/\1/; }" | grep ^[0-9] | head -1`
	let high_IP++
	low_IP=`/sbin/ifconfig | sed -e "/\($LAN_INTERFACE\).*/{ N; s/.*inet addr:[0-9]*\.\([0-9]*\.[0-9]*\.[0-9]*\).*/\1/; }" | grep ^[0-9] | head -1`
	IPADDR=${high_IP}.${low_IP}

	NETMASK=`/sbin/ifconfig | sed -e "/\($LAN_INTERFACE\).*/{ N; s/.*Mask:\([0-9]*\.[0-9]*\.[0-9]*\.[0-9]*\).*/\1/; }" | grep ^[0-9] | head -1`
	low_Bcast=`/sbin/ifconfig | sed -e "/\($LAN_INTERFACE\).*/{ N; s/.*Bcast:[0-9]*\.\([0-9]*\.[0-9]*\.[0-9]*\).*/\1/; }" | grep ^[0-9] | head -1`
	BROADCAST=${high_IP}.${low_Bcast}
	ONBOOT=1
	case "$NETMASK" in
		"255.255.255.0")
		    NETWORK=${IPADDR%*.*}.0
		;;
		"255.255.0.0")
		    NETWORK=${IPADDR%*.*.*}.0.0
		;;
		"255.0.0.0")
		    NETWORK=${IPADDR%*.*.*.*}.0.0.0
		;;
		*)
		    NETWORK=${IPADDR%*.*.*.*}.0.0.0
		;;
	esac

	IP_A=${IPADDR%*.*.*.*}
	IP_BCD=${IPADDR#*.}
	NET_BCD=${NETWORK#*.}
	BR_BCD=${BROADCAST#*.}

	for DEVICE in $DEVICES
	do
		target=${IB_CONFIG_DIR}/ifcfg-${DEVICE}
		cat $NULL > $target

		if [ "$DISTRIBUTION" == "SuSE" ]; then
		    echo BOOTPROTO=\'static\' >> $target
		    echo IPADDR=\'${IP_A}.${IP_BCD}\' >> $target
		    echo NETMASK=\'$NETMASK\' >> $target
		    echo NETWORK=\'${IP_A}.${NET_BCD}\' >> $target
		    echo BROADCAST=\'${IP_A}.${BR_BCD}\' >> $target
		    echo REMOTE_IPADDR=\'\' >> $target
		    if [ "$ONBOOT" == "1" ]; then
			echo STARTMODE=\'onboot\' >> $target
		    else
			echo STARTMODE=\'manual\' >> $target
		    fi
		    echo WIRELESS=\'no\' >> $target
		else
		    echo DEVICE=$DEVICE >> $target
		    echo BOOTPROTO=static >> $target
		    echo IPADDR=${IP_A}.${IP_BCD} >> $target
		    echo NETMASK=$NETMASK >> $target
		    echo NETWORK=${IP_A}.${NET_BCD} >> $target
		    echo BROADCAST=${IP_A}.${BR_BCD} >> $target
		    if [ "$ONBOOT" == "1" ]; then
			echo ONBOOT=yes >> $target
		    else
			echo ONBOOT=no >> $target
		    fi
		fi

		let IP_A++

	done


}

modules_conf()
{

	# Get nuber of IB ports in order to set ipoib configuration
	count_ib_ports
	ports_num=$?
	# Set default number of ports to 2 (In case that no IB devices currently installed)
	if [ $ports_num -eq 0 ]; then
		ports_num=2
	fi

    if [ "${KVERSION_TWO_DIGITS}" == "26" ]; then
	  printf "\n Updating /etc/modprobe.conf ..."
	if [ "x$(uname -m)" == "xx86_64" ] && cat /proc/cpuinfo | grep 'Intel' > /dev/null ; then
		disable_msix=0
	else
		disable_msix=1
	fi

	# Remove previous configuration if exist
        perl -ni -e 'if (/\#\# BEGIN OpenIB \#\#$/) { $filter = 1 }' -e 'if (!$filter) { print }' -e 'if (/\#\# END OpenIB \#\#$/){ $filter = 0 }' /etc/modprobe.conf

cat >> /etc/modprobe.conf << EOF
## BEGIN OpenIB ##
#
# InfiniBand module loading configuration
#

EOF

for (( i=0 ; i < $ports_num ; i++ ))
do
cat >> /etc/modprobe.conf << EOF
alias ib${i} ib_ipoib
EOF
done

cat >> /etc/modprobe.conf << EOF

alias net-pf-26 ib_sdp
alias kdapl dat_registry
alias kdapltest kdapltest_module
alias /dev/mosal mlxsys
alias /dev/vipkl mod_vip

options mod_thh disable_msix=${disable_msix}
options mod_rhh disable_msix=${disable_msix}
options ib_mad ib_outs_mad_recv_entries=1024
options ib_client_query ib_query_throttle_threshhold=0 ib_query_throttle_waitqueue_limit=10000
options ib_ipoib ipoib_use_interrupt_context=0

# mod_thh/mod_rhh
install mod_thh /usr/bin/mst start > /dev/null 2>&1 ; /sbin/modprobe -q --ignore-install mod_thh
install mod_rhh /usr/bin/mst start > /dev/null 2>&1 ; /sbin/modprobe -q --ignore-install mod_rhh

# mlxsys
install mlxsys /sbin/modprobe -q --ignore-install mlxsys && { rm -f /dev/mosal; [ "x\$(awk '(\$2=="mosal") {print \$1}' /proc/devices)" != "x" ] && mknod /dev/mosal c \`awk '(\$2=="mosal") {print \$1}' /proc/devices\` 0; }
remove mlxsys /sbin/modprobe -r -q --ignore-remove mlxsys && { /bin/rm -f /dev/mosal; }

# mod_vapi
install mod_vip /sbin/modprobe -q --ignore-install mod_vip && { rm -f /dev/vipkl; mknod /dev/vipkl c \`awk '(\$2=="mellanox_vipkl") {print \$1}' /proc/devices\` 1; }
remove mod_vip /sbin/modprobe -r -q --ignore-remove mod_vip && { /bin/rm -f /dev/vipkl; }

# Close HCAs after removing mod_thh/mod_rhh
remove mod_thh close_hcas && /sbin/modprobe -r -q --ignore-remove mod_thh
remove mod_rhh close_hcas && /sbin/modprobe -r -q --ignore-remove mod_rhh

install ib_ipoib /sbin/modprobe -i ib_ipoib && ${prefix}/utils/update_ipoib_intfs
install ib_ip2pr /sbin/modprobe -i ib_ip2pr && ${prefix}/utils/create_devs ib_ip2pr

install ib_dapl_srv /sbin/modprobe -i ib_dapl_srv && ${prefix}/utils/create_devs ib_udapl

install ib_useraccess /sbin/modprobe -i ib_useraccess && ${prefix}/utils/create_devs ib_useraccess
install ib_useraccess_cm /sbin/modprobe -i ib_useraccess_cm && ${prefix}/utils/create_devs ib_useraccess_cm


install kdapltest_module /sbin/modprobe -i kdapltest_module && ${prefix}/utils/create_devs kdapltest_module

# SRP
# install ib_srp /sbin/modprobe -i ib_srp && { ${prefix}/utils/srp_persistent_bind.sh & }
# remove ib_srp ${prefix}/utils/remove_srp_bind_persistent.sh && /sbin/modprobe -r -q --ignore-remove ib_srp
## END OpenIB ##
EOF

#	  perl -i.orig -ne 'print unless (m@include /etc/modprobe-openib.conf@)' /etc/modprobe.conf
#	  echo "include /etc/modprobe-openib.conf" >> /etc/modprobe.conf

    else # Kernel 2.4
	  printf "\n Updating /etc/modules.conf ..."
#	  /bin/rm -f /etc/modules-openib.conf

	# Remove previous configuration if exist
        perl -ni -e 'if (/\#\# BEGIN OpenIB \#\#$/) { $filter = 1 }' -e 'if (!$filter) { print }' -e 'if (/\#\# END OpenIB \#\#$/){ $filter = 0 }' /etc/modules.conf

cat >> /etc/modules.conf << EOF
## BEGIN OpenIB ##
#
# InfiniBand module loading configuration
#

EOF

for (( i=0 ; i < $ports_num ; i++ ))
do
cat >> /etc/modules.conf << EOF
alias ib${i} ib_ipoib
EOF
done

cat >> /etc/modules.conf << EOF

alias net-pf-26 ib_sdp

alias /dev/mosal mlxsys
alias /dev/vipkl mod_vapi
alias kdapl dat_registry
alias kdapltest kdapltest_module

options ib_mad ib_outs_mad_recv_entries=1024
options ib_client_query ib_query_throttle_threshhold=0 ib_query_throttle_waitqueue_limit=10000
options ib_ipoib ipoib_use_interrupt_context=0

post-install mlxsys { rm -f /dev/mosal; mknod /dev/mosal c \`awk '(\$2=="mosal") {print \$1}' /proc/devices\` 0; }
post-remove mlxsys /bin/rm -f /dev/mosal

post-install mod_vip { rm -f /dev/vipkl; mknod /dev/vipkl c \`awk '(\$2=="mellanox_vipkl") {print \$1}' /proc/devices\` 1 ; }
post-remove mod_vip /bin/rm -f /dev/vipkl

post-install mod_thh open_hcas
post-install mod_rhh open_hcas
pre-remove mod_thh close_hcas
pre-remove mod_rhh close_hcas

post-install ib_useraccess ${prefix}/utils/create_devs ib_useraccess
post-install ib_ipoib ${prefix}/utils/update_ipoib_intfs
# add above ib_ipoib ib_useraccess
# add above ib_ipoib ib_ip2pr
post-install ib_ip2pr ${prefix}/utils/create_devs ib_ip2pr
post-install ib_dapl_srv ${prefix}/utils/create_devs ib_udapl
# add above ib_dapl_srv ib_useraccess_cm
post-install ib_useraccess_cm ${prefix}/utils/create_devs ib_useraccess_cm
# add above ib_sa_client ib_useraccess
add below ib_ipoib ib_tavor
# post-install ib_srp { ${prefix}/utils/srp_persistent_bind.sh & }
# pre-remove ib_srp ${prefix}/utils/remove_srp_bind_persistent.sh
post-install dat_registry /sbin/modprobe ib_kdapl
post-install kdapltest_module ${prefix}/utils/create_devs kdapltest_module
## END OpenIB ##
EOF
#	  perl -i.orig -ne 'print unless (m@include /etc/modules-openib.conf@)' /etc/modules.conf
#	  echo "include /etc/modules-openib.conf" >> /etc/modules.conf
    fi

    printf "\r\t\t\t\t\t (done)"                                                    

}

#################################################################################################
#                                       MAIN area						#
#################################################################################################
main()
{

    if [ $SET_ENV -eq 1 ]; then
    	LOGFILE=${tmpdir}/check_prev_install.log
    	check_prev_install
    fi
    
    LOGFILE=${tmpdir}/pre_install.log

    set_deps
    
    print_selected_packages
    
    if [ ! -d ${BUILD_ROOT}/etc ]; then
	mkdir -p ${BUILD_ROOT}/etc
    fi
                      
    LOGFILE=${tmpdir}/create_kernel_tree.log
    create_kernel_tree
    
    LOGFILE=${tmpdir}/build_kernel_modules.log
    build_kernel_modules

    LOGFILE=${tmpdir}/build_user_level.log
    build_user_level

    LOGFILE=${tmpdir}/install_kernel_modules.log
    install_kernel_modules

    # Run this part to update running environment
    if [ $SET_ENV -eq 1 ]; then

	printf "\n Running depmod ..."
	if [ "${KVERSION_TWO_DIGITS}" == "24" ]; then
		 ex /sbin/depmod -r -ae -F /boot/System.map-${KVERSION} ${KVERSION}
	else
		 ex /sbin/depmod -r -ae ${KVERSION}
	fi
        printf "\r\t\t\t\t\t (done)"
        
	[ ! -d ${IB_CONFIG_DIR} ] && mkdir -p ${IB_CONFIG_DIR}

	install -m 0644 ${installdir}/openib.conf ${IB_CONFIG_DIR}
                                                          
	if [ "$ipoib" == "y" ]; then
	    printf "\n Configuring IPoIB..."
	    ipoib_config

	    echo >> ${IB_CONFIG_DIR}/openib.conf                                                
	    echo "# Load IPoIB" >> ${IB_CONFIG_DIR}/openib.conf
	    echo "IPOIB_LOAD=yes" >> ${IB_CONFIG_DIR}/openib.conf
            printf "\r\t\t\t\t\t (done)"

	fi
        
	if [ "$cm" == "y" ]; then
	    echo >> ${IB_CONFIG_DIR}/openib.conf
	    echo "# Load USER ACCESS CM module" >> ${IB_CONFIG_DIR}/openib.conf
	    echo "USER_ACCESS_CM_LOAD=no" >> ${IB_CONFIG_DIR}/openib.conf
	fi
        
	if [ "$sdp" == "y" ]; then
	    echo >> ${IB_CONFIG_DIR}/openib.conf
	    echo "# Load SDP module" >> ${IB_CONFIG_DIR}/openib.conf
	    echo "SDP_LOAD=no" >> ${IB_CONFIG_DIR}/openib.conf
	fi
        
	if [ "$dapl" == "y" ]; then
	    echo >> ${IB_CONFIG_DIR}/openib.conf
	    echo "# Load UDAPL module" >> ${IB_CONFIG_DIR}/openib.conf
	    echo "UDAPL_LOAD=no" >> ${IB_CONFIG_DIR}/openib.conf
            echo >> ${IB_CONFIG_DIR}/openib.conf
	    echo "# Load KDAPL module" >> ${IB_CONFIG_DIR}/openib.conf
	    echo "KDAPL_LOAD=no" >> ${IB_CONFIG_DIR}/openib.conf

	fi
        
	if [ "$srp" == "y" ]; then
	    echo >> ${IB_CONFIG_DIR}/openib.conf
	    echo "# Load SRP initiator module" >> ${IB_CONFIG_DIR}/openib.conf
	    echo "SRP_LOAD=no" >> ${IB_CONFIG_DIR}/openib.conf
	fi
        
	if [ "$srp_target" == "y" ]; then
	    echo >> ${IB_CONFIG_DIR}/openib.conf
	    echo "# Load SRP target module" >> ${IB_CONFIG_DIR}/openib.conf
	    echo "SRP_TARGET_LOAD=no" >> ${IB_CONFIG_DIR}/openib.conf
	fi

	# Create /etc/modprobe-openib.conf or /etc/modules-openib.conf                    
	modules_conf

	install -m 0755 ${installdir}/openibd /etc/init.d/
	perl -ni -e "s@(STACK_PREFIX=).*@\$1${STACK_PREFIX}@; print" /etc/init.d/openibd

	if [ -f /etc/SuSE-release ]; then
        perl -i -ne 'if (m@^#!/bin/bash@) {
        print q@#!/bin/bash
### BEGIN INIT INFO
# Provides:       openibd
# Required-Start: $local_fs $network
# Required-Stop: opensmd
# Default-Start:  2 3 5
# Default-Stop: 0 1 2 6
# Description:    Activates/Deactivates InfiniBand Driver to \
#                 start at boot time.
### END INIT INFO
@;
                 } else {
                     print;
                 }' /etc/init.d/openibd


		/sbin/insserv openibd
	else
perl -i -ne 'if (m@^#!/bin/bash@) {
        print q@#!/bin/bash
#
# Bring up/down openib
#
# chkconfig: 2345 15 85
# description: Activates/Deactivates InfiniBand Driver to \
#              start at boot time.
#
### BEGIN INIT INFO
# Provides:       openibd
### END INIT INFO
@;
                 } else {
                     print;
                 }' /etc/init.d/openibd

        	/sbin/chkconfig --del openibd
        	/sbin/chkconfig --add openibd

	fi

	# Add mod_thh/mod_rhh to the hotplug blacklist if exists
	if [ -e /etc/hotplug/blacklist ]; then
	    perl -i.orig -ne 'print unless (m@mod_thh@)' /etc/hotplug/blacklist
	    perl -i.orig -ne 'print unless (m@mod_rhh@)' /etc/hotplug/blacklist
	    echo mod_thh >> /etc/hotplug/blacklist
	    echo mod_rhh >> /etc/hotplug/blacklist
	fi

    fi

    printf "\n Installation finished successfully \n"
}

main
