#!/bin/bash
#
# A script for SRP persistent binding
#
# this script will create links to SCSI disks under /dev/srp/
# using name conventions : srp-<IOC_GUID>-l<LUN>p<partition>
#

SRP_HOST_DIR=/proc/scsi/srp/
DEV_DIR=/dev/scsi/
SRP_LINKS_DIR=/dev/srp/
VERBOUS=0
CLEAN=1
RUN_ONCE=0
TIME_FOR_SLEEP=60
BIN_DIR=/usr/local/ibgd/bin
TMP_FILE=/tmp/srp_bind.tmp

# CFG file handling
CFG_FILE=/usr/local/ibgd/etc/srp_persistnet_bind.conf
CFG_BIND_ALL=1
CFG_MASKED_CNT=0
if test -r $CFG_FILE ; then 
   if test "`grep "^m srp-ALL" $CFG_FILE`" = "m srp-ALL" ; then
      CFG_BIND_ALL=0
   fi
elif test $VERBOUS = 1; then
   echo "config file $CFG_FILE doesn't exist"
fi


function usage {
    echo "srp_persistent_bind.sh [ -srp_link_dir <DIR_NAME>] [ -v ] [ -dont_clean ] [ -sleep_time <sec>] [ -run_once ]"
    exit 0
}


# Input parsing
while test "$1" ; do
   case "$1" in
                -srp_link_dir)
                        SRP_LINKS_DIR=$2
                        shift 2
                        ;;
                -v)
                        VERBOUS=1
                        shift 1
                        ;;
                -dont_clean)
                        CLEAN=0
                        shift 1
                        ;;
                -run_once)
                        RUN_ONCE=1
                        shift 1
                        ;;
                -sleep_time)
                        TIME_FOR_SLEEP=$2
                        shift 2
                        ;;
                *)
                        echo "Wrong parameter error!"
                        usage
                        exit 1
                        ;;
   esac
done



# check if SRP module is already up
if test ! -d "$SRP_HOST_DIR"; then
   echo "No SRP hosts are loaded"
   exit 1
fi

# check if the dir where i'll create srp links exist
if test ! -d "$SRP_LINKS_DIR"; then
   mkdir $SRP_LINKS_DIR
else
   # On the first run i need to delete all links cause 
   # they may link to wrong but existing block device files
   if test $CLEAN = 1; then
      rm $SRP_LINKS_DIR/* &>/dev/null
      rm $SRP_LINKS_DIR/.masked.* &>/dev/null
   fi
fi

#clean also the devices on the first run
if test -d "$DEV_DIR" ; then
   rm $DEV_DIR/* &>/dev/null
fi

# start the loop that will run forever
while test 1 ; do

  # run scsidev utility
  # it will create devices under DEV_DIR
  if test -e /bin/scsidev ; then
      /bin/scsidev > /dev/null
  else
      $BIN_DIR/scsidev > /dev/null
  fi

  if test $? != 0 ; then
     echo "scsidev utility does not exist on this mashine. Please install it"
     exit 1
  fi
           
  
  # Define GUIDs list
  GUIDS_LIST=""

  # make links for all hosts/IOCs/LUNs/partitions
  cd $SRP_HOST_DIR
  for host_num in * ; do
           cd $SRP_HOST_DIR
        
	   cat $host_num | grep "SRP.T10" > $TMP_FILE
	   exec 3< $TMP_FILE
           # Using input redirection for while loop in order to save GUIDS_LIST value outside the loop
	   while read IOC_NUM IOC_SVC_NAME PORT_GUID_tmp IOC_GUID_tmp no_need <&3
           do
              IOC_GUID="${IOC_GUID_tmp//:/}"
              if test $VERBOUS = 1; then
                 echo "Host : $host_num  IOC number : $IOC_NUM   IOC GUID : $IOC_GUID  SCV name : $IOC_SVC_NAME"
              fi

	      GUIDS_LIST="$GUIDS_LIST $IOC_GUID"
              # now need to make a link srp<IOC_GUID>-l<LUN>-p<partition> to the
              # apropriate sdh<host>-0c0i<channel>l<LUN>p<partition>
              #             host_num-0c0i<IOC_NUM>.....as.is........
              # sdh1-0c0i1l0p4   <- srp-<IOC_GUID>-l<LUN>p<partition>
              cd $DEV_DIR
              FILE_PREFIX="sdh$host_num-0c0i$IOC_NUM*"
  
              for dev_file in $FILE_PREFIX ; do
                   SRP_LINK=srp-$IOC_GUID-l$( echo $dev_file | cut -dl -f2)
                   if test -L $SRP_LINKS_DIR$SRP_LINK ; then
                       if test -b $DEV_DIR$dev_file ; then
                          if test $VERBOUS = 1; then
                             echo "link    $SRP_LINK  ->  $dev_file  already exist"
                          fi
                       else
                          if test $VERBOUS = 1; then
                             echo "link  $SRP_LINK  exist BUT block device $DEV_DIR$dev_file does not (removing link)"
                          fi
                          rm $SRP_LINKS_DIR$SRP_LINK
                       fi
                   else # link does not exist
                       if test -b $DEV_DIR$dev_file ; then
                          if test $VERBOUS = 1; then
                             echo "making link    $SRP_LINK  ->  $dev_file "
                          fi
                          # I need to check if this link is not in the masked links
                          if test -r $CFG_FILE ; then 
                              if test "`grep "^b $SRP_LINK" $CFG_FILE`" = "b $SRP_LINK" ; then
                                 ln -snf $DEV_DIR$dev_file $SRP_LINKS_DIR$SRP_LINK
                              elif test "`grep "^m $SRP_LINK" $CFG_FILE`" = "m $SRP_LINK" ; then
                                 if test -L $SRP_LINKS_DIR/.masked.$SRP_LINK ; then
                                    if test $VERBOUS = 1; then
                                       echo "link    $SRP_LINK  ->  $dev_file is masked"
                                    fi
                                 else
                                    ln -snf $DEV_DIR$dev_file $SRP_LINKS_DIR/.masked.$SRP_LINK
                                 fi   
                              elif test $CFG_BIND_ALL = 1 ; then
                                 ln -snf $DEV_DIR$dev_file $SRP_LINKS_DIR$SRP_LINK
                              else
                                 # I can't do this cause there is no way to export
                                 # var to the parent prosses
                                 #CFG_MASKED_CNT=`expr $CFG_MASKED_CNT + 1`
                                 ln -snf $DEV_DIR$dev_file $SRP_LINKS_DIR/.masked.$SRP_LINK
                              fi
                          else
                              ln -snf $DEV_DIR$dev_file $SRP_LINKS_DIR$SRP_LINK
                          fi
                       else
                          if test $VERBOUS = 1; then
                             echo "No link and no block device ($SRP_LINK -> $dev_file) - do nothing"
                          fi
                       fi
                   fi
              done
           done
	   exec 3<&-
  done
  
  # remove non existent devices (GUID is not in GUIDS_LIST)
  NEED_RESCAN=0
  for srp_link in /dev/srp/*
  do
	match=0
	if test $VERBOUS = 1; then
		echo "The list of available SRP target(s) guids GUIDS_LIST=$GUIDS_LIST"
	fi
	for guid in $GUIDS_LIST
	do
		if ( echo -n ${srp_link} | grep ${guid} > /dev/null 2>&1 ); then
			match=1
			break
		fi
	done
	
	if [ $match -eq 0 ]; then
		# Extra link should be removed from /dev/srp and from /dev/scsi
		dev_scsi=$(readlink -f $srp_link)	
		if test $VERBOUS = 1; then
			echo "Removing extra devices: $srp_link -> $dev_scsi"
		fi
		rm -f $srp_link $dev_scsi
		NEED_RESCAN=1
	fi
  done 

  if [ $NEED_RESCAN -eq 1 ]; then
        if [ ! -d /proc/scsi/srp ]; then
                echo "No SRP devices found"
                return 1
        fi

	if test $VERBOUS = 1; then
		echo "Running rescan..."
	fi

        for srpf in /proc/scsi/srp/*
        do
                echo "force_rediscover" > $srpf
        done
  fi

  # remove all links that point to non existing devises
  cd $DEV_DIR
  NUM_OF_FILES_FROM_SCSIDEV=`ls | grep sdh | wc -l`
  cd $SRP_LINKS_DIR
  NUM_OF_LINKS=`ls | wc -l`
  if test $NUM_OF_LINKS -gt $NUM_OF_FILES_FROM_SCSIDEV ; then
     if test $VERBOUS = 1; then
        echo "There are more links then active devices - removing not active links"
     fi
     for cur_link in * ; do
        stat -L $cur_link 2>&1 | grep Size > /dev/null
        if test $? != 0 ; then
           if test $VERBOUS = 1; then
              echo "$cur_link is deleted"
           fi
           rm $cur_link
        fi
     done
  fi
  
  
  # check num of files = num of links
  #NUM_OF_FILES_FROM_SCSIDEV=`ls sdh* | wc | awk '{print $1}'`
  #NUM_OF_LINKS=`ls $SRP_LINKS_DIR | wc | awk '{print $1}'`
  NUM_OF_FILES_FROM_SCSIDEV=0
  cd $SRP_HOST_DIR
  for host_num in * ; do
      cd $DEV_DIR
      let "NUM_OF_FILES_FROM_SCSIDEV +=`ls | grep sdh$host_num | wc -l`"
      cd $SRP_HOST_DIR
  done
  NUM_OF_LINKS=`ls $SRP_LINKS_DIR 2>/dev/null | wc -l`
  CFG_MASKED_CNT=`ls $SRP_LINKS_DIR/.masked.* 2>/dev/null | wc -l`
  NUM_OF_LINKS_AND_MASK=`expr $NUM_OF_LINKS + $CFG_MASKED_CNT`
  if test ! $NUM_OF_LINKS_AND_MASK = $NUM_OF_FILES_FROM_SCSIDEV ; then
     echo "WARNING: Num of links $NUM_OF_LINKS_AND_MASK is different from sdsidev files number $NUM_OF_FILES_FROM_SCSIDEV"
     # exit 1
  fi

  if test $RUN_ONCE = 1 ; then
    break
  fi
  cd /
  sleep $TIME_FOR_SLEEP

done 
