#!/bin/bash 
#
# This software is available to you under a choice of one of two
# licenses.  You may choose to be licensed under the terms of the GNU
# General Public License (GPL) Version 2, available at
# <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
# license, available in the LICENSE.TXT file accompanying this
# software.  These details are also available at
# <http://openib.org/license.html>.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
#

#cd ${0%*/*}

# pointing to the dir where this install script resides
PACKAGE_ORIG_DIR=`cd ${0%*/*};pwd`

########################################################################
#
# FUNCTIONS:
#

# provide usage message
function print_usage() {
 echo "Usage: $0 [--prefix <install-dir>][--with-ibadm <ibadm-dir>][--batch]"
 echo " "
 echo "Options:"
 echo "   --prefix <dir> : directory to install into. "
 echo "   --with-ibadm <dir> : the directory holding ibadm instllation. "
 echo "   --batch option : do not start the setup wizard."
}

# Heruristically try and detect where ibadm is installed.
# Usage auto_detect_ibadm
function auto_detect_ibadm() {
    ibadm_file=`which ibadm` 1> /dev/null 2>&1
    if test $? != 0; then
        echo "-E- Could not find ibadm executable."
        echo "    Please use the flag --with-ibadm to point to the directory"
        echo "    IBADM was installed into. Such that <dir>/bin/ibadm exists."
        exit 1;
    fi
}

#Report: if install failed
function install_failed() {
	echo "IBGFVS install failed !"
   cleanup
	exit 1
}

# cleanup
function cleanup() {
    rm -rf $PACKAGE_DIR
}

# creates the tcl package directory under the lib $prefix/dir
function build_tcl_package() {
    mkdir -p $prefix/lib/ibgfvs1.0
    files="pkgIndex.tcl ibgfvsSetupMessages.tcl ibgfvsMain.tcl mpitest.c"
    for f in $files; do
        cp $PACKAGE_DIR/src/$f $prefix/lib/ibgfvs1.0/$f
        if [ $? != 0 ]; then 
            echo "-E- Failed to create IBGFVS TCL package"
            cleanup
            exit 1
        fi
    done
}

# based on the ibadm_file create and install a wrapper that
# invokes TCL and load the ibgfvs package
function install_wrapper() {
    # we simply modify the ibadm wrapper
    sed -e 's%TCLLIBPATH=\([^;]*\)%TCLLIBPATH="'$prefix/lib' \1"%' \
        -e 's/BLD_VER.*/BLD_VER "'$BLD_ID'"/' \
        -e 's/package require app-ibadm/package require ibgfvs/' \
        -e 's/IBADM Names Server/IBGranite Cluster Verification Suite application/' \
        $ibadm_file > $PACKAGE_DIR/ibgfvs.$$

    ibadm_tcl=`grep -w exec $ibadm_file | cut -f2 -d" "`

    check_cmd="if  ! [ -x $ibadm_tcl ] ; then echo -E- TCL $ibadm_tcl  not found. Probably IBADM is not installed.; exit 1; fi; \\\\"
    awk "{print \$0} /IBGranite Cluster Verification Suite application/ {print \"$check_cmd\"}" $PACKAGE_DIR/ibgfvs.$$ > $PACKAGE_DIR/ibgfvs

    \rm -rf $PACKAGE_DIR/ibgfvs.$$

    \mv $PACKAGE_DIR/ibgfvs $PACKAGE_DIR/ibgfvs.$$
    set_pref="set gInstallationPrefix {$prefix}"
    awk "{print \$0} /set BLD_VER/ {print \"$set_pref\"}" $PACKAGE_DIR/ibgfvs.$$ > $PACKAGE_DIR/ibgfvs


    mkdir -p $prefix/bin
    \mv -f $PACKAGE_DIR/ibgfvs $prefix/bin/ibgfvs
    chmod 555 $prefix/bin/ibgfvs

    if [ $? != 0 ]; then 
        echo "-E- Failed to prepare the ibgfvs wrapper"
        cleanup
        exit 1
    fi
   
}

# install the man pages
function install_mans() {
    mkdir -p $prefix/share/man/man1/
    files="ibgfvs.1"
    for f in $files; do
        cp $PACKAGE_DIR/src/$f $prefix/share/man/man1/$f
        if [ $? != 0 ]; then 
            echo "-E- Failed to install IBGFVS man pages"
            cleanup
            exit 1
        fi
    done
}

# copy the ibadm tar file.
function copy_ibadm_tar() {
    mkdir -p $prefix/share/ibgfvs
    cp -f ${PACKAGE_ORIG_DIR}/ibadm*.tgz $prefix/share/ibgfvs/ibadm.tgz
}

########################################################################

#
# MAIN FLOW
#

# Initialize default before parsing parameters
prefix=/usr
BATCH_BUILD=0

# parse command line
# parse parameters
while [ "$1" ]; do
  case $1 in
    "--prefix") 
          prefix=$2
          echo "<<<<<<<< INSTALLING INTO $prefix >>>>>>>>>>"
          shift
          ;;
    "--batch") 
          BATCH_BUILD=1
          echo "<<<<<<<< BATCH MODE - DO NOT RUN SETUP >>>>>>>>>>"
          ;;
    "--with-ibadm") 
          if test ! -f $2/bin/ibadm; then
            echo "-E- Given ibadm path $2 does not point to an ibadm"
            echo "    installation since $2/bin/ibadm does not exist"
            echi " "
            print_usage
            exit 1
          fi
          export PATH=$2:$PATH
          echo "<<<<<<<< Using IBADM from $2 >>>>>>>>>>"
          shift
          ;;
    *)
     print_usage
     exit 1
  esac
  shift
done

# double check for ibadm availablity. get ibadm_file if found
auto_detect_ibadm

# double check that the tar file exists in the package dir
tar_file=$PACKAGE_ORIG_DIR/ibgfvs.tgz
if test ! -f $tar_file; then
    echo "-E- Fail to find ibgfvs tar file in package dir. Bad package dir!"
    exit 1
fi

# check root
if [ $UID != 0 ]; then
  echo "-E- You must be root to install IBGFVS"
  exit 1
fi

# take the src BUILD_ID from the package
if test -e ${PACKAGE_ORIG_DIR}/BUILD_ID ; then
  BLD_ID=`cat ${PACKAGE_ORIG_DIR}/BUILD_ID`
else
  BLD_ID="<unknown>"
fi

# Done checking so Start installation
echo "-------------------------------------------------------------------------"
echo IBGranite Cluster Verification Suite installation - Build: $BLD_ID
echo Copyright \(C\) June 2005, Mellanox Technologies Ltd. ALL RIGHTS RESERVED.
echo Use of software subject to the terms and conditions detailed in the file:
echo \"$PACKAGE_ORIG_DIR/LICENSE.txt\".
echo "-------------------------------------------------------------------------"
echo " "

# we modify the srcs so we need to use a temporary dir
PACKAGE_DIR=/tmp/ibgfvs_install_tmp
if [ -d $PACKAGE_DIR ]; then
    rm -rf $PACKAGE_DIR
fi
mkdir -p $PACKAGE_DIR
# go there
cd $PACKAGE_DIR

# untar the package file
tar -zxf $tar_file
find . -exec touch {} \;

umask 022   # assure readibility to all users

# create the TCL package
build_tcl_package

# update and install the ibgfvs wrapper
install_wrapper

# install man page
install_mans

# copy the ibadm tar file.
copy_ibadm_tar

cleanup

echo IBGFVS installation done. Please run \"ibgfvs setup\".

echo "  "
exit 0
