/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) June 2000, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "LICENSE.txt".
 *
 */                                                                                   
 
 
/*
*-------------------------------------------------------------------
* Description: A test that check if the SA is alive (and available) by
* sending an SA query.
* If a device ID is not being provided, the test will autodetect the first 
* HCA that it finds in the host (if available).
*-------------------------------------------------------------------								 
* Version : $Id: cmd_pars.c,v 1.1 2003/10/29 17:10:11 dotanb Exp $									   
* Author  : Dotan Barak
*-------------------------------------------------------------------
*/

#include <stdlib.h>
#include <stdio.h>
#include <byteswap.h>
#include <vapi.h>
#include <vapi_common.h>
#include <mosal.h>
#include "types.h"
#include "cmd_pars.h"



/******************************************************************
                        global variables
******************************************************************/

config_t config = {
    "",                         /* hca_id */
    1,                          /* port_num */
    1,                          /* timeout */
    0                           /* verbose */
};

/*****************************************************************************
* Function: async_event_handler
*****************************************************************************/
static void async_event_handler(
    VAPI_hca_hndl_t hca_hndl,
    VAPI_event_record_t *event_record_p,
    void *private_data)
{
    printf("Got unexpected asynchronous event. Type: %s Syndrome: %s", 
           VAPI_event_record_sym(event_record_p->type), VAPI_event_syndrome_sym(event_record_p->syndrome));
    exit(1);
}


/*****************************************************************************
* Function: build_packet
*****************************************************************************/
static void build_packet(
    void *buf)
{
    /* Cast by Path Record Query */
    ib_sa_mad_t *p_pr = (ib_sa_mad_t*)buf;

    memset(buf, 0, MAX_PACKET_SIZE);

    /* First set the header type */
    p_pr->base_ver = 0x1;
    p_pr->mgmt_class = 0x3;
    p_pr->class_ver = 0x2;
    p_pr->method = 0x1; /* SubnAdmGet */
    p_pr->status = 0x0;
    p_pr->resv = 0x0;
    p_pr->trans_id=0;
    p_pr->attr_id = bswap_16(0x1); /* ClassPortInfo */
    p_pr->rmpp_version = 0x0;
    p_pr->rmpp_type = 0x0;
    p_pr->rmpp_flags = 0x0;
    p_pr->rmpp_status = 0x0;
    p_pr->seg_num = 0x0;
    p_pr->paylen_newwin=0;
    p_pr->sm_key = 0x0;
    p_pr->attr_offset = 0x0;
    p_pr->resv3 = 0x0;
    p_pr->comp_mask = 0x0;
}

/*****************************************************************************
* Function: resource_init
*****************************************************************************/
static void resource_init(
    resource_t *resources_p)
{
    resources_p->hca_hndl = VAPI_INVAL_HNDL;
    resources_p->async_hndl = VAPI_INVAL_HNDL;
    resources_p->pd_hndl = VAPI_INVAL_PD_HNDL;
    resources_p->cq_hndl = VAPI_INVAL_HNDL;
    resources_p->qp_hndl = VAPI_INVAL_HNDL;
    resources_p->av_hndl = VAPI_INVAL_HNDL;
    resources_p->snd_mr_hndl = VAPI_INVAL_HNDL;
    resources_p->snd_buf = NULL;
    resources_p->rcv_mr_hndl = VAPI_INVAL_HNDL;
    resources_p->rcv_buf = NULL;
}


/*****************************************************************************
* Function: get_hca_name
*****************************************************************************/
static VAPI_ret_t get_hca_name(void)
{
    VAPI_ret_t rc;
    VAPI_hca_id_t *hca_id_buf = NULL;
    u_int32_t local_num_of_hcas;

    rc = EVAPI_list_hcas(0, &local_num_of_hcas, NULL);
    if (rc != VAPI_EAGAIN) {
        printf("Error, VAPI_EAGAIN was expected, but got %s\n", VAPI_strerror(rc));
        return VAPI_EGEN;
    }

    hca_id_buf = MALLOC(sizeof(VAPI_hca_id_t)*local_num_of_hcas);
    if (hca_id_buf == NULL) {
        return VAPI_ENOMEM;
    }

    rc = EVAPI_list_hcas(local_num_of_hcas, &local_num_of_hcas, hca_id_buf);
    if (rc == VAPI_OK) {
        strncpy(config.hca_id, hca_id_buf[0], HCA_MAXNAME);

        if (config.verbose) {
            printf("Device %s was autodetected.\n", config.hca_id);
        }
    } else {
        printf("Error, EVAPI_list_hcas failed: %s\n", VAPI_strerror(rc));
    }
    FREE(hca_id_buf);
    
    return rc;
}

/*****************************************************************************
* Function: resource_create
*****************************************************************************/
static VAPI_ret_t resource_create(
    resource_t *resources_p)
{
    VAPI_ret_t rc;
    VAPI_qp_init_attr_t qp_init_attr;
    VAPI_qp_attr_t qp_attr;
    VAPI_qp_cap_t qp_cap;
    VAPI_qp_attr_mask_t qp_attr_mask;
    VAPI_ud_av_t av;
    VAPI_mrw_t mr_props;

    /* if no device ID was given, use auto detect */
    if (strlen(config.hca_id) == 0) {
        rc = get_hca_name();
        CHECK_RC("get_hca_name", rc, return rc);
    }

    if (config.verbose == 1) {
        printf("Opening device %s: port %d\n", config.hca_id, config.port_num);
    }

    /* get HCA handle */
    rc = EVAPI_get_hca_hndl(config.hca_id, &(resources_p->hca_hndl));
    CHECK_RC("EVAPI_get_hca_hndl", rc, return rc);

    /* Query port props */
    rc = VAPI_query_hca_port_prop(resources_p->hca_hndl, config.port_num, &(resources_p->hca_port));
    CHECK_RC("VAPI_query_hca_port_prop", rc, return rc);

    if (config.verbose == 1) {
        printf("SM lid = 0x%x\n", resources_p->hca_port.sm_lid);
        printf("SM SL = 0x%x\n", resources_p->hca_port.sm_sl);
    }

    /* set async event handler */
    rc = EVAPI_set_async_event_handler(resources_p->hca_hndl, async_event_handler, NULL, &(resources_p->async_hndl));
    CHECK_RC("EVAPI_set_async_event_handler", rc, return rc);

    /* allocate PD */
    rc = VAPI_alloc_pd(resources_p->hca_hndl, &(resources_p->pd_hndl));
    CHECK_RC("VAPI_alloc_pd", rc, return rc);

    /* create UD AV */
    av.port = config.port_num;
    av.dlid = resources_p->hca_port.sm_lid;
    av.sl = resources_p->hca_port.sm_sl;
    av.src_path_bits = 0;
    av.grh_flag = FALSE;
    av.static_rate = 0;

    rc = VAPI_create_addr_hndl(resources_p->hca_hndl, resources_p->pd_hndl, &av, &(resources_p->av_hndl));
    CHECK_RC("VAPI_create_addr_hndl", rc, return rc);

    /* create CQ (we will only use one CQ) */
    rc = VAPI_create_cq(resources_p->hca_hndl, NUM_OF_COMPLETION, &(resources_p->cq_hndl), &(resources_p->cq_size));
    CHECK_RC("VAPI_create_cq", rc, return rc);

    /* create QP */
    qp_init_attr.ts_type = VAPI_TS_UD;
    qp_init_attr.pd_hndl = resources_p->pd_hndl;
    qp_init_attr.rq_cq_hndl = qp_init_attr.sq_cq_hndl = resources_p->cq_hndl; /* use the same CQ for RQ and SQ */
    qp_init_attr.sq_sig_type = VAPI_SIGNAL_ALL_WR;
    qp_init_attr.rq_sig_type = VAPI_SIGNAL_ALL_WR;
    qp_init_attr.cap.max_oust_wr_rq = QP_OUST_WR_SIZE;
    qp_init_attr.cap.max_oust_wr_sq = QP_OUST_WR_SIZE;
    qp_init_attr.cap.max_sg_size_rq = QP_SG_SIZE;
    qp_init_attr.cap.max_sg_size_sq = QP_SG_SIZE;

    rc = VAPI_create_qp(resources_p->hca_hndl, &qp_init_attr, &(resources_p->qp_hndl), &(resources_p->qp_prop));
    CHECK_RC("VAPI_create_qp", rc, return rc);
    
    
    /* modify QP state to Init */
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);
    
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_QP_STATE);
    qp_attr.qp_state = VAPI_INIT;

    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_PKEY_IX );
    qp_attr.pkey_ix  = DEF_PKEY_IX; 

    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_QKEY);
    qp_attr.qkey = IB_QP1_WELL_KNOWN_Q_KEY;

    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_PORT );
    qp_attr.port = config.port_num;

    rc = VAPI_modify_qp(resources_p->hca_hndl, resources_p->qp_hndl, &qp_attr,
                        &qp_attr_mask, &qp_cap);
    CHECK_RC("VAPI_create_qp: 2init", rc, return rc);
   
    
    /* modify QP state to RTR */
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);

    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_QP_STATE);
    qp_attr.qp_state = VAPI_RTR;


    rc = VAPI_modify_qp(resources_p->hca_hndl, resources_p->qp_hndl, &qp_attr,
                        &qp_attr_mask, &qp_cap);
    CHECK_RC("VAPI_create_qp: 2RTR", rc, return rc);


    /* modify QP state to RTS */
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);

    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_QP_STATE);
    qp_attr.qp_state = VAPI_RTS;
    
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_SQ_PSN);
    qp_attr.sq_psn = 0;
    
    rc = VAPI_modify_qp(resources_p->hca_hndl, resources_p->qp_hndl, &qp_attr,
                        &qp_attr_mask, &qp_cap);
    CHECK_RC("VAPI_create_qp: 2RTS", rc, return rc);

    /* common code to both MRs */
    mr_props.type = VAPI_MR;
    mr_props.pd_hndl = resources_p->pd_hndl;
    
    /* create send MR */
    mr_props.size = MAX_PACKET_SIZE;
    resources_p->snd_buf = MALLOC(mr_props.size);
    if (resources_p->snd_buf == NULL) {
        printf("Error, failed to allocate memory for send buffer\n");
        return VAPI_ENOMEM;
    }

    mr_props.start = (VAPI_virt_addr_t)(unsigned long)resources_p->snd_buf;
    mr_props.acl = 0; /* we only need loacl read permission */
    
    rc = VAPI_register_mr(resources_p->hca_hndl, &mr_props,
                        &(resources_p->snd_mr_hndl), &(resources_p->snd_mr_props));
    CHECK_RC("VAPI_register_mr", rc, return rc);

    /* create receive MR */
    mr_props.size = MAX_PACKET_SIZE + VAPI_GRH_LEN;
    resources_p->rcv_buf = MALLOC(mr_props.size);
    if (resources_p->rcv_buf == NULL) {
        printf("Error, failed to allocate memory for receive buffer\n");
        return VAPI_ENOMEM;
    }

    mr_props.start = (VAPI_virt_addr_t)(unsigned long)resources_p->rcv_buf;
    mr_props.acl = VAPI_EN_LOCAL_WRITE; /* we only need local write permission */
    
    rc = VAPI_register_mr(resources_p->hca_hndl, &mr_props,
                        &(resources_p->rcv_mr_hndl), &(resources_p->rcv_mr_props));
    CHECK_RC("VAPI_register_mr", rc, return rc);


    return VAPI_OK;
}

/*****************************************************************************
* Function: resource_destroy
*****************************************************************************/
static VAPI_ret_t resource_destroy(
    resource_t *resources_p)
{
    VAPI_ret_t rc;
    VAPI_ret_t final_rc = VAPI_OK;
    
    if (resources_p->rcv_mr_hndl != VAPI_INVAL_HNDL) {
        rc = VAPI_deregister_mr(resources_p->hca_hndl, resources_p->rcv_mr_hndl);
        CHECK_RC("VAPI_deregister_mr", rc, final_rc = rc);
    }

    if (resources_p->rcv_buf != NULL) {
        FREE(resources_p->rcv_buf);
    }

    if (resources_p->snd_mr_hndl != VAPI_INVAL_HNDL) {
        rc = VAPI_deregister_mr(resources_p->hca_hndl, resources_p->snd_mr_hndl);
        CHECK_RC("VAPI_deregister_mr", rc, final_rc = rc);
    }

    if (resources_p->snd_buf != NULL) {
        FREE(resources_p->snd_buf);
    }

    if (resources_p->qp_hndl != VAPI_INVAL_HNDL) {
        rc = VAPI_destroy_qp(resources_p->hca_hndl, resources_p->qp_hndl);
        CHECK_RC("VAPI_destroy_qp", rc, final_rc = rc);
    }

    if (resources_p->cq_hndl != VAPI_INVAL_HNDL) {
        rc = VAPI_destroy_cq(resources_p->hca_hndl, resources_p->cq_hndl);
        CHECK_RC("VAPI_destroy_cq", rc, final_rc = rc);
    }

    if (resources_p->av_hndl != VAPI_INVAL_HNDL) {
        rc = VAPI_destroy_addr_hndl(resources_p->hca_hndl, resources_p->av_hndl);
        CHECK_RC("VAPI_destroy_addr_hndl", rc, final_rc = rc);
    }

    if (resources_p->pd_hndl != VAPI_INVAL_PD_HNDL) {
        rc = VAPI_dealloc_pd(resources_p->hca_hndl, resources_p->pd_hndl);
        CHECK_RC("VAPI_dealloc_pd", rc, final_rc = rc);
    }

    if (resources_p->async_hndl != VAPI_INVAL_HNDL) {
        rc = EVAPI_clear_async_event_handler(resources_p->hca_hndl, resources_p->async_hndl);
        CHECK_RC("EVAPI_clear_async_event_handler", rc, final_rc = rc);
    }

    if (resources_p->hca_hndl != VAPI_INVAL_HNDL) {
        rc = EVAPI_release_hca_hndl(resources_p->hca_hndl);
        CHECK_RC("EVAPI_release_hca_hndl", rc, return rc);
    }


    return VAPI_OK;
}

/*****************************************************************************
* Function: query_sa
*****************************************************************************/
static VAPI_ret_t query_sa(
    resource_t *resources_p)
{
    VAPI_ret_t rc;
    VAPI_sr_desc_t sr_desc;
    VAPI_rr_desc_t rr_desc;
    VAPI_sg_lst_entry_t snd_sg_entry;
    VAPI_sg_lst_entry_t rcv_sg_entry;
    VAPI_wc_desc_t wc_desc;
    int i;


    /* post the RR to the QP */
    rcv_sg_entry.addr = resources_p->rcv_mr_props.start;
    rcv_sg_entry.len = resources_p->rcv_mr_props.size;
    rcv_sg_entry.lkey = resources_p->rcv_mr_props.l_key;


    rr_desc.opcode = VAPI_RECEIVE;
    rr_desc.comp_type = VAPI_SIGNALED;
    rr_desc.sg_lst_len = 1;
    rr_desc.sg_lst_p = &rcv_sg_entry;
    rr_desc.id = 0;

    rc = VAPI_post_rr(resources_p->hca_hndl, resources_p->qp_hndl, &rr_desc);
    CHECK_RC("VAPI_post_sr", rc, return rc);


    /* post the SR to the QP */
    /* build the MAD */
    build_packet(resources_p->snd_buf);

    snd_sg_entry.addr = resources_p->snd_mr_props.start;
    snd_sg_entry.len = resources_p->snd_mr_props.size;
    snd_sg_entry.lkey = resources_p->snd_mr_props.l_key;

    sr_desc.opcode = VAPI_SEND;
    sr_desc.comp_type = VAPI_SIGNALED;
    sr_desc.sg_lst_len = 1;
    sr_desc.sg_lst_p = &snd_sg_entry;
    sr_desc.fence = FALSE;
    sr_desc.set_se = FALSE;
    sr_desc.remote_ah = resources_p->av_hndl;
    sr_desc.remote_qp = IB_QP1;
    sr_desc.remote_qkey = IB_QP1_WELL_KNOWN_Q_KEY;
    sr_desc.id = 0;

    rc = VAPI_post_sr(resources_p->hca_hndl, resources_p->qp_hndl, &sr_desc);
    CHECK_RC("VAPI_post_sr", rc, return rc);

    /* read the completions */
    for (i = 0; i < NUM_OF_COMPLETION; i ++) {
        unsigned int max_timeout;
        unsigned int sleep_time; /* sleep time (in usec) */
        int j = 0;

        max_timeout = config.timeout * SENSITIVITY;
        sleep_time = 1000000/SENSITIVITY;
        
        do  {
            j ++;

            rc = VAPI_poll_cq(resources_p->hca_hndl, resources_p->cq_hndl, &wc_desc);
            if (rc == VAPI_CQ_EMPTY) {
                MOSAL_usleep(sleep_time);
            }

        } while ((j < max_timeout) && (rc == VAPI_CQ_EMPTY));
        CHECK_RC("VAPI_poll_cq", rc, return rc);

        if (wc_desc.status != VAPI_SUCCESS) {
            printf("Error, got bad completion[%d]: %s\n", i, VAPI_wc_status_sym(wc_desc.status));
            return VAPI_EGEN;
        }
    }

    return VAPI_OK;
}



/*****************************************************************************
* Function: main
*****************************************************************************/
int main(
    int argc, 
    char *argv[])
{
    VAPI_ret_t rc;
    resource_t resources;
    int test_result = 1;


    resource_init(&resources);

    rc = parse_argv(argc, argv);
    if (rc) {
        goto cleanup;
    }

    rc = resource_create(&resources);
    if (rc != VAPI_OK) {
        goto cleanup;
    }

    rc = query_sa(&resources);
    if (rc != VAPI_OK) {
        goto cleanup;
    }

    test_result = 0;

cleanup:
    rc = resource_destroy(&resources);
    if (rc != VAPI_OK) {
        test_result = 1;
    }

    if (test_result) {
        printf("SA is dead (didn't get any response after %d seconds)\n", config.timeout);
    } else {
        printf("SA is alive\n");
    }
    return test_result;
}

